// Copyright (c) 2012 Leonhard Gruenschloss (leonhard@gruenschloss.org)
//
// Permission is hereby granted, free of charge, to any person obtaining a copy
// of this software and associated documentation files (the "Software"), to deal
// in the Software without restriction, including without limitation the rights to
// use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
// of the Software, and to permit persons to whom the Software is furnished to do
// so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
// IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
// OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
// SOFTWARE.

// These matrices are based on the following publication:
//
// S. Joe and F. Y. Kuo: "Constructing Sobol sequences with better
// two-dimensional projections", SIAM J. Sci. Comput. 30, 2635-2654 (2008).
//
// The tabulated direction numbers are available here:
// http://web.maths.unsw.edu.au/~fkuo/sobol/new-joe-kuo-6.21201

#include "sobolseq.h"

namespace sobol {

const uint32_t Matrices::matrices32[Matrices::num_dimensions * Matrices::size] =
{
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x8000000U,
    0x4000000U,
    0x2000000U,
    0x1000000U,
    0x800000U,
    0x400000U,
    0x200000U,
    0x100000U,
    0x80000U,
    0x40000U,
    0x20000U,
    0x10000U,
    0x8000U,
    0x4000U,
    0x2000U,
    0x1000U,
    0x800U,
    0x400U,
    0x200U,
    0x100U,
    0x80U,
    0x40U,
    0x20U,
    0x10U,
    0x8U,
    0x4U,
    0x2U,
    0x1U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x0U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0x88000000U,
    0xcc000000U,
    0xaa000000U,
    0xff000000U,
    0x80800000U,
    0xc0c00000U,
    0xa0a00000U,
    0xf0f00000U,
    0x88880000U,
    0xcccc0000U,
    0xaaaa0000U,
    0xffff0000U,
    0x80008000U,
    0xc000c000U,
    0xa000a000U,
    0xf000f000U,
    0x88008800U,
    0xcc00cc00U,
    0xaa00aa00U,
    0xff00ff00U,
    0x80808080U,
    0xc0c0c0c0U,
    0xa0a0a0a0U,
    0xf0f0f0f0U,
    0x88888888U,
    0xccccccccU,
    0xaaaaaaaaU,
    0xffffffffU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0x88000000U,
    0xcc000000U,
    0xaa000000U,
    0xff000000U,
    0x80800000U,
    0xc0c00000U,
    0xa0a00000U,
    0xf0f00000U,
    0x88880000U,
    0xcccc0000U,
    0xaaaa0000U,
    0xffff0000U,
    0x80008000U,
    0xc000c000U,
    0xa000a000U,
    0xf000f000U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0xe8000000U,
    0x5c000000U,
    0x8e000000U,
    0xc5000000U,
    0x68800000U,
    0x9cc00000U,
    0xee600000U,
    0x55900000U,
    0x80680000U,
    0xc09c0000U,
    0x60ee0000U,
    0x90550000U,
    0xe8808000U,
    0x5cc0c000U,
    0x8e606000U,
    0xc5909000U,
    0x6868e800U,
    0x9c9c5c00U,
    0xeeee8e00U,
    0x5555c500U,
    0x8000e880U,
    0xc0005cc0U,
    0x60008e60U,
    0x9000c590U,
    0xe8006868U,
    0x5c009c9cU,
    0x8e00eeeeU,
    0xc5005555U,
    0x68808000U,
    0x9cc0c000U,
    0xee606000U,
    0x55909000U,
    0x8068e800U,
    0xc09c5c00U,
    0x60ee8e00U,
    0x9055c500U,
    0xe880e880U,
    0x5cc05cc0U,
    0x8e608e60U,
    0xc590c590U,
    0x68686868U,
    0x9c9c9c9cU,
    0xeeeeeeeeU,
    0x55555555U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xf8000000U,
    0x74000000U,
    0xa2000000U,
    0x93000000U,
    0xd8800000U,
    0x25400000U,
    0x59e00000U,
    0xe6d00000U,
    0x78080000U,
    0xb40c0000U,
    0x82020000U,
    0xc3050000U,
    0x208f8000U,
    0x51474000U,
    0xfbea2000U,
    0x75d93000U,
    0xa0858800U,
    0x914e5400U,
    0xdbe79e00U,
    0x25db6d00U,
    0x58800080U,
    0xe54000c0U,
    0x79e00020U,
    0xb6d00050U,
    0x800800f8U,
    0xc00c0074U,
    0x200200a2U,
    0x50050093U,
    0xf80f80d8U,
    0x74074025U,
    0xa20a2059U,
    0x930930e6U,
    0xd88d8878U,
    0x254254b4U,
    0x59e59e82U,
    0xe6de6dc3U,
    0x780f80a0U,
    0xb4074091U,
    0x820a20dbU,
    0xc3093025U,
    0x208d8858U,
    0x514254e5U,
    0xfbe59e79U,
    0x75de6db6U,
    0xa08f8000U,
    0x91474000U,
    0xdbea2000U,
    0x25d93000U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0xf8000000U,
    0xdc000000U,
    0x7a000000U,
    0x9d000000U,
    0x5a800000U,
    0x2fc00000U,
    0xa1600000U,
    0xf0b00000U,
    0xda880000U,
    0x6fc40000U,
    0x81620000U,
    0x40bb0000U,
    0x22878000U,
    0xb3c9c000U,
    0xfb65a000U,
    0xddb2d000U,
    0x78022800U,
    0x9c0b3c00U,
    0x5a0fb600U,
    0x2d0ddb00U,
    0xa2878080U,
    0xf3c9c040U,
    0xdb65a020U,
    0x6db2d0b0U,
    0x800228f8U,
    0x400b3cdcU,
    0x200fb67aU,
    0xb00ddb9dU,
    0xf80780daU,
    0xdc09c06fU,
    0x7a05a081U,
    0x9d02d040U,
    0x5a8a2822U,
    0x2fcf3cb3U,
    0xa16db6fbU,
    0xf0b6dbddU,
    0xda8000f8U,
    0x6fc000dcU,
    0x8160007aU,
    0x40b0009dU,
    0x2288005aU,
    0xb3c4002fU,
    0xfb6200a1U,
    0xddbb00f0U,
    0x780780daU,
    0x9c09c06fU,
    0x5a05a081U,
    0x2d02d040U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0xc8000000U,
    0x24000000U,
    0x56000000U,
    0xfb000000U,
    0xe0800000U,
    0x70400000U,
    0xa8600000U,
    0x14300000U,
    0x9ec80000U,
    0xdf240000U,
    0xb6d60000U,
    0x8bbb0000U,
    0x48008000U,
    0x64004000U,
    0x36006000U,
    0xcb003000U,
    0x2880c800U,
    0x54402400U,
    0xfe605600U,
    0xef30fb00U,
    0x7e48e080U,
    0xaf647040U,
    0x1eb6a860U,
    0x9f8b1430U,
    0xd6c81ec8U,
    0xbb249f24U,
    0x80d6d6d6U,
    0x40bbbbbbU,
    0x60800000U,
    0x30400000U,
    0xc8600000U,
    0x24300000U,
    0x56c80000U,
    0xfb240000U,
    0xe0d60000U,
    0x70bb0000U,
    0xa8808000U,
    0x14404000U,
    0x9e606000U,
    0xdf303000U,
    0xb648c800U,
    0x8b642400U,
    0x48b65600U,
    0x648bfb00U,
    0x36486080U,
    0xcb643040U,
    0x28b6c860U,
    0x548b2430U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x58000000U,
    0x94000000U,
    0x3e000000U,
    0xe3000000U,
    0xbe800000U,
    0x23c00000U,
    0x1e200000U,
    0xf3100000U,
    0x46780000U,
    0x67840000U,
    0x78460000U,
    0x84670000U,
    0xc6788000U,
    0xa784c000U,
    0xd846a000U,
    0x5467d000U,
    0x9e78d800U,
    0x33845400U,
    0xe6469e00U,
    0xb7673300U,
    0x20f86680U,
    0x104477c0U,
    0xf8668020U,
    0x4477c010U,
    0x668020f8U,
    0x77c01044U,
    0x8020f866U,
    0xc0104477U,
    0xa0f86680U,
    0xd04477c0U,
    0x58668020U,
    0x9477c010U,
    0x3e8020f8U,
    0xe3c01044U,
    0xbe20f866U,
    0x23104477U,
    0x1e786680U,
    0xf38477c0U,
    0x46468020U,
    0x6767c010U,
    0x78f820f8U,
    0x84441044U,
    0xc666f866U,
    0xa7774477U,
    0xd800e680U,
    0x5400b7c0U,
    0x9e002020U,
    0x33001010U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x88000000U,
    0x24000000U,
    0x12000000U,
    0x2d000000U,
    0x76800000U,
    0x9e400000U,
    0x8200000U,
    0x64100000U,
    0xb2280000U,
    0x7d140000U,
    0xfea20000U,
    0xba490000U,
    0x1a248000U,
    0x491b4000U,
    0xc4b5a000U,
    0xe3739000U,
    0xf6800800U,
    0xde400400U,
    0xa8200a00U,
    0x34100500U,
    0x3a280880U,
    0x59140240U,
    0xeca20120U,
    0x974902d0U,
    0x6ca48768U,
    0xd75b49e4U,
    0xcc95a082U,
    0x87639641U,
    0x44a80322U,
    0xa35403d1U,
    0x568205eaU,
    0x8e590ea4U,
    0x200c8922U,
    0x100f46d1U,
    0x2817ad6bU,
    0x743a9ce7U,
    0x9a248000U,
    0x91b4000U,
    0x64b5a000U,
    0xb3739000U,
    0x7e800800U,
    0xfa400400U,
    0xba200a00U,
    0x19100500U,
    0x4ca80880U,
    0xc7540240U,
    0xe4820120U,
    0xf35902d0U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x28000000U,
    0xd4000000U,
    0x6a000000U,
    0x71000000U,
    0x38800000U,
    0x58400000U,
    0xea200000U,
    0x31100000U,
    0x98a80000U,
    0x8540000U,
    0xc22a0000U,
    0xe5250000U,
    0xf2b28000U,
    0x79484000U,
    0xfaa42000U,
    0xbd731000U,
    0x18a80800U,
    0x48540400U,
    0x622a0a00U,
    0xb5250500U,
    0xdab28280U,
    0xad484d40U,
    0x90a426a0U,
    0xcc731710U,
    0x20280b88U,
    0x10140184U,
    0x880a04a2U,
    0x84350611U,
    0x421a8b0aU,
    0xa51c4dc5U,
    0x528e2a82U,
    0x29561942U,
    0xd29a84a3U,
    0x695c4610U,
    0x72ae2b08U,
    0x39461dc6U,
    0x5ab28280U,
    0xed484d40U,
    0x30a426a0U,
    0x9c731710U,
    0x8280b88U,
    0xc4140184U,
    0xe20a04a2U,
    0xf5350611U,
    0x7a9a8b0aU,
    0xfd5c4dc5U,
    0xb8ae2a82U,
    0x18461942U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x98000000U,
    0x94000000U,
    0x8a000000U,
    0x5b000000U,
    0x33800000U,
    0xd9c00000U,
    0x72200000U,
    0x3f100000U,
    0xc1b80000U,
    0xa6ec0000U,
    0x53860000U,
    0x29f50000U,
    0xa3a8000U,
    0x1b2ac000U,
    0xd392e000U,
    0x69ff7000U,
    0xea380800U,
    0xab2c0400U,
    0x4ba60e00U,
    0xfde50b00U,
    0x60028980U,
    0xf006c940U,
    0x7834e8a0U,
    0x241a75b0U,
    0x123a8b38U,
    0xcf2ac99cU,
    0xb992e922U,
    0x82ff78f1U,
    0x41b80d9bU,
    0xe6ec072eU,
    0xb3860398U,
    0x99f50c2fU,
    0x923a8a1bU,
    0x8f2ac56eU,
    0x5992e2bbU,
    0x32ff70deU,
    0xd9b80980U,
    0x72ec0940U,
    0x398608a0U,
    0xc2f505b0U,
    0xa1ba8338U,
    0x56eacd9cU,
    0x2bb2e722U,
    0xdef73f1U,
    0x1800041bU,
    0xd4000e6eU,
    0x6a000b38U,
    0xeb00099fU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x8000000U,
    0x6c000000U,
    0x9e000000U,
    0x23000000U,
    0x57800000U,
    0xadc00000U,
    0x7fa00000U,
    0x91d00000U,
    0x49880000U,
    0xced40000U,
    0x880a0000U,
    0x2c0f0000U,
    0x3e0d8000U,
    0x3317c000U,
    0x5fb06000U,
    0xc1f8b000U,
    0xe18d8800U,
    0xb2d7c400U,
    0x1e106a00U,
    0x6328b100U,
    0xf7858880U,
    0xbdc3c2c0U,
    0x77ba63e0U,
    0xfdf7b330U,
    0xd7800df8U,
    0xedc0081cU,
    0xdfa0041aU,
    0x81d00a2dU,
    0x41880160U,
    0xa2d400f1U,
    0x160a069aU,
    0xf0f09edU,
    0x698d8200U,
    0x9ed7c500U,
    0x20106a81U,
    0x5028b7c2U,
    0xa8058160U,
    0x7c03c0f1U,
    0x961a669aU,
    0x4f27b9edU,
    0xc9880a00U,
    0x8ed40100U,
    0x280a0081U,
    0x3c0f06c2U,
    0x360d89e0U,
    0x5f17c231U,
    0xc1b0657aU,
    0xe2f8baddU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0x58000000U,
    0xac000000U,
    0x96000000U,
    0x2b000000U,
    0xd4800000U,
    0x9400000U,
    0xe2a00000U,
    0x52500000U,
    0x4e280000U,
    0xc71c0000U,
    0x629e0000U,
    0x12670000U,
    0x6e138000U,
    0xf731c000U,
    0x3a98a000U,
    0xbe449000U,
    0xf83b8800U,
    0xdc2dc400U,
    0xee06a200U,
    0xb7239300U,
    0x1aa80d80U,
    0x8e5c0ec0U,
    0xa03e0b60U,
    0x703701b0U,
    0x783b88c8U,
    0x9c2dca54U,
    0xce06a74aU,
    0x87239795U,
    0x42a801aaU,
    0x225c08e5U,
    0x363e0a03U,
    0x5b370703U,
    0xacbb8783U,
    0x956dc9c2U,
    0x2ca6ace0U,
    0xd5739872U,
    0xc800c2aU,
    0xe5400625U,
    0x54a00163U,
    0x495006b3U,
    0xc2a80f4bU,
    0x625c0396U,
    0x163e0baaU,
    0x6b370fe7U,
    0xf4bb8d80U,
    0x396dcec0U,
    0xbaa6ab60U,
    0xfe7391b0U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0xf8000000U,
    0x8c000000U,
    0xe2000000U,
    0x33000000U,
    0xf800000U,
    0x21400000U,
    0x95a00000U,
    0x5e700000U,
    0xd8080000U,
    0x1c240000U,
    0xba160000U,
    0xef370000U,
    0x15868000U,
    0x9e6fc000U,
    0x781b6000U,
    0x4c349000U,
    0x420e8800U,
    0x630bcc00U,
    0xf7ad6a00U,
    0xad739500U,
    0x77800780U,
    0x6d4004c0U,
    0xd7a00420U,
    0x3d700630U,
    0x2f880f78U,
    0xb1640ad4U,
    0xcdb6077aU,
    0x824706d7U,
    0xc20e8d78U,
    0xa30bc3d6U,
    0x57ad62fbU,
    0xfd739b14U,
    0x8f8004d8U,
    0xe1400424U,
    0x35a00620U,
    0xe700f30U,
    0x20080af8U,
    0x90240716U,
    0x581606dbU,
    0xdc370d24U,
    0x1a0683a0U,
    0xbf2fc2f0U,
    0xedbb6b5aU,
    0x12449ce7U,
    0x9a068000U,
    0x7f2fc000U,
    0x4dbb6000U,
    0x42449000U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x38000000U,
    0xc4000000U,
    0x42000000U,
    0xa3000000U,
    0xf1800000U,
    0xaa400000U,
    0xfce00000U,
    0x85100000U,
    0xe0080000U,
    0x500c0000U,
    0x58060000U,
    0x54090000U,
    0x7a038000U,
    0x670c4000U,
    0xb3842000U,
    0x94a3000U,
    0xd6f1800U,
    0x2f5aa400U,
    0x1ce7ce00U,
    0xd5145100U,
    0xb8000080U,
    0x40000c0U,
    0x22000060U,
    0x33000090U,
    0xc9800038U,
    0x6e4000c4U,
    0xbee00042U,
    0x261000a3U,
    0x118800f1U,
    0xfa4c00aaU,
    0xa4e600fcU,
    0xd1190085U,
    0x9a0b80e0U,
    0x37004050U,
    0xeb822058U,
    0x5d433054U,
    0x776c987aU,
    0x4856e467U,
    0xaf63eeb3U,
    0xdc5e6109U,
    0xb56f188dU,
    0x2b5aa4efU,
    0x3ee7ce7cU,
    0xe6145145U,
    0x71800000U,
    0x6a400000U,
    0x9ce00000U,
    0x15100000U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xa8000000U,
    0x54000000U,
    0x9a000000U,
    0x9d000000U,
    0x1e800000U,
    0x5cc00000U,
    0x7d200000U,
    0x8d100000U,
    0x24880000U,
    0x71c40000U,
    0xeba20000U,
    0x75df0000U,
    0x6ba28000U,
    0x35d14000U,
    0x4ba3a000U,
    0xc5d2d000U,
    0xe3a16800U,
    0x91db8c00U,
    0x79aef200U,
    0xcdf4100U,
    0x672a8080U,
    0x50154040U,
    0x1a01a020U,
    0xdd0dd0f0U,
    0x3e83e8a8U,
    0xaccacc54U,
    0xd52d529aU,
    0xd91d919dU,
    0xbe83e89eU,
    0xeccacc1cU,
    0xf52d525dU,
    0x291d917dU,
    0x1683e80cU,
    0xb8cacc65U,
    0x6f2d5251U,
    0xb41d9118U,
    0x803e85dU,
    0xe40acc7dU,
    0x120d528cU,
    0x390d9125U,
    0x2c8be8f1U,
    0x95cecca8U,
    0xf9af5255U,
    0x4cd29199U,
    0x4729681eU,
    0xa01f8c5cU,
    0xb20cf27dU,
    0x8900418dU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0xd8000000U,
    0xc4000000U,
    0x46000000U,
    0x85000000U,
    0xa5800000U,
    0x76c00000U,
    0xada00000U,
    0x6ab00000U,
    0x2da80000U,
    0xaabc0000U,
    0xdaa0000U,
    0x7ab10000U,
    0xd5a78000U,
    0xbebd4000U,
    0x93a3e000U,
    0x3bb51000U,
    0x3629b800U,
    0x4d727c00U,
    0x9b836200U,
    0x27c4d700U,
    0xb629b880U,
    0x8d727cc0U,
    0xbb836220U,
    0xf7c4d7d0U,
    0x6e29b858U,
    0x49727c04U,
    0xfd836266U,
    0x72c4d755U,
    0xcba9b8fdU,
    0x3fb27c72U,
    0x502362cbU,
    0x1874d73fU,
    0xe601b8d0U,
    0x950e7cd8U,
    0x5d8962c6U,
    0x62c5d745U,
    0x33a63805U,
    0x2bb33c66U,
    0xce2a8255U,
    0x5970c77eU,
    0x58f8033U,
    0x66c1402bU,
    0x55a9e0ceU,
    0x7eb41059U,
    0xb3a63805U,
    0xebb33c66U,
    0xee2a8255U,
    0x8970c77eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0x38000000U,
    0x14000000U,
    0xf6000000U,
    0x67000000U,
    0x8f800000U,
    0x50400000U,
    0x8aa00000U,
    0xff00000U,
    0x12a80000U,
    0xabf40000U,
    0xfcaa0000U,
    0x28fb0000U,
    0xbd298000U,
    0xbba4000U,
    0x4e06e000U,
    0x330c3000U,
    0x59861800U,
    0xc74d3400U,
    0x3d2cb200U,
    0x4bb2cb00U,
    0x6e061880U,
    0xc30d3440U,
    0x618cb220U,
    0xd342cbf0U,
    0xcb2e18b8U,
    0x2cb93454U,
    0xe186b2d6U,
    0x9349cb97U,
    0xeb2f9837U,
    0xdcb77404U,
    0xd98a525cU,
    0x874efb98U,
    0x1d280025U,
    0xbbb400afU,
    0x560a00a0U,
    0xd70b00b0U,
    0x97818018U,
    0xb44e40e4U,
    0x44ace0ceU,
    0x7cf73073U,
    0x6b2f9879U,
    0x9cb77437U,
    0xf98a5205U,
    0x774efb5fU,
    0x25280018U,
    0xafb400e4U,
    0xa00a00ceU,
    0xb00b0073U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0x68000000U,
    0x64000000U,
    0x36000000U,
    0x6d000000U,
    0x41800000U,
    0xe0400000U,
    0xd2e00000U,
    0x9bf00000U,
    0xce80000U,
    0x52fc0000U,
    0x5b6a0000U,
    0x2fb30000U,
    0xa00c8000U,
    0x30054000U,
    0x4807e000U,
    0x940f9000U,
    0x5e01f800U,
    0x90e9400U,
    0x778a5600U,
    0x8d416b00U,
    0x9369f880U,
    0x7bb294c0U,
    0xde005620U,
    0xc9026bf0U,
    0x578d78e8U,
    0x7d4bd4a4U,
    0xfb6db616U,
    0x1fbefb9dU,
    0xe80000a9U,
    0xa4000044U,
    0x160000c4U,
    0x9d000006U,
    0x29800025U,
    0x844000d6U,
    0xe4e000bfU,
    0xf6f000d9U,
    0x4d6800edU,
    0xb2bc0082U,
    0x898a00c1U,
    0xb4430020U,
    0xace480f2U,
    0x62f9406bU,
    0x136de064U,
    0xbbbc9036U,
    0xfe0d786dU,
    0x390bd442U,
    0x3f8db6e1U,
    0x194efbd0U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x98000000U,
    0xf4000000U,
    0xae000000U,
    0xbb000000U,
    0xe7800000U,
    0x95c00000U,
    0x1c200000U,
    0xd0300000U,
    0xdba80000U,
    0x55f40000U,
    0xff820000U,
    0x21c10000U,
    0x12238000U,
    0x3b3a4000U,
    0xa42b6000U,
    0x3430f000U,
    0x4da69800U,
    0x4af3ec00U,
    0x2e043a00U,
    0xfb0a1f00U,
    0x47851880U,
    0xc5c9ac40U,
    0x842f5aa0U,
    0x243aef50U,
    0x75a38018U,
    0xeefa40b4U,
    0x180b600eU,
    0xb400f0ebU,
    0xe0e987fU,
    0xeb07ec61U,
    0x7f863ab2U,
    0x61cb1f6bU,
    0xb22698bcU,
    0x6b33ec80U,
    0x3c243a43U,
    0xc03a1fa1U,
    0xe3ad18d1U,
    0xf1fdacdaU,
    0xc98d5a55U,
    0x6ecbeffeU,
    0x5ba800a0U,
    0x15f40050U,
    0x5f820098U,
    0x71c100f4U,
    0x8a2380aeU,
    0xcf3a40bbU,
    0xa2b60e7U,
    0x8f30f095U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xb8000000U,
    0x3c000000U,
    0xce000000U,
    0x41000000U,
    0x21800000U,
    0x51c00000U,
    0x9600000U,
    0x85700000U,
    0xf2780000U,
    0x8e9c0000U,
    0x60020000U,
    0x70030000U,
    0x58038000U,
    0x8c02c000U,
    0x7602e000U,
    0x7d00f000U,
    0xef833800U,
    0x10c10400U,
    0x28e08600U,
    0xd4b14700U,
    0xfb182580U,
    0xbee15c0U,
    0x9279c9e0U,
    0xfe9d3a70U,
    0x38000008U,
    0xfc00000cU,
    0x2e00000eU,
    0xf100000bU,
    0x9980000bU,
    0x6dc00003U,
    0xc760000cU,
    0xc4700004U,
    0xd3f80002U,
    0xdf5c0005U,
    0x69620000U,
    0xf5730008U,
    0xaa7b800fU,
    0x29ec008U,
    0x1600e006U,
    0xd03f007U,
    0xb780b805U,
    0x9cc3c408U,
    0x5ee26607U,
    0xa9b1b707U,
    0x149b1d8eU,
    0x1b2f11c1U,
    0xba994fe2U,
    0x2a2c7d7dU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xd0000000U,
    0x68000000U,
    0x3c000000U,
    0x8a000000U,
    0x51000000U,
    0xa9800000U,
    0xddc00000U,
    0x5ba00000U,
    0x39d00000U,
    0x95f80000U,
    0x56d40000U,
    0xa020000U,
    0x91030000U,
    0x49838000U,
    0xdc34000U,
    0x33a1a000U,
    0x5d0f000U,
    0x1ffa2800U,
    0x7d54400U,
    0xa380a600U,
    0x4cc07700U,
    0x1222ee80U,
    0x3413a740U,
    0xa65bf7e0U,
    0x5305ab50U,
    0x15f80008U,
    0x96d4000cU,
    0xea02000eU,
    0x4103000dU,
    0x21838006U,
    0x31c34003U,
    0xb9a1a008U,
    0x54d0f005U,
    0xb67a280aU,
    0xda15440dU,
    0xf820a605U,
    0x75107703U,
    0x87daee89U,
    0x62c7a745U,
    0xac59f7e0U,
    0xc206ab59U,
    0x5c7b800cU,
    0x9b17400cU,
    0xd9a3a00dU,
    0x44d3f00dU,
    0x3e79a807U,
    0x36160403U,
    0x1a210602U,
    0x18108701U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0x38000000U,
    0x8c000000U,
    0x7e000000U,
    0x71000000U,
    0xc8800000U,
    0x4c00000U,
    0x1ba00000U,
    0xbb700000U,
    0x4a980000U,
    0xc3bc0000U,
    0xa6020000U,
    0x6d010000U,
    0xee818000U,
    0x29c34000U,
    0x9520e000U,
    0x42b23000U,
    0xe7b9f800U,
    0xd0dc400U,
    0x3fb92200U,
    0x110d1300U,
    0x19bbee80U,
    0x3c0cadc0U,
    0x973a4a60U,
    0xc5cf7ef0U,
    0x3a180008U,
    0xb7c0004U,
    0xa3a20006U,
    0x7771000dU,
    0x54998003U,
    0x62bf4008U,
    0x5682e007U,
    0xe5c33007U,
    0x8b20780cU,
    0xe3b28400U,
    0x173bc201U,
    0x85ce230bU,
    0x5a1b9684U,
    0xdb7e29ccU,
    0x9ba1886aU,
    0xfb715df6U,
    0x2a9b9686U,
    0x13be29c6U,
    0x9e01886fU,
    0xe1015df9U,
    0x90839685U,
    0x58c229ccU,
    0x5da38862U,
    0x46705dfbU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0x8000000U,
    0x64000000U,
    0x6a000000U,
    0x89000000U,
    0xa5800000U,
    0xcb400000U,
    0x18200000U,
    0xad900000U,
    0xaf880000U,
    0x72f40000U,
    0x25820000U,
    0xb430000U,
    0xb8228000U,
    0x3d924000U,
    0xa7882000U,
    0x16f59000U,
    0x4f83a800U,
    0x82412400U,
    0x1da01600U,
    0xf6d16d00U,
    0xbfa84080U,
    0xbb672640U,
    0xe0091620U,
    0xf0b4efd0U,
    0x38228008U,
    0xfd92400cU,
    0x788200aU,
    0x86f59009U,
    0x4783a800U,
    0xe6412406U,
    0x77a01606U,
    0x7fd16d08U,
    0x1a28408aU,
    0x7027264cU,
    0xf8291621U,
    0x5d24efdaU,
    0x97aa8002U,
    0x8f66400bU,
    0x220a2008U,
    0x8db69009U,
    0xffa1280bU,
    0xdbd36405U,
    0xd028360cU,
    0x6924fd09U,
    0x55abe88eU,
    0xf2660244U,
    0xe5890020U,
    0xabf582d5U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x48000000U,
    0x8c000000U,
    0xd6000000U,
    0x39000000U,
    0xd5800000U,
    0x32400000U,
    0xb2a00000U,
    0x72100000U,
    0x53d80000U,
    0x82cc0000U,
    0xcb820000U,
    0x47430000U,
    0x91208000U,
    0xa9534000U,
    0x7cf92000U,
    0x4e9e3000U,
    0xfcf95800U,
    0x8e9fe400U,
    0xdcf9d600U,
    0x5e9c8900U,
    0x94f96a80U,
    0xd29fb840U,
    0x42f9b760U,
    0xeb9c9f30U,
    0x97788008U,
    0xd9df400cU,
    0x25db2002U,
    0xabcd300dU,
    0x7601d804U,
    0x2900a408U,
    0xbd82f60dU,
    0x6e41b903U,
    0x2ca0b28dU,
    0xc7131c43U,
    0x5059416bU,
    0x898e2637U,
    0xaca0b28dU,
    0x7131c44U,
    0x7059416eU,
    0x598e2639U,
    0xe4a0b285U,
    0x8b131c4eU,
    0xa6594168U,
    0x608e263aU,
    0x3120b28eU,
    0xb9531c4fU,
    0x14f94169U,
    0x129e263cU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xd8000000U,
    0xf4000000U,
    0x3e000000U,
    0x95000000U,
    0x8f800000U,
    0x3d400000U,
    0xf3200000U,
    0x2ef00000U,
    0xadc80000U,
    0xa0c0000U,
    0x8b220000U,
    0x4af30000U,
    0x6bc88000U,
    0x3b0d4000U,
    0xe2a16000U,
    0x16b0d000U,
    0x29687800U,
    0xbdbf1400U,
    0x33cb5e00U,
    0xf0c2500U,
    0xfca1b480U,
    0xd3b0afc0U,
    0x7eeb6920U,
    0x74fe4d30U,
    0xfee87808U,
    0xb4ff140cU,
    0xdeeb5e02U,
    0xe4fc2505U,
    0x6e9b48dU,
    0x10fcafcfU,
    0x38e96923U,
    0x85fd4d39U,
    0xb768f800U,
    0xb8be540fU,
    0x44483e0dU,
    0x964ff507U,
    0xe9814c87U,
    0x9c42fbcfU,
    0x62a1572bU,
    0xd6b2b83dU,
    0x969b486U,
    0xedbcafccU,
    0xebc96923U,
    0xfb0d4d36U,
    0xc2a0f80dU,
    0x46b25408U,
    0xf16a3e0aU,
    0x49bcf508U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x98000000U,
    0xa4000000U,
    0x7a000000U,
    0xd5000000U,
    0x2800000U,
    0x60400000U,
    0x51e00000U,
    0x88700000U,
    0x8c280000U,
    0x47c40000U,
    0xbe20000U,
    0xad710000U,
    0xb6aa8000U,
    0x3386c000U,
    0xb8006000U,
    0x54039000U,
    0x42036800U,
    0xc1019400U,
    0xe0826a00U,
    0x11431100U,
    0x2960af80U,
    0x3d3175c0U,
    0xdf4a3aa0U,
    0xaff49e10U,
    0xd62b6808U,
    0x62c59404U,
    0x31606a0aU,
    0xd932110bU,
    0x54a2f89U,
    0xcaf7b5caU,
    0x4caa5aa7U,
    0xa6870e1dU,
    0x1a800008U,
    0x84400002U,
    0x8be0000fU,
    0xed700003U,
    0x16a80001U,
    0x8384000eU,
    0x20020007U,
    0xf0010007U,
    0x3802800bU,
    0x1402c005U,
    0xe202600eU,
    0x7102900dU,
    0x7881e80cU,
    0xb5435408U,
    0x53600a0eU,
    0xe831810bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0x18000000U,
    0x34000000U,
    0x8a000000U,
    0x9d000000U,
    0x67800000U,
    0x82400000U,
    0x40e00000U,
    0x60f00000U,
    0x91480000U,
    0x29440000U,
    0x2d620000U,
    0xbfb30000U,
    0x162a8000U,
    0xfbf4c000U,
    0xe4ca6000U,
    0xc207d000U,
    0x2002a800U,
    0xf001b400U,
    0xb8037e00U,
    0x4021900U,
    0x92034b80U,
    0xa90327c0U,
    0xed81f320U,
    0x1f40d810U,
    0x27602808U,
    0xe2b1740cU,
    0xd1ab1e0aU,
    0x49b6c903U,
    0xbc2b6381U,
    0x96f653c3U,
    0x3b48ed28U,
    0x44451119U,
    0xf2e1cb8eU,
    0x39f3e7c4U,
    0xc4c9932eU,
    0x32040815U,
    0x98000000U,
    0xf400000dU,
    0x2a000000U,
    0xad000001U,
    0x7f800006U,
    0xb6400004U,
    0xcae00002U,
    0xfdf00003U,
    0xf6c8000dU,
    0xab040005U,
    0x6d82000dU,
    0xdf43000dU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0x8000000U,
    0x4c000000U,
    0x2000000U,
    0xb5000000U,
    0x36800000U,
    0xc2c00000U,
    0x14200000U,
    0x7500000U,
    0x1bf80000U,
    0x50340000U,
    0x48a20000U,
    0xac910000U,
    0xd35b8000U,
    0xbca74000U,
    0x7bfa2000U,
    0xc0343000U,
    0xa0a18800U,
    0x30909400U,
    0xd95b7a00U,
    0x45a57b00U,
    0x4f7a7880U,
    0xb7f6f940U,
    0x82013de0U,
    0xf502dfd0U,
    0xd6820808U,
    0x12c3d404U,
    0x1c235a0eU,
    0x4b504b0dU,
    0x19f87080U,
    0xe5352d44U,
    0x7e2267e0U,
    0x6e5294dbU,
    0xc77a788bU,
    0xbbf6f948U,
    0x60013defU,
    0x9002dfddU,
    0xe8020809U,
    0x9c03d405U,
    0xa035a0aU,
    0xf9004b0cU,
    0x3480708eU,
    0x77c12d43U,
    0x22a067e6U,
    0xc59394d7U,
    0xfd9f880U,
    0x5765b94eU,
    0x53591de6U,
    0xfca7efd3U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x68000000U,
    0x4c000000U,
    0x76000000U,
    0xf7000000U,
    0x36800000U,
    0xd7400000U,
    0x87e00000U,
    0xef300000U,
    0xa3a80000U,
    0xd5440000U,
    0x23aa0000U,
    0x15470000U,
    0xc3a98000U,
    0x45464000U,
    0xaba82000U,
    0x9477000U,
    0xdda9f800U,
    0xfe44ac00U,
    0xeb292200U,
    0x2907f100U,
    0x6ccb3d80U,
    0xc6344dc0U,
    0xcf61b320U,
    0x137318d0U,
    0xeccb3d88U,
    0x6344dccU,
    0x2f61b32eU,
    0x437318d5U,
    0x84cb3d8eU,
    0x4a344dc8U,
    0x5961b329U,
    0xb47318daU,
    0xb24b3d8dU,
    0x9d744dc5U,
    0xde81b321U,
    0x5b4318d4U,
    0x11e33d87U,
    0x48304dc8U,
    0xfd2bb323U,
    0x4e0418d5U,
    0xd24abd8bU,
    0xd760dccU,
    0x56839329U,
    0x474368d5U,
    0xfe34586U,
    0xf332a1c3U,
    0xbdaab127U,
    0x6e4499d7U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0xc8000000U,
    0x74000000U,
    0x52000000U,
    0x3000000U,
    0xeb800000U,
    0x6f400000U,
    0x64600000U,
    0xdaf00000U,
    0x17980000U,
    0x297c0000U,
    0xa59a0000U,
    0xfa7d0000U,
    0xe61b8000U,
    0x713f4000U,
    0x1878a000U,
    0xdcce9000U,
    0xb661e800U,
    0x99f29c00U,
    0x9c184600U,
    0xd63e2100U,
    0x9fa5780U,
    0x548e0ac0U,
    0xa380a9e0U,
    0x5b413f30U,
    0x56625788U,
    0x49f20ac4U,
    0x341aa9e6U,
    0x323c3f39U,
    0x93f9d784U,
    0x238d4ac3U,
    0x1a0209e3U,
    0x3702af39U,
    0xd9803f8aU,
    0xfc43d6c5U,
    0x47e04fe5U,
    0xc1b18e34U,
    0x21f9e80bU,
    0xf08e9c07U,
    0x5982460fU,
    0xbc43210bU,
    0x27e1d78dU,
    0x51b14ac4U,
    0xe9f809e8U,
    0x848faf3fU,
    0xb83bf82U,
    0xbf4096ceU,
    0xcc62efe2U,
    0x3ef21e3bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0xb8000000U,
    0x4000000U,
    0x6e000000U,
    0x97000000U,
    0xf2800000U,
    0xedc00000U,
    0x13600000U,
    0x5c900000U,
    0xdb580000U,
    0x31e40000U,
    0x9da0000U,
    0xcc270000U,
    0x2b88000U,
    0x44b44000U,
    0xfe26000U,
    0xe6505000U,
    0x9ab9d800U,
    0x50b50c00U,
    0x79e29200U,
    0xa552fb00U,
    0xbe38bf80U,
    0x2e77d940U,
    0xf6000ae0U,
    0x830112d0U,
    0x84803f88U,
    0xaec3994cU,
    0x37e26aeaU,
    0x225142ddU,
    0x54b9e783U,
    0x17b6954cU,
    0x3360f8ecU,
    0x4c93b9d4U,
    0xc359580cU,
    0xe5e54c02U,
    0xdfdaf20dU,
    0x5f25ab01U,
    0x9e39e789U,
    0x3e76954dU,
    0xee00f8e7U,
    0x5703b9d0U,
    0x5281580aU,
    0x3dc14c05U,
    0xab60f20bU,
    0x5892ab0aU,
    0xb559678fU,
    0xa6e6d542U,
    0xfb5898e1U,
    0x21e4e9d1U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x68000000U,
    0xec000000U,
    0x22000000U,
    0x2b000000U,
    0x36800000U,
    0x9d400000U,
    0x6a200000U,
    0x16700000U,
    0x4de80000U,
    0x330c0000U,
    0x936a0000U,
    0x824f0000U,
    0x3b498000U,
    0x8f3fc000U,
    0x28202000U,
    0xcd707000U,
    0xf36aa800U,
    0x724fdc00U,
    0xb34bf200U,
    0x533e6900U,
    0x62207a80U,
    0xa7140c0U,
    0xe7ea6520U,
    0xc40d90f0U,
    0xefe9fa88U,
    0xd80e80ccU,
    0x45ea452eU,
    0x2f0de0f3U,
    0x396b528eU,
    0x754d5cc2U,
    0x47cbb72cU,
    0xd57c89f1U,
    0x5682a80dU,
    0x6d43dc0bU,
    0xe221f20aU,
    0xca716900U,
    0x7e9fa81U,
    0xf40e80c4U,
    0x87ea452dU,
    0x340de0fbU,
    0x67eb528cU,
    0x40d5cceU,
    0xfebb723U,
    0xe80c89f6U,
    0x2deaa806U,
    0xc30fdc0eU,
    0x1b6bf20cU,
    0x5e4e6900U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x28000000U,
    0xd4000000U,
    0x8a000000U,
    0xff000000U,
    0x84800000U,
    0x73c00000U,
    0x13200000U,
    0xc2b00000U,
    0xfb380000U,
    0x361c0000U,
    0x401a0000U,
    0xe0af0000U,
    0x11228000U,
    0x19b3c000U,
    0xfdb82000U,
    0x5edf9000U,
    0x75b88800U,
    0x7adfac00U,
    0xf7baba00U,
    0x61ddf300U,
    0xd1387e80U,
    0x391e55c0U,
    0xcc9ba860U,
    0x776cbeb0U,
    0xa000f688U,
    0xf001f9ccU,
    0x8011262U,
    0xe4014db3U,
    0xa200880aU,
    0x2b03ac01U,
    0xe80ba0aU,
    0x8cc2f30cU,
    0x97a2fe8aU,
    0xb17195caU,
    0xe8198869U,
    0xf4ac2eb3U,
    0xbb22fe8fU,
    0xd6b195c8U,
    0x51398867U,
    0xf91c2eb1U,
    0xec9afe84U,
    0x476d95c3U,
    0x88038861U,
    0x24032eb7U,
    0x82007e84U,
    0x1b0255c0U,
    0x2681a86bU,
    0x58c3beb3U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0xb8000000U,
    0x84000000U,
    0x1a000000U,
    0xaf000000U,
    0xbd800000U,
    0xdfc00000U,
    0x14e00000U,
    0x43500000U,
    0xda380000U,
    0x4e1c0000U,
    0x4cda0000U,
    0x364d0000U,
    0x29608000U,
    0xdc904000U,
    0x6ed86000U,
    0x5d4f5000U,
    0x2ee08800U,
    0xfc51ac00U,
    0x7fb81e00U,
    0x45dc8300U,
    0xfa3a4580U,
    0x5e1d6240U,
    0x54dbd360U,
    0xe24ec930U,
    0x8b62cd88U,
    0xf790ce44U,
    0xc959cd6aU,
    0x2d8f4a35U,
    0x87800803U,
    0x60c1ec0cU,
    0xb1607e0bU,
    0x4893d30fU,
    0x6cdacd80U,
    0x264cce45U,
    0x3163cd60U,
    0x8924a3eU,
    0xccd8880eU,
    0x764dac0dU,
    0x89621e0fU,
    0x8c918302U,
    0xd6dac584U,
    0xd94d2241U,
    0x34e3b363U,
    0x5351993cU,
    0xc23a4583U,
    0x9a1d624bU,
    0xeedbd36aU,
    0x1d4ec930U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0x8000000U,
    0xf4000000U,
    0xf6000000U,
    0x8b000000U,
    0xc9800000U,
    0x55400000U,
    0x67200000U,
    0xf3f00000U,
    0x34780000U,
    0x57440000U,
    0x1ada0000U,
    0xb1f50000U,
    0xa9818000U,
    0x6540c000U,
    0x8f23a000U,
    0x77f21000U,
    0xca7bf800U,
    0x2845fc00U,
    0x255afe00U,
    0x6fb67900U,
    0x7233a80U,
    0xc3f25ac0U,
    0xdc7aed60U,
    0xd34482d0U,
    0xe4d94288U,
    0xcef766c4U,
    0x9603b36eU,
    0xbb00ebd7U,
    0x21818008U,
    0xd140c00bU,
    0x9923a001U,
    0x8cf2100fU,
    0xbfbf80cU,
    0x8905fc0aU,
    0xb47afe09U,
    0x17467907U,
    0xfadb3a8fU,
    0xc1f65ac0U,
    0xa180ed67U,
    0x914182d4U,
    0x7920c281U,
    0xfcf3a6c7U,
    0x3fa1367U,
    0x7d07fbdbU,
    0x427bf80eU,
    0x9c45fc0fU,
    0x335afe0bU,
    0x94b6790eU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x68000000U,
    0xf4000000U,
    0x62000000U,
    0xdf000000U,
    0x79800000U,
    0xdd400000U,
    0x76e00000U,
    0x2cf00000U,
    0xcfb80000U,
    0x51ec0000U,
    0xc8da0000U,
    0x845d0000U,
    0x9b818000U,
    0x42434000U,
    0xef622000U,
    0x61b19000U,
    0xd1582800U,
    0x891cac00U,
    0x65626e00U,
    0xab10900U,
    0x2adbbd80U,
    0x1b5d86c0U,
    0x2014560U,
    0xf032470U,
    0xf1821588U,
    0xb9426ac4U,
    0x7ce10b6eU,
    0x7f3bd79U,
    0xd439800eU,
    0x53af400bU,
    0xc7b82008U,
    0x75ec9004U,
    0x22d9a801U,
    0x3f5fec02U,
    0xe8004e01U,
    0xb400990fU,
    0x8203958bU,
    0x4f012ac8U,
    0x11832b6bU,
    0x29422d7aU,
    0x14e1a80dU,
    0xf3f3ec05U,
    0xb63a4e0cU,
    0x8cad9907U,
    0xbe3a1582U,
    0xa8ae6ac3U,
    0x543b0b6eU,
    0x13aebd7bU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x18000000U,
    0xdc000000U,
    0x42000000U,
    0x37000000U,
    0x20800000U,
    0xf1400000U,
    0x28600000U,
    0x94900000U,
    0x87880000U,
    0xa83c0000U,
    0x556a0000U,
    0xe6ef0000U,
    0xf8038000U,
    0x4c024000U,
    0x3a01e000U,
    0xbb023000U,
    0x7a816800U,
    0x1a43ac00U,
    0x4ae18a00U,
    0x52d31900U,
    0x8f682380U,
    0xcded9740U,
    0xfa80bfa0U,
    0xda43f2b0U,
    0x2ae2cb88U,
    0x2d07b4cU,
    0x976ad5a6U,
    0x11eddbb5U,
    0xb8800009U,
    0xed400001U,
    0xa600002U,
    0xf3900006U,
    0xbf080003U,
    0x857c0002U,
    0x3f0a0006U,
    0x457f000aU,
    0x5f0b800aU,
    0x157e4005U,
    0x470be007U,
    0xc97d3007U,
    0x50ae807U,
    0xfe7dec0eU,
    0x258a6a06U,
    0xf3e2905U,
    0xdeacb88U,
    0x9bac7b45U,
    0x8a60d5a7U,
    0x3392dbbeU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0xf8000000U,
    0x34000000U,
    0x62000000U,
    0xf5000000U,
    0xa8800000U,
    0xfcc00000U,
    0x8e200000U,
    0x53f00000U,
    0xc7780000U,
    0x95740000U,
    0xb8020000U,
    0xd4e50000U,
    0xb2808000U,
    0xfdc0c000U,
    0x64a02000U,
    0xaa30f000U,
    0x19d8f800U,
    0xe443400U,
    0x935a6200U,
    0xe761f500U,
    0x657a2880U,
    0x40913cc0U,
    0xe0022e20U,
    0xd0e563f0U,
    0x8809f78U,
    0xccc09174U,
    0x56200202U,
    0x97f0e5e5U,
    0x5d788000U,
    0x5474c000U,
    0x72822000U,
    0xdd25f000U,
    0x94207800U,
    0x52f0f400U,
    0x2df84200U,
    0x6cb40500U,
    0x66a25080U,
    0x4fd5c8c0U,
    0x99d86c20U,
    0xce4466f0U,
    0xb35a4ff8U,
    0x176199b4U,
    0x9d7a4e22U,
    0x74917315U,
    0x8202b7f8U,
    0x25e5adb4U,
    0xa0002c22U,
    0x30008615U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0xf8000000U,
    0xec000000U,
    0x7e000000U,
    0x61000000U,
    0x5c800000U,
    0xe6c00000U,
    0xdda00000U,
    0x2a700000U,
    0x93380000U,
    0x13cc0000U,
    0xd3ce0000U,
    0x73790000U,
    0x83a08000U,
    0x7b70c000U,
    0x97b8a000U,
    0xe90cf000U,
    0x886ef800U,
    0xd409ec00U,
    0x3218fe00U,
    0xef7ca100U,
    0xc556fc80U,
    0x56c516c0U,
    0x4556a5a0U,
    0x96c50670U,
    0xe556cd38U,
    0x66c542ccU,
    0x1d56574eU,
    0x8ac549b9U,
    0x6356f800U,
    0xebc5ec00U,
    0x3fd6fe00U,
    0xd05a100U,
    0xe2767c80U,
    0x2775d6c0U,
    0x714e05a0U,
    0x34b9f670U,
    0xa2803538U,
    0x47c0aeccU,
    0x2120a94eU,
    0x3cb0e8b9U,
    0xb6988480U,
    0xd5bc3ac0U,
    0x3ef6fba0U,
    0x1b55770U,
    0xceec9b8U,
    0xeec9b80cU,
    0xc9b80ceeU,
    0xb80ceec9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x58000000U,
    0x2c000000U,
    0x9a000000U,
    0xf9000000U,
    0x3c800000U,
    0xb2c00000U,
    0xad200000U,
    0x3a300000U,
    0x89980000U,
    0x448c0000U,
    0x2eea0000U,
    0x6f810000U,
    0xef208000U,
    0x2f30c000U,
    0xf182000U,
    0xbf4cb000U,
    0xe74a5800U,
    0xcb712c00U,
    0x51981a00U,
    0xa88c3900U,
    0x94ea1c80U,
    0x268102c0U,
    0x8ba07520U,
    0xb1f0d630U,
    0x38383398U,
    0x7c7c0d8cU,
    0x52524a6aU,
    0x3d3df141U,
    0xd2525800U,
    0xfd3d2c00U,
    0xf2521a00U,
    0x4d3d3900U,
    0xaa529c80U,
    0x613dc2c0U,
    0x30525520U,
    0x983d6630U,
    0xcd2eb98U,
    0x2afde18cU,
    0xa1f2706aU,
    0x10cd7841U,
    0x286a1c80U,
    0x544102c0U,
    0x6807520U,
    0x3bc0d630U,
    0xe9a03398U,
    0x14f00d8cU,
    0xe6b84a6aU,
    0xabbcf141U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0xd8000000U,
    0xac000000U,
    0x8e000000U,
    0x9000000U,
    0x9e800000U,
    0xa1c00000U,
    0xcaa00000U,
    0x33700000U,
    0x95780000U,
    0x85c0000U,
    0x24b60000U,
    0x6a350000U,
    0x43788000U,
    0x6d5cc000U,
    0x14362000U,
    0x72f5b000U,
    0xcf585800U,
    0x53ec6c00U,
    0xc5eeae00U,
    0x40d9b900U,
    0xe016c680U,
    0x9045cdc0U,
    0x6880e4a0U,
    0x74c04a70U,
    0x2220f3f8U,
    0x87b0b59cU,
    0x9758b816U,
    0x3fecfc45U,
    0x6beec680U,
    0xf9d9cdc0U,
    0xa696e4a0U,
    0x9d854a70U,
    0x2c2073f8U,
    0x4eb0759cU,
    0x29d89816U,
    0x2e2c4c45U,
    0x794e1e80U,
    0x66a961c0U,
    0xbdee6aa0U,
    0x9cd94370U,
    0x9616ed78U,
    0x8545d45cU,
    0xa000d2b6U,
    0x7000bf35U,
    0xf800abf8U,
    0x1c00d99cU,
    0x56001616U,
    0xa5004545U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0xa8000000U,
    0x2c000000U,
    0xa2000000U,
    0x2d000000U,
    0xda800000U,
    0xf9400000U,
    0xec600000U,
    0x2b00000U,
    0x3d480000U,
    0x825c0000U,
    0x7d4a0000U,
    0x62610000U,
    0x8dc88000U,
    0xca1c4000U,
    0xa1aae000U,
    0x6891f000U,
    0x8c602800U,
    0xb2b06c00U,
    0x75484200U,
    0x5e5cdd00U,
    0x774a7280U,
    0x6361d540U,
    0xf548ce60U,
    0x1e5c6fb0U,
    0x974a07c8U,
    0x93618b1cU,
    0x5d48b92aU,
    0x325c0cd1U,
    0x354af280U,
    0xbe619540U,
    0x87c82e60U,
    0xcb1c9fb0U,
    0xd92aafc8U,
    0xbcd1a71cU,
    0xba801b2aU,
    0x494021d1U,
    0xa4602800U,
    0xdeb06c00U,
    0x37484200U,
    0x835cdd00U,
    0x5ca7280U,
    0xb621d540U,
    0xbb28ce60U,
    0x31ec6fb0U,
    0x708207c8U,
    0xe87d8b1cU,
    0xcc62b92aU,
    0x528d0cd1U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0xc8000000U,
    0x7c000000U,
    0x82000000U,
    0x4f000000U,
    0xbe800000U,
    0xedc00000U,
    0x21600000U,
    0xab700000U,
    0x78680000U,
    0x746c0000U,
    0x1e9a0000U,
    0xfdcb0000U,
    0x39088000U,
    0x2f1cc000U,
    0x4ef2e000U,
    0xc5a73000U,
    0x6d924800U,
    0xe1d7bc00U,
    0x4b7ae200U,
    0x487bbf00U,
    0xbc801680U,
    0x62c061c0U,
    0x7fe08b60U,
    0x76b0a870U,
    0x91088ce8U,
    0xa31caaacU,
    0xe4f2037aU,
    0xc6a7f47bU,
    0x99125e80U,
    0x3f17ddc0U,
    0x569a6960U,
    0x41cb1770U,
    0x5b089a68U,
    0x501ccb6cU,
    0x3872881aU,
    0x54675c0bU,
    0xcef2d268U,
    0x5a7776cU,
    0x8d926a1aU,
    0xd1d7e30bU,
    0x837a44e8U,
    0x347bd6acU,
    0x3e80017aU,
    0x2dc07b7bU,
    0xc1608000U,
    0x9b70c000U,
    0xb068e000U,
    0x86c3000U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x98000000U,
    0x2c000000U,
    0x6000000U,
    0xcd000000U,
    0x8a800000U,
    0x1bc00000U,
    0xffa00000U,
    0xad500000U,
    0x7af80000U,
    0xb3dc0000U,
    0x5b2e0000U,
    0x1f290000U,
    0x9d588000U,
    0xf28cc000U,
    0x7d62000U,
    0x71f51000U,
    0xd4f61800U,
    0xda65ec00U,
    0x632ea600U,
    0xe3291d00U,
    0x2358b280U,
    0x38ce7c0U,
    0x135641a0U,
    0x8b355c50U,
    0xa7d6ee78U,
    0xa1f5891cU,
    0x6cf6880eU,
    0xe665b4b9U,
    0xfd2eaa80U,
    0x2290bc0U,
    0xafd8e7a0U,
    0xd54c4150U,
    0x66765cf8U,
    0x3da56edcU,
    0x228ec9aeU,
    0xbf79e8e9U,
    0x4d20c4f8U,
    0x4a9042dcU,
    0x3b584faeU,
    0xef8ce5e9U,
    0x3556ee78U,
    0x5635891cU,
    0xb556880eU,
    0x9635b4b9U,
    0x9556aa80U,
    0x86350bc0U,
    0xd56e7a0U,
    0xaa354150U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x98000000U,
    0x54000000U,
    0x3a000000U,
    0x9d000000U,
    0x7e800000U,
    0x7f400000U,
    0x17200000U,
    0xab500000U,
    0x6df80000U,
    0x96a40000U,
    0x83d20000U,
    0x71e10000U,
    0xc0d88000U,
    0xe0f44000U,
    0x30aaa000U,
    0x8059000U,
    0xcc2a1800U,
    0x6e451400U,
    0xa78a1a00U,
    0xe3554d00U,
    0x1d2c680U,
    0x68e1fb40U,
    0xbc589520U,
    0xc6b4b250U,
    0xfb0a1178U,
    0x1515b0e4U,
    0xf272c872U,
    0xb1f12cf1U,
    0x2000de80U,
    0xd000ef40U,
    0x38008f20U,
    0xc400ff50U,
    0xa20057f8U,
    0xc9000ba4U,
    0x4480fd52U,
    0xe2400ea1U,
    0x69a0d7f8U,
    0xd4104ba4U,
    0x7ad85d52U,
    0x3df49ea1U,
    0xee2a4ff8U,
    0xe7451fa4U,
    0x430ae752U,
    0x911543a1U,
    0xf0721178U,
    0xe8f1b0e4U,
    0xfc80c872U,
    0x66402cf1U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x8000000U,
    0x84000000U,
    0xb2000000U,
    0xb9000000U,
    0xbe800000U,
    0x4fc00000U,
    0x55600000U,
    0xf8f00000U,
    0xac280000U,
    0x66d40000U,
    0xb30a0000U,
    0x8bb50000U,
    0xc7c88000U,
    0x11e4c000U,
    0xaa42e000U,
    0xa591b000U,
    0xd0ea8800U,
    0x78854400U,
    0x6c80d200U,
    0x86c0c900U,
    0x3e05680U,
    0x83307bc0U,
    0x4348ef60U,
    0xa324c5f0U,
    0x13a2a0a8U,
    0x1ba19014U,
    0x9f22d8eaU,
    0x2d61fc85U,
    0x94c25e80U,
    0x2a51ffc0U,
    0x658add60U,
    0x3075bcf0U,
    0xc8a87e28U,
    0x6414afd4U,
    0x2eae58aU,
    0xb185f075U,
    0x3a00a8a8U,
    0xfd001414U,
    0xec80eaeaU,
    0x46c08585U,
    0xe3e08000U,
    0x3330c000U,
    0x4b48e000U,
    0x2724b000U,
    0xa1a20800U,
    0xa2a18400U,
    0x21a23200U,
    0x62a17900U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x78000000U,
    0x24000000U,
    0x9e000000U,
    0x47000000U,
    0x67800000U,
    0xf7400000U,
    0xdf200000U,
    0xb3100000U,
    0x71680000U,
    0x8c4c0000U,
    0x32520000U,
    0xe5d50000U,
    0xaa528000U,
    0x31d5c000U,
    0x2c52e000U,
    0x62d5f000U,
    0xadd29800U,
    0xf695d400U,
    0x8b720600U,
    0xf5c59300U,
    0x42ba6180U,
    0x3dd96440U,
    0xdea0bea0U,
    0xe750d750U,
    0x37c84fc8U,
    0xbf1c9b1cU,
    0x839a1d9aU,
    0x9c94ec9U,
    0xa8484fc8U,
    0xac5c9b1cU,
    0xa2ba1d9aU,
    0xcdd94ec9U,
    0xc6a04fc8U,
    0xf3509b1cU,
    0xd1c81d9aU,
    0xdc1c4ec9U,
    0x7a1acfc8U,
    0xb9895b1cU,
    0x10e8fd9aU,
    0xe80cbec9U,
    0xcf2d7c8U,
    0xf2854f1cU,
    0x859a1b9aU,
    0x9ac9ddc9U,
    0x49c8ae48U,
    0x81c3f5cU,
    0xfc1a433aU,
    0xea896999U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x78000000U,
    0x9c000000U,
    0xee000000U,
    0x1b000000U,
    0xcb800000U,
    0xc3400000U,
    0xc7a00000U,
    0x5100000U,
    0x88680000U,
    0xc4740000U,
    0x225a0000U,
    0x3da10000U,
    0x345a8000U,
    0x7aa1c000U,
    0xf1da6000U,
    0x12e17000U,
    0x85fa1800U,
    0x48b1ec00U,
    0x2432f600U,
    0x92d5f700U,
    0x45803d80U,
    0xa8403440U,
    0x94207a20U,
    0xea50f150U,
    0xd9c81248U,
    0x46648524U,
    0x8fb24812U,
    0x21952485U,
    0x1a201248U,
    0x81508524U,
    0x8a484812U,
    0xa9242485U,
    0xde129248U,
    0xa3854524U,
    0xb7c82812U,
    0x9d645485U,
    0xa4320a48U,
    0x52d56924U,
    0xa580be12U,
    0x1840d385U,
    0xec202fc8U,
    0x7650b164U,
    0x37c83232U,
    0x5d64d5d5U,
    0x44328000U,
    0xe2d5c000U,
    0xdd806000U,
    0x84407000U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x58000000U,
    0x7c000000U,
    0xc2000000U,
    0xe1000000U,
    0xd800000U,
    0xd7c00000U,
    0x2aa00000U,
    0xf5300000U,
    0x9ba80000U,
    0xc0f40000U,
    0x20c60000U,
    0x702f0000U,
    0x48668000U,
    0x241f4000U,
    0xbe4ee000U,
    0x232b5000U,
    0xec28b800U,
    0xda342c00U,
    0xfde6fa00U,
    0xdfdf8d00U,
    0x6eee1780U,
    0x5b1b0ac0U,
    0xe0000520U,
    0x500093f0U,
    0x38008488U,
    0x6c008e04U,
    0x9a000bceU,
    0x9d00d8ebU,
    0xcf803c88U,
    0x36c0a204U,
    0x2720f1ceU,
    0x22f055ebU,
    0xb108ab08U,
    0x35c4e8c4U,
    0xbb6e14eeU,
    0xb0db961bU,
    0x68a09780U,
    0x54304ac0U,
    0xf628e520U,
    0x734c3f0U,
    0x5266bc88U,
    0xf91fe204U,
    0x11ce11ceU,
    0x5eb05ebU,
    0x93081308U,
    0x84c4c4c4U,
    0x8eeeeeeeU,
    0xb1b1b1bU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0xb8000000U,
    0xac000000U,
    0x72000000U,
    0xb1000000U,
    0x3800000U,
    0xd2c00000U,
    0xc1600000U,
    0x9b900000U,
    0x4e480000U,
    0xb740000U,
    0x864e0000U,
    0x3f0b0000U,
    0x68068000U,
    0x447f4000U,
    0x7648a000U,
    0xe7747000U,
    0xd44e9800U,
    0xbe0b9c00U,
    0xd3864a00U,
    0x3abf5d00U,
    0xc528d180U,
    0xcde413c0U,
    0x99865ae0U,
    0x67bfd550U,
    0x94a8c528U,
    0x9e24cde4U,
    0xe3669986U,
    0x82ef67bfU,
    0x698014a8U,
    0xbfc0de24U,
    0x28e0c366U,
    0x6450b2efU,
    0x46a8d180U,
    0x5f2413c0U,
    0x78e65ae0U,
    0xcc2fd550U,
    0x62e0c528U,
    0x3950cde4U,
    0x17289986U,
    0xce467bfU,
    0x20694a8U,
    0x297f9e24U,
    0x9fc86366U,
    0x18b4c2efU,
    0xdcae4980U,
    0xea5b8fc0U,
    0x2d2e10e0U,
    0xc99b8850U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x88000000U,
    0x44000000U,
    0x4a000000U,
    0x47000000U,
    0xdd800000U,
    0x42400000U,
    0xc3200000U,
    0x77100000U,
    0x75b80000U,
    0x966c0000U,
    0x715e0000U,
    0xfc950000U,
    0xa6e68000U,
    0xd9f9c000U,
    0x28386000U,
    0x142cb000U,
    0x527e6800U,
    0xfb853400U,
    0x5b5e4200U,
    0xb95c300U,
    0x1366f780U,
    0xafb9b540U,
    0x2918f6a0U,
    0x603cc150U,
    0xb0469498U,
    0x68a9927cU,
    0x34a09b66U,
    0xc250ebb9U,
    0x3186318U,
    0x973c273cU,
    0x5c66dc6U,
    0x1ee92ae9U,
    0x35807780U,
    0xb6407540U,
    0xe12096a0U,
    0x4107150U,
    0x6a38fc98U,
    0xd72ca67cU,
    0x25fed966U,
    0x8ec528b9U,
    0xcdfe9498U,
    0x7ac5927cU,
    0xeffe9b66U,
    0x9c5ebb9U,
    0xf07ee318U,
    0x4885e73cU,
    0xa4de0dc6U,
    0x3ad59ae9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xd8000000U,
    0xfc000000U,
    0xf6000000U,
    0xd5000000U,
    0xbf800000U,
    0x2c400000U,
    0xeee00000U,
    0x9700000U,
    0x19080000U,
    0x21640000U,
    0xad6a0000U,
    0xd3130000U,
    0x22828000U,
    0x9707c000U,
    0x98e0a000U,
    0x1c709000U,
    0x8688f800U,
    0x5d24ac00U,
    0x9b8a2e00U,
    0x26632900U,
    0xcd8ac980U,
    0x63633940U,
    0x8a0af160U,
    0xe323b530U,
    0x4aea8fe8U,
    0xc3534414U,
    0x1a623a62U,
    0x1b774b77U,
    0xe668be68U,
    0xed54d154U,
    0x3302e502U,
    0x5247d747U,
    0x1f80f800U,
    0xbc40ac00U,
    0x16e02e00U,
    0xa5702900U,
    0x37084980U,
    0x864f940U,
    0xe4ea5160U,
    0x2a532530U,
    0x73e277e8U,
    0xb237e814U,
    0x6f081462U,
    0x34646277U,
    0x32ea77e8U,
    0xaf53e814U,
    0x14621462U,
    0x62776277U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x58000000U,
    0xac000000U,
    0x6a000000U,
    0x85000000U,
    0xfb800000U,
    0xa8c00000U,
    0x84200000U,
    0xae300000U,
    0x4b080000U,
    0xe0740000U,
    0x10860000U,
    0x388f0000U,
    0xfc2e8000U,
    0x320b4000U,
    0x2980e000U,
    0x91c01000U,
    0x2da03800U,
    0x7ff0fc00U,
    0x6a83200U,
    0xcf842900U,
    0x4e2e9180U,
    0x5b0b2dc0U,
    0xd800ffa0U,
    0xec0046f0U,
    0xa00af28U,
    0xd5001e44U,
    0xa380038eU,
    0x4c074fbU,
    0xee2086a8U,
    0x2b308f84U,
    0xb0882e2eU,
    0x48b40b0bU,
    0x94a68000U,
    0x96bf4000U,
    0xb726e000U,
    0xd27f1000U,
    0x3906b800U,
    0xa94fbc00U,
    0xd18ed200U,
    0x4dfb3900U,
    0x2f282980U,
    0x5e4491c0U,
    0x638e2da0U,
    0x24fb7ff0U,
    0xdea886a8U,
    0x23848f84U,
    0x442e2e2eU,
    0x8e0b0b0bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0xe8000000U,
    0x44000000U,
    0x5e000000U,
    0xad000000U,
    0xef800000U,
    0x68400000U,
    0x84600000U,
    0xfe500000U,
    0xfd280000U,
    0x7f40000U,
    0x2c620000U,
    0xda4f0000U,
    0x53068000U,
    0x12dfc000U,
    0x6f802000U,
    0xa8403000U,
    0x24602800U,
    0xae501400U,
    0x15283a00U,
    0x43f41100U,
    0x72621780U,
    0x774f2b40U,
    0xbc86bbe0U,
    0x7a9fda10U,
    0xebe00118U,
    0x56100f94U,
    0xd948174aU,
    0xa9a415fdU,
    0x394a3118U,
    0x99bb2793U,
    0x21648341U,
    0x6590eff7U,
    0xd3068000U,
    0xd2dfc000U,
    0xcf802000U,
    0xf8403000U,
    0xcc602800U,
    0xea501400U,
    0x4b283a00U,
    0xeef41100U,
    0x9de21780U,
    0x1f0f2b40U,
    0x38e6bbe0U,
    0x84cfda10U,
    0x16c80118U,
    0x51e40f94U,
    0xf52a174aU,
    0x73eb15fdU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x18000000U,
    0x4000000U,
    0xda000000U,
    0x9000000U,
    0x22800000U,
    0xe8400000U,
    0xbc600000U,
    0xe300000U,
    0x7b580000U,
    0x378c0000U,
    0x14c20000U,
    0x874d0000U,
    0x99d48000U,
    0xbfb94000U,
    0x18802000U,
    0x91403000U,
    0xe6e01800U,
    0x52702c00U,
    0x5380600U,
    0x34bc0100U,
    0x971a3680U,
    0x51810240U,
    0x13f688a0U,
    0xde847a10U,
    0x466c8f18U,
    0x1745738cU,
    0x91fa26d6U,
    0x73f111e3U,
    0x6ece9b30U,
    0x5e384cd3U,
    0x1376b6f5U,
    0x4bc45caeU,
    0x7a8c8000U,
    0x4c354000U,
    0xb6422000U,
    0xaf0d3000U,
    0x45b49800U,
    0x1896c00U,
    0x7bd82600U,
    0xa2cc3100U,
    0x28222e80U,
    0xdc3d2e40U,
    0xbe6c8ea0U,
    0x63457b10U,
    0x33fa3998U,
    0xcef131ccU,
    0x8e4e8e76U,
    0xbb785bf3U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0x88000000U,
    0x9c000000U,
    0x2e000000U,
    0x5000000U,
    0xab800000U,
    0x1c400000U,
    0x6e200000U,
    0x25100000U,
    0xfba80000U,
    0x94040000U,
    0xf26e0000U,
    0xb070000U,
    0xfeaa8000U,
    0x3fd1c000U,
    0xee202000U,
    0x65101000U,
    0xdba80800U,
    0xc4041400U,
    0x7a6e2200U,
    0x97072700U,
    0xd0aa8b80U,
    0x3ad1c140U,
    0x45a00ae0U,
    0x79501710U,
    0xb5881388U,
    0xe1141d44U,
    0x81c61ceaU,
    0x3030201U,
    0x22c4b71bU,
    0x31d6c381U,
    0xbb0ab54aU,
    0x4681d8e4U,
    0x5ba80800U,
    0x84041400U,
    0x5a6e2200U,
    0xc7072700U,
    0x58aa8b80U,
    0xa6d1c140U,
    0x6ba00ae0U,
    0x7c501710U,
    0x1e081388U,
    0xfd541d44U,
    0xefe61ceaU,
    0x26130201U,
    0xd96cb71bU,
    0xa5d2c381U,
    0x4964b54aU,
    0x4d86d8e4U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xc8000000U,
    0x3c000000U,
    0x3e000000U,
    0x67000000U,
    0xf9800000U,
    0xcc400000U,
    0x66600000U,
    0xb3100000U,
    0xaba80000U,
    0x5d240000U,
    0xc4fe0000U,
    0xb8cf0000U,
    0x66bb8000U,
    0x71a8c000U,
    0x10602000U,
    0x28103000U,
    0x4c280800U,
    0xa6641400U,
    0x931e3200U,
    0xfb9f0f00U,
    0x95738f80U,
    0xf89cd9c0U,
    0x86b61e60U,
    0x1bb0310U,
    0x880d9198U,
    0xdc13f8c4U,
    0x4e6db8eaU,
    0xff03e849U,
    0xdc596bfU,
    0xce27d3f3U,
    0x3f3bbdceU,
    0x2de8c47aU,
    0x9e000800U,
    0xf7001400U,
    0x11803200U,
    0xa0400f00U,
    0x90600f80U,
    0xe81019c0U,
    0x6c283e60U,
    0xf6643310U,
    0x5b1e1998U,
    0xc79f2cc4U,
    0xab73aaeaU,
    0x9f9cd749U,
    0x7f36113fU,
    0xcdfb1e33U,
    0xee6d91aeU,
    0x6f03c86aU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x58000000U,
    0x44000000U,
    0x7e000000U,
    0x69000000U,
    0x5b800000U,
    0xdc400000U,
    0x5a200000U,
    0x87100000U,
    0xdad80000U,
    0x9bec0000U,
    0xbc420000U,
    0xca0f0000U,
    0x6f7c8000U,
    0xc6d9c000U,
    0xa1a02000U,
    0xab501000U,
    0xf8f80800U,
    0xe8fc2c00U,
    0x409a1600U,
    0x7ce31100U,
    0xf6be9f80U,
    0xb996da40U,
    0xcf7cb6e0U,
    0x36d9e710U,
    0xd9a03e88U,
    0x5f501dc4U,
    0xdef828b6U,
    0xc5fc1bfbU,
    0x651a2690U,
    0xc9a339c3U,
    0xf71e92bfU,
    0xe2c6dce6U,
    0x4f848800U,
    0x2a25ec00U,
    0xbf3a3600U,
    0xeb30100U,
    0xdc69780U,
    0xc92af640U,
    0xabc6a0e0U,
    0xa42af610U,
    0xae46a108U,
    0xa16ac784U,
    0xf7e69e56U,
    0xbe3afcebU,
    0x91e9818U,
    0xcbc6e407U,
    0x34049a09U,
    0x4665d71dU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x48000000U,
    0x74000000U,
    0xc2000000U,
    0xe7000000U,
    0xb5800000U,
    0xba400000U,
    0x9b200000U,
    0xa3d00000U,
    0x2f180000U,
    0x81840000U,
    0xd82a0000U,
    0xcc190000U,
    0x5e078000U,
    0xe138c000U,
    0xd8982000U,
    0x9cc41000U,
    0x568a2800U,
    0x65892c00U,
    0xa23f9200U,
    0xb76cdd00U,
    0xedaa1080U,
    0x365929c0U,
    0x65278560U,
    0xf2e8c290U,
    0xbf8014c8U,
    0x694025f4U,
    0x4ca01346U,
    0x4e9035a1U,
    0x49b8096aU,
    0xec140096U,
    0xae1201c9U,
    0x94d297aU,
    0x1cb59080U,
    0x16e5e9c0U,
    0xc595a560U,
    0x1235d290U,
    0xff0dbcc8U,
    0x99f1c9f4U,
    0xf407a146U,
    0x8238f8a1U,
    0x471831eaU,
    0x5840556U,
    0xf22a16a9U,
    0xef1936eaU,
    0x618794c8U,
    0xc878e5f4U,
    0x34383346U,
    0x625425a1U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0x98000000U,
    0xb4000000U,
    0x52000000U,
    0x7000000U,
    0xbf800000U,
    0x5a400000U,
    0x3b200000U,
    0x91d00000U,
    0xd3380000U,
    0xfdec0000U,
    0x954a0000U,
    0x58f10000U,
    0xb5df8000U,
    0x91dc000U,
    0x86b82000U,
    0xa4ac1000U,
    0x7bea2800U,
    0xd0613c00U,
    0x2847a600U,
    0x8c61ed00U,
    0x166a3480U,
    0xcd2111c0U,
    0xce787e0U,
    0xb7f1ea90U,
    0x667208c8U,
    0x151d1974U,
    0x1895884eU,
    0x15ecc2bbU,
    0xf9678cb2U,
    0x1eb1fdacU,
    0x10d23f3fU,
    0x298d0bf3U,
    0xd70db480U,
    0x9790d1c0U,
    0xd635a7e0U,
    0x2d7cfa90U,
    0x5cffa0c8U,
    0xdfcde574U,
    0x4a000e4eU,
    0xf3003fbbU,
    0x4d801032U,
    0xad40106cU,
    0x1ca03edfU,
    0x7f901c63U,
    0xba1820c8U,
    0x6b3c2574U,
    0xf9f22e4eU,
    0xff5d2fbbU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0xf8000000U,
    0x4c000000U,
    0xa6000000U,
    0x89000000U,
    0x6e800000U,
    0x1a400000U,
    0x17600000U,
    0x4bf00000U,
    0xa2f80000U,
    0x7c5c0000U,
    0x7e360000U,
    0x551b0000U,
    0x40808000U,
    0x272d4000U,
    0x93982000U,
    0x7eac3000U,
    0x524e3800U,
    0x43071c00U,
    0xd1d6be00U,
    0x75c65300U,
    0xd7e08980U,
    0xacdd5240U,
    0xd16003a0U,
    0x72f02a90U,
    0xd47803d8U,
    0x5a1c1dfcU,
    0x37563f3eU,
    0xdbeb2e57U,
    0x2af8adadU,
    0xc8317196U,
    0x944e2e58U,
    0x7a072da7U,
    0xa756b180U,
    0x53864e40U,
    0x9e80bda0U,
    0x222d7990U,
    0xbb180a58U,
    0x9dec0fbcU,
    0xd3ae1c9eU,
    0x5eb734c7U,
    0xc24e9675U,
    0xcb6a706aU,
    0x65aeaf66U,
    0x9fda50f0U,
    0xf8b695adU,
    0x4b366d96U,
    0xa5989058U,
    0x7fc17ea7U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0xb8000000U,
    0x3c000000U,
    0xde000000U,
    0xdf000000U,
    0x29800000U,
    0x32400000U,
    0xe9200000U,
    0x62900000U,
    0x71d80000U,
    0x5e3c0000U,
    0x9f2e0000U,
    0x9e70000U,
    0x26b8000U,
    0x5176c000U,
    0x5ef82000U,
    0xafac1000U,
    0x81760800U,
    0xb69b0c00U,
    0x3be5ae00U,
    0xeb41cf00U,
    0x33eb9780U,
    0x2f36e7c0U,
    0xf1d82260U,
    0x1e3c1090U,
    0xbf2e1c48U,
    0x39e71ba4U,
    0xba6b85f6U,
    0x6d76ef4fU,
    0x80f83a2bU,
    0x70ac2929U,
    0xa8f638b2U,
    0x84db0c69U,
    0xd2c59f80U,
    0x89d1ebc0U,
    0x42338c60U,
    0x710adf90U,
    0x6ef60bc8U,
    0x17db3c64U,
    0xbd458796U,
    0x6891efdfU,
    0xe493ae63U,
    0xc2dafe8dU,
    0x18e3344U,
    0xc6373c26U,
    0x9313ba2bU,
    0x6f9ae929U,
    0xe12e18b2U,
    0xa6e71c69U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0xf8000000U,
    0x3c000000U,
    0x6e000000U,
    0x19000000U,
    0x50800000U,
    0xca400000U,
    0x7b200000U,
    0xafd00000U,
    0x97a80000U,
    0x4b9c0000U,
    0x55ae0000U,
    0x64ef0000U,
    0xf0288000U,
    0x68524000U,
    0x64082000U,
    0x820c1000U,
    0x8f262800U,
    0x75a33400U,
    0xf4aebe00U,
    0xa8614f00U,
    0x842ebb80U,
    0xf2215640U,
    0xa70e9c20U,
    0xb1f15690U,
    0xa6a6a8c8U,
    0xdf6d40f4U,
    0xcd88886aU,
    0x68c27fa7U,
    0x16002ccbU,
    0x650006ebU,
    0x9e803b62U,
    0x3403e30U,
    0xd3a01380U,
    0x59902240U,
    0x82880220U,
    0xfd4c0990U,
    0x92863b48U,
    0xe53322b4U,
    0xdea6aa4aU,
    0xa36d6637U,
    0x388bf83U,
    0xa1c2505fU,
    0xbe800f28U,
    0x93400707U,
    0x8ba03f83U,
    0xb590105fU,
    0x14882f28U,
    0xd84c1707U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0xc8000000U,
    0xbc000000U,
    0x4e000000U,
    0x57000000U,
    0x80800000U,
    0xa400000U,
    0xfd200000U,
    0x8db00000U,
    0xffa80000U,
    0xa6840000U,
    0x110e0000U,
    0x4bdf0000U,
    0x74d78000U,
    0xb8724000U,
    0x84082000U,
    0x8a741000U,
    0xbd061800U,
    0xedab3400U,
    0x2fd1b200U,
    0x6ed96f00U,
    0xad59b380U,
    0x5ed45c0U,
    0x23ff9820U,
    0x38b66690U,
    0x8e263548U,
    0x771b286cU,
    0x30f9866aU,
    0x121d6761U,
    0x8977a5e3U,
    0x7f827aa7U,
    0xe68029ddU,
    0x71403e20U,
    0x9ba02b80U,
    0xbcf031c0U,
    0x4080a20U,
    0xca741990U,
    0xdd061ec8U,
    0x3dab19acU,
    0xe7d18c4aU,
    0xd2d97ef1U,
    0xe359bb2bU,
    0x52ed630bU,
    0xa37fa597U,
    0x32f640d1U,
    0x730610abU,
    0xfaab12cbU,
    0xcf518fb7U,
    0xb4994941U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0xa8000000U,
    0xd4000000U,
    0xfa000000U,
    0xf9000000U,
    0x92800000U,
    0x19400000U,
    0x42a00000U,
    0x21500000U,
    0x8ef80000U,
    0xa7040000U,
    0x59920000U,
    0x36f90000U,
    0x2b2e8000U,
    0xffd04000U,
    0x51922000U,
    0x12f91000U,
    0x592e8800U,
    0x62d06c00U,
    0x91120a00U,
    0x26b92500U,
    0x730eb680U,
    0xa3c05240U,
    0xcfca2ea0U,
    0xb9ad2350U,
    0xe6c4a628U,
    0x136d5a14U,
    0x338e8d1eU,
    0xd7804a91U,
    0xc5ea104cU,
    0xc8bd07aaU,
    0x101cafd5U,
    0x58794965U,
    0x5c44a628U,
    0x9e2d5a14U,
    0xab2e8d1eU,
    0xbfd04a91U,
    0x7192104cU,
    0xa2f907aaU,
    0xf12eafd5U,
    0xb6d04965U,
    0x6b122628U,
    0xdfb91a14U,
    0xe18ead1eU,
    0xba805a91U,
    0x8d6a184cU,
    0x98fd2baaU,
    0x683c85d5U,
    0xb4697c65U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x58000000U,
    0x1c000000U,
    0x72000000U,
    0x4f000000U,
    0xa1800000U,
    0x77400000U,
    0x4da00000U,
    0xbd300000U,
    0xaef80000U,
    0x369c0000U,
    0x8ab60000U,
    0xa8850000U,
    0xfe18000U,
    0xea0dc000U,
    0xf3362000U,
    0x83c51000U,
    0xd041b800U,
    0xa83dec00U,
    0xa44e3600U,
    0xde191700U,
    0x6557a480U,
    0xf288ffc0U,
    0xa4d79e60U,
    0x75c8cad0U,
    0x517797e8U,
    0x64f8c08cU,
    0xd58f8ddeU,
    0x164eb77U,
    0x8cb9a345U,
    0x91a1edadU,
    0x6f7812a6U,
    0xb1dc0234U,
    0x7f1617e8U,
    0xb9b5008cU,
    0x8b19addeU,
    0x8f91fb77U,
    0xaa001b45U,
    0x130001adU,
    0x338024a6U,
    0x88401534U,
    0xb4203368U,
    0xd6703f4cU,
    0x915813beU,
    0xc4ac21a7U,
    0x85ce34adU,
    0xe9592d21U,
    0xc8f79f78U,
    0xffb8e943U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x88000000U,
    0x34000000U,
    0xa2000000U,
    0x3000000U,
    0x41800000U,
    0xf7400000U,
    0x3a00000U,
    0x4100000U,
    0x9a080000U,
    0x4f140000U,
    0xfb20000U,
    0xea550000U,
    0xd73b8000U,
    0x13a1c000U,
    0x2c122000U,
    0xfe451000U,
    0x6533a800U,
    0x38b5d400U,
    0x9a00200U,
    0x23101d00U,
    0x51880080U,
    0xdf5414c0U,
    0x67923260U,
    0x2e0530d0U,
    0xad13a868U,
    0xace5c1c4U,
    0xfb8816e2U,
    0xa8543e15U,
    0x24122b04U,
    0x8a452f91U,
    0x6733b14cU,
    0x6bb5da2dU,
    0xc0200068U,
    0xe05015c4U,
    0xf02814e2U,
    0xd8442315U,
    0xbc1a2b84U,
    0x96513b51U,
    0xa101832cU,
    0x42a0eafdU,
    0xb6bba800U,
    0xf4e1d400U,
    0x7b20200U,
    0x9e551d00U,
    0xd53b8080U,
    0x40a1d4c0U,
    0xe5921260U,
    0x3d0520d0U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0xb8000000U,
    0x1c000000U,
    0x82000000U,
    0xfb000000U,
    0xed800000U,
    0x87400000U,
    0xffa00000U,
    0x24300000U,
    0xde480000U,
    0x992c0000U,
    0xc6e60000U,
    0xd2dd0000U,
    0x64938000U,
    0x59a7c000U,
    0x1462000U,
    0xaaed1000U,
    0xd8dbb800U,
    0xeb8bf400U,
    0x92200e00U,
    0xe3701700U,
    0xc1e81880U,
    0x6d1c0ac0U,
    0xa0ae1560U,
    0x57f126d0U,
    0x20759f68U,
    0x707af7ccU,
    0x8855acf2U,
    0x740ad79bU,
    0x263d9651U,
    0x6556d9bbU,
    0x94b398b6U,
    0x91d7d322U,
    0x952e2768U,
    0x5cb103ccU,
    0x5d5a2f2U,
    0x634ac09bU,
    0x819d8ed1U,
    0x8d66d37bU,
    0x70fb8dd6U,
    0xeffbf5f2U,
    0x3c483800U,
    0xf22c3400U,
    0x73662e00U,
    0x999d0700U,
    0xa133a080U,
    0x9a97fec0U,
    0xb08e1b60U,
    0x4f8131d0U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x48000000U,
    0xac000000U,
    0x6000000U,
    0x95000000U,
    0x5800000U,
    0xc9400000U,
    0x3be00000U,
    0x8100000U,
    0xcc680000U,
    0xb6740000U,
    0xcd5e0000U,
    0xe1a70000U,
    0x635c8000U,
    0xa8e1c000U,
    0x98be2000U,
    0xb73000U,
    0x44b4a800U,
    0xfed5c400U,
    0x25803200U,
    0x19401b00U,
    0xd3e02980U,
    0xb4102140U,
    0x82681360U,
    0x8f741950U,
    0xcede0f78U,
    0xbde72744U,
    0x5d3cb27aU,
    0x69b1dfcdU,
    0x6f361dafU,
    0xbed30a6dU,
    0x458283cdU,
    0xa906c3a5U,
    0x8b82a778U,
    0x5006e344U,
    0x2802807aU,
    0x1c46c4cdU,
    0x5e62b42fU,
    0x7116eb2dU,
    0xafeab0adU,
    0x5a72eaf5U,
    0xab5c8000U,
    0xc4e1c000U,
    0x3ebe2000U,
    0x85b73000U,
    0x934a800U,
    0x9b95c400U,
    0x18603200U,
    0x84501b00U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0xf8000000U,
    0x34000000U,
    0x1a000000U,
    0xff000000U,
    0xf3800000U,
    0x93400000U,
    0x2da00000U,
    0x3e700000U,
    0x3d480000U,
    0x88cc0000U,
    0x52b20000U,
    0x8d910000U,
    0xce358000U,
    0x750cc000U,
    0x94922000U,
    0x84a11000U,
    0x5cdd9800U,
    0xd8b0f400U,
    0xeae81e00U,
    0xd9bc1d00U,
    0x47a1e80U,
    0x721d0bc0U,
    0x532782e0U,
    0xdede9d0U,
    0x8e6fade8U,
    0x1521e05cU,
    0x44dd8bb2U,
    0x7cb0e2e3U,
    0x68e819c4U,
    0xc2bc05f8U,
    0x15fa1c5fU,
    0x2a5d39b0U,
    0x9707b5e8U,
    0x5fddd45cU,
    0x6d07b5b2U,
    0x30ddefe3U,
    0x6879f44U,
    0x479dfa38U,
    0xc92780bfU,
    0xb2edcd60U,
    0x1def8680U,
    0x5661ffc0U,
    0x917d9ce0U,
    0x76c0f4d0U,
    0x4fa03368U,
    0xb5702b9cU,
    0xb4c82952U,
    0x348c1b33U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0x28000000U,
    0xfc000000U,
    0xb2000000U,
    0x5b000000U,
    0x3f800000U,
    0x7f400000U,
    0x89e00000U,
    0x22700000U,
    0xb3680000U,
    0xa3a40000U,
    0xdd360000U,
    0xfaad0000U,
    0xe1a38000U,
    0x7e6ec000U,
    0x71562000U,
    0xc09d3000U,
    0x36ab9800U,
    0xcbfac400U,
    0x81682a00U,
    0x38a40f00U,
    0x82b63480U,
    0x95ed12c0U,
    0x404385e0U,
    0xa01ee0d0U,
    0x703e2ef8U,
    0x38392e5cU,
    0xd41dbb3aU,
    0x4e17f339U,
    0xe92bbf35U,
    0x64baf937U,
    0x40880032U,
    0xf6d415b2U,
    0xabde36f8U,
    0x91492a5cU,
    0x10f5b13aU,
    0x7ef3cc39U,
    0x27fd93b5U,
    0x1b67eff7U,
    0x9fc38fd2U,
    0xf5eca62U,
    0xb1de3480U,
    0xf64912c0U,
    0xfd7585e0U,
    0x4ab3e0d0U,
    0xb99daef8U,
    0xba57ee5cU,
    0x174b9b3aU,
    0xd58ac339U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x8000000U,
    0x4c000000U,
    0xf6000000U,
    0x7f000000U,
    0x76800000U,
    0x19400000U,
    0x11a00000U,
    0x7bf00000U,
    0x8af80000U,
    0xa7540000U,
    0x42ae0000U,
    0xcb170000U,
    0xe4a58000U,
    0x8c124000U,
    0xd6562000U,
    0x2f431000U,
    0x4e8b9800U,
    0x5d454c00U,
    0xabd3a200U,
    0xf2e14300U,
    0x83058580U,
    0xc8e243c0U,
    0x4a2e27a0U,
    0xa1570950U,
    0x1585a3e8U,
    0xa1a25e3cU,
    0x338e209eU,
    0xa6a73345U,
    0x617dace3U,
    0x35f679a9U,
    0xf1a0205fU,
    0xbf0117dU,
    0xe2f82668U,
    0xdb541dfcU,
    0xbcae073eU,
    0xf8173a15U,
    0x64258f0bU,
    0xea526795U,
    0xb17620c1U,
    0x4df33238U,
    0xd5d3928bU,
    0x81e16855U,
    0x6385a561U,
    0x9ea27868U,
    0x250e34e3U,
    0x8fe735a9U,
    0x78dd825fU,
    0x206527dU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0xb8000000U,
    0x7c000000U,
    0x4a000000U,
    0xf3000000U,
    0x90800000U,
    0x81400000U,
    0x5fa00000U,
    0xfb900000U,
    0x5dd80000U,
    0x8cec0000U,
    0x5b360000U,
    0xc4b10000U,
    0xdf338000U,
    0x52974000U,
    0x166e2000U,
    0x891d1000U,
    0x7ba5a800U,
    0x1db65c00U,
    0x2c858e00U,
    0x2b664f00U,
    0x7cfd9a80U,
    0xa31a70c0U,
    0x18938220U,
    0xe5077350U,
    0x19b62368U,
    0xfaf11124U,
    0x4213a7d6U,
    0xd7477cabU,
    0x76963985U,
    0xf0211c58U,
    0xf86ba9f2U,
    0xdc3b49eaU,
    0x3a7839e8U,
    0x4b7c21e4U,
    0xecee05f6U,
    0xcb5d1ffbU,
    0xac85b2edU,
    0x6b66517cU,
    0xdcfd8024U,
    0xd31a7a41U,
    0xa0939aedU,
    0x99074d7cU,
    0x53b62e24U,
    0x9f12541U,
    0xd293a86dU,
    0x560761bcU,
    0x29362204U,
    0xbb11911U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x88000000U,
    0xd4000000U,
    0xea000000U,
    0xb7000000U,
    0xf5800000U,
    0xa5400000U,
    0xfea00000U,
    0x7e900000U,
    0x3eb80000U,
    0x9ef40000U,
    0x2e820000U,
    0xa6d90000U,
    0x729d8000U,
    0x98c9c000U,
    0x2fba2000U,
    0xda6d1000U,
    0x7f3fa800U,
    0x81c0ec00U,
    0xff3f8200U,
    0xc1c0e500U,
    0x5f3fb280U,
    0x71c0d1c0U,
    0xd73f9760U,
    0xa5c0e050U,
    0x3d3faf28U,
    0x12c0fb64U,
    0xc8bfa24eU,
    0xb780ea2dU,
    0x361f99e8U,
    0xc910fb82U,
    0x8a78141U,
    0x57e4d3bbU,
    0x26259da8U,
    0xf13deaa4U,
    0x54b8152eU,
    0x69f41a7dU,
    0x7b021ec0U,
    0xb3992ce6U,
    0x43d810fU,
    0x3259cc96U,
    0xfb021ec0U,
    0xf3992ce6U,
    0xa43d810fU,
    0x8259cc96U,
    0x73021ec0U,
    0x27992ce6U,
    0x4e3d810fU,
    0x3559cc96U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0x8000000U,
    0x34000000U,
    0x1a000000U,
    0xd1000000U,
    0xac800000U,
    0x57400000U,
    0x43a00000U,
    0x18d00000U,
    0xd480000U,
    0xb2b40000U,
    0xe4620000U,
    0x52010000U,
    0xc5668000U,
    0xe6e94000U,
    0x8e0a2000U,
    0xdb251000U,
    0x55ec8800U,
    0x9f8c5400U,
    0x6c6a200U,
    0xbe395d00U,
    0xa3422e80U,
    0x39913040U,
    0xb98ea120U,
    0xf98d4cd0U,
    0xd9a03468U,
    0x89d02f74U,
    0x81c826f2U,
    0xb5f4193dU,
    0xafc200d0U,
    0x7ed10a64U,
    0xd22ea463U,
    0x855d6763U,
    0xc6e812e8U,
    0xde640b34U,
    0xd32a05d2U,
    0x61b518edU,
    0x85849238U,
    0xd7a84150U,
    0x12cc81b1U,
    0xf41c6f8eU,
    0x7a2e88d0U,
    0xa15d5e64U,
    0xf4e80663U,
    0x6b643a63U,
    0x6daa3c68U,
    0xd3f53b74U,
    0x70a4a4f2U,
    0x4938543dU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x8000000U,
    0xe4000000U,
    0xe6000000U,
    0x7000000U,
    0x10800000U,
    0x7d400000U,
    0x5da00000U,
    0x8f00000U,
    0x21180000U,
    0x37940000U,
    0xfdfa0000U,
    0xd8ef0000U,
    0xb9258000U,
    0x2be14000U,
    0xf7c22000U,
    0xddcb1000U,
    0x48e79800U,
    0x412a7c00U,
    0xc7a5a200U,
    0xf5a16900U,
    0x3ce20180U,
    0x5f7b2dc0U,
    0x2cdf9e20U,
    0xe70e5a50U,
    0xa0e78ce8U,
    0x152a6afcU,
    0x49a58de6U,
    0xe6a17f75U,
    0xc2621636U,
    0xc13b3e57U,
    0x87ff92e7U,
    0x95be41e1U,
    0xccdf9568U,
    0x570e7b3cU,
    0xc8e791c6U,
    0x12a5c25U,
    0xa7a5835eU,
    0x5a1456bU,
    0x34e20321U,
    0xbb7b1dc4U,
    0xcadf9636U,
    0xe00e7e57U,
    0xb067b2e7U,
    0x686a51e1U,
    0x14058d68U,
    0xee51473cU,
    0xe37a13c6U,
    0xf6af2525U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x38000000U,
    0xac000000U,
    0xa2000000U,
    0xcf000000U,
    0x57800000U,
    0x2fc00000U,
    0x63a00000U,
    0x51b00000U,
    0x16e80000U,
    0xd5740000U,
    0xf4e20000U,
    0xfa130000U,
    0x33448000U,
    0x5dc74000U,
    0xc4c4a000U,
    0x2077000U,
    0xbf64a800U,
    0x4fb75c00U,
    0x338ca600U,
    0xf9c37700U,
    0x32ee8e80U,
    0xe31044c0U,
    0x358a1b60U,
    0xc0a70f30U,
    0x8406a388U,
    0x46646b5cU,
    0xd9680e32U,
    0x26b40201U,
    0x2d42150aU,
    0x78a30b85U,
    0xe82cb75bU,
    0xc4736834U,
    0xa606950aU,
    0x49644b85U,
    0xaee8175bU,
    0xb9741834U,
    0x76e23d0aU,
    0x85131785U,
    0x5cc4b15bU,
    0xde076f34U,
    0x564b38aU,
    0x9cb75345U,
    0xfe0caa3bU,
    0xb5036004U,
    0xa4ce9002U,
    0x52607819U,
    0x17420409U,
    0x6ba31205U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x78000000U,
    0x6c000000U,
    0x7e000000U,
    0xff000000U,
    0x18800000U,
    0xc0c00000U,
    0x7ca00000U,
    0x5ab00000U,
    0xd9b80000U,
    0xc7040000U,
    0x94f20000U,
    0x8eed0000U,
    0xebe28000U,
    0x5676c000U,
    0xb62a000U,
    0x3ab6f000U,
    0x29c2a800U,
    0x8f06f400U,
    0x90fab600U,
    0xe4c2ef00U,
    0x6a8a980U,
    0xcf9fd0c0U,
    0x2c722fa0U,
    0x9e2d20f0U,
    0xcf429088U,
    0x70c6c65cU,
    0xd4da8ee6U,
    0x6eb2c39dU,
    0xdbb0bddaU,
    0x3e2bff26U,
    0x1f3806a2U,
    0x28c40e7bU,
    0xa8d23ddaU,
    0x689d3f26U,
    0x48faa6a2U,
    0x58c2fe7bU,
    0x20a895daU,
    0x4c9fcb26U,
    0x32f210a2U,
    0xcded117bU,
    0xd562bc5aU,
    0x15b6dbe6U,
    0x69429f02U,
    0x33c6c18bU,
    0xea5a84d2U,
    0x2d72e9baU,
    0xb990a7e4U,
    0x375bed16U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xd8000000U,
    0xf4000000U,
    0xd2000000U,
    0xab000000U,
    0x98800000U,
    0x90c00000U,
    0xeca00000U,
    0x82f00000U,
    0xe7e80000U,
    0x2a040000U,
    0xaf3e0000U,
    0x32b70000U,
    0xfff28000U,
    0x7e46c000U,
    0x4d72a000U,
    0x4186f000U,
    0x93528800U,
    0x3cb6fc00U,
    0xa9abe00U,
    0x5b82c100U,
    0xe46c8a80U,
    0xfa01ebc0U,
    0x27682ca0U,
    0x8ec40ff0U,
    0x319e3788U,
    0x2b471f4cU,
    0x589aa672U,
    0x3082d9cdU,
    0xdcec9bbdU,
    0x5ac1d860U,
    0x13c838c1U,
    0xf8342131U,
    0x4761bbdU,
    0xaa431860U,
    0x6f2498c1U,
    0x92f5d131U,
    0xcfbe13bdU,
    0xa6772460U,
    0xb95286c1U,
    0x93b6e031U,
    0x381a913dU,
    0xa442f3a0U,
    0x9a4cb461U,
    0xb731dec1U,
    0x66a02435U,
    0x1df03b2cU,
    0xd6820b3U,
    0x21c439fcU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0x8000000U,
    0xc000000U,
    0x72000000U,
    0xf9000000U,
    0x4a800000U,
    0x86c00000U,
    0x14e00000U,
    0x7db00000U,
    0xf280000U,
    0x8dec0000U,
    0xe70a0000U,
    0x11830000U,
    0xad578000U,
    0xecdec000U,
    0x99b7a000U,
    0xe16ed000U,
    0x3e9f8800U,
    0x5082dc00U,
    0xa3958a00U,
    0xb401df00U,
    0x36421680U,
    0x271f2140U,
    0xf195a420U,
    0x3d01d0f0U,
    0xd4c22918U,
    0x9ddf139cU,
    0x9f75a0d2U,
    0xb5b1efe7U,
    0xe36a0f90U,
    0x6ff30ac7U,
    0x261fa0e5U,
    0x5f42f100U,
    0x55f58790U,
    0x7371d6c7U,
    0x578a2ae5U,
    0x22432e00U,
    0x21379110U,
    0xdeaef787U,
    0xc0ff8ec5U,
    0x9bf2fef0U,
    0xb05db808U,
    0x485de41bU,
    0xac602e17U,
    0x42701117U,
    0xf1483798U,
    0x469c2edcU,
    0xf4c22ef2U,
    0xeddf3017U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x28000000U,
    0xe4000000U,
    0x1e000000U,
    0xd000000U,
    0x4f800000U,
    0x3c00000U,
    0xb9e00000U,
    0xcad00000U,
    0xd8780000U,
    0xbc2c0000U,
    0xe27e0000U,
    0x8f410000U,
    0x90ef8000U,
    0xbb1c4000U,
    0xe68fa000U,
    0x320c5000U,
    0xe717b800U,
    0x14f04400U,
    0xf511b200U,
    0xc39d7d00U,
    0x99803580U,
    0xfac03e40U,
    0xa0600660U,
    0x70102eb0U,
    0x18183018U,
    0x9c3c0804U,
    0xd2660c06U,
    0xf77d1e0fU,
    0x5c89b319U,
    0x41617e9fU,
    0xf58624c2U,
    0x70ad00a8U,
    0xab718b19U,
    0xae8d7a9fU,
    0x861836c2U,
    0xd13c2da8U,
    0xfde60699U,
    0xa4bd00dfU,
    0xcd6982a2U,
    0x6fb17e18U,
    0x33fe0301U,
    0xc181369bU,
    0x68f88c4U,
    0x220c4ea7U,
    0xaf178000U,
    0xa0f04000U,
    0xc311a000U,
    0x2a9d5000U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x38000000U,
    0x2c000000U,
    0xd2000000U,
    0x8d000000U,
    0x70800000U,
    0x14c00000U,
    0xb2e00000U,
    0x51f00000U,
    0xf6280000U,
    0xb740000U,
    0x23c20000U,
    0x8b7b0000U,
    0x63858000U,
    0xab51c000U,
    0xd3e5a000U,
    0x9361d000U,
    0xffada800U,
    0x4125fc00U,
    0x72a7a600U,
    0x31daf700U,
    0x66481280U,
    0x83441440U,
    0x378a2ea0U,
    0x753f0170U,
    0x3c8f8a18U,
    0x56aef90cU,
    0xb78a1992U,
    0x353f20d1U,
    0x1c8f8de2U,
    0xe6aedd4fU,
    0x8f8a2f23U,
    0x193f05abU,
    0xce8fa5e2U,
    0x6baee14fU,
    0xff0a2923U,
    0xdff22abU,
    0x7c6f9f62U,
    0x3a5ec90fU,
    0x9220183U,
    0x68b04dbU,
    0x5fadaffaU,
    0xb125d843U,
    0x6aa790b1U,
    0xaddad27aU,
    0x8c483a80U,
    0x22442840U,
    0x950a28a0U,
    0xecff2670U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xd8000000U,
    0xec000000U,
    0xf2000000U,
    0x65000000U,
    0x87800000U,
    0x5c00000U,
    0x48a00000U,
    0xcb100000U,
    0x58f80000U,
    0xb3340000U,
    0x84d20000U,
    0xc9130000U,
    0xd5f58000U,
    0x50944000U,
    0x470da000U,
    0xfaa07000U,
    0xe5fb800U,
    0xef736400U,
    0x3e8a0e00U,
    0xf8371f00U,
    0x1c5f9280U,
    0x1a737640U,
    0x10a0b60U,
    0x41f71330U,
    0x7eff9748U,
    0x58637ef4U,
    0x2c7233f6U,
    0x92031479U,
    0x350d81a2U,
    0x5fa0610dU,
    0xe9dfb597U,
    0xbab340dcU,
    0xae2a1322U,
    0xdf27174dU,
    0xb6a7bef7U,
    0xcc4753ecU,
    0x258046aU,
    0x8d2429b9U,
    0xe3aa2d01U,
    0xc3e73795U,
    0x3387bdc8U,
    0xdb976cb4U,
    0xbf803696U,
    0x79c01849U,
    0x2a0046aU,
    0x121029b9U,
    0xf5782d01U,
    0x3ff43795U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0x58000000U,
    0xc4000000U,
    0x66000000U,
    0x3b000000U,
    0x39800000U,
    0xd7c00000U,
    0x10a00000U,
    0xbb700000U,
    0xf9f80000U,
    0x77f40000U,
    0x80a60000U,
    0xe30d0000U,
    0x3db48000U,
    0x11c64000U,
    0xbbcca000U,
    0xdaf27000U,
    0xea4a8800U,
    0x14f5400U,
    0xa61e00U,
    0x230d2500U,
    0x9db4a780U,
    0x81c65bc0U,
    0xe3cca1e0U,
    0x1ef27a30U,
    0x8c4a9bc8U,
    0x3a4f41ecU,
    0x39262a36U,
    0xf4cd23d1U,
    0x8d14bdffU,
    0x3ab65022U,
    0x1a34b0daU,
    0x69065b7fU,
    0xcec9a7fU,
    0xd9424be2U,
    0x492b13aU,
    0xe50b514fU,
    0x36d809b7U,
    0xe0441e0eU,
    0xf07e250cU,
    0x6849279eU,
    0xc4a9bc8U,
    0xfa4f41ecU,
    0x99262a36U,
    0x64cd23d1U,
    0xd514bdffU,
    0xfeb65022U,
    0x7c34b0daU,
    0x52065b7fU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0xb8000000U,
    0xb4000000U,
    0xfa000000U,
    0x47000000U,
    0xd1800000U,
    0x1fc00000U,
    0xe2e00000U,
    0x94100000U,
    0x4a580000U,
    0xf240000U,
    0xcd8e0000U,
    0xe9bb0000U,
    0xebe48000U,
    0xf8a64000U,
    0xc35ca000U,
    0x23925000U,
    0xa48a9800U,
    0xd50d5400U,
    0x3ae03600U,
    0x70103900U,
    0xe8582880U,
    0xec2438c0U,
    0x5e0e24e0U,
    0x57b3b30U,
    0x2284b258U,
    0x34767334U,
    0xba64be4eU,
    0xa766713dU,
    0xc1bc814dU,
    0xa78248a3U,
    0x56d2ad0cU,
    0x6e297e8eU,
    0xd6e31cdU,
    0xdeab2463U,
    0xd23cbfecU,
    0xb427cbeU,
    0x7fb2ab15U,
    0xb2f96f97U,
    0xcc562542U,
    0xee5f36b3U,
    0x4d0a9800U,
    0x3ecd5400U,
    0xc2003600U,
    0xb3003900U,
    0xcb802880U,
    0x48c038c0U,
    0x8b6024e0U,
    0x3fd03b30U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0xb8000000U,
    0x14000000U,
    0xd2000000U,
    0x6d000000U,
    0x25800000U,
    0x73c00000U,
    0x54e00000U,
    0x38500000U,
    0x54380000U,
    0xb2440000U,
    0x3d7e0000U,
    0x9dbf0000U,
    0x67958000U,
    0x86ad4000U,
    0x554da000U,
    0x71b95000U,
    0xc18bb800U,
    0x69824400U,
    0xa5801600U,
    0x33c00100U,
    0x34e00280U,
    0x68500a40U,
    0xec3813e0U,
    0xa64402b0U,
    0xef7e28d8U,
    0xf0bf09a4U,
    0x42158956U,
    0xf56d7e75U,
    0x1adaf69U,
    0x49e955eaU,
    0x95b3bbb4U,
    0xdbc66e55U,
    0x98fe15e9U,
    0xae7f1baaU,
    0x5375be54U,
    0xeefd6de5U,
    0xb975bfb1U,
    0xd7fd584eU,
    0x2ef584e2U,
    0x993d4120U,
    0xe7958000U,
    0xc6ad4000U,
    0x354da000U,
    0x21b95000U,
    0x798bb800U,
    0x7d824400U,
    0x77801600U,
    0x5ec00100U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x38000000U,
    0x2c000000U,
    0x86000000U,
    0x79000000U,
    0xe2800000U,
    0xd8c00000U,
    0xafe00000U,
    0xc0100000U,
    0xa0280000U,
    0x10140000U,
    0xc8720000U,
    0x14490000U,
    0xaa698000U,
    0xff0ec000U,
    0x9ba1a000U,
    0x3a0ad000U,
    0x777b9800U,
    0x6f97ec00U,
    0x60001600U,
    0xb0002700U,
    0xd8001780U,
    0xdc002940U,
    0xbe001720U,
    0x55002370U,
    0x648032d8U,
    0xa1c01874U,
    0x4d603b52U,
    0x18d00231U,
    0xfc831eeU,
    0xd0043113U,
    0x685a308cU,
    0x45d3ed4U,
    0x621bbe6eU,
    0xeb47f453U,
    0x31c831acU,
    0xc5043aa4U,
    0xecda1b36U,
    0x559d0567U,
    0x177bbddeU,
    0xdf97cbe5U,
    0xb8000000U,
    0x6c000000U,
    0x66000000U,
    0x89000000U,
    0xda800000U,
    0xf4c00000U,
    0x29e00000U,
    0xb9100000U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x48000000U,
    0x34000000U,
    0x3e000000U,
    0x1b000000U,
    0xe0800000U,
    0xe2c00000U,
    0xd3a00000U,
    0xc6500000U,
    0xa7080000U,
    0xacc0000U,
    0xf7e60000U,
    0x60010000U,
    0xf0188000U,
    0xa80ac000U,
    0x430a000U,
    0x7656f000U,
    0x2f7e9800U,
    0xdecbfc00U,
    0xf9880a00U,
    0x330c3100U,
    0x24c62580U,
    0x749107c0U,
    0xccb0a5a0U,
    0x5096f370U,
    0x6adea348U,
    0x79bffe4U,
    0xc8003d0aU,
    0xf4003797U,
    0xde000ad3U,
    0x2b002a27U,
    0xa88035bdU,
    0xd6c03b71U,
    0xeda03753U,
    0xdd5011e7U,
    0x47883a1dU,
    0xe80c0901U,
    0x2446299bU,
    0xa65115c3U,
    0x5710a8b7U,
    0xa2c6fce6U,
    0xf3d6a580U,
    0x1657c7c0U,
    0xdf6605a0U,
    0x76c10370U,
    0xfdb8bb48U,
    0x455ac3e4U,
    0xbb8970aU,
    0xaa5af697U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0x98000000U,
    0x9c000000U,
    0x4e000000U,
    0x59000000U,
    0x7800000U,
    0xddc00000U,
    0xdea00000U,
    0x1a300000U,
    0x23080000U,
    0x34a40000U,
    0xa13a0000U,
    0x8bc50000U,
    0xdb958000U,
    0x73d04000U,
    0x57bda000U,
    0x75847000U,
    0xfaafa800U,
    0x38154c00U,
    0xac280e00U,
    0xf6542b00U,
    0x35123d80U,
    0xd1910d40U,
    0x1887b460U,
    0x97414630U,
    0x9eba05c8U,
    0xfa0517bcU,
    0xf335b68aU,
    0x5ce040d5U,
    0xa5359124U,
    0x59e07e54U,
    0xccb5ade9U,
    0x2d20696cU,
    0x8d9584a4U,
    0x76d07f14U,
    0x3e3db789U,
    0x144745cU,
    0xbb8f94ecU,
    0x63e569e8U,
    0x1f801b63U,
    0x81c029b9U,
    0xb0a01580U,
    0xb3300140U,
    0xbc881a60U,
    0x75641d30U,
    0x319a1048U,
    0xc8f516fcU,
    0xff1daceaU,
    0x9ab45de5U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x8000000U,
    0x84000000U,
    0x92000000U,
    0x91000000U,
    0xbd800000U,
    0x8cc00000U,
    0x61600000U,
    0xc5b00000U,
    0x30d80000U,
    0x6f6c0000U,
    0x4af60000U,
    0xa530000U,
    0x5d2d8000U,
    0x8bc04000U,
    0x9fdba000U,
    0x45935000U,
    0x70f62800U,
    0x4f531400U,
    0x5aad8a00U,
    0x2006500U,
    0xd93b8680U,
    0x19e35540U,
    0xece23e0U,
    0xf84f1370U,
    0xfc63bd38U,
    0x2e4f775cU,
    0x9f5812eeU,
    0x32ac3ff7U,
    0xb6161d6bU,
    0x53231a3fU,
    0x495b0ceU,
    0xa51c5338U,
    0x77f5a053U,
    0xb1ac6d63U,
    0xdaada220U,
    0x42006ccfU,
    0xf93bbd38U,
    0x9e3775cU,
    0x6ce12eeU,
    0x7c4f3ff7U,
    0x6e639d6bU,
    0xbf4f5a3fU,
    0x22d810ceU,
    0xbe6c0338U,
    0xd7760853U,
    0x96933963U,
    0x344d8820U,
    0xca7059cfU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0x78000000U,
    0xac000000U,
    0x3a000000U,
    0xd000000U,
    0xf1800000U,
    0x6cc00000U,
    0xf5200000U,
    0x9df00000U,
    0x76a80000U,
    0x8640000U,
    0x141a0000U,
    0xb6230000U,
    0xc75f8000U,
    0x84944000U,
    0x3145a000U,
    0xa3b77000U,
    0x659a2800U,
    0x1ae30c00U,
    0x127f9600U,
    0xe9645700U,
    0x3fedb080U,
    0x7d35840U,
    0x4b801ae0U,
    0xa1c01470U,
    0x24a01728U,
    0x1302b4cU,
    0xfb883062U,
    0x39940d25U,
    0x58b22a4cU,
    0xb34737e1U,
    0x22c5b5f9U,
    0x5e7770e1U,
    0x33a3d64U,
    0xbad31cadU,
    0x2277859bU,
    0xb1307dc4U,
    0x63ff9728U,
    0x45a46b4cU,
    0xeacd9062U,
    0x6a237d25U,
    0x4528024cU,
    0x5a43be1U,
    0xaba23f9U,
    0xba1327e1U,
    0xcd578de4U,
    0xd1c044edU,
    0x9cd79f7bU,
    0x8d0069b4U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x98000000U,
    0x6c000000U,
    0xaa000000U,
    0x83000000U,
    0xd7800000U,
    0xc0c00000U,
    0xa1600000U,
    0x30d00000U,
    0x99280000U,
    0x8cf40000U,
    0x9b4a0000U,
    0xfbdb0000U,
    0x8ae88000U,
    0x12644000U,
    0x7f42a000U,
    0x35af5000U,
    0x87e21800U,
    0x28ef1c00U,
    0xb5429e00U,
    0xc6af5700U,
    0x28622c80U,
    0xb42f2bc0U,
    0x2622a760U,
    0x197f5cf0U,
    0xccca1bb8U,
    0x7b1b3704U,
    0xcb889c92U,
    0x12b443c9U,
    0x7e6ab378U,
    0xd55b66fbU,
    0xb6a81eb0U,
    0x50340a9bU,
    0xe82a30c0U,
    0x140b0dffU,
    0xf640bc22U,
    0xb1504e52U,
    0x38e8b738U,
    0xbd645cc4U,
    0xe2c29bf2U,
    0x466f4f39U,
    0x690230c0U,
    0xb4ff0dffU,
    0x270abc22U,
    0xf98b4e52U,
    0xfd803738U,
    0x3c01cc4U,
    0x96e03bf2U,
    0xc0101f39U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x28000000U,
    0x8c000000U,
    0x2e000000U,
    0xc3000000U,
    0xae800000U,
    0x79c00000U,
    0x9d200000U,
    0xe5d00000U,
    0xb680000U,
    0xd2ec0000U,
    0x1fa20000U,
    0xe2690000U,
    0x4d328000U,
    0x3dd8c000U,
    0xcf30a000U,
    0x40a1f000U,
    0xdaca3800U,
    0x3853c00U,
    0xb4109200U,
    0x1a71ef00U,
    0x19222180U,
    0xd7a923c0U,
    0x9e12b820U,
    0x2b08e2b0U,
    0x42d8a6e8U,
    0x678df404U,
    0x76481612U,
    0xc73c010fU,
    0x5cca3c92U,
    0x8c853d51U,
    0x5490b26cU,
    0x90b1dd58U,
    0x282227aU,
    0xc7b93555U,
    0x265a967eU,
    0xdf34c357U,
    0xf8928768U,
    0x2ec8d7c4U,
    0xb9f8ae32U,
    0xfd5de3bfU,
    0xd5a01a7aU,
    0x23100955U,
    0x5ec8047eU,
    0x31fc2c57U,
    0x216a26e8U,
    0xe3953404U,
    0x4458b612U,
    0x524df10fU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x48000000U,
    0x6c000000U,
    0x4e000000U,
    0x3b000000U,
    0x94800000U,
    0xc1c00000U,
    0xbe200000U,
    0xb3500000U,
    0x98880000U,
    0xffdc0000U,
    0xcd320000U,
    0x4bc10000U,
    0x17728000U,
    0x7aabc000U,
    0xeac8a000U,
    0x12b6f000U,
    0x56883800U,
    0x4dc2c00U,
    0x39b20a00U,
    0xfa010700U,
    0xe1528180U,
    0xa5fbd5c0U,
    0x3c4096a0U,
    0xd66aceb0U,
    0xf3a32a8U,
    0x8edd30a4U,
    0x90e083aaU,
    0x33fad423U,
    0x931234eeU,
    0x489108c7U,
    0xa7fa9830U,
    0x9977eeeaU,
    0x21fa87c6U,
    0xe77eda3U,
    0x9b7a8d3aU,
    0x84b7f479U,
    0xf9da8180U,
    0x9a27d5c0U,
    0x917296a0U,
    0xedabceb0U,
    0x5048b2a8U,
    0x9876f0a4U,
    0x342823aaU,
    0x1a4c2423U,
    0x511a0ceeU,
    0x8d8d24c7U,
    0x20689230U,
    0xd026e9eaU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x58000000U,
    0x44000000U,
    0x1a000000U,
    0xf1000000U,
    0x4e800000U,
    0xf5c00000U,
    0x32600000U,
    0x3d100000U,
    0x28f80000U,
    0xcaa40000U,
    0xcfee0000U,
    0x337f0000U,
    0xbbad8000U,
    0xc14bc000U,
    0xa6bba000U,
    0x1990d000U,
    0xa4783800U,
    0xca643400U,
    0xc90e0e00U,
    0x9aaf3500U,
    0xb7b59080U,
    0x873fed40U,
    0x69cdb520U,
    0x2c5bd130U,
    0xb643a738U,
    0x734c634U,
    0x299628a6U,
    0x4c1b18edU,
    0x2623b145U,
    0x5f24f736U,
    0x6dee3e30U,
    0x567f3cbdU,
    0xd72d86fdU,
    0x118bdc42U,
    0x985ba3b6U,
    0x6440f560U,
    0xea601080U,
    0x39102d40U,
    0x52f81520U,
    0xaba40130U,
    0xd96e1f38U,
    0x82bf3234U,
    0x93cd86a6U,
    0xd5bfdedU,
    0xc0c399c5U,
    0x26f4ee76U,
    0x59f62510U,
    0xc40b088dU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0xc8000000U,
    0xe4000000U,
    0x42000000U,
    0xbd000000U,
    0x6a800000U,
    0x5c00000U,
    0x2a200000U,
    0x89100000U,
    0xf0880000U,
    0x64dc0000U,
    0x2eb60000U,
    0x97830000U,
    0x4f578000U,
    0x3fe7c000U,
    0x9b69a000U,
    0x55b8f000U,
    0x32081800U,
    0xb51c0c00U,
    0x6e960a00U,
    0xb7930500U,
    0x5f5fa280U,
    0x7fbd640U,
    0xb77faa20U,
    0xf3ebde30U,
    0xcd778828U,
    0x62f7ef34U,
    0x1e19caaU,
    0x9864ce73U,
    0xfc3e0d0fU,
    0x7e5f2697U,
    0x23619858U,
    0xb9a4c129U,
    0x741e27a7U,
    0xba4f1fe3U,
    0x7169aed2U,
    0x3cb8d16aU,
    0x32882280U,
    0x19dc1640U,
    0xa4360a20U,
    0x62432e30U,
    0xad779028U,
    0x52f7e334U,
    0x29e196aaU,
    0x8c64cb73U,
    0x763e2f8fU,
    0x275f30d7U,
    0xbe19278U,
    0x164ef19U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0x48000000U,
    0xdc000000U,
    0x92000000U,
    0x53000000U,
    0x6c800000U,
    0x85c00000U,
    0x36600000U,
    0xe5500000U,
    0xc9f80000U,
    0xac6c0000U,
    0x8a6a0000U,
    0x27570000U,
    0x32e88000U,
    0xcfbc000U,
    0xd5faa000U,
    0x9e00d000U,
    0x29181800U,
    0x13fc1400U,
    0x23722a00U,
    0x74ab3300U,
    0xf19ab680U,
    0x6850e3c0U,
    0x6c601fa0U,
    0x2a5025b0U,
    0xd7782eb8U,
    0x6aac1c24U,
    0x988a2de6U,
    0x9bc7254fU,
    0x5f70b464U,
    0x16c7f60eU,
    0xfae89500U,
    0x90fbfb9bU,
    0xa7faac5cU,
    0xdd00c9eaU,
    0xd980746U,
    0x4a3c2b64U,
    0x87122e80U,
    0xc2fb37c0U,
    0x54e295a0U,
    0x41fcc6b0U,
    0xd06a0038U,
    0xe8572be4U,
    0x2c68b846U,
    0xca3be3ffU,
    0xc71ab45cU,
    0x2290ddeaU,
    0x44802d46U,
    0x9c01864U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x98000000U,
    0x6c000000U,
    0x2e000000U,
    0x71000000U,
    0x7c800000U,
    0xebc00000U,
    0xd2200000U,
    0x67500000U,
    0xd1d80000U,
    0xf1640000U,
    0xbc9a0000U,
    0x8bd10000U,
    0x2678000U,
    0xff1ac000U,
    0xbda5a000U,
    0xdf6ff000U,
    0xcdf83800U,
    0xf7340400U,
    0xe9c23e00U,
    0x2d752f00U,
    0xdaddad80U,
    0xe9bc740U,
    0x3c9a34a0U,
    0x4bd116b0U,
    0x6267b3a8U,
    0x2f1ad724U,
    0x25a586feU,
    0xb36fce8dU,
    0xe3f828d0U,
    0x863406edU,
    0x95420e9fU,
    0xc6b508c2U,
    0x8fdb6f8U,
    0x69cbd689U,
    0xed421cc1U,
    0xbab520ffU,
    0xdefd9580U,
    0xa4cbc340U,
    0x27c20aa0U,
    0x4c7539b0U,
    0x5e5d9e28U,
    0x595bd064U,
    0x58ba125eU,
    0x3181283dU,
    0xe13fa378U,
    0x44bed5c9U,
    0x379fb661U,
    0xb42ee94fU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xb8000000U,
    0xac000000U,
    0x6000000U,
    0xfd000000U,
    0xef800000U,
    0xf8c00000U,
    0x8c200000U,
    0xf6300000U,
    0xe5480000U,
    0x73c40000U,
    0x46ca0000U,
    0xdd750000U,
    0x1fcd8000U,
    0xe0814000U,
    0x106fa000U,
    0x48007000U,
    0xb4200800U,
    0x9a303c00U,
    0x43480600U,
    0xbec42700U,
    0x114a2f80U,
    0x89b51440U,
    0x95edba60U,
    0xebb14170U,
    0x1aa7b8e8U,
    0xc30473bcU,
    0x7eca125aU,
    0xb1751d7dU,
    0xb9cdaee0U,
    0x2d814cacU,
    0x47ef99c1U,
    0x1cc06b3dU,
    0x3e003188U,
    0x91001750U,
    0x498037fbU,
    0x35c01030U,
    0xdba00800U,
    0xa2f03c00U,
    0x6f680600U,
    0x78f42700U,
    0x4c022f80U,
    0x56711440U,
    0xd527ba60U,
    0xcbc44170U,
    0xeaea38e8U,
    0xdb4533bcU,
    0xe285b25aU,
    0xf456d7dU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x58000000U,
    0x14000000U,
    0x5a000000U,
    0x75000000U,
    0x6c800000U,
    0x87c00000U,
    0xdc600000U,
    0xf6700000U,
    0xcb780000U,
    0x4b840000U,
    0xd2660000U,
    0x79070000U,
    0x82c78000U,
    0xf8e4c000U,
    0x9db9a000U,
    0x917d000U,
    0xcae00800U,
    0x14b00400U,
    0x83983e00U,
    0x7e341100U,
    0xc77e0080U,
    0xa5f31840U,
    0xad598da0U,
    0x38a7fcb0U,
    0x7df80c38U,
    0xf9440c6cU,
    0xc2862dc6U,
    0x58b73b7dU,
    0xcddf9047U,
    0x5110c9a0U,
    0xdea7a2a8U,
    0x4e94ed38U,
    0xf6c194ffU,
    0x1293d98cU,
    0x40863cceU,
    0x79b73bf5U,
    0x5b5f8800U,
    0xf3d0c400U,
    0x36479e00U,
    0x2b24c100U,
    0xbbd98880U,
    0xda67dc40U,
    0x351813a0U,
    0xccf43db0U,
    0xd79e04b8U,
    0x8443102cU,
    0xe2419e66U,
    0x9153d6cdU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0xe8000000U,
    0x94000000U,
    0x42000000U,
    0x7b000000U,
    0x49800000U,
    0x3cc00000U,
    0x90200000U,
    0x58500000U,
    0x1c080000U,
    0xa64c0000U,
    0xd13e0000U,
    0xa6eb0000U,
    0x375c8000U,
    0xd7f94000U,
    0x81caa000U,
    0x78ce7000U,
    0x2a003800U,
    0x2f002c00U,
    0x6b802200U,
    0x37c03900U,
    0x31a02a80U,
    0xf0903bc0U,
    0xce2802e0U,
    0x851c11f0U,
    0x84b63668U,
    0x3c671924U,
    0x7642a30aU,
    0x29427f87U,
    0xaa9e134dU,
    0x97b3029U,
    0x7af4a198U,
    0xf1254044U,
    0x76dcb7a5U,
    0xcf397ecdU,
    0xdbea8272U,
    0xbf9e6733U,
    0xd5880000U,
    0x5a8c0000U,
    0x211e0000U,
    0x8ebb0000U,
    0xc3548000U,
    0xe5b54000U,
    0x12f4a000U,
    0xa5257000U,
    0x54dcb800U,
    0xc4396c00U,
    0x7a6a8200U,
    0x175e4900U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0x28000000U,
    0x14000000U,
    0x4a000000U,
    0xe3000000U,
    0x6f800000U,
    0x72c00000U,
    0x70200000U,
    0xe8300000U,
    0x34080000U,
    0xba3c0000U,
    0xcb0a0000U,
    0x7b850000U,
    0x38d28000U,
    0x9318c000U,
    0x87abe000U,
    0x46d4b000U,
    0xca000800U,
    0x23000c00U,
    0x4f800200U,
    0x82c00f00U,
    0x58200280U,
    0xfc300140U,
    0x7e0804a0U,
    0x593c0e30U,
    0xa48a06f8U,
    0x945072cU,
    0x48f28702U,
    0x7b28ce83U,
    0xb3a3e340U,
    0xfce8bba3U,
    0x10a04b0U,
    0x58850bb8U,
    0x7752818dU,
    0x11d8c631U,
    0xdf8beafaU,
    0xbae4b52dU,
    0xb4080000U,
    0x7a3c0000U,
    0xeb0a0000U,
    0x8b850000U,
    0x10d28000U,
    0x8718c000U,
    0xcdabe000U,
    0xa5d4b000U,
    0xa5800800U,
    0x51c00c00U,
    0x3fa00200U,
    0x6af00f00U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x28000000U,
    0x9c000000U,
    0x7e000000U,
    0xff000000U,
    0x43800000U,
    0x79c00000U,
    0xb8200000U,
    0x14100000U,
    0x52380000U,
    0xf9140000U,
    0x88a0000U,
    0xd8670000U,
    0x40ff8000U,
    0x108fc000U,
    0x7c78e000U,
    0x6ad27000U,
    0x5d800800U,
    0x96c00400U,
    0x33a00e00U,
    0xa1d00500U,
    0xbc180280U,
    0x8e0409c0U,
    0x673207e0U,
    0xa7b30ff0U,
    0xb3d58438U,
    0xa538c79cU,
    0xd69f6b82U,
    0x9759b141U,
    0x7b4aed23U,
    0xdd617b91U,
    0x26558688U,
    0xfff8cc86U,
    0xb33f688fU,
    0x5589bdc8U,
    0xfad2eaa7U,
    0xd5a5749dU,
    0xbac78000U,
    0x359bc000U,
    0xeaf2e000U,
    0x1db57000U,
    0x76ff8800U,
    0x638fc400U,
    0x89f8ee00U,
    0x20127500U,
    0xf0200a80U,
    0x98100dc0U,
    0xe43809e0U,
    0xca140af0U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0x48000000U,
    0x1c000000U,
    0xae000000U,
    0xf9000000U,
    0x6c800000U,
    0x95c00000U,
    0x7c200000U,
    0x3e300000U,
    0xe1080000U,
    0x489c0000U,
    0x6fd20000U,
    0x37270000U,
    0x59b8000U,
    0xe1764000U,
    0xcde72000U,
    0xb8277000U,
    0x94200800U,
    0x92300c00U,
    0x27080200U,
    0xdd9c0700U,
    0xe5520480U,
    0x47e701c0U,
    0xbb3b8ae0U,
    0xb3864f90U,
    0x3c4f26c8U,
    0x5b4b795cU,
    0x66da0fc2U,
    0xd3bb0fe3U,
    0xac498c10U,
    0x43514389U,
    0x42fca27dU,
    0x299132b2U,
    0xe76722b9U,
    0x78e77d87U,
    0x42800016U,
    0xacc009deU,
    0x30a00800U,
    0xdbf00c00U,
    0xd5280200U,
    0x6aac0700U,
    0x20da0480U,
    0x86bb01c0U,
    0x6c98ae0U,
    0x43914f90U,
    0xb45ca6c8U,
    0x6761395cU,
    0xb8cf2fc2U,
    0x628b7fe3U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0x48000000U,
    0xbc000000U,
    0xe000000U,
    0xe1000000U,
    0xb5800000U,
    0x3dc00000U,
    0x8c200000U,
    0xd6100000U,
    0x75180000U,
    0xd7b40000U,
    0x9ad20000U,
    0x648f0000U,
    0x50538000U,
    0x25c04000U,
    0x38296000U,
    0x4157000U,
    0x4a200800U,
    0xcb100400U,
    0xae980600U,
    0xdb740d00U,
    0xeb720480U,
    0x335f0bc0U,
    0xa76b80e0U,
    0xc5644e10U,
    0x62636b58U,
    0x8aee73dcU,
    0xc8180c2U,
    0x5c4f4961U,
    0xb3fae151U,
    0x2d15307bU,
    0x3a9652dU,
    0x98d57988U,
    0x13800be5U,
    0xf0c0054cU,
    0x1fa00a5aU,
    0x66d0055dU,
    0xab80800U,
    0x61640400U,
    0xd86a0600U,
    0xb9eb0d00U,
    0x86398480U,
    0x7d2b4bc0U,
    0xb90e0e0U,
    0x44fe3e10U,
    0xcd90e358U,
    0x59fe37dcU,
    0x1610e6c2U,
    0x553e3461U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xd0000000U,
    0x98000000U,
    0x34000000U,
    0x12000000U,
    0x43000000U,
    0x4800000U,
    0xb8400000U,
    0x46200000U,
    0x41300000U,
    0x3fb80000U,
    0x58f40000U,
    0x74460000U,
    0x701d0000U,
    0x680c8000U,
    0x9c1cc000U,
    0x6e132000U,
    0xfd051000U,
    0x61980800U,
    0xedc40c00U,
    0xb9fe0e00U,
    0xbbe90d00U,
    0x80ca8980U,
    0x6041c340U,
    0x523fa120U,
    0x6329d430U,
    0x34b32848U,
    0xf0751784U,
    0xea000262U,
    0x67000513U,
    0x6e80047bU,
    0x1f400bcfU,
    0xc8a00fe4U,
    0x8e700071U,
    0x6f1807e8U,
    0xe2840675U,
    0x95e02cbU,
    0xd1990047U,
    0x65d28180U,
    0xf5c5cf40U,
    0x4de1af20U,
    0x49f0d930U,
    0x13c1a1c8U,
    0xfcc0d4c4U,
    0xde79a342U,
    0x3734d123U,
    0x36bfac33U,
    0xcb69dc4bU,
    0xa932d86U,
    0x55451562U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0x98000000U,
    0xec000000U,
    0xe000000U,
    0x29000000U,
    0x9f800000U,
    0xa9400000U,
    0x52200000U,
    0x8f300000U,
    0x32a80000U,
    0x1cd40000U,
    0xa8460000U,
    0x89ab0000U,
    0xac5b8000U,
    0x63964000U,
    0x5f65e000U,
    0x673f5000U,
    0xd6880800U,
    0xc6e40c00U,
    0x336e0a00U,
    0xa93f0500U,
    0x5fbd8980U,
    0x94d4ec0U,
    0x23660e0U,
    0x170d1290U,
    0xdea3e1f8U,
    0x12d45694U,
    0x81738722U,
    0x160241f3U,
    0x503e0aaU,
    0x31a45a0dU,
    0xd07b8be4U,
    0x55a648caU,
    0xca4de9ddU,
    0x6eab5b3dU,
    0xbaee08acU,
    0x57f06d4U,
    0x3c1d8180U,
    0x563d42c0U,
    0x653e6ae0U,
    0xc1a91790U,
    0x186de878U,
    0x219b5854U,
    0x284607c2U,
    0x49ab0363U,
    0xc5b8952U,
    0x33964099U,
    0xc765e6c6U,
    0x8b3f5c39U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0xf8000000U,
    0xfc000000U,
    0x1e000000U,
    0x2b000000U,
    0x67800000U,
    0xc5400000U,
    0xab200000U,
    0x27900000U,
    0x65680000U,
    0x9b2c0000U,
    0xdfae0000U,
    0x99570000U,
    0x852b8000U,
    0xf4a4c000U,
    0xfecee000U,
    0x405ad000U,
    0x5fae0800U,
    0xd9570400U,
    0x252b8a00U,
    0xc4a4c300U,
    0x6ceef80U,
    0xbc5adfc0U,
    0x41ae09e0U,
    0xf25706b0U,
    0x42ab8c78U,
    0x1e4cf54U,
    0xadeee532U,
    0x9bcaddb9U,
    0x24c60fb6U,
    0x697b0f02U,
    0x9d058182U,
    0x98b3c6c1U,
    0x28c56d60U,
    0x6f6e12f3U,
    0xda08e05aU,
    0x2921db07U,
    0xc2ab8c78U,
    0x41e4cf54U,
    0xdeee532U,
    0xabcaddb9U,
    0xdcc60fb6U,
    0x957b0f02U,
    0x83058182U,
    0xb3b3c6c1U,
    0x4f456d60U,
    0xaa2e12f3U,
    0x7128e05aU,
    0xeb1db07U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x68000000U,
    0x9c000000U,
    0x6000000U,
    0x2f000000U,
    0xf8800000U,
    0x2a400000U,
    0x7f200000U,
    0x30900000U,
    0xc6780000U,
    0x81040000U,
    0xeb8a0000U,
    0xa4df0000U,
    0x82458000U,
    0x4321c000U,
    0x46b12000U,
    0x11571000U,
    0x8d8a0800U,
    0x5bdf0400U,
    0xf2c58e00U,
    0x6561c900U,
    0x57912680U,
    0x92c719c0U,
    0xb5720860U,
    0xdf9b06f0U,
    0x9eef8188U,
    0xdb6ecba4U,
    0x6c8ca172U,
    0x6072dac9U,
    0xde312407U,
    0xeb171ee0U,
    0x7aaa0730U,
    0x674f05e9U,
    0x5abd8756U,
    0x5765c7fbU,
    0x429b2eecU,
    0x33581235U,
    0xb0978188U,
    0x866acba4U,
    0x6106a172U,
    0x7baddac9U,
    0xccf4a407U,
    0x1e76dee0U,
    0x453b2730U,
    0x698815e9U,
    0xe9cf8f56U,
    0xa7fec3fbU,
    0x24f4a0ecU,
    0xc276db35U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0xa8000000U,
    0x44000000U,
    0xc2000000U,
    0x13000000U,
    0xcf800000U,
    0xe2400000U,
    0x71200000U,
    0x6cb00000U,
    0xa5c80000U,
    0xa77c0000U,
    0x77ba0000U,
    0x9e690000U,
    0xf048000U,
    0x2182c000U,
    0x5740e000U,
    0x1fa51000U,
    0xfa720800U,
    0xbd150c00U,
    0x9abe8200U,
    0xdcebc700U,
    0x3fc46a80U,
    0x9867d440U,
    0x1e12e420U,
    0xdd001d30U,
    0xa8486f8U,
    0x24c2c524U,
    0xa3e0ef92U,
    0xb655158bU,
    0x8b1a04fcU,
    0xc3990307U,
    0x346c85a3U,
    0x780ec1f2U,
    0x5c12e19aU,
    0xe001eb7U,
    0xe5048c1aU,
    0xb682c076U,
    0x7ac0ec78U,
    0x9ee51164U,
    0xecd20bb2U,
    0x77e508bbU,
    0x8c568204U,
    0x427c623U,
    0x22366a31U,
    0x4332d479U,
    0x178c6566U,
    0xe5bddb0U,
    0xf708e9b9U,
    0xbd991184U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0x18000000U,
    0x7c000000U,
    0x8e000000U,
    0x6f000000U,
    0x52800000U,
    0x1fc00000U,
    0x59200000U,
    0x71b00000U,
    0x2b780000U,
    0x5de40000U,
    0x90160000U,
    0xd8170000U,
    0x9c1f8000U,
    0x9e19c000U,
    0x770da000U,
    0x2ebb7000U,
    0x91ee0800U,
    0x36330c00U,
    0x23298e00U,
    0x34bec100U,
    0x4ea2180U,
    0xe186b7c0U,
    0xf355a0e0U,
    0xc1ef7af0U,
    0xe000328U,
    0xaf000cfcU,
    0xb2800a12U,
    0xfc001dbU,
    0x41200bd7U,
    0xdb004eeU,
    0xa57804c9U,
    0x32e40a8dU,
    0xc296097bU,
    0xc7d70f06U,
    0xc53f8055U,
    0xefa9c02eU,
    0x5c75a2a8U,
    0x735f7b3cU,
    0x1f80af2U,
    0xee240b2bU,
    0xbf3600ffU,
    0xaaa70412U,
    0x73e780dbU,
    0xcf3dca56U,
    0x62bba32cU,
    0xf7dc7c28U,
    0x2d29847cU,
    0x9bbec053U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x8000000U,
    0xf4000000U,
    0xa6000000U,
    0x77000000U,
    0x65800000U,
    0xd3c00000U,
    0x45200000U,
    0xe4900000U,
    0xd9680000U,
    0xbf4c0000U,
    0x28720000U,
    0x5de50000U,
    0x361d8000U,
    0x8f0bc000U,
    0x39a26000U,
    0x31ce7000U,
    0x9c3a0800U,
    0x2390400U,
    0xc9078a00U,
    0x5ea2cb00U,
    0xec4de080U,
    0xb3e0bf40U,
    0x1525e260U,
    0xfcacb370U,
    0x9557e458U,
    0xe549b23cU,
    0xd4a66d2U,
    0xe9427e09U,
    0xf7680576U,
    0x7c4c0cc4U,
    0x4bf2043fU,
    0x49250bd2U,
    0x1ebd8b8bU,
    0x4c5bc3b5U,
    0x3ea6666U,
    0x1d127cedU,
    0x8a00cd8U,
    0x3350097cU,
    0x92480eb2U,
    0x68dc0679U,
    0x3a9a01aeU,
    0xb26901b8U,
    0x98cf808dU,
    0x92bec6abU,
    0xf677eaa5U,
    0x36d9bd4dU,
    0x11a2648bU,
    0x35ce7936U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x48000000U,
    0xf4000000U,
    0x26000000U,
    0x61000000U,
    0x17800000U,
    0x8c00000U,
    0xbb200000U,
    0x4b00000U,
    0xe8580000U,
    0x5d540000U,
    0x1cc20000U,
    0x8d350000U,
    0x4d958000U,
    0xdbe64000U,
    0x3bbee000U,
    0x32d4b000U,
    0xb83a0800U,
    0xcc110c00U,
    0x2a2f8600U,
    0x2b374d00U,
    0xecb16480U,
    0xac53ff40U,
    0xe3536a60U,
    0xc1d6fa10U,
    0x489eef78U,
    0x264b18cU,
    0x16620132U,
    0x20450e0bU,
    0x696d8e65U,
    0x5ac24185U,
    0x3c04e654U,
    0x1205b10fU,
    0x27358473U,
    0xa6964769U,
    0x5746e1f4U,
    0x47f0bb3fU,
    0x998003f8U,
    0x8dc002ccU,
    0xa2a00d52U,
    0x4970091bU,
    0x2af80d9dU,
    0xc4240349U,
    0x7e3a0b06U,
    0xbd110814U,
    0x15af81eeU,
    0x7f74820U,
    0x39916cf2U,
    0x3de3fe2bU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0x78000000U,
    0x74000000U,
    0x7e000000U,
    0x5f000000U,
    0xd0800000U,
    0x75400000U,
    0x7d200000U,
    0x2d900000U,
    0x18f80000U,
    0x85fc0000U,
    0xd86e0000U,
    0xb8950000U,
    0x496b8000U,
    0xef0dc000U,
    0x8bb2000U,
    0x9179d000U,
    0xb360800U,
    0x7eb90400U,
    0xc25d8e00U,
    0xd1b4c700U,
    0x2ae6a780U,
    0x30cd1740U,
    0x59d0afe0U,
    0x3a7411f0U,
    0xe58d2b08U,
    0xb4c0d454U,
    0x1feb8652U,
    0xf14dc699U,
    0x3b1b2fefU,
    0xe6a9ddefU,
    0xc66e08eeU,
    0xd7950a6dU,
    0x1eb87acU,
    0x9e4dc98dU,
    0x739b2c5eU,
    0x97e9d385U,
    0xbd4e0488U,
    0xd1050714U,
    0xb79387b2U,
    0x31f1c069U,
    0x6552b67U,
    0x77acdafbU,
    0x91fd8f5cU,
    0x9664ca04U,
    0x7fbeaccbU,
    0x9de11376U,
    0x9c66a302U,
    0x5e8d1981U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0xc8000000U,
    0x24000000U,
    0x8e000000U,
    0x39000000U,
    0x6a800000U,
    0x60400000U,
    0x5aa00000U,
    0xf8700000U,
    0x96a80000U,
    0xc2540000U,
    0xe99a0000U,
    0xb5dd0000U,
    0x6d798000U,
    0xb6334000U,
    0xa5332000U,
    0xb8b35000U,
    0xab798800U,
    0x6b334c00U,
    0x61b32200U,
    0x71f35900U,
    0x53598480U,
    0xd7034e40U,
    0x23bb2ae0U,
    0x72d75a90U,
    0x46eb8228U,
    0xc0ca4844U,
    0xfdf8af4aU,
    0x89491517U,
    0x18092b42U,
    0xc1e5461U,
    0x1a2809d3U,
    0xef14024aU,
    0xbfba0795U,
    0xa0ed0a02U,
    0x8df18500U,
    0xf1574e81U,
    0x64212b42U,
    0x6e0a5461U,
    0x891209d3U,
    0x32b9024aU,
    0x8c6b8795U,
    0xf08a4a02U,
    0x4f58a500U,
    0xc5391e81U,
    0xc8a12342U,
    0xd34a5861U,
    0x17320bd3U,
    0x3890b4aU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0xf8000000U,
    0xbc000000U,
    0xca000000U,
    0x39000000U,
    0x13800000U,
    0x55400000U,
    0xbba00000U,
    0xd1700000U,
    0x6d880000U,
    0xf2440000U,
    0xbf360000U,
    0x8ab0000U,
    0x9be48000U,
    0x5b754000U,
    0x34986000U,
    0x91ec1000U,
    0xc2648800U,
    0xf7354c00U,
    0x3cb86a00U,
    0xc5dc1d00U,
    0xec4c8780U,
    0x680147c0U,
    0x240666a0U,
    0x6331e90U,
    0xdb1e06b8U,
    0x6e9f0294U,
    0x30da8d1aU,
    0x1dda4387U,
    0x406ae860U,
    0xfa0251b0U,
    0x713064e9U,
    0x2798120dU,
    0xb7a8c5eU,
    0xcaa4ec7U,
    0x8de2e680U,
    0xd8465243U,
    0x36066860U,
    0x933311b0U,
    0x5a9e04e9U,
    0x6edf020dU,
    0xaafa845eU,
    0x1cea42c7U,
    0x4fc2ec80U,
    0xb5765f43U,
    0xbae67e0U,
    0x59471a70U,
    0xf9800849U,
    0x7c40019dU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0xf8000000U,
    0xe4000000U,
    0xfa000000U,
    0xad000000U,
    0xb6800000U,
    0x89c00000U,
    0x92a00000U,
    0x53d00000U,
    0x6fb80000U,
    0x2d5c0000U,
    0xfa460000U,
    0xa1c50000U,
    0xfea88000U,
    0xd5d64000U,
    0xec992000U,
    0x34d23000U,
    0x8c088800U,
    0xf6064400U,
    0x1b212600U,
    0xcd8e3300U,
    0x744e8780U,
    0x1ec34a40U,
    0xa909a9a0U,
    0x3c9879d0U,
    0xbcf7ace8U,
    0xf00172dcU,
    0xd819288aU,
    0xb41238edU,
    0x32288c13U,
    0xb1164309U,
    0xa8b920aeU,
    0xdec238b1U,
    0x89108a59U,
    0x6c8a4784U,
    0x74df2a8fU,
    0xec173d20U,
    0xc6200413U,
    0xe3100709U,
    0x299806aeU,
    0x8e4c0bb1U,
    0xb3de0dd9U,
    0x1f890dc4U,
    0xb576832fU,
    0x2e5f44f0U,
    0xa3efa8fbU,
    0xb78d75d5U,
    0x99672e24U,
    0xc84b335cU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0x68000000U,
    0xe4000000U,
    0x86000000U,
    0x9d000000U,
    0xe1800000U,
    0xb0c00000U,
    0xeda00000U,
    0x12f00000U,
    0x16980000U,
    0x7e740000U,
    0x2fc20000U,
    0xc72d0000U,
    0x56b38000U,
    0x5e624000U,
    0xdfe7e000U,
    0xbf387000U,
    0xda938800U,
    0x3c524c00U,
    0xc4dfee00U,
    0xc3bc7100U,
    0x8bc98e80U,
    0x610b4240U,
    0x3bae6660U,
    0xc7f338d0U,
    0xe31de098U,
    0x3091794cU,
    0xd37a00baU,
    0x566905ffU,
    0xebc987f1U,
    0xb10b43abU,
    0xb3ae62c6U,
    0x33f33acdU,
    0xd1dea7aU,
    0x49917cddU,
    0xb4fa09c0U,
    0x7ba900a2U,
    0xe7e98ff1U,
    0x133b4fabU,
    0x48966cc6U,
    0x5f773bcdU,
    0x3447ecfaU,
    0xf0c8729dU,
    0xcd8b81a0U,
    0xe2e64972U,
    0x6ebde1e9U,
    0xf26174a7U,
    0x4de20a1cU,
    0xdc1d06e2U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0xa8000000U,
    0x34000000U,
    0xd2000000U,
    0x59000000U,
    0xd6800000U,
    0xf1400000U,
    0x9aa00000U,
    0x8f500000U,
    0xada80000U,
    0x96cc0000U,
    0xa9420000U,
    0x46a10000U,
    0x49468000U,
    0x56af4000U,
    0xb1672000U,
    0xbaa51000U,
    0xff668800U,
    0x5bf4400U,
    0xa2ef2200U,
    0x7b791700U,
    0x1fac8280U,
    0x9fc24740U,
    0xa7e3af20U,
    0x2beb5290U,
    0x35e72fe8U,
    0x52e51854U,
    0x93468e8aU,
    0xbaf4e65U,
    0x3de72f32U,
    0x56e51238U,
    0xc9468c9eU,
    0x16af4e4fU,
    0x91672c06U,
    0xcaa51182U,
    0x576687c0U,
    0x31bf4f61U,
    0x70ef2732U,
    0x22791638U,
    0xc92c8e9eU,
    0x6e82494fU,
    0x3d43a686U,
    0xa4bb52c2U,
    0x984f2ae0U,
    0xc4291af1U,
    0x3a04825aU,
    0x4d0e4d2cU,
    0x74a1ad34U,
    0x4a52baU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x88000000U,
    0xcc000000U,
    0x5a000000U,
    0x77000000U,
    0x4e800000U,
    0x23400000U,
    0xd4a00000U,
    0xb4500000U,
    0xaa080000U,
    0x8f340000U,
    0x3a8a0000U,
    0xad570000U,
    0xbd948000U,
    0x1bfec000U,
    0xeacd2000U,
    0x41411000U,
    0x379c8800U,
    0x44cac400U,
    0xf8472a00U,
    0xb0161900U,
    0x58080080U,
    0xe43408c0U,
    0x60a0fa0U,
    0xa5170e70U,
    0xf5b48c68U,
    0x37eecef4U,
    0x80e528eaU,
    0x2e651c35U,
    0x3d3e8648U,
    0xf1b9cec7U,
    0x1f1a4e2U,
    0x3ddbd890U,
    0x9fd3a1f9U,
    0x9ce8d18cU,
    0x5c452aa4U,
    0x163514b1U,
    0x6d3686c8U,
    0x998dc607U,
    0xfdfbab42U,
    0x7fccd6e0U,
    0xace72d91U,
    0x44461f78U,
    0x5200024eU,
    0xfb000884U,
    0xb4800080U,
    0xc44008c0U,
    0x12200fa0U,
    0x5b100e70U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x68000000U,
    0xb4000000U,
    0xb6000000U,
    0x9000000U,
    0x40800000U,
    0xb9400000U,
    0x3ea00000U,
    0x54700000U,
    0x30180000U,
    0x482c0000U,
    0x24220000U,
    0xae310000U,
    0xd5378000U,
    0x42af4000U,
    0x67da000U,
    0x770c1000U,
    0xadaf8800U,
    0xb7c34c00U,
    0x22ffae00U,
    0x404d1700U,
    0xd6000e80U,
    0xb9000740U,
    0xc8800560U,
    0x7d400790U,
    0xe0a00288U,
    0xe97000d4U,
    0xc698088aU,
    0xf86c05d7U,
    0x5a020709U,
    0x43010f16U,
    0xdb8f81a8U,
    0x5ef34fa4U,
    0x1247ae52U,
    0x911110a8U,
    0x5cba0a25U,
    0x5b5d0193U,
    0xf1b58989U,
    0x75ee4856U,
    0x7dd224c8U,
    0x79cf5834U,
    0x47d024daU,
    0x8ace5c7cU,
    0x145facafU,
    0x103d1344U,
    0xd8180000U,
    0x3c2c0000U,
    0x72220000U,
    0xd7310000U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0xb8000000U,
    0xe4000000U,
    0x86000000U,
    0x8d000000U,
    0x4b800000U,
    0x8ec00000U,
    0x79a00000U,
    0x1df00000U,
    0xab180000U,
    0xf6b40000U,
    0x8d560000U,
    0xbb5d0000U,
    0xbe5f8000U,
    0x59f24000U,
    0x1d1f6000U,
    0x33a2f000U,
    0x8ae78800U,
    0xefb64c00U,
    0x28d16e00U,
    0x48bf100U,
    0xfe4e0380U,
    0x79e90240U,
    0xed098660U,
    0x9baf49d0U,
    0xd6c0ef38U,
    0x8d90b6acU,
    0x23d8e7faU,
    0xc224b50fU,
    0x3b0ee809U,
    0x8eb9ba87U,
    0x897162cfU,
    0x2d7bf92aU,
    0x8b5604b4U,
    0xf65d0164U,
    0x15df805cU,
    0xc73247f3U,
    0xdcbf6389U,
    0xca52f4c7U,
    0xa7ff8aafU,
    0x940241faU,
    0xee07680cU,
    0x3116f588U,
    0x39b181c6U,
    0x3deb4b2cU,
    0x5b0eecb8U,
    0x5eb9b4ecU,
    0xd171619aU,
    0xd97bfcdfU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x88000000U,
    0xbc000000U,
    0xba000000U,
    0x6b000000U,
    0xbb800000U,
    0x27400000U,
    0x30a00000U,
    0x6cd00000U,
    0xfff80000U,
    0x505c0000U,
    0xa10a0000U,
    0x788b0000U,
    0xf0f88000U,
    0x95dbc000U,
    0x37d6000U,
    0x2eba9000U,
    0x59f28800U,
    0x1150c400U,
    0x2985e600U,
    0x60615b00U,
    0x690fe080U,
    0xa4aa5fc0U,
    0xfad765a0U,
    0x76e199b0U,
    0x4f20d38U,
    0xb3d706b4U,
    0x272862aU,
    0xc610c0bdU,
    0xf125e767U,
    0xb15a01U,
    0xa4f7e102U,
    0x23f65181U,
    0x5a5d6642U,
    0x422a97e1U,
    0x7f2a8052U,
    0x6d9cc26bU,
    0xce57ea5fU,
    0xd4265cb5U,
    0xf6256728U,
    0x3936913cU,
    0xdc808125U,
    0xaec7cde0U,
    0xc0d76150U,
    0x5de193eaU,
    0xdf720c1dU,
    0x24970b54U,
    0xbad2877aU,
    0x16c0c357U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x58000000U,
    0x54000000U,
    0x56000000U,
    0x33000000U,
    0x54800000U,
    0xe4c00000U,
    0x17a00000U,
    0x18700000U,
    0xcf780000U,
    0x5c40000U,
    0xbe1e0000U,
    0xaf290000U,
    0x6e8f8000U,
    0x85dbc000U,
    0x7e23a000U,
    0xcf057000U,
    0x3e918800U,
    0xddf2cc00U,
    0x2a2c2600U,
    0x991eb500U,
    0xd922d80U,
    0x8947b940U,
    0xcee5ab60U,
    0x8e987a30U,
    0x15f80ec8U,
    0x4604020cU,
    0xcb3e079aU,
    0x309902f7U,
    0xbad7895fU,
    0x28afc860U,
    0x4ee5aab3U,
    0x4e987609U,
    0x75f8062dU,
    0x16040bcaU,
    0x933e078eU,
    0x649902d9U,
    0xecd78797U,
    0x1bafca6cU,
    0x1a65ad29U,
    0xaa5874feU,
    0x62580f72U,
    0xe7403aaU,
    0x5c460d3dU,
    0x615d04d0U,
    0x52c989baU,
    0xb486cda6U,
    0x74ea2ca7U,
    0x2f83b327U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0x78000000U,
    0xdc000000U,
    0xca000000U,
    0x43000000U,
    0xe3800000U,
    0x9c400000U,
    0xb8a00000U,
    0x73d00000U,
    0x6c80000U,
    0x1c7c0000U,
    0xf8860000U,
    0xd3c30000U,
    0x36e88000U,
    0x6445c000U,
    0x24bb6000U,
    0x19c65000U,
    0x75ee8800U,
    0x87c6c400U,
    0xb8f3ea00U,
    0xa1539300U,
    0x61def80U,
    0x813c99c0U,
    0xa4bb6ea0U,
    0x59c65330U,
    0xd5ee8bb8U,
    0xb7c6c304U,
    0xc0f3eaaaU,
    0x7d539dcdU,
    0xcc1ded74U,
    0xc23c95f3U,
    0x473b649aU,
    0xc58650f5U,
    0x6d4e8330U,
    0xc416c3baU,
    0xc63be705U,
    0x612f90aaU,
    0x349be6ccU,
    0x11ff96f7U,
    0x71d3ee30U,
    0xa1c39d38U,
    0x49f5e644U,
    0xddd09249U,
    0xb3d5699fU,
    0xe6e9535fU,
    0x8c680a7cU,
    0xb0ac0c8dU,
    0x77ce0795U,
    0x20ff0ea2U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0x8000000U,
    0x5c000000U,
    0x3a000000U,
    0x2f000000U,
    0xac800000U,
    0x94c00000U,
    0x5fa00000U,
    0xc2700000U,
    0x44480000U,
    0xa1740000U,
    0x1afa0000U,
    0xe68b0000U,
    0x43f08000U,
    0x9732c000U,
    0xa8a4a000U,
    0x5add7000U,
    0x86aa8800U,
    0x73c9cc00U,
    0xf1c2a00U,
    0xfc9bb900U,
    0x3cf42880U,
    0x939fb9c0U,
    0xf04621a0U,
    0x3760b7f0U,
    0x37cca848U,
    0xa119798cU,
    0x15b884daU,
    0x1546ce17U,
    0x8cdea7acU,
    0xcb967d6bU,
    0x47a05bdU,
    0xc14b0033U,
    0x2ad086e9U,
    0x7e82cdffU,
    0x17cca0d3U,
    0xf1197179U,
    0xbdb887e4U,
    0xd946c8e7U,
    0xbedeab67U,
    0xb8967724U,
    0x92fa09c5U,
    0x7a8b0954U,
    0xd9f084ceU,
    0x2832c6baU,
    0xc24a945U,
    0x921d7c94U,
    0xe30a8f6eU,
    0x9eb9c84aU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x28000000U,
    0xc4000000U,
    0x3a000000U,
    0x9b000000U,
    0xa1800000U,
    0x93400000U,
    0xa0a00000U,
    0xf9f00000U,
    0x2a580000U,
    0x560c0000U,
    0xa5020000U,
    0xe0950000U,
    0xd9d88000U,
    0xba7dc000U,
    0xe07e000U,
    0x49049000U,
    0x1e828800U,
    0x78e4cc00U,
    0x80dd6e00U,
    0x3cec5700U,
    0x7addea80U,
    0x47dd9040U,
    0xab7805a0U,
    0xfcbc02b0U,
    0xcffa0698U,
    0x3f690274U,
    0x7e828b2aU,
    0xc8e4ca6fU,
    0x48dd6b1dU,
    0x88ec55e4U,
    0x68dde242U,
    0x18dd94a2U,
    0x30f80332U,
    0xf4fc0a58U,
    0xceda0495U,
    0x55d904b9U,
    0xf47a8705U,
    0x6718c7d0U,
    0xc7876cc8U,
    0x3e755c7dU,
    0x14076eb7U,
    0x42355dc8U,
    0xe7276dfdU,
    0x7855a74U,
    0xde5f6f2aU,
    0x6439586cU,
    0x6a256a1fU,
    0x23105c66U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x28000000U,
    0xa4000000U,
    0xfe000000U,
    0x3d000000U,
    0x82800000U,
    0xb3400000U,
    0x5a00000U,
    0x42f00000U,
    0x41780000U,
    0xa28c0000U,
    0x63620000U,
    0x3d8d0000U,
    0xbed98000U,
    0x33544000U,
    0xc5ba2000U,
    0x22fe1000U,
    0x31638800U,
    0x8aa54c00U,
    0xc779a600U,
    0xc3ab5700U,
    0x83e22a80U,
    0xb1c21640U,
    0x76d981e0U,
    0x275448d0U,
    0x73ba2ca8U,
    0xcbfe1674U,
    0x65e3853aU,
    0xa0e541bfU,
    0xbe59a5dfU,
    0xf1b55ccU,
    0x45ba24c4U,
    0xe2fe1a83U,
    0x51638e40U,
    0xfaa545e2U,
    0xef79a2d2U,
    0x67ab5dabU,
    0x7de22bf7U,
    0x8cc219f8U,
    0xf459861eU,
    0x941444ecU,
    0x761a2db7U,
    0x890e101aU,
    0x249b82ccU,
    0x2694e47U,
    0xdd3bacc0U,
    0x32965fa2U,
    0xfb63a532U,
    0xd1aa527bU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x88000000U,
    0x5c000000U,
    0xea000000U,
    0x43000000U,
    0x40800000U,
    0xfc400000U,
    0x9da00000U,
    0x19f00000U,
    0x90580000U,
    0xdf8c0000U,
    0x96ea0000U,
    0xc2d30000U,
    0xb4d48000U,
    0xfdcdc000U,
    0x8e49a000U,
    0xba835000U,
    0x87468800U,
    0xe922cc00U,
    0x238f2e00U,
    0x2ce19700U,
    0x99e9a080U,
    0x507359c0U,
    0x3f9e88a0U,
    0xe6eecf30U,
    0x4ac52288U,
    0xe8c29d04U,
    0x17e527faU,
    0xcd729c6fU,
    0xfa1d25adU,
    0x7b0e9d0cU,
    0x74af2a3cU,
    0x3a519cb6U,
    0xbc91abb2U,
    0x464f594bU,
    0xc68c8ca7U,
    0xfd41cf49U,
    0x5223afa5U,
    0xb71055c8U,
    0x66b20b66U,
    0xad5f08e9U,
    0x4a3e8417U,
    0x131ecc83U,
    0x589d27c1U,
    0x584e97a0U,
    0xa38f23b2U,
    0xece1954bU,
    0x79e9a2a7U,
    0x20735849U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0x58000000U,
    0x9c000000U,
    0x2a000000U,
    0xcf000000U,
    0x3d800000U,
    0x4c400000U,
    0xca00000U,
    0x70d00000U,
    0xb4780000U,
    0x60840000U,
    0xb2de0000U,
    0x6f6b0000U,
    0xb3188000U,
    0x878ac000U,
    0xbb4ee000U,
    0x7d285000U,
    0x4e9e8800U,
    0xebf5c400U,
    0x7d06600U,
    0x85dd9d00U,
    0x3eceed80U,
    0x3d685dc0U,
    0x503e8ca0U,
    0x1825c5f0U,
    0xfc286058U,
    0xfa199004U,
    0x9730efeaU,
    0xa193563dU,
    0x667e043fU,
    0xc3bb073cU,
    0x4d608bbfU,
    0xf80ec1ffU,
    0x6c10e79cU,
    0xc2035c4dU,
    0xdb2602a6U,
    0xd3af0b18U,
    0x356681e7U,
    0xd431cef8U,
    0xce2e6ef5U,
    0xc9269f32U,
    0x508e667bU,
    0x7af692b5U,
    0xcb766c53U,
    0xd532942aU,
    0x3a88679cU,
    0xd5c99c4dU,
    0x26c8e2a6U,
    0xc1575b18U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0xe8000000U,
    0xc000000U,
    0xbe000000U,
    0x2f000000U,
    0x73800000U,
    0x12400000U,
    0x3a600000U,
    0x16700000U,
    0x58680000U,
    0xf4c0000U,
    0x98f60000U,
    0x38b50000U,
    0x93fa8000U,
    0xd9014000U,
    0xe0b5a000U,
    0xc7dd3000U,
    0x7335a800U,
    0x59d3400U,
    0xc155ae00U,
    0x8fed3d00U,
    0xcf3da080U,
    0xa3a13dc0U,
    0x9a4bab60U,
    0xa6543f30U,
    0x40512c58U,
    0x7b657e14U,
    0xc2ec8ffeU,
    0xa5844f73U,
    0x71472278U,
    0x97e07b87U,
    0xbb1e05f7U,
    0xf9b9050cU,
    0x76c8c48U,
    0x44c4455cU,
    0xea72123U,
    0x20d07211U,
    0xfc960c48U,
    0xd1c5055cU,
    0x30128123U,
    0x580d4211U,
    0x1423a448U,
    0xca18315cU,
    0x75272f23U,
    0xee907f11U,
    0xf0f604c8U,
    0x74b50c9cU,
    0x4dfa8443U,
    0x66014021U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x8000000U,
    0x74000000U,
    0xea000000U,
    0x15000000U,
    0xdc800000U,
    0x40c00000U,
    0xe8e00000U,
    0xccd00000U,
    0xed80000U,
    0xfffc0000U,
    0xc17e0000U,
    0xe0a30000U,
    0x3ef98000U,
    0xa7ecc000U,
    0x9d63e000U,
    0xee9fd000U,
    0xc9e3e800U,
    0x1a5fdc00U,
    0xeb03ea00U,
    0x538fd500U,
    0x315bea80U,
    0x98b3d240U,
    0xf2c5e420U,
    0xa1c0d310U,
    0x1e6469e8U,
    0xb910171cU,
    0xaa998766U,
    0x7bfccbd1U,
    0x535bef8bU,
    0x39b3d82eU,
    0xe445e99cU,
    0x6400d324U,
    0x220468f3U,
    0x4100171aU,
    0xa6a18f05U,
    0x5dd0cfe3U,
    0x407de8f3U,
    0x662cd71aU,
    0xf3226f05U,
    0xef9f1fe3U,
    0x8f4600f3U,
    0xf78f0b1aU,
    0x335f8505U,
    0x49b3cae3U,
    0x5c646a73U,
    0x8810195aU,
    0xb4198125U,
    0xca3cc9f3U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0xa8000000U,
    0x9c000000U,
    0xfa000000U,
    0xd3000000U,
    0xdb800000U,
    0xb4c00000U,
    0x88e00000U,
    0x62d00000U,
    0x39d80000U,
    0xae6c0000U,
    0x48860000U,
    0x8f410000U,
    0xcc2f8000U,
    0x92034000U,
    0x6f39a000U,
    0xf1829000U,
    0xcfd9a800U,
    0xcf529c00U,
    0x2c01aa00U,
    0x623e9100U,
    0x1707a080U,
    0xc5bf98c0U,
    0xa9c82f20U,
    0xe66cd5f0U,
    0x24a98298U,
    0xd424ebcU,
    0x2b162216U,
    0x2f81d491U,
    0xf2e00b8bU,
    0x71d00f77U,
    0xc2580f83U,
    0xcaac0f43U,
    0x68660461U,
    0x71910f93U,
    0xff78a0aU,
    0xef6f4f37U,
    0xfc3fac61U,
    0xca039393U,
    0x8b16200aU,
    0x3f81de37U,
    0x7ae00ce1U,
    0x3dd00b53U,
    0x90580f2aU,
    0x85ac0bc7U,
    0x49e60e79U,
    0x165105efU,
    0x5c978d3cU,
    0x397f4f56U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x88000000U,
    0xfc000000U,
    0xe6000000U,
    0x59000000U,
    0x18800000U,
    0xc1400000U,
    0x95600000U,
    0x1f700000U,
    0xf8480000U,
    0xfdc40000U,
    0xb6260000U,
    0xa1390000U,
    0x8c908000U,
    0xeb7ec000U,
    0xa254e000U,
    0xa2d6b000U,
    0xc7b4e800U,
    0xf0e6b400U,
    0x249ce600U,
    0x8752b700U,
    0x1c72ee80U,
    0xbfefb8c0U,
    0x4d2460e0U,
    0x72987d50U,
    0x16480168U,
    0x18c40144U,
    0xa8a6003eU,
    0x97904b3U,
    0x897081baU,
    0xc94ec86fU,
    0x297ce258U,
    0x1962bdbcU,
    0x915aecf1U,
    0x6d5bb29bU,
    0x8b4a695cU,
    0xd26577a2U,
    0xcafe84f1U,
    0xb83c69bU,
    0x9ee26f5cU,
    0x819170a2U,
    0x79f08271U,
    0x840eca5bU,
    0x321ce9bcU,
    0x9312baf2U,
    0x1f92ed99U,
    0xf4dfb3dfU,
    0x568c6962U,
    0xf46c7311U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xd0000000U,
    0x58000000U,
    0x84000000U,
    0xca000000U,
    0x6b000000U,
    0x1f800000U,
    0x12400000U,
    0x98600000U,
    0xd3500000U,
    0xfce80000U,
    0x669c0000U,
    0x22ea0000U,
    0xbf9b0000U,
    0xe2418000U,
    0xf0474000U,
    0xdf6aa000U,
    0xead53000U,
    0x4382a800U,
    0x9c493c00U,
    0xd968a600U,
    0x77d23100U,
    0xa9292380U,
    0x68957940U,
    0xa3c38f20U,
    0xfb004ff0U,
    0x67a126d8U,
    0xa65972d4U,
    0xda41895eU,
    0x64474ee1U,
    0xad6aa510U,
    0xd5d531c8U,
    0xce02a81eU,
    0x610935c1U,
    0x9488a0e2U,
    0xddc23811U,
    0xd2212c4aU,
    0xcf19745dU,
    0xe5a18362U,
    0xf1574151U,
    0x5be2a36aU,
    0x8f193badU,
    0xc580a5baU,
    0xc14e3385U,
    0xd3c32a34U,
    0x530e754cU,
    0x8b8200aaU,
    0x6047024dU,
    0xa74b822aU,
    0x5ecc408dU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x68000000U,
    0xe4000000U,
    0x7e000000U,
    0x87000000U,
    0xda800000U,
    0x2c400000U,
    0x62600000U,
    0x3d700000U,
    0x9bd80000U,
    0x158c0000U,
    0x2f60000U,
    0xf83f0000U,
    0x6c388000U,
    0x6a354000U,
    0xc125e000U,
    0xd199f000U,
    0x6cfde800U,
    0xb715f400U,
    0x28be200U,
    0x506af100U,
    0xc0536480U,
    0x486fbf40U,
    0x5c4e8360U,
    0x1a4a4730U,
    0x117d66c8U,
    0x51dcb1f4U,
    0x848002eeU,
    0xbb400323U,
    0x70e00fd3U,
    0xa5300d3bU,
    0xefb8049cU,
    0x4bfc0188U,
    0x3dae0e97U,
    0x46f303dfU,
    0xd62e8eebU,
    0x833a47a5U,
    0x54a56a17U,
    0x9350bc9fU,
    0x34f60d8bU,
    0x8b3f0095U,
    0xb88cdfU,
    0x75754d6bU,
    0x6fc5ef65U,
    0xa3a9f3b6U,
    0x31c5eb0cU,
    0x34a9f450U,
    0x2345e9f9U,
    0xace9f33eU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0xd8000000U,
    0xfc000000U,
    0xba000000U,
    0x2f000000U,
    0xd0800000U,
    0x78c00000U,
    0xaae00000U,
    0x81f00000U,
    0x7680000U,
    0x42b40000U,
    0x33e20000U,
    0x5c6d0000U,
    0x1c0b8000U,
    0x8a244000U,
    0x972a6000U,
    0x5ca55000U,
    0x1ac26800U,
    0x79d15400U,
    0xeb406e00U,
    0x508c5300U,
    0x38c3e380U,
    0xcaec1cc0U,
    0xf1e38820U,
    0xdf504e30U,
    0xbea86d28U,
    0x89f85dbcU,
    0x7341e986U,
    0xccb116a3U,
    0xf2e00270U,
    0x3df00a48U,
    0xdd680f6eU,
    0x1db406beU,
    0x3b620e06U,
    0xd8ad0c60U,
    0xceb8d52U,
    0x24d449f9U,
    0x40c26d86U,
    0x66d150a0U,
    0x83c06572U,
    0xa44c57c9U,
    0xf023e8aeU,
    0x981c191cU,
    0x9c0b82f4U,
    0xca24426aU,
    0xf72a615eU,
    0x2ca55b94U,
    0xc2c26bbaU,
    0x85d159e4U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0xb8000000U,
    0x74000000U,
    0x2000000U,
    0xbf000000U,
    0x18800000U,
    0x5c00000U,
    0x5be00000U,
    0x6af00000U,
    0xdb680000U,
    0x6b2c0000U,
    0x2aa20000U,
    0x72ef0000U,
    0x9f578000U,
    0xa12c4000U,
    0x49bc6000U,
    0xa4741000U,
    0x53b46800U,
    0x5f681400U,
    0x811e6e00U,
    0x399b1500U,
    0xc63e580U,
    0x9f845240U,
    0x294205a0U,
    0x3c1f09b0U,
    0x9e3f8c28U,
    0x11004decU,
    0xc19e6f96U,
    0x185b1a43U,
    0x8d83e1a0U,
    0x2e745fb1U,
    0x50aa052aU,
    0x99f3026fU,
    0xf5fd8ed7U,
    0xb3df4b61U,
    0xc6c1e9d2U,
    0xbd6b5d79U,
    0x6e158357U,
    0xf9334d21U,
    0x6d83e272U,
    0x7e7451c9U,
    0x88aa02ffU,
    0xfdf3068dU,
    0x4ffd8644U,
    0x78df473aU,
    0xdc41e2f7U,
    0x7ab5290U,
    0x2d758758U,
    0x960343a6U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0xc8000000U,
    0xdc000000U,
    0x12000000U,
    0x65000000U,
    0x6d800000U,
    0x97c00000U,
    0xc1e00000U,
    0x72f00000U,
    0xec680000U,
    0x85140000U,
    0xdd860000U,
    0x3fe10000U,
    0xeddd8000U,
    0xa8e8c000U,
    0x5559a000U,
    0xfab75000U,
    0x2f51a800U,
    0x93935400U,
    0x8dfae00U,
    0x85565b00U,
    0xa28c2280U,
    0xcb7b96c0U,
    0x958603a0U,
    0xa3e10090U,
    0x9fdd8d78U,
    0x3de8cdecU,
    0xf0d9a766U,
    0xb17755c3U,
    0xfcb1ad20U,
    0x84635452U,
    0x8937a7d8U,
    0x97825efcU,
    0xbeea235dU,
    0x866a90ceU,
    0x94338b36U,
    0x8e1dce1aU,
    0x170229ddU,
    0xf8be920eU,
    0x32558696U,
    0x8a0cc58aU,
    0xa137ae25U,
    0xfb825d22U,
    0x4ea2c50U,
    0xcf6a9bd9U,
    0x23b384fdU,
    0xa0ddc65cU,
    0xa962214eU,
    0x788e9b76U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0x38000000U,
    0x4c000000U,
    0x66000000U,
    0xfb000000U,
    0xc4800000U,
    0x4cc00000U,
    0x47e00000U,
    0x7b500000U,
    0x5b980000U,
    0xa640000U,
    0xb23a0000U,
    0x61310000U,
    0x418c8000U,
    0x4f684000U,
    0x91b86000U,
    0xc77fd000U,
    0xc5826800U,
    0x9d4edc00U,
    0x608eea00U,
    0xeee69b00U,
    0x8ed68980U,
    0x1ec94fc0U,
    0x76cce7e0U,
    0x72e39c70U,
    0x60f801a8U,
    0xb1f403bcU,
    0xe8420c56U,
    0x9b050bc9U,
    0x54ae8e0fU,
    0x24fd471fU,
    0x43eeec84U,
    0x69769eb0U,
    0x8aae8ecaU,
    0x53fd4a8fU,
    0xc16ee05eU,
    0xaeb69e27U,
    0x31ce8f62U,
    0x286d4933U,
    0xbb16ec08U,
    0x248295eeU,
    0x1cec816dU,
    0xff84e2cU,
    0xf40608cU,
    0x718bdb5eU,
    0x974067a7U,
    0x8d8bd8a3U,
    0xe9406ad2U,
    0x4a8bde79U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0xc8000000U,
    0xdc000000U,
    0x22000000U,
    0x4b000000U,
    0xa8800000U,
    0xc00000U,
    0xbfe00000U,
    0x19500000U,
    0xd8880000U,
    0x68e40000U,
    0x33c60000U,
    0xc3590000U,
    0x87858000U,
    0x3e494000U,
    0x5a366000U,
    0x9f01d000U,
    0x36906800U,
    0x59c8dc00U,
    0x947dee00U,
    0x6d359f00U,
    0x45858280U,
    0x854942c0U,
    0x5ab668a0U,
    0x73c1da70U,
    0x63706c28U,
    0xd798d57cU,
    0xc675ed56U,
    0x4e119a29U,
    0x61238e7eU,
    0x5f8045d7U,
    0xba5bedeaU,
    0x3c3c915eU,
    0xd2080ae4U,
    0xe3240c23U,
    0x44a60bb1U,
    0xeac90008U,
    0x28ed8eccU,
    0x93fd455fU,
    0x937868e7U,
    0x7fbcd521U,
    0x2a53ea32U,
    0xa4189e48U,
    0xf62e03adU,
    0xd52d010fU,
    0xf1ab8e7eU,
    0x2b6445d7U,
    0xcb9dedeaU,
    0x8465915eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0xc8000000U,
    0x44000000U,
    0x82000000U,
    0x2d000000U,
    0xef800000U,
    0x67400000U,
    0x48600000U,
    0xcaf00000U,
    0x62380000U,
    0x9d2c0000U,
    0xd78e0000U,
    0x3b570000U,
    0x46638000U,
    0x21cac000U,
    0x22a82000U,
    0x38c57000U,
    0x171e2800U,
    0xdebe7400U,
    0xa6f3aa00U,
    0x1423b900U,
    0xa380680U,
    0x492c0d40U,
    0xbd8e06a0U,
    0x82570b90U,
    0xe3e38a58U,
    0x2f8ac0e4U,
    0x748225eU,
    0xb875780bU,
    0xd2c6245dU,
    0xee227709U,
    0x5b25a5ddU,
    0x78a8bfc8U,
    0xf9ed89bdU,
    0xce9dc53bU,
    0xecbad94U,
    0xa00fb9b5U,
    0x90360565U,
    0xe83b089fU,
    0x940d896aU,
    0x4a2dca2eU,
    0x6913ab60U,
    0x6d93bf72U,
    0x4a600ee9U,
    0xa7f00dedU,
    0xadb80680U,
    0x2a6c0d40U,
    0x57ee06a0U,
    0xb5a70b90U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xd8000000U,
    0xc4000000U,
    0xc6000000U,
    0xa1000000U,
    0x6a800000U,
    0xb5c00000U,
    0xcae00000U,
    0x65700000U,
    0xec080000U,
    0x9a340000U,
    0x43120000U,
    0x7d9b0000U,
    0xa6768000U,
    0xd188c000U,
    0xdc536000U,
    0xe2879000U,
    0x79c96800U,
    0x80e89400U,
    0x6e4de600U,
    0xfd8b5300U,
    0xe6600b80U,
    0x31b00f40U,
    0xac680fe0U,
    0x3a840150U,
    0xbdfa0748U,
    0x46df0d0cU,
    0xcf6c8666U,
    0x9727c71bU,
    0x53b7e4c6U,
    0xfb5454a8U,
    0xc90c8e5fU,
    0xd697cc2dU,
    0x27dfee96U,
    0x5d05462U,
    0xb2f68613U,
    0x3148cca9U,
    0x8233625eU,
    0x2737962eU,
    0x2ba16f95U,
    0xaf6c9ae2U,
    0xa737e9d0U,
    0x6b945b8aU,
    0x4f6c81acU,
    0xd727c2d4U,
    0xb3b7e800U,
    0x8b545400U,
    0x110c8600U,
    0x1297c300U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0xb8000000U,
    0x14000000U,
    0x56000000U,
    0x29000000U,
    0x7d800000U,
    0xd6400000U,
    0xe1600000U,
    0x69d00000U,
    0x88380000U,
    0x7c1c0000U,
    0x5a020000U,
    0x7b2f0000U,
    0xbabe8000U,
    0x96dac000U,
    0x1ca0a000U,
    0x77d0b000U,
    0x7d1aa800U,
    0x4ba3bc00U,
    0x4f462200U,
    0x34c67d00U,
    0x13800980U,
    0x2b400c40U,
    0x6ae004e0U,
    0x869002d0U,
    0xacd80938U,
    0xd78c0eb4U,
    0x655a0caeU,
    0x47e30969U,
    0x1504814dU,
    0x47a9cc38U,
    0x1d7c2335U,
    0xf3f576eeU,
    0x533c8e88U,
    0xd6b5c51eU,
    0x74fe25c1U,
    0x739a78a2U,
    0x9b620e30U,
    0xc2ff07eaU,
    0x2a868d8fU,
    0xeec6c31bU,
    0xa8a2ae45U,
    0x31ffb966U,
    0xec242014U,
    0x2239719cU,
    0xcf068000U,
    0xfc86c000U,
    0x7c2a000U,
    0x752fb000U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x58000000U,
    0xf4000000U,
    0x4e000000U,
    0x57000000U,
    0xbf800000U,
    0xd0c00000U,
    0xf5e00000U,
    0x4f500000U,
    0xf5080000U,
    0xf2b40000U,
    0x66520000U,
    0xe5970000U,
    0x79c88000U,
    0x25434000U,
    0xa425a000U,
    0xc6301000U,
    0xeb1fa800U,
    0x2d831c00U,
    0x65ed2e00U,
    0xa7735b00U,
    0x393a0b80U,
    0x38b30440U,
    0x27728760U,
    0xf9304d30U,
    0x58b72a98U,
    0x5750573cU,
    0xa1208846U,
    0xaca74089U,
    0x197fadf6U,
    0xf61311d2U,
    0x13052bebU,
    0xc99756c7U,
    0x3e00449U,
    0x5c500d56U,
    0x3c8803c3U,
    0xf1740522U,
    0x3a320d51U,
    0xd907022aU,
    0x88a082e5U,
    0x1f67439bU,
    0x7d1fa9bfU,
    0x4e831c84U,
    0xf46d2828U,
    0x50b353e5U,
    0x2b5a0918U,
    0x53230f7cU,
    0x699a8126U,
    0x13d446b9U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x68000000U,
    0x1c000000U,
    0x3a000000U,
    0x7000000U,
    0xfc800000U,
    0xe6c00000U,
    0xd2600000U,
    0xf8b00000U,
    0xe8c80000U,
    0x93440000U,
    0xc9160000U,
    0x9d950000U,
    0x476d8000U,
    0x3f094000U,
    0x58b3e000U,
    0x38d71000U,
    0x6b5e6800U,
    0x4d1e5400U,
    0xd38d8e00U,
    0x66794b00U,
    0xfe9be080U,
    0x45e31ec0U,
    0xf0e06d20U,
    0xa77f55b0U,
    0x8f3e0268U,
    0x50a1051cU,
    0xd4d3826eU,
    0x39684f27U,
    0x5600668aU,
    0x150f570fU,
    0x87960675U,
    0xd0550af1U,
    0x5b8d8c09U,
    0xca7944ceU,
    0xcc9be4d7U,
    0xaee31702U,
    0x5e606a83U,
    0x5abf53c1U,
    0x9bde02a2U,
    0x49d10df3U,
    0x12fb8e8aU,
    0xb45c430fU,
    0xa5be6875U,
    0xe36e51f1U,
    0xe1258489U,
    0xe18d4e0eU,
    0x8d45e7f7U,
    0x503219b2U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0x28000000U,
    0x7c000000U,
    0x5e000000U,
    0x9d000000U,
    0xde800000U,
    0x7d400000U,
    0xfbe00000U,
    0x2900000U,
    0x13480000U,
    0xbedc0000U,
    0x981e0000U,
    0x641d0000U,
    0xfa348000U,
    0x873ec000U,
    0x4985e000U,
    0x1ce07000U,
    0x9b316800U,
    0xc79ebc00U,
    0x49d48600U,
    0x29aecd00U,
    0xcccdec80U,
    0x533c76c0U,
    0xabaf6f60U,
    0x3fc3b210U,
    0xc8800e08U,
    0x4c400304U,
    0xb3600ed6U,
    0x8ed0013dU,
    0x402809eaU,
    0x200c05d4U,
    0xf03600bdU,
    0x381103a8U,
    0x54028474U,
    0x222fc9cfU,
    0xc30765b1U,
    0x438fb279U,
    0xa3d60d9eU,
    0x86810c1bU,
    0xf94a850cU,
    0x11f3c1d1U,
    0xad9961eaU,
    0x26d2b9d4U,
    0xfc0286bdU,
    0x9e2fcea8U,
    0x7d0768f4U,
    0xce8fbf0fU,
    0x55560ad1U,
    0x87c10069U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x8000000U,
    0xac000000U,
    0x12000000U,
    0x93000000U,
    0x39800000U,
    0x97400000U,
    0x82e00000U,
    0x47b00000U,
    0x36480000U,
    0x38640000U,
    0x916e0000U,
    0x87c10000U,
    0x431b8000U,
    0x61b64000U,
    0x2347a000U,
    0x44fa3000U,
    0x62bc2800U,
    0x22fc7400U,
    0x17b38e00U,
    0x2e624300U,
    0xec61a680U,
    0x275f3dc0U,
    0xaac9a9a0U,
    0xfb8b37f0U,
    0x5c6facb8U,
    0xf6e3344U,
    0x16fa2536U,
    0x91a974cfU,
    0x6b6e04eaU,
    0x88c1070dU,
    0x9b8e4aU,
    0xb9f6467eU,
    0x8227a731U,
    0xab0a3098U,
    0xfd9427f7U,
    0x596871edU,
    0xbf58bdbU,
    0x41374395U,
    0x8abc27bdU,
    0xbefc7493U,
    0x6db38a6aU,
    0x61624ecdU,
    0xcfe1a9eaU,
    0x8f1f328eU,
    0x3a9ad09U,
    0xb87b3e1cU,
    0xd147ab61U,
    0xe7fa32d2U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x28000000U,
    0x4000000U,
    0xee000000U,
    0xd3000000U,
    0xe3800000U,
    0xfa400000U,
    0x65e00000U,
    0x9d900000U,
    0xa1680000U,
    0x324c0000U,
    0x71ee0000U,
    0x3b830000U,
    0xa6668000U,
    0x17c94000U,
    0x5ca32000U,
    0x53def000U,
    0x12a5a800U,
    0x30c7bc00U,
    0x90e8e00U,
    0xe6854100U,
    0x86cd2480U,
    0x461dfd40U,
    0x17232460U,
    0x6d9ef170U,
    0xf945aed8U,
    0xae57b394U,
    0x63e684e6U,
    0x2a89473dU,
    0x7cc32628U,
    0x330efa8dU,
    0xf3ada750U,
    0xb25bbe48U,
    0xb1e885feU,
    0x5b9a4889U,
    0x764da95cU,
    0x3fcbbc58U,
    0x58808bd6U,
    0xbdd64e04U,
    0xc1a3a00cU,
    0xd348b310U,
    0xf36602a8U,
    0x835f07cdU,
    0x1b608330U,
    0xe7464f38U,
    0x254bab26U,
    0x1c44bb1dU,
    0xc2e80dbaU,
    0x80c0b65U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x68000000U,
    0x4c000000U,
    0x6000000U,
    0xf3000000U,
    0x25800000U,
    0xd2c00000U,
    0x51600000U,
    0xb00000U,
    0x2c480000U,
    0x8a2c0000U,
    0xd51e0000U,
    0xe6910000U,
    0xcf468000U,
    0xa7874000U,
    0x1bf6a000U,
    0xd9c45000U,
    0x70f82800U,
    0xa86f1400U,
    0xb0108e00U,
    0x783a4100U,
    0x442e2080U,
    0x1a1211c0U,
    0x9d280ee0U,
    0x9a9c0ff0U,
    0xf1560438U,
    0x70bd071cU,
    0x745881ceU,
    0xfe1649e7U,
    0xf73027b2U,
    0x5f831799U,
    0x1fee85cdU,
    0xa3db4fe7U,
    0xbdc0a2b3U,
    0x5ac9511aU,
    0xd68a10fU,
    0xe955f86U,
    0xc35ea581U,
    0xc1985743U,
    0xb8ce2a22U,
    0x94621f91U,
    0xee00030aU,
    0xff000145U,
    0x43800ae3U,
    0x71c009f0U,
    0x1ce00139U,
    0x9e70019fU,
    0x7b28050cU,
    0x799c0986U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x8000000U,
    0x4c000000U,
    0xb6000000U,
    0xf9000000U,
    0xb2800000U,
    0x93400000U,
    0x87e00000U,
    0x55900000U,
    0xc4c80000U,
    0x88040000U,
    0x8c2e0000U,
    0x56290000U,
    0x89228000U,
    0xba91c000U,
    0xdf41e000U,
    0x31c5f000U,
    0xacab6800U,
    0x76503c00U,
    0x1b448600U,
    0xbfccb00U,
    0x3ad6e80U,
    0x4ded33c0U,
    0x328003e0U,
    0x53400750U,
    0x67e00e48U,
    0x259001a4U,
    0xccc803d6U,
    0xc40408adU,
    0x3a2e0f52U,
    0xaf290249U,
    0x3ba28fa6U,
    0x29d1c4d6U,
    0x58a1e82eU,
    0x6455fb92U,
    0x6863612aU,
    0xfe543c35U,
    0x976a89fcU,
    0x5dd5ca1bU,
    0x8a8fed6cU,
    0xf77cffb3U,
    0xedc1ef9aU,
    0x6285f02dU,
    0xcb4b6f90U,
    0x53c03b2bU,
    0xd78c8134U,
    0xcff8c47fU,
    0x39836b5aU,
    0xe2c43b4eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0x8000000U,
    0x64000000U,
    0xda000000U,
    0xc5000000U,
    0x8b800000U,
    0x66c00000U,
    0x9f600000U,
    0x88b00000U,
    0x6e580000U,
    0xb10c0000U,
    0x19b60000U,
    0xe7d30000U,
    0x5eeb8000U,
    0x2375c000U,
    0x1eb2a000U,
    0x8571d000U,
    0x2d812800U,
    0x55c81400U,
    0x4fe58a00U,
    0xeadacd00U,
    0xd1572280U,
    0x2fab1f40U,
    0xdcd60520U,
    0xea630e10U,
    0x9b338718U,
    0x64b9c33cU,
    0x9064a1ceU,
    0x8e12d8a7U,
    0xc732aab3U,
    0x2b1dbcbU,
    0xc36125c6U,
    0xeeb81e01U,
    0x3d5d8703U,
    0xd1a6cf81U,
    0xe3d92dc0U,
    0x34c41260U,
    0x7e538f30U,
    0xf909cb0aU,
    0x5dbcad26U,
    0xadded271U,
    0x93e4a72bU,
    0xccd2d3b7U,
    0xa252ab28U,
    0xdf01d5b6U,
    0x2eb92028U,
    0x5d741c36U,
    0x618b86e8U,
    0x7bc5c7d6U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0xb8000000U,
    0xe4000000U,
    0x76000000U,
    0x87000000U,
    0x5f800000U,
    0x12c00000U,
    0xf3600000U,
    0xfa900000U,
    0xe4780000U,
    0xd140000U,
    0x7aa20000U,
    0x245f0000U,
    0x2d388000U,
    0x8aa24000U,
    0x9c472000U,
    0xc912d000U,
    0xfc87a800U,
    0x1b649c00U,
    0x96828a00U,
    0xee794300U,
    0xe825a980U,
    0x6c3b9140U,
    0xa3a04e0U,
    0xe51b0a30U,
    0x16828098U,
    0x2e79465cU,
    0xc825a14eU,
    0x9c3b90c5U,
    0xb23a0e11U,
    0x11b0b08U,
    0x60828e15U,
    0xa9794cb8U,
    0x97a5a1cfU,
    0x8efb9b06U,
    0x415a0a30U,
    0xfb8b009aU,
    0x84fa865eU,
    0xa46d414eU,
    0xed07a0c5U,
    0xaaa49612U,
    0x6c628f09U,
    0x71294014U,
    0x18bda1bbU,
    0x6d7f954dU,
    0x118006c6U,
    0xb1c00712U,
    0xfae0018bU,
    0x9f5006d7U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0xe8000000U,
    0x54000000U,
    0x4e000000U,
    0xfd000000U,
    0xbf800000U,
    0x57400000U,
    0x32e00000U,
    0x36b00000U,
    0x16f80000U,
    0xd0b40000U,
    0x1fd60000U,
    0xd1270000U,
    0x5daa8000U,
    0x87ec000U,
    0x7259a000U,
    0xd163b000U,
    0x2beb2800U,
    0x8f197400U,
    0x189c8200U,
    0xfbe9cb00U,
    0xd70b2480U,
    0x4a97a40U,
    0x11e48260U,
    0x8c1dce90U,
    0x123d2518U,
    0xb73e75a4U,
    0x14b60e56U,
    0xa9d7075fU,
    0xc0328ca1U,
    0xe03aceb0U,
    0x5017a5eaU,
    0x1800b02dU,
    0xbc0fabe9U,
    0x1a04b52eU,
    0xb321aa68U,
    0x4297beeeU,
    0xe8dd2bc8U,
    0x658e75deU,
    0x44e0fe2U,
    0x20630b53U,
    0xc66481b9U,
    0xcf5dcf14U,
    0xcedd29bcU,
    0x8c8e7c72U,
    0x55ce03c8U,
    0x7a2301deU,
    0xa3048de2U,
    0xfaadc053U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x98000000U,
    0xb4000000U,
    0x7a000000U,
    0x97000000U,
    0x63800000U,
    0xf5400000U,
    0x70e00000U,
    0xd4900000U,
    0x66e80000U,
    0x2dac0000U,
    0xd8420000U,
    0xe4730000U,
    0x12778000U,
    0x7b554000U,
    0x3dcc2000U,
    0x10013000U,
    0x3833a800U,
    0x84287c00U,
    0xb2358e00U,
    0xbb264900U,
    0xadbbaf80U,
    0x18547e40U,
    0x847f8620U,
    0x42694e30U,
    0xe3662798U,
    0x89de3324U,
    0x6a0620b6U,
    0xaf0e355dU,
    0xe78e28c0U,
    0x47723fe3U,
    0xcbc42a52U,
    0x793d3189U,
    0x7e99ab8dU,
    0xa9f77f8bU,
    0x9a00028dU,
    0x700000bU,
    0x9b8004cdU,
    0x11400228U,
    0x92e000ffU,
    0xf79006b2U,
    0x7f680f58U,
    0x4fec0cc7U,
    0xcb220ae4U,
    0xc5a304d4U,
    0x47f834dU,
    0x82694068U,
    0x836628dfU,
    0xd9de3182U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x48000000U,
    0xf4000000U,
    0xd6000000U,
    0x8d000000U,
    0x46800000U,
    0x400000U,
    0x3be00000U,
    0x25100000U,
    0x2980000U,
    0xee5c0000U,
    0xdac60000U,
    0x9bf0000U,
    0xadd08000U,
    0xc82ac000U,
    0x3424e000U,
    0x7616b000U,
    0x5d326800U,
    0xe837c00U,
    0xf4460200U,
    0xedff0100U,
    0xa8308e80U,
    0x443ac240U,
    0xee3ce9e0U,
    0xe10ab790U,
    0x2c946108U,
    0xaf6c74d4U,
    0x266e85d6U,
    0xeed9c355U,
    0x7faa6d17U,
    0xf0df7ef4U,
    0xc6800c65U,
    0xc04005eeU,
    0x9be00479U,
    0xf51009cdU,
    0x4a980ac9U,
    0x1a5c06f6U,
    0xcc60866U,
    0x84bf03edU,
    0xeb50837aU,
    0xc86ac04dU,
    0xfc4e108U,
    0x5306b4d4U,
    0x5faa65d6U,
    0xe0df7355U,
    0x2e800517U,
    0xe44002f4U,
    0x5e00e65U,
    0x8c1004eeU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0xd8000000U,
    0x64000000U,
    0xaa000000U,
    0x69000000U,
    0x9a800000U,
    0xf4c00000U,
    0x9f600000U,
    0x75100000U,
    0x24b80000U,
    0xcfe40000U,
    0xb8fa0000U,
    0x697f0000U,
    0x422b8000U,
    0x1524c000U,
    0x74aee000U,
    0x87d89000U,
    0xb4ff6800U,
    0x7f435c00U,
    0xe51a0a00U,
    0x4caf0700U,
    0x73f38f80U,
    0x76d0cd40U,
    0xaa6ce720U,
    0xb18390d0U,
    0x7b4eeb08U,
    0x1f0895dcU,
    0x6da76a5eU,
    0xe5775f1dU,
    0x94380c3dU,
    0x92240aefU,
    0x9d1a0ce4U,
    0x58af01b9U,
    0x21f38125U,
    0xcbd0c859U,
    0x42ece996U,
    0x484397a2U,
    0xd4aee610U,
    0xf7d8976aU,
    0x4cff6f2cU,
    0xab435906U,
    0x971a0308U,
    0x41af09dcU,
    0x4373805eU,
    0xeb10c81dU,
    0xaf8cebbdU,
    0x30539bafU,
    0xc096e1c4U,
    0xa5fc9669U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x98000000U,
    0x1c000000U,
    0xe6000000U,
    0xdf000000U,
    0xae800000U,
    0xac00000U,
    0x7f600000U,
    0xe1100000U,
    0xcda80000U,
    0x92740000U,
    0x82aa0000U,
    0xb4d10000U,
    0xdc468000U,
    0x19b24000U,
    0x80526000U,
    0x5fba1000U,
    0x8f5ee800U,
    0xd9095c00U,
    0x1820e00U,
    0x5c650700U,
    0xd98c8f80U,
    0xe0734ac0U,
    0xefbce7e0U,
    0x177c5c30U,
    0xc5268c88U,
    0xe7a2415cU,
    0x837a6b1eU,
    0x770e15bdU,
    0xea94eaacU,
    0x90c85af6U,
    0xf66c8490U,
    0x8a3489aU,
    0x75f4e086U,
    0x1d850baU,
    0xc3c48dd7U,
    0x36d74a41U,
    0x895eeea2U,
    0x76095710U,
    0x57020c59U,
    0xfaa50066U,
    0xd8ec8308U,
    0xc2634b9cU,
    0x6a94ecfeU,
    0x50c8598dU,
    0x966c8e24U,
    0xb8a347aaU,
    0xedf4e18eU,
    0x1dd85a27U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0xa8000000U,
    0x9c000000U,
    0xf6000000U,
    0x15000000U,
    0x89800000U,
    0xe7c00000U,
    0x75600000U,
    0x9a300000U,
    0x9b280000U,
    0x489c0000U,
    0x944e0000U,
    0xd9ad0000U,
    0xefca8000U,
    0x9960c000U,
    0x54332000U,
    0xba199000U,
    0xab1fa800U,
    0x90885400U,
    0x30480600U,
    0x1bac0d00U,
    0x90e60480U,
    0xf3f100c0U,
    0x2f6485e0U,
    0xa13dc890U,
    0xa3b1a7f8U,
    0xe4d55f2cU,
    0x79ca87ceU,
    0x5c60c3dfU,
    0x95b320e4U,
    0x51d9942aU,
    0x807fa276U,
    0x83b854d3U,
    0xd4e00ed8U,
    0xa1f0085cU,
    0xf84807a7U,
    0x57ac0c8bU,
    0x2ee60d44U,
    0xeaf1039aU,
    0xf8e487ffU,
    0xcffdc317U,
    0xa951ac80U,
    0x8c2554c0U,
    0x1e0283e0U,
    0x690cc590U,
    0xefb52378U,
    0x5ad89fecU,
    0x60d3222eU,
    0x8be99b4fU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0x78000000U,
    0xa4000000U,
    0x5e000000U,
    0xf3000000U,
    0x97800000U,
    0x6d400000U,
    0x4be00000U,
    0x18300000U,
    0x34280000U,
    0xc60c0000U,
    0x872a0000U,
    0xb1b70000U,
    0x3a7e8000U,
    0x824b4000U,
    0x86692000U,
    0xe876b000U,
    0xb375a800U,
    0xf8f6f400U,
    0x6f800600U,
    0x89400900U,
    0xf5e00980U,
    0x3b300740U,
    0xdba80260U,
    0xf4c0570U,
    0x92ca0c98U,
    0x5a8709e4U,
    0x99d68dc6U,
    0x29074757U,
    0x4aa327fcU,
    0x41f1bde3U,
    0xbd232330U,
    0xbcb1b37bU,
    0x6ec32dd5U,
    0xd081ba3fU,
    0x7ceb2a41U,
    0x418db5e3U,
    0x32412731U,
    0x6e7abd78U,
    0xd45fa0d7U,
    0x9941fdbfU,
    0x2dfe8000U,
    0xaf0b4000U,
    0x2d892000U,
    0x2046b000U,
    0xff5da800U,
    0x9afaf400U,
    0xb6aa0600U,
    0xcbf70900U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0x38000000U,
    0xc000000U,
    0x1e000000U,
    0x63000000U,
    0xcc800000U,
    0xb3c00000U,
    0xc6600000U,
    0xb0300000U,
    0x58180000U,
    0x5c140000U,
    0x56120000U,
    0x570d0000U,
    0xde848000U,
    0xcedf4000U,
    0x69cbe000U,
    0xcf6fd000U,
    0x2da56800U,
    0x2a599400U,
    0xbe000a00U,
    0x53000300U,
    0xd4800180U,
    0xcfc007c0U,
    0xe0600260U,
    0xdf3006f0U,
    0x8a980d28U,
    0x8cd40d0cU,
    0x5cf200aeU,
    0x54fd003fU,
    0x40fc89e7U,
    0x22fb4ec2U,
    0x67c1e0e0U,
    0xc476d0b1U,
    0xa533e089U,
    0xb38bdd9dU,
    0x94f6174U,
    0x52b0931bU,
    0x90ee8dc6U,
    0x2af64d93U,
    0x73c5695aU,
    0xa6699665U,
    0xe0180000U,
    0x10140000U,
    0x68120000U,
    0x440d0000U,
    0x2a048000U,
    0x711f4000U,
    0xb1abe000U,
    0x1c5fd000U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0xd8000000U,
    0xc4000000U,
    0xe2000000U,
    0x7b000000U,
    0x38800000U,
    0x10c00000U,
    0x3aa00000U,
    0xe0100000U,
    0x30180000U,
    0xb80c0000U,
    0x341e0000U,
    0x3a1b0000U,
    0xbf188000U,
    0xda9c4000U,
    0x6bcf6000U,
    0x2271000U,
    0xf0d21800U,
    0xabf5400U,
    0x58000200U,
    0x4000300U,
    0x82000180U,
    0x8b0003c0U,
    0xe0800360U,
    0xd4c00310U,
    0xd8a00388U,
    0x9b1001ecU,
    0x89800e2U,
    0xa8cc0043U,
    0xebe00eaU,
    0xda0b0380U,
    0x8f0080c0U,
    0x629042e0U,
    0x5fd160d0U,
    0x383c10e8U,
    0x4fca9afcU,
    0xd023176aU,
    0x33cf63afU,
    0x6271308U,
    0x72d21a43U,
    0x81bf57eaU,
    0xb8800000U,
    0xd0c00000U,
    0x5aa00000U,
    0x10100000U,
    0xe8180000U,
    0x7c0c0000U,
    0xd61e0000U,
    0x411b0000U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x18000000U,
    0x5c000000U,
    0xae000000U,
    0xa9000000U,
    0x3b800000U,
    0x78c00000U,
    0x18e00000U,
    0xb8100000U,
    0xcc180000U,
    0xe61c0000U,
    0x5d160000U,
    0xd1830000U,
    0x23cb8000U,
    0x2475c000U,
    0x52452000U,
    0x97bf7000U,
    0x3e231800U,
    0x31fd1c00U,
    0x43800200U,
    0x54c00300U,
    0x4ee00380U,
    0xfd1002c0U,
    0x41980060U,
    0x6bdc0170U,
    0xd07602b8U,
    0xb85302a4U,
    0xccb380eeU,
    0x2b9c1e3U,
    0xdbab2063U,
    0x183072e0U,
    0x8ce69b30U,
    0x2217df98U,
    0x6f18a374U,
    0x3689b046U,
    0x14dbb0fU,
    0x7f27ae51U,
    0x6e7e3b29U,
    0x995e6c2eU,
    0xe3351bc0U,
    0x207e1fa3U,
    0x804b8000U,
    0xc0b5c000U,
    0x4a52000U,
    0x36af7000U,
    0xd1bb1800U,
    0xf3211c00U,
    0xa8760200U,
    0x94530300U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0x38000000U,
    0x44000000U,
    0xda000000U,
    0xe5000000U,
    0x59800000U,
    0x35400000U,
    0x5ca00000U,
    0x74100000U,
    0x2080000U,
    0xf1140000U,
    0x4b9e0000U,
    0xac470000U,
    0x9b288000U,
    0x7e4b4000U,
    0xe234a000U,
    0xe9cf3000U,
    0x2070a800U,
    0x75e89400U,
    0xa6a00200U,
    0x21100100U,
    0xc3880280U,
    0x705403c0U,
    0xf53e00e0U,
    0x79570110U,
    0x1aa08368U,
    0x5f1f4394U,
    0xac8aa166U,
    0x4d830d5U,
    0xe5f02972U,
    0x8ea7d5d0U,
    0xd02a208U,
    0x958c32c4U,
    0x334e2baeU,
    0xd7b0d571U,
    0x2a82208cU,
    0xcfc371e9U,
    0x1b6c8a60U,
    0xb36ce7f3U,
    0xdf640b07U,
    0x2977a512U,
    0x9a78a800U,
    0x80fc9400U,
    0x973e0200U,
    0x98570100U,
    0x39208280U,
    0x7f5f43c0U,
    0x91aaa0e0U,
    0x1883110U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0x58000000U,
    0x44000000U,
    0x32000000U,
    0x87000000U,
    0xc9800000U,
    0x8b400000U,
    0xb3200000U,
    0xb4100000U,
    0xfa080000U,
    0x3b140000U,
    0x6f9a0000U,
    0x664b0000U,
    0xb9a88000U,
    0xc4d64000U,
    0x4578e000U,
    0xf5bd3000U,
    0xaad36800U,
    0x14616400U,
    0x59200200U,
    0x37100100U,
    0xa1880280U,
    0xe7540340U,
    0x4d3a0160U,
    0x1d1b0110U,
    0xc28080c8U,
    0xccd2421cU,
    0x196ae326U,
    0x23a2322dU,
    0xcfc1eaccU,
    0x2ec26d0U,
    0x5ff861e8U,
    0x3d6f71ecU,
    0x21b98b3eU,
    0x60c356d9U,
    0x761e986U,
    0xfabc2703U,
    0x7f50615dU,
    0xe92b738aU,
    0x5f0b8a6dU,
    0xcd8c542cU,
    0x195b6800U,
    0xe4356400U,
    0x259a0200U,
    0x354b0100U,
    0xba288280U,
    0x5c964340U,
    0x55d8e160U,
    0x9ed3110U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x8000000U,
    0xf4000000U,
    0x86000000U,
    0xd7000000U,
    0x5e800000U,
    0xec400000U,
    0x9b600000U,
    0xe6100000U,
    0x27180000U,
    0x66840000U,
    0x420000U,
    0xe1610000U,
    0x430e8000U,
    0x2888c000U,
    0x35ee000U,
    0x29e9d000U,
    0x704c8800U,
    0x1975ec00U,
    0x8f180200U,
    0x42840300U,
    0xae420080U,
    0xd2610040U,
    0x938e8020U,
    0xe7c8c3d0U,
    0x40bee218U,
    0xf4b9d35cU,
    0x92b4897aU,
    0x75a1efb1U,
    0x3322006dU,
    0x23710098U,
    0xca16801cU,
    0x7d0cc1daU,
    0xd39ce021U,
    0x7c8d055U,
    0x70a20b14U,
    0xecad2ffeU,
    0x6ebee377U,
    0x7b9d016U,
    0x62348b2cU,
    0xaae1efbdU,
    0x78c20080U,
    0xa210040U,
    0xaeee8020U,
    0xc6d8c3d0U,
    0x3126e218U,
    0x8a7dd35cU,
    0x8f96897aU,
    0xa5d0efb1U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0x88000000U,
    0x84000000U,
    0x12000000U,
    0xdd000000U,
    0xd6800000U,
    0x36400000U,
    0xd1e00000U,
    0x56100000U,
    0x6f080000U,
    0x3b9c0000U,
    0xb8da0000U,
    0xebb10000U,
    0x11e08000U,
    0xf61a4000U,
    0x5f0fa000U,
    0x639cd000U,
    0xb4cdc800U,
    0x7dbbbc00U,
    0xdee80200U,
    0xfd8c0100U,
    0xbfd20380U,
    0x842d0340U,
    0x333a8220U,
    0x44ab4210U,
    0x8a6f2048U,
    0x7ec69374U,
    0xeca26b5aU,
    0x7e776cd9U,
    0xd4cdc947U,
    0xedbbbc58U,
    0xb6e8003cU,
    0xa98c02aeU,
    0x25d20343U,
    0xdd2d02feU,
    0xf7ba822fU,
    0xafeb42bcU,
    0x8d0f226eU,
    0x1e969223U,
    0x524a6aceU,
    0x13fb6c77U,
    0x31fcb80U,
    0x3d96bf40U,
    0x1fd28020U,
    0xb4274310U,
    0x6b3d23c8U,
    0x48ab9034U,
    0x1c78e97aU,
    0xb1cc2ec9U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0xd8000000U,
    0xc000000U,
    0x1e000000U,
    0x5d000000U,
    0x2e800000U,
    0xd8400000U,
    0x83200000U,
    0xf2100000U,
    0xb3080000U,
    0xcb8c0000U,
    0x4ad60000U,
    0x9d7b0000U,
    0x20318000U,
    0x719bc000U,
    0xfdcfa000U,
    0x2cffd000U,
    0xfde60800U,
    0x8d72e400U,
    0x68280200U,
    0x759c0100U,
    0x87de0180U,
    0xbbf702c0U,
    0x9c678360U,
    0x38a0c030U,
    0x40de2078U,
    0xf2741174U,
    0x9da1a8baU,
    0xb2413761U,
    0x5c38080cU,
    0x9785e6c8U,
    0x7ccf814cU,
    0x747cc0eeU,
    0xaca023cbU,
    0x32d31085U,
    0x16e2b98U,
    0x363df482U,
    0x48982935U,
    0x1956f7a6U,
    0xbba1aac1U,
    0x1341355cU,
    0xd4b80980U,
    0xaec5e6c0U,
    0x176f8160U,
    0xf2cc130U,
    0xac0821f8U,
    0x8e0f13b4U,
    0x55102bdaU,
    0x4a9af751U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0xc8000000U,
    0xa4000000U,
    0xf2000000U,
    0x23000000U,
    0xba800000U,
    0x5ec00000U,
    0xc1600000U,
    0x1e100000U,
    0xb5180000U,
    0x8b9c0000U,
    0xb74e0000U,
    0xed390000U,
    0x25a48000U,
    0x98764000U,
    0x3906000U,
    0x335d5000U,
    0x8f20d800U,
    0xbeabac00U,
    0x4ef80200U,
    0xb4c0300U,
    0x23360380U,
    0x8b501c0U,
    0xeff28320U,
    0x5ad34290U,
    0x637ae3c8U,
    0x6512108cU,
    0x13943aeaU,
    0x4b40bd7bU,
    0x328b905U,
    0x98aaff78U,
    0x57f6db54U,
    0x36ceaceeU,
    0x3572807dU,
    0xb41343e4U,
    0x8a1ae3feU,
    0xaf0211bdU,
    0x9c8c3a0cU,
    0x47dcbfaaU,
    0xfce6ba1bU,
    0x853fc75U,
    0x9b25b80U,
    0xee68edc0U,
    0x429ae120U,
    0x12c21190U,
    0x76c3848U,
    0x770cbd4cU,
    0x409ebbcaU,
    0x39dffcebU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x58000000U,
    0x8c000000U,
    0x5a000000U,
    0xcd000000U,
    0x78800000U,
    0x2400000U,
    0x7600000U,
    0x5e100000U,
    0xe3180000U,
    0x938c0000U,
    0x25d20000U,
    0x64ab0000U,
    0xf5a18000U,
    0x2f234000U,
    0x9c7da000U,
    0x71991000U,
    0x34cd4800U,
    0x7e246c00U,
    0x26f80200U,
    0x52dc0300U,
    0x812a0180U,
    0xb7770240U,
    0x360b8160U,
    0xc7144230U,
    0x8d962168U,
    0x66dd5334U,
    0x72369e2U,
    0x18653c09U,
    0x9f916a1dU,
    0xbfde3e78U,
    0xc9a8e80cU,
    0x1d317f0eU,
    0x75674a77U,
    0x17136ee2U,
    0x7593805eU,
    0x1ad840f8U,
    0x9524221bU,
    0x166528bU,
    0x311ae924U,
    0x2a8a7d8dU,
    0x7b5ec980U,
    0x39fc2e40U,
    0x935c2360U,
    0xddfa5130U,
    0xed50e8e8U,
    0xeed7d74U,
    0xd6cd4a82U,
    0x6f246d39U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0x18000000U,
    0xcc000000U,
    0xe000000U,
    0xb1000000U,
    0x1a800000U,
    0xf3c00000U,
    0xb200000U,
    0x61100000U,
    0xb2980000U,
    0x7c40000U,
    0xb92e0000U,
    0xc6130000U,
    0xd5018000U,
    0xe085c000U,
    0xf0d82000U,
    0xb6a05000U,
    0xf557f800U,
    0x5ef8a400U,
    0x2f2e0200U,
    0x7b130300U,
    0xe1818080U,
    0xd245c1c0U,
    0xf9782060U,
    0xe8705330U,
    0x42eff838U,
    0x892ca6c4U,
    0x3e18026aU,
    0x490400cfU,
    0x868e00acU,
    0x95c30044U,
    0x6e3982aaU,
    0xb991c32fU,
    0xfe4e22dcU,
    0xa76751dcU,
    0x31607b3eU,
    0x8c6a644dU,
    0xb8f9a1efU,
    0x8a35905eU,
    0x8397d99fU,
    0xdd5cf4a4U,
    0x6af7f838U,
    0xbd28a6c4U,
    0xac16026aU,
    0x9e0700cfU,
    0xf91780acU,
    0xbe82c044U,
    0x29cfa2aaU,
    0x7822932fU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0xd8000000U,
    0x4000000U,
    0xe2000000U,
    0xe7000000U,
    0x95800000U,
    0xbdc00000U,
    0x6ba00000U,
    0xdf100000U,
    0x81880000U,
    0xe7cc0000U,
    0xd8aa0000U,
    0x708b0000U,
    0xdf488000U,
    0xfbf04000U,
    0x5d66e000U,
    0xf31b000U,
    0x5a433800U,
    0xc97f3400U,
    0x152a0200U,
    0x894b0100U,
    0x36e88180U,
    0x93e04140U,
    0x916ee360U,
    0x513db010U,
    0xb493b88U,
    0x81e4379cU,
    0xde6a8056U,
    0x28b743f7U,
    0xd884602eU,
    0x334af23cU,
    0x95ed5966U,
    0x847ec78fU,
    0x9bafdaeaU,
    0x7715875eU,
    0x6d89baabU,
    0x89d87558U,
    0x1a6603bU,
    0xe40df310U,
    0xf20fda07U,
    0x5f058676U,
    0xc181bb88U,
    0x87d4779cU,
    0x88ac6056U,
    0xa896f3f7U,
    0xdb4f582eU,
    0x19f9c63cU,
    0xba6d5b66U,
    0x9abec68fU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0xc8000000U,
    0x74000000U,
    0xa000000U,
    0xff000000U,
    0x45800000U,
    0xdec00000U,
    0xfde00000U,
    0xe7100000U,
    0xa9980000U,
    0xf8cc0000U,
    0x44fe0000U,
    0xeb890000U,
    0xb3d68000U,
    0x6b6d4000U,
    0x3a51e000U,
    0xc3a9b000U,
    0x26a95800U,
    0x6428fc00U,
    0x237e0200U,
    0x8e490300U,
    0xa9b68180U,
    0x29bd43c0U,
    0xe9a9e320U,
    0x89b5b1d0U,
    0x79af5828U,
    0xb1bdfffcU,
    0xc5ae8316U,
    0xcfb1407bU,
    0x30b7e277U,
    0x752cb05cU,
    0xabe1d986U,
    0x561cbe33U,
    0xb101633bU,
    0x1891f352U,
    0xe0483a59U,
    0xa4a90d16U,
    0x4f2e3a3eU,
    0xfcec0e42U,
    0x9786b994U,
    0xadc84d4fU,
    0x6e61d828U,
    0x48dcbffcU,
    0x2ce16316U,
    0xf81f07bU,
    0x81d03a77U,
    0x28650c5cU,
    0x1d03b86U,
    0xe8650d33U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x68000000U,
    0x24000000U,
    0xda000000U,
    0xc1000000U,
    0xd1800000U,
    0x17c00000U,
    0x2200000U,
    0x75100000U,
    0xa3980000U,
    0xd2c40000U,
    0x19a20000U,
    0xcbdd0000U,
    0x543c8000U,
    0x6a0f4000U,
    0x79022000U,
    0xbd933000U,
    0xf9c47800U,
    0x71360400U,
    0x41820200U,
    0xbfcd0300U,
    0x6248080U,
    0xbf0b4040U,
    0xa8021a0U,
    0x275e3090U,
    0xd460fb68U,
    0x8fd4704U,
    0xf0a22146U,
    0xde43335fU,
    0xa9fc7888U,
    0xd1220594U,
    0x9198012eU,
    0xf7c400dbU,
    0x3222038eU,
    0xd1d006bU,
    0xef9c83b6U,
    0x2cdf4167U,
    0x2ba2284U,
    0xdb473026U,
    0x927e7a2fU,
    0x7fef06d0U,
    0xe3c8368U,
    0x6b0f4304U,
    0x88822346U,
    0xba53305fU,
    0x93e47808U,
    0x202605d4U,
    0x381a008eU,
    0xac09004bU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x18000000U,
    0x1c000000U,
    0x3a000000U,
    0x29000000U,
    0x43800000U,
    0xd1c00000U,
    0x6c600000U,
    0xad100000U,
    0x25880000U,
    0x22dc0000U,
    0x96f20000U,
    0x27530000U,
    0x842b8000U,
    0xd9ab4000U,
    0x376f2000U,
    0x638f7000U,
    0xe1dfb800U,
    0x467cc00U,
    0x39120200U,
    0x1b830100U,
    0x2dc38380U,
    0xc6674240U,
    0x9c152060U,
    0x7a007070U,
    0xc90638e8U,
    0xd39f8ca4U,
    0xc9d6a30eU,
    0x70673247U,
    0x97131a31U,
    0xc9ffcf4U,
    0x61509af6U,
    0x4738bdfbU,
    0x4b25bb33U,
    0x2928cf79U,
    0xfc2b82feU,
    0x15ab4311U,
    0xf56f2264U,
    0xc68f71aeU,
    0x805fba97U,
    0xe0a7cf09U,
    0x2cf200e8U,
    0x4e5300a4U,
    0x27ab810eU,
    0x986b4347U,
    0x430f21b1U,
    0xd29f72b4U,
    0xfe57b896U,
    0xfbbcc8bU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0xc8000000U,
    0xc4000000U,
    0xf6000000U,
    0xd9000000U,
    0x38800000U,
    0xe3400000U,
    0x1ae00000U,
    0x9f100000U,
    0x19880000U,
    0xa7dc0000U,
    0x93b20000U,
    0xa6e90000U,
    0x55008000U,
    0x4a8ac000U,
    0x2c552000U,
    0x6b6c1000U,
    0x8ccde800U,
    0x242d1c00U,
    0xd43a0200U,
    0xcc350100U,
    0x20328380U,
    0x8223c240U,
    0x6535a320U,
    0x40b6d310U,
    0x6d70cbd8U,
    0x4dcd0f64U,
    0xf0adeae2U,
    0x957d1e8dU,
    0x31d201ebU,
    0x9ab9013cU,
    0xb66880c6U,
    0xe246c3cfU,
    0x6e6f22b6U,
    0xae5912efU,
    0xbc7f6a6eU,
    0x114eddc3U,
    0x95efa238U,
    0xe93d15cU,
    0x9a4a49beU,
    0x5272ce33U,
    0x244a4858U,
    0xf72cd24U,
    0xaca49c2U,
    0xa532cd9dU,
    0xe0aa4a33U,
    0x1d62ce58U,
    0x15c24a24U,
    0xfcaecd42U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0xb8000000U,
    0x3c000000U,
    0x56000000U,
    0x85000000U,
    0x6c800000U,
    0x51c00000U,
    0x70a00000U,
    0xeb100000U,
    0x95880000U,
    0x8b5c0000U,
    0x94660000U,
    0x4f270000U,
    0x135f8000U,
    0x387ec000U,
    0x713aa000U,
    0x22449000U,
    0x86eeb800U,
    0xcfe95400U,
    0xed6e0200U,
    0xd5bb0100U,
    0x17998380U,
    0xac49c0c0U,
    0x4fed22e0U,
    0xad6650f0U,
    0x35b21958U,
    0x278ac614U,
    0x145f3bb2U,
    0x73ec9447U,
    0xfb6d2042U,
    0xb0a6526cU,
    0x4b121b36U,
    0x459ac61dU,
    0x3573be9U,
    0x107096d8U,
    0x252b2127U,
    0xc0515272U,
    0xd1e598d4U,
    0x4c680652U,
    0x3239837U,
    0x1d5f055aU,
    0xb1741ad8U,
    0xf3bdc6d4U,
    0xca88b952U,
    0xccce54b7U,
    0xf031811aU,
    0x64c5c078U,
    0xe4232284U,
    0x66cd535aU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x68000000U,
    0xd4000000U,
    0x7e000000U,
    0x7b000000U,
    0xee800000U,
    0xb1c00000U,
    0xad600000U,
    0x51100000U,
    0xab880000U,
    0x444c0000U,
    0xc2260000U,
    0x25bd0000U,
    0x83e28000U,
    0xc0dbc000U,
    0x56ed6000U,
    0x4d4ad000U,
    0xd5afb800U,
    0xcbf1ec00U,
    0x44ce0200U,
    0x70e10100U,
    0x8a4c8180U,
    0x912ac0c0U,
    0x7f29e1a0U,
    0x7c2c1350U,
    0x7ea059f8U,
    0x3d60fdecU,
    0x290cd9baU,
    0x479a3fc7U,
    0xb64d3935U,
    0x7b2a2c84U,
    0x1a23608eU,
    0xd9abd381U,
    0x49e33ad0U,
    0xf5db2eeaU,
    0xab67e3edU,
    0xc60d1278U,
    0xb70cd82cU,
    0x4c9a3f9aU,
    0x50cd3b57U,
    0x2eea2fedU,
    0xa1436078U,
    0x27bbd12cU,
    0x72eb3a1aU,
    0x7b572d97U,
    0x2aa1e14dU,
    0x36011a8U,
    0x32065894U,
    0x991dff16U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x38000000U,
    0x74000000U,
    0x96000000U,
    0x77000000U,
    0x5a800000U,
    0xef400000U,
    0x1ee00000U,
    0x35100000U,
    0x6b880000U,
    0x6acc0000U,
    0x173e0000U,
    0x1eb70000U,
    0xd5768000U,
    0x8fc6c000U,
    0xd4b4e000U,
    0xc86e5000U,
    0x18526800U,
    0x47c9c00U,
    0xba560200U,
    0x856b0100U,
    0x67c08180U,
    0x8bdc3c0U,
    0xe27c60e0U,
    0xb55f91d0U,
    0xcbf00a58U,
    0xee840ddcU,
    0x59588b6aU,
    0xf9e5cebdU,
    0xa792ebfbU,
    0xc8c15e14U,
    0x962a62ceU,
    0xc33492bbU,
    0x58b08b64U,
    0xa79cfb6U,
    0x6944ea87U,
    0xe1ea5c8eU,
    0x438ae39bU,
    0x96d95354U,
    0xad24e9beU,
    0x7bba5cc3U,
    0x56e2e3d8U,
    0x3905521cU,
    0xe992e98aU,
    0x7bc15e6dU,
    0x2aa6023U,
    0xdf749008U,
    0xb2d08944U,
    0xd329cdd6U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0xd8000000U,
    0x64000000U,
    0xda000000U,
    0x97000000U,
    0x85800000U,
    0x18c00000U,
    0xb6a00000U,
    0x59100000U,
    0x70880000U,
    0xfe440000U,
    0x99f60000U,
    0xdc6b0000U,
    0xa4b48000U,
    0x82074000U,
    0xb308e000U,
    0x7f99b000U,
    0x3fc29800U,
    0xeb3a1400U,
    0x25de0200U,
    0x1c3f0100U,
    0x304a8080U,
    0x6ce842c0U,
    0x3aea6360U,
    0x8be5f190U,
    0x776fb68U,
    0x6720e65cU,
    0x3c27816U,
    0x1527a563U,
    0x48ca985aU,
    0xbebe16a4U,
    0x75080022U,
    0xa68401a9U,
    0xf5602efU,
    0x357b027dU,
    0xc3c808cU,
    0x184341a7U,
    0xf0fee168U,
    0x34f2b2a5U,
    0x1ef619f1U,
    0x71fd5552U,
    0x2076e3e8U,
    0x3ab6b09cU,
    0x7f001976U,
    0x799655f3U,
    0x86c263b2U,
    0x4bb1f238U,
    0x9388fb54U,
    0x89cfe55aU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0x28000000U,
    0x14000000U,
    0x6a000000U,
    0x91000000U,
    0x5800000U,
    0xb5400000U,
    0xbba00000U,
    0xa7100000U,
    0xda980000U,
    0x83dc0000U,
    0x7e7e0000U,
    0xee750000U,
    0xf66a8000U,
    0x2a794000U,
    0xa4782000U,
    0x776fb000U,
    0xf7fb2800U,
    0x2d37b400U,
    0xb2c60200U,
    0xabf90300U,
    0x632c8380U,
    0x81c043c0U,
    0xdb74a0a0U,
    0x91fff050U,
    0xa3789a8U,
    0xa8444644U,
    0xc8378a16U,
    0xed4445d5U,
    0x47b78b6eU,
    0x3904465cU,
    0xd1978a4aU,
    0x3f54469fU,
    0xdaaf8971U,
    0x8a98456dU,
    0x7bc98b47U,
    0x527144a8U,
    0x487d08e1U,
    0x316d0740U,
    0xc0f7aac5U,
    0xefa7f426U,
    0xed0aa228U,
    0x5b8af184U,
    0x5e5d08b6U,
    0xf73d0685U,
    0x2bcfa946U,
    0xaa6bf7d8U,
    0x646ca2fcU,
    0x9763f01aU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0x48000000U,
    0xac000000U,
    0x8e000000U,
    0xe5000000U,
    0x4e800000U,
    0x97c00000U,
    0xe5600000U,
    0x3e900000U,
    0xfd80000U,
    0xe17c0000U,
    0xc920000U,
    0x2cd10000U,
    0xe6e98000U,
    0x5bc3c000U,
    0xbb6da000U,
    0x73905000U,
    0xfd593800U,
    0xb0b50c00U,
    0xa0e98200U,
    0xd2c3c300U,
    0x9beda380U,
    0x11505040U,
    0x1eb93920U,
    0xb5e50eb0U,
    0xc4518038U,
    0xe82fc094U,
    0xd627a2baU,
    0x4b3d521fU,
    0x11a2bab5U,
    0xfc67ce4aU,
    0x960da007U,
    0x21005311U,
    0x9c813a88U,
    0xa4c90eacU,
    0xaafb832eU,
    0xc5d2c025U,
    0x1664236aU,
    0x9103919fU,
    0xe48c9bddU,
    0xb0c95e9eU,
    0xd0fabab5U,
    0x4adbce4aU,
    0x9fffa007U,
    0x23415311U,
    0x3db0ba88U,
    0xb276ceacU,
    0x9304232eU,
    0x7f939025U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0xc8000000U,
    0x4000000U,
    0xa000000U,
    0x1b000000U,
    0xeb800000U,
    0x57400000U,
    0xfe00000U,
    0x63900000U,
    0xf3580000U,
    0xb5ec0000U,
    0x608a0000U,
    0xd4c10000U,
    0xecac8000U,
    0x7e6b4000U,
    0x47d26000U,
    0xa329f000U,
    0x932da800U,
    0x4b299c00U,
    0xe72c8200U,
    0x392b4300U,
    0xe0326180U,
    0x14b9f340U,
    0xa275ab20U,
    0xe1c59c10U,
    0x66268228U,
    0xa1aa436cU,
    0xe8fee22eU,
    0x5e02b21dU,
    0x191fcb1fU,
    0x94906d9eU,
    0x66d929e5U,
    0x8baeddbbU,
    0x83f4e1acU,
    0xcd83b24eU,
    0x5253492dU,
    0x5d6b2f27U,
    0x10534bdaU,
    0x826b2f27U,
    0x91d34bc8U,
    0x1e2b2f0eU,
    0xbdb34b1fU,
    0x2efb2d9eU,
    0x4b0b49e5U,
    0xe3872dbbU,
    0x335949acU,
    0xd5ea2e4eU,
    0xb09fcb2dU,
    0x1cd06c27U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0xb8000000U,
    0x94000000U,
    0x12000000U,
    0xdd000000U,
    0x86800000U,
    0xe0400000U,
    0xd2a00000U,
    0xb2900000U,
    0xa2480000U,
    0x97a40000U,
    0xb0120000U,
    0xe81f0000U,
    0xc168000U,
    0x96004000U,
    0x770ca000U,
    0xcf825000U,
    0x74de4800U,
    0xefe1f400U,
    0xe6b68200U,
    0xf0904100U,
    0xe744a080U,
    0x95265040U,
    0xfa4c4ae0U,
    0x73bef650U,
    0x2a000248U,
    0x9000274U,
    0xb480029aU,
    0x2d4003c1U,
    0xec2001aaU,
    0xc6d0009aU,
    0x62e802c1U,
    0xf834012aU,
    0x94da00daU,
    0x9ffb0021U,
    0x6ea4837aU,
    0xcc8f4092U,
    0xd9522355U,
    0xce261160U,
    0xb3c0eb13U,
    0xc87ca41fU,
    0x9e7e49aaU,
    0x8971f49aU,
    0x76fe80c1U,
    0xaa34402aU,
    0x69d6a05aU,
    0x9795061U,
    0x36fac99aU,
    0x8a2eb6c2U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0xf8000000U,
    0x74000000U,
    0x66000000U,
    0x13000000U,
    0x9b800000U,
    0x8a400000U,
    0xf6200000U,
    0x5f900000U,
    0x54480000U,
    0xf12c0000U,
    0x92160000U,
    0x35170000U,
    0xe8828000U,
    0x21cac000U,
    0x847d6000U,
    0xddadf000U,
    0x6dc58800U,
    0xb678a400U,
    0xf8a28200U,
    0x2d5ac100U,
    0x2bb56180U,
    0x96c1f0c0U,
    0xf1f38be0U,
    0xa8ffa5d0U,
    0x22680398U,
    0xeebc014cU,
    0xa65e03eeU,
    0xf43b02e9U,
    0x82948038U,
    0x60ddc0aeU,
    0xaffe0c9U,
    0xef673388U,
    0x7238e826U,
    0xe19556fdU,
    0x63470a7aU,
    0xc4b264f9U,
    0x875fe0c0U,
    0x4ab73172U,
    0x4850e99fU,
    0xb29571dU,
    0x3b190838U,
    0x678964aeU,
    0x4b62c9U,
    0xc72af288U,
    0x490f09a6U,
    0x229e663dU,
    0x70c9e19aU,
    0xa2e03129U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x28000000U,
    0xdc000000U,
    0x22000000U,
    0x27000000U,
    0xed800000U,
    0xa7c00000U,
    0xbf200000U,
    0x9900000U,
    0x51d80000U,
    0xd63c0000U,
    0x991a0000U,
    0xc89d0000U,
    0xdd478000U,
    0x8d6e4000U,
    0xe571e000U,
    0x196e3000U,
    0x8b64b800U,
    0xd46d0400U,
    0xcde78200U,
    0x943e4300U,
    0x2e09e380U,
    0xed023140U,
    0xf686b8a0U,
    0x385c0770U,
    0x17e20288U,
    0x8731039cU,
    0xdd858036U,
    0x1fcf43dfU,
    0xab2c625cU,
    0xa79d7356U,
    0x7cd2d9cfU,
    0xc0ad77c4U,
    0xf152dbd2U,
    0xf76d76bdU,
    0x8672d809U,
    0x72fd7453U,
    0xddaadbf2U,
    0x5fc1760dU,
    0x8b30d9e1U,
    0x179c75dfU,
    0x4d75a5cU,
    0x34a23756U,
    0xf5ebbcfU,
    0xf26004c4U,
    0x4cf80052U,
    0x38ac03fdU,
    0xc54202a9U,
    0xe9610023U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0x8000000U,
    0x8c000000U,
    0x5e000000U,
    0x1b000000U,
    0xda800000U,
    0x62c00000U,
    0xcc600000U,
    0x2e900000U,
    0xe8c80000U,
    0xed7c0000U,
    0x2d120000U,
    0x1d890000U,
    0xe4d8000U,
    0x29a9c000U,
    0x6e726000U,
    0x3392d000U,
    0x4d539800U,
    0x9724ec00U,
    0xcead8200U,
    0xb2f9c100U,
    0x2e5a6380U,
    0x99bed040U,
    0x96699820U,
    0xa781ee30U,
    0x975a0378U,
    0x5e35016cU,
    0x513f80eaU,
    0x11b0c1cbU,
    0x5a77e311U,
    0x1987128aU,
    0x9c53fadbU,
    0x8caf3fd9U,
    0xbffb99deU,
    0x83c8ecfdU,
    0x6ff78288U,
    0x2bccc15cU,
    0x13e5e1baU,
    0x6dce110bU,
    0x8cfe78d9U,
    0x6556fc76U,
    0xab21fb11U,
    0x68b63e8aU,
    0x3dfe18dbU,
    0x2edd2ed9U,
    0x3277e25eU,
    0xc58710bdU,
    0x2a53f8a8U,
    0xbaf3e6cU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x68000000U,
    0x5c000000U,
    0x56000000U,
    0x87000000U,
    0xa3800000U,
    0x22c00000U,
    0x30a00000U,
    0x95900000U,
    0xd5c80000U,
    0x1b3c0000U,
    0xdb560000U,
    0xdb650000U,
    0x55a78000U,
    0xfb14c000U,
    0x6589e000U,
    0x1dc5f000U,
    0x97259800U,
    0xd5407400U,
    0x686f8200U,
    0xfc28c100U,
    0x8dfe380U,
    0x71a0f0c0U,
    0x90219a0U,
    0x5094b570U,
    0x6b466358U,
    0xf37d331cU,
    0xe9b2790eU,
    0x9d1c844bU,
    0x8a9b9b62U,
    0xe2497626U,
    0xe3f6038fU,
    0x62f503b0U,
    0x5e6f83c3U,
    0xb28c256U,
    0x235fe16cU,
    0x3f60f2d6U,
    0x7a21917U,
    0x1e04b68cU,
    0x4b0e617dU,
    0x4d8133e5U,
    0xa1c478e2U,
    0xf12986e6U,
    0x3a541a2fU,
    0x97f1b6c0U,
    0x88e1e09bU,
    0xbf69f14aU,
    0x47bb9862U,
    0xfe19769dU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x48000000U,
    0x64000000U,
    0xb6000000U,
    0x19000000U,
    0xd6800000U,
    0x36c00000U,
    0x40200000U,
    0xd0900000U,
    0xa7d80000U,
    0x12bc0000U,
    0x60560000U,
    0xb6f10000U,
    0x70a08000U,
    0x4758c000U,
    0xfd6e2000U,
    0x16fc5000U,
    0x60b71800U,
    0xdf5aa400U,
    0xe1788200U,
    0x8ce4c300U,
    0xabb82380U,
    0xa6cd50c0U,
    0x18379920U,
    0x2c926590U,
    0xdcea0d8U,
    0x91a49364U,
    0x7dd938daU,
    0xf9a6f41bU,
    0xc9cf9a20U,
    0x37be65d2U,
    0xfcc0a1c7U,
    0x332993eeU,
    0x650fb87bU,
    0xbc9f3750U,
    0x55d9393aU,
    0x6da6f4abU,
    0xd7cf9968U,
    0x7abe676eU,
    0xd440a2f9U,
    0x78e9906fU,
    0x45afb9a0U,
    0x43cf3512U,
    0x64a138e7U,
    0x994af67eU,
    0x506198a3U,
    0xe636434U,
    0x636e21e0U,
    0x9bfc50b0U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x68000000U,
    0x74000000U,
    0xee000000U,
    0xc9000000U,
    0x8a800000U,
    0x27400000U,
    0xff600000U,
    0x8c900000U,
    0xda480000U,
    0xfbec0000U,
    0x32da0000U,
    0xc7a50000U,
    0x24328000U,
    0xf9efc000U,
    0x79c02000U,
    0x4631d000U,
    0xe2e22800U,
    0x90574c00U,
    0xf4fa8200U,
    0x6543c100U,
    0xd47a2180U,
    0x5d04d140U,
    0x7498a9a0U,
    0xe6548dd0U,
    0x51e0a1b8U,
    0x2dd71224U,
    0x982a8baaU,
    0x13fc5dddU,
    0x6c2085dU,
    0x49b69ee2U,
    0xbd30ab51U,
    0x1b688c8bU,
    0x2a92a0c1U,
    0x574e1213U,
    0xa76a08f5U,
    0xc08a9e61U,
    0x2842a8bcU,
    0xa8f18d9eU,
    0x9f5222afU,
    0xa378d094U,
    0x768aa9ddU,
    0xad5d8fa2U,
    0xd06822f1U,
    0xeb0dd15bU,
    0xf1902979U,
    0x9ece4c37U,
    0x25ba015fU,
    0x7f3502bcU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xe8000000U,
    0xe4000000U,
    0x1a000000U,
    0x11000000U,
    0x53800000U,
    0xb8c00000U,
    0x80e00000U,
    0xd1900000U,
    0xc5d80000U,
    0xbd740000U,
    0xba560000U,
    0xba50000U,
    0x79b88000U,
    0x4cad4000U,
    0xa5232000U,
    0x3cecf000U,
    0x7935800U,
    0x8ad9bc00U,
    0xd5e08200U,
    0xa8194300U,
    0x84152280U,
    0x8a19f0c0U,
    0xc913dba0U,
    0x5f90ff90U,
    0x46cda268U,
    0x8be4b344U,
    0x9308fbceU,
    0x2e980e23U,
    0x8550fba3U,
    0xeb2c0fd6U,
    0x1fe6fbffU,
    0x1190d71U,
    0x4b867887U,
    0x14c04e9dU,
    0x4ef358adU,
    0x3289be60U,
    0x63588385U,
    0x4c3d40f1U,
    0x937b20f2U,
    0x958f0b4U,
    0xd5255923U,
    0xb4ecbf16U,
    0x7380005fU,
    0x48c002e1U,
    0xc8e002efU,
    0x59001d9U,
    0x37d80163U,
    0x48740343U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xe8000000U,
    0x54000000U,
    0xbe000000U,
    0x1b000000U,
    0x7f800000U,
    0xa9c00000U,
    0xbf600000U,
    0x3d900000U,
    0xf8d80000U,
    0x55ec0000U,
    0x48da0000U,
    0x2de50000U,
    0x94d68000U,
    0x97e5c000U,
    0xd9c26000U,
    0xa7777000U,
    0xb183f800U,
    0xaacc0c00U,
    0x4cee8200U,
    0xc659c300U,
    0x5ea06180U,
    0xe6ee7140U,
    0xc3577ba0U,
    0xa20cd50U,
    0x88a060f8U,
    0x69ee736cU,
    0x62d7787eU,
    0xe8e0cfe7U,
    0xa040635dU,
    0xa9be72a6U,
    0x9b6f789bU,
    0x9b9ccc7bU,
    0x6fc2617dU,
    0x78777040U,
    0xf803fbd6U,
    0x1c0c0f85U,
    0xda0e80daU,
    0x1d09c33dU,
    0xd8986260U,
    0x3c5273c6U,
    0xb3b57addU,
    0xe679cfe6U,
    0x1314e13bU,
    0xbb92b22bU,
    0x9fc19805U,
    0xa07b7e6cU,
    0x140d7a08U,
    0x1e05ce32U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0x48000000U,
    0x14000000U,
    0x2a000000U,
    0x47000000U,
    0x1e800000U,
    0xf0400000U,
    0x96200000U,
    0x8900000U,
    0xb9580000U,
    0x4dbc0000U,
    0x454e0000U,
    0xa3b10000U,
    0x704e8000U,
    0x56234000U,
    0xe88de000U,
    0x49475000U,
    0x5ae4800U,
    0x5156a400U,
    0x89b68200U,
    0x374f4300U,
    0x48bbe380U,
    0x18da53c0U,
    0xdf76c920U,
    0xd38e450U,
    0xe81be2a8U,
    0xc40a521cU,
    0x720ecbfaU,
    0xeb14e401U,
    0x688de178U,
    0x89475372U,
    0xe5ae49cdU,
    0xa156a7eaU,
    0xc1b6824fU,
    0x234f421fU,
    0x62bbe0b1U,
    0x5fda53a6U,
    0xc1f6c81dU,
    0xfd78e782U,
    0x7e3be1f3U,
    0xcc9a51f5U,
    0xcb56caf8U,
    0xa6a8e4b2U,
    0x2dc3e2edU,
    0x2af650baU,
    0x95e0cb67U,
    0xf775e7c3U,
    0x293b606bU,
    0x6a0810f7U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x78000000U,
    0xe4000000U,
    0x42000000U,
    0x5d000000U,
    0xe5800000U,
    0xd8c00000U,
    0x1be00000U,
    0xb900000U,
    0xebd80000U,
    0xd740000U,
    0x65ca0000U,
    0x4e770000U,
    0xdb588000U,
    0xbc2a4000U,
    0xd8e0a000U,
    0x75123000U,
    0xb9905800U,
    0xdeda5c00U,
    0xd4e08200U,
    0xcb0e4300U,
    0x3292a280U,
    0x8c413340U,
    0x28bad9e0U,
    0x4ba31f90U,
    0xe52aa308U,
    0xcf653274U,
    0xf8c8d916U,
    0xbf01c23U,
    0xd380238fU,
    0xdfdc70beU,
    0x3762f827U,
    0xdccb6e01U,
    0xe9e259e1U,
    0x5e895ecaU,
    0x424a028aU,
    0xbb703eaU,
    0x8538803aU,
    0xbf7a4252U,
    0x50d8a3f6U,
    0x97f63088U,
    0x7582590fU,
    0xc0d95ffeU,
    0x8ff203c7U,
    0xe1930291U,
    0x2aca8069U,
    0x4ee943feU,
    0xa212207cU,
    0xed1f7359U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0x88000000U,
    0xac000000U,
    0xfe000000U,
    0xbf000000U,
    0x21800000U,
    0x2c400000U,
    0x86200000U,
    0x5e900000U,
    0x6dc80000U,
    0x7a640000U,
    0xa0ae0000U,
    0x17490000U,
    0x45b98000U,
    0x9bd5c000U,
    0x2970e000U,
    0x5f34b000U,
    0x7409c800U,
    0x6a05e400U,
    0x45178200U,
    0x3c9cc100U,
    0xc4c96080U,
    0x98e173c0U,
    0x2b792a20U,
    0x263156b0U,
    0xee9e49f8U,
    0xc5d927fcU,
    0x267ee206U,
    0xd6adb271U,
    0x45848b8U,
    0x9a24240aU,
    0x881626fU,
    0x8ec570a5U,
    0x2df7297cU,
    0x9fe855d0U,
    0x4eefcba8U,
    0x8868e414U,
    0x51a00272U,
    0x21d001d3U,
    0x146800c3U,
    0x47b401ddU,
    0xe2c60140U,
    0xb3fd03f6U,
    0xf0ff8069U,
    0x1768c2d4U,
    0x802f61c4U,
    0x858c71daU,
    0x3e4ea9c7U,
    0xe73d94b1U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xb8000000U,
    0x1c000000U,
    0x2a000000U,
    0xc7000000U,
    0x25800000U,
    0x49400000U,
    0xc9600000U,
    0xf2900000U,
    0x44c80000U,
    0xf4240000U,
    0x45ee0000U,
    0xff4f0000U,
    0x646f8000U,
    0x300f4000U,
    0xd819e000U,
    0xcc143000U,
    0x6211e800U,
    0x63138c00U,
    0x13818200U,
    0xa4404100U,
    0x2bf66080U,
    0x9e5b73c0U,
    0xc4e80ae0U,
    0xcfd7bc70U,
    0xf3b86aa8U,
    0x4fa7ce1cU,
    0xd5b1e216U,
    0x8aa033e5U,
    0x4337e945U,
    0x98788c7aU,
    0x8a00015bU,
    0x770001bcU,
    0xbd8001a9U,
    0xa5400204U,
    0x5b6000c2U,
    0x2990025fU,
    0x4b4801feU,
    0x7a640036U,
    0xa90e011aU,
    0x449f0084U,
    0xe9c783edU,
    0x36bb4266U,
    0xd93fe34dU,
    0xc77f3259U,
    0x439068ecU,
    0xac53ce7eU,
    0xaff7e199U,
    0x585b33e3U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xd0000000U,
    0x78000000U,
    0x9c000000U,
    0x2a000000U,
    0x95000000U,
    0x20800000U,
    0xb9400000U,
    0x27e00000U,
    0xb9900000U,
    0xebd80000U,
    0xa7bc0000U,
    0x3cea0000U,
    0x9a070000U,
    0xfd178000U,
    0x24834000U,
    0xdf47a000U,
    0xe0f89000U,
    0x9011e800U,
    0x580b1c00U,
    0xac1d8200U,
    0x82144300U,
    0x71082380U,
    0x9687d340U,
    0x65c49e0U,
    0x92648e70U,
    0x2043eaa8U,
    0x75601d54U,
    0x39d80102U,
    0x6ebc02a5U,
    0xd66a02ffU,
    0x664703d6U,
    0x827782e7U,
    0xb85342baU,
    0x397fa0d9U,
    0x6bd492d9U,
    0x67a3ebe9U,
    0xdcf01c91U,
    0x4a00024dU,
    0x850000fbU,
    0xb880029cU,
    0xf54002d6U,
    0x75e00057U,
    0xb0900282U,
    0xe15801e5U,
    0x8bfc031fU,
    0x3b8a01a6U,
    0x9ad7024fU,
    0x312f80eeU,
    0x3aaf405bU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xb8000000U,
    0x34000000U,
    0xca000000U,
    0x59000000U,
    0x8a800000U,
    0x81c00000U,
    0x58e00000U,
    0x37900000U,
    0x9580000U,
    0xec3c0000U,
    0xabe60000U,
    0x941f0000U,
    0x5a028000U,
    0xb1134000U,
    0xb68a2000U,
    0xc7c0d000U,
    0xffeff800U,
    0x2e097c00U,
    0x5b048200U,
    0xbf9c4300U,
    0x4550a380U,
    0xf22f92c0U,
    0x88e3dae0U,
    0xff86acd0U,
    0x6551f928U,
    0xa22a7e64U,
    0x80e003aaU,
    0x739003c7U,
    0x9b580003U,
    0x313c02ceU,
    0x536601edU,
    0x78df0204U,
    0x426283adU,
    0x5e434133U,
    0x6db22221U,
    0x9dacd01dU,
    0x5b1fb9cU,
    0x61ba7c71U,
    0xa3b80375U,
    0x76ac0258U,
    0x23e012bU,
    0x10e300aaU,
    0x9b848047U,
    0xa75c42c3U,
    0x7730a02eU,
    0xf47f913dU,
    0x615bd92cU,
    0x102aaec9U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0x98000000U,
    0x6c000000U,
    0xaa000000U,
    0xcb000000U,
    0xdc800000U,
    0x18400000U,
    0xece00000U,
    0x93900000U,
    0x5ad80000U,
    0x19bc0000U,
    0x32ee0000U,
    0x5e9b0000U,
    0x7498000U,
    0xe6774000U,
    0x82c46000U,
    0x55bd1000U,
    0x88f7c800U,
    0xfd87bc00U,
    0x2fc78200U,
    0x383c4300U,
    0xf35e380U,
    0xd1a653c0U,
    0x16e5aa60U,
    0x108dadb0U,
    0x624fc8a8U,
    0xafebbc2cU,
    0x4b1182f2U,
    0x1c8b40a1U,
    0xf84a6253U,
    0x1cf6103eU,
    0xb8649a3U,
    0x36dcfefaU,
    0xb3b5e311U,
    0xf9e65137U,
    0x8205ab5cU,
    0x1f1dae76U,
    0xa97c8b3U,
    0x1157bdceU,
    0xf7f80abU,
    0x91504106U,
    0xcf63e2fbU,
    0x71515012U,
    0x3f7a2951U,
    0xe94ded5bU,
    0x53742ac2U,
    0x4346eec9U,
    0x9865aa9fU,
    0x9fcdae3cU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0xe8000000U,
    0x64000000U,
    0x2a000000U,
    0x9b000000U,
    0x5800000U,
    0x2fc00000U,
    0x18a00000U,
    0xd4900000U,
    0x61580000U,
    0xcae40000U,
    0x5ff60000U,
    0xb750000U,
    0xaa208000U,
    0x40cf4000U,
    0x2f346000U,
    0xcc433000U,
    0x26f3800U,
    0x68b0b400U,
    0xdc8e8200U,
    0x155e4300U,
    0x38e2e080U,
    0x48f970c0U,
    0x40fbdba0U,
    0x34fcc590U,
    0xc6f5daa8U,
    0xd1fdc56cU,
    0x9a7b5a96U,
    0x4a3877fU,
    0x8299b942U,
    0x7c5af502U,
    0x2a62600dU,
    0x2ca63317U,
    0xc697bae1U,
    0x665bf7aeU,
    0x596ce3d4U,
    0x4d387212U,
    0xc3555965U,
    0xe5f2855bU,
    0x886f3927U,
    0x3b0b7d9U,
    0x110e836aU,
    0x6e9e40aeU,
    0xe242e13bU,
    0x636971f8U,
    0xe23d90bU,
    0x4ad8c7c0U,
    0x8423d94fU,
    0x21d8c67aU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0xd8000000U,
    0x2c000000U,
    0xa2000000U,
    0x65000000U,
    0xe4800000U,
    0xa8c00000U,
    0xd2e00000U,
    0x2b900000U,
    0x9d480000U,
    0xf0bc0000U,
    0xc6ba0000U,
    0x3da30000U,
    0x26218000U,
    0x737a4000U,
    0x97d9a000U,
    0x8f7c1000U,
    0xadd9d800U,
    0x267e3c00U,
    0xbb538200U,
    0x33a54100U,
    0xa1222380U,
    0x12f55140U,
    0x8b89fb60U,
    0x2d546cb0U,
    0x78a1f888U,
    0x32b86c94U,
    0xb3b3fa12U,
    0xe1376ee3U,
    0xf2e079abU,
    0xdb922c5eU,
    0xf5425a1dU,
    0x54a77ea6U,
    0x90aba0e0U,
    0xd6a31331U,
    0x5a25a29U,
    0x5a377fa4U,
    0x963a31bU,
    0xdedf13d6U,
    0xc9f85a09U,
    0x60047d74U,
    0x100a22a3U,
    0x3819518aU,
    0x7c1bfb6fU,
    0x7a1b6cf5U,
    0x491279c3U,
    0x468d2ffbU,
    0xcdd9d826U,
    0x367e3ee1U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xc8000000U,
    0x14000000U,
    0x82000000U,
    0xc1000000U,
    0x14800000U,
    0x8dc00000U,
    0x61a00000U,
    0xe3900000U,
    0x4a580000U,
    0xd1e40000U,
    0x5f7a0000U,
    0xfb10000U,
    0xcc8e8000U,
    0xb1c64000U,
    0x87af2000U,
    0x38831000U,
    0x43c27800U,
    0xdea2d400U,
    0xf00c8200U,
    0x58034300U,
    0xfc03a080U,
    0xc6105140U,
    0x8b19db20U,
    0xc1968450U,
    0x1b4f5808U,
    0x2d74c404U,
    0x96ba7ad2U,
    0x2416d477U,
    0xfa0e8026U,
    0x6d06419eU,
    0x1a8f2081U,
    0x12d31153U,
    0x223a7807U,
    0x39d6d55dU,
    0x73ae82eeU,
    0xca96437aU,
    0x1ad72263U,
    0x1637135cU,
    0xebc07bfdU,
    0x7aa7d7ddU,
    0xca0000aeU,
    0x150000daU,
    0xb6800373U,
    0x1cc00274U,
    0xbd2000a9U,
    0x7a500187U,
    0xa9f8039dU,
    0xf374020eU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0x58000000U,
    0x3c000000U,
    0xe2000000U,
    0x4d000000U,
    0xcd800000U,
    0xadc00000U,
    0x8ae00000U,
    0x66900000U,
    0x43580000U,
    0x6da40000U,
    0xf53a0000U,
    0x49630000U,
    0xfc4b8000U,
    0x552a4000U,
    0xd97ca000U,
    0x84573000U,
    0x39397800U,
    0x637d3c00U,
    0xf5498200U,
    0x16bd4300U,
    0x83b52080U,
    0xb22a7140U,
    0xddec5960U,
    0x4a074cf0U,
    0xb91dd988U,
    0x6b8e0e34U,
    0x6ecaf9b6U,
    0xb0637cf7U,
    0x77d721cbU,
    0x4bed732aU,
    0x611ddb65U,
    0x978e0d32U,
    0xaccaf80aU,
    0xad637d56U,
    0xe25723e4U,
    0xda2d710dU,
    0x9fdda95U,
    0xbc1e0d3aU,
    0x2212f8feU,
    0x6d077e40U,
    0x9d8d20c3U,
    0xf5de705eU,
    0xb6ee59b3U,
    0x84904e35U,
    0xe5458c9U,
    0xa0334d08U,
    0x58ffda57U,
    0xc3890f38U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x58000000U,
    0xd4000000U,
    0xee000000U,
    0xab000000U,
    0x1b800000U,
    0x4a400000U,
    0x3fa00000U,
    0xca500000U,
    0x7fa80000U,
    0xaa5c0000U,
    0xefba0000U,
    0xf2570000U,
    0x3ba08000U,
    0x1c4ac000U,
    0x90b5e000U,
    0x7c9b000U,
    0xdafca800U,
    0x386e4400U,
    0x10b5e200U,
    0x47c9b100U,
    0xbafca980U,
    0xa86e4640U,
    0x48b5e360U,
    0x93c9b250U,
    0x54fcaa38U,
    0x36e44ecU,
    0x5335e30eU,
    0xd989b379U,
    0x6b5caac6U,
    0xc93e47c5U,
    0x2c9de2f0U,
    0x73d5b1d0U,
    0x84e6a978U,
    0x3b69440cU,
    0x173d621eU,
    0x6f9f71a1U,
    0x14534b3aU,
    0x3ca0f413U,
    0xcdc1c975U,
    0x57f13540U,
    0x4e6a978U,
    0x7b69440cU,
    0x773d621eU,
    0xff9f71a1U,
    0x4c534b3aU,
    0xe8a0f413U,
    0x23c1c975U,
    0xfcf13540U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x58000000U,
    0x9c000000U,
    0xe2000000U,
    0x8b000000U,
    0x52800000U,
    0x56c00000U,
    0x4a600000U,
    0xb6d00000U,
    0x3a680000U,
    0xded40000U,
    0xee6a0000U,
    0xf8df0000U,
    0x1b798000U,
    0xc3434000U,
    0x9428e000U,
    0x8d691000U,
    0x3e523800U,
    0x4bb85c00U,
    0x6c28e200U,
    0x41691100U,
    0xa4523880U,
    0x4cb85c40U,
    0x84a8e360U,
    0xa91370U,
    0x5eb23b08U,
    0x27a85e6cU,
    0xa620e22aU,
    0x3e6d122bU,
    0xc0d03a21U,
    0xb7735cb7U,
    0x695b62c3U,
    0xdb255150U,
    0xa1eb5998U,
    0x1860f54U,
    0xd8583aaeU,
    0xdeb75e5dU,
    0x67b961c8U,
    0x863e5061U,
    0x2e70d8d7U,
    0x98de4e33U,
    0x2b6b5998U,
    0x8b460f54U,
    0x50383aaeU,
    0xf3675e5dU,
    0x575161c8U,
    0x922a5061U,
    0x687ad8d7U,
    0x5dd14e33U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x48000000U,
    0x44000000U,
    0xca000000U,
    0xd000000U,
    0x28800000U,
    0x51400000U,
    0xd8a00000U,
    0x59500000U,
    0xbcb80000U,
    0xc34c0000U,
    0x49ae0000U,
    0xe7c50000U,
    0x96e68000U,
    0xf87c4000U,
    0xea372000U,
    0x3d015000U,
    0xa0910800U,
    0xf5444400U,
    0xa2b72200U,
    0x1c415300U,
    0xd0310b80U,
    0x581446c0U,
    0x9c0f2320U,
    0x960d5210U,
    0x7b1f08a8U,
    0xe39146f4U,
    0xfac9a382U,
    0x66611355U,
    0xf5302bcaU,
    0x448c1791U,
    0xaf4ea970U,
    0xb7ac5458U,
    0x88cf88ecU,
    0x4774070eU,
    0x3ae00abU,
    0x2ac50179U,
    0x5e668064U,
    0x193c41eaU,
    0x7a972081U,
    0x20515358U,
    0xd62908ecU,
    0x3b08470eU,
    0xc39920abU,
    0xaac45179U,
    0x9e778864U,
    0xf93805eaU,
    0xca800281U,
    0x68400058U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0xa8000000U,
    0xac000000U,
    0xe6000000U,
    0x9000000U,
    0x38800000U,
    0xe2c00000U,
    0x50a00000U,
    0xead00000U,
    0x2cb80000U,
    0x84d40000U,
    0x99ba0000U,
    0xf2410000U,
    0xde6f8000U,
    0x7c7dc000U,
    0xdf7ea000U,
    0x38e2d000U,
    0x53a6f800U,
    0xdd4ac400U,
    0xcffea200U,
    0xb622d300U,
    0xc506f880U,
    0x2e9ac440U,
    0x73c6a0a0U,
    0x7c36d1b0U,
    0xea1cfb18U,
    0x3f0bc464U,
    0xb9912042U,
    0x665f123bU,
    0xfc785a5aU,
    0x1f7817cfU,
    0x18e058f0U,
    0x43bc1428U,
    0x75425a3cU,
    0x63f91706U,
    0x502fdb09U,
    0xcc11d599U,
    0x1604fbe1U,
    0x910fc555U,
    0x2c9321dfU,
    0xca12c8U,
    0x13adda3cU,
    0x3d44d706U,
    0xfff17b09U,
    0xe230599U,
    0xc11a03e1U,
    0x64910155U,
    0x9cd783dfU,
    0x4da9c1c8U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x48000000U,
    0x64000000U,
    0xea000000U,
    0xc5000000U,
    0x4f800000U,
    0x75c00000U,
    0x3b600000U,
    0xa1d00000U,
    0xd9680000U,
    0xe0cc0000U,
    0x4cee0000U,
    0x18050000U,
    0x5c028000U,
    0x26084000U,
    0x7b122000U,
    0x289db000U,
    0x9b5a9800U,
    0xeb33dc00U,
    0x2a722200U,
    0xc4db100U,
    0xedb29b80U,
    0x4e3fdcc0U,
    0x15fc2320U,
    0xd198b090U,
    0x82d81828U,
    0x4dfb9fd4U,
    0x6d80021eU,
    0x94c00147U,
    0x7ee000c5U,
    0x21100153U,
    0xe588037bU,
    0x50dc03c4U,
    0x44e60276U,
    0x9c1901f3U,
    0x8604832bU,
    0xab0142ccU,
    0x509ea032U,
    0xb740f245U,
    0x65223878U,
    0x56a2cb7U,
    0x86d49a76U,
    0xd7e6ddf3U,
    0x98a12bU,
    0x8f49f3ccU,
    0xa92ebbb2U,
    0xbb776e85U,
    0xe1cc3b58U,
    0x396f2c27U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x58000000U,
    0x54000000U,
    0x72000000U,
    0xcf000000U,
    0xf2800000U,
    0x99400000U,
    0xcfa00000U,
    0xfb500000U,
    0x98b80000U,
    0x7dd40000U,
    0xe3ea0000U,
    0x257f0000U,
    0xbe308000U,
    0x56894000U,
    0xc35d6000U,
    0xbcad1000U,
    0xe7c94800U,
    0xb0f06400U,
    0xa9e56200U,
    0xce791300U,
    0xd6a34880U,
    0x9ccf6640U,
    0x8075e360U,
    0x37a05250U,
    0xff462948U,
    0x92b6757cU,
    0x76d6a8aaU,
    0x3b6f3435U,
    0x2733c876U,
    0xf1162591U,
    0xfb908248U,
    0x66d94382U,
    0xa3656099U,
    0x53391354U,
    0x13034bf8U,
    0x6c9f64a4U,
    0xc04de3beU,
    0x483453d3U,
    0x538c2b71U,
    0x1ad975a6U,
    0x6d7e2819U,
    0x72227514U,
    0x509ca898U,
    0xee4036f4U,
    0xd93b4af6U,
    0xd80b66afU,
    0x9407e3dbU,
    0x521b5193U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xd8000000U,
    0xdc000000U,
    0x62000000U,
    0xdf000000U,
    0x2c800000U,
    0x15400000U,
    0xbe600000U,
    0x2b500000U,
    0x43680000U,
    0x38cc0000U,
    0xaa60000U,
    0x4bf70000U,
    0x438d8000U,
    0x41c34000U,
    0xe7326000U,
    0x5f2cd000U,
    0xb321c800U,
    0xe9219c00U,
    0x9a3a6200U,
    0xcb0d100U,
    0x1aefcb80U,
    0xea1a9dc0U,
    0xb11e160U,
    0xda849270U,
    0xdc502a08U,
    0x2bf50ebcU,
    0x738249d2U,
    0x79c9de25U,
    0x4b2b82f1U,
    0xe5344311U,
    0xb03fe15fU,
    0xa7af9206U,
    0x5073a83bU,
    0x9e5d4d8eU,
    0x24f3ab39U,
    0x171d4fcdU,
    0x1893a87dU,
    0x934d4c6bU,
    0xaf7baad6U,
    0x62c14d5dU,
    0x79bdabbbU,
    0xdd664c4eU,
    0x55d82859U,
    0xb1290cbdU,
    0x62c49f5U,
    0x8ea2df17U,
    0xb5e80264U,
    0x1e8c0108U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0x98000000U,
    0xa4000000U,
    0x5a000000U,
    0x33000000U,
    0xdf800000U,
    0x4ac00000U,
    0xa9600000U,
    0xa4d00000U,
    0x28680000U,
    0xd4440000U,
    0x732e0000U,
    0xc6f70000U,
    0x360a8000U,
    0x85044000U,
    0x1a80e000U,
    0xf045b000U,
    0x6928b800U,
    0x55e11c00U,
    0xd988e200U,
    0x57d1b100U,
    0x17eeba80U,
    0xe921cc0U,
    0x724c6060U,
    0x5e32f390U,
    0xe06cdbe8U,
    0x3847ec0cU,
    0x85223b1eU,
    0xa3e55ebbU,
    0xbc88034dU,
    0xdd54039fU,
    0x4fa6033fU,
    0x5ba3002aU,
    0xd9ac8061U,
    0xeea743d4U,
    0x5b2c626fU,
    0xbae2f0a2U,
    0x6804d87dU,
    0xdc03ef62U,
    0x6e0c3978U,
    0xc1125f81U,
    0xd08282e1U,
    0x6b504314U,
    0x8aa6e00fU,
    0xe126b232U,
    0x19e43915U,
    0x1f965faeU,
    0xaacc8206U,
    0x397742aaU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x8000000U,
    0x84000000U,
    0xea000000U,
    0x8f000000U,
    0x9c800000U,
    0x2ec00000U,
    0x86200000U,
    0x18d00000U,
    0xaf380000U,
    0x45540000U,
    0xa8620000U,
    0xf6350000U,
    0xc28000U,
    0x3204000U,
    0xfb4f6000U,
    0xc56dd000U,
    0x61b1b800U,
    0x589c9400U,
    0x24d76200U,
    0xf929d300U,
    0x5c4bb880U,
    0xded9740U,
    0xad6fe220U,
    0xd5b89110U,
    0x4a9c5b28U,
    0xf7c1057cU,
    0x3b1b852U,
    0x939c95abU,
    0x72576330U,
    0x88e9d21fU,
    0x4eebb86eU,
    0xbffd94feU,
    0x6e77e131U,
    0x73c9097U,
    0xd1465be5U,
    0x6a6004deU,
    0x2d313b06U,
    0x7e59d505U,
    0x26e28009U,
    0xbf04264U,
    0x7c7761b1U,
    0xd439d3d7U,
    0x2bd3bbc5U,
    0xa5a996ceU,
    0x5295e0aeU,
    0x5bc99339U,
    0xbda4d87bU,
    0xfe9045dfU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0x28000000U,
    0x2c000000U,
    0x82000000U,
    0x61000000U,
    0xbc800000U,
    0x5e400000U,
    0xbba00000U,
    0x88500000U,
    0x8ca80000U,
    0x43cc0000U,
    0xf9660000U,
    0xfda90000U,
    0xe7538000U,
    0x43214000U,
    0x980ba000U,
    0xa404d000U,
    0x3e1b6800U,
    0xbb1b2400U,
    0xd983a200U,
    0x4cd8d100U,
    0x6f56b80U,
    0xee6e25c0U,
    0xe63e22a0U,
    0x2a8c91b0U,
    0x89434988U,
    0x3eb444U,
    0xd5936852U,
    0x1ec724c9U,
    0xdfeda166U,
    0x26edd365U,
    0x3e68ebe0U,
    0x5e2a6506U,
    0x5e8001a3U,
    0xf4001e3U,
    0xcf2002d5U,
    0x8a10013eU,
    0x9d080139U,
    0x869c028eU,
    0x4b4e01b1U,
    0x8125014aU,
    0x19158223U,
    0x68984023U,
    0x48502275U,
    0x2cb9918eU,
    0xd3decb31U,
    0xa17af70aU,
    0xf9ad4943U,
    0x494bb533U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0xa8000000U,
    0x8c000000U,
    0xbe000000U,
    0x41000000U,
    0x31800000U,
    0x5c00000U,
    0x9ae00000U,
    0xe4d00000U,
    0x7b780000U,
    0x69140000U,
    0x7d9a0000U,
    0x9bd10000U,
    0xcbf38000U,
    0x7d52c000U,
    0xf2a1a000U,
    0x4de5f000U,
    0xd8429800U,
    0xd13c6c00U,
    0xa73ba200U,
    0x9a34f300U,
    0x8db11880U,
    0xfd6eac40U,
    0xcc1a00a0U,
    0x5e110130U,
    0x71138278U,
    0x8982c144U,
    0x21d9a266U,
    0xa8f1f227U,
    0x1bd89893U,
    0xbed6e97U,
    0x5d48232bU,
    0xe2a632b3U,
    0xe5f0bb1dU,
    0x545b5dbcU,
    0x6f209a62U,
    0xe6396d61U,
    0xabb22244U,
    0x8867311cU,
    0x679b3aa8U,
    0x28cd9cd6U,
    0x25633965U,
    0x18199cf8U,
    0xf4193804U,
    0xba189f46U,
    0x6312bad7U,
    0x669e5f8bU,
    0xf9491983U,
    0x10baae65U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x18000000U,
    0x94000000U,
    0xbe000000U,
    0xa7000000U,
    0x39800000U,
    0x6a400000U,
    0x6c600000U,
    0x55500000U,
    0x81e80000U,
    0x21040000U,
    0xda9a0000U,
    0x55d70000U,
    0xa5208000U,
    0xa6a4c000U,
    0x8dfe6000U,
    0x4b1cf000U,
    0x6397a800U,
    0x5b4fbc00U,
    0xeee46200U,
    0x348bf100U,
    0xad72a80U,
    0x38bb7e40U,
    0xfaf20260U,
    0xca930350U,
    0xdda8078U,
    0x9123c0dcU,
    0x88b6e086U,
    0x32fc33f9U,
    0xe693cb49U,
    0xb7d44fc9U,
    0x903b48e1U,
    0x82248c2dU,
    0xdf372a5bU,
    0x47ab7f42U,
    0xb77a0073U,
    0x2ec7030eU,
    0x4ea88145U,
    0x71f0c2f7U,
    0x490c6164U,
    0x668ff292U,
    0xf7cd2a23U,
    0x302c7f9eU,
    0x123280f5U,
    0xc727c0f7U,
    0xd3ace20cU,
    0x96b313eU,
    0x89d34b85U,
    0x77208fbfU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x8000000U,
    0x94000000U,
    0x36000000U,
    0xbd000000U,
    0x28800000U,
    0xa9c00000U,
    0x60a00000U,
    0x2cd00000U,
    0xc4380000U,
    0x4f1c0000U,
    0x6b820000U,
    0xc84d0000U,
    0x7e88000U,
    0x50b7c000U,
    0x34cd2000U,
    0x3837d000U,
    0x3d049800U,
    0xe88a0400U,
    0xc9cf2200U,
    0xb0bad300U,
    0x24cc1980U,
    0x502dc740U,
    0x791a0220U,
    0xd6810150U,
    0xe0d28158U,
    0xae26c1b4U,
    0x3007a082U,
    0x181d10f7U,
    0xfc19395aU,
    0x72161647U,
    0x830499b2U,
    0x18a049bU,
    0xb74f23acU,
    0x747ad3d2U,
    0x64ec1aafU,
    0x413dc56eU,
    0xeb820385U,
    0x84d0065U,
    0x67e88111U,
    0x80b7c1f3U,
    0x3ccd22f4U,
    0xac37d266U,
    0xb049a2dU,
    0x558a0599U,
    0xe14f22dfU,
    0x197ad222U,
    0x446c1aa3U,
    0x7cfdc668U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x8000000U,
    0xf4000000U,
    0xc6000000U,
    0xe9000000U,
    0xd6800000U,
    0xf2400000U,
    0x6200000U,
    0x4f500000U,
    0x86a80000U,
    0x8c1c0000U,
    0x1a1e0000U,
    0xb0d0000U,
    0x41888000U,
    0x31d9c000U,
    0x10e3a000U,
    0x94321000U,
    0xf05b6800U,
    0x613e1400U,
    0xe4dda200U,
    0xd46f1100U,
    0x997be980U,
    0x71fbd6c0U,
    0xf6a00220U,
    0xa41002d0U,
    0x3e080298U,
    0x750c0164U,
    0x5496037aU,
    0xd5410019U,
    0xcdbe8300U,
    0xad88c299U,
    0x9bd52140U,
    0x13f6d1f9U,
    0x21b84870U,
    0x799c511U,
    0x98d3691cU,
    0xa67217f7U,
    0xd66ba3abU,
    0xfe7e11bcU,
    0xda6d6b67U,
    0xa46f1453U,
    0xb16b22e8U,
    0x55ebd075U,
    0x88b8c866U,
    0xb10c06eeU,
    0xda86c92bU,
    0xa85105e5U,
    0x6d264807U,
    0xbed4c77aU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0xc8000000U,
    0xfc000000U,
    0xc6000000U,
    0xa5000000U,
    0x7a800000U,
    0xc6400000U,
    0x8ee00000U,
    0x2f500000U,
    0x9a680000U,
    0xe0140000U,
    0x10060000U,
    0x981d0000U,
    0x44118000U,
    0xf201c000U,
    0x9f0b2000U,
    0x19979000U,
    0x19cc6800U,
    0x323e8c00U,
    0x67ed2200U,
    0x3bda9100U,
    0x5535e880U,
    0x6a6b4dc0U,
    0x68000120U,
    0xcc0002f0U,
    0x2e000398U,
    0x29000354U,
    0x748002caU,
    0x9f4000e9U,
    0x32600123U,
    0x4c100229U,
    0x6e080383U,
    0x9040099U,
    0x48e027bU,
    0x575902ddU,
    0xce7f8379U,
    0x8a08c048U,
    0xcb1ca23cU,
    0x738b5206U,
    0xc2d6c977U,
    0xd9a8df00U,
    0xe12a69e3U,
    0x10738d89U,
    0x2b14a3b3U,
    0x638f51a1U,
    0x5ad8cb9fU,
    0x9db1ddefU,
    0x1335ebd4U,
    0x8f6b4d69U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x58000000U,
    0xac000000U,
    0x6e000000U,
    0x41000000U,
    0x23800000U,
    0x46c00000U,
    0x22200000U,
    0x33d00000U,
    0x23a80000U,
    0x4a040000U,
    0x8b1e0000U,
    0xe89f0000U,
    0xe438000U,
    0x9c66c000U,
    0xf7ada000U,
    0x7811b000U,
    0x5c1cf800U,
    0x96023400U,
    0x5d13a200U,
    0x159eb100U,
    0xabd77a80U,
    0x6fb0f6c0U,
    0x34080360U,
    0x221403b0U,
    0x3f160338U,
    0x8a8b03c4U,
    0x915581eeU,
    0xa6edc3abU,
    0x3ef82130U,
    0x72fc71cbU,
    0xce4d800U,
    0xa5fe4433U,
    0x727778a4U,
    0xf6a0f66dU,
    0xfb800037U,
    0xaac00099U,
    0xec200212U,
    0xc2d00232U,
    0x58280191U,
    0xa0c4039dU,
    0xc73e039cU,
    0x9a4f01a9U,
    0xe6b83d9U,
    0x90a2c232U,
    0x5e93a1a2U,
    0xa35eb139U,
    0x71f778f1U,
    0x4060f51eU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x48000000U,
    0xc000000U,
    0xca000000U,
    0xfb000000U,
    0xb1800000U,
    0x5ec00000U,
    0xc9600000U,
    0xc1d00000U,
    0xaee80000U,
    0x46040000U,
    0x71020000U,
    0xea9d0000U,
    0xbf4b8000U,
    0xdfb04000U,
    0x4aee000U,
    0xa5377000U,
    0x13f8b800U,
    0x86891c00U,
    0xc54ce200U,
    0x9cba7100U,
    0xa13b3a80U,
    0x75ed5d40U,
    0xe7880320U,
    0xc7d40130U,
    0x7fea01a8U,
    0xfc9902acU,
    0x864983e6U,
    0x392d414bU,
    0x71e5600dU,
    0x818730ebU,
    0xa6d6587dU,
    0x7d7e6c63U,
    0x1fd458e1U,
    0xdbe36dadU,
    0xca9fd8c6U,
    0xaf532e26U,
    0x37b13876U,
    0x58a45d4eU,
    0x2729803aU,
    0xe4fd42f8U,
    0xfd0d6149U,
    0x60833301U,
    0xe4545920U,
    0x3e236e6dU,
    0x127fdafbU,
    0x60432ce5U,
    0x18393967U,
    0xd3705eabU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xa8000000U,
    0xfc000000U,
    0xaa000000U,
    0x63000000U,
    0x18800000U,
    0xbb400000U,
    0xa6e00000U,
    0x7a500000U,
    0x91680000U,
    0xfb940000U,
    0xe3d60000U,
    0x3db10000U,
    0x2cb18000U,
    0x43384000U,
    0x96efe000U,
    0xb2545000U,
    0xbd7c0800U,
    0x9974c00U,
    0xd4d1e200U,
    0x73315100U,
    0x5efb8880U,
    0x9e4e0fc0U,
    0x4f6780a0U,
    0x3e8942f0U,
    0x9a5e6228U,
    0x16c124cU,
    0x8393eac2U,
    0x47c31d1dU,
    0xc3adea33U,
    0x19a61d65U,
    0x92aa68a7U,
    0x563f5fc3U,
    0xb77c0934U,
    0xda974c2fU,
    0x4451e07fU,
    0xc471525eU,
    0xfa1b893eU,
    0x7b1e0d6eU,
    0x6c8f82b6U,
    0x1d5d4152U,
    0xc7e8639cU,
    0xfdcd11a3U,
    0x98aa6a1dU,
    0x853f5db3U,
    0x27fc0925U,
    0x6dd74e47U,
    0xe0b1e2d3U,
    0x2121528cU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0xd8000000U,
    0x24000000U,
    0xf2000000U,
    0x2b000000U,
    0x7f800000U,
    0xb2c00000U,
    0x24200000U,
    0x3d00000U,
    0xbca80000U,
    0xb0840000U,
    0xdf520000U,
    0xcdef0000U,
    0x506e8000U,
    0xf1bd4000U,
    0xfe0de000U,
    0xb517b000U,
    0xfa863800U,
    0xe04ef400U,
    0xe877e200U,
    0xe5bcb100U,
    0xa41aba80U,
    0xb20cb740U,
    0x8b1c8160U,
    0xaf824190U,
    0x6acb6148U,
    0x2ef3ecU,
    0xf1d9da9eU,
    0x97b6465bU,
    0xcf1f5958U,
    0x6d8f05a3U,
    0xe9c0bb8cU,
    0x53b7b649U,
    0x4d08000dU,
    0x4e9401e8U,
    0x6a5a019bU,
    0x377b0278U,
    0xb03481d3U,
    0x19c64314U,
    0x1bb9600dU,
    0x1111f22fU,
    0x489f5bc5U,
    0x6b4f0544U,
    0x47e0ba65U,
    0x8f67b5b3U,
    0xa42001c3U,
    0x43d0015bU,
    0x1ca8031fU,
    0x6084033dU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xd8000000U,
    0x4c000000U,
    0x22000000U,
    0xdf000000U,
    0x26800000U,
    0x8ec00000U,
    0xb0e00000U,
    0xefd00000U,
    0x73780000U,
    0xf4940000U,
    0x19d20000U,
    0x227f0000U,
    0x5f138000U,
    0xe6874000U,
    0xaed66000U,
    0xe0ea9000U,
    0x37c13800U,
    0x3f634c00U,
    0xd69c6200U,
    0xc6d19300U,
    0x4f8b880U,
    0xd1cf0d40U,
    0x566b8160U,
    0x41134230U,
    0x93846008U,
    0xc355923cU,
    0x26b2bbfaU,
    0xf4f40f0bU,
    0x99d2004bU,
    0xe27f03c3U,
    0x7f1381d7U,
    0xb68740a9U,
    0x76d661c4U,
    0xacea9206U,
    0x15c13839U,
    0xe0634f34U,
    0xf01c6046U,
    0x48119231U,
    0xb418bbe0U,
    0x3e1f0d98U,
    0x2513814cU,
    0xb587417aU,
    0x8a5660a3U,
    0xe12a910fU,
    0x79a13885U,
    0x12734e8eU,
    0x370460adU,
    0x295900aU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0x68000000U,
    0xec000000U,
    0xfa000000U,
    0xfb000000U,
    0x61800000U,
    0x89c00000U,
    0xd9200000U,
    0x34d00000U,
    0x1a80000U,
    0x33840000U,
    0x7ece0000U,
    0xd2b90000U,
    0x1e0b8000U,
    0x3d1a4000U,
    0x9896e000U,
    0xa7557000U,
    0x4f635800U,
    0x893a9400U,
    0x8cd0e200U,
    0x15b87100U,
    0x7d8eda80U,
    0xfbddd4c0U,
    0x5e2383a0U,
    0xf75e42b0U,
    0xf778e168U,
    0x9d3c732cU,
    0xc2c0d826U,
    0x90a4d597U,
    0xf108028cU,
    0x1294023fU,
    0x94460300U,
    0xf2ed0229U,
    0x92ed833fU,
    0x42e74029U,
    0xbaf36396U,
    0xee633d6U,
    0x70f63a1fU,
    0x9de1a7f0U,
    0xfd635821U,
    0xee3a9703U,
    0xdf50e0d7U,
    0xbb7873c5U,
    0x572eda10U,
    0x51cdd5f1U,
    0x1d2b837bU,
    0x82ca4223U,
    0x30bee0a7U,
    0xc111720dU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0x68000000U,
    0x6c000000U,
    0x62000000U,
    0x69000000U,
    0xc2800000U,
    0xf2c00000U,
    0x5e600000U,
    0x7dd00000U,
    0x3bf80000U,
    0x269c0000U,
    0xccc60000U,
    0x5d610000U,
    0x4a438000U,
    0x1b94000U,
    0x8f3ca000U,
    0x4fe21000U,
    0x809a7800U,
    0xcbd41c00U,
    0x64e2a200U,
    0x7b0f1300U,
    0x387f980U,
    0x3c405fc0U,
    0x3ea383a0U,
    0x22a942b0U,
    0xe8a4a008U,
    0x8dae1264U,
    0xdd247aaaU,
    0x2ee91e7bU,
    0xde0722f1U,
    0x33075353U,
    0x3f80da65U,
    0xa6470c91U,
    0xa3a359eaU,
    0x862e4e32U,
    0x7d67fbceU,
    0x1a505df8U,
    0xf9bb8025U,
    0x7b2542f1U,
    0x29faa3faU,
    0xe783110aU,
    0x259f862U,
    0x3dad5f96U,
    0x153e00c4U,
    0xd2fd0233U,
    0x2405835cU,
    0x5e184006U,
    0xf31f2295U,
    0x5f8b5250U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0xb8000000U,
    0xac000000U,
    0x46000000U,
    0x6b000000U,
    0xdf800000U,
    0xbec00000U,
    0x1fa00000U,
    0x37d00000U,
    0x99380000U,
    0x87840000U,
    0x82c60000U,
    0xb1a30000U,
    0xb8db8000U,
    0x14a7c000U,
    0xb85ce000U,
    0x6ff53000U,
    0x8c29b800U,
    0x5f1b8400U,
    0x582e200U,
    0x8bc23300U,
    0xf72c3a80U,
    0xa88b47c0U,
    0xff5b80e0U,
    0x3667c1b0U,
    0xf9fce398U,
    0x6f25326cU,
    0x3491b99eU,
    0xa15f844bU,
    0x164e366U,
    0xd8713273U,
    0x16efbb9aU,
    0xb2b88725U,
    0x43596295U,
    0x5865f3a9U,
    0xd6f0db1fU,
    0x12be7420U,
    0xb3523b74U,
    0xe06c4432U,
    0x7ae60135U,
    0x54b3009dU,
    0xd843838dU,
    0x3ff3c205U,
    0xc422e061U,
    0x4b1232dbU,
    0xef94390aU,
    0xa6cf47edU,
    0x43bd81d1U,
    0xc9d4c143U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0xa8000000U,
    0xc000000U,
    0x22000000U,
    0x27000000U,
    0xf8800000U,
    0x65c00000U,
    0x4c600000U,
    0xdd500000U,
    0x49b80000U,
    0x81240000U,
    0x8cfe0000U,
    0x6c890000U,
    0x23db8000U,
    0xe5734000U,
    0x90d66000U,
    0x5bf7d000U,
    0x5c099800U,
    0x5a183c00U,
    0x930de200U,
    0xce849300U,
    0xc4dffa80U,
    0xbdefeec0U,
    0x450478a0U,
    0xbf9caf30U,
    0x8d521808U,
    0x31ab7d5cU,
    0x353b81c2U,
    0xbae34067U,
    0xcd8e6199U,
    0xd243d119U,
    0x7d2f98ecU,
    0x66e53f3fU,
    0x57906268U,
    0xe15ad0ccU,
    0x3ac19faU,
    0xa227db3U,
    0xe66000c7U,
    0x8650029cU,
    0x9b3800e2U,
    0x7fe40317U,
    0xb21e03f1U,
    0xff1901d5U,
    0xfc838116U,
    0xfbc7428cU,
    0xe17062afU,
    0xecad250U,
    0xf6f41918U,
    0x8f967ea4U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0x78000000U,
    0x7c000000U,
    0xe2000000U,
    0x11000000U,
    0xb800000U,
    0x7ec00000U,
    0xf9600000U,
    0x25500000U,
    0xbfa80000U,
    0x402c0000U,
    0xa8e20000U,
    0x968b0000U,
    0x9f418000U,
    0xa2b54000U,
    0xe1a8a000U,
    0x1325b000U,
    0x827a3800U,
    0x4bde3c00U,
    0x1ce92200U,
    0xe090f100U,
    0x88529b80U,
    0xf63b8f40U,
    0x1ff31be0U,
    0xf21eccf0U,
    0xc913ba08U,
    0x47877e04U,
    0x34c3804eU,
    0xec6e424bU,
    0xb0c1220dU,
    0x327cf261U,
    0xa3d098b8U,
    0xf8e08d4fU,
    0xae9a98e0U,
    0x3478e70U,
    0x90b91ac8U,
    0x88b9cca4U,
    0x64b0385eU,
    0x1eb93c33U,
    0xa3aaa3c1U,
    0x323eb30bU,
    0x21f3baedU,
    0x31177c11U,
    0x7b8b8270U,
    0x36d241ebU,
    0x2d6b20beU,
    0xc34bf243U,
    0x30bb1909U,
    0xb8a2cfafU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x18000000U,
    0x3c000000U,
    0x3e000000U,
    0x99000000U,
    0xd3800000U,
    0x13c00000U,
    0x3ee00000U,
    0x58500000U,
    0x51280000U,
    0xb8ac0000U,
    0x60660000U,
    0x22070000U,
    0xf70d8000U,
    0x4286c000U,
    0x1449a000U,
    0x37393000U,
    0xfda11800U,
    0x7df14c00U,
    0xb0c42200U,
    0x267ff100U,
    0x3708bb80U,
    0xe2987fc0U,
    0x44d3a60U,
    0xdf22bdf0U,
    0xd9aa9978U,
    0x7fe08ca4U,
    0x17c802aeU,
    0x6cfc027fU,
    0xf74e03e3U,
    0xcfab0035U,
    0x62eb8292U,
    0xd641c039U,
    0x302422ccU,
    0xa72ff3c2U,
    0x55a0b92dU,
    0xb9f47ec6U,
    0x42cb3a6fU,
    0x9975bed7U,
    0x418f196fU,
    0x1cca4c57U,
    0xb067a12fU,
    0x6a0233f7U,
    0xc3029bbfU,
    0xa88c8effU,
    0x974e00a3U,
    0x7fab0115U,
    0x9aeb8242U,
    0x1a41c391U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x98000000U,
    0x64000000U,
    0x2e000000U,
    0xab000000U,
    0xfc800000U,
    0xe6c00000U,
    0xf8a00000U,
    0x26500000U,
    0xf4680000U,
    0x2bec0000U,
    0x8daa0000U,
    0x39d70000U,
    0xf22a8000U,
    0xef88c000U,
    0x4e50e000U,
    0xd879d000U,
    0x19e2f800U,
    0xfcbb2400U,
    0xd85a6200U,
    0x27611100U,
    0x237a1b80U,
    0xdd7ef640U,
    0xe7a9860U,
    0x6e13490U,
    0x5620f938U,
    0x618025ecU,
    0x955ae012U,
    0x2cfed14bU,
    0x3207abaU,
    0x4e1fe5e5U,
    0x7b0080fbU,
    0x849fc108U,
    0x12da611eU,
    0x4ea11309U,
    0xe95a1a69U,
    0x26eef583U,
    0x66329984U,
    0x899d341cU,
    0xf942f83aU,
    0xfeeb24dfU,
    0xe23263a4U,
    0x378d12ecU,
    0xca501892U,
    0x6669f58bU,
    0x2af01b1aU,
    0x6439f555U,
    0x10981833U,
    0x74d5f4ccU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0xa8000000U,
    0x6c000000U,
    0xf2000000U,
    0xdf000000U,
    0x42800000U,
    0x15c00000U,
    0x57200000U,
    0x63500000U,
    0x4e80000U,
    0xed640000U,
    0x5c2a0000U,
    0x57d90000U,
    0xf0238000U,
    0x45dd4000U,
    0x1f396000U,
    0x3f4a9000U,
    0xcef09800U,
    0xf6761400U,
    0x80bae200U,
    0x6f07d100U,
    0x3a81fa80U,
    0x71c88640U,
    0x592878a0U,
    0xe65cc4b0U,
    0xf57a9b48U,
    0x483f143cU,
    0xc1d1632aU,
    0x212e92a7U,
    0x825a9ab4U,
    0xfb6f1501U,
    0xcd396271U,
    0x304a902eU,
    0x8470986eU,
    0x1fb615b9U,
    0x8d9ae325U,
    0xbf57d3c8U,
    0x8ee9f90bU,
    0x566c86a6U,
    0x10a27832U,
    0xc715c4c3U,
    0x56911adaU,
    0x83d654b8U,
    0x862a0354U,
    0xa4d902e6U,
    0xe0a383e5U,
    0x1f1d415fU,
    0xa29961b7U,
    0x25da92bbU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x48000000U,
    0x4c000000U,
    0xd6000000U,
    0x63000000U,
    0x9f800000U,
    0xbf400000U,
    0xb6a00000U,
    0x76d00000U,
    0x2ae80000U,
    0xffe40000U,
    0xb3720000U,
    0x2ba50000U,
    0x764f8000U,
    0x6425c000U,
    0xc102e000U,
    0x3e8bb000U,
    0x31d5c800U,
    0x3f700400U,
    0x1da56200U,
    0xe54a7100U,
    0xe3a52a80U,
    0x7a5eb540U,
    0x123f2b20U,
    0xf21fb430U,
    0xe902abd8U,
    0xc29f75ccU,
    0x5fcfcbdeU,
    0x871078dU,
    0x5038e022U,
    0xa31ab127U,
    0x7f804a2dU,
    0x4f54c4feU,
    0xaeba0211U,
    0x72d10234U,
    0xb0f5800eU,
    0x4af4c185U,
    0x4ff760e6U,
    0xb7f71bdU,
    0x7fa2aafaU,
    0xa44f766bU,
    0x9d27c8b3U,
    0xeb950653U,
    0x7d4ae103U,
    0xa7bfb14bU,
    0x404fc92fU,
    0xf7310645U,
    0x4698e12aU,
    0x85cab063U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x98000000U,
    0x64000000U,
    0x6000000U,
    0x65000000U,
    0x55800000U,
    0xb6400000U,
    0x17600000U,
    0x15d00000U,
    0x7c280000U,
    0x633c0000U,
    0x71be0000U,
    0x91670000U,
    0x30de8000U,
    0x49b1c000U,
    0xe5692000U,
    0xfecbf000U,
    0xe0a12800U,
    0x4af2bc00U,
    0x4f9fa200U,
    0xc1463100U,
    0xd9f60980U,
    0xc71e4cc0U,
    0x16800860U,
    0xd6d54c90U,
    0x5ca88b98U,
    0x48ef8c54U,
    0x1c8928b6U,
    0x69cebe89U,
    0x5e21a1a5U,
    0x60213193U,
    0x71288b5eU,
    0xeaaf8f91U,
    0xf5e928b5U,
    0x4d1ebdcfU,
    0xe989a3a0U,
    0xb45d3270U,
    0x44768948U,
    0xbd588fecU,
    0xfbffab52U,
    0xc4037e97U,
    0x16168090U,
    0xad1dc09cU,
    0x999f209eU,
    0x4c40f171U,
    0x1069aa65U,
    0x23587e77U,
    0xfaf60044U,
    0x978b026eU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x68000000U,
    0x5c000000U,
    0xaa000000U,
    0x5d000000U,
    0x91800000U,
    0x34400000U,
    0x4a00000U,
    0xf7d00000U,
    0x19e80000U,
    0xf4fc0000U,
    0x3d7a0000U,
    0x4d390000U,
    0xc79d8000U,
    0xf35e4000U,
    0xd0232000U,
    0x7613f000U,
    0xb716e800U,
    0xac8bc400U,
    0xb5d6a200U,
    0x58f1b100U,
    0xaf6fc980U,
    0x44313440U,
    0xb015cba0U,
    0x58083470U,
    0x44084928U,
    0x8e167434U,
    0xc30b6866U,
    0x3a9584e1U,
    0x52d5809aU,
    0xfc72429bU,
    0x56b120a9U,
    0xded6f346U,
    0x1e7169a9U,
    0x27ac863eU,
    0x9d480185U,
    0x432c0364U,
    0x44920006U,
    0xa9c50071U,
    0x92e780e2U,
    0xe2674137U,
    0xbdbea0b3U,
    0xd84db0e5U,
    0xf6b5ca8cU,
    0xeed83452U,
    0x6604887U,
    0x3aa767bU,
    0x3516b01U,
    0xe83c87b2U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xc8000000U,
    0x64000000U,
    0x9a000000U,
    0xfd000000U,
    0xa6800000U,
    0xecc00000U,
    0x86200000U,
    0x8500000U,
    0xabe80000U,
    0xcef40000U,
    0x74660000U,
    0x8ea50000U,
    0xfb8e8000U,
    0xfa584000U,
    0x82f02000U,
    0x1a7c1000U,
    0xf1bcf800U,
    0xfc187400U,
    0x8616a200U,
    0xeb105100U,
    0x558ad880U,
    0xe55167c0U,
    0x556cd920U,
    0x5e346490U,
    0x344258e8U,
    0x4dfc2534U,
    0xf5fa7b3aU,
    0x19e434e3U,
    0xc7e881d0U,
    0x50fd4085U,
    0xc37ea1fdU,
    0xed24512cU,
    0x1dccd81bU,
    0x6ea467ccU,
    0x6b8a59abU,
    0xe25825b4U,
    0xdef47b77U,
    0x2c75343aU,
    0xf2a000eeU,
    0x3d900269U,
    0x3148014bU,
    0x27640089U,
    0x372e0176U,
    0xc0c10008U,
    0x38208384U,
    0x4f5942c2U,
    0x70a07fU,
    0x4cb5523eU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x88000000U,
    0x74000000U,
    0x6e000000U,
    0x2f000000U,
    0x7f800000U,
    0x4c400000U,
    0xfea00000U,
    0xc9d00000U,
    0x23e80000U,
    0x8ffc0000U,
    0x3dea0000U,
    0xf8e50000U,
    0xde688000U,
    0x3ebac000U,
    0xe9dc6000U,
    0xd3ebb000U,
    0x97f02800U,
    0xc1ed8400U,
    0xe2fce200U,
    0x9f7d7100U,
    0x6e2e4980U,
    0xda1f3640U,
    0x610e4820U,
    0xa08f34d0U,
    0x2bc64a38U,
    0x4ee337fcU,
    0x2d64485eU,
    0xab2a37a1U,
    0xfc8ecbe2U,
    0x81c9f50bU,
    0x77f02969U,
    0x11ed8522U,
    0xafce2ebU,
    0x7b7d73d9U,
    0x882e4beaU,
    0x811f37ffU,
    0x708e4843U,
    0xc3cf34e9U,
    0xaae64962U,
    0xcb7337cbU,
    0xf02c4889U,
    0xed063492U,
    0xe28cc8a3U,
    0xf6d0f70dU,
    0x9472a910U,
    0xd7b2468cU,
    0x3d4802f6U,
    0x962c0345U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0xe8000000U,
    0xfc000000U,
    0x12000000U,
    0x67000000U,
    0xf4800000U,
    0xeac00000U,
    0xbe600000U,
    0x40500000U,
    0xf9b80000U,
    0x89bc0000U,
    0x31ba0000U,
    0x45b90000U,
    0x3ba98000U,
    0xa6a94000U,
    0xc9386000U,
    0xc5685000U,
    0xf6c77800U,
    0xfc637c00U,
    0xaf49e200U,
    0x612d1300U,
    0x997d1980U,
    0x14ce2e40U,
    0xe37d19a0U,
    0x4fce2cf0U,
    0x65fd1bc8U,
    0x520e2cdcU,
    0xc71d1bf2U,
    0x49e2d1bU,
    0x92c51a91U,
    0xaa722e2dU,
    0xae47185cU,
    0x8cb72ea1U,
    0x1a349985U,
    0x2ff76c90U,
    0x111d7a4bU,
    0xc58a7d3aU,
    0x1f586147U,
    0xb93851a3U,
    0x7d7f7b56U,
    0x82df7cddU,
    0x8273e294U,
    0x325410fdU,
    0xeb499b7U,
    0x95376c6bU,
    0x277d7a8aU,
    0xe9da7fafU,
    0x1ce0634fU,
    0xab84500cU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0x38000000U,
    0x7c000000U,
    0xb2000000U,
    0xaf000000U,
    0xb8800000U,
    0x54c00000U,
    0x4a600000U,
    0x8a500000U,
    0x77b80000U,
    0xc1ac0000U,
    0x80b60000U,
    0x35330000U,
    0xbe658000U,
    0x5c5b4000U,
    0x6a76000U,
    0xac3bd000U,
    0x87f7b800U,
    0x5d017c00U,
    0x379ae200U,
    0xfc5c9300U,
    0xd6bedb80U,
    0x8435afc0U,
    0x33e6dae0U,
    0xab09aef0U,
    0x5688db48U,
    0x59c6ae7cU,
    0x95e35b82U,
    0x2202ec63U,
    0xa717ba81U,
    0x1c917da5U,
    0x6ac2e094U,
    0x6f6091d9U,
    0xe1d0d861U,
    0x29faaf22U,
    0x700d58c4U,
    0xf80ded41U,
    0x9c1c3b45U,
    0x42053c64U,
    0x970e0191U,
    0xc49f019dU,
    0xe6d38160U,
    0xe5684147U,
    0x32c2e0b0U,
    0x23609168U,
    0x8bd0dbecU,
    0xafaaefaU,
    0x428d5a37U,
    0x7fcded4fU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xd0000000U,
    0xc8000000U,
    0x14000000U,
    0x8e000000U,
    0xcf000000U,
    0x7d800000U,
    0x5bc00000U,
    0x53200000U,
    0x49500000U,
    0x85780000U,
    0xb9ac0000U,
    0x93120000U,
    0x279d0000U,
    0xfac48000U,
    0x31b2c000U,
    0xa7126000U,
    0x9984f000U,
    0x2ddf5800U,
    0x90263400U,
    0x66c4e200U,
    0x8bab3300U,
    0xd609bb80U,
    0x33100740U,
    0xd789d920U,
    0x2c9f750U,
    0x3db203b8U,
    0xf50d037cU,
    0xcc9c8156U,
    0x114ec27fU,
    0x79786054U,
    0x13b5f249U,
    0xca09d9dbU,
    0x4909f4b5U,
    0x469200f6U,
    0x785d03d4U,
    0xfe48332U,
    0x73e2c045U,
    0x19ea61e5U,
    0xa0e8f04eU,
    0x1e6d5b28U,
    0x6a2b3524U,
    0x37d8631aU,
    0x1125f2e1U,
    0x3451dbbfU,
    0x7df5f68fU,
    0x16f80147U,
    0x3d6c01d3U,
    0x95b201c1U,
    0xf10d01d4U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0xa8000000U,
    0x64000000U,
    0x96000000U,
    0x23000000U,
    0xf800000U,
    0x50400000U,
    0xc8600000U,
    0xcd00000U,
    0x9fb80000U,
    0xa0fc0000U,
    0x550e0000U,
    0xdc810000U,
    0x97c68000U,
    0xcc21c000U,
    0xfaade000U,
    0xd477b000U,
    0xa6d8a800U,
    0x86ae9c00U,
    0x4e656200U,
    0x17d77300U,
    0xc33c980U,
    0x9ab8ecc0U,
    0xe47028a0U,
    0xede5e90U,
    0xe2b601d8U,
    0xd87d034cU,
    0x34c8831eU,
    0x3a0c011U,
    0xcaeb63d9U,
    0x2c16712fU,
    0x2154818U,
    0x7d092f4eU,
    0x7885c84bU,
    0x61c5ed00U,
    0xdf38a880U,
    0x5d3e9e40U,
    0xe03d61e0U,
    0xf8bb70f0U,
    0xa965c8e8U,
    0xde55ef84U,
    0xe760aa4aU,
    0x91529f63U,
    0x4aeb6386U,
    0xec16711fU,
    0x62154a72U,
    0x4d092cdfU,
    0xd085c870U,
    0x5c5ee8aU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0x58000000U,
    0x4000000U,
    0x82000000U,
    0x5000000U,
    0x5800000U,
    0x52400000U,
    0x42e00000U,
    0xcad00000U,
    0x71280000U,
    0xe8740000U,
    0x340a0000U,
    0xa070000U,
    0x79168000U,
    0x13894000U,
    0x8d4b2000U,
    0x46707000U,
    0x1f0aa800U,
    0x64804c00U,
    0x65d7a200U,
    0xddbe3100U,
    0xd1b70880U,
    0x6fa97e40U,
    0x5cbe2b60U,
    0x163a0d10U,
    0x18e00088U,
    0x8bd00354U,
    0xd6a801f6U,
    0x2f340319U,
    0x2b6a0263U,
    0x9697032fU,
    0xc8de83daU,
    0x342d42fcU,
    0xcde921cdU,
    0xf643724aU,
    0x10f6292bU,
    0xb7de0cc4U,
    0xe0a20121U,
    0x60330390U,
    0x77fc81c8U,
    0x475e43b4U,
    0x5f75a326U,
    0xbc8d3231U,
    0x21cb8b07U,
    0x7fb73d21U,
    0x44ab8a8fU,
    0x32273ef5U,
    0xae38939U,
    0xd6c33c8cU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0x98000000U,
    0x34000000U,
    0x22000000U,
    0x2b000000U,
    0xbd800000U,
    0x79400000U,
    0x2fe00000U,
    0x57d00000U,
    0xaf280000U,
    0x1b640000U,
    0xf81e0000U,
    0xe40f0000U,
    0x2a148000U,
    0xb7114000U,
    0x3390e000U,
    0x44557000U,
    0x9b6e0800U,
    0xb8072c00U,
    0x441a6200U,
    0x1a0b3100U,
    0x2f0a6a80U,
    0x7931cc0U,
    0x664c8860U,
    0xb07d6dd0U,
    0x5800008U,
    0x3d40016cU,
    0x35e00216U,
    0x78d001f5U,
    0xa8a80257U,
    0x7d240123U,
    0x487e00c2U,
    0xe19f0124U,
    0x175c81a9U,
    0x82e5432aU,
    0x4b46e2abU,
    0xecee73eeU,
    0xe64c8992U,
    0xf07d6eecU,
    0xa5800325U,
    0xd4002acU,
    0xade003f6U,
    0x4cd00365U,
    0x8aa8007fU,
    0x5624033fU,
    0xf5fe026cU,
    0x98df00e5U,
    0x38bc834cU,
    0xd5354166U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x68000000U,
    0x24000000U,
    0xaa000000U,
    0x93000000U,
    0xf4800000U,
    0xc4c00000U,
    0x8da00000U,
    0xb7500000U,
    0xcff80000U,
    0x51240000U,
    0x37920000U,
    0x98410000U,
    0xcd6a8000U,
    0x40f24000U,
    0x83b3e000U,
    0x4e535000U,
    0x869b800U,
    0xb16e4400U,
    0x16eb6200U,
    0x86b01300U,
    0x1fc8da80U,
    0x4d2b5740U,
    0x11833ba0U,
    0x255c0790U,
    0xd0f88228U,
    0x8bb3420cU,
    0x1a5960f2U,
    0x1a6110c3U,
    0x6e7a593eU,
    0x6c6d1401U,
    0x7b7ad845U,
    0xf5fa571bU,
    0xca31b9baU,
    0x971a47efU,
    0x4e81631cU,
    0x9fd5124aU,
    0x8d305867U,
    0xb1981780U,
    0xf55a5910U,
    0xb8fd1768U,
    0xafa2daacU,
    0xb04e55e2U,
    0x897bb9abU,
    0x9aef4692U,
    0xa8a1e263U,
    0xf6d2532eU,
    0x42a339e9U,
    0x5cc05a9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x28000000U,
    0xcc000000U,
    0x4a000000U,
    0x81000000U,
    0x61800000U,
    0x4a400000U,
    0xdee00000U,
    0x2ed00000U,
    0x83380000U,
    0x69740000U,
    0x678a0000U,
    0x41590000U,
    0x1e7f8000U,
    0x151e4000U,
    0x3f94a000U,
    0x5558f000U,
    0x80722800U,
    0x2a12cc00U,
    0x71012200U,
    0x798fb300U,
    0xbe410880U,
    0x70f07c40U,
    0x29d5aaa0U,
    0x29a88c30U,
    0xc3a783a8U,
    0x92ba4144U,
    0xfb26a1a6U,
    0x6d75f259U,
    0xd187aaf3U,
    0xb2558d8fU,
    0x1aea00a2U,
    0xb8c90188U,
    0x602783e3U,
    0x45fa40d7U,
    0x646a1aeU,
    0x54e5f27aU,
    0x8fdfaa44U,
    0xf2b18ff1U,
    0xb380060U,
    0x757403d0U,
    0x258a02b8U,
    0x1c59001cU,
    0x1dff82aaU,
    0x125e42abU,
    0xcaf4a3d4U,
    0xb0c8f2e9U,
    0xbc2a29ccU,
    0x27f6cf12U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x98000000U,
    0xe4000000U,
    0x76000000U,
    0xb5000000U,
    0xe3800000U,
    0xae400000U,
    0xfe200000U,
    0x94d00000U,
    0x9e80000U,
    0x1bbc0000U,
    0x57860000U,
    0x90570000U,
    0xa72e8000U,
    0x8d53c000U,
    0x18b12000U,
    0x21123000U,
    0x4d9ba800U,
    0x9f4a1400U,
    0x9bb9a200U,
    0x1786f100U,
    0xf04c0980U,
    0x5727e7c0U,
    0x155d2860U,
    0xfca5d690U,
    0x570e8258U,
    0xf883c014U,
    0x7cd9206eU,
    0x35ee32e9U,
    0xe9bda840U,
    0x648d1717U,
    0x5edf2071U,
    0xef93103U,
    0xab3328a8U,
    0xc74ed56bU,
    0x5fa60123U,
    0xf1870298U,
    0x2d468143U,
    0xc8afc3bfU,
    0x2917231aU,
    0xc1953320U,
    0x955d2ab0U,
    0xbca5d668U,
    0x370e833cU,
    0x883c072U,
    0xe4d9202bU,
    0xd1ee3203U,
    0x9fbda828U,
    0xd18d172bU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0xd8000000U,
    0x24000000U,
    0x82000000U,
    0x91000000U,
    0xa2800000U,
    0x2f400000U,
    0x19200000U,
    0xed00000U,
    0x1df80000U,
    0x5e2c0000U,
    0xb74a0000U,
    0x1d3f0000U,
    0x7cca8000U,
    0x44f14000U,
    0xa1a000U,
    0x3e0a9000U,
    0x17188800U,
    0x419fd400U,
    0xd4d92200U,
    0x68e8d300U,
    0x9eb9ab80U,
    0x191b0440U,
    0x7e8a8960U,
    0x795cd790U,
    0xc221a388U,
    0xc14a9104U,
    0x36388a6aU,
    0x7b4fd46dU,
    0x9321230cU,
    0x83c4d1efU,
    0x973ab75U,
    0xba6407c1U,
    0xb9e00953U,
    0x1c3d9762U,
    0xc6580268U,
    0xafbc0154U,
    0x8b9200c2U,
    0x79c30199U,
    0x8c78827eU,
    0xb2e2428eU,
    0xeba12257U,
    0xd984d3c9U,
    0xd0d3aa17U,
    0x1af40568U,
    0xc7b80915U,
    0x47819610U,
    0xf7ca0109U,
    0x737f0076U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0xf8000000U,
    0x5c000000U,
    0x26000000U,
    0x5000000U,
    0xf2800000U,
    0x91400000U,
    0x43600000U,
    0x8d00000U,
    0x61b80000U,
    0x82740000U,
    0x90560000U,
    0xfbe30000U,
    0x62818000U,
    0x795dc000U,
    0x37702000U,
    0x7adf1000U,
    0xe6b14800U,
    0xfe91c00U,
    0xd09fa200U,
    0xde55d300U,
    0x4af6e880U,
    0xc218cfc0U,
    0x9f0749e0U,
    0x19a1e70U,
    0x6ec62218U,
    0x84ac10d4U,
    0xa0e8c8aaU,
    0x910dcf5U,
    0x1c818075U,
    0x105dc247U,
    0x3bf02254U,
    0x429f1398U,
    0x89514b66U,
    0xcf791cf9U,
    0x26c7a399U,
    0xc0b1d0dbU,
    0xaf8e850U,
    0x221fcd88U,
    0x4f08c8bcU,
    0x980dc66U,
    0xcad9818bU,
    0xfeb9c1aaU,
    0x83fe2087U,
    0xfe9811c6U,
    0x7f5ec95bU,
    0xc263dfe2U,
    0x705803dbU,
    0x2be400d0U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0x58000000U,
    0x3c000000U,
    0x82000000U,
    0x53000000U,
    0x5c800000U,
    0xd3c00000U,
    0x67e00000U,
    0xb6500000U,
    0x4fa80000U,
    0x58ec0000U,
    0x70c20000U,
    0x137f0000U,
    0x31918000U,
    0x1e494000U,
    0x5bbae000U,
    0x8ef65000U,
    0x15d35800U,
    0xfafe8c00U,
    0x53c16200U,
    0x27ec1100U,
    0x565a3b80U,
    0x9fae9f40U,
    0xf15b60U,
    0x4cd18df0U,
    0x9178e388U,
    0x6289530cU,
    0x42c2db92U,
    0x8877ccffU,
    0xe91b80f7U,
    0xa38a4265U,
    0xb54161c4U,
    0xb2c1020U,
    0x573a384eU,
    0x453e9d2bU,
    0xae39586fU,
    0x1ead8f8fU,
    0x1772e161U,
    0x1f8a521aU,
    0x77595a2dU,
    0xb83d8eb6U,
    0xdbbae0b1U,
    0xcef65302U,
    0xf5d35959U,
    0x2afe8ea0U,
    0xbc16250U,
    0x1bec1258U,
    0xd45a3a14U,
    0xccae9ce6U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0x38000000U,
    0xdc000000U,
    0x92000000U,
    0xf000000U,
    0x97800000U,
    0x7fc00000U,
    0xee200000U,
    0x49500000U,
    0x8b780000U,
    0xa8240000U,
    0x50460000U,
    0x8deb0000U,
    0x95718000U,
    0xbd334000U,
    0x8cc16000U,
    0x5bb93000U,
    0xa1915800U,
    0x5ec45400U,
    0x34aee200U,
    0xe6157300U,
    0x691fba80U,
    0x3e8125c0U,
    0xe1495ae0U,
    0x7f705470U,
    0xe30e2c8U,
    0x594a72fcU,
    0x6370383eU,
    0x3c2d674fU,
    0x2647bb10U,
    0xccf52569U,
    0x9ff75abbU,
    0x5a7f55a3U,
    0x8aa760c7U,
    0x8302335dU,
    0x8d98dbe8U,
    0x34d317fdU,
    0x47a983a9U,
    0x9387404aU,
    0x21df60bcU,
    0x9b26309eU,
    0x45dedb9fU,
    0x15381508U,
    0x78d8018dU,
    0xfdb40261U,
    0xa89e0036U,
    0xb05f0142U,
    0x9def8331U,
    0x7d6c41ffU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0x38000000U,
    0x54000000U,
    0x5a000000U,
    0xf000000U,
    0xc6800000U,
    0x60400000U,
    0x4aa00000U,
    0x17d00000U,
    0xd1780000U,
    0x38a40000U,
    0xb4da0000U,
    0xb9eb0000U,
    0x9dfc8000U,
    0xff64000U,
    0x5cf06000U,
    0x3c771000U,
    0x652fa800U,
    0xe813e400U,
    0xc0ee200U,
    0x3e1e5300U,
    0xfd014a80U,
    0xa59db640U,
    0xa8d7aae0U,
    0x7ff7e650U,
    0x94f4e1e8U,
    0x6065517cU,
    0xc325c97aU,
    0xd11ff691U,
    0xcb85caa2U,
    0xcdcff633U,
    0x1e7dc9b7U,
    0x5e2bf41fU,
    0x4487c883U,
    0xcb50f729U,
    0xde234910U,
    0x8492b4aeU,
    0x6b512af7U,
    0x4e3aa759U,
    0xbc800208U,
    0x3f40008aU,
    0x14200149U,
    0xb3900226U,
    0xf9d8013dU,
    0x74740370U,
    0xf92203beU,
    0xee0f017fU,
    0xa5068135U,
    0xc18d40faU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0xb8000000U,
    0x34000000U,
    0x92000000U,
    0xdd000000U,
    0xff800000U,
    0xdcc00000U,
    0x9e200000U,
    0xee300000U,
    0x76280000U,
    0x92340000U,
    0x88320000U,
    0xd9250000U,
    0x80a38000U,
    0x1371c000U,
    0xafca6000U,
    0x62a59000U,
    0x56760800U,
    0xb441cc00U,
    0xa4760a00U,
    0x9941cf00U,
    0x3f60980U,
    0x4181cfc0U,
    0xb7d60b60U,
    0x46b1cf10U,
    0xac7e0928U,
    0xd545cc64U,
    0x45ec0ad6U,
    0x3e90cf17U,
    0x2d4788aeU,
    0x51e50cafU,
    0x7c97e976U,
    0x78519ee7U,
    0x22706356U,
    0x6449183U,
    0xa96f8b54U,
    0xeed10eceU,
    0x5325e86bU,
    0x79b49c84U,
    0x16f3e052U,
    0x12055245U,
    0x1d0de86bU,
    0x1f809c84U,
    0xecc1e052U,
    0x26205245U,
    0xda2e686bU,
    0xe4315c84U,
    0x4f2b8052U,
    0x77b5c245U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x8000000U,
    0xcc000000U,
    0x92000000U,
    0xb9000000U,
    0x10800000U,
    0xf3c00000U,
    0xb4200000U,
    0xe4300000U,
    0x1c380000U,
    0xb8340000U,
    0x56220000U,
    0x75230000U,
    0x10ba8000U,
    0x61e8c000U,
    0x9f1fe000U,
    0xdf8df000U,
    0xd45b8800U,
    0xc4610c00U,
    0xce5b8a00U,
    0xf1610d00U,
    0x2cdb8b80U,
    0xba10fc0U,
    0x807b8ba0U,
    0xd0510cf0U,
    0xba6389e8U,
    0x35550e14U,
    0xe0f989aaU,
    0xb820ddbU,
    0x12410b7aU,
    0xfb79ce4bU,
    0x81dc6b0aU,
    0xfd28fcabU,
    0x9cba0252U,
    0x93e7017fU,
    0x96008010U,
    0xc70fc1f8U,
    0xeb9f63ecU,
    0xe2423046U,
    0x9364e89dU,
    0xfdd33c67U,
    0x672763ecU,
    0xe9b63046U,
    0x1166e89dU,
    0xdcc03c67U,
    0x63a5e3ecU,
    0xfc6af046U,
    0x4a5b089dU,
    0xcf6ecc67U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x98000000U,
    0xec000000U,
    0x2a000000U,
    0x27000000U,
    0xa9800000U,
    0x9400000U,
    0x11e00000U,
    0x99f00000U,
    0x2de80000U,
    0x8be40000U,
    0x16f60000U,
    0x650000U,
    0x4cb88000U,
    0x7e1dc000U,
    0xd1092000U,
    0xcc841000U,
    0x63c93800U,
    0xef390400U,
    0x60493a00U,
    0x1790700U,
    0xb8293b80U,
    0x1c90540U,
    0x1c2139e0U,
    0xffdd06f0U,
    0xd3f3948U,
    0x535c066cU,
    0xfef1bbeeU,
    0x2464c649U,
    0x72a01ba9U,
    0xf0d142eU,
    0xed88031eU,
    0xa7540201U,
    0xf8fe0145U,
    0x49710100U,
    0x6c2681f7U,
    0x37dcc2b8U,
    0x1927a0d3U,
    0x54cd23aU,
    0x6bf099f4U,
    0x46f4d51aU,
    0xf87720d3U,
    0x30b5123aU,
    0xcc0fb9f4U,
    0x1a15c51aU,
    0x4f0698d3U,
    0x4d91d63aU,
    0x574fa3f4U,
    0xf0e8d21aU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0xf8000000U,
    0x84000000U,
    0x26000000U,
    0xa7000000U,
    0x3a800000U,
    0x9ec00000U,
    0x4fe00000U,
    0x87f00000U,
    0x4bf80000U,
    0xe1e40000U,
    0xce60000U,
    0xcb790000U,
    0x6a298000U,
    0x8d4000U,
    0xcbc9a000U,
    0xec66f000U,
    0x10b92800U,
    0x7e5ad400U,
    0x94392a00U,
    0x939ad500U,
    0xdf592980U,
    0xb9aad540U,
    0x3fc12a60U,
    0x627ed750U,
    0x2bbf2af8U,
    0xf6d3d5ccU,
    0x73e8aa12U,
    0xc5f397b7U,
    0xdaf68b2dU,
    0x847525a8U,
    0x2cb18202U,
    0x3c59422fU,
    0x537a031U,
    0xf60bf002U,
    0xff0eaa19U,
    0x4e8a9440U,
    0x50df08d6U,
    0x94f864e1U,
    0x1f78238cU,
    0x543fb3c4U,
    0x338e88d6U,
    0x2f5124e1U,
    0x51b7838cU,
    0x43d043c4U,
    0xc06620d6U,
    0xaaa2b0e1U,
    0x6b41098cU,
    0xd7a566c4U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0x28000000U,
    0x34000000U,
    0x76000000U,
    0x53000000U,
    0xb1800000U,
    0x57400000U,
    0xf5e00000U,
    0x4bf00000U,
    0x7cf80000U,
    0x837c0000U,
    0xab3a0000U,
    0x855f0000U,
    0x88e98000U,
    0x957f4000U,
    0xe8392000U,
    0x2ccd1000U,
    0xc3b4f800U,
    0x2291c400U,
    0x86ccfa00U,
    0xb2adc500U,
    0xa616fa80U,
    0x2b02c5c0U,
    0xbd877a20U,
    0xd418510U,
    0xcce45bf8U,
    0xeb63945cU,
    0xbf21213eU,
    0xc3411101U,
    0xf3f6f8e9U,
    0x10f2c52eU,
    0xc97f7b9aU,
    0x8a3d86e3U,
    0x39de5916U,
    0x93c94e6U,
    0xf048a24dU,
    0x527e532fU,
    0x5fafd8f3U,
    0x208fd68dU,
    0x83d3806cU,
    0x602043b5U,
    0x68d0a273U,
    0xbdb2534dU,
    0x758dda4cU,
    0x695cd6a5U,
    0x82f8018bU,
    0x947c0311U,
    0x64ba0172U,
    0x851f02a4U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0x98000000U,
    0xc000000U,
    0x6a000000U,
    0x85000000U,
    0x30800000U,
    0xd7c00000U,
    0x7ae00000U,
    0x9cf00000U,
    0x33f80000U,
    0xa67c0000U,
    0x11360000U,
    0x24050000U,
    0xce148000U,
    0xb11c000U,
    0x1b856000U,
    0x9c497000U,
    0x7ea08800U,
    0xee5e5400U,
    0xb7b88a00U,
    0x94d25500U,
    0xb5768a80U,
    0x46ab5440U,
    0xf25408e0U,
    0x65bf9470U,
    0x4dc5e948U,
    0x77e72664U,
    0x1860018aU,
    0xd230013bU,
    0xab980261U,
    0x344c0248U,
    0x9aae022eU,
    0x40490191U,
    0xccba800aU,
    0x4758c131U,
    0xbd3fe29aU,
    0x5e11b309U,
    0xf31f6a36U,
    0x678fe627U,
    0x3e47e027U,
    0x6fadb2edU,
    0xb8c968b6U,
    0x8f7ae667U,
    0x5bab62c7U,
    0xcec0739dU,
    0xd87a0b7eU,
    0xb2369443U,
    0xdb9f6badU,
    0xfc4fe6d6U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x48000000U,
    0xa4000000U,
    0xb6000000U,
    0x69000000U,
    0x56800000U,
    0x18400000U,
    0x4e200000U,
    0xcc300000U,
    0xbb280000U,
    0xd0ac0000U,
    0xb0760000U,
    0x631f0000U,
    0xa59b8000U,
    0x15d54000U,
    0xffc6000U,
    0xeddcb000U,
    0xd3feb800U,
    0x47cffc00U,
    0x30f6ba00U,
    0x2253ff00U,
    0x528ba80U,
    0xfda0fc40U,
    0xa0e53ba0U,
    0x8a5abed0U,
    0x512ad978U,
    0xd3bf4f74U,
    0x25fe0022U,
    0xcec30315U,
    0x9665831aU,
    0xa2164325U,
    0xa719e176U,
    0x238af327U,
    0xc4c75a97U,
    0x65750c3bU,
    0xaf99e0d9U,
    0xe6caf2c8U,
    0x2675b3cU,
    0xac050d5eU,
    0xf211e2ebU,
    0x2f16f1a1U,
    0xa79959bcU,
    0xa2c60d1eU,
    0x4474634bU,
    0x5d00b071U,
    0x4880ba44U,
    0x254cfe2aU,
    0x36b338c9U,
    0x5175bdb4U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0x38000000U,
    0xc4000000U,
    0xe6000000U,
    0x85000000U,
    0x23800000U,
    0x39c00000U,
    0x84e00000U,
    0x5ef00000U,
    0x19f80000U,
    0xc96c0000U,
    0x5e2a0000U,
    0x3c8d0000U,
    0xfd458000U,
    0x5db14000U,
    0x7545e000U,
    0xe1bf9000U,
    0xf753a800U,
    0xb2b1b400U,
    0x69cbaa00U,
    0x6cedb500U,
    0x32f9a880U,
    0x33fcb5c0U,
    0x566e2860U,
    0xdaacf6d0U,
    0x45c44bf8U,
    0xe6ef24c4U,
    0x9df78376U,
    0x8f604023U,
    0xeb326365U,
    0x271fd058U,
    0x81c982U,
    0x3f5e67bdU,
    0xaeb2621aU,
    0xdbdfd15fU,
    0x7e1caf7U,
    0x686e67fdU,
    0xbaa63ccU,
    0x8843d24cU,
    0xa633c9faU,
    0x188f6539U,
    0x8b45e34cU,
    0xd0bf938cU,
    0xaad3ab9aU,
    0xfa71b6e9U,
    0x10aba834U,
    0x4addb688U,
    0x6a61aa8cU,
    0x18a0b51aU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x88000000U,
    0xac000000U,
    0x5e000000U,
    0x2d000000U,
    0xf8800000U,
    0x88c00000U,
    0x5ea00000U,
    0x13b00000U,
    0xfb380000U,
    0x1be40000U,
    0xb94e0000U,
    0x7cef0000U,
    0x6c28000U,
    0xbbad4000U,
    0xa726e000U,
    0x7df6f000U,
    0xe0556800U,
    0xfe6bfc00U,
    0x511b6a00U,
    0x6e84fd00U,
    0xe9d9e980U,
    0xa829bc40U,
    0x3e7f0ba0U,
    0xf11f4ef0U,
    0xde8a62d8U,
    0x31c4b244U,
    0x8c290b3aU,
    0xcc644d67U,
    0x821ee3c0U,
    0xb12f169U,
    0x419b6b8cU,
    0x5a44fff6U,
    0x8179ebb1U,
    0x6a99bc41U,
    0xebc70890U,
    0xe33b4fc8U,
    0x9fe463ccU,
    0xfb5bb256U,
    0xd7f38a41U,
    0xf75d0e19U,
    0x39ee0114U,
    0x25f0012U,
    0xe57a817bU,
    0x7889437eU,
    0xc8c8e2d4U,
    0xbea9f17bU,
    0x43afeaf7U,
    0x7322bc88U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0xa8000000U,
    0xf4000000U,
    0xf6000000U,
    0xbf000000U,
    0x7c800000U,
    0x7c00000U,
    0x4ee00000U,
    0x81f00000U,
    0x95680000U,
    0xc6bc0000U,
    0x8e420000U,
    0x18a70000U,
    0x5518000U,
    0x32334000U,
    0x4d86a000U,
    0xa85e3000U,
    0x3fa74800U,
    0x85c1d400U,
    0xc7e54a00U,
    0xe266d700U,
    0x1e34c880U,
    0x279594c0U,
    0xb5526a20U,
    0xfa3ba710U,
    0xe99d21f8U,
    0x64671ecU,
    0xdcba6a0aU,
    0xfb47a6f3U,
    0x893f23b1U,
    0x6f117034U,
    0xe483ebc4U,
    0xfbc8e43eU,
    0xbcfb8005U,
    0x60e841b0U,
    0xa2f5201aU,
    0x4bfa708bU,
    0x78786a1dU,
    0xab20a6deU,
    0x160ea307U,
    0x2f1233e7U,
    0x848d4be5U,
    0xabdad732U,
    0xe4f6c90dU,
    0x3cf29514U,
    0xa0e3e854U,
    0x2f8e706U,
    0xbbf382c9U,
    0xd064412aU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0x88000000U,
    0x1c000000U,
    0x16000000U,
    0xb9000000U,
    0x3f800000U,
    0x2a400000U,
    0xf3e00000U,
    0x62f00000U,
    0xf1680000U,
    0x35340000U,
    0x7bc60000U,
    0xa4a70000U,
    0xcc8c8000U,
    0x9ada4000U,
    0x2f3b6000U,
    0x9cd39000U,
    0xde3e5800U,
    0x5f446400U,
    0x72785a00U,
    0x9da36700U,
    0x3314d980U,
    0x50892540U,
    0x4cc7bba0U,
    0x762eb530U,
    0x335fe3f8U,
    0x7c7dd3d4U,
    0xb0a33906U,
    0x680f47dU,
    0xd5c28349U,
    0x39b940b6U,
    0x119e32cU,
    0x1b9ad152U,
    0xd84fbb3bU,
    0x58eab714U,
    0x6671e08fU,
    0x57aed042U,
    0x7c09badaU,
    0x460db4b7U,
    0xd11d62f6U,
    0xb384910cU,
    0xb45ad922U,
    0x56ea2763U,
    0x4b653bf0U,
    0x6227f571U,
    0xf94e026bU,
    0x336303d5U,
    0xa62282dcU,
    0x9b494323U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0x58000000U,
    0x2c000000U,
    0x12000000U,
    0x31000000U,
    0xe800000U,
    0x1f400000U,
    0x62e00000U,
    0x1bf00000U,
    0x81780000U,
    0xd02c0000U,
    0x19d60000U,
    0x49ad0000U,
    0xe7108000U,
    0x699ec000U,
    0x36dee000U,
    0xf43a3000U,
    0x3fdd5800U,
    0xe6a5c400U,
    0x1a8b5a00U,
    0x1148c500U,
    0x69fbd880U,
    0x80660640U,
    0xf6bd39e0U,
    0xa28036f0U,
    0x4d4e61a8U,
    0xf3e4f234U,
    0xc563bb92U,
    0xc62ff749U,
    0x9ece0299U,
    0x10310166U,
    0xf9de817cU,
    0x39afc0d6U,
    0x6f0062b3U,
    0x1d95f2c4U,
    0x8dd383dU,
    0xd730376bU,
    0x56631fU,
    0xf778f16dU,
    0x672db8d4U,
    0x685ef7e5U,
    0xf37082b7U,
    0xd12ec359U,
    0x3f46e346U,
    0xf2e630acU,
    0xe3f3582eU,
    0x7d64c63fU,
    0x9a2dd83aU,
    0x4cb057aU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xf8000000U,
    0xa4000000U,
    0x1a000000U,
    0xa7000000U,
    0x88800000U,
    0x6b400000U,
    0x7200000U,
    0x8300000U,
    0x6cb80000U,
    0x61fc0000U,
    0x8fc20000U,
    0x4a6d0000U,
    0xc6018000U,
    0x591ec000U,
    0x15982000U,
    0xb4cc3000U,
    0x1cf4d800U,
    0xc04cfc00U,
    0x79b6da00U,
    0xa261fd00U,
    0xa175a80U,
    0xcf0f3ec0U,
    0x4977960U,
    0xdd4f0c50U,
    0xe639a108U,
    0x99a2f2ccU,
    0x3274f92aU,
    0x220cce61U,
    0x631803b6U,
    0x428c0081U,
    0xc45a0364U,
    0xd3a10156U,
    0xdd7b8253U,
    0xd68fc3b3U,
    0x465ba0e1U,
    0xbff376U,
    0x47ed7ae1U,
    0x46de0eb4U,
    0x77fa231eU,
    0xded130bfU,
    0x3ed5be9U,
    0xccde3c78U,
    0xf8f4fa34U,
    0xfa4ccedeU,
    0x2eb8005fU,
    0xd2fc00f9U,
    0xc5420350U,
    0xd22d0288U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xd8000000U,
    0x1c000000U,
    0x46000000U,
    0xd1000000U,
    0x20800000U,
    0x42400000U,
    0xae200000U,
    0xfb300000U,
    0xe1a80000U,
    0xa4fc0000U,
    0x135a0000U,
    0x4eb10000U,
    0x197a8000U,
    0x7f9e4000U,
    0x87c66000U,
    0xeee77000U,
    0xac53b800U,
    0xdb233c00U,
    0x11a9ba00U,
    0xdce23f00U,
    0x3f5b3880U,
    0xd0b07fc0U,
    0xd46f5be0U,
    0x191a0fb0U,
    0x149c62f8U,
    0x605670f4U,
    0x1529387aU,
    0x94bd7efdU,
    0x4e6fda42U,
    0x2e054ed1U,
    0xd5088024U,
    0xda9340f2U,
    0xe546e011U,
    0x87b831fcU,
    0x85e75b6eU,
    0x4bd60e8fU,
    0x20ee622bU,
    0x295b70a1U,
    0x49a9b83cU,
    0xe23e36U,
    0xd95b38d3U,
    0x31b07c55U,
    0x2cef5a46U,
    0x475a0fcbU,
    0xfcbc6211U,
    0x4a667144U,
    0xd4013b82U,
    0x72017f89U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0xf8000000U,
    0x4c000000U,
    0x6a000000U,
    0x8f000000U,
    0x43800000U,
    0x2400000U,
    0x7fa00000U,
    0xf4b00000U,
    0x59280000U,
    0xfa6c0000U,
    0x47560000U,
    0xa3390000U,
    0x6d788000U,
    0xb8c9c000U,
    0xe3762000U,
    0x29d35000U,
    0x29e99800U,
    0x5d14ac00U,
    0x609f9a00U,
    0xdbddaf00U,
    0xdaef1a80U,
    0x7c886c40U,
    0x89c73960U,
    0x39fe3d70U,
    0x850820c8U,
    0xfc86534cU,
    0x49c71bc6U,
    0x19e46c45U,
    0x551138b8U,
    0x4873fd7U,
    0x5d0a2bcU,
    0x73ff924eU,
    0xda193be9U,
    0x471b3cceU,
    0x78ea07aU,
    0xc5a9058U,
    0x2ebfba67U,
    0x1e37fe54U,
    0xfdfe0292U,
    0x4b150057U,
    0x8d8e82afU,
    0x7340c118U,
    0x4526a354U,
    0xa8769312U,
    0xa449b897U,
    0x5abefe8fU,
    0x182e8288U,
    0xc8f0c32cU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0xf8000000U,
    0x14000000U,
    0xe6000000U,
    0x99000000U,
    0x47800000U,
    0xef400000U,
    0x4de00000U,
    0x12f00000U,
    0xbc780000U,
    0xdca40000U,
    0x421e0000U,
    0x17110000U,
    0xd29b8000U,
    0x8adb4000U,
    0x25b26000U,
    0x95959000U,
    0xe052f800U,
    0x5b738400U,
    0x8634fa00U,
    0x14c68500U,
    0xe8b17980U,
    0xd40cc640U,
    0x46189a60U,
    0xa9021610U,
    0x6f9801f8U,
    0x3540074U,
    0xbfe602e6U,
    0x6df502c9U,
    0x62e58051U,
    0x747a41c2U,
    0xe0b1e140U,
    0x481ad2e0U,
    0x7c069bd0U,
    0xea1314d8U,
    0xbb038384U,
    0xc08f404eU,
    0x25d463a5U,
    0x3209013U,
    0x29577ab5U,
    0xa4f9c51cU,
    0x9d7d1bddU,
    0x6f385627U,
    0x9b49e233U,
    0x3bfed1c5U,
    0x93f89a74U,
    0x3ff21791U,
    0x2de00195U,
    0x82f001ecU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0xc8000000U,
    0x4c000000U,
    0x3a000000U,
    0x1f000000U,
    0x94800000U,
    0xebc00000U,
    0x37600000U,
    0x82700000U,
    0x31f80000U,
    0x6aac0000U,
    0x701a0000U,
    0x28130000U,
    0x5c128000U,
    0x4201c000U,
    0x9b142000U,
    0xe29fb000U,
    0xcecb4800U,
    0xbceae400U,
    0xfd294a00U,
    0xed55e500U,
    0xdfa1c880U,
    0xc38727c0U,
    0x73476ba0U,
    0x46a956f0U,
    0x1a000348U,
    0xef00028cU,
    0xfc80031aU,
    0x17c00023U,
    0xc5600147U,
    0xd17000eaU,
    0x9f7802a0U,
    0x9e6c0070U,
    0xd3fa0388U,
    0x41a300acU,
    0x5a8a80aaU,
    0xaaddc20bU,
    0xdaf6221bU,
    0xa020b388U,
    0xe2c3c9e7U,
    0xd6f824d1U,
    0x3a2fe9d3U,
    0x4dcb96c4U,
    0x8a7ea15dU,
    0x9df27202U,
    0x40a56bdcU,
    0x171654a2U,
    0x388880e7U,
    0xc1d2c251U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xa8000000U,
    0x2c000000U,
    0xfe000000U,
    0x51000000U,
    0xef800000U,
    0x6400000U,
    0x57600000U,
    0x10700000U,
    0x22e80000U,
    0x5240000U,
    0xfd0e0000U,
    0xd18b0000U,
    0x37568000U,
    0xe8e54000U,
    0xbe356000U,
    0x59859000U,
    0xeb5cd800U,
    0x8efbf400U,
    0x3b3ada00U,
    0xcc14f700U,
    0x6e025b80U,
    0x990ab640U,
    0x9389b920U,
    0x504e66f0U,
    0x2a6e02d8U,
    0x1fb03b4U,
    0x75be8366U,
    0xbdc140adU,
    0xeb3b63bbU,
    0xa40e91acU,
    0x220a5990U,
    0x371eb508U,
    0x6a8fbb9cU,
    0x93d1640aU,
    0xd23e817fU,
    0x78142f0U,
    0x8a5b627aU,
    0x997e9067U,
    0xb9625b24U,
    0x497ab64cU,
    0xd161bb22U,
    0x56a6593U,
    0x7f600162U,
    0xfc700011U,
    0xbce80241U,
    0x424038bU,
    0xba8e0194U,
    0xfbcb02b4U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0xc8000000U,
    0x7c000000U,
    0xc2000000U,
    0x13000000U,
    0x61800000U,
    0xc3400000U,
    0x58200000U,
    0xa1300000U,
    0x5fb80000U,
    0x97740000U,
    0xd3020000U,
    0x418b0000U,
    0x134c8000U,
    0x2038c000U,
    0x153ca000U,
    0xe1a19000U,
    0x467a7800U,
    0xa19f3c00U,
    0xe3407a00U,
    0x88203d00U,
    0xd92efb80U,
    0xeba3ffc0U,
    0x2966d8a0U,
    0x20eae30U,
    0x330203a8U,
    0xb18b027cU,
    0xbb4c802eU,
    0xec38c071U,
    0x1f3ca0ceU,
    0x8ea19035U,
    0xe5fa7a90U,
    0x71df3f98U,
    0xdae078d4U,
    0xea503fd2U,
    0xdeb6f89fU,
    0xdde7fe9fU,
    0xa5dcda8bU,
    0xd4f1ae6dU,
    0xf34c81e4U,
    0xd038c2faU,
    0xbd3ca2a3U,
    0x2da193d1U,
    0x4c7a796aU,
    0xce9f3cbbU,
    0x40c079c5U,
    0x58603d98U,
    0xe08ef9d4U,
    0x89d3fe52U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0xf8000000U,
    0x2c000000U,
    0x96000000U,
    0xcf000000U,
    0x62800000U,
    0xdd400000U,
    0xfea00000U,
    0x4bb00000U,
    0x90380000U,
    0x10ec0000U,
    0xf9860000U,
    0x45c90000U,
    0xdae48000U,
    0x48924000U,
    0x9c59a000U,
    0x8f36f000U,
    0xca66d800U,
    0xa8d6cc00U,
    0x7d78da00U,
    0xa643cd00U,
    0x96225880U,
    0xc7f48c40U,
    0x47197b60U,
    0xa6993cf0U,
    0x3f428138U,
    0xdbab43ccU,
    0x882522b2U,
    0xccf8b1b9U,
    0xc7817bc8U,
    0x5ec53dd7U,
    0x27c8068U,
    0xccce4024U,
    0xcf67a156U,
    0x7b53f24fU,
    0x79a45ad7U,
    0xd3d8de8U,
    0x5f7df8e4U,
    0x634b7c76U,
    0xa5bb20dfU,
    0x332db39fU,
    0x447bf95cU,
    0xbbc27ffaU,
    0x21ffa30dU,
    0x600ff3d6U,
    0xb01a5b2cU,
    0x8188da1U,
    0x841f7ab7U,
    0x42103ebbU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x48000000U,
    0x5c000000U,
    0x3a000000U,
    0xe7000000U,
    0x99800000U,
    0x4400000U,
    0xbb200000U,
    0xa300000U,
    0x46b80000U,
    0x407c0000U,
    0xec8a0000U,
    0xf6d30000U,
    0xc4f08000U,
    0xf643c000U,
    0x40312000U,
    0x89b71000U,
    0x15e61800U,
    0x2acd3c00U,
    0xbef41a00U,
    0x31523d00U,
    0x69b69a80U,
    0xc5feffc0U,
    0xf2dd3ba0U,
    0x1ae92eb0U,
    0x1f428348U,
    0xe8acc12cU,
    0x816ba32eU,
    0x8817d03dU,
    0x3c1fb942U,
    0xaa05eed5U,
    0x1f092378U,
    0x8d8b13a4U,
    0x624c1aa2U,
    0x662e3c23U,
    0x74bc9877U,
    0xdb6dfd5bU,
    0xff0dbad3U,
    0x5d9aef79U,
    0xba4ba030U,
    0xc227d02eU,
    0x5aa7bb1bU,
    0x5a79ed95U,
    0xbb8322beU,
    0x275813a3U,
    0x9cbc9b91U,
    0x776dfdacU,
    0xad0db948U,
    0x569aee8aU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x58000000U,
    0xac000000U,
    0xde000000U,
    0xe9000000U,
    0x99800000U,
    0x82c00000U,
    0x9d600000U,
    0x3e700000U,
    0x6cf80000U,
    0x8ac0000U,
    0x239a0000U,
    0xb1c10000U,
    0x67fc8000U,
    0xe234000U,
    0x85c06000U,
    0xf5fc5000U,
    0x29378800U,
    0x1d48e400U,
    0x1ab58a00U,
    0x4495e500U,
    0x894b0880U,
    0x38aba7c0U,
    0x3b95e9e0U,
    0xbdd9b570U,
    0x9e48298U,
    0xbf3f42d4U,
    0xb04262feU,
    0xa92151dfU,
    0x5d490a0bU,
    0xbab6a7e6U,
    0xf48b68d8U,
    0xd157f474U,
    0x94a261aeU,
    0xe59153d7U,
    0x54d10937U,
    0x906aa654U,
    0x3de96aadU,
    0x2d3af51eU,
    0x9744e23cU,
    0x31b31332U,
    0xb20de8b5U,
    0xd705b40aU,
    0x60868322U,
    0xf352429dU,
    0xeba4e0a6U,
    0xd10312f8U,
    0x6595eae4U,
    0x14d9b746U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x8000000U,
    0x54000000U,
    0xd6000000U,
    0xe5000000U,
    0xab800000U,
    0xe9400000U,
    0x1b200000U,
    0x70b00000U,
    0x39f80000U,
    0x92c40000U,
    0xba7e0000U,
    0xdf870000U,
    0xcf498000U,
    0x462d4000U,
    0xd736e000U,
    0x5aaa1000U,
    0xeefa7800U,
    0x525ffc00U,
    0x41b3fa00U,
    0x472bd00U,
    0x7e851980U,
    0x5ad8ac40U,
    0x4e7f61a0U,
    0xb9875110U,
    0x724c98f8U,
    0xb1b5ef84U,
    0xbc6983d6U,
    0x729d4061U,
    0xd0cee21aU,
    0x296e10b3U,
    0x21047a05U,
    0xd598fd7cU,
    0xe85a7b3aU,
    0x3eaffe63U,
    0x20ebfa1dU,
    0xb46bfc8U,
    0x38231b44U,
    0xd62bafdeU,
    0x7f30e1c5U,
    0xbea911f4U,
    0x60edf95eU,
    0xeb45bf6dU,
    0x68349bc0U,
    0xde31ec88U,
    0x2b378264U,
    0x68aa410eU,
    0x85ff61ddU,
    0x40c75340U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x18000000U,
    0x54000000U,
    0xfa000000U,
    0x83000000U,
    0xca800000U,
    0x95c00000U,
    0xb7a00000U,
    0x2d300000U,
    0x70f80000U,
    0x1b5c0000U,
    0x307a0000U,
    0x75810000U,
    0x89578000U,
    0x9f72c000U,
    0x8103a000U,
    0xcd9d7000U,
    0x2d550800U,
    0x5d745c00U,
    0xe6028a00U,
    0xe5069d00U,
    0x6f812a80U,
    0xda5befc0U,
    0xddf422e0U,
    0xe8dfb290U,
    0xcc2eab08U,
    0x2d752d9cU,
    0x9e0d82a2U,
    0x2103c00bU,
    0x3d8c229cU,
    0x1543b22fU,
    0xb974a857U,
    0x4042fdeU,
    0x32020234U,
    0x5f1d0203U,
    0xcc8d808dU,
    0xa0c3c351U,
    0x502c225eU,
    0xb73b179U,
    0x1b0caabfU,
    0xde982e5fU,
    0x4fd800c2U,
    0x84ac0356U,
    0xffa202e8U,
    0xb12d0181U,
    0x56f582f6U,
    0x9e5fc155U,
    0xcff62265U,
    0x7c2b2d0U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0x48000000U,
    0x9c000000U,
    0x42000000U,
    0x51000000U,
    0xc2800000U,
    0x25c00000U,
    0x65600000U,
    0x9ff00000U,
    0xfe280000U,
    0xad5c0000U,
    0xeda60000U,
    0x8a070000U,
    0xd088000U,
    0x2086c000U,
    0x24dbe000U,
    0xeff33000U,
    0x26354800U,
    0xd9500400U,
    0x63bdca00U,
    0xd116c700U,
    0x2862880U,
    0x85d5f640U,
    0x357b61a0U,
    0xd7e9f030U,
    0x6228aaa8U,
    0xef543674U,
    0xbca80122U,
    0x489c00a3U,
    0x28c60117U,
    0x45f702ecU,
    0xbb208047U,
    0x11dac22dU,
    0x8b7de0d3U,
    0x34f433a6U,
    0xe9bdcb70U,
    0xdc16c488U,
    0x22062884U,
    0xa115f46aU,
    0xda9b6347U,
    0xf1d9f0adU,
    0xbb60a993U,
    0x8cf83686U,
    0x6da60300U,
    0x4a070280U,
    0xad088140U,
    0x7086c120U,
    0x6cdbe270U,
    0x73f33108U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0x38000000U,
    0x4000000U,
    0x76000000U,
    0xdf000000U,
    0x84800000U,
    0x84400000U,
    0xfca00000U,
    0x5c300000U,
    0x7e780000U,
    0xf5c40000U,
    0x617a0000U,
    0xd1450000U,
    0x75218000U,
    0x65e54000U,
    0x1503a000U,
    0x6996b000U,
    0x49c97800U,
    0x53784c00U,
    0x9848fa00U,
    0x5ead0d00U,
    0x5d335980U,
    0x69ffbcc0U,
    0xbf002160U,
    0xb482f0d0U,
    0x5c495ab8U,
    0x88babeacU,
    0x1221a12aU,
    0xa567b07dU,
    0x74afbb8U,
    0x3a2c0dddU,
    0xa968daf9U,
    0xad5ffda6U,
    0xe72203d6U,
    0xbcf1025eU,
    0x768380daU,
    0x6d5442c4U,
    0x472022cfU,
    0x2cf2f168U,
    0x3e915a94U,
    0x514ebfc6U,
    0x3523a177U,
    0x85e6b0b5U,
    0x65117a6dU,
    0x518c4f60U,
    0x4dcafb21U,
    0x256c0e2bU,
    0x4748dbd7U,
    0xda2ffd74U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x48000000U,
    0x5c000000U,
    0xaa000000U,
    0xb5000000U,
    0x17800000U,
    0x42400000U,
    0x6200000U,
    0xddb00000U,
    0x8de80000U,
    0xacc0000U,
    0x6fa0000U,
    0x9e5b0000U,
    0x6c3c8000U,
    0x48a3c000U,
    0xaa666000U,
    0x841000U,
    0x5ccd9800U,
    0xe9f43c00U,
    0x54d11a00U,
    0x55e7ff00U,
    0xeedf7a80U,
    0x98efefc0U,
    0xed48e3a0U,
    0x2cb0d2b0U,
    0xf46d7b08U,
    0x5388ef64U,
    0x2c4661d6U,
    0x1d3411adU,
    0xf1259a6eU,
    0xd3383c6bU,
    0x1a2b1951U,
    0x97bcfc24U,
    0x28e3fa1cU,
    0x654c2c30U,
    0x50ae8062U,
    0x6e74c0c9U,
    0xae80e012U,
    0x67ccd34bU,
    0xf57f7b8bU,
    0x421fec01U,
    0x1900e27cU,
    0xd58cd320U,
    0x9b5f78daU,
    0xf3afef25U,
    0x76e8e2e0U,
    0x3640d0d0U,
    0x20257b18U,
    0x9ab4ed5cU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0x98000000U,
    0x5c000000U,
    0x12000000U,
    0x7000000U,
    0x4e800000U,
    0x4ac00000U,
    0xf7e00000U,
    0x2b700000U,
    0xa6a80000U,
    0xbfdc0000U,
    0x6e7a0000U,
    0xa7210000U,
    0x6f868000U,
    0xb556c000U,
    0xaaab2000U,
    0xe5de7000U,
    0xbd60e800U,
    0x8fb3b400U,
    0xf44e6a00U,
    0x45397700U,
    0x909f4880U,
    0xb7c605c0U,
    0xda7922e0U,
    0xd12370b0U,
    0x969c68a8U,
    0x76c477acU,
    0xf5e3c912U,
    0x471c747U,
    0x2c34822dU,
    0xbb1bc05aU,
    0x8c9fa09fU,
    0xc5c5b0c9U,
    0xcd7f4ad4U,
    0xb7b605a2U,
    0x185121efU,
    0xcf3f7081U,
    0xcb866ac8U,
    0xeb557498U,
    0x97ad48c4U,
    0x684b075eU,
    0xf72da065U,
    0x2788b102U,
    0x614bcabbU,
    0xccadc727U,
    0x34ce8219U,
    0xbafac30cU,
    0x9ef92246U,
    0xe370e1U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0x78000000U,
    0x14000000U,
    0x2a000000U,
    0x69000000U,
    0xe800000U,
    0x54c00000U,
    0x82200000U,
    0xc6b00000U,
    0x4b680000U,
    0xcfcc0000U,
    0x61b60000U,
    0xeef10000U,
    0xb08f8000U,
    0xd7d4c000U,
    0xc5a4e000U,
    0xccfcb000U,
    0xe583e800U,
    0x9f40cc00U,
    0xc6646a00U,
    0x3580f00U,
    0x10768a80U,
    0xc55be40U,
    0x19fae360U,
    0x6d01b210U,
    0xbc9a69c8U,
    0x59d50cb4U,
    0x8ea70b72U,
    0x977c7fa7U,
    0xf9c7801aU,
    0x1ea8c1adU,
    0xcf7ae27fU,
    0xbdc1b066U,
    0x4cba6bebU,
    0xb2650e6aU,
    0x994f08fbU,
    0x71707c2cU,
    0x3ed180e0U,
    0xb29c3deU,
    0xb83d6007U,
    0x37a97324U,
    0x21e08a7aU,
    0x9914bc33U,
    0xe69d6118U,
    0x18d97152U,
    0xec288859U,
    0xfda8beedU,
    0x38e36291U,
    0xbf9472d9U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x28000000U,
    0x84000000U,
    0x82000000U,
    0x55000000U,
    0x37800000U,
    0xb0400000U,
    0x3be00000U,
    0x6700000U,
    0x37380000U,
    0xb15c0000U,
    0xc66e0000U,
    0x57330000U,
    0x814f8000U,
    0xfe69c000U,
    0xfb3c6000U,
    0x87431000U,
    0x2963d800U,
    0x99a14400U,
    0x945a00U,
    0xa2d48500U,
    0xa4263a80U,
    0x31d49540U,
    0x24b26220U,
    0xd3001350U,
    0xa0945a28U,
    0xf2d48704U,
    0xac263876U,
    0xa5d49485U,
    0x8eb260b9U,
    0x20012ccU,
    0x151458cdU,
    0x1794864dU,
    0xa0463b82U,
    0x13e495c4U,
    0x826a6196U,
    0xb52c12b5U,
    0xe4425ae1U,
    0xf1fb8430U,
    0xe767b9d7U,
    0xbabe571aU,
    0xf81980d8U,
    0xcc06c3bcU,
    0x361de33aU,
    0xef19d207U,
    0xce903972U,
    0x81cb957cU,
    0x5cabe0daU,
    0x5f06d137U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0x48000000U,
    0x8c000000U,
    0x9a000000U,
    0x57000000U,
    0x4b800000U,
    0x50400000U,
    0xca00000U,
    0x65300000U,
    0xe6780000U,
    0x975c0000U,
    0x1f220000U,
    0xa1730000U,
    0xc4ce8000U,
    0xfb67c000U,
    0xf3d22000U,
    0xe0ffd000U,
    0x9b851800U,
    0x28462c00U,
    0xe8b39a00U,
    0x33ded00U,
    0x6363ba80U,
    0x7c13cc0U,
    0x9ef023a0U,
    0xaa8cd2f0U,
    0xe6cb9bc8U,
    0x2861eeacU,
    0x8e41b866U,
    0xdb23f2dU,
    0xc3bea1f4U,
    0xdaab1149U,
    0xc839bb25U,
    0xaaee3d78U,
    0xb49ca29fU,
    0x87d81320U,
    0xdef73b30U,
    0x8a89fee8U,
    0x96ce829cU,
    0x6067c10eU,
    0x25223f1U,
    0x97bfd39aU,
    0x94a519e8U,
    0x91362c87U,
    0x986b9b74U,
    0xa651ec12U,
    0xd1b9b8bfU,
    0x61ae3d0bU,
    0x49bca222U,
    0x95a810d7U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0x88000000U,
    0x4000000U,
    0x66000000U,
    0xc5000000U,
    0x57800000U,
    0xcac00000U,
    0x99a00000U,
    0x6c300000U,
    0xe5f80000U,
    0x3ecc0000U,
    0x37b20000U,
    0xd3b0000U,
    0xa4678000U,
    0xb9114000U,
    0xfd86e000U,
    0xcdd59000U,
    0x1d348800U,
    0xbc6ddc00U,
    0xe50b0a00U,
    0x67809d00U,
    0xe2c7e880U,
    0xeda20cc0U,
    0x8226e2a0U,
    0x24e590d0U,
    0xf4c8938U,
    0x3861dec4U,
    0xc31908e6U,
    0x28b9d2fU,
    0xc5586a20U,
    0xaf7f4d4fU,
    0x1f92020fU,
    0x2ecb02b0U,
    0x2fbf8017U,
    0x512d43bbU,
    0x3e6ce33eU,
    0x96129204U,
    0xd190b19U,
    0xf38b9c68U,
    0xdcd86bbcU,
    0xd4bf4c42U,
    0x3fb201b9U,
    0x493b01e7U,
    0x62678113U,
    0xc114322U,
    0x2206e016U,
    0x3159058U,
    0xe294886bU,
    0x155ddcd9U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0xb8000000U,
    0x94000000U,
    0xaa000000U,
    0xb9000000U,
    0x77800000U,
    0x87c00000U,
    0x5b600000U,
    0xc2f00000U,
    0xbe280000U,
    0x97c40000U,
    0xd36a0000U,
    0x8ef50000U,
    0x70338000U,
    0x3cd4c000U,
    0x89fae000U,
    0xd4b99000U,
    0x159aa800U,
    0xd2d1a400U,
    0x72e12a00U,
    0xa6316700U,
    0x93d9cb80U,
    0x3179f4c0U,
    0x1bfae160U,
    0x39b99290U,
    0xa81aa9c8U,
    0x1c11a774U,
    0xe6012b96U,
    0x770165abU,
    0xdc91c99bU,
    0xdd4df730U,
    0x138e32bU,
    0xa74891dbU,
    0x50232910U,
    0x6cc0641bU,
    0xa1e04be3U,
    0x8a8353cU,
    0x939b8131U,
    0x55d0c2aaU,
    0x1670e29aU,
    0xef7c9231U,
    0x38e12b2aU,
    0x2f31655aU,
    0x3c59cad1U,
    0xd2b9f67aU,
    0x529ae072U,
    0xd64992f5U,
    0xcbb2abc4U,
    0xb515a47dU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0x38000000U,
    0xdc000000U,
    0xde000000U,
    0x6d000000U,
    0x90800000U,
    0x6dc00000U,
    0x1e600000U,
    0x34f00000U,
    0xf6280000U,
    0xebcc0000U,
    0x3f720000U,
    0x127d0000U,
    0x62fa8000U,
    0x5f2b4000U,
    0x94fe000U,
    0xd9b17000U,
    0x2d948800U,
    0x754d2400U,
    0xb7a60a00U,
    0x589a6700U,
    0x49d3ea80U,
    0x3c6a16c0U,
    0xf7e7e260U,
    0x33bd71b0U,
    0xca868b18U,
    0x92c02704U,
    0x55f48bdaU,
    0x30bd2473U,
    0x2f0e0bc3U,
    0xc3966610U,
    0xc041eb83U,
    0xab27177bU,
    0xa3556244U,
    0xdeaa3331U,
    0x9a136bf4U,
    0xdf005429U,
    0x3b8882f0U,
    0x3c5642f3U,
    0x5356303U,
    0xf65a3170U,
    0x923b6bb3U,
    0x29cc5623U,
    0xac7a80e0U,
    0x9feb40bbU,
    0xa7afe06fU,
    0xf08170d6U,
    0xfddc88aeU,
    0x76712651U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x88000000U,
    0xd4000000U,
    0x46000000U,
    0xd9000000U,
    0x81800000U,
    0xd5400000U,
    0xd0600000U,
    0xc0f00000U,
    0xe0380000U,
    0xf3040000U,
    0x2820000U,
    0xffdb0000U,
    0xabbd8000U,
    0x15534000U,
    0x70616000U,
    0x90efd000U,
    0xd83fb800U,
    0xaf059400U,
    0x90803a00U,
    0x60cdd500U,
    0xf33cd980U,
    0x418147c0U,
    0x755a03a0U,
    0x806f0090U,
    0xf8e780b8U,
    0xbc3c42f4U,
    0x6106e13eU,
    0x9d939361U,
    0xa759585fU,
    0xff660732U,
    0x106160c7U,
    0x60efd29aU,
    0xb03fbb4bU,
    0xcb0595f0U,
    0x5e803904U,
    0x6dcdd636U,
    0x34bcdbfdU,
    0x4dc1454dU,
    0x24ba01a9U,
    0x95df036fU,
    0xc8bf831aU,
    0x8fc8408bU,
    0x83bce050U,
    0x914c9294U,
    0xe66d80eU,
    0x15ee4649U,
    0xcbbd8393U,
    0xe5534338U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x8000000U,
    0xfc000000U,
    0xae000000U,
    0x11000000U,
    0x24800000U,
    0x8d400000U,
    0x88600000U,
    0x45f00000U,
    0x68280000U,
    0x870c0000U,
    0x918e0000U,
    0xd3d70000U,
    0xd62c8000U,
    0x5e1b4000U,
    0x69106000U,
    0x60977000U,
    0x47451800U,
    0xc3653400U,
    0x22679a00U,
    0x7ee97700U,
    0xb7bb7a80U,
    0xf5d54540U,
    0xfb2602a0U,
    0xf49b02b0U,
    0x65428218U,
    0xc47c41f4U,
    0x73f4e20aU,
    0x8d303181U,
    0xf193780bU,
    0xa3d94766U,
    0x6e280113U,
    0x3a0c023eU,
    0x330e0147U,
    0x23970204U,
    0xdccc809eU,
    0x7bab43dbU,
    0x3d86202U,
    0x3e2b73f1U,
    0x12031b73U,
    0x5f0e35c2U,
    0x858d1811U,
    0x31d93563U,
    0xf121986aU,
    0x9f82767dU,
    0x92d1f95dU,
    0xdaa50401U,
    0xbf5ce3e7U,
    0x477c3098U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x58000000U,
    0xa4000000U,
    0x2000000U,
    0x39000000U,
    0x93800000U,
    0x41c00000U,
    0x80a00000U,
    0x3e300000U,
    0xf6780000U,
    0xca140000U,
    0xd5020000U,
    0x3d810000U,
    0xb6d98000U,
    0xd43fc000U,
    0x7363e000U,
    0x9f8d1000U,
    0x9fd96800U,
    0x4fa05c00U,
    0xfea2ea00U,
    0xe12e9d00U,
    0x3ee08b80U,
    0x11484dc0U,
    0xdff802e0U,
    0x66d40350U,
    0xfc2200e8U,
    0xef7102b4U,
    0x9818326U,
    0xfcdbc373U,
    0x4139e044U,
    0x2ee8111bU,
    0x195aeb95U,
    0x13fa9ed7U,
    0x98c2888fU,
    0x63394e26U,
    0x47f981d6U,
    0xe2cfc14eU,
    0xae3be0a2U,
    0xbe691388U,
    0x66036801U,
    0x1b055e64U,
    0xfa81680eU,
    0xba445c27U,
    0x3d78eab2U,
    0x188b9ec0U,
    0xf3430b45U,
    0x96e28edaU,
    0xcd406211U,
    0x29e7d209U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0xd8000000U,
    0xb4000000U,
    0xda000000U,
    0xa1000000U,
    0xcd800000U,
    0x69400000U,
    0xae600000U,
    0x7ef00000U,
    0x8c380000U,
    0x340c0000U,
    0x9a1e0000U,
    0x10d0000U,
    0xdd8f8000U,
    0xb156c000U,
    0x1a79e000U,
    0xa4e6d000U,
    0x2d38b800U,
    0xf98d5400U,
    0xf3493a00U,
    0xaf669500U,
    0xa3675880U,
    0x3d6a8540U,
    0x2e6003e0U,
    0x3ef00390U,
    0x2c380288U,
    0x240c0014U,
    0x421e033eU,
    0xb50d01f1U,
    0x78f82e7U,
    0x1056c2daU,
    0xd7f9e3bfU,
    0xcda6d08eU,
    0x8358bae1U,
    0x877d552fU,
    0x7f71392eU,
    0x9b6a9711U,
    0x397959f7U,
    0x3c678512U,
    0xf3ef804bU,
    0x8fa6c020U,
    0x3641e198U,
    0x80ead3dcU,
    0x6f26bacaU,
    0x4c80545fU,
    0xf4c6b91eU,
    0xbf305569U,
    0x749eb9bbU,
    0xf0cc5550U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0xc8000000U,
    0x8c000000U,
    0x5e000000U,
    0xbf000000U,
    0xa9800000U,
    0x68400000U,
    0xd7e00000U,
    0x15700000U,
    0xbea80000U,
    0x41c0000U,
    0xb21a0000U,
    0xf11b0000U,
    0x6e9e8000U,
    0x85c34000U,
    0x6dada000U,
    0x23991000U,
    0xbd57d800U,
    0x7b76c400U,
    0x9b35a00U,
    0x819e8700U,
    0x34487880U,
    0x51e8d5c0U,
    0x5e6003a0U,
    0xd3003f0U,
    0x14800d8U,
    0x2d6c020cU,
    0x9ab200feU,
    0xc607016dU,
    0x2b048081U,
    0xa3984308U,
    0x7d532303U,
    0xdb6a51e4U,
    0xb9b27b6dU,
    0x4983d781U,
    0xb8568388U,
    0xfef42c3U,
    0xe17fa044U,
    0xa4ae139dU,
    0x691b59d9U,
    0xfa828544U,
    0x8fd2781dU,
    0x78b3d699U,
    0x2f1e8024U,
    0x118342cdU,
    0x8c4da371U,
    0xb5e913d0U,
    0x3c7fd90fU,
    0x242ac51aU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0xa8000000U,
    0x4000000U,
    0xba000000U,
    0xfb000000U,
    0xaf800000U,
    0x36400000U,
    0xdce00000U,
    0x42700000U,
    0xc8b80000U,
    0x298c0000U,
    0xcb420000U,
    0xce7f0000U,
    0xc6bb8000U,
    0x609ac000U,
    0x6bd4a000U,
    0x45b7d000U,
    0xd3175800U,
    0xeb952400U,
    0x2c56da00U,
    0xd7fce500U,
    0x45fbf880U,
    0xfaeef6c0U,
    0x4f638220U,
    0x7226c2d0U,
    0xa94ea2c8U,
    0xd974d03cU,
    0x7736daf6U,
    0xe7cce725U,
    0x4ba3f925U,
    0x9a12f73cU,
    0x4b19826fU,
    0xa795c3b6U,
    0xc257235cU,
    0xcee11146U,
    0xbd61fb5dU,
    0xdd2df5e1U,
    0xe4c203ceU,
    0xb83f0198U,
    0xba5b82b4U,
    0xd2eac12aU,
    0xb6ca0a3U,
    0x683bd058U,
    0xa255580dU,
    0xdeea2769U,
    0x456d5812U,
    0x8126261eU,
    0xf2cf5bc9U,
    0xfd29261bU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x18000000U,
    0x1c000000U,
    0x96000000U,
    0x9b000000U,
    0x9c800000U,
    0x34c00000U,
    0x3d600000U,
    0x94f00000U,
    0xaba80000U,
    0x60840000U,
    0xd2c20000U,
    0xde790000U,
    0xa4668000U,
    0x117a4000U,
    0xdae62000U,
    0x4ca03000U,
    0xda176800U,
    0x65102400U,
    0x639bea00U,
    0x92576700U,
    0x1cb94b80U,
    0xb20415c0U,
    0x10683e0U,
    0x598a41b0U,
    0x874e23b8U,
    0x72430dcU,
    0x8c55684aU,
    0x93a926cfU,
    0x6c9d6adfU,
    0x8cdd27ecU,
    0xf1776ba9U,
    0xaae026c2U,
    0x34b3eaf0U,
    0x76136518U,
    0xeb1b4accU,
    0xe48d1562U,
    0x98c8017bU,
    0xb3740101U,
    0x13ea03adU,
    0xa13d02bfU,
    0xcf44839cU,
    0x733343f1U,
    0x7e48a3aeU,
    0x32ae7302U,
    0xe51b480bU,
    0xa38d15d9U,
    0xf2480081U,
    0xacb4016dU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x58000000U,
    0x24000000U,
    0xca000000U,
    0xa1000000U,
    0xe2800000U,
    0xe4400000U,
    0x85600000U,
    0x44f00000U,
    0x3f280000U,
    0x62840000U,
    0x245e0000U,
    0xe56d0000U,
    0x14f58000U,
    0x673ec000U,
    0x468de000U,
    0xee4f3000U,
    0x4479f800U,
    0xf665f400U,
    0x837a7a00U,
    0xc3f23700U,
    0xaabc1b80U,
    0x7b5ec640U,
    0x94f582e0U,
    0xa73ec2d0U,
    0x268de3c8U,
    0xbe4f3354U,
    0x1c79f9c2U,
    0xd265f585U,
    0x497a78b7U,
    0x62f23706U,
    0x483c1ae3U,
    0x9f1ec548U,
    0x11958098U,
    0xe3cec1dcU,
    0x19a5e3f6U,
    0xdccb30d7U,
    0x3827f81aU,
    0x3708f435U,
    0x5d8ff8efU,
    0x5ccf6faU,
    0xeb1f8e5U,
    0x7151f627U,
    0x55ec792eU,
    0x15ab34e7U,
    0x9adf9b02U,
    0x1f3907a9U,
    0x929be3f9U,
    0x4c5633fdU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x78000000U,
    0xb4000000U,
    0xfe000000U,
    0x8d000000U,
    0x51800000U,
    0xb5c00000U,
    0xc7a00000U,
    0xd1300000U,
    0x6780000U,
    0xc8940000U,
    0x2a460000U,
    0xf7610000U,
    0x9b0b8000U,
    0x508e4000U,
    0xce54e000U,
    0x11665000U,
    0xb2122800U,
    0x871afc00U,
    0x4287aa00U,
    0x7951bd00U,
    0x5be6cb80U,
    0x174cad40U,
    0xeeed8260U,
    0x92df4390U,
    0xb5276398U,
    0xa07c10a4U,
    0x8180cb5eU,
    0xedddae33U,
    0x3be03a0U,
    0x473502a7U,
    0x476d83c1U,
    0xd31f42f1U,
    0xec87624eU,
    0xec4c116bU,
    0xae78c924U,
    0x2489af89U,
    0x105802baU,
    0xec64010dU,
    0x8b9e0103U,
    0xfec5001fU,
    0xcf358102U,
    0xdb7b424eU,
    0x811961fcU,
    0x3b89124dU,
    0xb6cd48f4U,
    0x7332ec66U,
    0xf9616227U,
    0x3e1d1296U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x8000000U,
    0x84000000U,
    0x7e000000U,
    0x49000000U,
    0x24800000U,
    0x55400000U,
    0xd3600000U,
    0x64f00000U,
    0xf2280000U,
    0x1840000U,
    0x3bda0000U,
    0xd9230000U,
    0x2e0f8000U,
    0x91064000U,
    0xb89a6000U,
    0xd74c7000U,
    0x68661800U,
    0xf374d400U,
    0xb4fb9a00U,
    0xea259700U,
    0xfd9c7b80U,
    0xc9c8a6c0U,
    0x6a3583a0U,
    0x3d9540d0U,
    0xa9dde058U,
    0x1a3e32f4U,
    0x358e794aU,
    0x2ddfa661U,
    0x64280027U,
    0x7c8402e2U,
    0x95a0197U,
    0x31630100U,
    0xafef804aU,
    0x6db640e1U,
    0xc3526267U,
    0xae787002U,
    0x56741a27U,
    0x1a63d728U,
    0x80661ab6U,
    0xc774d447U,
    0xa2fb99d8U,
    0x572595feU,
    0xaf1c7bcbU,
    0x5188a4f6U,
    0xe3d5826dU,
    0x452543c9U,
    0xac15e2d1U,
    0x2a0a3345U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x78000000U,
    0x9c000000U,
    0x6a000000U,
    0xeb000000U,
    0x26800000U,
    0x18c00000U,
    0x76a00000U,
    0x15300000U,
    0x26680000U,
    0x379c0000U,
    0xcd560000U,
    0xa4e10000U,
    0xb9568000U,
    0x4ae9c000U,
    0xd447a000U,
    0x4576f000U,
    0xe511a800U,
    0xfb989c00U,
    0x1f592a00U,
    0x73fc5f00U,
    0x25d60880U,
    0xc12e6c40U,
    0x18688160U,
    0x2294c230U,
    0x5ec722c8U,
    0xffbe329cU,
    0xeca088d2U,
    0x9637adbfU,
    0x44e722e8U,
    0x94e309bU,
    0x42e88b59U,
    0xe05bafa9U,
    0xcb792196U,
    0xf8033369U,
    0x5c000a61U,
    0xca0f6f8aU,
    0x3b1e01fbU,
    0x5e8d03beU,
    0x84c880d2U,
    0x1ca4c368U,
    0xfe2f215bU,
    0xe231f9U,
    0x2f568839U,
    0xbbe6ac4eU,
    0xb1d9a06dU,
    0x9f3bf0d7U,
    0x7d792ab3U,
    0x190c5ce2U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0xf8000000U,
    0x44000000U,
    0xc2000000U,
    0xd000000U,
    0x6b800000U,
    0x4b400000U,
    0xf0e00000U,
    0x9bb00000U,
    0xfb580000U,
    0x38fc0000U,
    0x57be0000U,
    0x555d0000U,
    0x4beb8000U,
    0xb72ec000U,
    0xba84e000U,
    0x96c75000U,
    0xfcbad800U,
    0x61db9400U,
    0x3e3e3a00U,
    0xf31cc500U,
    0xa084e080U,
    0x4fc75240U,
    0xd3ad960U,
    0xb39b9650U,
    0x9f5e3ae8U,
    0x6aecc424U,
    0x2bce226U,
    0xaacb5359U,
    0x3abcdb0dU,
    0xeca9513U,
    0x88b3bac6U,
    0x2bd306a9U,
    0x5f358095U,
    0xe683c2efU,
    0x60d7637cU,
    0x1ba5929aU,
    0xbb583b03U,
    0x98fdc72eU,
    0x87b1630dU,
    0xad449273U,
    0xfedbb96U,
    0x753e05c1U,
    0xb7860071U,
    0xfd5100e9U,
    0xb7ed8095U,
    0x913fc2efU,
    0xa589637cU,
    0x848929aU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0x8000000U,
    0x94000000U,
    0xb6000000U,
    0x5d000000U,
    0xdc800000U,
    0x3ec00000U,
    0x8d600000U,
    0xcbb00000U,
    0x8ad80000U,
    0x8b740000U,
    0x6eae0000U,
    0xda5b0000U,
    0x9fa78000U,
    0x9cca4000U,
    0x26766000U,
    0xaa2fb000U,
    0x78820800U,
    0x30de1400U,
    0xbc746a00U,
    0x2d31a500U,
    0xc3166180U,
    0xa59fb140U,
    0xac5a0ba0U,
    0x62aa1610U,
    0xb05a68f8U,
    0xaaa624U,
    0xbb51e02aU,
    0x9125f0cfU,
    0x51146ae7U,
    0xb681a5c5U,
    0xa1ce60e6U,
    0xaaebb327U,
    0x7cf409bbU,
    0x71f1148bU,
    0x457de8a3U,
    0xffa0e59fU,
    0xccc781a1U,
    0xce7a41ecU,
    0x2e2e6238U,
    0xc69bb1c4U,
    0xf9cc081aU,
    0xd6f51567U,
    0x4eebea9bU,
    0x92ffe45bU,
    0x50f601bbU,
    0xabef008bU,
    0x226982a3U,
    0x5421409fU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0xc8000000U,
    0x14000000U,
    0xd2000000U,
    0x45000000U,
    0x28800000U,
    0xebc00000U,
    0xb6600000U,
    0xc9b00000U,
    0x6dc80000U,
    0x1640000U,
    0xd43e0000U,
    0xf6830000U,
    0x80df8000U,
    0x6df6c000U,
    0xc5e9e000U,
    0x81fe3000U,
    0x7bf4a800U,
    0xbaff5400U,
    0x887d4a00U,
    0x32b16700U,
    0x7e41e180U,
    0x192a3140U,
    0xea02a8a0U,
    0xc9185610U,
    0xc69cca68U,
    0xf8c4a644U,
    0xc1f7826aU,
    0x5be2c3fbU,
    0xcaffe2dbU,
    0xd0693299U,
    0xeebd2b06U,
    0xb85e9667U,
    0x8e3d2b0dU,
    0x879e9664U,
    0xa5d2aa8U,
    0x9b2e97a4U,
    0x871528daU,
    0x658a9583U,
    0x374b29f7U,
    0x1fb99737U,
    0xf2dcabd7U,
    0x98eb57e7U,
    0x556b495fU,
    0x26266793U,
    0xc388628dU,
    0xf04bf324U,
    0x5a22c808U,
    0xb587a6b4U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xf8000000U,
    0xac000000U,
    0x1a000000U,
    0xd1000000U,
    0xd800000U,
    0xfe400000U,
    0x3ea00000U,
    0xccf00000U,
    0x78480000U,
    0xcda40000U,
    0x5c7a0000U,
    0x21810000U,
    0x244e8000U,
    0xfbcc000U,
    0x71606000U,
    0x7e191000U,
    0x5f0cb800U,
    0x8a8b7c00U,
    0x88ccda00U,
    0xe4626f00U,
    0xad886180U,
    0x6e4d11c0U,
    0x96beb860U,
    0x28ee7c70U,
    0x36585988U,
    0xaaafaef4U,
    0x9aee81deU,
    0x34cc17dU,
    0xe92860acU,
    0x3bd10baU,
    0xfb76ba93U,
    0x70a7df1U,
    0xb6825aceU,
    0x3adeac0dU,
    0xd16800acU,
    0xee1401b2U,
    0xf712002fU,
    0x6e9501b3U,
    0xc6dc83a1U,
    0x8369c196U,
    0x6b1ce2e1U,
    0x4c80d1beU,
    0x5bd8590dU,
    0x24efaf24U,
    0xbc4e834eU,
    0xd3bcc1cdU,
    0x736062ccU,
    0xb31912c2U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x8000000U,
    0x84000000U,
    0x3e000000U,
    0xe9000000U,
    0x50800000U,
    0x7ec00000U,
    0x30600000U,
    0xfab00000U,
    0x5cd80000U,
    0xdb740000U,
    0x11320000U,
    0xd1870000U,
    0x1a458000U,
    0x98b54000U,
    0x57d5e000U,
    0xc0eeb000U,
    0x947e8800U,
    0x94b53400U,
    0xadcb6a00U,
    0xf7eb8500U,
    0xddede180U,
    0xb2eab2c0U,
    0x67748ba0U,
    0x33635d0U,
    0x9284ead8U,
    0x65ddc4b4U,
    0xd3f7803aU,
    0xf3f2409fU,
    0xa3f06223U,
    0xbbebf3c1U,
    0xc7f3696aU,
    0x75ef8533U,
    0x26e7e28dU,
    0xa169b374U,
    0x663b0a2eU,
    0x780074f5U,
    0xcc1e8aa4U,
    0x5a0536f6U,
    0x27136b41U,
    0xb19f841eU,
    0xaa5fe029U,
    0x70adb182U,
    0x23d10b6fU,
    0xf6f375ebU,
    0xf9690b0dU,
    0xfa3775b4U,
    0x3a030b8eU,
    0x97047425U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x28000000U,
    0xd4000000U,
    0x42000000U,
    0xb1000000U,
    0xfb800000U,
    0x9cc00000U,
    0xf0e00000U,
    0xd0300000U,
    0x16c80000U,
    0x5f40000U,
    0x21b60000U,
    0xbf070000U,
    0x648b8000U,
    0x854c000U,
    0xe0b16000U,
    0x9c86d000U,
    0x34462800U,
    0x16bfe400U,
    0x3f974a00U,
    0x56c93700U,
    0xa5f96380U,
    0x11b2d240U,
    0xc7102920U,
    0x9888e610U,
    0x9e54cba8U,
    0x13a9f694U,
    0xd61e0166U,
    0x530302f7U,
    0x7a95810dU,
    0x1f57c323U,
    0x9024e230U,
    0xb6d112c3U,
    0x35e2ca3bU,
    0x59aef65cU,
    0x43158339U,
    0xf297c0eeU,
    0xfb44e008U,
    0xaa211044U,
    0xfbcacb2eU,
    0x586af793U,
    0x366b8233U,
    0xf964c218U,
    0x45f96217U,
    0x81b2d07dU,
    0x8f1029bbU,
    0x1c88e41cU,
    0xf454ca19U,
    0x76a9f6feU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x38000000U,
    0xf4000000U,
    0xee000000U,
    0x1f000000U,
    0xe4800000U,
    0x39400000U,
    0xea600000U,
    0x2f300000U,
    0xa7580000U,
    0x6d740000U,
    0x6fba0000U,
    0xf8090000U,
    0x540d8000U,
    0xbe094000U,
    0x970f2000U,
    0x2886f000U,
    0x2358f800U,
    0x1b644400U,
    0xd4b7da00U,
    0x7a92b500U,
    0xbe572180U,
    0xaaf2f3c0U,
    0x7062fb60U,
    0x9e2d4510U,
    0xbcda5958U,
    0xabf6acU,
    0x848001aaU,
    0xc9400359U,
    0x3260035bU,
    0x6b300149U,
    0x7158026cU,
    0x867402a5U,
    0x653a0309U,
    0xde4902ccU,
    0x5aed80d5U,
    0xa8794141U,
    0xda372238U,
    0x6ac2f353U,
    0xebbaf9f2U,
    0x8e19469aU,
    0xef005931U,
    0x3c92f59fU,
    0x7d558327U,
    0x3c7d42dbU,
    0xc4352289U,
    0xadcff10cU,
    0x4b357bb5U,
    0x615d0451U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0x98000000U,
    0x14000000U,
    0x6a000000U,
    0xc9000000U,
    0x4e800000U,
    0xdd400000U,
    0x84200000U,
    0x53700000U,
    0xe580000U,
    0xdbbc0000U,
    0x74ba0000U,
    0xcd350000U,
    0x5dfd8000U,
    0x731dc000U,
    0xaf856000U,
    0xbfcb7000U,
    0xd770b800U,
    0xec5ecc00U,
    0xceadda00U,
    0x2c29bd00U,
    0x3f676080U,
    0x204272c0U,
    0x10b738e0U,
    0x7f360f90U,
    0x20f539c8U,
    0xa78f0e74U,
    0x23cab812U,
    0x616bce91U,
    0x335059caU,
    0xaf347ca8U,
    0x8e20161U,
    0x8b890397U,
    0xadc78199U,
    0x5a68c39bU,
    0xa0d8e047U,
    0x56e6b3b4U,
    0x988dd972U,
    0x1259bd41U,
    0x2dbf6122U,
    0xbbe722cU,
    0x5cad386bU,
    0xe1330c1aU,
    0xd3f0b985U,
    0x481ece08U,
    0x3c0dd8d1U,
    0x4619bf2fU,
    0x471f60b5U,
    0x758e72b5U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0x8000000U,
    0x8c000000U,
    0x66000000U,
    0x71000000U,
    0x7c800000U,
    0x6a400000U,
    0x7e200000U,
    0xa2700000U,
    0x85480000U,
    0xa7a40000U,
    0xb6be0000U,
    0x9a3b0000U,
    0x18648000U,
    0x7a5b4000U,
    0xb635e000U,
    0xce6f3000U,
    0x73449800U,
    0xdeb4ec00U,
    0x46397a00U,
    0x967fdf00U,
    0x1743e180U,
    0xa4b03140U,
    0x593e1ba0U,
    0x17e4ac70U,
    0x67801bb8U,
    0xc1dfacf4U,
    0xf9e499eaU,
    0x5a84ed2dU,
    0x3b5179aaU,
    0x72abdc50U,
    0x5035e3d5U,
    0x7f6f32a3U,
    0xefc49b05U,
    0x24f4ecebU,
    0x30197909U,
    0xb80fdc55U,
    0xf40be0beU,
    0x721433eaU,
    0x93001870U,
    0xe79fafe5U,
    0x1c49a3bU,
    0x19f4ee41U,
    0xca997ad9U,
    0x334fdfc0U,
    0xfeabe13dU,
    0x3624315fU,
    0xe681b43U,
    0x934bad08U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0xa8000000U,
    0xd4000000U,
    0x4a000000U,
    0x9b000000U,
    0x3b800000U,
    0x56400000U,
    0x4ee00000U,
    0xdab00000U,
    0x6b480000U,
    0xc86c0000U,
    0xca6a0000U,
    0x95730000U,
    0x9cf38000U,
    0x5da1c000U,
    0x1edd2000U,
    0x97399000U,
    0x58017800U,
    0xbc00b400U,
    0x3e145a00U,
    0xa1152700U,
    0x89f2080U,
    0xb9d692c0U,
    0x52b0f820U,
    0xf4e7690U,
    0x8a78f988U,
    0xf562763cU,
    0x4cf2f946U,
    0x85a175f5U,
    0x62c97bf5U,
    0x92cb7a3U,
    0x891e5b9eU,
    0x1c9627b7U,
    0x53c4a162U,
    0xb9ab53d1U,
    0x9ccfd987U,
    0x828e5ecU,
    0xf18003f8U,
    0xd400004U,
    0xd56001e2U,
    0xfcf00047U,
    0x8da8021aU,
    0xc6dc0115U,
    0xeb220045U,
    0xc61f033bU,
    0x6d19806aU,
    0x9e92c32dU,
    0xcccea2e1U,
    0x10285189U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0x78000000U,
    0x7c000000U,
    0xda000000U,
    0x45000000U,
    0xfb800000U,
    0x94c00000U,
    0xeae00000U,
    0x4c300000U,
    0xf7c80000U,
    0x4e6c0000U,
    0xc66a0000U,
    0x2690000U,
    0x47c8000U,
    0xeb66c000U,
    0x2deee000U,
    0x3ea55000U,
    0x9a842800U,
    0x795ea400U,
    0x3922ca00U,
    0x1457f700U,
    0x76ace080U,
    0xfe9052c0U,
    0xf5aab60U,
    0xae3d6430U,
    0xaedaa888U,
    0xbffd67e4U,
    0x1fbaa806U,
    0x170d6587U,
    0x7a92a925U,
    0x69516453U,
    0x9130aa7cU,
    0x6054662dU,
    0xa8a62920U,
    0x1d9ba547U,
    0x6bd44945U,
    0x846836e3U,
    0x2b768134U,
    0x8dffc329U,
    0x4eba6256U,
    0xe29f9268U,
    0x548c894U,
    0xe33ef4aeU,
    0x51506073U,
    0x8d3690bbU,
    0x6a544928U,
    0xe5a83463U,
    0xe2168023U,
    0x590fc054U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0xc8000000U,
    0x8c000000U,
    0x82000000U,
    0x83000000U,
    0xc9800000U,
    0xb0400000U,
    0x3ee00000U,
    0xc8b00000U,
    0xbb480000U,
    0xdb740000U,
    0x8afe0000U,
    0xdeb30000U,
    0x9e4c8000U,
    0x8ff7c000U,
    0x3a272000U,
    0x968bd000U,
    0x93c3f800U,
    0xf1ae7400U,
    0x65ccda00U,
    0x4a1a700U,
    0x99592380U,
    0x2878d340U,
    0x9b6f7aa0U,
    0x2ae9b670U,
    0xaea3fb28U,
    0x465e773cU,
    0xcbe4daaeU,
    0x3425a78dU,
    0x978f237dU,
    0xd94fd093U,
    0x8875fabeU,
    0xeb697433U,
    0xf2fe58c8U,
    0xeaa565eaU,
    0x48528207U,
    0xcaf4c1eaU,
    0x7ea3a101U,
    0xee48109bU,
    0x57fadbb2U,
    0x7e26a695U,
    0x988ba249U,
    0x92cc1231U,
    0xbb2cd915U,
    0x1c11a60fU,
    0x8a1123e0U,
    0x6f0cd2d6U,
    0x5b9178a9U,
    0xfb5ab667U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x98000000U,
    0x84000000U,
    0x16000000U,
    0x99000000U,
    0x16800000U,
    0x9e400000U,
    0x3e600000U,
    0x9a300000U,
    0x17480000U,
    0xe40000U,
    0xf8620000U,
    0xcb350000U,
    0x8dd78000U,
    0xe4b84000U,
    0x5500e000U,
    0x4c847000U,
    0x9d4c5800U,
    0xcbfb6400U,
    0x81e4ba00U,
    0xaaeb1700U,
    0xa362e180U,
    0x5ab172c0U,
    0xf01bd9e0U,
    0x580327d0U,
    0x64045838U,
    0x661f6774U,
    0x106b982U,
    0x929e16ddU,
    0x88556343U,
    0xa77930c1U,
    0x8cb3391cU,
    0x891357dfU,
    0x3e8203deU,
    0x62450222U,
    0xdc7f82adU,
    0x8d2c430bU,
    0x1ccae1fdU,
    0x9e2570a2U,
    0xc151d83cU,
    0x79e2253eU,
    0x9ef9dba3U,
    0xcd762440U,
    0x37b3d971U,
    0x689726e5U,
    0xdb4e5be6U,
    0x5afe6556U,
    0xfb7b3b2fU,
    0x3eb755d6U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x58000000U,
    0xc000000U,
    0x3a000000U,
    0x5d000000U,
    0xf3800000U,
    0x8400000U,
    0x57e00000U,
    0x81b00000U,
    0x75480000U,
    0x54740000U,
    0xe1f60000U,
    0x26a90000U,
    0xbbce8000U,
    0xffb8c000U,
    0xe651e000U,
    0xfced5000U,
    0x752eb800U,
    0x83849400U,
    0xa0575a00U,
    0x33edc700U,
    0xbfa7e380U,
    0x46445240U,
    0xcce038e0U,
    0x7d3c5570U,
    0xd786bb88U,
    0x96409644U,
    0x54e958a6U,
    0x1130c715U,
    0xbd9f6071U,
    0x93559057U,
    0xab7f5882U,
    0x2369c753U,
    0xb779e214U,
    0xe16953aeU,
    0xd670ba91U,
    0x20e99577U,
    0xd727d952U,
    0xb288044bU,
    0x39ce8228U,
    0x3eb8c104U,
    0x17d1e0c6U,
    0xf5ad51a5U,
    0xb34eba59U,
    0x5b749603U,
    0x4b7f599cU,
    0xb369c5eaU,
    0x8f79e237U,
    0xbd695262U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0xa8000000U,
    0xd4000000U,
    0xd6000000U,
    0xb3000000U,
    0xc1800000U,
    0xe7c00000U,
    0xdba00000U,
    0xf700000U,
    0x55580000U,
    0xdde40000U,
    0xa3020000U,
    0xd9830000U,
    0xfbc38000U,
    0x29be4000U,
    0xc269a000U,
    0xf3ccd000U,
    0x2dba0800U,
    0x2c73ec00U,
    0xccd1aa00U,
    0x463c3d00U,
    0x7aa82380U,
    0xb3f192c0U,
    0xdc102b20U,
    0xd2017c90U,
    0x5d0203f8U,
    0xfe83009cU,
    0x8c4381deU,
    0x8d7e4293U,
    0x7049a2c8U,
    0xcf7cd0f2U,
    0x75420963U,
    0x4de7ed86U,
    0xfb0baadfU,
    0xa59b3e8eU,
    0xf9c9a18dU,
    0x4cbcd13dU,
    0xb0e20a93U,
    0xd597ec8eU,
    0x11d3aacbU,
    0xf8bf3c84U,
    0x96eba3b4U,
    0xce8fd29cU,
    0x4598a98U,
    0xc97daf2aU,
    0xfe400befU,
    0x64eee0U,
    0xb6c82930U,
    0xcf257e28U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0xa8000000U,
    0xb4000000U,
    0x66000000U,
    0x93000000U,
    0x57800000U,
    0x4f400000U,
    0x2d200000U,
    0x99700000U,
    0x77d80000U,
    0xd7e40000U,
    0xc6120000U,
    0xc3130000U,
    0xef8f8000U,
    0x23544000U,
    0x572ba000U,
    0xd87e1000U,
    0xd557d800U,
    0x5c3a2c00U,
    0xf3ee7a00U,
    0x38173d00U,
    0x2c162380U,
    0x1a0950c0U,
    0x310bfb20U,
    0xa847f10U,
    0x3fd80138U,
    0x53e4019cU,
    0x68120146U,
    0x941302b1U,
    0x760f834aU,
    0x4b144188U,
    0x4b8ba0f3U,
    0x9d4e1376U,
    0xd82fdb99U,
    0x5dee2d6eU,
    0x6f047a12U,
    0xb5903ce4U,
    0x7253a0bdU,
    0x2daa1073U,
    0x4fbdda31U,
    0x32bd2e0dU,
    0x522bfa1cU,
    0xf4f47c81U,
    0x19800165U,
    0x28400058U,
    0x7ca001ebU,
    0x3530017aU,
    0xa57801a7U,
    0xb5d40023U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0x38000000U,
    0x34000000U,
    0x1e000000U,
    0xf3000000U,
    0x86800000U,
    0xc6c00000U,
    0xf5a00000U,
    0x68700000U,
    0x7580000U,
    0xebec0000U,
    0x881a0000U,
    0x6c010000U,
    0x7a058000U,
    0xa50b4000U,
    0x799aa000U,
    0x6a483000U,
    0xde6bc800U,
    0xe8480c00U,
    0x2f6b6a00U,
    0xdfc13d00U,
    0xfe252280U,
    0x8b3270c0U,
    0xfaace860U,
    0xbce77d10U,
    0x38800298U,
    0x5c0021cU,
    0x6b2002e2U,
    0xeab00117U,
    0xd4f8012cU,
    0x449c026aU,
    0x17c203b3U,
    0xb22d01c2U,
    0x813f8027U,
    0x67ba4104U,
    0xf167201eU,
    0x4cdf7285U,
    0x28b36813U,
    0x5ed3c92U,
    0x31f22bfU,
    0xfe8372b8U,
    0xd2d16aacU,
    0x9bb03f8aU,
    0xa378a043U,
    0xb5d5312aU,
    0x332c4b93U,
    0x8eae4d72U,
    0x82ee4b4fU,
    0xbb834f50U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x98000000U,
    0x84000000U,
    0x72000000U,
    0x73000000U,
    0xdd800000U,
    0x86400000U,
    0x4e600000U,
    0xff300000U,
    0x3ed80000U,
    0x4a40000U,
    0x61960000U,
    0x510000U,
    0xe76f8000U,
    0x1da4c000U,
    0xb0b2000U,
    0x49815000U,
    0xc433800U,
    0xe96e6400U,
    0xc8be1a00U,
    0x4f8e3500U,
    0xe54aa380U,
    0x6be09040U,
    0xad699be0U,
    0x9abef750U,
    0xc8f81a8U,
    0xf0d4c2dcU,
    0xf1b3233eU,
    0x15155295U,
    0x648d382fU,
    0x9cdb67f5U,
    0xafa79a0aU,
    0x580bf722U,
    0xa416018bU,
    0x4211024aU,
    0xbb0f80c2U,
    0xc194c0dbU,
    0x705320e2U,
    0x4f65529eU,
    0x51b53b25U,
    0x650f6557U,
    0xcc8998c1U,
    0xd0cef608U,
    0xc1b782d9U,
    0xdd00c219U,
    0x789d238cU,
    0x6ad05323U,
    0xaeacb816U,
    0xf68aa73cU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x18000000U,
    0x9c000000U,
    0x32000000U,
    0x3d000000U,
    0x5c800000U,
    0xb3400000U,
    0x83200000U,
    0x59700000U,
    0x90c80000U,
    0x19f40000U,
    0xd18e0000U,
    0xc7c90000U,
    0x447c8000U,
    0xfc57c000U,
    0x42bae000U,
    0x46a3b000U,
    0x28a1b800U,
    0x47a6ac00U,
    0xb6355a00U,
    0x41fc1f00U,
    0xed806380U,
    0x5c97140U,
    0xf169d9e0U,
    0x24dbdc30U,
    0x5ff282a8U,
    0xca9ec184U,
    0x1046603aU,
    0x38b47079U,
    0x9fbb5b1eU,
    0x4a351dd8U,
    0xe3fce287U,
    0xc89eb2f6U,
    0xc55339f7U,
    0x70386ff5U,
    0x9af33a3bU,
    0x5a086cfbU,
    0xe91b3910U,
    0x8a8c6f38U,
    0xb05d387cU,
    0xc8b16ee6U,
    0x17afba4fU,
    0xce2fad49U,
    0x4de9d87dU,
    0xc79bdd64U,
    0xa4d283c1U,
    0x9feec2e9U,
    0xaa8e6166U,
    0x80407244U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x38000000U,
    0x94000000U,
    0xd6000000U,
    0x99000000U,
    0x3800000U,
    0x43400000U,
    0xc5a00000U,
    0x59f00000U,
    0x2ec80000U,
    0x4be40000U,
    0xa9ca0000U,
    0x9d790000U,
    0x9b198000U,
    0x6c964000U,
    0x9d2e000U,
    0x8d67d000U,
    0x7305d800U,
    0xf0897c00U,
    0x73d53a00U,
    0x5673af00U,
    0x3f836180U,
    0x295593c0U,
    0xf6bd3b60U,
    0x8967ae90U,
    0x8d0161b8U,
    0x15889334U,
    0x3a4eb8d6U,
    0xf638efa9U,
    0x22a201f8U,
    0x7f6d013aU,
    0xc41b8375U,
    0x1e0b42ecU,
    0xd50162f2U,
    0x71889007U,
    0x344eb885U,
    0xcb38ecc4U,
    0xcf2200feU,
    0x312d00a5U,
    0xd43b83daU,
    0x9dbb4225U,
    0x3de96234U,
    0x20dc9056U,
    0x76ecb8e9U,
    0x4455efd8U,
    0xd339828aU,
    0x8b2642fdU,
    0xef3ae090U,
    0xe133d0b8U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x38000000U,
    0xa4000000U,
    0xb6000000U,
    0x29000000U,
    0x48800000U,
    0x7a400000U,
    0x9ba00000U,
    0x16f00000U,
    0xd80000U,
    0xeef40000U,
    0x4ca0000U,
    0x88e90000U,
    0x25df8000U,
    0x5c6c4000U,
    0x4d98e000U,
    0x58d6b000U,
    0x3aeb9800U,
    0x6adb9c00U,
    0x35e17a00U,
    0xe3502d00U,
    0xab3f6180U,
    0x28bef1c0U,
    0x1d617b60U,
    0xe9102e50U,
    0xe89f6038U,
    0xaa4ef334U,
    0x93b9787aU,
    0x8ae42d8dU,
    0x12d561acU,
    0x71e7f222U,
    0x6546fb35U,
    0xba386f84U,
    0xc43583d2U,
    0xd13543e1U,
    0x5bbf63b2U,
    0xb6fef36dU,
    0xd0c17860U,
    0xe6e02cd0U,
    0x98c76378U,
    0x9afaf314U,
    0xbad37a0aU,
    0x3dfd2dc5U,
    0x7f52e3d0U,
    0xb93fb3a4U,
    0xb7b41bfeU,
    0x7cf7dc03U,
    0xdbd999e7U,
    0x9769fb9U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x38000000U,
    0x1c000000U,
    0xd2000000U,
    0x51000000U,
    0xc9800000U,
    0x65c00000U,
    0x23600000U,
    0x21b00000U,
    0xb4580000U,
    0x8abc0000U,
    0x48d60000U,
    0x48fd0000U,
    0x1d608000U,
    0xeab3c000U,
    0x38db6000U,
    0x40fe9000U,
    0xa97ec800U,
    0x1cac7400U,
    0xa7cbaa00U,
    0xa63e500U,
    0x543be080U,
    0x938d5340U,
    0xc0c5aa60U,
    0xbce2e630U,
    0x4b6d61a8U,
    0xe5b39034U,
    0x4a464aeeU,
    0xa1a3b693U,
    0xf446c94bU,
    0x2aa07461U,
    0xd8c5a994U,
    0x70e2e4a8U,
    0x16d60b4U,
    0x38b391aeU,
    0x69c648f3U,
    0x8963b57bU,
    0xcca6c8c9U,
    0x3fd07520U,
    0x867dab06U,
    0xbe2ee747U,
    0xde836155U,
    0xdb4293faU,
    0x8828c97bU,
    0xa191779fU,
    0xa1cb29ffU,
    0xdd6025cfU,
    0xab88167U,
    0x8cfc1d3U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0xd8000000U,
    0xdc000000U,
    0x3a000000U,
    0x4d000000U,
    0x60800000U,
    0xf0c00000U,
    0x76a00000U,
    0xe5700000U,
    0xc0580000U,
    0xfe740000U,
    0x3fde0000U,
    0x57210000U,
    0x772c8000U,
    0x4736c000U,
    0xef31e000U,
    0x7b283000U,
    0x45384800U,
    0xee2fec00U,
    0xf9afaa00U,
    0x24e2dd00U,
    0xc21d6180U,
    0xa11ef340U,
    0xf289a8e0U,
    0x29c7dd30U,
    0x2837e188U,
    0xbebd3344U,
    0xa16acaeaU,
    0xc6482cb7U,
    0x196ac938U,
    0xca482e26U,
    0x1b6acbb3U,
    0xcb482e18U,
    0x99eac89cU,
    0xaa882d56U,
    0xb5cacaf1U,
    0xf2382ed1U,
    0x63b2caebU,
    0x19fc2dceU,
    0xea94c807U,
    0x55d92e5aU,
    0x623e4a7fU,
    0xbbbaec9cU,
    0xc5fd2afcU,
    0xd0851c0cU,
    0x18d801a4U,
    0x2b40370U,
    0x4b7e0142U,
    0xb35100c9U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0xc8000000U,
    0xcc000000U,
    0xd6000000U,
    0x3f000000U,
    0xc9800000U,
    0xe5400000U,
    0x65e00000U,
    0xb9300000U,
    0x11980000U,
    0x915c0000U,
    0x17f20000U,
    0xac2f0000U,
    0xf9188000U,
    0x4e9f4000U,
    0x88cf2000U,
    0xaaa1d000U,
    0xfdd51800U,
    0x923b5400U,
    0x7a02ba00U,
    0x3905c500U,
    0xae988080U,
    0x18df41c0U,
    0x52af21a0U,
    0xf9d1d3f0U,
    0x882d1878U,
    0x931757ccU,
    0xcf88ba7eU,
    0x8246c669U,
    0x986a0231U,
    0x8e7300b1U,
    0x516a8071U,
    0x8f04351U,
    0x15b7a061U,
    0x744e9339U,
    0xd7623bc5U,
    0x2ff684e3U,
    0x483da176U,
    0x730d93e1U,
    0x5f90ba85U,
    0x7a5ac483U,
    0x9c780326U,
    0x946c0069U,
    0xb86a0131U,
    0xfe730331U,
    0x396a8231U,
    0xf4f040b1U,
    0xbb7a071U,
    0x874e9351U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x38000000U,
    0xf4000000U,
    0x26000000U,
    0x33000000U,
    0xe4800000U,
    0x95c00000U,
    0x42e00000U,
    0xbbb00000U,
    0xe0980000U,
    0x2bd40000U,
    0x55fa0000U,
    0x71310000U,
    0x8a418000U,
    0x5fae4000U,
    0x6e88a000U,
    0x4d3f000U,
    0x6b6ba800U,
    0x3df6a400U,
    0xcd228a00U,
    0x804b1500U,
    0x8ea18180U,
    0xa71e41c0U,
    0xb290a360U,
    0x7ec7f310U,
    0x6271a878U,
    0x3077a71cU,
    0x657b090aU,
    0x52f1579bU,
    0x13b32319U,
    0xfc8cb179U,
    0xf1c289e9U,
    0x8cfb1551U,
    0x74b980adU,
    0xee0a4377U,
    0x5f0aa3f4U,
    0xe686f2e1U,
    0x88c829eaU,
    0x597de604U,
    0x18f1a806U,
    0x22b7a769U,
    0x51b0a5eU,
    0x7f81540aU,
    0x6d4b221bU,
    0xd28b0d9U,
    0x20408b19U,
    0x5eae1579U,
    0xaf1a03e9U,
    0x7e810051U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0xb8000000U,
    0xcc000000U,
    0xd2000000U,
    0x99000000U,
    0x77800000U,
    0x35c00000U,
    0xe6e00000U,
    0x50b00000U,
    0x5d880000U,
    0x40c40000U,
    0xb37e0000U,
    0x54650000U,
    0xb8e68000U,
    0x3bb7c000U,
    0x431f6000U,
    0xaa8cd000U,
    0x1458c800U,
    0xfc2ca400U,
    0xfec12a00U,
    0xc867b700U,
    0xf2ee8180U,
    0x9eb3c340U,
    0x4e816160U,
    0x7259d370U,
    0x9b3649a8U,
    0xd25f6554U,
    0xab204896U,
    0x9a4e65b3U,
    0xcf30cbc5U,
    0x3c58a6d5U,
    0x4837298dU,
    0xc6b431U,
    0x93760393U,
    0xa4610246U,
    0x10f883d8U,
    0x4fa2c07cU,
    0x5d11e002U,
    0xe18f1285U,
    0xfad1a926U,
    0xbe7177a8U,
    0x2de96054U,
    0x7e2dd216U,
    0xffc04973U,
    0x83fe64e5U,
    0xbd38ca45U,
    0x955ca595U,
    0x77a928edU,
    0x5113b741U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0xd8000000U,
    0x5c000000U,
    0x9e000000U,
    0x31000000U,
    0xf7800000U,
    0xb400000U,
    0xf1e00000U,
    0x29300000U,
    0x61980000U,
    0xfe440000U,
    0xa4660000U,
    0xdd610000U,
    0x8ee88000U,
    0x77ac4000U,
    0xa14be000U,
    0xcaedd000U,
    0x95b91800U,
    0x3e4e1400U,
    0x847a7a00U,
    0xed7f8500U,
    0x6f08380U,
    0xf3a84040U,
    0x634de2e0U,
    0x65fcd030U,
    0x53299918U,
    0xc29655b4U,
    0x7ecf9ba6U,
    0x35b755e9U,
    0x4e471b99U,
    0x6c6b14c9U,
    0x3974faa1U,
    0x1cf2c66dU,
    0xb3e18fU,
    0x3bd9d178U,
    0xa8271becU,
    0x5e1b1672U,
    0x110cfbbfU,
    0xc786c6c8U,
    0x834de0b4U,
    0x75fcd026U,
    0xeb2999a9U,
    0xce965779U,
    0x38cf9bf9U,
    0x58b75539U,
    0x27c71819U,
    0x562b1489U,
    0x3f14f841U,
    0x3e82c65dU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x58000000U,
    0xbc000000U,
    0x6e000000U,
    0x85000000U,
    0xf7800000U,
    0xf9c00000U,
    0xb6a00000U,
    0x9ef00000U,
    0x40980000U,
    0x354c0000U,
    0xf5f20000U,
    0x720d0000U,
    0x1b148000U,
    0xa9e4000U,
    0x24b6000U,
    0x79655000U,
    0x11464800U,
    0x47ec2c00U,
    0x901aa00U,
    0x819b3d00U,
    0xe0de8280U,
    0x9f2f43c0U,
    0xba35e1e0U,
    0xbdba1030U,
    0x3c6ba8d8U,
    0x86da3ce4U,
    0x2e380066U,
    0x57bc00f3U,
    0xfb6a0184U,
    0x7241035cU,
    0x416682b8U,
    0x3d53405eU,
    0xc1ffe36dU,
    0x680b1309U,
    0xa4152865U,
    0xf2057e05U,
    0x5b15e0bfU,
    0x2a8a12b4U,
    0xb253a9ceU,
    0x21663eefU,
    0xad5200d6U,
    0x29fd0121U,
    0x8c0c82f3U,
    0x76124084U,
    0x191963dcU,
    0x29985078U,
    0x24cac93eU,
    0xfd3e6d9dU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0xd8000000U,
    0xcc000000U,
    0xba000000U,
    0x11000000U,
    0xb1800000U,
    0xbec00000U,
    0xd1600000U,
    0x98300000U,
    0xa4980000U,
    0x11440000U,
    0xb0b20000U,
    0x6fd50000U,
    0x40ea8000U,
    0xb6f2c000U,
    0x5df36000U,
    0x9d7b1000U,
    0x62326800U,
    0xd58bb400U,
    0x50d38a00U,
    0xd6766500U,
    0x72b88380U,
    0x62d7c2c0U,
    0x361e2e0U,
    0xfd3dd0f0U,
    0x930b0988U,
    0xdc91a774U,
    0xd41622eU,
    0x42ae13bfU,
    0x9ad8e803U,
    0x5f79775bU,
    0x6f20e8b7U,
    0x960d7415U,
    0x1b0ae938U,
    0x189c777eU,
    0x335269e7U,
    0xdbbb46fU,
    0x944b8939U,
    0x3732677aU,
    0x1a0a82a3U,
    0xc102c229U,
    0xf98b6382U,
    0x5acf119fU,
    0x7f786b13U,
    0xff2ab7a3U,
    0xbe13098bU,
    0xf15a42fU,
    0x6e936099U,
    0x984b12aaU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x8000000U,
    0xc000000U,
    0x9e000000U,
    0x77000000U,
    0xf9800000U,
    0x4400000U,
    0x7c600000U,
    0x83b00000U,
    0xea880000U,
    0xbfdc0000U,
    0x89220000U,
    0x45570000U,
    0xef88000U,
    0xc8664000U,
    0xd9b32000U,
    0x2f9cd000U,
    0x9f443800U,
    0xbffc400U,
    0xc8e79a00U,
    0x6b695700U,
    0xaa3a8280U,
    0xf6c143c0U,
    0xf7a3a0a0U,
    0x909690f0U,
    0x4add1a58U,
    0xa1a815ecU,
    0xcb99239eU,
    0x1d57d2cdU,
    0xdafeba97U,
    0x427e86dfU,
    0x4a4387bU,
    0x3b0fc7c1U,
    0x878f9976U,
    0x634555f6U,
    0xbdf08359U,
    0x83fa41d5U,
    0x4f12074U,
    0xd57bd0bdU,
    0xed34bb0fU,
    0x7458793U,
    0xfff6b915U,
    0x72e28454U,
    0xbe663a8dU,
    0xb2a8c677U,
    0x701f18cfU,
    0xe80f1733U,
    0x1c09a1e5U,
    0xa61d920cU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x78000000U,
    0xac000000U,
    0x5a000000U,
    0x11000000U,
    0x53800000U,
    0xf3400000U,
    0x63e00000U,
    0xed300000U,
    0xca980000U,
    0x4cc0000U,
    0x2ea60000U,
    0x1dd0000U,
    0xeb328000U,
    0x898ac000U,
    0xa2436000U,
    0x10635000U,
    0xe739800U,
    0xd16e5400U,
    0x45fa7a00U,
    0xbe3bc500U,
    0x14148280U,
    0x9617c140U,
    0x7b11e160U,
    0x2a9992f0U,
    0x74c8f888U,
    0x66b106f4U,
    0xc5d7e226U,
    0x65349289U,
    0x6e827841U,
    0xbac7c439U,
    0xa1aa8145U,
    0xcd46c017U,
    0xace56278U,
    0x1be5010U,
    0x9d411bb8U,
    0xf4e4975cU,
    0xbdb91bc2U,
    0xbf589497U,
    0x49e71a54U,
    0xb439959aU,
    0x5d0b9b7bU,
    0x799255feU,
    0xaa447bc9U,
    0xf46ac521U,
    0x906000c9U,
    0x4e7002cdU,
    0xf1780363U,
    0x55fc029eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0xd8000000U,
    0xc4000000U,
    0xb2000000U,
    0x1d000000U,
    0xf1800000U,
    0xe4400000U,
    0xfce00000U,
    0x34300000U,
    0xbd180000U,
    0x818c0000U,
    0x1c420000U,
    0x8fd0000U,
    0x5e338000U,
    0x6416c000U,
    0xc219e000U,
    0xe51f9000U,
    0x59fb800U,
    0x8e409400U,
    0x25f7da00U,
    0x77b4c500U,
    0x44420280U,
    0x8cfd01c0U,
    0xcc3383e0U,
    0x4916c3d0U,
    0xeb99e1a8U,
    0xc55f9364U,
    0x4b7fb90eU,
    0xa77097e5U,
    0x696fda35U,
    0x1278c641U,
    0xa4e00387U,
    0xb0300316U,
    0x2f180165U,
    0xac8c01f5U,
    0x35c200a1U,
    0x28bd02d7U,
    0x10d3837eU,
    0x4d26c1e1U,
    0x8e81e32bU,
    0x80d392ecU,
    0xe53db986U,
    0xb28d9531U,
    0xc6dc5903U,
    0x922e04c8U,
    0x9a19e1a8U,
    0x611f9364U,
    0x979fb90eU,
    0xa34097e5U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x28000000U,
    0x2c000000U,
    0x96000000U,
    0x7000000U,
    0x14800000U,
    0xd4c00000U,
    0x4c600000U,
    0x23300000U,
    0xc3180000U,
    0x8e840000U,
    0x15ca0000U,
    0x47e30000U,
    0xe7708000U,
    0xe1af4000U,
    0x70c62000U,
    0xe6607000U,
    0xca270800U,
    0xe48d9400U,
    0xdcdbaa00U,
    0x3061a500U,
    0xad2a0380U,
    0xc01301c0U,
    0x20088120U,
    0x501b4070U,
    0x181422f8U,
    0x40770acU,
    0xba1d8a0aU,
    0x9101d74fU,
    0x138d0b63U,
    0xc05e96dfU,
    0x98b329bdU,
    0x6f4ae776U,
    0xe026207bU,
    0x4d907255U,
    0x9b5f08b2U,
    0x523996f5U,
    0xa089aaf4U,
    0x6c6a6e0U,
    0x91708250U,
    0x96af4208U,
    0x2c462314U,
    0x2ea072c6U,
    0x38470955U,
    0xecbd94c4U,
    0x9d43aaf8U,
    0x6d25a4acU,
    0xe000000aU,
    0x7000024fU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x38000000U,
    0xe4000000U,
    0x4e000000U,
    0x7b000000U,
    0x80800000U,
    0x46c00000U,
    0x3f600000U,
    0xda300000U,
    0x17080000U,
    0xd29c0000U,
    0xb7c20000U,
    0xbae10000U,
    0x5f6b8000U,
    0xa3d4000U,
    0x3f0aa000U,
    0x3e857000U,
    0x25ddc800U,
    0x6bf2b400U,
    0xeafeea00U,
    0xb76b8700U,
    0xc6220280U,
    0x9d1103c0U,
    0xd7838060U,
    0x74514350U,
    0x98a0a1d8U,
    0x68c8727cU,
    0x947c4b3aU,
    0x72b2f4f7U,
    0xbdddc8ffU,
    0x7ff2b573U,
    0xbcfeea21U,
    0x186b87e2U,
    0x30a200e3U,
    0x44d101f1U,
    0x266382faU,
    0x93a1407fU,
    0x3048a25bU,
    0x26a473b5U,
    0xbd64bacU,
    0xc0fff6caU,
    0x427c4b9fU,
    0x1db2f54bU,
    0x6b5dc84dU,
    0x9632b5a0U,
    0x751ee9d8U,
    0x1b9b867cU,
    0xd64a013aU,
    0x71bd03f7U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0xd8000000U,
    0x5c000000U,
    0xe2000000U,
    0xe5000000U,
    0x5d800000U,
    0x4ac00000U,
    0x10a00000U,
    0xcaf00000U,
    0xe9080000U,
    0x17940000U,
    0x4bda0000U,
    0x432d0000U,
    0x5f208000U,
    0x1d3fc000U,
    0x2824e000U,
    0x1da51000U,
    0xd36be800U,
    0x7dc9cc00U,
    0xb0358a00U,
    0xe1be1f00U,
    0xe17a0380U,
    0xf0dd01c0U,
    0x69a882e0U,
    0x156bc3b0U,
    0xf6dee0e8U,
    0x82b812e4U,
    0x97e36afeU,
    0x58920ebfU,
    0x67436b34U,
    0xe2620d00U,
    0x364b6866U,
    0x19f60eb5U,
    0x47916a6bU,
    0xe3db0de2U,
    0xa731eb3bU,
    0x5124ce5cU,
    0xc2350bc2U,
    0xccb1de0dU,
    0xe8f6e1b1U,
    0x6c1c10a5U,
    0xfa116a75U,
    0xd91b0c0bU,
    0xf91e812U,
    0x77d4cef3U,
    0x113d08e8U,
    0x6225dee4U,
    0x1cace0feU,
    0x80f111bfU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0x48000000U,
    0xfc000000U,
    0xca000000U,
    0x4d000000U,
    0xe1800000U,
    0x2a400000U,
    0x2a200000U,
    0xf0f00000U,
    0x2b180000U,
    0x6e840000U,
    0x1ad60000U,
    0xfbf30000U,
    0xf5868000U,
    0xec4ec000U,
    0x553ce000U,
    0x387f9000U,
    0x9ec87800U,
    0x15fb8c00U,
    0xa6841a00U,
    0xa6c9dd00U,
    0x51ee0380U,
    0x68870240U,
    0x45c880a0U,
    0x8379c0b0U,
    0xd84c6208U,
    0x233252c4U,
    0xdf6a18aeU,
    0xe24edd9dU,
    0x9626812eU,
    0x5afec26aU,
    0xb604e184U,
    0xc70b9139U,
    0xcc8679e7U,
    0x1bcc8da5U,
    0x48749a95U,
    0x26c41c6aU,
    0x11ec6284U,
    0x88251b9U,
    0x95d21b27U,
    0xcb7adfc5U,
    0x24488045U,
    0xe939c0b2U,
    0x926c6298U,
    0x3c2520bU,
    0xbc721988U,
    0x70cadc84U,
    0x46f0820eU,
    0xec0dc02dU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x18000000U,
    0xec000000U,
    0x5a000000U,
    0xc3000000U,
    0x72800000U,
    0x67c00000U,
    0x33200000U,
    0x59700000U,
    0xcd080000U,
    0x17840000U,
    0x445a0000U,
    0x21650000U,
    0xd10e8000U,
    0xb58ec000U,
    0x5b4f2000U,
    0x11f03000U,
    0x99d1a800U,
    0xae2f1c00U,
    0xa6ea0a00U,
    0x9d44ef00U,
    0xb0f20180U,
    0x145103c0U,
    0xc97c83e0U,
    0xc51fc370U,
    0x3393a108U,
    0xda5ff0bcU,
    0x4c6a08a2U,
    0xd684ee93U,
    0x39d20306U,
    0xbe2103faU,
    0x6ef48078U,
    0x595bc1fbU,
    0x1ee9a245U,
    0x614af36bU,
    0x22ec8b3dU,
    0x134e2d5fU,
    0x15e721e3U,
    0xd7c4300eU,
    0xeb23a846U,
    0x557e1f5aU,
    0xa71689a8U,
    0xcc9b2c23U,
    0xdac1a1a1U,
    0x1cbef26dU,
    0x213e8888U,
    0x9e6f2f7cU,
    0xf193a142U,
    0x355ff2e3U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0x58000000U,
    0x4c000000U,
    0x8e000000U,
    0x5d000000U,
    0x15800000U,
    0xd6400000U,
    0x2e200000U,
    0x84f00000U,
    0x5b080000U,
    0xc8c0000U,
    0xc9d20000U,
    0xf6d0000U,
    0x78528000U,
    0x83204000U,
    0x697e6000U,
    0xf1471000U,
    0xf4bf7800U,
    0x5c348c00U,
    0xcfe19a00U,
    0xff8edf00U,
    0x715a0080U,
    0x34a103c0U,
    0xfc2083e0U,
    0xfffd41f0U,
    0xa784e358U,
    0x3d5b5044U,
    0xbabb1a6eU,
    0xa1229e2dU,
    0xea7662eeU,
    0x71cb124aU,
    0x136d7b54U,
    0x3e598ee1U,
    0xfa33192bU,
    0xe6ee9e3fU,
    0xb804607eU,
    0x1c161085U,
    0x4617fbf2U,
    0x7905cf60U,
    0x8f977b97U,
    0x49488fe2U,
    0xe8bb9b6fU,
    0x1a2fdf51U,
    0xb6fa80d3U,
    0x701c40abU,
    0x380463d8U,
    0xdc161384U,
    0xe617f98eU,
    0x4905cfddU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0x98000000U,
    0x84000000U,
    0xe6000000U,
    0x13000000U,
    0x78800000U,
    0xafc00000U,
    0x1ee00000U,
    0xccb00000U,
    0xde080000U,
    0x27040000U,
    0x768e0000U,
    0x30c10000U,
    0x9c638000U,
    0x127c4000U,
    0x4d76a000U,
    0x2ff97000U,
    0x613ce800U,
    0x6a432400U,
    0xe6afca00U,
    0x17031700U,
    0x5e880180U,
    0x5cc40140U,
    0xf66e01e0U,
    0xfb710050U,
    0x44eb81f8U,
    0xdb8425cU,
    0xbb98a27aU,
    0x6f4872f3U,
    0x45376999U,
    0x3c4b658dU,
    0x1dbf6903U,
    0xe38f64aeU,
    0xb516b22U,
    0x333e651fU,
    0xb75ae84bU,
    0xe13627daU,
    0xaa4a4b3cU,
    0x6ba57eaU,
    0x8713212bU,
    0xc68030b5U,
    0xd8c7c9ffU,
    0x10771464U,
    0xe86e03f9U,
    0x3c710162U,
    0xa26b8180U,
    0xa5784140U,
    0xa3f8a1e0U,
    0x9b387050U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0x28000000U,
    0x8c000000U,
    0x2a000000U,
    0x9b000000U,
    0xe7800000U,
    0xfb400000U,
    0x86200000U,
    0x43f00000U,
    0xae080000U,
    0x4d0c0000U,
    0xde820000U,
    0xbbd30000U,
    0x98f48000U,
    0x2984c000U,
    0xa64e2000U,
    0xc0bb9000U,
    0x5c259800U,
    0x90e2ac00U,
    0xd5953a00U,
    0x442ff00U,
    0x67a80080U,
    0xb9bc01c0U,
    0x7caa0220U,
    0x9e2f00f0U,
    0x27fe8008U,
    0xa81bc05cU,
    0x4c18a336U,
    0x8a1c5181U,
    0x2b173986U,
    0xcf91fee2U,
    0x775c82a0U,
    0xac38c33bU,
    0xd8e42342U,
    0x4994911bU,
    0xb65b19b9U,
    0x58b96c79U,
    0xf82d9a52U,
    0x36eeaec8U,
    0x649738f7U,
    0x78d1ff6cU,
    0x7b7c801eU,
    0xc4c8c0adU,
    0xb96c22f8U,
    0x73d8915fU,
    0xc4f91880U,
    0x3b9a6dc0U,
    0x29511820U,
    0xa9266ff0U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0xc8000000U,
    0x3c000000U,
    0xda000000U,
    0x53000000U,
    0x2e800000U,
    0x11400000U,
    0x94a00000U,
    0x7e700000U,
    0xac080000U,
    0x2040000U,
    0x771a0000U,
    0x309f0000U,
    0x6c5e8000U,
    0xf3cc000U,
    0xc8a12000U,
    0x5467b000U,
    0xd713f800U,
    0x2092bc00U,
    0x745e5a00U,
    0xcb22cf00U,
    0xe6a00180U,
    0xe17003c0U,
    0x708800a0U,
    0x4c440030U,
    0xdf3a0048U,
    0x30af01bcU,
    0xa07683d2U,
    0x3108c109U,
    0xa99322e8U,
    0x9ccb1bcU,
    0xf4ff78e2U,
    0x63457df1U,
    0xbb3f96cU,
    0xa2e2bfcaU,
    0xe25659edU,
    0xaa26cf3eU,
    0x773a0033U,
    0xfcaf025dU,
    0x527683b6U,
    0x6e08c0afU,
    0x9513227fU,
    0x778cb257U,
    0x94df7bcbU,
    0x5f757c59U,
    0x1d9bf800U,
    0xcfd6bc00U,
    0xade45a00U,
    0xe6cdcf00U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0x28000000U,
    0x54000000U,
    0xba000000U,
    0x85000000U,
    0x43800000U,
    0x29c00000U,
    0x70e00000U,
    0x81b00000U,
    0x4c080000U,
    0xa60c0000U,
    0xbb1a0000U,
    0x248b0000U,
    0x934f8000U,
    0xf4bdc000U,
    0xe797e000U,
    0x7bc0d000U,
    0x1f94800U,
    0x8352400U,
    0xc8d32a00U,
    0xc17f3700U,
    0xa9600080U,
    0x9d7002c0U,
    0xf7680220U,
    0x2a7c0090U,
    0x15f20248U,
    0xd2370344U,
    0x9ddd83e6U,
    0xdafac2b3U,
    0xfca260cdU,
    0x38612a1U,
    0x49c928f3U,
    0xa0f4349eU,
    0xd9af81dcU,
    0x300dc10aU,
    0x481fe221U,
    0x840cd280U,
    0xe2034973U,
    0xf90e276dU,
    0xad94ab71U,
    0x16cef4dbU,
    0xb66de2aaU,
    0xebfbd142U,
    0x153eca65U,
    0x5744e7e6U,
    0x76bec800U,
    0xce84e400U,
    0x8e5eca00U,
    0x6b34e700U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0x18000000U,
    0xec000000U,
    0xf2000000U,
    0xa1000000U,
    0x8e800000U,
    0xcbc00000U,
    0xdd200000U,
    0xe7900000U,
    0x2a080000U,
    0xed040000U,
    0xec8e0000U,
    0x2c70000U,
    0xa7a18000U,
    0x325ec000U,
    0xa4272000U,
    0x251e1000U,
    0x83cee800U,
    0xf92fbc00U,
    0x71925200U,
    0xfd0fb900U,
    0xc4880080U,
    0x16c40040U,
    0xc9ae00e0U,
    0x79570070U,
    0x67a98018U,
    0x925ac0ecU,
    0x342920f2U,
    0x4d1910a1U,
    0x77cf688eU,
    0xe7217ccbU,
    0x229d72ddU,
    0xd285a9e7U,
    0x81c0e8aaU,
    0x28bcadU,
    0xf313d20cU,
    0xb4c17972U,
    0xa0a7203fU,
    0x93de109eU,
    0xda6ee8b6U,
    0xe87fbcf4U,
    0xe23a5215U,
    0x715bb9deU,
    0xa3ae00deU,
    0x745700fbU,
    0xfb298080U,
    0x889ac040U,
    0x7f8920e0U,
    0x8d491070U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x18000000U,
    0xbc000000U,
    0xce000000U,
    0xab000000U,
    0xbe800000U,
    0x98400000U,
    0x17200000U,
    0x17500000U,
    0xe2080000U,
    0xbd0c0000U,
    0x89820000U,
    0xf8c10000U,
    0x3c618000U,
    0x2fbc000U,
    0xa314e000U,
    0xe32eb000U,
    0x9d55e800U,
    0xb028400U,
    0x6e8df200U,
    0xa0493500U,
    0xbb280080U,
    0xc15c00c0U,
    0xf50a0020U,
    0xcd8d0010U,
    0xbac38018U,
    0x516ac0bcU,
    0xb37d60ceU,
    0xf7d970abU,
    0x94308beU,
    0x88ad3498U,
    0xd8999a17U,
    0xbee07117U,
    0x94b91262U,
    0x3f37857dU,
    0x5af5e829U,
    0xff128428U,
    0x1d25f204U,
    0x3e5535aeU,
    0x118200f5U,
    0x84c10034U,
    0xd26180cdU,
    0xb9fbc028U,
    0x594e0dfU,
    0xc76eb093U,
    0x4475e880U,
    0xb75284c0U,
    0x3205f220U,
    0x85053510U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0xb8000000U,
    0x84000000U,
    0xd6000000U,
    0x83000000U,
    0xdc800000U,
    0x6f400000U,
    0x56200000U,
    0x80b00000U,
    0x5a080000U,
    0xc90c0000U,
    0x9d860000U,
    0xaecf0000U,
    0x5aeb8000U,
    0x37584000U,
    0xb6d56000U,
    0x1a1c3000U,
    0x3dffc800U,
    0x992bf400U,
    0x763be200U,
    0x844acb00U,
    0xe6a80080U,
    0x95fc00c0U,
    0x952e0060U,
    0xfc3300f0U,
    0xa54580b8U,
    0xd72b4084U,
    0x2130e0d6U,
    0xa6c77083U,
    0xc6e728dcU,
    0xf550846fU,
    0xabd2ca56U,
    0x79994f80U,
    0xdc374adaU,
    0xf5420f09U,
    0xff2faa7dU,
    0xed397f9eU,
    0x4cc68202U,
    0x17eafb83U,
    0x7cd9c838U,
    0x9b14f42dU,
    0x9c7862efU,
    0x65ee8b01U,
    0x79d36012U,
    0xec9330acU,
    0xe3b44880U,
    0xb683b4c0U,
    0x7e468260U,
    0xbfaafbf0U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x38000000U,
    0xd4000000U,
    0xca000000U,
    0xa7000000U,
    0xd9800000U,
    0xc8c00000U,
    0x9ce00000U,
    0xec500000U,
    0x76080000U,
    0xf10c0000U,
    0x38860000U,
    0x184b0000U,
    0xd8a38000U,
    0x12fd4000U,
    0xddf4a000U,
    0x987e7000U,
    0x16bf9800U,
    0x83558c00U,
    0xe3834e00U,
    0xf7c10500U,
    0x41680080U,
    0x669c00c0U,
    0xa1ee0060U,
    0xdad700b0U,
    0xeb4d8038U,
    0xb2a40d4U,
    0x1d3920caU,
    0x289430a7U,
    0x84e6b8d9U,
    0x2851bcc8U,
    0x540df69cU,
    0xa0cb9ecU,
    0xc70bf6f6U,
    0x6987b931U,
    0xf0c876d8U,
    0x48eaf968U,
    0x2654d600U,
    0xd10889b6U,
    0x28854ecfU,
    0xf04a059bU,
    0x84ab80ddU,
    0x34f14048U,
    0x64f2a034U,
    0x2cf57010U,
    0xa0fc1880U,
    0xef8ccc0U,
    0x5bffee60U,
    0xf17375b0U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x78000000U,
    0xec000000U,
    0x52000000U,
    0x13000000U,
    0x43800000U,
    0xd0c00000U,
    0x9de00000U,
    0x1f900000U,
    0x5b080000U,
    0x97840000U,
    0xceca0000U,
    0x6ceb0000U,
    0x37178000U,
    0x2446c000U,
    0x88a12000U,
    0x5fbb3000U,
    0x6bf73800U,
    0x2b568c00U,
    0xe0e11e00U,
    0x3511d900U,
    0x5f4a0080U,
    0xef2b0040U,
    0x497780a0U,
    0x5b16c0b0U,
    0x36492078U,
    0x3baf30ecU,
    0xac353852U,
    0xc3398c13U,
    0x5abc9e43U,
    0xad7c19d0U,
    0x7d1ca01dU,
    0x8b46f05fU,
    0xf12998fbU,
    0xb87f7c27U,
    0x739506b6U,
    0x49036580U,
    0x2489a6e5U,
    0x3d459577U,
    0xc4203eebU,
    0x46fae97fU,
    0x69d538aeU,
    0xc0a98c28U,
    0x8bb49e31U,
    0x75f819edU,
    0xda56a078U,
    0xc86df0ecU,
    0x4a5e1852U,
    0x4069bc13U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0xa8000000U,
    0xc4000000U,
    0x2e000000U,
    0xdb000000U,
    0x3f800000U,
    0xf0400000U,
    0x85200000U,
    0x18100000U,
    0xdd080000U,
    0x60840000U,
    0x1ce0000U,
    0x1e6b0000U,
    0xaba8000U,
    0xf1544000U,
    0x16aee000U,
    0xa257b000U,
    0xfd2ef800U,
    0x841a8400U,
    0x7f089200U,
    0xe1802100U,
    0x63460080U,
    0xceaf0040U,
    0x6e5480e0U,
    0x472f40b0U,
    0x691c60a8U,
    0xf787f0c4U,
    0xc44e182eU,
    0xe32634dbU,
    0xb71cea3fU,
    0x648ee5f0U,
    0x8fc07205U,
    0x95689158U,
    0x2d3478bdU,
    0x6d1ec490U,
    0x798e7249U,
    0x4f43916aU,
    0xc4aef80cU,
    0x2b5a84aeU,
    0xba892e7U,
    0x62d02139U,
    0xacee006fU,
    0x827b00e8U,
    0x19b280e9U,
    0xfad040f2U,
    0x80e0e028U,
    0x887cb084U,
    0x5cb478ceU,
    0xb65ec46bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0x38000000U,
    0xbc000000U,
    0xea000000U,
    0xd9000000U,
    0x16800000U,
    0xb6c00000U,
    0xcce00000U,
    0x2dd00000U,
    0x3f080000U,
    0x7d8c0000U,
    0x954a0000U,
    0x4eaf0000U,
    0xaf38000U,
    0x9eb3c000U,
    0xea9aa000U,
    0xd4a39000U,
    0x2bf86800U,
    0x9435ec00U,
    0x465f8e00U,
    0xf94a3d00U,
    0xaca20080U,
    0xf7f300c0U,
    0x2e3180a0U,
    0x5750c0f0U,
    0x6bc32038U,
    0x4c6f50bcU,
    0x81948eaU,
    0xcc69bcd9U,
    0xc8174616U,
    0x6c6341b6U,
    0x381e664cU,
    0x546311edU,
    0x8414ae1fU,
    0xbe696d4dU,
    0x5d11488dU,
    0xa8e5bcc2U,
    0xebdd4678U,
    0x640c410bU,
    0xc60de6aeU,
    0x5b00d1c7U,
    0xbb860e3bU,
    0xce46fde6U,
    0xf52320bbU,
    0xc4bf5093U,
    0x6b9148b8U,
    0x2e25bc7cU,
    0xbf3d464aU,
    0x5dc4129U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0xd8000000U,
    0xe4000000U,
    0xb6000000U,
    0x57000000U,
    0x28800000U,
    0x8c00000U,
    0x9ea00000U,
    0x3ef00000U,
    0x8880000U,
    0x18c40000U,
    0x16ae0000U,
    0xb2fb0000U,
    0x82858000U,
    0x1dca4000U,
    0xdf256000U,
    0xc5be7000U,
    0x3c670800U,
    0xb55acc00U,
    0x77f48a00U,
    0xeb019f00U,
    0x3a858080U,
    0x9ca4040U,
    0x512560e0U,
    0xc6be70b0U,
    0x7ae708d8U,
    0xe9acce4U,
    0x77d48ab6U,
    0x8a319f57U,
    0x842d80a8U,
    0x273e4048U,
    0xd1a3607eU,
    0x5271708eU,
    0xe64488d0U,
    0xb96f8cfcU,
    0x3cda6aa0U,
    0xe0beafe5U,
    0xe5ea682aU,
    0x4a10bc55U,
    0x451582a1U,
    0xc994534bU,
    0x97528a6cU,
    0xeefe9f09U,
    0x208e00b7U,
    0xc4cb00feU,
    0xf4ad80a8U,
    0x8bfe4048U,
    0x1903607eU,
    0x8b81708eU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x78000000U,
    0x94000000U,
    0x6a000000U,
    0xdb000000U,
    0x60800000U,
    0xea400000U,
    0xed600000U,
    0xe9500000U,
    0x8880000U,
    0x96440000U,
    0xbb660000U,
    0x45b0000U,
    0x350f8000U,
    0x398d4000U,
    0x6dc0a000U,
    0x51a6b000U,
    0xb3718800U,
    0x493fe400U,
    0x4c547600U,
    0x99092500U,
    0xa78f8080U,
    0x6ccd4040U,
    0x9220a060U,
    0xadb6b0b0U,
    0x24198878U,
    0x936be494U,
    0x185a766aU,
    0xd30625dbU,
    0xac8e00e0U,
    0xc44f00aaU,
    0x9461808dU,
    0xbed24059U,
    0x7c492070U,
    0x2060f002U,
    0x4d6a8d1U,
    0x6f4014dfU,
    0xaceb5e55U,
    0x109f71d3U,
    0x332bfe00U,
    0xfe39c1f8U,
    0x92da765bU,
    0xa246252fU,
    0x416e006fU,
    0x775f00f9U,
    0x98980e0U,
    0x55c640aaU,
    0xa5a7208dU,
    0x697ff059U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0x68000000U,
    0x2c000000U,
    0x36000000U,
    0x13000000U,
    0x63800000U,
    0x62400000U,
    0x78a00000U,
    0x89300000U,
    0xf880000U,
    0x344c0000U,
    0xfbaa0000U,
    0xb2b30000U,
    0x29ce8000U,
    0x56eec000U,
    0x57996000U,
    0xcdba3000U,
    0x1c44b800U,
    0xb7a2e400U,
    0x14bdea00U,
    0x62c42300U,
    0x716e8080U,
    0x2fdec0c0U,
    0x901160a0U,
    0xe5f63030U,
    0xb9eeb868U,
    0x3a11e42cU,
    0x68f36a36U,
    0x456ae313U,
    0x3dd7e0e3U,
    0x914f0a2U,
    0xfb7dd8d8U,
    0xef28d4b9U,
    0x59715267U,
    0xde2ac718U,
    0xcbf96a4dU,
    0x8ee9e361U,
    0xd391606aU,
    0x77b630c4U,
    0x94eb8e7U,
    0xaf21e458U,
    0x397b6a4dU,
    0x4e26e3bcU,
    0x93fde03aU,
    0xcae7f061U,
    0xc99358e3U,
    0x52b614a2U,
    0x79c032d8U,
    0xaeecf7b9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x28000000U,
    0x4000000U,
    0x4a000000U,
    0xad000000U,
    0xce800000U,
    0xffc00000U,
    0x45200000U,
    0x19900000U,
    0x92880000U,
    0x79cc0000U,
    0xb6220000U,
    0x28130000U,
    0x22ca8000U,
    0x43ac4000U,
    0xf256a000U,
    0x85a1d000U,
    0xe1526800U,
    0xa425bc00U,
    0x491ff200U,
    0xb243c900U,
    0x436a8080U,
    0xf8fc40c0U,
    0xe37ea020U,
    0x2e3dd030U,
    0xe2d86828U,
    0x456abc04U,
    0xcbff724aU,
    0xf2f089adU,
    0xae74a04eU,
    0xf0b2d03fU,
    0x518e865U,
    0x2c49fc29U,
    0x9c6952baU,
    0x8772197dU,
    0xb430e87cU,
    0x77d5fc45U,
    0x97e352ccU,
    0x523d198cU,
    0x54d0681fU,
    0x9e66bc58U,
    0xfe7d7219U,
    0x8b38940U,
    0x39162099U,
    0x7a4290a1U,
    0x5764484eU,
    0xaaf72c3fU,
    0x6273ba65U,
    0xaeb8e529U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0x78000000U,
    0x74000000U,
    0xf6000000U,
    0x5f000000U,
    0x7f800000U,
    0x5d400000U,
    0xe0e00000U,
    0xf0100000U,
    0x99880000U,
    0xea4c0000U,
    0xc3620000U,
    0x27570000U,
    0xdc6f8000U,
    0xb8db4000U,
    0xd1256000U,
    0x503ef000U,
    0x2d3a7800U,
    0x4bb19400U,
    0x6e70ee00U,
    0xa85cb100U,
    0x6be78080U,
    0xc99740c0U,
    0xc3c76020U,
    0x7129f070U,
    0xe035f878U,
    0x753ad474U,
    0x4fbd8ef6U,
    0xe07e415fU,
    0x8357f8ffU,
    0xe26dd49dU,
    0xcbd20ec0U,
    0x5ca50180U,
    0xdc729861U,
    0x9953245eU,
    0x6f687615U,
    0x15549508U,
    0x2d62765bU,
    0x9c5f9551U,
    0x7de7f667U,
    0xc69fd54fU,
    0xb44f1613U,
    0x206d6538U,
    0x82df8e63U,
    0x2c2941e4U,
    0xf6b8787fU,
    0x28f6945dU,
    0xfd976ee0U,
    0xd5cbf1f0U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0xa8000000U,
    0xe4000000U,
    0x66000000U,
    0x11000000U,
    0xff800000U,
    0x30c00000U,
    0x90a00000U,
    0x72700000U,
    0xd1880000U,
    0x75cc0000U,
    0xd12a0000U,
    0xcfbd0000U,
    0x94a28000U,
    0xa4724000U,
    0x18846000U,
    0x16401000U,
    0xcbef7800U,
    0xcc104c00U,
    0x6e5dea00U,
    0xabbc7700U,
    0x32aa8080U,
    0x157e40c0U,
    0x370e60a0U,
    0x8e8d10d0U,
    0xbf45f8a8U,
    0xd86e0ce4U,
    0xaed38a66U,
    0x21f16711U,
    0xd34f787fU,
    0x4a604cf0U,
    0xd1d5ea30U,
    0xfb7077a2U,
    0xd20080f9U,
    0x1f034051U,
    0xaa8ce017U,
    0x794f500eU,
    0x196998c3U,
    0xf9521c70U,
    0xf53ef2eeU,
    0x25e02b75U,
    0x29121265U,
    0xffdf7b49U,
    0xbe738a0fU,
    0x93816706U,
    0xa2c778ffU,
    0xefac4c30U,
    0xa8ffea90U,
    0xd0cd7772U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xe8000000U,
    0x4c000000U,
    0xe000000U,
    0x49000000U,
    0x93800000U,
    0x81c00000U,
    0x49600000U,
    0xdc300000U,
    0x51880000U,
    0x86c40000U,
    0xb3e20000U,
    0x3e7f0000U,
    0x71268000U,
    0x5f90c000U,
    0x305ae000U,
    0xa0379000U,
    0xd789b800U,
    0x53c5dc00U,
    0x3667f600U,
    0x52bc1300U,
    0xe9ce8080U,
    0x4564c040U,
    0xf230e020U,
    0xe88c90f0U,
    0xfd4d38e8U,
    0x7e2a1c4cU,
    0x791b960eU,
    0xe41b4349U,
    0x9d9dd813U,
    0x37568cc1U,
    0x5abeae69U,
    0x35c55f2cU,
    0x736c4e39U,
    0xef36cf8aU,
    0xd107f61dU,
    0x778c13c7U,
    0xe3c680aaU,
    0xfe60c022U,
    0xeeb2e0bfU,
    0xfc39089U,
    0x4063b8f3U,
    0x6fbedc51U,
    0x20437651U,
    0x27a3d378U,
    0x23dae093U,
    0x61f79081U,
    0xbee9b849U,
    0x7ff5dcdcU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0x68000000U,
    0x8c000000U,
    0xe6000000U,
    0xbf000000U,
    0x3d800000U,
    0x85c00000U,
    0x59a00000U,
    0x67d00000U,
    0x47880000U,
    0x54c40000U,
    0x272a0000U,
    0x819d0000U,
    0xc0ee8000U,
    0x3d3cc000U,
    0xdbbc6000U,
    0x49faf000U,
    0x6c935800U,
    0x6b9c00U,
    0xf07e7a00U,
    0xb35ccd00U,
    0x18c68080U,
    0x2128c040U,
    0x4e9e60a0U,
    0x4563f0d0U,
    0x5cf7d868U,
    0xe81a5c8cU,
    0x77249ae6U,
    0xa99efdbfU,
    0xece3b8bdU,
    0xb34acc5U,
    0xcb5c2f9U,
    0xf87c61b7U,
    0xf5142afU,
    0xe6cda198U,
    0xaa2ba2e1U,
    0x711f91aeU,
    0x15a69ab5U,
    0x61d7fda4U,
    0x888f382cU,
    0xd1416c8dU,
    0x46e52238U,
    0xb2335132U,
    0xfe32fabdU,
    0xf8390da3U,
    0x373e603dU,
    0xb2b3f085U,
    0xd37fd859U,
    0xe0de5c67U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xb8000000U,
    0x54000000U,
    0x5e000000U,
    0xfb000000U,
    0xbd800000U,
    0xe6400000U,
    0x8be00000U,
    0x12900000U,
    0xa6880000U,
    0xcbcc0000U,
    0xb5a60000U,
    0x9d750000U,
    0x52138000U,
    0xc2414000U,
    0x9defe000U,
    0x359cb000U,
    0xf90d5800U,
    0x648ca400U,
    0x72ce1e00U,
    0x712e7900U,
    0xdfb58080U,
    0xcb3440c0U,
    0xf1fc6060U,
    0x5cddf050U,
    0x6162b8b8U,
    0xe3501454U,
    0x5e23465eU,
    0xfc32ddfbU,
    0xb6739e3dU,
    0x97963926U,
    0x100fe06bU,
    0x180cb082U,
    0x6405587eU,
    0xb600a4cfU,
    0x17081e53U,
    0xb78b7932U,
    0x434e0031U,
    0xcd69001fU,
    0x495d80cbU,
    0xcb284091U,
    0x52b2606cU,
    0x1b4f0e9U,
    0xf03f383fU,
    0x2c7854dcU,
    0xea912685U,
    0x52862d72U,
    0xa5cca635U,
    0xa6ae6d79U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0xe8000000U,
    0xd4000000U,
    0xf2000000U,
    0xc9000000U,
    0x43800000U,
    0x30400000U,
    0x41600000U,
    0xb7f00000U,
    0xf6880000U,
    0xf5cc0000U,
    0xba260000U,
    0x4190000U,
    0xeb68000U,
    0x54694000U,
    0x82752000U,
    0x75cb9000U,
    0x7a2fb800U,
    0x6416c400U,
    0x9eb77600U,
    0xbc624f00U,
    0x56708080U,
    0x87c040c0U,
    0xb32ba060U,
    0x279ed090U,
    0xaef498e8U,
    0xfd0854d4U,
    0xe1884ef2U,
    0x7144cbc9U,
    0x46e456c3U,
    0x9db0dff0U,
    0xaae9b8a1U,
    0xf3bfc4e7U,
    0xb5e9f67eU,
    0xf3370fb1U,
    0x332ba020U,
    0xe79ed0d9U,
    0xcef498dfU,
    0x6d08543dU,
    0x9884e68U,
    0xa544cb26U,
    0xb4e456bfU,
    0x54b0df2fU,
    0xe969b871U,
    0xc3ffc42dU,
    0xf489f62bU,
    0x44c70f24U,
    0xc5a3a053U,
    0x1252d02eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0xe8000000U,
    0xb4000000U,
    0xfa000000U,
    0x9000000U,
    0x31800000U,
    0xd8c00000U,
    0x35200000U,
    0x25900000U,
    0xda880000U,
    0xfc440000U,
    0xb2620000U,
    0x4b770000U,
    0x1ab68000U,
    0x5e174000U,
    0x8341a000U,
    0x1ce19000U,
    0x2d389800U,
    0xf1d74c00U,
    0xbce53200U,
    0x1d302900U,
    0x39dc8080U,
    0x78e44040U,
    0xf352020U,
    0x84d1d070U,
    0xb367b8e8U,
    0xdef59cb4U,
    0x80760afaU,
    0x4e35f509U,
    0x3155aab1U,
    0x59a36598U,
    0x26dbb295U,
    0x86636915U,
    0xf17f2092U,
    0x33b2d078U,
    0x1f9b3880U,
    0xb381dc86U,
    0x9dcb2a39U,
    0xf2a0253bU,
    0x225012fdU,
    0x7121f9a8U,
    0x779b3853U,
    0x4781dcc0U,
    0x47cb2a29U,
    0x8ba0254bU,
    0xfbd012d9U,
    0x1de1f96cU,
    0xb8bb384fU,
    0x6b11dc6cU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x68000000U,
    0x3c000000U,
    0x86000000U,
    0xb5000000U,
    0x36800000U,
    0x49400000U,
    0xa3200000U,
    0xc0700000U,
    0x9d880000U,
    0xeecc0000U,
    0xbee20000U,
    0xdd190000U,
    0xaf5e8000U,
    0xc3f7c000U,
    0x224e6000U,
    0x24ac5000U,
    0x4dbee800U,
    0xe8e0d400U,
    0xe01b1200U,
    0x15d6f700U,
    0xf4b48080U,
    0x6062c0c0U,
    0x6852e020U,
    0xaa729090U,
    0x26860868U,
    0xe147443cU,
    0xbf219a86U,
    0xd67f73b5U,
    0x4085fab6U,
    0xe4462389U,
    0x71a79203U,
    0xcb383790U,
    0x59a46055U,
    0x17395082U,
    0x6fa26870U,
    0x5a3e14c4U,
    0xd23f277U,
    0xa9716703U,
    0x9d0e08b1U,
    0xea8b44d8U,
    0x7f439a65U,
    0xee26731fU,
    0xa2fb7a88U,
    0x6ec1e31aU,
    0x7ee1f25eU,
    0xfd186775U,
    0x3f5888a5U,
    0xabf084b5U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x28000000U,
    0x8c000000U,
    0x1e000000U,
    0xd1000000U,
    0x98800000U,
    0x15c00000U,
    0x5fe00000U,
    0xaeb00000U,
    0x65c80000U,
    0x27ec0000U,
    0x6ab60000U,
    0x67c10000U,
    0xc0ee8000U,
    0xaf3ec000U,
    0xf480e000U,
    0x5bc39000U,
    0xc6e74800U,
    0x2a39bc00U,
    0x46026e00U,
    0x2507a300U,
    0x4280e080U,
    0xc6c390c0U,
    0x20674860U,
    0x7ef9bc90U,
    0xa9626e28U,
    0x1277a38cU,
    0x66a8e09eU,
    0x9e9f9011U,
    0xb79948f8U,
    0x2b14bc85U,
    0x5cdaee77U,
    0x74386322U,
    0x370e80fbU,
    0x4d8ec036U,
    0xef48e092U,
    0x3d2f90e2U,
    0x1cd148b7U,
    0xd438bc8dU,
    0xc70cee8fU,
    0xf58963adU,
    0x4b4800b4U,
    0xaf2c0098U,
    0xd3d600b9U,
    0x9db100b4U,
    0x4a4680fbU,
    0xbfa2c036U,
    0xba1ee092U,
    0x645e90e2U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0xd8000000U,
    0xe4000000U,
    0xe2000000U,
    0x7b000000U,
    0x59800000U,
    0xa0c00000U,
    0x12a00000U,
    0x20b00000U,
    0xbcc80000U,
    0x4ac0000U,
    0xb1b60000U,
    0x12430000U,
    0x23e18000U,
    0x7ed84000U,
    0xe0dd2000U,
    0x3dde3000U,
    0x2d59d800U,
    0xc71d2c00U,
    0x10f21a00U,
    0x842d5300U,
    0x93fd2080U,
    0x29ae30c0U,
    0xd931d860U,
    0xcc012cb0U,
    0x7e0c1ad8U,
    0xad0253e4U,
    0xa88aa062U,
    0xbe4570bbU,
    0xede57839U,
    0xbdb5c10U,
    0xac56e2caU,
    0x619e4fc4U,
    0xbbbee2deU,
    0x95424fbfU,
    0x1c60e208U,
    0xf71d4fc2U,
    0x8ff6209U,
    0x2a0fcaU,
    0xc1f5c286U,
    0x8aaf7fd6U,
    0x64b0ba1fU,
    0x8ec4239aU,
    0x17ae58a2U,
    0xd4366c95U,
    0xb486badeU,
    0xa84723bfU,
    0x7cefd808U,
    0xa55e2cc2U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x58000000U,
    0x44000000U,
    0x86000000U,
    0xdf000000U,
    0x1f800000U,
    0xa4400000U,
    0x29e00000U,
    0xced00000U,
    0xa8480000U,
    0x3e40000U,
    0xebd60000U,
    0x5ac70000U,
    0xb9218000U,
    0xe2f24000U,
    0x803f6000U,
    0x8d147000U,
    0xb1efb800U,
    0xaadf2400U,
    0xbe4bee00U,
    0x74e7d500U,
    0xe85f6080U,
    0x3c847040U,
    0xc9c7b860U,
    0xecab24f0U,
    0x93b5ee58U,
    0x354d544U,
    0xbd00e006U,
    0x9685309fU,
    0xacc7587fU,
    0x7e2e1454U,
    0xd972b671U,
    0x527ac18aU,
    0x23f256aeU,
    0x24bff19cU,
    0x20d50e14U,
    0x4b41e54eU,
    0x4e6fb828U,
    0xbe9f24d8U,
    0xafabee96U,
    0xe37d5e5U,
    0x9e17601bU,
    0xa460708fU,
    0xbb91b8d7U,
    0xcd2c24d3U,
    0x1cf46eaeU,
    0x8b36959cU,
    0xbc978014U,
    0xd6a5404eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0xc8000000U,
    0x74000000U,
    0xda000000U,
    0x9f000000U,
    0x13800000U,
    0x80400000U,
    0x47600000U,
    0x6e100000U,
    0xce480000U,
    0xf2640000U,
    0x9a920000U,
    0x71810000U,
    0x53488000U,
    0x82e54000U,
    0x8754a000U,
    0x63ad7000U,
    0x3eb27800U,
    0xc57fa400U,
    0xa6588600U,
    0xbd255900U,
    0x667ca080U,
    0x5bd97040U,
    0x58687820U,
    0x3d9aa490U,
    0x6e0206c8U,
    0x25011974U,
    0x3c80805aU,
    0xcbc140dfU,
    0x7b26a033U,
    0x877c7010U,
    0xe552f88fU,
    0xb0aee41aU,
    0xfb362614U,
    0x2c3d296dU,
    0xbbc5809U,
    0x71f694b1U,
    0x3996de34U,
    0x8c02cd7cU,
    0xb60afe01U,
    0xd90afda5U,
    0xb28c26deU,
    0x1ec829fbU,
    0x3faed88eU,
    0x60b6d41bU,
    0xf878fe94U,
    0x6dbfd2dU,
    0x48eca629U,
    0x90596921U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x28000000U,
    0xfc000000U,
    0x96000000U,
    0x13000000U,
    0xe3800000U,
    0x96400000U,
    0xf600000U,
    0x2b300000U,
    0x14480000U,
    0x56640000U,
    0x3db60000U,
    0x4a890000U,
    0xf8c68000U,
    0xe5a9c000U,
    0xff986000U,
    0x615bb000U,
    0xeffdf800U,
    0x47ea0400U,
    0xcaf2c600U,
    0x87610b00U,
    0xe7306080U,
    0xda4fb040U,
    0x8163f860U,
    0x9c370410U,
    0xa5ca4628U,
    0x9125cbfcU,
    0xa8588016U,
    0x9174c053U,
    0x8520e083U,
    0xe25f7086U,
    0x64751827U,
    0x4da174d7U,
    0x4399de82U,
    0x975d7f45U,
    0xecf13e5eU,
    0x8c660f9cU,
    0xa0b22617U,
    0x1e0e7b9eU,
    0xdf0d7823U,
    0x2d8ac49bU,
    0x414c260cU,
    0xaee37bf2U,
    0xc47df8efU,
    0x7daa045bU,
    0x1b92c602U,
    0x53510b05U,
    0xaef8603eU,
    0xf56bb08cU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0xa8000000U,
    0x14000000U,
    0x5a000000U,
    0xa9000000U,
    0xbc800000U,
    0x80400000U,
    0xf3e00000U,
    0xa0500000U,
    0xb6480000U,
    0xece40000U,
    0x43d20000U,
    0xf58b0000U,
    0x6cce8000U,
    0xcba34000U,
    0xdfb6a000U,
    0xf0181000U,
    0xbda68800U,
    0xe0bba400U,
    0x239cae00U,
    0x56644d00U,
    0x749ea080U,
    0xf1ec1040U,
    0xd55c8820U,
    0x54c4a430U,
    0x7a82ea8U,
    0x19b80d14U,
    0x271c80daU,
    0x5a2840e9U,
    0x6178209cU,
    0xb6bb50b0U,
    0x2c90285bU,
    0x2de3b4b4U,
    0x8b5a26ecU,
    0x3fcfe945U,
    0xae2a0e7fU,
    0x6b7c5d35U,
    0xa7b8283fU,
    0x1c17b41bU,
    0x4ba026e1U,
    0x9fb0e938U,
    0xd01e8e8cU,
    0x8da01de8U,
    0x48ba0809U,
    0x3793e4c4U,
    0xc648e6cU,
    0xdd9f1d05U,
    0x4d6e885fU,
    0x551fa405U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x18000000U,
    0x84000000U,
    0xee000000U,
    0x45000000U,
    0x7e800000U,
    0xe2c00000U,
    0x64a00000U,
    0x5a900000U,
    0xbc80000U,
    0x28240000U,
    0xf7560000U,
    0xae30000U,
    0xb3f58000U,
    0x80764000U,
    0x1039e000U,
    0x345fd000U,
    0x8d6ea800U,
    0x95b84c00U,
    0x51eda00U,
    0x89826100U,
    0x934fe080U,
    0x83ecd040U,
    0x4f732860U,
    0x35ba0cb0U,
    0x95193a18U,
    0x418ab184U,
    0xbf4ac86eU,
    0xf1e5dc05U,
    0x6077921eU,
    0xe032fd52U,
    0x4c5412fcU,
    0xb967bd9eU,
    0x63b87205U,
    0xc41e2d9dU,
    0x19073a71U,
    0x348db19cU,
    0x99c948c1U,
    0xf7209cebU,
    0x5ad3f2f3U,
    0xe7af6d8fU,
    0xbd1d5a16U,
    0x9d8721a6U,
    0xb54b80a9U,
    0xeae1402cU,
    0x43f2601dU,
    0xf87e9019U,
    0x243cc81fU,
    0xc256dc99U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0xf8000000U,
    0x3c000000U,
    0x22000000U,
    0x7000000U,
    0xf9800000U,
    0xf400000U,
    0x55e00000U,
    0x74b00000U,
    0xaa480000U,
    0xeb640000U,
    0x227a0000U,
    0x20e50000U,
    0x323b8000U,
    0xc081c000U,
    0x43cb2000U,
    0xfc25f000U,
    0x3b915800U,
    0x5e555400U,
    0x5b7aee00U,
    0xcc6cf300U,
    0x4bf92080U,
    0x77a4f040U,
    0x8bd0d8a0U,
    0x6e3194d0U,
    0xd28a4ef8U,
    0xccc8c33cU,
    0x11a358a2U,
    0xd2d45447U,
    0x12bb6e59U,
    0xf34833dfU,
    0x97e9802dU,
    0x3b0c008U,
    0x7bc2a028U,
    0x2025303cU,
    0x699bf823U,
    0x71546413U,
    0x66fb1645U,
    0xdd2d97b3U,
    0x3b11b690U,
    0xfd9ca758U,
    0xd7584e6cU,
    0xdff9c39aU,
    0xd1aad81bU,
    0x32d4944bU,
    0x62b1ce50U,
    0xdb490340U,
    0x53e87821U,
    0x1db1a484U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x88000000U,
    0x84000000U,
    0x92000000U,
    0x4d000000U,
    0x95800000U,
    0x3cc00000U,
    0x50200000U,
    0x92900000U,
    0x13c80000U,
    0x20ac0000U,
    0xefde0000U,
    0xd5230000U,
    0xa3148000U,
    0xed0e4000U,
    0x58e2000U,
    0xe4c25000U,
    0xec2a1800U,
    0xc926c00U,
    0x48c19200U,
    0x6a255100U,
    0xb982080U,
    0x2c4d50c0U,
    0x5d6098e0U,
    0x3c7f2cb0U,
    0xa47b3288U,
    0x38794184U,
    0xf6741812U,
    0x95716c8dU,
    0xd3f51275U,
    0x21bb118cU,
    0x7de0058U,
    0x212300d6U,
    0x59148061U,
    0x940e40ddU,
    0x8a0e20f2U,
    0x1102506dU,
    0xbb8a1861U,
    0xefc26c32U,
    0x9ea99283U,
    0xe4d951f8U,
    0xa7ae20d3U,
    0x4b52508bU,
    0x2621818U,
    0x24fe6c17U,
    0xed3f9269U,
    0xe4965199U,
    0xbcc4a000U,
    0x902f1050U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x48000000U,
    0x8c000000U,
    0x3e000000U,
    0xeb000000U,
    0xb3800000U,
    0xa1c00000U,
    0x2f600000U,
    0x48900000U,
    0x82480000U,
    0xd6a40000U,
    0xf3fe0000U,
    0x80d30000U,
    0x45e88000U,
    0xcfd2c000U,
    0xb46ee000U,
    0x13153000U,
    0xff817800U,
    0x3fce3c00U,
    0x94652600U,
    0x312c100U,
    0xe7866080U,
    0x4bc7f040U,
    0x6e6f98e0U,
    0x5a1b0cb0U,
    0x81045e48U,
    0xb28cfd8cU,
    0x534b46beU,
    0x9c2131abU,
    0x64bff8d3U,
    0xaefbfc51U,
    0xf4554687U,
    0x39a23174U,
    0x527f7874U,
    0xa41d3cf1U,
    0xca0da69eU,
    0x5100017aU,
    0xa888011U,
    0x7742c0eaU,
    0x9e26e0c7U,
    0xf9b130d6U,
    0x7a7f7875U,
    0xd81d3c44U,
    0x5c0da633U,
    0x860001efU,
    0xcf0880a7U,
    0xb182c0a0U,
    0x3cc6e019U,
    0xfbe1300eU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0xf8000000U,
    0x64000000U,
    0xa000000U,
    0x43000000U,
    0x10800000U,
    0xb400000U,
    0x36a00000U,
    0xfe300000U,
    0xf3c80000U,
    0x91ec0000U,
    0x3e960000U,
    0xd0f70000U,
    0xc3ab8000U,
    0x93bc4000U,
    0xc901a000U,
    0x9388b000U,
    0x7bcec800U,
    0xcde95400U,
    0x30987a00U,
    0x69f49b00U,
    0x682a2080U,
    0xec74f0c0U,
    0xfeef6860U,
    0x1811e4f0U,
    0x66beb2f8U,
    0xa481cf64U,
    0xa94c5a8aU,
    0x79ab6b83U,
    0x88b0c8f0U,
    0x2d82543bU,
    0x9acdfaaeU,
    0x9263db6aU,
    0x4d5e0081U,
    0x811b0076U,
    0x9d3d8044U,
    0xb34b4068U,
    0xf2aa201dU,
    0x6434f002U,
    0xb8cf6806U,
    0x1d61e4bfU,
    0x5bd6b2a6U,
    0xaf5dcf87U,
    0x6e125a7bU,
    0x7bb06b14U,
    0x650d4871U,
    0x6589144dU,
    0xa6c7daeaU,
    0x6c672b02U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x38000000U,
    0x44000000U,
    0x4a000000U,
    0x57000000U,
    0xa6800000U,
    0x3f400000U,
    0xbda00000U,
    0xb6900000U,
    0x5dc80000U,
    0x88e40000U,
    0x3c360000U,
    0xfdd30000U,
    0xd26f8000U,
    0x4d764000U,
    0x3d71a000U,
    0x2571f000U,
    0xb177a800U,
    0x73721400U,
    0x58788a00U,
    0xf0f2b100U,
    0xd2be2080U,
    0x9e97b040U,
    0xb1ce0860U,
    0x6ee7e4b0U,
    0xed392238U,
    0x3a53a544U,
    0x12a92acaU,
    0x74134117U,
    0xf8818846U,
    0xea41a4cfU,
    0x502082e5U,
    0x91565542U,
    0xfa2082afU,
    0x365655dbU,
    0x4a08230U,
    0xfd165565U,
    0xcb008211U,
    0x58865570U,
    0x7a48824fU,
    0xb82255f8U,
    0x5d5e82e4U,
    0xcc2155daU,
    0x6f59023fU,
    0xbf23152aU,
    0x33d6a2e9U,
    0xef65e514U,
    0x6cf88ad5U,
    0x2cb2b127U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x38000000U,
    0x3c000000U,
    0xd6000000U,
    0xbd000000U,
    0x4c800000U,
    0xc0400000U,
    0xbe200000U,
    0xd6300000U,
    0xf8c80000U,
    0xec640000U,
    0xdf1e0000U,
    0xf9770000U,
    0x1d6f8000U,
    0x3d99c000U,
    0x58346000U,
    0xc9cb5000U,
    0xaeee8800U,
    0x6e53a400U,
    0xe5d9d200U,
    0x8a1e3b00U,
    0xe1fbe080U,
    0x1f229040U,
    0x1cb2e8e0U,
    0xad8cf4f0U,
    0xeac15a38U,
    0x1b6e9f3cU,
    0xe89bb256U,
    0xb66bfdU,
    0x2b8ce82cU,
    0x7fcbf470U,
    0xa3e6da66U,
    0xfad35f1aU,
    0xe991d296U,
    0xda3a3b2dU,
    0xb6c5e025U,
    0x7d659034U,
    0x8d956837U,
    0x803134e7U,
    0x5cb3af0U,
    0x40e2cf72U,
    0xef52ba93U,
    0x7f580f46U,
    0xf75f5ab5U,
    0x13599fe8U,
    0x95432baU,
    0x5a5fab5dU,
    0x97d08843U,
    0xcd14a42eU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x28000000U,
    0xf4000000U,
    0xe2000000U,
    0x25000000U,
    0x92800000U,
    0x62400000U,
    0xe200000U,
    0x2b00000U,
    0x1cc80000U,
    0xe26c0000U,
    0x679a0000U,
    0x47f50000U,
    0x15668000U,
    0x48194000U,
    0x1b312000U,
    0x9709d000U,
    0x5f82e800U,
    0xa4c04400U,
    0x5e631200U,
    0x1969300U,
    0x78f7a080U,
    0x7ee090c0U,
    0xaedbc8a0U,
    0xb05594d0U,
    0x2c13fa28U,
    0xb13fd7f4U,
    0x36003262U,
    0xd70643e5U,
    0x3f89c8b2U,
    0xd4cc9472U,
    0xa66f7a86U,
    0xdd939726U,
    0x6ef79256U,
    0xb9e6d3f3U,
    0x19520037U,
    0x40990010U,
    0x407c8001U,
    0xbdac400cU,
    0x2877a0c3U,
    0x29a090a6U,
    0xba7bc854U,
    0xf4a59446U,
    0xf4fbfab0U,
    0x80e3d76eU,
    0x3dd232e4U,
    0x35df4381U,
    0xd1d548b1U,
    0xbbd0d436U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xe8000000U,
    0xac000000U,
    0xd6000000U,
    0x1f000000U,
    0x53800000U,
    0x24c00000U,
    0x25a00000U,
    0x3fd00000U,
    0xf1480000U,
    0x46640000U,
    0x2df20000U,
    0x60570000U,
    0xb3828000U,
    0xb4ccc000U,
    0x1da06000U,
    0x8bd07000U,
    0x63497800U,
    0x2361ec00U,
    0xb77cca00U,
    0x8103500U,
    0xe16ae080U,
    0x8a78b040U,
    0xf69b1820U,
    0x3269cf0U,
    0xf99732e8U,
    0x28ad19acU,
    0x495e4a56U,
    0x6f0cf55fU,
    0xfb8280f3U,
    0xa8ccc094U,
    0x3a060edU,
    0xc8d07063U,
    0xec978cfU,
    0xb4a1ecf5U,
    0x175ccaa8U,
    0xc00355bU,
    0x6602e045U,
    0xd70cb0efU,
    0xf8118e9U,
    0x1ac59c02U,
    0x96afb257U,
    0xba52d9a9U,
    0xca8eaa7fU,
    0x30474583U,
    0x36e918bcU,
    0xb4b19c21U,
    0xca35b265U,
    0x6871d9c8U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x18000000U,
    0x4000000U,
    0x46000000U,
    0x5d000000U,
    0x61800000U,
    0x2d400000U,
    0xfd600000U,
    0x5d300000U,
    0x96c80000U,
    0x8f240000U,
    0xfad20000U,
    0xa350000U,
    0xb04d8000U,
    0xfce64000U,
    0x807b6000U,
    0xa3a15000U,
    0xce13d800U,
    0x2e1cf400U,
    0x9e1f6600U,
    0x76199b00U,
    0xba1ee080U,
    0xe0131040U,
    0xff12b820U,
    0x859ca4d0U,
    0x945b3e18U,
    0x25f22f04U,
    0xa8e566c6U,
    0x9e789b1dU,
    0xdaa960c1U,
    0x399450bdU,
    0x465e58c5U,
    0x6fab489U,
    0x406406c8U,
    0x8cb8cbd6U,
    0x538d385dU,
    0xbe4fe43aU,
    0xfdedde0cU,
    0x83f53f5cU,
    0x45edde73U,
    0x17f53f35U,
    0x3bedde44U,
    0x9ef53f66U,
    0x46dde0dU,
    0xeab53f39U,
    0xde8dde89U,
    0xc7c53f2bU,
    0xd4a5deb8U,
    0x38913f63U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x68000000U,
    0x84000000U,
    0x9a000000U,
    0xc7000000U,
    0x4c800000U,
    0x37400000U,
    0xd4600000U,
    0xe5700000U,
    0x64c80000U,
    0x83ac0000U,
    0x78560000U,
    0x38db0000U,
    0xe1928000U,
    0x60fa4000U,
    0xba08a000U,
    0x5703f000U,
    0x748f0800U,
    0xeb471400U,
    0xa2665600U,
    0x3c770f00U,
    0x754c2080U,
    0x3f62b0c0U,
    0xd7f52860U,
    0x3e8ea430U,
    0xb449fe68U,
    0x62efeb84U,
    0x253b7e1aU,
    0x9425ab07U,
    0xd59bdeacU,
    0x52fa5bc7U,
    0x990ad65cU,
    0xb18a4f91U,
    0xecc80076U,
    0xf7ac0030U,
    0xea5600a6U,
    0x4bdb007cU,
    0x5f12808bU,
    0x14ba40f1U,
    0xb8e8a0dcU,
    0x4233f0c1U,
    0x88a708f0U,
    0xbadb1482U,
    0x6a98568bU,
    0x62700f02U,
    0x8840a0b2U,
    0xe4eff073U,
    0xf43908e0U,
    0x31ac14eaU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x38000000U,
    0xc4000000U,
    0xf6000000U,
    0x9b000000U,
    0xf7800000U,
    0xaa400000U,
    0x66600000U,
    0x32900000U,
    0x42c80000U,
    0xadac0000U,
    0x15ba0000U,
    0x9b50000U,
    0x8bb78000U,
    0xaeb24000U,
    0xee3c6000U,
    0xb9f93000U,
    0xdcd3b800U,
    0x12a90400U,
    0xa4347600U,
    0x70fa4100U,
    0x3651e080U,
    0x3c6e70c0U,
    0x639058a0U,
    0xbc4e74d0U,
    0x4d61ae38U,
    0x8d1f75c4U,
    0xc481ae76U,
    0xc1cf755bU,
    0x8e29aed7U,
    0xd1f375baU,
    0xe0dbae7eU,
    0x80aa75e6U,
    0x19362eacU,
    0x247d35e2U,
    0xaf95ce0cU,
    0x64a45ecU,
    0xac6816f4U,
    0x3b937179U,
    0x84a58d3U,
    0x536b74c3U,
    0xea1e2e74U,
    0x10135a1U,
    0x687cea7U,
    0x84c3457dU,
    0xbead9643U,
    0x6e38319cU,
    0x79fbb804U,
    0x7cd50451U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0xd8000000U,
    0xf4000000U,
    0xea000000U,
    0xbd000000U,
    0x5b800000U,
    0xddc00000U,
    0x4d200000U,
    0xfd500000U,
    0xbf480000U,
    0x396c0000U,
    0x84ba0000U,
    0x4eb70000U,
    0xc3b98000U,
    0xe0314000U,
    0x39ffa000U,
    0x46d75000U,
    0xf2867800U,
    0xfc48bc00U,
    0xa5e66200U,
    0x37f2dd00U,
    0x59dc2080U,
    0x680110c0U,
    0x4c0858a0U,
    0x4e02acf0U,
    0x7f05bad8U,
    0xca8a31f4U,
    0x984dba6aU,
    0xc7e6317dU,
    0x56f7ba7bU,
    0xc45131edU,
    0x16ce3ab5U,
    0xda07139U,
    0x88119aadU,
    0xb272140U,
    0x9e5fe2cdU,
    0x13c39deaU,
    0xf22380c7U,
    0x97d640b9U,
    0xd70e2002U,
    0x268a10e6U,
    0x2643d8edU,
    0x60e8ece2U,
    0x68799aa1U,
    0xb61b213dU,
    0x6c2de20eU,
    0xd89d59U,
    0x91800012U,
    0x50c000aeU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0xf8000000U,
    0xec000000U,
    0xfa000000U,
    0xeb000000U,
    0xc2800000U,
    0x5c400000U,
    0x37e00000U,
    0x40f00000U,
    0xd7e80000U,
    0x50f40000U,
    0x7fea0000U,
    0x34f30000U,
    0x91e58000U,
    0xc9fdc000U,
    0x426a2000U,
    0xbc337000U,
    0xeb460800U,
    0x9766b400U,
    0x4bbd7600U,
    0x508ebb00U,
    0xb3460880U,
    0x4b66b440U,
    0x69bd7620U,
    0x278ebb70U,
    0x73c60878U,
    0x1026b4acU,
    0x66dd76daU,
    0xd03ebb9bU,
    0x514e08baU,
    0x5c62b4f0U,
    0xf93f76edU,
    0xf4c9bbdbU,
    0x68a9886dU,
    0xf19874a0U,
    0x555ad692U,
    0xb5f40befU,
    0x506020fcU,
    0x13307069U,
    0xb7cb88d0U,
    0xce2f7453U,
    0x43dd5697U,
    0x9fbecbbeU,
    0x268d803bU,
    0xf249c033U,
    0xeae020fcU,
    0xe3707069U,
    0x5a2b88d0U,
    0x15df7453U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x28000000U,
    0x9c000000U,
    0xe6000000U,
    0x59000000U,
    0x7c800000U,
    0x5e400000U,
    0x35e00000U,
    0xea300000U,
    0x1e80000U,
    0x503c0000U,
    0x1ee20000U,
    0x65bd0000U,
    0x14208000U,
    0x9214c000U,
    0x2dfee000U,
    0x20c95000U,
    0xad2800U,
    0x3052b400U,
    0x4e117600U,
    0xabf19700U,
    0xc9cd2880U,
    0x4422b4c0U,
    0xda1976a0U,
    0x1fd9710U,
    0xfec728a8U,
    0xeda3b45cU,
    0x36dbf646U,
    0xaf545749U,
    0xbb9948d4U,
    0x13e2402U,
    0x2e683e73U,
    0x55ffb3a3U,
    0xf4cd16d5U,
    0xcaa10752U,
    0xb75660edU,
    0xdf919006U,
    0xf339c8e1U,
    0x1d6ae410U,
    0xbe76dec8U,
    0x1306e36aU,
    0x3b883e0fU,
    0x6fcfb335U,
    0x7d251614U,
    0x169d078aU,
    0x67b460e1U,
    0x7f2c9010U,
    0x7d9948c8U,
    0x883e246aU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0x38000000U,
    0xbc000000U,
    0x76000000U,
    0x9d000000U,
    0x26800000U,
    0x6f400000U,
    0x57600000U,
    0xd6f00000U,
    0x33680000U,
    0x2cf40000U,
    0x906a0000U,
    0x43710000U,
    0x2ba98000U,
    0x2c5ac000U,
    0x67bee000U,
    0x1d0b1000U,
    0x66808800U,
    0x4f43e400U,
    0x766fe00U,
    0xeefa0b00U,
    0x8f608880U,
    0x5af3e440U,
    0xd6efe20U,
    0x65fe0b50U,
    0x44e288b8U,
    0x7b36e4fcU,
    0xb14d7e56U,
    0x2e65cbcdU,
    0x4a7de89eU,
    0xdf233493U,
    0x44111601U,
    0xc5583f1bU,
    0xa3301eadU,
    0x3d451bbfU,
    0x106a0011U,
    0x3710018U,
    0xba98026U,
    0x7c5ac083U,
    0x5fbee0eeU,
    0xa10b10a9U,
    0x1080882eU,
    0xd243e4bdU,
    0x21e6fe81U,
    0x81ba0b09U,
    0xd8008826U,
    0x8c03e483U,
    0x3e06feeeU,
    0x490a0ba9U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x58000000U,
    0x54000000U,
    0x26000000U,
    0x4b000000U,
    0x5800000U,
    0x95c00000U,
    0xc4e00000U,
    0x2df00000U,
    0x12e80000U,
    0x4ef40000U,
    0x2b6e0000U,
    0xa1310000U,
    0x8e8b8000U,
    0xb0444000U,
    0x6121e000U,
    0xe114f000U,
    0x331fb800U,
    0x2e14ac00U,
    0x8947600U,
    0xc4d03300U,
    0xbff7b880U,
    0x6fe0ac40U,
    0x387a7660U,
    0xbf213350U,
    0x8e1c38d8U,
    0x7894ec14U,
    0xacd39646U,
    0xe3f1c31bU,
    0x45e580ddU,
    0x1754081U,
    0xd7aa6082U,
    0x5550b036U,
    0x2c3e584fU,
    0xd0005c8fU,
    0x180bcec9U,
    0x34049fc7U,
    0x7603ce19U,
    0x13009f2bU,
    0x5185ce6eU,
    0xb3c59f7dU,
    0x8fe04e17U,
    0x2870df94U,
    0x872a2edcU,
    0x8a106f8bU,
    0x69c7699U,
    0xb3d4336bU,
    0xc071b80eU,
    0x9b25ac2dU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0x98000000U,
    0x34000000U,
    0x52000000U,
    0x45000000U,
    0x63800000U,
    0x23c00000U,
    0xdc200000U,
    0xc8f00000U,
    0x6280000U,
    0xd1fc0000U,
    0x3ba20000U,
    0xe13f0000U,
    0xb0b8000U,
    0x188c4000U,
    0xc346a000U,
    0xbb641000U,
    0x899a9800U,
    0x66bd2c00U,
    0xb2cada00U,
    0xbdace300U,
    0xa6329880U,
    0x25812cc0U,
    0xc4c8daa0U,
    0xc2a3e330U,
    0x8cb11818U,
    0x33c16cf4U,
    0xb4247af2U,
    0xc4f4f375U,
    0x502a007bU,
    0x5ef300d7U,
    0x2921802eU,
    0xf37f40bdU,
    0xb1e720fdU,
    0x6fdb50c6U,
    0x2e5db8b5U,
    0xb0967c6cU,
    0xab3f626eU,
    0xba069feaU,
    0x890ffaa4U,
    0x9588b352U,
    0x9cc4a0a4U,
    0x56ab105fU,
    0xeeb91872U,
    0xeecd6ce3U,
    0xe3ae7aeeU,
    0xb537f32aU,
    0xc9038004U,
    0xf5804062U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0x98000000U,
    0xac000000U,
    0x32000000U,
    0x29000000U,
    0x14800000U,
    0x2c00000U,
    0xcde00000U,
    0xadf00000U,
    0xb3e80000U,
    0xa6fc0000U,
    0xc6620000U,
    0xf4bb0000U,
    0xf4b8000U,
    0x23a1c000U,
    0x1ed0a000U,
    0x9f75000U,
    0xa5efe800U,
    0x59fe7c00U,
    0x8debb600U,
    0xcdfae300U,
    0x63e7e880U,
    0x4ef27cc0U,
    0xf261b6a0U,
    0x6abde370U,
    0x144e6818U,
    0x1e28bc6cU,
    0x89a9692U,
    0xc6db7359U,
    0xc5f9200cU,
    0x47ed906eU,
    0x98f4c85fU,
    0xad68ecf4U,
    0x5134fe3fU,
    0xb503cf08U,
    0xde8bb6b9U,
    0x57cae3b0U,
    0x36fe888U,
    0xb23e7c37U,
    0xf48bb62dU,
    0x12cae38cU,
    0x85efe833U,
    0xe9fe7c99U,
    0xb5ebb6d3U,
    0x11fae36bU,
    0xc9e7e808U,
    0xcbf27cf7U,
    0xd4e1b68dU,
    0x417de3fcU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x98000000U,
    0xb4000000U,
    0xde000000U,
    0x61000000U,
    0xca800000U,
    0x63c00000U,
    0x39200000U,
    0xc9300000U,
    0x6b280000U,
    0x923c0000U,
    0xa6aa0000U,
    0x7e750000U,
    0x70c38000U,
    0x38ae4000U,
    0xff766000U,
    0xa4c5000U,
    0x5364c800U,
    0xea596c00U,
    0xa973da00U,
    0x87413f00U,
    0xd3ecc880U,
    0xa6956cc0U,
    0xa3d1da20U,
    0x53383f90U,
    0x3c2d4818U,
    0xa5b22c74U,
    0x6dee3afeU,
    0xb79f2ff1U,
    0x415460d2U,
    0x8cf55017U,
    0xf70548c7U,
    0x678e2c38U,
    0x73443a39U,
    0xedea2f45U,
    0x7797e0c1U,
    0x615b1016U,
    0x1cf328f1U,
    0x6f027c59U,
    0xd380f2f8U,
    0xad434309U,
    0x8cec3a16U,
    0xbd162fe1U,
    0x29de09aU,
    0x25de10c0U,
    0xa638a871U,
    0xb8a03c99U,
    0x3f7492d8U,
    0x2a461399U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0xb8000000U,
    0x3c000000U,
    0x82000000U,
    0x79000000U,
    0x1d800000U,
    0xf6400000U,
    0x67a00000U,
    0x8d300000U,
    0x82a80000U,
    0x22b40000U,
    0xc5ee0000U,
    0xcc950000U,
    0xb8958000U,
    0xee9ec000U,
    0xf199a000U,
    0x3b1f5000U,
    0x93d97800U,
    0x5abbb400U,
    0xd9ecc200U,
    0x3e934700U,
    0x69977880U,
    0x771eb440U,
    0xb9d14260U,
    0xa7b98710U,
    0x7a60d838U,
    0x33d4e47cU,
    0x6abdbae2U,
    0x11ecf369U,
    0xaa9dba25U,
    0x6f9cf38aU,
    0xb015ba05U,
    0x5f58f3a4U,
    0x3573bac7U,
    0xf609f3b8U,
    0x2f003a78U,
    0x2863354U,
    0x3cca1afdU,
    0xcf66a32fU,
    0x44514294U,
    0x1f98799U,
    0xc5c0d809U,
    0x92e4e4f8U,
    0xd215ba4fU,
    0x7658f3c5U,
    0xf0f3bac5U,
    0x2c49f353U,
    0x72a03a76U,
    0xcab633f0U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x88000000U,
    0x34000000U,
    0xd6000000U,
    0xe5000000U,
    0x25800000U,
    0x89c00000U,
    0xa2600000U,
    0x93b00000U,
    0xe3680000U,
    0xd8340000U,
    0x53ae0000U,
    0xcdd30000U,
    0xf2d68000U,
    0x78584000U,
    0xc31fe000U,
    0x6301000U,
    0xc2aab800U,
    0x5e580c00U,
    0xee17de00U,
    0x77b25700U,
    0xed64b880U,
    0x913b0c40U,
    0x6c295e60U,
    0xcb1e1770U,
    0x72355808U,
    0x74a81c74U,
    0xcb5d66b6U,
    0x439a5b95U,
    0xca7b662dU,
    0x990d5bfdU,
    0xe78be694U,
    0xaac21b46U,
    0x9ae486aeU,
    0x83fd4b55U,
    0x2ac15e4fU,
    0xdaea17bfU,
    0xe3fb588aU,
    0x5acb1cc8U,
    0x52e3e629U,
    0xd7f61b70U,
    0x8cca868aU,
    0xb7ee4b75U,
    0xf277de2cU,
    0x50257abU,
    0x158cb882U,
    0x61cf0cbcU,
    0xe6675e9fU,
    0xcdbd17e5U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0xa8000000U,
    0xe4000000U,
    0x2e000000U,
    0xc7000000U,
    0xfe800000U,
    0x22c00000U,
    0xfae00000U,
    0xf2300000U,
    0x7be80000U,
    0xcfb40000U,
    0x79a20000U,
    0x28910000U,
    0xc3f88000U,
    0x49874000U,
    0x64466000U,
    0x342d3000U,
    0x92d28800U,
    0xc8d29c00U,
    0x69d0e600U,
    0x44559f00U,
    0x4e188880U,
    0x1d379c40U,
    0x216a66a0U,
    0x6373df50U,
    0x144e6828U,
    0x8c29eca4U,
    0x7edc8e8eU,
    0x52dd7397U,
    0x28d486d6U,
    0x99d9af86U,
    0xbc5a80f4U,
    0x2164025U,
    0xd73ee08dU,
    0xc86a7059U,
    0x5af4e805U,
    0xc80facf9U,
    0x540a6e68U,
    0x76030363U,
    0xdb026e99U,
    0x9c87038cU,
    0x2fc86ef0U,
    0xed6203acU,
    0xe972ee87U,
    0x5d444363U,
    0xe5a40ec0U,
    0xa9a3387U,
    0x6ef2e6b7U,
    0xe049f4bU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x98000000U,
    0xd4000000U,
    0x1e000000U,
    0xe5000000U,
    0x6b800000U,
    0xb9400000U,
    0x56e00000U,
    0x33700000U,
    0xed680000U,
    0x423c0000U,
    0x678a0000U,
    0x93470000U,
    0x3deb8000U,
    0x95f64000U,
    0x73a06000U,
    0x679f1000U,
    0xf91d5800U,
    0xc3dac400U,
    0x407c5600U,
    0x5fe8e300U,
    0xe2f6d880U,
    0x172c84c0U,
    0xfe5c3620U,
    0x37f3b0U,
    0xa08b8018U,
    0x6fc64014U,
    0x7028603eU,
    0x72d31055U,
    0xf5ff58f3U,
    0x23a1c46dU,
    0x6f9dd648U,
    0x519a3d6U,
    0x91dd3806U,
    0x6f75d43bU,
    0xcf690e11U,
    0xd53e2710U,
    0x93088ec8U,
    0x228f67c3U,
    0x88c36eaaU,
    0x3ca63761U,
    0xa71e5697U,
    0xc6d3e34bU,
    0xbbf75843U,
    0xceadc44aU,
    0xf81fd63bU,
    0xee52a3aeU,
    0xe834b8e2U,
    0xcc8894b7U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0xe8000000U,
    0xc000000U,
    0x22000000U,
    0xa3000000U,
    0xd0800000U,
    0x8bc00000U,
    0x44a00000U,
    0x13f00000U,
    0xc280000U,
    0xac3c0000U,
    0x6e860000U,
    0x22cd0000U,
    0xdb208000U,
    0x7ab1c000U,
    0x8844a000U,
    0x27e6f000U,
    0xd8912800U,
    0x657b8c00U,
    0xdae3c200U,
    0xb11b4300U,
    0x9931a880U,
    0xfb0a4cc0U,
    0x48762e0U,
    0xbdcdb310U,
    0x31a88068U,
    0xa67dc0ccU,
    0xda6aa0c2U,
    0x32d7f0b3U,
    0xc197a838U,
    0x22f74c87U,
    0xa7afe266U,
    0xe37073b0U,
    0x77e2205cU,
    0x109a30e7U,
    0x997508caU,
    0xecbc21U,
    0xf6164a3fU,
    0x67b63fdaU,
    0xf1cb4244U,
    0x33a68366U,
    0xf57b08b3U,
    0xf2edbc44U,
    0x5d18ca1eU,
    0xab36ff99U,
    0xb0096207U,
    0xd80cb35dU,
    0x14060022U,
    0xd60d00d6U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x88000000U,
    0x24000000U,
    0x5e000000U,
    0x85000000U,
    0xb6800000U,
    0x1c00000U,
    0xf2200000U,
    0x42f00000U,
    0x6aa80000U,
    0x5e3c0000U,
    0xa20a0000U,
    0xb7010000U,
    0x698a8000U,
    0xbc4f4000U,
    0x486d6000U,
    0x239a1000U,
    0x85b68800U,
    0x14c4c00U,
    0x62ea4a00U,
    0x205b3300U,
    0xd0940880U,
    0x823f0cc0U,
    0x400d2a20U,
    0xe00023d0U,
    0xf0080008U,
    0x580c00e4U,
    0xac02007eU,
    0x7a0d0055U,
    0xdb08803eU,
    0x33824025U,
    0xb745e0acU,
    0xf3e850c7U,
    0xb0db68dcU,
    0x28581c5fU,
    0x349b2250U,
    0xfc322ff5U,
    0x150daa83U,
    0xde8e6322U,
    0xd5cfe04aU,
    0xf4295084U,
    0x6bf1e844U,
    0xa6275c49U,
    0x84fe42fcU,
    0x63a43fa2U,
    0x42b9223dU,
    0xf0cf2fc7U,
    0x52ad2ac6U,
    0xc2302393U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0x98000000U,
    0x94000000U,
    0x6e000000U,
    0xa3000000U,
    0x96800000U,
    0x3e400000U,
    0x56200000U,
    0x91700000U,
    0xbba80000U,
    0xcdbc0000U,
    0x6dc20000U,
    0x41eb0000U,
    0xdd938000U,
    0xf3b64000U,
    0xe6cb6000U,
    0x7b657000U,
    0x69d28800U,
    0xf09f5400U,
    0xd638ca00U,
    0x60b5300U,
    0x6f030880U,
    0xcc8214c0U,
    0x83402aa0U,
    0xfba86370U,
    0xadb2e018U,
    0xbdc43054U,
    0xa9e868ceU,
    0xd19764d3U,
    0x9baa20eU,
    0x2bcb37aaU,
    0x4ee82ab8U,
    0xc11463f2U,
    0x98f0e00dU,
    0x116f3043U,
    0x2cdbe803U,
    0x19112434U,
    0x6cf9c290U,
    0xaf624709U,
    0x67d0a2f3U,
    0x839c3767U,
    0xa8b9aa54U,
    0x3449231cU,
    0xc3280065U,
    0x90fc0005U,
    0xd620086U,
    0x9edb00f7U,
    0xa81b8085U,
    0x3b7a4046U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x58000000U,
    0x5c000000U,
    0x7a000000U,
    0xcd000000U,
    0x80800000U,
    0x5e400000U,
    0x4ce00000U,
    0x5b100000U,
    0x5b680000U,
    0x58dc0000U,
    0x31420000U,
    0x9d610000U,
    0x5fdf8000U,
    0x54c0c000U,
    0x97ae2000U,
    0x33731000U,
    0xa83aa800U,
    0xa058b400U,
    0x23070600U,
    0x538cf500U,
    0x45cf2880U,
    0xad2574c0U,
    0x34a6a0U,
    0xb45e25d0U,
    0xfd0420d8U,
    0x688e109cU,
    0x2a4728daU,
    0xe2e9741dU,
    0xe81ea6d8U,
    0x30e32502U,
    0x3119a0b6U,
    0x6e6fd056U,
    0x5456887bU,
    0x4d0aa4d6U,
    0x40822e25U,
    0xfe44819aU,
    0x9cee0efeU,
    0x3169101U,
    0x76b2606U,
    0x22dee5e0U,
    0xfc4a00c3U,
    0x1ded00e3U,
    0x195806dU,
    0x182dc079U,
    0xccbba07eU,
    0x681ed05fU,
    0xf0e108caU,
    0x9116647bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0xb8000000U,
    0x94000000U,
    0x1a000000U,
    0x4b000000U,
    0xa5800000U,
    0x7bc00000U,
    0x48600000U,
    0xd2700000U,
    0xe4e80000U,
    0xa33c0000U,
    0x60c20000U,
    0x35ed0000U,
    0x8db18000U,
    0x1e804000U,
    0x16462000U,
    0xafa7f000U,
    0xbc1af800U,
    0x4b908c00U,
    0x27539e00U,
    0x4cb53b00U,
    0x88017880U,
    0xfc01ccc0U,
    0x96063ea0U,
    0xd50f8b10U,
    0x4c842038U,
    0x14af054U,
    0x8c2b78baU,
    0xaad0cc5bU,
    0xdb75be1dU,
    0xee62cbefU,
    0x6f7380d2U,
    0x246d4059U,
    0x3c77a0e1U,
    0x5e7b0c8U,
    0xe5bcd890U,
    0x92877c73U,
    0x88416673U,
    0x46a9b7f6U,
    0xc698e653U,
    0x8fd5f721U,
    0x5ffcc659U,
    0x732f0757U,
    0xc55fbe4dU,
    0xf3b3cb04U,
    0x878000d6U,
    0x64c0008dU,
    0x57e0009bU,
    0xf2b00033U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0x8000000U,
    0x5c000000U,
    0x62000000U,
    0x81000000U,
    0xed800000U,
    0x87c00000U,
    0x6e600000U,
    0x9e300000U,
    0x58e80000U,
    0xf1740000U,
    0x4aca0000U,
    0xa9ed0000U,
    0xe4f28000U,
    0x2904c000U,
    0x6182a000U,
    0xcdcfd000U,
    0x2361f800U,
    0x19b0ec00U,
    0x225fe00U,
    0x1093bf00U,
    0xd2317880U,
    0x72ed2c40U,
    0xc7fde20U,
    0x5545af90U,
    0xa728a088U,
    0x3b12d01cU,
    0x52fb7842U,
    0xd6002c11U,
    0x4f0d5ee5U,
    0xb6816fdbU,
    0x2f4a008cU,
    0x322d005fU,
    0xc8928095U,
    0xa634c0e6U,
    0xdceaa02cU,
    0xe77bd06bU,
    0x65cbf8deU,
    0xaf6dec59U,
    0x53bf7ec6U,
    0x4f237fe3U,
    0x9719d829U,
    0x88fffceeU,
    0x9304a61bU,
    0x948583bcU,
    0x8e45fe33U,
    0x4fa3bfdeU,
    0x47597828U,
    0x94592c3dU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xb8000000U,
    0xec000000U,
    0x66000000U,
    0x23000000U,
    0xab800000U,
    0xadc00000U,
    0x1b600000U,
    0x17100000U,
    0x9e80000U,
    0xc5c0000U,
    0xabca0000U,
    0x88690000U,
    0x74998000U,
    0xa023c000U,
    0x2531e000U,
    0x15d6f000U,
    0x6f0dd800U,
    0x5d8d6c00U,
    0x16c2ce00U,
    0xce34d00U,
    0x64d65880U,
    0xdd8bacc0U,
    0xd6c8ae20U,
    0x2ce37d50U,
    0x34d9e038U,
    0x658af02cU,
    0x3ac7d846U,
    0x4ae46c73U,
    0x17db4e13U,
    0xce008d41U,
    0x9707b8fdU,
    0x518d5cf4U,
    0xcd7682U,
    0xc7e211f1U,
    0x9b592e88U,
    0xfa4cbdb3U,
    0x88aa00bbU,
    0xb379001fU,
    0x3b7180bdU,
    0xdf7fc08cU,
    0x9d7be0deU,
    0xdc7ff025U,
    0x66f458f7U,
    0xc9beac9bU,
    0x919b2e10U,
    0xb8a9bdb2U,
    0xbb718026U,
    0x1f7fc05bU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0xd8000000U,
    0x44000000U,
    0x7e000000U,
    0x61000000U,
    0x23800000U,
    0x7ec00000U,
    0xf1200000U,
    0x26500000U,
    0x1fe80000U,
    0xcf7c0000U,
    0x1dbe0000U,
    0xbe950000U,
    0x6bcf8000U,
    0xc4af4000U,
    0xd59d6000U,
    0x3f46d000U,
    0x14e01800U,
    0x8bf45c00U,
    0x15f5da00U,
    0xa4f82100U,
    0x3f7d7880U,
    0x15b28cc0U,
    0x4295c260U,
    0x41cc7d10U,
    0x3a8a258U,
    0xd31aad84U,
    0x1c00ba9eU,
    0xfa02f1b1U,
    0x7f03601bU,
    0x5283d0eaU,
    0x85479837U,
    0xcbe71c13U,
    0xa976ba9aU,
    0x58bbf194U,
    0xf31ae031U,
    0xac059047U,
    0x520b78c6U,
    0xf30b8c43U,
    0xb08c42feU,
    0xde4a3d2cU,
    0xf7644283U,
    0x95363d9fU,
    0xf4da4275U,
    0x5aa33d4fU,
    0x6495c25cU,
    0xa4cc7dd7U,
    0x3e28a2cfU,
    0xdcdaad6bU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0xa8000000U,
    0x7c000000U,
    0x16000000U,
    0x6d000000U,
    0x7b800000U,
    0x66400000U,
    0xe3200000U,
    0x87d00000U,
    0xf3680000U,
    0xf9f40000U,
    0x17320000U,
    0xa8dd0000U,
    0xcfec8000U,
    0x3f38c000U,
    0x94d8e000U,
    0x79e69000U,
    0x42359800U,
    0x47551400U,
    0x63a9ba00U,
    0xb71c3100U,
    0xad857880U,
    0xeb478440U,
    0x28ae22a0U,
    0x59942510U,
    0xc4404228U,
    0x1423753cU,
    0x9f53ba36U,
    0x47a5313dU,
    0xc513f8f3U,
    0x1286444aU,
    0x3fc042fdU,
    0x32637586U,
    0xdc73bab6U,
    0xd07531ceU,
    0x9e7bf8b9U,
    0x97724474U,
    0x3ef242acU,
    0xf7be754bU,
    0x681f3aadU,
    0x890df1feU,
    0xe9831824U,
    0x6944d472U,
    0x8fafdabfU,
    0x491f61d4U,
    0x1c84809aU,
    0x96ccc0c5U,
    0x8beae0b4U,
    0xbd3b909aU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0xb8000000U,
    0x24000000U,
    0xe000000U,
    0xd1000000U,
    0x96800000U,
    0x5c00000U,
    0x12600000U,
    0x69b00000U,
    0x41a80000U,
    0x3ed40000U,
    0x689e0000U,
    0x7fb30000U,
    0x64a18000U,
    0x8e5f4000U,
    0xd05a6000U,
    0x6957d000U,
    0xf3da4800U,
    0x941aec00U,
    0xc57b3e00U,
    0xc14e7f00U,
    0xcca82880U,
    0xa2593c40U,
    0x225f7660U,
    0x625793f0U,
    0x25a9638U,
    0xf25c0364U,
    0x4a533eeeU,
    0x6e5a7f61U,
    0x605628ceU,
    0xb15a3c91U,
    0x27d6f6c4U,
    0x221cd36cU,
    0x307ef6e1U,
    0x59c8d38eU,
    0x1860f602U,
    0x26bbd372U,
    0x4e2176f9U,
    0x31949348U,
    0x553316bdU,
    0xdb674304U,
    0xb3f5e7eU,
    0x626aaf1bU,
    0x91b3e0cbU,
    0x5ac9070U,
    0xc0d62898U,
    0x19a3c86U,
    0xcd36f6dfU,
    0x6f6cd386U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x28000000U,
    0x4c000000U,
    0x4a000000U,
    0xe5000000U,
    0x7c800000U,
    0x45400000U,
    0x8b600000U,
    0x90500000U,
    0x7c280000U,
    0xda340000U,
    0x76f60000U,
    0x1c590000U,
    0x96208000U,
    0x8f3bc000U,
    0x72752000U,
    0x3d111000U,
    0x1b472800U,
    0xb0656400U,
    0x97dc0e00U,
    0xdeef8900U,
    0xa39a0880U,
    0xb7007440U,
    0xed8d26e0U,
    0xf7c3ed50U,
    0x7d2e86a8U,
    0xe0b03d0cU,
    0x7cbc0e2aU,
    0x5ebf89f5U,
    0x17b20834U,
    0x71347419U,
    0xf97b2609U,
    0x429aed69U,
    0xdd8e06e2U,
    0xcfcbfd76U,
    0xf9292eabU,
    0xb6be993cU,
    0xfbbd2055U,
    0x8b3510dcU,
    0x64792871U,
    0x5a186492U,
    0x9ec28e99U,
    0xeba94915U,
    0xf0f1a838U,
    0xf357a400U,
    0xefa1ae37U,
    0xe6f159eaU,
    0x945e003aU,
    0x6a2d00feU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0x28000000U,
    0xfc000000U,
    0xa000000U,
    0xf9000000U,
    0x8e800000U,
    0x6bc00000U,
    0x2ae00000U,
    0x9a300000U,
    0xe4280000U,
    0x24540000U,
    0xe4da0000U,
    0xc8190000U,
    0x137c8000U,
    0xe704c000U,
    0xe1812000U,
    0x6645d000U,
    0x4ea00800U,
    0x8194cc00U,
    0xc932b600U,
    0x83a4c700U,
    0xc91b2880U,
    0x9f81c40U,
    0x5ac63e20U,
    0x6860cb50U,
    0xa3f2bea8U,
    0xb3c00bbcU,
    0xeee19eaaU,
    0x6438dbe9U,
    0xbb2f1606U,
    0x71d5d787U,
    0x49200a8U,
    0xdbd008fU,
    0x316e8068U,
    0x3d79c01aU,
    0xd00fa062U,
    0x680c103cU,
    0xdc07a859U,
    0x5a0cdccbU,
    0xd10f1e45U,
    0x72811b7bU,
    0x61c0b6f7U,
    0xd3e9c750U,
    0x14bda89bU,
    0x8fe5dcfeU,
    0xebb9e99U,
    0x7ee1db6dU,
    0x2c33968dU,
    0x372117aeU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x38000000U,
    0xb4000000U,
    0x46000000U,
    0x4b000000U,
    0xfc800000U,
    0x86400000U,
    0x3a00000U,
    0xb6700000U,
    0x52e80000U,
    0x73540000U,
    0xc3da0000U,
    0xd7970000U,
    0x493d8000U,
    0xc0ce4000U,
    0x64eae000U,
    0x60557000U,
    0x1b506800U,
    0x9fdc7400U,
    0x8d9a6200U,
    0x78391b00U,
    0xbd408880U,
    0xa72e0440U,
    0x143f8a20U,
    0x9f4f2fb0U,
    0x2220ab8U,
    0x9fb16ff4U,
    0x2f80eae6U,
    0x3ec01fbbU,
    0xdbe28264U,
    0xfadf6bc2U,
    0x661f60ddU,
    0x58ff30b9U,
    0x9a888f0U,
    0x6f7a04baU,
    0x33658a42U,
    0xca982f18U,
    0x36bf8a22U,
    0x160f2f88U,
    0xa3020aaaU,
    0xe0816f04U,
    0xfc48ea58U,
    0x82a41f09U,
    0xf3f082efU,
    0x252c6b73U,
    0x4130e06aU,
    0xcc270c6U,
    0xb6ede80eU,
    0xdd5234a7U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0x58000000U,
    0x9c000000U,
    0x1e000000U,
    0x3f000000U,
    0x83800000U,
    0x2400000U,
    0x18e00000U,
    0x1bd00000U,
    0xe1280000U,
    0xad7c0000U,
    0x4120000U,
    0x48b0000U,
    0x8dc38000U,
    0xcacc000U,
    0x83b6000U,
    0xa377b000U,
    0xf3141800U,
    0x130c3400U,
    0xa5834600U,
    0xf1492100U,
    0x2d6cf880U,
    0xe29744c0U,
    0xf84c3ea0U,
    0x95e2a5f0U,
    0x4153a6d8U,
    0x33ee515cU,
    0x7251803eU,
    0xa667c00fU,
    0x7b18e07bU,
    0xe70b706eU,
    0xdf87785eU,
    0xbc4784b8U,
    0xd7e55e7cU,
    0xc05e1590U,
    0x7f643e1fU,
    0xab9ea5ddU,
    0x20c1a654U,
    0x6625518aU,
    0x22f20095U,
    0x105b00a6U,
    0x176b809cU,
    0xcf90c048U,
    0x52c960e5U,
    0x1f2cb0ffU,
    0x27f9853U,
    0x8f9cf474U,
    0x32ca26d4U,
    0x4f2591c3U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0xd8000000U,
    0xfc000000U,
    0x6a000000U,
    0xab000000U,
    0x71800000U,
    0xfc00000U,
    0x83200000U,
    0x33b00000U,
    0x95680000U,
    0x5b5c0000U,
    0xd3fe0000U,
    0xc870000U,
    0x1f478000U,
    0x66ac000U,
    0x26d82000U,
    0xe30f000U,
    0xe8aa7800U,
    0xef76ec00U,
    0xcbc89a00U,
    0xe5265f00U,
    0x4ab5d880U,
    0x83ecdcc0U,
    0x971ac260U,
    0x8104330U,
    0xa79f3a58U,
    0x1d506f3cU,
    0x3af1808aU,
    0x9201c05bU,
    0xc709a0c9U,
    0x938130c3U,
    0xb0cbd831U,
    0xbcabdc64U,
    0x17d428eU,
    0xa6ca83ffU,
    0xbdaf1aa1U,
    0xbbfc9ff0U,
    0xe885f8e9U,
    0xe9402c9eU,
    0xf76ebad8U,
    0x2451af95U,
    0x8c782016U,
    0xe40f072U,
    0xf4e27889U,
    0x1c9aec79U,
    0x24de9aaeU,
    0x413d5fa2U,
    0x6f2c5848U,
    0x11b11c01U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0x38000000U,
    0x2c000000U,
    0xe6000000U,
    0xb7000000U,
    0x59800000U,
    0xea400000U,
    0xdf200000U,
    0x17d00000U,
    0x4ae80000U,
    0xb5b40000U,
    0x6b9e0000U,
    0xeac10000U,
    0x5fe18000U,
    0xe939c000U,
    0xc558e000U,
    0x7263000U,
    0xabdb3800U,
    0xf4e8b400U,
    0xfeb99a00U,
    0xec150100U,
    0x9b8a5880U,
    0x3f434440U,
    0xa3a44260U,
    0x91a85d0U,
    0x4f097ab8U,
    0x5587316cU,
    0xbc4f6006U,
    0x802af027U,
    0x5a5c5801U,
    0x6aa64416U,
    0x3b93c201U,
    0x92c6458cU,
    0x13e61a75U,
    0xdf3dc1a8U,
    0x4a5b380dU,
    0x72a8b487U,
    0xa7999a92U,
    0x9cc501b7U,
    0xb0e258baU,
    0x4cb74483U,
    0xf11a4240U,
    0x430b857aU,
    0x380fa5aU,
    0xe34af193U,
    0xcda98066U,
    0x7a1dc049U,
    0xe48ee0d6U,
    0xc2c33058U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x68000000U,
    0x3c000000U,
    0xd6000000U,
    0x51000000U,
    0x1a800000U,
    0x49c00000U,
    0xdf200000U,
    0xead00000U,
    0xfa680000U,
    0x5d3c0000U,
    0x9d1a0000U,
    0x884f0000U,
    0x5b688000U,
    0xcfbac000U,
    0x78d8e000U,
    0x59645000U,
    0xb4b10800U,
    0xc55dec00U,
    0xeb20ca00U,
    0x50d0d900U,
    0x85696880U,
    0xf2bf7cc0U,
    0x4c532220U,
    0xa5a665b0U,
    0xf3102ae8U,
    0xdd4889fcU,
    0x53e26076U,
    0xe57d9021U,
    0x6e736852U,
    0x5bf07cc5U,
    0x45bba261U,
    0xafdca587U,
    0xeae8cab6U,
    0x3fcd985U,
    0xd1bb68f8U,
    0x65dc7c5bU,
    0xdde9a236U,
    0x207fa5f4U,
    0xbef24a21U,
    0x25351921U,
    0xb911081eU,
    0x2a4deceaU,
    0x5068ca45U,
    0x3a3cd9e9U,
    0x469b6852U,
    0x30c7c74U,
    0x9981a298U,
    0x1043a54dU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x58000000U,
    0x14000000U,
    0xda000000U,
    0xf3000000U,
    0x1800000U,
    0x7e400000U,
    0x28600000U,
    0x55d00000U,
    0x21a80000U,
    0x2efc0000U,
    0x4c9e0000U,
    0x12490000U,
    0x966f8000U,
    0x94da4000U,
    0x4f2d2000U,
    0xc737f000U,
    0x27b2f800U,
    0x477c7400U,
    0x6eda2e00U,
    0xcc2ce100U,
    0x2eb85880U,
    0xc5fdc4c0U,
    0xf913f660U,
    0x60265d0U,
    0xc5090ed8U,
    0x8c8211d4U,
    0xdecd203aU,
    0xe8a7f0e3U,
    0xcc7af839U,
    0xb5074baU,
    0x806c2eaaU,
    0x69d9e1b2U,
    0x7a9d87aU,
    0x9bfe8463U,
    0x28195605U,
    0x5083d5e9U,
    0xe8c0d647U,
    0x65ac953bU,
    0x6cfc7618U,
    0xcb982518U,
    0x19c42e44U,
    0x4325e1bbU,
    0xa937d8fdU,
    0xbeb7843dU,
    0x3df6d684U,
    0x5d199522U,
    0x540df6d7U,
    0x7a0b6593U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0xa8000000U,
    0xcc000000U,
    0x22000000U,
    0xb1000000U,
    0xbe800000U,
    0xea400000U,
    0x9de00000U,
    0xf3100000U,
    0x48280000U,
    0x523c0000U,
    0x609e0000U,
    0xe6ef0000U,
    0x90908000U,
    0xeee9c000U,
    0x8c9fa000U,
    0xd4e35000U,
    0x59920800U,
    0x84631400U,
    0x20d2e600U,
    0x160cdd00U,
    0x87032880U,
    0x298684c0U,
    0xbbcfce60U,
    0xf22559b0U,
    0x473c6628U,
    0x501a1d0cU,
    0x2fa408c2U,
    0xaaf014c1U,
    0x853c6676U,
    0x911a1d96U,
    0x59240897U,
    0x3cb0144eU,
    0x92dc66b4U,
    0x1f0a1db9U,
    0x8d8c086bU,
    0x35cc14f3U,
    0xd1226607U,
    0xe0b51d4eU,
    0xc8d48832U,
    0x7a09d486U,
    0x750bc676U,
    0x80854d7eU,
    0x614800e2U,
    0xf66c0098U,
    0x49d600edU,
    0xac83006dU,
    0x3346800cU,
    0x8f6ac0faU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0xd8000000U,
    0x64000000U,
    0xaa000000U,
    0xc1000000U,
    0x47800000U,
    0x8f400000U,
    0x94a00000U,
    0xbb300000U,
    0x49680000U,
    0xc55c0000U,
    0x34b20000U,
    0xba2d0000U,
    0x3ffb8000U,
    0x748d4000U,
    0xadc22000U,
    0x566dd000U,
    0xf1d19800U,
    0xf97bc400U,
    0xdc436200U,
    0x862b9f00U,
    0x31fa3880U,
    0xbf8654c0U,
    0xbb435aa0U,
    0x46accb90U,
    0x8e38e258U,
    0x7ce6dfa4U,
    0xef18188aU,
    0x4d9b8491U,
    0x4f5ac23fU,
    0xa5bb0f7bU,
    0x85a18066U,
    0x44bc40deU,
    0x922ba004U,
    0xb3fd901bU,
    0xf280383fU,
    0x7ac754eaU,
    0x1ee2da48U,
    0x52108bcbU,
    0x64134217U,
    0x3b1b4f66U,
    0x6f98204eU,
    0x925cd0dcU,
    0xbc38187fU,
    0xb9eb8495U,
    0x3292c22bU,
    0x4bd70f0fU,
    0x207b80c4U,
    0x5fcd40c3U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0x68000000U,
    0x24000000U,
    0xa2000000U,
    0x15000000U,
    0x4f800000U,
    0xee400000U,
    0xb600000U,
    0x6d700000U,
    0xb3a80000U,
    0x21dc0000U,
    0x9cf20000U,
    0x76ef0000U,
    0x8e308000U,
    0x1b4b4000U,
    0x94eba000U,
    0xfb3f9000U,
    0x44cf9800U,
    0xa2af1400U,
    0xbc5d5e00U,
    0xafb5eb00U,
    0xa606b880U,
    0xc704c4c0U,
    0x9281e6a0U,
    0x35c22fb0U,
    0x2f25dee8U,
    0x5712abe4U,
    0x33571882U,
    0x69385465U,
    0xf9c4fe87U,
    0x69257b7aU,
    0x4019a041U,
    0xd9d0909cU,
    0xa0ff18feU,
    0x38e4546aU,
    0xad36feb0U,
    0x8bca7bd1U,
    0x4292014U,
    0xf39bd082U,
    0xd994b854U,
    0x389bc4f2U,
    0xad196635U,
    0xaa556f86U,
    0xbc7e37U,
    0x10823bbbU,
    0x50c80085U,
    0xa8ac0076U,
    0x2d5a0027U,
    0xf23300daU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0xd8000000U,
    0xf4000000U,
    0x16000000U,
    0xfd000000U,
    0xc3800000U,
    0x53400000U,
    0x8a200000U,
    0x27b00000U,
    0x5be80000U,
    0x17540000U,
    0x323e0000U,
    0xd5af0000U,
    0xeff8000U,
    0xaec24000U,
    0x286be000U,
    0xdc1b1000U,
    0x2c191800U,
    0xd41a0400U,
    0x701e0a00U,
    0x8e131700U,
    0x5f137880U,
    0x7e9c5440U,
    0xc6db7260U,
    0x72744330U,
    0xc5098a58U,
    0x478557b4U,
    0xfd4698f6U,
    0xb328448dU,
    0x2a3dea7bU,
    0x1ac0797U,
    0x48fc60c4U,
    0xbbcd506eU,
    0xc7ecf8eeU,
    0x6d5e1489U,
    0x4d3092a3U,
    0xcd2f5355U,
    0xbb309249U,
    0x402f5323U,
    0xc0b09262U,
    0xd76f533dU,
    0x849092a2U,
    0xf9df530cU,
    0xaf89223U,
    0x40cb5392U,
    0x71669284U,
    0xe1945389U,
    0xae5112f3U,
    0x7fb2138bU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0x48000000U,
    0xe4000000U,
    0xbe000000U,
    0x51000000U,
    0xd1800000U,
    0xb7c00000U,
    0x1f600000U,
    0x47500000U,
    0x16280000U,
    0xaf7c0000U,
    0x485e0000U,
    0xa6ab0000U,
    0xe13e8000U,
    0xf4ff4000U,
    0x8946000U,
    0x54dd000U,
    0x222db800U,
    0x9780c00U,
    0x25543e00U,
    0x95273100U,
    0x15f15880U,
    0x311d9cc0U,
    0xc68d6660U,
    0x3b46adf0U,
    0xb3223ec8U,
    0xb8f03124U,
    0x6291d85eU,
    0xc249dc61U,
    0xb6a786f9U,
    0x99343da3U,
    0x38fbe669U,
    0xa295ed32U,
    0xa2405e99U,
    0x46aae179U,
    0xd13ce02eU,
    0xdcf59082U,
    0x1c91587eU,
    0xf34d9c59U,
    0x97256671U,
    0x66faad0eU,
    0xc39c3e7bU,
    0x5bcb313aU,
    0xe56758f4U,
    0x385a9c0bU,
    0x2ea5e656U,
    0x653eed67U,
    0xbafede08U,
    0x1195a12fU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0xf8000000U,
    0x9c000000U,
    0x76000000U,
    0xd000000U,
    0x9f800000U,
    0xc9c00000U,
    0x9b200000U,
    0x98100000U,
    0x76180000U,
    0x971c0000U,
    0xb6960000U,
    0x865f0000U,
    0x90f08000U,
    0x26294000U,
    0x2f962000U,
    0x7bdaf000U,
    0x763d0800U,
    0xad829400U,
    0xfec5a600U,
    0x27ae6700U,
    0xcf5da680U,
    0x35726740U,
    0xcceba620U,
    0x2a3d67d0U,
    0x3b8326d8U,
    0xc3c8274cU,
    0xd02306aeU,
    0x629dd741U,
    0x44568e31U,
    0x2ffa0388U,
    0xceab08aaU,
    0xfadd9410U,
    0x27b5265cU,
    0x554727c7U,
    0x1c6b86caU,
    0x83789791U,
    0x21eeae02U,
    0x85b3f3bbU,
    0x9a4880a3U,
    0x3ce54051U,
    0xa23820bcU,
    0x6f89f078U,
    0x41c388e6U,
    0xcf28d49eU,
    0x1a15063cU,
    0x6912d738U,
    0xef9e0ec6U,
    0x9bdf434eU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xd8000000U,
    0x34000000U,
    0xd6000000U,
    0x8d000000U,
    0xd3800000U,
    0xea400000U,
    0xfd600000U,
    0xca100000U,
    0x74180000U,
    0x85140000U,
    0xc920000U,
    0xd9d70000U,
    0x743a8000U,
    0x3ee9c000U,
    0x585ca000U,
    0x8d707000U,
    0x11074800U,
    0x1982ec00U,
    0x6d49ba00U,
    0xc9eddb00U,
    0x44d1ba80U,
    0xdfb9dbc0U,
    0xd0a3bae0U,
    0xfb3edb90U,
    0x26613a38U,
    0x54931ba4U,
    0x2dd79aeeU,
    0xc2306b29U,
    0xe3e0523dU,
    0x53d847c3U,
    0x533dc8c0U,
    0x3a6b2c09U,
    0x5e951a34U,
    0xcaddab4cU,
    0xa6b6f2d8U,
    0xb52b3705U,
    0xcc720014U,
    0xd087005fU,
    0x40c280c2U,
    0x72adc0abU,
    0x36a0b6U,
    0x48e37085U,
    0xe557c82dU,
    0xd6f82c58U,
    0x17459a36U,
    0x6e76b45U,
    0xfc5ad2cdU,
    0xe37187c8U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x48000000U,
    0x64000000U,
    0xfa000000U,
    0x17000000U,
    0xdd800000U,
    0xd0c00000U,
    0x9ba00000U,
    0x34f00000U,
    0xebf80000U,
    0x92740000U,
    0xd8b20000U,
    0xc41b0000U,
    0x656f8000U,
    0x3a51c000U,
    0xc98a6000U,
    0x32c91000U,
    0xc0a7c800U,
    0xe7714400U,
    0x9e3b3e00U,
    0xa7591100U,
    0x1b093e80U,
    0x738211c0U,
    0x85c6bee0U,
    0xed23d150U,
    0xefb4dea8U,
    0x299ec134U,
    0xda11652U,
    0x1df48523U,
    0x2b75a88fU,
    0x503c54f3U,
    0x6256f694U,
    0x25875507U,
    0xcccf809fU,
    0xbda1c0c5U,
    0x65f260efU,
    0xf77d1035U,
    0x1635c858U,
    0xa35a44ecU,
    0x710cbefeU,
    0x2c8cd17dU,
    0x3c495e02U,
    0xc7e401a9U,
    0x631cf651U,
    0xc0e855e8U,
    0x369200aaU,
    0x142b009dU,
    0xc7378003U,
    0x7fd5c0cbU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x88000000U,
    0x44000000U,
    0xa6000000U,
    0x6b000000U,
    0xcd800000U,
    0x47400000U,
    0x69e00000U,
    0x62300000U,
    0xa1380000U,
    0x78bc0000U,
    0x1f20000U,
    0x3f1d0000U,
    0xfaa58000U,
    0x95d9c000U,
    0x840fa000U,
    0x60b1000U,
    0x7b02c800U,
    0xb58f3c00U,
    0x8b4a6200U,
    0x8beae100U,
    0xaf386280U,
    0x7b7e140U,
    0xf27de260U,
    0x2f5e21b0U,
    0x34ca42e8U,
    0x39a931f4U,
    0x4c5a8a4eU,
    0xfd4b0d9fU,
    0x38ed6883U,
    0xbeb42cd8U,
    0xdaf0aa6aU,
    0x9a99ddfaU,
    0x9e000abU,
    0xd2300032U,
    0xc9380042U,
    0xccbc00f9U,
    0x2ff20076U,
    0x101d00b3U,
    0x91258011U,
    0xb999c0ddU,
    0x206fa068U,
    0x237b1026U,
    0xb3dac8e6U,
    0xaf033cf0U,
    0x2b806280U,
    0xcc4be1d2U,
    0x546fe2a8U,
    0xad73216fU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0xe8000000U,
    0xf4000000U,
    0x4a000000U,
    0x51000000U,
    0xae800000U,
    0xb5c00000U,
    0xb5a00000U,
    0x7a500000U,
    0x67580000U,
    0xdfdc0000U,
    0xe51a0000U,
    0x35370000U,
    0xed298000U,
    0xd09ec000U,
    0x3f766000U,
    0x23c17000U,
    0x7aa89800U,
    0x7fd44400U,
    0x1510de00U,
    0x6d359b00U,
    0x212ade80U,
    0x86929b40U,
    0xd07b5ee0U,
    0x96405b90U,
    0x30ef3e08U,
    0xd13a2b64U,
    0x6f2c2642U,
    0xc59baf35U,
    0xe3f918ecU,
    0xdb068480U,
    0xdf84bed9U,
    0xab4febbaU,
    0x3869c65eU,
    0xd3f31ff5U,
    0xa30e6033U,
    0xa38d70e4U,
    0xc54a98fcU,
    0xcb6f44d1U,
    0xd87b5e47U,
    0xf2405bc6U,
    0xf2ef3e4eU,
    0xa43a2bc2U,
    0x63ac26a3U,
    0xd55baf8bU,
    0xb2d91846U,
    0x459684a6U,
    0xa3fcbee1U,
    0xbb03ebbeU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xf8000000U,
    0x34000000U,
    0x76000000U,
    0x4d000000U,
    0x88800000U,
    0x93c00000U,
    0x9e200000U,
    0x5fd00000U,
    0x40d80000U,
    0x9b540000U,
    0x511a0000U,
    0xde710000U,
    0x7b4e8000U,
    0x7de5c000U,
    0xd9fea000U,
    0x9b0c7000U,
    0x558bf800U,
    0xb34f4400U,
    0x71e0e600U,
    0x7bf6fb00U,
    0x1002e680U,
    0x6803fbc0U,
    0x9c0e6660U,
    0xa033b70U,
    0xf7044698U,
    0x878b8b44U,
    0x20479eeeU,
    0xc8697f09U,
    0xdab02066U,
    0x12e9b09aU,
    0x1a755878U,
    0xd5433405U,
    0x54eb1ed8U,
    0x6f79bf2eU,
    0x89c200f1U,
    0x1250004U,
    0x4454809cU,
    0x6a94c044U,
    0xa43020d5U,
    0xc29b0c8U,
    0xecd55890U,
    0xa953345dU,
    0x72131e5fU,
    0x9ffdbf63U,
    0xee000088U,
    0x90000d9U,
    0xe68000d1U,
    0x5ac0001aU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0x68000000U,
    0x44000000U,
    0x32000000U,
    0x19000000U,
    0x5d800000U,
    0x52400000U,
    0x2b600000U,
    0x5c100000U,
    0xd9980000U,
    0x7dc0000U,
    0xcab20000U,
    0xa5a50000U,
    0x14bb8000U,
    0x6aa3c000U,
    0xf232e000U,
    0x64667000U,
    0xfa902800U,
    0x3752cc00U,
    0x7f7c9a00U,
    0xf1802b00U,
    0x64471a80U,
    0xe063eb40U,
    0x6895fa60U,
    0xbe559b30U,
    0x3afdd208U,
    0x8fcb5774U,
    0x392b483aU,
    0x97727c6dU,
    0xf585d2e7U,
    0xb647577fU,
    0x896148acU,
    0x5d1b7c13U,
    0xa814527dU,
    0x239d9760U,
    0xcada288dU,
    0xbd3bcca8U,
    0xc2ed1afeU,
    0x145aebfdU,
    0x4ffc7a40U,
    0x44435bfaU,
    0x306cb2afU,
    0x9092e7e9U,
    0xe25b80efU,
    0x24f3c0e9U,
    0xe0cae048U,
    0x4faa7096U,
    0x81ba2843U,
    0xd12bccfaU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x28000000U,
    0xcc000000U,
    0x4a000000U,
    0x63000000U,
    0x6e800000U,
    0x26400000U,
    0x3ee00000U,
    0xa1700000U,
    0x4bf80000U,
    0x5bb40000U,
    0x845a0000U,
    0x94ad0000U,
    0xf5978000U,
    0x198f4000U,
    0x26c1e000U,
    0x1da2d000U,
    0xa81fb800U,
    0xe94c9c00U,
    0xaa619200U,
    0x2c37d500U,
    0xa7961280U,
    0xbe8895c0U,
    0x5e4ff260U,
    0x4aee45b0U,
    0x7f724a48U,
    0x86fbd97cU,
    0xd03e5802U,
    0xe59e4c1fU,
    0x81862aecU,
    0x22cf49f9U,
    0xebad80b2U,
    0xa91240e8U,
    0x28ce60b1U,
    0xa8a990cfU,
    0x579c58b7U,
    0x56874c84U,
    0x724baa4eU,
    0x70ed0914U,
    0x347be0f2U,
    0x247fd02dU,
    0xbc703821U,
    0xb877dca8U,
    0x4e7a72b2U,
    0x4f780556U,
    0x8efe2acdU,
    0x8c3b4951U,
    0xf7978000U,
    0x68f40beU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x98000000U,
    0x6c000000U,
    0x16000000U,
    0x45000000U,
    0x5a800000U,
    0x77c00000U,
    0x93e00000U,
    0xc9d00000U,
    0x5580000U,
    0xf7940000U,
    0x1ef60000U,
    0x70eb0000U,
    0x7e5a8000U,
    0x2c184000U,
    0x21b72000U,
    0xe3cc7000U,
    0xd9e55800U,
    0xf2d16400U,
    0x3ed5fa00U,
    0x38dd5700U,
    0xc5d77a80U,
    0x235117c0U,
    0xda965aa0U,
    0x60766710U,
    0xe5298238U,
    0xd2bf437cU,
    0xec4b582eU,
    0x4cae6439U,
    0xaaf97af4U,
    0xeaee178eU,
    0xdd5adac7U,
    0x7b952757U,
    0xf8fc227aU,
    0x7def731cU,
    0xd0d720eaU,
    0x21dc7045U,
    0xe15d5858U,
    0x3595649bU,
    0xf1fbfa90U,
    0xe1625716U,
    0x7a1bfac7U,
    0x84b25776U,
    0x4943fad4U,
    0xe626575cU,
    0x9535fab3U,
    0x8d0d5738U,
    0x6e8f7ab3U,
    0x2dc5171bU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0xb8000000U,
    0x54000000U,
    0xce000000U,
    0x93000000U,
    0xa1800000U,
    0xe3400000U,
    0x76600000U,
    0x13b00000U,
    0xc8380000U,
    0x66740000U,
    0xfa920000U,
    0x2fef0000U,
    0x4fc8000U,
    0xc05dc000U,
    0x8ecb2000U,
    0x7f211000U,
    0x30d18800U,
    0xd50f7c00U,
    0xce88ba00U,
    0xcdd100U,
    0x4c2c3a80U,
    0x215411c0U,
    0x5e4d1ae0U,
    0x54ee0110U,
    0x89721258U,
    0xd913bd44U,
    0x75ad0896U,
    0xe612bcd7U,
    0xee239ab7U,
    0xe85cc1f4U,
    0xc2c5b221U,
    0x552f6df7U,
    0x5d7a031U,
    0xb8cd015U,
    0x1642a83dU,
    0xc8ea6c3dU,
    0x1b733236U,
    0xb819ad5dU,
    0x652a0057U,
    0xaddb0046U,
    0x878e80f7U,
    0x5c42c065U,
    0x2defa072U,
    0x7df8d04bU,
    0x34d0a8c0U,
    0x63056c51U,
    0x698fb2b3U,
    0xbf446dacU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0xb8000000U,
    0x4000000U,
    0x3e000000U,
    0x3b000000U,
    0x2e800000U,
    0x80400000U,
    0x5a00000U,
    0x8df00000U,
    0x51780000U,
    0x88340000U,
    0xb2160000U,
    0x1c250000U,
    0x61368000U,
    0xf592c000U,
    0x1bef2000U,
    0x43559000U,
    0x6a0bf800U,
    0x6d0e3c00U,
    0xb980c600U,
    0x1cc2c900U,
    0x3cee4680U,
    0xa7d409c0U,
    0x774f66a0U,
    0x492099f0U,
    0x1cbc1e18U,
    0xf75d65f4U,
    0x2c0b7826U,
    0x7209fccfU,
    0x99016688U,
    0x1f81998fU,
    0x13c49e2dU,
    0x946ea5f2U,
    0x189cd8e4U,
    0x626fac4eU,
    0x7f9d3e50U,
    0x66e9f5adU,
    0x72d80081U,
    0xcac40013U,
    0x6bee00a9U,
    0x1b5100a9U,
    0x7e008031U,
    0xdb07c0ceU,
    0x3e81a0c3U,
    0x84350f9U,
    0xb9aad839U,
    0xb7faac81U,
    0x5473be4eU,
    0x9dbf35fbU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x38000000U,
    0xac000000U,
    0xda000000U,
    0x3b000000U,
    0x73800000U,
    0xa400000U,
    0xbd600000U,
    0x95100000U,
    0x8d980000U,
    0xac540000U,
    0x77e0000U,
    0xdd0f0000U,
    0x6a848000U,
    0x4ec24000U,
    0xea5e000U,
    0x21375000U,
    0x7a6be800U,
    0x64958c00U,
    0xa0ddfa00U,
    0x90b01b00U,
    0x37277a80U,
    0xe87d5bc0U,
    0xb8861a20U,
    0x31c84bd0U,
    0x93281218U,
    0x767a977cU,
    0x6d8600c2U,
    0x1f4b0047U,
    0x33e28031U,
    0x25d9408dU,
    0xe63f602cU,
    0x8bea1008U,
    0x49d28899U,
    0x9c349c08U,
    0xa0edf244U,
    0x25dc7eeU,
    0x3a75e8ddU,
    0xc78a8c6aU,
    0xac417a4eU,
    0xc4665b0eU,
    0x619c9aa1U,
    0x16550bbaU,
    0x8c717244U,
    0x868b8770U,
    0xf4ce0888U,
    0x85a2dc4bU,
    0x7ab612aaU,
    0xe425973fU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xf8000000U,
    0xe4000000U,
    0x62000000U,
    0xdf000000U,
    0x8d800000U,
    0xfd400000U,
    0x1600000U,
    0x2bd00000U,
    0xe5580000U,
    0xd3940000U,
    0xb2b60000U,
    0x9a830000U,
    0xdcc48000U,
    0xf226c000U,
    0xdbb2e000U,
    0xe2021000U,
    0x1f014800U,
    0xad8edc00U,
    0xad4fea00U,
    0xf96a2700U,
    0xcfdd6a80U,
    0x875fe7c0U,
    0xc930aa0U,
    0x3f3f3790U,
    0x67cea258U,
    0xdda4fb74U,
    0xd9f2803aU,
    0x3ee5c0abU,
    0x31966037U,
    0xadb4d096U,
    0x370ba816U,
    0x7188ccedU,
    0xb40a20bU,
    0x1463fbdaU,
    0x6558005bU,
    0x1394005fU,
    0x92b600aaU,
    0xca8300c0U,
    0x24c480a8U,
    0x1626c0bdU,
    0xb9b2e0c8U,
    0x3d0210f8U,
    0x92814838U,
    0x50cedcb9U,
    0xac2feaa7U,
    0xd2ba271aU,
    0x2a856a14U,
    0x54cbe7ffU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x98000000U,
    0xd4000000U,
    0x6e000000U,
    0x1f000000U,
    0xf800000U,
    0x4ac00000U,
    0x3de00000U,
    0x2b500000U,
    0x47d80000U,
    0x709c0000U,
    0x9ab60000U,
    0x36870000U,
    0xf6468000U,
    0x72afc000U,
    0xa5732000U,
    0xdebd000U,
    0xe35ea800U,
    0xabd14400U,
    0xda9b7a00U,
    0x73babf00U,
    0xf20bfa80U,
    0xdd027f40U,
    0x1a865a20U,
    0xbc4a6f50U,
    0x2ba5d2b8U,
    0xe9fbfb84U,
    0x6aa880d6U,
    0x3174c09bU,
    0xc3e3a059U,
    0xec531091U,
    0x3c5308c4U,
    0x445954aaU,
    0x2058f21bU,
    0xc65b2b0eU,
    0x955ea86fU,
    0x20d14467U,
    0x1b1b7ab6U,
    0x367abf0fU,
    0x586bfa96U,
    0x68927fadU,
    0xebe5a64U,
    0xf8866fadU,
    0xf94bd2ddU,
    0xe520fb69U,
    0x3bb80085U,
    0x5e0c00b8U,
    0xd70e00cfU,
    0xe38b0058U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x28000000U,
    0x84000000U,
    0x9a000000U,
    0xfd000000U,
    0xd5800000U,
    0xc5c00000U,
    0x5b600000U,
    0x3fb00000U,
    0x9d380000U,
    0x98740000U,
    0x37520000U,
    0x9c4b0000U,
    0xb0a18000U,
    0xdbd5c000U,
    0xb6026000U,
    0x2300b000U,
    0xd2856800U,
    0x7d4e7400U,
    0x6329e200U,
    0x251f0900U,
    0x63e26280U,
    0xc7f5c9c0U,
    0xe79382e0U,
    0xf6abb950U,
    0x40d50ac8U,
    0xc880bdd4U,
    0x4043e052U,
    0xd6a57029U,
    0x70df0807U,
    0x108ac42cU,
    0x7c468a3cU,
    0xe0ae7d83U,
    0x93d80009U,
    0xa204005fU,
    0x910a0044U,
    0xab8f006eU,
    0x32cb8069U,
    0x5beac064U,
    0xabf1e036U,
    0x999e7001U,
    0x1a68836U,
    0x405b04dcU,
    0x5cceeaecU,
    0x44e1cd9fU,
    0x4f76e879U,
    0x7d0b4e4U,
    0xd00a0262U,
    0x880a7965U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x8000000U,
    0x24000000U,
    0xee000000U,
    0xcb000000U,
    0x1a800000U,
    0xdb400000U,
    0xd5600000U,
    0xdab00000U,
    0xf5380000U,
    0x67f40000U,
    0xf5a0000U,
    0xa9c10000U,
    0x5a298000U,
    0x9d7c000U,
    0x708f2000U,
    0x6e4c9000U,
    0x5cefb800U,
    0x87754c00U,
    0x39157a00U,
    0x692aa100U,
    0x1f5efa80U,
    0x31c861c0U,
    0x26225a60U,
    0x7bd23170U,
    0x798b4268U,
    0x75cc2d54U,
    0xb826a086U,
    0x58db509fU,
    0xf700981cU,
    0x4889dc84U,
    0x8242c229U,
    0x76ebedeeU,
    0x12718054U,
    0xe093c06dU,
    0x8ced20d5U,
    0xff79907fU,
    0xf51c387dU,
    0x13238c39U,
    0x3253da3eU,
    0xc441f162U,
    0x9e662e3U,
    0xbef5bd52U,
    0x6cda98faU,
    0x8108dc23U,
    0xff8b4217U,
    0xeacc2d42U,
    0xa4a6a035U,
    0xdc9b5022U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x68000000U,
    0x9c000000U,
    0x6e000000U,
    0x9f000000U,
    0xd3800000U,
    0x1c400000U,
    0xcea00000U,
    0xd7f00000U,
    0xc9780000U,
    0xa7bc0000U,
    0x41da0000U,
    0x134d0000U,
    0x45238000U,
    0x5fbe4000U,
    0x85d6a000U,
    0xe94c5000U,
    0x5829e800U,
    0x1532ec00U,
    0x49106600U,
    0x862e2900U,
    0xc231e680U,
    0x16916940U,
    0xec66c6e0U,
    0x67927930U,
    0xe0e00e88U,
    0xc85f85acU,
    0x420ca0e6U,
    0xb9015033U,
    0xc08a68b5U,
    0xb9ccac6fU,
    0xb9e6c61bU,
    0x38d279c8U,
    0xf3c00e3aU,
    0xecef85b3U,
    0xfe54a075U,
    0x490d504fU,
    0xe888680bU,
    0x45cdacd0U,
    0xa7e7464eU,
    0xcfdd3901U,
    0xbc4f2ee4U,
    0x9ea09573U,
    0xaff0682cU,
    0x4d71acdeU,
    0x3dbd4659U,
    0x2cd039f0U,
    0x31ccae31U,
    0x15eed515U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0xa8000000U,
    0xbc000000U,
    0x66000000U,
    0x3b000000U,
    0x1b800000U,
    0x66c00000U,
    0x74600000U,
    0xfb00000U,
    0x51780000U,
    0x111c0000U,
    0xf4aa0000U,
    0xd70000U,
    0x25418000U,
    0x76a04000U,
    0x39d12000U,
    0x47c31000U,
    0x12eae800U,
    0x79f9a400U,
    0xa05bf200U,
    0x1804e100U,
    0x40ad280U,
    0x1207f140U,
    0xb1003ae0U,
    0xee8e5510U,
    0xfa43c848U,
    0x6f26b4acU,
    0x261b1aaeU,
    0x312a45d7U,
    0x2910a055U,
    0xc0a350a1U,
    0xeadbc869U,
    0x804ab402U,
    0x42291a96U,
    0x9e9145c4U,
    0x8632037U,
    0x49b81065U,
    0x5a7968fbU,
    0xf292e451U,
    0x86615234U,
    0xaeb0b182U,
    0x77f09ab9U,
    0x75d0522U,
    0x55800084U,
    0xa1c00050U,
    0x69e0002fU,
    0x27000e6U,
    0x969800b3U,
    0xc46c0035U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0xf8000000U,
    0x64000000U,
    0xa2000000U,
    0x73000000U,
    0xca800000U,
    0x3bc00000U,
    0x5ca00000U,
    0x1c700000U,
    0x33b80000U,
    0x15140000U,
    0xce6a0000U,
    0x5d510000U,
    0x9c8a8000U,
    0x2c6c000U,
    0xa92de000U,
    0xe3385000U,
    0x19d63800U,
    0x2a4a6400U,
    0xc1e2fe00U,
    0xa51d8100U,
    0xf66f1e80U,
    0xd955d1c0U,
    0xe812660U,
    0x89cbb5f0U,
    0x7a9d898U,
    0x7af73494U,
    0x367446baU,
    0xfcb02527U,
    0xc9928010U,
    0xeca2c0ecU,
    0x247fe0d4U,
    0xb7bd5064U,
    0x8716b85dU,
    0x456da456U,
    0xf3dd9e03U,
    0x5471127U,
    0x2d66c62bU,
    0x7fd2e5b1U,
    0xeb4d60c7U,
    0x506f9050U,
    0xf8515807U,
    0xaf04f48aU,
    0x2c8126e5U,
    0x3acbb526U,
    0x2d29d85aU,
    0x713734dcU,
    0x92d446e6U,
    0x84c02501U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0xf8000000U,
    0xc000000U,
    0xca000000U,
    0x53000000U,
    0xef800000U,
    0x7ac00000U,
    0xde600000U,
    0x8b100000U,
    0xc5d80000U,
    0x3dbc0000U,
    0x9fa60000U,
    0xeff0000U,
    0x520c8000U,
    0xaf0a4000U,
    0x6d89e000U,
    0xcdc85000U,
    0xfe8e800U,
    0x64dca400U,
    0x6d301200U,
    0x70ebcb00U,
    0x1159f280U,
    0x1cf39b40U,
    0x5d091a20U,
    0xd2833f50U,
    0x584708d8U,
    0xce2bf45cU,
    0xa1347a92U,
    0x5aed2f4fU,
    0xf258005dU,
    0x1b7c0065U,
    0xd3c6005bU,
    0xcaef00b2U,
    0x4a54800cU,
    0xb77640c0U,
    0x9cfe04eU,
    0x61e750ebU,
    0xa9dc68e7U,
    0x7bae4e6U,
    0x84a7f27eU,
    0x5709be7U,
    0x16c39a75U,
    0xe4667f7bU,
    0x901a68b0U,
    0xce55e487U,
    0x79737279U,
    0xd4c6dbbbU,
    0xf36c7afeU,
    0x51912f6cU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x48000000U,
    0x94000000U,
    0xd6000000U,
    0xa1000000U,
    0x65800000U,
    0x7dc00000U,
    0x5fe00000U,
    0xd6b00000U,
    0x59780000U,
    0xf9940000U,
    0x1fa20000U,
    0xb8990000U,
    0x8a298000U,
    0xed52c000U,
    0x71cea000U,
    0xa5e13000U,
    0x1db3a800U,
    0xeff9d400U,
    0xded99a00U,
    0x6d0e1900U,
    0xff8f3a80U,
    0x86cb29c0U,
    0xa16692e0U,
    0x65fffdf0U,
    0xfdd488a8U,
    0x1f8a2464U,
    0x76c492feU,
    0x8966fd05U,
    0xc1fd087bU,
    0x63d8e488U,
    0x2a8a328cU,
    0xc547cd3aU,
    0x55aea0abU,
    0xfb913006U,
    0xc8aba8afU,
    0x2c1dd406U,
    0xea639ae1U,
    0x13731945U,
    0xba9cbaedU,
    0x5d24e947U,
    0x79dbb23aU,
    0x11810dd3U,
    0x5bc2005aU,
    0xd6e9002dU,
    0x17318011U,
    0xbab6c015U,
    0x9374a015U,
    0x7a9c30dbU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0x8000000U,
    0x6c000000U,
    0x7a000000U,
    0x77000000U,
    0x74800000U,
    0x9c00000U,
    0xabe00000U,
    0xff900000U,
    0x51580000U,
    0xa6340000U,
    0x4c6a0000U,
    0x54db0000U,
    0x1e7f8000U,
    0x61cb4000U,
    0x97e3e000U,
    0x6d911000U,
    0xda561800U,
    0xa0b2e400U,
    0x5ea60a00U,
    0xf1be0b00U,
    0x9f2fea80U,
    0xeff41bc0U,
    0xc7867260U,
    0xdb4dbf50U,
    0xd7239868U,
    0xa3f2a43cU,
    0xcd826a92U,
    0x34405b8bU,
    0xc7a81286U,
    0xbc38efd2U,
    0x6b63e0c5U,
    0xc85110d1U,
    0xebb61866U,
    0xb822e4acU,
    0x737e0ac4U,
    0x324a0b16U,
    0x2a5ea8dU,
    0x33bf1b2dU,
    0xfc21f27aU,
    0x1572fffeU,
    0xa74a78c1U,
    0x6528b4c9U,
    0x58f3f29cU,
    0x530dff45U,
    0x4287f8cfU,
    0x74ccf459U,
    0x306592caU,
    0xe6dcafd8U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x8000000U,
    0x7c000000U,
    0x5a000000U,
    0x33000000U,
    0xb8800000U,
    0x92c00000U,
    0x2fe00000U,
    0xafd00000U,
    0xdc180000U,
    0x447c0000U,
    0x3be60000U,
    0xf9d70000U,
    0x8d1b8000U,
    0x2bf24000U,
    0x3fa4e000U,
    0x51f65000U,
    0xdca69800U,
    0x417bd400U,
    0xa260ba00U,
    0x3c9d7f00U,
    0x42ba5a80U,
    0x2002f40U,
    0xa7014220U,
    0xae8ebbd0U,
    0x63c69828U,
    0xd06bd4acU,
    0xa398baf2U,
    0x8317fdfU,
    0x4fc45a6aU,
    0x626b2f9dU,
    0xdc9cc2edU,
    0x72bbfbdeU,
    0x9a01f863U,
    0xd303c4d5U,
    0x8884c23aU,
    0xac7fbcdU,
    0x5be7f800U,
    0x89d4c40bU,
    0xb51f42e4U,
    0xcff5bb1dU,
    0x11a318adU,
    0x44f2947eU,
    0xd21da73U,
    0x58326f9dU,
    0xa7c5a266U,
    0x2e68eb47U,
    0x1e98009bU,
    0x35bc005fU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0xf8000000U,
    0x1c000000U,
    0x62000000U,
    0x77000000U,
    0xf3800000U,
    0x93c00000U,
    0xcb600000U,
    0x3500000U,
    0x19980000U,
    0xf4740000U,
    0x6a0000U,
    0xe2d70000U,
    0xb55c8000U,
    0x789f4000U,
    0x16f26000U,
    0x392f9000U,
    0xa4f6e800U,
    0x762c3400U,
    0xab7bd200U,
    0xd7ecaf00U,
    0xef1bb280U,
    0x3b303fc0U,
    0x743da60U,
    0x73204b90U,
    0xf7fce898U,
    0x1bab348cU,
    0x7dbf527aU,
    0x6007ef3bU,
    0x9003d2e9U,
    0x1808af38U,
    0x4c09b23aU,
    0x9a033f77U,
    0x6b0d5ab9U,
    0x918c0be8U,
    0xe4c00828U,
    0x38e8e4eeU,
    0x90955a25U,
    0xd2f80bc6U,
    0xf72a0881U,
    0x19ffe498U,
    0x16a9daa4U,
    0xb5374b74U,
    0x9a406840U,
    0xa3a474c2U,
    0x41b53205U,
    0xb20c7fd0U,
    0x4f073a72U,
    0xf879b87U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0x48000000U,
    0xbc000000U,
    0x2000000U,
    0xdf000000U,
    0xb8800000U,
    0xb1c00000U,
    0xc8600000U,
    0x77d00000U,
    0x8f180000U,
    0xcefc0000U,
    0xc5620000U,
    0xf85b0000U,
    0xca538000U,
    0xed524000U,
    0x51dd6000U,
    0x6151000U,
    0xe771d800U,
    0x49212400U,
    0x12787200U,
    0xc2a57700U,
    0xb13f1280U,
    0xc8076740U,
    0xfc074a60U,
    0xe2030350U,
    0xcf0bd828U,
    0xf08624ecU,
    0xdc9f2aaU,
    0xca6c3773U,
    0xa8d1f272U,
    0x37903792U,
    0x7f33f212U,
    0xd0b3749U,
    0x8f8072d7U,
    0x454977e4U,
    0x23a512a8U,
    0x94b06772U,
    0xfe4eca12U,
    0x2d264359U,
    0xa47f3804U,
    0x43a674e6U,
    0xc4b4ca29U,
    0x564143d9U,
    0x812eb854U,
    0xee7f345eU,
    0x20a22a56U,
    0x7e3a1391U,
    0x38800036U,
    0xf1c0000fU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xa8000000U,
    0x94000000U,
    0xde000000U,
    0x11000000U,
    0xf1800000U,
    0xb8c00000U,
    0x95200000U,
    0xad700000U,
    0x94380000U,
    0x71d40000U,
    0x4d8e0000U,
    0x32c50000U,
    0xfa238000U,
    0x7df0c000U,
    0x75762000U,
    0xc831b000U,
    0xabdde800U,
    0x1a888400U,
    0xee424600U,
    0xf962c900U,
    0x9f17e680U,
    0xada3b9c0U,
    0xb83c2e20U,
    0x63da8df0U,
    0x1e838088U,
    0xa840c064U,
    0xd46e20d6U,
    0x2495b0b5U,
    0xda6be887U,
    0xcd99843dU,
    0x87efc6baU,
    0xa7570904U,
    0xe1c24670U,
    0xa0a2c9a4U,
    0xf3b7e66cU,
    0x1c13b91eU,
    0xcf242eabU,
    0x3a7e8d5aU,
    0xe8b58094U,
    0x4291c0f2U,
    0x763a003U,
    0x7e1070a2U,
    0x54264822U,
    0xa4fcf4bbU,
    0x80f20ef4U,
    0x36ff3d3bU,
    0x13f068f4U,
    0xc7d44a1U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x88000000U,
    0x2c000000U,
    0x66000000U,
    0x75000000U,
    0xe3800000U,
    0x2fc00000U,
    0xbaa00000U,
    0xeed00000U,
    0xea980000U,
    0xf3f40000U,
    0xd1820000U,
    0x58cd0000U,
    0x6298000U,
    0x419d4000U,
    0x61742000U,
    0x47c7f000U,
    0x26a50800U,
    0x60da9400U,
    0xc390ee00U,
    0xfe79fb00U,
    0xa74d4e80U,
    0x67e34bc0U,
    0xb23c66e0U,
    0xf22e2fb0U,
    0x53918068U,
    0xc679409cU,
    0x634e208eU,
    0x5deef029U,
    0x2936880dU,
    0x48aed476U,
    0xf9d74e3fU,
    0x261a4bb4U,
    0xd4b7e633U,
    0x766e6ff2U,
    0x8af42061U,
    0x8407f0f5U,
    0x9a0508d5U,
    0x8b0a94b6U,
    0x4288ee58U,
    0xe4dfb07U,
    0xaa6f4e44U,
    0xa4fe4bfcU,
    0xbd0de63dU,
    0x6f876fa8U,
    0x59c7a07aU,
    0x37a3b078U,
    0xa95aa863U,
    0x8f5d24e9U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0xb8000000U,
    0x9c000000U,
    0xc6000000U,
    0xb5000000U,
    0xab800000U,
    0x79400000U,
    0xc600000U,
    0x78b00000U,
    0xe2780000U,
    0xc35c0000U,
    0x65860000U,
    0x38470000U,
    0xe5e88000U,
    0x7efb4000U,
    0x8e962000U,
    0x4022f000U,
    0x9f538800U,
    0x438d7c00U,
    0xbd4b6a00U,
    0x66689900U,
    0x23b5ca80U,
    0xd8f12940U,
    0x4b906220U,
    0x63aea5d0U,
    0x52108098U,
    0xade7404cU,
    0x2af020deU,
    0x5c95f0b9U,
    0x872308d5U,
    0x63da3c50U,
    0x7e434a61U,
    0x10e169b4U,
    0x7570c2c5U,
    0x67dc1582U,
    0xf443a8abU,
    0x7be88c53U,
    0x67f0627aU,
    0xcb1ea589U,
    0xa86880f0U,
    0x62bb406fU,
    0x31762039U,
    0x4dd2f021U,
    0xf4b88b3U,
    0xd1617cf8U,
    0x57356a29U,
    0x513399f3U,
    0x43b4a79U,
    0x9fbd691fU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0xe8000000U,
    0xc4000000U,
    0x3e000000U,
    0x47000000U,
    0xf4800000U,
    0x83c00000U,
    0xdc600000U,
    0xb5d00000U,
    0x74980000U,
    0x38340000U,
    0x23060000U,
    0x7a890000U,
    0xccb8000U,
    0x9ce5c000U,
    0xe0112000U,
    0x2bf3d000U,
    0x476cd800U,
    0x935d0400U,
    0x6a57b200U,
    0x19d32700U,
    0x4e951280U,
    0x613137c0U,
    0xe08eeaa0U,
    0xe5c6e3b0U,
    0x97660048U,
    0xcb590074U,
    0x665380f6U,
    0x93d1c0f3U,
    0xdf972022U,
    0x16bad000U,
    0xa9c75816U,
    0xfd68c471U,
    0xa5e925cU,
    0x89d4f70eU,
    0xf69fca0bU,
    0x3d353337U,
    0xf28ad8fbU,
    0x58c4046eU,
    0x1ae43207U,
    0xfb12e779U,
    0xcd7a32fdU,
    0x79afe767U,
    0xc2b7b2a7U,
    0xefc327d5U,
    0xc66d1203U,
    0x9cd537a9U,
    0x5f10ea1aU,
    0x637be323U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x28000000U,
    0x14000000U,
    0x92000000U,
    0xbd000000U,
    0xa0800000U,
    0xe400000U,
    0x94e00000U,
    0xe6500000U,
    0x5d980000U,
    0x33bc0000U,
    0x7d460000U,
    0x5f650000U,
    0x5f938000U,
    0x56b0c000U,
    0x41c6e000U,
    0x672db000U,
    0xcc79f800U,
    0xb960bc00U,
    0x20946200U,
    0x733daf00U,
    0x63070280U,
    0x5385df40U,
    0x85cb9a20U,
    0x8d2813f0U,
    0x5d78e008U,
    0x7e4b0e4U,
    0x3dd4781aU,
    0x7c597c19U,
    0x49f021aU,
    0x8939dfa7U,
    0x6a0d9a26U,
    0xd10d1315U,
    0x1e8b6049U,
    0xad44702bU,
    0x276a983cU,
    0x7398cc2eU,
    0x48b8fac5U,
    0x52c06393U,
    0xfcac78d1U,
    0x4db57c9dU,
    0x3e410246U,
    0xfce0dfd0U,
    0x52581accU,
    0x7f98d32cU,
    0xa6be001dU,
    0xc9c900b8U,
    0xc32d80ccU,
    0x7679c00eU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0x68000000U,
    0xb4000000U,
    0xfa000000U,
    0xcf000000U,
    0xb9800000U,
    0x67c00000U,
    0x27600000U,
    0x3d700000U,
    0xf8380000U,
    0xeb1c0000U,
    0x61c60000U,
    0x16610000U,
    0x79f38000U,
    0xad7ac000U,
    0x803e6000U,
    0x671e1000U,
    0xb7c1c800U,
    0xff669400U,
    0x41775200U,
    0x4632bd00U,
    0xb61cb280U,
    0xa3476d40U,
    0x32289a20U,
    0xf65929b0U,
    0xee66048U,
    0x22b21004U,
    0x535fc832U,
    0x6b948bU,
    0x70fad22bU,
    0xc5f57d1cU,
    0xf37752e4U,
    0x8d32bdd5U,
    0x3d9cb246U,
    0x4f876d01U,
    0x3ec89af6U,
    0xd7e929c4U,
    0x923e6052U,
    0x5c1e109bU,
    0x5441c813U,
    0xa7a69430U,
    0xb7975242U,
    0xa882bd14U,
    0x9344b26aU,
    0xba2b6de0U,
    0x12569a67U,
    0x6ce4290dU,
    0x31b3e04aU,
    0xa4d9d07aU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xc8000000U,
    0xec000000U,
    0x42000000U,
    0x1f000000U,
    0x77800000U,
    0xe5c00000U,
    0xe8600000U,
    0xa9500000U,
    0x70180000U,
    0x9c340000U,
    0x71c20000U,
    0x1e6f0000U,
    0xb05f8000U,
    0xe69f4000U,
    0x9b726000U,
    0xf8e91000U,
    0x8f1da800U,
    0x7bbf5400U,
    0x3c04e200U,
    0x4a062f00U,
    0x930b0280U,
    0x658f7fc0U,
    0x32c34ae0U,
    0x73e27b90U,
    0xe926028U,
    0x8779107cU,
    0x42e5a8eaU,
    0xe41b54a3U,
    0x6a3ee2fdU,
    0x68cd2f16U,
    0x88ee825dU,
    0xb71b3f93U,
    0x6fb4aa0fU,
    0xa0f2b40U,
    0x330028c1U,
    0x558f1457U,
    0xaac9025bU,
    0x57e07fc7U,
    0xa09cca27U,
    0xda7d3bd6U,
    0x2a600099U,
    0x765000a6U,
    0x6798006fU,
    0x29f400a5U,
    0x51a20043U,
    0x5b3f008cU,
    0x82478019U,
    0x65ab40c2U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0xd8000000U,
    0xf4000000U,
    0x2e000000U,
    0x73000000U,
    0x1e800000U,
    0x67400000U,
    0x9f600000U,
    0xde700000U,
    0x7ab80000U,
    0x2f140000U,
    0xb7ce0000U,
    0x3b2b0000U,
    0xf79a8000U,
    0x3f87c000U,
    0xc8cc2000U,
    0xefa09000U,
    0xb5d93800U,
    0xe9652c00U,
    0x997d2a00U,
    0xea328500U,
    0xeb5d8a80U,
    0xee2ad5c0U,
    0x761c1220U,
    0x3c43a910U,
    0xbdeea0f8U,
    0x1b3350e4U,
    0x9cdb1856U,
    0x6aeebc57U,
    0x33be92e8U,
    0xdc9069e0U,
    0x150c802fU,
    0x2188c00aU,
    0xd3c0a05bU,
    0xad285046U,
    0x99988aU,
    0x70d7c3eU,
    0xf084b214U,
    0xb44ff958U,
    0x51e13816U,
    0x61312c4cU,
    0x11d32ab5U,
    0xdf698537U,
    0xbe7f0ac6U,
    0xab915f7U,
    0x271e327eU,
    0x9bc839b5U,
    0xe12d1815U,
    0xaa91bc84U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0x28000000U,
    0xec000000U,
    0xca000000U,
    0x51000000U,
    0x17800000U,
    0xf1c00000U,
    0x65200000U,
    0xc7500000U,
    0xd2180000U,
    0xcc740000U,
    0x524a0000U,
    0x5aed0000U,
    0x35738000U,
    0x69cb4000U,
    0x812ce000U,
    0xf1597000U,
    0x911fc800U,
    0x4efdf400U,
    0x920cf600U,
    0x55049b00U,
    0xd1819680U,
    0x9acfabc0U,
    0xbab3e60U,
    0xcd1d6f30U,
    0x8cff6048U,
    0x3f0230dcU,
    0x508b2802U,
    0xb540844dU,
    0xce613ef5U,
    0xa6306f4cU,
    0x3cace038U,
    0x619970a7U,
    0x2bbfc840U,
    0x646df40aU,
    0xc734f6cdU,
    0xe3209bbdU,
    0x8c5396ffU,
    0xac96ab6eU,
    0x1e32be91U,
    0xc8ab2f03U,
    0x8f980008U,
    0xacb400cdU,
    0xc0ea0052U,
    0x9c7d006bU,
    0xaa4b8075U,
    0x8eef4097U,
    0xcb7ee0c5U,
    0x36c0700cU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0xf8000000U,
    0x3c000000U,
    0x4e000000U,
    0x19000000U,
    0x56800000U,
    0x7e400000U,
    0x65200000U,
    0xa7500000U,
    0xe2780000U,
    0xb52c0000U,
    0xf5e0000U,
    0x36790000U,
    0x8f288000U,
    0x9c524000U,
    0xbfe2000U,
    0xf0edf000U,
    0xc8b6b800U,
    0xe14cac00U,
    0xaea44600U,
    0x5d9e9f00U,
    0x1e12fe80U,
    0xc0d23340U,
    0x5eb6b820U,
    0x344cacf0U,
    0xee244658U,
    0xb6de9f8cU,
    0x9bb2fe36U,
    0x3cc23365U,
    0xc16eb838U,
    0x4170ac97U,
    0x30a2466bU,
    0xec9b9f55U,
    0x9c9c7eb1U,
    0x84957377U,
    0xa89e18d5U,
    0xae9a1c14U,
    0x739c5eebU,
    0x571183ffU,
    0x7e582035U,
    0x54f8f01bU,
    0xdb603896U,
    0x2277ec45U,
    0x552ae62dU,
    0x1f5d2ff3U,
    0x7e7c667dU,
    0x4b266fbaU,
    0xee52c618U,
    0x5cf5dfe8U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x98000000U,
    0x34000000U,
    0xbe000000U,
    0x59000000U,
    0xff800000U,
    0x1fc00000U,
    0xae600000U,
    0xb7700000U,
    0x1180000U,
    0xc26c0000U,
    0xdd7e0000U,
    0xd61b0000U,
    0x6cec8000U,
    0x9134c000U,
    0xedb3a000U,
    0x92f2f000U,
    0x81d6a800U,
    0xce89a400U,
    0x3c4bea00U,
    0x93a23900U,
    0xc21d4280U,
    0x2eb9d40U,
    0xf036a820U,
    0xb639a4d0U,
    0xab33ea38U,
    0x42be39a4U,
    0x387b42a6U,
    0x7b9c9d2dU,
    0x242861U,
    0xb7d66496U,
    0x7b8cca69U,
    0xe9c8090cU,
    0xfb664a09U,
    0x52fbc958U,
    0x61df6abdU,
    0xfe8af982U,
    0x3448e258U,
    0x3fae6d0bU,
    0x48128028U,
    0xe5efc0cbU,
    0x56bf20c9U,
    0x567630aaU,
    0x1a9d087bU,
    0x5ba754dfU,
    0x8e1b4291U,
    0xb8ec9da1U,
    0x1f3c2853U,
    0xbcba6414U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0xf8000000U,
    0xb4000000U,
    0x12000000U,
    0x39000000U,
    0x2d800000U,
    0x4bc00000U,
    0xa1e00000U,
    0xc8f00000U,
    0xf1180000U,
    0x3dec0000U,
    0xd6fe0000U,
    0xce170000U,
    0x6b668000U,
    0x71bac000U,
    0xbdbd6000U,
    0x5bb43000U,
    0xd0b38800U,
    0x943f0400U,
    0xa709200U,
    0x54501900U,
    0x2f431a80U,
    0x2faf1d40U,
    0xced38820U,
    0x6e0f0410U,
    0xd7089258U,
    0xba8c19e4U,
    0x51451a6aU,
    0xa0a41dcdU,
    0x9053081fU,
    0xd54ec462U,
    0xfaad72d4U,
    0x1555e967U,
    0xf6cd723aU,
    0xd665e938U,
    0x6357238U,
    0x7379e991U,
    0xd9d372e9U,
    0x3482e998U,
    0x764bf24cU,
    0xb22f2942U,
    0x8d1012adU,
    0xd3e1d9eaU,
    0x41fefa21U,
    0xd49aed61U,
    0x6a25e044U,
    0x2919f072U,
    0x99e8686dU,
    0x9cf1f423U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x98000000U,
    0x3c000000U,
    0xe2000000U,
    0xbb000000U,
    0x6c800000U,
    0x7a400000U,
    0xa0a00000U,
    0x70b00000U,
    0xcc180000U,
    0xcea40000U,
    0x71ba0000U,
    0x479d0000U,
    0xa6e8000U,
    0x88504000U,
    0xc700e000U,
    0x2e831000U,
    0x91413800U,
    0x642f8c00U,
    0xaef1be00U,
    0xb2bbb900U,
    0xe7108680U,
    0x6a2435c0U,
    0x9ff938e0U,
    0xa13b8cf0U,
    0x1fd3bef8U,
    0x71c2b90cU,
    0xa46406faU,
    0xe9597547U,
    0x7c8f586eU,
    0x724cdc31U,
    0x54a86634U,
    0x6b32506U,
    0x311a0096U,
    0xc72d00f9U,
    0x3e768053U,
    0x4af44078U,
    0xccbae02fU,
    0xee1e10f9U,
    0x15afb843U,
    0x2d3fcca3U,
    0xa5d15e87U,
    0x96c8a9b9U,
    0x1ae9be64U,
    0xb01fb9a9U,
    0x8caa86a8U,
    0x9ab93540U,
    0x8317b827U,
    0xd42bcc0aU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x68000000U,
    0x74000000U,
    0x72000000U,
    0x8b000000U,
    0x7f800000U,
    0x31400000U,
    0x43200000U,
    0x88700000U,
    0x49580000U,
    0x722c0000U,
    0x38f20000U,
    0x6d9f0000U,
    0x874b8000U,
    0xe223c000U,
    0x10f76000U,
    0x79901000U,
    0xe5428800U,
    0x1268c00U,
    0x1b7fb600U,
    0x3adc7d00U,
    0x2d653e80U,
    0xf6d6f140U,
    0x636888e0U,
    0xbd58c50U,
    0x9de63608U,
    0xd210bd64U,
    0x9601de9aU,
    0xd10921bfU,
    0xe08f60edU,
    0x2ccc10eaU,
    0x91688834U,
    0xc0d58cddU,
    0x82663610U,
    0xf350bd05U,
    0xbd21defcU,
    0x2d7921e5U,
    0xdbd76063U,
    0xd5e01066U,
    0xd61a8895U,
    0x9c0a8c05U,
    0x460db69eU,
    0x99037d98U,
    0xe48ebeccU,
    0x26c5314eU,
    0x667e87dU,
    0xb9599cbeU,
    0x4a2ebeb9U,
    0x44f5311bU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0x58000000U,
    0x1c000000U,
    0x6a000000U,
    0xff000000U,
    0xe9800000U,
    0x2400000U,
    0x5ea00000U,
    0xc5f00000U,
    0x4580000U,
    0x38240000U,
    0xb63e0000U,
    0x64b50000U,
    0x397e8000U,
    0xd0924000U,
    0xad4e6000U,
    0xdf2dd000U,
    0xa3b82800U,
    0xdcfb5c00U,
    0x30dcde00U,
    0xb8e9bd00U,
    0xdd5af680U,
    0x2ca7e1c0U,
    0xc6f8a8a0U,
    0x57dc1c30U,
    0x6d6c3e78U,
    0x45162decU,
    0xd50cbeb2U,
    0xf6816d23U,
    0x3bc45e23U,
    0xf4eafdcdU,
    0xdf54164fU,
    0xb7ad71ebU,
    0xa97060c8U,
    0x9898d02eU,
    0x3946a849U,
    0x1291c4dU,
    0x72b2be0aU,
    0xbc746d90U,
    0x4e1adecdU,
    0xf288bdcfU,
    0xadc276abU,
    0xb1e4a1a8U,
    0x61d648beU,
    0x18668c01U,
    0x839cf6d9U,
    0x96c6e1d4U,
    0xee6028c1U,
    0xd69f5cedU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0xa8000000U,
    0x34000000U,
    0x12000000U,
    0xbd000000U,
    0xb3800000U,
    0x50c00000U,
    0x6200000U,
    0x14d00000U,
    0xd7f80000U,
    0xbca40000U,
    0x259a0000U,
    0x5c110000U,
    0xedd18000U,
    0x1e7bc000U,
    0x56f6000U,
    0x72357000U,
    0xd082800U,
    0xbb85d400U,
    0xf4c05e00U,
    0x4c2ca700U,
    0x35d27680U,
    0x427873c0U,
    0x4363a8e0U,
    0x2d3f1430U,
    0xab86bec8U,
    0x6cc617c4U,
    0xb02f3e3aU,
    0xe3d9d749U,
    0x457a5e41U,
    0x79eda7ddU,
    0xdc7bf6fdU,
    0x4067b340U,
    0xcdb6c88bU,
    0x23cb6411U,
    0x8ea7169bU,
    0xc89c03c9U,
    0xd79a0047U,
    0xd111004fU,
    0x965180f8U,
    0x8abbc0d9U,
    0xb94f60e7U,
    0xefe57059U,
    0x7b7028f2U,
    0xeae1d4a0U,
    0x64fa5e68U,
    0x542da7d2U,
    0x9dbf630U,
    0xf477b330U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x38000000U,
    0x94000000U,
    0x2a000000U,
    0x77000000U,
    0xc8800000U,
    0x27400000U,
    0xcf600000U,
    0x42d00000U,
    0xeb80000U,
    0xdee40000U,
    0x58120000U,
    0xa51d0000U,
    0x8a9c8000U,
    0xd50c000U,
    0xf97ae000U,
    0xf2c75000U,
    0x77a83800U,
    0x32739c00U,
    0x2c429200U,
    0xf1e74300U,
    0x7498aa80U,
    0x7059dfc0U,
    0x16feb860U,
    0x950a5c70U,
    0x338ef2d8U,
    0x1c9d324U,
    0x3d24f292U,
    0x9e30d323U,
    0x1daa7282U,
    0x657d1320U,
    0xb4cc125fU,
    0x5eaa8381U,
    0x17fecab3U,
    0x8c8e4f0fU,
    0x454c60ecU,
    0xf46e90cfU,
    0x845c5827U,
    0x4cf90c9eU,
    0x3a0cca68U,
    0xdf034fa7U,
    0x3488e075U,
    0x114a5069U,
    0x3e6cb860U,
    0x43575c60U,
    0xbc72728aU,
    0x894913d3U,
    0x3a66129aU,
    0x15383e4U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0x88000000U,
    0x84000000U,
    0xe6000000U,
    0xe5000000U,
    0x4a800000U,
    0x25c00000U,
    0xfd600000U,
    0x11700000U,
    0xb5180000U,
    0xe4ec0000U,
    0x233e0000U,
    0x8b350000U,
    0x3f338000U,
    0xf136c000U,
    0xe0372000U,
    0xc4b9b000U,
    0x807b3800U,
    0xd1995400U,
    0x2c7a00U,
    0x2d39900U,
    0xd2894280U,
    0xb9cfcd40U,
    0xc76eb820U,
    0x6e769430U,
    0xf096da28U,
    0xca9e9f4U,
    0xb216daeeU,
    0xed69e921U,
    0x976da8cU,
    0x6919e9f0U,
    0x7eeeda1fU,
    0x2c35e980U,
    0x46b0da06U,
    0x5370e9a4U,
    0x861b5a92U,
    0x636a291bU,
    0xb8727ae5U,
    0x3d96991fU,
    0xb222c293U,
    0xf9d50db4U,
    0x7107980fU,
    0xb48a241fU,
    0x1cc66250U,
    0x2dea7d68U,
    0x3bb380c2U,
    0x85f6c0f4U,
    0x99d7202dU,
    0xe109b0fdU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0x28000000U,
    0xc4000000U,
    0x9a000000U,
    0x9b000000U,
    0xbf800000U,
    0x60400000U,
    0xf8200000U,
    0x3b100000U,
    0xe2380000U,
    0x4dac0000U,
    0xf85e0000U,
    0x199d0000U,
    0x90f18000U,
    0xf48c000U,
    0x15a4e000U,
    0x54579000U,
    0xe79b8800U,
    0x61f69c00U,
    0x3c3d600U,
    0xe66db00U,
    0x563e5e80U,
    0xbfa14740U,
    0x7520820U,
    0xcc125cb0U,
    0x43b93688U,
    0x8cec4b34U,
    0xd4f45632U,
    0xd54f1b1fU,
    0x2ead3e05U,
    0x1bdf174bU,
    0xafdae04fU,
    0x5dda902fU,
    0xa2d20808U,
    0x77525cd9U,
    0xa4193617U,
    0x27bc4b6bU,
    0xe6ec5635U,
    0x67f31b96U,
    0xaecb3e00U,
    0xd4ee17fdU,
    0x78f56018U,
    0x2b4f502eU,
    0xdfa7687cU,
    0x175d0c9bU,
    0xb41e5e25U,
    0x5fb147ccU,
    0xfaea086eU,
    0x11fe5cc9U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x38000000U,
    0xdc000000U,
    0xaa000000U,
    0x29000000U,
    0xb2800000U,
    0x83c00000U,
    0xd6600000U,
    0x65300000U,
    0x80580000U,
    0x86e40000U,
    0xbf20000U,
    0x42b50000U,
    0xeb148000U,
    0x22444000U,
    0xe6266000U,
    0xb616b000U,
    0xceccf800U,
    0xd2ebf400U,
    0x85feaa00U,
    0xfdb32d00U,
    0xea985280U,
    0x209d9c0U,
    0xdd007860U,
    0x2c8bb4f0U,
    0xd4cacad8U,
    0x3e09decU,
    0xb782a12U,
    0xe4726d35U,
    0x2df2b278U,
    0x9bf295aU,
    0x7498e03cU,
    0x5503f0caU,
    0x88c1824U,
    0x42cc0426U,
    0xb0e0b2abU,
    0xa8fa29b2U,
    0x93460aeU,
    0xd253b06fU,
    0x93e07809U,
    0x237bb4c4U,
    0xd072ca4dU,
    0x53f49d19U,
    0x6eb22a01U,
    0x99136d20U,
    0xe74c32bbU,
    0xc6aa695aU,
    0xc0d8007aU,
    0x240021U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0xa8000000U,
    0xb4000000U,
    0xe6000000U,
    0x15000000U,
    0x63800000U,
    0xf1c00000U,
    0xc5a00000U,
    0xedf00000U,
    0x6b580000U,
    0x682c0000U,
    0xfd320000U,
    0x4b7f0000U,
    0x71178000U,
    0x984bc000U,
    0x8b662000U,
    0xc4d2b000U,
    0x43efa800U,
    0xf928400U,
    0x8e093600U,
    0x810e8d00U,
    0xf58c9e80U,
    0x5ccf0940U,
    0xba2028e0U,
    0x4e354450U,
    0x5dfd16c8U,
    0xf3533da4U,
    0x42cb6ceU,
    0x173a4de1U,
    0xa47d3e65U,
    0x559679b4U,
    0xf909a0eeU,
    0xc98570f8U,
    0x7ec38880U,
    0xf1233434U,
    0x24bb1ebbU,
    0x9db4c973U,
    0x343e084cU,
    0x5afbf49aU,
    0x3d8be57U,
    0x9362b938U,
    0xe8d800b3U,
    0xc9ec00f1U,
    0xf09200acU,
    0x28f0030U,
    0x544f80b7U,
    0x5167c08fU,
    0xf3d420d5U,
    0x6b6db0b9U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0xa8000000U,
    0x9c000000U,
    0xea000000U,
    0xbf000000U,
    0x54800000U,
    0x12400000U,
    0x33a00000U,
    0xd5900000U,
    0x47380000U,
    0x2f2c0000U,
    0x6bd20000U,
    0x56990000U,
    0xa9b18000U,
    0x3a694000U,
    0x48f26000U,
    0x4f46d000U,
    0x142ef800U,
    0xc15f8c00U,
    0x9dd18e00U,
    0x4398fb00U,
    0x22357680U,
    0xca27740U,
    0xc11f78e0U,
    0x3576cc30U,
    0x6c83eec8U,
    0x164e2becU,
    0x8da38ec2U,
    0x6c91fb63U,
    0xeebcf65eU,
    0x4ae7379dU,
    0xf4bf182fU,
    0x1de91c6bU,
    0x5c3c9656U,
    0x95a8e7a9U,
    0x789860baU,
    0x88b3d008U,
    0xc7ed782fU,
    0x2b3fcc64U,
    0x2d2a6e69U,
    0x28db6b01U,
    0x981bee29U,
    0xacf22bddU,
    0xc1498eb8U,
    0x6524fbe3U,
    0x84df764eU,
    0xba177715U,
    0xfffcf8f3U,
    0xd7c68cb1U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x8000000U,
    0x7c000000U,
    0x72000000U,
    0xc3000000U,
    0x73800000U,
    0x50400000U,
    0x2de00000U,
    0x40300000U,
    0xfe580000U,
    0x5e2c0000U,
    0x75960000U,
    0xcf870000U,
    0x42498000U,
    0x1ee54000U,
    0xebbde000U,
    0xda139000U,
    0x7dc0c800U,
    0x84a31c00U,
    0x8154ea00U,
    0x3fa46500U,
    0x16dda280U,
    0x49e23940U,
    0x2634a8a0U,
    0xc355cc90U,
    0x44a9c228U,
    0xe154e9acU,
    0xcfa980faU,
    0xced540ffU,
    0x3de5e021U,
    0x283f9043U,
    0x7256c856U,
    0xf4241c6cU,
    0xc29d6aa1U,
    0xb20125ddU,
    0xa3004278U,
    0x8381a981U,
    0x884c60baU,
    0x59ead001U,
    0x4e33287fU,
    0x4f5b8c36U,
    0xeeaba299U,
    0x5655392eU,
    0xb225283fU,
    0x2f9c8c76U,
    0xa0822202U,
    0x4bc0796cU,
    0xf1a0c816U,
    0xc7d31c2cU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x58000000U,
    0x74000000U,
    0x5e000000U,
    0xa1000000U,
    0x23800000U,
    0x68c00000U,
    0xd7200000U,
    0x1b500000U,
    0xdbf80000U,
    0xd6640000U,
    0xabb20000U,
    0x2d050000U,
    0x198a8000U,
    0x1fccc000U,
    0x49a1a000U,
    0x221a7000U,
    0xef176800U,
    0xc6955400U,
    0x815bee00U,
    0xbcf7ff00U,
    0xc0e60680U,
    0x8efe6bc0U,
    0x6be44860U,
    0x8277e4f0U,
    0xe9a726b8U,
    0x3219db44U,
    0x67120086U,
    0xda950015U,
    0xf352809dU,
    0x37f8c0f9U,
    0x1c6ba0acU,
    0x64bb7007U,
    0xf78fe852U,
    0x26cc946cU,
    0xfe28ce53U,
    0x24d54f93U,
    0xc13acee5U,
    0x8a404fe9U,
    0x6c684eebU,
    0x1cb88f1bU,
    0x5383ee35U,
    0x10c3ff4dU,
    0x732c0611U,
    0x2d5f6b40U,
    0x56fcc84dU,
    0xdfee245dU,
    0x3c740656U,
    0x78ab6b5cU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x58000000U,
    0x24000000U,
    0x56000000U,
    0x41000000U,
    0x4e800000U,
    0x400000U,
    0xe7600000U,
    0x33700000U,
    0xc980000U,
    0xdeac0000U,
    0xabda0000U,
    0x2e450000U,
    0x12628000U,
    0xa3f9c000U,
    0x1dc2000U,
    0x5d48f000U,
    0xabe8b800U,
    0x9c3c8c00U,
    0xf074e200U,
    0xad126900U,
    0x2566da80U,
    0x3c7b2540U,
    0xc7141860U,
    0xf664bc70U,
    0x55f8fab8U,
    0x90dad514U,
    0x7bc4208eU,
    0xd7a4f025U,
    0x952b8f8U,
    0xd4098c71U,
    0xae0e62f1U,
    0xb507a917U,
    0x7080fa3dU,
    0x3d46d5ecU,
    0xdbe62089U,
    0x243df080U,
    0xe47238e6U,
    0xa3194caaU,
    0x406ac211U,
    0x24f39945U,
    0x8656e299U,
    0x5f8b69eeU,
    0x66c65a80U,
    0x7b2be5d1U,
    0x9610b807U,
    0x70e08c75U,
    0xf9b6e200U,
    0x6dbb69f3U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0xa8000000U,
    0x3c000000U,
    0x3e000000U,
    0x65000000U,
    0x92800000U,
    0x4ac00000U,
    0xfe00000U,
    0x52900000U,
    0xd9f80000U,
    0xd8ac0000U,
    0x82360000U,
    0x77cf0000U,
    0x196b8000U,
    0x925b4000U,
    0xe11d2000U,
    0xebb33000U,
    0xda82b800U,
    0x46cd6400U,
    0xe9ef8200U,
    0xf390bf00U,
    0xe17eba80U,
    0xf56a9b40U,
    0x445a18a0U,
    0x98161410U,
    0x17359a88U,
    0x5d46ab6cU,
    0xe7ab2016U,
    0xcfbc3019U,
    0x4089388cU,
    0xf9c6247fU,
    0xe46aa235U,
    0xe4df8f24U,
    0x51d20268U,
    0x2b54ffafU,
    0x7d901a69U,
    0x1c7eebf5U,
    0x83eb809cU,
    0xf49b409bU,
    0x58fd2097U,
    0xb02330a8U,
    0x7fab8bdU,
    0x8da164c9U,
    0xc8b98289U,
    0xf90fbf5aU,
    0xbc8d3a2dU,
    0xa7cddb6dU,
    0xf169388bU,
    0x8e5624c6U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0xd8000000U,
    0x4000000U,
    0x16000000U,
    0xe5000000U,
    0x78800000U,
    0xb0400000U,
    0x5600000U,
    0x5c300000U,
    0x3fd80000U,
    0x4aac0000U,
    0xea9a0000U,
    0x77470000U,
    0xe6e88000U,
    0x5df3c000U,
    0x28782000U,
    0x8cbeb000U,
    0x39b9800U,
    0x65ceec00U,
    0x51a38200U,
    0x1b10dd00U,
    0x85889a80U,
    0x6cc1f140U,
    0x53293860U,
    0x74589c50U,
    0xc76aba38U,
    0xd7384114U,
    0x565a204eU,
    0x5c65b0a1U,
    0x66b1188eU,
    0x44962c45U,
    0xc64922a5U,
    0xb06aade8U,
    0x1cbba2acU,
    0x9b9e6db3U,
    0x81cb02cdU,
    0x57a31dddU,
    0x2610bab1U,
    0xf90f4122U,
    0xca8aa05fU,
    0xb34a702aU,
    0x50eb3897U,
    0x48f39cdfU,
    0x98f83a07U,
    0xe0fc812eU,
    0x14f28028U,
    0xcaf4c0f8U,
    0xf3f0a09dU,
    0x997d70bcU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0x28000000U,
    0x3c000000U,
    0x92000000U,
    0xb7000000U,
    0x60800000U,
    0xdc00000U,
    0xa8200000U,
    0x3100000U,
    0x2db80000U,
    0xea640000U,
    0xe2720000U,
    0xfd470000U,
    0x4def8000U,
    0x773f4000U,
    0xaca3e000U,
    0x58dcd000U,
    0x5498b800U,
    0x80faf400U,
    0xaf89ae00U,
    0x594c7b00U,
    0xfbe69680U,
    0xb63dcfc0U,
    0xed26d860U,
    0x49e64d0U,
    0x88fd76c8U,
    0x63851f2cU,
    0xf34c603aU,
    0x58e3904bU,
    0x78bb5812U,
    0xc5e624aaU,
    0x7b3116e0U,
    0xe6a68f32U,
    0xebd73897U,
    0xb215b49eU,
    0x94324e4fU,
    0x4224ab0aU,
    0x80142eafU,
    0x13343b75U,
    0xfaaaf63eU,
    0x89de5fa2U,
    0x3d1d8032U,
    0xe0b840c4U,
    0xe1ec6044U,
    0xd339036U,
    0xc7a3580fU,
    0xba5224dbU,
    0xc65b16faU,
    0x74558f76U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x98000000U,
    0xc000000U,
    0x6e000000U,
    0x15000000U,
    0x54800000U,
    0xa5c00000U,
    0xd1600000U,
    0x8900000U,
    0x78780000U,
    0x71a40000U,
    0x857a0000U,
    0x112d0000U,
    0x12b68000U,
    0x13474000U,
    0xcfa92000U,
    0xf8783000U,
    0xb1afd800U,
    0xe57c7400U,
    0x2128da00U,
    0x8ab13700U,
    0x1f4b8280U,
    0xa1a703c0U,
    0xed7cf8e0U,
    0xe52944f0U,
    0x40b182f8U,
    0xf04a033cU,
    0x822a7876U,
    0x673e04d9U,
    0xd000a25aU,
    0x68063380U,
    0xf407a09dU,
    0x520b7061U,
    0xe304f8a7U,
    0x4d8d445cU,
    0x9f4b8231U,
    0x61a703c9U,
    0x8d7cf8a8U,
    0xd52944dfU,
    0xd8b18291U,
    0xfc4a030fU,
    0xec2a7832U,
    0x723e04c8U,
    0x8480a2f4U,
    0xcdc63328U,
    0x2567a038U,
    0x5a9b70abU,
    0x9b7cf88eU,
    0x3c29449fU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x38000000U,
    0x54000000U,
    0x8a000000U,
    0x43000000U,
    0xba800000U,
    0xf1400000U,
    0x9f200000U,
    0x8e700000U,
    0xb8d80000U,
    0x56c0000U,
    0xa7d60000U,
    0xe9e70000U,
    0x39a8000U,
    0xb94cc000U,
    0x6324e000U,
    0x487d1000U,
    0xf5dfd800U,
    0x4ee39c00U,
    0xfb119600U,
    0xb7051d00U,
    0xa082ce80U,
    0xaa4d4140U,
    0xa1ad38a0U,
    0xcd398c90U,
    0x30f4ce18U,
    0xbb9a4184U,
    0xad4fb832U,
    0xc9294c57U,
    0xdb7e2e10U,
    0x775c5162U,
    0xeba4e09dU,
    0xee3d106bU,
    0x7a7fd88dU,
    0xa2d39c18U,
    0x5e69969dU,
    0x99591d49U,
    0xaaacce31U,
    0x8bb641ddU,
    0x7b9b8c2U,
    0x69be4c95U,
    0x68bcaeb6U,
    0x2d3c91e5U,
    0x80f60087U,
    0x73970012U,
    0x1142808fU,
    0x2f20c0deU,
    0x4672e0eaU,
    0x4da10bbU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xa8000000U,
    0x5c000000U,
    0x22000000U,
    0x2b000000U,
    0xaa800000U,
    0xd3400000U,
    0x3a200000U,
    0xd1b00000U,
    0xf0d80000U,
    0x774c0000U,
    0x3c2a0000U,
    0xbcb90000U,
    0xd7558000U,
    0xf3834000U,
    0x2acde000U,
    0x27e99000U,
    0x9ed59800U,
    0xe640dc00U,
    0xa9ad7600U,
    0xcb702900U,
    0x9f350e80U,
    0xc5996540U,
    0x506de060U,
    0x62199030U,
    0x94ad9848U,
    0x94fcdc2cU,
    0x7fff768aU,
    0x75752977U,
    0x36328e08U,
    0xd41f25b8U,
    0xf1a780f0U,
    0x7f764032U,
    0x1326082U,
    0xbc93d08aU,
    0x79edf846U,
    0x47da0cbcU,
    0x5fcd0e63U,
    0x546565b7U,
    0xf41fe0edU,
    0x21ac9096U,
    0x4772189eU,
    0x85369c17U,
    0x6a9f1653U,
    0x2ce3f9baU,
    0xe458f67fU,
    0xd03692aU,
    0x1780eef8U,
    0xccccf590U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0xc8000000U,
    0x14000000U,
    0xa2000000U,
    0x65000000U,
    0x74800000U,
    0x22c00000U,
    0x2ce00000U,
    0x80900000U,
    0x96b80000U,
    0x26c40000U,
    0x26e60000U,
    0x19f0000U,
    0xd8318000U,
    0xbd09c000U,
    0xc88d6000U,
    0x64c37000U,
    0x73e89800U,
    0x4d18a400U,
    0x76f38200U,
    0xefe2b700U,
    0x9b167a80U,
    0x41f963c0U,
    0x666d6020U,
    0x11537090U,
    0xf9d09868U,
    0xd1ca444U,
    0x16f5826aU,
    0x1fedb771U,
    0x31ffa56U,
    0x9df4a387U,
    0xd0660078U,
    0xd65f0032U,
    0xe85180d2U,
    0x5b59c0e2U,
    0x18d560daU,
    0xb3977056U,
    0x15369828U,
    0x2d83a4fbU,
    0xd0440248U,
    0xf124777bU,
    0x31f29a81U,
    0x3e67d34fU,
    0x6d569895U,
    0x1fd3a415U,
    0x21c027bU,
    0x13707756U,
    0xebac9a07U,
    0x683cd338U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x78000000U,
    0x9c000000U,
    0x6e000000U,
    0x3000000U,
    0xeb800000U,
    0xbf400000U,
    0x8ee00000U,
    0xf6500000U,
    0x4bf80000U,
    0xb54c0000U,
    0x3fe20000U,
    0x10dd0000U,
    0xac348000U,
    0x17624000U,
    0x641d2000U,
    0x4d93d000U,
    0xa7de0800U,
    0x3db10c00U,
    0x412d3a00U,
    0x407fdb00U,
    0xfa0e1280U,
    0xa90d0740U,
    0xca8520e0U,
    0xf1cfd0f0U,
    0x3d240818U,
    0xde700c2cU,
    0xe103ba16U,
    0x6e8c9b9fU,
    0x63c5b205U,
    0xe02197fcU,
    0x4ef28885U,
    0x29cf4cb9U,
    0x312a1addU,
    0x187d0b6fU,
    0xb6069ae2U,
    0xf034b7aU,
    0x2d81baf6U,
    0xe8419b4bU,
    0xef693235U,
    0xb81fd77fU,
    0x4395a8f0U,
    0x14dd9c5cU,
    0xae3a920fU,
    0x626f4748U,
    0xa09800dbU,
    0xf5c0078U,
    0x97a00d8U,
    0xc081004dU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x48000000U,
    0xe4000000U,
    0xa2000000U,
    0x4f000000U,
    0x6b800000U,
    0x6c400000U,
    0x29a00000U,
    0xf0f00000U,
    0x77180000U,
    0x514c0000U,
    0x152a0000U,
    0x4b310000U,
    0xccf38000U,
    0xc916c000U,
    0xf0466000U,
    0x87abd000U,
    0xe9f7c800U,
    0x499dbc00U,
    0xb588be00U,
    0x7d470500U,
    0x13211680U,
    0x463d6940U,
    0x18746060U,
    0x86d6d0b0U,
    0x51ae48a8U,
    0xfcfa7c14U,
    0x211d5eeaU,
    0x44a15abU,
    0x9da8be49U,
    0xaf70563U,
    0x64191622U,
    0x34c1692cU,
    0x70666076U,
    0xe41bd0f5U,
    0x74cfc868U,
    0x9061bc41U,
    0x141abe58U,
    0x3cca0545U,
    0x7460965cU,
    0xb616a9ddU,
    0x73cb804cU,
    0x1feac0b3U,
    0xda5460bfU,
    0x5a66d0caU,
    0xef1648e2U,
    0xad467c43U,
    0xb2f5eebU,
    0xfa3715e6U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x98000000U,
    0xb4000000U,
    0xa6000000U,
    0x1f000000U,
    0x68800000U,
    0x5ac00000U,
    0x57a00000U,
    0x51700000U,
    0xe2180000U,
    0xfcc0000U,
    0xd22e0000U,
    0x2a3f0000U,
    0x50f48000U,
    0x175ec000U,
    0x7e272000U,
    0x78309000U,
    0x49f9f800U,
    0x30d8e400U,
    0xd46e7600U,
    0xc15bbd00U,
    0xe928ae80U,
    0x9cbfc940U,
    0x51312020U,
    0x46739010U,
    0x469b7838U,
    0xb70524e4U,
    0xf48bd63eU,
    0x20c6edabU,
    0x92acf64eU,
    0xacf67d05U,
    0xdd550e1fU,
    0xd32e991bU,
    0xf9bb780dU,
    0xfcb524faU,
    0xe133d6aeU,
    0x2e7aeddeU,
    0x3a9af654U,
    0x3d057d8cU,
    0xf98f8e97U,
    0xf14f596bU,
    0xbfe8d852U,
    0xc91b74e8U,
    0x814d0e64U,
    0x37e2997eU,
    0x4515788bU,
    0xc34a24deU,
    0x16e7567dU,
    0x86942ddbU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x8000000U,
    0x34000000U,
    0x6e000000U,
    0xa3000000U,
    0xd9800000U,
    0xc9c00000U,
    0x23e00000U,
    0x7ef00000U,
    0x6ed80000U,
    0x8ec40000U,
    0x5c620000U,
    0xc03b0000U,
    0xaabd8000U,
    0x92f7c000U,
    0xfcdfa000U,
    0x97c3d000U,
    0x18e02800U,
    0x7b79c400U,
    0xad168200U,
    0x682ec100U,
    0x6c110a80U,
    0xd0a0d5c0U,
    0x6d5da060U,
    0x4a08d090U,
    0xe505a8e8U,
    0xfe8a0464U,
    0x264b2266U,
    0x75261197U,
    0xde94a237U,
    0x14ead1aaU,
    0xb976829aU,
    0xbc1ec127U,
    0x18a90a25U,
    0xb954d554U,
    0x7407a034U,
    0x4e07d049U,
    0x130228a9U,
    0x8182c45cU,
    0xf5cb02caU,
    0x79e90125U,
    0xb3f6aa9eU,
    0x145705b6U,
    0x9e87881dU,
    0xb64e147aU,
    0x9d2caa92U,
    0xba98057eU,
    0x72e00883U,
    0x2e76d486U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x18000000U,
    0xc000000U,
    0x22000000U,
    0xc5000000U,
    0x13800000U,
    0x2a400000U,
    0x3d200000U,
    0x6e900000U,
    0xff80000U,
    0x17440000U,
    0x92aa0000U,
    0x1ed10000U,
    0x8bde8000U,
    0x50554000U,
    0x7e10e000U,
    0x53b7000U,
    0x98a9e800U,
    0x6fdfdc00U,
    0xa65b1600U,
    0xf5190b00U,
    0x81ba1e80U,
    0x7c69a7c0U,
    0x673ae0e0U,
    0xfdaa70b0U,
    0x6c576878U,
    0xe41a9c7cU,
    0xdc33f63aU,
    0xc1267bc9U,
    0x9499f6b1U,
    0xa6f77b2fU,
    0x56c776ceU,
    0x65e23bf4U,
    0x3c7796caU,
    0xed094bc5U,
    0xa7867e47U,
    0x14429770U,
    0xb22f68d0U,
    0xac1e9c1dU,
    0x1839f681U,
    0xe7277ba8U,
    0x179f767dU,
    0x46763b30U,
    0x8405964fU,
    0x860c4ba2U,
    0x930afee7U,
    0x8882d728U,
    0xc6cb0869U,
    0xcde1accaU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0xe8000000U,
    0xac000000U,
    0xae000000U,
    0xa1000000U,
    0x90800000U,
    0x54c00000U,
    0x9a200000U,
    0x93100000U,
    0xcf80000U,
    0xd4c0000U,
    0xb2620000U,
    0x93f30000U,
    0xf4c38000U,
    0xea204000U,
    0x4b162000U,
    0x18fdf000U,
    0xe7419800U,
    0x11648400U,
    0xc750e00U,
    0x91890300U,
    0xb440b680U,
    0x16e37740U,
    0x1d37a0e0U,
    0x75eeb010U,
    0xa2b43888U,
    0x20a934fcU,
    0x44dab646U,
    0x395c770dU,
    0x431620beU,
    0xa4fdf0b5U,
    0xc141986aU,
    0x4c648497U,
    0xdaf50e7eU,
    0xc8490332U,
    0x10e0b690U,
    0x7033777fU,
    0x1b6fa036U,
    0xbf72b03dU,
    0x860e38adU,
    0x2d06349dU,
    0xe8336e6U,
    0x4dc33709U,
    0x4ea180e7U,
    0xc5d340a6U,
    0x99d5a026U,
    0xafddb0faU,
    0x7ad7b89cU,
    0x50597449U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0xa8000000U,
    0x24000000U,
    0x9a000000U,
    0x1000000U,
    0xe2800000U,
    0x29c00000U,
    0xcd600000U,
    0x8b300000U,
    0x90980000U,
    0xf440000U,
    0x2eaa0000U,
    0x8f5f0000U,
    0x4d2b8000U,
    0x8414c000U,
    0xc985e000U,
    0x524cb000U,
    0x7a22b800U,
    0xe997f400U,
    0xc1c56a00U,
    0x496a3500U,
    0xe1303280U,
    0xa99a71c0U,
    0x61ce60e0U,
    0xb9687050U,
    0xd93f58c8U,
    0x259f44b4U,
    0xdfcdd232U,
    0x2262c125U,
    0x3abed8f8U,
    0xeed484e8U,
    0x3b63b24fU,
    0x643ab132U,
    0x21198075U,
    0x710fc060U,
    0x1a846044U,
    0xc5c77011U,
    0xe36cd829U,
    0xb83f8406U,
    0x571a32d3U,
    0x5e05715cU,
    0xcb05e041U,
    0x2b8cb05dU,
    0x7f42b838U,
    0xd6a7f434U,
    0x635d6a55U,
    0x632e354fU,
    0xb71a327dU,
    0xe05710cU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0x78000000U,
    0x1c000000U,
    0x86000000U,
    0x21000000U,
    0x60800000U,
    0x3c400000U,
    0x80e00000U,
    0x92300000U,
    0xb6180000U,
    0xc48c0000U,
    0x64a0000U,
    0xd7ef0000U,
    0xbbb68000U,
    0x8ede4000U,
    0xe225a000U,
    0xb3907000U,
    0xb2431800U,
    0x95e74c00U,
    0xf0b38a00U,
    0x21518d00U,
    0xfbe3b280U,
    0xf5b8f140U,
    0x3bd68060U,
    0x70ae4050U,
    0x4cdda098U,
    0xe92c700cU,
    0xfc1118feU,
    0xbb844c3dU,
    0xabcf0ae6U,
    0x6520cd1dU,
    0x42109260U,
    0x6686c1eeU,
    0x5d483856U,
    0x657c06U,
    0xbe7f3228U,
    0x4ef9b11fU,
    0x4abda0c3U,
    0x365c7024U,
    0x3269185fU,
    0xdd784c30U,
    0x657d0ac8U,
    0xd973cd94U,
    0xaf741272U,
    0xe67b8183U,
    0xe2f118bbU,
    0x44b44cabU,
    0x63570a65U,
    0xb0eccd42U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x8000000U,
    0xfc000000U,
    0x5a000000U,
    0x41000000U,
    0x84800000U,
    0xa5400000U,
    0x6a200000U,
    0x7bb00000U,
    0xc8580000U,
    0x68840000U,
    0x874e0000U,
    0xff290000U,
    0xe93b8000U,
    0x7e174000U,
    0x3b2e6000U,
    0x873fb000U,
    0xb91d5800U,
    0x14ad9c00U,
    0xeffdb600U,
    0xefa7900U,
    0xfa750e80U,
    0x773f15c0U,
    0x511b80a0U,
    0x78a74070U,
    0xdf66028U,
    0xbbfbb04cU,
    0xd8f35852U,
    0xc9749cbdU,
    0xfebe36deU,
    0x22d939e4U,
    0xacd6e6eU,
    0xc2eda51eU,
    0xec535802U,
    0x96849c63U,
    0xf8463667U,
    0x24ad39dbU,
    0x27fb6ebcU,
    0xd2f0a5fcU,
    0x107ed8acU,
    0x3e3edc6dU,
    0x299dd644U,
    0x87ecc901U,
    0x26d3d626U,
    0x44c5c974U,
    0xb5e85624U,
    0xcbd2897aU,
    0x2463683U,
    0x15ad39dcU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x98000000U,
    0xc4000000U,
    0x6000000U,
    0xb000000U,
    0x4f800000U,
    0x35400000U,
    0x52600000U,
    0xb1700000U,
    0x3dd80000U,
    0xdd840000U,
    0x604e0000U,
    0xf2ef0000U,
    0x1db48000U,
    0xc9fdc000U,
    0xb5182000U,
    0xdfa6d000U,
    0x4bd74800U,
    0x5e892c00U,
    0x53c2b200U,
    0xd5a99100U,
    0xdad95a80U,
    0xf80badc0U,
    0xf40c80a0U,
    0x6e09c010U,
    0x870e20b8U,
    0x158dd014U,
    0xfc4dc89eU,
    0x10ebeccfU,
    0xc0b61249U,
    0x1576813eU,
    0xbd8329dU,
    0xbe8b5144U,
    0xa3cdfacfU,
    0x9da4bd7cU,
    0x86d5e865U,
    0x3a0d3cfbU,
    0xf9015ac3U,
    0x2a8fad24U,
    0xfdc280f9U,
    0x72a6c02cU,
    0x1f5aa0c5U,
    0x9c4010deU,
    0x20ede8a0U,
    0xa8b93c16U,
    0x99775a80U,
    0x51d4adb8U,
    0x778000a1U,
    0xe1400085U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0x48000000U,
    0x74000000U,
    0x36000000U,
    0xe3000000U,
    0x7e800000U,
    0x70400000U,
    0xaf600000U,
    0xd6300000U,
    0xfd980000U,
    0x95840000U,
    0x1ac60000U,
    0xf9210000U,
    0x625f8000U,
    0x692bc000U,
    0x5a582000U,
    0xf52f5000U,
    0xa85ed800U,
    0x68266c00U,
    0x41ddb200U,
    0x50e27500U,
    0x7dfcca80U,
    0x573489c0U,
    0x26198020U,
    0x904ac070U,
    0x7f67a0e8U,
    0x8e3490c4U,
    0x719ef87eU,
    0x9f8d3c97U,
    0xbbc56a48U,
    0x52a51993U,
    0x8f9ef851U,
    0xc88d3c66U,
    0x53456af2U,
    0x71e519f3U,
    0x167ef82fU,
    0x1afd3cd8U,
    0x37bd6aeeU,
    0xd15119c3U,
    0x8fa0f88eU,
    0x6183c58U,
    0xe044eacbU,
    0x976bd918U,
    0x4a3f58a2U,
    0xf98ac69U,
    0x884121fU,
    0xf343e57bU,
    0xc1e3b272U,
    0x5e7775ffU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x38000000U,
    0xc000000U,
    0x9a000000U,
    0x8f000000U,
    0x8a800000U,
    0xc2c00000U,
    0xbb600000U,
    0x1db00000U,
    0xc9980000U,
    0x530c0000U,
    0x8820000U,
    0x31c70000U,
    0x83e28000U,
    0x64734000U,
    0x6e716000U,
    0x997d5000U,
    0x7ffa5800U,
    0x373f8400U,
    0x3b5f1a00U,
    0xa0622d00U,
    0x3134a280U,
    0x5654b940U,
    0xc9e200e0U,
    0xf3770050U,
    0x98fa8058U,
    0xe9bf401cU,
    0x7f9360a2U,
    0x360a5083U,
    0x2500d810U,
    0x4d80c44dU,
    0x6c4c7a31U,
    0xd7a87ddfU,
    0xbfd47af2U,
    0x4ba47d0eU,
    0x5dd67a21U,
    0xa8a37d32U,
    0x5d54fa53U,
    0xdd603d09U,
    0x60bd9aafU,
    0x98116d2eU,
    0x9d45c2d9U,
    0x5c29e9bfU,
    0x9e985837U,
    0x8588849bU,
    0x8459a30U,
    0x19ad6de9U,
    0xb6dfc22aU,
    0xec22e9dbU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0x58000000U,
    0x5c000000U,
    0xae000000U,
    0x83000000U,
    0xc4800000U,
    0xd4400000U,
    0x18200000U,
    0xc7500000U,
    0xd8b80000U,
    0xdd0c0000U,
    0xf860000U,
    0x34c70000U,
    0x6e608000U,
    0xaa704000U,
    0x4672000U,
    0xbb76d000U,
    0x9dee7800U,
    0x20b00c00U,
    0x91034600U,
    0xd9893700U,
    0xbbcc9e80U,
    0x5ce8ab40U,
    0xa13e00a0U,
    0x76cb0010U,
    0x2b668078U,
    0x99f7400cU,
    0xeea7a0f6U,
    0x5e1690dfU,
    0xf711586aU,
    0x29adc57U,
    0x1f533edcU,
    0xc4b23b13U,
    0x53095840U,
    0xdc86dc5aU,
    0xa84d3ef7U,
    0xe6293bb9U,
    0x1c57d8b9U,
    0x403d9c82U,
    0xa74c9ee4U,
    0x94a8abc2U,
    0x371e0085U,
    0x629b0053U,
    0x6f5e809aU,
    0xccbb40edU,
    0x5701a06eU,
    0x2e8190d0U,
    0x8549d8a3U,
    0xa1a69c5aU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0x18000000U,
    0xc000000U,
    0xb2000000U,
    0x9000000U,
    0xe0800000U,
    0x3cc00000U,
    0xbba00000U,
    0x6cb00000U,
    0x86580000U,
    0x5b040000U,
    0x39860000U,
    0xd4410000U,
    0x236a8000U,
    0x71114000U,
    0x5de3e000U,
    0x86d7b000U,
    0xb7cc9800U,
    0x8a2ac400U,
    0x1cf3f600U,
    0x33eb100U,
    0x9d100e80U,
    0x3fe385c0U,
    0x87de0020U,
    0xf3450070U,
    0x10ec80b8U,
    0x105040bcU,
    0x340960aaU,
    0xce06f005U,
    0x30f7852U,
    0x558d7435U,
    0x76476e5bU,
    0x82607550U,
    0x859df8bdU,
    0xdf2834f7U,
    0x867a8e1fU,
    0x32f2c53fU,
    0xd03de082U,
    0xc092b069U,
    0xeda0186cU,
    0xa3ba844eU,
    0xc1da9612U,
    0x9448413cU,
    0x43677622U,
    0x611af12cU,
    0xf5e76ec7U,
    0x92d0752bU,
    0x9c5f8a0U,
    0x312c3408U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0xe8000000U,
    0x74000000U,
    0xe2000000U,
    0xf000000U,
    0x20800000U,
    0x98c00000U,
    0x10e00000U,
    0xe6700000U,
    0xb3d80000U,
    0xda040000U,
    0x730e0000U,
    0x68b0000U,
    0xdc78000U,
    0xa369c000U,
    0xc8396000U,
    0xd6f6d000U,
    0x316f800U,
    0x9eed8400U,
    0xe77f8200U,
    0x5252e900U,
    0x83419a80U,
    0xaaaf7dc0U,
    0xd15800a0U,
    0x1dc40050U,
    0x8b6e00c8U,
    0x9c3b00e4U,
    0xa4ff800aU,
    0xe41dc07bU,
    0xca6f60c2U,
    0x9db9d097U,
    0x4dbf7830U,
    0x45bf447eU,
    0x81b962a3U,
    0x1bb9f93cU,
    0x88b80240U,
    0x3e3b291cU,
    0x4bf8fadeU,
    0x7499adf5U,
    0x2aaef88dU,
    0x11598451U,
    0x3dc982e1U,
    0x1b6de9a3U,
    0x74301a4eU,
    0xd0f9bdefU,
    0x610e016U,
    0xc56410d9U,
    0xbd301819U,
    0xd57694f7U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0xa8000000U,
    0x4c000000U,
    0x4e000000U,
    0x97000000U,
    0xc5800000U,
    0x7d400000U,
    0xa7600000U,
    0x3bf00000U,
    0x55280000U,
    0xf8040000U,
    0x84020000U,
    0xb2010000U,
    0xe1048000U,
    0xb6854000U,
    0xbac26000U,
    0x3227000U,
    0xce14b800U,
    0xd61e2c00U,
    0x770fea00U,
    0xe0933b00U,
    0x58dadf80U,
    0xfe2b2940U,
    0x2b800020U,
    0x3a400010U,
    0xeae00008U,
    0x9ab00024U,
    0x1448002aU,
    0x18f40013U,
    0x5aaa0013U,
    0xa0450025U,
    0x7e68031U,
    0x4234401fU,
    0xa98ee029U,
    0x7653300eU,
    0xa5fcd815U,
    0x9f395c3eU,
    0xdc1dd221U,
    0x3209572cU,
    0x7413d5b8U,
    0xab1f226dU,
    0x78c078eU,
    0x11577550U,
    0x787b521bU,
    0x967d1701U,
    0xd17d35bfU,
    0x9cfc124fU,
    0x3db8df8fU,
    0x7cda2959U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0x38000000U,
    0x8c000000U,
    0x5a000000U,
    0x21000000U,
    0x3b800000U,
    0x38400000U,
    0xcbe00000U,
    0x69f00000U,
    0x2d980000U,
    0x2c040000U,
    0x6a060000U,
    0xe9070000U,
    0x9f808000U,
    0xd641c000U,
    0x3ce46000U,
    0x2972d000U,
    0xf590800U,
    0xe425dc00U,
    0xf057c200U,
    0x66895f00U,
    0xf7efcf80U,
    0xbddaacc0U,
    0x93e00020U,
    0x35f00030U,
    0xaf980038U,
    0x91040004U,
    0x3386000eU,
    0x7c470023U,
    0x35e08016U,
    0xa6f1c008U,
    0xe11c600eU,
    0x54c6d00eU,
    0x83270832U,
    0x48d6dc1aU,
    0x2849420bU,
    0x75cb9f0bU,
    0x3e8d2f9aU,
    0xabeebcfaU,
    0x3fdde807U,
    0x2ee6cc05U,
    0x6c72aa37U,
    0x3ada530eU,
    0x3b67058dU,
    0x43312fdaU,
    0xa1b88d8aU,
    0x485233e1U,
    0x2a8bafabU,
    0x4de87cd5U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0xc8000000U,
    0xdc000000U,
    0x4a000000U,
    0x4f000000U,
    0x53800000U,
    0xe2c00000U,
    0x86600000U,
    0x10f00000U,
    0x7ed80000U,
    0xb4040000U,
    0xe6020000U,
    0xed010000U,
    0xf0818000U,
    0x33464000U,
    0xf7a6e000U,
    0x8f125000U,
    0xc3ea7800U,
    0x1f3e9c00U,
    0xa0311600U,
    0xf3bf3300U,
    0xb9740780U,
    0x5b1e36c0U,
    0x9de00020U,
    0x6e300010U,
    0x92b80008U,
    0xdbf4000cU,
    0x35a0032U,
    0x67c50037U,
    0xdae38012U,
    0x81b74013U,
    0x2a7f6014U,
    0xea901038U,
    0x542e9821U,
    0x6dddcc04U,
    0xed82ee1fU,
    0x6bc3ef2dU,
    0x8e1f1b9U,
    0xcab255fbU,
    0x17ffff9cU,
    0x1956eadcU,
    0x58cff615U,
    0xc969633fU,
    0xeb7c7f8aU,
    0xf811aafcU,
    0xe4689628U,
    0x7bfd7308U,
    0xdb50e7a0U,
    0xebcd66f1U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x28000000U,
    0x34000000U,
    0x6a000000U,
    0xa5000000U,
    0xda800000U,
    0x10c00000U,
    0xd5200000U,
    0xc6900000U,
    0xfcf80000U,
    0x2040000U,
    0xf1060000U,
    0xc0850000U,
    0x4dc68000U,
    0x13a14000U,
    0x8851a000U,
    0xe6db5000U,
    0xbb112800U,
    0xc7b8d400U,
    0x7618600U,
    0xaf762900U,
    0xe94a7480U,
    0x3b6cc7c0U,
    0x6780020U,
    0x2c40030U,
    0xac260028U,
    0xe2150034U,
    0xf33e800aU,
    0x80a5400dU,
    0xc9d7a01aU,
    0x939e5029U,
    0xf977a836U,
    0x2499404U,
    0xa6e82635U,
    0x8d397931U,
    0x5fa55cbfU,
    0x3e5513c0U,
    0xbdd9061cU,
    0x7e966900U,
    0x80fb54bbU,
    0x2c03d7f0U,
    0xc6060808U,
    0xa302c404U,
    0xd9810e1cU,
    0x1945ad2cU,
    0xe462dabdU,
    0x16f63ae2U,
    0x800df2b5U,
    0x270feee6U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0xd8000000U,
    0xe4000000U,
    0xbe000000U,
    0x37000000U,
    0xdd800000U,
    0x3d400000U,
    0xf3200000U,
    0x6bd00000U,
    0x3c480000U,
    0xba040000U,
    0x99060000U,
    0x52830000U,
    0xd4c68000U,
    0xa866c000U,
    0x4bf72000U,
    0x341df000U,
    0x8c0db800U,
    0xda4ec00U,
    0x9d14ea00U,
    0x492a1900U,
    0xad771e80U,
    0x46de0240U,
    0x97680020U,
    0xf5d40030U,
    0x7b4e0018U,
    0xf870034U,
    0x48408036U,
    0x23a5c039U,
    0xd211a02fU,
    0xc0ab300dU,
    0x26329837U,
    0xbefd1c0fU,
    0x7b3f523cU,
    0x7dddf51aU,
    0xd8ed748fU,
    0x1d96db6eU,
    0xe8ee3e86U,
    0xd594f264U,
    0xb4ed380dU,
    0x5f912c2eU,
    0xe5ed4a3cU,
    0x51152900U,
    0xbb2b86baU,
    0xa8741e77U,
    0xfe5fd21fU,
    0xc2a83520U,
    0x8b34d4b8U,
    0x8a79eb70U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x28000000U,
    0x84000000U,
    0xda000000U,
    0x23000000U,
    0x31800000U,
    0xcec00000U,
    0x2aa00000U,
    0x96100000U,
    0x4b580000U,
    0xae040000U,
    0xa1060000U,
    0xe810000U,
    0x19448000U,
    0xb1614000U,
    0xf0342000U,
    0xdb0ed000U,
    0x13bd1800U,
    0x4cf38c00U,
    0x96a97600U,
    0x4e28d500U,
    0x7a6b7080U,
    0xb8c3ac0U,
    0x33f80020U,
    0xcf140030U,
    0x9de0008U,
    0x59450024U,
    0x5162800aU,
    0x40304021U,
    0x6308a036U,
    0xbfbb9008U,
    0x12f7380cU,
    0x6fa85c33U,
    0x5caeee0aU,
    0x852f1925U,
    0xefeca692U,
    0x8f4e7febU,
    0x1258c888U,
    0xec8526f3U,
    0x5642ce2eU,
    0xfee5c938U,
    0x57f7bebeU,
    0xcb2cf3f3U,
    0xfeed3e98U,
    0x39c8b3caU,
    0xa71b9eb7U,
    0x3e623fcU,
    0x5f7626baU,
    0x376a3fe1U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x38000000U,
    0x84000000U,
    0x56000000U,
    0x27000000U,
    0x2b800000U,
    0xc7c00000U,
    0x4f600000U,
    0xd9900000U,
    0xdfb80000U,
    0x77040000U,
    0xf3820000U,
    0x73c50000U,
    0xb1648000U,
    0x4291c000U,
    0x263c2000U,
    0xc1c6b000U,
    0xb0633800U,
    0x46145c00U,
    0xe67abe00U,
    0xa367bb00U,
    0xd390ec80U,
    0x1ab94dc0U,
    0x85820020U,
    0x84c50010U,
    0x2e48028U,
    0x9151c024U,
    0x75c200eU,
    0xbb56b021U,
    0x125b3815U,
    0xd1d05c09U,
    0x7118be0aU,
    0xcef2bb31U,
    0xf22c6c93U,
    0xf6bc8df6U,
    0x8f842017U,
    0x41c2b00dU,
    0xf0613814U,
    0xe6115c38U,
    0x767e3e22U,
    0x9b667b31U,
    0x5794cc9cU,
    0x4cbbfdf9U,
    0xa2833806U,
    0xaf445c30U,
    0xc522be02U,
    0xde33bb3aU,
    0xdecaec8dU,
    0x64e84dfaU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0xf8000000U,
    0x3c000000U,
    0x5a000000U,
    0x1b000000U,
    0xa8800000U,
    0x1c400000U,
    0x7b600000U,
    0x81100000U,
    0x4f080000U,
    0xab040000U,
    0x80860000U,
    0xf8410000U,
    0xd608000U,
    0x3817c000U,
    0xc089e000U,
    0x45c6d000U,
    0xfca6d800U,
    0xaab44400U,
    0xdf386200U,
    0xa7781b00U,
    0x6a1fe880U,
    0x720ba840U,
    0xa2860020U,
    0x1f410030U,
    0xdfe08008U,
    0x2f57c004U,
    0xeb69e03eU,
    0xe496d00fU,
    0x92ced816U,
    0x9ba04406U,
    0xb836622aU,
    0xe87d1b07U,
    0x9c99689eU,
    0x334d6860U,
    0x20676033U,
    0xc994101aU,
    0x6349b828U,
    0x3864543aU,
    0xc597da3dU,
    0xd14d4f01U,
    0xe760b2a6U,
    0xb152757U,
    0x8c0952b5U,
    0x2f83f75dU,
    0x3ec78a81U,
    0xa423b37dU,
    0x7ef1e89fU,
    0x705ea84dU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x28000000U,
    0x3c000000U,
    0xd2000000U,
    0x7b000000U,
    0xef800000U,
    0x48400000U,
    0xc9200000U,
    0x88100000U,
    0x90a80000U,
    0xa3040000U,
    0x8b820000U,
    0xbe450000U,
    0xe4248000U,
    0xda914000U,
    0xa269e000U,
    0x64e29000U,
    0x1671d800U,
    0xc81a7c00U,
    0xdefac200U,
    0x23890900U,
    0x2c16a080U,
    0x86a99540U,
    0xbe020020U,
    0x61050010U,
    0x98848028U,
    0xcdc14024U,
    0xee61e00aU,
    0x40b6900fU,
    0xf9fbd834U,
    0x6e0b7c1eU,
    0x754423bU,
    0x24094912U,
    0xb851c0b2U,
    0xad8b4562U,
    0xf514b804U,
    0xa2dac38U,
    0x9dc0fa0aU,
    0x2664e50bU,
    0x4cb13ab3U,
    0x93ffa079U,
    0x10d82bcU,
    0x6d60c57U,
    0xc54f78b6U,
    0xe5f7e954U,
    0x825ac207U,
    0xe4d90911U,
    0xe81ea0b9U,
    0xefd957cU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0xf8000000U,
    0xbc000000U,
    0xc2000000U,
    0x57000000U,
    0xc0800000U,
    0x30c00000U,
    0x82200000U,
    0x28b00000U,
    0x66380000U,
    0xcb040000U,
    0x72860000U,
    0x3fc50000U,
    0x4ea58000U,
    0x2277c000U,
    0xf1de000U,
    0xe1321000U,
    0x737cb800U,
    0xb6670400U,
    0x24520600U,
    0xf9aed100U,
    0xa20d2580U,
    0xd7f8e1c0U,
    0x88a60020U,
    0x6b750030U,
    0x4a9d8028U,
    0xe73c02cU,
    0x451be03eU,
    0x5237102fU,
    0x7df93830U,
    0xeba0c415U,
    0x8df7e630U,
    0xe358c10cU,
    0x21d79da0U,
    0x76eae5caU,
    0x84698639U,
    0x7ba81102U,
    0x950b45b4U,
    0x77c31e3U,
    0xf064d80dU,
    0xad52d417U,
    0x1c2b5e1bU,
    0xfd4fc501U,
    0x199d9b92U,
    0xd0f034ceU,
    0xf0daa3b9U,
    0xf91f0d1U,
    0xe10ec589U,
    0x417bf1d7U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x88000000U,
    0xc000000U,
    0xca000000U,
    0x37000000U,
    0x2b800000U,
    0x9d400000U,
    0xd2e00000U,
    0x56b00000U,
    0x55c80000U,
    0xeb040000U,
    0x39820000U,
    0x3e430000U,
    0xc7638000U,
    0xb2f44000U,
    0x5da86000U,
    0xf6725000U,
    0x3f6bb800U,
    0xe7d65c00U,
    0xb23d6a00U,
    0xb62dd700U,
    0x4b375580U,
    0xfd89be40U,
    0x49620020U,
    0x63f30010U,
    0x9b2b8018U,
    0x8fb0401cU,
    0xdf4a6022U,
    0x38c15003U,
    0x9ea03832U,
    0x42d61c0dU,
    0x7abf0a0aU,
    0x5ee88727U,
    0xd956db4U,
    0x2b5ca255U,
    0xdde8a35U,
    0x441fc72aU,
    0x4c3e8d96U,
    0xef2ab253U,
    0x81b55233U,
    0x4e4fcb3fU,
    0x9e425fbdU,
    0x9766396cU,
    0x2af6ed87U,
    0xa9a8e24dU,
    0xb876ea12U,
    0xce6d9739U,
    0x315535b7U,
    0x33fcee7dU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0x38000000U,
    0x54000000U,
    0x12000000U,
    0xed000000U,
    0x9c800000U,
    0x89400000U,
    0x8a600000U,
    0x7500000U,
    0xc1380000U,
    0x91040000U,
    0x82860000U,
    0xd2450000U,
    0x81e28000U,
    0x9391c000U,
    0x189aa000U,
    0x6a749000U,
    0xbf096800U,
    0x90ede400U,
    0x507cca00U,
    0xdf629300U,
    0xfd31a80U,
    0xfa78d9c0U,
    0xa6660020U,
    0x21550030U,
    0xce3a8028U,
    0x8885c014U,
    0xfb44a00eU,
    0x97659015U,
    0xf3d5e804U,
    0x247d243bU,
    0x5d62ea27U,
    0x7ad2c322U,
    0x62fa52a2U,
    0x5246dc1U,
    0x12370210U,
    0x47efe714U,
    0x2df8b888U,
    0x7ca6aee0U,
    0x9bf5508eU,
    0x6acf8ac1U,
    0x33c9ba8aU,
    0x8d4c49c5U,
    0x210f680eU,
    0xbe8e437U,
    0xfbfe4a14U,
    0x1ba3530cU,
    0xee71ba9eU,
    0x550849d6U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x48000000U,
    0x14000000U,
    0x6e000000U,
    0xc9000000U,
    0xf3800000U,
    0xd4c00000U,
    0x89e00000U,
    0xbb100000U,
    0x7eb80000U,
    0xe5040000U,
    0x11820000U,
    0x4fc30000U,
    0x59648000U,
    0x73524000U,
    0xd18a000U,
    0x9eb77000U,
    0x400c4800U,
    0x15089c00U,
    0xbc8c2600U,
    0x3f4b0f00U,
    0x542cf880U,
    0x947b45c0U,
    0xcde20020U,
    0xbd130010U,
    0x53bc8018U,
    0xa0864024U,
    0x5042a012U,
    0x4fa07005U,
    0x91b2c81bU,
    0xbe8ddc32U,
    0x744a063cU,
    0xaca93f35U,
    0xd83e90a2U,
    0x9845e9eeU,
    0x1ba64e3fU,
    0x9fb1a329U,
    0xe78ab69cU,
    0xcfcae6f7U,
    0x6c68b6a4U,
    0x3fd9e6c9U,
    0xea5436a0U,
    0x969fa6e1U,
    0xae769686U,
    0x7fefd6c1U,
    0x3b1c5ebcU,
    0x4bb60ae2U,
    0xa98c5887U,
    0xf6c835eaU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0xf8000000U,
    0x4c000000U,
    0x76000000U,
    0x7000000U,
    0xb5800000U,
    0x5f400000U,
    0x91e00000U,
    0x80900000U,
    0xb2980000U,
    0xb1040000U,
    0x52820000U,
    0xdac50000U,
    0xa6a48000U,
    0xa577c000U,
    0x80a6000U,
    0x729fb000U,
    0x51063800U,
    0x6282ac00U,
    0xb2c37a00U,
    0x12a7cf00U,
    0x9f71a480U,
    0x790c44c0U,
    0xc01a0020U,
    0xbbc10010U,
    0xac268028U,
    0xa3b2c024U,
    0x20aee03eU,
    0x9ce87013U,
    0x9a8c581dU,
    0x485d1c01U,
    0xc7a5422dU,
    0xaff56317U,
    0xecadea4U,
    0x5a3f8be0U,
    0xbf71a48cU,
    0xa90c44fcU,
    0x981a001cU,
    0x67c10002U,
    0x2226803fU,
    0xe8b2c01eU,
    0xe32ee021U,
    0xc4a8700eU,
    0xbeec5824U,
    0x978d1c0eU,
    0xe4dd4225U,
    0x9e616333U,
    0xeed0deafU,
    0x31fe8bd2U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0xb8000000U,
    0xfc000000U,
    0x26000000U,
    0xe3000000U,
    0xc1800000U,
    0xcec00000U,
    0x3ce00000U,
    0x41900000U,
    0xe9780000U,
    0xb5040000U,
    0x5a860000U,
    0x93430000U,
    0x24278000U,
    0x2675c000U,
    0x956fe000U,
    0xb4bd3000U,
    0x30651800U,
    0x49d30c00U,
    0x90daf600U,
    0x8bb7a700U,
    0xa80e2c80U,
    0x5bee9bc0U,
    0x15fe0020U,
    0x5470030U,
    0xdf218018U,
    0x8bf6c03cU,
    0x35a8602eU,
    0x2f58f03fU,
    0x6a72f809U,
    0xab6a3c38U,
    0x3bb9ee30U,
    0x9fe7ab33U,
    0x20135a8fU,
    0xb7bcfcd0U,
    0xc1e7cc9aU,
    0x5f10abddU,
    0xa03c982eU,
    0x5421cc28U,
    0x5e74963fU,
    0x96c570aU,
    0x62bb5482U,
    0x6b6167eaU,
    0x74500e15U,
    0x78199b29U,
    0x5451c283U,
    0x281a30fdU,
    0x7c52da91U,
    0x9c1a3ce4U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xb8000000U,
    0x64000000U,
    0x7a000000U,
    0xf5000000U,
    0xb5800000U,
    0xd7c00000U,
    0x78600000U,
    0xdff00000U,
    0xca580000U,
    0x6b040000U,
    0xfa860000U,
    0xb7410000U,
    0x4a228000U,
    0xc855c000U,
    0x9cb2000U,
    0x4afd000U,
    0xaedda800U,
    0x9346ac00U,
    0xd0233e00U,
    0x4d558300U,
    0x544a9f80U,
    0xf6f62c0U,
    0xc8be0020U,
    0xc3b50030U,
    0x5afc8008U,
    0x4410c014U,
    0x16fa02eU,
    0x1fbb1019U,
    0x9734081eU,
    0xaabcbc3dU,
    0xc2b5b62dU,
    0xd7cff35U,
    0x52d4099eU,
    0xe8c4df7U,
    0x868fa192U,
    0xea8be1eaU,
    0x3c8e1f96U,
    0x3f8ba2c9U,
    0xd90fa014U,
    0x504b100fU,
    0xc56c083aU,
    0xf5b8bc31U,
    0xfa33b636U,
    0x2b3dff35U,
    0xd7768999U,
    0xe4198decU,
    0x42a481b4U,
    0xe61431ebU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x28000000U,
    0x6c000000U,
    0xa2000000U,
    0x55000000U,
    0x89800000U,
    0xc400000U,
    0x1d600000U,
    0xcdb00000U,
    0x25580000U,
    0x73040000U,
    0xf2820000U,
    0xc2c50000U,
    0xf5a08000U,
    0x11114000U,
    0x37cb6000U,
    0x71891000U,
    0xb5eca800U,
    0xe3df4c00U,
    0xa7c3e200U,
    0x4262b00U,
    0xf951cd80U,
    0xf4ae9ac0U,
    0x633a0020U,
    0x20310010U,
    0xb71a8028U,
    0x11604004U,
    0x9fb1e00aU,
    0xe859501bU,
    0xae854828U,
    0x18c21c15U,
    0x44a4aa22U,
    0x46913703U,
    0xe48de787U,
    0xdc6aedf3U,
    0x5f9e87a9U,
    0x82a7fdccU,
    0x8d902fb4U,
    0x120db1e4U,
    0x70ab4dbfU,
    0x4d3edacbU,
    0x6733e027U,
    0x539c5006U,
    0xd0a5c805U,
    0x40935c31U,
    0xcf8fca2eU,
    0x9ae8273fU,
    0xcb594fbaU,
    0xd401a1c4U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0xb8000000U,
    0xd4000000U,
    0xaa000000U,
    0x6b000000U,
    0xff800000U,
    0xc2c00000U,
    0x9e200000U,
    0x44500000U,
    0xf5980000U,
    0xaf840000U,
    0x4ac20000U,
    0x52250000U,
    0x4a538000U,
    0x8c9dc000U,
    0xef02a000U,
    0xdd875000U,
    0x65c65800U,
    0x6fa47c00U,
    0xff93d600U,
    0x2b3a5100U,
    0x7c937280U,
    0x88bef4c0U,
    0xd8538020U,
    0x739dc010U,
    0x1a82a028U,
    0x447501cU,
    0xbc66582eU,
    0x3d347c35U,
    0x3e2bd62aU,
    0xabee511aU,
    0x3c4972bfU,
    0xb7dff4f0U,
    0x46e20007U,
    0xe9750001U,
    0x4a4b8015U,
    0xfad9c037U,
    0x7c60a03cU,
    0xdd325021U,
    0xee2dd838U,
    0x63edbc39U,
    0x504b7604U,
    0xc9dc0107U,
    0x87e4aabeU,
    0x7df248caU,
    0x77097622U,
    0xa6390131U,
    0xa6172aa5U,
    0x6cff88eaU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0x48000000U,
    0x54000000U,
    0x96000000U,
    0xdb000000U,
    0x1d800000U,
    0x79400000U,
    0xd9600000U,
    0xf0500000U,
    0x22a80000U,
    0xb5840000U,
    0xfd460000U,
    0x67630000U,
    0x6f508000U,
    0xe12a4000U,
    0x43c6a000U,
    0xafa2b000U,
    0x3575d800U,
    0x89186c00U,
    0xddbd8a00U,
    0x454e6b00U,
    0x8d123280U,
    0x85ca4d40U,
    0xbad08020U,
    0xc6a4030U,
    0x6ca6a018U,
    0x94f2b004U,
    0x425dd812U,
    0x11dc6c15U,
    0x6f9b8a25U,
    0x97d6b36U,
    0xdd6ab287U,
    0xa8240d5eU,
    0xdd302016U,
    0x34fbf00cU,
    0x142bf810U,
    0x49449c29U,
    0x2160f22dU,
    0x9c53b70cU,
    0xb8ace0beU,
    0xe4850a4eU,
    0xf9c118b7U,
    0x4ea29645U,
    0x9f16a83U,
    0xacdb616aU,
    0x311b2a15U,
    0x11bcdb1dU,
    0xaf4feab3U,
    0x8416215fU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x68000000U,
    0xa4000000U,
    0x26000000U,
    0x8f000000U,
    0x92800000U,
    0x61c00000U,
    0xfce00000U,
    0x6ff00000U,
    0x89780000U,
    0xa6840000U,
    0xefc60000U,
    0x77e50000U,
    0xcb708000U,
    0xfbb4000U,
    0x6ce12000U,
    0xc7f73000U,
    0x8d797800U,
    0x90801400U,
    0x8c64e00U,
    0x4167c700U,
    0x8cb2cf80U,
    0x7c5b73c0U,
    0x91908020U,
    0x2f4b4030U,
    0xd7192028U,
    0x10b33004U,
    0xf65f781aU,
    0x2c951429U,
    0x6ccece09U,
    0x67588723U,
    0x9d15efa4U,
    0xad8943d8U,
    0x2b79781fU,
    0xdf80142bU,
    0x3a464e0aU,
    0x30a7c72dU,
    0x1852cfa1U,
    0xb7ab73f4U,
    0x3ee8801bU,
    0x6cf4010U,
    0xaa5f2017U,
    0x696302dU,
    0xc1cff806U,
    0x4cde5416U,
    0x8957ee29U,
    0x62bb71aU,
    0xffaa97bcU,
    0x4aec57daU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0xa8000000U,
    0x34000000U,
    0x2a000000U,
    0xc3000000U,
    0x6b800000U,
    0x67400000U,
    0x41600000U,
    0x83300000U,
    0x49280000U,
    0xf840000U,
    0x35460000U,
    0xe670000U,
    0xaeb78000U,
    0x5b6d4000U,
    0xf865a000U,
    0x53b75000U,
    0xc9e91800U,
    0x6fa0dc00U,
    0x82107a00U,
    0x72fcab00U,
    0xddec980U,
    0xef0b1140U,
    0xdd78020U,
    0x1f5d4030U,
    0xb8cda038U,
    0x3c73503cU,
    0x544f182aU,
    0x72b7dc0dU,
    0x656ffa0aU,
    0x165eb30U,
    0xa335699aU,
    0x592f4159U,
    0x57871830U,
    0xa943dc10U,
    0x1061fa2aU,
    0x47b6eb3fU,
    0xf3ece9a7U,
    0xf4a1014eU,
    0x75933801U,
    0xbbecc16U,
    0xa5bac21cU,
    0xc4bc2731U,
    0x48382b88U,
    0x36fe2676U,
    0x4fda9398U,
    0xf80aaa5fU,
    0xbc52718cU,
    0x131c9d6dU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0xe8000000U,
    0xcc000000U,
    0x5e000000U,
    0x39000000U,
    0x43800000U,
    0x8fc00000U,
    0x75e00000U,
    0x68900000U,
    0x10e80000U,
    0x97840000U,
    0x95c20000U,
    0xcee10000U,
    0x94118000U,
    0xc1af4000U,
    0xbc226000U,
    0x6070f000U,
    0xb9f8c800U,
    0xb7af9c00U,
    0x29253e00U,
    0x6df6cf00U,
    0xd7393480U,
    0xa5cb8f40U,
    0x5c718020U,
    0x6fff4010U,
    0x52aa6008U,
    0xeca4f00cU,
    0xff32c83aU,
    0x735a9c33U,
    0x259ebe17U,
    0x43fc8f0eU,
    0x5ca8d490U,
    0x2da53f63U,
    0xa8b2a83dU,
    0x869b6c0aU,
    0xfb7ff60cU,
    0xfe85329U,
    0xb9058a9fU,
    0x3860040U,
    0xafc0d492U,
    0x45e13f6eU,
    0x8090a817U,
    0xdcea6c27U,
    0xc9867601U,
    0xacc31308U,
    0x8d65eaa3U,
    0x1bd7f043U,
    0xb4499cb7U,
    0xd4b1e367U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x48000000U,
    0x34000000U,
    0xe2000000U,
    0xaf000000U,
    0xdf800000U,
    0x1cc00000U,
    0x52200000U,
    0xfed00000U,
    0xbaa80000U,
    0x7d840000U,
    0xd3c60000U,
    0x3da50000U,
    0xba108000U,
    0x948a4000U,
    0x5555a000U,
    0x24691000U,
    0x30a47800U,
    0xaa907c00U,
    0xe7cfa600U,
    0x43b7b100U,
    0xf49ca680U,
    0xa2d80d40U,
    0x20b88020U,
    0x390e4030U,
    0x6e93a028U,
    0x3dcc1004U,
    0x20b4f812U,
    0xa51a3c0dU,
    0x8f1a0638U,
    0xd41ea12bU,
    0x4998deb7U,
    0xea587147U,
    0x2fff2634U,
    0xf9edf10fU,
    0xf3610686U,
    0xdf351d5bU,
    0x69daf806U,
    0x353b3c32U,
    0xecc863eU,
    0x5931e10aU,
    0xcddfeb0U,
    0xc1bb2143U,
    0x778efe00U,
    0xc4d49d0eU,
    0xa9aad888U,
    0xd402d04cU,
    0x9201f8afU,
    0x57008065U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x88000000U,
    0xa4000000U,
    0x4a000000U,
    0x5f000000U,
    0x94800000U,
    0x90c00000U,
    0x71e00000U,
    0x89f00000U,
    0x4a980000U,
    0x9a840000U,
    0x55c20000U,
    0xa2610000U,
    0x11b58000U,
    0xcdbc4000U,
    0xd3912000U,
    0x17285000U,
    0x293ff800U,
    0xab572400U,
    0x52cbc600U,
    0xd2ce2f00U,
    0x92c9df80U,
    0xb2ceccc0U,
    0x2cd8020U,
    0x8ac84010U,
    0x2ecb2008U,
    0x64cd502cU,
    0x3bc87822U,
    0xaf4a6429U,
    0x3f8f6612U,
    0x4e6a3f17U,
    0xc79f07a5U,
    0x8d05b8e4U,
    0x1783be3cU,
    0x42444b32U,
    0xe026b99aU,
    0xf194f3caU,
    0x3c2a8797U,
    0xefb9f8d1U,
    0xf8929e3eU,
    0xd1ac1b18U,
    0x7af941bbU,
    0x2833d7e4U,
    0xfaf94186U,
    0x6833d7e6U,
    0xdaf94199U,
    0xd833d7ddU,
    0x52f94182U,
    0x7c33d7d7U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xb8000000U,
    0x2c000000U,
    0x2e000000U,
    0xe7000000U,
    0x2e800000U,
    0xa6c00000U,
    0x58e00000U,
    0x61b00000U,
    0xf8c80000U,
    0x5c840000U,
    0xb7c60000U,
    0x2d650000U,
    0x4ff18000U,
    0x75edc000U,
    0x49556000U,
    0x417f7000U,
    0x184e3800U,
    0x4d42f400U,
    0xe9a3f600U,
    0x4e172700U,
    0x55de3d80U,
    0xadd85e40U,
    0xe1d98020U,
    0x5fd9c030U,
    0x30db6028U,
    0x8a5e700cU,
    0x2e99b82eU,
    0xbf7a340bU,
    0x174f160bU,
    0x67c19739U,
    0xc563658bU,
    0x4bf0da69U,
    0xffedce36U,
    0x1455d328U,
    0x8afdcb96U,
    0x590f795bU,
    0x7ae7bda3U,
    0x58b19e70U,
    0x294ae010U,
    0xc8c3b018U,
    0xdfe4d81fU,
    0xdf31443eU,
    0xd60f2e35U,
    0x90626339U,
    0xd4771384U,
    0x1caf3d54U,
    0xdaf71385U,
    0x4a6f3d61U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x88000000U,
    0xec000000U,
    0x1a000000U,
    0xb5000000U,
    0x46800000U,
    0xf7c00000U,
    0xe3200000U,
    0x1c900000U,
    0x35780000U,
    0x88840000U,
    0x7cc20000U,
    0x6a70000U,
    0x32d18000U,
    0x5a1c4000U,
    0x2a736000U,
    0x5bcad000U,
    0xf1eea800U,
    0xa9fdb400U,
    0x6a42fe00U,
    0x964f900U,
    0x55f2f480U,
    0x808ea3c0U,
    0x38098020U,
    0xccc84010U,
    0x26696038U,
    0xeb9d00cU,
    0xe1a52822U,
    0x8d52f43bU,
    0x215a1e06U,
    0x3391692dU,
    0x86fcbc91U,
    0xd9c287fdU,
    0x58273618U,
    0x71179d17U,
    0xf73ca2b5U,
    0xfde0eeeeU,
    0x6b300a9dU,
    0x672a5aeaU,
    0x325b7492U,
    0x2216e3e7U,
    0x6b8e021U,
    0x4da5900cU,
    0x77564818U,
    0xa4582437U,
    0xfd14b622U,
    0x9d3cdd0bU,
    0x20e642a5U,
    0xf1b27ef6U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xf8000000U,
    0xec000000U,
    0xa2000000U,
    0xcd000000U,
    0x7a800000U,
    0x86400000U,
    0x83200000U,
    0x3e500000U,
    0x38b80000U,
    0x67840000U,
    0xc4c60000U,
    0x89630000U,
    0x4f728000U,
    0x33efc000U,
    0xc9bb6000U,
    0x87071000U,
    0x3836800U,
    0x82c25400U,
    0xc2617200U,
    0x1ef4b900U,
    0xb42b8280U,
    0x355ebdc0U,
    0xa9b48020U,
    0xb78cc030U,
    0x9c49e018U,
    0x62a8d014U,
    0xb118083eU,
    0xd795443bU,
    0x5b5a1a28U,
    0x36b2ed33U,
    0xc80cf09eU,
    0x248904e1U,
    0xd1cd8280U,
    0x8f6dbdffU,
    0xc4fe0036U,
    0x35a7003dU,
    0xea148017U,
    0xbe9cc00dU,
    0x6751e005U,
    0xcc3cd004U,
    0xeec60804U,
    0xa0624433U,
    0xb3f69a3eU,
    0x9eaa2d0eU,
    0x4b1b109eU,
    0xc696d4d1U,
    0x2bd90ab2U,
    0x69f039c6U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x88000000U,
    0x34000000U,
    0x12000000U,
    0x4f000000U,
    0x8800000U,
    0x2e400000U,
    0xa00000U,
    0x8a300000U,
    0xc3b80000U,
    0x6f840000U,
    0x62c60000U,
    0x4e70000U,
    0x29928000U,
    0xef0c4000U,
    0xeb7da000U,
    0x58629000U,
    0x85d37800U,
    0xc8adc400U,
    0xdcbb200U,
    0x2f9fe500U,
    0x91f06180U,
    0x3adf75c0U,
    0xabd48020U,
    0x61ab4030U,
    0xb84f2038U,
    0x165ed014U,
    0x3f96d822U,
    0xaa0b540dU,
    0xf0feca04U,
    0x8a52113U,
    0x7e315382U,
    0x31b8d0cbU,
    0x708741a0U,
    0xe245a5d2U,
    0x1ea45828U,
    0x3b37143fU,
    0x2a3b6a02U,
    0x2043b118U,
    0x19a42bacU,
    0xe7b214e5U,
    0x467e739cU,
    0xe7e600feU,
    0xaf1199bbU,
    0x184ef1ffU,
    0x665a9209U,
    0x792351bU,
    0x460a3992U,
    0x5efb61deU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xd0000000U,
    0x38000000U,
    0xd4000000U,
    0x7a000000U,
    0x91000000U,
    0x6800000U,
    0x47400000U,
    0x96600000U,
    0x3b300000U,
    0x5ba80000U,
    0x1f840000U,
    0xbdc60000U,
    0xa7270000U,
    0x42568000U,
    0xb519c000U,
    0x40eaa000U,
    0xfe1d000U,
    0x61f58800U,
    0x1509b400U,
    0xcf117a00U,
    0x9a7dd300U,
    0x715fa980U,
    0x6d090540U,
    0x3b108020U,
    0xd07ec030U,
    0x85c2038U,
    0x87881034U,
    0xd257280eU,
    0xad1c6435U,
    0xa4eaf21eU,
    0x9de76724U,
    0x1cf65381U,
    0xbd8e1651U,
    0x63550985U,
    0x9b98d57eU,
    0xbad080eU,
    0xe7837403U,
    0x89c35a19U,
    0xd26c328U,
    0xeb500180U,
    0x679fa17cU,
    0x7da8522fU,
    0x882b706U,
    0x5c45db84U,
    0x9e0a26eU,
    0x46f2f3aeU,
    0x1c8cc66aU,
    0x8dd001a3U,
    0x30dfa155U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0xf8000000U,
    0x4c000000U,
    0x8a000000U,
    0xd9000000U,
    0x17800000U,
    0x6e400000U,
    0xeca00000U,
    0x2a100000U,
    0x8b180000U,
    0x6a840000U,
    0xd7c20000U,
    0xb5e70000U,
    0x6318000U,
    0x3fcfc000U,
    0x6bfe6000U,
    0x51325000U,
    0x8f4bc800U,
    0xed3a3c00U,
    0xa7533200U,
    0x113fc500U,
    0x9555e080U,
    0x643a80c0U,
    0x38d38020U,
    0x2b78c010U,
    0x8ff7e038U,
    0xe529900cU,
    0x314fa83eU,
    0xb23b6c13U,
    0x3d37a22U,
    0x81f93936U,
    0xf8333285U,
    0x80cbd5dbU,
    0xff78489bU,
    0x39f6ecdaU,
    0xae297a3aU,
    0x83ca3906U,
    0xb9f8b2b3U,
    0x143715f2U,
    0xdacda8bdU,
    0xee787cfaU,
    0x9a775225U,
    0x66e9525U,
    0x3c6da89bU,
    0x5d687cf5U,
    0xe6ef520cU,
    0x32aa9521U,
    0xff0fa8adU,
    0x8e9f7ccfU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0xc8000000U,
    0x24000000U,
    0x2e000000U,
    0xe1000000U,
    0xc0800000U,
    0x9d400000U,
    0x4b600000U,
    0xea300000U,
    0x84880000U,
    0x8f840000U,
    0xfcc20000U,
    0x36a10000U,
    0x2c108000U,
    0xedde4000U,
    0xc53d2000U,
    0xd9cf7000U,
    0xa4e00800U,
    0x26f58400U,
    0x6a2b2a00U,
    0x4f91bb00U,
    0x1b1b4180U,
    0x3c9e7c40U,
    0xd45a8020U,
    0x14fb4010U,
    0x35efa008U,
    0x12b03004U,
    0x85cda832U,
    0x36e4b409U,
    0x25f6020bU,
    0x51ab4f38U,
    0x155063b0U,
    0xc8ba4367U,
    0xee0aebb2U,
    0x8dc4876aU,
    0x7e266189U,
    0xb5510c77U,
    0x98ba8825U,
    0x60ec410U,
    0xb9c48a3aU,
    0x98218b0eU,
    0x7056e9b8U,
    0x763ac860U,
    0x7a4c8220U,
    0x32200f3cU,
    0xef57c391U,
    0xbfbe7369U,
    0x138d4398U,
    0x2053342U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x28000000U,
    0x94000000U,
    0x3e000000U,
    0xed000000U,
    0xd7800000U,
    0x13c00000U,
    0xbe200000U,
    0x1cb00000U,
    0x4ed80000U,
    0x64840000U,
    0x59460000U,
    0xe2610000U,
    0x9d548000U,
    0x6e494000U,
    0x71e8a000U,
    0x779bf000U,
    0xf6e46800U,
    0x5c153c00U,
    0xa02c5e00U,
    0x6eb89100U,
    0x5ed6f580U,
    0x838f4ec0U,
    0x2ca8020U,
    0x2cac4030U,
    0x83fa2008U,
    0x12b3b024U,
    0xdbd8480aU,
    0x2f078c25U,
    0x7080960fU,
    0xa7465d3bU,
    0x2f66c3b5U,
    0xdad6e3c4U,
    0x558e2b8fU,
    0x5bce9ff7U,
    0x552cd5bbU,
    0x553cfecdU,
    0xef12c834U,
    0xeaabcc09U,
    0x32fab62aU,
    0xdf35ed21U,
    0xa31e8bacU,
    0x17a16fc7U,
    0x276bd88U,
    0x977cc2dfU,
    0xd3f41609U,
    0x15bf1d15U,
    0x305663bfU,
    0x59c913c7U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x48000000U,
    0xac000000U,
    0xd2000000U,
    0xb3000000U,
    0x2800000U,
    0x38400000U,
    0x63e00000U,
    0x6f100000U,
    0x98e80000U,
    0x33840000U,
    0x31c60000U,
    0x5d270000U,
    0x7ab18000U,
    0xbd1a4000U,
    0xd9f96000U,
    0x726c9000U,
    0xb7c09800U,
    0x18209400U,
    0x55350200U,
    0xd6ddbf00U,
    0x7bdec880U,
    0xe858af40U,
    0xc99f8020U,
    0xcb94030U,
    0xa20ee038U,
    0x2911d00cU,
    0x3de87812U,
    0x2c06442bU,
    0x1204fa34U,
    0x5305bb2cU,
    0x32855280U,
    0x7046844eU,
    0xcfe3cab8U,
    0xbd11106bU,
    0x2bef48beU,
    0x3102ef70U,
    0x986e006U,
    0x3ec5d000U,
    0x15a67800U,
    0x25f54424U,
    0xea7b7a10U,
    0x43acfb15U,
    0xeae3b281U,
    0x6293547fU,
    0xe82db28bU,
    0xf20545cU,
    0x9b232b4U,
    0x5f991446U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0xe8000000U,
    0xb4000000U,
    0x3a000000U,
    0x9f000000U,
    0x85800000U,
    0x3dc00000U,
    0x38e00000U,
    0x34f00000U,
    0x7a280000U,
    0x78840000U,
    0x4b420000U,
    0x96a30000U,
    0xd8d08000U,
    0xabf4000U,
    0x9f19a000U,
    0x6c4fd000U,
    0x8e31f800U,
    0xcb4dac00U,
    0x77b12e00U,
    0x889a700U,
    0x12559780U,
    0x7af9d940U,
    0x5e3a8020U,
    0xa2584010U,
    0x3eb2018U,
    0xc2639004U,
    0xdcb0d83aU,
    0xf7093c2dU,
    0xd093760eU,
    0xd79cdb27U,
    0xbf0fc1a1U,
    0x5495924fU,
    0x759d19aeU,
    0x6c0fae5dU,
    0xb716efa6U,
    0x5958354aU,
    0x746a8e10U,
    0xa4a5771cU,
    0xa3d4ef9aU,
    0x3d3b354cU,
    0x99da0e12U,
    0x32a370eU,
    0xac054f9aU,
    0x3600e553U,
    0x9901f636U,
    0x84809b23U,
    0xf546e190U,
    0xaba5024bU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0xc8000000U,
    0x54000000U,
    0xf2000000U,
    0x39000000U,
    0xe9800000U,
    0x44c00000U,
    0x59e00000U,
    0x5f100000U,
    0xe4b80000U,
    0xb6840000U,
    0x46420000U,
    0xdda30000U,
    0x13b38000U,
    0x3a8e4000U,
    0xf2cea000U,
    0x80699000U,
    0x195cc800U,
    0x9391cc00U,
    0xaffee600U,
    0xa2af00U,
    0xe0300880U,
    0x34e9dc0U,
    0x8a98020U,
    0x3eb94010U,
    0xeb872018U,
    0xf5c0d01cU,
    0x8463e832U,
    0x99d51c15U,
    0x4c5f0e3cU,
    0x5414b30eU,
    0x143c86baU,
    0xbc46ed1U,
    0xe363a6b6U,
    0x4750bec7U,
    0xe1a4e81U,
    0xa3b2a2e1U,
    0xd28cc09bU,
    0xb6cf51eeU,
    0xca6f6632U,
    0xbc5fef19U,
    0xdc152890U,
    0x203d4df6U,
    0x89c1e804U,
    0x12661c29U,
    0xfad48e1aU,
    0xb8def339U,
    0xc350268eU,
    0x641efec9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x58000000U,
    0x8c000000U,
    0xee000000U,
    0x7b000000U,
    0x28800000U,
    0xf4c00000U,
    0x94600000U,
    0xe2700000U,
    0x86280000U,
    0xe9840000U,
    0x5460000U,
    0xfd210000U,
    0xe1518000U,
    0x747ac000U,
    0x61786000U,
    0xf2fd7000U,
    0xeb8d800U,
    0x5e1bc400U,
    0x700fe600U,
    0x98508300U,
    0x19f84380U,
    0x8e3ab940U,
    0x5edf8020U,
    0xb66fc030U,
    0xb727e008U,
    0x4052b00cU,
    0x55ff3816U,
    0x40397423U,
    0x15d8de3bU,
    0xc6edf71eU,
    0xcfe69d8aU,
    0x3a364e7dU,
    0xcc889d85U,
    0xee934e48U,
    0x89f1d89U,
    0x57c88e46U,
    0xd0b6fdbfU,
    0x5d4f3e60U,
    0x517645bdU,
    0x8aa98a5cU,
    0x7417b97U,
    0xb822cd40U,
    0xfad6de29U,
    0x4c38f713U,
    0x3bd91db7U,
    0x9de98e68U,
    0xd7677d84U,
    0x96f5fe6eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0x78000000U,
    0x64000000U,
    0xf2000000U,
    0xeb000000U,
    0xc800000U,
    0x79c00000U,
    0x6d200000U,
    0x6fb00000U,
    0x3dc80000U,
    0xb9840000U,
    0x10420000U,
    0x35610000U,
    0x66d38000U,
    0xbd1bc000U,
    0x659f2000U,
    0xf6599000U,
    0x44785800U,
    0xdd4ce400U,
    0x6e448e00U,
    0x20630900U,
    0x3f568d80U,
    0xbd5a4640U,
    0x78f98020U,
    0xfc8ec010U,
    0x1766a008U,
    0xc5d7501cU,
    0x8d9ef81eU,
    0xfa5bb419U,
    0xfa7a763cU,
    0xa848bd3aU,
    0x67c4fb83U,
    0x2826fb5eU,
    0x3e377bbbU,
    0x218d3b4bU,
    0x32e05ba7U,
    0x390ab62U,
    0x81fa03b2U,
    0xb0d4f58U,
    0x4a250d93U,
    0xbd318640U,
    0x610ea010U,
    0xd523500cU,
    0x6bb4f823U,
    0x9fceb40dU,
    0x5a83f62eU,
    0xc67d00U,
    0xdaa25ba0U,
    0x12f1ab73U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x18000000U,
    0x1c000000U,
    0x22000000U,
    0x87000000U,
    0xac800000U,
    0x58400000U,
    0x2fe00000U,
    0xfab00000U,
    0x50680000U,
    0x70440000U,
    0x6be20000U,
    0xe4b10000U,
    0x456d8000U,
    0x43c4c000U,
    0x8322e000U,
    0xb1101000U,
    0x1739b800U,
    0x4599a400U,
    0xfb4c2200U,
    0x85576f00U,
    0x62dd6d80U,
    0xb42fa740U,
    0x8da2e020U,
    0x2e501010U,
    0xb459b808U,
    0x5769a42cU,
    0x9cc42206U,
    0x13a36f07U,
    0x7b576d88U,
    0xa7daa761U,
    0xfad600bU,
    0xd165d006U,
    0x5df6d803U,
    0x5f0d7412U,
    0x58b77a12U,
    0x976edb1bU,
    0xfcc0f792U,
    0x83a56c58U,
    0xd351af9aU,
    0xa3d9d856U,
    0x31abb583U,
    0x7462d36eU,
    0x76759a3fU,
    0xabcecb36U,
    0x2f114fa2U,
    0x4238c852U,
    0x561f8d8cU,
    0xa38fb75dU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0xd8000000U,
    0x34000000U,
    0xc2000000U,
    0x7b000000U,
    0x20800000U,
    0xa8c00000U,
    0x76e00000U,
    0x50500000U,
    0xccf80000U,
    0xfcc40000U,
    0x84e60000U,
    0x13570000U,
    0x287f8000U,
    0x4e02c000U,
    0xbd07a000U,
    0xa1811000U,
    0x37445800U,
    0xe4a7c400U,
    0xc174e600U,
    0x8cda700U,
    0xb3ad5a80U,
    0xe2bd63c0U,
    0x93e7a020U,
    0x5d11030U,
    0xd9bc5838U,
    0x9363c43cU,
    0xbd12e636U,
    0x875aa70dU,
    0x2f32dab0U,
    0x87efa3deU,
    0xc2980028U,
    0xf054001aU,
    0x1cfe0025U,
    0x34c30028U,
    0x98e18005U,
    0x3d51c032U,
    0xa57e2011U,
    0xd784d01aU,
    0x4e447802U,
    0x5f201439U,
    0xb9311e32U,
    0xeec733cU,
    0xa11a64beU,
    0xbc11c0c6U,
    0xd8dfbcb1U,
    0x5d77c4f6U,
    0xe6cd7abeU,
    0xdeaab3e7U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x38000000U,
    0x4000000U,
    0x8e000000U,
    0x13000000U,
    0xa2800000U,
    0xbf400000U,
    0x2a200000U,
    0x4ad00000U,
    0x54480000U,
    0x23440000U,
    0xb0260000U,
    0x3fd70000U,
    0xe9c98000U,
    0x8c85c000U,
    0xbc462000U,
    0x60a37000U,
    0xf9911800U,
    0xcc6cd400U,
    0xf091da00U,
    0xdbeb2100U,
    0x1251ce80U,
    0x6e083640U,
    0x6ce62020U,
    0x61337030U,
    0xd1f91838U,
    0x82f8d40cU,
    0x7fda0eU,
    0x6f382101U,
    0x4d1e4ea3U,
    0x3bcaf644U,
    0xe5818008U,
    0x5bc1c01fU,
    0x5a602032U,
    0x7874701eU,
    0x4d8981bU,
    0xe8a91409U,
    0x4a77fa0fU,
    0x5dd8510bU,
    0x3728d692U,
    0x74b0e24cU,
    0x5239fa05U,
    0xec9b513aU,
    0xce0f5693U,
    0x7ce62277U,
    0x89305a3eU,
    0xddfae108U,
    0x30f9eea5U,
    0x99784670U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0xf8000000U,
    0x94000000U,
    0x16000000U,
    0x47000000U,
    0x54800000U,
    0x46c00000U,
    0xa600000U,
    0xed900000U,
    0xdd580000U,
    0x1ac40000U,
    0xa0660000U,
    0x70950000U,
    0x5cd88000U,
    0xa183c000U,
    0x6b42a000U,
    0x9325b000U,
    0x67b2b800U,
    0xd56d6400U,
    0xa7ea9600U,
    0x8e2ae300U,
    0x1cca5480U,
    0x259bcec0U,
    0x5ba2a020U,
    0x2b75b030U,
    0x520ab828U,
    0x33f96404U,
    0x7c34963eU,
    0xdb2be325U,
    0x214cd485U,
    0x44d90ed1U,
    0x45868035U,
    0x3542c021U,
    0xa824202aU,
    0x4937703fU,
    0x56ae9809U,
    0xa80e1423U,
    0x36fa0e2dU,
    0xd9b5f70dU,
    0x9e6eda82U,
    0xc168f9f9U,
    0x71ee5ab8U,
    0x692f39efU,
    0x584afaa6U,
    0x9b5f89c7U,
    0xc5c0c2b7U,
    0xd0e12deeU,
    0xc850f4a3U,
    0x7dba7ee1U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x88000000U,
    0x74000000U,
    0xa6000000U,
    0x41000000U,
    0x89800000U,
    0xa9c00000U,
    0xdaa00000U,
    0xfa700000U,
    0xf2280000U,
    0xcfc40000U,
    0xbba20000U,
    0xe3f30000U,
    0x23ef8000U,
    0xe9604000U,
    0x93d1a000U,
    0xf6de3000U,
    0x24a98800U,
    0x72860c00U,
    0x347ee00U,
    0xf5e76500U,
    0xc5961b80U,
    0xdcbc3d40U,
    0x81f9a020U,
    0x891a3010U,
    0x770b8818U,
    0x15750c3cU,
    0xea86e22U,
    0x2987251dU,
    0x79c7bba9U,
    0xc2a20d50U,
    0xf6702802U,
    0xa82c3c3aU,
    0x5cc4662eU,
    0xd5266902U,
    0x82b4759eU,
    0xd90c186eU,
    0x60739ba7U,
    0x412b7d78U,
    0x31458012U,
    0xa2e7403cU,
    0xe5142028U,
    0x78f97022U,
    0xdc9da83eU,
    0x9ccf7c22U,
    0x50d24625U,
    0xc05c193fU,
    0x29ee5dadU,
    0x22672459U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0xa8000000U,
    0x2c000000U,
    0x52000000U,
    0x5f000000U,
    0x76800000U,
    0x5a400000U,
    0xe3600000U,
    0xadf00000U,
    0x4a780000U,
    0x70440000U,
    0x18620000U,
    0xd5730000U,
    0xa5388000U,
    0x3ea14000U,
    0xe7536000U,
    0x2b299000U,
    0xb26e7800U,
    0xed88f400U,
    0xb13eb200U,
    0xa8a00b00U,
    0xa6535780U,
    0x40ae47c0U,
    0x69ab6020U,
    0xe2d9010U,
    0xf7ec7818U,
    0xf34bf404U,
    0x471e322aU,
    0x38b54b0bU,
    0x379a3794U,
    0xcbf0d7d7U,
    0xf37f983dU,
    0xffc72406U,
    0xada12a20U,
    0x93d72f2fU,
    0xe66a7db8U,
    0x1b8d68d7U,
    0xe03b1db2U,
    0x6b27f8f2U,
    0x5195e58cU,
    0xf10a4ccdU,
    0x707a3793U,
    0x2340d7feU,
    0x4ce79838U,
    0x6833242fU,
    0xb4db2a11U,
    0xb7102f11U,
    0x4348fd9bU,
    0xdf1b28feU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xe8000000U,
    0xe4000000U,
    0xde000000U,
    0xbb000000U,
    0xe8800000U,
    0x24400000U,
    0x4ba00000U,
    0x7c300000U,
    0xfd680000U,
    0xc2440000U,
    0x8ca60000U,
    0xa6b70000U,
    0xe82d8000U,
    0xec634000U,
    0x63512000U,
    0x2b9f000U,
    0x82bc5800U,
    0x42ba0400U,
    0xa2bb0200U,
    0x12bc2d00U,
    0xfab9f980U,
    0x1eb84f40U,
    0xc0b92020U,
    0x7bbdf030U,
    0x933a5838U,
    0xb77d042cU,
    0xfcde823aU,
    0x80eb6d39U,
    0x7d86d9b7U,
    0xbfc2bf6eU,
    0x3366f81aU,
    0x95d7b439U,
    0x7dfbfa2aU,
    0x919c9933U,
    0xf2cf8385U,
    0xf0779649U,
    0x72cf83b4U,
    0x30779677U,
    0x92cf8398U,
    0x8077966eU,
    0x7acf83a8U,
    0x6477967aU,
    0xa4cf838aU,
    0xdf77966bU,
    0x4c4f83a4U,
    0xfb37966fU,
    0x7ef83b9U,
    0x87079661U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x98000000U,
    0x3c000000U,
    0x6e000000U,
    0x53000000U,
    0xb2800000U,
    0xfa400000U,
    0x50600000U,
    0x5ed00000U,
    0xcc080000U,
    0x80440000U,
    0x89620000U,
    0x8d530000U,
    0x89c98000U,
    0x53e0c000U,
    0xf313e000U,
    0xdba87000U,
    0x98371800U,
    0x7a9b5400U,
    0x41293200U,
    0x86767300U,
    0x18fdae80U,
    0x12fc9440U,
    0x33fbe020U,
    0xec7c7010U,
    0x5fbd1818U,
    0xe31c540cU,
    0x45eab226U,
    0xba51b30fU,
    0x494dce9bU,
    0xc4a32454U,
    0xeaf4980cU,
    0x69bc942eU,
    0xac19520cU,
    0xc969c31bU,
    0xbb12d695U,
    0xfac706fU,
    0x3237aa19U,
    0x4b9de727U,
    0x56af7c96U,
    0xa1b69766U,
    0x6e5b56aeU,
    0xb50cb07eU,
    0x43c44a36U,
    0x4a59736U,
    0xcaf064a0U,
    0x39b9c37fU,
    0x41864adU,
    0x6d6dc34cU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xe8000000U,
    0x8c000000U,
    0x92000000U,
    0x17000000U,
    0x26800000U,
    0x9ac00000U,
    0xbe600000U,
    0x6bd00000U,
    0xfe880000U,
    0x5cc40000U,
    0xef660000U,
    0xdc510000U,
    0xf3ca8000U,
    0xbfe34000U,
    0x8f926000U,
    0xce2d9000U,
    0x2f763800U,
    0x9b387400U,
    0x519db600U,
    0x7929e300U,
    0x99f7a680U,
    0xc9ff2040U,
    0x33fa6020U,
    0x68f99030U,
    0xfc783808U,
    0x21bd7414U,
    0x5159363aU,
    0x2c4fa323U,
    0xfea146a4U,
    0xdeb4f045U,
    0xe9dab829U,
    0xf38a3416U,
    0x51455627U,
    0x5927330eU,
    0xdbf3fe85U,
    0xf6fac474U,
    0xb979ee3fU,
    0x303c0702U,
    0xbd1c28bdU,
    0xe0eeb75dU,
    0xa7107096U,
    0xd9ef537aU,
    0x5c95feb8U,
    0x66abc450U,
    0xf8b36e34U,
    0xc8df4708U,
    0xfc0e48a3U,
    0x38032775U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x98000000U,
    0x5c000000U,
    0xe000000U,
    0xc5000000U,
    0xe0800000U,
    0x54400000U,
    0x59a00000U,
    0xb6900000U,
    0x9d80000U,
    0x41440000U,
    0x61220000U,
    0x1ed50000U,
    0xe7a8000U,
    0xaad0c000U,
    0xd478e000U,
    0x5d17000U,
    0xf3fea800U,
    0xe212c400U,
    0x8798c200U,
    0xb7e77d00U,
    0xf031fc80U,
    0xf0cefac0U,
    0x3b5ae020U,
    0x9e047010U,
    0xbd042828U,
    0x4c820414U,
    0x92402226U,
    0x58a60d17U,
    0x4175483U,
    0x96983ef1U,
    0x3d602218U,
    0x63760d05U,
    0x1aef54beU,
    0xa00c3ef9U,
    0x74ba2204U,
    0x82370d17U,
    0x2bcfd4b3U,
    0x26ddfee2U,
    0x26c2423dU,
    0x26e7bd38U,
    0x3ab11c88U,
    0xc38b8ac9U,
    0x817e4800U,
    0x9d57b43aU,
    0xbbc6a04U,
    0x7db4b922U,
    0x3009bea3U,
    0xcb847c4U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0x38000000U,
    0xfc000000U,
    0xea000000U,
    0x61000000U,
    0x30800000U,
    0xcb400000U,
    0xf0a00000U,
    0xb0100000U,
    0x31580000U,
    0x72440000U,
    0xcc260000U,
    0x99550000U,
    0xf4fc8000U,
    0xdcd5c000U,
    0x4139e000U,
    0xa0365000U,
    0x12098800U,
    0x7fb94400U,
    0x4c773a00U,
    0xc12a1500U,
    0x396ce880U,
    0xa4481ec0U,
    0xa59fe020U,
    0x5e235030U,
    0x64550828U,
    0x1e7c8424U,
    0xde96da0eU,
    0x4518453fU,
    0xcd6360baU,
    0x58f45ad8U,
    0xec6c5a2cU,
    0x4ac88502U,
    0x11de0094U,
    0x4103cac8U,
    0x60823222U,
    0x53469113U,
    0x9ca232a1U,
    0x62145bdaU,
    0xac5a809fU,
    0xa8c20acaU,
    0x6665d21eU,
    0x5971c108U,
    0x8fa93aa2U,
    0x1d29dfeaU,
    0x836eda88U,
    0x5d4e8fedU,
    0xf91dd281U,
    0x47670bfcU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0xf8000000U,
    0x1c000000U,
    0x32000000U,
    0xeb000000U,
    0x73800000U,
    0x21400000U,
    0x7de00000U,
    0x55b00000U,
    0x30880000U,
    0x42440000U,
    0x8a620000U,
    0xcaf10000U,
    0xdc6d8000U,
    0x6b73c000U,
    0xc62ae000U,
    0xd149000U,
    0xe6d8d800U,
    0x96795c00U,
    0x342a6a00U,
    0x8612bf00U,
    0x55b9580U,
    0xff3cc840U,
    0xadc8e020U,
    0xfda59010U,
    0xecd55808U,
    0x25fa9c2cU,
    0x75688a3eU,
    0x6bf22f07U,
    0x18e94d8cU,
    0x2bb0947aU,
    0xc18d0a3cU,
    0xaec5ef18U,
    0xbea1ad97U,
    0xab550479U,
    0x42b85212U,
    0x858f7307U,
    0x70c127a6U,
    0xbfa32b64U,
    0x9fd31fb5U,
    0xda7ee745U,
    0x2e29ad8aU,
    0x19110450U,
    0x10da523fU,
    0xe37e7306U,
    0x66aca78aU,
    0x23d0eb57U,
    0x1879ffb4U,
    0x1d2a7750U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x98000000U,
    0x1c000000U,
    0x5e000000U,
    0xab000000U,
    0x85800000U,
    0x3cc00000U,
    0xa6600000U,
    0x23700000U,
    0xbc880000U,
    0xd3c40000U,
    0xc1e20000U,
    0x5ab70000U,
    0xfc6a8000U,
    0xcdf0c000U,
    0xa5cae000U,
    0xf961f000U,
    0x8cf5d800U,
    0x414fec00U,
    0xd1278600U,
    0x2ad10300U,
    0x101ba380U,
    0xe6fdb840U,
    0x27c8e020U,
    0xac66f010U,
    0x12775838U,
    0x300b2c14U,
    0xcf876626U,
    0xedc3f307U,
    0x7ae6fb97U,
    0xa735946aU,
    0x8caf0601U,
    0xad96c31fU,
    0x31bbc391U,
    0x9e6c885cU,
    0xc8f7d829U,
    0xa348ec23U,
    0x9425061fU,
    0xcc55c328U,
    0x2d5b43b8U,
    0x3e5f487bU,
    0x57ddb837U,
    0x1d1adc0cU,
    0xc47abe15U,
    0xb8881f3fU,
    0xd1c37d8bU,
    0xd4e39753U,
    0x243625a3U,
    0xdd2fbb44U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x38000000U,
    0xc4000000U,
    0x6e000000U,
    0x59000000U,
    0x36800000U,
    0xb7c00000U,
    0xb5e00000U,
    0xc7900000U,
    0xc2f80000U,
    0x62c40000U,
    0xe1620000U,
    0xb550000U,
    0x9a988000U,
    0xef15c000U,
    0xe43c2000U,
    0xada27000U,
    0x13304800U,
    0xb74d7400U,
    0xfa4f1e00U,
    0x7ac91f00U,
    0xe009d480U,
    0x256f53c0U,
    0xf03e2020U,
    0x1ba77010U,
    0x3e30c828U,
    0xc7ccb404U,
    0xe8893e0eU,
    0x53aa6f31U,
    0xa75b1c9bU,
    0x3df6e7d6U,
    0xd42f9e2dU,
    0x3a18df3dU,
    0x25d7f485U,
    0xb95823f5U,
    0xacf6e81eU,
    0xaeafc439U,
    0xfdbf60bU,
    0x5b33db10U,
    0xbb4aa2a5U,
    0xd84948e7U,
    0xa1c8a2afU,
    0x1d8c48dcU,
    0x572822b1U,
    0xf39d88ccU,
    0xb19602adU,
    0xffaf8c8U,
    0xa246cab3U,
    0xdba64cd9U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0x8000000U,
    0xf4000000U,
    0x12000000U,
    0x67000000U,
    0x1800000U,
    0xe8400000U,
    0xb9200000U,
    0xb6f00000U,
    0x5c680000U,
    0x4d440000U,
    0xe7a20000U,
    0x83310000U,
    0x9b0b8000U,
    0x3f944000U,
    0x9fbda000U,
    0x69dd9000U,
    0x78eb800U,
    0x55574c00U,
    0x85586200U,
    0x654b9900U,
    0xbbb4ef80U,
    0x9bce9f40U,
    0xbe77a020U,
    0x74a89010U,
    0x48a73808U,
    0x16b20c1cU,
    0x314e4202U,
    0x99b2493dU,
    0xd4cff784U,
    0x3bf04359U,
    0x86eb7a20U,
    0x6281452aU,
    0xb3c235a6U,
    0xe2624a71U,
    0x6c532db5U,
    0x6dd9967eU,
    0x5d8e7795U,
    0xe6510375U,
    0xe6df5a0cU,
    0xe2099504U,
    0xf712ad87U,
    0xd778d64fU,
    0x3c3a57b0U,
    0x8e99d348U,
    0x1b2fc21aU,
    0x61e3093bU,
    0xfb13d7bbU,
    0x797c934cU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0x88000000U,
    0x54000000U,
    0x96000000U,
    0x31000000U,
    0x7f800000U,
    0xfec00000U,
    0x91200000U,
    0x85500000U,
    0x24c80000U,
    0xbdc40000U,
    0xf1a20000U,
    0x99150000U,
    0xaeae8000U,
    0x2c704000U,
    0x4d98a000U,
    0xa30db000U,
    0xe7610800U,
    0x6c30bc00U,
    0x45fed600U,
    0x35bef900U,
    0xc5dd4280U,
    0x5f6fca40U,
    0x37d2a020U,
    0xf78cb010U,
    0x5d258828U,
    0x2f51fc34U,
    0xa7caf622U,
    0x3d460915U,
    0x9de26aa5U,
    0x4b76864cU,
    0xe31fde3fU,
    0x524e452fU,
    0xd083948cU,
    0xa4413355U,
    0x2667e28bU,
    0xbb77a6aU,
    0xc73d2811U,
    0xb69c4c0eU,
    0x6f8bfe3eU,
    0x6126b515U,
    0x9d54bcb0U,
    0x58cc7f50U,
    0xafc09caaU,
    0xdea48f57U,
    0x8397b49eU,
    0xb9e9c351U,
    0x7290ca9dU,
    0x266a3663U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x68000000U,
    0x3c000000U,
    0x52000000U,
    0xf7000000U,
    0x69800000U,
    0x66400000U,
    0x91600000U,
    0x20b00000U,
    0x52580000U,
    0xf7440000U,
    0xb1e60000U,
    0xf2710000U,
    0x5ff98000U,
    0x4f574000U,
    0xad2fe000U,
    0x2a3f9000U,
    0x93743800U,
    0x377f0c00U,
    0x91965200U,
    0x9a8adb00U,
    0x59291d80U,
    0xb4392ec0U,
    0x7e71e020U,
    0xa5fa9030U,
    0xa453b808U,
    0xa6ad4c0cU,
    0xd37e321aU,
    0x57970b0fU,
    0x238d4594U,
    0xa5aff2fdU,
    0x24fdea3aU,
    0xdcd39729U,
    0x20692facU,
    0xb0db25c4U,
    0x4a8325aeU,
    0xa1c322c2U,
    0x72203210U,
    0x1f520b3dU,
    0x152ac59fU,
    0x6e3db2c9U,
    0xa5758a21U,
    0xc67e4700U,
    0x611577b6U,
    0x3049f9d8U,
    0x30892f8cU,
    0xa2b25f4U,
    0xfbbb258dU,
    0xb13722caU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0xa8000000U,
    0xec000000U,
    0xb6000000U,
    0x95000000U,
    0xef800000U,
    0x61c00000U,
    0x55200000U,
    0x90700000U,
    0xc3480000U,
    0x36c40000U,
    0xfda60000U,
    0x41370000U,
    0xe32b8000U,
    0xacd14000U,
    0x2ff96000U,
    0x332eb000U,
    0x54d52800U,
    0xfbfb3c00U,
    0xb12cee00U,
    0x33d3d900U,
    0xdb7b0b80U,
    0x1c6b3640U,
    0x7db76020U,
    0x906db030U,
    0x7bb0a838U,
    0x4d697c1cU,
    0x8300e2aU,
    0x32af293bU,
    0x2412c3adU,
    0xfb5cfa65U,
    0x139a461bU,
    0x773da528U,
    0x1f8885adU,
    0xa0615f78U,
    0x5a12c3baU,
    0x325cfa46U,
    0x21a460aU,
    0x1ffda519U,
    0xbb288591U,
    0x28d15f54U,
    0x95fac3b9U,
    0x6028fa46U,
    0x8654462bU,
    0x99bea522U,
    0x33cd05afU,
    0x63831f47U,
    0x67c6238cU,
    0x88240a6bU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0xd8000000U,
    0xd4000000U,
    0x2a000000U,
    0xd3000000U,
    0x62800000U,
    0xcbc00000U,
    0x59e00000U,
    0x64500000U,
    0x7f80000U,
    0x89440000U,
    0x62220000U,
    0x45b50000U,
    0x7ad8000U,
    0x7cbac000U,
    0xec60a000U,
    0x6f165000U,
    0x5ade1800U,
    0x40745400U,
    0x3f4c3e00U,
    0xdd6c3f00U,
    0xd8df6a80U,
    0x7f7073c0U,
    0xb3cd2020U,
    0x87ac9010U,
    0x3cbeb828U,
    0x4c62042cU,
    0xdf122636U,
    0x82d86b35U,
    0x9473548aU,
    0x154c4cf4U,
    0xe6a4ab8U,
    0xba58e3e2U,
    0xb4b1981eU,
    0xea2b9425U,
    0xe3f91e1fU,
    0x3b44af3bU,
    0xc523d2a4U,
    0xbd3777d0U,
    0xc76a8633U,
    0x93da3b09U,
    0x69f74c9dU,
    0xe20918dcU,
    0xd549f497U,
    0xee6b1cceU,
    0xaa5bd294U,
    0xdcb377ccU,
    0xe6288609U,
    0x1dff3b2aU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x48000000U,
    0x84000000U,
    0x76000000U,
    0x3000000U,
    0xa1800000U,
    0x400000U,
    0xca00000U,
    0x46f00000U,
    0xfd480000U,
    0x85c40000U,
    0x6ae20000U,
    0x51530000U,
    0xe6398000U,
    0xc2ce4000U,
    0x12822000U,
    0xa9c2b000U,
    0xe8e29800U,
    0xc56cc00U,
    0xf0be6200U,
    0x5d0bf500U,
    0x99679f80U,
    0x3413fe40U,
    0x501ba020U,
    0xd9fcf010U,
    0x5928b818U,
    0x97507c0cU,
    0xed3efa12U,
    0x874e3921U,
    0x54c07d9dU,
    0xee664b40U,
    0x8b961f88U,
    0x8759be50U,
    0x7adb803bU,
    0xd09d400dU,
    0x353ba035U,
    0x5b4cf00cU,
    0xbec0b815U,
    0x67647c35U,
    0x9314fa0cU,
    0xd7993930U,
    0xa2bbfdafU,
    0x380b0b7bU,
    0x23e5bfa6U,
    0x69d14e63U,
    0xe6f93821U,
    0x62aa3c3eU,
    0x3616da0fU,
    0x4d1b8904U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0xe8000000U,
    0x14000000U,
    0x22000000U,
    0x8d000000U,
    0x19800000U,
    0xf7c00000U,
    0x69200000U,
    0x14700000U,
    0x78980000U,
    0xaa440000U,
    0xf4e60000U,
    0x94530000U,
    0xaf688000U,
    0x541b4000U,
    0x1802a000U,
    0x2c041000U,
    0xbe07e800U,
    0xab070c00U,
    0x5e805e00U,
    0x7742f900U,
    0xa565db80U,
    0x79760c0U,
    0x1c4a2020U,
    0x316f5030U,
    0x4f1d4818U,
    0xde871c04U,
    0xb741b63aU,
    0xc566f505U,
    0x17950588U,
    0xf44ad9e3U,
    0x256b5ba6U,
    0x6d1f20cdU,
    0x53800022U,
    0xaec00031U,
    0x32a0003cU,
    0x7eb0002bU,
    0xe038000fU,
    0x5df40003U,
    0xc75e0025U,
    0xa767000bU,
    0x3a96801aU,
    0x9dcc4033U,
    0x2aac202bU,
    0xf83c5005U,
    0x71f5c834U,
    0x795c5c12U,
    0xc63163fU,
    0x6412e531U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0x58000000U,
    0x4c000000U,
    0x2e000000U,
    0x59000000U,
    0x57800000U,
    0x2b400000U,
    0x14a00000U,
    0xb4100000U,
    0x3ac80000U,
    0x32c40000U,
    0xd6e60000U,
    0xdf310000U,
    0xe19a8000U,
    0x26aec000U,
    0x6b306000U,
    0x139e7000U,
    0x29ad4800U,
    0x79b0fc00U,
    0x195efa00U,
    0x34c3500U,
    0xa0845880U,
    0xe5c60240U,
    0x4862e020U,
    0xdbf4b030U,
    0xc27b2808U,
    0xf01f8c14U,
    0xf0e93216U,
    0x62120913U,
    0xbfcac28bU,
    0x33444756U,
    0xb8a3f0b5U,
    0xea164e7aU,
    0x6bc9322dU,
    0x71420909U,
    0x9fa2c290U,
    0xbc90474bU,
    0x5b0df088U,
    0x60a34e42U,
    0x6615b20eU,
    0x65cdc931U,
    0x78402294U,
    0x9024f764U,
    0xdbd6d8bcU,
    0x61acc24aU,
    0x8db4800bU,
    0xb5bc003U,
    0x7c4ce02bU,
    0xba01b027U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0x68000000U,
    0xcc000000U,
    0xae000000U,
    0x27000000U,
    0xf7800000U,
    0x98c00000U,
    0xa7a00000U,
    0x5fb00000U,
    0xdba80000U,
    0x9440000U,
    0x64620000U,
    0xe9910000U,
    0x7df8000U,
    0x44cf4000U,
    0x41506000U,
    0xa87e7000U,
    0x77bb800U,
    0xacfd7c00U,
    0x223ee600U,
    0xf69ead00U,
    0x54ad7580U,
    0x52c5f940U,
    0x82a7e020U,
    0xf1353010U,
    0xbe9d808U,
    0x65a20c3cU,
    0x66b2de1aU,
    0x43289133U,
    0x5a01f3abU,
    0x9d042449U,
    0xfa86ad9dU,
    0xda46f576U,
    0x29e2be01U,
    0x7c56e13bU,
    0xe2fa4ba4U,
    0x9539584dU,
    0xb9184b88U,
    0x98685850U,
    0xff67cbbfU,
    0x98171852U,
    0x541fabb6U,
    0x4aed6844U,
    0xc9261387U,
    0xeff11468U,
    0x674f7583U,
    0x9394f975U,
    0x6ad86000U,
    0xd64a7018U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x68000000U,
    0xdc000000U,
    0x96000000U,
    0xb9000000U,
    0x1c800000U,
    0xf400000U,
    0xefa00000U,
    0x80300000U,
    0xba680000U,
    0xddc40000U,
    0xade20000U,
    0x59130000U,
    0xaa1d8000U,
    0x458f4000U,
    0x3fd0e000U,
    0x387db000U,
    0xf85f4800U,
    0x64a82400U,
    0x4ba0da00U,
    0x3231ad00U,
    0xf168b980U,
    0xea44df40U,
    0x25256020U,
    0x5676f010U,
    0xf94da818U,
    0x3ab6942cU,
    0x242a121aU,
    0x92e2c937U,
    0x49283a5U,
    0xdddfc26eU,
    0x376d11a7U,
    0xcb464b53U,
    0xeda57203U,
    0xa333391cU,
    0x51e8abacU,
    0xc301166cU,
    0x1b8063b4U,
    0x16c1725fU,
    0xda67d9b0U,
    0xb1d52f40U,
    0x157f4811U,
    0x7ed82426U,
    0x9ce8da0cU,
    0xb585ad3bU,
    0xebc2b9afU,
    0x84e3df77U,
    0xfd92e004U,
    0xa15eb020U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x38000000U,
    0xd4000000U,
    0x16000000U,
    0xd9000000U,
    0x6a800000U,
    0xe9400000U,
    0x73200000U,
    0xe1700000U,
    0xa1b80000U,
    0x71c40000U,
    0xdd620000U,
    0x3dd50000U,
    0x690a8000U,
    0x691dc000U,
    0x4f70a000U,
    0xecbfb000U,
    0xad454800U,
    0xbd221400U,
    0x1c752a00U,
    0xb53f0900U,
    0xad84b380U,
    0x6c431c0U,
    0xfae22020U,
    0x8167010U,
    0x7a6fe828U,
    0xa9cca414U,
    0x32f8e20eU,
    0x59e5dd35U,
    0x3993b985U,
    0x96ad48f6U,
    0xd0297bbaU,
    0x1b6ee5eaU,
    0x574d2a14U,
    0x6dbb093cU,
    0xa3c6b38eU,
    0x8a6131fdU,
    0xa50a01cU,
    0xc4cfb01cU,
    0xfe7d482dU,
    0xa1a61402U,
    0x9c372a1bU,
    0x649a090cU,
    0x19b6339aU,
    0x2e5df1fdU,
    0x71508022U,
    0xd14cc014U,
    0x6cb82032U,
    0xed477038U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x88000000U,
    0xd4000000U,
    0xce000000U,
    0xfb000000U,
    0xb8800000U,
    0xc3400000U,
    0xa4e00000U,
    0xe8500000U,
    0x8ab80000U,
    0x69c40000U,
    0x46a60000U,
    0xa1330000U,
    0xb22a8000U,
    0xb4584000U,
    0x7494a000U,
    0xef9b7000U,
    0xe3325800U,
    0x1b2d0400U,
    0x8ddbfa00U,
    0x6a557700U,
    0x43bfaa80U,
    0x4569c0U,
    0xd0662020U,
    0xd9173030U,
    0x9f58f818U,
    0x84117414U,
    0x98dd2222U,
    0xd9d73335U,
    0x807c70b3U,
    0x6be02efeU,
    0xed5f28eU,
    0x5afc6dc0U,
    0xf1a3da11U,
    0x2bb5471eU,
    0xa06bd298U,
    0x453f5dfbU,
    0x40052220U,
    0xa0033313U,
    0x300270a9U,
    0xd8072ef7U,
    0x5c017298U,
    0x1a032df4U,
    0x3503fa38U,
    0x4381773cU,
    0x7bc1aa84U,
    0x67a269f9U,
    0x4cb2a035U,
    0x62e87004U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x8000000U,
    0x9c000000U,
    0x32000000U,
    0xeb000000U,
    0xa0800000U,
    0xc1400000U,
    0x1f600000U,
    0xc2700000U,
    0xc7e80000U,
    0x37c40000U,
    0x83260000U,
    0x90950000U,
    0xf15e8000U,
    0x5f8c4000U,
    0x7f36e000U,
    0x5a8a9000U,
    0x66b0d800U,
    0xa04dc400U,
    0xc013aa00U,
    0x281d0b00U,
    0x74e99b80U,
    0xc3440340U,
    0x4c606020U,
    0x86f2d030U,
    0x70a83828U,
    0x65a65434U,
    0xa4d3f202U,
    0xefbd8f27U,
    0x533c518cU,
    0x607ed87aU,
    0x341f4388U,
    0x86ecc740U,
    0x88454a0fU,
    0x3ce79b34U,
    0x4fb1439bU,
    0xf3cdc75eU,
    0x95d5ca0eU,
    0x883adb09U,
    0x78ff23b0U,
    0x115e1769U,
    0xef8d7212U,
    0x731cf01U,
    0x1e8ab18bU,
    0xc0b4485bU,
    0xe54f9babU,
    0xb9910344U,
    0xa2dee002U,
    0xa4e900dU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0xf8000000U,
    0x14000000U,
    0x32000000U,
    0x1d000000U,
    0xa2800000U,
    0x60400000U,
    0xa4200000U,
    0x7f00000U,
    0x93080000U,
    0xffc40000U,
    0xf6e20000U,
    0xeb970000U,
    0xecd98000U,
    0xb53bc000U,
    0xa12ea000U,
    0x6f309000U,
    0x2e6c6800U,
    0xd4905400U,
    0x3b5fe200U,
    0x4bf8d100U,
    0x8b4e6780U,
    0xfe63440U,
    0x23152020U,
    0x559c5010U,
    0x499b4838U,
    0x379b040cU,
    0xdc9d2a3eU,
    0x99181505U,
    0xaadded8cU,
    0xea3eb147U,
    0x6aca588U,
    0x89f6b548U,
    0x200a0fb1U,
    0x7e42605dU,
    0x3f20c202U,
    0x5a778126U,
    0x84ceafb7U,
    0x7521f074U,
    0x13772a01U,
    0xf44b1527U,
    0x38666d9dU,
    0xedd27149U,
    0x87fb85b0U,
    0x3d4de546U,
    0x8e0c7aaU,
    0x4096a44dU,
    0x4959480eU,
    0xb6fc042aU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0x78000000U,
    0x6c000000U,
    0x1e000000U,
    0x55000000U,
    0x77800000U,
    0x3cc00000U,
    0xcc200000U,
    0xd9100000U,
    0x51e80000U,
    0xce440000U,
    0x4f620000U,
    0x9df50000U,
    0xc6df8000U,
    0x75bfc000U,
    0x9e4d6000U,
    0x4993f000U,
    0xb1aa2800U,
    0x4c26fc00U,
    0x9914c600U,
    0xf1eb3100U,
    0x3e447080U,
    0x3762ab40U,
    0xf1f0e020U,
    0xd8d93010U,
    0x20b8c828U,
    0xe9cacc3cU,
    0x75538e1eU,
    0x7d8e3d1bU,
    0x95329e87U,
    0xc8fb6655U,
    0x3faa56bdU,
    0x7121aa5fU,
    0xaa91d8bbU,
    0x372b975aU,
    0xad614602U,
    0xbef5f104U,
    0xa05c90a2U,
    0xc4ff9b55U,
    0x31aa2815U,
    0xc26fc0cU,
    0x3914c626U,
    0x1eb3102U,
    0x46447088U,
    0x5b62ab42U,
    0xeff0e035U,
    0x8dd9301bU,
    0x5738c80cU,
    0xd50acc37U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x48000000U,
    0x4c000000U,
    0xe2000000U,
    0xd000000U,
    0x44800000U,
    0x29400000U,
    0x3200000U,
    0xb8d00000U,
    0xebe80000U,
    0xc8c40000U,
    0xd2e60000U,
    0x58b30000U,
    0x21198000U,
    0x757e4000U,
    0x99886000U,
    0xe7361000U,
    0x37dce800U,
    0xa09ee400U,
    0x94baea00U,
    0x31ee2900U,
    0x61c05e80U,
    0x6864c340U,
    0xe6f7e020U,
    0x6fbb5030U,
    0xe6d0818U,
    0xb706b40cU,
    0xdd866212U,
    0xdbc2dd13U,
    0xf1605cb8U,
    0x14770e43U,
    0x9dfcd4b1U,
    0xa2cbfa7aU,
    0xc114d6b8U,
    0xb78c3752U,
    0xc831e210U,
    0x1e5b9d1dU,
    0xfd5fbc86U,
    0xf6d85e4aU,
    0xc61fdc93U,
    0xd0fa4e77U,
    0x464d34b7U,
    0xb853aa62U,
    0xcca85e8cU,
    0x74e0c351U,
    0x5bb1e039U,
    0xda98502aU,
    0x6dbc883dU,
    0xf36cf435U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x58000000U,
    0x54000000U,
    0x6000000U,
    0x61000000U,
    0xd0800000U,
    0x2dc00000U,
    0x96a00000U,
    0x59d00000U,
    0x7b80000U,
    0x88440000U,
    0x8de60000U,
    0x93b10000U,
    0x70498000U,
    0xf7eec000U,
    0xc6b8a000U,
    0xa8c33000U,
    0xd8258800U,
    0x6114c400U,
    0x77988e00U,
    0xc550e500U,
    0x987d5680U,
    0x69e081c0U,
    0xdb72020U,
    0x654cf030U,
    0x116ca808U,
    0xd27d340cU,
    0xbae3a616U,
    0xaa361115U,
    0x9109d081U,
    0x768a60d8U,
    0x62ca5894U,
    0x152fa4fbU,
    0xcd9b568dU,
    0x6e5181eaU,
    0x5bfea03fU,
    0xc3a2300bU,
    0x5f54083cU,
    0x37e0425U,
    0xf2662e3fU,
    0xf3f2d53bU,
    0x31a95ea1U,
    0x165e85efU,
    0xe1f10e06U,
    0xb6ae2517U,
    0xb7ddf692U,
    0x74b7b1edU,
    0x1cca82eU,
    0x2aad3404U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x88000000U,
    0x54000000U,
    0xfa000000U,
    0x37000000U,
    0x21800000U,
    0x60c00000U,
    0x30200000U,
    0x91300000U,
    0x1bb80000U,
    0x92440000U,
    0x43620000U,
    0x42d70000U,
    0x51298000U,
    0xc7084000U,
    0x8f3ca000U,
    0x9806d000U,
    0x4c053800U,
    0xf603cc00U,
    0x21046600U,
    0x3080b100U,
    0xd8404180U,
    0xbc67b9c0U,
    0xd7572020U,
    0xfbe99010U,
    0x48281838U,
    0x6b895c0cU,
    0x197ffe22U,
    0xd962ad15U,
    0x5d09fbeU,
    0x18a884cdU,
    0xc3cda7a8U,
    0xcd1f48c8U,
    0x6a33c194U,
    0x8c3cf9f8U,
    0x3380003cU,
    0x73c0002dU,
    0x83a00034U,
    0xa2f00004U,
    0x78180000U,
    0xb40031U,
    0xb37a0019U,
    0x16630017U,
    0x18538011U,
    0xe66b4025U,
    0xb6ef2012U,
    0x1ead901cU,
    0xcaca1834U,
    0x799e5c2aU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0xe8000000U,
    0x2c000000U,
    0xce000000U,
    0xdf000000U,
    0x74800000U,
    0xa0400000U,
    0x10a00000U,
    0x19f00000U,
    0x28880000U,
    0x4fc40000U,
    0xe6660000U,
    0x18150000U,
    0xda5a8000U,
    0x427b4000U,
    0x1dcb6000U,
    0x76e17000U,
    0xca577800U,
    0x93fb6400U,
    0x8c0da200U,
    0xc782d500U,
    0x7ac22780U,
    0x9fe5e840U,
    0xc5d7e020U,
    0x553f3030U,
    0x95ee9828U,
    0xa1555414U,
    0x7d7fba3aU,
    0x5949c10bU,
    0xcea47db3U,
    0x4ef21977U,
    0x800985bdU,
    0x99823d58U,
    0xedc7478cU,
    0x97659862U,
    0x43941838U,
    0xb69e143cU,
    0x369cda18U,
    0xf69cb12eU,
    0x569d05a2U,
    0x6987d44U,
    0xee98a78dU,
    0xc29ea874U,
    0xc9c8029U,
    0xd39e403dU,
    0xa719e006U,
    0x75e303dU,
    0x17fa1812U,
    0xe0f141eU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x8000000U,
    0x24000000U,
    0xa000000U,
    0xbd000000U,
    0x75800000U,
    0xfbc00000U,
    0xaa600000U,
    0xf9100000U,
    0x1bc80000U,
    0x3f440000U,
    0x1a260000U,
    0xc7b30000U,
    0x1c3d8000U,
    0x75dc4000U,
    0xfc0b2000U,
    0x73215000U,
    0x10356800U,
    0xaefd6c00U,
    0x183f3e00U,
    0x6fde6300U,
    0x990a5080U,
    0x9aa24a40U,
    0xcdf0a020U,
    0x979e1030U,
    0x23abc818U,
    0x47547c2cU,
    0x826f7602U,
    0x28355f09U,
    0x42fb8682U,
    0x563e056fU,
    0x68db6ebdU,
    0x598b294eU,
    0x30e17092U,
    0x96531a62U,
    0x79edc83cU,
    0xb4f77c1aU,
    0x981af61eU,
    0xd6d1f3bU,
    0xcb6a69aU,
    0x82bc5545U,
    0x9b86aaU,
    0xfb2e0543U,
    0x11136e8dU,
    0x4fcf2978U,
    0x5d477096U,
    0x33201a4eU,
    0xb0304819U,
    0x7efb3c3fU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0xb8000000U,
    0x4c000000U,
    0x26000000U,
    0xb7000000U,
    0x86800000U,
    0x64c00000U,
    0xcd600000U,
    0xde700000U,
    0xa2280000U,
    0x71440000U,
    0xdd260000U,
    0xdad70000U,
    0x8ebc8000U,
    0xac59c000U,
    0x4b686000U,
    0x90643000U,
    0x9f03800U,
    0x176a7400U,
    0xce640600U,
    0x12f6fb00U,
    0x27ef1b80U,
    0xa5a52d40U,
    0x1512e020U,
    0xbb5af030U,
    0x7decd838U,
    0x3ca38424U,
    0xa0925e2eU,
    0x1b1bbf13U,
    0xb149a589U,
    0x7433626dU,
    0xaf8b1d81U,
    0x5053d669U,
    0xec7dfbabU,
    0xa63fdd63U,
    0x3b9e381eU,
    0xa289741bU,
    0xdfd68628U,
    0x853c3b2cU,
    0x871dfba5U,
    0xf4fdd4cU,
    0xff363829U,
    0x270d7417U,
    0x77908600U,
    0xcd9b3b13U,
    0x8d897bb1U,
    0x65521d4aU,
    0xeff8581cU,
    0x97e4435U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0xe8000000U,
    0xb4000000U,
    0x22000000U,
    0x45000000U,
    0xed800000U,
    0x25400000U,
    0xa1600000U,
    0xa8100000U,
    0xee480000U,
    0x9bc40000U,
    0xcea60000U,
    0x64b70000U,
    0x1ff98000U,
    0xb1bb4000U,
    0x65fa000U,
    0x558c1000U,
    0x8566a800U,
    0x7210ac00U,
    0x274d4a00U,
    0x38437b00U,
    0xd0e29880U,
    0xfb518640U,
    0x18282020U,
    0x73545030U,
    0x5c2e8838U,
    0x5954fc0cU,
    0x9d2bc23aU,
    0x7ed3872dU,
    0x496f5a88U,
    0x10350151U,
    0x813efa9bU,
    0xb29a1179U,
    0x97afd2b0U,
    0x3312fd56U,
    0xc0cab899U,
    0xc05d677U,
    0x8e06a819U,
    0xdb00ac19U,
    0xe854a36U,
    0x47c77b15U,
    0x98a498a0U,
    0xe3b6864aU,
    0x8779a021U,
    0x917b1038U,
    0x367f282aU,
    0x7efbec1eU,
    0x863aea29U,
    0xea1b6b2aU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0xd8000000U,
    0x94000000U,
    0x52000000U,
    0x49000000U,
    0x7e800000U,
    0xfa400000U,
    0x35e00000U,
    0x45700000U,
    0xf6780000U,
    0x5bc40000U,
    0x82260000U,
    0x11570000U,
    0xf4298000U,
    0xe46ac000U,
    0xfc8ea000U,
    0xad7f9000U,
    0x4c46c800U,
    0xd6e5b400U,
    0x36f33200U,
    0xc8b8df00U,
    0x91649380U,
    0x4a363fc0U,
    0x43192020U,
    0x5bf65030U,
    0xfc3fe838U,
    0x5623e40cU,
    0x6354da36U,
    0x6d2f3b25U,
    0x72ee4994U,
    0x4aca04d2U,
    0x5e98e9bfU,
    0x123194ceU,
    0x17182195U,
    0xe9f320edU,
    0x853a9393U,
    0xf0a53fcfU,
    0xd16a01aU,
    0xacb900fU,
    0x7e98c838U,
    0xc236b43cU,
    0xff1cb20fU,
    0xa5f51f33U,
    0x433bb3a8U,
    0xeba76fdeU,
    0x3a974822U,
    0x8e0b743fU,
    0xb13b9231U,
    0xb2a04f24U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x18000000U,
    0x1c000000U,
    0xae000000U,
    0x8f000000U,
    0x90800000U,
    0x7bc00000U,
    0x4ee00000U,
    0xbc300000U,
    0xc0380000U,
    0x1e440000U,
    0xc0a20000U,
    0xc4970000U,
    0xa8aa8000U,
    0x50ecc000U,
    0x1b4ee000U,
    0xfd5c7000U,
    0xf37f800U,
    0xebbc400U,
    0x42827e00U,
    0x4ac7a700U,
    0xa963c980U,
    0x1b71edc0U,
    0x431e6020U,
    0x7f93b010U,
    0x82299838U,
    0xee287414U,
    0xa82be606U,
    0x932fd307U,
    0xf9a82fabU,
    0xa76e3ee3U,
    0x440e4f84U,
    0x32798eceU,
    0xd8e5d78bU,
    0xdf36faebU,
    0x56bcb1aeU,
    0xbe81e9c4U,
    0xb4c07e05U,
    0x3e60a711U,
    0x97f1499bU,
    0x96d92dfaU,
    0xbe72803cU,
    0xae98c03aU,
    0x55d4e017U,
    0xf88f7002U,
    0xefbf7821U,
    0xfd000412U,
    0x11861e38U,
    0xd4471728U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x28000000U,
    0xb4000000U,
    0x22000000U,
    0xa7000000U,
    0xb800000U,
    0xf3c00000U,
    0x64200000U,
    0x7b700000U,
    0xd5880000U,
    0x74240000U,
    0x33720000U,
    0x91890000U,
    0xae228000U,
    0x8734000U,
    0xc0ca000U,
    0xd8e39000U,
    0xc0d67800U,
    0x8f3cfc00U,
    0x9a8c0e00U,
    0xaba15900U,
    0x12b52780U,
    0xaaafa240U,
    0x42d67820U,
    0xb83cfc10U,
    0x990c0e08U,
    0x3c615934U,
    0x7c95278aU,
    0xc2dfa26dU,
    0xbede7828U,
    0x98d8fc39U,
    0xc5de0e0aU,
    0x25585908U,
    0x631fa793U,
    0xc5f8e273U,
    0x5428d81dU,
    0xa5966c24U,
    0x9858f606U,
    0x339ee52cU,
    0x5bbd89b8U,
    0xbec96b71U,
    0x8a47279eU,
    0x58e6a252U,
    0x80d4f816U,
    0xaf3fbc1fU,
    0x4a88ae16U,
    0x83a6c92fU,
    0xa6b15fb0U,
    0x88aa5e45U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x38000000U,
    0x54000000U,
    0x4e000000U,
    0x17000000U,
    0x6a800000U,
    0x6b400000U,
    0xa2a00000U,
    0x80b00000U,
    0x6280000U,
    0xdaa40000U,
    0x34b60000U,
    0xd8290000U,
    0x45a18000U,
    0x3237c000U,
    0xa96ba000U,
    0xbe03f000U,
    0xcf077800U,
    0xae80f400U,
    0xad44aa00U,
    0xd9a46d00U,
    0xf035f180U,
    0x34691b40U,
    0x5877820U,
    0xb5c0f430U,
    0x17e4aa08U,
    0xbd146d2cU,
    0x801df18eU,
    0xadcd1b55U,
    0x15b17833U,
    0x11a9f435U,
    0x9a652a12U,
    0x64d3ad36U,
    0x8dfe51a6U,
    0x49daeb75U,
    0xe8280032U,
    0xbda40003U,
    0x4636001fU,
    0x57690000U,
    0x91018037U,
    0xf187c039U,
    0x8bc3a038U,
    0x18e7f01cU,
    0x33917824U,
    0x9d59f407U,
    0x4c6d2a12U,
    0xb187ad1aU,
    0x6bc05197U,
    0x88e7eb54U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0x18000000U,
    0xe4000000U,
    0xf2000000U,
    0x85000000U,
    0xe800000U,
    0xb1c00000U,
    0xf5200000U,
    0x9c100000U,
    0x5e80000U,
    0x3d240000U,
    0x30160000U,
    0x9beb0000U,
    0x46238000U,
    0xb596c000U,
    0xb7ac2000U,
    0x7b441000U,
    0xe6e2e800U,
    0x64345400U,
    0x9b7d9e00U,
    0x870f4100U,
    0x6b11b480U,
    0x4e6f3140U,
    0xe262e820U,
    0x84f45430U,
    0xeadd9e18U,
    0x3bdf413cU,
    0x7f59b486U,
    0x3f9b3179U,
    0xde3ce81cU,
    0x8aeb5411U,
    0x62a01e1bU,
    0x5568110U,
    0x4e0814bbU,
    0xcb96e15eU,
    0xfcafa01dU,
    0x86c2d01eU,
    0xdea6c817U,
    0xb3544436U,
    0xa909762aU,
    0xa8101533U,
    0xbfefaa90U,
    0xd426b070U,
    0xc0977c96U,
    0xa12b7563U,
    0x2e839e32U,
    0xe1c04127U,
    0x7d2434a8U,
    0x9012f147U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x58000000U,
    0xdc000000U,
    0xe6000000U,
    0x89000000U,
    0x74800000U,
    0xa8400000U,
    0x7b200000U,
    0xb2100000U,
    0xb8b80000U,
    0x47240000U,
    0x44120000U,
    0x9bb0000U,
    0xbfa68000U,
    0x5250c000U,
    0x279de000U,
    0x9f31b000U,
    0xcbac8800U,
    0x6f1a4400U,
    0xc176f200U,
    0x18c5100U,
    0x258d5480U,
    0x6f89f7c0U,
    0x688c8820U,
    0x410a4410U,
    0xffcef218U,
    0x9fa85114U,
    0x4d1f5496U,
    0x1272f7f7U,
    0x4a0a0819U,
    0x284a8432U,
    0x146b1205U,
    0xeffde13eU,
    0xb981dc88U,
    0xc6c3b3dbU,
    0x8c627a37U,
    0x3cb21523U,
    0x5269a694U,
    0x16fea6fcU,
    0xa5075c87U,
    0xba8373dfU,
    0x95479a26U,
    0x3da7a510U,
    0xf1572eb0U,
    0x41fe2d0U,
    0x46f72e8eU,
    0xd24fe2fdU,
    0x5b6f2eb2U,
    0x227be2e1U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0xe8000000U,
    0xb4000000U,
    0x2000000U,
    0xf1000000U,
    0xa4800000U,
    0x50c00000U,
    0x73a00000U,
    0xb0500000U,
    0x68180000U,
    0xb7a40000U,
    0x1a520000U,
    0xcd1f0000U,
    0xe1238000U,
    0x53954000U,
    0xae3aa000U,
    0x3b39000U,
    0xb92ac800U,
    0xd4b8400U,
    0x213b1600U,
    0xb0325500U,
    0x466f8680U,
    0xd7aa16c0U,
    0x488ac820U,
    0xc1b8410U,
    0xda31638U,
    0xa756553cU,
    0xc79d86baU,
    0x1ee516edU,
    0xc3b14820U,
    0x192ac42cU,
    0x1d4bb611U,
    0x393ac528U,
    0xec34cea6U,
    0xf06bd2c1U,
    0x24a8fe3aU,
    0x1d0f0101U,
    0xf85cf897U,
    0x2ec457dbU,
    0x64a690beU,
    0x1fd743c5U,
    0xc15ecea9U,
    0x6e40d2fdU,
    0xce617e23U,
    0x31714125U,
    0x8b8fd8aeU,
    0x119987f9U,
    0x5de77893U,
    0x9e3517cdU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0x68000000U,
    0x7c000000U,
    0xa2000000U,
    0x69000000U,
    0x63800000U,
    0x24400000U,
    0x18600000U,
    0x55500000U,
    0x9ff80000U,
    0x5c640000U,
    0xd3520000U,
    0xe0fd0000U,
    0xf8e78000U,
    0x37914000U,
    0xaa5ee000U,
    0x50129000U,
    0xa81a0800U,
    0x4771fc00U,
    0x31cdb200U,
    0x4049cb00U,
    0x850fd680U,
    0x52e88dc0U,
    0x99fa0820U,
    0x6361fc10U,
    0xd7d5b228U,
    0x743dcb1cU,
    0xda45d69aU,
    0xe3618ddfU,
    0x97d78808U,
    0xd439bc0aU,
    0xaa46d230U,
    0x8b671b15U,
    0xebd4be9cU,
    0x763ea1caU,
    0xc341522fU,
    0xe8e65b1dU,
    0xcf925e84U,
    0x6e5831edU,
    0x96115a02U,
    0x771ea717U,
    0x93f26c8dU,
    0xbd09bac5U,
    0x76edecbcU,
    0x8ffcfaefU,
    0xa4610c94U,
    0x17536adeU,
    0x26fc84a6U,
    0x27e3d6cdU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x18000000U,
    0xdc000000U,
    0x16000000U,
    0x1f000000U,
    0x3a800000U,
    0x55c00000U,
    0x9e200000U,
    0x72d00000U,
    0x4c980000U,
    0x10240000U,
    0x71d60000U,
    0x801f0000U,
    0xeae18000U,
    0xcd76c000U,
    0x7b09e000U,
    0x2e593000U,
    0xb0823800U,
    0x7cc63400U,
    0xeba09e00U,
    0xb5914900U,
    0x437c6280U,
    0x36d29ac0U,
    0x469a3820U,
    0xf9223430U,
    0xe4569e38U,
    0xf75e492cU,
    0xfd05e286U,
    0x37805af7U,
    0x5a45d825U,
    0x48640437U,
    0x84b52636U,
    0x132ebd39U,
    0xf38c9ca1U,
    0xde9823ebU,
    0xe52382b6U,
    0x1254aaf3U,
    0x5859800aU,
    0xdf82c029U,
    0xbe47e023U,
    0xd2623034U,
    0x25b5b80eU,
    0xf2aff43fU,
    0x56c8fe3bU,
    0x1c7eb92dU,
    0x2c57baa9U,
    0x435d9ed7U,
    0xcf009e1aU,
    0x5281492eU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0xe8000000U,
    0xe4000000U,
    0x7a000000U,
    0xb3000000U,
    0xd6800000U,
    0x93c00000U,
    0x16200000U,
    0x79900000U,
    0x43780000U,
    0x24240000U,
    0x7e920000U,
    0x17fb0000U,
    0x78e78000U,
    0xc8354000U,
    0xd0ac2000U,
    0x323c5000U,
    0x63c4d800U,
    0xbe209400U,
    0xfd97ce00U,
    0x4979e900U,
    0x7f211880U,
    0x4c1075c0U,
    0xfe3cd820U,
    0xddc49410U,
    0x6725ce18U,
    0x12e91cU,
    0x3e98baU,
    0x64c135f9U,
    0xeaa2f83eU,
    0xa153c43cU,
    0xffde962dU,
    0xb8733d38U,
    0x6cf76bfU,
    0x55cb8ce7U,
    0xb34ab8aeU,
    0xa88965f5U,
    0x2aac2012U,
    0xc13c502dU,
    0xd544d839U,
    0x5de09409U,
    0x3b7ce00U,
    0xd4e9e910U,
    0x465918acU,
    0xdb3475e2U,
    0x562ed831U,
    0x59ff940aU,
    0x9e24e26U,
    0xb1b7a914U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0xb8000000U,
    0xc4000000U,
    0xca000000U,
    0x89000000U,
    0xa9800000U,
    0x50c00000U,
    0x81600000U,
    0x6d500000U,
    0xf1c80000U,
    0xf640000U,
    0xee560000U,
    0x714d0000U,
    0x46218000U,
    0x87f3c000U,
    0x597b2000U,
    0x10bfd000U,
    0xf05b8800U,
    0x11cc6c00U,
    0xdf665600U,
    0x46530300U,
    0xbd4d4680U,
    0xf0213840U,
    0xf38820U,
    0xb3f86c30U,
    0x60f85628U,
    0x887a032cU,
    0xad3ac6aeU,
    0xc29ff871U,
    0xd5a92812U,
    0xadb47c12U,
    0x6058fe02U,
    0xd9c9bf38U,
    0x367188eU,
    0xf850976aU,
    0x464a38aeU,
    0x14a24751U,
    0x33b03099U,
    0xeb5deb54U,
    0x254d4697U,
    0x44213857U,
    0xd2f3883eU,
    0x4ef86c08U,
    0xbb785619U,
    0x95ba0333U,
    0x4fdac6b0U,
    0x760ff86cU,
    0xc81281dU,
    0x9f407c26U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0xa8000000U,
    0x8c000000U,
    0x6e000000U,
    0x4f000000U,
    0x9800000U,
    0x43400000U,
    0x83e00000U,
    0xa5b00000U,
    0xdd580000U,
    0x11e40000U,
    0x3cb60000U,
    0x9fd90000U,
    0x69228000U,
    0x75134000U,
    0x730d6000U,
    0x3509f000U,
    0x360d3800U,
    0x718f2c00U,
    0xadccea00U,
    0x8faea700U,
    0xe559c180U,
    0xd5e320c0U,
    0x2eb53820U,
    0xc6db2c30U,
    0xba2ea08U,
    0xddd3a734U,
    0x92ad41aaU,
    0x55d960e3U,
    0x422d83bU,
    0x7d959c23U,
    0xa6ccb20aU,
    0x54287b24U,
    0xdf18138aU,
    0xe482abcaU,
    0x8bc0138cU,
    0x7626abe7U,
    0x149613a5U,
    0x9c4fabf3U,
    0x8ec93b8U,
    0x74b8ebd3U,
    0x2957f3b2U,
    0x91681bfdU,
    0x7ff84b99U,
    0xd6b477dfU,
    0x22d9c18aU,
    0xc9a320f3U,
    0x2cd53829U,
    0x7c2b2c31U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x48000000U,
    0x24000000U,
    0xbe000000U,
    0xa7000000U,
    0xda800000U,
    0x5a400000U,
    0x94200000U,
    0xff300000U,
    0x12d80000U,
    0xd6240000U,
    0x9a360000U,
    0x6d5b0000U,
    0x93e18000U,
    0xfbd44000U,
    0xbb8a2000U,
    0x73ec7000U,
    0x73fd7800U,
    0xe96f400U,
    0x712fa200U,
    0xb8995900U,
    0xaa05cd80U,
    0x9101d4c0U,
    0x39857820U,
    0x8ec2f430U,
    0x2161a218U,
    0x9596592cU,
    0x9aa4d92U,
    0xf7da94c9U,
    0x69a0d80fU,
    0xbff5c419U,
    0xbcb97a2eU,
    0xd379d3aU,
    0x3fdd37b7U,
    0x8da209f6U,
    0x61f26f8bU,
    0xabbc8decU,
    0x9fb6b5a8U,
    0x419820d1U,
    0xa7855a2bU,
    0x39c0ed14U,
    0x63e1cf8fU,
    0x13d0bde5U,
    0x4f8fed91U,
    0x35eda4deU,
    0xb8f8003cU,
    0x4e14000cU,
    0x326e0019U,
    0x513f0001U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0xf8000000U,
    0xf4000000U,
    0x9a000000U,
    0x41000000U,
    0xf4800000U,
    0xa4400000U,
    0x56600000U,
    0x97700000U,
    0x81980000U,
    0xd7640000U,
    0x3f20000U,
    0x75d90000U,
    0x9078000U,
    0x9885c000U,
    0x9a46a000U,
    0x5675000U,
    0x2ef5c800U,
    0xbf580400U,
    0xfe45f200U,
    0x7762bb00U,
    0x33f21f80U,
    0xaddaaec0U,
    0x8d07c820U,
    0xfa810410U,
    0x2f427208U,
    0x6be77b1cU,
    0xcbb4bfbeU,
    0x1dbdfefdU,
    0xcd720006U,
    0xa0990000U,
    0x73e78035U,
    0x2fb5c035U,
    0x2fbea02bU,
    0xf0735018U,
    0xc21fc806U,
    0xf8a50425U,
    0x5650723dU,
    0xa90e7b34U,
    0x774b3f97U,
    0x702c3ec3U,
    0xbbdea026U,
    0xe2035024U,
    0xf507c803U,
    0x4e81042eU,
    0x95427203U,
    0x5ae77b06U,
    0xc734bf8cU,
    0x4dfdfefaU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0xa8000000U,
    0x94000000U,
    0xae000000U,
    0xbf000000U,
    0x9b800000U,
    0x71400000U,
    0xfba00000U,
    0x33d00000U,
    0x51d80000U,
    0xdca40000U,
    0xf4560000U,
    0xca9d0000U,
    0xee078000U,
    0xdf034000U,
    0x4b81a000U,
    0xa946f000U,
    0xc7a0b800U,
    0x9d07c00U,
    0x40defa00U,
    0xf8262500U,
    0x1e934280U,
    0x407c04c0U,
    0x2676b820U,
    0xbd0d7c30U,
    0xc6f97a28U,
    0x81b5651cU,
    0xf96ae2aaU,
    0x2d4ef4e5U,
    0x71d8000bU,
    0x6ca4001fU,
    0xfc56000eU,
    0x2e9d0000U,
    0xe8078014U,
    0xf4034029U,
    0x7e01a03fU,
    0x6706f018U,
    0xa780b81bU,
    0x4b407c2eU,
    0xeaa6fa05U,
    0x1752253bU,
    0xbb1d4286U,
    0x564504ddU,
    0x3c27382cU,
    0xa8933c00U,
    0x637f5a03U,
    0xf7f0d515U,
    0x754bfabfU,
    0x8dd878dfU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x88000000U,
    0xbc000000U,
    0x46000000U,
    0x65000000U,
    0x2800000U,
    0xcb400000U,
    0x90a00000U,
    0x4ff00000U,
    0xe6280000U,
    0xcda40000U,
    0xa9720000U,
    0xff6d0000U,
    0x52008000U,
    0x4f064000U,
    0x980e000U,
    0x80c6b000U,
    0x97676800U,
    0xcd15f400U,
    0x7dfc6a00U,
    0xca3bad00U,
    0x10dc6b80U,
    0x1f0a6b40U,
    0x72956820U,
    0x6038f410U,
    0x5bdcea28U,
    0xf48ded24U,
    0xe5548ba2U,
    0x6ad8db6fU,
    0x7c080031U,
    0xb5140009U,
    0x39fa0028U,
    0x88390016U,
    0x97da8006U,
    0xa8f403cU,
    0x24526028U,
    0x1a5df02aU,
    0x284f082aU,
    0x431040dU,
    0xdac9e230U,
    0x3775e900U,
    0x4e6fe9bbU,
    0x9a867253U,
    0xbf4789afU,
    0xcaa2825cU,
    0xfcf20183U,
    0x9acc655U,
    0xb86183acU,
    0xb490df7cU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0x58000000U,
    0x8c000000U,
    0xe2000000U,
    0xc7000000U,
    0xb8800000U,
    0xb4400000U,
    0x70a00000U,
    0x5bf00000U,
    0xf3c80000U,
    0x43a40000U,
    0xfd720000U,
    0x9a8d0000U,
    0xf0858000U,
    0x1040c000U,
    0xe6a16000U,
    0xc2f69000U,
    0x364cf800U,
    0x465a400U,
    0x6310b200U,
    0xc298bd00U,
    0x909a9b80U,
    0x4f9efc40U,
    0xdb1ef820U,
    0xbd58a410U,
    0x52fd3228U,
    0x3d8c7d0cU,
    0x9801fb96U,
    0x6c016c63U,
    0x72058018U,
    0xaf00c021U,
    0x6c816006U,
    0xda469021U,
    0x55a4f80aU,
    0x2471a435U,
    0xff0ab204U,
    0x8741bd21U,
    0xd6251b91U,
    0x32b73c4bU,
    0x40e81800U,
    0xae93f402U,
    0x815d2a2dU,
    0xa8fb892cU,
    0xd68ed1a4U,
    0xf287e544U,
    0x4746d1bcU,
    0x3623e555U,
    0xa2b4d1bbU,
    0x28eee552U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0x48000000U,
    0xf4000000U,
    0xa6000000U,
    0x9d000000U,
    0xc3800000U,
    0xb8c00000U,
    0xf4a00000U,
    0x84b00000U,
    0x87a80000U,
    0xe5a40000U,
    0xed360000U,
    0xc8690000U,
    0x5868000U,
    0x95c44000U,
    0xf26a000U,
    0xd077b000U,
    0x690ea800U,
    0xae15bc00U,
    0x921ef600U,
    0xcb8e4d00U,
    0x6752b980U,
    0x957cbb40U,
    0xfd18a820U,
    0xa30cbc30U,
    0xe9107608U,
    0xbe9e0d14U,
    0xdbca1992U,
    0x3db60b7dU,
    0x6a2e8009U,
    0x14604017U,
    0x2c10a038U,
    0x211eb03aU,
    0x4108282fU,
    0xea11fc1cU,
    0xc185608U,
    0xba89fd1eU,
    0xbed4118bU,
    0xe2bd075cU,
    0xf1385e1cU,
    0xc1fff134U,
    0xe15a4faaU,
    0x93ebf659U,
    0xc1449193U,
    0x13604754U,
    0xe0907e24U,
    0x41580114U,
    0x3ec47b0U,
    0xa944fa56U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0x98000000U,
    0x4c000000U,
    0xde000000U,
    0xe1000000U,
    0xbf800000U,
    0x36c00000U,
    0x7ca00000U,
    0x2cd00000U,
    0xea880000U,
    0xc3a40000U,
    0xb2560000U,
    0xc3cd0000U,
    0x39c18000U,
    0x7a22c000U,
    0x49976000U,
    0xceeb7000U,
    0xead7c800U,
    0x87899c00U,
    0x2264600U,
    0xd5976d00U,
    0x98ed5a80U,
    0x6fd50240U,
    0x3209c820U,
    0x47e09c30U,
    0x29b1c628U,
    0xdc78ad2cU,
    0x70bbbaa6U,
    0x971cb253U,
    0x4f496017U,
    0xef827008U,
    0x7ec04807U,
    0xb8a65c21U,
    0x96d0a639U,
    0x18edd18U,
    0xf25f28dU,
    0x413ee48U,
    0xd72e4623U,
    0xe0f36d0dU,
    0xf61b5ab9U,
    0x30c80279U,
    0x7940482eU,
    0xb2665c20U,
    0xc70a609U,
    0x305edd05U,
    0x1c2df287U,
    0x5c77ee75U,
    0x7858463cU,
    0xd82e6d2bU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x68000000U,
    0xfc000000U,
    0x46000000U,
    0x7f000000U,
    0x68800000U,
    0xcfc00000U,
    0x3ea00000U,
    0xbc900000U,
    0x6d980000U,
    0x67a40000U,
    0xfb160000U,
    0x32d90000U,
    0xb2c18000U,
    0xab254000U,
    0x26d6e000U,
    0xe97fb000U,
    0xefd4b800U,
    0x76faa400U,
    0x24914e00U,
    0x19998d00U,
    0x4da0e080U,
    0x1a1172c0U,
    0xb15ab820U,
    0xaf87a430U,
    0x6346ce08U,
    0xb365cd2cU,
    0x9f37809aU,
    0x148b82ffU,
    0x6878e031U,
    0x3c52b02fU,
    0x93bb3812U,
    0xc8f2e41fU,
    0x5e282e15U,
    0xfeee7d10U,
    0xb0cd388aU,
    0x779c26c6U,
    0x2ea62e04U,
    0xa4937d0fU,
    0xd99ab8abU,
    0x6da066e9U,
    0xaa114e28U,
    0xd9598d2cU,
    0x5380e09cU,
    0x254172e2U,
    0xcc62b805U,
    0xf7b3a410U,
    0xdb48ce0cU,
    0x56d8cd20U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0xe8000000U,
    0x1c000000U,
    0x16000000U,
    0xa3000000U,
    0x28800000U,
    0x8cc00000U,
    0xb4a00000U,
    0xc8f00000U,
    0xc7480000U,
    0xd1a40000U,
    0x8b760000U,
    0xf0f0000U,
    0x87c08000U,
    0xa0214000U,
    0x2237e000U,
    0xa8ec3000U,
    0xc5d65800U,
    0xcaffa400U,
    0xdf0fd600U,
    0x4fc63d00U,
    0xcc214380U,
    0x4c313c40U,
    0xffe85820U,
    0xe754a430U,
    0xf3395638U,
    0xe0287d24U,
    0x237623baU,
    0x330c4c47U,
    0xe1c1e025U,
    0x7b233018U,
    0xfeb6d832U,
    0x2e2ee407U,
    0xc4703617U,
    0x898e0d35U,
    0xbc011bb4U,
    0xa601985cU,
    0x3b070e08U,
    0x4c83d910U,
    0x6ec7f5b4U,
    0x1da1717eU,
    0x557623a3U,
    0xc00c4c69U,
    0xc141e00eU,
    0x7be3300dU,
    0xb416d81eU,
    0x59dee43bU,
    0x3db83629U,
    0x77ea0d24U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xc8000000U,
    0xc000000U,
    0x4e000000U,
    0xc3000000U,
    0xf5800000U,
    0xb400000U,
    0x68e00000U,
    0xfb500000U,
    0x99180000U,
    0xa9e40000U,
    0xa3d60000U,
    0xf4df0000U,
    0x47c18000U,
    0xff204000U,
    0xbc776000U,
    0x16ff000U,
    0x5a8c5800U,
    0xbc5dcc00U,
    0x72072a00U,
    0x8501df00U,
    0xda84b680U,
    0xc732c0U,
    0x68a25820U,
    0x6a36cc30U,
    0x2708aa38U,
    0x9b1a9f2cU,
    0x4e456b2U,
    0xc55782c3U,
    0x7218e033U,
    0xe064b000U,
    0x2e94b805U,
    0x53397c2eU,
    0x7139228U,
    0xae78a33dU,
    0x88f724b5U,
    0x52af91daU,
    0x5fad7cb5U,
    0xa92d5dcfU,
    0xb7ebd693U,
    0x644cc2ddU,
    0x3ff80023U,
    0x51b40035U,
    0x2f4e0027U,
    0xe67b000dU,
    0x44f78011U,
    0xfcaf402fU,
    0x2caee02dU,
    0x94abb00cU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0x68000000U,
    0x14000000U,
    0x86000000U,
    0x8f000000U,
    0x3a800000U,
    0x7e400000U,
    0xa0200000U,
    0xe5f00000U,
    0xb1f80000U,
    0xeaa40000U,
    0x33b60000U,
    0xe5df0000U,
    0x79538000U,
    0x654d4000U,
    0x21fba000U,
    0xb2a1b000U,
    0xafb63800U,
    0x87d9f400U,
    0x18570200U,
    0xc4cc9900U,
    0xe3396b80U,
    0xe3c3b5c0U,
    0xd0e5b820U,
    0xad94b430U,
    0xe32ca238U,
    0xf82d293cU,
    0x84af539aU,
    0x95ea41c5U,
    0xff4aba01U,
    0xcfc2d13U,
    0x923c9b6U,
    0x80719ce3U,
    0x8d396b92U,
    0xb8c3b5ccU,
    0x8c65b815U,
    0xacd4b415U,
    0x118ca220U,
    0x779d291fU,
    0x1377538dU,
    0x15be41c6U,
    0x4784ba2bU,
    0x7dc72d1aU,
    0xe3e649b9U,
    0xe513dcc2U,
    0x64694ba6U,
    0x858b45deU,
    0x319e2029U,
    0x7c73f038U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0x48000000U,
    0x4c000000U,
    0x2a000000U,
    0xdd000000U,
    0x49800000U,
    0xbec00000U,
    0x55600000U,
    0x8a300000U,
    0x12680000U,
    0xf4e40000U,
    0xc8f60000U,
    0x550f0000U,
    0x37d08000U,
    0xbd1c4000U,
    0xa4296000U,
    0x14c7d000U,
    0x4865a800U,
    0x23b3ac00U,
    0x3caa6600U,
    0xe9806300U,
    0xec32d80U,
    0x6d65dd40U,
    0x1e352820U,
    0x3c6fec30U,
    0x4fe30638U,
    0x7677b324U,
    0x7f4e8592U,
    0x95f27153U,
    0xdc894e2aU,
    0x69108f07U,
    0x7878abaaU,
    0x661a2e4bU,
    0x4aaccda7U,
    0x96814d41U,
    0x36416016U,
    0x2023d01eU,
    0x6093a80aU,
    0xe6bcac0dU,
    0x437ae600U,
    0x189c2329U,
    0x80ea4db5U,
    0xa4a20d60U,
    0x1fd08007U,
    0xa11c4020U,
    0x2629602bU,
    0x15c7d02eU,
    0x63e5a80eU,
    0xc73ac25U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0xa8000000U,
    0xbc000000U,
    0x6a000000U,
    0x6d000000U,
    0x95800000U,
    0xc9c00000U,
    0x90200000U,
    0x4b900000U,
    0xf8c80000U,
    0x1da40000U,
    0x86520000U,
    0x76eb0000U,
    0xbd368000U,
    0x441f4000U,
    0xe60ae000U,
    0x9d07d000U,
    0x4d872800U,
    0xedc0cc00U,
    0xbe22de00U,
    0x58974900U,
    0xd6484080U,
    0x46e5b040U,
    0x2731a820U,
    0xf11f8c10U,
    0x57883e18U,
    0x7ac09914U,
    0xcea768aaU,
    0x88d17c6fU,
    0x1da9763aU,
    0xe457c50bU,
    0xd7ecfebdU,
    0xdab56966U,
    0xa4d8a0aeU,
    0x9dad606dU,
    0xa452001cU,
    0xb7eb0018U,
    0x8ab68036U,
    0xcdf4014U,
    0x21aae03bU,
    0xce57d037U,
    0xdaef2838U,
    0x1f34cc09U,
    0xc518de33U,
    0xb1884906U,
    0x85c4c09aU,
    0x2225f052U,
    0x297c833U,
    0x43481c3cU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x48000000U,
    0x7c000000U,
    0x5a000000U,
    0x2b000000U,
    0xfc800000U,
    0xa1c00000U,
    0x8ce00000U,
    0xa7900000U,
    0x99a80000U,
    0x8c640000U,
    0x9c560000U,
    0xde4f0000U,
    0x67718000U,
    0xec3c4000U,
    0xa2c8e000U,
    0x6b55000U,
    0x80da1800U,
    0xb55a8400U,
    0xd71a1e00U,
    0x70b99f00U,
    0x730b6880U,
    0x22508b40U,
    0xeb4b9820U,
    0x3ef6c430U,
    0xc7afe38U,
    0x4a68cf2cU,
    0x27877092U,
    0x35450f5fU,
    0x1200636U,
    0x89731b3aU,
    0x21397687U,
    0xcf4d1444U,
    0xc0f6f091U,
    0x19794f46U,
    0x43e8e628U,
    0x3fc64b35U,
    0xe9e36eb2U,
    0x617906cU,
    0x4decee9eU,
    0x42c0d077U,
    0xcc638ebfU,
    0xbc56c05cU,
    0x8e48f691U,
    0x9f715458U,
    0xd83e10b5U,
    0x84cc1f64U,
    0x77b2fe16U,
    0x575ccf3cU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0xf8000000U,
    0xa4000000U,
    0x7e000000U,
    0x71000000U,
    0xc7800000U,
    0xb5c00000U,
    0x54200000U,
    0xa1b00000U,
    0xa4580000U,
    0xe7a40000U,
    0x22720000U,
    0x7d7d0000U,
    0xab938000U,
    0xc4edc000U,
    0xdc7c2000U,
    0x24127000U,
    0x8d2b4800U,
    0xa25d9c00U,
    0x72a01e00U,
    0x3bf4c900U,
    0x49bf6180U,
    0xc034b4c0U,
    0x7498c820U,
    0x52005c10U,
    0x13043e28U,
    0xcc82b93cU,
    0x8d4629beU,
    0xaf6428e9U,
    0x23d3563fU,
    0x9a0d550cU,
    0xc76d7f99U,
    0x7bd7dd1U,
    0xa934298bU,
    0x471928d1U,
    0x91c0d63eU,
    0x6a209509U,
    0x70b15fafU,
    0x93df0de7U,
    0xaa6761beU,
    0xd250b4fcU,
    0xa2cac838U,
    0x7ecd5c2dU,
    0xe4cfbe19U,
    0x4bcb7927U,
    0xd4809abU,
    0x870b58dbU,
    0xc2eb9e0aU,
    0x497d093eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x88000000U,
    0xd4000000U,
    0x46000000U,
    0x63000000U,
    0x1c800000U,
    0x3cc00000U,
    0xb4e00000U,
    0x77f00000U,
    0x58780000U,
    0xa4640000U,
    0x11320000U,
    0x3d990000U,
    0x3a128000U,
    0x820e4000U,
    0xd95ba000U,
    0x92f2b000U,
    0x47fc5800U,
    0x7425c400U,
    0xc1131600U,
    0xa28d5f00U,
    0x879d4b80U,
    0x5b104dc0U,
    0x138ed820U,
    0x9e1b8410U,
    0xc850b608U,
    0xf2bef34U,
    0xb5cb13a2U,
    0x153889f5U,
    0xe7c54e31U,
    0x44659b08U,
    0xa136dd8fU,
    0x459e52fbU,
    0xb610b3afU,
    0x980a39f8U,
    0x2859162fU,
    0xab705f15U,
    0x4bdcba9U,
    0xe0870dc1U,
    0x3ec7f830U,
    0x39e77400U,
    0xc774e3eU,
    0xb63c9b2eU,
    0x9b445d9dU,
    0x88a012f5U,
    0x4dd313b4U,
    0x6e6c89f4U,
    0x7c6f4e2aU,
    0x19689b3fU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xb8000000U,
    0xcc000000U,
    0x1a000000U,
    0x9000000U,
    0x18800000U,
    0x70400000U,
    0x7ca00000U,
    0xc0100000U,
    0x89b80000U,
    0x86240000U,
    0x55560000U,
    0xa79f0000U,
    0x37718000U,
    0xcccbc000U,
    0x47ed6000U,
    0x843f5000U,
    0x3ce08800U,
    0xc4b22400U,
    0x65ac1200U,
    0xa5998d00U,
    0x6277b480U,
    0x264837c0U,
    0xfaa90820U,
    0x8e1de430U,
    0x53b77238U,
    0x3a29dd2cU,
    0x425ebcaeU,
    0x6d15d3f3U,
    0x8f3e7a26U,
    0x71643932U,
    0x5e71cebeU,
    0xa4480ef0U,
    0x6faec691U,
    0x849aeaf3U,
    0x5ef034bcU,
    0x880cf7ffU,
    0xe10de825U,
    0x898d741aU,
    0xa1cc9a02U,
    0xa16ba91eU,
    0xc37ba69dU,
    0x8fc1baedU,
    0xb66bca5U,
    0x2771d3c6U,
    0xe4c87a37U,
    0x63eb390bU,
    0x5a384eb7U,
    0x5be7cecfU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0xb8000000U,
    0x84000000U,
    0xb2000000U,
    0x29000000U,
    0xcf800000U,
    0xd400000U,
    0x57200000U,
    0x45900000U,
    0x2e880000U,
    0xb2a40000U,
    0x55d60000U,
    0x1c2d0000U,
    0x27778000U,
    0xe9fbc000U,
    0x4b5d2000U,
    0x768c1000U,
    0x26a38800U,
    0x8fd2dc00U,
    0x792eba00U,
    0x66f59100U,
    0xfbbe2080U,
    0x48f8e640U,
    0xd8dc0820U,
    0x9dcd1c30U,
    0x22859a28U,
    0x10c4811cU,
    0xa36228aeU,
    0x1535fa61U,
    0x5a59920cU,
    0xfd099d3aU,
    0x39e7b29bU,
    0x81f17b5fU,
    0x4b3bba9bU,
    0xc13c6740U,
    0xac3e20afU,
    0x71b8e64aU,
    0x25fc0800U,
    0x55d1c09U,
    0xc98d9a36U,
    0x220811cU,
    0xdc1428a8U,
    0x68c8fa59U,
    0xcb061236U,
    0xee865d0cU,
    0x5ec49292U,
    0x1c646b41U,
    0x31b1b2a9U,
    0x99c7b64U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x8000000U,
    0x34000000U,
    0x3a000000U,
    0xd000000U,
    0x7f800000U,
    0xbb400000U,
    0x9a00000U,
    0xd8700000U,
    0x37080000U,
    0xb8240000U,
    0x74320000U,
    0x5c2b0000U,
    0xdc978000U,
    0xcdda4000U,
    0x3bdea000U,
    0x70dc5000U,
    0x6c5d2800U,
    0xef9cdc00U,
    0x27faca00U,
    0x6ef4500U,
    0x61738f80U,
    0x328eba40U,
    0x6e62a820U,
    0x72129c10U,
    0x371e6a18U,
    0xd63c151cU,
    0x978b2782U,
    0x75e3264dU,
    0xeb51422eU,
    0x7fbbc913U,
    0x77ce6d87U,
    0x18c22372U,
    0x8ce66da0U,
    0x8ed6236bU,
    0xd9fc6dbbU,
    0x79e92361U,
    0x47f1edb8U,
    0xbc4c6378U,
    0xd982cd81U,
    0xe8457347U,
    0x5924c5a8U,
    0x5db4bf5eU,
    0xfc6a07afU,
    0xd1b13676U,
    0xda68ca1aU,
    0xc2b44506U,
    0xeaec0f8fU,
    0x3770fa63U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x68000000U,
    0x3c000000U,
    0xba000000U,
    0xbf000000U,
    0x9a800000U,
    0xd9400000U,
    0x5c600000U,
    0x41100000U,
    0x5f680000U,
    0xa8e40000U,
    0x61560000U,
    0xfc8b0000U,
    0x65378000U,
    0x499d4000U,
    0xac5ee000U,
    0xfc7c5000U,
    0xbc8fb800U,
    0xc537b400U,
    0x599eda00U,
    0xb4593b00U,
    0xa8782c80U,
    0x3a8fb940U,
    0xc0303820U,
    0x7c1ef430U,
    0xf79e3a18U,
    0x2d5a6b1cU,
    0x27fe149aU,
    0xde4a4d4fU,
    0x8b91820eU,
    0x3e2ddf1fU,
    0xb080cebeU,
    0xbe43766aU,
    0xf2e1aeadU,
    0x6e56666fU,
    0x6e0ef681U,
    0xf0728269U,
    0xc7fe149cU,
    0x6e4a4d6aU,
    0x83918202U,
    0x722ddf0eU,
    0x6280ceb6U,
    0x3d43765fU,
    0xd261ae92U,
    0x8166678U,
    0xa8eef6bfU,
    0x68228270U,
    0xc4f6149dU,
    0x87be4d61U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x28000000U,
    0x4c000000U,
    0xda000000U,
    0xd3000000U,
    0xad800000U,
    0xd1c00000U,
    0x6fa00000U,
    0xb8300000U,
    0xf1b80000U,
    0x9c240000U,
    0x9cf60000U,
    0x629d0000U,
    0x3f528000U,
    0xc0af4000U,
    0xbdeb6000U,
    0x210a5000U,
    0xd8d8d800U,
    0x60368c00U,
    0x85bc9e00U,
    0x82214500U,
    0x9f5dd80U,
    0xee1f6bc0U,
    0xdc125820U,
    0x94dcc30U,
    0x68b9fe28U,
    0xaaa21534U,
    0xdcb1858aU,
    0xfeffa7d3U,
    0x5fc12616U,
    0xb2a49904U,
    0x8b51b83U,
    0x30fae2c0U,
    0xe2c2fbb1U,
    0x7226f2cdU,
    0x31f5c382U,
    0x2a186ed7U,
    0x9a10658eU,
    0x284eb7fbU,
    0x5a3c9e20U,
    0xce1453dU,
    0xb1d5ddacU,
    0x84ef6bc8U,
    0xc78a583dU,
    0xb099cc0dU,
    0xb057fe3cU,
    0x3f2b152dU,
    0x232d058cU,
    0x1129e7c5U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0xd8000000U,
    0xc000000U,
    0xee000000U,
    0x9d000000U,
    0xf2800000U,
    0x18c00000U,
    0xd4e00000U,
    0xf3700000U,
    0x11e80000U,
    0xdb640000U,
    0x29360000U,
    0xf5c90000U,
    0xf8f28000U,
    0xf1a8c000U,
    0x40456000U,
    0x8b27f000U,
    0x5a132800U,
    0x73d97400U,
    0xbd2d7600U,
    0xdd802f00U,
    0x81407f80U,
    0x8fa32140U,
    0xf3d7a820U,
    0x6bb8b430U,
    0x139a9608U,
    0xe60f1f34U,
    0x5f9637b6U,
    0xfe9da543U,
    0xdc89f61bU,
    0xb351ef17U,
    0x187f9fb4U,
    0x84f81172U,
    0xc13fe023U,
    0xc75b300fU,
    0xfce84837U,
    0xe1e38425U,
    0xc5f2de00U,
    0x132c9b38U,
    0xa084e9b0U,
    0x83c13e67U,
    0x9f651f88U,
    0xf334d162U,
    0x7acc8001U,
    0x7175c019U,
    0x52e9e02bU,
    0x9ce2300aU,
    0xc772c837U,
    0x3ef4418U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x8000000U,
    0x5c000000U,
    0xf6000000U,
    0x69000000U,
    0xf800000U,
    0x59c00000U,
    0x22600000U,
    0x6ef00000U,
    0x73280000U,
    0xb4e40000U,
    0x10b60000U,
    0x48f0000U,
    0xb6718000U,
    0x306e4000U,
    0x66c1e000U,
    0xd4e43000U,
    0x60b40800U,
    0xec8f1c00U,
    0x5a713e00U,
    0xce6ebb00U,
    0x53c7bb80U,
    0x2d619f40U,
    0x50738820U,
    0xc16e5c30U,
    0x6d415e38U,
    0x9f24cb2cU,
    0x89d25382U,
    0x4efab357U,
    0xe01ebe1dU,
    0x67abfb2aU,
    0x59a1dbbbU,
    0x4f94ef7aU,
    0x6d5fe02aU,
    0x488f303cU,
    0xd8738819U,
    0x5d6e5c1bU,
    0x7b415e05U,
    0x4624cb00U,
    0x8e525398U,
    0x4b3ab34dU,
    0x347ebe26U,
    0x605bfb3eU,
    0x2509dbacU,
    0xa2b0ef4dU,
    0x5f89e005U,
    0x22f03035U,
    0x1d2a080cU,
    0xd9e41c09U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xd8000000U,
    0x2c000000U,
    0xfa000000U,
    0x3000000U,
    0xce800000U,
    0x31c00000U,
    0xff600000U,
    0xb4f00000U,
    0x5c080000U,
    0xae40000U,
    0xf7b60000U,
    0xd0ab0000U,
    0x3e768000U,
    0xf3c8c000U,
    0x98866000U,
    0x8c35000U,
    0x52e5d800U,
    0x1bb59400U,
    0x4aa89e00U,
    0x6d728300U,
    0xe54a0380U,
    0x85447640U,
    0xda55820U,
    0xe5165430U,
    0x89387e18U,
    0x71891314U,
    0x65a1bbb6U,
    0x8116b24bU,
    0xe73e1e1eU,
    0xf48a4330U,
    0x8a2463abU,
    0x51532658U,
    0x1d1e8029U,
    0x8ddcc016U,
    0x53386031U,
    0x828c5016U,
    0x43255838U,
    0x14d65433U,
    0x16587e0eU,
    0x95791311U,
    0xe1a9bb82U,
    0xa7f2b24cU,
    0xea881e36U,
    0x2721432fU,
    0x7ad2e392U,
    0x935be671U,
    0x7af8e034U,
    0x31ef902cU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x78000000U,
    0xc4000000U,
    0xf2000000U,
    0x9f000000U,
    0x74800000U,
    0x87c00000U,
    0x18200000U,
    0xfa700000U,
    0xa080000U,
    0x3ba40000U,
    0x75360000U,
    0xc3eb0000U,
    0x84f58000U,
    0x16cdc000U,
    0xc5012000U,
    0xa7801000U,
    0xdd443800U,
    0x6e40400U,
    0x83d06e00U,
    0x583a1900U,
    0x28cfe780U,
    0xb4029640U,
    0x2a07b820U,
    0xb02c430U,
    0xce84ce18U,
    0xa4c7c90cU,
    0x5aa2ff9eU,
    0x10b28271U,
    0xf9adee1cU,
    0x3293d917U,
    0xe0d8c785U,
    0x1a19866dU,
    0x1be0038U,
    0xe48f003fU,
    0xf638026U,
    0x796c025U,
    0x5f5ca01eU,
    0x3359d020U,
    0x8d5b181bU,
    0x3c5b141cU,
    0xd1dfd62eU,
    0x379cdd06U,
    0x87fd29acU,
    0x64ee5f4fU,
    0x9470c7a3U,
    0x330d866dU,
    0xa20001eU,
    0x95700008U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x48000000U,
    0xbc000000U,
    0xae000000U,
    0x1000000U,
    0x19800000U,
    0xfcc00000U,
    0xc7600000U,
    0x8bf00000U,
    0xca180000U,
    0x67e40000U,
    0xeab20000U,
    0x9bbb0000U,
    0x20768000U,
    0xe9d84000U,
    0xc002e000U,
    0x2005f000U,
    0x30014800U,
    0x1807ac00U,
    0xf400d600U,
    0x1203d300U,
    0xaf01e380U,
    0x1881bec0U,
    0xe545c820U,
    0x3ba4ec10U,
    0x4c94b618U,
    0x41ee6314U,
    0xadfa4b92U,
    0x8d57e2efU,
    0x710e560bU,
    0xbbcf9310U,
    0xc9a9039eU,
    0x29db4eebU,
    0xe0000003U,
    0x1000001dU,
    0x28000001U,
    0xec00000dU,
    0xe600002eU,
    0xbd000036U,
    0xb7800012U,
    0xfdc00018U,
    0xdee00004U,
    0x7730002eU,
    0xd780007U,
    0xec140002U,
    0x20aa0007U,
    0xfc5f0027U,
    0xcac48021U,
    0x72634028U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0x8000000U,
    0x84000000U,
    0xae000000U,
    0x89000000U,
    0xea800000U,
    0xbb400000U,
    0xb0a00000U,
    0x32500000U,
    0xd5080000U,
    0x67240000U,
    0x75960000U,
    0xabef0000U,
    0xbd508000U,
    0x828e4000U,
    0xe0e72000U,
    0xeb77f000U,
    0xe19b0800U,
    0x424bf400U,
    0x4d82ea00U,
    0x18c4ed00U,
    0x1603680U,
    0xa9b3fa40U,
    0x5ffd8820U,
    0x47ab430U,
    0x69bd4a38U,
    0x30595d24U,
    0x89aa1e82U,
    0x2f70fe61U,
    0x6f9cea0bU,
    0xbb4fed12U,
    0x3f06b682U,
    0x2f82ba4aU,
    0x9bc2280eU,
    0xbce7041dU,
    0xe9776226U,
    0x329a591fU,
    0x17cb7ca5U,
    0x1945a765U,
    0xd3a71688U,
    0x1fd40a7eU,
    0x9dce802fU,
    0x8e454019U,
    0xc821a033U,
    0x5216b033U,
    0xb42ca820U,
    0xafb2443cU,
    0xe2fec234U,
    0x38f8e936U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x98000000U,
    0xb4000000U,
    0xbe000000U,
    0xe3000000U,
    0xe800000U,
    0xa9400000U,
    0x9f600000U,
    0xd4300000U,
    0x8b180000U,
    0x2ee40000U,
    0xf1f60000U,
    0xec390000U,
    0x52328000U,
    0xcc1ac000U,
    0x2660a000U,
    0xdfb57000U,
    0x5bdad800U,
    0x78475400U,
    0x68e27a00U,
    0x96f74300U,
    0x34bda580U,
    0xe476cac0U,
    0xcffe5820U,
    0xb2149430U,
    0x98485a08U,
    0xe5ccf334U,
    0x3a09dda6U,
    0x4829eeedU,
    0xf57a7a0fU,
    0x65534308U,
    0xc42ba58bU,
    0xef7fcadeU,
    0x8054d821U,
    0x4daa5428U,
    0x6ebefa05U,
    0xe1708307U,
    0xb6798599U,
    0x7bd07afcU,
    0xd56ea03cU,
    0x1418700eU,
    0x72665820U,
    0x91b09422U,
    0xf0de5a2bU,
    0x5ac5f33dU,
    0xcba35d83U,
    0x54972ee5U,
    0xd0cda1bU,
    0xc8af332fU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x38000000U,
    0xfc000000U,
    0x42000000U,
    0x47000000U,
    0xdd800000U,
    0x8c400000U,
    0xa3e00000U,
    0xbfd00000U,
    0xd8f80000U,
    0x1b640000U,
    0xb9120000U,
    0xf2df0000U,
    0x27168000U,
    0x7dbc000U,
    0x3594e000U,
    0xc21a9000U,
    0x9c30f800U,
    0x27298c00U,
    0x239cb200U,
    0xf2744700U,
    0x73cdc280U,
    0x29cefdc0U,
    0x62cc7820U,
    0x75494c10U,
    0x2a0cd238U,
    0xd26a1714U,
    0x18ff5a8eU,
    0xbb6621ffU,
    0x914d230U,
    0x9ade1701U,
    0xe3155a8fU,
    0xb9dd21f7U,
    0x30905206U,
    0x589ad700U,
    0xcdf73a9eU,
    0x88c71c3U,
    0x3fac4a37U,
    0x955dcb30U,
    0xb0517089U,
    0x8bbabaebU,
    0x2901baa2U,
    0xa087b1faU,
    0xac0aa29U,
    0xe0235b0fU,
    0xef7388b6U,
    0xe54c36f3U,
    0xb20b88a9U,
    0x9e6836d7U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x48000000U,
    0x94000000U,
    0x2e000000U,
    0x45000000U,
    0x6800000U,
    0x67c00000U,
    0xb8e00000U,
    0x4ad00000U,
    0x1ae80000U,
    0xe5640000U,
    0x96920000U,
    0xef490000U,
    0x52928000U,
    0xc9484000U,
    0xa391a000U,
    0x31cdf000U,
    0xc9556800U,
    0x1ba81400U,
    0xca450e00U,
    0xfca7cf00U,
    0x63705280U,
    0x171be140U,
    0xd1bde820U,
    0x99cd5410U,
    0xed542e08U,
    0x8dab7f34U,
    0x83461a92U,
    0x8264565U,
    0xfbb42e2bU,
    0xc27b7f01U,
    0xbf2e1a89U,
    0x5a82456dU,
    0x9dc62e1cU,
    0xf3e27f27U,
    0xd9549a85U,
    0x33ae055cU,
    0xae458e16U,
    0x4aa68f17U,
    0xfa7372b9U,
    0xab9e5145U,
    0xdd79201dU,
    0x62a8b019U,
    0xc6c4483aU,
    0x4864a40aU,
    0x94134633U,
    0xce8a6b08U,
    0xeb7194b0U,
    0xe319ca71U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0xd8000000U,
    0x54000000U,
    0x1a000000U,
    0xe9000000U,
    0x34800000U,
    0xc2400000U,
    0x2b200000U,
    0x98b00000U,
    0x22880000U,
    0x14a40000U,
    0xa3760000U,
    0x8d6b0000U,
    0xd0728000U,
    0x90e8c000U,
    0x46b5a000U,
    0xa9885000U,
    0x2d268800U,
    0x47b38400U,
    0x10b0200U,
    0x19663100U,
    0x1b957180U,
    0x8639ee40U,
    0x702a0820U,
    0xc8d44430U,
    0xbd9a2218U,
    0x58dda134U,
    0x747cd9b6U,
    0x930efa55U,
    0xbc642226U,
    0x112a10aU,
    0x67785995U,
    0x5e8d3a44U,
    0xf2a3021cU,
    0x6c723103U,
    0x16eb7196U,
    0x59b6ee4bU,
    0xea0e8833U,
    0xf0e78436U,
    0x27550220U,
    0x5e59310cU,
    0xa339f1a2U,
    0xaaae2e4aU,
    0x4993280fU,
    0x433bd41eU,
    0xbaad8a29U,
    0xf195b523U,
    0xc73e7383U,
    0x78afdf5bU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0xb8000000U,
    0x34000000U,
    0x2a000000U,
    0x93000000U,
    0x6e800000U,
    0x89400000U,
    0x9200000U,
    0xaa900000U,
    0xc4380000U,
    0x9aa40000U,
    0xec520000U,
    0xdddf0000U,
    0xfcd18000U,
    0x1d9fc000U,
    0x5072a000U,
    0xd4ced000U,
    0x862a5800U,
    0x4ad95400U,
    0x4051da00U,
    0xd3ddf100U,
    0xadd43080U,
    0xdc1d13c0U,
    0x11b1d820U,
    0xd9ad9410U,
    0xf218fa38U,
    0x50b7e12cU,
    0xf02f48aeU,
    0x8fda57cdU,
    0xfbd17a2aU,
    0x91c2134U,
    0x4237e8a3U,
    0x3f6f87ceU,
    0xbbf8a20cU,
    0x3ec5b537U,
    0x8de51283U,
    0x1c7366eeU,
    0x6acc6aaeU,
    0x6f2b22c8U,
    0xbf5d48afU,
    0x2b9557f5U,
    0x4db8fa08U,
    0xb767e100U,
    0x4f374884U,
    0xa8ee57caU,
    0x2f3b7a01U,
    0x6027210eU,
    0x1f1468afU,
    0x4fff47f5U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xc8000000U,
    0x6c000000U,
    0x1e000000U,
    0x2d000000U,
    0x71800000U,
    0x9c00000U,
    0x9600000U,
    0x4a700000U,
    0x6f580000U,
    0xc0a40000U,
    0xe3120000U,
    0xb52f0000U,
    0x17ff8000U,
    0x87b04000U,
    0x243e6000U,
    0x91d07000U,
    0xcccea800U,
    0xdbcdac00U,
    0xb54e5e00U,
    0x1e0e9300U,
    0x32295780U,
    0xc17c88c0U,
    0x88f0c820U,
    0x7a1ddc10U,
    0x5180f638U,
    0xd9c33f1cU,
    0x516709b2U,
    0x9e721bdbU,
    0xd5599fa7U,
    0x9fa154dbU,
    0xa1903e04U,
    0xe06ee30eU,
    0x66dfff88U,
    0xcd6524d5U,
    0x874960eU,
    0x74584f30U,
    0x8024219bU,
    0x4d50f7f8U,
    0xf48e2189U,
    0x3b6bf7faU,
    0x865ba18fU,
    0xf320b7c1U,
    0xf1d041a6U,
    0xfccb87feU,
    0xf3cd09bfU,
    0xa9491bc5U,
    0xc80c1faaU,
    0x732a14dbU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0x68000000U,
    0xc4000000U,
    0xc6000000U,
    0xa7000000U,
    0x71800000U,
    0x3400000U,
    0x2c200000U,
    0x41700000U,
    0xfc580000U,
    0xf640000U,
    0x5d520000U,
    0x752d0000U,
    0x473f8000U,
    0xfc354000U,
    0x4b792000U,
    0x8594b000U,
    0x1f4ff800U,
    0xeaefec00U,
    0x8cdd8a00U,
    0x4aa22900U,
    0xbbb61f80U,
    0xc73f84c0U,
    0xbc36d820U,
    0xeb7b5c10U,
    0xf5927228U,
    0x774dc51cU,
    0x2eeb959aU,
    0x4addadf1U,
    0xeda0c791U,
    0xca34d8f9U,
    0xc47caa14U,
    0x9012990cU,
    0xaa0be7b9U,
    0x9cd68fdU,
    0x782cd234U,
    0x73b8350bU,
    0x3ff74d92U,
    0xaa9ff1e8U,
    0x360735bcU,
    0x8f055dceU,
    0x15839f9fU,
    0xb543c4d7U,
    0xe3227821U,
    0xf4f7ac19U,
    0x391b2a2dU,
    0x8443d924U,
    0x6da0c7a5U,
    0x8a34d8f4U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0xf8000000U,
    0x24000000U,
    0x36000000U,
    0x7000000U,
    0xe6800000U,
    0xb8400000U,
    0xc2e00000U,
    0xcab00000U,
    0x1a80000U,
    0x2a40000U,
    0x6c520000U,
    0x9d1b0000U,
    0xd40a8000U,
    0x7071c000U,
    0x6d082000U,
    0xbdf03000U,
    0x69cb7800U,
    0xfa56d400U,
    0x6a19d200U,
    0x7a8f5f00U,
    0xc430f980U,
    0x45ed8b40U,
    0x62435820U,
    0xbfe6e410U,
    0xa132aa18U,
    0x9a698b34U,
    0x57812bbeU,
    0x19c6d449U,
    0xfc21a1adU,
    0x60106f51U,
    0xf1fb7201U,
    0xedbeaf0aU,
    0x595ba196U,
    0xf4ef6f4fU,
    0xc3c3f213U,
    0x81246f18U,
    0xb9101b7U,
    0x6a3a9f7cU,
    0xb89aaa22U,
    0x2ccd8b19U,
    0x95d32bbeU,
    0x77ddd468U,
    0xab21a6U,
    0x8b21af6eU,
    0x8e935231U,
    0x25be9f32U,
    0x1558d980U,
    0x7eedbb71U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0xa8000000U,
    0xa4000000U,
    0xc2000000U,
    0xa5000000U,
    0x55800000U,
    0xcd400000U,
    0x49600000U,
    0x85900000U,
    0x47a80000U,
    0x90240000U,
    0x76f60000U,
    0x4b390000U,
    0xf40c8000U,
    0xe4974000U,
    0x202a2000U,
    0x60639000U,
    0x7616e800U,
    0xcdebcc00U,
    0x39c1da00U,
    0xc3a23300U,
    0x2cb78880U,
    0x6ad86b40U,
    0xf55cc820U,
    0xe9185c30U,
    0xbfff3208U,
    0xe62dff04U,
    0xd76052aaU,
    0xe935869U,
    0xe12fc090U,
    0xd7e37759U,
    0xe57da3dU,
    0x790b3307U,
    0xd51308b0U,
    0xf6b2b4cU,
    0x9c00e82bU,
    0xe02cc14U,
    0xe3055a30U,
    0x2281730cU,
    0x95c3a890U,
    0x75a6fb42U,
    0xc3b0a01bU,
    0x2e5dd004U,
    0x798480cU,
    0x6c3b1c32U,
    0xe48b1216U,
    0x4c536f13U,
    0x1c0c3a9fU,
    0xa096d476U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x88000000U,
    0xa4000000U,
    0x3a000000U,
    0xbb000000U,
    0x2b800000U,
    0x95c00000U,
    0xb7600000U,
    0x57300000U,
    0xb9580000U,
    0xdea40000U,
    0xa6560000U,
    0xd3690000U,
    0x51798000U,
    0x6bb64000U,
    0x591a2000U,
    0x5d015000U,
    0x66821800U,
    0xb413c00U,
    0xb025de00U,
    0xc915f300U,
    0x794b1d80U,
    0x936de8c0U,
    0xb1783820U,
    0xfbb06c30U,
    0x611fc608U,
    0x7100cf2cU,
    0xf880c3a2U,
    0x8a451be9U,
    0x20a4a5aeU,
    0x7756c4deU,
    0x5bea5e22U,
    0x733ab339U,
    0x5f10bd87U,
    0x9c4ef8d0U,
    0x19ee0022U,
    0x43d0000U,
    0x7a978025U,
    0xb08b4013U,
    0x120da011U,
    0x734a1003U,
    0x606fb81bU,
    0x1efb2c1bU,
    0xe4726614U,
    0xe57adf24U,
    0xa9b77bbcU,
    0xee1a37c9U,
    0x5880c392U,
    0xfa451beeU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0x48000000U,
    0xc4000000U,
    0xb2000000U,
    0xcd000000U,
    0xcc800000U,
    0x71400000U,
    0x7a00000U,
    0x78300000U,
    0x27480000U,
    0x94e40000U,
    0x9a920000U,
    0xe7f90000U,
    0xa8e88000U,
    0x48d44000U,
    0xbfdd2000U,
    0x661c1000U,
    0x72fea800U,
    0xa86d0400U,
    0x6f93ba00U,
    0x577e0500U,
    0x57ae2d80U,
    0x2c77a440U,
    0xf86b8820U,
    0x47951410U,
    0x237f1208U,
    0x3daa0124U,
    0x6d751792U,
    0x42ede171U,
    0x49d0858cU,
    0x255aa063U,
    0xf858321bU,
    0x3cdb1128U,
    0xa9993f9bU,
    0x8c39a54bU,
    0x710c9f97U,
    0x5ec1f577U,
    0x9671791U,
    0x5954e162U,
    0xa1805a2U,
    0xf4fee055U,
    0x2f6d1231U,
    0x72130114U,
    0xd4bd97b6U,
    0x9a49a14cU,
    0xe065a5bbU,
    0xd7d2b06eU,
    0x4e5c9a2fU,
    0x23db1520U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x48000000U,
    0xac000000U,
    0xb2000000U,
    0x8f000000U,
    0xd7800000U,
    0x14c00000U,
    0x9b600000U,
    0x70300000U,
    0x70780000U,
    0x35a40000U,
    0xe8560000U,
    0xb5cf0000U,
    0xa61d8000U,
    0x75104000U,
    0x276c6000U,
    0xc5cf1000U,
    0xce1db800U,
    0x9149c00U,
    0xed68d600U,
    0xaecdf300U,
    0x79de580U,
    0x20d229c0U,
    0x2e89d820U,
    0x1dbf8c30U,
    0xf8436e38U,
    0xfe266f0cU,
    0xc690b392U,
    0xedab9aebU,
    0x83ae5d8cU,
    0xd6adb5d3U,
    0xf22a8e2dU,
    0xfb6d3f39U,
    0x1fcf6b8cU,
    0x3d1b16fbU,
    0x1490b382U,
    0x92ab9af5U,
    0xfc2e5dbbU,
    0x5e6db5f7U,
    0x934a8e30U,
    0xa85d3f13U,
    0xa376bb5U,
    0x937f16e4U,
    0xb026b3beU,
    0x43949adfU,
    0xb12bdd8fU,
    0x6ee9f5ddU,
    0x2c08ee24U,
    0xedf92f27U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x18000000U,
    0x64000000U,
    0x76000000U,
    0xdb000000U,
    0x15800000U,
    0x37c00000U,
    0x63a00000U,
    0xa5b00000U,
    0xbbf80000U,
    0x6a640000U,
    0xa1160000U,
    0xb9cd0000U,
    0x9b5a8000U,
    0x7e52c000U,
    0xfb282000U,
    0xbb0c3000U,
    0x927a9800U,
    0x2fa44c00U,
    0x1fb4ae00U,
    0x1afd3500U,
    0x7ee23980U,
    0x52529dc0U,
    0x312ab820U,
    0xd20c7c30U,
    0xfaf83628U,
    0xcee47934U,
    0x3a541786U,
    0x9d2968d9U,
    0xd80ea1bdU,
    0x33fbd1c6U,
    0x7664960dU,
    0xc3138909U,
    0xdcca2fb6U,
    0x29ded4c4U,
    0x4e123795U,
    0x24c58c5U,
    0xb398b9b8U,
    0xae705dd1U,
    0xaa5a9835U,
    0x12d44c26U,
    0x1c6cae3bU,
    0x8de93506U,
    0x7c2c39bcU,
    0xac8b9dddU,
    0xb0be3829U,
    0x3687bc15U,
    0xf644961cU,
    0x4a63892fU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0x38000000U,
    0xdc000000U,
    0x82000000U,
    0xcf000000U,
    0x6a800000U,
    0x4dc00000U,
    0xa0e00000U,
    0x4f500000U,
    0xca180000U,
    0xc3240000U,
    0xdab20000U,
    0x5ccd0000U,
    0x5c7b8000U,
    0x3b7c000U,
    0xdf49e000U,
    0x4a3a9000U,
    0x39163800U,
    0xe0bb7400U,
    0x94d61e00U,
    0x105b2f00U,
    0x13817680U,
    0x3e401ac0U,
    0x8ea7d820U,
    0xa9f5e410U,
    0x77ea2628U,
    0xd2c95b3cU,
    0x997ee88eU,
    0xe231f5f7U,
    0x560cce80U,
    0xd21caee3U,
    0xaf202612U,
    0xc0b05b3fU,
    0xbfcf688eU,
    0x8cff35d8U,
    0x5d74aebcU,
    0x9728fee4U,
    0x87ae7e04U,
    0x396f7f1bU,
    0x210f2e83U,
    0x249f3ecbU,
    0xc0e79e23U,
    0x5f55ef25U,
    0xa2191698U,
    0xd7244ac3U,
    0xbcb18018U,
    0xcdcec034U,
    0x7bf86036U,
    0xebf45031U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xa8000000U,
    0xe4000000U,
    0x3e000000U,
    0xb7000000U,
    0xaf800000U,
    0xd6400000U,
    0x71e00000U,
    0x69300000U,
    0xdf980000U,
    0x75a40000U,
    0xe5d60000U,
    0x682f0000U,
    0x98f98000U,
    0xea574000U,
    0x2e6e2000U,
    0x111f7000U,
    0x7f677800U,
    0x83f61c00U,
    0x9bae200U,
    0xbc330700U,
    0x251fed80U,
    0x9960e2c0U,
    0x98f15820U,
    0xcc3d6c30U,
    0x4b739a38U,
    0xa87e1b2cU,
    0xb7928faaU,
    0x57cfa5f9U,
    0xecf15afU,
    0xce4abeddU,
    0x1b8a1a33U,
    0xd6295b29U,
    0xeffcaf8eU,
    0xa5d0d5cfU,
    0x48286db2U,
    0xc8fca2f9U,
    0xf250f805U,
    0x626a5c1eU,
    0xcb1b4203U,
    0xf6643700U,
    0x9b773585U,
    0x707aced7U,
    0x1b94e215U,
    0x3dc8070fU,
    0x2fc86d92U,
    0x32cca2ccU,
    0x5c48f805U,
    0xc68e5c16U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0xd8000000U,
    0x4000000U,
    0x2a000000U,
    0x2f000000U,
    0xc1800000U,
    0x5fc00000U,
    0xc7200000U,
    0x7300000U,
    0x76e80000U,
    0x9ee40000U,
    0x79160000U,
    0xa15d0000U,
    0x1b498000U,
    0x4d94c000U,
    0xa81b2000U,
    0x562dd000U,
    0x90403800U,
    0xede3ec00U,
    0x56930e00U,
    0x5b9d5100U,
    0x82edad80U,
    0xce43540U,
    0x42131820U,
    0x22da3c30U,
    0xb78d3628U,
    0x6537bd2cU,
    0xf5e923b6U,
    0x964a441U,
    0x97d215aaU,
    0x7afe197bU,
    0xa93ab638U,
    0x481a7d0bU,
    0x862d83afU,
    0x5840b46cU,
    0x81e08d81U,
    0xa094e55dU,
    0x5a9aa02cU,
    0x466d1028U,
    0xbda51819U,
    0x1b773c28U,
    0xbd0cb63aU,
    0x1777d39U,
    0x8a0c039dU,
    0x64f07468U,
    0x4fcdadacU,
    0x54d4354bU,
    0x8d7b1828U,
    0x57fe3c3fU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x68000000U,
    0x5c000000U,
    0xae000000U,
    0xad000000U,
    0x5a800000U,
    0x9ac00000U,
    0xcea00000U,
    0xf5b00000U,
    0xbdc80000U,
    0x99640000U,
    0x91960000U,
    0x8ab90000U,
    0xce0c8000U,
    0x67414000U,
    0xf9e5e000U,
    0x8454f000U,
    0x559ea800U,
    0x1aff7400U,
    0x12694600U,
    0x30556d00U,
    0x679f9180U,
    0x39fd2c40U,
    0xf5ed4820U,
    0x98128430U,
    0x6ffb6e08U,
    0xaceb5904U,
    0x1093379aU,
    0xe63cb157U,
    0xfdcc718bU,
    0x7960dc5bU,
    0xa197603eU,
    0xf2b8b012U,
    0xfa09c821U,
    0x9547c42eU,
    0xfae08e1eU,
    0x73d2a91aU,
    0x95df1f99U,
    0x4e9f856fU,
    0x297a57b6U,
    0x78290172U,
    0x4337398aU,
    0x310b584eU,
    0xeec08e2cU,
    0xdca2a921U,
    0xc6b71faaU,
    0x324b854fU,
    0x6d245793U,
    0x37f4015cU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x88000000U,
    0xac000000U,
    0x4a000000U,
    0x3d000000U,
    0x99800000U,
    0x9b400000U,
    0xdf200000U,
    0x3700000U,
    0x3980000U,
    0x31640000U,
    0x49d20000U,
    0x61af0000U,
    0x8de8000U,
    0xc6424000U,
    0x56a06000U,
    0xf037d000U,
    0x20bf2800U,
    0xf011ec00U,
    0xdb49ea00U,
    0x834e4100U,
    0xe749bf80U,
    0xe14d4ac0U,
    0xc64d4820U,
    0xeac93c10U,
    0x44084238U,
    0x4a6ded14U,
    0xab3835a2U,
    0x3250dbebU,
    0x9be9dfb2U,
    0x3c7a9adfU,
    0x1772603eU,
    0x7d98d022U,
    0x8261a82dU,
    0x5b53ac3fU,
    0x9c698a30U,
    0x44399128U,
    0x52d697a6U,
    0x2f2ca6f1U,
    0x871ca233U,
    0xc3a37d07U,
    0x35b3fd9aU,
    0xc9ffa7e4U,
    0x6633fd98U,
    0x2fbfa7d3U,
    0xc093fdacU,
    0xe78fa7c2U,
    0x942bfdb2U,
    0x799ba7fcU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0xc8000000U,
    0xf4000000U,
    0xf6000000U,
    0xf000000U,
    0x76800000U,
    0xa7c00000U,
    0xb8200000U,
    0x1d300000U,
    0x3b980000U,
    0xa8e40000U,
    0x2f960000U,
    0x636f0000U,
    0x225b8000U,
    0x2bc44000U,
    0x5226a000U,
    0x373000U,
    0x8c19b800U,
    0x4aa65400U,
    0x44f32e00U,
    0x98bd0900U,
    0xc550a580U,
    0x74cf16c0U,
    0xf1291820U,
    0x413e6430U,
    0xc4111638U,
    0xb32f1d3cU,
    0x983d2bb2U,
    0xad912ffdU,
    0x5a6e059dU,
    0xbbdc26f3U,
    0xa86a005U,
    0x85c73025U,
    0x5121b824U,
    0x5cb25406U,
    0xd6dd2e01U,
    0xb5060914U,
    0x738525abU,
    0x3c4056f2U,
    0x6062381bU,
    0xb3521414U,
    0xbbcd8e0cU,
    0x67ae3911U,
    0x16ff1d94U,
    0xb43642ddU,
    0x5a19b60fU,
    0x55a32d07U,
    0xa711388U,
    0x3783bc2U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0xe8000000U,
    0xe4000000U,
    0x96000000U,
    0xed000000U,
    0x2a800000U,
    0xaec00000U,
    0x13e00000U,
    0xdf900000U,
    0x95180000U,
    0xdc240000U,
    0xd4f20000U,
    0x7f4f0000U,
    0xcd5e8000U,
    0xbb014000U,
    0x67822000U,
    0x34413000U,
    0x522a800U,
    0xc0713400U,
    0x388ee600U,
    0x323f4700U,
    0xcf56c080U,
    0x2ffd8ac0U,
    0xf328820U,
    0xba2f0410U,
    0x960ace38U,
    0x1afb3314U,
    0x39b006baU,
    0x6ee8fdf9U,
    0x4aea6085U,
    0x7cedfaebU,
    0x21ea0012U,
    0xb36b002fU,
    0x11ac8006U,
    0x704e401aU,
    0xd4dca03aU,
    0x86407035U,
    0xde20881aU,
    0xb7f0040fU,
    0x4cc4e3dU,
    0x831e7332U,
    0x712026b8U,
    0x1e76cdcdU,
    0x818e48b1U,
    0x36b98ec8U,
    0x8094c610U,
    0x649a7739U,
    0x566689bU,
    0xfb53bec5U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0xc8000000U,
    0x4c000000U,
    0x66000000U,
    0x1000000U,
    0x49800000U,
    0x7c00000U,
    0x6fe00000U,
    0xa3900000U,
    0x13880000U,
    0xd240000U,
    0x7ff20000U,
    0x88d90000U,
    0xdc98000U,
    0x7d004000U,
    0xa7816000U,
    0x6ac7b000U,
    0xf063c800U,
    0x6d566c00U,
    0x79e96e00U,
    0xcf744700U,
    0x29d5880U,
    0x6fefee40U,
    0x6670a820U,
    0x9718dc10U,
    0x762b2608U,
    0x8c966b2cU,
    0x570f56b2U,
    0x5611953U,
    0xa6d5b8b9U,
    0xc5281e50U,
    0xf612003aU,
    0x6a49003dU,
    0x77c18021U,
    0xc7e44017U,
    0x7f93602fU,
    0xd8eb010U,
    0x88224814U,
    0x1c722c23U,
    0xe81a0e0aU,
    0x2aaaf736U,
    0x90d710b6U,
    0xdc29c271U,
    0xab90a607U,
    0x3f8f2b2aU,
    0xfb27b68aU,
    0x6f6e962U,
    0x455f10b8U,
    0x200dc25dU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0xe8000000U,
    0x3c000000U,
    0xe6000000U,
    0xaf000000U,
    0xc5800000U,
    0x61c00000U,
    0x8ea00000U,
    0x72300000U,
    0x24e80000U,
    0xa4640000U,
    0x9b160000U,
    0x4a1f0000U,
    0xeeae8000U,
    0x48054000U,
    0x4c00e000U,
    0xbe02b000U,
    0x2b00b800U,
    0xf7808c00U,
    0x14c5de00U,
    0x2263d00U,
    0x7970c580U,
    0xe0c0540U,
    0x39365820U,
    0x436d3c30U,
    0xb8a3e638U,
    0xf537f114U,
    0x3d6bfbbaU,
    0x73a3884fU,
    0x52b6a599U,
    0xc1aff55bU,
    0x4d800029U,
    0xcdc0003cU,
    0x60a00021U,
    0xb1300007U,
    0xef68000aU,
    0x56a4000dU,
    0x3636002eU,
    0xf6ef0021U,
    0x81668029U,
    0xff914016U,
    0x7d5ee039U,
    0x2249b01aU,
    0x7a503825U,
    0x51fecc06U,
    0x2d7dbe18U,
    0xbe3ecd00U,
    0xf0de1db6U,
    0xf8b797eU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0x88000000U,
    0x1c000000U,
    0x16000000U,
    0xc1000000U,
    0x2c800000U,
    0x3cc00000U,
    0x46200000U,
    0xecb00000U,
    0x8e380000U,
    0xede40000U,
    0x27160000U,
    0xa2cd0000U,
    0xe13f8000U,
    0xf4664000U,
    0xed53e000U,
    0xf72e3000U,
    0x322dc800U,
    0x8cafc400U,
    0x236eb600U,
    0x7ac80900U,
    0xc538c980U,
    0xd664d3c0U,
    0x56502820U,
    0x84a8f430U,
    0xff6afe28U,
    0xcccc8d1cU,
    0x743a1fa2U,
    0x72e4aac7U,
    0x7696c9a5U,
    0xd48dd3c0U,
    0xd2d9a803U,
    0x6e73b423U,
    0xa51e9e1bU,
    0x13d0fd20U,
    0x38ea3784U,
    0xbb8c5eccU,
    0xcb5c37a7U,
    0xa4315ee7U,
    0xf0fbb7aaU,
    0xc0c31ee5U,
    0x402657a3U,
    0x75b42ed6U,
    0x46ba1f89U,
    0x5324aaebU,
    0xaa36c99dU,
    0xb5fdd3dfU,
    0xbe41a807U,
    0x4fe7b41aU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x38000000U,
    0xe4000000U,
    0xe000000U,
    0x7d000000U,
    0xc9800000U,
    0xecc00000U,
    0x9da00000U,
    0x19900000U,
    0xfc980000U,
    0x16640000U,
    0xe6b60000U,
    0xc9490000U,
    0xd4198000U,
    0x3923c000U,
    0x74d02000U,
    0x71fcf000U,
    0xc853a800U,
    0x48beac00U,
    0x73353600U,
    0xf8cf500U,
    0x1ab9a880U,
    0x24366ec0U,
    0x950d8820U,
    0xf2ff5c30U,
    0xa4d11e08U,
    0x19fc992cU,
    0x84533e8eU,
    0x2abcabf9U,
    0xdc3628a3U,
    0x510caeefU,
    0x4cfc281aU,
    0xe1d46c27U,
    0x347c9621U,
    0x6693c513U,
    0xca1a20b2U,
    0xc2432e3U,
    0x41531628U,
    0xc7390526U,
    0x1ef380a8U,
    0x1ab02dfU,
    0xa4e89e2bU,
    0x158f5916U,
    0xb1bb1ebfU,
    0xe4b45bd6U,
    0x364b808cU,
    0x229f02cdU,
    0x3669e22U,
    0x6332591eU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0xa8000000U,
    0xfc000000U,
    0xe6000000U,
    0xe9000000U,
    0x73800000U,
    0xe6400000U,
    0xda600000U,
    0xe3300000U,
    0x4eb80000U,
    0x47240000U,
    0x29d20000U,
    0x3f4d0000U,
    0x833a8000U,
    0x8c634000U,
    0x9230e000U,
    0x193cb000U,
    0xc3608800U,
    0xa8b33c00U,
    0x5cfae200U,
    0x57442b00U,
    0x6de48d80U,
    0x17743bc0U,
    0xb7da6820U,
    0x70968c10U,
    0xe7aaea28U,
    0xd1cd5734U,
    0x9cfe0faaU,
    0xb746e0ffU,
    0x1de60d99U,
    0x6f737beaU,
    0xe3d88814U,
    0x6a973c1dU,
    0xe8a8e234U,
    0x4b492b33U,
    0x93e0d80U,
    0x8b677bd4U,
    0x24b2880fU,
    0xc2fe3c1cU,
    0xea406222U,
    0x4676b16U,
    0xfe346dabU,
    0xf7388bd0U,
    0x662e030U,
    0x9531b01eU,
    0xafba0809U,
    0x18a07c34U,
    0x81920203U,
    0xf02c9b27U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x18000000U,
    0x6c000000U,
    0x1e000000U,
    0x5b000000U,
    0xfc800000U,
    0x93c00000U,
    0xb5600000U,
    0xf2900000U,
    0x5da80000U,
    0x2240000U,
    0x58320000U,
    0x145d0000U,
    0x5e198000U,
    0x8f3ec000U,
    0xaf886000U,
    0xe2167000U,
    0x90689800U,
    0x4c40c400U,
    0xe6204e00U,
    0xc2312300U,
    0x855cb880U,
    0x3b9f7f40U,
    0xbd797820U,
    0x42a87410U,
    0xd4a0b628U,
    0x42f7972cU,
    0xa8bc6e86U,
    0x80ca985bU,
    0x64b78ea7U,
    0x679b2846U,
    0xfb7cf6b7U,
    0xd5ae5c58U,
    0x8625c0a3U,
    0x92370b5bU,
    0x8d59ce1eU,
    0xff9fe331U,
    0xd77cd880U,
    0x6bad0f40U,
    0x6d23e024U,
    0x76b5b026U,
    0x72997801U,
    0x54f8741aU,
    0x7ee8b631U,
    0xca839715U,
    0xfcc66e85U,
    0x9be3987cU,
    0x94540e83U,
    0x574ce864U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0xa8000000U,
    0x24000000U,
    0xda000000U,
    0x99000000U,
    0x9e800000U,
    0x85400000U,
    0x4a200000U,
    0xf7f00000U,
    0x57680000U,
    0x45e40000U,
    0xa920000U,
    0x37bb0000U,
    0x11f98000U,
    0xac5b4000U,
    0x886c2000U,
    0x84665000U,
    0xe550800U,
    0x195e1400U,
    0x78ef1a00U,
    0x3a222900U,
    0x7ff17380U,
    0xa36efd40U,
    0x87e0a820U,
    0x1f930410U,
    0x573e3218U,
    0xd7be6d2cU,
    0xe1f961aaU,
    0x6459c049U,
    0x1c6f4196U,
    0xf6609076U,
    0xb351c99fU,
    0x5ddec45dU,
    0x64ab7380U,
    0xee81fd48U,
    0xd432831U,
    0xbe274402U,
    0x35f19221U,
    0x426c7d2fU,
    0x2563c9adU,
    0xccd5c458U,
    0x7a1af394U,
    0xfdcebd5aU,
    0x34550819U,
    0x705e143aU,
    0x2e6f1a2dU,
    0x2b622921U,
    0x47d173b2U,
    0xe99efd51U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x48000000U,
    0xe4000000U,
    0xf2000000U,
    0xf5000000U,
    0x6d800000U,
    0xd8c00000U,
    0x78e00000U,
    0x31700000U,
    0xddb80000U,
    0xf1a40000U,
    0x6f560000U,
    0xc72b0000U,
    0xefea8000U,
    0xaf084000U,
    0x827e2000U,
    0xd1c01000U,
    0x8b666800U,
    0x6ab38c00U,
    0x1fdab600U,
    0xc8929f00U,
    0x9b4bff80U,
    0x55dc7fc0U,
    0xe192c820U,
    0x18cbdc30U,
    0xf61afe18U,
    0x5ff50334U,
    0x7ff92192U,
    0xe1026cf9U,
    0xa787019cU,
    0x31c27ccdU,
    0x9b6169a3U,
    0x42b1f0f2U,
    0x2bdbdfb4U,
    0x72936ff1U,
    0x8a482021U,
    0xca5b103cU,
    0xcc54e836U,
    0xdafcc37U,
    0x1faa9610U,
    0xfaad8f15U,
    0xbf2917aeU,
    0x53e8b3cfU,
    0x910ade2bU,
    0xb17a131dU,
    0xf343c996U,
    0x2f22a0ecU,
    0x551117a3U,
    0x7b8cb3ceU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0xe8000000U,
    0x54000000U,
    0xc6000000U,
    0x23000000U,
    0xbb800000U,
    0xefc00000U,
    0x29200000U,
    0xff700000U,
    0xe0180000U,
    0xcb640000U,
    0x2920000U,
    0x81c90000U,
    0xd1c88000U,
    0x59c94000U,
    0xbdcda000U,
    0x3cdb000U,
    0x9ccdd800U,
    0xb54bbc00U,
    0xbf8f2e00U,
    0xe280100U,
    0xa51b6f80U,
    0x83e4eac0U,
    0x9ed0f820U,
    0x832b4c10U,
    0x299d5608U,
    0x31a70d24U,
    0xeb3119baU,
    0x6fe17d5U,
    0x8d311991U,
    0xf5fe17d8U,
    0xfeb11986U,
    0xde3e17cfU,
    0xf9911998U,
    0x564e17deU,
    0x6409199bU,
    0x51ea17dbU,
    0xf43b198dU,
    0xc09317f2U,
    0xeccb99b3U,
    0xad4e57fbU,
    0xb38c3990U,
    0xe42ee7dfU,
    0xfc1b6198U,
    0x89651bd2U,
    0x2f916fa4U,
    0xdd49eaeaU,
    0xab8a780dU,
    0xe82b0c09U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0x58000000U,
    0x44000000U,
    0x22000000U,
    0x61000000U,
    0x2c800000U,
    0x21c00000U,
    0x4ee00000U,
    0xee100000U,
    0x65a80000U,
    0xd9a40000U,
    0x54320000U,
    0x23d90000U,
    0x9ed88000U,
    0xb45cc000U,
    0x669d2000U,
    0xadfa9000U,
    0x52aa4800U,
    0x3e220400U,
    0x4777e00U,
    0x967d0f00U,
    0x4ec6c80U,
    0x7887bf40U,
    0x1bc7e820U,
    0x4be05410U,
    0x70921608U,
    0x7d6c9b24U,
    0xffc1da96U,
    0xb9e07451U,
    0x6993daa8U,
    0x85e97448U,
    0xa4035aa3U,
    0xd201b47cU,
    0x89047aadU,
    0xa086245eU,
    0x1fc4b2afU,
    0x49e1e04bU,
    0x81966ca8U,
    0x9eabf7dU,
    0x9a056809U,
    0xd5019423U,
    0xe685b60eU,
    0xccc3cb03U,
    0x5c66329eU,
    0xc8d0204bU,
    0x44c9cc9aU,
    0x5bf1ef4dU,
    0x72388003U,
    0x7b4cc021U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xa8000000U,
    0xcc000000U,
    0xb6000000U,
    0xf9000000U,
    0xe5800000U,
    0xc8400000U,
    0x5ea00000U,
    0x93100000U,
    0x3f880000U,
    0x3d640000U,
    0x28f20000U,
    0xe5b90000U,
    0x2abb8000U,
    0x763b4000U,
    0x7bf96000U,
    0x1d1d3000U,
    0x78aa0800U,
    0x18348c00U,
    0xacdc5200U,
    0x40492d00U,
    0x68836c80U,
    0x6fc2ca40U,
    0x6de0e820U,
    0xf336fc10U,
    0xaa5d3a08U,
    0xf289913cU,
    0xbae6b6aaU,
    0xebb02b73U,
    0x2d9cb68dU,
    0x736d2b6eU,
    0xfd53691U,
    0x10ef6b5eU,
    0xf697d695U,
    0xb7c91b7bU,
    0x43c4bea0U,
    0x4be0a76eU,
    0x7232e49cU,
    0xdbdd066aU,
    0xe8cdda0bU,
    0x6746e12aU,
    0xd2265eb0U,
    0xc6d6d76fU,
    0x3d698c92U,
    0x22d0ba7bU,
    0x769800fU,
    0x4dd2403fU,
    0xebeae007U,
    0xc812702eU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0xb8000000U,
    0x74000000U,
    0x9a000000U,
    0x35000000U,
    0xc7800000U,
    0xbfc00000U,
    0xe3200000U,
    0x87b00000U,
    0x32e80000U,
    0xa5640000U,
    0xac560000U,
    0xe0fd0000U,
    0xda788000U,
    0x9e3fc000U,
    0xa4dca000U,
    0x334f5000U,
    0x2297e800U,
    0x165fdc00U,
    0x2b0ace00U,
    0x4e728100U,
    0x5aca1980U,
    0x52510d40U,
    0x37fbc820U,
    0xfcfb4c30U,
    0xec7f8628U,
    0x753b0d24U,
    0xe459bfaeU,
    0xb20e905dU,
    0x57f73f86U,
    0x620c507dU,
    0x8ff31fb9U,
    0x260cc07bU,
    0x3df0579eU,
    0xdf084c68U,
    0x1473718cU,
    0xcfcc116dU,
    0x5d5261aU,
    0x30395d13U,
    0x6bded79fU,
    0xf1ca8c5dU,
    0x72d75184U,
    0x86bc814bU,
    0xa19e6e08U,
    0x5429d103U,
    0x3fc3f194U,
    0x2327d152U,
    0x27b7863eU,
    0xa2ef0d2fU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0xe8000000U,
    0x14000000U,
    0x56000000U,
    0x79000000U,
    0x8800000U,
    0xd6c00000U,
    0xb8e00000U,
    0xec900000U,
    0xcf880000U,
    0xca40000U,
    0x99b20000U,
    0x8c7b0000U,
    0x30fe8000U,
    0xc0394000U,
    0xd9dda000U,
    0x91cd3000U,
    0x22830800U,
    0xadc42400U,
    0xff626600U,
    0x1bd1df00U,
    0x4129b880U,
    0x9a155640U,
    0x54482820U,
    0xdac45410U,
    0x62e6ce18U,
    0xf97cb14U,
    0x440c56baU,
    0x1e6ed45U,
    0xc412d6b5U,
    0x294fad4eU,
    0xc4776baU,
    0x89269d71U,
    0x1f67eacU,
    0x1359b97aU,
    0x280a98a4U,
    0x8be1265cU,
    0xcf168033U,
    0xb6cd4009U,
    0x5707a017U,
    0xad823031U,
    0xac478806U,
    0xf9226432U,
    0xd9f34624U,
    0xbf5eaf1bU,
    0x8209908aU,
    0xb0e50258U,
    0xe894e631U,
    0x118c9f39U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x38000000U,
    0xf4000000U,
    0x5a000000U,
    0x6d000000U,
    0xb9800000U,
    0xd0400000U,
    0xb600000U,
    0x9900000U,
    0x8e480000U,
    0x4a40000U,
    0x69b20000U,
    0x963b0000U,
    0xadbf8000U,
    0xd4ffc000U,
    0x641aa000U,
    0x54cc5000U,
    0xa642800U,
    0x4e112c00U,
    0x410ab200U,
    0x67400300U,
    0x9fe05880U,
    0x522440U,
    0x25690820U,
    0x3e56bc10U,
    0x1a6e3a18U,
    0x46d27f1cU,
    0xadab428eU,
    0x92f3cb7dU,
    0x8fdcc2b6U,
    0x26e80b4bU,
    0xb3146296U,
    0xf08f5b78U,
    0x1b07cab4U,
    0x4a85b773U,
    0x6c5d883U,
    0x8ba2e46bU,
    0x9362814U,
    0xc5fa2c3bU,
    0x6b9d3207U,
    0x578bc339U,
    0xc780f88eU,
    0x8f417444U,
    0x3e0a021U,
    0x8653500aU,
    0xae69a836U,
    0x7cd5ec22U,
    0xb0af9220U,
    0x1373933bU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x38000000U,
    0xdc000000U,
    0xa2000000U,
    0x7b000000U,
    0xc2800000U,
    0x22c00000U,
    0xf1a00000U,
    0x37b00000U,
    0x37e80000U,
    0xaae40000U,
    0x6520000U,
    0x33b0000U,
    0xd3198000U,
    0x48efc000U,
    0xe663e000U,
    0xa5919000U,
    0x271c9800U,
    0x9ee8b400U,
    0xb652600U,
    0xea16a500U,
    0xfa5fe180U,
    0x8a0d2540U,
    0xf434f820U,
    0x80ade410U,
    0x4103de18U,
    0xf580412cU,
    0xaa45bf8eU,
    0xfc62a477U,
    0x4292a788U,
    0x279ed04eU,
    0x772de188U,
    0x462574U,
    0x2365780aU,
    0x9e162406U,
    0x6c5a3e33U,
    0x70ed12fU,
    0xbb2a797U,
    0x65eed061U,
    0x89e5e1b0U,
    0xa8d22559U,
    0xbbff7800U,
    0x85b9243aU,
    0x1fd9be2fU,
    0x884e111bU,
    0x3c52c79cU,
    0x343f806bU,
    0x5b99198aU,
    0x452bc15fU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0xc8000000U,
    0x44000000U,
    0xba000000U,
    0xa5000000U,
    0xfb800000U,
    0xb3400000U,
    0xc2200000U,
    0x5c300000U,
    0x32d80000U,
    0x67e40000U,
    0x52d60000U,
    0x458f0000U,
    0x9e2b8000U,
    0x685c4000U,
    0xb5272000U,
    0x96b35000U,
    0x181be800U,
    0xcf803c00U,
    0x1422a00U,
    0xc323c900U,
    0xedb24580U,
    0xec98b8c0U,
    0x1a414820U,
    0x7a02c30U,
    0x10756238U,
    0x5e7ce53cU,
    0xbb14a7b2U,
    0x556c1dd1U,
    0x67ccf8eU,
    0x71461d9U,
    0x836fc5a6U,
    0xad7bf8e0U,
    0x2b95e81aU,
    0x2aab3c0aU,
    0x9d9faa28U,
    0xa3c0891dU,
    0x7de6e5b6U,
    0x27d3a8c5U,
    0x660b8027U,
    0x516c4038U,
    0x9c7f2017U,
    0xb217503bU,
    0x40ede81cU,
    0x923f3c3dU,
    0x17b1aa11U,
    0x699b890bU,
    0xf1c3659bU,
    0x8ce4e8f9U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x68000000U,
    0xfc000000U,
    0x1a000000U,
    0xe1000000U,
    0xb2800000U,
    0x8ec00000U,
    0x4200000U,
    0xfe900000U,
    0x76980000U,
    0xa5640000U,
    0x9cf20000U,
    0xc3e90000U,
    0x4bca8000U,
    0x575d4000U,
    0x74c0e000U,
    0x55255000U,
    0xd413c800U,
    0xbc5fb400U,
    0x2f440600U,
    0x65621900U,
    0xfcf4b880U,
    0x33ebdec0U,
    0xf3cba820U,
    0xc35ea410U,
    0x92c5ae08U,
    0xae21bd34U,
    0x8791969aU,
    0x801a23ffU,
    0xa5a25ea6U,
    0x9fd597e8U,
    0x74fe5884U,
    0xe0138ec7U,
    0xca58e033U,
    0x9c415024U,
    0x1ae1c809U,
    0xb2b6b41aU,
    0xa40e8617U,
    0xa1ff5914U,
    0x249458b9U,
    0xf79e8ef1U,
    0xe7e0600dU,
    0xaa351030U,
    0x53cba806U,
    0x535ea43dU,
    0xdac5ae39U,
    0x8221bd2aU,
    0xf5919691U,
    0x9d1a23c4U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0x8000000U,
    0x7c000000U,
    0xd6000000U,
    0x91000000U,
    0xcb800000U,
    0xf0400000U,
    0x3ea00000U,
    0xbab00000U,
    0x3380000U,
    0xd0640000U,
    0x5dd20000U,
    0x2c6b0000U,
    0x6bc98000U,
    0x96fe4000U,
    0xdb86e000U,
    0x68423000U,
    0x9aa24800U,
    0xa8b07c00U,
    0x303a9200U,
    0x20e42d00U,
    0x2113b180U,
    0xb80e4fc0U,
    0xd41f2820U,
    0xe1970c10U,
    0x8ccfba18U,
    0x5c7c212cU,
    0x38478b82U,
    0x62a72edfU,
    0xbcb7c395U,
    0x2a3c52f4U,
    0x6fe4d18aU,
    0x7963fc0U,
    0x55c98035U,
    0x8bfe400dU,
    0xae06e02fU,
    0xc5023023U,
    0xb1824812U,
    0xf407c0cU,
    0x10229212U,
    0x2b702d13U,
    0x8ad9b18bU,
    0xef14ffcU,
    0xdf9ca81eU,
    0x31d64c2bU,
    0x626ada19U,
    0x5ecf513eU,
    0x4f78a3b1U,
    0x18c022c5U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0xb8000000U,
    0xfc000000U,
    0x62000000U,
    0xc1000000U,
    0x56800000U,
    0x46400000U,
    0x3200000U,
    0x9a100000U,
    0xb9380000U,
    0x90e40000U,
    0xf8f20000U,
    0x78cf0000U,
    0xa2ac8000U,
    0x6e1bc000U,
    0x6cf6e000U,
    0xfece9000U,
    0xda84800U,
    0x339e1400U,
    0xf9320200U,
    0x732f6100U,
    0x71dd3c80U,
    0x8417e7c0U,
    0x5a382820U,
    0xe7604410U,
    0xf8322a38U,
    0x85ab2504U,
    0xc79d16aeU,
    0x2f33c2ffU,
    0x8429beb8U,
    0xc05846e0U,
    0xcbd5748dU,
    0xead9f3c5U,
    0xaf922a36U,
    0xc8fb250dU,
    0x730516a0U,
    0x8f87c2efU,
    0x1cc3bebdU,
    0x8f6346d0U,
    0x1c33f4bdU,
    0xeba933c8U,
    0xcc9a4a16U,
    0xfcb17507U,
    0xaef3e9dU,
    0x263886dbU,
    0xc5651485U,
    0xd937a3f3U,
    0xc32a0204U,
    0x39db6108U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0x8000000U,
    0x4000000U,
    0xaa000000U,
    0x4d000000U,
    0x88800000U,
    0xad400000U,
    0x2ba00000U,
    0x4d700000U,
    0x8cb80000U,
    0x39640000U,
    0xa6120000U,
    0x3baf0000U,
    0xcc4c8000U,
    0x341e4000U,
    0x43912000U,
    0xae68d000U,
    0xb2a9a800U,
    0x4ecaa400U,
    0x24595a00U,
    0xf8b7c500U,
    0x425cd780U,
    0xbbb4c3c0U,
    0xddde0820U,
    0xcb773410U,
    0xdfbf5238U,
    0xaee4f104U,
    0xd2d18582U,
    0x93cf32c1U,
    0xe4db0d8aU,
    0xb1f246c3U,
    0xd3ff7fbaU,
    0xb3c567ffU,
    0xdde1d210U,
    0xf555b106U,
    0x1f0c25b3U,
    0x50b9a2d8U,
    0x3f638583U,
    0xb51032e7U,
    0x4c2f8da9U,
    0xa88806f2U,
    0x947c5f81U,
    0x6f02b7c8U,
    0x8184fa1cU,
    0x6fc15506U,
    0xdbe45f82U,
    0xe656b7ceU,
    0x688efa3aU,
    0x347a553cU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x38000000U,
    0xf4000000U,
    0x12000000U,
    0x79000000U,
    0xc800000U,
    0x59400000U,
    0xb4600000U,
    0x19500000U,
    0xf4980000U,
    0x62a40000U,
    0x1df20000U,
    0xc26d0000U,
    0x75498000U,
    0x8d7fc000U,
    0x9a34a000U,
    0xb64a1000U,
    0xc4fe8800U,
    0xb1f58400U,
    0x746b3a00U,
    0x364ddb00U,
    0x84f9b680U,
    0x11f3a2c0U,
    0xc469a820U,
    0xe495410U,
    0x70fa9228U,
    0x3f08f2cU,
    0xbd69248eU,
    0x2ca2dfdU,
    0x29bb0ca4U,
    0xb791b9ceU,
    0xa43cbe8bU,
    0xf654e6eaU,
    0x4b1fb22bU,
    0xaa615f07U,
    0x66510c9fU,
    0x8318b9d7U,
    0xc6673eaeU,
    0x305626d5U,
    0xd01a923fU,
    0x47e08f11U,
    0x77912493U,
    0x443e2df4U,
    0xe6510c8cU,
    0xc318b9f1U,
    0x66673eb5U,
    0x805626f2U,
    0xe81a9235U,
    0xb3e08f2bU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0x58000000U,
    0xd4000000U,
    0x16000000U,
    0xcf000000U,
    0x48800000U,
    0x34c00000U,
    0xd4e00000U,
    0x43900000U,
    0x42280000U,
    0x55a40000U,
    0x1a320000U,
    0xda1f0000U,
    0xb6ba8000U,
    0xfb0cc000U,
    0x557a000U,
    0x87cb3000U,
    0xfdb13800U,
    0x365aa400U,
    0xee9e7600U,
    0x687e4f00U,
    0x4aed5080U,
    0x8b46ab40U,
    0xeca61820U,
    0xdb65410U,
    0x1e586e38U,
    0x52981b34U,
    0x227d3e96U,
    0x1feab075U,
    0xcec126a5U,
    0x49e7e463U,
    0x6a11c88aU,
    0xf56c3f69U,
    0xb301d63bU,
    0xa2817f01U,
    0x51c668bbU,
    0x19670f67U,
    0x6ad0ee27U,
    0xe78bdb2dU,
    0x67901eb7U,
    0x7c2d4064U,
    0x26a7be80U,
    0x18b6704cU,
    0xbbde8681U,
    0x27d8d469U,
    0x1ddaf09cU,
    0x20dd9b5bU,
    0x395f2006U,
    0x618f039U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0xd8000000U,
    0x5c000000U,
    0x26000000U,
    0x9000000U,
    0x94800000U,
    0x45c00000U,
    0xb9200000U,
    0x8e100000U,
    0xc7980000U,
    0x43640000U,
    0xd5720000U,
    0x6eeb0000U,
    0xdb898000U,
    0x6ff8c000U,
    0x1197e000U,
    0xf5ddb000U,
    0xc040800U,
    0xe070400U,
    0xed043e00U,
    0x5e83d100U,
    0xb2c76c80U,
    0x78a21ec0U,
    0x79506820U,
    0x327d7410U,
    0xe0d45618U,
    0xd93ea52cU,
    0x53333ab6U,
    0xc38cbbd7U,
    0x13fb52a9U,
    0xe795cfd2U,
    0x94dd049dU,
    0x1c806aedU,
    0x31c7be20U,
    0x7b241108U,
    0x4d130c96U,
    0xa4186ee9U,
    0xc720000fU,
    0x9b10000fU,
    0x1518002cU,
    0xbfa40007U,
    0x20d20008U,
    0xf93b0008U,
    0x83318020U,
    0xab8cc029U,
    0x97fde02cU,
    0x9d92b016U,
    0xbbdf8805U,
    0x8104c400U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x78000000U,
    0xb4000000U,
    0x6000000U,
    0x29000000U,
    0x98800000U,
    0x15c00000U,
    0xe2e00000U,
    0xd8500000U,
    0xa7280000U,
    0xf8540000U,
    0x572e0000U,
    0xb0570000U,
    0xcb2c8000U,
    0x7a54c000U,
    0x50292000U,
    0xcdd4f000U,
    0xf4686800U,
    0xa2703400U,
    0xdb18c600U,
    0x3fef2300U,
    0x5c320480U,
    0x37b81a40U,
    0x4318c620U,
    0x1bef2330U,
    0x42320498U,
    0xfab81a54U,
    0xa598c63eU,
    0x932f231dU,
    0x3e520499U,
    0x1e281a5eU,
    0x78d0c618U,
    0xa6eb2318U,
    0x2cb404a1U,
    0x8e7b1a68U,
    0x43fa4631U,
    0x94bce326U,
    0xe09fa4b4U,
    0x89ac2a44U,
    0x2c978e03U,
    0x814ce738U,
    0x9fc62aa0U,
    0xd9e7cd77U,
    0x5fd524beU,
    0x2b6bea50U,
    0x5bf42e36U,
    0xca5fd708U,
    0x2cde291U,
    0xfd84c979U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x48000000U,
    0x94000000U,
    0x9e000000U,
    0x7d000000U,
    0x81800000U,
    0x8c400000U,
    0x87e00000U,
    0xffb00000U,
    0x18c80000U,
    0xe7b40000U,
    0x4ce0000U,
    0x8db70000U,
    0x37c88000U,
    0x39354000U,
    0xae0c2000U,
    0xac91b000U,
    0xab5fc800U,
    0xca6dbc00U,
    0xd867aa00U,
    0xaef38300U,
    0xaea95780U,
    0xc44c540U,
    0x47e7aa20U,
    0x9fb38330U,
    0xc8c95798U,
    0xafb4c574U,
    0x90cfaa32U,
    0x13b78315U,
    0x4acf57bfU,
    0xb8b7c56bU,
    0x22492a12U,
    0x2b75c336U,
    0x54ebf79eU,
    0xd2a33554U,
    0x3fd2c214U,
    0xaa3dcf0fU,
    0x231d959fU,
    0x3b8a0a77U,
    0x7ed4bf99U,
    0x31bfc941U,
    0x645f4814U,
    0x4ecfc2cU,
    0x5aa58a0bU,
    0xcbd53337U,
    0xe43e1f88U,
    0x161c3966U,
    0x2e0c202dU,
    0x6c91b00aU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0xf8000000U,
    0xdc000000U,
    0xe000000U,
    0x91000000U,
    0xad800000U,
    0xe8400000U,
    0x49a00000U,
    0xadf00000U,
    0xb8880000U,
    0xc5f40000U,
    0x8c8e0000U,
    0x3ff10000U,
    0x478d8000U,
    0x2976c000U,
    0x414b6000U,
    0xba16b000U,
    0x71dfe800U,
    0xa30bdc00U,
    0xb4aa00U,
    0xb6ab9100U,
    0x6342c580U,
    0x3f20d540U,
    0x7b34aa20U,
    0xa3eb9130U,
    0xd162c5a8U,
    0x4790d55cU,
    0xd19caa1eU,
    0x6eaf9107U,
    0xf44c5abU,
    0x6925d578U,
    0x46372a35U,
    0xf86c513dU,
    0x742425b9U,
    0xadb0a553U,
    0x528221bU,
    0x3202fd0cU,
    0x8f07879aU,
    0x1481985cU,
    0x11c7458aU,
    0x8be21546U,
    0xa751ca2aU,
    0x27fc2102U,
    0xa9b8ad9eU,
    0x4519c942U,
    0xbf6d603cU,
    0x24a3b004U,
    0xbe74683bU,
    0x99c81c1eU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x18000000U,
    0x64000000U,
    0x2000000U,
    0xb5000000U,
    0x6800000U,
    0x3cc00000U,
    0x51a00000U,
    0x8bb00000U,
    0xe8a80000U,
    0x2fb40000U,
    0x4aaa0000U,
    0x8ab30000U,
    0xa42d8000U,
    0xca73c000U,
    0x938ea000U,
    0xf6c3d000U,
    0xc8a30800U,
    0xe337e400U,
    0xef68ee00U,
    0xb3966900U,
    0x28593380U,
    0xbefeec40U,
    0xdec8ee20U,
    0x88266910U,
    0x38f133b8U,
    0x54aec7cU,
    0x8e62ee26U,
    0xd3956909U,
    0x985cb3b8U,
    0x46f92c51U,
    0x4acc4e27U,
    0x9226b906U,
    0xe9f7bbacU,
    0x1cac873U,
    0x7a6a01dU,
    0x84b7d00dU,
    0x2f29083eU,
    0xfff4e411U,
    0xeecd6e14U,
    0x3021a92fU,
    0x4cf593a2U,
    0xef4a3c50U,
    0x47646600U,
    0xb3154d1eU,
    0x1398fda1U,
    0xa4db956dU,
    0xfb39f58dU,
    0x17eb7167U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0x88000000U,
    0xd4000000U,
    0x82000000U,
    0x7000000U,
    0x10800000U,
    0x91c00000U,
    0xaa600000U,
    0x52100000U,
    0x48780000U,
    0x8e140000U,
    0xde7e0000U,
    0xab110000U,
    0x99fc8000U,
    0xa2554000U,
    0x765a2000U,
    0xd8215000U,
    0x6b321800U,
    0xecdd400U,
    0xaa9c9600U,
    0xd1c78700U,
    0xca638680U,
    0xa21404c0U,
    0x907c9620U,
    0xd2178730U,
    0x887b86a8U,
    0x2e1004d4U,
    0x8e7a9602U,
    0x23128705U,
    0x4df90688U,
    0x205444d5U,
    0x715c3606U,
    0xc8a69721U,
    0xfaf13ea2U,
    0xa4a8c0c1U,
    0xf88ab814U,
    0x99b8c402U,
    0x44702e15U,
    0x7c6a432bU,
    0x3b692892U,
    0x4bee07daU,
    0x2a2d1e80U,
    0x584c90f9U,
    0x565a2022U,
    0x48215008U,
    0x43321820U,
    0x8acdd41dU,
    0xa09c961eU,
    0x2c78714U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0xd8000000U,
    0xbc000000U,
    0xc2000000U,
    0xc9000000U,
    0x5d800000U,
    0x50400000U,
    0x7e00000U,
    0x5eb00000U,
    0xf7680000U,
    0xab40000U,
    0xf16e0000U,
    0xadb70000U,
    0xfed8000U,
    0xd71c000U,
    0x7bc86000U,
    0xf67d000U,
    0x41762800U,
    0xa1c93c00U,
    0x1a66aa00U,
    0xaef40300U,
    0xe088df80U,
    0xfc04f840U,
    0x6206aa20U,
    0xd9040330U,
    0xf580df98U,
    0x3440f85cU,
    0x79e0aa16U,
    0x55b7031fU,
    0x63eb5fa8U,
    0x772386eU,
    0xa6cb4a01U,
    0xf4e6130bU,
    0xa63097a9U,
    0xf0a91479U,
    0x8015a83cU,
    0x53bffc09U,
    0xe32b4a15U,
    0xa3561312U,
    0x6cd8979fU,
    0xda5d147aU,
    0xae9ba833U,
    0x1cb8fc3dU,
    0xd9aeca21U,
    0x6d93d321U,
    0xbbfef793U,
    0x28cdc47bU,
    0xe7e00006U,
    0xeeb00014U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0xb8000000U,
    0x24000000U,
    0x6000000U,
    0xa5000000U,
    0x1e800000U,
    0x4c00000U,
    0xdfa00000U,
    0xec700000U,
    0x5ee80000U,
    0xc0740000U,
    0xd4ea0000U,
    0x9f770000U,
    0x1d688000U,
    0x9e32c000U,
    0x5989a000U,
    0xb1a2f000U,
    0x95748800U,
    0x26e0400U,
    0x37b2ae00U,
    0x8c9f900U,
    0x4d444d80U,
    0x9ee44e40U,
    0xc892ae20U,
    0x2179f910U,
    0xb48c4d98U,
    0x43204e54U,
    0x3bb0ae0eU,
    0xb2caf919U,
    0xfa46cd99U,
    0xcb618e7dU,
    0xebd38e09U,
    0xc699c918U,
    0x209965aeU,
    0x9598ba46U,
    0x531c881eU,
    0x23da0428U,
    0x4278ae1bU,
    0x2f0ef921U,
    0x6964cdb9U,
    0x8d28e5fU,
    0x1d190e35U,
    0x6ad80909U,
    0xf6fa45aaU,
    0x44cb8a42U,
    0x97432017U,
    0x79e3302bU,
    0x2517a83fU,
    0x263d340dU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x78000000U,
    0xac000000U,
    0xfa000000U,
    0xbf000000U,
    0x21800000U,
    0x44400000U,
    0xc1e00000U,
    0x8b300000U,
    0xa4280000U,
    0xd1340000U,
    0xb2e0000U,
    0xa8b50000U,
    0x33688000U,
    0xeb56c000U,
    0xab5de000U,
    0x94f91000U,
    0x812c1800U,
    0x7fb61c00U,
    0xe6eb3a00U,
    0x29109300U,
    0x57bda380U,
    0x2d4e5240U,
    0xbac33a20U,
    0x94249330U,
    0x8693a388U,
    0xeafb5274U,
    0xd02bba3eU,
    0x9732531bU,
    0x162e43b6U,
    0x4a32425bU,
    0xd4afa236U,
    0x7df04f0aU,
    0x3a0b7986U,
    0x40a7d179U,
    0x1452819fU,
    0x6adcdd7eU,
    0x20bba3a4U,
    0xe8cf5263U,
    0x2085ba1bU,
    0x14c75330U,
    0xbd26c3b0U,
    0xc214825dU,
    0xe03a420dU,
    0xc0d5f34U,
    0x35a1619fU,
    0x2d0cd6dU,
    0xb1f3b96U,
    0x8b1f8e72U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0xc8000000U,
    0xd4000000U,
    0xd6000000U,
    0x21000000U,
    0x95800000U,
    0x3e400000U,
    0xb8600000U,
    0x8b900000U,
    0x1c580000U,
    0x69940000U,
    0x5b5a0000U,
    0xe5150000U,
    0x3c9c8000U,
    0xa9334000U,
    0x2c6a2000U,
    0x5c7df000U,
    0xd3672800U,
    0x7d175c00U,
    0x6099da00U,
    0x8b331f00U,
    0xb6e4680U,
    0x20fc9dc0U,
    0xaca1da20U,
    0x2d371f10U,
    0x726c4688U,
    0x897d9df4U,
    0x40e75a12U,
    0xaa515f25U,
    0x997ae6bdU,
    0xe8e32dfcU,
    0x8e525237U,
    0x577ff32aU,
    0xd5e61493U,
    0x19d66edeU,
    0x9e3b4eb0U,
    0xd90231f0U,
    0x39852825U,
    0x4465c37U,
    0x83675a17U,
    0xf5115f3eU,
    0x949ae6b4U,
    0x8d332df1U,
    0xe26a520cU,
    0x617bf338U,
    0x44e41486U,
    0xb4576ed5U,
    0x6c7dce82U,
    0xab6471e0U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0xb8000000U,
    0x94000000U,
    0xaa000000U,
    0x11000000U,
    0x67800000U,
    0xa0c00000U,
    0x53200000U,
    0xb7500000U,
    0xa2780000U,
    0x11540000U,
    0x5d7e0000U,
    0xd5d10000U,
    0xf3e8000U,
    0x53b4c000U,
    0xce8a2000U,
    0xd8399000U,
    0x7b36a800U,
    0x2c8ac00U,
    0xf558ae00U,
    0x7403b500U,
    0xda025480U,
    0x1900c6c0U,
    0x5b80ae20U,
    0x26c7b530U,
    0x7c2454a8U,
    0x6bd5c6c4U,
    0x74382e0eU,
    0x85367515U,
    0x19cef482U,
    0x93d996c0U,
    0xb422617U,
    0xbde7893dU,
    0x5774d296U,
    0x5faa1fedU,
    0xf268f4bfU,
    0x90cc96f9U,
    0xb05aa621U,
    0x19864928U,
    0x7bc672b4U,
    0x95a24fd9U,
    0xd814fc94U,
    0x7d986affU,
    0xcb20000eU,
    0xf3500008U,
    0x10780012U,
    0x8454003aU,
    0x28fe0013U,
    0xf011001aU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xc8000000U,
    0xe4000000U,
    0x76000000U,
    0x29000000U,
    0x7800000U,
    0xbbc00000U,
    0x5de00000U,
    0x6d500000U,
    0xdc680000U,
    0x24540000U,
    0xebea0000U,
    0xb7930000U,
    0x22098000U,
    0x64c54000U,
    0x3366a000U,
    0x3112f000U,
    0x4a4de800U,
    0xdce1cc00U,
    0xbed1b600U,
    0x39ae6300U,
    0xc4b07c80U,
    0x3f387040U,
    0x1d3bb620U,
    0x2a3d6310U,
    0x70b9fcb8U,
    0x2fd305cU,
    0xe1dd1612U,
    0x44ef9329U,
    0x111414a5U,
    0x9a4cfc56U,
    0x84e4a013U,
    0xe2d5f007U,
    0x63ae6832U,
    0x7fb78c0dU,
    0x67be9604U,
    0x8879d31eU,
    0xcb9b34b0U,
    0xfbcb4c7cU,
    0xee27e81eU,
    0x74b2cc2eU,
    0x57383619U,
    0x693b2326U,
    0xe43edc9fU,
    0x75be805eU,
    0x977c5e24U,
    0x51faf35U,
    0x3789ca92U,
    0xc0021353U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0x68000000U,
    0x64000000U,
    0xfe000000U,
    0x5d000000U,
    0x8c800000U,
    0x99400000U,
    0x6d600000U,
    0x8af00000U,
    0xc6480000U,
    0xe3f40000U,
    0x7cce0000U,
    0x33b70000U,
    0x1b2e8000U,
    0x71044000U,
    0xa685a000U,
    0x2247b000U,
    0xa0e14800U,
    0x6d30d400U,
    0xf56d0200U,
    0x2be4ff00U,
    0x78b47580U,
    0x6eaa3d40U,
    0xfac30220U,
    0x1fa3ff30U,
    0x2152f598U,
    0x111a7d44U,
    0xdbe8a23aU,
    0xbda34f29U,
    0x2e553da7U,
    0x2a9ae953U,
    0x152e0019U,
    0x9407000fU,
    0x4606803cU,
    0x31004031U,
    0x683a008U,
    0x5244b007U,
    0xd8e1c83bU,
    0x61339439U,
    0x6f6e2214U,
    0x88e30f02U,
    0xa9369dadU,
    0x7b6e5966U,
    0xee7c83fU,
    0xf830940fU,
    0x6deea213U,
    0x34a04f0aU,
    0x44d5bdaaU,
    0xf2d9a945U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0xe8000000U,
    0xdc000000U,
    0x9e000000U,
    0xf9000000U,
    0x15800000U,
    0x4f400000U,
    0x42a00000U,
    0x3e300000U,
    0x3e380000U,
    0xb340000U,
    0x2dba0000U,
    0x29710000U,
    0x89b8000U,
    0xd7064000U,
    0x44856000U,
    0x26c6b000U,
    0xa3e0a800U,
    0x6d931c00U,
    0xc98b5200U,
    0xe44a4900U,
    0x9d2a4380U,
    0x2f7848c0U,
    0x50915220U,
    0x960b4910U,
    0x8009c3a8U,
    0x750a08ccU,
    0x868e321aU,
    0x34ccf927U,
    0xc9eaeb8fU,
    0x39b54f2U,
    0xd82001fU,
    0x9b450034U,
    0x90a1801fU,
    0x8137403dU,
    0xe6bee030U,
    0x13f0f026U,
    0x475dc83cU,
    0xb961ac3bU,
    0x5251fa08U,
    0xefe85524U,
    0x1e9a918aU,
    0x220441ccU,
    0xb7067185U,
    0x9481b1fcU,
    0x5ec239a0U,
    0xa7e35de6U,
    0xc79723bfU,
    0x728af8ceU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0x88000000U,
    0xc4000000U,
    0x9e000000U,
    0x1f000000U,
    0x60800000U,
    0x3a400000U,
    0xb3a00000U,
    0xa3700000U,
    0x19980000U,
    0x20740000U,
    0xfb1e0000U,
    0x47370000U,
    0x15398000U,
    0x73834000U,
    0xa0c7a000U,
    0x98e3b000U,
    0x17515800U,
    0xac2bb400U,
    0xea0b0a00U,
    0x4ab86900U,
    0x79453180U,
    0x3124d5c0U,
    0x34350a20U,
    0x7fbf6930U,
    0xd0c4b198U,
    0xd0e395fcU,
    0xb354aa02U,
    0xc22fd901U,
    0x7d0a69bfU,
    0xee3861fbU,
    0xdd07801aU,
    0x9d84400fU,
    0xf7c62013U,
    0x5c64f013U,
    0x4310f83cU,
    0x888b0437U,
    0xedfdd235U,
    0xf7679d3eU,
    0xf5901b9bU,
    0x9cc4cdaU,
    0x50dec39aU,
    0xf957b8efU,
    0xfb2de9beU,
    0x2e8c21f3U,
    0x1ef9a036U,
    0x5de4b018U,
    0x36d0d803U,
    0x89ecf431U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0xc8000000U,
    0xc000000U,
    0xf6000000U,
    0x91000000U,
    0xbc800000U,
    0x2400000U,
    0x38e00000U,
    0x4a700000U,
    0x7b880000U,
    0xfd740000U,
    0x2e0a0000U,
    0x27b10000U,
    0x1ead8000U,
    0xfa274000U,
    0xed5e000U,
    0x341cf000U,
    0x760f6800U,
    0xf3b21400U,
    0x7cabfa00U,
    0xa9251300U,
    0x11550980U,
    0xe1dcdf40U,
    0x9729fa20U,
    0xaee01310U,
    0xcb7289a8U,
    0x5f0a9f7cU,
    0xcb319a12U,
    0x24eba313U,
    0x6c00195U,
    0xbea07b58U,
    0x2897083dU,
    0xa6f9a413U,
    0x6379721bU,
    0x61bff70aU,
    0xc51b9b83U,
    0xda8ad87cU,
    0x69f28998U,
    0x704a9f6fU,
    0xd1519a3eU,
    0x1dba32dU,
    0xc728019cU,
    0x96e47b69U,
    0xf75082bU,
    0xa50ca40aU,
    0xac36f22dU,
    0x96db736U,
    0xb801fb8dU,
    0x8404686bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0x78000000U,
    0x4c000000U,
    0x52000000U,
    0xe7000000U,
    0x28800000U,
    0x62400000U,
    0x600000U,
    0x1300000U,
    0x4b480000U,
    0xec340000U,
    0x38ce0000U,
    0x2cf10000U,
    0xc1ef8000U,
    0x2f22c000U,
    0x42d5e000U,
    0xf2bc5000U,
    0x1258d800U,
    0x852a1400U,
    0x70014a00U,
    0x18015700U,
    0xdc036380U,
    0x1a033d40U,
    0xd3074a20U,
    0x36845730U,
    0xd742e3a8U,
    0xcfe0fd4cU,
    0x4bf52a3eU,
    0x296ec723U,
    0xed61dbbcU,
    0x72b7b975U,
    0x8b8b3834U,
    0x1513443bU,
    0x3b18123cU,
    0xafc88335U,
    0x1c704986U,
    0x7fa8fa70U,
    0x354711baU,
    0x90e02e42U,
    0x8f73bba0U,
    0x292d2954U,
    0x72000008U,
    0x17000018U,
    0xf0800004U,
    0x1e400037U,
    0x2a600002U,
    0xaa300003U,
    0x31c8001dU,
    0x6974000dU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0xd8000000U,
    0xac000000U,
    0xe2000000U,
    0xb3000000U,
    0x3800000U,
    0xd5400000U,
    0x76600000U,
    0xaab00000U,
    0x9480000U,
    0x93b40000U,
    0xfdca0000U,
    0xa3770000U,
    0x6bea8000U,
    0x5921c000U,
    0xe3d7e000U,
    0x327ad000U,
    0x34bc7800U,
    0x6f9fcc00U,
    0x6f4b5200U,
    0xe6b13f00U,
    0xbb492780U,
    0x98b7c540U,
    0x42495220U,
    0x4c323f10U,
    0x289a798U,
    0x12110544U,
    0x8cdc3216U,
    0xd42d2f3bU,
    0xb6803fa0U,
    0xb3c71968U,
    0x3da39816U,
    0x4b111c0eU,
    0x685d2a3dU,
    0x3ce9f302U,
    0x8ca0f5b4U,
    0xab933a7aU,
    0x69d95baU,
    0x23c82a7eU,
    0x6a760da0U,
    0x176d3653U,
    0x7de12794U,
    0x2bf3c571U,
    0x412b5223U,
    0xa0013f3aU,
    0x3001278aU,
    0xa803c56bU,
    0x6403520aU,
    0x96053f23U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0xe8000000U,
    0x9c000000U,
    0x6a000000U,
    0xbf000000U,
    0xdd800000U,
    0xac400000U,
    0x42200000U,
    0xd0100000U,
    0xe8380000U,
    0x59140000U,
    0x80ba0000U,
    0x7d50000U,
    0xeb5f8000U,
    0x4da24000U,
    0x67536000U,
    0xf99d1000U,
    0x34409800U,
    0x5620fc00U,
    0xb6157a00U,
    0x653a2d00U,
    0x4f955a80U,
    0x77d5cc0U,
    0x3cb77a20U,
    0x1beb2d10U,
    0x448da88U,
    0x381e1cecU,
    0xb2019a1aU,
    0x8b017d37U,
    0xb84a292U,
    0xc941b0c3U,
    0xc8a7182dU,
    0x3dd6bc1cU,
    0x6c5c1a02U,
    0x74223d37U,
    0x651242b7U,
    0x1abbe0daU,
    0x70d3600aU,
    0x1add100aU,
    0x63e0983fU,
    0x670fc02U,
    0x9e0d7a13U,
    0xcf3e2d2cU,
    0x90975a97U,
    0x4afc5cffU,
    0xc8f2fa03U,
    0x2dcc6d1dU,
    0x225c3abcU,
    0x5254cdfU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x18000000U,
    0x74000000U,
    0x26000000U,
    0x43000000U,
    0xdc800000U,
    0x1c400000U,
    0x76a00000U,
    0xad100000U,
    0x82480000U,
    0x6c140000U,
    0xe3ce0000U,
    0x7bd30000U,
    0x23ac8000U,
    0xe723c000U,
    0xd2d72000U,
    0x9e2ef000U,
    0xeae73800U,
    0x1bb11400U,
    0xcf5e5a00U,
    0xbe5ae100U,
    0x97dd5280U,
    0xec1fe340U,
    0x7e785a20U,
    0x878de130U,
    0xe977d2b8U,
    0x50bb2354U,
    0x26dfa26U,
    0x5c43d12dU,
    0x56a34ab1U,
    0x1d130744U,
    0xca4db811U,
    0x15d42aU,
    0xb1cbfa2cU,
    0x1ed4d12fU,
    0xbc29ca91U,
    0x27e7c741U,
    0xb830180cU,
    0x459fe415U,
    0xc5b96227U,
    0xf5ebf531U,
    0x40830897U,
    0x2645024fU,
    0xcfa508bbU,
    0x28920244U,
    0x4b8f8890U,
    0xcb76c275U,
    0x9dba2894U,
    0xa1e8f256U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0x58000000U,
    0xdc000000U,
    0xaa000000U,
    0xa9000000U,
    0x6b800000U,
    0xcbc00000U,
    0x64600000U,
    0x92900000U,
    0x7bf80000U,
    0xe5940000U,
    0xf77e0000U,
    0x8ad10000U,
    0xd85b8000U,
    0xf163c000U,
    0x83156000U,
    0x513a9000U,
    0xde722800U,
    0x802acc00U,
    0x4a4cf600U,
    0xc7daaf00U,
    0x74a24280U,
    0x2873c0c0U,
    0x432af620U,
    0x28cfaf30U,
    0xd79fc2a8U,
    0x830500ecU,
    0x2821636U,
    0x43ff07U,
    0x1fa50a82U,
    0xaef39cc6U,
    0x356c282cU,
    0x346bcc35U,
    0xbbef761bU,
    0x162d6f22U,
    0x99492292U,
    0x4d5850fcU,
    0xe0e35e2eU,
    0xa9d6a31cU,
    0x6ade549aU,
    0xa9213fdbU,
    0x40b47cb2U,
    0x55cff3d5U,
    0x421e8abdU,
    0xdfc05cd3U,
    0x3a61480bU,
    0xb5955c1eU,
    0x7f7b5e11U,
    0x1ed2a303U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x28000000U,
    0xb4000000U,
    0xde000000U,
    0x5000000U,
    0xd1800000U,
    0x5dc00000U,
    0x25e00000U,
    0x6bd00000U,
    0x8c480000U,
    0xaa540000U,
    0xb98a0000U,
    0xdbb30000U,
    0x9c5e8000U,
    0x7afa4000U,
    0x538f2000U,
    0x40b6b000U,
    0xa8dc0800U,
    0x66bebc00U,
    0x3aca600U,
    0xba84b300U,
    0x91434880U,
    0x45a7d940U,
    0xc1f22620U,
    0x857ef310U,
    0xf34c68b8U,
    0xc8d16964U,
    0x64ce2e2aU,
    0x3c104f3dU,
    0xa2a8ce8fU,
    0xdd01da65U,
    0x9d87669eU,
    0xffc4966aU,
    0xdae468a6U,
    0x4955696fU,
    0xb10c2e05U,
    0xddf74f24U,
    0xaf7c4ea2U,
    0xc8489a64U,
    0x8c56c6adU,
    0xc088665fU,
    0xf03740a8U,
    0x329d655eU,
    0x3f9c8001U,
    0x6a1d4039U,
    0xb1dba015U,
    0x9d3ff030U,
    0x656da82dU,
    0xdbe24c31U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0xe8000000U,
    0xac000000U,
    0x9e000000U,
    0x13000000U,
    0xb800000U,
    0x68400000U,
    0x2fa00000U,
    0x21d00000U,
    0x99d80000U,
    0x3a540000U,
    0x899e0000U,
    0x1f10000U,
    0x82498000U,
    0x512e4000U,
    0x9298e000U,
    0x3674b000U,
    0xcc097800U,
    0xd98d6c00U,
    0x62cc7a00U,
    0xe6e8f100U,
    0x5f7bc480U,
    0x10647c0U,
    0xb685fa20U,
    0x38c6b130U,
    0x106324a8U,
    0x1032f7fcU,
    0x28ac821aU,
    0x17dbdd1bU,
    0x51575e8fU,
    0x961e06f8U,
    0x43b14698U,
    0x64e89ac1U,
    0x5a7d24a4U,
    0xe583f7c0U,
    0x53450216U,
    0x28259d33U,
    0x2797be9cU,
    0x4d7eb6dbU,
    0xbc063eb9U,
    0xe604f6dfU,
    0x700dea0U,
    0x218146d7U,
    0xa140268eU,
    0x5226ad9U,
    0xcf14bca6U,
    0xc6be2bd1U,
    0x57e60017U,
    0x76750015U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x18000000U,
    0x7c000000U,
    0xea000000U,
    0x27000000U,
    0x3a800000U,
    0x4c400000U,
    0x99600000U,
    0x4c700000U,
    0x4780000U,
    0xeef40000U,
    0x743a0000U,
    0xfd970000U,
    0xef4d8000U,
    0xdb6fc000U,
    0x31fa6000U,
    0xdc329000U,
    0xa71c5800U,
    0xd86c400U,
    0xdec5fa00U,
    0xd1226f00U,
    0xb3120980U,
    0x9d0b94c0U,
    0x8b087a20U,
    0x60daf10U,
    0x7b886998U,
    0xfd4904fcU,
    0x306c2226U,
    0x997f6b0fU,
    0x3b7793a2U,
    0xf6fc6bf5U,
    0x56b3aba8U,
    0x935b3fdcU,
    0x18e589a4U,
    0x60b354f6U,
    0x8e5f9a31U,
    0x8d60ff1bU,
    0x8276519fU,
    0xd57950c6U,
    0xe9778028U,
    0x3df8c018U,
    0x6e37e01bU,
    0x9c1d503dU,
    0x2d06383dU,
    0xad845411U,
    0xec1a23bU,
    0x5920ab1dU,
    0x2715f3acU,
    0x130afbc5U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x28000000U,
    0xbc000000U,
    0xae000000U,
    0xef000000U,
    0x5a800000U,
    0x58400000U,
    0xc3200000U,
    0xcf100000U,
    0x3c080000U,
    0xed940000U,
    0x104a0000U,
    0xcb10000U,
    0x265e8000U,
    0x11384000U,
    0xdb0f6000U,
    0xdb173000U,
    0x6e091800U,
    0xc936400U,
    0x35cf9a00U,
    0x26741d00U,
    0x13dd280U,
    0xb30f6540U,
    0xc7111a20U,
    0x500c5d10U,
    0xcb92b2a8U,
    0xd3485564U,
    0xd030022aU,
    0x2d1b393fU,
    0x269f2883U,
    0xa359485fU,
    0x7eb950bcU,
    0x14cd1c69U,
    0xe3f7d293U,
    0x1bfe657cU,
    0x2c6f9a1bU,
    0xf1241d36U,
    0x5e15d2bdU,
    0x18b6540U,
    0xc3531a11U,
    0xd295d2dU,
    0x53863297U,
    0x21c1154fU,
    0x77e1e211U,
    0xbf744926U,
    0x50b95083U,
    0xbbcd1c6bU,
    0x1977d2a5U,
    0xd3be657aU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0x18000000U,
    0x3c000000U,
    0x6a000000U,
    0xdd000000U,
    0xdc800000U,
    0xf6c00000U,
    0xf9a00000U,
    0x9e100000U,
    0xfbd80000U,
    0x2e940000U,
    0xdf1e0000U,
    0x46370000U,
    0xff8e8000U,
    0xb22fc000U,
    0xa23b6000U,
    0xb2e29000U,
    0xe7f28800U,
    0xd7aef400U,
    0x7e7cfe00U,
    0x7a815900U,
    0x89c21e80U,
    0xbc20cfc0U,
    0xf2527e20U,
    0x5bbe9930U,
    0x74a17e98U,
    0xfa965fc4U,
    0x411ef626U,
    0xb5376d3fU,
    0xb80b0082U,
    0x376cc6f3U,
    0x90d98891U,
    0x951232c2U,
    0x505d769cU,
    0xb9d76be4U,
    0x1cf96837U,
    0xcf44a40dU,
    0x5fe3960dU,
    0x9376fd0aU,
    0x756908aaU,
    0xd9daf2e2U,
    0x779096acU,
    0x259e3be0U,
    0xff68009U,
    0x23abc03cU,
    0x507d601eU,
    0xe1819004U,
    0xfa420831U,
    0x77663414U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x78000000U,
    0xe4000000U,
    0xde000000U,
    0x8b000000U,
    0x44800000U,
    0xdcc00000U,
    0xeb200000U,
    0xb0700000U,
    0xb0e80000U,
    0x8f40000U,
    0xa62a0000U,
    0x2ed30000U,
    0x8ddf8000U,
    0xa078c000U,
    0xddcca000U,
    0xc7843000U,
    0xa443f800U,
    0x5de59400U,
    0xc653de00U,
    0x531f6d00U,
    0x1a5aed80U,
    0xcc3c5240U,
    0x1c2c5e20U,
    0x3bd7ad10U,
    0xa25e4db8U,
    0x883c626cU,
    0x922da63eU,
    0x78d53929U,
    0x7ad8138fU,
    0x6ef8cf4eU,
    0x2c0c6bafU,
    0x7219b5eU,
    0x52711595U,
    0x51edc672U,
    0xd175803bU,
    0x296bc030U,
    0x67b32002U,
    0x5f4cf010U,
    0x84475837U,
    0x4de5a406U,
    0xee522602U,
    0x7f1df903U,
    0x585cb38dU,
    0x7d38ff66U,
    0xdad9399U,
    0x28930f50U,
    0xd13f4b91U,
    0xfad6b79U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0xa8000000U,
    0x2c000000U,
    0xa000000U,
    0xd000000U,
    0xd800000U,
    0xf3400000U,
    0x49600000U,
    0x53300000U,
    0x2ae80000U,
    0x6ab40000U,
    0x67ae0000U,
    0x50d10000U,
    0xea1c8000U,
    0x7fc4000U,
    0x1c09e000U,
    0x83c21000U,
    0x8fa08800U,
    0xcd12fc00U,
    0x46b95200U,
    0xbc6a5700U,
    0xb765b80U,
    0xa9084840U,
    0x6a45d220U,
    0xaae61730U,
    0x7b77bba8U,
    0x910e5854U,
    0x4e435a0aU,
    0xdce1eb3bU,
    0xd47469aaU,
    0xbd8d4f57U,
    0xba866189U,
    0x6bc2f347U,
    0xc3a25398U,
    0x3712f473U,
    0xb3bb602bU,
    0x35ef5009U,
    0xde35e80bU,
    0xe76cac1cU,
    0x39f03a1bU,
    0x7ecabb2fU,
    0xabe7818dU,
    0xecf4e35fU,
    0x674cdba2U,
    0x76a10870U,
    0xde96b21fU,
    0x96fd4731U,
    0x838c538aU,
    0x983f449U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0xe8000000U,
    0xcc000000U,
    0xba000000U,
    0x1d000000U,
    0x6c800000U,
    0x37400000U,
    0x6e00000U,
    0x9ad00000U,
    0x3dd80000U,
    0xcc540000U,
    0x579a0000U,
    0x46330000U,
    0x4a0f8000U,
    0x950c4000U,
    0xd08de000U,
    0x454b9000U,
    0x47ea0800U,
    0x345d7400U,
    0xa394f200U,
    0xa83f2300U,
    0xe907f480U,
    0x28125c0U,
    0x54437220U,
    0xb1676310U,
    0x8e1014b8U,
    0xb1f9b5ecU,
    0xb1a6fa1aU,
    0xdc365723U,
    0x47090696U,
    0x418d06ebU,
    0x73cb0681U,
    0xddaa06eeU,
    0xb63e86b7U,
    0xc20546ddU,
    0x2904e6b6U,
    0xa28696f1U,
    0x4410e98U,
    0xe96772ffU,
    0xaa12748aU,
    0xc7fe65d3U,
    0x16a11237U,
    0xadb0b323U,
    0x1ccffcbaU,
    0x702b51e8U,
    0xeffa0028U,
    0x7aa30016U,
    0x47b7802fU,
    0x59c8400aU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x18000000U,
    0x2c000000U,
    0x62000000U,
    0x4b000000U,
    0xbd800000U,
    0x5cc00000U,
    0xb4200000U,
    0x31900000U,
    0xee880000U,
    0x7e140000U,
    0x414e0000U,
    0xebb70000U,
    0x86188000U,
    0xce1fc000U,
    0x3a1de000U,
    0x141ad000U,
    0xed1cb800U,
    0x399bc00U,
    0xcede9e00U,
    0x443d7d00U,
    0x8a8c0280U,
    0xe8105140U,
    0x244e1e20U,
    0xaf36bd30U,
    0x345fe298U,
    0xb77d8174U,
    0x816a2626U,
    0x3420c13bU,
    0xf1949c80U,
    0x8e8e2c66U,
    0xae149c89U,
    0x594e2c6cU,
    0xc7b49c8dU,
    0xe41e2c5aU,
    0x851c9caaU,
    0x879a2c47U,
    0x48da9c9bU,
    0x59392c6bU,
    0x320c1c93U,
    0x2051ec66U,
    0x3a297cbaU,
    0xcbc4fc79U,
    0x3a02485U,
    0xa253906aU,
    0x612c02b2U,
    0xe405177U,
    0xa3661e37U,
    0x6c72bd2aU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x98000000U,
    0xb4000000U,
    0xb2000000U,
    0xf9000000U,
    0x2d800000U,
    0xa2400000U,
    0xd3e00000U,
    0x39300000U,
    0xaf980000U,
    0x9eb40000U,
    0xd0da0000U,
    0x6ad50000U,
    0xd6ac8000U,
    0x1129c000U,
    0x9e6d2000U,
    0x220e5000U,
    0x87be800U,
    0x7581bc00U,
    0x76463a00U,
    0x91e48300U,
    0x88316380U,
    0xae19b7c0U,
    0x3af2ba20U,
    0x48394310U,
    0x87664388U,
    0xf6f2e7f4U,
    0xfe3dd206U,
    0xa4653f3dU,
    0xde7759a4U,
    0xeffd34caU,
    0x2ac3d98dU,
    0x5220f4d5U,
    0xf94f9b0U,
    0x47cba4d4U,
    0x54db91aeU,
    0xf0d7d8c6U,
    0xf3aa8ba2U,
    0x72a80be3U,
    0x5b2c8037U,
    0x2369c03cU,
    0xf58d202eU,
    0x7f3e5013U,
    0x8de3e829U,
    0xa635bc0bU,
    0x391c3a36U,
    0xa0718309U,
    0xa0fde3b5U,
    0x244077e9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0x38000000U,
    0x74000000U,
    0x2000000U,
    0x4d000000U,
    0x9c800000U,
    0xdfc00000U,
    0x40e00000U,
    0x30300000U,
    0x9380000U,
    0x82b40000U,
    0x7dfe0000U,
    0x79d50000U,
    0x7f8d8000U,
    0x3d4cc000U,
    0x692a2000U,
    0xeadad000U,
    0xa5014800U,
    0x6081b400U,
    0xb1c47600U,
    0xbe71300U,
    0xdbb76780U,
    0x737f72c0U,
    0x8311f620U,
    0x9fefd330U,
    0x9cbb4788U,
    0x1bf4a2dcU,
    0x8bdb3e2eU,
    0xdf83a72dU,
    0x98469188U,
    0x1ca0a1cfU,
    0x3692b189U,
    0x76af71daU,
    0x529e79b8U,
    0xc9a205f3U,
    0x6e102f83U,
    0xb36fc6e6U,
    0x5b7e0029U,
    0x5f150030U,
    0x81ed803cU,
    0xefbcc00aU,
    0x8472203cU,
    0xf39ed011U,
    0x93274803U,
    0xe6d0b41fU,
    0x260ff61cU,
    0x120ad33dU,
    0xf00ec78dU,
    0xed0c62f9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0x28000000U,
    0xc000000U,
    0x2000000U,
    0x39000000U,
    0xa5800000U,
    0x1b400000U,
    0x3ca00000U,
    0xb2500000U,
    0xd5780000U,
    0x61d40000U,
    0xe13e0000U,
    0xc7f50000U,
    0x2aa98000U,
    0x83ec4000U,
    0x4e4de000U,
    0x859a5000U,
    0x91a52800U,
    0xd1d4fc00U,
    0x293eb200U,
    0x1bf67900U,
    0xf0abb080U,
    0x9eeedac0U,
    0xe5cf3220U,
    0xa5de3930U,
    0x31805088U,
    0xdd458afcU,
    0xdba59a2aU,
    0xf4d78533U,
    0x76bc8288U,
    0xedb4e3f2U,
    0x4f896283U,
    0xfab3f5U,
    0x92924aa7U,
    0xf19b4feeU,
    0x87a578beU,
    0xeed176d1U,
    0x4bbb2815U,
    0xb631fc1cU,
    0x47cf3214U,
    0xacde390fU,
    0x9c005087U,
    0x3a058af7U,
    0xcd059a1dU,
    0x73878518U,
    0x44482baU,
    0xae20e3d8U,
    0x379762baU,
    0x6e1fb3c5U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0x88000000U,
    0xc000000U,
    0x46000000U,
    0x63000000U,
    0x7a800000U,
    0xf2c00000U,
    0x68e00000U,
    0x9cd00000U,
    0x57980000U,
    0x5540000U,
    0x9fda0000U,
    0xbf750000U,
    0x9bed8000U,
    0xfc394000U,
    0xc521e000U,
    0x20b57000U,
    0x868df800U,
    0x7eae2400U,
    0x139eae00U,
    0x2f55c300U,
    0x6aded880U,
    0xdef7f7c0U,
    0x97a92e20U,
    0x2c198310U,
    0x1412b888U,
    0xf27bc7fcU,
    0x5c053602U,
    0x7e02d713U,
    0xc701ee99U,
    0xc08020e4U,
    0x5fc5409cU,
    0x7d60e3efU,
    0x52961823U,
    0xaebe5413U,
    0x8be6d621U,
    0xa656a702U,
    0x855b968eU,
    0x9db244d3U,
    0x8b0f8e9cU,
    0xe4ed10caU,
    0xd8bed8baU,
    0xe7f7d0U,
    0xd12e02U,
    0x499d8329U,
    0x1250b8bcU,
    0x275ac7caU,
    0x64b2b637U,
    0xac8e970fU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x48000000U,
    0x94000000U,
    0x72000000U,
    0xfb000000U,
    0xa2800000U,
    0x4f400000U,
    0x42200000U,
    0x52100000U,
    0xc7580000U,
    0x3940000U,
    0x1e9e0000U,
    0x2cf70000U,
    0x8daf8000U,
    0xc73d4000U,
    0xe0272000U,
    0x1114d000U,
    0xd9d9f800U,
    0x9ad14400U,
    0x9dbd4200U,
    0xb3673500U,
    0xd5332280U,
    0x32cfa7c0U,
    0x4c0cc220U,
    0xe7ed7530U,
    0xd29b8298U,
    0x12f637ccU,
    0x88aa1a32U,
    0xbce115U,
    0x58e13884U,
    0xa3f746f2U,
    0xad2bfa9aU,
    0x5b7933c6U,
    0xa481f834U,
    0x36454416U,
    0x33a34213U,
    0x1bd0353aU,
    0xf03ca299U,
    0x7ca2e7c4U,
    0x5b53e206U,
    0x5c7da52cU,
    0x70047a9fU,
    0xe80473c4U,
    0xc406d806U,
    0xa019416U,
    0x2702ba2dU,
    0x44857111U,
    0xc647e09dU,
    0x1ba6d2e6U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xc8000000U,
    0xec000000U,
    0x2000000U,
    0x77000000U,
    0x2f800000U,
    0x3c00000U,
    0x63600000U,
    0x94700000U,
    0xc9f80000U,
    0x14f40000U,
    0x51ba0000U,
    0x82550000U,
    0x7ad8000U,
    0xa15b4000U,
    0x6fe6e000U,
    0xde355000U,
    0x105a5800U,
    0x4964ac00U,
    0x9f708600U,
    0xfc7c1700U,
    0xb4349580U,
    0x3b5bd0c0U,
    0x8ce70620U,
    0xf7b25710U,
    0x7a9ff588U,
    0xb885c0fcU,
    0xbc43be12U,
    0x78a7ab2bU,
    0x23172b88U,
    0x7b8c7be1U,
    0x44082d99U,
    0x1fca2cebU,
    0x88add830U,
    0xcadaec14U,
    0x5223e603U,
    0xf0560702U,
    0xf8aa2dbeU,
    0x42db2ce3U,
    0x9e225818U,
    0x250ac20U,
    0x47aa8614U,
    0x8159170eU,
    0x9fe1159dU,
    0x163490e1U,
    0xfc5be63aU,
    0x4b620720U,
    0xe8702dbfU,
    0xd3fe2ce8U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0x8000000U,
    0xc4000000U,
    0xea000000U,
    0x15000000U,
    0x17800000U,
    0x2cc00000U,
    0x6be00000U,
    0x5500000U,
    0x34280000U,
    0xf3d40000U,
    0x5d6e0000U,
    0x1df50000U,
    0x525c8000U,
    0x334d4000U,
    0xd2c6a000U,
    0x5ce01000U,
    0xe3d50800U,
    0xb56d2c00U,
    0xa9f3ae00U,
    0xe05c9300U,
    0xea4bb480U,
    0xeb426a40U,
    0x8f212e20U,
    0x8ab4d330U,
    0x8b799488U,
    0xdafb3a54U,
    0x7d3c8622U,
    0x865cef01U,
    0x314bb2b2U,
    0x73c3c551U,
    0xf9643ca7U,
    0x1613064aU,
    0xe70ea008U,
    0x7f641020U,
    0xbd130802U,
    0x278c2c12U,
    0xc5212e35U,
    0xfb4d332U,
    0xb4f994a6U,
    0x623b3a7eU,
    0xf4dc861cU,
    0x520cef2aU,
    0xf8e3b2b3U,
    0xb9d7c548U,
    0xd86a3ca6U,
    0x22760655U,
    0xea9a2013U,
    0xbaad503dU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x48000000U,
    0x94000000U,
    0x26000000U,
    0xdd000000U,
    0xe3800000U,
    0xf3400000U,
    0x47600000U,
    0x59300000U,
    0x19880000U,
    0x1bb40000U,
    0xab4e0000U,
    0xcc930000U,
    0x679f8000U,
    0x66e94000U,
    0xbe872000U,
    0xd0c27000U,
    0x5424c800U,
    0x2e55ec00U,
    0x8bf9200U,
    0x963cb700U,
    0x96f88b80U,
    0xbade0340U,
    0x488e1220U,
    0xf236f730U,
    0x9f082bb8U,
    0x3771334cU,
    0x9d6bfa32U,
    0x61c66b15U,
    0x8da2f1b1U,
    0x5212287bU,
    0x675c438aU,
    0x4acbef69U,
    0x4cd18000U,
    0x6a7a401dU,
    0x3918a014U,
    0x862b3013U,
    0xa2a3e820U,
    0x6a979c37U,
    0x7a9b5a2eU,
    0x65695b14U,
    0x7dc719acU,
    0xdfa2b46dU,
    0x991699bbU,
    0x11d8f472U,
    0xce0e3999U,
    0xdef3c458U,
    0xa92dd189U,
    0x9a245856U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0x88000000U,
    0x24000000U,
    0x86000000U,
    0xd5000000U,
    0x3f800000U,
    0xddc00000U,
    0xa2600000U,
    0xb6100000U,
    0xe8d80000U,
    0xa0940000U,
    0x909a0000U,
    0x350000U,
    0x3c6d8000U,
    0x67b94000U,
    0xc500a000U,
    0x27837000U,
    0xa1c10800U,
    0x78660c00U,
    0x49152600U,
    0xa059ff00U,
    0x11d5c680U,
    0x5bf01c0U,
    0xf602a620U,
    0x1d05bf10U,
    0x3b80e688U,
    0xabc131fcU,
    0xff610e02U,
    0xad91c319U,
    0xb31b48a9U,
    0xd7f682c9U,
    0x190ece8dU,
    0x352c0deeU,
    0x3e9a0021U,
    0x41350034U,
    0x2ded801fU,
    0xbb79402aU,
    0x56e0a00fU,
    0xbd537011U,
    0x52f9081bU,
    0x66220c33U,
    0xacb72632U,
    0x6b28ff13U,
    0xf79a46b5U,
    0x74b741c2U,
    0x772d861dU,
    0xfd9e8f19U,
    0xf3b6ce8fU,
    0x93a80dfaU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0xd8000000U,
    0x6c000000U,
    0x52000000U,
    0x2b000000U,
    0x55800000U,
    0x9bc00000U,
    0x56600000U,
    0x18700000U,
    0x98a80000U,
    0x74f40000U,
    0x9dee0000U,
    0x1cd50000U,
    0x40388000U,
    0x7e6bc000U,
    0x8a17e000U,
    0x76db5000U,
    0x159b800U,
    0x799a7c00U,
    0xee7b2600U,
    0x83cc6f00U,
    0xec049580U,
    0x9201a240U,
    0xb05a620U,
    0x8586af30U,
    0x43c5f588U,
    0x3a643274U,
    0x4a737e16U,
    0xb3ac432bU,
    0x21708b9cU,
    0x629717eU,
    0x4ab57583U,
    0x584bf24dU,
    0xe6c29e29U,
    0xfee61308U,
    0xeb37b38dU,
    0x1d89cd54U,
    0x39a733b0U,
    0x90160d60U,
    0x9ded39fU,
    0x9ad85d6aU,
    0x935feb9bU,
    0x7299e165U,
    0x6bfb2da9U,
    0xc00ade4fU,
    0xd6600007U,
    0xd8700014U,
    0xb8a80026U,
    0xa4f40027U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x88000000U,
    0x44000000U,
    0xf6000000U,
    0xef000000U,
    0x47800000U,
    0x38c00000U,
    0x8fe00000U,
    0x20700000U,
    0xcc780000U,
    0x96f40000U,
    0x283e0000U,
    0xe8550000U,
    0x4c688000U,
    0xf6394000U,
    0x5356a000U,
    0x55ecf000U,
    0xb5fed800U,
    0xe5354c00U,
    0xe65fde00U,
    0x47e05300U,
    0x8474cd80U,
    0xca7838c0U,
    0x21f15e20U,
    0xa3b81330U,
    0x6294ed88U,
    0xda8888f4U,
    0x7ec9a602U,
    0xe06cef21U,
    0x743d4bb5U,
    0x325467cfU,
    0xe16ced93U,
    0xf0bc88efU,
    0xf17a636U,
    0x1749ef37U,
    0x3badcb88U,
    0x2e5927ceU,
    0xe3e44d96U,
    0x827578e8U,
    0x7d79fe24U,
    0xaa71e329U,
    0x297ab586U,
    0xf47084ccU,
    0x527ed811U,
    0xcdf54c33U,
    0xc1bfde30U,
    0xf390531aU,
    0x360ccd85U,
    0xf78c38dbU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x38000000U,
    0xfc000000U,
    0xc2000000U,
    0x27000000U,
    0x32800000U,
    0x8ac00000U,
    0x84200000U,
    0x17d00000U,
    0xf6980000U,
    0x44540000U,
    0x77de0000U,
    0x2d330000U,
    0x136c8000U,
    0x341ac000U,
    0x86976000U,
    0x87fed000U,
    0x6ce75800U,
    0xc4f2c400U,
    0x5bcc4e00U,
    0xac0ac500U,
    0xb1a80080U,
    0x5439d9c0U,
    0x1f46ce20U,
    0xfc670530U,
    0xf935e0b8U,
    0x356ac9d4U,
    0x9d1c762eU,
    0x3916d10fU,
    0xdcbb1688U,
    0x6e85d8ddU,
    0x38c480a2U,
    0x2b2319edU,
    0xb151ae09U,
    0x7a59d528U,
    0xd972b887U,
    0xb7880dd8U,
    0x44e83822U,
    0x6b581408U,
    0x5af5163dU,
    0xcecb0123U,
    0x3188ce9fU,
    0x9de9dcf2U,
    0x5cd9ae93U,
    0x95b00cd1U,
    0xcaac7683U,
    0xd4bb08d6U,
    0x3a83d899U,
    0xdec1ddfcU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x18000000U,
    0x6c000000U,
    0xaa000000U,
    0x81000000U,
    0x50800000U,
    0xdac00000U,
    0xf3200000U,
    0xdd500000U,
    0x15b80000U,
    0xf2d40000U,
    0x8fe0000U,
    0x18b10000U,
    0x51ce8000U,
    0x1b39c000U,
    0xaf11e000U,
    0x205e9000U,
    0xdea6e800U,
    0x25119400U,
    0x715c7e00U,
    0x3627ad00U,
    0x83d7a480U,
    0x307fd8c0U,
    0x674fe20U,
    0x6ceb6d30U,
    0x996ec4a8U,
    0xad2d88c4U,
    0x73cb7626U,
    0x2e38a92bU,
    0x99953282U,
    0xc19de1e4U,
    0xe40124b2U,
    0x1e0218ddU,
    0xe7031e1eU,
    0x6b80fd23U,
    0x1340acbfU,
    0x1560dcd5U,
    0x5eb66832U,
    0xbac9543aU,
    0x6abb1e07U,
    0x3554fd28U,
    0x11beac87U,
    0x9cd1dcd4U,
    0x47f8e810U,
    0x17309414U,
    0x9c8afe0cU,
    0x495a6d0dU,
    0x8a2044bbU,
    0x91d448e9U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xa8000000U,
    0xec000000U,
    0x32000000U,
    0xb7000000U,
    0x11800000U,
    0x6400000U,
    0x7b600000U,
    0x5b300000U,
    0x5a80000U,
    0x1d740000U,
    0xdece0000U,
    0x76470000U,
    0x23648000U,
    0x2f364000U,
    0x23ade000U,
    0xdc70d000U,
    0xa64cd800U,
    0xe4861c00U,
    0xf8c10a00U,
    0x18a51700U,
    0x170b80U,
    0xe4ff2640U,
    0x3b6cea20U,
    0x54d5c730U,
    0x6e5bd398U,
    0xbc793a54U,
    0x59ade00aU,
    0x1770d00bU,
    0x4dccd814U,
    0xe9c61c39U,
    0x8210a2eU,
    0x1ed5173aU,
    0x5d5f0b92U,
    0x13fb267bU,
    0x8aeaea25U,
    0x6296c736U,
    0xed3953b1U,
    0xa34c7a5fU,
    0x82028003U,
    0x8f054007U,
    0x3587602bU,
    0x88419013U,
    0x5665b80fU,
    0x11b08c08U,
    0x20683224U,
    0xd757db0aU,
    0xef1cd9acU,
    0x2e9f2d46U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x8000000U,
    0x3c000000U,
    0xfe000000U,
    0xed000000U,
    0xae800000U,
    0x54400000U,
    0xba600000U,
    0x94f00000U,
    0xf4180000U,
    0xf0b40000U,
    0xf67e0000U,
    0x20410000U,
    0x18608000U,
    0xf7f54000U,
    0x6f9f6000U,
    0x3e73b000U,
    0xda588800U,
    0x1dd38400U,
    0x3b0d4a00U,
    0x999e700U,
    0x87736580U,
    0xc6dec3c0U,
    0x82922a20U,
    0xf6ea5730U,
    0xa5abeda8U,
    0x8e4d47f4U,
    0x5b7f6022U,
    0xeec3b03fU,
    0x3c208817U,
    0x9597840fU,
    0xcf6b4a29U,
    0x86ce71aU,
    0x39ede591U,
    0xa82a83deU,
    0xe18dca36U,
    0xffdca719U,
    0x5e1405bbU,
    0xc9a973d9U,
    0xb84ca219U,
    0x2a7cd33eU,
    0x6e4027b1U,
    0x8d64e0c8U,
    0xbd75e5b9U,
    0xf1de83d0U,
    0xab13ca0cU,
    0xf32da717U,
    0xfe0c85abU,
    0x8b1833e4U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0xf8000000U,
    0x14000000U,
    0xda000000U,
    0x33000000U,
    0xcc800000U,
    0xb7400000U,
    0xae600000U,
    0xc8700000U,
    0xbd080000U,
    0xf7340000U,
    0x6f6e0000U,
    0xa9410000U,
    0x7f648000U,
    0x83f2c000U,
    0x4cca2000U,
    0x25151000U,
    0xd1ff3800U,
    0xb179b400U,
    0xa83ca200U,
    0x1f5bd700U,
    0x80aed980U,
    0xaa640b40U,
    0x8a768220U,
    0x6a0ec730U,
    0x49b1e1a8U,
    0x72dbf54U,
    0x5a2201eU,
    0xe9511035U,
    0x6f19381eU,
    0x84cb418U,
    0xee56220dU,
    0xe9981728U,
    0xd408799dU,
    0xc8b7db7eU,
    0xb4ad9a3cU,
    0xe0636320U,
    0xe176fb98U,
    0xe28d1c4cU,
    0xdcf27baeU,
    0x8e4fdc68U,
    0xdb505bb6U,
    0xd01ecc4bU,
    0xfac963aeU,
    0x18127840U,
    0x547f4192U,
    0xa9ba6f51U,
    0x859f3808U,
    0xda09b43bU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0x68000000U,
    0x8c000000U,
    0xe2000000U,
    0x19000000U,
    0x83800000U,
    0x61400000U,
    0x25600000U,
    0x9d900000U,
    0x37b80000U,
    0xf0d40000U,
    0xb0da0000U,
    0xd4410000U,
    0x74e18000U,
    0x2dd64000U,
    0x6d5fe000U,
    0x86005000U,
    0x3706a800U,
    0xd8860c00U,
    0x8bc04200U,
    0xafa67700U,
    0x55b38680U,
    0x6d4f86c0U,
    0x43ffa220U,
    0xf4362710U,
    0xf50d2ea8U,
    0x351d8adcU,
    0x10e5e03aU,
    0x3d15033U,
    0x365f2810U,
    0x6c844c1aU,
    0xbdc5a23aU,
    0x10a7273bU,
    0xd134aeb1U,
    0x1c8fcae1U,
    0x1158000dU,
    0x4c040034U,
    0x202000dU,
    0xc905000eU,
    0x9b83800aU,
    0x85434014U,
    0x4b64600fU,
    0x66971012U,
    0xad38c830U,
    0x12101c25U,
    0xf4f90a30U,
    0x6cb02b2dU,
    0xdecd6c95U,
    0xeabbfdd1U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x78000000U,
    0xbc000000U,
    0xc2000000U,
    0x49000000U,
    0x1b800000U,
    0x12c00000U,
    0xe200000U,
    0xa8700000U,
    0x1ff80000U,
    0x3eb40000U,
    0x4fde0000U,
    0x4dc30000U,
    0xcaa38000U,
    0xbe34c000U,
    0xe7986000U,
    0xdea1d000U,
    0xc831a800U,
    0x4899cc00U,
    0x42235e00U,
    0x32756d00U,
    0xbaff3580U,
    0x2f371ec0U,
    0xd01b3e20U,
    0x2664bd30U,
    0xfb169db8U,
    0xfd6ad2ecU,
    0x781e603eU,
    0x9266d01fU,
    0x1d142808U,
    0x7a6a0c3eU,
    0x909ebe18U,
    0x6e277d2bU,
    0xd8737db3U,
    0x87fcc2f8U,
    0x32b42821U,
    0xf5da0c3bU,
    0xb8c6be28U,
    0x13237d15U,
    0xe5f57dabU,
    0xf23bc2cfU,
    0x6411a83aU,
    0xd9e9cc06U,
    0xde5b5e3eU,
    0x12016d15U,
    0x410135b4U,
    0x3f841efeU,
    0xdcc0be0bU,
    0xfd247d34U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x18000000U,
    0x34000000U,
    0x8a000000U,
    0xd3000000U,
    0x90800000U,
    0x78c00000U,
    0xa6e00000U,
    0x31300000U,
    0x15280000U,
    0xc5f40000U,
    0xddce0000U,
    0x95c50000U,
    0xaf668000U,
    0x2a71c000U,
    0x4b892000U,
    0x32e09000U,
    0x6b313800U,
    0xfe29d400U,
    0x7171de00U,
    0x370f7300U,
    0xd426ef80U,
    0x84d74e40U,
    0x9498fe20U,
    0xb81fe330U,
    0xa6dfd788U,
    0x4d3a9a44U,
    0xa58f2026U,
    0x93e1903dU,
    0x2cb1b82aU,
    0x94691430U,
    0x50507e22U,
    0x8e5b2313U,
    0xd1f8778bU,
    0xb16fca78U,
    0x55d13801U,
    0x3b19d41fU,
    0xce59de16U,
    0x31fb732dU,
    0x8168efa8U,
    0x5dd24e76U,
    0x171e7e27U,
    0x705e232dU,
    0x68fef79dU,
    0xc2ee0a6fU,
    0xb5901801U,
    0xc93d4430U,
    0xe78ee600U,
    0x4ce3a71fU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x88000000U,
    0x1c000000U,
    0x4a000000U,
    0x4d000000U,
    0xfc800000U,
    0x22400000U,
    0x8b200000U,
    0x15900000U,
    0x5cf80000U,
    0xadd40000U,
    0x62da0000U,
    0xb0c70000U,
    0x7a638000U,
    0x19374000U,
    0xa0a96000U,
    0x61ca1000U,
    0x937c8800U,
    0x7810b400U,
    0xe83bfa00U,
    0x19b17d00U,
    0xb8ee9e80U,
    0x4fef4ec0U,
    0x566a9a20U,
    0xc4af6d10U,
    0x9fc81698U,
    0xd878faccU,
    0x3b92e002U,
    0xeff95017U,
    0x1a57e80aU,
    0x319a41fU,
    0x3c26f21dU,
    0x6c15890fU,
    0xfe3d84b0U,
    0x7eb063d6U,
    0x39686c88U,
    0x192ac7f3U,
    0xe30f1e82U,
    0x101b0ef5U,
    0x9ba27a13U,
    0xd5d13d36U,
    0xd6dc7e98U,
    0xb6c61eecU,
    0xc5657215U,
    0x3cb2c914U,
    0x286ce485U,
    0xcfae73ccU,
    0xbc4ee482U,
    0xefbd73fcU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x18000000U,
    0x2c000000U,
    0xbe000000U,
    0xc7000000U,
    0xe2800000U,
    0x1b400000U,
    0x3be00000U,
    0xe8d00000U,
    0x7780000U,
    0x31940000U,
    0x1d9a0000U,
    0xf0c50000U,
    0x8d248000U,
    0x3ef1c000U,
    0x8d0fe000U,
    0x84d93000U,
    0xca215800U,
    0x9c71e400U,
    0xb6496200U,
    0x6f3c8b00U,
    0x3af03c80U,
    0xb70e74c0U,
    0x39de8220U,
    0xb5a1bb10U,
    0x28b36488U,
    0x216e90f4U,
    0x3cc96026U,
    0xd079f01bU,
    0xab103827U,
    0xdadc1405U,
    0xbd235a3eU,
    0x96f59f0dU,
    0x490fe6a1U,
    0xdede2bcbU,
    0x87248499U,
    0x2bf3a0daU,
    0x368a3821U,
    0x6a191422U,
    0x1007da2cU,
    0x78045f3cU,
    0xdc00068dU,
    0x76071be7U,
    0xf305dca6U,
    0x708244cfU,
    0x62435a10U,
    0x1e659f22U,
    0x1117e6bdU,
    0x27da2bc3U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x98000000U,
    0xd4000000U,
    0x6000000U,
    0x91000000U,
    0xb3800000U,
    0xd7c00000U,
    0x2b600000U,
    0xef900000U,
    0x5f180000U,
    0x3a540000U,
    0x2b7a0000U,
    0xab430000U,
    0xaa258000U,
    0xb6b3c000U,
    0xb5292000U,
    0x5e38f000U,
    0xa4642800U,
    0x5910ac00U,
    0xdd5ab600U,
    0xa7f38700U,
    0xc409a280U,
    0x180e33c0U,
    0xe20b9620U,
    0xd90f7710U,
    0xd98f8ab8U,
    0xb2c99ffcU,
    0xb8eea006U,
    0x1e5c3025U,
    0xc3728839U,
    0xd1489c18U,
    0xa72a3e0aU,
    0x693c1b00U,
    0x96e41c8bU,
    0x5556e8dfU,
    0x6df92a9bU,
    0xb101afebU,
    0xe3862818U,
    0x5fc7ac3dU,
    0x97653623U,
    0xa593472aU,
    0x1c1d0280U,
    0x1ed103ecU,
    0xde3c9e0eU,
    0xe4642b3aU,
    0xb9149484U,
    0x2d5974ccU,
    0x3ff49483U,
    0x100974c0U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0xc8000000U,
    0xfc000000U,
    0xfe000000U,
    0x81000000U,
    0x3e800000U,
    0x30c00000U,
    0x13600000U,
    0x8a300000U,
    0xfec80000U,
    0xf0f40000U,
    0x2eaa0000U,
    0xdb430000U,
    0xaf258000U,
    0xfe914000U,
    0xd4186000U,
    0xbc4db000U,
    0x7e35e800U,
    0x5ccb0400U,
    0x9ff10e00U,
    0xa92aaf00U,
    0xe1077780U,
    0x8e82fd40U,
    0x18c16e20U,
    0x1f631f10U,
    0xbc309fb8U,
    0x83cef97cU,
    0x3077e012U,
    0x9f6ff02fU,
    0xf6a00807U,
    0x15d3f41cU,
    0x133e863dU,
    0xaedb1b33U,
    0x6c2c11bbU,
    0x55811642U,
    0xdd42f790U,
    0xba23bd60U,
    0xd2110e17U,
    0xa3daaf38U,
    0x24af779eU,
    0xf846fd40U,
    0xfea36e13U,
    0x49d41f16U,
    0xfd3f1fbcU,
    0x17dcb978U,
    0x66aa000bU,
    0x67430004U,
    0xb125801bU,
    0x8f914036U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0x48000000U,
    0xc4000000U,
    0xf2000000U,
    0x73000000U,
    0x24800000U,
    0x1dc00000U,
    0xe7600000U,
    0x2d700000U,
    0x1780000U,
    0x5eb40000U,
    0x4b1a0000U,
    0x22430000U,
    0x6a248000U,
    0x5dd54000U,
    0xe468a000U,
    0x2839d000U,
    0x4c13b800U,
    0x750bb400U,
    0x644ef600U,
    0xcaecdf00U,
    0x5afce580U,
    0x11f100c0U,
    0xd0be5620U,
    0x6fd10f10U,
    0x376d5db8U,
    0x3cbdb4f4U,
    0xc9d62032U,
    0x1e6f9021U,
    0x7f3f9804U,
    0x4a972428U,
    0x53cdee1bU,
    0x63aabb36U,
    0x85bab85U,
    0xcfe56bd7U,
    0x46b0c5a9U,
    0x871d90c0U,
    0xf4454e33U,
    0x3b236b27U,
    0x425013b0U,
    0x192adfc5U,
    0x209c33a5U,
    0xf5064fffU,
    0x7d872ba4U,
    0x26442bf5U,
    0x3822659aU,
    0x1ed740e1U,
    0x58ea760aU,
    0xb9f99f0cU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x68000000U,
    0x2c000000U,
    0x8e000000U,
    0x7000000U,
    0x16800000U,
    0x93c00000U,
    0x10a00000U,
    0x96700000U,
    0xadd80000U,
    0xffb40000U,
    0x607e0000U,
    0xd2470000U,
    0x3de08000U,
    0x59104000U,
    0x750fe000U,
    0xab1b3000U,
    0x8935800U,
    0x5049c400U,
    0x987d2600U,
    0x3647f100U,
    0x2fe70d80U,
    0x68157bc0U,
    0x208ac620U,
    0x8c58c130U,
    0x2f25598U,
    0xc71fbff4U,
    0x2691603aU,
    0xe74c703bU,
    0x36fc383bU,
    0xe182b435U,
    0x9d419e1fU,
    0x7765052fU,
    0x9cd173a7U,
    0xf6af4ee4U,
    0xe1eeed8eU,
    0x938d4bebU,
    0x20df1e24U,
    0xec324515U,
    0xdc3e13a6U,
    0xf8a43ed6U,
    0x7a72559cU,
    0x43dfbfe4U,
    0x28b16013U,
    0x1efc7023U,
    0x6d843838U,
    0xa346b419U,
    0xc8679e1fU,
    0xce56051eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0xb8000000U,
    0x74000000U,
    0xae000000U,
    0x59000000U,
    0xfa800000U,
    0x11c00000U,
    0xdbe00000U,
    0x45f00000U,
    0x15780000U,
    0xd9340000U,
    0xa1a0000U,
    0x4c050000U,
    0x9a048000U,
    0xd704c000U,
    0x73852000U,
    0x5340b000U,
    0xbe216800U,
    0x3012a400U,
    0x98a4600U,
    0x36cd0b00U,
    0xc2eca180U,
    0x9df974c0U,
    0x93f56620U,
    0x5878bb10U,
    0x7db14988U,
    0x2adb10f4U,
    0xa160000eU,
    0x1430000dU,
    0xee980023U,
    0x4cc40022U,
    0xa7620010U,
    0xe1310019U,
    0x3e1e801dU,
    0xc201c007U,
    0x1301a03bU,
    0x95847032U,
    0x16444834U,
    0x26a21402U,
    0xa2d32e23U,
    0xdfebaf03U,
    0xc17ce79eU,
    0xe7317fc7U,
    0xcb1d47a9U,
    0x12850fe9U,
    0x9dc10fb6U,
    0x21e31bfdU,
    0x62f02199U,
    0xef9b4d2U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0x68000000U,
    0x1c000000U,
    0x1a000000U,
    0xb7000000U,
    0x1a800000U,
    0xedc00000U,
    0xeaa00000U,
    0xf9700000U,
    0xd3980000U,
    0x4fb40000U,
    0x11ba0000U,
    0x10030000U,
    0x58018000U,
    0x94044000U,
    0x76036000U,
    0xc5039000U,
    0xb1855800U,
    0xed400400U,
    0xb062d600U,
    0x9578100U,
    0xc728a580U,
    0x768dc9c0U,
    0xa77bb620U,
    0xd2271110U,
    0x7b47db8U,
    0xddbd8ddcU,
    0xf200003aU,
    0xeb000017U,
    0xe080003eU,
    0x2ac00031U,
    0x9820001cU,
    0x8b0003cU,
    0x2338003cU,
    0x1c40013U,
    0xd8a20032U,
    0xb2770028U,
    0xa31b803eU,
    0x7d77403aU,
    0xfd9ae022U,
    0x1eb3d036U,
    0xd63c3819U,
    0x38409435U,
    0x59e60e3aU,
    0x7013c51eU,
    0x14913b1U,
    0xbad9d8f3U,
    0xd1d64bbdU,
    0x49eadcf6U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x38000000U,
    0x4000000U,
    0x36000000U,
    0x2f000000U,
    0x1e800000U,
    0x5400000U,
    0x72e00000U,
    0x3c100000U,
    0x22180000U,
    0xaa540000U,
    0xbc7e0000U,
    0x6a030000U,
    0x6d038000U,
    0xdf86c000U,
    0x80c5a000U,
    0x67207000U,
    0x60b15800U,
    0x44694400U,
    0x9e1d5200U,
    0x58516300U,
    0x4578ca80U,
    0xeb8407c0U,
    0x4ec6f220U,
    0xac221330U,
    0xf83212b8U,
    0x562f83ecU,
    0xf678002eU,
    0x57040031U,
    0x3a860035U,
    0x63470027U,
    0xd5e58009U,
    0x1e91c000U,
    0x15582011U,
    0xc1b5b004U,
    0xb1ef780fU,
    0x53dbf42bU,
    0xb0f7aa33U,
    0x2d0b5715U,
    0xf4cf4093U,
    0x826ee0ddU,
    0xa918caafU,
    0x12d407f3U,
    0xfe3ef21bU,
    0x8a66133aU,
    0x5e5412b1U,
    0x927883e2U,
    0x1105802aU,
    0xcd81c011U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0x8000000U,
    0x6c000000U,
    0x2000000U,
    0x5000000U,
    0x1f800000U,
    0x27400000U,
    0xa2200000U,
    0x41900000U,
    0xb8280000U,
    0x41d40000U,
    0x908e0000U,
    0x90810000U,
    0x91c48000U,
    0x38e54000U,
    0x91f1e000U,
    0x4c185000U,
    0xf32f4800U,
    0xc952ec00U,
    0x54a1200U,
    0x36666d00U,
    0x5f37f680U,
    0xa7ef040U,
    0x341df220U,
    0x72b3d30U,
    0xf523ea8U,
    0x3e485c54U,
    0x26e28022U,
    0xef0402bU,
    0x929b6028U,
    0xf5ec1015U,
    0x43322805U,
    0x907bfc12U,
    0xf51aba28U,
    0x26add111U,
    0x27162ca9U,
    0xeaef3159U,
    0x5db1f6acU,
    0x36bbf044U,
    0x2f7f7202U,
    0xbb9b7d3aU,
    0x88695eaaU,
    0x91744c78U,
    0x75d8a83aU,
    0xf7cfbc16U,
    0x52a7da0aU,
    0x2054c102U,
    0xd8ce8486U,
    0xb4208d72U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0x98000000U,
    0x84000000U,
    0x22000000U,
    0x85000000U,
    0xd4800000U,
    0xc7400000U,
    0xd7200000U,
    0x9900000U,
    0x3b80000U,
    0x51d40000U,
    0xb11e0000U,
    0x81830000U,
    0xabc28000U,
    0xe463c000U,
    0xf4b0a000U,
    0x932dd000U,
    0x20ec0800U,
    0x808db400U,
    0xb63f1200U,
    0x30119d00U,
    0x5c7f3680U,
    0x9fb1c4c0U,
    0xdca9b220U,
    0x602b4d30U,
    0xeb6fbeb8U,
    0xe24cb0e4U,
    0x58dc8006U,
    0xb0e0c011U,
    0xf3f22030U,
    0xa40e1005U,
    0xb97ca813U,
    0x1b306410U,
    0x63eb1a3dU,
    0xa3082923U,
    0x58fe24b5U,
    0xe0f359e5U,
    0xff8c04b9U,
    0x43bd49d2U,
    0x71d0acb2U,
    0xc11d2df9U,
    0x8983b6afU,
    0xb7c104f1U,
    0x42639214U,
    0x53b15d0dU,
    0xc2ad16b3U,
    0x332fd4fdU,
    0x90ed1a11U,
    0x688f2904U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x68000000U,
    0x94000000U,
    0x2000000U,
    0xd7000000U,
    0xfe800000U,
    0xfac00000U,
    0x7a00000U,
    0x93100000U,
    0xeab80000U,
    0xe4d40000U,
    0x989e0000U,
    0xf5870000U,
    0x42458000U,
    0x50604000U,
    0x31372000U,
    0x83eb5000U,
    0xb20e1800U,
    0xc37de400U,
    0xc7f06e00U,
    0xba494900U,
    0x1c8e80U,
    0x724108c0U,
    0x98614e20U,
    0xb5311930U,
    0x99e91698U,
    0x990bacdcU,
    0xabfb803aU,
    0xe8374015U,
    0x946aa018U,
    0x974f1029U,
    0x659f3825U,
    0xe805b41bU,
    0x5405f601U,
    0x6203ed11U,
    0xa7064085U,
    0x968751c7U,
    0x6ec2f887U,
    0x5a5a5e9U,
    0x4415ae92U,
    0x143d58c1U,
    0x1e12d608U,
    0x9f38bd14U,
    0x669058beU,
    0xa8feb5e8U,
    0xb4b496b8U,
    0xa9afecd0U,
    0x766aa02aU,
    0xf04f102fU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0xe8000000U,
    0x4c000000U,
    0x3e000000U,
    0x4f000000U,
    0x63800000U,
    0x8ac00000U,
    0x97200000U,
    0x64300000U,
    0xbe480000U,
    0x3f40000U,
    0x73ea0000U,
    0x7d830000U,
    0x75c28000U,
    0xeca04000U,
    0x5af2e000U,
    0xb36eb000U,
    0x5ac41800U,
    0xdf20ec00U,
    0xd8312e00U,
    0x784f2500U,
    0xe8f43780U,
    0x62692b40U,
    0x8641ce20U,
    0xce669510U,
    0x61d0afb8U,
    0xf95e8744U,
    0x438a801aU,
    0xf3544003U,
    0x1f18e037U,
    0xddedb017U,
    0x9a869822U,
    0xba40ac31U,
    0x4863ce2aU,
    0x6ad1950aU,
    0xf8d82fb7U,
    0x504dc762U,
    0x4f2e019U,
    0xac6eb016U,
    0x3144180aU,
    0x9e0ec3fU,
    0x99112e30U,
    0x1f7f2520U,
    0xb3c3784U,
    0xa45d2b5fU,
    0x10bce18U,
    0x5d159500U,
    0x3d7a2fb1U,
    0x723ac772U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x28000000U,
    0x7c000000U,
    0xde000000U,
    0x7f000000U,
    0xa6800000U,
    0x6b400000U,
    0xe8e00000U,
    0xf5500000U,
    0xa9180000U,
    0x2b140000U,
    0xda7e0000U,
    0x82850000U,
    0x11468000U,
    0x2de04000U,
    0xb6d36000U,
    0x91dfb000U,
    0xd3341800U,
    0xab0dc400U,
    0x84eda200U,
    0x4338f300U,
    0x366b080U,
    0x71966fc0U,
    0xaf38c220U,
    0x35664330U,
    0x52922888U,
    0xc7baebc4U,
    0x926802aU,
    0x60f0402fU,
    0x87ab603fU,
    0x37dbb01bU,
    0x18321823U,
    0x9f8cc405U,
    0x76ad220dU,
    0x2659b322U,
    0x99f55083U,
    0xfa289fd0U,
    0x401f3a2cU,
    0x7e95772dU,
    0xd1b872a6U,
    0x3a212cc5U,
    0x20726ab1U,
    0xc7e9e8ebU,
    0xdbb948afU,
    0x27215bcaU,
    0x7f49807U,
    0x652c8405U,
    0xd69e423bU,
    0x2dd6030fU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x28000000U,
    0x14000000U,
    0xf6000000U,
    0xbf000000U,
    0x17800000U,
    0xf8400000U,
    0x21a00000U,
    0xfd300000U,
    0x41f80000U,
    0x92f40000U,
    0xd81a0000U,
    0xae630000U,
    0x34d08000U,
    0x556e4000U,
    0x6a3e2000U,
    0xd510f000U,
    0x4f885800U,
    0xa32e8c00U,
    0x889ffa00U,
    0x3da15100U,
    0xef30bb80U,
    0x90fd8ac0U,
    0x56715a20U,
    0xfddfe110U,
    0x2a86c3b8U,
    0xb6c3f6d4U,
    0xb966f82aU,
    0x77503c15U,
    0xbb298205U,
    0xa49f2d3bU,
    0x17a7198fU,
    0xe23257fbU,
    0xe67e1b95U,
    0x12b33ac0U,
    0xc5bf220dU,
    0xb1559d1eU,
    0x3c2b618cU,
    0x6f1f2bc5U,
    0x5de1398aU,
    0xa296a7cbU,
    0xa7cc438cU,
    0x4a8eb6cbU,
    0x31a85801U,
    0xa25e8c3dU,
    0xd6c7fa2fU,
    0xa965511fU,
    0xbf52bbbcU,
    0xff2a8ad7U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0x8000000U,
    0x24000000U,
    0xae000000U,
    0x5000000U,
    0x17800000U,
    0x73c00000U,
    0x6c200000U,
    0x71500000U,
    0xe5e80000U,
    0x3d140000U,
    0x6e0a0000U,
    0x26610000U,
    0x2bb58000U,
    0x4f1d4000U,
    0x2a6aa000U,
    0xd2d33000U,
    0x202e0800U,
    0x28334c00U,
    0x7ade7600U,
    0x374bbd00U,
    0xe205c480U,
    0x9f065640U,
    0xe4815620U,
    0xbd45cd10U,
    0x8a616ca8U,
    0xe1b62a7cU,
    0x44192822U,
    0xc8e93c19U,
    0xbe90de03U,
    0x1bcac13dU,
    0x9b403a87U,
    0xeb63e755U,
    0xf8304490U,
    0xc2db164dU,
    0xbb4bf634U,
    0x9806fd3fU,
    0x3c0764aaU,
    0xd2016660U,
    0x77055e3aU,
    0x90878104U,
    0xeb429a97U,
    0xa364d75cU,
    0x7c344cbeU,
    0x9cd95a40U,
    0xb6480015U,
    0xab840007U,
    0xe1c2003dU,
    0xbb250002U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0xb8000000U,
    0xfc000000U,
    0xf6000000U,
    0x53000000U,
    0xff800000U,
    0xc2400000U,
    0x7e00000U,
    0x50500000U,
    0x8dc80000U,
    0xd1940000U,
    0x5e6e0000U,
    0xc5250000U,
    0xf4778000U,
    0x1f3cc000U,
    0x766a6000U,
    0x71207000U,
    0x66727800U,
    0x7638cc00U,
    0x94ecca00U,
    0xe3e7e300U,
    0xaa53c080U,
    0xb0c9f940U,
    0xb1112a20U,
    0x2dab5330U,
    0xaf83d888U,
    0xaa45454cU,
    0x53e1980eU,
    0xd2517c0fU,
    0x6ccb5235U,
    0x77179f18U,
    0xc6a91291U,
    0xac03a64fU,
    0x9e03d89cU,
    0x7054570U,
    0x7d819834U,
    0x23417c38U,
    0xa1635230U,
    0xc8939f16U,
    0xe4ef128dU,
    0xbbe2a658U,
    0x465258a8U,
    0xcec88579U,
    0xa6127832U,
    0xd828cc02U,
    0xc8c4ca3cU,
    0x123e337U,
    0x3e75c094U,
    0x9a38f95fU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0xd8000000U,
    0x64000000U,
    0x3e000000U,
    0x41000000U,
    0xdc800000U,
    0x84400000U,
    0xe2e00000U,
    0x76500000U,
    0xa180000U,
    0xaa940000U,
    0xe2be0000U,
    0xc7270000U,
    0x56758000U,
    0x5edc000U,
    0xd9bca000U,
    0xb4a53000U,
    0x87b42800U,
    0x73cf9400U,
    0x4b4f6a00U,
    0x310e1300U,
    0xf2ec4080U,
    0x283e3fc0U,
    0x7e664a20U,
    0x5216e330U,
    0xf8fcc898U,
    0x74c09bdcU,
    0x54a30816U,
    0x37b06429U,
    0xcbca6217U,
    0x5f4d770cU,
    0xd70da281U,
    0xd7e988f8U,
    0xcabac897U,
    0xbb239bfdU,
    0x6c70880dU,
    0xaeea427U,
    0x9c3d422eU,
    0x88628729U,
    0xdf10aaaeU,
    0xa67eece2U,
    0xeb852a96U,
    0x15c32cddU,
    0x58218a83U,
    0x7bf21ccdU,
    0xe52ba290U,
    0xab1a88e3U,
    0xc61148a6U,
    0xdef95be8U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x58000000U,
    0x4c000000U,
    0xe000000U,
    0x5000000U,
    0xc6800000U,
    0xcac00000U,
    0x5e600000U,
    0xb4300000U,
    0x83c80000U,
    0x4740000U,
    0x316e0000U,
    0x4f230000U,
    0x8a148000U,
    0x1d5dc000U,
    0x60ece000U,
    0xd060b000U,
    0x71334800U,
    0xa54bc400U,
    0x9eb23e00U,
    0x370a6700U,
    0xb712e780U,
    0x7da3540U,
    0x1c2a5e20U,
    0x97071730U,
    0x55854fb8U,
    0xa5454154U,
    0xc252836U,
    0x7d95b423U,
    0xe319163bU,
    0xf74cd315U,
    0xedb771a7U,
    0x88f2661U,
    0xbd57cfb4U,
    0x827f815cU,
    0x72fb4809U,
    0x613fc417U,
    0x4fdc3e35U,
    0x28296739U,
    0x65066797U,
    0x5687f554U,
    0x72c6be06U,
    0x4267a724U,
    0xe23607a0U,
    0xcace8565U,
    0xccf7163eU,
    0xfeafd336U,
    0xd7c3f1a6U,
    0xf4e2e66fU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x48000000U,
    0x54000000U,
    0x1a000000U,
    0x7000000U,
    0x87800000U,
    0x86c00000U,
    0xdd200000U,
    0xcab00000U,
    0xa6680000U,
    0xb7f40000U,
    0x838e0000U,
    0x5650000U,
    0xa3d08000U,
    0x13bb4000U,
    0x85482000U,
    0x18409000U,
    0x49659800U,
    0x65d36c00U,
    0x7ab8ce00U,
    0xcfce3500U,
    0x56021380U,
    0xc1043240U,
    0xee806e20U,
    0xcc45e530U,
    0x9367ab88U,
    0x42d3ce74U,
    0x2d3b3832U,
    0x109bc25U,
    0xdf23f60eU,
    0x11b78935U,
    0x4fe9e5b3U,
    0xfc37bb74U,
    0x962f8b91U,
    0x9a935e43U,
    0x445ea020U,
    0xb4dad008U,
    0xed9b380dU,
    0x8a79bc12U,
    0x3ebf615U,
    0x3a33893dU,
    0xff2fe5a6U,
    0xd016bb67U,
    0xa190b8cU,
    0x3cb91e7eU,
    0x66c80009U,
    0x3c84002eU,
    0x5f460014U,
    0x2ee10024U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0xf8000000U,
    0x34000000U,
    0x5e000000U,
    0x83000000U,
    0x5a800000U,
    0x72400000U,
    0xc2e00000U,
    0xa6700000U,
    0xcfe80000U,
    0x10b40000U,
    0x5c4a0000U,
    0xdea30000U,
    0xaf928000U,
    0xf18c000U,
    0x2f1d2000U,
    0x9f1b3000U,
    0x971cb800U,
    0x4b19a400U,
    0xd919ca00U,
    0x301b2300U,
    0xb79bb180U,
    0x1c596040U,
    0xf67e6a20U,
    0xe0acd310U,
    0x4bd029b8U,
    0x32f8f444U,
    0xb1e9981eU,
    0x23b6941dU,
    0xecf722fU,
    0x70e18724U,
    0xff70fba8U,
    0x406a8351U,
    0x6770fb87U,
    0x246a8359U,
    0x2170fbbdU,
    0x836a834cU,
    0xddf0fb81U,
    0x462a8347U,
    0x1b90fba9U,
    0x111a837eU,
    0x4c18fbbbU,
    0xd59e8375U,
    0x1d5afb99U,
    0xbdf9835dU,
    0x216a7b99U,
    0x7cf64363U,
    0xfdafdbbbU,
    0x3256b36aU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0xe8000000U,
    0xdc000000U,
    0x32000000U,
    0xcb000000U,
    0xd1800000U,
    0x16c00000U,
    0x36a00000U,
    0x48700000U,
    0xb1d80000U,
    0x4d340000U,
    0x72be0000U,
    0x5e670000U,
    0xdfd58000U,
    0xb72c4000U,
    0xa5a96000U,
    0xeecd000U,
    0x9489b800U,
    0x8d5d9c00U,
    0x7f732e00U,
    0x82581900U,
    0x68f64580U,
    0xd19fd3c0U,
    0x6d7ce20U,
    0x7dac8930U,
    0x9ae89d98U,
    0x6a899fdcU,
    0xec58d81aU,
    0x51f54c07U,
    0xab1c9614U,
    0xb196852eU,
    0xb5ceeb8eU,
    0x383c8af2U,
    0xa1256ba1U,
    0xdbb7cad0U,
    0xbff98ba0U,
    0x91075afaU,
    0x6815393U,
    0x954216deU,
    0x9b65c595U,
    0x475093d7U,
    0x336d2e0fU,
    0xf84f1938U,
    0xbcfbc585U,
    0x4c8793c0U,
    0xda40ae35U,
    0x2ce75912U,
    0x494a5a2U,
    0xc54843eaU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0xe8000000U,
    0x4c000000U,
    0x8e000000U,
    0xd9000000U,
    0xa8800000U,
    0x52400000U,
    0x3ea00000U,
    0x3ab00000U,
    0x44e80000U,
    0x32740000U,
    0x70a0000U,
    0x83e70000U,
    0x91948000U,
    0x269c4000U,
    0x9d78e000U,
    0xa8ed3000U,
    0x4c71a800U,
    0x860eb400U,
    0x5f629a00U,
    0xe9d51100U,
    0x3391780U,
    0x58498a40U,
    0xcfc6fa20U,
    0xba606110U,
    0xd7525f98U,
    0x9c790e74U,
    0x346b481aU,
    0x54308403U,
    0x73adb23bU,
    0x7250e502U,
    0xc2ffed90U,
    0x7b29eb47U,
    0xd014a58cU,
    0xa4596f68U,
    0x35991783U,
    0xbef98a7cU,
    0x8d2efa0cU,
    0xcd146119U,
    0x1ed85fb4U,
    0xd8de0e54U,
    0xbddfc835U,
    0xc35cc42bU,
    0x3c1d5229U,
    0x8039d530U,
    0xf3cc45beU,
    0x76045f55U,
    0x5d00bf8fU,
    0xda833e73U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x8000000U,
    0x6c000000U,
    0x22000000U,
    0xc9000000U,
    0x30800000U,
    0x47400000U,
    0x9ea00000U,
    0xfcb00000U,
    0x44f80000U,
    0x7d740000U,
    0x221a0000U,
    0xd0e70000U,
    0x72928000U,
    0x8f0b4000U,
    0xbf2fe000U,
    0xc7df5000U,
    0xf8002800U,
    0x34011400U,
    0x2607b200U,
    0x9706d100U,
    0xd387ef80U,
    0xd2c2eac0U,
    0xcb62d220U,
    0xec56c110U,
    0x166a2798U,
    0x827faec4U,
    0x8535c822U,
    0x7339440bU,
    0xc4951a10U,
    0x500c8536U,
    0x60afbd8eU,
    0x271b6bcaU,
    0xf265158fU,
    0x84d53fddU,
    0x552f47a5U,
    0x42dfbedaU,
    0x9a800015U,
    0xa2400004U,
    0xec20001fU,
    0x62f00011U,
    0xe2d80021U,
    0xaa84003dU,
    0xda420001U,
    0x98230032U,
    0x24f0801cU,
    0x65d84025U,
    0x71076034U,
    0x6483102dU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x18000000U,
    0xa4000000U,
    0x1a000000U,
    0xc7000000U,
    0x46800000U,
    0x20400000U,
    0xbd600000U,
    0xc3900000U,
    0xba880000U,
    0xc2540000U,
    0x312a0000U,
    0xa4e50000U,
    0x41528000U,
    0x1da9c000U,
    0x6ba1a000U,
    0x16b21000U,
    0x587a1800U,
    0x154e6400U,
    0x8cf1fa00U,
    0x7219ff00U,
    0xa85a4e80U,
    0xa2384740U,
    0xf28da20U,
    0x39e72f10U,
    0xe0d37688U,
    0x6b6df344U,
    0xee829826U,
    0xcc42a439U,
    0x3b62da0eU,
    0xaa922f35U,
    0x3909f697U,
    0xc7903361U,
    0xf0893809U,
    0x3d55b411U,
    0xc3aa4237U,
    0x86a58b3cU,
    0x9f31aca5U,
    0x82bfdc70U,
    0x366b6e98U,
    0x2f029756U,
    0x8a83e22fU,
    0xb6439b30U,
    0xcc61b4a7U,
    0xe414b85fU,
    0xa54814b7U,
    0xc4f2a861U,
    0xee180c8fU,
    0x659cc69U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x18000000U,
    0x44000000U,
    0xee000000U,
    0x29000000U,
    0x87800000U,
    0xe9400000U,
    0x6da00000U,
    0x2bd00000U,
    0x44a80000U,
    0x54140000U,
    0x33ce0000U,
    0xfd270000U,
    0xd4128000U,
    0xf3cbc000U,
    0x9d22a000U,
    0x4411b000U,
    0xebc86800U,
    0xd923cc00U,
    0xaa10c200U,
    0xc2ce7100U,
    0x5ea1e480U,
    0x4356c040U,
    0xaf6ee220U,
    0x75730130U,
    0x7f9ac98U,
    0xfb7f7c64U,
    0x46bce826U,
    0xfadb0c21U,
    0x2f6ee223U,
    0xb573012eU,
    0x67f9aca7U,
    0x6b7f7c6bU,
    0x5ebce800U,
    0xbedb0c30U,
    0xc16ee228U,
    0x9c73010bU,
    0xe079ac91U,
    0x823f7c64U,
    0x331ce83fU,
    0x950b0c13U,
    0x85c6e212U,
    0xc8670130U,
    0xd3b7ac92U,
    0x7f187c51U,
    0xe70e6807U,
    0x66c0cc0eU,
    0x18e44218U,
    0x8c76b107U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0x78000000U,
    0xbc000000U,
    0x6000000U,
    0xc3000000U,
    0x39800000U,
    0xcd400000U,
    0x5aa00000U,
    0x34b00000U,
    0xbdb80000U,
    0xcb740000U,
    0xb7da0000U,
    0x82250000U,
    0xa9718000U,
    0xfedac000U,
    0x26a26000U,
    0x52b1f000U,
    0x2ebaf800U,
    0xfaf67c00U,
    0xbe99b200U,
    0x6283f900U,
    0x58c67f80U,
    0xb9e3e0c0U,
    0x19105220U,
    0x728dc910U,
    0xc28f6788U,
    0xda8eacdcU,
    0x3689783eU,
    0x388dbc3fU,
    0x3f885209U,
    0xbc09c92cU,
    0xb44d6790U,
    0x146facdcU,
    0xd41af837U,
    0xfe467c2dU,
    0x5b21b229U,
    0x65f7f91dU,
    0x911c7fa5U,
    0x44c6e0c2U,
    0x8fe1d224U,
    0x8217092dU,
    0x870d07a2U,
    0x71cf5cdbU,
    0xff2b8009U,
    0x3dbfc033U,
    0x8b73e000U,
    0x97db3012U,
    0xf2209820U,
    0xd1738c2cU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0x88000000U,
    0x84000000U,
    0xaa000000U,
    0x41000000U,
    0x94800000U,
    0x8ec00000U,
    0x8c600000U,
    0xab700000U,
    0xb2f80000U,
    0xa4340000U,
    0x36de0000U,
    0xe0e10000U,
    0xe9b18000U,
    0xb0994000U,
    0x1447a000U,
    0xd9a71000U,
    0xda932800U,
    0x77ce7400U,
    0x366e3e00U,
    0x70fee700U,
    0xf1337d80U,
    0xd05f7cc0U,
    0xd3261e20U,
    0x5751b730U,
    0xfa2e75a8U,
    0x555b58fcU,
    0x8da2a802U,
    0xc8973411U,
    0x9ac99e02U,
    0x7ce9f72cU,
    0x9d385587U,
    0x2cd508c2U,
    0xcb6e2029U,
    0x227a500aU,
    0x40f28821U,
    0x5bbc6436U,
    0x2f12960cU,
    0x5d88d333U,
    0x1ccb63bcU,
    0x23efcbd5U,
    0xcab9eb8dU,
    0x393afddU,
    0x7f4b7da7U,
    0x4eab7cd7U,
    0xeb981e0eU,
    0x29c0b72eU,
    0x17e7f5a7U,
    0xa3618dcU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0xd8000000U,
    0x3c000000U,
    0x4a000000U,
    0xeb000000U,
    0xcf800000U,
    0x49400000U,
    0xb3e00000U,
    0xab300000U,
    0x82580000U,
    0xe2f40000U,
    0x217a0000U,
    0x2610000U,
    0xf7778000U,
    0xab3bc000U,
    0x3a006000U,
    0xd3069000U,
    0xe383b800U,
    0x6b41ac00U,
    0xbce3d200U,
    0x12b20300U,
    0x6a1d4380U,
    0x7592d5c0U,
    0xc8e3220U,
    0x7a9e5310U,
    0xd511ba8U,
    0xa32a29ecU,
    0x58cc3816U,
    0xe7fe6c1fU,
    0x6ba1b23aU,
    0xa4519316U,
    0x33ab7b85U,
    0x600db9ddU,
    0xe858001eU,
    0xf9f40000U,
    0x96fa003bU,
    0xc7210006U,
    0xd6978012U,
    0xd70bc025U,
    0x3dd8600fU,
    0x93b29028U,
    0xbe99b835U,
    0x8b50ac26U,
    0x7a2c5229U,
    0xf04dc31aU,
    0xf33f2381U,
    0x460145c8U,
    0x39000a0aU,
    0xb8853f17U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x8000000U,
    0xdc000000U,
    0x8e000000U,
    0x23000000U,
    0x99800000U,
    0x2c00000U,
    0x32200000U,
    0x75900000U,
    0x1d280000U,
    0xbbd40000U,
    0x834e0000U,
    0xd5230000U,
    0x6158000U,
    0x86ef4000U,
    0x11716000U,
    0xc89a3000U,
    0x9e6cf800U,
    0xef313c00U,
    0x7ef84e00U,
    0xbc9b3d00U,
    0x9c6e0a80U,
    0xda33c040U,
    0x607aae20U,
    0x7dd94d30U,
    0xdd8812b8U,
    0xa9048c5cU,
    0xb082f822U,
    0x72423c07U,
    0xa065ce1bU,
    0xa5f07d14U,
    0xb0d96aa4U,
    0xd70ef077U,
    0xda45d60fU,
    0x2c603125U,
    0xb3f2bc99U,
    0x37ddc172U,
    0x148aea8eU,
    0xa986b06fU,
    0x2ac7363dU,
    0x7e22412bU,
    0x8394a4b8U,
    0xea2a8d6cU,
    0x7052bcabU,
    0x2c8dc148U,
    0x5d82ea81U,
    0xe8c2b057U,
    0xab213610U,
    0xed15413eU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0xa8000000U,
    0x94000000U,
    0x6000000U,
    0xb1000000U,
    0x3b800000U,
    0x54c00000U,
    0x45e00000U,
    0xf9700000U,
    0x6b680000U,
    0x15340000U,
    0x86ce0000U,
    0xffe50000U,
    0x1a738000U,
    0xc7ec4000U,
    0x7b752000U,
    0x46b7000U,
    0x83b42800U,
    0xcb8a6c00U,
    0xf1c76e00U,
    0xb864d300U,
    0x3cb11d80U,
    0xa50ddc40U,
    0x1087ce20U,
    0x3c42e330U,
    0x15a59588U,
    0x46d1806cU,
    0x16bc280aU,
    0xb30e6c15U,
    0x79816e09U,
    0x1bc5d300U,
    0x63649da4U,
    0x1c309c70U,
    0xf14f6e30U,
    0xb120d322U,
    0xcc971d9cU,
    0x1a1cdc7cU,
    0x5a5a4e3aU,
    0x3dfba314U,
    0x5c2b3595U,
    0xd812b074U,
    0x385b203bU,
    0xc2fe703eU,
    0xd2afa83eU,
    0xa9522c3dU,
    0xa47c4e1cU,
    0xd7eaa31aU,
    0xa376b59eU,
    0x186bf076U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x88000000U,
    0x44000000U,
    0xc2000000U,
    0x77000000U,
    0x96800000U,
    0x2a400000U,
    0x13a00000U,
    0xcc300000U,
    0x46a80000U,
    0x79f40000U,
    0xc0ce0000U,
    0x3a70000U,
    0x84328000U,
    0x62af4000U,
    0x2bf2a000U,
    0x3fcc5000U,
    0xd1269800U,
    0x6c702400U,
    0x60daa00U,
    0x71468900U,
    0x53236380U,
    0xbb703040U,
    0x608b8a20U,
    0x43069930U,
    0x8c83db98U,
    0xf1400464U,
    0x93269802U,
    0xdb702421U,
    0xf08daa28U,
    0xcb068939U,
    0xc8836387U,
    0x3340305bU,
    0xe4238a14U,
    0x4df2991eU,
    0xdacddbacU,
    0xd8a70470U,
    0x4b41836U,
    0x75ef6412U,
    0x9dd70a18U,
    0x8d3ed93fU,
    0xd96bfbbdU,
    0x5c971454U,
    0x661ca001U,
    0x5e1b501dU,
    0xa21c1827U,
    0x5c1b6423U,
    0xb5190a36U,
    0x5a99d917U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0xe8000000U,
    0xc000000U,
    0xde000000U,
    0x65000000U,
    0xb1800000U,
    0xd9400000U,
    0x2ee00000U,
    0xa3100000U,
    0xe6880000U,
    0x3df40000U,
    0x9d9e0000U,
    0xcf7f0000U,
    0xa6e8000U,
    0xde64000U,
    0xc391e000U,
    0xd44e3000U,
    0xed548800U,
    0xf2ea5c00U,
    0x8b247200U,
    0x4933a300U,
    0x9c393680U,
    0x4b8de540U,
    0xb070fa20U,
    0x82d9ff30U,
    0xf09d4498U,
    0x22fe4664U,
    0x8529cc9aU,
    0xda041a73U,
    0xe705be8fU,
    0x8283b94dU,
    0x91c2882eU,
    0x70215c21U,
    0x3db4f21eU,
    0x4fae316U,
    0xc42e5698U,
    0xd9819563U,
    0x15439217U,
    0x90e69304U,
    0x56153e84U,
    0xbf0af976U,
    0xe8b5e83fU,
    0x6d7d2c12U,
    0x96f9a08U,
    0x5d618f3eU,
    0xe9502ca1U,
    0x70ee2a62U,
    0xb82736b2U,
    0x1b2e559U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0x38000000U,
    0x54000000U,
    0xf6000000U,
    0x63000000U,
    0xab800000U,
    0xad400000U,
    0xf0e00000U,
    0x64500000U,
    0x99d80000U,
    0x40b40000U,
    0xcb8e0000U,
    0x1a690000U,
    0xb0be8000U,
    0x44a6c000U,
    0xe352000U,
    0x664ff000U,
    0xb048f800U,
    0x834b8c00U,
    0xb0cc6200U,
    0xd98df300U,
    0xe76a4c80U,
    0xb43b9fc0U,
    0x70649a20U,
    0xfe967f30U,
    0x6e7e2ea8U,
    0xbd026ce4U,
    0x6480d6aeU,
    0x4c4e0e5U,
    0x58a434b5U,
    0x6432d3ccU,
    0xaf4bd82cU,
    0x72c97c2aU,
    0x7c8c1a27U,
    0x87edbf30U,
    0xdefb8ebfU,
    0x8fc25cf6U,
    0xff238eb9U,
    0x5b765cdcU,
    0x62ad8e94U,
    0xb21f5ce7U,
    0x41930eadU,
    0xff99cd7U,
    0x49462eafU,
    0x6ee66cfeU,
    0xcb56d6a0U,
    0x59e0caU,
    0x40f4b497U,
    0xa7ed13eeU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x68000000U,
    0x7c000000U,
    0xc6000000U,
    0x11000000U,
    0x25800000U,
    0x60400000U,
    0xc0600000U,
    0x61100000U,
    0xdd280000U,
    0x4d740000U,
    0x323a0000U,
    0xcd5f0000U,
    0xa4cc8000U,
    0x9a264000U,
    0x36706000U,
    0xe8bcf000U,
    0x859ea800U,
    0xb769bc00U,
    0x22135a00U,
    0x13aa1700U,
    0xbfb76e80U,
    0xf69ae440U,
    0x41edf220U,
    0xc4d3ab10U,
    0x6d0c3498U,
    0xe404f354U,
    0xa2009cbaU,
    0xf3064f4fU,
    0xb6854689U,
    0x86c51840U,
    0x2ea6c82bU,
    0x33b14c03U,
    0x289ff203U,
    0x74e8ab17U,
    0x6352b495U,
    0xce49b350U,
    0xaae67ca4U,
    0x51d3ff67U,
    0xba8f0e9fU,
    0x1f421471U,
    0xa8e15a24U,
    0xd2d11719U,
    0x5409ee80U,
    0xdd87a464U,
    0x54431222U,
    0x4a621b38U,
    0x8e147cbaU,
    0xfda8ff5cU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0xf8000000U,
    0x8c000000U,
    0xd6000000U,
    0xb000000U,
    0xcb800000U,
    0xfec00000U,
    0xa4a00000U,
    0xe0900000U,
    0xe1180000U,
    0x78340000U,
    0xbf8e0000U,
    0x862f0000U,
    0x6388000U,
    0xe266c000U,
    0x6833e000U,
    0xe788f000U,
    0x3a2d7800U,
    0x78394c00U,
    0x9d647e00U,
    0xf9b21100U,
    0xc448b680U,
    0x5e0ea7c0U,
    0xade90620U,
    0x261b5d30U,
    0xc5b4c898U,
    0x7a48b6d4U,
    0x410fb09eU,
    0x6c6afad3U,
    0xfddd4e8dU,
    0x2d912be6U,
    0x799e9834U,
    0xf171bc38U,
    0x9be9063aU,
    0xbd1b5d0dU,
    0x9634c881U,
    0x5888b6c0U,
    0xcbafb0a1U,
    0xbfafad4U,
    0x1454e9aU,
    0xa0652bc5U,
    0xa9309822U,
    0x690ebc39U,
    0xd869860bU,
    0xc7d99d2cU,
    0xa09128aaU,
    0x411b46c0U,
    0x483448aaU,
    0x178a76fdU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x88000000U,
    0x84000000U,
    0xc6000000U,
    0xf5000000U,
    0xc8800000U,
    0xef400000U,
    0x13a00000U,
    0x79d00000U,
    0x31180000U,
    0x96740000U,
    0xf2ce0000U,
    0x286f0000U,
    0x433c8000U,
    0xc2654000U,
    0xe837a000U,
    0xffeff000U,
    0xa3fd4800U,
    0xfd021400U,
    0x8c87c200U,
    0x4942c100U,
    0x36a79880U,
    0x39504bc0U,
    0x5a5a8a20U,
    0x43d0d530U,
    0x7e185a98U,
    0xd1f68af4U,
    0x3a0b1282U,
    0x230b9ed1U,
    0xb9885089U,
    0x8dc81ff9U,
    0xf7ee6828U,
    0xe7fca41eU,
    0x5b032a0fU,
    0xa9802536U,
    0x9c192adU,
    0x5de5dec2U,
    0xecf5709bU,
    0xd689afe2U,
    0xba4e0012U,
    0x72f0034U,
    0x309c8034U,
    0x6bb5401bU,
    0x512fa01bU,
    0xed9bf03cU,
    0x97334836U,
    0x206d1411U,
    0x73b4239U,
    0x6467811bU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0xd8000000U,
    0x3c000000U,
    0x16000000U,
    0x6d000000U,
    0xd6800000U,
    0x4cc00000U,
    0xeee00000U,
    0xa8d00000U,
    0x1ae80000U,
    0xd0340000U,
    0x1f3e0000U,
    0x3c5d0000U,
    0x33cf8000U,
    0x1585c000U,
    0x9b436000U,
    0x2aa2f000U,
    0x33f1c800U,
    0x78592400U,
    0x69cc5e00U,
    0x96801b00U,
    0xacc5fc80U,
    0x7ee7f240U,
    0xc0d59620U,
    0xfeed3f30U,
    0xfa37a288U,
    0x643ae96cU,
    0x87dfea96U,
    0xa98f0d7fU,
    0xb7a154adU,
    0xdd752647U,
    0x9899802bU,
    0xf92cc000U,
    0xb752e000U,
    0x4aaa3012U,
    0x99152800U,
    0x8a8a1433U,
    0xf020f60cU,
    0x7136cf2dU,
    0xe7bfea84U,
    0xb19f0d48U,
    0x75a954bcU,
    0x7891266bU,
    0x93cf8029U,
    0x6585c03cU,
    0x63436009U,
    0xa6a2f006U,
    0xfdf1c83eU,
    0x2959241aU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0x68000000U,
    0x8c000000U,
    0x6e000000U,
    0xff000000U,
    0x9b800000U,
    0xef400000U,
    0xe1600000U,
    0xa7100000U,
    0xa2580000U,
    0xd8740000U,
    0x24a0000U,
    0xa5a90000U,
    0x7788000U,
    0xbf804000U,
    0xed45e000U,
    0x63b000U,
    0x7b965800U,
    0x32988c00U,
    0x34539200U,
    0xd53d5f00U,
    0xc4606680U,
    0x9953fc0U,
    0x6b9dca20U,
    0xcd1d310U,
    0xa7f9f4a8U,
    0xb94160f4U,
    0xba652cbaU,
    0x7e94acf3U,
    0xc19de93U,
    0x359703dbU,
    0x5d988034U,
    0x67d0403cU,
    0x667de011U,
    0x2307b001U,
    0xdd84580fU,
    0x3c458c11U,
    0xc4e11205U,
    0xdf541f04U,
    0x54bd06b7U,
    0xfe26cfcbU,
    0xbe76720bU,
    0x414eef26U,
    0x482e3e9fU,
    0x2339b3c8U,
    0x4f64580dU,
    0xb8158c00U,
    0x49d91226U,
    0x8f301f2aU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x98000000U,
    0xc000000U,
    0xb2000000U,
    0x25000000U,
    0x69800000U,
    0xc3c00000U,
    0xa9600000U,
    0xdef00000U,
    0xdef80000U,
    0xd1940000U,
    0xdb0a0000U,
    0xd1ed0000U,
    0x35db8000U,
    0x91c3c000U,
    0x5c66e000U,
    0xbf75d000U,
    0x39380800U,
    0x5ef67c00U,
    0x9ef8c600U,
    0xf197bf00U,
    0x6b0d0d80U,
    0x49ef2ec0U,
    0x39d8ce20U,
    0x23c5c310U,
    0x7967cb88U,
    0xd6f191ecU,
    0xfafc4386U,
    0xf7902dd3U,
    0x40086584U,
    0x2f6f42f5U,
    0xba9ee014U,
    0x92e1d00fU,
    0xe8320828U,
    0x161b7c11U,
    0xe8a34627U,
    0x8a947f0eU,
    0x458bedaaU,
    0xceaafefaU,
    0x1ef8c62aU,
    0xb197bf07U,
    0x4b0d0db4U,
    0xf9ef2ef4U,
    0xa1d8ce04U,
    0x2fc5c320U,
    0xcb67cbafU,
    0xf3f191e6U,
    0x937c4392U,
    0x34502dedU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0xa8000000U,
    0xc4000000U,
    0x9e000000U,
    0x11000000U,
    0xfe800000U,
    0x52c00000U,
    0x16e00000U,
    0xef700000U,
    0x1ab80000U,
    0x78940000U,
    0x434a0000U,
    0x4e90000U,
    0x4b3c8000U,
    0xd5d64000U,
    0xf1eba000U,
    0x3bba3000U,
    0x9e176800U,
    0x8d8f6400U,
    0xb00b4e00U,
    0x474afb00U,
    0x7aeaf180U,
    0xaa3e50c0U,
    0xd3562620U,
    0xcf2c9f10U,
    0x775d3fa8U,
    0xfe62ebd4U,
    0x78b7778aU,
    0x64d8ffe1U,
    0x402471afU,
    0x879510c0U,
    0x14cb061dU,
    0x64a9ef21U,
    0x51d77a0U,
    0xb41ffdaU,
    0x2fa0f194U,
    0x6bd750efU,
    0xf0eaa605U,
    0xcd3adf2eU,
    0x58d69facU,
    0xad68dbe1U,
    0x8a781f82U,
    0x3d739bcfU,
    0x41bd3fb3U,
    0x112ebddU,
    0x6a0f778aU,
    0x884cfff1U,
    0x356e7191U,
    0x567c10e6U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0xe8000000U,
    0x4c000000U,
    0xaa000000U,
    0x6d000000U,
    0x34800000U,
    0x82c00000U,
    0x69a00000U,
    0xd3900000U,
    0xe9a80000U,
    0xe7340000U,
    0x76be0000U,
    0xc85b0000U,
    0x4e288000U,
    0x42724000U,
    0x7a59e000U,
    0xf72b9000U,
    0xc8f6c800U,
    0xcb98d400U,
    0xc909ce00U,
    0x26231100U,
    0x3ad46680U,
    0x4f4b0240U,
    0xb5410620U,
    0xc4e0c530U,
    0xc37528b8U,
    0x94da5354U,
    0x40ec809aU,
    0x61d05763U,
    0x94cae6b2U,
    0xc206427fU,
    0xe106662fU,
    0x7e861517U,
    0xbfc480b2U,
    0xb5245768U,
    0x1d54e687U,
    0x2a0d4261U,
    0x59a6e638U,
    0xab905519U,
    0xadab60bcU,
    0x5130c77aU,
    0x59bcae93U,
    0xddd8d67aU,
    0x52684839U,
    0xc4159436U,
    0xf4eeae37U,
    0x4fd5c105U,
    0xafcdce81U,
    0x89810649U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0xc8000000U,
    0x44000000U,
    0xfa000000U,
    0x83000000U,
    0x1a800000U,
    0xbd400000U,
    0x1fa00000U,
    0xbad00000U,
    0x5bd80000U,
    0x4e740000U,
    0x8f8a0000U,
    0xbaed0000U,
    0x695d8000U,
    0xa7314000U,
    0xf22aa000U,
    0x7f3f1000U,
    0x5e05f800U,
    0x69006400U,
    0x1868200U,
    0x5bc62100U,
    0xd4e09080U,
    0x9872edc0U,
    0x82897a20U,
    0x316b4510U,
    0x799b9288U,
    0x2d558cdcU,
    0x2d1b4a92U,
    0x6112b8c1U,
    0xcf3d1096U,
    0xf603adecU,
    0x7d03da1cU,
    0x43845532U,
    0x54c66a83U,
    0x7061e8c3U,
    0x5c37c8bcU,
    0x4a999ddU,
    0x2c780034U,
    0x80a40003U,
    0xc6520037U,
    0x43990036U,
    0xce578002U,
    0x679c403dU,
    0x64572000U,
    0x5c9e5001U,
    0xf2d75804U,
    0x5fdb740fU,
    0x94717a15U,
    0x7c8f4537U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x98000000U,
    0x5c000000U,
    0xd2000000U,
    0x27000000U,
    0xaf800000U,
    0xa9400000U,
    0x1a600000U,
    0x19100000U,
    0xfd380000U,
    0x38740000U,
    0x39ae0000U,
    0xbb0f0000U,
    0xbc3e8000U,
    0xf6f5c000U,
    0x6e6b6000U,
    0x9ea9d000U,
    0xd4899800U,
    0x757ccc00U,
    0x7c958200U,
    0xef7f4700U,
    0x3f946680U,
    0x3efd3cc0U,
    0x6bd21a20U,
    0x681c8b30U,
    0xfa076498U,
    0xd303bbe4U,
    0x49831c86U,
    0x84767e7U,
    0x84e2668cU,
    0xb6d63cddU,
    0x149a9a35U,
    0x1ac24b29U,
    0xc124848cU,
    0xd474abfcU,
    0x43a964a6U,
    0xa80cbbcaU,
    0x95bd9cafU,
    0x6eb2a7dfU,
    0x728906a3U,
    0x747fecdcU,
    0x12130206U,
    0x48be870bU,
    0x1231068aU,
    0x924becd0U,
    0x665d023aU,
    0x8fe1870dU,
    0x35786b0U,
    0x3eda2cc7U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x28000000U,
    0x2c000000U,
    0x22000000U,
    0x3000000U,
    0xca800000U,
    0x75400000U,
    0xc4200000U,
    0xb3100000U,
    0x61e80000U,
    0xa8340000U,
    0x627e0000U,
    0xb990000U,
    0xf2ea8000U,
    0x8ab64000U,
    0x9b3ae000U,
    0xfdb9d000U,
    0x6afab800U,
    0xdd86400U,
    0x644c4200U,
    0x58e7bd00U,
    0x18f3de80U,
    0xeb9c6040U,
    0x82e8fa20U,
    0x12b6d930U,
    0x8f3d1ca8U,
    0xdbb99d44U,
    0x67ffc4aaU,
    0xe65a697bU,
    0xd88dde80U,
    0x23056074U,
    0x1a827a30U,
    0xfd409922U,
    0xf827fc9bU,
    0xb9104d63U,
    0x4eed7c88U,
    0x40b60d4cU,
    0x143f9c9bU,
    0x53bdd47U,
    0x34bb24a5U,
    0x2f7ab946U,
    0x801de69dU,
    0xfe2b4444U,
    0xc954d81eU,
    0x9d4ef42bU,
    0xe5669a32U,
    0x31304937U,
    0x6fffc49fU,
    0x1a5a6974U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0x78000000U,
    0x14000000U,
    0xea000000U,
    0xb3000000U,
    0x45800000U,
    0x24400000U,
    0x3ba00000U,
    0x47100000U,
    0x4f480000U,
    0x93b40000U,
    0x5bda0000U,
    0xadbf0000U,
    0x2b488000U,
    0xf1b6c000U,
    0x34df2000U,
    0xbe3d1000U,
    0x3a0e7800U,
    0xc2931c00U,
    0x4b0cca00U,
    0x18137900U,
    0xb4cb4e80U,
    0x8ef39940U,
    0x46f8b220U,
    0xf36f6510U,
    0x6fe70498U,
    0xf4b22074U,
    0xd45edcbeU,
    0x8afaec55U,
    0xfd6b4e82U,
    0xfee39948U,
    0xbe30b237U,
    0x339b6528U,
    0xd81d04b2U,
    0x9d5d204cU,
    0x247e5c86U,
    0x38a82c54U,
    0xa3066e8bU,
    0x1d85897bU,
    0x80444a20U,
    0x79a5b921U,
    0x98146ebaU,
    0xf4ce896cU,
    0xeef6ca14U,
    0x96fc7933U,
    0x8b6bcebeU,
    0x7be1597eU,
    0x1eb59216U,
    0x67597529U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x28000000U,
    0x54000000U,
    0x12000000U,
    0x5d000000U,
    0x69800000U,
    0xf0c00000U,
    0x96e00000U,
    0xd0f00000U,
    0x99780000U,
    0x69140000U,
    0xad0a0000U,
    0xa12b0000U,
    0x663a8000U,
    0x70304000U,
    0x879b2000U,
    0x9de65000U,
    0x9e778800U,
    0xe93f3c00U,
    0xe4b7de00U,
    0x6edec300U,
    0x43c10580U,
    0xa462a6c0U,
    0xe6b25620U,
    0x5bdeff10U,
    0x9e465bb8U,
    0xd6a725e4U,
    0x552f3aaU,
    0xb6aa49c5U,
    0xe5f8a59cU,
    0x7450b6e3U,
    0x112cfe08U,
    0x7e38931dU,
    0x7c368d93U,
    0x799d9ad2U,
    0xaee58812U,
    0xecf03c34U,
    0x3f7f5e28U,
    0xb6118313U,
    0xb18aa58aU,
    0x6c6fb6dbU,
    0x121c7e2dU,
    0x6b23d326U,
    0xa7972d91U,
    0x9d4b8accU,
    0xcc892025U,
    0x35e95039U,
    0xbadf0815U,
    0x11c07c01U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0x68000000U,
    0xac000000U,
    0x22000000U,
    0x35000000U,
    0x28800000U,
    0xc6400000U,
    0xc4600000U,
    0x9db00000U,
    0x73180000U,
    0xa6d40000U,
    0x1f2a0000U,
    0x390d0000U,
    0x1eda8000U,
    0x3f744000U,
    0xa33ae000U,
    0x1203d000U,
    0x8d010800U,
    0xcc877400U,
    0xd8476a00U,
    0xbb649f00U,
    0x2c375380U,
    0xbfdbddc0U,
    0x91f46220U,
    0x6e7aeb10U,
    0x11e0b988U,
    0xf7602e4U,
    0x1b3b51baU,
    0xf602a6fbU,
    0x930533a0U,
    0xb3854df9U,
    0x69c70a38U,
    0x77a30f2eU,
    0x1b14bbabU,
    0xce8b79e5U,
    0xb9180024U,
    0x7fd4002eU,
    0x35aa003cU,
    0x5a4d0031U,
    0x9a3a803aU,
    0xc884401bU,
    0x3642e00aU,
    0x1c67d009U,
    0xc9b30835U,
    0x951e742aU,
    0x1dd7ea2fU,
    0x20addf0cU,
    0xe2cf339bU,
    0x34784dfeU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0xc8000000U,
    0x54000000U,
    0xa000000U,
    0x3b000000U,
    0xf4800000U,
    0xf6c00000U,
    0xf2a00000U,
    0xab00000U,
    0xbbc80000U,
    0x21140000U,
    0x5afe0000U,
    0xea9f0000U,
    0x88c8000U,
    0x2af74000U,
    0xc0a92000U,
    0x8f879000U,
    0xa245b800U,
    0xb4601400U,
    0xe0121e00U,
    0x2d7ff900U,
    0xc4df3f80U,
    0x4ae80940U,
    0x7fe1a620U,
    0xe054ed30U,
    0x261fa198U,
    0x124fb074U,
    0xfe533992U,
    0xb7183465U,
    0x6dcc9fbaU,
    0xe413d94aU,
    0x2f7fbe37U,
    0xcbdb292cU,
    0x646da794U,
    0x7a208d6dU,
    0xb2728003U,
    0xd068401eU,
    0x6025a00dU,
    0x2170d03fU,
    0xa0ec981fU,
    0x54e78428U,
    0xbcd7a636U,
    0x54dfed1dU,
    0x22ed21afU,
    0x9be7f068U,
    0xf2569980U,
    0x8118e46aU,
    0xb8c807bdU,
    0x39905d5aU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x8000000U,
    0x9c000000U,
    0xb2000000U,
    0xab000000U,
    0xc9800000U,
    0x74c00000U,
    0xdce00000U,
    0xd5f00000U,
    0x57380000U,
    0xc140000U,
    0x924e0000U,
    0x2d6f0000U,
    0x8ff98000U,
    0xcaf14000U,
    0x60ba6000U,
    0x79d15000U,
    0xc82c7800U,
    0x63d87c00U,
    0xf5e60e00U,
    0x1f752300U,
    0x667a7d80U,
    0x6e33e4c0U,
    0xa45c7620U,
    0x48265f30U,
    0xb913f398U,
    0xeacc87dcU,
    0xb72beba2U,
    0x245aabd7U,
    0x88201d94U,
    0xd912b4c6U,
    0x9ac80e08U,
    0xbf2a2316U,
    0xb85bfd81U,
    0x3a26a4e4U,
    0x72101629U,
    0x534c0f23U,
    0xcbe80b95U,
    0x64babbd5U,
    0xefd60591U,
    0x252b98ceU,
    0x5f5a780aU,
    0x59a37c25U,
    0x49d18e08U,
    0x602b6304U,
    0xefd99dbbU,
    0x3fe3f4d8U,
    0x20726e3bU,
    0x81fb732bU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x98000000U,
    0xc4000000U,
    0xb6000000U,
    0x7d000000U,
    0x51800000U,
    0xfd400000U,
    0xb200000U,
    0xb2100000U,
    0xd4080000U,
    0x16340000U,
    0x4c9a0000U,
    0x7bff0000U,
    0x314a8000U,
    0x9911c000U,
    0x288ba000U,
    0x42f2f000U,
    0xeeff8800U,
    0xeccadc00U,
    0xee551200U,
    0xbca94d00U,
    0xf266be80U,
    0x2f346f40U,
    0xeb189a20U,
    0x9bb89110U,
    0xfbeb2c98U,
    0x4e47e264U,
    0x33a50486U,
    0x50d0ce61U,
    0xb6d1e95U,
    0xd0869f6bU,
    0x6ec7120aU,
    0x55624d3aU,
    0x59b63e91U,
    0x20daaf66U,
    0x3b59ba0aU,
    0x791ba104U,
    0x80bf04b0U,
    0x2f6fce76U,
    0xb6879ea9U,
    0xebc75f51U,
    0x50e4b23bU,
    0x8af4bd0cU,
    0x92fbb6bbU,
    0x6ecb736cU,
    0x2d542807U,
    0xcc282c15U,
    0x2ca29a15U,
    0x12579107U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x48000000U,
    0x2c000000U,
    0x1e000000U,
    0xf1000000U,
    0x15800000U,
    0x4a400000U,
    0xd6e00000U,
    0x60500000U,
    0x38e80000U,
    0x25b40000U,
    0xa3a0000U,
    0xaf990000U,
    0x60aa8000U,
    0xcc514000U,
    0x66eae000U,
    0x74b73000U,
    0x8fba2800U,
    0xadda7c00U,
    0x9a4eca00U,
    0xb2002300U,
    0xaf049480U,
    0x4486be40U,
    0xcfc6e220U,
    0xd4a75f10U,
    0x9ab2dea8U,
    0x46badd64U,
    0xec5816b2U,
    0x3a0e915bU,
    0xefe6748fU,
    0x19d58e48U,
    0xccaeca1fU,
    0x9250232dU,
    0x37ec9488U,
    0xf132be4bU,
    0x8dfce23eU,
    0x573e5f3cU,
    0xe4185e9dU,
    0x7beb9d79U,
    0x9f32f694U,
    0x4f9a16bU,
    0xb6bc5cafU,
    0xd45ff26fU,
    0x6e080028U,
    0x5e40017U,
    0x92d2002dU,
    0x1a2d0022U,
    0x22908017U,
    0x4fc84038U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0xe8000000U,
    0x5c000000U,
    0xce000000U,
    0xef000000U,
    0x5f800000U,
    0x8400000U,
    0x82600000U,
    0xe4100000U,
    0xcd980000U,
    0xa9f40000U,
    0xf9ce0000U,
    0xb20b0000U,
    0xd62d8000U,
    0x4f5a4000U,
    0x21506000U,
    0xf97fb000U,
    0xb9265800U,
    0x37f0ac00U,
    0x2ecbea00U,
    0x498e8700U,
    0x5469d680U,
    0x583859c0U,
    0xe7c03220U,
    0x42246b30U,
    0x2a725cb8U,
    0xff896efcU,
    0x3f6fbc9aU,
    0x9dbc9ee7U,
    0x628184abU,
    0x56c782f7U,
    0x2aa236b5U,
    0x29b2a9d9U,
    0x17ad8a31U,
    0x901e3719U,
    0x58b40e95U,
    0x9f2db5f8U,
    0x63d83820U,
    0x58941c0bU,
    0x6c58323aU,
    0x68d06b05U,
    0x4a3c5c88U,
    0x6c26ed4U,
    0x12a23ca5U,
    0x8db6deeaU,
    0x9da9e4b4U,
    0x51c32f7U,
    0x7a326ebaU,
    0xe9ed05f8U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0xa8000000U,
    0xcc000000U,
    0xf6000000U,
    0x37000000U,
    0x6f800000U,
    0xa1400000U,
    0xb6e00000U,
    0xfb500000U,
    0xb8180000U,
    0x9a340000U,
    0x460e0000U,
    0xcecf0000U,
    0x766e8000U,
    0xc05c4000U,
    0x2053e000U,
    0x119b7000U,
    0x54711800U,
    0x6b6d6c00U,
    0x6df2200U,
    0x75902300U,
    0xfbfe480U,
    0x278259c0U,
    0x3d40ba20U,
    0x88e10f30U,
    0x90532698U,
    0x89990ae4U,
    0xc076468aU,
    0x696a3ac3U,
    0xa3dabe85U,
    0x171326d9U,
    0x7a0489U,
    0x68a229cfU,
    0xbe312202U,
    0xdc0f2314U,
    0xffc964baU,
    0x16ea19d4U,
    0x6a9d5a10U,
    0xdf57f2bU,
    0x9d2cbea0U,
    0x4fb826e9U,
    0x8782848bU,
    0xcd4569c4U,
    0xb0e2421fU,
    0xf4571318U,
    0xb39d1c82U,
    0x17445c3U,
    0x31ee8008U,
    0x6d1c4024U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0x18000000U,
    0xa4000000U,
    0x36000000U,
    0xeb000000U,
    0xc3800000U,
    0x4bc00000U,
    0xfca00000U,
    0xc7300000U,
    0xedb80000U,
    0xac140000U,
    0xf4a0000U,
    0x52090000U,
    0xe1ed8000U,
    0x83bc000U,
    0x84d5a000U,
    0x5c6df000U,
    0x24f83800U,
    0x91f7cc00U,
    0xd99e5600U,
    0x3360b100U,
    0x3d95d380U,
    0x798b72c0U,
    0x562bee20U,
    0xc69cbd10U,
    0xeee625a8U,
    0x815233dcU,
    0xa4ac05a6U,
    0x1fd903f9U,
    0x5b061d85U,
    0x3b81fff6U,
    0x3fc053beU,
    0xa2a4b2f7U,
    0x90344e3cU,
    0xbc384d2eU,
    0x3ad39da8U,
    0xdb6e3ff6U,
    0x1d7ff3a7U,
    0xbb3042f4U,
    0x9fb9f609U,
    0xa110413cU,
    0x4fca6bb8U,
    0x664e7efeU,
    0xc48d981dU,
    0xeeaa3c2bU,
    0x86de6e02U,
    0x95837d12U,
    0x30c18595U,
    0xf722c3f1U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0x28000000U,
    0xc4000000U,
    0x62000000U,
    0x1d000000U,
    0xbc800000U,
    0x3ec00000U,
    0xebe00000U,
    0x3f900000U,
    0x28a80000U,
    0xdcf40000U,
    0x93fa0000U,
    0x5ebd0000U,
    0xd21a8000U,
    0x482e4000U,
    0xbc35a000U,
    0xcb1dd000U,
    0xf6ae0800U,
    0x2ff31400U,
    0x247e6e00U,
    0x1bfee300U,
    0x4abdd680U,
    0xb81a8040U,
    0x12ae620U,
    0x4ab3b710U,
    0x2cde1888U,
    0xc3cdb364U,
    0x33a338aaU,
    0x5bb72361U,
    0xc65810b0U,
    0x1a0aa773U,
    0x5c7568dU,
    0x6864c04aU,
    0x80d74600U,
    0xd10a670dU,
    0x1e4210bfU,
    0xab27a76aU,
    0xf775d683U,
    0xa8be8071U,
    0xe518e61bU,
    0x9daab716U,
    0xe4769896U,
    0xef3af347U,
    0x385e18afU,
    0x790db350U,
    0x9a4338a3U,
    0xe927237aU,
    0x7a7010a7U,
    0x3c3ea757U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x68000000U,
    0xa4000000U,
    0x36000000U,
    0x2f000000U,
    0x6f800000U,
    0x5c00000U,
    0x40a00000U,
    0xc9b00000U,
    0xd2580000U,
    0x8a940000U,
    0x9c2a0000U,
    0x396f0000U,
    0x638f8000U,
    0xe75e4000U,
    0x2811a000U,
    0xe76a7000U,
    0xa8899800U,
    0xdedb2c00U,
    0x32d38600U,
    0xa04c1100U,
    0xc0f81780U,
    0x7a275dc0U,
    0x5ef59e20U,
    0x81b97d10U,
    0x76c23198U,
    0xc5253cccU,
    0x497611baU,
    0x187e0cf9U,
    0xfc61a9b5U,
    0xd99110d7U,
    0x49aa17b9U,
    0xe1ac5df4U,
    0x65a81e3fU,
    0x3ac3d0cU,
    0x74ae11a0U,
    0xd72a0ccdU,
    0x40eba99bU,
    0x194e10cfU,
    0x907d9792U,
    0x28661dffU,
    0xe793be30U,
    0xf2a94d13U,
    0xd02809beU,
    0xeb6f60feU,
    0x1a898fa1U,
    0x97d871f7U,
    0x2a541811U,
    0x60e6c03U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0xf8000000U,
    0x64000000U,
    0x5e000000U,
    0x9000000U,
    0xff800000U,
    0x9bc00000U,
    0xffe00000U,
    0x4ef00000U,
    0x2f380000U,
    0x2940000U,
    0x980a0000U,
    0x114b0000U,
    0x94ee8000U,
    0x573ec000U,
    0x2692a000U,
    0x2608b000U,
    0x84f2800U,
    0x936f6c00U,
    0xa8fb8600U,
    0x87701b00U,
    0x61f9ad80U,
    0xd8f28cc0U,
    0xa3a2e20U,
    0xcf11b710U,
    0xd8c88bb8U,
    0xda2e27c4U,
    0x8d5eab9eU,
    0xb4a357c9U,
    0xf855a38fU,
    0x446e4bd6U,
    0x1e79ad99U,
    0x3328cebU,
    0x15da2e0eU,
    0x91e1b71cU,
    0xff08ba5U,
    0xbcba27f9U,
    0x4b54abafU,
    0xace857faU,
    0x933b23a1U,
    0x88908bf6U,
    0xc70b0d82U,
    0x6bca3cf5U,
    0x32ad0604U,
    0x1adb29U,
    0x3f010d85U,
    0x2a813cc0U,
    0xbe438619U,
    0x23241b34U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x78000000U,
    0x5c000000U,
    0x6a000000U,
    0x1000000U,
    0x82800000U,
    0x94c00000U,
    0xaa00000U,
    0xf100000U,
    0xdc380000U,
    0xd340000U,
    0xf6ea0000U,
    0x412f0000U,
    0xcc098000U,
    0x735ec000U,
    0xd2006000U,
    0x7d061000U,
    0x7885a800U,
    0x1dc1e400U,
    0xac26de00U,
    0xadd5f900U,
    0xbd9b9280U,
    0x81a20ec0U,
    0x3c92f620U,
    0xd27edd10U,
    0x3f572c98U,
    0xe15ee7fcU,
    0xcf054cbeU,
    0xf583f7c7U,
    0x1d4364a2U,
    0xede3d3ecU,
    0x6bf45a86U,
    0x114afadeU,
    0xbeb8003eU,
    0x29f40038U,
    0xe44a000bU,
    0xe23f000dU,
    0x231801bU,
    0x236ac021U,
    0xcc6a603cU,
    0xa9e91039U,
    0x3cac281cU,
    0xf54f242aU,
    0xa8bebe25U,
    0xd2f7e90bU,
    0xefcc3aa6U,
    0xd078eafbU,
    0xaa57a82bU,
    0x4ddae424U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x88000000U,
    0x14000000U,
    0xa2000000U,
    0x49000000U,
    0x4a800000U,
    0xc3400000U,
    0xfe600000U,
    0xc5b00000U,
    0x84b80000U,
    0xab540000U,
    0xf74e0000U,
    0xfd0f0000U,
    0x89ef8000U,
    0xef1d4000U,
    0x25c42000U,
    0xc321d000U,
    0xa4d02800U,
    0x508e2400U,
    0x6729a200U,
    0x3fb85f00U,
    0x50d43b80U,
    0x828d50c0U,
    0xc62e0a20U,
    0xd13f3b30U,
    0xb997b998U,
    0x21ebdfecU,
    0xeb1d9982U,
    0xdfc50ff5U,
    0xa6223190U,
    0xd0566bceU,
    0x6ccfb388U,
    0x38cfe4e9U,
    0x3acba00dU,
    0xa3cc903aU,
    0xd14c0821U,
    0x8e0bf421U,
    0xc66f8a1aU,
    0xe8dd7b29U,
    0x6fe41986U,
    0x4d734fc2U,
    0x871f9198U,
    0x41c1fbd4U,
    0x89223b90U,
    0x49d650d7U,
    0x300f8a34U,
    0xf96d7b08U,
    0x295c19b9U,
    0x1f274ffbU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0xe8000000U,
    0xec000000U,
    0x62000000U,
    0x4f000000U,
    0xc6800000U,
    0xe1400000U,
    0x8a200000U,
    0x88f00000U,
    0xc4280000U,
    0xea540000U,
    0xd09e0000U,
    0x55db0000U,
    0xe1f98000U,
    0x4c0a4000U,
    0x9da1e000U,
    0xda36d000U,
    0xc2ced800U,
    0x31428400U,
    0x32204200U,
    0x1cf36d00U,
    0xd22f9180U,
    0xc3527640U,
    0xd71f1a20U,
    0x5f1fa930U,
    0x31833b8U,
    0x6918cb5cU,
    0xba19d39aU,
    0xf69e1b4bU,
    0xb4df0b80U,
    0x9a789f7fU,
    0xbcc94993U,
    0x3444f26fU,
    0xb7a15818U,
    0x2937c426U,
    0x1e4e221aU,
    0xeb00fd36U,
    0x18872985U,
    0xb4476267U,
    0x77a1e022U,
    0xc936d010U,
    0x6e4ed809U,
    0x302841fU,
    0xf4804215U,
    0xd6436d29U,
    0x38a7918eU,
    0xfb6765eU,
    0x8f091a3eU,
    0x8920a904U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x68000000U,
    0x6c000000U,
    0x32000000U,
    0xad000000U,
    0x35800000U,
    0x12400000U,
    0x9e200000U,
    0x24900000U,
    0xa4180000U,
    0x59340000U,
    0x11ce0000U,
    0xc18f0000U,
    0x3aaf8000U,
    0x97ba4000U,
    0xcbe7e000U,
    0xa5f25000U,
    0xfb2bc800U,
    0x13f99c00U,
    0x76c64a00U,
    0xf3e7c500U,
    0xd1f26280U,
    0xad2b47c0U,
    0x50fa0220U,
    0xb4401930U,
    0x1525c898U,
    0x5a15d2ecU,
    0xa8da28baU,
    0x7bd382ebU,
    0x7fbfe0b4U,
    0x67e51ef7U,
    0xf7f62aafU,
    0xe6289bc3U,
    0x4e7ba809U,
    0x8858c25U,
    0x5fc4622aU,
    0x5863090eU,
    0x3cb060b8U,
    0x1b8f5ecdU,
    0x3ba9caa1U,
    0xf03ecbe5U,
    0xc4a6600dU,
    0x6657101eU,
    0xa1fba838U,
    0x1bc58c39U,
    0xa664621aU,
    0x73b30926U,
    0x5b0860b3U,
    0x186b5efcU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0xe8000000U,
    0x5c000000U,
    0x92000000U,
    0xe5000000U,
    0x75800000U,
    0xddc00000U,
    0xb7a00000U,
    0xfe900000U,
    0x50f80000U,
    0xa9b40000U,
    0xae2a0000U,
    0x2f2b0000U,
    0x2ca98000U,
    0x226e4000U,
    0xd34c6000U,
    0x469d5000U,
    0x22840800U,
    0x1d401c00U,
    0xe7e07600U,
    0xd0f78900U,
    0x3fccae80U,
    0x2cdf20c0U,
    0x55e7fe20U,
    0x65f2d510U,
    0xb24938b8U,
    0xb51bb9fcU,
    0xc443389aU,
    0xb060b9c7U,
    0xe032b8bcU,
    0x51eaf9d5U,
    0x758cd8bfU,
    0xbb8a9ccU,
    0xd55350abU,
    0x7459f5fdU,
    0xf1a4c68fU,
    0x65926cdfU,
    0x5a7b801dU,
    0x48f14001U,
    0xcbcfe009U,
    0x2d81038U,
    0xd2e1e825U,
    0x1d730c16U,
    0x46081e28U,
    0x4d7ac51fU,
    0x6870d089U,
    0x6b8cb5c4U,
    0xc4b9268aU,
    0xe1d57cdbU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x88000000U,
    0x3c000000U,
    0xa6000000U,
    0xad000000U,
    0x67800000U,
    0xdbc00000U,
    0xd4a00000U,
    0x5e500000U,
    0x580000U,
    0x70740000U,
    0xda4a0000U,
    0x3b4f0000U,
    0xa2ca8000U,
    0x94000U,
    0xb9286000U,
    0xbfb8f000U,
    0x9dc14800U,
    0xa9a62c00U,
    0xd1d4a600U,
    0x779ce100U,
    0x8ad6e880U,
    0x1518fdc0U,
    0xfa956e20U,
    0xa47c8d10U,
    0x6960ae98U,
    0x7f5ace4U,
    0xc98aae82U,
    0xb3eaacdfU,
    0xca982e91U,
    0x2557ecdfU,
    0x92da4e83U,
    0x18301ccdU,
    0xdf298686U,
    0x32bb70c7U,
    0xa4740aaU,
    0x6a6461faU,
    0xb1706029U,
    0xb3ccf03fU,
    0x818b4817U,
    0xafe92c3dU,
    0x9c9e263fU,
    0x9055a138U,
    0x415e88bcU,
    0x59f00df2U,
    0x8c2608U,
    0xa66ea111U,
    0xb65e0898U,
    0x15764de1U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x28000000U,
    0x94000000U,
    0x82000000U,
    0xa1000000U,
    0x13800000U,
    0x69400000U,
    0x90200000U,
    0x7d700000U,
    0xa7580000U,
    0x6bd40000U,
    0x52ea0000U,
    0x8cef0000U,
    0x2bec8000U,
    0x536c4000U,
    0x4cab2000U,
    0xc4ad000U,
    0xbd5f6800U,
    0x66d6ec00U,
    0x1f6c3200U,
    0x2ad6500U,
    0xa3484580U,
    0x61db95c0U,
    0x8315da20U,
    0x4288c910U,
    0xb63bd798U,
    0x328460d4U,
    0x3ac3d7aaU,
    0x996060f5U,
    0xbd51d798U,
    0xf22b60ecU,
    0x588f57b6U,
    0xbb3c20fbU,
    0x7f02f7b6U,
    0xb485b0d6U,
    0x11c23fbfU,
    0x8fe1ccc9U,
    0x3d9045a6U,
    0x164f95efU,
    0xb05fda11U,
    0x2b57c902U,
    0x912f57bcU,
    0x8a0c20e6U,
    0x91faf7b2U,
    0xae61b0ecU,
    0xcdd03f9bU,
    0x49eaccc7U,
    0xe26ec5b1U,
    0x1728d5cbU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0xb8000000U,
    0x94000000U,
    0xa000000U,
    0xb000000U,
    0x4800000U,
    0x7cc00000U,
    0x2200000U,
    0x9d00000U,
    0xb3980000U,
    0xe0740000U,
    0xd00a0000U,
    0xe20b0000U,
    0x3d0b8000U,
    0xd38cc000U,
    0x14482000U,
    0xaae91000U,
    0x4bfdb800U,
    0xf0459400U,
    0x12e20e00U,
    0x75f3dd00U,
    0xb4f6480U,
    0xe46b27c0U,
    0xdc3e3620U,
    0x46e18910U,
    0xdff6cab8U,
    0x50492aecU,
    0xe8eeca8eU,
    0x8cfd2af5U,
    0xdac4ca9aU,
    0xd7262afeU,
    0xec574ab7U,
    0x70deeac6U,
    0x22156ab4U,
    0x333cfad9U,
    0x50635299U,
    0x6c35aefcU,
    0x54e97c81U,
    0xbaff63feU,
    0xa7c3a016U,
    0x4ea5d00aU,
    0xb9159808U,
    0x9fbc843cU,
    0x50a7b628U,
    0xf8124932U,
    0x703f6a95U,
    0x98e7fadcU,
    0x3ef0d2a3U,
    0xefcd6ed6U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xd0000000U,
    0xc8000000U,
    0x24000000U,
    0x62000000U,
    0x7000000U,
    0x13800000U,
    0xec400000U,
    0xa8a00000U,
    0x6df00000U,
    0x2d080000U,
    0x5dd40000U,
    0xe93e0000U,
    0x793b0000U,
    0x913c8000U,
    0x853b4000U,
    0xff3aa000U,
    0x143ed000U,
    0x41bcd800U,
    0xc8ff6c00U,
    0x74d8aa00U,
    0xe6ea1100U,
    0x8f035980U,
    0x17801b40U,
    0xbe46f220U,
    0xb7a53d30U,
    0x9275d3b8U,
    0x874b9a74U,
    0x9075d392U,
    0x904b9a79U,
    0xabf5d380U,
    0x880b9a45U,
    0xa955d3aeU,
    0xc6fb9a76U,
    0xf5ddd398U,
    0x706f9a57U,
    0xa7c3d3bdU,
    0x88e49a60U,
    0xb35753a6U,
    0x3dfbda72U,
    0x885bf3b3U,
    0xd2a0a5aU,
    0xb1e5ab92U,
    0x39d52660U,
    0xab3b2199U,
    0x4e3aa776U,
    0x9abe002bU,
    0x857b0038U,
    0x119c8025U,
    0x1ccb4006U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0xa8000000U,
    0x3c000000U,
    0xe2000000U,
    0x51000000U,
    0x84800000U,
    0xd1400000U,
    0x29a00000U,
    0xe2900000U,
    0xbfa80000U,
    0xb40000U,
    0x5bfa0000U,
    0x2790000U,
    0x5dbf8000U,
    0x685e4000U,
    0xf5eb6000U,
    0x8090d000U,
    0xaea8a800U,
    0x24371400U,
    0xbabc3200U,
    0x83dd5100U,
    0x832e4280U,
    0x35f5c640U,
    0xa4591a20U,
    0x5fe90510U,
    0x7d9490a8U,
    0x502b074cU,
    0x3071108aU,
    0xc99c475fU,
    0x30df0b0U,
    0xc0a6d748U,
    0x82143883U,
    0xb8e81367U,
    0x96172290U,
    0x26e8167fU,
    0x85143214U,
    0xc3695133U,
    0x78d44287U,
    0x78cc65bU,
    0x51e69a21U,
    0xbb7451dU,
    0x6a7ff0bfU,
    0x81bbd774U,
    0x1a59b8a4U,
    0x3ceb5340U,
    0x9011c287U,
    0xa1eb866aU,
    0xbe927a1bU,
    0x8da9d519U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x68000000U,
    0x6c000000U,
    0x92000000U,
    0x6d000000U,
    0x94800000U,
    0x51c00000U,
    0x25600000U,
    0xcb500000U,
    0xd9680000U,
    0x1b40000U,
    0x847e0000U,
    0xe7fd0000U,
    0x73bf8000U,
    0xf9984000U,
    0x9f6de000U,
    0xaab15000U,
    0x7bf9c800U,
    0x69b89400U,
    0x689ad600U,
    0x41e95700U,
    0x92726880U,
    0x341ff240U,
    0xc2a9e20U,
    0x6d108330U,
    0x4e4cde88U,
    0xbbc6b56cU,
    0x3c655ebaU,
    0xe9d7f56bU,
    0xaba93e8cU,
    0x2753e547U,
    0x8b6a9697U,
    0x4cb66153U,
    0xa0fa689fU,
    0xde3bf26eU,
    0x3adc9e05U,
    0xa0c98308U,
    0x2b055ebbU,
    0x3f87f573U,
    0xae413e98U,
    0xab27e568U,
    0xd0749690U,
    0x611b6172U,
    0xcade88bU,
    0x1ad7b27cU,
    0x902f7e29U,
    0x7715d313U,
    0xdf4b1685U,
    0x65432142U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0x98000000U,
    0x9c000000U,
    0xf6000000U,
    0x49000000U,
    0x20800000U,
    0x1400000U,
    0x84a00000U,
    0x88b00000U,
    0x9df80000U,
    0xa7940000U,
    0x1f8a0000U,
    0x440f0000U,
    0x66c88000U,
    0xcca9c000U,
    0xe07f6000U,
    0x5450f000U,
    0x96de800U,
    0x9f9f1400U,
    0x1f455200U,
    0x39a1f900U,
    0xea31d380U,
    0xf7be7cc0U,
    0xb8b23a20U,
    0x95fc2d10U,
    0xc3916198U,
    0xa58db5f4U,
    0x630be186U,
    0x934f75f7U,
    0x1b6e0185U,
    0x2c9d45f6U,
    0x78c60996U,
    0x1d6061c3U,
    0x21535382U,
    0x2fe8bcf3U,
    0xffddda0cU,
    0x6f611d38U,
    0x4251e9adU,
    0xd06951f2U,
    0x4719dbacU,
    0x520358ddU,
    0xd3076034U,
    0xb784f039U,
    0xbcc7e82cU,
    0x5760141dU,
    0x2e55d23cU,
    0xe6c3907U,
    0xda1cb38cU,
    0x80858cd6U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x58000000U,
    0x7c000000U,
    0x2a000000U,
    0x73000000U,
    0xe2800000U,
    0xd6400000U,
    0x58e00000U,
    0x1ef00000U,
    0x11280000U,
    0x30540000U,
    0x673a0000U,
    0x178f0000U,
    0x72468000U,
    0x4ee3c000U,
    0x27f06000U,
    0x90ab1000U,
    0x5c96b800U,
    0x959fa400U,
    0x7b9b7a00U,
    0xe98cb00U,
    0x8d1e9180U,
    0xa6da9c40U,
    0xb8fda220U,
    0x27ac7f10U,
    0xb8135398U,
    0x22ddf344U,
    0xdef849b6U,
    0x56ae284fU,
    0x1d906032U,
    0x181b1008U,
    0x755eb836U,
    0xab3ba42eU,
    0x55897a0aU,
    0x5543cb04U,
    0xb26211b6U,
    0x8cb65c65U,
    0xfcb4217U,
    0x2fa4af27U,
    0x9b958b96U,
    0x3919475cU,
    0xe8dd8ba2U,
    0xfdfd4762U,
    0x8c2f8ba0U,
    0xa7d6477bU,
    0x32fb0b9dU,
    0x64ae874aU,
    0x7297eb8fU,
    0xc0995778U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x78000000U,
    0x74000000U,
    0x9e000000U,
    0x67000000U,
    0x17800000U,
    0xb2400000U,
    0xd8e00000U,
    0xc3100000U,
    0xc5880000U,
    0x91b40000U,
    0x8a7e0000U,
    0x992b0000U,
    0xb1418000U,
    0x9962c000U,
    0xfa572000U,
    0x90ecb000U,
    0x61e75800U,
    0x9391ec00U,
    0xf04d5a00U,
    0xf312c300U,
    0x9d8ef680U,
    0x75b5d940U,
    0x1c7d2220U,
    0xf22f9f30U,
    0x4cc4f4b8U,
    0xd226f65cU,
    0x52368ebeU,
    0xf63c856dU,
    0xcec9203fU,
    0x19d7b035U,
    0x7caed823U,
    0x3e072c2dU,
    0x37047a0fU,
    0xff857334U,
    0xce402e93U,
    0x4ae2f551U,
    0x4e115831U,
    0x2b0eec0aU,
    0x5372da13U,
    0xf75b032eU,
    0x309856b3U,
    0x6f3ba97bU,
    0xe4d5a11U,
    0x2412c30dU,
    0x120ef693U,
    0xc3f5d94eU,
    0x229d2209U,
    0x223f9f00U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0xe8000000U,
    0x24000000U,
    0xc2000000U,
    0x21000000U,
    0xeb800000U,
    0xb0400000U,
    0x29200000U,
    0x5dd00000U,
    0x2a280000U,
    0xcb40000U,
    0x6ada0000U,
    0xf1490000U,
    0x19448000U,
    0x56a04000U,
    0x3796a000U,
    0xbe0bd000U,
    0xb4e36800U,
    0x37b26c00U,
    0xdc5f9a00U,
    0x54899d00U,
    0xff217080U,
    0xe6d3ba40U,
    0xdcaa5220U,
    0x9702110U,
    0xdcbd82a8U,
    0xa9384b54U,
    0xd67cb89aU,
    0x23de0659U,
    0x7ecca038U,
    0xcb02d00cU,
    0xe87e828U,
    0x69c22c31U,
    0x613a00U,
    0x53764d16U,
    0x21b81880U,
    0xecb8d667U,
    0xf139480aU,
    0xea7dfc17U,
    0x7dd8522eU,
    0x51cd2115U,
    0xe78302a7U,
    0xb6410b59U,
    0x3a2698a1U,
    0x2f51967cU,
    0xbdeb680eU,
    0x73d66c07U,
    0xfd2d9a07U,
    0x4c349d11U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0xa8000000U,
    0x24000000U,
    0xde000000U,
    0xd1000000U,
    0x23800000U,
    0xbc00000U,
    0x3d600000U,
    0x85f00000U,
    0x38280000U,
    0x7540000U,
    0x36be0000U,
    0x2f8f0000U,
    0xe4428000U,
    0x6ca04000U,
    0xc792a000U,
    0xf75b3000U,
    0x9b3ee800U,
    0xfd4ccc00U,
    0xaea12600U,
    0x8493ff00U,
    0x9fdbad80U,
    0xac789440U,
    0x916d6e20U,
    0x1a740330U,
    0xd1ec6398U,
    0xc933a744U,
    0x3c9e58aU,
    0xbce06879U,
    0xdd30200fU,
    0xc5cb7000U,
    0x91e4483aU,
    0xbcb3fc0fU,
    0x4d09ce12U,
    0xa404331cU,
    0x1e060bb1U,
    0xb1042b77U,
    0x3386e3aaU,
    0xa3c7e758U,
    0x19654580U,
    0x5bf4584fU,
    0xe92c4820U,
    0x24d7fc0aU,
    0x3d7fce17U,
    0x12ef3316U,
    0x61b28bb6U,
    0x548f6b78U,
    0xc0c0c3a0U,
    0xc1e7974eU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x58000000U,
    0x9c000000U,
    0x2a000000U,
    0x87000000U,
    0xde800000U,
    0x40c00000U,
    0x4be00000U,
    0xee300000U,
    0xaa480000U,
    0x7f140000U,
    0x109e0000U,
    0x19ef0000U,
    0x9f038000U,
    0xe287c000U,
    0x3ac66000U,
    0xa4e21000U,
    0xf4b21800U,
    0x5c8b4400U,
    0x99f36e00U,
    0xa72e9300U,
    0x2de48380U,
    0xeb328b40U,
    0x57cf1620U,
    0x6753c730U,
    0xeebbf598U,
    0x79b85c4cU,
    0x6f3b7bb6U,
    0xbbf81f57U,
    0x2e1e0032U,
    0xa92f001dU,
    0xece38039U,
    0xa0b7c03bU,
    0xe28e600eU,
    0xc0f6103dU,
    0x10ac1817U,
    0x82a4442eU,
    0x9390ee21U,
    0xeb59532dU,
    0xf68ae39fU,
    0xdef49b5cU,
    0x19ab0e00U,
    0x5d238300U,
    0xf8d51b86U,
    0x25fe0f65U,
    0xe71a1839U,
    0x16af442fU,
    0xb7a56e2aU,
    0x36159329U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x58000000U,
    0x64000000U,
    0xe2000000U,
    0xfb000000U,
    0xc5800000U,
    0x4f400000U,
    0x27a00000U,
    0x5b00000U,
    0x3dd80000U,
    0x3f540000U,
    0x8cca0000U,
    0x8ebb0000U,
    0x8f028000U,
    0x3f87c000U,
    0x3040a000U,
    0x90215000U,
    0xe9f63800U,
    0xbbfa1c00U,
    0x97a6b200U,
    0x6db2c900U,
    0x71db7b80U,
    0x715229c0U,
    0xa9c82a20U,
    0x363d8510U,
    0x1cc1f1b8U,
    0x69e1fce4U,
    0x989763b6U,
    0xc0aaa5c9U,
    0xc96a0020U,
    0x300b000aU,
    0x975a801fU,
    0xdf93c02eU,
    0xc32aa027U,
    0x7f2a5026U,
    0xb92cb806U,
    0x4029dc22U,
    0xeeac121bU,
    0x3c689931U,
    0x308fc3a5U,
    0xf09ff5c1U,
    0x33f63800U,
    0xf4fa1c18U,
    0x826b202U,
    0x2df2c926U,
    0x29fb7b9cU,
    0xa4a229e9U,
    0x94302a32U,
    0xb8998538U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x18000000U,
    0x54000000U,
    0xce000000U,
    0x85000000U,
    0xc2800000U,
    0xebc00000U,
    0x9a600000U,
    0x6d500000U,
    0xdcb80000U,
    0x1af40000U,
    0x28e0000U,
    0x439d0000U,
    0xec068000U,
    0xca07c000U,
    0x73072000U,
    0x83823000U,
    0xff464800U,
    0xa0a2c400U,
    0xfbb50600U,
    0xdf2eef00U,
    0x9eae2080U,
    0x9eef68c0U,
    0x7cc6e20U,
    0xd83a1b30U,
    0xf8336e88U,
    0xe06e43e4U,
    0x1e09c8a6U,
    0xf5585ce5U,
    0xc0e0001bU,
    0x12900035U,
    0xa8d8003eU,
    0x62a4000bU,
    0x4b60013U,
    0xe6a9002fU,
    0xbae88034U,
    0x61cac01dU,
    0x8139a016U,
    0xb8b1f031U,
    0x14af682aU,
    0xdedf42fU,
    0x344dce2bU,
    0xaf7feb36U,
    0x14920684U,
    0x81deb7caU,
    0x8a2286b9U,
    0x2c7077d5U,
    0x774d26acU,
    0x64f887fbU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x8000000U,
    0x2c000000U,
    0xa000000U,
    0xf000000U,
    0xf8800000U,
    0x77400000U,
    0x3be00000U,
    0xc0700000U,
    0xd9c80000U,
    0x1ed40000U,
    0x395a0000U,
    0xb1ef0000U,
    0xd0058000U,
    0x58014000U,
    0x3401e000U,
    0x5e039000U,
    0x21015800U,
    0xd186f400U,
    0x8ac39200U,
    0xbb266300U,
    0x74527d80U,
    0x551cab40U,
    0x3c8b2a20U,
    0x3e370710U,
    0x4faab798U,
    0x46633c5cU,
    0xb745a2U,
    0xde88f5bU,
    0xe200003aU,
    0x1300000fU,
    0x9a800024U,
    0x2440000aU,
    0xc160000eU,
    0x94300038U,
    0x10a8000aU,
    0xa6e40019U,
    0x23f2003eU,
    0x180b0001U,
    0xb77800cU,
    0x374a4021U,
    0x4966035U,
    0xa939d03cU,
    0xfc5f3832U,
    0x666b240eU,
    0x4fc6aa1bU,
    0x6ca2472fU,
    0xeb915798U,
    0x61bfac76U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x78000000U,
    0x9c000000U,
    0xd2000000U,
    0x83000000U,
    0xed800000U,
    0x74400000U,
    0x16a00000U,
    0x36100000U,
    0x77280000U,
    0x9af40000U,
    0x469a0000U,
    0xd54f0000U,
    0x2878000U,
    0x13c2c000U,
    0x2de66000U,
    0x5734d000U,
    0x527b3800U,
    0x58f91c00U,
    0xcf3f2a00U,
    0x9cdec100U,
    0xe2e9fe80U,
    0x4217ee40U,
    0x512a7220U,
    0x7ff70d10U,
    0x2e1fec98U,
    0xf98b334cU,
    0x40e126beU,
    0xfbb3e277U,
    0x60bd800cU,
    0x39dc03cU,
    0xadc9e03dU,
    0x6421036U,
    0x85a7582fU,
    0xf392cc0aU,
    0xd76b9212U,
    0xba511d1bU,
    0xbd8ab4afU,
    0x1ee2ff65U,
    0x82b734b8U,
    0xda3f3f7fU,
    0xac5ed4bfU,
    0x22d2f5aU,
    0x9a718ca3U,
    0xde5be35cU,
    0x91281eaaU,
    0x5ff1fe6dU,
    0x7e1f2a0eU,
    0xb18ec115U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x98000000U,
    0xfc000000U,
    0xf6000000U,
    0x83000000U,
    0xcf800000U,
    0x83400000U,
    0x39200000U,
    0x9a700000U,
    0x61f80000U,
    0x7a140000U,
    0x8faa0000U,
    0x411b0000U,
    0x26828000U,
    0xe9c1c000U,
    0x8ee46000U,
    0x8397f000U,
    0x2fef3800U,
    0xc7ba2c00U,
    0x37b1b200U,
    0xd51ead00U,
    0x64813580U,
    0x18c49cc0U,
    0x5862ea20U,
    0x22577110U,
    0x250dbfb8U,
    0xde2f1de4U,
    0xf55aed86U,
    0xa22780efU,
    0xe9f08025U,
    0xf4bec014U,
    0x5034e02dU,
    0x4259303bU,
    0x5fa3d815U,
    0x93b71c39U,
    0x7f186a16U,
    0x6182b11eU,
    0x5c43df9eU,
    0x70a3edd6U,
    0x923755bfU,
    0xf35c6cd2U,
    0x69255200U,
    0xa2779d1aU,
    0xedfaedbfU,
    0x841780c0U,
    0x68a8801cU,
    0x849ac020U,
    0xd0c6e007U,
    0x9c663033U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0xc8000000U,
    0x8c000000U,
    0xf6000000U,
    0x2d000000U,
    0xaa800000U,
    0xc9400000U,
    0x3ca00000U,
    0xfd100000U,
    0xdcc80000U,
    0x75f40000U,
    0xaffe0000U,
    0xc96b0000U,
    0x78658000U,
    0x71774000U,
    0x3fbce000U,
    0xd4a7000U,
    0xeb314800U,
    0x289f8400U,
    0x759c1200U,
    0xd71e1900U,
    0x725bd680U,
    0x48f81940U,
    0x50efba20U,
    0xaaa0ed30U,
    0xe0130cb8U,
    0x5e4ec47cU,
    0xc0b49e92U,
    0xad5c9d53U,
    0x957e2825U,
    0xf82db407U,
    0xe0c23a20U,
    0x663ad2dU,
    0xf071ec80U,
    0x733fb457U,
    0xd1085680U,
    0x39105954U,
    0x66c8da16U,
    0x4ef6dd33U,
    0xd87b24acU,
    0x62ac7061U,
    0x18524bcU,
    0x46c77074U,
    0x3360a480U,
    0xef03051U,
    0xf87c4483U,
    0x72aa4048U,
    0x39850c86U,
    0x2c1c45dU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0xe8000000U,
    0x24000000U,
    0x5e000000U,
    0xe3000000U,
    0x19800000U,
    0xd8c00000U,
    0xea200000U,
    0x63900000U,
    0x46f80000U,
    0xda740000U,
    0xb2ca0000U,
    0x59dd0000U,
    0x77648000U,
    0xa3744000U,
    0x264aa000U,
    0xe39b3000U,
    0xc0077800U,
    0x60001c00U,
    0x7000ae00U,
    0xb8054500U,
    0xcc037680U,
    0x7a0070c0U,
    0xbd077620U,
    0xfa836910U,
    0xc1402088U,
    0x32e169d4U,
    0x89b60e9aU,
    0x25696cd9U,
    0x9c89583fU,
    0x68bf6c3cU,
    0xeb157634U,
    0x2eba692bU,
    0xd416a097U,
    0x853c29e9U,
    0xc5d22ea0U,
    0x239b1ce8U,
    0xa0000038U,
    0x1000003cU,
    0xc8000004U,
    0x7400001bU,
    0xb6000019U,
    0xc7000029U,
    0x47800010U,
    0x3bc00028U,
    0xf3a00034U,
    0xbb500010U,
    0xacd80021U,
    0xb9e40010U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0xf8000000U,
    0x24000000U,
    0x8e000000U,
    0x75000000U,
    0x8e800000U,
    0xb1c00000U,
    0x68a00000U,
    0x59b00000U,
    0xc4180000U,
    0x2dd40000U,
    0xf8a0000U,
    0x67bb0000U,
    0xd7e18000U,
    0xc0d2c000U,
    0x1509a000U,
    0x40783000U,
    0x1e434800U,
    0x61e42400U,
    0x31d2d600U,
    0x58cd900U,
    0x8cb9d780U,
    0x2463bbc0U,
    0x23923e20U,
    0x666bcd10U,
    0x8a69c9b8U,
    0x786986fcU,
    0xb7689f9eU,
    0xcae89fd9U,
    0x5cab683bU,
    0xa14ed431U,
    0xda983e25U,
    0xd410cd19U,
    0x5b284987U,
    0x640b46c2U,
    0x10f93f91U,
    0x3284afeaU,
    0xabc22022U,
    0x8ba1f00fU,
    0x76336818U,
    0x8d5ad418U,
    0xa5b23e2aU,
    0x3e1bcd16U,
    0x3ed1c9b1U,
    0xd80d86c0U,
    0xafa9fa5U,
    0xd1879fcbU,
    0x8440e812U,
    0xc2e71430U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x28000000U,
    0x4c000000U,
    0x9e000000U,
    0x19000000U,
    0x96800000U,
    0x67c00000U,
    0xe9e00000U,
    0x28500000U,
    0x65980000U,
    0x27740000U,
    0x86ae0000U,
    0xd77d0000U,
    0xf6a28000U,
    0x67764000U,
    0x66abe000U,
    0xe7793000U,
    0xcea1e800U,
    0x3752400U,
    0xb4ae8600U,
    0x607ef300U,
    0x4127e880U,
    0xf236cdc0U,
    0x3a8a8e20U,
    0xa1cfe730U,
    0xcea0688U,
    0xb0db5ac4U,
    0x9b5000aaU,
    0xf01ae9e3U,
    0x2d30880fU,
    0x210e5432U,
    0x9a8a8e27U,
    0x71cfe70eU,
    0x4ea0697U,
    0xecdb5adfU,
    0x2d5000bbU,
    0xa51ae9e6U,
    0x25b08816U,
    0x5fce5412U,
    0xe5ea8e1cU,
    0x3e5fe70dU,
    0x8892068cU,
    0xe3ff5ad0U,
    0xce66008dU,
    0x5513e9c3U,
    0x55bc083eU,
    0xefc5141fU,
    0x75e3ee1bU,
    0xbe509711U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x88000000U,
    0xcc000000U,
    0xc6000000U,
    0xb9000000U,
    0xdf800000U,
    0xac00000U,
    0x41600000U,
    0xdb00000U,
    0xd680000U,
    0x43140000U,
    0x303e0000U,
    0x2c0d0000U,
    0x9ea38000U,
    0xa5534000U,
    0x1b1ee000U,
    0x2c9ff000U,
    0x3a5de800U,
    0x253dac00U,
    0xe5892e00U,
    0xb567ff00U,
    0x4fb26780U,
    0x9e6ea640U,
    0xd3942620U,
    0xd478a330U,
    0x742d2188U,
    0x1eb3b54cU,
    0x5def8f82U,
    0x87530a43U,
    0x581d0819U,
    0x241f5c12U,
    0xaa1f461dU,
    0xff1d133dU,
    0xd69ba983U,
    0xed5ba95eU,
    0xbfb829a5U,
    0x7ec8e94cU,
    0x8bc6c983U,
    0x1ae71976U,
    0x2d7321beU,
    0x30eb575U,
    0xa0240fb2U,
    0x24144a71U,
    0xe2bde829U,
    0x5b4dac08U,
    0x56012e24U,
    0xc103ff1fU,
    0xbb8467acU,
    0x30c7a669U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0xe8000000U,
    0x34000000U,
    0x66000000U,
    0x5d000000U,
    0x3800000U,
    0xf8c00000U,
    0x5ea00000U,
    0xbc900000U,
    0xd6880000U,
    0xdbf40000U,
    0x853e0000U,
    0x822d0000U,
    0xb0648000U,
    0x47304000U,
    0x959e2000U,
    0x77bef000U,
    0xd36e0800U,
    0xd1018c00U,
    0x79879e00U,
    0xd7c0c100U,
    0xb6225e80U,
    0xf2d65740U,
    0x4769b620U,
    0xa702bd30U,
    0xec874888U,
    0x30435a74U,
    0xc0e4569aU,
    0xc573db7dU,
    0x9e782831U,
    0xd44b7c13U,
    0x6a579632U,
    0xf12c4d07U,
    0xcae140b4U,
    0x1276d675U,
    0x8afdc88cU,
    0x180e1a6fU,
    0x28b6f69eU,
    0xe7596b75U,
    0x3e1e001cU,
    0x5b7d0001U,
    0x1bcc801aU,
    0x8944025U,
    0x7088201aU,
    0xa6f7f021U,
    0x56bc8819U,
    0x92e8cc3fU,
    0xdac33e1eU,
    0x9da37132U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x38000000U,
    0x64000000U,
    0x96000000U,
    0x11000000U,
    0x4d800000U,
    0xf5400000U,
    0x99e00000U,
    0x4cd00000U,
    0x3e580000U,
    0x1d740000U,
    0x2cea0000U,
    0x19bf0000U,
    0xaca18000U,
    0xf530c000U,
    0xe28fa000U,
    0xeb2ef000U,
    0x6d9c6800U,
    0xc750fc00U,
    0x321d1200U,
    0x5176b00U,
    0xaf78c080U,
    0x6503c7c0U,
    0x5384da20U,
    0xd8466710U,
    0xd6603a98U,
    0x69090fcU,
    0x9f3ca8aeU,
    0x7f673bc9U,
    0x6f13c81dU,
    0x9c7e0c28U,
    0x7817a05U,
    0x56479718U,
    0x3365d28dU,
    0x1514acceU,
    0x277c1a99U,
    0x5905a0eeU,
    0x5184e086U,
    0x6746f7c4U,
    0xeee49234U,
    0x2853ab1cU,
    0xe29d608fU,
    0xe7d237daU,
    0xe8d93223U,
    0x26b65b3aU,
    0x7aca8897U,
    0x5d4d0bd0U,
    0x9b0a001cU,
    0x606f0032U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xd8000000U,
    0xb4000000U,
    0x8e000000U,
    0x49000000U,
    0x70800000U,
    0xbec00000U,
    0x2f200000U,
    0x47300000U,
    0x59780000U,
    0x19d40000U,
    0x34ee0000U,
    0xbbd90000U,
    0x7ea38000U,
    0x2575c000U,
    0x52992000U,
    0x86c73000U,
    0xcb272800U,
    0x81375400U,
    0x4c7a4e00U,
    0x8b527500U,
    0xf92feb80U,
    0x237b34c0U,
    0xbed24620U,
    0x9d6f1130U,
    0x9d9f0da8U,
    0x8342d5ccU,
    0x77654396U,
    0x8bd0a0ddU,
    0xffeaa82bU,
    0x285b942eU,
    0x9fe0ee02U,
    0xe110850eU,
    0xcec9e39eU,
    0xce6f50eeU,
    0xda192009U,
    0x7c07300bU,
    0x12072832U,
    0xb07541cU,
    0x33824e34U,
    0xd1467533U,
    0x1c61ebacU,
    0x285234dbU,
    0xc6a9c61cU,
    0x583ed118U,
    0x8db02dbbU,
    0xe0b8e5d7U,
    0xaf77eb98U,
    0x8d9f34d2U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x8000000U,
    0x9c000000U,
    0x8a000000U,
    0x71000000U,
    0xe800000U,
    0x5cc00000U,
    0x83600000U,
    0x31900000U,
    0x75280000U,
    0x51340000U,
    0x865a0000U,
    0x50cf0000U,
    0x90668000U,
    0x82114000U,
    0xc56f6000U,
    0x73909000U,
    0xb82d2800U,
    0xa5b7c400U,
    0xb39fbe00U,
    0x4928a700U,
    0xeb317e80U,
    0x8f584440U,
    0xda4ff620U,
    0xd2a4f310U,
    0x66776898U,
    0x17d6744U,
    0x25fad6a2U,
    0x383ec077U,
    0xcddf281aU,
    0x248cc408U,
    0xccc33e19U,
    0xcb66e734U,
    0xcd909e80U,
    0xef2d9477U,
    0x2837be1cU,
    0x14dca70fU,
    0x860b7e9aU,
    0x6207444bU,
    0xbd01763bU,
    0xec81b30fU,
    0xa1c208b7U,
    0xfe2f744U,
    0x80517e8aU,
    0x72c8447dU,
    0x4d67f603U,
    0x7e90f32aU,
    0x6cad68a1U,
    0xe0726776U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0xb8000000U,
    0xf4000000U,
    0x36000000U,
    0xb7000000U,
    0x99800000U,
    0x6bc00000U,
    0xa8200000U,
    0xac500000U,
    0x7df80000U,
    0x92b40000U,
    0xbd0a0000U,
    0x1c5f0000U,
    0xad208000U,
    0xfcd4c000U,
    0xc4ba2000U,
    0x62d6f000U,
    0x1fbdd800U,
    0x8955dc00U,
    0x9d7bd600U,
    0x43f6f500U,
    0x169d080U,
    0x7cee80c0U,
    0xb92e2e20U,
    0xe20ed910U,
    0x89dd5e98U,
    0x64e269f4U,
    0x5f4888eU,
    0xfe6f9cedU,
    0x196fd835U,
    0x78eedc09U,
    0x87295610U,
    0xc9093513U,
    0x6a597089U,
    0x7a23b0fbU,
    0xb5515614U,
    0x177d3503U,
    0x3ef370bbU,
    0x85ecb0c2U,
    0x43a9d610U,
    0x964df51dU,
    0x953b5094U,
    0xb51140e2U,
    0x7d9e8e2bU,
    0x3803e92dU,
    0xb4022682U,
    0x560185c5U,
    0x6702a681U,
    0x218545f6U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x28000000U,
    0xec000000U,
    0xea000000U,
    0xc5000000U,
    0x97800000U,
    0xb4400000U,
    0x6da00000U,
    0x2d300000U,
    0x32080000U,
    0x3540000U,
    0x56de0000U,
    0xb9cb0000U,
    0xf6b08000U,
    0x1c484000U,
    0x33f6e000U,
    0x786a9000U,
    0x11830800U,
    0x5b400c00U,
    0xdf229a00U,
    0x5ef41900U,
    0xc3ee6780U,
    0xafc21740U,
    0xa3e7f220U,
    0x4d96c530U,
    0x34b915b8U,
    0x471c9254U,
    0x112807aaU,
    0xbfa0c74bU,
    0xc321a22U,
    0x138c5915U,
    0xe0108797U,
    0xadfc8742U,
    0xb13afa0bU,
    0x325dc931U,
    0xd0b0fabU,
    0x59d0cb48U,
    0x29188001U,
    0xf22c4003U,
    0xb720e028U,
    0x92f59029U,
    0x99ed883aU,
    0x12c34c26U,
    0xf064fa0dU,
    0xffd6c905U,
    0x761b8fafU,
    0x38a88b63U,
    0xe66016U,
    0x6512d036U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0xd8000000U,
    0x84000000U,
    0xc6000000U,
    0xed000000U,
    0xf2800000U,
    0xb1c00000U,
    0x53200000U,
    0xc0500000U,
    0x9f880000U,
    0x40340000U,
    0xa53e0000U,
    0x68c90000U,
    0x6cd28000U,
    0x274dc000U,
    0x9795a000U,
    0xf1ae7000U,
    0xae48800U,
    0x1f730400U,
    0x65df1600U,
    0x60bac500U,
    0xe48ee880U,
    0xe7b7a340U,
    0xd27cbe20U,
    0x842a7130U,
    0x4ea0d6a8U,
    0xc8101244U,
    0xd2e9c896U,
    0x58041351U,
    0x44059639U,
    0x6603050fU,
    0xfd0548a2U,
    0x2a80d349U,
    0x35c2b633U,
    0x9520b51aU,
    0x2d546082U,
    0x6d0da743U,
    0xf1f12817U,
    0xf61d7417U,
    0xa89b9e31U,
    0xf359c116U,
    0x6779fea1U,
    0x32a9667bU,
    0x996456b5U,
    0x6630d26fU,
    0x3838e8b4U,
    0xf24aa365U,
    0x91103e1dU,
    0xee6eb13dU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x68000000U,
    0x24000000U,
    0xce000000U,
    0xc9000000U,
    0xb800000U,
    0xedc00000U,
    0xfda00000U,
    0xa500000U,
    0x41680000U,
    0xf9b40000U,
    0x3f5a0000U,
    0xe12f0000U,
    0x54d68000U,
    0x52a4000U,
    0xbad6a000U,
    0xbc2fb000U,
    0xc9532800U,
    0x1ded8c00U,
    0xdef31600U,
    0x10ba5900U,
    0x5d1d5080U,
    0xf4b6040U,
    0x72201e20U,
    0x19922510U,
    0x6dcbce98U,
    0xa4630544U,
    0x11f570baU,
    0xc63a9059U,
    0x6d5f960bU,
    0x8a2f1926U,
    0xc557080U,
    0x9c6a9076U,
    0x24379636U,
    0x479b1929U,
    0x950f70b9U,
    0x90459073U,
    0xb5611608U,
    0x66715918U,
    0xd9f9d091U,
    0xcbfa204bU,
    0xc0fa3e09U,
    0x8878d50bU,
    0x7938c6bcU,
    0xc3db794aU,
    0xf66beea4U,
    0x6336f576U,
    0xe518f8b0U,
    0x634cac77U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x28000000U,
    0x74000000U,
    0x6e000000U,
    0x55000000U,
    0x6e800000U,
    0xa9400000U,
    0x4c600000U,
    0xe1d00000U,
    0x8eb80000U,
    0xa2740000U,
    0x4e4e0000U,
    0xb7790000U,
    0x5a548000U,
    0x897a4000U,
    0xe7542000U,
    0x33fab000U,
    0xf0114800U,
    0x29b2400U,
    0xb44d200U,
    0x4b609f00U,
    0xb0560080U,
    0x8a7cd4c0U,
    0x20d53a20U,
    0x3e3b4b30U,
    0x637baa8U,
    0xe0addff4U,
    0xcdeea0aaU,
    0xd78824edU,
    0x9f5e5213U,
    0xb563df11U,
    0xad56a0b9U,
    0xe0fc24c3U,
    0x1f905222U,
    0x7b5adf00U,
    0x9362208bU,
    0xfc5664e3U,
    0x187c7231U,
    0xbfd46f12U,
    0x43bd6895U,
    0xe0f440d3U,
    0x50c2038U,
    0x9c1eb00eU,
    0x9a074819U,
    0xfb062416U,
    0x9b86521fU,
    0x17c7df23U,
    0xcd20a0b2U,
    0xd9b124eeU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x38000000U,
    0xf4000000U,
    0x12000000U,
    0xe9000000U,
    0x82800000U,
    0x8c400000U,
    0x77200000U,
    0xfb700000U,
    0x4c580000U,
    0xbe940000U,
    0xe24a0000U,
    0x799d0000U,
    0x4e768000U,
    0x50dac000U,
    0x1dd22000U,
    0xfcebd000U,
    0x5d2c2800U,
    0x8fcda400U,
    0xfc5d7a00U,
    0xd690c500U,
    0xe4d7d80U,
    0x8f9911c0U,
    0x8d75f220U,
    0xcf5c7110U,
    0x1160f88U,
    0xee8ba0c4U,
    0x53fbddaeU,
    0xb4a101edU,
    0x79b77a2cU,
    0x713dc52eU,
    0xd9c3fda6U,
    0x6e7d1daU,
    0x7195d237U,
    0xfbcea129U,
    0xae5ea7bfU,
    0x1f95c4f2U,
    0x9cc807abU,
    0x3bddd4eeU,
    0xe528fb1U,
    0x262860ebU,
    0xa44d7daaU,
    0xd29911c4U,
    0x3df5f203U,
    0xba1c7104U,
    0xccb60f9aU,
    0x6dbba0d1U,
    0x7a83ddbcU,
    0x184501ebU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x8000000U,
    0x44000000U,
    0xce000000U,
    0x47000000U,
    0x6f800000U,
    0x35400000U,
    0x68200000U,
    0x6d300000U,
    0xdab80000U,
    0x11d40000U,
    0x6dee0000U,
    0xc1ff0000U,
    0x2ef48000U,
    0x75d4000U,
    0xba03a000U,
    0x9106b000U,
    0x4835800U,
    0xc1ac00U,
    0xdc65a200U,
    0x69902500U,
    0x234fcf80U,
    0xd10a06c0U,
    0x16a9da20U,
    0x2ede7930U,
    0x4c449598U,
    0x38a23ff4U,
    0xcbf46fa2U,
    0xd5d8b6e1U,
    0xd1c4820bU,
    0x78e0d515U,
    0xd955b781U,
    0x272f5ae8U,
    0xf498000bU,
    0xbfe4002bU,
    0x76d60006U,
    0x726b002cU,
    0x4cba8018U,
    0xdad24037U,
    0xe86f2014U,
    0xadbff00fU,
    0x6656f818U,
    0x74ac1c1dU,
    0xf3dc7a0bU,
    0xc2c3c92aU,
    0xb1654da4U,
    0x3f15d3fbU,
    0xe688ed9cU,
    0xafec63d0U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0xc8000000U,
    0xd4000000U,
    0x72000000U,
    0xbb000000U,
    0xc800000U,
    0xadc00000U,
    0x96e00000U,
    0xff900000U,
    0x59e80000U,
    0xa340000U,
    0x8e5e0000U,
    0xb2290000U,
    0xc5d58000U,
    0x5e4d4000U,
    0xa6802000U,
    0x9ac45000U,
    0xa4603800U,
    0x7752d400U,
    0x168d6600U,
    0x49e3ab00U,
    0x29176680U,
    0xcb2bce40U,
    0xda50fe20U,
    0x7b0c6f30U,
    0xd7a418a8U,
    0x54b7e17cU,
    0xa89f4692U,
    0x404b9e45U,
    0xb386c614U,
    0x2b43bb22U,
    0x3c22fe99U,
    0x33f00a62U,
    0xabd802bU,
    0xa6b94014U,
    0xa8be2021U,
    0xe5bd503bU,
    0x183db837U,
    0x1e7b940dU,
    0x8fdb461eU,
    0x826afb04U,
    0x574de9fU,
    0x8795a63U,
    0xeede383eU,
    0xb1ebd429U,
    0xee30e623U,
    0x945aeb1dU,
    0x2d29469bU,
    0x73569e50U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0x38000000U,
    0x7c000000U,
    0xae000000U,
    0x81000000U,
    0xf6800000U,
    0x81c00000U,
    0xc2600000U,
    0x48300000U,
    0x90e80000U,
    0x23140000U,
    0x727a0000U,
    0x6b2b0000U,
    0xca768000U,
    0x7fccc000U,
    0xb3806000U,
    0xe5463000U,
    0x8ca12800U,
    0xedd1a400U,
    0x979cb600U,
    0x8c5a4f00U,
    0x753a8b80U,
    0x408b9240U,
    0xe4237e20U,
    0x65151b10U,
    0x177d75b8U,
    0xbfad4944U,
    0x4cb2ebaeU,
    0x8e29a24fU,
    0x5ef05633U,
    0x190bbf34U,
    0x5265438bU,
    0x3034c67bU,
    0xcec8035U,
    0x9d17c03dU,
    0xcb7ee00aU,
    0xe1aef00cU,
    0xe5b3481fU,
    0x3ca89437U,
    0xd311e0dU,
    0xf46c2b25U,
    0x6dd0dda9U,
    0xd79b2d4bU,
    0x6c58bd9dU,
    0x65391d7cU,
    0x788b9587U,
    0x9827b959U,
    0xcb13a398U,
    0x967e3643U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xa8000000U,
    0x44000000U,
    0x2e000000U,
    0x1d000000U,
    0x9800000U,
    0x3fc00000U,
    0x73e00000U,
    0x6ff00000U,
    0x60a80000U,
    0xdc540000U,
    0xf27a0000U,
    0xe36d0000U,
    0xddb18000U,
    0x628e4000U,
    0xd40a000U,
    0xaca1b000U,
    0x8c568800U,
    0xea7c9c00U,
    0x2f692600U,
    0x47b71b00U,
    0xf988bd80U,
    0x5dc30c40U,
    0xb4e68e20U,
    0xdd707710U,
    0xffedb388U,
    0x1ff43b7cU,
    0x88a81d8aU,
    0xb852bc41U,
    0x2c780623U,
    0x5668eb2bU,
    0x903695a0U,
    0x734a2062U,
    0x63a3201dU,
    0xcad6f01dU,
    0xd33da819U,
    0x45ce6c03U,
    0xb2e68e08U,
    0xc470773dU,
    0xf86db3b8U,
    0xcd343b73U,
    0x5ac81d85U,
    0xac62bc7bU,
    0x1130061cU,
    0xf8cceb2dU,
    0xb649585U,
    0x73b32078U,
    0x3f88a023U,
    0x24c5b038U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x88000000U,
    0x84000000U,
    0xbe000000U,
    0x81000000U,
    0x57800000U,
    0x73400000U,
    0x23e00000U,
    0xa4900000U,
    0x48a80000U,
    0xa5b40000U,
    0x819a0000U,
    0x7fef0000U,
    0x18528000U,
    0x540d4000U,
    0x78c7a000U,
    0xc2a73000U,
    0xcc776800U,
    0x9cbfec00U,
    0x44da2200U,
    0x190d8b00U,
    0x3544f280U,
    0x7ee14840U,
    0x21106a20U,
    0x10ec1710U,
    0x66d41898U,
    0x4cb1f64U,
    0x8e635282U,
    0xc8d67871U,
    0x4dcf0217U,
    0x3de7fb14U,
    0x95943aafU,
    0x67299459U,
    0x4af52025U,
    0x107a7038U,
    0x6078c830U,
    0x387cdc31U,
    0x547f4a1fU,
    0xfe79670bU,
    0x497e50b3U,
    0x1bfa8375U,
    0x813bb899U,
    0x50982f4aU,
    0x806e3aa5U,
    0x1f169470U,
    0xd1efa001U,
    0x5153301fU,
    0xe78d6815U,
    0x2580ec32U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0xb8000000U,
    0x4c000000U,
    0x7e000000U,
    0xd1000000U,
    0x7c800000U,
    0x91c00000U,
    0xaba00000U,
    0x7f700000U,
    0x45880000U,
    0x2b940000U,
    0x1f9e0000U,
    0xf490000U,
    0xdfb18000U,
    0x82aec000U,
    0xad20e000U,
    0x75b3b000U,
    0xd1ac2800U,
    0x66a5f400U,
    0x75f23600U,
    0x514f5100U,
    0xbeb0e180U,
    0xe62eb740U,
    0xe77e20U,
    0x1813d530U,
    0x33d81fa8U,
    0x21aaa25cU,
    0x1ea6018eU,
    0x99f00763U,
    0x5f4cd617U,
    0xd7b5e118U,
    0xd6ad4999U,
    0xef25836bU,
    0x62b5a83bU,
    0x302f3408U,
    0xf5e4d639U,
    0x9e91e106U,
    0xf91b499cU,
    0x1588836bU,
    0x23922833U,
    0x4b9cf431U,
    0x4d4bb60aU,
    0xc8b5912cU,
    0x632e0180U,
    0x3e640776U,
    0x9ed2d62fU,
    0x79fce127U,
    0xcd9cc980U,
    0xb04b4346U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x28000000U,
    0x4c000000U,
    0x9e000000U,
    0xf1000000U,
    0xe5800000U,
    0x58c00000U,
    0xb5e00000U,
    0xcf100000U,
    0x71b80000U,
    0xc5b40000U,
    0xd48e0000U,
    0xf67f0000U,
    0xdcd08000U,
    0xa2dc4000U,
    0xe8e7e000U,
    0x24913000U,
    0x907fa800U,
    0x89d73c00U,
    0xf55efe00U,
    0x8f206b00U,
    0xadf3e280U,
    0xae292cc0U,
    0x8b4e3620U,
    0xbb1e2730U,
    0x835498U,
    0xb7444bd4U,
    0xd62202aaU,
    0xc4731ce3U,
    0x8ef1e1fU,
    0x9faa5b18U,
    0xb98acaabU,
    0x65f950d1U,
    0x5911a820U,
    0x3cb83c2cU,
    0xa6367e1aU,
    0x49482b0fU,
    0x5c1a02abU,
    0x68071ccbU,
    0xec011e01U,
    0xae055b26U,
    0x89024ab2U,
    0x818110c7U,
    0x8ac04821U,
    0xdae20c27U,
    0xdb975613U,
    0xccfc571bU,
    0x28939cb3U,
    0xae7a07f8U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0xa8000000U,
    0xa4000000U,
    0xc2000000U,
    0x77000000U,
    0x40800000U,
    0xc0400000U,
    0xc0a00000U,
    0xda700000U,
    0xf1e80000U,
    0x25140000U,
    0x1ffa0000U,
    0x846b0000U,
    0x92d28000U,
    0xbf984000U,
    0xcefaa000U,
    0xd1e9d000U,
    0x1512d800U,
    0xc7f8b400U,
    0x18696a00U,
    0x5cd55900U,
    0xae99a580U,
    0x3b788040U,
    0x26299220U,
    0x55777d10U,
    0x1d6a37b8U,
    0xf354fd64U,
    0x5259258aU,
    0x659fc079U,
    0x85fbb228U,
    0x2f6ded29U,
    0xfc50cf82U,
    0xfeddd97dU,
    0x53d8378aU,
    0x585bfd52U,
    0x6699a584U,
    0x4f788050U,
    0xac299215U,
    0x16777d3aU,
    0x37ea378aU,
    0xe014fd40U,
    0x107925a2U,
    0x8afc048U,
    0xf433b22eU,
    0x1049ed10U,
    0xd2e2cf9fU,
    0x85d2d94eU,
    0x2f18b7acU,
    0x46bcbd71U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x48000000U,
    0x74000000U,
    0x46000000U,
    0xff000000U,
    0x46800000U,
    0xea400000U,
    0xdc600000U,
    0x2900000U,
    0x2b880000U,
    0x99340000U,
    0xefba0000U,
    0xd10b0000U,
    0x1738000U,
    0x6adf4000U,
    0xda1d2000U,
    0xce3bf000U,
    0xb1cf1800U,
    0xad572400U,
    0xc9285a00U,
    0x44844f00U,
    0x5b412280U,
    0xf1e3c540U,
    0x5853e220U,
    0xaaacdb10U,
    0x7540c0b8U,
    0x6ae01e6cU,
    0x40d2a2b2U,
    0x33ec854dU,
    0x95a6c209U,
    0x4f332b03U,
    0x48bdd8bbU,
    0xfb883a4bU,
    0x21337894U,
    0x33bc8a72U,
    0x53094092U,
    0xf0745e5fU,
    0xa75c0297U,
    0x30d8356fU,
    0x71cfa26U,
    0x29bbff22U,
    0x6e089aadU,
    0xa7f45155U,
    0x309b8012U,
    0x4e7b400dU,
    0xb8af2001U,
    0xdc44f037U,
    0xcb669809U,
    0x60136411U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0x68000000U,
    0xcc000000U,
    0x7a000000U,
    0x9d000000U,
    0x58800000U,
    0xd3c00000U,
    0xe4a00000U,
    0xdd300000U,
    0x8e80000U,
    0x53d40000U,
    0x983e0000U,
    0x786b0000U,
    0x6c128000U,
    0xf69a4000U,
    0x405de000U,
    0x987a1000U,
    0xc08cc800U,
    0xdbc2f400U,
    0xf8a36600U,
    0x2f33a900U,
    0x49ecc380U,
    0x1957e440U,
    0x1afece20U,
    0xbe4a0d30U,
    0xffe40db8U,
    0x4252e944U,
    0x1d7e43baU,
    0xec0da443U,
    0x36032e06U,
    0x27001d13U,
    0x2580c5b4U,
    0x9b441d43U,
    0x5f63259dU,
    0xf5950d53U,
    0xafdd6d92U,
    0xc63db973U,
    0x936beb8bU,
    0x3390005aU,
    0xf0dde03cU,
    0x47ba103eU,
    0xbe2cc81eU,
    0x8bf2f408U,
    0xc0cb6626U,
    0x6327a902U,
    0x4f72c386U,
    0x210ce460U,
    0x26844e19U,
    0xc8c44d2fU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xa8000000U,
    0x34000000U,
    0x12000000U,
    0x37000000U,
    0xc6800000U,
    0x4c00000U,
    0xbae00000U,
    0x4e700000U,
    0xe2180000U,
    0x1dd40000U,
    0xb2ca0000U,
    0xa9d0000U,
    0xb8118000U,
    0x63ae4000U,
    0x75ab2000U,
    0x98ac5000U,
    0x8d2a5800U,
    0x56f8400U,
    0x104cce00U,
    0x45e3b00U,
    0x89f09180U,
    0x3d5ba9c0U,
    0x1e763620U,
    0xfa1eaf10U,
    0xa1d4a788U,
    0x64cc06fcU,
    0x8799118aU,
    0x7d91e9ddU,
    0xb3ef162cU,
    0xfc8bff21U,
    0xaabd7f93U,
    0x2584c2f0U,
    0x40457fa0U,
    0xf120c2c3U,
    0x5e177fa5U,
    0x26a9c2e7U,
    0x542cffb1U,
    0x2ea82f1U,
    0xaf0e5fbeU,
    0xd7fc92e0U,
    0x992527beU,
    0xa1246ceU,
    0xe4aa31b8U,
    0x3b29b9e5U,
    0x586f4e16U,
    0x8dc97b06U,
    0x481831b1U,
    0x26d0b9c3U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xc8000000U,
    0x94000000U,
    0x6a000000U,
    0xed000000U,
    0x9f800000U,
    0x31400000U,
    0x21a00000U,
    0x4b500000U,
    0x2e680000U,
    0x13340000U,
    0x1f5a0000U,
    0x356b0000U,
    0x3fb58000U,
    0xa69b4000U,
    0xef0c2000U,
    0xa7809000U,
    0x2d408800U,
    0x3fa78c00U,
    0xbc507200U,
    0x94e87900U,
    0x29f4dc80U,
    0x65bfcc40U,
    0xb29b5a20U,
    0xc50b2510U,
    0xaa8786b8U,
    0xc2c0e95cU,
    0xc6e6dc92U,
    0x9f0cc75U,
    0xb5bcda02U,
    0xea9f6537U,
    0xe90c26adU,
    0x9c843975U,
    0xd1c5f480U,
    0xde67905cU,
    0x4a35802eU,
    0x3adb403fU,
    0xb12c2022U,
    0xad90902aU,
    0xea888824U,
    0xf3c38c21U,
    0xe762720aU,
    0x5fb77917U,
    0x969b5cacU,
    0xc70f8c75U,
    0x4382fa38U,
    0x8f40f523U,
    0x46a32e8eU,
    0x49d3f56aU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0xb8000000U,
    0x44000000U,
    0x46000000U,
    0xf7000000U,
    0xa1800000U,
    0xc9c00000U,
    0x1d200000U,
    0xe6100000U,
    0x89b80000U,
    0x4a740000U,
    0xa1ca0000U,
    0x4ab90000U,
    0x55f18000U,
    0x7b0ac000U,
    0x401ca000U,
    0x35267000U,
    0xa175800U,
    0x63bf9c00U,
    0xf7758600U,
    0x4d4fbd00U,
    0xd7fd6280U,
    0x9196e2c0U,
    0x1f7dfe20U,
    0x5ad19110U,
    0xc4189ca8U,
    0x932373fcU,
    0xad17628eU,
    0x8a3fe2c1U,
    0xc2b47e19U,
    0x526f5111U,
    0x80ee3caeU,
    0x4eac03dfU,
    0x3d49ba98U,
    0x2ffebec5U,
    0x75975813U,
    0xa97f9c31U,
    0x15d58601U,
    0x219fbd34U,
    0x1ce5628fU,
    0x4732e2d9U,
    0xcdaffe03U,
    0x82cc9112U,
    0x53b1cb8U,
    0xc734b3e3U,
    0x8da84293U,
    0x22ce52eaU,
    0xf53c0634U,
    0x7f317d20U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x48000000U,
    0xec000000U,
    0xaa000000U,
    0x69000000U,
    0xf6800000U,
    0xbec00000U,
    0xcce00000U,
    0x18700000U,
    0x2d580000U,
    0x13d40000U,
    0x228a0000U,
    0x935f0000U,
    0x54d58000U,
    0xeb0b4000U,
    0x201be000U,
    0x1d729000U,
    0x31df6800U,
    0xa4110400U,
    0x68e82e00U,
    0x7de8bb00U,
    0x596d3180U,
    0x4aaba3c0U,
    0x46cb2620U,
    0x6b7b6f10U,
    0xf9461798U,
    0x10a0ccdcU,
    0x45531b2U,
    0xa6cfa3ebU,
    0x5b792612U,
    0xd1406f16U,
    0x8ca197b7U,
    0xe6508cc8U,
    0x23c951abU,
    0x7fd73f7U,
    0x6822e24U,
    0xb6c7bb3dU,
    0x40e0b18cU,
    0xc274e3eaU,
    0xc5ac635U,
    0x956ff09U,
    0x364cff84U,
    0x36ba88ceU,
    0xba26ffadU,
    0x789588f5U,
    0xff2b7fadU,
    0x278ac8f6U,
    0x8fda9f86U,
    0xe31758c4U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x90000000U,
    0x78000000U,
    0x84000000U,
    0xa000000U,
    0xfd000000U,
    0x40800000U,
    0x9c400000U,
    0xbee00000U,
    0xae900000U,
    0xbd480000U,
    0xcfb40000U,
    0xaefe0000U,
    0xdf4f0000U,
    0xeb68000U,
    0x7ec000U,
    0x240fa000U,
    0x75d51000U,
    0xf62d0800U,
    0xb1e5f400U,
    0x57121a00U,
    0xb8cfb00U,
    0x80176280U,
    0x860e8740U,
    0xd4d03220U,
    0xc8a9df30U,
    0x32a75098U,
    0xa8775864U,
    0xf7df62beU,
    0xe8fa8751U,
    0xe44e323aU,
    0xb536df2bU,
    0x47b9d096U,
    0x4d2d9862U,
    0xca66c293U,
    0x70d49745U,
    0xf2abba05U,
    0x47a2eb3fU,
    0x84f2ea92U,
    0x9d9ab37eU,
    0x251b083fU,
    0xfd5ef436U,
    0x29ba9a07U,
    0xea2d3b25U,
    0x6fe642b4U,
    0xb8115776U,
    0xa20c9a3dU,
    0x2ed63b08U,
    0xddaec298U,
    0x8e209757U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x38000000U,
    0xbc000000U,
    0x2a000000U,
    0xa1000000U,
    0x75800000U,
    0x14c00000U,
    0x31a00000U,
    0xdeb00000U,
    0xba080000U,
    0x21540000U,
    0xd35e0000U,
    0x860d0000U,
    0xcb528000U,
    0x525ec000U,
    0xe3886000U,
    0xe7921000U,
    0xdffe2800U,
    0x173a9c00U,
    0xfc9fbe00U,
    0x8b280900U,
    0xd0a2cf80U,
    0x4b36e340U,
    0x9ecd7620U,
    0x38f74530U,
    0x89efb988U,
    0xaa01a644U,
    0x6102cfaeU,
    0x5586e35fU,
    0x4c57622U,
    0x9a3451cU,
    0x62b1b99bU,
    0x900ca66eU,
    0x80504fa8U,
    0xa6d82340U,
    0x92cd1611U,
    0xfaf15511U,
    0x8cef91a3U,
    0x59863a78U,
    0xc6c7f1b4U,
    0xca42a5cU,
    0x9131d99dU,
    0x37cab649U,
    0xd9706786U,
    0x332fbf59U,
    0xaca0283bU,
    0x41379c03U,
    0x2fcd3e30U,
    0x7576c901U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x18000000U,
    0xa4000000U,
    0x32000000U,
    0x75000000U,
    0xbb800000U,
    0xc2c00000U,
    0xa2200000U,
    0xe8300000U,
    0xfda80000U,
    0xd8540000U,
    0xa5fe0000U,
    0xbdab0000U,
    0xf8578000U,
    0xf5f9c000U,
    0x15aba000U,
    0x4456d000U,
    0x63ff4800U,
    0x52afec00U,
    0x8ad36e00U,
    0x1abf8500U,
    0x324ebb80U,
    0xc0c19240U,
    0xf260620U,
    0x17b07930U,
    0xbd68bdb8U,
    0x1771eb6cU,
    0x524ebba6U,
    0xb0c19259U,
    0xf7260614U,
    0x3b07901U,
    0x9768bd90U,
    0xc671eb75U,
    0xdbcebb82U,
    0x701927eU,
    0xee860605U,
    0x29407907U,
    0xc8e0bd91U,
    0xf615eb60U,
    0x8398bb83U,
    0x62fe9275U,
    0xb32f8629U,
    0x6112b939U,
    0x251c9db7U,
    0x47bafb49U,
    0xf5cc53a4U,
    0x7407ae62U,
    0x5a03a03dU,
    0x2902d01dU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0xf8000000U,
    0x4c000000U,
    0xce000000U,
    0xad000000U,
    0xed800000U,
    0x2f400000U,
    0xfe600000U,
    0xb6900000U,
    0x22980000U,
    0x829c0000U,
    0xd29e0000U,
    0x3a9d0000U,
    0xfe9e8000U,
    0x84994000U,
    0xab9b2000U,
    0x251d5000U,
    0x4ad83800U,
    0x767f5400U,
    0x11ce2e00U,
    0x7ba1dd00U,
    0x6d306980U,
    0x1fab7d40U,
    0x753069a0U,
    0x63ab7d50U,
    0x233069b8U,
    0xf2ab7d4cU,
    0xf8b06986U,
    0x3ceb7d5fU,
    0x255069b5U,
    0x83b7d74U,
    0x1428698eU,
    0x13777d7fU,
    0x1a4e69b1U,
    0x1de67d52U,
    0x14d6e9b9U,
    0x217e3d72U,
    0x934d498dU,
    0xba672d7cU,
    0x8c90d1b2U,
    0x6d9c695dU,
    0x7c1de798U,
    0x455fb054U,
    0x4a3b980aU,
    0x572a4409U,
    0x43f5b62eU,
    0x508b9907U,
    0x78c5dfadU,
    0xde20e45cU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0xc8000000U,
    0x1c000000U,
    0xba000000U,
    0xff000000U,
    0x99800000U,
    0x1c400000U,
    0xeca00000U,
    0x8b900000U,
    0xd8380000U,
    0x703c0000U,
    0x1c3e0000U,
    0x8e390000U,
    0x5d388000U,
    0xb6bec000U,
    0x49fe2000U,
    0x86dbf000U,
    0xee0c0800U,
    0x4310c400U,
    0xa4fc2600U,
    0x8c5d4300U,
    0xf9cc1f80U,
    0x60f45e40U,
    0x284c1fa0U,
    0x20b45e50U,
    0x9eec1f98U,
    0x64245e5cU,
    0x17541faaU,
    0x14585e5bU,
    0x5dca1f84U,
    0xeef15e64U,
    0x414a9fa2U,
    0x34339e63U,
    0xf82abf99U,
    0xb7416e41U,
    0x932637afU,
    0x32d36a5dU,
    0x621a31a8U,
    0xb66cd97eU,
    0x28e2a63fU,
    0x23368313U,
    0xedacbf8dU,
    0x9046e62U,
    0x6480b7aeU,
    0x4ec4aa6eU,
    0xff64918dU,
    0xf675e946U,
    0xc58e8e2fU,
    0x1c54b726U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0x48000000U,
    0xe4000000U,
    0x9e000000U,
    0x8f000000U,
    0x52800000U,
    0x1400000U,
    0xada00000U,
    0xc100000U,
    0x15a80000U,
    0x2dac0000U,
    0x41aa0000U,
    0x9baf0000U,
    0x7aa88000U,
    0xef2ac000U,
    0x58ede000U,
    0x6a091000U,
    0x44bd6800U,
    0xf87fc00U,
    0x36c3f600U,
    0xf7609b00U,
    0x2175fb80U,
    0xd5dad6c0U,
    0x6df5fba0U,
    0x9b9ad6f0U,
    0x32d5fb88U,
    0x66cad6ccU,
    0xc2ddfb9aU,
    0xa376d6f5U,
    0x8dffbbdU,
    0x9a75d6d6U,
    0x615d7ba9U,
    0xefb016d6U,
    0xeeb81b82U,
    0x6683c6d5U,
    0xe7409387U,
    0x2ea12adeU,
    0xd4940d97U,
    0x4de94df8U,
    0xb98a8009U,
    0x8079c003U,
    0xade7603fU,
    0xa430d029U,
    0xb67a0806U,
    0x56e72c16U,
    0x70b1fe19U,
    0x643bb712U,
    0x9bc605b6U,
    0x88e261c4U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0xe8000000U,
    0xb4000000U,
    0xb6000000U,
    0x31000000U,
    0x5e800000U,
    0x73400000U,
    0x36e00000U,
    0x50300000U,
    0xf4d80000U,
    0xb0dc0000U,
    0x9eda0000U,
    0x43df0000U,
    0xe75f8000U,
    0x4f18c000U,
    0x147d6000U,
    0xee88d000U,
    0x30715800U,
    0xcbbfac00U,
    0xb42e3600U,
    0xd5248100U,
    0xb8978080U,
    0x6e89e640U,
    0xf07780a0U,
    0x6bb9e670U,
    0xa42f8088U,
    0x3d25e674U,
    0xc9580b2U,
    0xd88ae659U,
    0xc172009fU,
    0x353e2655U,
    0xd76d6088U,
    0xbc5f649U,
    0x5ca1b885U,
    0x2c519a5dU,
    0x71af6eb8U,
    0xabe60b71U,
    0x94b1561fU,
    0xec9f5121U,
    0x13bb5886U,
    0x382d8a42U,
    0x9f21568bU,
    0x9b92774dU,
    0x5f0bb815U,
    0x58b3bc16U,
    0xc6980e0fU,
    0x80bcfd32U,
    0xf1af6eb3U,
    0x6be60b7dU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0x58000000U,
    0xf4000000U,
    0x52000000U,
    0xe7000000U,
    0xba800000U,
    0xadc00000U,
    0x31600000U,
    0x39300000U,
    0xa8880000U,
    0xaa8c0000U,
    0x558e0000U,
    0x7b0d0000U,
    0x54cb8000U,
    0xdaa94000U,
    0xad1ee000U,
    0xfa547000U,
    0x86bf4800U,
    0x50849c00U,
    0x2ec4f200U,
    0xe1e2f900U,
    0x57f47080U,
    0x296ad4c0U,
    0x2d7c70a0U,
    0x24e6d4d0U,
    0xa27270b8U,
    0x222bd4e4U,
    0x9fd9f0aeU,
    0x35b294d9U,
    0xc84f10baU,
    0x826ae4e0U,
    0xa1fe5894U,
    0x42378cbU,
    0xea912a98U,
    0x658c1c5U,
    0xb8f3ba32U,
    0x7fea652fU,
    0x46be8287U,
    0x70852de1U,
    0x9ec3802aU,
    0x69e54033U,
    0xfbf0e02fU,
    0x8f697014U,
    0x987cc814U,
    0x7961dc30U,
    0xb5341233U,
    0xbe8b8927U,
    0x9788b890U,
    0xa40b08f1U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0x28000000U,
    0xcc000000U,
    0x86000000U,
    0x35000000U,
    0xf1800000U,
    0x17c00000U,
    0xf2e00000U,
    0xf700000U,
    0xb0a80000U,
    0x26ac0000U,
    0xabaa0000U,
    0xee2d0000U,
    0x2be88000U,
    0x8e09c000U,
    0xfffa000U,
    0xea139000U,
    0xed1e3800U,
    0x5d63b400U,
    0xe9b60a00U,
    0xa9cd1d00U,
    0x151aa680U,
    0xc96383c0U,
    0x6bb2a6a0U,
    0x66cf83f0U,
    0x7f98a698U,
    0xd62283dcU,
    0xf9102692U,
    0xb99b43efU,
    0xc32786b3U,
    0x5894d3e2U,
    0x365bbe8fU,
    0xd5c667e7U,
    0x1de734b3U,
    0x35f3bae4U,
    0x37e8b21fU,
    0x500b692dU,
    0x3ef90c95U,
    0xe1900ee6U,
    0x61deb807U,
    0xa7067419U,
    0x8683aa16U,
    0xc9438d0cU,
    0x3fa41ebeU,
    0xcad5f7c9U,
    0xe1790cb3U,
    0xcf500ee8U,
    0x4beb82fU,
    0x3ab6743dU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x28000000U,
    0xe4000000U,
    0x5a000000U,
    0xeb000000U,
    0x44800000U,
    0xc6400000U,
    0x85a00000U,
    0x67d00000U,
    0x30a80000U,
    0x2eac0000U,
    0xefaa0000U,
    0xd82d0000U,
    0xc6ef8000U,
    0x130ec000U,
    0xa479e000U,
    0x6833000U,
    0x71406800U,
    0xb7244c00U,
    0xa4957e00U,
    0x688bd300U,
    0xc4bdad80U,
    0x826508c0U,
    0x39b5ada0U,
    0xbf1908f0U,
    0x74b7ad98U,
    0x169808e4U,
    0x6b722d92U,
    0x14fbc8ddU,
    0x92c44d84U,
    0x4be638e7U,
    0xb675c595U,
    0x757d44d6U,
    0x3902d3b4U,
    0x3b83dbcfU,
    0x40c78038U,
    0x34e2c004U,
    0x30f3e023U,
    0xc43e3002U,
    0x6e27e82aU,
    0xcf168c12U,
    0x4e9e09U,
    0x1459e32aU,
    0xe8904583U,
    0x168e84fdU,
    0x95bcb39bU,
    0x7de22be6U,
    0x93740833U,
    0x18f8bc0bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xc8000000U,
    0xdc000000U,
    0xce000000U,
    0xb1000000U,
    0x28800000U,
    0xd7400000U,
    0x8fe00000U,
    0x8f00000U,
    0xef380000U,
    0x853c0000U,
    0x2e3a0000U,
    0x5bf0000U,
    0xb57e8000U,
    0xbc58c000U,
    0x188e2000U,
    0x69e05000U,
    0x35f1d800U,
    0x51be9c00U,
    0xc7792600U,
    0x535ad900U,
    0x7909de80U,
    0x8221c740U,
    0x7f51dea0U,
    0xa9adc770U,
    0xb933de88U,
    0xc69ec77cU,
    0x8aaf5ebaU,
    0x2eb5074bU,
    0x285dfe89U,
    0xca8e9767U,
    0xb6e68683U,
    0x9c779b52U,
    0x667ed8a0U,
    0x43db4e50U,
    0x1a49d81bU,
    0x2fc29c31U,
    0xc0a32630U,
    0x315d900U,
    0xdcf5eb5U,
    0x1050763U,
    0x2085fea1U,
    0xab42977eU,
    0x71e486b7U,
    0x71f49b4aU,
    0x1bba58beU,
    0x9c7c8e5cU,
    0x10d97830U,
    0x25ca0c12U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0xa8000000U,
    0xa4000000U,
    0x86000000U,
    0x83000000U,
    0x7e800000U,
    0x6bc00000U,
    0x9ea00000U,
    0xc6d00000U,
    0x6b280000U,
    0x512c0000U,
    0x682e0000U,
    0x6fad0000U,
    0x63ed8000U,
    0xe108c000U,
    0xd3de000U,
    0x2e21d000U,
    0xda11c800U,
    0xc5083c00U,
    0xcb3c6200U,
    0xcd221700U,
    0xb4966180U,
    0x6ca98c0U,
    0xf19e61a0U,
    0x8df698d0U,
    0x5cb861b8U,
    0x296798d4U,
    0xf273e192U,
    0x7cfe58fdU,
    0xf98581b3U,
    0xa34648ddU,
    0xae6229acU,
    0x3af664c7U,
    0x4c39e38bU,
    0x5a45ff6U,
    0x84544811U,
    0xfaecfc22U,
    0xd68f822bU,
    0xd97ec729U,
    0xb0c229abU,
    0xbc2664d5U,
    0x4711e3a2U,
    0x44885fe7U,
    0x447a482dU,
    0x3141fc32U,
    0x3362020aU,
    0xbb760722U,
    0xc37fc9b9U,
    0xf9c7b4f6U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0xa8000000U,
    0x1c000000U,
    0x9a000000U,
    0xdd000000U,
    0x80800000U,
    0xa5c00000U,
    0x8e00000U,
    0xb8500000U,
    0x4d880000U,
    0xb88c0000U,
    0xe40a0000U,
    0xfbcb0000U,
    0x3e2b8000U,
    0xaefec000U,
    0x5ab62000U,
    0x70def000U,
    0xf1821800U,
    0xc7451c00U,
    0xe427b200U,
    0xa6333500U,
    0xbf1bd780U,
    0xbc64cbc0U,
    0x5b91d7a0U,
    0x226fcbf0U,
    0x4d5a57a8U,
    0x24c10bf4U,
    0xf2647782U,
    0xf093fbf3U,
    0x7dec6fa4U,
    0x111de7c4U,
    0x27605d84U,
    0xbc1012edU,
    0xca2daa06U,
    0xc8fa2903U,
    0x2db665b5U,
    0x955cfeddU,
    0x60c18024U,
    0xec65c017U,
    0x1395a029U,
    0xe6c300fU,
    0x6f5e383bU,
    0x4dc0ec27U,
    0xf4e62a2bU,
    0x1254e90fU,
    0x28884596U,
    0x8c0e0ed6U,
    0xc7c99828U,
    0xb42bdc20U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0x18000000U,
    0x5c000000U,
    0x5a000000U,
    0x6b000000U,
    0x95800000U,
    0xb8c00000U,
    0x7ea00000U,
    0x300000U,
    0x62e80000U,
    0xd5ec0000U,
    0xda6a0000U,
    0xa9af0000U,
    0x728e8000U,
    0xd27e4000U,
    0x9234a000U,
    0x1dea7000U,
    0xce6b2800U,
    0x27ac8400U,
    0x378dc200U,
    0x32ff7700U,
    0xd2755680U,
    0x7f0a71c0U,
    0xf6bf56a0U,
    0x169571f0U,
    0x46d9d688U,
    0x210731fcU,
    0x1687768eU,
    0xc94241ebU,
    0xf062de98U,
    0x579085f1U,
    0xc05bbcbdU,
    0xc04582dfU,
    0xa2e5c222U,
    0x98d3771fU,
    0x13f569aU,
    0x695571daU,
    0x4ff9d6a4U,
    0x9ef731ccU,
    0xdd4f7696U,
    0x935e41e3U,
    0xf9c0dea2U,
    0xf82385e9U,
    0xe1773ca2U,
    0xd688c2d4U,
    0x847de215U,
    0x2b344736U,
    0x976c5e86U,
    0xfd2ec5ecU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x18000000U,
    0x94000000U,
    0xda000000U,
    0x25000000U,
    0x79800000U,
    0xf5c00000U,
    0x98a00000U,
    0x69b00000U,
    0x78e80000U,
    0xf1ec0000U,
    0x966e0000U,
    0x24ab0000U,
    0x648b8000U,
    0x437b4000U,
    0xe7f7e000U,
    0x448d3000U,
    0x137a1800U,
    0x5ff6d400U,
    0xc088fa00U,
    0xd17a7f00U,
    0xeef25280U,
    0x630a65c0U,
    0x1bc52a0U,
    0xfd165d0U,
    0xff7fd288U,
    0xe1f625d4U,
    0x978e328eU,
    0x11fc15f1U,
    0x34b1aab8U,
    0x656a81f8U,
    0x762d30a6U,
    0xb7ca8ec5U,
    0x629c9a00U,
    0x82a00f1fU,
    0xacb1aabeU,
    0xb16a81f3U,
    0xc2d3093U,
    0x82ca8eeeU,
    0x31c9a24U,
    0xe3600f1fU,
    0xee11aa8dU,
    0xfdda81f2U,
    0xd4530a9U,
    0x86e68ec5U,
    0xdd29a0fU,
    0xae7b0f37U,
    0xf2722a9cU,
    0x4f4dc1f5U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0xf8000000U,
    0x1c000000U,
    0xa000000U,
    0xc1000000U,
    0xc6800000U,
    0xd8c00000U,
    0xbae00000U,
    0xd6b00000U,
    0xe8880000U,
    0xfd8c0000U,
    0x650e0000U,
    0x62cb0000U,
    0x61a98000U,
    0x865e4000U,
    0x1d76a000U,
    0x59eaf000U,
    0x677af800U,
    0x62a3cc00U,
    0xc7968a00U,
    0x3f5c1d00U,
    0x87f48880U,
    0x2b2ffcc0U,
    0x4c9a88a0U,
    0x8a94fcd0U,
    0xebdb0888U,
    0x7236bcc4U,
    0xf54ba8b6U,
    0x67eb4cc3U,
    0x87ed0b4U,
    0xd321c0f3U,
    0x42d17a85U,
    0xa2fe6dc5U,
    0xe9e62a2bU,
    0x7d31ed30U,
    0xf4c9f0b1U,
    0xeaa970dfU,
    0x21db22a0U,
    0x533051c3U,
    0xc3cdd82eU,
    0x172b7c31U,
    0x569cd235U,
    0x13922100U,
    0x615f7ab7U,
    0x58f56dcaU,
    0x92afaa07U,
    0x7ddfad11U,
    0xf93750b8U,
    0x52cf80e5U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x88000000U,
    0xec000000U,
    0xc2000000U,
    0xe9000000U,
    0xc2800000U,
    0x95400000U,
    0x7ea00000U,
    0x65f00000U,
    0xf180000U,
    0xa41c0000U,
    0xcf9e0000U,
    0xb85f0000U,
    0x43bf8000U,
    0xd0ebc000U,
    0x5604a000U,
    0x3f06b000U,
    0xbd83f800U,
    0x8c4ec00U,
    0x6660c600U,
    0x8b977300U,
    0x688b3c80U,
    0xe915440U,
    0x280d3ca0U,
    0x52d25450U,
    0x842cbca8U,
    0x2a669454U,
    0x19979c8aU,
    0x298be46fU,
    0x3010c4baU,
    0xf749b855U,
    0x2973fa8aU,
    0xe15a2770U,
    0x59380015U,
    0x1ac0029U,
    0x16a60036U,
    0xf9f30010U,
    0x7519800dU,
    0x3918c03aU,
    0xab1d203fU,
    0xea1e7035U,
    0xd49ed81aU,
    0xbda9c00U,
    0x707e1e05U,
    0x150def27U,
    0x66552283U,
    0x7e6cbb5cU,
    0x41401e3aU,
    0x88a2ef26U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x38000000U,
    0xa4000000U,
    0xd6000000U,
    0x35000000U,
    0x37800000U,
    0x1f400000U,
    0xbd200000U,
    0x78100000U,
    0xeef80000U,
    0xf5fc0000U,
    0x337e0000U,
    0x5dbf0000U,
    0x42da8000U,
    0x2c6b4000U,
    0xd7c66000U,
    0x92e4d000U,
    0x2ef77800U,
    0x860fc400U,
    0xcef6be00U,
    0x560c9700U,
    0x66f57780U,
    0x7a0e3040U,
    0x2cf377a0U,
    0x3d0d3050U,
    0xf877f7a8U,
    0x20c9707cU,
    0x6d9317a6U,
    0xfabae055U,
    0x465c0f93U,
    0x99aef458U,
    0x6e27499eU,
    0x296e75fU,
    0x423ae031U,
    0x319c9001U,
    0xfb4d982aU,
    0x58d3542cU,
    0x909b262eU,
    0x62cfc317U,
    0xf69651b0U,
    0x7c3df375U,
    0x89b2610U,
    0x36cfc30cU,
    0x389651beU,
    0x5d3df356U,
    0xd11b2604U,
    0xb88fc308U,
    0x643651b5U,
    0xf6df373U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0xf8000000U,
    0x74000000U,
    0xea000000U,
    0x3000000U,
    0xb6800000U,
    0xf2c00000U,
    0x8e600000U,
    0x84b00000U,
    0xb8c80000U,
    0x19cc0000U,
    0xe04a0000U,
    0x5a090000U,
    0xffa88000U,
    0xb8ff4000U,
    0xb4c4e000U,
    0x8367b000U,
    0x7331a800U,
    0x438f1c00U,
    0xb5ed0a00U,
    0x8558b900U,
    0xb7943c80U,
    0x9d5b2ac0U,
    0x33963ca0U,
    0xef5e2af0U,
    0x7494bcb8U,
    0x4bd86afcU,
    0xf150dc86U,
    0x993c9ae1U,
    0x842794bcU,
    0x191136e1U,
    0xeb99b691U,
    0xcff0d3ddU,
    0x2cece032U,
    0x68dbb03cU,
    0xd7d3a83cU,
    0xc3fa1c0aU,
    0xb6478a1cU,
    0xfba2f910U,
    0xce525c9dU,
    0x89badae3U,
    0x8be3f4aaU,
    0xf8f5c6f3U,
    0xd06efe83U,
    0xc91d7fd3U,
    0x2732c233U,
    0x398a550eU,
    0x1eeefeaaU,
    0x8fdd7fdfU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0xb8000000U,
    0xc000000U,
    0x2000000U,
    0xa7000000U,
    0xed800000U,
    0x9fc00000U,
    0xe2600000U,
    0xf0100000U,
    0x60480000U,
    0x694c0000U,
    0xb1ce0000U,
    0xfc8f0000U,
    0xb0a98000U,
    0x871c4000U,
    0x5e242000U,
    0x48b03000U,
    0xc8bf0800U,
    0xe3556400U,
    0x71691a00U,
    0x387cc900U,
    0xc8b6a980U,
    0x88bbfbc0U,
    0xc350a9a0U,
    0xa168fbd0U,
    0x807f29a8U,
    0xc4b7bbe4U,
    0x8abc8986U,
    0x6454cbe7U,
    0x4cee2186U,
    0x1fbddfceU,
    0x26d413bdU,
    0x7aa842e9U,
    0x41aa805U,
    0x25a51415U,
    0xae74323dU,
    0xda5a9d1fU,
    0xca073b99U,
    0x830116d4U,
    0x7b82ba3bU,
    0xe6c3b926U,
    0x12e20184U,
    0x2951eff9U,
    0xf26d1b8bU,
    0x43fe26c1U,
    0x2e743209U,
    0x9a5a9d2dU,
    0xea073b96U,
    0x530116c8U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0x28000000U,
    0xbc000000U,
    0x82000000U,
    0x7b000000U,
    0x89800000U,
    0xc3c00000U,
    0xbe00000U,
    0x7700000U,
    0xed380000U,
    0xf83c0000U,
    0x70be0000U,
    0x45fb0000U,
    0x225a8000U,
    0x800bc000U,
    0xad262000U,
    0xc163000U,
    0x530fe800U,
    0xd8a2a400U,
    0xedd33a00U,
    0x13ed2500U,
    0x7e507e80U,
    0x7f2dc740U,
    0x62367ea0U,
    0x729ac750U,
    0x5e6afe88U,
    0x7b96075cU,
    0x89c85e82U,
    0xb1c0f773U,
    0xb8e316a2U,
    0xa2f3a36dU,
    0xb4ffe480U,
    0x1cdd125dU,
    0xc04fc802U,
    0x9203941cU,
    0x23005219U,
    0xbd834132U,
    0xcdc1e4adU,
    0xdae6126fU,
    0x49f5480fU,
    0x25785430U,
    0x4b1e720cU,
    0xf5a97131U,
    0x6c700c98U,
    0x3cbfb654U,
    0xffcf22bU,
    0x755eb130U,
    0x93882c85U,
    0x81e28663U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0x68000000U,
    0x84000000U,
    0xba000000U,
    0xed000000U,
    0xe9800000U,
    0x6bc00000U,
    0x58200000U,
    0xfad00000U,
    0x7e180000U,
    0x379c0000U,
    0x6c5a0000U,
    0x7c7b0000U,
    0xf2ac8000U,
    0x5eb7c000U,
    0x2fe000U,
    0x3ff03000U,
    0xc54c5800U,
    0x86479400U,
    0x49e30a00U,
    0x6ff62b00U,
    0x9d4dcd80U,
    0x4a406540U,
    0x87e14da0U,
    0x50f7a570U,
    0x1dceada8U,
    0x7207954cU,
    0x5902f5b2U,
    0x1b80016dU,
    0xf2c1ff9cU,
    0x63a62a56U,
    0xf8143226U,
    0x75ba4f0cU,
    0x4b8f7f90U,
    0xa3e6ea42U,
    0xdaf55227U,
    0xb8cabf03U,
    0xef82478eU,
    0xe0c18e71U,
    0x2aa3602eU,
    0x5b97f030U,
    0x9b7bb808U,
    0x7e2ba433U,
    0x88f55229U,
    0x11cabf10U,
    0x9c0247bcU,
    0x96018e78U,
    0xf303603dU,
    0x4e87f017U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x98000000U,
    0x1c000000U,
    0xca000000U,
    0xe7000000U,
    0x4d800000U,
    0xb8400000U,
    0xaca00000U,
    0xd0f00000U,
    0x40180000U,
    0xfd9c0000U,
    0xadda0000U,
    0x357f0000U,
    0xbb8f8000U,
    0x52934000U,
    0xd38fa000U,
    0xa693f000U,
    0x2d8bc800U,
    0x1f962400U,
    0xc908a600U,
    0xdb566300U,
    0xbd6d1480U,
    0xb5c5dc40U,
    0x74e294a0U,
    0xcc569c70U,
    0x18ed3488U,
    0x39856c74U,
    0x8646fcaeU,
    0xb5a34873U,
    0x24765a9cU,
    0x1cd92b4aU,
    0x95f94e0fU,
    0x94ff724U,
    0x1bf65a84U,
    0x8f992b60U,
    0x86d94e1cU,
    0x8afff72bU,
    0xe8ce5a99U,
    0xe1b52b55U,
    0x403b4e2bU,
    0xcdacf70bU,
    0x5fa3da90U,
    0x13756b46U,
    0x69596e3fU,
    0x21bf4711U,
    0xf7ea32b9U,
    0x3000ff48U,
    0x48002009U,
    0x2400b007U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x78000000U,
    0xc4000000U,
    0x5a000000U,
    0xa3000000U,
    0xd8800000U,
    0xbf400000U,
    0xfde00000U,
    0xc900000U,
    0x50e80000U,
    0x146c0000U,
    0xe52e0000U,
    0x39cb0000U,
    0x1adf8000U,
    0xff44000U,
    0x3e3e6000U,
    0xa8e2b000U,
    0xd114a800U,
    0xaa29ec00U,
    0xe74a5600U,
    0xa89ed900U,
    0x3b948d80U,
    0x4a6f62c0U,
    0xfc2b0da0U,
    0x724b22d0U,
    0x951d6d88U,
    0xf15592f4U,
    0x514fc596U,
    0xd9b7ec5U,
    0xee141380U,
    0x3caae7edU,
    0x79097e36U,
    0x7aff7502U,
    0x72c6bba9U,
    0x3f240bd1U,
    0x9bb2a835U,
    0xd75eec20U,
    0xcd33d61aU,
    0xa91d992bU,
    0x2f536dbcU,
    0x84e92d5U,
    0xe61845adU,
    0xf1d33ed9U,
    0xba0c73aeU,
    0xd27f57d3U,
    0x85845614U,
    0xeec5d921U,
    0x71230dbeU,
    0xbab722c3U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0x98000000U,
    0xa4000000U,
    0x3e000000U,
    0x59000000U,
    0x2e800000U,
    0x9bc00000U,
    0xd200000U,
    0x2e900000U,
    0x89c80000U,
    0x394c0000U,
    0x4b8a0000U,
    0x802d0000U,
    0x99798000U,
    0x1f954000U,
    0xcb4c6000U,
    0x408f9000U,
    0x15ac5800U,
    0x7f3eac00U,
    0x58f23200U,
    0xa0bfbb00U,
    0x61b01f80U,
    0x445a69c0U,
    0x89019fa0U,
    0x968329f0U,
    0x4fc7ff98U,
    0x3b21b9d4U,
    0x4b9227beU,
    0x3d4a55fdU,
    0xc58c75b1U,
    0x312a7eebU,
    0x1bf8323aU,
    0x8652bb0dU,
    0x5699f99U,
    0x279f29f6U,
    0x7025ffa3U,
    0xfe10b9ecU,
    0xab09a7afU,
    0x8aee15c1U,
    0x66db95a0U,
    0xa0c1aef9U,
    0xb0a38a16U,
    0x84d7c719U,
    0x5ca815bbU,
    0x9b9eedbU,
    0x87366a14U,
    0x39d1722U,
    0x8e202d82U,
    0x4714d2e0U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x58000000U,
    0xb4000000U,
    0xfe000000U,
    0x55000000U,
    0xc2800000U,
    0x6dc00000U,
    0x67a00000U,
    0x1dd00000U,
    0xf8e80000U,
    0xf46c0000U,
    0xc4ae0000U,
    0x2d8b0000U,
    0x79b8000U,
    0x6bd14000U,
    0xa1e8e000U,
    0xcce91000U,
    0x4a6cb800U,
    0x31af2c00U,
    0xff0e0200U,
    0x325a5f00U,
    0xb870e180U,
    0x423e7ec0U,
    0x610361a0U,
    0x7c833ed0U,
    0x98c58188U,
    0xb5212ed4U,
    0x2812b99eU,
    0x2b4f42f9U,
    0x17bc5ba1U,
    0x69400decU,
    0x1b660211U,
    0xaef65f37U,
    0x21fee1a8U,
    0xbba57ee0U,
    0x5fd0e19eU,
    0x1fee7ec9U,
    0x39eb6191U,
    0x98ef3eebU,
    0x46b81b9U,
    0x2caa2ec9U,
    0xd189399eU,
    0x159e02f4U,
    0x74d4bba5U,
    0xc8691de4U,
    0x36aaba3dU,
    0x82897312U,
    0x2618e3bcU,
    0x7d9321e4U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0x68000000U,
    0xa4000000U,
    0x2000000U,
    0xcf000000U,
    0xd9800000U,
    0x92400000U,
    0x26e00000U,
    0x6bd00000U,
    0x55880000U,
    0x360c0000U,
    0xf74a0000U,
    0x56290000U,
    0x5ebe8000U,
    0xf555c000U,
    0x8a4ce000U,
    0x88aef000U,
    0x217d8800U,
    0xdbf7c400U,
    0xc47b9600U,
    0xc976c100U,
    0x1db9af80U,
    0x2ad3ebc0U,
    0x450f2fa0U,
    0x20ca2bf0U,
    0xcd69cfb8U,
    0x9cdddbccU,
    0x7242c7a2U,
    0x16e3dfe5U,
    0x63d7b1a2U,
    0x18eeed6U,
    0x5c0f1614U,
    0x9c4a0132U,
    0x8dabcfbdU,
    0x3f8dbd8U,
    0xa364790U,
    0x73df1fd9U,
    0xfbc5d1afU,
    0x7ca5ded9U,
    0x79367e22U,
    0xa45f353eU,
    0x6087d196U,
    0xbec0defbU,
    0xfe22fe19U,
    0x38f3f53fU,
    0x8bfdb196U,
    0x9e37eeffU,
    0x79d9961bU,
    0x60c3c13bU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x68000000U,
    0xc4000000U,
    0x2a000000U,
    0x1f000000U,
    0x62800000U,
    0xc7c00000U,
    0xb0600000U,
    0xb8f00000U,
    0xdee80000U,
    0x126c0000U,
    0x40aa0000U,
    0xdd490000U,
    0x98f88000U,
    0x9db3c000U,
    0x6949e000U,
    0x7afab000U,
    0x16b74800U,
    0x79cadc00U,
    0xe3c2200U,
    0x2a51c100U,
    0x33fdf680U,
    0x1d3269c0U,
    0x258d76a0U,
    0xda1da9f0U,
    0xda8696b8U,
    0x9bc219fcU,
    0x56635ea2U,
    0x79f205cdU,
    0xe16e1ca8U,
    0x6c2ab4caU,
    0x7f8d4230U,
    0xd18b13bU,
    0x2c035ebcU,
    0x2e0205e5U,
    0x55061cb3U,
    0x4d86b4ddU,
    0xcd474221U,
    0xb3a1b117U,
    0x2293deb7U,
    0x791dc5f6U,
    0xae05fc87U,
    0x950504e2U,
    0x2d808a34U,
    0xfd44ad1eU,
    0xdba41c9cU,
    0xe693b4f8U,
    0x531dc21dU,
    0xb1077106U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x38000000U,
    0xac000000U,
    0x12000000U,
    0x8d000000U,
    0x42800000U,
    0x90c00000U,
    0x84600000U,
    0x5d00000U,
    0x1de80000U,
    0x596c0000U,
    0x3aaa0000U,
    0x434b0000U,
    0x29da8000U,
    0x5a134000U,
    0xbe0a2000U,
    0xaf8b000U,
    0x85e3d800U,
    0x90954c00U,
    0x7acf8a00U,
    0x109a0900U,
    0x2f305780U,
    0x7af880c0U,
    0x5de2d7a0U,
    0x2c97c0f0U,
    0x60caf7a8U,
    0x99870fcU,
    0xd3b1afa6U,
    0x75397cd7U,
    0x160485a2U,
    0xfb0385f4U,
    0x6987aa32U,
    0xd145b910U,
    0xf1230fb3U,
    0x72f58ce1U,
    0x5c1dfd9cU,
    0xd8f639cbU,
    0xbd1b5814U,
    0xa8710c3cU,
    0x90dd2a34U,
    0x5696f912U,
    0xa9c92f87U,
    0xdd1d3cdbU,
    0x787625b9U,
    0xb8df75c1U,
    0xf296d20dU,
    0x2fcc053dU,
    0xee1dfdb1U,
    0xa5f639daU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0x78000000U,
    0xcc000000U,
    0x2e000000U,
    0xd5000000U,
    0xbe800000U,
    0xd4c00000U,
    0x57600000U,
    0x8d500000U,
    0x6ad80000U,
    0x1e5c0000U,
    0xa59a0000U,
    0x8b7d0000U,
    0xff6d8000U,
    0xc9964000U,
    0x85bc2000U,
    0xbf0f5000U,
    0xf140f800U,
    0x38268c00U,
    0x28756a00U,
    0xb82d7100U,
    0xf8b28b80U,
    0x454fc540U,
    0xcae70ba0U,
    0xce158570U,
    0xf1f92b98U,
    0x61abd574U,
    0x31f65386U,
    0x916a1947U,
    0x3c90998dU,
    0xcb3f7872U,
    0x3c94a22U,
    0x12222107U,
    0x57727397U,
    0xb9a94954U,
    0xcdf26195U,
    0x3768f467U,
    0x4d93a01aU,
    0xefb81032U,
    0xe00b5810U,
    0xc29c36U,
    0x6564322dU,
    0xee52ed1dU,
    0xdd5b39b5U,
    0x221b6876U,
    0x3f38121bU,
    0xc1cdbd39U,
    0x1923c1a1U,
    0x54f1e44cU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0xc8000000U,
    0x8c000000U,
    0x86000000U,
    0x8d000000U,
    0x11800000U,
    0x66400000U,
    0xb1e00000U,
    0x16d00000U,
    0x33680000U,
    0x63ec0000U,
    0xf22a0000U,
    0x988f0000U,
    0x763b8000U,
    0x71c5c000U,
    0x9aa5a000U,
    0x16b79000U,
    0xd3fef800U,
    0x2966f400U,
    0xb118600U,
    0xdfcb6900U,
    0x5a6f80U,
    0xe4535cc0U,
    0x15abefa0U,
    0x4bc99cf0U,
    0xa25dcf88U,
    0x5357ccd4U,
    0x572c97baU,
    0xf509a8f7U,
    0x6e78699bU,
    0x9921f5d4U,
    0x5ff6201fU,
    0xc49e500dU,
    0x1b715813U,
    0x375e6438U,
    0x66d4fe37U,
    0xb685d04U,
    0x17ee49b9U,
    0xa02fa5c1U,
    0xd78f7825U,
    0xe0bc340eU,
    0xd07a60fU,
    0xd1853906U,
    0xc643378dU,
    0x21e138ceU,
    0xded511b0U,
    0xbf6ec1d1U,
    0xe5e80601U,
    0x7f2da910U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x78000000U,
    0x84000000U,
    0x36000000U,
    0xb5000000U,
    0x2e800000U,
    0x15c00000U,
    0xdc200000U,
    0xdb700000U,
    0x2a480000U,
    0x5cc0000U,
    0xb08a0000U,
    0x28690000U,
    0x87bc8000U,
    0xae474000U,
    0xf7676000U,
    0xba90d000U,
    0x611e7800U,
    0x3ef68c00U,
    0x980fe200U,
    0x30ac2300U,
    0x911bc880U,
    0xe6f272c0U,
    0xcc0d48a0U,
    0xceac32f0U,
    0xd81ea8b8U,
    0x7a77a2dcU,
    0x5acdb0a6U,
    0x8908fefdU,
    0x3828aaabU,
    0x99d911d0U,
    0x5851e020U,
    0xc8be9015U,
    0x3fc59837U,
    0xab211c13U,
    0x6ef6fa25U,
    0x900a7f3eU,
    0xacaa5297U,
    0x9318ddd5U,
    0x1df1621bU,
    0xd38e6323U,
    0xc3ea2895U,
    0xa4fce2d4U,
    0x53a0d09cU,
    0xbe312ed0U,
    0x7aaa52bbU,
    0x5618ddfdU,
    0x2b71622bU,
    0xf24e6307U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xc8000000U,
    0xfc000000U,
    0xde000000U,
    0x89000000U,
    0xa0800000U,
    0xccc00000U,
    0x58200000U,
    0x60b00000U,
    0xe3380000U,
    0x1ebc0000U,
    0xfa0000U,
    0x9b1f0000U,
    0x8e0e8000U,
    0x8bc0c000U,
    0x49a32000U,
    0x98f3d000U,
    0x65da9800U,
    0xb92a1c00U,
    0x56f43a00U,
    0xd4d81b00U,
    0x8dae6e80U,
    0x88307f40U,
    0x13faeea0U,
    0x389fbf70U,
    0xd94f4e88U,
    0x10a0af7cU,
    0xa074f6baU,
    0xd1a6343U,
    0x3b0ed48dU,
    0x5547a461U,
    0x3c612025U,
    0xa850d012U,
    0x35ae1813U,
    0xdc35dc3aU,
    0xa1f99a03U,
    0x979b0b31U,
    0xc4cfd6b1U,
    0x5ee5b368U,
    0x63964c85U,
    0x1c0eb858U,
    0xd4c19a17U,
    0x3c270b01U,
    0x1ab5d68aU,
    0xb03ab378U,
    0xdd38cc98U,
    0xc7be7854U,
    0xf87aba09U,
    0x53d8db10U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x28000000U,
    0xe4000000U,
    0xc6000000U,
    0x3b000000U,
    0xcd800000U,
    0xbd400000U,
    0xab600000U,
    0xf7100000U,
    0x73780000U,
    0xe3fc0000U,
    0x783a0000U,
    0xab9b0000U,
    0xbeab8000U,
    0x7fa2c000U,
    0xc731e000U,
    0x2409d000U,
    0x92150800U,
    0x99f9e400U,
    0xb93e4e00U,
    0x671bb500U,
    0xef6afd80U,
    0x95065e40U,
    0xf2837da0U,
    0x86c39e70U,
    0x15231da8U,
    0x5df38e74U,
    0x1b2c75a2U,
    0xc6e17a4dU,
    0x9a56d393U,
    0xd01afb43U,
    0xe0ef6820U,
    0xe542f42cU,
    0x9762a62aU,
    0xcd178121U,
    0x527bbbbeU,
    0x1f7f0f5dU,
    0x11fc4e0aU,
    0x8d3cb53eU,
    0xa91b7dbcU,
    0x206f9e5bU,
    0xb6811d9cU,
    0x90c48e70U,
    0x2625f5a1U,
    0x6474ba58U,
    0x486eb39bU,
    0xb286eb61U,
    0x66c20016U,
    0x25270032U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0x58000000U,
    0x54000000U,
    0x32000000U,
    0x93000000U,
    0x77800000U,
    0x71c00000U,
    0x58a00000U,
    0x3b900000U,
    0xe0380000U,
    0x6bc0000U,
    0xbfa0000U,
    0xa11f0000U,
    0x31ea8000U,
    0x79e04000U,
    0x8776e000U,
    0x29c97000U,
    0x4e34e800U,
    0x9928d400U,
    0xe1c69a00U,
    0x70a47f00U,
    0x97907380U,
    0x363b0a40U,
    0xffb8f3a0U,
    0xbb784a70U,
    0x955e9388U,
    0x8b8e7a5cU,
    0x6d561b9eU,
    0x2d1f9e49U,
    0x97ea8992U,
    0x78e6456dU,
    0xd3f2880fU,
    0x778ee431U,
    0x33561239U,
    0xc8199b0fU,
    0xa96ee1a9U,
    0xe1a1d162U,
    0xeb10721dU,
    0xc47fab2fU,
    0x54de6995U,
    0x134c3549U,
    0x6d76e037U,
    0x2ec97016U,
    0xabb4e820U,
    0xcbe8d413U,
    0x96e69a0fU,
    0x6ef47f0fU,
    0x1d0873b9U,
    0x98170a49U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0x28000000U,
    0xd4000000U,
    0xb6000000U,
    0xcd000000U,
    0xaa800000U,
    0xfc00000U,
    0x5a600000U,
    0x61d00000U,
    0xdef80000U,
    0x977c0000U,
    0xa33a0000U,
    0xae190000U,
    0x396e8000U,
    0x47a7c000U,
    0xa8b56000U,
    0x6caed000U,
    0xaac7b800U,
    0x84e76c00U,
    0x48135200U,
    0x919ba100U,
    0xa02fdf80U,
    0x1006a040U,
    0x18035fa0U,
    0x9c046070U,
    0x9202bfb8U,
    0x5304704cU,
    0xb386e7b2U,
    0x13460c79U,
    0x98a56d9fU,
    0x9131114aU,
    0xb0ead835U,
    0x13e5bc09U,
    0x5596ea03U,
    0xd3d9cd21U,
    0x80dacU,
    0xddf3c178U,
    0x410f602eU,
    0xfd77d016U,
    0x81c93835U,
    0x8690ac38U,
    0xa05e3227U,
    0x73497129U,
    0xb55267bcU,
    0xf838cc56U,
    0x999e8dbeU,
    0x24280157U,
    0x1e00001aU,
    0xd900001fU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x58000000U,
    0x54000000U,
    0xc6000000U,
    0x4f000000U,
    0x1f800000U,
    0x94400000U,
    0xfae00000U,
    0xe2d00000U,
    0xa4b80000U,
    0xe63c0000U,
    0xa4fe0000U,
    0x7ad90000U,
    0x192e8000U,
    0x69e44000U,
    0x6752e000U,
    0xa5fed000U,
    0x765fa800U,
    0x48ead400U,
    0x7243b600U,
    0xe5e54100U,
    0xb554a780U,
    0xccfe5a40U,
    0xd6dc27a0U,
    0x9b2f1a50U,
    0x48e64798U,
    0x3bd08a6cU,
    0xbc3b8f8eU,
    0x91f9ce79U,
    0x5bf1bfU,
    0xbfeccb6aU,
    0xc9c1a838U,
    0xefa3d40fU,
    0x54b53626U,
    0xf7ad0127U,
    0xa92047b7U,
    0x98758a62U,
    0x9a8b0fa8U,
    0x1f548e79U,
    0xa1ff91a7U,
    0xf85a5b4cU,
    0xfbeae02fU,
    0xa7c2d015U,
    0xaca1a81dU,
    0xd933d43cU,
    0xeaed3617U,
    0x3410125U,
    0xf16647bbU,
    0x50908a65U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0x48000000U,
    0x54000000U,
    0x5a000000U,
    0xaf000000U,
    0xe2800000U,
    0x6cc00000U,
    0xe7e00000U,
    0xa5900000U,
    0xd9780000U,
    0xcafc0000U,
    0x59ba0000U,
    0xbb190000U,
    0x75298000U,
    0x31e3c000U,
    0x8c956000U,
    0xfaf9b000U,
    0xa1bbc800U,
    0xd71a5c00U,
    0x23287a00U,
    0xd8e56500U,
    0xcf14d980U,
    0x81bbca40U,
    0x671f59a0U,
    0x1b2d0a70U,
    0xd4e1b9b8U,
    0x89127a74U,
    0x20bc91aaU,
    0x709b5661U,
    0x3a6bc3bcU,
    0xbd421f4aU,
    0xa7a3c804U,
    0xbbb65c11U,
    0xc68a7a1dU,
    0x70506508U,
    0x951f59b3U,
    0x302d0a7eU,
    0x4461b997U,
    0xed27a75U,
    0x37dc9184U,
    0x42cb5656U,
    0xbc73c38dU,
    0x11ee1f66U,
    0x2201c80aU,
    0x3035c11U,
    0xd481fa2cU,
    0x95c6a53eU,
    0xec61b9a8U,
    0x8ad27a5dU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x78000000U,
    0x84000000U,
    0xf2000000U,
    0x79000000U,
    0x7d800000U,
    0x6bc00000U,
    0x8ce00000U,
    0xce900000U,
    0x83380000U,
    0xe1bc0000U,
    0xb4fe0000U,
    0xcb5d0000U,
    0x196a8000U,
    0x40e14000U,
    0xe0916000U,
    0x1c3cf000U,
    0x9f383800U,
    0x27be5400U,
    0x27fdca00U,
    0xbbda0d00U,
    0x85282980U,
    0x5ac3d840U,
    0x2d64a9a0U,
    0x43539850U,
    0xcd949b8U,
    0xd7af286cU,
    0x578491a6U,
    0x46c1cc4dU,
    0xeb62839aU,
    0xd0542553U,
    0x7c583805U,
    0x4bee5409U,
    0x4da5ca06U,
    0x8b360d2aU,
    0xb40e29beU,
    0x43b2d86eU,
    0x8c4829adU,
    0x3b93d86dU,
    0xb0bca98fU,
    0xe57f985fU,
    0x3e9f49b6U,
    0x128e2863U,
    0x84f0118eU,
    0x7aed8c47U,
    0xec2163b8U,
    0x6f5954bU,
    0x3bef6017U,
    0x75a1f01cU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0xf8000000U,
    0x3c000000U,
    0x8e000000U,
    0x1d000000U,
    0xac800000U,
    0x61c00000U,
    0x90e00000U,
    0xaf900000U,
    0xac80000U,
    0x314c0000U,
    0x330e0000U,
    0x4daf0000U,
    0xca9b8000U,
    0xf6644000U,
    0x2151a000U,
    0x85afd000U,
    0xde9a3800U,
    0x9c61dc00U,
    0x1e501200U,
    0x422bf500U,
    0x32d83980U,
    0x4fc7e7c0U,
    0x5de5b9a0U,
    0xdb10a7d0U,
    0xc70999a8U,
    0xd7a837f4U,
    0xbd9f8196U,
    0x65e17bfbU,
    0x87118bb5U,
    0xf90fc2fcU,
    0xa2a9b81eU,
    0xd5199c24U,
    0xb627b21aU,
    0x84f7251fU,
    0xe71f8190U,
    0x65217bf7U,
    0x15718b82U,
    0xba5fc2fbU,
    0x6c01b81bU,
    0x16059c25U,
    0x9101b226U,
    0x4a842503U,
    0xb8c2019eU,
    0x8e663becU,
    0x5d55abaeU,
    0x2bab52f7U,
    0x5399a021U,
    0xc8e3d03fU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0xb8000000U,
    0x74000000U,
    0xbe000000U,
    0x97000000U,
    0xa3800000U,
    0xb5c00000U,
    0x50e00000U,
    0x65700000U,
    0x5cd80000U,
    0xaa5c0000U,
    0x691a0000U,
    0x3abd0000U,
    0xec6d8000U,
    0x76604000U,
    0xfe37e000U,
    0x6b7c1000U,
    0x7a8ba800U,
    0x10103400U,
    0xaf6b5a00U,
    0x5be0fd00U,
    0x94f6e680U,
    0x441a9840U,
    0xd83966a0U,
    0x412bd870U,
    0xeac10698U,
    0xef668874U,
    0x42b2ceb6U,
    0x413bec69U,
    0xfdaa5c99U,
    0xc086754cU,
    0x68442831U,
    0x92217421U,
    0x86133a05U,
    0x346dad08U,
    0xd2652eaaU,
    0x1837fc66U,
    0x9879f4beU,
    0xdf0a4145U,
    0x46d57212U,
    0xe20c8939U,
    0x1c505c8bU,
    0x3b4b7563U,
    0x3ef1a836U,
    0xbd1d342fU,
    0xb4beda14U,
    0x336cbd17U,
    0x9e30686U,
    0x39f7886dU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0x48000000U,
    0x74000000U,
    0x9e000000U,
    0x5b000000U,
    0x3f800000U,
    0xe3400000U,
    0xa1a00000U,
    0x52300000U,
    0xa6180000U,
    0xb55c0000U,
    0x1cfe0000U,
    0xdac90000U,
    0x92d28000U,
    0x348cc000U,
    0x63f16000U,
    0xc47c9000U,
    0xac0cc800U,
    0xf531f400U,
    0xd39f8e00U,
    0x31a1300U,
    0xe1da4a80U,
    0x683f8e40U,
    0x9c2b2aa0U,
    0xb7431e50U,
    0xefa7e298U,
    0xd132ea4cU,
    0xd5986c8aU,
    0xf418f951U,
    0xc5a262dU,
    0x727b7707U,
    0xb30f0c82U,
    0xfcb1696fU,
    0x6fda6e32U,
    0x4b3f4337U,
    0x7fab6291U,
    0xce072a4bU,
    0xc3038cbbU,
    0x9384a97aU,
    0x31418e1dU,
    0x58a31318U,
    0xdcb0cab7U,
    0xbfdf4e5eU,
    0x933c4aadU,
    0x33aa8e78U,
    0x6c07aabaU,
    0x7206de4eU,
    0x69040295U,
    0xb682ba5dU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x68000000U,
    0x94000000U,
    0xb2000000U,
    0x49000000U,
    0x4a800000U,
    0x77400000U,
    0xeea00000U,
    0xd0f00000U,
    0x10680000U,
    0xb72c0000U,
    0x618e0000U,
    0x5d7b0000U,
    0xb3168000U,
    0x339c000U,
    0x47336000U,
    0xdc8df000U,
    0xf5f93800U,
    0xf552ac00U,
    0xeb1a7600U,
    0xae84f300U,
    0x4d414c80U,
    0x23a263c0U,
    0xf0722ca0U,
    0xd22f93d0U,
    0x350b1488U,
    0x3d3d3ffcU,
    0xa0316292U,
    0xe909ccd9U,
    0xeb382e3eU,
    0x9337af0bU,
    0xce8c028cU,
    0xcff3cc6U,
    0xd7d7963fU,
    0x85cc30eU,
    0xf22514a9U,
    0xd4b63ffaU,
    0x794fe28fU,
    0x301c0ce6U,
    0x5d054e2fU,
    0xb8815f20U,
    0x9e43ba81U,
    0x142450d9U,
    0xcfb68035U,
    0x6ac9c00eU,
    0xd5db6028U,
    0x38e1f02fU,
    0xa0d7382eU,
    0xa5d9ac32U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0x28000000U,
    0xec000000U,
    0x6e000000U,
    0xff000000U,
    0x6f800000U,
    0x72c00000U,
    0xfd600000U,
    0xed00000U,
    0x83680000U,
    0x89ac0000U,
    0xca0000U,
    0x41f0000U,
    0x8a748000U,
    0xd559c000U,
    0xdbd4a000U,
    0x51e8d000U,
    0xa5eba800U,
    0x6fe91400U,
    0xc2ea2200U,
    0x646b1900U,
    0x4229c780U,
    0x200d48c0U,
    0x307d67a0U,
    0xd02598f0U,
    0xc976cf88U,
    0x4cdc8cecU,
    0x3214ed82U,
    0x310b95d7U,
    0x60ff2a19U,
    0x6d65dd28U,
    0x36d4cda2U,
    0x976a85c4U,
    0x53a82215U,
    0x55c8192bU,
    0x169f47b7U,
    0x63788deU,
    0xca7f47aeU,
    0x352788f7U,
    0x53f747aeU,
    0x5c9b88c6U,
    0x6b3547adU,
    0xccf888c9U,
    0x6363c79bU,
    0x19d248d0U,
    0xa0e9e7afU,
    0xe56c58f5U,
    0x2aaa6f82U,
    0x89485ce8U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0xd8000000U,
    0x2c000000U,
    0xbe000000U,
    0x41000000U,
    0xa2800000U,
    0xd0c00000U,
    0x71600000U,
    0xed900000U,
    0x76580000U,
    0x6a9c0000U,
    0xb5fe0000U,
    0x916f0000U,
    0xc1b48000U,
    0x31ee4000U,
    0x26712000U,
    0x70bd000U,
    0x83a55800U,
    0xdbf20c00U,
    0x44baa00U,
    0x7884f100U,
    0xcbc64580U,
    0x88e3bb40U,
    0xe4d765a0U,
    0x7e786b50U,
    0x572a3da8U,
    0xd216674cU,
    0x221f979eU,
    0x863d9647U,
    0x458d5231U,
    0x43606d17U,
    0x229317b9U,
    0x6ddfd673U,
    0x44da720dU,
    0x7258bd04U,
    0x609ccf8eU,
    0x46fc9a59U,
    0xbcec7812U,
    0x2af5dc2aU,
    0x1ec8f23bU,
    0xd445fd05U,
    0xaca76f97U,
    0xd0760a4cU,
    0xa20c8027U,
    0x3b224024U,
    0x70372025U,
    0xdca8d017U,
    0x3457d81dU,
    0x41bf4c1fU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0xd8000000U,
    0x84000000U,
    0xc6000000U,
    0xaf000000U,
    0x19800000U,
    0xe4400000U,
    0x29600000U,
    0x73b00000U,
    0x81280000U,
    0x96c0000U,
    0xfa0e0000U,
    0x44bf0000U,
    0x5128000U,
    0x86b84000U,
    0x2c126000U,
    0x103b5000U,
    0x61d4a800U,
    0x5182400U,
    0xbf076600U,
    0x31807d00U,
    0x9844c880U,
    0xdb6525c0U,
    0xc2b6a8a0U,
    0xb3ae75d0U,
    0x32aa00a8U,
    0x982a51fcU,
    0x7eb669eU,
    0x13c92cddU,
    0x279b2e2fU,
    0xacc74936U,
    0x2fa36689U,
    0x29152cdfU,
    0x7cbd2e2bU,
    0x5114493fU,
    0x8bfe695U,
    0x6f126ce0U,
    0x93bdce04U,
    0x68975907U,
    0x5cf92eacU,
    0x9e7118e9U,
    0x640e0004U,
    0x2fbf003eU,
    0xfa92801fU,
    0x7df8402eU,
    0xc4f2602dU,
    0x3cb5032U,
    0xf9ca81fU,
    0xd0c42419U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0x68000000U,
    0x94000000U,
    0x3e000000U,
    0x63000000U,
    0x88800000U,
    0x66400000U,
    0xe4600000U,
    0xf0500000U,
    0xab980000U,
    0xe3dc0000U,
    0x6cba0000U,
    0x706f0000U,
    0xbbb58000U,
    0x230ac000U,
    0x9d646000U,
    0x7dd29000U,
    0x2a580800U,
    0x7b7c8c00U,
    0xbfccde00U,
    0x8e40ed00U,
    0xb067a780U,
    0x6e511740U,
    0x989bc7a0U,
    0x35f8770U,
    0x9ef9cf88U,
    0xaa0c0b64U,
    0x28e09192U,
    0x162641U,
    0x18fb561dU,
    0xf509a119U,
    0xaa62999fU,
    0x6b55aa70U,
    0x7f1a082eU,
    0x7f9f8c28U,
    0xbddb5e16U,
    0xffb92d11U,
    0xc0ec47b0U,
    0x21f64764U,
    0x6d6a2f89U,
    0x30375b78U,
    0x3dcb7992U,
    0x2741fa4eU,
    0x95e46033U,
    0xdb929028U,
    0x6e38083bU,
    0xdb2c8c3bU,
    0x7c54de02U,
    0xf99ced01U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0xb8000000U,
    0x4c000000U,
    0xd6000000U,
    0x55000000U,
    0x8d800000U,
    0x82c00000U,
    0x3fa00000U,
    0x1c300000U,
    0xd280000U,
    0x21ec0000U,
    0x274e0000U,
    0x70fd0000U,
    0x62168000U,
    0x5df4000U,
    0x26632000U,
    0xe7109000U,
    0x6058b800U,
    0x40a59400U,
    0x8ab3a200U,
    0xb6a4f00U,
    0xf085680U,
    0x271c0bc0U,
    0x49c376a0U,
    0x53209bd0U,
    0xe975ceb8U,
    0x43480fccU,
    0x62f8ec96U,
    0x751100dfU,
    0x375d9a23U,
    0x12209b0aU,
    0xcef0d4a0U,
    0x9a8ad4faU,
    0xb9dd9817U,
    0x8640431U,
    0x9e139a02U,
    0x4bdd9b26U,
    0x2f665488U,
    0x749594e0U,
    0x439eb808U,
    0x68849417U,
    0x97432220U,
    0xe1640f10U,
    0x3d93f6a1U,
    0xb01edbe0U,
    0x3b466ebdU,
    0x766dfe7U,
    0xb093f4a2U,
    0x19a44d2U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0x68000000U,
    0x24000000U,
    0x66000000U,
    0x51000000U,
    0x2b800000U,
    0x9ac00000U,
    0x7a00000U,
    0xaf300000U,
    0xa4b80000U,
    0xc07c0000U,
    0x1ada0000U,
    0x7c6f0000U,
    0x5148000U,
    0xd04bc000U,
    0xf0e0e000U,
    0xd354d000U,
    0xefaed800U,
    0x5364400U,
    0x77b97200U,
    0x7cff1f00U,
    0x8a9a7e80U,
    0xa28a7f40U,
    0x13429ea0U,
    0xba62af70U,
    0x6a164688U,
    0x86cbeb6cU,
    0xd3a3b492U,
    0x61333465U,
    0x71bb2a0bU,
    0xfdfe9b31U,
    0xb919eca1U,
    0x744eb067U,
    0x56e2b828U,
    0x22555420U,
    0xb42d4a13U,
    0xf7f28b35U,
    0x541954beU,
    0xb5c8e46bU,
    0x7f217216U,
    0x49731f14U,
    0x93587eb0U,
    0xc1a97f55U,
    0xc0341ebbU,
    0xf23a6f4cU,
    0xe33826a3U,
    0xa8bbfb56U,
    0xe2790ca6U,
    0xfdda605eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0x48000000U,
    0xb4000000U,
    0x46000000U,
    0xdb000000U,
    0xc8800000U,
    0x43c00000U,
    0x77200000U,
    0x28900000U,
    0xd5f80000U,
    0xc43c0000U,
    0x9e1e0000U,
    0xcd0f0000U,
    0xd7b08000U,
    0x296ec000U,
    0x87c56000U,
    0x39205000U,
    0x67920800U,
    0xab788c00U,
    0x14fef200U,
    0x95ba9b00U,
    0xe05b9780U,
    0xae681440U,
    0xd46f7a0U,
    0x8fe44450U,
    0x4732ffa8U,
    0x6afc86cU,
    0x9ae28dbaU,
    0x80b49341U,
    0x1becfa2bU,
    0xcd021737U,
    0xeb8565b9U,
    0x57428f77U,
    0x36e5600cU,
    0x8ab05011U,
    0x5aea0803U,
    0xdc848c21U,
    0xb5c0f22fU,
    0xc4259b35U,
    0xa41317abU,
    0x983ad474U,
    0xdc1d9794U,
    0x380b1446U,
    0x8030779fU,
    0xac29847bU,
    0xdc211f87U,
    0xe812587bU,
    0x23b6589U,
    0x51d8f6eU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x48000000U,
    0xb4000000U,
    0x22000000U,
    0xdb000000U,
    0x54800000U,
    0xd1400000U,
    0xb9600000U,
    0x64500000U,
    0x5aa80000U,
    0xd0ec0000U,
    0x7d0e0000U,
    0x28190000U,
    0x7bd48000U,
    0x876ac000U,
    0x14c96000U,
    0xce3d7000U,
    0x13e12800U,
    0xe8104c00U,
    0x564b2600U,
    0xa6795d00U,
    0xc5023780U,
    0xb985dec0U,
    0x2cc557a0U,
    0x1fa1aed0U,
    0xa4f0ff98U,
    0xc0db22fcU,
    0xae72b98aU,
    0x1c9f0fd1U,
    0x5a91a602U,
    0x960a9d27U,
    0x749fd7b7U,
    0xfe926ec3U,
    0xac0d1f81U,
    0xe39c92e6U,
    0x7412f1bdU,
    0x304e33d3U,
    0xf77da838U,
    0x4e868c02U,
    0x5644461bU,
    0x2be12d31U,
    0xb4119f9dU,
    0x904a52c3U,
    0xa77d9193U,
    0xb68643e9U,
    0xaa46003fU,
    0xbde50024U,
    0x4d128028U,
    0x1fcfc028U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xa8000000U,
    0xac000000U,
    0x4a000000U,
    0x21000000U,
    0x91800000U,
    0x59c00000U,
    0x8e00000U,
    0xb5300000U,
    0xdbe80000U,
    0xcb2c0000U,
    0x5e4e0000U,
    0x68b90000U,
    0x42b28000U,
    0xa92bc000U,
    0x9348a000U,
    0x533ff000U,
    0x4a727800U,
    0x984c3400U,
    0xd3b8da00U,
    0xca32ad00U,
    0x156cdc80U,
    0x8c6f47c0U,
    0x29ea7ca0U,
    0xde29b7d0U,
    0x61ca8498U,
    0x867e43ccU,
    0x18d2feb2U,
    0x595f1ee7U,
    0xc2025a20U,
    0x5d056d2fU,
    0xe3827ca4U,
    0xcc5b7e9U,
    0xd76484beU,
    0x2bf743c8U,
    0x29887e9aU,
    0x9758decdU,
    0x4504fa35U,
    0x47839d34U,
    0x7ac284b3U,
    0x646243d0U,
    0x9f74fea9U,
    0x7ca1ef3U,
    0xad7eda1dU,
    0x4857ad3dU,
    0xe1185c90U,
    0x3b2187d4U,
    0x4cd65cbcU,
    0xc75887d5U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0xf8000000U,
    0xd4000000U,
    0x4a000000U,
    0xf000000U,
    0x4a800000U,
    0xed400000U,
    0x44e00000U,
    0xd7500000U,
    0xaa780000U,
    0xe03c0000U,
    0x25a0000U,
    0xc6490000U,
    0xbdd78000U,
    0x373c4000U,
    0xecd8e000U,
    0x290fb000U,
    0xa324800U,
    0xcced5c00U,
    0x3ae15600U,
    0x82568100U,
    0x77f99380U,
    0x43f8ac40U,
    0x19fb73a0U,
    0x8efe1c70U,
    0xc07ebbb8U,
    0x9f3f0044U,
    0xdf0d86U,
    0xd70a3171U,
    0x9f36d614U,
    0x716ac132U,
    0x492173a6U,
    0x61f71c79U,
    0xeb493b8fU,
    0x74534048U,
    0xbeffed83U,
    0x28798171U,
    0xd33e9e2fU,
    0x4ede9d36U,
    0x6a0fa5a0U,
    0xeb1dd73U,
    0x9caa4820U,
    0xef815c33U,
    0xb8c3561fU,
    0x7b23813cU,
    0x7af4138fU,
    0x8bcdec57U,
    0x461413a3U,
    0x489dec46U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0x98000000U,
    0xb4000000U,
    0xa2000000U,
    0xdf000000U,
    0x60800000U,
    0x6dc00000U,
    0xffe00000U,
    0xf2700000U,
    0x18a80000U,
    0x26c0000U,
    0x10a0000U,
    0x60bb0000U,
    0x86f18000U,
    0x47694000U,
    0x9a8ee000U,
    0x77fcf000U,
    0x2fd18800U,
    0x1078cc00U,
    0x7690c200U,
    0x505c8d00U,
    0x6c017680U,
    0xf6010fc0U,
    0xd0596a0U,
    0x7786fff0U,
    0x21459ea8U,
    0x842773e4U,
    0x7093bc8eU,
    0x555b0ec9U,
    0x17814226U,
    0x9145cd3eU,
    0x6c27969eU,
    0xc91ffd5U,
    0xdb5e1e89U,
    0xde8533cdU,
    0x8cc4dca1U,
    0xbe62bef4U,
    0xfe342a2fU,
    0xbb0af136U,
    0xcbbf5c9eU,
    0x6470fefbU,
    0xa5ab4a37U,
    0x9def4125U,
    0x5c48348bU,
    0x6a98c2d4U,
    0x3f600036U,
    0xfb00026U,
    0x5f480032U,
    0x141c0037U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x58000000U,
    0xfc000000U,
    0x46000000U,
    0x3d000000U,
    0x7f800000U,
    0x8cc00000U,
    0x39600000U,
    0xbaf00000U,
    0x68f80000U,
    0x953c0000U,
    0x8dde0000U,
    0x9ae90000U,
    0xc2f08000U,
    0x4fe4000U,
    0x5b3fa000U,
    0x24d89000U,
    0x76ff800U,
    0xc9374400U,
    0x391a0200U,
    0x2f881900U,
    0x86031680U,
    0x9d00c540U,
    0xaf82b6a0U,
    0xe4c15550U,
    0x9d65ce98U,
    0xf4515cU,
    0x13fe6c8eU,
    0xd7bdd863U,
    0x7e9a021fU,
    0x2f48192cU,
    0x416316a0U,
    0xd6f0c55fU,
    0xa6fab6b6U,
    0x3c3d556dU,
    0x105bce82U,
    0x912d517bU,
    0xff16eca6U,
    0x704f986dU,
    0xf9e3a218U,
    0xbeb58900U,
    0x61da6ebfU,
    0x14ecc154U,
    0x8bf194b2U,
    0xa97a9c71U,
    0x8f8002bU,
    0xe53c000aU,
    0x35de003eU,
    0x56e9002aU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0x68000000U,
    0xc4000000U,
    0x5e000000U,
    0xb9000000U,
    0x77800000U,
    0x37400000U,
    0xbfa00000U,
    0xe3300000U,
    0x3fe80000U,
    0x7dac0000U,
    0x5f8e0000U,
    0x34fb0000U,
    0x46378000U,
    0x6a68c000U,
    0xa1e8e000U,
    0x24ad1000U,
    0x18095800U,
    0xfbbd2400U,
    0x5591fe00U,
    0x135cd300U,
    0x7902f280U,
    0x978679c0U,
    0x74412a0U,
    0x47a069d0U,
    0x4f32ca88U,
    0xa5e98df4U,
    0x9aadd492U,
    0x910f4ec5U,
    0x743ffe05U,
    0xced7d32bU,
    0x36fd72b8U,
    0x7d32b9f6U,
    0x66eaf29fU,
    0x3a2a79faU,
    0x90ca12a2U,
    0x275b69f0U,
    0x3f054ab8U,
    0xb2814de2U,
    0x12c534a3U,
    0x3be25edbU,
    0xa416a639U,
    0xe11af727U,
    0xe3248cbfU,
    0xf0f26aeeU,
    0x7f8e003aU,
    0xe4fb000cU,
    0x8e378030U,
    0x3e68c03bU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0xc8000000U,
    0x94000000U,
    0xfe000000U,
    0xdf000000U,
    0x19800000U,
    0x59400000U,
    0x3c600000U,
    0xefd00000U,
    0xd6c80000U,
    0x7c8c0000U,
    0xdb6a0000U,
    0x8fd0000U,
    0x22d48000U,
    0x724ec000U,
    0x69486000U,
    0xb6ce9000U,
    0x8c8ee800U,
    0xf369e400U,
    0xacf9ce00U,
    0x14d37d00U,
    0x39490b80U,
    0x8ecb89c0U,
    0x308b6ba0U,
    0xa96819f0U,
    0x45f90398U,
    0x46533dfcU,
    0x878aadaaU,
    0x34efd0d9U,
    0xfabbce15U,
    0xc9327d2eU,
    0xd3df8bb3U,
    0xe6449c8U,
    0x82d58ba4U,
    0x624949cfU,
    0x71490b9bU,
    0xdacb89f9U,
    0x2e8b6ba0U,
    0x466819e9U,
    0x94790391U,
    0x8b133dd4U,
    0x45eaad82U,
    0x43fd0d4U,
    0x35f3ce1bU,
    0xecfe7d0cU,
    0x34d58b93U,
    0xe94949e4U,
    0x76c90bbbU,
    0x6c8b89e5U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0xf8000000U,
    0x94000000U,
    0xe2000000U,
    0x1f000000U,
    0xc1800000U,
    0xd0400000U,
    0x9b600000U,
    0x60700000U,
    0xacf80000U,
    0x41bc0000U,
    0xa45e0000U,
    0xa5e90000U,
    0x2a358000U,
    0xfc9e4000U,
    0xe24ce000U,
    0x57e6f000U,
    0x46b18800U,
    0xaa5ec400U,
    0x8ce82600U,
    0xf6b1cd00U,
    0x6259e680U,
    0x90ec5540U,
    0xe8b306a0U,
    0xf35fa550U,
    0x38690e98U,
    0x4762174U,
    0x66f848a6U,
    0xd2bf5c51U,
    0x83dcc61eU,
    0x70ab3d16U,
    0xfbd66e8eU,
    0x6f2b9172U,
    0x1a96a0b0U,
    0x2ccc285eU,
    0xfe22081dU,
    0x7995840fU,
    0x234f4612U,
    0xa7607d14U,
    0x4e710ea8U,
    0xb5fa217cU,
    0x153e4898U,
    0xc69a5c70U,
    0x594f4604U,
    0xfc607d31U,
    0x75f10e8bU,
    0x7eba2174U,
    0x55de48b6U,
    0xfdaa5c6dU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0xb8000000U,
    0xec000000U,
    0xee000000U,
    0xdd000000U,
    0x24800000U,
    0x2e400000U,
    0xfae00000U,
    0xb4500000U,
    0x50180000U,
    0x955c0000U,
    0xa3a0000U,
    0x52ab0000U,
    0xcf178000U,
    0x8a7e4000U,
    0x954b2000U,
    0xefc35000U,
    0xdc243800U,
    0xaef25c00U,
    0x53a88e00U,
    0xed97c900U,
    0xd3c3380U,
    0xd92d1f40U,
    0xfed513a0U,
    0x63594f70U,
    0x6b3caba8U,
    0x2e5354U,
    0x40508586U,
    0x2188a6fU,
    0x3e59ae3dU,
    0x8fbf9918U,
    0xe6f8b94U,
    0x2ab10363U,
    0xe4ebd82U,
    0x8241d65aU,
    0xf4e6a030U,
    0xd9561020U,
    0x5c989807U,
    0xef1f4c08U,
    0xf2df960eU,
    0xd5fac532U,
    0x668a85b2U,
    0x15e38a54U,
    0x4bd62e34U,
    0xf3ddd930U,
    0xf77eab97U,
    0xe1c95365U,
    0x5985059dU,
    0x5ac1ca67U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xd0000000U,
    0x38000000U,
    0x84000000U,
    0x6a000000U,
    0xb3000000U,
    0x21800000U,
    0x2e400000U,
    0xc0a00000U,
    0x70f00000U,
    0x91980000U,
    0x4edc0000U,
    0xc0fe0000U,
    0xa1cf0000U,
    0xefb18000U,
    0x3fbc4000U,
    0xffee6000U,
    0xb6863000U,
    0xf5c02800U,
    0x45e46400U,
    0xc5d1f600U,
    0x9d2e0300U,
    0x28e68c80U,
    0xb554b440U,
    0x6deeeca0U,
    0xe1818450U,
    0x4e4144a8U,
    0x30a6a064U,
    0x98f752a6U,
    0x2d9ed345U,
    0xa0d9963cU,
    0x19fb3329U,
    0x33492494U,
    0xe0739072U,
    0xd158fa8cU,
    0x4fb9f74aU,
    0x57ef802eU,
    0x2a83400cU,
    0xcbc7e02aU,
    0xa4e67039U,
    0xd350480bU,
    0xf8ed5425U,
    0x75005e02U,
    0x2486273eU,
    0xa2c11ab1U,
    0xfe60874cU,
    0xb096483cU,
    0xc00e540bU,
    0x9857de25U,
    0x3d696739U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0xf8000000U,
    0x94000000U,
    0x2e000000U,
    0xd7000000U,
    0x86800000U,
    0x59400000U,
    0x19e00000U,
    0x28500000U,
    0x83d80000U,
    0xd79c0000U,
    0xe1fa0000U,
    0xa96b0000U,
    0x30968000U,
    0xcdfd4000U,
    0xf36be000U,
    0x59959000U,
    0xd47ba800U,
    0x40afdc00U,
    0xa3317600U,
    0x1ccce900U,
    0xbaa22380U,
    0x28b4d140U,
    0x6a0bc3a0U,
    0xc3864170U,
    0x1ac4eba8U,
    0xc323dd5cU,
    0xcb72fd96U,
    0xf8ace479U,
    0xd736161dU,
    0xa2cf390cU,
    0x25a4eb9cU,
    0xc233dd6aU,
    0x894afdb2U,
    0x2360e47cU,
    0x63141624U,
    0x9f383900U,
    0x5c486b89U,
    0x28e59d4aU,
    0xd5d79d87U,
    0xc6183446U,
    0xde3c5e2aU,
    0x79ce7505U,
    0x4920b5acU,
    0xa71a84aU,
    0x1d28481eU,
    0x96f64c2dU,
    0xcd68de35U,
    0x694351bU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x98000000U,
    0x6c000000U,
    0xd6000000U,
    0x37000000U,
    0x2d800000U,
    0x11c00000U,
    0xd3e00000U,
    0x71700000U,
    0x2c80000U,
    0x1e8c0000U,
    0xacaa0000U,
    0x26390000U,
    0x71878000U,
    0x4fc74000U,
    0x70e22000U,
    0x36f79000U,
    0x6a08a800U,
    0xbbe90400U,
    0x379cea00U,
    0xd1d7c500U,
    0xe01f7680U,
    0x8110ccc0U,
    0x997ad6a0U,
    0x84201cf0U,
    0xad905eb8U,
    0x823e88c4U,
    0x93841c9eU,
    0xa2c049dfU,
    0x6c67802bU,
    0xa5b74012U,
    0xe9aa2000U,
    0xdebb9026U,
    0xa0c2a82cU,
    0x9160040eU,
    0x41336a0aU,
    0xc6ec8512U,
    0x131f5694U,
    0x9e925cd6U,
    0xfdbffea3U,
    0x274758f1U,
    0x99a11493U,
    0x34559dd4U,
    0xc5dbe22fU,
    0x88f71126U,
    0xd90e949cU,
    0x46edde6U,
    0xe35bc20bU,
    0x3ab58115U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x18000000U,
    0x4c000000U,
    0x46000000U,
    0xf7000000U,
    0x18800000U,
    0x84400000U,
    0xbe600000U,
    0xcf500000U,
    0xc0380000U,
    0xa4fc0000U,
    0x62de0000U,
    0x2ded0000U,
    0x69828000U,
    0x4bc64000U,
    0xb2a62000U,
    0x1d771000U,
    0xe50f2800U,
    0xa1932c00U,
    0xb71b6e00U,
    0x449ad00U,
    0xba76ef80U,
    0x35886a40U,
    0xe1d24fa0U,
    0xeb793a50U,
    0x2e1b47b8U,
    0x9fcd066cU,
    0xbcb701beU,
    0x56ab877fU,
    0x2b64802fU,
    0x72d74002U,
    0xc1faa009U,
    0xc75c5033U,
    0x602b883eU,
    0x4c227c3cU,
    0xbe326608U,
    0xcbed9106U,
    0x4e82a981U,
    0xdb42eb70U,
    0x92e7ce3dU,
    0xb114fd13U,
    0xd759e786U,
    0xc82d564bU,
    0x782289b1U,
    0x4434fb75U,
    0x62ec660bU,
    0xed00912dU,
    0x180298eU,
    0x5fc4ab6dU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xb8000000U,
    0x64000000U,
    0xd6000000U,
    0x8b000000U,
    0xf1800000U,
    0x7dc00000U,
    0x9fa00000U,
    0xf300000U,
    0xf6380000U,
    0x4e7c0000U,
    0x121e0000U,
    0x45890000U,
    0x37028000U,
    0x73834000U,
    0x78c26000U,
    0xf123d000U,
    0xed740800U,
    0xfb5e4400U,
    0xf96a8a00U,
    0x1dd2c700U,
    0x93ede680U,
    0x19905b40U,
    0x4d8d06a0U,
    0x4b00cb50U,
    0x51836e98U,
    0xedc15f4cU,
    0x57a3ecb6U,
    0xd334dc55U,
    0x443e0003U,
    0x13790037U,
    0x689a801fU,
    0xc9cf4038U,
    0xd5646000U,
    0xe316d027U,
    0x81c88813U,
    0x4964042dU,
    0xb1126a22U,
    0xacce571eU,
    0x8be58e93U,
    0x5954cf5dU,
    0x98a9048aU,
    0x90730872U,
    0x51d86234U,
    0x5d291303U,
    0x53b58480U,
    0x979485cU,
    0xd9988204U,
    0x59498328U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x18000000U,
    0x3c000000U,
    0x16000000U,
    0xc5000000U,
    0x43800000U,
    0xd1c00000U,
    0x7b200000U,
    0x11300000U,
    0xacb80000U,
    0xb0fc0000U,
    0x231a0000U,
    0x7c890000U,
    0x5c78000U,
    0x39234000U,
    0x5634a000U,
    0xc83f1000U,
    0xd5bbb800U,
    0x257f6400U,
    0x57db0a00U,
    0x942ee300U,
    0xdd308280U,
    0xd2ba1b40U,
    0xe1fba2a0U,
    0x829a4b70U,
    0x5a4ebab8U,
    0xd2633f44U,
    0x695188beU,
    0xd6edf84bU,
    0x8094a03bU,
    0xe1cf103aU,
    0x2fa3b80bU,
    0x7c73643eU,
    0xeed90a0dU,
    0x61abe30eU,
    0x79f502a0U,
    0x221c5b6dU,
    0x50a828dU,
    0x7f031b73U,
    0x28842284U,
    0xdb450b7dU,
    0x37601a89U,
    0x5ad52f44U,
    0xaf2db092U,
    0xfb1dc50U,
    0xc2fb0a1fU,
    0x6c1ee31eU,
    0x5c088297U,
    0x9a861b50U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xc8000000U,
    0xbc000000U,
    0xbe000000U,
    0x15000000U,
    0x38800000U,
    0x4cc00000U,
    0x7d200000U,
    0xbf900000U,
    0x9380000U,
    0x1b7c0000U,
    0xbb9a0000U,
    0x2baf0000U,
    0xf6c68000U,
    0xc224c000U,
    0x172000U,
    0x2e7fd000U,
    0x731af800U,
    0xf6b7c00U,
    0x7e1fa00U,
    0xbb26300U,
    0xa02f7780U,
    0xb3821a40U,
    0x9146d7a0U,
    0x61650a70U,
    0x2710f88U,
    0xbd8ea67cU,
    0x4d48dbaU,
    0x2f5f7953U,
    0xd28f2015U,
    0x2353d016U,
    0xec18f83bU,
    0x40e87c35U,
    0x48a57a0cU,
    0xefd5a316U,
    0x62dcd7bcU,
    0x36ca0a7fU,
    0xeab78f9dU,
    0x5aaa6640U,
    0xf443ad86U,
    0xf1e0a96fU,
    0x62b5d82bU,
    0x86a8ac24U,
    0xda41023aU,
    0x1ce61f14U,
    0x2e300db3U,
    0xc868b943U,
    0xc64802bU,
    0x8ef7c02fU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x98000000U,
    0x4c000000U,
    0xc6000000U,
    0xd5000000U,
    0xb5800000U,
    0x70400000U,
    0x5be00000U,
    0xb2300000U,
    0x4ec80000U,
    0x790c0000U,
    0xddae0000U,
    0x6afd0000U,
    0x5ac18000U,
    0xad214000U,
    0x9112000U,
    0x6adc1000U,
    0x7a523800U,
    0x89bd5400U,
    0x21a1b600U,
    0xbad4d900U,
    0xfc7ea080U,
    0x62045ac0U,
    0x170600a0U,
    0x12850ad0U,
    0xac318b8U,
    0x15254ef4U,
    0xb517169eU,
    0xa4dc83e7U,
    0x7b56a02fU,
    0xb63c5012U,
    0x42e49822U,
    0x81b1041eU,
    0x8b8d2e2cU,
    0x769dd16U,
    0x87dd8e99U,
    0x33d087dbU,
    0x57fa0e3fU,
    0xeb44cd24U,
    0xaf6036bcU,
    0x9df193e7U,
    0xe66b1836U,
    0x685d4436U,
    0xa8958e14U,
    0xb0988d18U,
    0xc2b09689U,
    0x530cc3f2U,
    0x7ea8002cU,
    0x627c0007U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x28000000U,
    0xc4000000U,
    0x7e000000U,
    0x3d000000U,
    0x4e800000U,
    0xd6400000U,
    0x3e200000U,
    0xaf100000U,
    0xda580000U,
    0xf79c0000U,
    0x257e0000U,
    0xc40d0000U,
    0xce618000U,
    0x9034c000U,
    0xc9c8e000U,
    0xec841000U,
    0xc9477800U,
    0x2fa54c00U,
    0x8d25e00U,
    0xd3bd9f00U,
    0x796a6f80U,
    0xb5d11940U,
    0xdd3d0fa0U,
    0xcf2cc950U,
    0x1bf317b8U,
    0x5a295574U,
    0xd176d1b2U,
    0x92699645U,
    0x4257982dU,
    0x5bfd5c0aU,
    0x8a4b261eU,
    0xec45d32fU,
    0x3d21b189U,
    0xbc944660U,
    0xd2998015U,
    0x37f8c03cU,
    0x704ee03bU,
    0xcf451036U,
    0xdea0f809U,
    0xc508c32U,
    0xfefd3e21U,
    0xd8cc4f1cU,
    0x380277a9U,
    0xac058550U,
    0xda07499eU,
    0xd305ca5bU,
    0x5b833e12U,
    0x5cc14f20U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0x8000000U,
    0xa4000000U,
    0xda000000U,
    0x77000000U,
    0xf9800000U,
    0xeac00000U,
    0x6c600000U,
    0xe5b00000U,
    0xfab80000U,
    0xc6fc0000U,
    0x75de0000U,
    0xd8cf0000U,
    0x9ea38000U,
    0x65d7c000U,
    0x4108e000U,
    0x1433000U,
    0x35a20800U,
    0x1a568400U,
    0x1ac97200U,
    0x5a40100U,
    0xc253f480U,
    0x96cb1d40U,
    0xeba694a0U,
    0x6750ed50U,
    0xbc4ffca8U,
    0x22e29974U,
    0x96f466aaU,
    0xcc1f2c5dU,
    0xd72ae81cU,
    0xc6d5b400U,
    0x9a8b7a02U,
    0x8082852aU,
    0x46428691U,
    0x14231c77U,
    0x7c93600dU,
    0x98a8f03fU,
    0x4514e80eU,
    0x126ab420U,
    0x9970fa1dU,
    0x6fd9450eU,
    0xcfcc66a2U,
    0xd7232c43U,
    0x1714e839U,
    0x816ab403U,
    0x9af0fa2fU,
    0x62194503U,
    0x522c6697U,
    0x7c532c7aU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x68000000U,
    0x84000000U,
    0xda000000U,
    0x7b000000U,
    0x90800000U,
    0xc00000U,
    0x9ee00000U,
    0xad500000U,
    0x85480000U,
    0xba0c0000U,
    0xe6aa0000U,
    0x1adb0000U,
    0x9ea38000U,
    0x9b74c000U,
    0x12fee000U,
    0x58153000U,
    0xcaec6800U,
    0x52f8ac00U,
    0xb8130a00U,
    0xfaefed00U,
    0x2afcbf80U,
    0x5412a1c0U,
    0xa4ebdfa0U,
    0x8bf851f0U,
    0xbf92d7a8U,
    0x34ad0df4U,
    0x15d955b2U,
    0x8c247cd5U,
    0x1cb10804U,
    0x2a995c0bU,
    0xd0818200U,
    0xe0c2713bU,
    0xaee3ddafU,
    0xd555e0d4U,
    0x694c6a34U,
    0xe4091d2bU,
    0x47afb793U,
    0xf15cfdc7U,
    0xee3dda2U,
    0x555e0f4U,
    0x214c6a25U,
    0x70091d32U,
    0xf5afb79eU,
    0xe5cfdc7U,
    0x4463dd94U,
    0x7e95e0feU,
    0x2f2c6a09U,
    0xdd991d2dU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x48000000U,
    0xe4000000U,
    0xa6000000U,
    0xdd000000U,
    0x2f800000U,
    0xac00000U,
    0x94a00000U,
    0x1c900000U,
    0xdee80000U,
    0xbaac0000U,
    0xc14a0000U,
    0xfcbb0000U,
    0x4ee68000U,
    0x8670c000U,
    0x7699e000U,
    0xd531d000U,
    0x2df8b800U,
    0x4e808400U,
    0x5f417200U,
    0x69602700U,
    0x88b7cb80U,
    0xd43d90c0U,
    0xeca2aba0U,
    0x709780f0U,
    0xbced73a8U,
    0x71aa14dcU,
    0x7bcf59baU,
    0x3dfc67e5U,
    0x7685d813U,
    0xd3469412U,
    0x9b64aa38U,
    0xbbb6b320U,
    0xc2bb6195U,
    0x6fe723cbU,
    0x33f3ca10U,
    0x1b5ba32cU,
    0xb91039a1U,
    0x8b2d77f8U,
    0x5a8c8003U,
    0x959bc03eU,
    0xdbb76028U,
    0x92bd101eU,
    0xb7e3582cU,
    0x2ff65403U,
    0x115d4a1eU,
    0x2617632dU,
    0xdfabd99fU,
    0xa2cba7e5U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0x58000000U,
    0xc4000000U,
    0x42000000U,
    0xdb000000U,
    0x72800000U,
    0x48400000U,
    0x55a00000U,
    0xd1f00000U,
    0x8c780000U,
    0xafbc0000U,
    0x5bde0000U,
    0x5490000U,
    0x75208000U,
    0x8ab64000U,
    0xe75a6000U,
    0xd80b7000U,
    0x87062800U,
    0x54847c00U,
    0x1439e00U,
    0x9426e300U,
    0x77311480U,
    0x179e3340U,
    0xb2edf4a0U,
    0xafd60350U,
    0x1c4f3c98U,
    0x9ca64f5cU,
    0x50706a8eU,
    0xcab9e06dU,
    0x45ea81eU,
    0x1e8e3c1bU,
    0x3547fe22U,
    0x2e249319U,
    0xb837bc8fU,
    0x4f1c0f61U,
    0xbdac0a9aU,
    0xef7906bU,
    0x1cfe0024U,
    0xecf90010U,
    0x14f88013U,
    0xfa4001U,
    0x2afc6017U,
    0x6dfe7039U,
    0x9978a81fU,
    0x483b3c24U,
    0xb4197e0cU,
    0x5f2bd302U,
    0xceb55cb3U,
    0xe55d3f5cU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x98000000U,
    0xac000000U,
    0x86000000U,
    0xc1000000U,
    0x14800000U,
    0xc0c00000U,
    0x32e00000U,
    0x5100000U,
    0x79980000U,
    0x8adc0000U,
    0x7d7a0000U,
    0x16cb0000U,
    0x9c628000U,
    0xead24000U,
    0x8af92000U,
    0x41881000U,
    0x29c19800U,
    0xd260fc00U,
    0xcfd41a00U,
    0x3c7c7d00U,
    0xc24abd80U,
    0x7ca72540U,
    0xe8331da0U,
    0x17ea7570U,
    0x941325a8U,
    0x251bd97cU,
    0x6e1d078eU,
    0xcd9d0857U,
    0x60db182fU,
    0x1a7ebc27U,
    0xf34f3a0aU,
    0xd0236d27U,
    0xb4f3a58eU,
    0x3b0e996aU,
    0xfc04a796U,
    0xce005853U,
    0xe5032020U,
    0x9683102aU,
    0xb3c31804U,
    0xcd62bc3bU,
    0x96553a03U,
    0x5e386d16U,
    0x696925b4U,
    0xf3d0d95bU,
    0xd27f8795U,
    0x174f4850U,
    0x72223811U,
    0xf7f6ac2bU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x28000000U,
    0x74000000U,
    0x66000000U,
    0xaf000000U,
    0x94800000U,
    0x86c00000U,
    0x200000U,
    0x8c500000U,
    0xc9380000U,
    0x5a7c0000U,
    0x631e0000U,
    0x53290000U,
    0xfba48000U,
    0x3e11c000U,
    0x1fd86000U,
    0x4e085000U,
    0x7a72a800U,
    0xd26c4c00U,
    0xbc05a00U,
    0x35a32b00U,
    0xa5111180U,
    0xd5b0c40U,
    0xd7cbf1a0U,
    0xc6d79c50U,
    0xacfbb998U,
    0xa4db407cU,
    0x4c8dab92U,
    0x9bb1b761U,
    0x82c8280bU,
    0x4f548c0aU,
    0x27bcba0eU,
    0x81babb3bU,
    0x8ebbd9b6U,
    0x4a3f1074U,
    0x54f903aeU,
    0x8d8fb6fU,
    0x968af217U,
    0x96b3673cU,
    0xf74f4ba6U,
    0x8611275fU,
    0x53de6020U,
    0x240d5038U,
    0x5f70283bU,
    0xd3e88c1aU,
    0xa482ba0aU,
    0xeec3bb3aU,
    0x94275999U,
    0x5a52d077U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xf0000000U,
    0xa8000000U,
    0x84000000U,
    0xea000000U,
    0xb7000000U,
    0xff800000U,
    0x7a400000U,
    0xa0600000U,
    0xcd100000U,
    0xbad80000U,
    0xa61c0000U,
    0xfeba0000U,
    0xb10d0000U,
    0xbb628000U,
    0x9c93c000U,
    0x7d99e000U,
    0xbefdb000U,
    0x1e698800U,
    0x55f43400U,
    0x9a084600U,
    0xa2e55500U,
    0x2fd7a580U,
    0x277c5f40U,
    0x6d2ec5a0U,
    0xef132f70U,
    0xb9de2d98U,
    0xfb986b4cU,
    0x47fe83b2U,
    0xacea7a6dU,
    0x3f338808U,
    0x93a93400U,
    0xdbd2c617U,
    0xa57a952eU,
    0xbe2c4587U,
    0xa90ef61U,
    0xdc9fcda3U,
    0x2079db59U,
    0x1aad0b9eU,
    0x1534e64U,
    0xabb94e25U,
    0xad8fa13cU,
    0x6ca48388U,
    0x2db77a7dU,
    0x29690806U,
    0x6a76f413U,
    0x492605U,
    0xf2862509U,
    0x4ac54da6U,
    0x19a61b7eU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x50000000U,
    0x48000000U,
    0xc000000U,
    0xfa000000U,
    0x93000000U,
    0xb3800000U,
    0x89400000U,
    0x2e200000U,
    0xbe900000U,
    0x4b280000U,
    0x74ec0000U,
    0x990a0000U,
    0x87d0000U,
    0xb5278000U,
    0x2111c000U,
    0x886fa000U,
    0xd1cb7000U,
    0xd01b6800U,
    0x3c13b400U,
    0x86e91e00U,
    0x260fff00U,
    0xf1fff380U,
    0xb76266c0U,
    0xf8b5d3a0U,
    0x49b9d6f0U,
    0xddc49b98U,
    0x3ae1d2e4U,
    0xd4f4cd8aU,
    0x5d1a29e7U,
    0x9a916834U,
    0x2d2eb403U,
    0xb5ee9e38U,
    0xe58e3f11U,
    0x80b8538bU,
    0x8d4516eaU,
    0xf824bbabU,
    0x679762c4U,
    0x73aa058bU,
    0xa2fedf6U,
    0xc86c9e18U,
    0xf1cf3f1fU,
    0x601dd3afU,
    0x2415d6d0U,
    0xc2ee9bbdU,
    0xd00cd2c4U,
    0x98fb4d9bU,
    0x97e7e9f7U,
    0xc274c834U,
    0xeed8c428U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x70000000U,
    0x98000000U,
    0x14000000U,
    0x82000000U,
    0x61000000U,
    0xf1800000U,
    0x5e400000U,
    0xbe600000U,
    0x5b100000U,
    0x5d980000U,
    0x7d5c0000U,
    0x8bfa0000U,
    0x6a4f0000U,
    0xe0648000U,
    0x84144000U,
    0x631ae000U,
    0x559ff000U,
    0xe1597800U,
    0x55ff3400U,
    0x754ee600U,
    0x7ee3c500U,
    0xdcd6b580U,
    0x91b9c540U,
    0x7e2ad5a0U,
    0xf6717570U,
    0xce0fcd88U,
    0x3b06f16cU,
    0xe48433aeU,
    0xfdc2b069U,
    0x6921780eU,
    0x98f33431U,
    0xcecce612U,
    0x7da0c516U,
    0x3d303595U,
    0x13ee855cU,
    0x1756b584U,
    0xaf9c576U,
    0x8bcad5aeU,
    0xf6217568U,
    0xc677cd8cU,
    0x360af15bU,
    0xff06339bU,
    0x8e81b040U,
    0x10c7f808U,
    0xea4741dU,
    0x25b08638U,
    0xe028753eU,
    0x89752d88U,
    0x80890178U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x58000000U,
    0x54000000U,
    0x52000000U,
    0x71000000U,
    0x8a800000U,
    0x20c00000U,
    0xb8600000U,
    0x41300000U,
    0x49f80000U,
    0x1cbc0000U,
    0xc19e0000U,
    0x558b0000U,
    0x24638000U,
    0x5f354000U,
    0x9efde000U,
    0xa53e7000U,
    0x42df7800U,
    0x13acec00U,
    0xaff7ca00U,
    0x9e9e5900U,
    0xb7880U,
    0x1120f340U,
    0x721318a0U,
    0xe2ecc350U,
    0xc8d40088U,
    0x678c1f54U,
    0xa564d29eU,
    0x3db29a41U,
    0xa23f780aU,
    0x435cec1dU,
    0x74efca08U,
    0xf3d25905U,
    0x680d788eU,
    0x4d27f351U,
    0x8c1698aaU,
    0x85ee8353U,
    0x395460a6U,
    0xe8cc2f5eU,
    0xe545ca9bU,
    0x95254654U,
    0x98122a06U,
    0x77ec2931U,
    0x58520086U,
    0x3a4b1f6aU,
    0x918152bbU,
    0x7f40da5cU,
    0xa827181eU,
    0xb490dc19U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x10000000U,
    0x78000000U,
    0x44000000U,
    0xca000000U,
    0x45000000U,
    0xdd800000U,
    0xd8400000U,
    0x66200000U,
    0x58d00000U,
    0xa3780000U,
    0x555c0000U,
    0xb58a0000U,
    0x32f10000U,
    0xbdad8000U,
    0x35244000U,
    0x9e576000U,
    0x31bcb000U,
    0x27bad800U,
    0x3cbb5400U,
    0x76392600U,
    0x62fccf00U,
    0xb31f5d80U,
    0x54aa6040U,
    0xdea585a0U,
    0xed113470U,
    0x151ca3b8U,
    0x47adfb74U,
    0xb823fe26U,
    0xfd79b25U,
    0xffe7bb4U,
    0xf29aaf44U,
    0x72e8d83bU,
    0x5f865406U,
    0x1146a604U,
    0x3da58f35U,
    0x2397bd98U,
    0xabdf9064U,
    0xc34fbd8eU,
    0xfa13906eU,
    0x199dbda5U,
    0x906e9056U,
    0x17423d9aU,
    0x5ea7d07eU,
    0x2d12ddbdU,
    0x751e206fU,
    0x57aae5bdU,
    0xc0218455U,
    0x4bd47badU,
    0xc5fbaf5bU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x18000000U,
    0xdc000000U,
    0xbe000000U,
    0x63000000U,
    0x25800000U,
    0x30c00000U,
    0xf1200000U,
    0xf5b00000U,
    0xb6f80000U,
    0xafdc0000U,
    0xbe6e0000U,
    0x3a950000U,
    0x60488000U,
    0x5ca64000U,
    0x1175a000U,
    0x6dd8d000U,
    0x736d8800U,
    0x34110400U,
    0x810c5200U,
    0xe7445300U,
    0x14e1fa80U,
    0xd6119f40U,
    0x1c0c72a0U,
    0xc1c09b50U,
    0x31a020b8U,
    0xff4c85cU,
    0xc499da3eU,
    0x3c89572bU,
    0x4b83a8b1U,
    0x3bc0cc63U,
    0x70a50800U,
    0x97774433U,
    0x82d9f202U,
    0xc0ec8325U,
    0xa3d4729eU,
    0x37ac9b6dU,
    0xff3620b1U,
    0x15bdc850U,
    0x993f5a2bU,
    0xd57a1722U,
    0x501e8891U,
    0xacce5c44U,
    0x70e52024U,
    0xa4129015U,
    0x890e2829U,
    0x7340d439U,
    0x5ee75a12U,
    0xcf161710U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x98000000U,
    0x5c000000U,
    0xba000000U,
    0xd5000000U,
    0x78800000U,
    0x89c00000U,
    0xa3600000U,
    0x24900000U,
    0xf1880000U,
    0x6eec0000U,
    0x207a0000U,
    0x8cf30000U,
    0x6e9e8000U,
    0xb9204000U,
    0xa5b1a000U,
    0xf83af000U,
    0x94d41800U,
    0xdaa7c00U,
    0xdbdec200U,
    0xb584c100U,
    0x6d451980U,
    0xb0a217c0U,
    0x62f101a0U,
    0x35986bf0U,
    0x4aa7c3a8U,
    0x57f0aafcU,
    0x5d18da0eU,
    0x6b61bd2bU,
    0x30975b80U,
    0x9f8996eeU,
    0xf5e93836U,
    0xf3ffcc30U,
    0x4eb7fa10U,
    0x53bb0d12U,
    0x1a92e3acU,
    0x22891ae7U,
    0x696be212U,
    0xac3d7134U,
    0x1ad621b9U,
    0x86aedbf8U,
    0xa0587b94U,
    0xb3c326ddU,
    0xb664802dU,
    0x7c134001U,
    0x484f2002U,
    0x558ab00aU,
    0x58edb833U,
    0xc77c8c07U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0xb8000000U,
    0x3c000000U,
    0x6a000000U,
    0xf9000000U,
    0xe9800000U,
    0x81c00000U,
    0xbe00000U,
    0x2e500000U,
    0xcc80000U,
    0xed2c0000U,
    0xfa7a0000U,
    0x3f330000U,
    0xe3da8000U,
    0xaa454000U,
    0x8727a000U,
    0x2367000U,
    0xbc58f800U,
    0x28821400U,
    0x94436a00U,
    0xc0214d00U,
    0x6cb76b80U,
    0x731bc7c0U,
    0x5c2793a0U,
    0x76b5d3f0U,
    0x121ef9a8U,
    0x39a79edcU,
    0x25731206U,
    0xdcf91913U,
    0x947321bcU,
    0x517abaddU,
    0xd3b5202eU,
    0xc19f3021U,
    0x6965582aU,
    0xf3176439U,
    0x9d691215U,
    0x3c5a191fU,
    0xe881a1a5U,
    0xb443fac1U,
    0x7020801fU,
    0xd4b6400eU,
    0x4f1d202dU,
    0x36233002U,
    0x8fb7581bU,
    0xfb986414U,
    0xb861920fU,
    0x2e90590bU,
    0xf2ae8193U,
    0x98bfcaecU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xe8000000U,
    0x94000000U,
    0x56000000U,
    0xcb000000U,
    0x74800000U,
    0xdb400000U,
    0x9a00000U,
    0x78d00000U,
    0x55480000U,
    0x5aec0000U,
    0xb13e0000U,
    0xccf70000U,
    0x7f5c8000U,
    0xb2c5c000U,
    0x2d666000U,
    0xfeb05000U,
    0xf7fc800U,
    0x96d40c00U,
    0x524d8600U,
    0x246a8300U,
    0x937a8980U,
    0xc4d2c5c0U,
    0xf74d41a0U,
    0x17eac9d0U,
    0x16bec7a8U,
    0x1f374aecU,
    0x6fb8ce12U,
    0x5eb04f09U,
    0xbf7befa7U,
    0x5ed6d6ebU,
    0x364ce012U,
    0x9a6e9031U,
    0xcc7b2802U,
    0xe6569c26U,
    0xe708ae30U,
    0x6acb1f1bU,
    0xb52ea7aeU,
    0x43dc1ac9U,
    0x4d85861bU,
    0xbac6832fU,
    0x296489b2U,
    0x90b5c5c2U,
    0x4879c198U,
    0xc85309d5U,
    0x800ea78dU,
    0xc44c1afeU,
    0x8f6d8623U,
    0x37fa830eU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xd8000000U,
    0x9c000000U,
    0x56000000U,
    0xa1000000U,
    0x7f800000U,
    0x3c00000U,
    0xb6e00000U,
    0x34100000U,
    0xf780000U,
    0x279c0000U,
    0x668a0000U,
    0x54750000U,
    0x732c8000U,
    0x6fc04000U,
    0x18e1a000U,
    0x59179000U,
    0x4efed800U,
    0xc15e8400U,
    0x65e80a00U,
    0x94a1e500U,
    0x14358180U,
    0xbd040U,
    0xc43359a0U,
    0xa8095470U,
    0x60315398U,
    0xd20db15cU,
    0x4f30522eU,
    0x908a213bU,
    0x6570ab9bU,
    0x84ade563U,
    0x9801a01cU,
    0xbc07903fU,
    0x606d81fU,
    0xc9028409U,
    0x3b820a07U,
    0xc9c4e523U,
    0x41e101a2U,
    0xea979079U,
    0x7338f9b9U,
    0x92bbc477U,
    0xe47b0bb6U,
    0x6f1f7549U,
    0x5bcbf824U,
    0x2ed5540fU,
    0x2a1d720eU,
    0x7e4df135U,
    0x5217d3afU,
    0xb678f141U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x18000000U,
    0xd4000000U,
    0xa2000000U,
    0x43000000U,
    0xbb800000U,
    0xcdc00000U,
    0xede00000U,
    0xbf500000U,
    0xc580000U,
    0xa7bc0000U,
    0xa1ee0000U,
    0x61310000U,
    0xa2c88000U,
    0x78064000U,
    0x2401e000U,
    0x5a047000U,
    0x27042800U,
    0x1821400U,
    0x5ac70a00U,
    0xf461cd00U,
    0x3196e380U,
    0x5a3f2b40U,
    0xd52acba0U,
    0x40513f50U,
    0x79dbc198U,
    0xf7df27cU,
    0xbe8ba21eU,
    0x27259909U,
    0x43300996U,
    0xa1ca9649U,
    0x23800020U,
    0x59c00006U,
    0xafe00005U,
    0x4c500020U,
    0xafd80010U,
    0xbe7c0000U,
    0xee0e0038U,
    0x9d61000eU,
    0x1510802dU,
    0x127a401cU,
    0x680fe032U,
    0x8465703fU,
    0x8994a82bU,
    0xde385421U,
    0xdf28ea3eU,
    0xcf54bd33U,
    0xb45a4bbbU,
    0x23bb7f75U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x78000000U,
    0x94000000U,
    0xba000000U,
    0xe9000000U,
    0x7b800000U,
    0x7d400000U,
    0xa5200000U,
    0xccb00000U,
    0xa9080000U,
    0xba2c0000U,
    0xc19e0000U,
    0x8170000U,
    0xa1fb8000U,
    0x95864000U,
    0x4e45e000U,
    0x7a3b000U,
    0xa276b800U,
    0x9d6a8400U,
    0xfdbfde00U,
    0xf726cf00U,
    0x89b4ee80U,
    0xdc8c58c0U,
    0x846a56a0U,
    0x8e3adcd0U,
    0x566388a8U,
    0x6a9713d4U,
    0xaf3ae636U,
    0x91e60b31U,
    0x61d0d0b8U,
    0xdd1e27dbU,
    0xdd3800eU,
    0xb31a4010U,
    0x7ed3e011U,
    0x3198b012U,
    0x133803U,
    0x7dfbc427U,
    0xd381be1dU,
    0xf1433f25U,
    0xbb27b6acU,
    0xf7b56cd6U,
    0xd78b3082U,
    0x92ea97c5U,
    0x4afeb811U,
    0xf3068439U,
    0xc281de22U,
    0x5ec1cf38U,
    0x5c676eadU,
    0xab9618d7U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x38000000U,
    0x2c000000U,
    0xa000000U,
    0x69000000U,
    0x45800000U,
    0xf5c00000U,
    0x9ae00000U,
    0x21300000U,
    0x79480000U,
    0xcdac0000U,
    0x3b9a0000U,
    0x58510000U,
    0xebe8000U,
    0x62854000U,
    0x17452000U,
    0x4da3f000U,
    0xc91f800U,
    0x5d8f400U,
    0xf073aa00U,
    0xe7ef0500U,
    0xb5be3880U,
    0xd205b8c0U,
    0xf507c0a0U,
    0x17814cf0U,
    0xc0c66ab8U,
    0xed6349ecU,
    0x91f4d236U,
    0xac2eb127U,
    0x705ab294U,
    0x9db44dcdU,
    0x410c803dU,
    0xd688401cU,
    0xd649a02dU,
    0x5b2bb033U,
    0xbad85827U,
    0x2ef3440dU,
    0x72abf234U,
    0xe51c4137U,
    0xcd15ca82U,
    0x5e19f9c0U,
    0x7d920a0eU,
    0xbc58b525U,
    0x27b46083U,
    0x700bfce5U,
    0xcf08b29aU,
    0x11894dc6U,
    0x84c8002bU,
    0xd46c0022U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x30000000U,
    0x98000000U,
    0xc4000000U,
    0xd6000000U,
    0xe3000000U,
    0xb4800000U,
    0x19400000U,
    0x2c600000U,
    0xa5900000U,
    0xe6b80000U,
    0x49dc0000U,
    0x88ce0000U,
    0x7f330000U,
    0x768e8000U,
    0xd1d44000U,
    0xb9596000U,
    0x78d5000U,
    0x10500800U,
    0x579fcc00U,
    0xb62cd600U,
    0x18e1fd00U,
    0xfcd67d80U,
    0x6adace40U,
    0xdc4875a0U,
    0xf6760250U,
    0xf2ea2388U,
    0x2843bf5cU,
    0x4de53e2eU,
    0x7b54212dU,
    0xcf1d43bbU,
    0x856def45U,
    0xf583b61eU,
    0x40c3ad1fU,
    0x6a6f5bbU,
    0xc232425bU,
    0xd0b4399U,
    0x5612ef5cU,
    0x4d7b362dU,
    0x97f8ed00U,
    0xd9bf15a5U,
    0xaf585242U,
    0x48caba9U,
    0x34d4335aU,
    0xe6de881eU,
    0xc64b8c10U,
    0xaf75b60dU,
    0x2f6cad1fU,
    0x748675aaU,
    0xf945024fU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0xc8000000U,
    0xa4000000U,
    0xca000000U,
    0x79000000U,
    0x68800000U,
    0x49400000U,
    0xf2200000U,
    0x8d300000U,
    0xada80000U,
    0x2e8c0000U,
    0xcf3e0000U,
    0x51d10000U,
    0x6c7f8000U,
    0x2af7c000U,
    0xf1cbe000U,
    0xf21e3000U,
    0x8b643800U,
    0xa7159c00U,
    0xac9a6600U,
    0xbd200f00U,
    0x9ab31a80U,
    0x2b69a1c0U,
    0xd66922a0U,
    0x84ed3dd0U,
    0xccacc498U,
    0x6a0af2c4U,
    0x87c3e2aU,
    0x40f1632dU,
    0x38cf24b8U,
    0x298c2e3U,
    0xae26063aU,
    0x3b35ff35U,
    0x92aac2acU,
    0x10f0dfbU,
    0x95fefc9fU,
    0x5f326ec3U,
    0xf8afd80eU,
    0xc80bac26U,
    0x657e5e13U,
    0x7a75933aU,
    0x64897ca8U,
    0xf639aef6U,
    0xd9523826U,
    0x75389c08U,
    0x10d3e618U,
    0xd8facf0bU,
    0x95b17a91U,
    0xdced51c7U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x38000000U,
    0x4000000U,
    0x12000000U,
    0xd000000U,
    0xe7800000U,
    0xfec00000U,
    0x7c600000U,
    0xc5700000U,
    0x3580000U,
    0xba3c0000U,
    0xe4ce0000U,
    0x6f550000U,
    0xc60a8000U,
    0x35354000U,
    0x44786000U,
    0xa8ee7000U,
    0xcd055800U,
    0xc7870400U,
    0xeec00200U,
    0x3461f900U,
    0xf973b680U,
    0x155a5e40U,
    0xa538eea0U,
    0xe485a50U,
    0x76126cb8U,
    0x44ace34cU,
    0x8c21ba36U,
    0x8254cd0dU,
    0x118a0c92U,
    0x93f2935eU,
    0x4c1ce213U,
    0x479fc92dU,
    0xc75c0ebaU,
    0x883a6a51U,
    0xf9cbd488U,
    0xc0d5d771U,
    0x4ce001aU,
    0x5f55000bU,
    0x9e0a8027U,
    0x4135401aU,
    0x6e78600cU,
    0xa1ee700fU,
    0x3885582eU,
    0x34470404U,
    0x75200206U,
    0xfd1f903U,
    0x864bb698U,
    0x6a165e6bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0x28000000U,
    0x8c000000U,
    0xfe000000U,
    0x61000000U,
    0x82800000U,
    0xc5400000U,
    0x64e00000U,
    0x32700000U,
    0xc3880000U,
    0x6a6c0000U,
    0x549a0000U,
    0xbb570000U,
    0xb5e8000U,
    0x7ef5c000U,
    0x8fcea000U,
    0xa00bd000U,
    0xedaf1800U,
    0x553eac00U,
    0x5dc7d600U,
    0xd2a12900U,
    0x5b123680U,
    0x4f3fecc0U,
    0x2ec72ea0U,
    0xdf2640f0U,
    0xe5d67888U,
    0xc81ea9dcU,
    0xcf10ee02U,
    0xd53d953fU,
    0x9dc6589dU,
    0x72a0b9d7U,
    0xeb115635U,
    0x6738e93aU,
    0xa2c696aeU,
    0x21233cc9U,
    0x84d6b6a3U,
    0x4a9d2cf4U,
    0xa570e8bU,
    0xb1d850fcU,
    0xafb7c085U,
    0xb12bd5e5U,
    0x81782010U,
    0x33a21032U,
    0x1993b817U,
    0x2a7e7c1eU,
    0xfa244e16U,
    0xc5514521U,
    0xaa5fc08eU,
    0x5c77d5d2U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x58000000U,
    0x34000000U,
    0x5a000000U,
    0x5b000000U,
    0xd0800000U,
    0x19400000U,
    0xf6a00000U,
    0xe3d00000U,
    0x41d80000U,
    0xb07c0000U,
    0xa52e0000U,
    0x70b10000U,
    0x83ef8000U,
    0x55d34000U,
    0x34dbe000U,
    0x81fab000U,
    0x67ed2800U,
    0x17d54c00U,
    0xbbdab600U,
    0x5b786b00U,
    0x7dad5980U,
    0x5f49140U,
    0x1b4e71a0U,
    0xb700dd50U,
    0xfe834798U,
    0xf847f644U,
    0x2d23fe0eU,
    0x7214d709U,
    0x9a3927b8U,
    0xff0f064fU,
    0x5c22b614U,
    0x51946b0dU,
    0x8ffb59a7U,
    0x76e9917cU,
    0x2457f1afU,
    0x161e9d4fU,
    0xb699278aU,
    0x7df065aU,
    0x2d7ab623U,
    0xa8a86b2aU,
    0x84755981U,
    0xd188915eU,
    0xbc6071b0U,
    0xa8b1dd51U,
    0xf7ecc794U,
    0xefd4b643U,
    0x3fd81e1fU,
    0x97e671cU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x28000000U,
    0xfc000000U,
    0x72000000U,
    0x31000000U,
    0x65800000U,
    0x27400000U,
    0xe600000U,
    0x57100000U,
    0xfda80000U,
    0x40cc0000U,
    0x2b5e0000U,
    0x8350000U,
    0x63da8000U,
    0x8f5c000U,
    0x33ff2000U,
    0xb8023000U,
    0xe4047800U,
    0xc6037400U,
    0xcf07fe00U,
    0xe80b500U,
    0x8fc59980U,
    0x3ea59ac0U,
    0x4f37e1a0U,
    0xc15feed0U,
    0xcd349fb8U,
    0x985f9bccU,
    0x34b4a632U,
    0xbd1df133U,
    0x23d01f8eU,
    0x738f5befU,
    0x8b39062fU,
    0xc26012aU,
    0x92f5479eU,
    0xdefb1fccU,
    0xbf80000fU,
    0xaa400023U,
    0x79e00032U,
    0x31500017U,
    0xbe48003eU,
    0xcc9c0026U,
    0xaa960001U,
    0x2ee90037U,
    0xd0ac800bU,
    0x674cc004U,
    0x751ba00fU,
    0xefd2f00bU,
    0x4989d80eU,
    0x3638842fU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0x18000000U,
    0xac000000U,
    0x1e000000U,
    0x9b000000U,
    0x3f800000U,
    0x7a400000U,
    0x91a00000U,
    0xe7500000U,
    0x98180000U,
    0x84bc0000U,
    0x136a0000U,
    0x8cb70000U,
    0x2ee98000U,
    0x5bf34000U,
    0x5fc92000U,
    0x3645000U,
    0x1d35c800U,
    0xe7a81400U,
    0xb1d51600U,
    0x38d89300U,
    0x3bd8dd80U,
    0x685efb40U,
    0xac1f15a0U,
    0xf6bdef70U,
    0x36698388U,
    0x78313c64U,
    0x7329fe0eU,
    0x1c14d70fU,
    0x373803a9U,
    0xe72e7c59U,
    0x3e12de2eU,
    0xa3b8723U,
    0xbfae4b84U,
    0x7dd22855U,
    0xd6df682bU,
    0xe8d8041dU,
    0xe3d87e29U,
    0x645b9709U,
    0xe21b23a0U,
    0x75bd2c61U,
    0xa5ee961bU,
    0x1c70d31aU,
    0x798a7d9aU,
    0xc4c2eb44U,
    0xd5607dbaU,
    0xf235eb76U,
    0xca29fd87U,
    0x1e96ab58U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xe8000000U,
    0x74000000U,
    0x8a000000U,
    0xd7000000U,
    0xcf800000U,
    0x5bc00000U,
    0x2e600000U,
    0x76d00000U,
    0xbdd80000U,
    0x9abc0000U,
    0xaeee0000U,
    0x28710000U,
    0xe4ee8000U,
    0x5f77c000U,
    0xbb6d6000U,
    0x9cb3b000U,
    0x908c800U,
    0x1460d400U,
    0xe9d03a00U,
    0x9659c100U,
    0xd37a8380U,
    0xcb8c4740U,
    0x6f244ba0U,
    0x5ff19350U,
    0xf32cf198U,
    0xf712924cU,
    0x7cbb9222U,
    0x2feba511U,
    0x2ef4f1a0U,
    0x5dae9274U,
    0xda55920bU,
    0x39aa52eU,
    0xa81a71a2U,
    0xa1d95262U,
    0x24b8f20dU,
    0x13e91520U,
    0x40f2b995U,
    0x98a98658U,
    0x5ed0c814U,
    0x69dcd420U,
    0x80be3a2dU,
    0xe1e8c10eU,
    0x7bf403b3U,
    0x412b8758U,
    0x2c112bb4U,
    0xd53e2375U,
    0xb52a39b1U,
    0xe6134679U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0x58000000U,
    0x84000000U,
    0xb2000000U,
    0x95000000U,
    0xd8800000U,
    0xb0400000U,
    0xb3a00000U,
    0x63d00000U,
    0xfc80000U,
    0xf76c0000U,
    0xf73e0000U,
    0x13370000U,
    0x283a8000U,
    0xaab64000U,
    0x8afd6000U,
    0x88567000U,
    0x680e9800U,
    0x45083c00U,
    0x498c4e00U,
    0x3ce8b00U,
    0xc96d0980U,
    0x1c39e9c0U,
    0xb591a0U,
    0xbbfad5d0U,
    0xd2d55fa8U,
    0x15491edcU,
    0xaa29b63eU,
    0x281bc73dU,
    0x2aa35fb2U,
    0x85521ec8U,
    0x548d362cU,
    0x574a8738U,
    0x72cbfbeU,
    0x649e2eddU,
    0x40e0ce2fU,
    0x7ff3cb38U,
    0xd9dce9b6U,
    0xcfc2d9c8U,
    0xe62e9b4U,
    0x98b5d9ceU,
    0x5ff869bdU,
    0x70d399edU,
    0xe84d09b9U,
    0xaea9e9f5U,
    0xae5d91a9U,
    0xbe06d5f7U,
    0xab035f97U,
    0x33821ef5U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x68000000U,
    0x94000000U,
    0x26000000U,
    0x77000000U,
    0x2800000U,
    0x91400000U,
    0xfea00000U,
    0x66100000U,
    0x41980000U,
    0x8dbc0000U,
    0xc2ea0000U,
    0x21d70000U,
    0x687b8000U,
    0xd20ac000U,
    0x9fe56000U,
    0x80b5d000U,
    0x6b8ba800U,
    0x6255c00U,
    0xe256c600U,
    0x24bbfb00U,
    0x8b6e4780U,
    0xa0142540U,
    0x469e6fa0U,
    0x573bb970U,
    0xbfadc988U,
    0x5759274U,
    0xcb682612U,
    0xc014eb11U,
    0xf6988fbbU,
    0x2f38a97cU,
    0x43a901b3U,
    0xb7721e6cU,
    0x9a6ec83eU,
    0xb5908c14U,
    0x655d6e1dU,
    0x40dea73fU,
    0xdb18818cU,
    0x90ffde4aU,
    0x56482813U,
    0xfac39c16U,
    0x8661a61aU,
    0x9752b25U,
    0x616c6f92U,
    0xdd10b94aU,
    0x491c498cU,
    0x11f8526eU,
    0xebcec622U,
    0xed07fb0cU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xd0000000U,
    0x28000000U,
    0xc4000000U,
    0xce000000U,
    0xf5000000U,
    0x6b800000U,
    0x80c00000U,
    0x10a00000U,
    0x7d700000U,
    0x3c680000U,
    0x1f4c0000U,
    0xfefa0000U,
    0xc0350000U,
    0xa50f8000U,
    0xdb1cc000U,
    0x1023e000U,
    0x8b301000U,
    0xe38e7800U,
    0xec594400U,
    0x72459200U,
    0x31e68f00U,
    0x36173080U,
    0x1f39b9c0U,
    0x9d16c8a0U,
    0x89bc3df0U,
    0x8250ba98U,
    0x2fdaa2c4U,
    0xf681f212U,
    0xf465f35U,
    0x6e60a881U,
    0x60d5edf8U,
    0xc69aa283U,
    0xa9e336dcU,
    0xca13f835U,
    0xdd3c8406U,
    0xb213f223U,
    0xf13f5f3dU,
    0x5815288dU,
    0x1a3c2de4U,
    0xee96c290U,
    0x357fe6ecU,
    0xb0f66014U,
    0xada9d01fU,
    0x9a6a182cU,
    0xe499432U,
    0x6b7a0a29U,
    0x4df6db2eU,
    0x32295a96U,
    0x2cafb2ebU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x8000000U,
    0xc000000U,
    0x32000000U,
    0xc5000000U,
    0x7e800000U,
    0x60c00000U,
    0xd2600000U,
    0x8db00000U,
    0x6e80000U,
    0x60c0000U,
    0x897a0000U,
    0xb0f30000U,
    0xd9cb8000U,
    0xc99c000U,
    0xa106a000U,
    0x3086b000U,
    0x2fc25800U,
    0xfde08c00U,
    0xa5f6fa00U,
    0x4f4f9100U,
    0xd05a4080U,
    0xd96515c0U,
    0x7c3398a0U,
    0x69ac59f0U,
    0x6c2bc2a8U,
    0x456978d4U,
    0x6c49da2aU,
    0xbddfe117U,
    0x8c273886U,
    0xdf15e9d6U,
    0xe9381ab1U,
    0x155334daU,
    0x8d9a000fU,
    0x183001eU,
    0x17438020U,
    0xde25c01dU,
    0x6a14a019U,
    0x2fb9b03cU,
    0xe193d806U,
    0x69fa4c07U,
    0x7733da17U,
    0x982ce118U,
    0x36cb8a0U,
    0x2f4c29d2U,
    0xa05eba89U,
    0x616584fdU,
    0xe830582aU,
    0x5faf8c21U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0xa8000000U,
    0x24000000U,
    0x92000000U,
    0xc9000000U,
    0xca800000U,
    0xf9c00000U,
    0x36a00000U,
    0xf2500000U,
    0xbce80000U,
    0x4ccc0000U,
    0x315a0000U,
    0x40170000U,
    0xd0c8000U,
    0x1bfbc000U,
    0xfc41e000U,
    0x96e3f000U,
    0x4ab17800U,
    0x815e4c00U,
    0x8817b200U,
    0x790b9f00U,
    0xd1fa3180U,
    0xb9465cc0U,
    0xea67c9a0U,
    0xe873d0f0U,
    0xb4799b88U,
    0x4907bfccU,
    0xa80d222U,
    0x59c4af05U,
    0xc6a629a6U,
    0x5a5020c7U,
    0x98e8e39cU,
    0xdec9f3f5U,
    0xf85f6013U,
    0x8a93303cU,
    0xf4ce1812U,
    0x2d5d7c24U,
    0xe11aa20U,
    0x2a0ae304U,
    0x6799bb6U,
    0xb007bfc9U,
    0xc800d22fU,
    0x7404af01U,
    0xca0629a4U,
    0x450020d0U,
    0x7c80e3a6U,
    0xa2c5f3e1U,
    0x35256027U,
    0xc1143038U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0xf8000000U,
    0xdc000000U,
    0x32000000U,
    0x53000000U,
    0xd2800000U,
    0x62c00000U,
    0xe1e00000U,
    0x85100000U,
    0xf3a80000U,
    0x44cc0000U,
    0x7d1e0000U,
    0x86550000U,
    0xf2098000U,
    0xdaf84000U,
    0x2647a000U,
    0x7e21b000U,
    0xb7766800U,
    0x85fadc00U,
    0x1ec23a00U,
    0xe3e7fd00U,
    0x5e103980U,
    0x152836c0U,
    0x300fd1a0U,
    0x21faaad0U,
    0xb0c24bb8U,
    0xe2e0e7d4U,
    0xcf921a06U,
    0x5d6b0d23U,
    0x6f2871aaU,
    0x770b1ae7U,
    0x957c23a6U,
    0x87063bebU,
    0x94862038U,
    0x47c5f039U,
    0x9e67c816U,
    0xac526c1bU,
    0x2d0bd229U,
    0xa27c6132U,
    0xdb8223a7U,
    0x87433bcaU,
    0xdfa7a006U,
    0x7731b021U,
    0xeede6804U,
    0x5e36dc3cU,
    0x7b5c3a3eU,
    0x8872fd01U,
    0xad79b992U,
    0x7b0076dfU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0xb8000000U,
    0x54000000U,
    0xf2000000U,
    0xb7000000U,
    0x3e800000U,
    0x56c00000U,
    0x11e00000U,
    0xe6500000U,
    0x86e80000U,
    0xbb8c0000U,
    0xec1e0000U,
    0x9d950000U,
    0xd2898000U,
    0xe39a4000U,
    0xc0d1e000U,
    0xb429f000U,
    0x906f0800U,
    0x41489c00U,
    0x3bfc9a00U,
    0x2ec37d00U,
    0x65e36380U,
    0x645326c0U,
    0x99edeba0U,
    0x690dfad0U,
    0x1cde91b8U,
    0xc97277d4U,
    0xbd5b7a16U,
    0xd338d01U,
    0x3c7beb8aU,
    0xaf84fafcU,
    0x6d4111bdU,
    0xaea137fdU,
    0x8ff51a0fU,
    0x9f993d15U,
    0x9ed2839cU,
    0xc52ad6d3U,
    0xfbeae3acU,
    0x260966d8U,
    0x165c0ba5U,
    0x7db40addU,
    0xa3b999a2U,
    0x81e6ebdfU,
    0x1e51e02fU,
    0xb2e9f03eU,
    0x598f0805U,
    0xe3189c16U,
    0xf7149a25U,
    0x764f7d09U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0x88000000U,
    0xb4000000U,
    0x22000000U,
    0xdf000000U,
    0x85800000U,
    0xc6c00000U,
    0xe2e00000U,
    0x51d00000U,
    0x4d80000U,
    0x51bc0000U,
    0x1aa0000U,
    0xe110000U,
    0xe6bc8000U,
    0xc02ac000U,
    0x2d26000U,
    0x5f5df000U,
    0xbe7cc800U,
    0xabcc0400U,
    0x4b840200U,
    0xf7c53d00U,
    0xc6628780U,
    0x6b970b40U,
    0xd07acfa0U,
    0xcacdcf70U,
    0xd706adb8U,
    0xf184027cU,
    0x84c4621aU,
    0xde5cd11U,
    0x5c50cfb2U,
    0x761ccf56U,
    0x915a2dabU,
    0x8f7ec25bU,
    0x8f4e0229U,
    0x71c43d2eU,
    0x236607a2U,
    0x5d11cb50U,
    0xbd3aafb3U,
    0x2fed3f69U,
    0xa8b4e5a9U,
    0x1acfc65dU,
    0xaf04803aU,
    0xad86c010U,
    0x22c0601dU,
    0x78e0f01bU,
    0xb2d24837U,
    0x175bc43cU,
    0x6a78e224U,
    0xb9cf0d2cU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0x88000000U,
    0xec000000U,
    0xa000000U,
    0x85000000U,
    0x7800000U,
    0xb2c00000U,
    0xbbe00000U,
    0x9900000U,
    0x5a380000U,
    0x345c0000U,
    0xde0e0000U,
    0x76570000U,
    0x6a5c8000U,
    0x550ec000U,
    0xc2d2a000U,
    0x28187000U,
    0xbaa99800U,
    0x23e18400U,
    0xfd91a200U,
    0xe439c500U,
    0x2f5bbd80U,
    0x328e7640U,
    0xa016a5a0U,
    0xebfd3250U,
    0xd73ba7a8U,
    0x9fdb876cU,
    0x46cd020aU,
    0xb8b6b517U,
    0xec4ea5a8U,
    0x51f13266U,
    0x476da781U,
    0x7a808766U,
    0x91478205U,
    0xcf237523U,
    0xcd768593U,
    0x592c8264U,
    0x37241f8fU,
    0xe970b358U,
    0x9f299810U,
    0x48218424U,
    0xe3f1a218U,
    0xc669c52bU,
    0x4b03bd9cU,
    0xd4827642U,
    0x9240a5a3U,
    0x97a63276U,
    0x853127b0U,
    0x338e4777U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0xd8000000U,
    0x34000000U,
    0xee000000U,
    0x4d000000U,
    0x79800000U,
    0xf8c00000U,
    0xe7600000U,
    0xe8700000U,
    0x92b80000U,
    0x665c0000U,
    0x5dee0000U,
    0xd0b70000U,
    0x15db8000U,
    0xbe2b4000U,
    0x751e000U,
    0x5aeb1000U,
    0x92313800U,
    0xcd9a9400U,
    0x260c0200U,
    0x65836100U,
    0x8ac76680U,
    0x98656940U,
    0xcef5dea0U,
    0x5cf8bd50U,
    0x977e3ca8U,
    0x70bbcc54U,
    0x815de21eU,
    0x6f687119U,
    0xe0f65e85U,
    0x71fffd5aU,
    0xdef9dcb3U,
    0x407bdc70U,
    0x8a395a34U,
    0x5d1ea503U,
    0xd1483cbdU,
    0x2320cc54U,
    0x2506227U,
    0x8f683119U,
    0x90f23e9bU,
    0x99f8ad6bU,
    0x22fa84a2U,
    0x427d1854U,
    0x1d3b802eU,
    0x879b402eU,
    0x1d09e03fU,
    0x45071026U,
    0xf5873829U,
    0x12c19433U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x50000000U,
    0xf8000000U,
    0xc4000000U,
    0x56000000U,
    0xa5000000U,
    0x18800000U,
    0x98400000U,
    0x9e200000U,
    0x47500000U,
    0xfc80000U,
    0xa36c0000U,
    0xab7a0000U,
    0xa1130000U,
    0xc8ed8000U,
    0x4eb84000U,
    0x9ff16000U,
    0xd9dc5000U,
    0x60063800U,
    0xb0041c00U,
    0x8801ce00U,
    0x6c06e700U,
    0x6a025180U,
    0x370579c0U,
    0xeb81e9a0U,
    0x25c525f0U,
    0x1ee347a8U,
    0x413092e4U,
    0xd6b8ae16U,
    0xebf6b715U,
    0x7de69a3U,
    0xa90265ccU,
    0xc285a78dU,
    0x274782ceU,
    0x19a2761cU,
    0x896bb3aU,
    0x2628ff94U,
    0x9dccefaU,
    0x58000005U,
    0x54000039U,
    0x8e000000U,
    0x31000008U,
    0xb6800020U,
    0xf940000bU,
    0xd0a0003cU,
    0x7a100019U,
    0x89680032U,
    0x7c7c0027U,
    0x3a92001fU,
    0x452f0010U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0x58000000U,
    0xfc000000U,
    0x2a000000U,
    0xfd000000U,
    0x72800000U,
    0x92400000U,
    0xc0e00000U,
    0xe8700000U,
    0x81680000U,
    0x4d0c0000U,
    0x343e0000U,
    0x44310000U,
    0xa28b8000U,
    0x44ff4000U,
    0xae95a000U,
    0xa6197000U,
    0x5fe4d800U,
    0x63f36c00U,
    0x87aace00U,
    0x40ad3100U,
    0x372fae80U,
    0x1befd3c0U,
    0xeb48f6a0U,
    0x9adfffd0U,
    0x57419898U,
    0xde66bee4U,
    0xe8376e0eU,
    0x9088411bU,
    0xa5fd76a4U,
    0x2611bff4U,
    0x915fb8a0U,
    0x11808ef4U,
    0x33c6163eU,
    0x3b225d05U,
    0x975360a2U,
    0x103fe2c6U,
    0xd232d821U,
    0x958e6c28U,
    0x1b7f4e2bU,
    0x1653713cU,
    0x8b98eb4U,
    0x5d75e3c8U,
    0xf7ef8eaeU,
    0x3948e3e6U,
    0xabda0eb4U,
    0x87c6a3e3U,
    0x15242ea0U,
    0x8c5093efU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xb8000000U,
    0x6c000000U,
    0x22000000U,
    0x83000000U,
    0x93800000U,
    0xdfc00000U,
    0xbaa00000U,
    0x58300000U,
    0x65780000U,
    0xa75c0000U,
    0x172a0000U,
    0x1350000U,
    0x39fc8000U,
    0x991b4000U,
    0xf9492000U,
    0xa422b000U,
    0x9572e800U,
    0x409ed400U,
    0x1895a00U,
    0x9b020700U,
    0x6f868280U,
    0x15c72b40U,
    0x5da2eaa0U,
    0x3db7bf70U,
    0x77be1098U,
    0x2f7c485cU,
    0x805b5a36U,
    0x92ab0707U,
    0xa3f0029eU,
    0x9d96b57U,
    0x626f4aa2U,
    0x48d24f7cU,
    0xbd2fd89aU,
    0x96352c5dU,
    0x47c6825U,
    0x57dc940fU,
    0xeb6efa2cU,
    0xdc52f728U,
    0xb76bca90U,
    0x86550f43U,
    0xb86cf897U,
    0x87d29c7eU,
    0x5caa001eU,
    0xc2f5000bU,
    0xf95c803fU,
    0x9e2b4017U,
    0x95b1200aU,
    0x33beb007U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0xe8000000U,
    0x1c000000U,
    0xc2000000U,
    0xf000000U,
    0x38800000U,
    0x6f400000U,
    0x6b600000U,
    0x5cb00000U,
    0x23080000U,
    0xe3ec0000U,
    0x6a9e0000U,
    0xe2b30000U,
    0x9e0b8000U,
    0xc6ac000U,
    0x595d2000U,
    0xd092b000U,
    0x78dd0800U,
    0x86556c00U,
    0xc17be600U,
    0x65415500U,
    0xa862cc80U,
    0xee354ac0U,
    0xcf4a44a0U,
    0x9a09e6d0U,
    0x4a6f0288U,
    0x505cc3f4U,
    0x4113e632U,
    0x161d5533U,
    0x6bf4cca2U,
    0x6c6a4ae0U,
    0x295fc4a4U,
    0x89026cfU,
    0x1cd9a2acU,
    0xb054b3fbU,
    0x107bce2eU,
    0x90c68907U,
    0xf0a422aaU,
    0xd29173f7U,
    0x97db6e11U,
    0x8ed1f90cU,
    0xd6398aa7U,
    0xfa236fe4U,
    0x2ad52029U,
    0x3eb035U,
    0x1b230828U,
    0xa7566c3cU,
    0xacf86605U,
    0xf987951cU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x90000000U,
    0x68000000U,
    0xc4000000U,
    0xc6000000U,
    0x6b000000U,
    0xa1800000U,
    0x32400000U,
    0x55200000U,
    0xbff00000U,
    0xa5080000U,
    0x86ac0000U,
    0x9a0000U,
    0x8bf30000U,
    0x7b0b8000U,
    0xf1aac000U,
    0x5b1be000U,
    0xb8b05000U,
    0xe6a8c800U,
    0x709aa400U,
    0x53f42600U,
    0x470a6700U,
    0x9bac0c80U,
    0x321fdac0U,
    0xb43544a0U,
    0x1e6cbef0U,
    0xb67902a8U,
    0x8b6049d4U,
    0x8d42632U,
    0x7fa6725U,
    0x11240ca3U,
    0xb9f3dacfU,
    0xee0f44a3U,
    0xb72fbed7U,
    0x5ada82a4U,
    0x1ad689edU,
    0x2fdc626U,
    0x3fa53727U,
    0x7c35449fU,
    0x8a6cbee3U,
    0x38790294U,
    0xb46049e8U,
    0x754262cU,
    0x9aba671dU,
    0x23840cb7U,
    0x5f43dad9U,
    0xbfa7448cU,
    0xbc33bed4U,
    0xaa68828dU,
    0xa87989fcU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0x48000000U,
    0x4c000000U,
    0x7a000000U,
    0xa1000000U,
    0x3800000U,
    0x15c00000U,
    0x49600000U,
    0x89700000U,
    0x57d80000U,
    0x963c0000U,
    0x500e0000U,
    0x7cf70000U,
    0xf91c8000U,
    0xadd8c000U,
    0x7739a000U,
    0x738cf000U,
    0x99343800U,
    0xf8780400U,
    0x68af1a00U,
    0x5ae08900U,
    0x44b35680U,
    0xcabe9a40U,
    0x914deea0U,
    0xd8d55e50U,
    0x7e49d4a8U,
    0x6456e76cU,
    0x2f0b9a3aU,
    0x5874493fU,
    0x5c5cf684U,
    0xeff96a47U,
    0x936b56acU,
    0x1c829a6eU,
    0xe143eea4U,
    0x54225e63U,
    0xcf5554bdU,
    0x858e277dU,
    0x22323a37U,
    0x8af8b90fU,
    0xc6e8cebbU,
    0x2416e53U,
    0xb2a44c89U,
    0xcf121364U,
    0xf228b834U,
    0x8a0c40fU,
    0xe16ba28U,
    0x21ac7906U,
    0xa5676e8cU,
    0x43769e52U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0x78000000U,
    0x4c000000U,
    0xbe000000U,
    0x2d000000U,
    0x95800000U,
    0xcf400000U,
    0x7aa00000U,
    0xd9700000U,
    0xf380000U,
    0xeb1c0000U,
    0x39aa0000U,
    0xb0f10000U,
    0xa67a8000U,
    0x60bb4000U,
    0x435c2000U,
    0x698d5000U,
    0x50407800U,
    0xdc217400U,
    0x3eb69e00U,
    0xa75ad300U,
    0xb38a5c80U,
    0x67453340U,
    0x1ea2a4a0U,
    0xc3720770U,
    0x58389ab8U,
    0xd99fc44cU,
    0xbc6c1e26U,
    0x391931fU,
    0x726e7ca9U,
    0x96946364U,
    0xcbe8dcb4U,
    0x17d2735bU,
    0xe44c84acU,
    0x17225742U,
    0x9a30629eU,
    0xab98f060U,
    0xef6e201aU,
    0xdb105034U,
    0xd8a8f828U,
    0x5b773419U,
    0x243a3e3dU,
    0x7f9dc333U,
    0x2d6c84baU,
    0x50125741U,
    0xdc28629dU,
    0xc7b4f067U,
    0xf0dc2035U,
    0xf7cd5033U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0x18000000U,
    0x94000000U,
    0xe000000U,
    0x13000000U,
    0x71800000U,
    0x3400000U,
    0x31a00000U,
    0x20100000U,
    0x27480000U,
    0x726c0000U,
    0xffbe0000U,
    0x21970000U,
    0xcc0f8000U,
    0x7fc94000U,
    0xcdaee000U,
    0x97db1000U,
    0x46e7c800U,
    0xefb42400U,
    0xbc5aee00U,
    0x38a68b00U,
    0xc961f80U,
    0x668f73c0U,
    0xa10857a0U,
    0xf54917d0U,
    0x836dd9a8U,
    0xb73accf4U,
    0xc6556e2eU,
    0x106fcb11U,
    0x86b8ff8dU,
    0xe51463c5U,
    0xb1cf9fb9U,
    0xbead33f1U,
    0x565f37bbU,
    0xcda047f8U,
    0x521571b7U,
    0x64bb8c5U,
    0x2e92817U,
    0x1c7f3409U,
    0xebf52614U,
    0xb27eaf0eU,
    0x28f2f1aaU,
    0xfbfef8dbU,
    0x2fb1c82bU,
    0xdc5f2401U,
    0x88a36e23U,
    0x8494cb28U,
    0xea897fb2U,
    0x3b0a23caU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0x78000000U,
    0x34000000U,
    0x22000000U,
    0x6f000000U,
    0x43800000U,
    0xe8400000U,
    0x8f200000U,
    0x32b00000U,
    0x38480000U,
    0x4dec0000U,
    0x799e0000U,
    0x5a330000U,
    0x89888000U,
    0xd18d4000U,
    0x558be000U,
    0xaf887000U,
    0x48f4800U,
    0x1d0f3400U,
    0xae48c600U,
    0x40eed700U,
    0xf51e2680U,
    0x1f2bac0U,
    0x96efeea0U,
    0x581fced0U,
    0x7d724888U,
    0xb52b29ecU,
    0x737e4616U,
    0xd6e09721U,
    0xf2d546beU,
    0x2a5b8aeaU,
    0x495546a6U,
    0xb61b8afcU,
    0x447546b3U,
    0x1bab8ac1U,
    0x47bd46abU,
    0x8a078ac4U,
    0x930346adU,
    0x8d848af3U,
    0x6143c6bcU,
    0xf9a5caf1U,
    0xc2762684U,
    0x3eaebac5U,
    0x7739ee95U,
    0xbfc0cec7U,
    0xf564c882U,
    0xa9569f5U,
    0x8d7d2608U,
    0xc7e5a72aU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0xd8000000U,
    0xe4000000U,
    0x3e000000U,
    0x13000000U,
    0x8b800000U,
    0x19c00000U,
    0x56a00000U,
    0x27700000U,
    0x6a880000U,
    0xceac0000U,
    0x4e9a0000U,
    0x37f30000U,
    0x14ce8000U,
    0x93ca4000U,
    0x1e4a6000U,
    0xe08b1000U,
    0xebacc800U,
    0xfa197400U,
    0x97b76a00U,
    0x4aa8b700U,
    0xa09f4880U,
    0xecf78ec0U,
    0x934f00a0U,
    0x880bbaf0U,
    0x81ee8aa8U,
    0x617d5dfcU,
    0x63ea1eU,
    0xf051f705U,
    0xb13ba8b1U,
    0x6606def1U,
    0x3701a8bbU,
    0x9585decbU,
    0x3ac72890U,
    0x5239ef7U,
    0xdab748b3U,
    0x22b8ef9U,
    0xfadd008dU,
    0xaf94baf6U,
    0xe09a0abfU,
    0xcf41dddU,
    0x834f0a37U,
    0x600ca71fU,
    0xbdef00a4U,
    0xbb7bbaefU,
    0x2d668abfU,
    0x68d15de8U,
    0x2379ea2bU,
    0x2962f73aU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x78000000U,
    0x34000000U,
    0x86000000U,
    0x53000000U,
    0xb800000U,
    0xc7c00000U,
    0xd8e00000U,
    0x80f00000U,
    0x69180000U,
    0xab7c0000U,
    0x3ace0000U,
    0x68770000U,
    0x8d5b8000U,
    0xd85e4000U,
    0xc0dce000U,
    0x2c995000U,
    0x23b88800U,
    0x3aab400U,
    0xde411a00U,
    0x9a257500U,
    0x5891b680U,
    0x1ad9c40U,
    0x2344bea0U,
    0xaea26850U,
    0x36d4c4a8U,
    0xb38b0d54U,
    0x17d49a36U,
    0xd0c3519U,
    0xb896d6b7U,
    0x71aa8c5dU,
    0x6b40d69dU,
    0xf2a18c78U,
    0xfcd556bdU,
    0x5288cc51U,
    0xc9523698U,
    0x924fdc62U,
    0xac365eb5U,
    0xee7c385dU,
    0x5a4fcca2U,
    0xb033f97aU,
    0x47f603eU,
    0xab4b103bU,
    0x16b26802U,
    0xaf38e41bU,
    0x39ec1210U,
    0xda668115U,
    0x95b7cc8bU,
    0xbcbff965U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x18000000U,
    0x4000000U,
    0x7a000000U,
    0x1f000000U,
    0x14800000U,
    0x85400000U,
    0xfd200000U,
    0xfe300000U,
    0xd9980000U,
    0x21fc0000U,
    0xe2ea0000U,
    0xa5410000U,
    0x2d268000U,
    0xf6304000U,
    0xb59b6000U,
    0x47ff3000U,
    0x83e90800U,
    0xd4c28c00U,
    0xa3e2f600U,
    0x9ad32700U,
    0x33cf5280U,
    0x9d71e440U,
    0x85bd3aa0U,
    0xce4c5870U,
    0xc5b6c4b8U,
    0xf05df36cU,
    0x4d9b603eU,
    0xf3ff302dU,
    0x81e90800U,
    0xbfc28c1aU,
    0xd562f61dU,
    0x4932727U,
    0xa06f52a4U,
    0xf901e459U,
    0xb5853aacU,
    0x94c05866U,
    0x3e4c489U,
    0x8ad0f372U,
    0x5bcfe03bU,
    0x81727031U,
    0xfbbee81eU,
    0xab4cfc1fU,
    0xce361e1bU,
    0x619edb3eU,
    0x35ffcc81U,
    0xf0ef7f5bU,
    0xa6419628U,
    0x47a01712U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0x88000000U,
    0x6c000000U,
    0xa2000000U,
    0x4f000000U,
    0x35800000U,
    0x33400000U,
    0x3a00000U,
    0x6100000U,
    0x98980000U,
    0x187c0000U,
    0x75ce0000U,
    0x57430000U,
    0x8da48000U,
    0x4b12c000U,
    0x921ca000U,
    0x96be3000U,
    0x292a0800U,
    0xf0f5c400U,
    0x5c2ece00U,
    0xfe76eb00U,
    0xccef1780U,
    0xb797d940U,
    0xf5d9bfa0U,
    0xaedc2d50U,
    0x9d5d7988U,
    0x8f1f024cU,
    0x243ca02aU,
    0xe3ee3017U,
    0x12120822U,
    0x9e99c414U,
    0x3978ce07U,
    0xdd49eb04U,
    0x968597a5U,
    0xe4c6195eU,
    0xdfe19f8bU,
    0x4030dd5bU,
    0x25cbd1b3U,
    0xef44f65eU,
    0xc9a06632U,
    0xf5111f16U,
    0xf71dd1b8U,
    0x803bf676U,
    0x8deae62aU,
    0x8f10df1fU,
    0x6c1df1b4U,
    0x53bb0664U,
    0x4faa4e07U,
    0xdc342b2eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0x98000000U,
    0xc4000000U,
    0xb6000000U,
    0xc9000000U,
    0x8d800000U,
    0xed400000U,
    0x51600000U,
    0xeeb00000U,
    0x9e880000U,
    0xaac0000U,
    0xd97e0000U,
    0x33470000U,
    0xd4618000U,
    0xa9374000U,
    0x98c86000U,
    0x314b5000U,
    0x5a0bd800U,
    0x2a689400U,
    0x4558ce00U,
    0xdc90cd00U,
    0x3c59f480U,
    0xf911dac0U,
    0xfd1a4ca0U,
    0x12721ed0U,
    0x30a95aa8U,
    0xba7a47c4U,
    0x25c0600eU,
    0x9ba75035U,
    0xc995d803U,
    0x83df9417U,
    0x4b514e28U,
    0xb8fb8d38U,
    0x407949aU,
    0xd6018af2U,
    0xb90614b0U,
    0x4586cad2U,
    0xb14674adU,
    0x23619af1U,
    0x91b3acaaU,
    0xda0e0ed1U,
    0x6a6ae286U,
    0x655983edU,
    0x8c937605U,
    0xa45f0910U,
    0x3d14e295U,
    0x4b1e83e6U,
    0xdb72f633U,
    0xbd284912U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x58000000U,
    0x34000000U,
    0x92000000U,
    0xc5000000U,
    0x12800000U,
    0x36c00000U,
    0xf1600000U,
    0xd2900000U,
    0xbc080000U,
    0x2fac0000U,
    0x4e5e0000U,
    0x5dc10000U,
    0x78e18000U,
    0xe5d14000U,
    0xf62fe000U,
    0x811bf000U,
    0x53e51800U,
    0x8c572c00U,
    0x516c8a00U,
    0x933ce900U,
    0xc955c980U,
    0x3eccfc0U,
    0x45ff31a0U,
    0xa83013d0U,
    0x897ea398U,
    0xcdf7d6f4U,
    0x1599e00eU,
    0x226f039U,
    0x82b2980aU,
    0x5bbb6c18U,
    0x1694ea36U,
    0xa60b5931U,
    0x66a8b19cU,
    0x6adc53c8U,
    0x6406c3b7U,
    0x6a0066e2U,
    0x81049818U,
    0x48866c1cU,
    0x9fc36a0cU,
    0x45e7190cU,
    0xb350d182U,
    0x9aebe3f2U,
    0xd97bbbb7U,
    0x35f0fad6U,
    0x519d6a1dU,
    0x5826192aU,
    0x2bb1518fU,
    0xef3aa3d1U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x58000000U,
    0x1c000000U,
    0x96000000U,
    0xcd000000U,
    0x3800000U,
    0xccc00000U,
    0x31e00000U,
    0x29500000U,
    0xcab80000U,
    0xfb9c0000U,
    0x7d2e0000U,
    0x7c10000U,
    0x17628000U,
    0x42124000U,
    0x4d9da000U,
    0xe02a9000U,
    0x4c443800U,
    0x6fa7c400U,
    0xa1f09200U,
    0x23cab500U,
    0x72108380U,
    0xb598c040U,
    0xec291ba0U,
    0x72459450U,
    0xe6a5b198U,
    0x2874e56cU,
    0xb40ba00eU,
    0x8d77902bU,
    0x5388b80bU,
    0xdbb48408U,
    0x916fb233U,
    0x78626507U,
    0x1e911ba9U,
    0x25d9944aU,
    0xb58bb19eU,
    0xeb5e557U,
    0x6ee92039U,
    0xd2a5d019U,
    0xba751806U,
    0x130e1414U,
    0x25f38a3aU,
    0x9c9a11dU,
    0x391789a0U,
    0xd31e2173U,
    0x676fb204U,
    0x5626502U,
    0xa5111ba1U,
    0x5199476U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x30000000U,
    0xd8000000U,
    0x94000000U,
    0xbe000000U,
    0x29000000U,
    0x99800000U,
    0x29400000U,
    0x86a00000U,
    0xd9d00000U,
    0x11980000U,
    0x827c0000U,
    0x8f0a0000U,
    0x14410000U,
    0x61248000U,
    0xb9944000U,
    0x3ebbe000U,
    0xaaeb9000U,
    0x8c321800U,
    0xf1ed7400U,
    0xd6b24e00U,
    0xb628f300U,
    0xf9d6d180U,
    0x819ca740U,
    0xa7f29a0U,
    0xf30a4370U,
    0xe6477fb8U,
    0x6223c47cU,
    0xb711e00eU,
    0xa77a9019U,
    0x9c8e9801U,
    0xfa053423U,
    0xbf03ae3fU,
    0x9c826325U,
    0xaac049a8U,
    0xe0e59376U,
    0x7bf687a3U,
    0xc689206cU,
    0x1503b63eU,
    0xcb82172fU,
    0x7a448791U,
    0x60242062U,
    0xc15360dU,
    0xdfb571eU,
    0x6349e7a5U,
    0x7166f074U,
    0xeeb24e08U,
    0xd228f312U,
    0xffd6d1baU,
    0xc9ca761U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x30000000U,
    0xa8000000U,
    0xc4000000U,
    0x72000000U,
    0xb3000000U,
    0x36800000U,
    0x9cc00000U,
    0xdfa00000U,
    0x1af00000U,
    0x5be80000U,
    0xda8c0000U,
    0xc4da0000U,
    0xd7470000U,
    0x88e28000U,
    0x5211c000U,
    0xa9fae000U,
    0x4374f000U,
    0x2fae2800U,
    0x3ceebc00U,
    0xc60b4a00U,
    0x2719ef00U,
    0xb8626c80U,
    0x67d0a5c0U,
    0x97dea4a0U,
    0xb1c6e9f0U,
    0x2c21c688U,
    0xd1b6bafcU,
    0xad0a6022U,
    0x1d9e300dU,
    0x2a4481eU,
    0xe1708c11U,
    0x64af023bU,
    0xf669633aU,
    0x74cd6ea6U,
    0x55b9c6cdU,
    0x9113ca0cU,
    0x577f2f27U,
    0x8bb20cb8U,
    0x1a0995c9U,
    0xf9186c8fU,
    0x1d67a5d6U,
    0xa05424adU,
    0x269b29e1U,
    0x2126acU,
    0x7b54ae1U,
    0xfc0ec819U,
    0x1d4c31U,
    0xc4e7e211U,
    0x1416931fU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x58000000U,
    0x4c000000U,
    0x56000000U,
    0x89000000U,
    0x4a800000U,
    0xa9c00000U,
    0xc0e00000U,
    0xf500000U,
    0x2b280000U,
    0xcc0c0000U,
    0x86be0000U,
    0x5e410000U,
    0xc9a78000U,
    0x51774000U,
    0x631a2000U,
    0x49361000U,
    0x9bbe9800U,
    0x4ec7ac00U,
    0x5f63d200U,
    0x214dd00U,
    0xd70bd380U,
    0x473b54c0U,
    0xce076ba0U,
    0x6506e8d0U,
    0xac842198U,
    0x28c499c4U,
    0x9e63a00eU,
    0xbc905017U,
    0x34cb383bU,
    0xdafc25U,
    0x4ed16a11U,
    0xfb68611bU,
    0x80af19b7U,
    0x2b4e65cbU,
    0x4d9aca08U,
    0x52f4310aU,
    0x205a21a3U,
    0x901599f3U,
    0x5c0c202fU,
    0x7ebb1033U,
    0xa2471826U,
    0x97a1ec27U,
    0xcc767208U,
    0x33998d2cU,
    0x3ff16b8cU,
    0x98dbe8e0U,
    0xa2d5a1bcU,
    0x1d6ed9cbU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x50000000U,
    0x78000000U,
    0xc000000U,
    0x8a000000U,
    0x51000000U,
    0x43800000U,
    0x27c00000U,
    0xaa600000U,
    0xd4900000U,
    0xc6980000U,
    0xa93c0000U,
    0x92ca0000U,
    0xb5470000U,
    0xe218000U,
    0xf9344000U,
    0xe8692000U,
    0xc333f000U,
    0xf168b800U,
    0x4b5c400U,
    0x78ad2a00U,
    0x1d4b100U,
    0x38be2980U,
    0x780b5340U,
    0xea7b1a0U,
    0xf0716770U,
    0x24482388U,
    0xef071264U,
    0xf482a016U,
    0x4f40b027U,
    0xb7201814U,
    0x6eb27403U,
    0x19acb22cU,
    0xca52851eU,
    0x3b7bbb98U,
    0x2c6a2668U,
    0xd34b22bU,
    0xae6e8526U,
    0xb831bb97U,
    0xb3ed2666U,
    0x18f53228U,
    0xf90ac50aU,
    0xf5209b97U,
    0x7bb2d663U,
    0x542f8a2bU,
    0x12940135U,
    0x7d9e31b8U,
    0x4bb92744U,
    0xde8b03b3U,
    0x4ce3e24fU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x48000000U,
    0xf4000000U,
    0x62000000U,
    0x35000000U,
    0x9f800000U,
    0x6400000U,
    0xcde00000U,
    0x51100000U,
    0x13380000U,
    0x469c0000U,
    0x196a0000U,
    0x23c70000U,
    0x82a38000U,
    0x3577c000U,
    0x36d6000U,
    0x7ac2f000U,
    0x63231800U,
    0x80334400U,
    0x4e0de600U,
    0x7314e100U,
    0x463d7580U,
    0x691e6e40U,
    0x472b0da0U,
    0x5223da70U,
    0x45b7f388U,
    0x714f7f74U,
    0xef76e01aU,
    0xfa693009U,
    0x8b447822U,
    0x9e66b434U,
    0x39557e2dU,
    0x25dc6531U,
    0x630ff38cU,
    0x30937f58U,
    0x667ce030U,
    0xabfe3034U,
    0xbff819U,
    0x4edd7409U,
    0x9f8a1e30U,
    0x5e559529U,
    0x175d6bbeU,
    0xe6ccfb56U,
    0x7d35e63fU,
    0xe588e102U,
    0xb7577590U,
    0xaed96e72U,
    0xef888dacU,
    0xa6541a7aU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xb0000000U,
    0xf8000000U,
    0xac000000U,
    0x12000000U,
    0x65000000U,
    0x23800000U,
    0xe5c00000U,
    0x29e00000U,
    0xc4b00000U,
    0xb5f80000U,
    0x1fdc0000U,
    0x598a0000U,
    0x24470000U,
    0x31268000U,
    0x72d74000U,
    0xce8f2000U,
    0xb2c63000U,
    0x1f60d800U,
    0x5af7a400U,
    0x65d89a00U,
    0x188cbd00U,
    0xe1c72480U,
    0xbfe15540U,
    0xf7b0dca0U,
    0x57cc170U,
    0xfa9a9e88U,
    0x6d2cd85cU,
    0x7851a036U,
    0xcacd7037U,
    0xf3e5f812U,
    0x35b6941eU,
    0x887ec232U,
    0x3d1c590bU,
    0xcee89e86U,
    0x8ab7d851U,
    0x5afd200fU,
    0xe55d303fU,
    0xc5cc5819U,
    0xd967e413U,
    0x41f13a09U,
    0x915dcd39U,
    0xbc8dcadU,
    0xf660c160U,
    0x1b709e90U,
    0xbd1bd86fU,
    0xeef202fU,
    0x2ab63011U,
    0xeaf8d81dU,
    0x1d5ba43fU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0x78000000U,
    0x14000000U,
    0x3a000000U,
    0x6d000000U,
    0xff800000U,
    0xec400000U,
    0x6c600000U,
    0xe5d00000U,
    0xd8180000U,
    0xcd3c0000U,
    0x6f0a0000U,
    0xa4830000U,
    0xa0c78000U,
    0x60a24000U,
    0x8b73e000U,
    0x146ab000U,
    0x33d61800U,
    0x5b1b4400U,
    0x45bef600U,
    0x1c97d00U,
    0xa323ef80U,
    0xd135ad40U,
    0xb0c17a0U,
    0x16875970U,
    0x91c379a8U,
    0xd9272074U,
    0x5c358036U,
    0xc48d4031U,
    0x92c66018U,
    0x91a7f01aU,
    0x12f0782fU,
    0xd32cb425U,
    0x9b368e2aU,
    0xee09c919U,
    0xb5076192U,
    0x3b836471U,
    0x8e46f616U,
    0x5657d13U,
    0x5851ef8dU,
    0x4d5aad46U,
    0x64d997afU,
    0xb9a1941U,
    0xef7d19a0U,
    0x646cd064U,
    0xbbd7f836U,
    0x71ef419U,
    0x3bd6e36U,
    0x3acf7901U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0x88000000U,
    0xc4000000U,
    0xb6000000U,
    0x6f000000U,
    0x6b800000U,
    0x7400000U,
    0xeae00000U,
    0x6100000U,
    0x4e480000U,
    0xf4ec0000U,
    0x9f9a0000U,
    0x32810000U,
    0xc3c28000U,
    0xe620c000U,
    0x1bb5e000U,
    0x2abfd000U,
    0x78b4e800U,
    0x933b9400U,
    0x36f71e00U,
    0x55db100U,
    0x10241d80U,
    0xd4b2b5c0U,
    0xd13f15a0U,
    0x7f7f1f0U,
    0x35de63b8U,
    0x42e114ccU,
    0x9210801aU,
    0xe04dc03dU,
    0x27ed6017U,
    0xe1e1016U,
    0x28c38815U,
    0x9ba4842bU,
    0xe3f6160dU,
    0xd3d9f51bU,
    0x35e7ebb3U,
    0x459490fdU,
    0x1d0c162eU,
    0xd008f500U,
    0xba8d6b9cU,
    0x654850f4U,
    0xe96bf613U,
    0x97da253eU,
    0x43e1039dU,
    0x4a9204f0U,
    0x868b880dU,
    0x5f48842fU,
    0x946c1618U,
    0xd558f536U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x98000000U,
    0xf4000000U,
    0x16000000U,
    0x6f000000U,
    0x51800000U,
    0xc2400000U,
    0x42a00000U,
    0x45900000U,
    0xa6d80000U,
    0xa13c0000U,
    0x8b8a0000U,
    0xf0830000U,
    0xe8c18000U,
    0x37634000U,
    0xe275e000U,
    0x9c6bd000U,
    0x62306800U,
    0xb7cf4400U,
    0xa122e600U,
    0xa6d60b00U,
    0xefbeb80U,
    0xb1682f40U,
    0x5ab463a0U,
    0xf30fbb70U,
    0x54476da8U,
    0xeda5b444U,
    0x3413800eU,
    0xb49c4039U,
    0x7b9e602bU,
    0x3a1b9012U,
    0x405c0837U,
    0x26fbd416U,
    0xed6d6e01U,
    0xd0b19f3aU,
    0x3608e5adU,
    0x10c22049U,
    0xf3608e3bU,
    0x9c764f35U,
    0x4f6a8db9U,
    0x49b2646dU,
    0xf889e826U,
    0xcb00041bU,
    0x1f850617U,
    0x7942db28U,
    0xd52003baU,
    0x70d72b6eU,
    0x41fae586U,
    0x30ed2052U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x88000000U,
    0xdc000000U,
    0xea000000U,
    0x5b000000U,
    0xe3800000U,
    0x48400000U,
    0x45a00000U,
    0x39100000U,
    0x8b380000U,
    0xf1dc0000U,
    0xdcee0000U,
    0xc9810000U,
    0xb3418000U,
    0xd627c000U,
    0x69562000U,
    0x9a9df000U,
    0xfecf6800U,
    0xe6d49400U,
    0x80da8600U,
    0xc468f300U,
    0x12476780U,
    0xa6a01c40U,
    0xfe902fa0U,
    0xed787850U,
    0x517c4198U,
    0x6b7fdf74U,
    0x5879803aU,
    0xf7fbc003U,
    0xddb82020U,
    0x1f1cf005U,
    0x2f8ee820U,
    0xb7f35433U,
    0xe00ca613U,
    0x4db5030aU,
    0x4aa80f8bU,
    0x31248860U,
    0xb0d2a9a4U,
    0xe1dc8b50U,
    0x14ed2621U,
    0xf582c30fU,
    0xc9462f9bU,
    0x5257872U,
    0x56d3c1b1U,
    0x38d91f5bU,
    0xe06e201fU,
    0x3c41f01eU,
    0x43a16828U,
    0x7015942eU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0xc8000000U,
    0x4c000000U,
    0xf6000000U,
    0x25000000U,
    0x800000U,
    0x45400000U,
    0x59a00000U,
    0xb9500000U,
    0x8ba80000U,
    0x744c0000U,
    0x993a0000U,
    0xf2030000U,
    0x5f048000U,
    0xf384c000U,
    0x10c7a000U,
    0xd4639000U,
    0xe9b4f800U,
    0x9d5e9c00U,
    0x63b70600U,
    0x665ee300U,
    0x5a302d80U,
    0x6d9edac0U,
    0xe7d175a0U,
    0x37ecd6f0U,
    0x436c0ba8U,
    0xb42b69ecU,
    0x5288001aU,
    0x485c003fU,
    0x6b320007U,
    0x4f1f0006U,
    0x8596800fU,
    0x39cbc00bU,
    0x20f9202bU,
    0xf0e45036U,
    0xa6f7d834U,
    0xfff9cc22U,
    0xc3645e12U,
    0x9633ef0fU,
    0x5b9bd393U,
    0xe2d2a5ddU,
    0x476c5e15U,
    0xce2fef05U,
    0xa189d3a1U,
    0x1ddda5f2U,
    0xe6f2de34U,
    0x1ff82f28U,
    0x9362f381U,
    0x2e36f5f0U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x98000000U,
    0x5c000000U,
    0xae000000U,
    0xeb000000U,
    0x68800000U,
    0x72400000U,
    0xff200000U,
    0xb4d00000U,
    0x38680000U,
    0x280c0000U,
    0x9c7a0000U,
    0x74050000U,
    0xaa008000U,
    0xf9024000U,
    0x1d87a000U,
    0x69c35000U,
    0xb9e0c800U,
    0x97b16c00U,
    0x989f8e00U,
    0xcc332f00U,
    0xfe5fcd80U,
    0x3f509cc0U,
    0xf6aaa5a0U,
    0x1f6ba0f0U,
    0x668f6398U,
    0x113ea3fcU,
    0xb9a0003eU,
    0xed90002bU,
    0x4fc80035U,
    0xde9c0021U,
    0xc3320017U,
    0xb4d90031U,
    0xa012800eU,
    0x4e0b400bU,
    0x437d202bU,
    0x9684103bU,
    0xf147e816U,
    0xb3a07c3aU,
    0x490e603U,
    0x1a4d1329U,
    0x435a8b8aU,
    0x10d7dff9U,
    0x3a6a6623U,
    0x150a5331U,
    0x73fdabb7U,
    0x58c6cfc0U,
    0x90650e31U,
    0x74746f3eU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x28000000U,
    0xac000000U,
    0x2e000000U,
    0x3d000000U,
    0xaf800000U,
    0xc5c00000U,
    0x7600000U,
    0x7d900000U,
    0x5ae80000U,
    0x434c0000U,
    0x973a0000U,
    0xbc010000U,
    0xa6078000U,
    0xa100c000U,
    0xf986a000U,
    0x7cc55000U,
    0x2ae5f800U,
    0xab55e400U,
    0xcf0aba00U,
    0x549f6900U,
    0xf77f980U,
    0x85bc00c0U,
    0x1646a1a0U,
    0x4a1b4f0U,
    0x8bf463b8U,
    0xf6faf9e4U,
    0xc9600032U,
    0xd090000fU,
    0xbd680019U,
    0x7a8c0030U,
    0x965a002aU,
    0x50910015U,
    0x7d6f8021U,
    0x1a8cc021U,
    0xc65ca03cU,
    0x7894503eU,
    0xd16a7832U,
    0x34892420U,
    0xfb5e1a1aU,
    0xd7173917U,
    0x14af81b2U,
    0x33e824d8U,
    0x86cd3bbdU,
    0x8dfb4dedU,
    0x57e0c23fU,
    0xa4d64d04U,
    0x3ac9e3a8U,
    0x2bfb39e1U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0x28000000U,
    0x4000000U,
    0x36000000U,
    0xe7000000U,
    0xc0800000U,
    0x40400000U,
    0x9da00000U,
    0xa8d00000U,
    0x38780000U,
    0xf09c0000U,
    0xf46e0000U,
    0x13050000U,
    0x9e828000U,
    0xc341c000U,
    0x1b22e000U,
    0x2790d000U,
    0x615ea800U,
    0x2e0c8c00U,
    0xb6b53e00U,
    0x8bc9eb00U,
    0x12d72980U,
    0xbd78a840U,
    0xd91d61a0U,
    0xcfadf450U,
    0x146277b8U,
    0x3eb0534cU,
    0xdfce0032U,
    0x1cd5000dU,
    0x67a801fU,
    0x39dc024U,
    0x5aece037U,
    0x9845d028U,
    0xf1a4281aU,
    0xfad14c03U,
    0x9979de19U,
    0xff1c3b06U,
    0x70ab0194U,
    0xf8e5e452U,
    0x4cf2bfb4U,
    0x9368cf58U,
    0x9385f624U,
    0xbec17731U,
    0x2ee4dfabU,
    0x9bf0df6fU,
    0x1bed3e14U,
    0xa7c5eb08U,
    0x3d612994U,
    0x6531a873U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x10000000U,
    0xd8000000U,
    0x4c000000U,
    0xae000000U,
    0xfd000000U,
    0x41800000U,
    0xab400000U,
    0xd9e00000U,
    0xf8f00000U,
    0x52d80000U,
    0x4f7c0000U,
    0x54ea0000U,
    0x4d030000U,
    0x49878000U,
    0x1f474000U,
    0x2be42000U,
    0x73f1d000U,
    0xa258f800U,
    0xbbc9400U,
    0xdb4abe00U,
    0x2d929900U,
    0x48ebd280U,
    0xdb00de40U,
    0xc8850aa0U,
    0x38c29a70U,
    0xe9a2cca8U,
    0x994d774U,
    0xa2ea001eU,
    0x3c030027U,
    0x86078015U,
    0x59074028U,
    0x6b84200dU,
    0x6c41d006U,
    0x8760f80dU,
    0x4130942bU,
    0x9cf8be32U,
    0x84ad9907U,
    0x8c6652b7U,
    0x71b49e7eU,
    0xf83eaab3U,
    0x1b080a77U,
    0x34f414a5U,
    0x3cda937cU,
    0x927fc62eU,
    0x56a4d3eU,
    0x3e42cc81U,
    0xdc64d77dU,
    0x49b20021U,
    0x843f0009U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xa8000000U,
    0x44000000U,
    0x16000000U,
    0x51000000U,
    0x5f800000U,
    0xa0400000U,
    0xe3a00000U,
    0xbbd00000U,
    0xe2980000U,
    0x9efc0000U,
    0x16ca0000U,
    0xe4250000U,
    0xd7948000U,
    0x33fc000U,
    0x7a2ee000U,
    0x16d5d000U,
    0xd7191800U,
    0xc5be4400U,
    0xb1efb600U,
    0xb328300U,
    0x2a69d180U,
    0x7d742f40U,
    0x78caa9a0U,
    0x29207b70U,
    0x9212e798U,
    0x10796c5cU,
    0x290d9832U,
    0x47c1840dU,
    0x3fe15617U,
    0xf2775329U,
    0x4248c9b8U,
    0x88e66b6dU,
    0x74f71f8aU,
    0xa98bf862U,
    0xc505b60fU,
    0x9187833dU,
    0xfd455197U,
    0x5e27ef4dU,
    0xf4964987U,
    0x4bbcab6bU,
    0xcced7fa4U,
    0xe6b1e85cU,
    0x7d2a4e03U,
    0x3850173fU,
    0x78d9ff90U,
    0x9e5e2856U,
    0x59cae36U,
    0x79c72fU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x78000000U,
    0x7c000000U,
    0x52000000U,
    0xd5000000U,
    0x1b800000U,
    0xe8c00000U,
    0xf4200000U,
    0x3100000U,
    0xbca80000U,
    0xcbcc0000U,
    0x5c3a0000U,
    0x23a10000U,
    0x51d78000U,
    0xf98a4000U,
    0xad5ea000U,
    0xf750b000U,
    0xaecad800U,
    0xafb96400U,
    0xaf627200U,
    0xdbf51b00U,
    0x5980f80U,
    0xa37324c0U,
    0x9ddbf7a0U,
    0x3b10b0f0U,
    0x60adfdb8U,
    0xa9cc7fe4U,
    0xa13d5826U,
    0x3c23243bU,
    0x9714d212U,
    0x8aa9ab3eU,
    0x60c8d78cU,
    0xb8bb40d0U,
    0x79e6058fU,
    0x4b3ebffU,
    0x3379522cU,
    0xfd42eb2cU,
    0x6be1f783U,
    0x71b1b0f1U,
    0x18fa7d86U,
    0x3d863fdeU,
    0x9bc3f802U,
    0x5ca39420U,
    0x23560a29U,
    0x38cccf19U,
    0x34b8a5b3U,
    0x63e35bf8U,
    0x85b38a0bU,
    0x7efb8f30U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x38000000U,
    0x6c000000U,
    0x82000000U,
    0xdf000000U,
    0x38800000U,
    0x13c00000U,
    0xea600000U,
    0x38f00000U,
    0xaa280000U,
    0x630c0000U,
    0x565e0000U,
    0xd0e10000U,
    0xb4328000U,
    0x98c94000U,
    0xb8392000U,
    0x2e10b000U,
    0xe71c8800U,
    0xca47ac00U,
    0xf423e200U,
    0x62d3a900U,
    0x75fefa80U,
    0x174edc0U,
    0x4469d2a0U,
    0x12ab1d0U,
    0xe58f9898U,
    0x309e04ecU,
    0x79060816U,
    0x7182ec37U,
    0xfa44c216U,
    0x6c221910U,
    0x1ed0f2a0U,
    0x3ffa01e8U,
    0x8a7310b4U,
    0x92e9a8edU,
    0x4fedea06U,
    0xe86d453eU,
    0x232c38b3U,
    0x2a8bf4c5U,
    0xc01da01cU,
    0x3ec4f036U,
    0x75e12800U,
    0x9fb35c16U,
    0x218aca00U,
    0x569cf50eU,
    0x9002309aU,
    0x880518c0U,
    0xb4076208U,
    0x1e07e912U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0xb8000000U,
    0xdc000000U,
    0x66000000U,
    0xa7000000U,
    0xf7800000U,
    0x6400000U,
    0xc6200000U,
    0xd7700000U,
    0x18f80000U,
    0x731c0000U,
    0xc20a0000U,
    0x8ca50000U,
    0xb7b28000U,
    0xa0194000U,
    0x878ca000U,
    0xcfe21000U,
    0xeb11e800U,
    0xffad3400U,
    0xd390f600U,
    0x3a6c9f00U,
    0x81f66880U,
    0xfe3bc540U,
    0x1cf9a0a0U,
    0x791da170U,
    0x5b0c1e98U,
    0x70221a6cU,
    0x98716836U,
    0x1b7d741bU,
    0x75cd60fU,
    0xa12ecf02U,
    0xb153a08aU,
    0x9f88a15fU,
    0x43e69eb5U,
    0xa5175a6dU,
    0x4cafc812U,
    0x26166404U,
    0x212dbe11U,
    0x7153bb21U,
    0x7f8f76acU,
    0x33e66e49U,
    0x1d153e38U,
    0x90affb20U,
    0x40115682U,
    0x862d3e4cU,
    0x86d07616U,
    0x79ccdf2cU,
    0xf5c248b9U,
    0xca659544U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x70000000U,
    0xd8000000U,
    0x3c000000U,
    0xbe000000U,
    0xc1000000U,
    0xa8800000U,
    0x67c00000U,
    0x79e00000U,
    0xc0d00000U,
    0x1e380000U,
    0x49c0000U,
    0x8c6e0000U,
    0x6e650000U,
    0xee928000U,
    0xb89e4000U,
    0x7268a000U,
    0xcf64d000U,
    0x36131800U,
    0x75e3c00U,
    0x37893200U,
    0xb1b49500U,
    0xe92be780U,
    0xab424040U,
    0xdc22dfa0U,
    0xa636ec50U,
    0xc76855b8U,
    0xde4954cU,
    0xaad7980eU,
    0xe5397c03U,
    0x9f1d1201U,
    0xe42b0523U,
    0x25c2dfb3U,
    0x26e6ec66U,
    0xb95055bbU,
    0x79789549U,
    0xfeb9980bU,
    0xb75c7c24U,
    0xcf8f9214U,
    0x9db54539U,
    0xff2a7f86U,
    0x8e423c4bU,
    0xf6a34da1U,
    0xbef6a97fU,
    0xd708aa26U,
    0x274e939U,
    0xaaca75abU,
    0x5892057cU,
    0xcd9a2027U,
    0x90ea9032U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xf0000000U,
    0x98000000U,
    0x1c000000U,
    0xba000000U,
    0x29000000U,
    0xd8800000U,
    0x89c00000U,
    0x32200000U,
    0xc6100000U,
    0x7c180000U,
    0xeb7c0000U,
    0xe88e0000U,
    0xfa10000U,
    0xe5528000U,
    0x7afc4000U,
    0x9cfa000U,
    0xcc47f000U,
    0xfd617800U,
    0xe2f2dc00U,
    0xaba88200U,
    0xad333b00U,
    0xba8cdf80U,
    0xe2a3d940U,
    0x7bd087a0U,
    0x903ab550U,
    0x526edd98U,
    0x4710a26cU,
    0x809df83eU,
    0xd4bf9c2bU,
    0x21ada230U,
    0xbc348b31U,
    0x6e0c07beU,
    0xb967f56fU,
    0xe4f3fdb4U,
    0x28ab1249U,
    0x34b3203dU,
    0x6fcab005U,
    0xff445837U,
    0x1ce56c33U,
    0xdbb05a14U,
    0x944b172aU,
    0xff01a599U,
    0x93837e46U,
    0x8447fa1eU,
    0x2960e716U,
    0xccf6ddbcU,
    0x4caca257U,
    0x62b3f818U,
    0x64ce9c05U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x68000000U,
    0x3c000000U,
    0x6a000000U,
    0x45000000U,
    0xd8800000U,
    0xec00000U,
    0x8ba00000U,
    0xd1f00000U,
    0xc2b80000U,
    0x755c0000U,
    0x24ce0000U,
    0x84a10000U,
    0x5c708000U,
    0x2cf9c000U,
    0x143b2000U,
    0x409ff000U,
    0x84e99800U,
    0x1c133c00U,
    0xee2f5a00U,
    0xe9b6cf00U,
    0xccdcd480U,
    0x6c8c1f40U,
    0x42c6eca0U,
    0x29a51350U,
    0xf8f50e98U,
    0xa83e107cU,
    0x6a9f9802U,
    0x21ee3c33U,
    0x7491da38U,
    0x88ee0f32U,
    0x5e1774b6U,
    0x772a2f5dU,
    0xeb3454aeU,
    0xef19df76U,
    0x3abccbcU,
    0xf7f7e356U,
    0x25ba1699U,
    0x84d9ec7bU,
    0x408d6236U,
    0x10c3c315U,
    0x88a13695U,
    0x1e761c4bU,
    0xb5fcfa07U,
    0x16bcff25U,
    0x63586c99U,
    0xebcdd37aU,
    0xa926aebeU,
    0xe2342072U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0xb8000000U,
    0x14000000U,
    0xe6000000U,
    0xc9000000U,
    0x70800000U,
    0x1b400000U,
    0xede00000U,
    0xf3900000U,
    0xc3c80000U,
    0x83ec0000U,
    0xdfde0000U,
    0x1ae10000U,
    0x2e148000U,
    0x160b4000U,
    0x9848e000U,
    0xef2c5000U,
    0xecf8d800U,
    0x15527c00U,
    0x6bec5600U,
    0x3dbd100U,
    0xb0e5f980U,
    0x2517f5c0U,
    0xf18941a0U,
    0x2e8e99d0U,
    0x8f0b2f98U,
    0x20ca64dcU,
    0xb86ed836U,
    0xc31f7c19U,
    0x71c6d62fU,
    0x5a1913bU,
    0xdf1998bU,
    0x509ce5f1U,
    0x510779beU,
    0xf481b5d8U,
    0x9543218fU,
    0xb8e489c6U,
    0xc911978aU,
    0xe38e08c3U,
    0x318e3603U,
    0x8e8dc108U,
    0x5f0941a9U,
    0xa8ce99c5U,
    0x146b2f86U,
    0x311a64e6U,
    0x5ec6d831U,
    0xbc237c0aU,
    0x6630d63bU,
    0xa63c9119U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xb0000000U,
    0x88000000U,
    0x4000000U,
    0x4e000000U,
    0x8f000000U,
    0xb4800000U,
    0xa4c00000U,
    0x8a200000U,
    0x1700000U,
    0x99d80000U,
    0x5bc0000U,
    0xc12e0000U,
    0xb9270000U,
    0x37f68000U,
    0xd818c000U,
    0x4a1de000U,
    0xd7181000U,
    0xbe992800U,
    0x53d92c00U,
    0x84ba3a00U,
    0xdaaa3300U,
    0x1965ec80U,
    0x91147c40U,
    0x574fa4a0U,
    0x9cb18050U,
    0xc47f56a8U,
    0xce8d8f7cU,
    0x8f17280aU,
    0x804e2c3dU,
    0x3434ba39U,
    0x12bef30eU,
    0x9ae0c9cU,
    0xffe76c4bU,
    0xc8d60cbcU,
    0x24eb6c5aU,
    0x48800c81U,
    0x6c06c54U,
    0xdf208cbeU,
    0x4cf3ac52U,
    0x8a9dec82U,
    0xd5d87c47U,
    0x2fb9a49bU,
    0x902a805aU,
    0xbaa7d6a6U,
    0xabb24f60U,
    0xbcfc482cU,
    0x4efc21U,
    0x74307228U,
    0x32bfcf0eU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0xd8000000U,
    0xe4000000U,
    0xaa000000U,
    0xe3000000U,
    0x82800000U,
    0x1dc00000U,
    0xeb200000U,
    0xf7d00000U,
    0xdab80000U,
    0xb8dc0000U,
    0x54ee0000U,
    0x79270000U,
    0xd78000U,
    0x9a3b4000U,
    0xea196000U,
    0xab4ab000U,
    0x3e326800U,
    0xfac9dc00U,
    0xa974a200U,
    0x102cc300U,
    0x6b81b980U,
    0xac45da40U,
    0x8c6531a0U,
    0x7531f650U,
    0xb44c9ba8U,
    0x92b55974U,
    0x1e0a681eU,
    0x6bd5dc0dU,
    0x44baa214U,
    0x9dbc325U,
    0x416e39bcU,
    0x94629a46U,
    0xf1325190U,
    0xae4c4652U,
    0x39b17384U,
    0xa08bc560U,
    0x3811aa07U,
    0xe698af00U,
    0x9f88f3b7U,
    0x4978573U,
    0xda5f4a33U,
    0xb9295f04U,
    0xfa03fb8dU,
    0x5b04e954U,
    0xd6818025U,
    0xffc0402dU,
    0x3420e004U,
    0x3b56f000U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x78000000U,
    0x1c000000U,
    0xd6000000U,
    0x11000000U,
    0x9d800000U,
    0xc400000U,
    0xc6e00000U,
    0x38900000U,
    0xf3680000U,
    0xb64c0000U,
    0xcf7a0000U,
    0x70e50000U,
    0x99928000U,
    0xf6ef4000U,
    0xd60ca000U,
    0xa79f9000U,
    0x4573d800U,
    0x21785400U,
    0x5de74600U,
    0x42141900U,
    0xe3ad0180U,
    0xe96930c0U,
    0x5948f9a0U,
    0xb7fdb4f0U,
    0x2322c798U,
    0xf03769ecU,
    0x4e1bd806U,
    0x4b34542bU,
    0xa49d4613U,
    0x53f1191fU,
    0x9fbf818cU,
    0xfc670c0U,
    0x9fa45983U,
    0x39f224f9U,
    0x8b91fb2U,
    0x6b433dccU,
    0x1a669e38U,
    0x41554d14U,
    0x2dcac7baU,
    0xfa3b69d3U,
    0xdf81d80cU,
    0x6f415438U,
    0x6067c619U,
    0x1e525909U,
    0xcd492187U,
    0xc5fce0d6U,
    0x18250195U,
    0xdab530e9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0xe8000000U,
    0xc000000U,
    0x5e000000U,
    0xed000000U,
    0x4e800000U,
    0xc0400000U,
    0x68e00000U,
    0xc7f00000U,
    0x63d80000U,
    0x58fc0000U,
    0x712a0000U,
    0x21a30000U,
    0x73948000U,
    0xf76dc000U,
    0x8040a000U,
    0x88e3d000U,
    0x97f4f800U,
    0xfbdc6c00U,
    0xbcf85e00U,
    0x232d5100U,
    0x92a28a80U,
    0xd0158e40U,
    0x79a852a0U,
    0x28e4f270U,
    0x27f0d4a8U,
    0x33dbdf6cU,
    0xc0fe5812U,
    0x952cbc2fU,
    0x73a02625U,
    0xc090fd24U,
    0x54e874beU,
    0xe840f48U,
    0x2040a03eU,
    0x38e3d00aU,
    0x5ff4f819U,
    0x87dc6c0bU,
    0xaf85e24U,
    0xc22d512cU,
    0x82228a8fU,
    0xfd558e58U,
    0x5fc85293U,
    0x2f54f257U,
    0x2cc8d4abU,
    0xacd7df7cU,
    0xd20c5820U,
    0xec73bc2dU,
    0x711ea617U,
    0x165e3d2eU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0xc8000000U,
    0x9c000000U,
    0x7e000000U,
    0xf000000U,
    0x37800000U,
    0xd6400000U,
    0x6e200000U,
    0x5e100000U,
    0xefb80000U,
    0x475c0000U,
    0xbaae0000U,
    0x65650000U,
    0x9eb68000U,
    0x536ac000U,
    0x61012000U,
    0x40835000U,
    0xd5c46800U,
    0xaa65b400U,
    0x89313600U,
    0x15289f00U,
    0x3720c380U,
    0x4a948440U,
    0xd87d0ba0U,
    0x9c3da050U,
    0xb1ff5b8U,
    0x91891b6cU,
    0x11f3480aU,
    0x29cfe40bU,
    0x8d5de35U,
    0x745eeb18U,
    0x232855a0U,
    0xe4208b51U,
    0x63160029U,
    0xf1390031U,
    0xd988035U,
    0x1f4fc00fU,
    0x6e17a03cU,
    0x7b99038U,
    0x4b5d482dU,
    0xfcaae41fU,
    0x3e635e1cU,
    0x4b342b28U,
    0xf4297591U,
    0x37a3db79U,
    0xff526814U,
    0x821cb43eU,
    0xdd09b62cU,
    0x82375f18U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x10000000U,
    0x18000000U,
    0x8c000000U,
    0x52000000U,
    0xd7000000U,
    0x31800000U,
    0xdfc00000U,
    0x80200000U,
    0x5cb00000U,
    0x93180000U,
    0xa27c0000U,
    0x44ae0000U,
    0xeae10000U,
    0x92158000U,
    0x45ef4000U,
    0xb8802000U,
    0x2b443000U,
    0x59674800U,
    0x22d52400U,
    0x88490a00U,
    0x61f31300U,
    0x7b788a80U,
    0x482fc440U,
    0xafa462a0U,
    0xa6709050U,
    0x5bf8098U,
    0x188dd754U,
    0x1c51681eU,
    0x4a0c1437U,
    0x9c15c22aU,
    0x78e87712U,
    0x2204209eU,
    0x9f07a761U,
    0x45838018U,
    0x11c24005U,
    0x1d23a030U,
    0x3636700cU,
    0x2f5ce813U,
    0x2a9f5426U,
    0xa9bbe205U,
    0xfa8d4736U,
    0x2356e897U,
    0xff8dc356U,
    0x85d2aa1cU,
    0xf1c96314U,
    0xca326286U,
    0x355d9076U,
    0x899c009dU,
    0x563f974bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x8000000U,
    0x7c000000U,
    0x32000000U,
    0xf7000000U,
    0x38800000U,
    0xd2400000U,
    0x31a00000U,
    0x72100000U,
    0xd9c80000U,
    0xc9ac0000U,
    0x32da0000U,
    0x3de70000U,
    0x4eb38000U,
    0xb85dc000U,
    0xd8a72000U,
    0xe7943000U,
    0x808c800U,
    0x38cb6c00U,
    0x202aae00U,
    0xd11a3b00U,
    0x3b829f80U,
    0x64c0abc0U,
    0x86e4f7a0U,
    0x9c3537f0U,
    0x491a3188U,
    0x8f8190f4U,
    0x8ac7e80aU,
    0x53e35c2bU,
    0x5bb06626U,
    0xdfda5726U,
    0x5e61b180U,
    0x3e7050d6U,
    0xc0bac82eU,
    0x99906c25U,
    0x150b2e1cU,
    0x234cfb1eU,
    0x94ec3fb4U,
    0xfffe5bfeU,
    0xcbb09f85U,
    0x17dbabdeU,
    0x826577afU,
    0x1c73f7d8U,
    0x3fbc91a1U,
    0xdd1360eeU,
    0xf549800cU,
    0xe5eac027U,
    0xde7ca020U,
    0xf475f03eU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x50000000U,
    0x98000000U,
    0xa4000000U,
    0x62000000U,
    0xf3000000U,
    0x70800000U,
    0xf3400000U,
    0x1e200000U,
    0xaf00000U,
    0x20880000U,
    0xcc6c0000U,
    0xb87e0000U,
    0xf7e50000U,
    0x8938000U,
    0xb2dd4000U,
    0x4c576000U,
    0xf538f000U,
    0xccc6a800U,
    0x1be78c00U,
    0x16944600U,
    0xe7db4f00U,
    0xfdd4ec80U,
    0x1df91ac0U,
    0xf520a4a0U,
    0x9e7226d0U,
    0xd1c8aab8U,
    0x714e55c4U,
    0x5a0a481eU,
    0x802e3c2dU,
    0xb8db8e26U,
    0xf3513301U,
    0xebbd82a2U,
    0xc28499e9U,
    0xd8416e23U,
    0xaaa18312U,
    0xcbb7ca8bU,
    0x55aaa5d9U,
    0x121ae027U,
    0x9b0b035U,
    0xd6aa4824U,
    0xca9e3c0aU,
    0x96f38e05U,
    0x968d3339U,
    0xf56b82bdU,
    0x57fd99e9U,
    0x2a24ee11U,
    0xd0f5c31dU,
    0x478d2ab8U,
    0x16ea15d3U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0x78000000U,
    0x3c000000U,
    0xd2000000U,
    0x8d000000U,
    0x29800000U,
    0x12c00000U,
    0xd1600000U,
    0xdc300000U,
    0xa3180000U,
    0x863c0000U,
    0x552e0000U,
    0xd4270000U,
    0x8908000U,
    0x6ec94000U,
    0x5c55a000U,
    0x94aeb000U,
    0x92e61800U,
    0x1ff23400U,
    0xd9ff7200U,
    0xcdc93100U,
    0x4ad3ac80U,
    0x72ebfa40U,
    0x70c694a0U,
    0x44653e50U,
    0x99b4dea8U,
    0x6bdecb4cU,
    0x9e5b3836U,
    0x4e99c403U,
    0x3afaca22U,
    0xfb4eb530U,
    0xdc9446a0U,
    0xe8cebf68U,
    0x17576a2aU,
    0xd62c0500U,
    0xb2a45e9eU,
    0x96d78b4eU,
    0xb0ee9837U,
    0xd5c7743cU,
    0x59e4d203U,
    0x1d708130U,
    0xd9bd3494U,
    0x34ec8e40U,
    0x5bc24691U,
    0x56e5bf7dU,
    0xb1f1ea0fU,
    0xa6fe4538U,
    0x194f7eb7U,
    0x9977b79U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x68000000U,
    0x6c000000U,
    0x5e000000U,
    0xc7000000U,
    0xde800000U,
    0xd2400000U,
    0x77e00000U,
    0x48500000U,
    0x45080000U,
    0x4f2c0000U,
    0x975e0000U,
    0x5f210000U,
    0x2ff78000U,
    0x3bbec000U,
    0xff6a000U,
    0xcbb8d000U,
    0xd7f4a800U,
    0x9fbabc00U,
    0x8df61e00U,
    0x6ab83900U,
    0xca75df80U,
    0xa17e0dc0U,
    0xb15657a0U,
    0x4c8fa1d0U,
    0xcbebc198U,
    0xefa34c4U,
    0x56958802U,
    0x89acac1fU,
    0xb11c1635U,
    0xc2c6553eU,
    0xaaa169baU,
    0x3d3188deU,
    0xb91c161dU,
    0xbec65507U,
    0x7ca16983U,
    0xc63188e1U,
    0x519c1613U,
    0xc786551dU,
    0x8bc16990U,
    0x9b2188efU,
    0xbdf41620U,
    0x12ba5514U,
    0x2e77698fU,
    0xc37c88ecU,
    0x40559609U,
    0x39099523U,
    0x992849b3U,
    0x6c5b98f5U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0xb8000000U,
    0x7c000000U,
    0x3e000000U,
    0x51000000U,
    0x2a800000U,
    0x8400000U,
    0x62200000U,
    0xa4f00000U,
    0xf3b80000U,
    0x845c0000U,
    0xdd4a0000U,
    0x1ce70000U,
    0xc0948000U,
    0x7c6f4000U,
    0x7595a000U,
    0x24e9d000U,
    0xad19800U,
    0xb14bac00U,
    0xae75e00U,
    0x5906700U,
    0x5cef4480U,
    0xd6d34340U,
    0x7f4dfca0U,
    0xe3e57f70U,
    0x53101a88U,
    0x6aaf244cU,
    0xe5f0b826U,
    0xf13d3c13U,
    0x181b6609U,
    0xb56e1b37U,
    0x131382abU,
    0xaaf8849U,
    0xb5f1663dU,
    0xb9391b0fU,
    0xdc1f0286U,
    0xf76cc876U,
    0x7c16c61dU,
    0x712bcb20U,
    0x97301a9cU,
    0xd35f2451U,
    0x1ac8b81eU,
    0xa0213c3cU,
    0xbf16610U,
    0x28391b3bU,
    0x569f0285U,
    0xf2cc858U,
    0xa636c617U,
    0xa9dbcb12U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xd0000000U,
    0xa8000000U,
    0xc000000U,
    0x5e000000U,
    0x63000000U,
    0x75800000U,
    0xebc00000U,
    0xd8600000U,
    0x9a300000U,
    0xf6980000U,
    0xd5bc0000U,
    0x5eae0000U,
    0x2e230000U,
    0x45918000U,
    0x9acec000U,
    0xa217a000U,
    0xac0dd000U,
    0x1cf08800U,
    0xd07ec400U,
    0x7dcbaa00U,
    0x1592d100U,
    0x72cdfc80U,
    0xe115e40U,
    0x220d54a0U,
    0xd7f38a50U,
    0xa9fe5688U,
    0xc80f8f64U,
    0xaef6a822U,
    0x9d7dd427U,
    0x134c8215U,
    0x2fd1c52fU,
    0x136edea0U,
    0x71414b61U,
    0x5825021cU,
    0x6a930514U,
    0x114f7ebeU,
    0xfad39b5dU,
    0xa9ea0a16U,
    0x6100010bU,
    0xa082f484U,
    0x51425a74U,
    0xc820deb9U,
    0x62924b6fU,
    0xcd4c8217U,
    0xcd1c523U,
    0xc6eede92U,
    0x4a814b62U,
    0x28450238U,
    0xfca30514U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x30000000U,
    0x88000000U,
    0xd4000000U,
    0xaa000000U,
    0xdb000000U,
    0x24800000U,
    0x20c00000U,
    0xa7600000U,
    0x36b00000U,
    0x25380000U,
    0xa31c0000U,
    0x78e0000U,
    0x4c270000U,
    0x6f928000U,
    0xd72cc000U,
    0x5f776000U,
    0x61ddb000U,
    0xcfe82800U,
    0xe4152c00U,
    0x969ba00U,
    0xabd73900U,
    0x308f1880U,
    0xdea49f40U,
    0xca54d0a0U,
    0x15cbc350U,
    0x3e86a2a8U,
    0x33c3a65cU,
    0x77e3c80aU,
    0x8c735c29U,
    0xd15c7200U,
    0x652f650fU,
    0xa877ea81U,
    0x935c3a4bU,
    0x7a28da2aU,
    0x9ef18923U,
    0x349bb0a1U,
    0x87ca7361U,
    0xf9808aa3U,
    0x49418a53U,
    0x1620f211U,
    0x5c94a524U,
    0x37aa0aaeU,
    0x6db14a4fU,
    0x41b9923eU,
    0xa3de150aU,
    0x90e8a284U,
    0xf294a651U,
    0x9ea9481eU,
    0xde339c17U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0x88000000U,
    0xfc000000U,
    0xe2000000U,
    0x7000000U,
    0x5c800000U,
    0x8b400000U,
    0x5da00000U,
    0xdeb00000U,
    0x9b680000U,
    0x400c0000U,
    0xedde0000U,
    0x17630000U,
    0x67d48000U,
    0x71bac000U,
    0x46342000U,
    0x662eb000U,
    0xd4a9c800U,
    0x72e84c00U,
    0x28c9c600U,
    0xb3bb3300U,
    0x71339280U,
    0x12afdbc0U,
    0xd3ecfaa0U,
    0x454ce7d0U,
    0xd7a5488U,
    0x9654e8ccU,
    0xbd7f682aU,
    0xfe533c33U,
    0x917eae32U,
    0xe4540f22U,
    0xe27b3c8dU,
    0xa6d4d4dcU,
    0x8c3d4618U,
    0xa0f1f304U,
    0x854fb28eU,
    0xed7d6bd9U,
    0x46533288U,
    0x457babcdU,
    0x8a5112b4U,
    0x8f7a1bc4U,
    0x1525a8eU,
    0xb9fb97d1U,
    0x71133cb8U,
    0x5ad8d4dcU,
    0x23e3460dU,
    0xc092f329U,
    0x361b32aaU,
    0xeb87abc6U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0x78000000U,
    0x54000000U,
    0x56000000U,
    0x6f000000U,
    0x4d800000U,
    0xac400000U,
    0x2b200000U,
    0x1a700000U,
    0x70680000U,
    0x2a1c0000U,
    0xfa720000U,
    0x406d0000U,
    0x32188000U,
    0x5e764000U,
    0x6e6c6000U,
    0x9189000U,
    0x45f78800U,
    0xad2d7400U,
    0x6fbf9e00U,
    0xf3c3b300U,
    0xf664fe80U,
    0x5fd0cec0U,
    0x79db60a0U,
    0x9c137df0U,
    0x97bf9e38U,
    0x67c3b30cU,
    0xc064fe86U,
    0xc0d0cee9U,
    0x4c5b60abU,
    0x64537dfeU,
    0xea9f9e3eU,
    0x12b3b33cU,
    0xfd8cfe9fU,
    0x468ccec4U,
    0x9d0960bdU,
    0x3e4e7df2U,
    0xa8ef1e1cU,
    0x66d9f326U,
    0x69929eadU,
    0xff95ec3U,
    0xeae668aaU,
    0xcd1549e7U,
    0xa93ce016U,
    0x9c02d00fU,
    0xda01e827U,
    0xfd04e414U,
    0xfc82962cU,
    0xa2c5873cU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x78000000U,
    0x74000000U,
    0xba000000U,
    0x5000000U,
    0xb4800000U,
    0xb2400000U,
    0x56e00000U,
    0xf6100000U,
    0xd480000U,
    0xdb5c0000U,
    0x86160000U,
    0x854f0000U,
    0x67598000U,
    0x20154000U,
    0x364f6000U,
    0x18dff000U,
    0x99d0b800U,
    0x636aec00U,
    0xbeefc600U,
    0x862f9d00U,
    0x158beb80U,
    0x18fa1cc0U,
    0x53642da0U,
    0xaad581d0U,
    0x5cefc608U,
    0xa72f9d14U,
    0xc30bebb6U,
    0xcbba1cd9U,
    0x73042d90U,
    0x9f8581ccU,
    0x9c7c60bU,
    0x3d239d39U,
    0xaa35eb8eU,
    0xd1f91cc8U,
    0xc9e3ad86U,
    0x1793c1c7U,
    0xd38f2629U,
    0x5bfa2d2eU,
    0xe4e5b3b6U,
    0xf1640e1U,
    0xbfca3392U,
    0xca1c00c9U,
    0xd774d382U,
    0xbe9ab0f0U,
    0x85350b9fU,
    0x587cacc8U,
    0x1b25f590U,
    0x59339deaU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x68000000U,
    0xa4000000U,
    0xe6000000U,
    0xaf000000U,
    0x9d800000U,
    0xea400000U,
    0xf600000U,
    0xb1900000U,
    0x8b480000U,
    0xf2dc0000U,
    0x8d920000U,
    0x61490000U,
    0xffd88000U,
    0x2912c000U,
    0xd18f6000U,
    0x237a9000U,
    0x9ba08800U,
    0x4b36f400U,
    0x76fd2e00U,
    0x85e1a100U,
    0x5a540380U,
    0xd46f78c0U,
    0xb9292da0U,
    0xafced9f0U,
    0x821d2e18U,
    0xb131a12cU,
    0xa3fc03a2U,
    0xd6378f5U,
    0xd8932d83U,
    0x29cbd9f2U,
    0x9d1fae06U,
    0xa4b6613dU,
    0x9db9e386U,
    0x8c0228e3U,
    0x620445a2U,
    0xd9057de2U,
    0x2a856807U,
    0xbbc7a419U,
    0xa722461aU,
    0xf7f30531U,
    0x85cc589U,
    0xe57bdd0U,
    0x9a6a082dU,
    0x522d341fU,
    0xa44ace3bU,
    0xaf59f12bU,
    0x87d3ebbbU,
    0x1e2f1cefU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0xc8000000U,
    0x14000000U,
    0xce000000U,
    0x35000000U,
    0xd3800000U,
    0xcac00000U,
    0x52200000U,
    0x18f00000U,
    0xe2c80000U,
    0x623c0000U,
    0xacf60000U,
    0xbcc90000U,
    0x4f3a8000U,
    0x1375c000U,
    0x640fe000U,
    0xf21c5000U,
    0x23021800U,
    0xaa80d400U,
    0xdb409a00U,
    0xfe0df00U,
    0xbe538280U,
    0xc79cbc0U,
    0x309318a0U,
    0xa85914d0U,
    0xa1609a38U,
    0xf310df3cU,
    0x3a9b82aaU,
    0xff45cbe9U,
    0x49e518a1U,
    0xff5014d8U,
    0xa1fa1a3fU,
    0x7551f03U,
    0x3dfc628aU,
    0xbd559bddU,
    0x76f9008dU,
    0x93d5c0c6U,
    0x7bbe002cU,
    0xc5350034U,
    0x46c8032U,
    0xec8cc036U,
    0x4e5d6027U,
    0x7065902eU,
    0x4693f805U,
    0x6159841eU,
    0x78e60226U,
    0xbad0cb36U,
    0x523878bfU,
    0x24f584e6U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0x8000000U,
    0x34000000U,
    0xca000000U,
    0xd3000000U,
    0xbf800000U,
    0x51400000U,
    0xc200000U,
    0x6a700000U,
    0x8a680000U,
    0x341c0000U,
    0x4760000U,
    0xdb6b0000U,
    0x4c9d8000U,
    0x30b6c000U,
    0x5d0ae000U,
    0xa14a7000U,
    0x4bee0800U,
    0x3c5d8400U,
    0x44d70600U,
    0x4fd9e300U,
    0x93943680U,
    0x56fd46c0U,
    0x736330a0U,
    0x1754a5d0U,
    0x889f0628U,
    0x82b5e314U,
    0x820a368aU,
    0xf8ca46c9U,
    0x3fa8b090U,
    0xa2f965e9U,
    0xd9606635U,
    0xf4555334U,
    0x2f18dea6U,
    0xeff6b2c7U,
    0x702c3eb1U,
    0x8bbbc2feU,
    0xd941b696U,
    0x782786cfU,
    0x8077d094U,
    0x4969d5edU,
    0x839a8e3dU,
    0x6135a71eU,
    0x1d4a50b1U,
    0xf5ef15f5U,
    0x5586e2dU,
    0x3853d701U,
    0xa91a5885U,
    0xfaf591d1U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0xf8000000U,
    0xb4000000U,
    0xa6000000U,
    0x9d000000U,
    0xfe800000U,
    0xaf400000U,
    0x69200000U,
    0x45100000U,
    0x71d80000U,
    0x38cc0000U,
    0x7b120000U,
    0x8d90000U,
    0x784f8000U,
    0x6d514000U,
    0xc17b6000U,
    0xbb1f1000U,
    0x2d289800U,
    0x8ee47c00U,
    0x26779e00U,
    0xf76bab00U,
    0x95409d80U,
    0x3e227840U,
    0x449703a0U,
    0x7319d370U,
    0xd12f9e18U,
    0x14e7ab14U,
    0xc1729d86U,
    0x7eeb7849U,
    0x4b8083b7U,
    0x3dc4937aU,
    0xa466fe0eU,
    0x7f31bb18U,
    0x1b4d85a3U,
    0xad24473U,
    0xbebe7db1U,
    0x3c792865U,
    0x15997b8cU,
    0x9a6aff54U,
    0xc3c47803U,
    0x3d662c1aU,
    0xafb66607U,
    0xf509c704U,
    0x77701b91U,
    0xdbecef60U,
    0xe103602eU,
    0xa483102aU,
    0xa8429803U,
    0x70a17c17U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x70000000U,
    0x8000000U,
    0x24000000U,
    0x8a000000U,
    0x11000000U,
    0x40800000U,
    0x46400000U,
    0x91200000U,
    0xb8100000U,
    0xf0680000U,
    0x727c0000U,
    0xfb160000U,
    0xbdeb0000U,
    0xf2be8000U,
    0xe375c000U,
    0x475a6000U,
    0x8903f000U,
    0x7c851800U,
    0xb042ec00U,
    0x5626c600U,
    0x4f922300U,
    0x49aae580U,
    0x7e9a25c0U,
    0xc5e423a0U,
    0x627406d0U,
    0x1fd8c628U,
    0xb345230cU,
    0x3ba2658aU,
    0xbf54e5d5U,
    0x59c8c380U,
    0xa02e36ddU,
    0x3459be1aU,
    0xec833f19U,
    0xa8413bbeU,
    0x2a26eaeaU,
    0x99960018U,
    0xfeab002fU,
    0x811e8038U,
    0x5825c03aU,
    0xe4926034U,
    0x302ff021U,
    0x2c5b1817U,
    0x9085ec2cU,
    0x7e46461cU,
    0x9d20e32cU,
    0x661005acU,
    0x636b15c6U,
    0x7fbdbb6U,
    0x77d7dadaU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x10000000U,
    0x58000000U,
    0x74000000U,
    0x66000000U,
    0x8f000000U,
    0x26800000U,
    0x6c400000U,
    0xd7600000U,
    0x90d00000U,
    0xb0480000U,
    0x8e9c0000U,
    0x5dd20000U,
    0x7c90000U,
    0xcf5b8000U,
    0xee704000U,
    0x8f386000U,
    0xc231000U,
    0xc5330800U,
    0x92db2c00U,
    0x2b64600U,
    0x5b1e5100U,
    0x7d17d180U,
    0xd1eee2c0U,
    0xf6997a0U,
    0x272cb3f0U,
    0x3e4c4618U,
    0x1d9b5134U,
    0xc15651aeU,
    0xbe8ba2d9U,
    0xfbb877afU,
    0xa966e3ceU,
    0x4bd4ae08U,
    0xcf2d0cU,
    0x25d97f92U,
    0x8034cfe6U,
    0xe9596836U,
    0x41713c09U,
    0x59bece13U,
    0x28653d1eU,
    0x3e51f78eU,
    0x100fa3d8U,
    0x5bff4e30U,
    0x7c007d2cU,
    0x6a001798U,
    0x6d05f3daU,
    0x8787a616U,
    0xe9c40101U,
    0x8fa73982U,
    0xc2f69edeU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x10000000U,
    0xa8000000U,
    0xac000000U,
    0x1a000000U,
    0x43000000U,
    0x58800000U,
    0x7e400000U,
    0xcba00000U,
    0x4f300000U,
    0x98a80000U,
    0xf19c0000U,
    0xd8360000U,
    0xbe2b0000U,
    0x4add8000U,
    0xaa13c000U,
    0x90db2000U,
    0x89103000U,
    0xf85d0800U,
    0x4f55fc00U,
    0x37fe0600U,
    0xb6672700U,
    0xf6531880U,
    0x5c7b1ac0U,
    0x8a51ea0U,
    0x57b03dd0U,
    0xc6e80628U,
    0x2a3c2714U,
    0x3f0698a2U,
    0x8a84daefU,
    0xa1403e8cU,
    0x31270debU,
    0x76768e38U,
    0xbd8d1b1bU,
    0x99e83e86U,
    0x90bb0de3U,
    0x26408e18U,
    0xbfa61b04U,
    0x6135be88U,
    0xd5a8cdd8U,
    0xf41bae0aU,
    0xbf62b32U,
    0xa48b6acU,
    0xab8d31f3U,
    0x90eda816U,
    0x33d0c06U,
    0xbc85ae2bU,
    0xb8412b3cU,
    0xaa3369cU,
    0x40b5f1faU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0xd8000000U,
    0xfc000000U,
    0xde000000U,
    0x2f000000U,
    0x74800000U,
    0xc3400000U,
    0x38a00000U,
    0x68f00000U,
    0x64480000U,
    0x76bc0000U,
    0x5ff60000U,
    0xccc90000U,
    0xbbfa8000U,
    0x20554000U,
    0xf4b82000U,
    0x3ef39000U,
    0x9f4ec800U,
    0x983cfc00U,
    0x21b72e00U,
    0xa9ef9700U,
    0xb7cd3e80U,
    0x157e6bc0U,
    0x3e1210a0U,
    0x219dfcd0U,
    0xeac12e38U,
    0x5666973cU,
    0x8c97beaeU,
    0x11db2bd3U,
    0xa8e230a1U,
    0xbad26ce4U,
    0x80f9e612U,
    0xeed36b07U,
    0x5afa10bdU,
    0x33d1fcf9U,
    0xb77f2e36U,
    0x8f139723U,
    0x1a1b3ebcU,
    0xed076bf9U,
    0x758090bcU,
    0x20c4bcfeU,
    0xc3670e0bU,
    0x95100737U,
    0xe71df6a7U,
    0xd08797e4U,
    0xa141beb9U,
    0xa9a22bc9U,
    0x8370b0bdU,
    0xb8b2cf2U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xd0000000U,
    0xe8000000U,
    0x6c000000U,
    0xee000000U,
    0x8d000000U,
    0xf6800000U,
    0xf2c00000U,
    0x39a00000U,
    0x48700000U,
    0x13c80000U,
    0x7abc0000U,
    0x51f60000U,
    0xe4890000U,
    0x42598000U,
    0x9867c000U,
    0x72552000U,
    0xbbf99000U,
    0xf016d800U,
    0xd19dfc00U,
    0xf9433200U,
    0x25e1c300U,
    0xb7141280U,
    0xd81f1fc0U,
    0xc60120a0U,
    0xc107dcd0U,
    0xa884b238U,
    0x47c30324U,
    0x4b26b2a2U,
    0x38b44fefU,
    0x496d58a1U,
    0x494d70e8U,
    0x467df83eU,
    0x55516c3bU,
    0x627a6a11U,
    0x2752ff01U,
    0x597b808bU,
    0x8d08cf4U,
    0x373eca36U,
    0xccb0af28U,
    0xd36e78afU,
    0x1e4de0c5U,
    0x87faa01cU,
    0xb617500fU,
    0x509a7813U,
    0x31c3ac2aU,
    0xb220ca01U,
    0x1435af05U,
    0x8ca9f8aaU,
    0x616f20c6U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0xb8000000U,
    0x14000000U,
    0x9e000000U,
    0xd7000000U,
    0x51800000U,
    0x25400000U,
    0xe1a00000U,
    0x95f00000U,
    0x1e80000U,
    0x6f1c0000U,
    0xf5760000U,
    0x96290000U,
    0xf4788000U,
    0x3b644000U,
    0x1e146000U,
    0xa038d000U,
    0x8440800U,
    0x1235400U,
    0x4234f600U,
    0xfa8bef00U,
    0x510e5680U,
    0xc549a6c0U,
    0xd66ca0a0U,
    0x595b49d0U,
    0xf3527638U,
    0xad9aaf2cU,
    0x77b4b6b6U,
    0xb9cc36f9U,
    0x13ac48a9U,
    0x6f388de0U,
    0x9dc0e80bU,
    0x9263c410U,
    0x28929e1aU,
    0xcf96b15U,
    0x28a628b1U,
    0x53755dceU,
    0x152a602eU,
    0x5bfdd012U,
    0x39228834U,
    0x1632143bU,
    0x48e162cU,
    0x760e7f1dU,
    0x2ccebe9fU,
    0xe72a62e3U,
    0x26fe3ea6U,
    0xb1a222ccU,
    0xfdf45ebeU,
    0x3deff2e0U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x98000000U,
    0x14000000U,
    0x2000000U,
    0x27000000U,
    0xcc800000U,
    0x8c400000U,
    0x58a00000U,
    0x33d00000U,
    0x2a680000U,
    0x22bc0000U,
    0x2a520000U,
    0xc5ab0000U,
    0x16df8000U,
    0xb604000U,
    0x39b76000U,
    0xd25b3000U,
    0x6b276800U,
    0x7153c00U,
    0x8c8fb600U,
    0x26cd0f00U,
    0x736ac780U,
    0x573d6bc0U,
    0x639771a0U,
    0x3ccb64f0U,
    0x806a3608U,
    0x39ba4f04U,
    0x18d0278eU,
    0x30ed1bf1U,
    0x2bf8f986U,
    0x497528fcU,
    0x2ebd883bU,
    0xfc554c2eU,
    0x40a83e2bU,
    0x2d5f431eU,
    0xd3a779baU,
    0xf95568c8U,
    0xbb2ae813U,
    0x559e7c23U,
    0xc3475602U,
    0x8267f28U,
    0x9192cfb1U,
    0x93cf67f7U,
    0x60edafb4U,
    0x73f857f5U,
    0xfd70c78dU,
    0xfcba6bdcU,
    0x4352f18cU,
    0x982c24e7U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0x98000000U,
    0xac000000U,
    0x12000000U,
    0xb1000000U,
    0xe0800000U,
    0x6fc00000U,
    0x9200000U,
    0x83b00000U,
    0xb1280000U,
    0xc19c0000U,
    0x79320000U,
    0xfbed0000U,
    0x26398000U,
    0x7e404000U,
    0x2664a000U,
    0x55507000U,
    0x73bf5800U,
    0xc7072c00U,
    0xb7878a00U,
    0xe0452f00U,
    0xf5605880U,
    0x2cd12ec0U,
    0x18fdd2a0U,
    0x87e501f0U,
    0x32160a38U,
    0x91596f04U,
    0xbb96f8beU,
    0x121c5edfU,
    0x7f730a82U,
    0x618e6df7U,
    0x996f2004U,
    0xc0fd3033U,
    0x8be27804U,
    0x90171c24U,
    0x685c720cU,
    0x6f127317U,
    0xc3d88a8eU,
    0xd5532df9U,
    0xb3ba000dU,
    0xa701002aU,
    0x67838036U,
    0x7841402eU,
    0x59672001U,
    0x3ed13001U,
    0xa9f87822U,
    0x67661c04U,
    0x5dd7f228U,
    0x987f3322U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0xb8000000U,
    0x34000000U,
    0x7e000000U,
    0x6d000000U,
    0x65800000U,
    0x94400000U,
    0x67a00000U,
    0xfb100000U,
    0xd7080000U,
    0xcf1c0000U,
    0xd8920000U,
    0xda4b0000U,
    0x13398000U,
    0xcac74000U,
    0xb7676000U,
    0x8777d000U,
    0xdcfdd800U,
    0x34202400U,
    0xfe53de00U,
    0x972a0700U,
    0x2f482180U,
    0x92b92ec0U,
    0x3881ffa0U,
    0x99c429f0U,
    0xdfe25e08U,
    0xc6b14734U,
    0x871d4186U,
    0x2495fec9U,
    0x984da791U,
    0x9c3f4de6U,
    0x3044e00eU,
    0x61a7900aU,
    0x2113806U,
    0xdc8bb412U,
    0x8e58e61dU,
    0xeeb6b31bU,
    0xcb1b4785U,
    0xce93ddfbU,
    0x9b4c5812U,
    0x2cbb642fU,
    0xf586be2dU,
    0xec46d71aU,
    0xf3a479b7U,
    0x95124ae4U,
    0x20f4185U,
    0x9e9efec7U,
    0x32d4278cU,
    0xd0e80dc3U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0x18000000U,
    0x24000000U,
    0x6a000000U,
    0xaf000000U,
    0xc3800000U,
    0x2a400000U,
    0xa0600000U,
    0xed100000U,
    0x8b180000U,
    0x70c0000U,
    0xca960000U,
    0x2b5f0000U,
    0xd86a8000U,
    0x8c054000U,
    0xb606e000U,
    0x3104f000U,
    0x6e856800U,
    0x7ac28400U,
    0x7a53e00U,
    0xa2327900U,
    0xaae90f80U,
    0x88c07e40U,
    0xcca231a0U,
    0xabb10750U,
    0x5fafbe08U,
    0xf327390cU,
    0x2ff7efaeU,
    0xeac88e55U,
    0x1ab159bcU,
    0x712c8372U,
    0x29e00022U,
    0x5850002dU,
    0x50f80036U,
    0x945c0024U,
    0x93ee0016U,
    0x4a430018U,
    0x3064803aU,
    0x2516402aU,
    0xc71a601cU,
    0x510eb018U,
    0x2b910811U,
    0x2ddf3415U,
    0x9ea8b626U,
    0xc5a70d19U,
    0xd135d99eU,
    0xf76ac35fU,
    0xf826010U,
    0x3c42b035U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x70000000U,
    0xe8000000U,
    0xac000000U,
    0x12000000U,
    0x5000000U,
    0x68800000U,
    0x13400000U,
    0x73e00000U,
    0xb7d00000U,
    0xfe880000U,
    0xda5c0000U,
    0x17560000U,
    0x71cf0000U,
    0xf3ba8000U,
    0x7982c000U,
    0x1dc52000U,
    0x93a37000U,
    0x67b47800U,
    0xcd994c00U,
    0x85f22e00U,
    0x60f9bd00U,
    0x4a644d80U,
    0xa1014c0U,
    0x1d2863a0U,
    0x55eaa9d0U,
    0x76c8ae08U,
    0xe63b7d0cU,
    0x96416d92U,
    0x5b6364f7U,
    0x4941b9eU,
    0xfd6fe5faU,
    0x858c8016U,
    0x45ddc009U,
    0xb997a014U,
    0xe1edb01eU,
    0x60cf581dU,
    0xfd393c31U,
    0x99c2d624U,
    0xcda1312fU,
    0xa0b7c3bfU,
    0x381b19ebU,
    0xba31f638U,
    0xbe5d4131U,
    0x99513b8cU,
    0x2ecc95e1U,
    0x4238f81bU,
    0xf8448c12U,
    0xd4658e24U,
    0x2d140d3dU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xf0000000U,
    0x88000000U,
    0xc4000000U,
    0xa000000U,
    0x8f000000U,
    0x41800000U,
    0xc0c00000U,
    0xd2a00000U,
    0x39d00000U,
    0x43f80000U,
    0xc32c0000U,
    0xc8520000U,
    0x6b3b0000U,
    0x858e8000U,
    0x8381c000U,
    0xebc16000U,
    0xc924f000U,
    0xe17e800U,
    0x9cdc1400U,
    0xf43e7a00U,
    0xdc0ebb00U,
    0x5f462f80U,
    0x8f635b40U,
    0x39f255a0U,
    0x4d6ae070U,
    0x5bb0fa08U,
    0xe08f7b0cU,
    0xdd074f8aU,
    0xb287ab4dU,
    0x6745bd80U,
    0xa366f462U,
    0xa7f6801aU,
    0xb06dc01fU,
    0xd933602eU,
    0xafcff033U,
    0x4a616804U,
    0x9771d42cU,
    0x9c2d1a30U,
    0x31d14b39U,
    0x47ff47b7U,
    0x692e8f44U,
    0xb7554f9fU,
    0xa2bcab5cU,
    0x814b3d93U,
    0x5b273473U,
    0x5d17e007U,
    0xcb593017U,
    0xdfc8839U,
    0x862fe428U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xb0000000U,
    0xd8000000U,
    0x6c000000U,
    0x36000000U,
    0x9f000000U,
    0x50800000U,
    0xfdc00000U,
    0xc6600000U,
    0xfd100000U,
    0x3df80000U,
    0x106c0000U,
    0x90560000U,
    0x26590000U,
    0xfb1a8000U,
    0x38bd4000U,
    0xbfc8a000U,
    0x21225000U,
    0x27b70800U,
    0xad0d5c00U,
    0x4fc0fa00U,
    0x9b633500U,
    0x8967780U,
    0x89bba540U,
    0x3c4c0da0U,
    0x6a65d050U,
    0xeb12da38U,
    0x72fc253cU,
    0x28e9dfaeU,
    0xd994a977U,
    0xba3bff9bU,
    0xaf0bb96cU,
    0xcac457a1U,
    0x2e4b564U,
    0x8ad3a5abU,
    0x3a9adc67U,
    0x96fd283dU,
    0x5aee4c3bU,
    0xa491521cU,
    0x9fb93922U,
    0x734d0584U,
    0x52e18c7bU,
    0xa2d0a016U,
    0xee9e5035U,
    0x7cf90830U,
    0x2be85c02U,
    0x7147a08U,
    0x4fb752aU,
    0xd7ea57a6U,
    0x3911b568U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0xb8000000U,
    0xe4000000U,
    0xee000000U,
    0x43000000U,
    0xf0800000U,
    0x90400000U,
    0x30a00000U,
    0xcf700000U,
    0xa6680000U,
    0xf99c0000U,
    0x1fb20000U,
    0x7e8f0000U,
    0xc24c8000U,
    0xa8ae4000U,
    0xda79e000U,
    0x30e2d000U,
    0xb7d62800U,
    0x151ffc00U,
    0x65f31600U,
    0xf32c8f00U,
    0x36bf3280U,
    0x3206a6c0U,
    0x8900a4a0U,
    0x7d8469f0U,
    0x13c67628U,
    0xe8601f3cU,
    0x8b90faa6U,
    0xb7bb8af5U,
    0xd3859ab5U,
    0xb0c71ad9U,
    0x8e252afU,
    0x93d636c8U,
    0xdb1d6cb6U,
    0x16f645d2U,
    0xbbafc818U,
    0x42fd2c18U,
    0xeca53e0fU,
    0x5737320U,
    0x2b6c2492U,
    0x7a1a29cbU,
    0xc7779604U,
    0x3a6ecf2fU,
    0xd39cd28dU,
    0x82b776f3U,
    0x75080ca8U,
    0x468ad5c8U,
    0xe648002fU,
    0x66ac001bU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x10000000U,
    0x98000000U,
    0x9c000000U,
    0xae000000U,
    0x3d000000U,
    0xa4800000U,
    0xa4400000U,
    0xdda00000U,
    0x1f00000U,
    0x35580000U,
    0xc02c0000U,
    0xa9320000U,
    0x98bb0000U,
    0x16798000U,
    0x829e4000U,
    0xc24f2000U,
    0xdea61000U,
    0xc073d800U,
    0x9996c00U,
    0x7c9d600U,
    0xd0657700U,
    0x9490be80U,
    0xe948bb40U,
    0xb20e8a0U,
    0x56b38c70U,
    0xc17f7608U,
    0x491d2734U,
    0xe18c468eU,
    0xb6c7c763U,
    0x18e2e6adU,
    0x76d39758U,
    0xbdec1e8aU,
    0x1457eb52U,
    0x50af90b0U,
    0x7775b07aU,
    0x7219d83fU,
    0xac0e6c2cU,
    0xbc02562bU,
    0x7e003713U,
    0x5061ebcU,
    0x2880eb47U,
    0x924410adU,
    0x7ca0f071U,
    0xb777831U,
    0xac1a3c3eU,
    0xb90cae2aU,
    0xc844b23U,
    0x704590bbU,
    0xd7a2b067U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x28000000U,
    0x64000000U,
    0x1e000000U,
    0xf1000000U,
    0x41800000U,
    0xcbc00000U,
    0xdae00000U,
    0xbd500000U,
    0xfb380000U,
    0xefec0000U,
    0x9b160000U,
    0xc0190000U,
    0x2f5f8000U,
    0x19ffc000U,
    0x688f6000U,
    0xdf865000U,
    0x7ac7b800U,
    0xfb646400U,
    0x66964200U,
    0x9dd9100U,
    0x36b96980U,
    0x7e2f1940U,
    0xdef0aba0U,
    0xf5cd4850U,
    0x1226a238U,
    0x9d340114U,
    0x7b29b192U,
    0xe9712d5dU,
    0xcb0f51adU,
    0x8741bd75U,
    0x32a009afU,
    0x7f70495aU,
    0x6e089394U,
    0x80c6ec50U,
    0x5467803dU,
    0xb613c025U,
    0x9399601dU,
    0xf9f5018U,
    0x7d983806U,
    0x869ba40eU,
    0x1019220eU,
    0x275bc12bU,
    0xdfed1baU,
    0x4e8b7d5fU,
    0x6286e982U,
    0x4140d94eU,
    0xdfa7cba5U,
    0xcf71873U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xd0000000U,
    0xd8000000U,
    0x34000000U,
    0x9a000000U,
    0x57000000U,
    0xda800000U,
    0xe400000U,
    0x4da00000U,
    0x40700000U,
    0xf6980000U,
    0x6c0000U,
    0x96b20000U,
    0x767d0000U,
    0xfd398000U,
    0x711a4000U,
    0xfbae6000U,
    0x8bd2f000U,
    0xe7eb2800U,
    0xbd728c00U,
    0xb31d9600U,
    0x58ac1700U,
    0xab55f280U,
    0x6eaeb940U,
    0xf251e4a0U,
    0xc928ee70U,
    0x79927638U,
    0xbc48a704U,
    0xd402baaeU,
    0x8a03c579U,
    0xef06dab0U,
    0x3e803568U,
    0x4c4672a8U,
    0x2ea5f952U,
    0xf404abU,
    0xafdd5e5aU,
    0x974ebe06U,
    0xd8829b0cU,
    0xcd4264a6U,
    0xbd23ae72U,
    0x1137960fU,
    0x8dbd173aU,
    0xe05e72a7U,
    0xe089f94cU,
    0xbbe604acU,
    0x49d05e7eU,
    0x44ef3e3fU,
    0x9df4db39U,
    0x3a5e0487U,
    0x178c5e54U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0xb8000000U,
    0xcc000000U,
    0xea000000U,
    0x25000000U,
    0xa5800000U,
    0xba400000U,
    0x4ba00000U,
    0xc6b00000U,
    0x93480000U,
    0xaa7c0000U,
    0x2e720000U,
    0xee2f0000U,
    0xf2ab8000U,
    0x7f694000U,
    0x4788a000U,
    0x551cb000U,
    0x27a63800U,
    0x3cb62400U,
    0x2e4b8e00U,
    0x13f8d700U,
    0xc6369c80U,
    0x4c8d49c0U,
    0x7b9e92a0U,
    0x7360def0U,
    0x3d2ae28U,
    0x79e2704U,
    0x116184a6U,
    0xa2d19dc7U,
    0xec1804b4U,
    0x1027ddcaU,
    0x4ff324bbU,
    0xfa6e2de3U,
    0xf20fbcbdU,
    0x775bb9e8U,
    0x70018a8aU,
    0xa8000ae5U,
    0x5406383aU,
    0xf6062401U,
    0x77038e10U,
    0x4c84d735U,
    0xf5c49ca6U,
    0xd4e249c5U,
    0x28951296U,
    0xefb99eedU,
    0x72920e02U,
    0x42be971eU,
    0x5315bc8cU,
    0xb6f8b9faU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x50000000U,
    0x98000000U,
    0x44000000U,
    0xc6000000U,
    0xa7000000U,
    0x2b800000U,
    0xb4400000U,
    0x5c200000U,
    0x31500000U,
    0x43a80000U,
    0xbb7c0000U,
    0x9d960000U,
    0x9e4f0000U,
    0x214b8000U,
    0xec84000U,
    0xdc8de000U,
    0x77ad9000U,
    0xf57e4800U,
    0x46930400U,
    0x67c9ae00U,
    0x6c092b00U,
    0x9a6b5f80U,
    0xb41981c0U,
    0x7b6171a0U,
    0x8f4ead0U,
    0x1db9ce08U,
    0x5d73fb04U,
    0x5afb778eU,
    0xd4d355c5U,
    0x6ced77b7U,
    0xbedc55e8U,
    0xb186f7b3U,
    0xd14415c0U,
    0xb6a317b3U,
    0x599585f4U,
    0x184b5f98U,
    0x264981deU,
    0x754971a2U,
    0xf0c8ead8U,
    0x6f8fce2dU,
    0x22cfb39U,
    0xa238f78dU,
    0xf73715fdU,
    0x45de9797U,
    0x5802c5ccU,
    0xa4053fbfU,
    0x360051ccU,
    0x6f04d984U,
    0xf7857ed5U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0xd0000000U,
    0x8000000U,
    0xcc000000U,
    0xee000000U,
    0xd3000000U,
    0x52800000U,
    0x4ac00000U,
    0xa5e00000U,
    0xbed00000U,
    0x68780000U,
    0x562c0000U,
    0x51920000U,
    0x47db0000U,
    0x42df8000U,
    0xb75cc000U,
    0xf5192000U,
    0x33389000U,
    0x83882800U,
    0xbc63ac00U,
    0x2941200U,
    0xd55b3100U,
    0xa81f0580U,
    0xc2bbbf40U,
    0x43cc97a0U,
    0x97404e70U,
    0x3ba0b208U,
    0x3ef46104U,
    0x17c98daaU,
    0xdd404347U,
    0xbaa38d99U,
    0x89774377U,
    0x188e0da7U,
    0x42e08362U,
    0xd650adb7U,
    0x7bb8d34aU,
    0x804ba582U,
    0xc204ef79U,
    0x8d021fabU,
    0xf987b255U,
    0xc45a808U,
    0xc5246c02U,
    0xea32b21bU,
    0xb92f613aU,
    0xf5160d97U,
    0xba1c835dU,
    0x47baad98U,
    0x764fd34aU,
    0x750625bcU,
    0x2d832f59U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0xe8000000U,
    0xb4000000U,
    0x22000000U,
    0x81000000U,
    0x21800000U,
    0x56400000U,
    0xc2200000U,
    0xa9100000U,
    0xe6480000U,
    0x2fdc0000U,
    0x5e560000U,
    0xb5ef0000U,
    0x988e8000U,
    0x3e3ec000U,
    0xf921e000U,
    0x15943000U,
    0xa788d800U,
    0xe8bc6400U,
    0x7de64e00U,
    0x39731b00U,
    0xe77dcd80U,
    0xa5810cc0U,
    0x7c4303a0U,
    0xa723d7d0U,
    0xb291ae08U,
    0x6d082b2cU,
    0x5a7b9592U,
    0x203a8d1U,
    0x3104ad92U,
    0x6984fcddU,
    0x1244bb88U,
    0x82543d9U,
    0x9c16d80aU,
    0xe5cf643eU,
    0xf89ece13U,
    0xbdf2db10U,
    0x4abaad94U,
    0xbce7fcfeU,
    0xb8f43ba1U,
    0x413883edU,
    0x8fa7b811U,
    0x61569412U,
    0x636f761bU,
    0x1c4b4f22U,
    0x12db5b8cU,
    0xb9d273d0U,
    0x58ae8009U,
    0xa62ec025U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x90000000U,
    0xb8000000U,
    0x1c000000U,
    0x1a000000U,
    0xd1000000U,
    0x3800000U,
    0x65c00000U,
    0x27a00000U,
    0xc0500000U,
    0xe3b80000U,
    0x516c0000U,
    0x4960000U,
    0x8f9b0000U,
    0xe0f98000U,
    0xee8f4000U,
    0x71a22000U,
    0xd357b000U,
    0x9d3e1800U,
    0x52b3400U,
    0x33f0c200U,
    0x6defab00U,
    0x6dd0fb80U,
    0xa97fe540U,
    0x824fb9a0U,
    0xf9840e50U,
    0x44c4e228U,
    0x2c231b34U,
    0x919763a6U,
    0x7a1b9163U,
    0x463d5b88U,
    0x2fac1563U,
    0x69b201aeU,
    0xa58bca59U,
    0xe326180fU,
    0x9517341aU,
    0x4f5ec211U,
    0x5ad8ab33U,
    0xc1f7b91U,
    0xc53ba55dU,
    0xe92c1996U,
    0x21f0fe55U,
    0x88eeda3dU,
    0xd0549f06U,
    0x1bb839adU,
    0x6d6c4e51U,
    0x8e914222U,
    0xe69ceb35U,
    0xff7cdbbcU,
    0x914f5554U,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0x48000000U,
    0xec000000U,
    0xb2000000U,
    0x3000000U,
    0xf3800000U,
    0x70c00000U,
    0x3f200000U,
    0x9c300000U,
    0xec580000U,
    0x28ec0000U,
    0x9bf60000U,
    0x76fb0000U,
    0xb91f8000U,
    0x5a0ec000U,
    0xa622000U,
    0x9ed37000U,
    0x65cc4800U,
    0x64c42400U,
    0x1926aa00U,
    0xf9319700U,
    0x5add8c80U,
    0xbea9c140U,
    0x7a92a6a0U,
    0xe7ad9650U,
    0x76128a28U,
    0xbee9e704U,
    0x86f6449aU,
    0x147f256fU,
    0x69582c9eU,
    0xae68716bU,
    0x35b2ceb8U,
    0x639dc258U,
    0x3e49e809U,
    0xc8059404U,
    0xac06c202U,
    0x9201c302U,
    0x5306ee9fU,
    0xbb85b270U,
    0x9cc22037U,
    0x8d237020U,
    0x9f34482cU,
    0x1fd82414U,
    0x5828aa1bU,
    0xa4d69735U,
    0xeacc0c80U,
    0x55400166U,
    0x72e10686U,
    0x91972671U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0xa8000000U,
    0x4000000U,
    0x2a000000U,
    0xcb000000U,
    0x32800000U,
    0x68400000U,
    0xada00000U,
    0xabf00000U,
    0x2ba80000U,
    0xfddc0000U,
    0x66b20000U,
    0xb58f0000U,
    0x97ee8000U,
    0x627cc000U,
    0x8a466000U,
    0xf2a5f000U,
    0x4b70c800U,
    0x54e92c00U,
    0xe4fae600U,
    0x40030700U,
    0xe001ae80U,
    0x50020cc0U,
    0xd807c8a0U,
    0xac02cbf0U,
    0x2e068628U,
    0xe105f72cU,
    0xf985e6a2U,
    0x5ac4e0ddU,
    0xc5e7ce80U,
    0x657fcc3U,
    0x805f00aeU,
    0xd677e7c4U,
    0x9b6e6025U,
    0xd339f01dU,
    0x2262c809U,
    0xf5962c3aU,
    0xe83c663fU,
    0x78e3c73bU,
    0xb9d5ce91U,
    0x1f98fcf7U,
    0xb011808aU,
    0xa4fb27dbU,
    0xa0000020U,
    0xb000003dU,
    0x8800003fU,
    0x74000012U,
    0x8200003fU,
    0xcf00003eU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xd0000000U,
    0x38000000U,
    0x7c000000U,
    0x6000000U,
    0xf7000000U,
    0xae800000U,
    0x79400000U,
    0x9d600000U,
    0x22500000U,
    0xc5880000U,
    0x625c0000U,
    0x86120000U,
    0xff690000U,
    0xfccd8000U,
    0xdf3d4000U,
    0x94472000U,
    0xd2e45000U,
    0xda97c800U,
    0x232e6c00U,
    0x9a28ae00U,
    0x77a94900U,
    0xd26a0580U,
    0x534db8c0U,
    0xf7fd2ba0U,
    0x4ae0b1f0U,
    0x56958e18U,
    0x2d281904U,
    0xf92a4db6U,
    0x9b2b94ebU,
    0x262d25afU,
    0x91a9e8d2U,
    0xf56ae3b2U,
    0xc5cedde7U,
    0xf2bd201aU,
    0xd181500dU,
    0x83c04821U,
    0x46262c06U,
    0xe2300e34U,
    0x8059591bU,
    0xfb176d93U,
    0xa8eac4e3U,
    0x480d6d87U,
    0xd79fc4fdU,
    0xdf32edbcU,
    0xf7db84d7U,
    0x7fd04da9U,
    0x484e94deU,
    0x477aa5beU,
    0x88a1a8d3U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x70000000U,
    0xd8000000U,
    0x84000000U,
    0xd6000000U,
    0xc5000000U,
    0x92800000U,
    0xb6400000U,
    0xdb200000U,
    0xa0b00000U,
    0x3380000U,
    0x600c0000U,
    0x79f60000U,
    0xd9d0000U,
    0xbf788000U,
    0x6c2b4000U,
    0xb8c3e000U,
    0x9f641000U,
    0xf892d800U,
    0x5c0bdc00U,
    0x1bf17600U,
    0x269dd300U,
    0x1cfcd080U,
    0x46ef9240U,
    0x94a326a0U,
    0xa0740150U,
    0x285c9618U,
    0x7698c30cU,
    0x14f8888eU,
    0x8aee0e7dU,
    0x6ea730a3U,
    0xef778240U,
    0x2ddffe89U,
    0x415edd50U,
    0x2e1b603cU,
    0xd5bf5028U,
    0xa049380dU,
    0xf7d3cc3cU,
    0x97adae39U,
    0x56070f26U,
    0x85032696U,
    0x7284016cU,
    0xc6449633U,
    0x324c331U,
    0x24b68880U,
    0xd53f0e5fU,
    0xa509b09fU,
    0xeb71c26fU,
    0xbbdc9eb3U,
    0x645d8d5aU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0xc8000000U,
    0x8c000000U,
    0xa000000U,
    0xfd000000U,
    0x44800000U,
    0x35400000U,
    0xff200000U,
    0xd3b00000U,
    0x38880000U,
    0x88bc0000U,
    0x94f20000U,
    0x2eab0000U,
    0xf18d8000U,
    0x7238c000U,
    0x22b52000U,
    0xb60bb000U,
    0xe0fc7800U,
    0xbf516c00U,
    0x5afe00U,
    0xba200b00U,
    0x23303880U,
    0x73cc1fc0U,
    0x549d46a0U,
    0xb0c3d4f0U,
    0x1ce7de08U,
    0xd7bb14U,
    0x469e409aU,
    0x69c673c7U,
    0xee623890U,
    0x9a971fecU,
    0x4438c69bU,
    0x4db714e6U,
    0xeb88fe14U,
    0x873b0b0eU,
    0x4a35b8a9U,
    0x1948dfeeU,
    0x2ada66aaU,
    0xa46364c8U,
    0x796263dU,
    0x30be1705U,
    0x20f19e8eU,
    0x50adc8edU,
    0xd28e7834U,
    0x85ba6c33U,
    0x28777e14U,
    0xcfe8cb22U,
    0x40ad1885U,
    0xda8bafd2U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0x78000000U,
    0x8c000000U,
    0x86000000U,
    0x53000000U,
    0x93800000U,
    0x2bc00000U,
    0x68e00000U,
    0xb0f00000U,
    0x1a380000U,
    0x5c4c0000U,
    0x45360000U,
    0xdddd0000U,
    0xb13d8000U,
    0x3cb4000U,
    0x8f6a000U,
    0x363bd000U,
    0xea48e800U,
    0xbe312400U,
    0xba5b5200U,
    0x90fd5100U,
    0xbe2fb480U,
    0x7884ac40U,
    0x944766a0U,
    0xf523bd50U,
    0x2315f218U,
    0xcd4a8124U,
    0x67b15ca6U,
    0xa4988857U,
    0xe899b49fU,
    0xce99ac67U,
    0xad9ae69dU,
    0x9618fd7aU,
    0x49db5238U,
    0x2b3d5121U,
    0x4ecfb49bU,
    0x9474ac78U,
    0x707f66acU,
    0x766fbd4dU,
    0x73a3f23aU,
    0x68578121U,
    0x2decdcb0U,
    0x3c63c86bU,
    0x92b714aaU,
    0x141e7c49U,
    0x18dc0eb2U,
    0xa9b8d951U,
    0x78b8006U,
    0x65d64021U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xb0000000U,
    0x18000000U,
    0x44000000U,
    0x6e000000U,
    0x8d000000U,
    0x7b800000U,
    0xf7c00000U,
    0xea00000U,
    0xed700000U,
    0x26280000U,
    0xbdc0000U,
    0xc8b20000U,
    0xff8d0000U,
    0x62a8000U,
    0x1bdf4000U,
    0xb3e000U,
    0x1389d000U,
    0x342d5800U,
    0xbcdad400U,
    0x9834f200U,
    0x12ca7300U,
    0xb6c99a80U,
    0xa8cb30c0U,
    0x5dcde8a0U,
    0xd24f03f0U,
    0x538f1238U,
    0x942fa31cU,
    0x6cdec29eU,
    0x3030e4fdU,
    0x4ec99abdU,
    0x9ccb30c2U,
    0x4bcde8bdU,
    0xab4f03c2U,
    0x5e0f1223U,
    0xaaefa30bU,
    0x77fec28fU,
    0xa780e4e7U,
    0x1dc19a98U,
    0x8da730c2U,
    0xabf7e88dU,
    0xb26e03c0U,
    0xb6bf922bU,
    0x4561e30bU,
    0xb9d5a2b0U,
    0x505b74f3U,
    0x2f75a2b2U,
    0x392b74efU,
    0x75da286U,
    0xff774eaU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x58000000U,
    0x94000000U,
    0xae000000U,
    0xe9000000U,
    0xe4800000U,
    0xa3400000U,
    0x70600000U,
    0xb7500000U,
    0x61080000U,
    0x91dc0000U,
    0x58160000U,
    0x6b6f0000U,
    0xc1898000U,
    0x249c4000U,
    0xe5756000U,
    0xdebe5000U,
    0x26c17800U,
    0x1ba48c00U,
    0xee742e00U,
    0xa9388100U,
    0x8605e580U,
    0xd50564c0U,
    0x36864ba0U,
    0x3842a5d0U,
    0x1fe14e08U,
    0x2396d114U,
    0x11ac9dbeU,
    0xf82de8e1U,
    0xb9ec659dU,
    0xffc924cfU,
    0xae7b2b9aU,
    0x5060f5c6U,
    0xe756363bU,
    0x990d5d34U,
    0x15d93391U,
    0xae1529d2U,
    0x166ae02aU,
    0x8b0d1019U,
    0x6edd9800U,
    0x71969c28U,
    0xcaa83622U,
    0x37ae5d3aU,
    0x3d2eb389U,
    0xd76a69d8U,
    0x53880015U,
    0x1f9c0023U,
    0x9af60031U,
    0x127f0012U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x18000000U,
    0x7c000000U,
    0x86000000U,
    0xa7000000U,
    0xda800000U,
    0x2400000U,
    0x51600000U,
    0x45500000U,
    0x23980000U,
    0x594c0000U,
    0x2f160000U,
    0x56f90000U,
    0x86188000U,
    0xdd894000U,
    0x6a362000U,
    0xdc95000U,
    0x5e514800U,
    0x5f1d4c00U,
    0x6c0ed600U,
    0x4cf1f100U,
    0x75eebd80U,
    0xee3a840U,
    0xe096eba0U,
    0x1b3e1950U,
    0xdab8f638U,
    0xa478a12cU,
    0xc05ff59eU,
    0xe6eee463U,
    0xe6603d87U,
    0x37d3e866U,
    0x25d84bafU,
    0x22e0945U,
    0x53479e1cU,
    0x74e0bd22U,
    0x81966ba3U,
    0x6bb594dU,
    0xf278d632U,
    0xef58f125U,
    0xa86e3db1U,
    0x626e866U,
    0x3bb6cbb7U,
    0x9b0e497cU,
    0x5e713e03U,
    0x83acad03U,
    0x4d870386U,
    0xe0c64553U,
    0xbf27c825U,
    0x7a310c06U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x38000000U,
    0xfc000000U,
    0x86000000U,
    0x11000000U,
    0x5800000U,
    0xdcc00000U,
    0xe7e00000U,
    0x79500000U,
    0x94780000U,
    0x4bac0000U,
    0x5960000U,
    0x43990000U,
    0x7aff8000U,
    0x45ef4000U,
    0x4a332000U,
    0x146a5000U,
    0x94f2f800U,
    0xf88c0400U,
    0x6f23da00U,
    0x9cb00100U,
    0xe62aa180U,
    0x5d524b40U,
    0xbe78fba0U,
    0x84a80a50U,
    0xfd10fa38U,
    0x24da5114U,
    0xacd85996U,
    0x58de4f7bU,
    0x6adb218fU,
    0x29d80b6bU,
    0x7f5a5b98U,
    0xcd181a67U,
    0x64b8a20eU,
    0x324a4502U,
    0xe1c5dbabU,
    0x7c675a73U,
    0x3893822aU,
    0xd81c151eU,
    0x3b39239eU,
    0x98e5e68U,
    0x3aa1d83fU,
    0x18765400U,
    0xedc92213U,
    0x9a000527U,
    0xc7077b8bU,
    0x4c874a42U,
    0x3543da17U,
    0x8520013eU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x90000000U,
    0x98000000U,
    0x84000000U,
    0xe2000000U,
    0x75000000U,
    0xf7800000U,
    0x9c00000U,
    0xe0600000U,
    0xa0100000U,
    0xac780000U,
    0xa5ac0000U,
    0x59b20000U,
    0xb4090000U,
    0xcdc78000U,
    0x22664000U,
    0xa517e000U,
    0x53ffb000U,
    0xb06c7800U,
    0xdfd23400U,
    0x8319e200U,
    0xe33b6500U,
    0x798fff80U,
    0x1505cd40U,
    0xa781fda0U,
    0x71c11870U,
    0xf4627a18U,
    0xda16e114U,
    0x5d7a659eU,
    0xb02c9c45U,
    0x2577e03eU,
    0xa3efb00cU,
    0x6414783dU,
    0x6e7e343bU,
    0xa0abe223U,
    0xa632650bU,
    0xa1c87fabU,
    0x4ba38d5cU,
    0x15761da0U,
    0x8beea876U,
    0x8160205U,
    0x78d519U,
    0x2ba987afU,
    0x42b2f970U,
    0xc88d9fadU,
    0x20853d42U,
    0xab45e5a2U,
    0x6e26dc4cU,
    0x41b2002cU,
    0xf0090036U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x50000000U,
    0x8000000U,
    0xc000000U,
    0xce000000U,
    0x25000000U,
    0x64800000U,
    0x37c00000U,
    0x89a00000U,
    0x36d00000U,
    0xca980000U,
    0x178c0000U,
    0xd2b60000U,
    0x3aed0000U,
    0xd8418000U,
    0xefe0c000U,
    0x4372000U,
    0x1e2c9000U,
    0xa4631800U,
    0x10706c00U,
    0x9fcbea00U,
    0x35525b00U,
    0x32dc9f80U,
    0x86d3c40U,
    0x6e8055a0U,
    0x70c3f750U,
    0xb627d218U,
    0x1892a704U,
    0x147a6dbaU,
    0x5d3e0b57U,
    0x4c58a011U,
    0x90ad501aU,
    0x84a3b812U,
    0xae513c10U,
    0x8b5e5239U,
    0xef2e6703U,
    0x4ae34da3U,
    0x20b39b41U,
    0xc9ec383fU,
    0x7dc0fc16U,
    0x2ea6f218U,
    0x5953372cU,
    0xecd8f587U,
    0xc56ea77aU,
    0x56046a0fU,
    0x81039b14U,
    0x16843f99U,
    0x84c06c6bU,
    0xcc23edaaU,
    0xa792cb56U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x90000000U,
    0xe8000000U,
    0xc4000000U,
    0x66000000U,
    0x21000000U,
    0xdb800000U,
    0xe5400000U,
    0x6ba00000U,
    0x52900000U,
    0x42b80000U,
    0xc16c0000U,
    0x16760000U,
    0xb88f0000U,
    0xedc58000U,
    0x1a61c000U,
    0xd2f46000U,
    0x5b499000U,
    0x52a17800U,
    0xd513f400U,
    0xddffc600U,
    0x51cdff00U,
    0xf4665580U,
    0x47f68b40U,
    0x3ecdf3a0U,
    0x2ae2e450U,
    0x2f32de08U,
    0xe92b9b34U,
    0xe456eb92U,
    0x785b8055U,
    0x465fe003U,
    0x9b5b5029U,
    0xaade9807U,
    0x5c98a415U,
    0x43b95e16U,
    0xcae95b1cU,
    0xbb310b93U,
    0x872cd06cU,
    0x31577803U,
    0xbddcf437U,
    0xae1a460cU,
    0x897c3f18U,
    0x5a0a35a5U,
    0x8f031b51U,
    0xee828b9cU,
    0x10c21055U,
    0xfbe69826U,
    0x6cb4a42aU,
    0x106f5e36U,
    0x55f65b12U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0x8000000U,
    0x14000000U,
    0x22000000U,
    0x5d000000U,
    0x6c800000U,
    0x9ac00000U,
    0x4ca00000U,
    0x54d00000U,
    0xbe780000U,
    0xbc6c0000U,
    0xb60000U,
    0xeb090000U,
    0x62468000U,
    0xb463c000U,
    0x2573e000U,
    0xb6283000U,
    0x90d7b800U,
    0xe47b2c00U,
    0x9d69a200U,
    0x6a33db00U,
    0x6cae780U,
    0xb65a340U,
    0x34f0a5a0U,
    0x106e4850U,
    0x7eb5fa38U,
    0xe00cc71cU,
    0x41c2fd9aU,
    0x17245449U,
    0x8b95602aU,
    0x169bf002U,
    0x6bdc580bU,
    0xde3f1c3dU,
    0x5081a0aU,
    0x7141f736U,
    0xdbe5c5b4U,
    0x8835b86dU,
    0x7bc9a214U,
    0x37e3db23U,
    0x9632e793U,
    0x40c9a34eU,
    0x1c66a59cU,
    0x2177487aU,
    0xcc2b7a1dU,
    0xe1d30725U,
    0xb6ff1d9aU,
    0x6ca96461U,
    0x35125816U,
    0xf95a1c22U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x18000000U,
    0x54000000U,
    0xa000000U,
    0x29000000U,
    0x7800000U,
    0x2bc00000U,
    0x69600000U,
    0xb7900000U,
    0xd6080000U,
    0xd85c0000U,
    0xda320000U,
    0x987f0000U,
    0x26808000U,
    0x10434000U,
    0x24a36000U,
    0x19f77000U,
    0x5d19d800U,
    0x5f17d400U,
    0x4d4f3200U,
    0x9e7e5d00U,
    0xb1814b80U,
    0xa4c592c0U,
    0x89e519a0U,
    0x10d0bff0U,
    0xf62f8a28U,
    0x74edf91cU,
    0xda0d218eU,
    0x665c5bf9U,
    0x69328024U,
    0x1efc402fU,
    0x46c3e00bU,
    0xce4302cU,
    0xd952b81dU,
    0x56eca432U,
    0x7f0cea2dU,
    0x1fda8911U,
    0xd1f4f982U,
    0x411b8fdaU,
    0x8915b205U,
    0x924e1d3fU,
    0xf6f8abbdU,
    0xeac2a2d8U,
    0xc2e5218eU,
    0x52505bd2U,
    0x34688027U,
    0xd4f4022U,
    0x7e796039U,
    0x2184703bU,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x70000000U,
    0x48000000U,
    0xe4000000U,
    0x6e000000U,
    0x7f000000U,
    0x27800000U,
    0x47c00000U,
    0x37a00000U,
    0x89d00000U,
    0xf9a80000U,
    0x81bc0000U,
    0x2cb20000U,
    0xa15d0000U,
    0xb6848000U,
    0xdf47c000U,
    0x37e66000U,
    0x89b7d000U,
    0xfdac800U,
    0x15c1e400U,
    0x3aa67200U,
    0x73537b00U,
    0x1cef4f80U,
    0x8f5ffdc0U,
    0x69875da0U,
    0xe8c756f0U,
    0x4820da38U,
    0x12144f2cU,
    0xc0d758aU,
    0xfd6ba2e5U,
    0xe39e8029U,
    0x3fa6c016U,
    0xcdd0e00aU,
    0x87ad1012U,
    0xc6b82816U,
    0xa731f40aU,
    0x6c9ada28U,
    0x90254f1dU,
    0xe13f5bcU,
    0xae0d62d9U,
    0x786e600bU,
    0x9d1bd00dU,
    0xc4e0c806U,
    0xf430e429U,
    0x118f217U,
    0xa6e5bb36U,
    0x1137afbbU,
    0xf9eedf1U,
    0x15a575adU,
    0xccd7a2fcU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xd0000000U,
    0xb8000000U,
    0x44000000U,
    0xba000000U,
    0x81000000U,
    0x4e800000U,
    0x21c00000U,
    0xda600000U,
    0x71500000U,
    0x82b80000U,
    0xac2c0000U,
    0x8ff60000U,
    0x7d0d0000U,
    0xcf058000U,
    0x5d814000U,
    0xf240e000U,
    0x9b217000U,
    0x9f41800U,
    0xa20c5c00U,
    0x8a82aa00U,
    0xdbc1d500U,
    0xbb602d80U,
    0xefd01140U,
    0x1b7a67a0U,
    0x324cb450U,
    0x44a05218U,
    0x7eb1f924U,
    0x2dab1f96U,
    0xf3b0d865U,
    0x552e0020U,
    0x25710021U,
    0xd6cb8005U,
    0xfc60401cU,
    0x9e53600bU,
    0xaf3d3005U,
    0xd6697838U,
    0x38d06c17U,
    0x32f85210U,
    0x120df93cU,
    0xa2851fa6U,
    0xc7c1d854U,
    0x9565800fU,
    0x6cd1400aU,
    0x90f8e025U,
    0xe70d7025U,
    0x3e02183dU,
    0x9b015c18U,
    0xff872a1eU,
    0x740950bU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0xf8000000U,
    0xec000000U,
    0xc6000000U,
    0x21000000U,
    0xd4800000U,
    0x52400000U,
    0x6ee00000U,
    0x16f00000U,
    0x37080000U,
    0xb9bc0000U,
    0x37560000U,
    0xe59b0000U,
    0x27078000U,
    0x95874000U,
    0xf6c1a000U,
    0x94a3d000U,
    0x6c10f800U,
    0xbf8cc00U,
    0x69b60600U,
    0x7b6aed00U,
    0x54086480U,
    0xfe3f1dc0U,
    0xca97c2a0U,
    0x42ba20d0U,
    0xecd15e28U,
    0x765af104U,
    0xb5211ab6U,
    0xa0d67cefU,
    0xa05e002fU,
    0x4c270023U,
    0xc851800cU,
    0xcc1c4008U,
    0xefc62010U,
    0xcc24902aU,
    0x88515827U,
    0xec1b1c23U,
    0xbfc6fe35U,
    0x34222103U,
    0x6456629fU,
    0x2a19f0d9U,
    0x9ec1a619U,
    0xe0a23d3eU,
    0x36171ca3U,
    0x44fc91caU,
    0x88366496U,
    0xd7a81dd2U,
    0x8fae42abU,
    0x73aa60c8U,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x90000000U,
    0x88000000U,
    0xd4000000U,
    0x9a000000U,
    0x8b000000U,
    0xd9800000U,
    0x60c00000U,
    0x6200000U,
    0xe4100000U,
    0x37680000U,
    0xb5bc0000U,
    0xe1f60000U,
    0x8dd0000U,
    0xf4078000U,
    0xaa02c000U,
    0x7305e000U,
    0x1586f000U,
    0xa6c4c800U,
    0xc3243400U,
    0x2c911a00U,
    0x52f6f00U,
    0xadddd80U,
    0x63039dc0U,
    0xdd8127a0U,
    0x92c602d0U,
    0xc9263218U,
    0x2f90ab34U,
    0x8af8f9aU,
    0xf01a06d1U,
    0xee200024U,
    0xe0100007U,
    0xc5680028U,
    0x7abc001eU,
    0x2a760015U,
    0x371d0014U,
    0xb1a7802aU,
    0xa5d2c036U,
    0x9bcde003U,
    0x24eaf037U,
    0x767ac804U,
    0x9a553411U,
    0xe089a2cU,
    0x124caf02U,
    0x6c29bd8dU,
    0xd45aadf7U,
    0xfc478fbfU,
    0xee6606cbU,
    0x3076000dU,
    0xfc1d0028U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0xf0000000U,
    0xa8000000U,
    0xe4000000U,
    0x5a000000U,
    0xa1000000U,
    0xa5800000U,
    0xb3c00000U,
    0x22200000U,
    0xd9700000U,
    0x82680000U,
    0x1edc0000U,
    0x91920000U,
    0xc0bd0000U,
    0x70008000U,
    0x6804c000U,
    0x84016000U,
    0xaa027000U,
    0x9029800U,
    0x41870400U,
    0xe9c26200U,
    0x83218700U,
    0x7cf2f680U,
    0x31af46c0U,
    0x3cf9f4a0U,
    0x48e0b1f0U,
    0x42d39a38U,
    0x6ed9f30cU,
    0xf9928cb2U,
    0x44bd05c5U,
    0xda00001cU,
    0x61000021U,
    0xc5800027U,
    0x43c00038U,
    0x8a20000bU,
    0x3d700023U,
    0xd868003eU,
    0xbfdc0019U,
    0x3412002dU,
    0x737d001bU,
    0x52208030U,
    0xb174c01cU,
    0x669601dU,
    0xb4de7037U,
    0x98909807U,
    0x813a040aU,
    0x99c2e224U,
    0xeb25472aU,
    0xf8f39684U,
    0x9bad36c6U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x90000000U,
    0x18000000U,
    0xc000000U,
    0xd2000000U,
    0xc9000000U,
    0xba800000U,
    0xdb400000U,
    0x72e00000U,
    0x4ef00000U,
    0xcfc80000U,
    0x837c0000U,
    0xdcd60000U,
    0xa6190000U,
    0x85e38000U,
    0x3f764000U,
    0x1f0ba000U,
    0xc5e7000U,
    0xf1861800U,
    0x90c2fc00U,
    0xef221e00U,
    0xbd574500U,
    0x2edc9480U,
    0x66c023c0U,
    0x50232aa0U,
    0xf8d016d0U,
    0xd01a2638U,
    0x7ae48934U,
    0x1af0b29eU,
    0x71cdaae7U,
    0x8781812U,
    0xb757fc21U,
    0xcbdf9e02U,
    0xfe440520U,
    0xca62b4b4U,
    0x24b113e6U,
    0x9bad12bbU,
    0xdccadad1U,
    0xa4fd8012U,
    0x313400bU,
    0x84be2014U,
    0xd2713012U,
    0xd38e3827U,
    0x281acc1dU,
    0xa6e7a616U,
    0xb0f7c90aU,
    0x24ce9294U,
    0x78fc9ae1U,
    0xa9162038U,
    0xd1bd3016U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xf0000000U,
    0x68000000U,
    0xac000000U,
    0x1a000000U,
    0x63000000U,
    0x24800000U,
    0x28c00000U,
    0x2c200000U,
    0xdf100000U,
    0xaf080000U,
    0x8fdc0000U,
    0x36720000U,
    0xbb7f0000U,
    0xa3a28000U,
    0x4f52c000U,
    0xf9ef2000U,
    0x43ee3000U,
    0x10ea8800U,
    0x7c6c5c00U,
    0x8ac3200U,
    0x568cff00U,
    0x469c1d80U,
    0xd710b6c0U,
    0x130d0fa0U,
    0x2ddd79f0U,
    0x61711a28U,
    0x41ff530cU,
    0x5e650792U,
    0x3ef3e5d7U,
    0x49ba083cU,
    0xcc819c03U,
    0x44c19207U,
    0x16200f2eU,
    0x4c11b598U,
    0xe38edad6U,
    0xb19b586U,
    0x52dafcU,
    0x76bb587U,
    0x282ddacbU,
    0xe849359cU,
    0xe3bf1aedU,
    0x27861589U,
    0x1c412adfU,
    0xbce49d8aU,
    0xc73176d4U,
    0xae1aaf98U,
    0xf5d289dbU,
    0xe42c3230U,
    0xe24cff11U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x30000000U,
    0x28000000U,
    0x24000000U,
    0x8a000000U,
    0xb1000000U,
    0xe0800000U,
    0xbc00000U,
    0x61a00000U,
    0x46500000U,
    0x84a80000U,
    0xe43c0000U,
    0x77b20000U,
    0x39190000U,
    0x4c228000U,
    0xb16c000U,
    0x21cf6000U,
    0x48f000U,
    0xe30a9800U,
    0x7d6eac00U,
    0xc89f8a00U,
    0xb7631b00U,
    0x9976db80U,
    0x3a3cf9c0U,
    0x44b431a0U,
    0x9c9e12f0U,
    0x3562f218U,
    0xdc76873cU,
    0x48bc29b2U,
    0xf2767ec5U,
    0xb3ba1808U,
    0x43f16c15U,
    0x68fa6a22U,
    0x1912b22U,
    0x6b0923aaU,
    0x496fa5daU,
    0xba99a3bbU,
    0x3a6065f4U,
    0xfff4439cU,
    0x2efe55c9U,
    0xfe91bb84U,
    0x808809e8U,
    0x3b2ca98aU,
    0x1479bee2U,
    0x9c57f825U,
    0xedaf5c0aU,
    0xe8bf9229U,
    0xe277770aU,
    0x4bbc31b5U,
    0x7ff212ecU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0x50000000U,
    0x18000000U,
    0xb4000000U,
    0x2000000U,
    0x7b000000U,
    0xcc800000U,
    0xe1c00000U,
    0x19a00000U,
    0x80900000U,
    0xeea80000U,
    0x6fc0000U,
    0x4e760000U,
    0x695b0000U,
    0xd7e78000U,
    0x40704000U,
    0xb45c6000U,
    0x5a633000U,
    0xf2310800U,
    0x9ab91c00U,
    0xed97fa00U,
    0x4b2edb00U,
    0x68ba0d80U,
    0x9e90efc0U,
    0xbaf97a0U,
    0x6f7a04d0U,
    0xe6351228U,
    0x88b8b704U,
    0xae969f8eU,
    0x63af18f9U,
    0x937ce826U,
    0x48316c23U,
    0x45bd123bU,
    0x1b14b732U,
    0x35e89fb3U,
    0xa79818f5U,
    0x30c5681dU,
    0x32262c2fU,
    0x73d0f213U,
    0x93ccc737U,
    0x864a77a6U,
    0xc00e74c7U,
    0xcce9fa03U,
    0x5019db09U,
    0x3d838db9U,
    0x9a47afc6U,
    0xf2e27792U,
    0x49f274f9U,
    0x6c1ffa1fU,
    0xf382db38U,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0x98000000U,
    0xbc000000U,
    0xae000000U,
    0x87000000U,
    0xf0800000U,
    0xc7c00000U,
    0xca00000U,
    0xd5100000U,
    0x46080000U,
    0xa5dc0000U,
    0x44f20000U,
    0x9d7b0000U,
    0xdbe58000U,
    0xee764000U,
    0x37bc2000U,
    0x82c01000U,
    0x2d238800U,
    0xe9535c00U,
    0xdbe84a00U,
    0x18aaa300U,
    0x39cea980U,
    0x4578d6c0U,
    0x7e0c3a0U,
    0x707565f0U,
    0xb8ba6208U,
    0x5643af14U,
    0xf8e34b8eU,
    0xccf639cbU,
    0x797a282dU,
    0x69e50c3eU,
    0x5777e23fU,
    0xd839ef34U,
    0x9856b99U,
    0x484129ebU,
    0xb7e62019U,
    0xb877100dU,
    0x3cbc0814U,
    0xd4421c1bU,
    0x49e3ea02U,
    0x777f306U,
    0xe03c8181U,
    0x2581dad3U,
    0x7e45218bU,
    0x8ce78afcU,
    0xe6f28980U,
    0x7c78c6e6U,
    0x28634b8aU,
    0x5b3639e5U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0x18000000U,
    0x64000000U,
    0x1e000000U,
    0x1b000000U,
    0x5a800000U,
    0xf1c00000U,
    0x53a00000U,
    0xa1100000U,
    0xe2880000U,
    0xf5c0000U,
    0x8bf20000U,
    0xa8ff0000U,
    0xcce38000U,
    0xc5704000U,
    0x4f3ae000U,
    0xd8415000U,
    0x94e6e800U,
    0x41702c00U,
    0xa13b2a00U,
    0xb410b00U,
    0xb2606180U,
    0xcab462c0U,
    0xf79baba0U,
    0xebd739f0U,
    0xfbaca228U,
    0x678c3704U,
    0x8edd438eU,
    0xb715edU,
    0xda9f8821U,
    0x26513c2fU,
    0xe2ef2219U,
    0xbcec770eU,
    0x47efa384U,
    0xed6a45cdU,
    0xd4ab600bU,
    0xfb0e101dU,
    0x201f8804U,
    0xc7913c21U,
    0x894f2203U,
    0xa9fc772aU,
    0xa367a39fU,
    0x9d3645c2U,
    0x1bd96022U,
    0xb9311000U,
    0xe5dc0822U,
    0x52317c3fU,
    0x775dc229U,
    0xdff1271cU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0x30000000U,
    0xf8000000U,
    0x14000000U,
    0xce000000U,
    0x35000000U,
    0x800000U,
    0xe3c00000U,
    0x97200000U,
    0xe7100000U,
    0xd0e80000U,
    0xc73c0000U,
    0xc9760000U,
    0xf81d0000U,
    0xb2628000U,
    0xeaf44000U,
    0xaa586000U,
    0x20035000U,
    0x90021800U,
    0x28060c00U,
    0xdc00ea00U,
    0x2205cf00U,
    0xef032580U,
    0xfb80d1c0U,
    0xd645afa0U,
    0x74674ed0U,
    0x93f01218U,
    0xa0d8d31cU,
    0xf0c7b786U,
    0xdea142c9U,
    0xf650f82dU,
    0x830d1c18U,
    0xa08c920bU,
    0xf2cd9339U,
    0x60ab579bU,
    0x1a5b52c4U,
    0x98000002U,
    0x64000015U,
    0xd600003eU,
    0x1100003bU,
    0x36800005U,
    0xc2c00038U,
    0x59a00001U,
    0x31d0000eU,
    0x4748003cU,
    0xc3ec0004U,
    0x8ebe001aU,
    0xd8310001U,
    0xabfc8012U,
    0xd5d5400fU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x10000000U,
    0x28000000U,
    0x14000000U,
    0x62000000U,
    0x87000000U,
    0x8c800000U,
    0xbac00000U,
    0x5f200000U,
    0xb9d00000U,
    0xf8d80000U,
    0xeacc0000U,
    0x74b60000U,
    0xdee90000U,
    0x3fe78000U,
    0x1df3c000U,
    0xdb8ae000U,
    0x5dd73000U,
    0xe2dc6800U,
    0x31cffc00U,
    0x9e342200U,
    0xa92b5700U,
    0x7f447380U,
    0xf0670e40U,
    0x41b4b1a0U,
    0xdd6e6950U,
    0x6da52a38U,
    0xa2105b14U,
    0x867ed992U,
    0xe6589541U,
    0xaf0e8832U,
    0xfb14cc34U,
    0xbbfe4a23U,
    0x551dab0bU,
    0x516fd1beU,
    0x3a39945U,
    0xb142211U,
    0xc3fb5725U,
    0x91c7398U,
    0x376b0e67U,
    0xcea2b1afU,
    0x1497694bU,
    0x97baaa27U,
    0x6bff9b1fU,
    0x5d1a39a2U,
    0x356aa543U,
    0x59a36035U,
    0xb011f03bU,
    0x397f0813U,
    0x56de0c30U,
    0x80000000U,
    0xc0000000U,
    0x60000000U,
    0x50000000U,
    0x98000000U,
    0xf4000000U,
    0xfa000000U,
    0xc7000000U,
    0x95800000U,
    0x80c00000U,
    0xaaa00000U,
    0x17100000U,
    0xdb980000U,
    0x554c0000U,
    0xd4f20000U,
    0xd0ed0000U,
    0xca658000U,
    0x2db04000U,
    0x34886000U,
    0x2ad6f000U,
    0xe3bb2800U,
    0x371a3400U,
    0x750ffa00U,
    0xa012ab00U,
    0x26196180U,
    0x898cf240U,
    0xb854fba0U,
    0x96f9a970U,
    0x4d793238U,
    0x50be2f24U,
    0x989dd3beU,
    0x52ce9d69U,
    0xcf334838U,
    0x29ccc43cU,
    0xcb4d203U,
    0x89f05U,
    0x5e969ba9U,
    0x5d5e5958U,
    0xceed9a02U,
    0xcf655b11U,
    0xbb35c9beU,
    0x13cb8655U,
    0xabb6e1abU,
    0xc58db277U,
    0x46531b8fU,
    0x3fe1941U,
    0x23fdfa3bU,
    0xd3ffab0aU,
    0x7bfce1bdU,
    0x47fcb273U,
    0xd1fc9bb2U,
    0x18ff5949U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0x28000000U,
    0xa4000000U,
    0x7e000000U,
    0xe5000000U,
    0x8d800000U,
    0x8ec00000U,
    0x12600000U,
    0x1bf00000U,
    0x14880000U,
    0xa6bc0000U,
    0xdad20000U,
    0xbdf0000U,
    0x17a68000U,
    0xd914c000U,
    0xd03d6000U,
    0xac915000U,
    0x62fb9800U,
    0x6cf36400U,
    0x360e6a00U,
    0x46ff8d00U,
    0xd2f08080U,
    0xf309bdc0U,
    0xbb798aa0U,
    0x743460f0U,
    0x456e1228U,
    0xde89792cU,
    0x85b81282U,
    0x6e5404f5U,
    0x8a9cf835U,
    0x9c013420U,
    0x62017234U,
    0x47072906U,
    0xea850aadU,
    0x1443a0ccU,
    0x2e27f23eU,
    0x91d3e926U,
    0xfb586a98U,
    0xb8e2f0d4U,
    0xefb46a1dU,
    0x6aac8d2bU,
    0x6f6c00a3U,
    0xad8e7dd1U,
    0x69386a93U,
    0x6312f0c8U,
    0xdb3c6a23U,
    0xbc108d2fU,
    0x9dbe00a9U,
    0x2517df9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xb0000000U,
    0x28000000U,
    0x3c000000U,
    0x6e000000U,
    0x7000000U,
    0x81800000U,
    0xf400000U,
    0x3600000U,
    0x59500000U,
    0xaff80000U,
    0xd06c0000U,
    0x8fb20000U,
    0x4bef0000U,
    0x85f08000U,
    0x180ac000U,
    0xf165a000U,
    0x40511000U,
    0x217b0800U,
    0xd2ac5400U,
    0xc5944600U,
    0x2adeb900U,
    0x9cdd2d80U,
    0x6fd93740U,
    0x2c5c4ba0U,
    0xd21c5e70U,
    0x43ffce28U,
    0xc6682d1cU,
    0xcb4cb82U,
    0x6a9e63U,
    0xf7b06e31U,
    0xcfea3d3eU,
    0x4ff543b3U,
    0x650f0a6eU,
    0x25e3080aU,
    0xa790542aU,
    0xabde4638U,
    0x65db90fU,
    0x3f1fada2U,
    0xf7cf76aU,
    0x35a96bb9U,
    0xd4178e73U,
    0xbd99663eU,
    0x8bb9693dU,
    0x64898595U,
    0xeaa77343U,
    0x84f18582U,
    0x428b7372U,
    0x81a3859aU,
    0xd374734dU,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0x10000000U,
    0x18000000U,
    0xd4000000U,
    0x66000000U,
    0xfd000000U,
    0x4d800000U,
    0x8a400000U,
    0xa1a00000U,
    0xe9700000U,
    0xcd380000U,
    0xe78c0000U,
    0x9f560000U,
    0xa90f0000U,
    0x31918000U,
    0x96a4000U,
    0x1e076000U,
    0xd9071000U,
    0x83811800U,
    0xab43a400U,
    0x4625ce00U,
    0x2c301700U,
    0xba1b1a80U,
    0x343e9840U,
    0x340834a0U,
    0x8c13df50U,
    0x2b2d5608U,
    0x63a5f314U,
    0x9a77b4aeU,
    0x1ba9f71U,
    0x9a4db63fU,
    0xe3b7a31aU,
    0xe3584ca2U,
    0x49186b59U,
    0xb8b8001dU,
    0xe9cc003dU,
    0xe0f6002aU,
    0x797f0038U,
    0xcf298001U,
    0x4da64004U,
    0xb71601aU,
    0xee38102dU,
    0x9308982bU,
    0x2695e42fU,
    0xabecae28U,
    0x52c4070cU,
    0x5a65828eU,
    0xd8947c5aU,
    0xc2ed1abfU,
    0xd941984aU,
    0x80000000U,
    0xc0000000U,
    0xa0000000U,
    0x90000000U,
    0x28000000U,
    0xa4000000U,
    0x2a000000U,
    0x9f000000U,
    0x32800000U,
    0xc4400000U,
    0xf7a00000U,
    0xed700000U,
    0xf0680000U,
    0xffdc0000U,
    0x6e520000U,
    0x385b0000U,
    0x67958000U,
    0x93bc000U,
    0x1056000U,
    0x6f86f000U,
    0x6dc1e800U,
    0xbb64cc00U,
    0x6997a600U,
    0x7c384700U,
    0xc841280U,
    0x94432c0U,
    0x762354a0U,
    0x5b145f0U,
    0x5e8bce08U,
    0xec8b4b14U,
    0xaf8cd4a2U,
    0x9b0d85cdU,
    0xde492e20U,
    0x866d7b3eU,
    0xa6dddcaeU,
    0x7dd479f2U,
    0xc39a001bU,
    0x75f70013U,
    0x6b2f8023U,
    0x9afcc007U,
    0xd762e021U,
    0x17963029U,
    0xd1390835U,
    0xbd05fc10U,
    0x59812e29U,
    0x54c17b17U,
    0x98e7dcbdU,
    0x2a5379c0U,
    0xe25d8004U,
    0xe097c03cU,
    0x27bf602aU,
    0x6341f02cU,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0x70000000U,
    0x48000000U,
    0xc000000U,
    0xbe000000U,
    0x69000000U,
    0x40800000U,
    0x35400000U,
    0x97e00000U,
    0x40300000U,
    0xb2180000U,
    0xa3ec0000U,
    0xb1520000U,
    0xba690000U,
    0xac908000U,
    0x84cfc000U,
    0xfd412000U,
    0x5be01000U,
    0x1e33f800U,
    0xab1aac00U,
    0xab681e00U,
    0x88143d00U,
    0x93884a80U,
    0x85a5f340U,
    0x7651f4a0U,
    0x6bee1e70U,
    0x7d536618U,
    0xe46d512cU,
    0xb59374aaU,
    0x8c48de5fU,
    0xc402c61dU,
    0x72028129U,
    0x3701ac87U,
    0x5982624bU,
    0x3dc12007U,
    0xaea0102eU,
    0x69d3f81eU,
    0x9b2aac3eU,
    0x51701e02U,
    0x27f83d08U,
    0x9cda4a8dU,
    0x56ccf345U,
    0x9a417488U,
    0xda61de62U,
    0x17f24629U,
    0xffbd412eU,
    0x19b88c80U,
    0x84be725eU,
    0xde38d800U,
    0x407fbc12U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0x30000000U,
    0xe8000000U,
    0x34000000U,
    0x12000000U,
    0xfd000000U,
    0xe0800000U,
    0x50400000U,
    0x2aa00000U,
    0xb3f00000U,
    0x5580000U,
    0x4a6c0000U,
    0xebd60000U,
    0x5be90000U,
    0xe5168000U,
    0x720cc000U,
    0x77c1e000U,
    0x7362b000U,
    0x6b933800U,
    0xe948cc00U,
    0xeae65600U,
    0x61579f00U,
    0x3aa85e80U,
    0xb1355f40U,
    0x92b968a0U,
    0x8bbcb050U,
    0xe13bee38U,
    0xf07f931cU,
    0x2c59e8a2U,
    0xc8e97041U,
    0xce948e1eU,
    0x95cde322U,
    0xfca530a4U,
    0x94f60c67U,
    0x7cd96028U,
    0x30ab7025U,
    0xb0325817U,
    0x43fbc0dU,
    0xecfa0e39U,
    0xa1d231dU,
    0xd54ad08dU,
    0xbce1bc59U,
    0x652d83dU,
    0x232a7c16U,
    0xfbf56e2fU,
    0x215f5310U,
    0xe06e08a2U,
    0xfad2c061U,
    0x75693637U,
    0x6e55ef11U,
    0x80000000U,
    0xc0000000U,
    0xe0000000U,
    0xb0000000U,
    0xa8000000U,
    0x2c000000U,
    0x96000000U,
    0x23000000U,
    0x15800000U,
    0xf1400000U,
    0x54600000U,
    0x8a900000U,
    0xe0380000U,
    0xf06c0000U,
    0x9a720000U,
    0x67690000U,
    0x5df68000U,
    0xef2ec000U,
    0x19172000U,
    0xa7e5000U,
    0x658be800U,
    0x43a75400U,
    0xe237d600U,
    0xab8fd100U,
    0xc4a62b80U,
    0x1db7e740U,
    0xf7c85da0U,
    0x2244a670U,
    0xc9e0be18U,
    0x7d3451cU,
    0x3a5a5d92U,
    0xddfda667U,
    0xd5ce3e2fU,
    0xd3418533U,
    0xa5677d98U,
    0xe616f658U,
    0x1f9561aU,
    0xcbcd1125U,
    0x8c430bb0U,
    0x3ee0b77aU,
    0xb0553591U,
    0xfa1d3262U,
    0x5184819U,
    0xe9ec414U,
    0xa0dd9e3eU,
    0x6f38151eU,
    0xb3ed35a3U,
    0xa831326aU,
    0x968a480eU,
    0x8e27c439U,
    0x77731e01U,
    0xf5ead504U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xf0000000U,
    0xf8000000U,
    0x14000000U,
    0x3a000000U,
    0x89000000U,
    0x9b800000U,
    0xbbc00000U,
    0x5e00000U,
    0x55700000U,
    0xb9080000U,
    0x7e3c0000U,
    0x5c160000U,
    0x673f0000U,
    0x2f968000U,
    0x60fa4000U,
    0xc766000U,
    0x6a8cf000U,
    0x89fb5800U,
    0x87f1b400U,
    0x79481200U,
    0x901f9900U,
    0x481c880U,
    0x674695c0U,
    0xc6a13aa0U,
    0xf1d3bcd0U,
    0x25dbca08U,
    0x2676d2cU,
    0x78b73a96U,
    0xbbecbcf9U,
    0xf3cd4a10U,
    0xb45d2d37U,
    0x28a15a80U,
    0xa2d04cf0U,
    0x675e1239U,
    0x2a20992eU,
    0x2a974889U,
    0xc57cd5d4U,
    0xacb75a81U,
    0x61ef4cfdU,
    0x2ac8921bU,
    0x27dad91aU,
    0x7f612896U,
    0x893025caU,
    0x812c02a7U,
    0x81aef8c0U,
    0x74e88039U,
    0x27494022U,
    0x9b1ee03aU,
    0xa205b03cU,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x88000000U,
    0x4000000U,
    0x5a000000U,
    0x45000000U,
    0x1b800000U,
    0xa2400000U,
    0x68600000U,
    0x23900000U,
    0xedf80000U,
    0xe9ac0000U,
    0xf1760000U,
    0x9da90000U,
    0x63708000U,
    0xbca9c000U,
    0xd2f36000U,
    0xd3ef7000U,
    0xa513b800U,
    0x83f5400U,
    0x658c5e00U,
    0xd985d900U,
    0xdb451380U,
    0xc5e5e940U,
    0x6ad2ada0U,
    0x91a8050U,
    0x32f96638U,
    0x972a4d2cU,
    0xd8b2adbaU,
    0x2f8a807dU,
    0xa4816614U,
    0x2cc64d00U,
    0xc9a4ad88U,
    0x95b38045U,
    0x7009e61eU,
    0x3cc38d31U,
    0xf1a1cdafU,
    0x79b5f05fU,
    0xde0ade03U,
    0xabc51924U,
    0xab26f3b0U,
    0x9a73596dU,
    0x512a7593U,
    0xfbb6a458U,
    0xc7088015U,
    0xf645c024U,
    0x6a65602fU,
    0x7a967034U,
    0xd07b3809U,
    0xd8ea940cU,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0xd0000000U,
    0xf8000000U,
    0x1c000000U,
    0x2e000000U,
    0xf000000U,
    0xc3800000U,
    0x24400000U,
    0xe6a00000U,
    0x72100000U,
    0x2bd80000U,
    0x2c0c0000U,
    0xf4320000U,
    0x580f0000U,
    0xde338000U,
    0x950f4000U,
    0x14b22000U,
    0x99c91000U,
    0x27d3c800U,
    0x83e0400U,
    0x543c8e00U,
    0x9a3a0f00U,
    0x653f0980U,
    0x8ebef440U,
    0x4efa27a0U,
    0x9a5eab70U,
    0xc94ec628U,
    0x2e144b04U,
    0xe5da27b6U,
    0xd30eab73U,
    0x1fb6c615U,
    0x98484b34U,
    0xa9027b3U,
    0xc61dab7eU,
    0xf3ef4617U,
    0x52040b22U,
    0x110387b0U,
    0xc484fb60U,
    0xd3c52e19U,
    0x8e05f23U,
    0xa9b4e1b0U,
    0xbb4ae058U,
    0xef14e196U,
    0xd95ae076U,
    0x1ccce1abU,
    0x3956e069U,
    0x3efee187U,
    0x7259e040U,
    0xd4d6190U,
    0xcc16a069U,
    0x80000000U,
    0x40000000U,
    0x60000000U,
    0xf0000000U,
    0x28000000U,
    0x24000000U,
    0x9e000000U,
    0x4d000000U,
    0x34800000U,
    0x63c00000U,
    0xf1a00000U,
    0x9c700000U,
    0x95680000U,
    0xb45c0000U,
    0x98560000U,
    0xa4590000U,
    0x508000U,
    0xc85cc000U,
    0x4a576000U,
    0x33585000U,
    0x17d0d800U,
    0xfa1f6c00U,
    0xc8b00a00U,
    0xa34a5b00U,
    0xa2687a80U,
    0x73d99ec0U,
    0x121790a0U,
    0xdabb55d0U,
    0x2465238U,
    0x8ae0f72cU,
    0x24179092U,
    0xf3bb55f5U,
    0xc8c6520dU,
    0x5420f70aU,
    0xc93790b2U,
    0x280b55f7U,
    0x320e521bU,
    0x310cf726U,
    0xf08990a6U,
    0x5bce55e5U,
    0x5ba8d224U,
    0xc1793705U,
    0x2fe670a5U,
    0x1496c5d3U,
    0x9e796a10U,
    0xac670b12U,
    0xf0d622b2U,
    0x859f32d3U,
    0xbef67ab5U,
    0x4fac9ec8U,
    0x5779108dU,
    0xd6e295c9U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x30000000U,
    0x88000000U,
    0xbc000000U,
    0x26000000U,
    0xd000000U,
    0x35800000U,
    0x76c00000U,
    0x5ee00000U,
    0xb7500000U,
    0x75880000U,
    0x139c0000U,
    0x19320000U,
    0xdd9f0000U,
    0x78348000U,
    0x561bc000U,
    0x7f4a000U,
    0x2f7b5000U,
    0x51601800U,
    0x1a93a400U,
    0xbbee6a00U,
    0x6b894900U,
    0xea999180U,
    0x96b328c0U,
    0xb05fdba0U,
    0x9656f1f0U,
    0x1e08f228U,
    0x7c5e2d3cU,
    0xd857dbaaU,
    0xbf0af1e3U,
    0xd7daf20bU,
    0x97912d1cU,
    0x4e6b5b8cU,
    0x3d4d31d2U,
    0x847c5218U,
    0xa9e57d2fU,
    0x79d7c38bU,
    0xa3c955e6U,
    0xa3c9814U,
    0x94446431U,
    0xd6a0ca0eU,
    0xb7f11900U,
    0x677f09adU,
    0xcd644ce8U,
    0xc9191adU,
    0x3eef28e4U,
    0xe20ddb9cU,
    0xba59f1ecU,
    0xc554720eU,
    0x3289ed3dU,
    0x80000000U,
    0x40000000U,
    0xe0000000U,
    0xb0000000U,
    0x18000000U,
    0x1c000000U,
    0xf2000000U,
    0xd1000000U,
    0xc6800000U,
    0xdb400000U,
    0x24a00000U,
    0x34f00000U,
    0xdbf80000U,
    0x24cc0000U,
    0x2fd60000U,
    0xb6cd0000U,
    0xed68000U,
    0x884f4000U,
    0x79932000U,
    0x46e8f000U,
    0x8062d800U,
    0xa997dc00U,
    0xaee8be00U,
    0x64601700U,
    0xf793e680U,
    0x95e8ca40U,
    0x6fe6f8a0U,
    0x18536d50U,
    0xbb0ae618U,
    0xb9358b3cU,
    0x2c1ef8beU,
    0x609f6d6bU,
    0x86dce61aU,
    0x6ef88b23U,
    0xfc4878b5U,
    0x2f902d6eU,
    0x29efc61eU,
    0xcde07b2cU,
    0x6152a0bbU,
    0x798bf15cU,
    0x8c71782aU,
    0x2bbd6c05U,
    0x43efc605U,
    0x40e07b1aU,
    0xb5d2a09cU,
    0xc3cbf14bU,
    0x76517820U,
    0xd80d6c3fU,
    0x4eb7c637U,
    0x81dc7b02U,
    0x877ca08cU,
    0x8a8af14aU,
    0x80000000U,
    0x40000000U,
    0x20000000U,
    0xf0000000U,
    0xe8000000U,
    0xb4000000U,
    0xa000000U,
    0x6d000000U,
    0xa7800000U,
    0x73400000U,
    0x11e00000U,
    0x3a300000U,
    0x7e680000U,
    0xcd9c0000U,
    0xfed60000U,
    0xfdb0000U,
    0xb6b28000U,
    0x1aac000U,
    0xe3bfe000U,
    0xcac47000U,
    0xb5a77800U,
    0x29557400U,
    0xd49e3600U,
    0x33569d00U,
    0xc19b9280U,
    0x18d780c0U,
    0x14dec4a0U,
    0x3234add0U,
    0x5a6f4e28U,
    0xff98e92cU,
    0x3fd724b2U,
    0xfe5bddd1U,
    0x16f2b618U,
    0xd0cb5d26U,
    0x6048f2a3U,
    0x718e30fbU,
    0x20aadcb7U,
    0x823869efU,
    0x52856031U,
    0xc8c2b00aU,
    0xfca6983bU,
    0xbcd6043dU,
    0x66ddce2eU,
    0xd332293bU,
    0x5be8c4a1U,
    0xb7dfadc6U,
    0x5ab5ce27U,
    0x77ae2939U,
    0xc0bec4a6U,
    0xe244adc7U,
    0xb8674e07U,
    0xe674e935U,
    0x80000000U,
    0xc0000000U,
    0x20000000U,
    0x70000000U,
    0x58000000U,
    0xe4000000U,
    0x92000000U,
    0x59000000U,
    0x64800000U,
    0x2b400000U,
    0x64600000U,
    0xf7100000U,
    0xa2a80000U,
    0xeb7c0000U,
    0x1e720000U,
    0x1b3f0000U,
    0x8f968000U,
    0x4fef4000U,
    0x3c18e000U,
    0x40e23000U,
    0x3c539800U,
    0x96cc4c00U,
    0x346bae00U,
    0xdf5100U,
    0x86470f80U,
    0x5ae12fc0U,
    0x6950c1a0U,
    0xfc4c0ef0U,
    0x402a3628U,
    0x933c1d2cU,
    0x8392219eU,
    0x41ed3ee5U,
    0x631d2e12U,
    0xb760111fU,
    0xce97efb5U,
    0x2f6f1fc0U,
    0xd55959b6U,
    0x558342feU,
    0x13c5183dU,
    0x69230c3dU,
    0x70734e28U,
    0xd43d6113U,
    0x7014978fU,
    0x712d63c7U,
    0xabbb6f99U,
    0x8ed35fd5U,
    0xc68d39afU,
    0x158d32e8U,
    0x2c0ae02eU,
    0xa1cd3024U,
    0x9fed1836U,
    0xd41f0c09U,
    0x80000000U,
    0x40000000U,
    0xa0000000U,
    0xb0000000U,
    0x58000000U,
    0x84000000U,
    0x4a000000U,
    0x1d000000U,
    0x83800000U,
    0xfec00000U,
    0x52200000U,
    0x5ad00000U,
    0xf3f80000U,
    0x6d6c0000U,
    0xbf60000U,
    0x75af0000U,
    0x32d48000U,
    0x6ffd4000U,
    0x36fe000U,
    0xecf5f000U,
    0xb32f7800U,
    0xcb945400U,
    0x891cfa00U,
    0x169a6f00U,
    0xc659ab80U,
    0xd379c740U,
    0xff2831a0U,
    0xad941850U,
    0xa21d8208U,
    0xf61d3b3cU,
    0xb41fd1beU,
    0x651de84dU,
    0xc09cfa24U,
    0xb55a6f36U,
    0xb7f9ab9aU,
    0xc769c754U,
    0x6f03182U,
    0x1e281858U,
    0x10138232U,
    0xf3de3b36U,
    0xebd51a0U,
    0x818fa874U,
    0xa3079a3eU,
    0x6c82df26U,
    0xf44133aaU,
    0x8d64636fU,
    0x3735b3acU,
    0xb689234cU,
    0x6d8253b0U,
    0x59c0d35aU,
    0x34a32b93U,
    0x1397876eU,
};

const uint64_t Matrices::matrices64[Matrices::num_dimensions * Matrices::size] =
{
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x400000000000ULL,
    0x200000000000ULL,
    0x100000000000ULL,
    0x80000000000ULL,
    0x40000000000ULL,
    0x20000000000ULL,
    0x10000000000ULL,
    0x8000000000ULL,
    0x4000000000ULL,
    0x2000000000ULL,
    0x1000000000ULL,
    0x800000000ULL,
    0x400000000ULL,
    0x200000000ULL,
    0x100000000ULL,
    0x80000000ULL,
    0x40000000ULL,
    0x20000000ULL,
    0x10000000ULL,
    0x8000000ULL,
    0x4000000ULL,
    0x2000000ULL,
    0x1000000ULL,
    0x800000ULL,
    0x400000ULL,
    0x200000ULL,
    0x100000ULL,
    0x80000ULL,
    0x40000ULL,
    0x20000ULL,
    0x10000ULL,
    0x8000ULL,
    0x4000ULL,
    0x2000ULL,
    0x1000ULL,
    0x800ULL,
    0x400ULL,
    0x200ULL,
    0x100ULL,
    0x80ULL,
    0x40ULL,
    0x20ULL,
    0x10ULL,
    0x8ULL,
    0x4ULL,
    0x2ULL,
    0x1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0xcc00000000000ULL,
    0xaa00000000000ULL,
    0xff00000000000ULL,
    0x8080000000000ULL,
    0xc0c0000000000ULL,
    0xa0a0000000000ULL,
    0xf0f0000000000ULL,
    0x8888000000000ULL,
    0xcccc000000000ULL,
    0xaaaa000000000ULL,
    0xffff000000000ULL,
    0x8000800000000ULL,
    0xc000c00000000ULL,
    0xa000a00000000ULL,
    0xf000f00000000ULL,
    0x8800880000000ULL,
    0xcc00cc0000000ULL,
    0xaa00aa0000000ULL,
    0xff00ff0000000ULL,
    0x8080808000000ULL,
    0xc0c0c0c000000ULL,
    0xa0a0a0a000000ULL,
    0xf0f0f0f000000ULL,
    0x8888888800000ULL,
    0xcccccccc00000ULL,
    0xaaaaaaaa00000ULL,
    0xffffffff00000ULL,
    0x8000000080000ULL,
    0xc0000000c0000ULL,
    0xa0000000a0000ULL,
    0xf0000000f0000ULL,
    0x8800000088000ULL,
    0xcc000000cc000ULL,
    0xaa000000aa000ULL,
    0xff000000ff000ULL,
    0x8080000080800ULL,
    0xc0c00000c0c00ULL,
    0xa0a00000a0a00ULL,
    0xf0f00000f0f00ULL,
    0x8888000088880ULL,
    0xcccc0000cccc0ULL,
    0xaaaa0000aaaa0ULL,
    0xffff0000ffff0ULL,
    0x8000800080008ULL,
    0xc000c000c000cULL,
    0xa000a000a000aULL,
    0xf000f000f000fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0x5c00000000000ULL,
    0x8e00000000000ULL,
    0xc500000000000ULL,
    0x6880000000000ULL,
    0x9cc0000000000ULL,
    0xee60000000000ULL,
    0x5590000000000ULL,
    0x8068000000000ULL,
    0xc09c000000000ULL,
    0x60ee000000000ULL,
    0x9055000000000ULL,
    0xe880800000000ULL,
    0x5cc0c00000000ULL,
    0x8e60600000000ULL,
    0xc590900000000ULL,
    0x6868e80000000ULL,
    0x9c9c5c0000000ULL,
    0xeeee8e0000000ULL,
    0x5555c50000000ULL,
    0x8000e88000000ULL,
    0xc0005cc000000ULL,
    0x60008e6000000ULL,
    0x9000c59000000ULL,
    0xe800686800000ULL,
    0x5c009c9c00000ULL,
    0x8e00eeee00000ULL,
    0xc500555500000ULL,
    0x6880800080000ULL,
    0x9cc0c000c0000ULL,
    0xee60600060000ULL,
    0x5590900090000ULL,
    0x8068e800e8000ULL,
    0xc09c5c005c000ULL,
    0x60ee8e008e000ULL,
    0x9055c500c5000ULL,
    0xe880e88068800ULL,
    0x5cc05cc09cc00ULL,
    0x8e608e60ee600ULL,
    0xc590c59055900ULL,
    0x6868686880680ULL,
    0x9c9c9c9cc09c0ULL,
    0xeeeeeeee60ee0ULL,
    0x5555555590550ULL,
    0x8000000068808ULL,
    0xc00000009cc0cULL,
    0x60000000ee606ULL,
    0x9000000055909ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0x7400000000000ULL,
    0xa200000000000ULL,
    0x9300000000000ULL,
    0xd880000000000ULL,
    0x2540000000000ULL,
    0x59e0000000000ULL,
    0xe6d0000000000ULL,
    0x7808000000000ULL,
    0xb40c000000000ULL,
    0x8202000000000ULL,
    0xc305000000000ULL,
    0x208f800000000ULL,
    0x5147400000000ULL,
    0xfbea200000000ULL,
    0x75d9300000000ULL,
    0xa085880000000ULL,
    0x914e540000000ULL,
    0xdbe79e0000000ULL,
    0x25db6d0000000ULL,
    0x5880008000000ULL,
    0xe54000c000000ULL,
    0x79e0002000000ULL,
    0xb6d0005000000ULL,
    0x800800f800000ULL,
    0xc00c007400000ULL,
    0x200200a200000ULL,
    0x5005009300000ULL,
    0xf80f80d880000ULL,
    0x7407402540000ULL,
    0xa20a2059e0000ULL,
    0x930930e6d0000ULL,
    0xd88d887808000ULL,
    0x254254b40c000ULL,
    0x59e59e8202000ULL,
    0xe6de6dc305000ULL,
    0x780f80a08f800ULL,
    0xb407409147400ULL,
    0x820a20dbea200ULL,
    0xc3093025d9300ULL,
    0x208d885885880ULL,
    0x514254e54e540ULL,
    0xfbe59e79e79e0ULL,
    0x75de6db6db6d0ULL,
    0xa08f800000008ULL,
    0x914740000000cULL,
    0xdbea200000002ULL,
    0x25d9300000005ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xf800000000000ULL,
    0xdc00000000000ULL,
    0x7a00000000000ULL,
    0x9d00000000000ULL,
    0x5a80000000000ULL,
    0x2fc0000000000ULL,
    0xa160000000000ULL,
    0xf0b0000000000ULL,
    0xda88000000000ULL,
    0x6fc4000000000ULL,
    0x8162000000000ULL,
    0x40bb000000000ULL,
    0x2287800000000ULL,
    0xb3c9c00000000ULL,
    0xfb65a00000000ULL,
    0xddb2d00000000ULL,
    0x7802280000000ULL,
    0x9c0b3c0000000ULL,
    0x5a0fb60000000ULL,
    0x2d0ddb0000000ULL,
    0xa287808000000ULL,
    0xf3c9c04000000ULL,
    0xdb65a02000000ULL,
    0x6db2d0b000000ULL,
    0x800228f800000ULL,
    0x400b3cdc00000ULL,
    0x200fb67a00000ULL,
    0xb00ddb9d00000ULL,
    0xf80780da80000ULL,
    0xdc09c06fc0000ULL,
    0x7a05a08160000ULL,
    0x9d02d040b0000ULL,
    0x5a8a282288000ULL,
    0x2fcf3cb3c4000ULL,
    0xa16db6fb62000ULL,
    0xf0b6dbddbb000ULL,
    0xda8000f807800ULL,
    0x6fc000dc09c00ULL,
    0x8160007a05a00ULL,
    0x40b0009d02d00ULL,
    0x2288005a8a280ULL,
    0xb3c4002fcf3c0ULL,
    0xfb6200a16db60ULL,
    0xddbb00f0b6db0ULL,
    0x780780da80008ULL,
    0x9c09c06fc0004ULL,
    0x5a05a08160002ULL,
    0x2d02d040b000bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0x2400000000000ULL,
    0x5600000000000ULL,
    0xfb00000000000ULL,
    0xe080000000000ULL,
    0x7040000000000ULL,
    0xa860000000000ULL,
    0x1430000000000ULL,
    0x9ec8000000000ULL,
    0xdf24000000000ULL,
    0xb6d6000000000ULL,
    0x8bbb000000000ULL,
    0x4800800000000ULL,
    0x6400400000000ULL,
    0x3600600000000ULL,
    0xcb00300000000ULL,
    0x2880c80000000ULL,
    0x5440240000000ULL,
    0xfe60560000000ULL,
    0xef30fb0000000ULL,
    0x7e48e08000000ULL,
    0xaf64704000000ULL,
    0x1eb6a86000000ULL,
    0x9f8b143000000ULL,
    0xd6c81ec800000ULL,
    0xbb249f2400000ULL,
    0x80d6d6d600000ULL,
    0x40bbbbbb00000ULL,
    0x6080000080000ULL,
    0x3040000040000ULL,
    0xc860000060000ULL,
    0x2430000030000ULL,
    0x56c80000c8000ULL,
    0xfb24000024000ULL,
    0xe0d6000056000ULL,
    0x70bb0000fb000ULL,
    0xa8808000e0800ULL,
    0x1440400070400ULL,
    0x9e606000a8600ULL,
    0xdf30300014300ULL,
    0xb648c8009ec80ULL,
    0x8b642400df240ULL,
    0x48b65600b6d60ULL,
    0x648bfb008bbb0ULL,
    0x3648608048008ULL,
    0xcb64304064004ULL,
    0x28b6c86036006ULL,
    0x548b2430cb003ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0x9400000000000ULL,
    0x3e00000000000ULL,
    0xe300000000000ULL,
    0xbe80000000000ULL,
    0x23c0000000000ULL,
    0x1e20000000000ULL,
    0xf310000000000ULL,
    0x4678000000000ULL,
    0x6784000000000ULL,
    0x7846000000000ULL,
    0x8467000000000ULL,
    0xc678800000000ULL,
    0xa784c00000000ULL,
    0xd846a00000000ULL,
    0x5467d00000000ULL,
    0x9e78d80000000ULL,
    0x3384540000000ULL,
    0xe6469e0000000ULL,
    0xb767330000000ULL,
    0x20f8668000000ULL,
    0x104477c000000ULL,
    0xf866802000000ULL,
    0x4477c01000000ULL,
    0x668020f800000ULL,
    0x77c0104400000ULL,
    0x8020f86600000ULL,
    0xc010447700000ULL,
    0xa0f8668080000ULL,
    0xd04477c0c0000ULL,
    0x58668020a0000ULL,
    0x9477c010d0000ULL,
    0x3e8020f858000ULL,
    0xe3c0104494000ULL,
    0xbe20f8663e000ULL,
    0x23104477e3000ULL,
    0x1e7866803e800ULL,
    0xf38477c0e3c00ULL,
    0x46468020be200ULL,
    0x6767c01023100ULL,
    0x78f820f81e780ULL,
    0x84441044f3840ULL,
    0xc666f86646460ULL,
    0xa777447767670ULL,
    0xd800e680f8f88ULL,
    0x5400b7c04444cULL,
    0x9e0020206666aULL,
    0x330010107777dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x8800000000000ULL,
    0x2400000000000ULL,
    0x1200000000000ULL,
    0x2d00000000000ULL,
    0x7680000000000ULL,
    0x9e40000000000ULL,
    0x820000000000ULL,
    0x6410000000000ULL,
    0xb228000000000ULL,
    0x7d14000000000ULL,
    0xfea2000000000ULL,
    0xba49000000000ULL,
    0x1a24800000000ULL,
    0x491b400000000ULL,
    0xc4b5a00000000ULL,
    0xe373900000000ULL,
    0xf680080000000ULL,
    0xde40040000000ULL,
    0xa8200a0000000ULL,
    0x3410050000000ULL,
    0x3a28088000000ULL,
    0x5914024000000ULL,
    0xeca2012000000ULL,
    0x974902d000000ULL,
    0x6ca4876800000ULL,
    0xd75b49e400000ULL,
    0xcc95a08200000ULL,
    0x8763964100000ULL,
    0x44a8032280000ULL,
    0xa35403d140000ULL,
    0x568205ea20000ULL,
    0x8e590ea490000ULL,
    0x200c892248000ULL,
    0x100f46d1b4000ULL,
    0x2817ad6b5a000ULL,
    0x743a9ce739000ULL,
    0x9a24800000800ULL,
    0x91b400000400ULL,
    0x64b5a00000a00ULL,
    0xb373900000500ULL,
    0x7e80080000880ULL,
    0xfa40040000240ULL,
    0xba200a0000120ULL,
    0x19100500002d0ULL,
    0x4ca8088000768ULL,
    0xc7540240009e4ULL,
    0xe482012000082ULL,
    0xf35902d000641ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0xd400000000000ULL,
    0x6a00000000000ULL,
    0x7100000000000ULL,
    0x3880000000000ULL,
    0x5840000000000ULL,
    0xea20000000000ULL,
    0x3110000000000ULL,
    0x98a8000000000ULL,
    0x854000000000ULL,
    0xc22a000000000ULL,
    0xe525000000000ULL,
    0xf2b2800000000ULL,
    0x7948400000000ULL,
    0xfaa4200000000ULL,
    0xbd73100000000ULL,
    0x18a8080000000ULL,
    0x4854040000000ULL,
    0x622a0a0000000ULL,
    0xb525050000000ULL,
    0xdab2828000000ULL,
    0xad484d4000000ULL,
    0x90a426a000000ULL,
    0xcc73171000000ULL,
    0x20280b8800000ULL,
    0x1014018400000ULL,
    0x880a04a200000ULL,
    0x8435061100000ULL,
    0x421a8b0a80000ULL,
    0xa51c4dc540000ULL,
    0x528e2a82a0000ULL,
    0x2956194250000ULL,
    0xd29a84a328000ULL,
    0x695c461084000ULL,
    0x72ae2b0842000ULL,
    0x39461dc631000ULL,
    0x5ab2828000800ULL,
    0xed484d4000400ULL,
    0x30a426a000a00ULL,
    0x9c73171000500ULL,
    0x8280b8800280ULL,
    0xc414018400d40ULL,
    0xe20a04a2006a0ULL,
    0xf535061100710ULL,
    0x7a9a8b0a80388ULL,
    0xfd5c4dc540584ULL,
    0xb8ae2a82a0ea2ULL,
    0x1846194250311ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x9800000000000ULL,
    0x9400000000000ULL,
    0x8a00000000000ULL,
    0x5b00000000000ULL,
    0x3380000000000ULL,
    0xd9c0000000000ULL,
    0x7220000000000ULL,
    0x3f10000000000ULL,
    0xc1b8000000000ULL,
    0xa6ec000000000ULL,
    0x5386000000000ULL,
    0x29f5000000000ULL,
    0xa3a800000000ULL,
    0x1b2ac00000000ULL,
    0xd392e00000000ULL,
    0x69ff700000000ULL,
    0xea38080000000ULL,
    0xab2c040000000ULL,
    0x4ba60e0000000ULL,
    0xfde50b0000000ULL,
    0x6002898000000ULL,
    0xf006c94000000ULL,
    0x7834e8a000000ULL,
    0x241a75b000000ULL,
    0x123a8b3800000ULL,
    0xcf2ac99c00000ULL,
    0xb992e92200000ULL,
    0x82ff78f100000ULL,
    0x41b80d9b80000ULL,
    0xe6ec072ec0000ULL,
    0xb386039860000ULL,
    0x99f50c2f50000ULL,
    0x923a8a1ba8000ULL,
    0x8f2ac56eac000ULL,
    0x5992e2bb2e000ULL,
    0x32ff70def7000ULL,
    0xd9b8098000800ULL,
    0x72ec094000400ULL,
    0x398608a000e00ULL,
    0xc2f505b000b00ULL,
    0xa1ba833800980ULL,
    0x56eacd9c00940ULL,
    0x2bb2e722008a0ULL,
    0xdef73f1005b0ULL,
    0x1800041b80338ULL,
    0xd4000e6ec0d9cULL,
    0x6a000b3860722ULL,
    0xeb00099f503f1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x6c00000000000ULL,
    0x9e00000000000ULL,
    0x2300000000000ULL,
    0x5780000000000ULL,
    0xadc0000000000ULL,
    0x7fa0000000000ULL,
    0x91d0000000000ULL,
    0x4988000000000ULL,
    0xced4000000000ULL,
    0x880a000000000ULL,
    0x2c0f000000000ULL,
    0x3e0d800000000ULL,
    0x3317c00000000ULL,
    0x5fb0600000000ULL,
    0xc1f8b00000000ULL,
    0xe18d880000000ULL,
    0xb2d7c40000000ULL,
    0x1e106a0000000ULL,
    0x6328b10000000ULL,
    0xf785888000000ULL,
    0xbdc3c2c000000ULL,
    0x77ba63e000000ULL,
    0xfdf7b33000000ULL,
    0xd7800df800000ULL,
    0xedc0081c00000ULL,
    0xdfa0041a00000ULL,
    0x81d00a2d00000ULL,
    0x4188016080000ULL,
    0xa2d400f140000ULL,
    0x160a069aa0000ULL,
    0xf0f09edf0000ULL,
    0x698d820058000ULL,
    0x9ed7c5003c000ULL,
    0x20106a81a6000ULL,
    0x5028b7c27b000ULL,
    0xa805816080800ULL,
    0x7c03c0f140400ULL,
    0x961a669aa0a00ULL,
    0x4f27b9edf0100ULL,
    0xc9880a0058080ULL,
    0x8ed401003c6c0ULL,
    0x280a0081a69e0ULL,
    0x3c0f06c27b230ULL,
    0x360d89e080d78ULL,
    0x5f17c23140edcULL,
    0xc1b0657aa0dfaULL,
    0xe2f8baddf081dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0xac00000000000ULL,
    0x9600000000000ULL,
    0x2b00000000000ULL,
    0xd480000000000ULL,
    0x940000000000ULL,
    0xe2a0000000000ULL,
    0x5250000000000ULL,
    0x4e28000000000ULL,
    0xc71c000000000ULL,
    0x629e000000000ULL,
    0x1267000000000ULL,
    0x6e13800000000ULL,
    0xf731c00000000ULL,
    0x3a98a00000000ULL,
    0xbe44900000000ULL,
    0xf83b880000000ULL,
    0xdc2dc40000000ULL,
    0xee06a20000000ULL,
    0xb723930000000ULL,
    0x1aa80d8000000ULL,
    0x8e5c0ec000000ULL,
    0xa03e0b6000000ULL,
    0x703701b000000ULL,
    0x783b88c800000ULL,
    0x9c2dca5400000ULL,
    0xce06a74a00000ULL,
    0x8723979500000ULL,
    0x42a801aa80000ULL,
    0x225c08e5c0000ULL,
    0x363e0a03e0000ULL,
    0x5b37070370000ULL,
    0xacbb8783b8000ULL,
    0x956dc9c2dc000ULL,
    0x2ca6ace06a000ULL,
    0xd573987239000ULL,
    0xc800c2a80800ULL,
    0xe5400625c0400ULL,
    0x54a00163e0200ULL,
    0x495006b370300ULL,
    0xc2a80f4bb8580ULL,
    0x625c0396dcac0ULL,
    0x163e0baa6a960ULL,
    0x6b370fe7392b0ULL,
    0xf4bb8d8000548ULL,
    0x396dcec000494ULL,
    0xbaa6ab6000c2aULL,
    0xfe7391b000625ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0x8c00000000000ULL,
    0xe200000000000ULL,
    0x3300000000000ULL,
    0xf80000000000ULL,
    0x2140000000000ULL,
    0x95a0000000000ULL,
    0x5e70000000000ULL,
    0xd808000000000ULL,
    0x1c24000000000ULL,
    0xba16000000000ULL,
    0xef37000000000ULL,
    0x1586800000000ULL,
    0x9e6fc00000000ULL,
    0x781b600000000ULL,
    0x4c34900000000ULL,
    0x420e880000000ULL,
    0x630bcc0000000ULL,
    0xf7ad6a0000000ULL,
    0xad73950000000ULL,
    0x7780078000000ULL,
    0x6d4004c000000ULL,
    0xd7a0042000000ULL,
    0x3d70063000000ULL,
    0x2f880f7800000ULL,
    0xb1640ad400000ULL,
    0xcdb6077a00000ULL,
    0x824706d700000ULL,
    0xc20e8d7880000ULL,
    0xa30bc3d640000ULL,
    0x57ad62fb60000ULL,
    0xfd739b1470000ULL,
    0x8f8004d8e8000ULL,
    0xe1400424bc000ULL,
    0x35a00620d6000ULL,
    0xe700f3039000ULL,
    0x20080af880800ULL,
    0x9024071640c00ULL,
    0x581606db60a00ULL,
    0xdc370d2470500ULL,
    0x1a0683a0e8f80ULL,
    0xbf2fc2f0bc8c0ULL,
    0xedbb6b5ad6e20ULL,
    0x12449ce739330ULL,
    0x9a068000008f8ULL,
    0x7f2fc00000e14ULL,
    0x4dbb60000035aULL,
    0x42449000000e7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x3800000000000ULL,
    0xc400000000000ULL,
    0x4200000000000ULL,
    0xa300000000000ULL,
    0xf180000000000ULL,
    0xaa40000000000ULL,
    0xfce0000000000ULL,
    0x8510000000000ULL,
    0xe008000000000ULL,
    0x500c000000000ULL,
    0x5806000000000ULL,
    0x5409000000000ULL,
    0x7a03800000000ULL,
    0x670c400000000ULL,
    0xb384200000000ULL,
    0x94a300000000ULL,
    0xd6f180000000ULL,
    0x2f5aa40000000ULL,
    0x1ce7ce0000000ULL,
    0xd514510000000ULL,
    0xb800008000000ULL,
    0x40000c000000ULL,
    0x2200006000000ULL,
    0x3300009000000ULL,
    0xc980003800000ULL,
    0x6e4000c400000ULL,
    0xbee0004200000ULL,
    0x261000a300000ULL,
    0x118800f180000ULL,
    0xfa4c00aa40000ULL,
    0xa4e600fce0000ULL,
    0xd119008510000ULL,
    0x9a0b80e008000ULL,
    0x370040500c000ULL,
    0xeb82205806000ULL,
    0x5d43305409000ULL,
    0x776c987a03800ULL,
    0x4856e4670c400ULL,
    0xaf63eeb384200ULL,
    0xdc5e61094a300ULL,
    0xb56f188d6f180ULL,
    0x2b5aa4ef5aa40ULL,
    0x3ee7ce7ce7ce0ULL,
    0xe614514514510ULL,
    0x7180000000008ULL,
    0x6a4000000000cULL,
    0x9ce0000000006ULL,
    0x1510000000009ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0x5400000000000ULL,
    0x9a00000000000ULL,
    0x9d00000000000ULL,
    0x1e80000000000ULL,
    0x5cc0000000000ULL,
    0x7d20000000000ULL,
    0x8d10000000000ULL,
    0x2488000000000ULL,
    0x71c4000000000ULL,
    0xeba2000000000ULL,
    0x75df000000000ULL,
    0x6ba2800000000ULL,
    0x35d1400000000ULL,
    0x4ba3a00000000ULL,
    0xc5d2d00000000ULL,
    0xe3a1680000000ULL,
    0x91db8c0000000ULL,
    0x79aef20000000ULL,
    0xcdf410000000ULL,
    0x672a808000000ULL,
    0x5015404000000ULL,
    0x1a01a02000000ULL,
    0xdd0dd0f000000ULL,
    0x3e83e8a800000ULL,
    0xaccacc5400000ULL,
    0xd52d529a00000ULL,
    0xd91d919d00000ULL,
    0xbe83e89e80000ULL,
    0xeccacc1cc0000ULL,
    0xf52d525d20000ULL,
    0x291d917d10000ULL,
    0x1683e80c88000ULL,
    0xb8cacc65c4000ULL,
    0x6f2d5251a2000ULL,
    0xb41d9118df000ULL,
    0x803e85d22800ULL,
    0xe40acc7d11400ULL,
    0x120d528c83a00ULL,
    0x390d9125c2d00ULL,
    0x2c8be8f1a9680ULL,
    0x95cecca8df8c0ULL,
    0xf9af52552cf20ULL,
    0x4cd2919910410ULL,
    0x4729681e80008ULL,
    0xa01f8c5cc0004ULL,
    0xb20cf27d20002ULL,
    0x8900418d1000fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xd800000000000ULL,
    0xc400000000000ULL,
    0x4600000000000ULL,
    0x8500000000000ULL,
    0xa580000000000ULL,
    0x76c0000000000ULL,
    0xada0000000000ULL,
    0x6ab0000000000ULL,
    0x2da8000000000ULL,
    0xaabc000000000ULL,
    0xdaa000000000ULL,
    0x7ab1000000000ULL,
    0xd5a7800000000ULL,
    0xbebd400000000ULL,
    0x93a3e00000000ULL,
    0x3bb5100000000ULL,
    0x3629b80000000ULL,
    0x4d727c0000000ULL,
    0x9b83620000000ULL,
    0x27c4d70000000ULL,
    0xb629b88000000ULL,
    0x8d727cc000000ULL,
    0xbb83622000000ULL,
    0xf7c4d7d000000ULL,
    0x6e29b85800000ULL,
    0x49727c0400000ULL,
    0xfd83626600000ULL,
    0x72c4d75500000ULL,
    0xcba9b8fd80000ULL,
    0x3fb27c72c0000ULL,
    0x502362cba0000ULL,
    0x1874d73fb0000ULL,
    0xe601b8d028000ULL,
    0x950e7cd87c000ULL,
    0x5d8962c60a000ULL,
    0x62c5d74501000ULL,
    0x33a638058f800ULL,
    0x2bb33c66c1400ULL,
    0xce2a8255a9e00ULL,
    0x5970c77eb4100ULL,
    0x58f8033a6380ULL,
    0x66c1402bb33c0ULL,
    0x55a9e0ce2a820ULL,
    0x7eb4105970c70ULL,
    0xb3a638058f808ULL,
    0xebb33c66c140cULL,
    0xee2a8255a9e02ULL,
    0x8970c77eb410dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0x1400000000000ULL,
    0xf600000000000ULL,
    0x6700000000000ULL,
    0x8f80000000000ULL,
    0x5040000000000ULL,
    0x8aa0000000000ULL,
    0xff0000000000ULL,
    0x12a8000000000ULL,
    0xabf4000000000ULL,
    0xfcaa000000000ULL,
    0x28fb000000000ULL,
    0xbd29800000000ULL,
    0xbba400000000ULL,
    0x4e06e00000000ULL,
    0x330c300000000ULL,
    0x5986180000000ULL,
    0xc74d340000000ULL,
    0x3d2cb20000000ULL,
    0x4bb2cb0000000ULL,
    0x6e06188000000ULL,
    0xc30d344000000ULL,
    0x618cb22000000ULL,
    0xd342cbf000000ULL,
    0xcb2e18b800000ULL,
    0x2cb9345400000ULL,
    0xe186b2d600000ULL,
    0x9349cb9700000ULL,
    0xeb2f983780000ULL,
    0xdcb7740440000ULL,
    0xd98a525ca0000ULL,
    0x874efb98f0000ULL,
    0x1d28002528000ULL,
    0xbbb400afb4000ULL,
    0x560a00a00a000ULL,
    0xd70b00b00b000ULL,
    0x9781801801800ULL,
    0xb44e40e40e400ULL,
    0x44ace0ce0ce00ULL,
    0x7cf7307307300ULL,
    0x6b2f987987980ULL,
    0x9cb7743743740ULL,
    0xf98a520520520ULL,
    0x774efb5fb5fb0ULL,
    0x2528001801808ULL,
    0xafb400e40e404ULL,
    0xa00a00ce0ce02ULL,
    0xb00b00730730fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x6800000000000ULL,
    0x6400000000000ULL,
    0x3600000000000ULL,
    0x6d00000000000ULL,
    0x4180000000000ULL,
    0xe040000000000ULL,
    0xd2e0000000000ULL,
    0x9bf0000000000ULL,
    0xce8000000000ULL,
    0x52fc000000000ULL,
    0x5b6a000000000ULL,
    0x2fb3000000000ULL,
    0xa00c800000000ULL,
    0x3005400000000ULL,
    0x4807e00000000ULL,
    0x940f900000000ULL,
    0x5e01f80000000ULL,
    0x90e940000000ULL,
    0x778a560000000ULL,
    0x8d416b0000000ULL,
    0x9369f88000000ULL,
    0x7bb294c000000ULL,
    0xde00562000000ULL,
    0xc9026bf000000ULL,
    0x578d78e800000ULL,
    0x7d4bd4a400000ULL,
    0xfb6db61600000ULL,
    0x1fbefb9d00000ULL,
    0xe80000a980000ULL,
    0xa400004440000ULL,
    0x160000c4e0000ULL,
    0x9d000006f0000ULL,
    0x2980002568000ULL,
    0x844000d6bc000ULL,
    0xe4e000bf8a000ULL,
    0xf6f000d943000ULL,
    0x4d6800ed64800ULL,
    0xb2bc0082b9400ULL,
    0x898a00c18de00ULL,
    0xb44300204c900ULL,
    0xace480f2e5780ULL,
    0x62f9406bf7d40ULL,
    0x136de064e7b60ULL,
    0xbbbc9036fdfb0ULL,
    0xfe0d786d64808ULL,
    0x390bd442b940cULL,
    0x3f8db6e18de02ULL,
    0x194efbd04c90fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0xf400000000000ULL,
    0xae00000000000ULL,
    0xbb00000000000ULL,
    0xe780000000000ULL,
    0x95c0000000000ULL,
    0x1c20000000000ULL,
    0xd030000000000ULL,
    0xdba8000000000ULL,
    0x55f4000000000ULL,
    0xff82000000000ULL,
    0x21c1000000000ULL,
    0x1223800000000ULL,
    0x3b3a400000000ULL,
    0xa42b600000000ULL,
    0x3430f00000000ULL,
    0x4da6980000000ULL,
    0x4af3ec0000000ULL,
    0x2e043a0000000ULL,
    0xfb0a1f0000000ULL,
    0x4785188000000ULL,
    0xc5c9ac4000000ULL,
    0x842f5aa000000ULL,
    0x243aef5000000ULL,
    0x75a3801800000ULL,
    0xeefa40b400000ULL,
    0x180b600e00000ULL,
    0xb400f0eb00000ULL,
    0xe0e987f80000ULL,
    0xeb07ec61c0000ULL,
    0x7f863ab220000ULL,
    0x61cb1f6b30000ULL,
    0xb22698bc28000ULL,
    0x6b33ec8034000ULL,
    0x3c243a43a2000ULL,
    0xc03a1fa1f1000ULL,
    0xe3ad18d18b800ULL,
    0xf1fdacdace400ULL,
    0xc98d5a55a9600ULL,
    0x6ecbeffef1f00ULL,
    0x5ba800a005180ULL,
    0x15f4005009ac0ULL,
    0x5f8200980f5a0ULL,
    0x71c100f40aef0ULL,
    0x8a2380ae0b808ULL,
    0xcf3a40bb0e404ULL,
    0xa2b60e78960aULL,
    0x8f30f095c1f05ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xb800000000000ULL,
    0x3c00000000000ULL,
    0xce00000000000ULL,
    0x4100000000000ULL,
    0x2180000000000ULL,
    0x51c0000000000ULL,
    0x960000000000ULL,
    0x8570000000000ULL,
    0xf278000000000ULL,
    0x8e9c000000000ULL,
    0x6002000000000ULL,
    0x7003000000000ULL,
    0x5803800000000ULL,
    0x8c02c00000000ULL,
    0x7602e00000000ULL,
    0x7d00f00000000ULL,
    0xef83380000000ULL,
    0x10c1040000000ULL,
    0x28e0860000000ULL,
    0xd4b1470000000ULL,
    0xfb18258000000ULL,
    0xbee15c000000ULL,
    0x9279c9e000000ULL,
    0xfe9d3a7000000ULL,
    0x3800000800000ULL,
    0xfc00000c00000ULL,
    0x2e00000e00000ULL,
    0xf100000b00000ULL,
    0x9980000b80000ULL,
    0x6dc00003c0000ULL,
    0xc760000ce0000ULL,
    0xc470000410000ULL,
    0xd3f8000218000ULL,
    0xdf5c00051c000ULL,
    0x6962000096000ULL,
    0xf573000857000ULL,
    0xaa7b800f27800ULL,
    0x29ec008e9c00ULL,
    0x1600e00600200ULL,
    0xd03f00700300ULL,
    0xb780b80580380ULL,
    0x9cc3c408c02c0ULL,
    0x5ee26607602e0ULL,
    0xa9b1b707d00f0ULL,
    0x149b1d8ef8338ULL,
    0x1b2f11c10c104ULL,
    0xba994fe28e086ULL,
    0x2a2c7d7d4b147ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x6800000000000ULL,
    0x3c00000000000ULL,
    0x8a00000000000ULL,
    0x5100000000000ULL,
    0xa980000000000ULL,
    0xddc0000000000ULL,
    0x5ba0000000000ULL,
    0x39d0000000000ULL,
    0x95f8000000000ULL,
    0x56d4000000000ULL,
    0xa02000000000ULL,
    0x9103000000000ULL,
    0x4983800000000ULL,
    0xdc3400000000ULL,
    0x33a1a00000000ULL,
    0x5d0f00000000ULL,
    0x1ffa280000000ULL,
    0x7d5440000000ULL,
    0xa380a60000000ULL,
    0x4cc0770000000ULL,
    0x1222ee8000000ULL,
    0x3413a74000000ULL,
    0xa65bf7e000000ULL,
    0x5305ab5000000ULL,
    0x15f8000800000ULL,
    0x96d4000c00000ULL,
    0xea02000e00000ULL,
    0x4103000d00000ULL,
    0x2183800680000ULL,
    0x31c34003c0000ULL,
    0xb9a1a008a0000ULL,
    0x54d0f00510000ULL,
    0xb67a280a98000ULL,
    0xda15440ddc000ULL,
    0xf820a605ba000ULL,
    0x751077039d000ULL,
    0x87daee895f800ULL,
    0x62c7a7456d400ULL,
    0xac59f7e0a0200ULL,
    0xc206ab5910300ULL,
    0x5c7b800c98380ULL,
    0x9b17400cdc340ULL,
    0xd9a3a00d3a1a0ULL,
    0x44d3f00d5d0f0ULL,
    0x3e79a8077fa28ULL,
    0x36160403bd544ULL,
    0x1a210602980a6ULL,
    0x18108701dc077ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x3800000000000ULL,
    0x8c00000000000ULL,
    0x7e00000000000ULL,
    0x7100000000000ULL,
    0xc880000000000ULL,
    0x4c0000000000ULL,
    0x1ba0000000000ULL,
    0xbb70000000000ULL,
    0x4a98000000000ULL,
    0xc3bc000000000ULL,
    0xa602000000000ULL,
    0x6d01000000000ULL,
    0xee81800000000ULL,
    0x29c3400000000ULL,
    0x9520e00000000ULL,
    0x42b2300000000ULL,
    0xe7b9f80000000ULL,
    0xd0dc40000000ULL,
    0x3fb9220000000ULL,
    0x110d130000000ULL,
    0x19bbee8000000ULL,
    0x3c0cadc000000ULL,
    0x973a4a6000000ULL,
    0xc5cf7ef000000ULL,
    0x3a18000800000ULL,
    0xb7c000400000ULL,
    0xa3a2000600000ULL,
    0x7771000d00000ULL,
    0x5499800380000ULL,
    0x62bf4008c0000ULL,
    0x5682e007e0000ULL,
    0xe5c3300710000ULL,
    0x8b20780c88000ULL,
    0xe3b284004c000ULL,
    0x173bc201ba000ULL,
    0x85ce230bb7000ULL,
    0x5a1b9684a9800ULL,
    0xdb7e29cc3bc00ULL,
    0x9ba1886a60200ULL,
    0xfb715df6d0100ULL,
    0x2a9b9686e8180ULL,
    0x13be29c69c340ULL,
    0x9e01886f520e0ULL,
    0xe1015df92b230ULL,
    0x90839685fb9f8ULL,
    0x58c229cc10dc4ULL,
    0x5da388621b922ULL,
    0x46705dfb00d13ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0x6400000000000ULL,
    0x6a00000000000ULL,
    0x8900000000000ULL,
    0xa580000000000ULL,
    0xcb40000000000ULL,
    0x1820000000000ULL,
    0xad90000000000ULL,
    0xaf88000000000ULL,
    0x72f4000000000ULL,
    0x2582000000000ULL,
    0xb43000000000ULL,
    0xb822800000000ULL,
    0x3d92400000000ULL,
    0xa788200000000ULL,
    0x16f5900000000ULL,
    0x4f83a80000000ULL,
    0x8241240000000ULL,
    0x1da0160000000ULL,
    0xf6d16d0000000ULL,
    0xbfa8408000000ULL,
    0xbb67264000000ULL,
    0xe009162000000ULL,
    0xf0b4efd000000ULL,
    0x3822800800000ULL,
    0xfd92400c00000ULL,
    0x788200a00000ULL,
    0x86f5900900000ULL,
    0x4783a80080000ULL,
    0xe641240640000ULL,
    0x77a01606a0000ULL,
    0x7fd16d0890000ULL,
    0x1a28408a58000ULL,
    0x7027264cb4000ULL,
    0xf829162182000ULL,
    0x5d24efdad9000ULL,
    0x97aa8002f8800ULL,
    0x8f66400b2f400ULL,
    0x220a200858200ULL,
    0x8db69009b4300ULL,
    0xffa1280b02280ULL,
    0xdbd3640599240ULL,
    0xd028360cd8820ULL,
    0x6924fd09ff590ULL,
    0x55abe88ea03a8ULL,
    0xf266024490124ULL,
    0xe589002058016ULL,
    0xabf582d5b416dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0x8c00000000000ULL,
    0xd600000000000ULL,
    0x3900000000000ULL,
    0xd580000000000ULL,
    0x3240000000000ULL,
    0xb2a0000000000ULL,
    0x7210000000000ULL,
    0x53d8000000000ULL,
    0x82cc000000000ULL,
    0xcb82000000000ULL,
    0x4743000000000ULL,
    0x9120800000000ULL,
    0xa953400000000ULL,
    0x7cf9200000000ULL,
    0x4e9e300000000ULL,
    0xfcf9580000000ULL,
    0x8e9fe40000000ULL,
    0xdcf9d60000000ULL,
    0x5e9c890000000ULL,
    0x94f96a8000000ULL,
    0xd29fb84000000ULL,
    0x42f9b76000000ULL,
    0xeb9c9f3000000ULL,
    0x9778800800000ULL,
    0xd9df400c00000ULL,
    0x25db200200000ULL,
    0xabcd300d00000ULL,
    0x7601d80480000ULL,
    0x2900a408c0000ULL,
    0xbd82f60d60000ULL,
    0x6e41b90390000ULL,
    0x2ca0b28d58000ULL,
    0xc7131c4324000ULL,
    0x5059416b2a000ULL,
    0x898e263721000ULL,
    0xaca0b28d3d800ULL,
    0x7131c442cc00ULL,
    0x7059416eb8200ULL,
    0x598e263974300ULL,
    0xe4a0b28592080ULL,
    0x8b131c4e55340ULL,
    0xa6594168af920ULL,
    0x608e263a79e30ULL,
    0x3120b28e17958ULL,
    0xb9531c4f0dfe4ULL,
    0x14f94169859d6ULL,
    0x129e263c58c89ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xd800000000000ULL,
    0xf400000000000ULL,
    0x3e00000000000ULL,
    0x9500000000000ULL,
    0x8f80000000000ULL,
    0x3d40000000000ULL,
    0xf320000000000ULL,
    0x2ef0000000000ULL,
    0xadc8000000000ULL,
    0xa0c000000000ULL,
    0x8b22000000000ULL,
    0x4af3000000000ULL,
    0x6bc8800000000ULL,
    0x3b0d400000000ULL,
    0xe2a1600000000ULL,
    0x16b0d00000000ULL,
    0x2968780000000ULL,
    0xbdbf140000000ULL,
    0x33cb5e0000000ULL,
    0xf0c250000000ULL,
    0xfca1b48000000ULL,
    0xd3b0afc000000ULL,
    0x7eeb692000000ULL,
    0x74fe4d3000000ULL,
    0xfee8780800000ULL,
    0xb4ff140c00000ULL,
    0xdeeb5e0200000ULL,
    0xe4fc250500000ULL,
    0x6e9b48d80000ULL,
    0x10fcafcf40000ULL,
    0x38e96923e0000ULL,
    0x85fd4d3950000ULL,
    0xb768f800f8000ULL,
    0xb8be540fd4000ULL,
    0x44483e0d32000ULL,
    0x964ff507ef000ULL,
    0xe9814c875c800ULL,
    0x9c42fbcfe0c00ULL,
    0x62a1572b52200ULL,
    0xd6b2b83dff300ULL,
    0x969b48644880ULL,
    0xedbcafcc64d40ULL,
    0xebc9692318160ULL,
    0xfb0d4d36840d0ULL,
    0xc2a0f80dca078ULL,
    0x46b254083b314ULL,
    0xf16a3e0a6e95eULL,
    0x49bcf5080ff25ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x9800000000000ULL,
    0xa400000000000ULL,
    0x7a00000000000ULL,
    0xd500000000000ULL,
    0x280000000000ULL,
    0x6040000000000ULL,
    0x51e0000000000ULL,
    0x8870000000000ULL,
    0x8c28000000000ULL,
    0x47c4000000000ULL,
    0xbe2000000000ULL,
    0xad71000000000ULL,
    0xb6aa800000000ULL,
    0x3386c00000000ULL,
    0xb800600000000ULL,
    0x5403900000000ULL,
    0x4203680000000ULL,
    0xc101940000000ULL,
    0xe0826a0000000ULL,
    0x1143110000000ULL,
    0x2960af8000000ULL,
    0x3d3175c000000ULL,
    0xdf4a3aa000000ULL,
    0xaff49e1000000ULL,
    0xd62b680800000ULL,
    0x62c5940400000ULL,
    0x31606a0a00000ULL,
    0xd932110b00000ULL,
    0x54a2f8980000ULL,
    0xcaf7b5ca40000ULL,
    0x4caa5aa7a0000ULL,
    0xa6870e1d50000ULL,
    0x1a80000828000ULL,
    0x8440000204000ULL,
    0x8be0000f1e000ULL,
    0xed70000387000ULL,
    0x16a8000142800ULL,
    0x8384000e3c400ULL,
    0x200200071e200ULL,
    0xf001000787100ULL,
    0x3802800b42a80ULL,
    0x1402c0053c6c0ULL,
    0xe202600e9e060ULL,
    0x7102900dc7390ULL,
    0x7881e80ce2b68ULL,
    0xb54354086c594ULL,
    0x53600a0eb606aULL,
    0xe831810bc3211ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x1800000000000ULL,
    0x3400000000000ULL,
    0x8a00000000000ULL,
    0x9d00000000000ULL,
    0x6780000000000ULL,
    0x8240000000000ULL,
    0x40e0000000000ULL,
    0x60f0000000000ULL,
    0x9148000000000ULL,
    0x2944000000000ULL,
    0x2d62000000000ULL,
    0xbfb3000000000ULL,
    0x162a800000000ULL,
    0xfbf4c00000000ULL,
    0xe4ca600000000ULL,
    0xc207d00000000ULL,
    0x2002a80000000ULL,
    0xf001b40000000ULL,
    0xb8037e0000000ULL,
    0x402190000000ULL,
    0x92034b8000000ULL,
    0xa90327c000000ULL,
    0xed81f32000000ULL,
    0x1f40d81000000ULL,
    0x2760280800000ULL,
    0xe2b1740c00000ULL,
    0xd1ab1e0a00000ULL,
    0x49b6c90300000ULL,
    0xbc2b638180000ULL,
    0x96f653c340000ULL,
    0x3b48ed28a0000ULL,
    0x44451119d0000ULL,
    0xf2e1cb8e78000ULL,
    0x39f3e7c424000ULL,
    0xc4c9932e0e000ULL,
    0x320408150f000ULL,
    0x9800000094800ULL,
    0xf400000dd4400ULL,
    0x2a00000076200ULL,
    0xad0000012b300ULL,
    0x7f8000069aa80ULL,
    0xb6400004db4c0ULL,
    0xcae00002e2a60ULL,
    0xfdf00003ff7d0ULL,
    0xf6c8000d6caa8ULL,
    0xab040005b05b4ULL,
    0x6d82000d5817eULL,
    0xdf43000db4119ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x800000000000ULL,
    0x4c00000000000ULL,
    0x200000000000ULL,
    0xb500000000000ULL,
    0x3680000000000ULL,
    0xc2c0000000000ULL,
    0x1420000000000ULL,
    0x750000000000ULL,
    0x1bf8000000000ULL,
    0x5034000000000ULL,
    0x48a2000000000ULL,
    0xac91000000000ULL,
    0xd35b800000000ULL,
    0xbca7400000000ULL,
    0x7bfa200000000ULL,
    0xc034300000000ULL,
    0xa0a1880000000ULL,
    0x3090940000000ULL,
    0xd95b7a0000000ULL,
    0x45a57b0000000ULL,
    0x4f7a788000000ULL,
    0xb7f6f94000000ULL,
    0x82013de000000ULL,
    0xf502dfd000000ULL,
    0xd682080800000ULL,
    0x12c3d40400000ULL,
    0x1c235a0e00000ULL,
    0x4b504b0d00000ULL,
    0x19f8708080000ULL,
    0xe5352d44c0000ULL,
    0x7e2267e020000ULL,
    0x6e5294db50000ULL,
    0xc77a788b68000ULL,
    0xbbf6f9482c000ULL,
    0x60013def42000ULL,
    0x9002dfdd75000ULL,
    0xe80208093f800ULL,
    0x9c03d405c3400ULL,
    0xa035a0aaa200ULL,
    0xf9004b0c99100ULL,
    0x3480708eddb80ULL,
    0x77c12d4326740ULL,
    0x22a067e6dda20ULL,
    0xc59394d726430ULL,
    0xfd9f880dd988ULL,
    0x5765b94e26494ULL,
    0x53591de65d97aULL,
    0xfca7efd3e647bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x6800000000000ULL,
    0x4c00000000000ULL,
    0x7600000000000ULL,
    0xf700000000000ULL,
    0x3680000000000ULL,
    0xd740000000000ULL,
    0x87e0000000000ULL,
    0xef30000000000ULL,
    0xa3a8000000000ULL,
    0xd544000000000ULL,
    0x23aa000000000ULL,
    0x1547000000000ULL,
    0xc3a9800000000ULL,
    0x4546400000000ULL,
    0xaba8200000000ULL,
    0x947700000000ULL,
    0xdda9f80000000ULL,
    0xfe44ac0000000ULL,
    0xeb29220000000ULL,
    0x2907f10000000ULL,
    0x6ccb3d8000000ULL,
    0xc6344dc000000ULL,
    0xcf61b32000000ULL,
    0x137318d000000ULL,
    0xeccb3d8800000ULL,
    0x6344dcc00000ULL,
    0x2f61b32e00000ULL,
    0x437318d500000ULL,
    0x84cb3d8e80000ULL,
    0x4a344dc8c0000ULL,
    0x5961b32960000ULL,
    0xb47318da70000ULL,
    0xb24b3d8de8000ULL,
    0x9d744dc5b4000ULL,
    0xde81b3211e000ULL,
    0x5b4318d483000ULL,
    0x11e33d87d2800ULL,
    0x48304dc8e0400ULL,
    0xfd2bb32324a00ULL,
    0x4e0418d5d7700ULL,
    0xd24abd8be8180ULL,
    0xd760dccb4240ULL,
    0x568393299e220ULL,
    0x474368d543070ULL,
    0xfe3458632878ULL,
    0xf332a1c3506ecULL,
    0xbdaab1272cb02ULL,
    0x6e4499d7d3781ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0xc800000000000ULL,
    0x7400000000000ULL,
    0x5200000000000ULL,
    0x300000000000ULL,
    0xeb80000000000ULL,
    0x6f40000000000ULL,
    0x6460000000000ULL,
    0xdaf0000000000ULL,
    0x1798000000000ULL,
    0x297c000000000ULL,
    0xa59a000000000ULL,
    0xfa7d000000000ULL,
    0xe61b800000000ULL,
    0x713f400000000ULL,
    0x1878a00000000ULL,
    0xdcce900000000ULL,
    0xb661e80000000ULL,
    0x99f29c0000000ULL,
    0x9c18460000000ULL,
    0xd63e210000000ULL,
    0x9fa578000000ULL,
    0x548e0ac000000ULL,
    0xa380a9e000000ULL,
    0x5b413f3000000ULL,
    0x5662578800000ULL,
    0x49f20ac400000ULL,
    0x341aa9e600000ULL,
    0x323c3f3900000ULL,
    0x93f9d78480000ULL,
    0x238d4ac340000ULL,
    0x1a0209e320000ULL,
    0x3702af3930000ULL,
    0xd9803f8a38000ULL,
    0xfc43d6c5b4000ULL,
    0x47e04fe566000ULL,
    0xc1b18e349f000ULL,
    0x21f9e80b41800ULL,
    0xf08e9c0723c00ULL,
    0x5982460f3fa00ULL,
    0xbc43210b38d00ULL,
    0x27e1d78d20380ULL,
    0x51b14ac430340ULL,
    0xe9f809e8b82a0ULL,
    0x848faf3ff4390ULL,
    0xb83bf82c6268ULL,
    0xbf4096ceef1dcULL,
    0xcc62efe259ae6ULL,
    0x3ef21e3ba7db1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xb800000000000ULL,
    0x400000000000ULL,
    0x6e00000000000ULL,
    0x9700000000000ULL,
    0xf280000000000ULL,
    0xedc0000000000ULL,
    0x1360000000000ULL,
    0x5c90000000000ULL,
    0xdb58000000000ULL,
    0x31e4000000000ULL,
    0x9da000000000ULL,
    0xcc27000000000ULL,
    0x2b8800000000ULL,
    0x44b4400000000ULL,
    0xfe2600000000ULL,
    0xe650500000000ULL,
    0x9ab9d80000000ULL,
    0x50b50c0000000ULL,
    0x79e2920000000ULL,
    0xa552fb0000000ULL,
    0xbe38bf8000000ULL,
    0x2e77d94000000ULL,
    0xf6000ae000000ULL,
    0x830112d000000ULL,
    0x84803f8800000ULL,
    0xaec3994c00000ULL,
    0x37e26aea00000ULL,
    0x225142dd00000ULL,
    0x54b9e78380000ULL,
    0x17b6954c40000ULL,
    0x3360f8ece0000ULL,
    0x4c93b9d470000ULL,
    0xc359580ca8000ULL,
    0xe5e54c029c000ULL,
    0xdfdaf20dd6000ULL,
    0x5f25ab01b9000ULL,
    0x9e39e7891d800ULL,
    0x3e76954d82400ULL,
    0xee00f8e74ba00ULL,
    0x5703b9d07b700ULL,
    0x5281580ab6080ULL,
    0x3dc14c0589040ULL,
    0xab60f20b55860ULL,
    0x5892ab0a6e750ULL,
    0xb559678fb5958ULL,
    0xa6e6d5421e54cULL,
    0xfb5898e11daf2ULL,
    0x21e4e9d1825abULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x6800000000000ULL,
    0xec00000000000ULL,
    0x2200000000000ULL,
    0x2b00000000000ULL,
    0x3680000000000ULL,
    0x9d40000000000ULL,
    0x6a20000000000ULL,
    0x1670000000000ULL,
    0x4de8000000000ULL,
    0x330c000000000ULL,
    0x936a000000000ULL,
    0x824f000000000ULL,
    0x3b49800000000ULL,
    0x8f3fc00000000ULL,
    0x2820200000000ULL,
    0xcd70700000000ULL,
    0xf36aa80000000ULL,
    0x724fdc0000000ULL,
    0xb34bf20000000ULL,
    0x533e690000000ULL,
    0x62207a8000000ULL,
    0xa7140c000000ULL,
    0xe7ea652000000ULL,
    0xc40d90f000000ULL,
    0xefe9fa8800000ULL,
    0xd80e80cc00000ULL,
    0x45ea452e00000ULL,
    0x2f0de0f300000ULL,
    0x396b528e80000ULL,
    0x754d5cc2c0000ULL,
    0x47cbb72c20000ULL,
    0xd57c89f1b0000ULL,
    0x5682a80de8000ULL,
    0x6d43dc0b14000ULL,
    0xe221f20a82000ULL,
    0xca716900d7000ULL,
    0x7e9fa8136800ULL,
    0xf40e80c424c00ULL,
    0x87ea452db4a00ULL,
    0x340de0fbf3f00ULL,
    0x67eb528c02180ULL,
    0x40d5cce173c0ULL,
    0xfebb72316a20ULL,
    0xe80c89f694f70ULL,
    0x2deaa806dcb28ULL,
    0xc30fdc0e27c1cULL,
    0x1b6bf20ca01d2ULL,
    0x5e4e690070119ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x2800000000000ULL,
    0xd400000000000ULL,
    0x8a00000000000ULL,
    0xff00000000000ULL,
    0x8480000000000ULL,
    0x73c0000000000ULL,
    0x1320000000000ULL,
    0xc2b0000000000ULL,
    0xfb38000000000ULL,
    0x361c000000000ULL,
    0x401a000000000ULL,
    0xe0af000000000ULL,
    0x1122800000000ULL,
    0x19b3c00000000ULL,
    0xfdb8200000000ULL,
    0x5edf900000000ULL,
    0x75b8880000000ULL,
    0x7adfac0000000ULL,
    0xf7baba0000000ULL,
    0x61ddf30000000ULL,
    0xd1387e8000000ULL,
    0x391e55c000000ULL,
    0xcc9ba86000000ULL,
    0x776cbeb000000ULL,
    0xa000f68800000ULL,
    0xf001f9cc00000ULL,
    0x801126200000ULL,
    0xe4014db300000ULL,
    0xa200880a80000ULL,
    0x2b03ac0140000ULL,
    0xe80ba0aa0000ULL,
    0x8cc2f30cf0000ULL,
    0x97a2fe8ac8000ULL,
    0xb17195ca7c000ULL,
    0xe819886992000ULL,
    0xf4ac2eb3db000ULL,
    0xbb22fe8ffb800ULL,
    0xd6b195c85dc00ULL,
    0x5139886733a00ULL,
    0xf91c2eb121f00ULL,
    0xec9afe8421a80ULL,
    0x476d95c3bafc0ULL,
    0x880388617a220ULL,
    0x24032eb717090ULL,
    0x82007e8481a08ULL,
    0x1b0255c04ac6cULL,
    0x2681a86b3229aULL,
    0x58c3beb32b263ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xb800000000000ULL,
    0x8400000000000ULL,
    0x1a00000000000ULL,
    0xaf00000000000ULL,
    0xbd80000000000ULL,
    0xdfc0000000000ULL,
    0x14e0000000000ULL,
    0x4350000000000ULL,
    0xda38000000000ULL,
    0x4e1c000000000ULL,
    0x4cda000000000ULL,
    0x364d000000000ULL,
    0x2960800000000ULL,
    0xdc90400000000ULL,
    0x6ed8600000000ULL,
    0x5d4f500000000ULL,
    0x2ee0880000000ULL,
    0xfc51ac0000000ULL,
    0x7fb81e0000000ULL,
    0x45dc830000000ULL,
    0xfa3a458000000ULL,
    0x5e1d624000000ULL,
    0x54dbd36000000ULL,
    0xe24ec93000000ULL,
    0x8b62cd8800000ULL,
    0xf790ce4400000ULL,
    0xc959cd6a00000ULL,
    0x2d8f4a3500000ULL,
    0x8780080380000ULL,
    0x60c1ec0c40000ULL,
    0xb1607e0ba0000ULL,
    0x4893d30ff0000ULL,
    0x6cdacd8058000ULL,
    0x264cce45bc000ULL,
    0x3163cd60ee000ULL,
    0x8924a3ec5000ULL,
    0xccd8880e7b800ULL,
    0x764dac0d1dc00ULL,
    0x89621e0f83a00ULL,
    0x8c91830251d00ULL,
    0xd6dac584b5880ULL,
    0xd94d224168c40ULL,
    0x34e3b36380260ULL,
    0x5351993c40250ULL,
    0xc23a4583a0008ULL,
    0x9a1d624bf01ecULL,
    0xeedbd36a5807eULL,
    0x1d4ec930bc3d3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0xf400000000000ULL,
    0xf600000000000ULL,
    0x8b00000000000ULL,
    0xc980000000000ULL,
    0x5540000000000ULL,
    0x6720000000000ULL,
    0xf3f0000000000ULL,
    0x3478000000000ULL,
    0x5744000000000ULL,
    0x1ada000000000ULL,
    0xb1f5000000000ULL,
    0xa981800000000ULL,
    0x6540c00000000ULL,
    0x8f23a00000000ULL,
    0x77f2100000000ULL,
    0xca7bf80000000ULL,
    0x2845fc0000000ULL,
    0x255afe0000000ULL,
    0x6fb6790000000ULL,
    0x7233a8000000ULL,
    0xc3f25ac000000ULL,
    0xdc7aed6000000ULL,
    0xd34482d000000ULL,
    0xe4d9428800000ULL,
    0xcef766c400000ULL,
    0x9603b36e00000ULL,
    0xbb00ebd700000ULL,
    0x2181800880000ULL,
    0xd140c00b40000ULL,
    0x9923a00160000ULL,
    0x8cf2100fb0000ULL,
    0xbfbf80c18000ULL,
    0x8905fc0a14000ULL,
    0xb47afe0912000ULL,
    0x174679078f000ULL,
    0xfadb3a8fdf800ULL,
    0xc1f65ac020400ULL,
    0xa180ed67dfa00ULL,
    0x914182d420500ULL,
    0x7920c281df980ULL,
    0xfcf3a6c7204c0ULL,
    0x3fa13675f9a0ULL,
    0x7d07fbdb60710ULL,
    0x427bf80ebfa78ULL,
    0x9c45fc0f9053cULL,
    0x335afe0bc795eULL,
    0x94b6790e34469ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x6800000000000ULL,
    0xf400000000000ULL,
    0x6200000000000ULL,
    0xdf00000000000ULL,
    0x7980000000000ULL,
    0xdd40000000000ULL,
    0x76e0000000000ULL,
    0x2cf0000000000ULL,
    0xcfb8000000000ULL,
    0x51ec000000000ULL,
    0xc8da000000000ULL,
    0x845d000000000ULL,
    0x9b81800000000ULL,
    0x4243400000000ULL,
    0xef62200000000ULL,
    0x61b1900000000ULL,
    0xd158280000000ULL,
    0x891cac0000000ULL,
    0x65626e0000000ULL,
    0xab1090000000ULL,
    0x2adbbd8000000ULL,
    0x1b5d86c000000ULL,
    0x201456000000ULL,
    0xf03247000000ULL,
    0xf182158800000ULL,
    0xb9426ac400000ULL,
    0x7ce10b6e00000ULL,
    0x7f3bd7900000ULL,
    0xd439800e80000ULL,
    0x53af400b40000ULL,
    0xc7b8200820000ULL,
    0x75ec9004f0000ULL,
    0x22d9a80118000ULL,
    0x3f5fec0294000ULL,
    0xe8004e014e000ULL,
    0xb400990f3f000ULL,
    0x8203958b63800ULL,
    0x4f012ac8cac00ULL,
    0x11832b6be3a00ULL,
    0x29422d7a8ad00ULL,
    0x14e1a80d43980ULL,
    0xf3f3ec053af40ULL,
    0xb63a4e0c7b820ULL,
    0x8cad99075ec90ULL,
    0xbe3a15822d9a8ULL,
    0xa8ae6ac3f5fecULL,
    0x543b0b6e8004eULL,
    0x13aebd7b40099ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x1800000000000ULL,
    0xdc00000000000ULL,
    0x4200000000000ULL,
    0x3700000000000ULL,
    0x2080000000000ULL,
    0xf140000000000ULL,
    0x2860000000000ULL,
    0x9490000000000ULL,
    0x8788000000000ULL,
    0xa83c000000000ULL,
    0x556a000000000ULL,
    0xe6ef000000000ULL,
    0xf803800000000ULL,
    0x4c02400000000ULL,
    0x3a01e00000000ULL,
    0xbb02300000000ULL,
    0x7a81680000000ULL,
    0x1a43ac0000000ULL,
    0x4ae18a0000000ULL,
    0x52d3190000000ULL,
    0x8f68238000000ULL,
    0xcded974000000ULL,
    0xfa80bfa000000ULL,
    0xda43f2b000000ULL,
    0x2ae2cb8800000ULL,
    0x2d07b4c00000ULL,
    0x976ad5a600000ULL,
    0x11eddbb500000ULL,
    0xb880000980000ULL,
    0xed400001c0000ULL,
    0xa60000220000ULL,
    0xf390000670000ULL,
    0xbf08000388000ULL,
    0x857c0002d4000ULL,
    0x3f0a0006a6000ULL,
    0x457f000a39000ULL,
    0x5f0b800a70800ULL,
    0x157e400597c00ULL,
    0x470be007d0a00ULL,
    0xc97d300727f00ULL,
    0x50ae807f8b80ULL,
    0xfe7dec0e43e40ULL,
    0x258a6a06f6be0ULL,
    0xf3e2905ded30ULL,
    0xdeacb88282e8ULL,
    0x9bac7b45641ecULL,
    0x8a60d5a70e06aULL,
    0x3392dbbe9d129ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xf800000000000ULL,
    0x3400000000000ULL,
    0x6200000000000ULL,
    0xf500000000000ULL,
    0xa880000000000ULL,
    0xfcc0000000000ULL,
    0x8e20000000000ULL,
    0x53f0000000000ULL,
    0xc778000000000ULL,
    0x9574000000000ULL,
    0xb802000000000ULL,
    0xd4e5000000000ULL,
    0xb280800000000ULL,
    0xfdc0c00000000ULL,
    0x64a0200000000ULL,
    0xaa30f00000000ULL,
    0x19d8f80000000ULL,
    0xe44340000000ULL,
    0x935a620000000ULL,
    0xe761f50000000ULL,
    0x657a288000000ULL,
    0x40913cc000000ULL,
    0xe0022e2000000ULL,
    0xd0e563f000000ULL,
    0x8809f7800000ULL,
    0xccc0917400000ULL,
    0x5620020200000ULL,
    0x97f0e5e500000ULL,
    0x5d78800080000ULL,
    0x5474c000c0000ULL,
    0x7282200020000ULL,
    0xdd25f000f0000ULL,
    0x94207800f8000ULL,
    0x52f0f40034000ULL,
    0x2df8420062000ULL,
    0x6cb40500f5000ULL,
    0x66a25080a8800ULL,
    0x4fd5c8c0fcc00ULL,
    0x99d86c208e200ULL,
    0xce4466f053f00ULL,
    0xb35a4ff8c7780ULL,
    0x176199b495740ULL,
    0x9d7a4e22b8020ULL,
    0x74917315d4e50ULL,
    0x8202b7f832808ULL,
    0x25e5adb43dc0cULL,
    0xa0002c2244a02ULL,
    0x300086155a30fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xf800000000000ULL,
    0xec00000000000ULL,
    0x7e00000000000ULL,
    0x6100000000000ULL,
    0x5c80000000000ULL,
    0xe6c0000000000ULL,
    0xdda0000000000ULL,
    0x2a70000000000ULL,
    0x9338000000000ULL,
    0x13cc000000000ULL,
    0xd3ce000000000ULL,
    0x7379000000000ULL,
    0x83a0800000000ULL,
    0x7b70c00000000ULL,
    0x97b8a00000000ULL,
    0xe90cf00000000ULL,
    0x886ef80000000ULL,
    0xd409ec0000000ULL,
    0x3218fe0000000ULL,
    0xef7ca10000000ULL,
    0xc556fc8000000ULL,
    0x56c516c000000ULL,
    0x4556a5a000000ULL,
    0x96c5067000000ULL,
    0xe556cd3800000ULL,
    0x66c542cc00000ULL,
    0x1d56574e00000ULL,
    0x8ac549b900000ULL,
    0x6356f80080000ULL,
    0xebc5ec00c0000ULL,
    0x3fd6fe00a0000ULL,
    0xd05a100f0000ULL,
    0xe2767c80f8000ULL,
    0x2775d6c0ec000ULL,
    0x714e05a07e000ULL,
    0x34b9f67061000ULL,
    0xa28035385c800ULL,
    0x47c0aecce6c00ULL,
    0x2120a94edda00ULL,
    0x3cb0e8b92a700ULL,
    0xb698848013380ULL,
    0xd5bc3ac0d3cc0ULL,
    0x3ef6fba073ce0ULL,
    0x1b5577083790ULL,
    0xceec9b87ba08ULL,
    0xeec9b80c9770cULL,
    0xc9b80ceee9b8aULL,
    0xb80ceec9880cfULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x5800000000000ULL,
    0x2c00000000000ULL,
    0x9a00000000000ULL,
    0xf900000000000ULL,
    0x3c80000000000ULL,
    0xb2c0000000000ULL,
    0xad20000000000ULL,
    0x3a30000000000ULL,
    0x8998000000000ULL,
    0x448c000000000ULL,
    0x2eea000000000ULL,
    0x6f81000000000ULL,
    0xef20800000000ULL,
    0x2f30c00000000ULL,
    0xf18200000000ULL,
    0xbf4cb00000000ULL,
    0xe74a580000000ULL,
    0xcb712c0000000ULL,
    0x51981a0000000ULL,
    0xa88c390000000ULL,
    0x94ea1c8000000ULL,
    0x268102c000000ULL,
    0x8ba0752000000ULL,
    0xb1f0d63000000ULL,
    0x3838339800000ULL,
    0x7c7c0d8c00000ULL,
    0x52524a6a00000ULL,
    0x3d3df14100000ULL,
    0xd252580080000ULL,
    0xfd3d2c00c0000ULL,
    0xf2521a0020000ULL,
    0x4d3d3900b0000ULL,
    0xaa529c8058000ULL,
    0x613dc2c02c000ULL,
    0x305255209a000ULL,
    0x983d6630f9000ULL,
    0xcd2eb983c800ULL,
    0x2afde18cb2c00ULL,
    0xa1f2706aad200ULL,
    0x10cd78413a300ULL,
    0x286a1c8009980ULL,
    0x544102c0848c0ULL,
    0x68075200eea0ULL,
    0x3bc0d630df810ULL,
    0xe9a03398b7208ULL,
    0x14f00d8c0330cULL,
    0xe6b84a6a95182ULL,
    0xabbcf141464cbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0xac00000000000ULL,
    0x8e00000000000ULL,
    0x900000000000ULL,
    0x9e80000000000ULL,
    0xa1c0000000000ULL,
    0xcaa0000000000ULL,
    0x3370000000000ULL,
    0x9578000000000ULL,
    0x85c000000000ULL,
    0x24b6000000000ULL,
    0x6a35000000000ULL,
    0x4378800000000ULL,
    0x6d5cc00000000ULL,
    0x1436200000000ULL,
    0x72f5b00000000ULL,
    0xcf58580000000ULL,
    0x53ec6c0000000ULL,
    0xc5eeae0000000ULL,
    0x40d9b90000000ULL,
    0xe016c68000000ULL,
    0x9045cdc000000ULL,
    0x6880e4a000000ULL,
    0x74c04a7000000ULL,
    0x2220f3f800000ULL,
    0x87b0b59c00000ULL,
    0x9758b81600000ULL,
    0x3fecfc4500000ULL,
    0x6beec68080000ULL,
    0xf9d9cdc0c0000ULL,
    0xa696e4a020000ULL,
    0x9d854a70b0000ULL,
    0x2c2073f8d8000ULL,
    0x4eb0759cac000ULL,
    0x29d898168e000ULL,
    0x2e2c4c4509000ULL,
    0x794e1e801e800ULL,
    0x66a961c061c00ULL,
    0xbdee6aa0eaa00ULL,
    0x9cd9437083700ULL,
    0x9616ed784d780ULL,
    0x8545d45ca45c0ULL,
    0xa000d2b6aab60ULL,
    0x7000bf3563350ULL,
    0xf800abf85df88ULL,
    0x1c00d99c0c9ccULL,
    0x56001616fe962ULL,
    0xa5004545f185bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0x2c00000000000ULL,
    0xa200000000000ULL,
    0x2d00000000000ULL,
    0xda80000000000ULL,
    0xf940000000000ULL,
    0xec60000000000ULL,
    0x2b0000000000ULL,
    0x3d48000000000ULL,
    0x825c000000000ULL,
    0x7d4a000000000ULL,
    0x6261000000000ULL,
    0x8dc8800000000ULL,
    0xca1c400000000ULL,
    0xa1aae00000000ULL,
    0x6891f00000000ULL,
    0x8c60280000000ULL,
    0xb2b06c0000000ULL,
    0x7548420000000ULL,
    0x5e5cdd0000000ULL,
    0x774a728000000ULL,
    0x6361d54000000ULL,
    0xf548ce6000000ULL,
    0x1e5c6fb000000ULL,
    0x974a07c800000ULL,
    0x93618b1c00000ULL,
    0x5d48b92a00000ULL,
    0x325c0cd100000ULL,
    0x354af28080000ULL,
    0xbe61954040000ULL,
    0x87c82e60e0000ULL,
    0xcb1c9fb0f0000ULL,
    0xd92aafc8a8000ULL,
    0xbcd1a71c2c000ULL,
    0xba801b2aa2000ULL,
    0x494021d12d000ULL,
    0xa46028005a800ULL,
    0xdeb06c00b9400ULL,
    0x374842000c600ULL,
    0x835cdd00f2b00ULL,
    0x5ca728095480ULL,
    0xb621d540ae5c0ULL,
    0xbb28ce60df4a0ULL,
    0x31ec6fb04f610ULL,
    0x708207c857488ULL,
    0xe87d8b1c335c4ULL,
    0xcc62b92a4dcaeULL,
    0x528d0cd16a21fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0x7c00000000000ULL,
    0x8200000000000ULL,
    0x4f00000000000ULL,
    0xbe80000000000ULL,
    0xedc0000000000ULL,
    0x2160000000000ULL,
    0xab70000000000ULL,
    0x7868000000000ULL,
    0x746c000000000ULL,
    0x1e9a000000000ULL,
    0xfdcb000000000ULL,
    0x3908800000000ULL,
    0x2f1cc00000000ULL,
    0x4ef2e00000000ULL,
    0xc5a7300000000ULL,
    0x6d92480000000ULL,
    0xe1d7bc0000000ULL,
    0x4b7ae20000000ULL,
    0x487bbf0000000ULL,
    0xbc80168000000ULL,
    0x62c061c000000ULL,
    0x7fe08b6000000ULL,
    0x76b0a87000000ULL,
    0x91088ce800000ULL,
    0xa31caaac00000ULL,
    0xe4f2037a00000ULL,
    0xc6a7f47b00000ULL,
    0x99125e8080000ULL,
    0x3f17ddc0c0000ULL,
    0x569a6960e0000ULL,
    0x41cb177030000ULL,
    0x5b089a68c8000ULL,
    0x501ccb6c7c000ULL,
    0x3872881a82000ULL,
    0x54675c0b4f000ULL,
    0xcef2d2683e800ULL,
    0x5a7776c2dc00ULL,
    0x8d926a1ac1600ULL,
    0xd1d7e30b9b700ULL,
    0x837a44e830680ULL,
    0x347bd6acc86c0ULL,
    0x3e80017a7c9a0ULL,
    0x2dc07b7b82cb0ULL,
    0xc1608000cf888ULL,
    0x9b70c0007edccULL,
    0xb068e0000d92eULL,
    0x86c300011d73ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x9800000000000ULL,
    0x2c00000000000ULL,
    0x600000000000ULL,
    0xcd00000000000ULL,
    0x8a80000000000ULL,
    0x1bc0000000000ULL,
    0xffa0000000000ULL,
    0xad50000000000ULL,
    0x7af8000000000ULL,
    0xb3dc000000000ULL,
    0x5b2e000000000ULL,
    0x1f29000000000ULL,
    0x9d58800000000ULL,
    0xf28cc00000000ULL,
    0x7d6200000000ULL,
    0x71f5100000000ULL,
    0xd4f6180000000ULL,
    0xda65ec0000000ULL,
    0x632ea60000000ULL,
    0xe3291d0000000ULL,
    0x2358b28000000ULL,
    0x38ce7c000000ULL,
    0x135641a000000ULL,
    0x8b355c5000000ULL,
    0xa7d6ee7800000ULL,
    0xa1f5891c00000ULL,
    0x6cf6880e00000ULL,
    0xe665b4b900000ULL,
    0xfd2eaa8080000ULL,
    0x2290bc0c0000ULL,
    0xafd8e7a020000ULL,
    0xd54c415010000ULL,
    0x66765cf898000ULL,
    0x3da56edc2c000ULL,
    0x228ec9ae06000ULL,
    0xbf79e8e9cd000ULL,
    0x4d20c4f80a800ULL,
    0x4a9042dcdbc00ULL,
    0x3b584faedfa00ULL,
    0xef8ce5e9bd500ULL,
    0x3556ee7862f80ULL,
    0x5635891c5fdc0ULL,
    0xb556880e7d2e0ULL,
    0x9635b4b9c2290ULL,
    0x9556aa800fd88ULL,
    0x86350bc0054ccULL,
    0xd56e7a0de762ULL,
    0xaa35415001a51ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0x5400000000000ULL,
    0x3a00000000000ULL,
    0x9d00000000000ULL,
    0x7e80000000000ULL,
    0x7f40000000000ULL,
    0x1720000000000ULL,
    0xab50000000000ULL,
    0x6df8000000000ULL,
    0x96a4000000000ULL,
    0x83d2000000000ULL,
    0x71e1000000000ULL,
    0xc0d8800000000ULL,
    0xe0f4400000000ULL,
    0x30aaa00000000ULL,
    0x805900000000ULL,
    0xcc2a180000000ULL,
    0x6e45140000000ULL,
    0xa78a1a0000000ULL,
    0xe3554d0000000ULL,
    0x1d2c68000000ULL,
    0x68e1fb4000000ULL,
    0xbc58952000000ULL,
    0xc6b4b25000000ULL,
    0xfb0a117800000ULL,
    0x1515b0e400000ULL,
    0xf272c87200000ULL,
    0xb1f12cf100000ULL,
    0x2000de8080000ULL,
    0xd000ef4040000ULL,
    0x38008f20a0000ULL,
    0xc400ff5090000ULL,
    0xa20057f898000ULL,
    0xc9000ba454000ULL,
    0x4480fd523a000ULL,
    0xe2400ea19d000ULL,
    0x69a0d7f8fe800ULL,
    0xd4104ba43f400ULL,
    0x7ad85d52b7200ULL,
    0x3df49ea13b500ULL,
    0xee2a4ff875f80ULL,
    0xe7451fa482a40ULL,
    0x430ae75219d20ULL,
    0x911543a17ce10ULL,
    0xf0721178a6588ULL,
    0xe8f1b0e48bb44ULL,
    0xfc80c872bd8aaULL,
    0x66402cf1ae559ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x800000000000ULL,
    0x8400000000000ULL,
    0xb200000000000ULL,
    0xb900000000000ULL,
    0xbe80000000000ULL,
    0x4fc0000000000ULL,
    0x5560000000000ULL,
    0xf8f0000000000ULL,
    0xac28000000000ULL,
    0x66d4000000000ULL,
    0xb30a000000000ULL,
    0x8bb5000000000ULL,
    0xc7c8800000000ULL,
    0x11e4c00000000ULL,
    0xaa42e00000000ULL,
    0xa591b00000000ULL,
    0xd0ea880000000ULL,
    0x7885440000000ULL,
    0x6c80d20000000ULL,
    0x86c0c90000000ULL,
    0x3e0568000000ULL,
    0x83307bc000000ULL,
    0x4348ef6000000ULL,
    0xa324c5f000000ULL,
    0x13a2a0a800000ULL,
    0x1ba1901400000ULL,
    0x9f22d8ea00000ULL,
    0x2d61fc8500000ULL,
    0x94c25e8080000ULL,
    0x2a51ffc0c0000ULL,
    0x658add60e0000ULL,
    0x3075bcf0b0000ULL,
    0xc8a87e2808000ULL,
    0x6414afd484000ULL,
    0x2eae58ab2000ULL,
    0xb185f075b9000ULL,
    0x3a00a8a83e800ULL,
    0xfd0014148fc00ULL,
    0xec80eaeab5600ULL,
    0x46c0858548f00ULL,
    0xe3e0800024280ULL,
    0x3330c00022d40ULL,
    0x4b48e000e10a0ULL,
    0x2724b00082b50ULL,
    0xa1a2080071488ULL,
    0xa2a18400da24cULL,
    0x21a232004d22eULL,
    0x62a17900e461bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x7800000000000ULL,
    0x2400000000000ULL,
    0x9e00000000000ULL,
    0x4700000000000ULL,
    0x6780000000000ULL,
    0xf740000000000ULL,
    0xdf20000000000ULL,
    0xb310000000000ULL,
    0x7168000000000ULL,
    0x8c4c000000000ULL,
    0x3252000000000ULL,
    0xe5d5000000000ULL,
    0xaa52800000000ULL,
    0x31d5c00000000ULL,
    0x2c52e00000000ULL,
    0x62d5f00000000ULL,
    0xadd2980000000ULL,
    0xf695d40000000ULL,
    0x8b72060000000ULL,
    0xf5c5930000000ULL,
    0x42ba618000000ULL,
    0x3dd9644000000ULL,
    0xdea0bea000000ULL,
    0xe750d75000000ULL,
    0x37c84fc800000ULL,
    0xbf1c9b1c00000ULL,
    0x839a1d9a00000ULL,
    0x9c94ec900000ULL,
    0xa8484fc880000ULL,
    0xac5c9b1cc0000ULL,
    0xa2ba1d9a60000ULL,
    0xcdd94ec930000ULL,
    0xc6a04fc8f8000ULL,
    0xf3509b1ce4000ULL,
    0xd1c81d9afe000ULL,
    0xdc1c4ec977000ULL,
    0x7a1acfc89f800ULL,
    0xb9895b1c13400ULL,
    0x10e8fd9a21200ULL,
    0xe80cbec9c4100ULL,
    0xcf2d7c8eee80ULL,
    0xf2854f1c9f0c0ULL,
    0x859a1b9a13720ULL,
    0x9ac9ddc921c50ULL,
    0x49c8ae4844ba8ULL,
    0x81c3f5caed9cULL,
    0xfc1a433a3f20eULL,
    0xea8969994310fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x7800000000000ULL,
    0x9c00000000000ULL,
    0xee00000000000ULL,
    0x1b00000000000ULL,
    0xcb80000000000ULL,
    0xc340000000000ULL,
    0xc7a0000000000ULL,
    0x510000000000ULL,
    0x8868000000000ULL,
    0xc474000000000ULL,
    0x225a000000000ULL,
    0x3da1000000000ULL,
    0x345a800000000ULL,
    0x7aa1c00000000ULL,
    0xf1da600000000ULL,
    0x12e1700000000ULL,
    0x85fa180000000ULL,
    0x48b1ec0000000ULL,
    0x2432f60000000ULL,
    0x92d5f70000000ULL,
    0x45803d8000000ULL,
    0xa840344000000ULL,
    0x94207a2000000ULL,
    0xea50f15000000ULL,
    0xd9c8124800000ULL,
    0x4664852400000ULL,
    0x8fb2481200000ULL,
    0x2195248500000ULL,
    0x1a20124880000ULL,
    0x81508524c0000ULL,
    0x8a484812e0000ULL,
    0xa9242485b0000ULL,
    0xde129248f8000ULL,
    0xa38545245c000ULL,
    0xb7c828120e000ULL,
    0x9d645485ab000ULL,
    0xa4320a4833800ULL,
    0x52d569249f400ULL,
    0xa580be12c9a00ULL,
    0x1840d385ae100ULL,
    0xec202fc8bbe80ULL,
    0x7650b1645b340ULL,
    0x37c83232ebfa0ULL,
    0x5d64d5d593b10ULL,
    0x443280008fb28ULL,
    0xe2d5c0002195cULL,
    0xdd8060001a206ULL,
    0x8440700081507ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x5800000000000ULL,
    0x7c00000000000ULL,
    0xc200000000000ULL,
    0xe100000000000ULL,
    0xd80000000000ULL,
    0xd7c0000000000ULL,
    0x2aa0000000000ULL,
    0xf530000000000ULL,
    0x9ba8000000000ULL,
    0xc0f4000000000ULL,
    0x20c6000000000ULL,
    0x702f000000000ULL,
    0x4866800000000ULL,
    0x241f400000000ULL,
    0xbe4ee00000000ULL,
    0x232b500000000ULL,
    0xec28b80000000ULL,
    0xda342c0000000ULL,
    0xfde6fa0000000ULL,
    0xdfdf8d0000000ULL,
    0x6eee178000000ULL,
    0x5b1b0ac000000ULL,
    0xe000052000000ULL,
    0x500093f000000ULL,
    0x3800848800000ULL,
    0x6c008e0400000ULL,
    0x9a000bce00000ULL,
    0x9d00d8eb00000ULL,
    0xcf803c8880000ULL,
    0x36c0a20440000ULL,
    0x2720f1ce60000ULL,
    0x22f055eb10000ULL,
    0xb108ab08d8000ULL,
    0x35c4e8c43c000ULL,
    0xbb6e14eea2000ULL,
    0xb0db961bf1000ULL,
    0x68a09780d5800ULL,
    0x54304ac0ebc00ULL,
    0xf628e52088a00ULL,
    0x734c3f004300ULL,
    0x5266bc88ce280ULL,
    0xf91fe2046b340ULL,
    0x11ce11cec8660ULL,
    0x5eb05eb641f0ULL,
    0x930813085e4e8ULL,
    0x84c4c4c4732b4ULL,
    0x8eeeeeeed428eULL,
    0xb1b1b1bb6345ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0xac00000000000ULL,
    0x7200000000000ULL,
    0xb100000000000ULL,
    0x380000000000ULL,
    0xd2c0000000000ULL,
    0xc160000000000ULL,
    0x9b90000000000ULL,
    0x4e48000000000ULL,
    0xb74000000000ULL,
    0x864e000000000ULL,
    0x3f0b000000000ULL,
    0x6806800000000ULL,
    0x447f400000000ULL,
    0x7648a00000000ULL,
    0xe774700000000ULL,
    0xd44e980000000ULL,
    0xbe0b9c0000000ULL,
    0xd3864a0000000ULL,
    0x3abf5d0000000ULL,
    0xc528d18000000ULL,
    0xcde413c000000ULL,
    0x99865ae000000ULL,
    0x67bfd55000000ULL,
    0x94a8c52800000ULL,
    0x9e24cde400000ULL,
    0xe366998600000ULL,
    0x82ef67bf00000ULL,
    0x698014a880000ULL,
    0xbfc0de2440000ULL,
    0x28e0c36620000ULL,
    0x6450b2ef30000ULL,
    0x46a8d18038000ULL,
    0x5f2413c0ec000ULL,
    0x78e65ae052000ULL,
    0xcc2fd55081000ULL,
    0x62e0c528bb800ULL,
    0x3950cde47ec00ULL,
    0x17289986b3600ULL,
    0xce467bf2a900ULL,
    0x20694a8cdc80ULL,
    0x297f9e2499b40ULL,
    0x9fc86366672e0ULL,
    0x18b4c2ef949b0ULL,
    0xdcae49801e4e8ULL,
    0xea5b8fc0a30b4ULL,
    0x2d2e10e0a206aULL,
    0xc99b8850597f7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0x4400000000000ULL,
    0x4a00000000000ULL,
    0x4700000000000ULL,
    0xdd80000000000ULL,
    0x4240000000000ULL,
    0xc320000000000ULL,
    0x7710000000000ULL,
    0x75b8000000000ULL,
    0x966c000000000ULL,
    0x715e000000000ULL,
    0xfc95000000000ULL,
    0xa6e6800000000ULL,
    0xd9f9c00000000ULL,
    0x2838600000000ULL,
    0x142cb00000000ULL,
    0x527e680000000ULL,
    0xfb85340000000ULL,
    0x5b5e420000000ULL,
    0xb95c30000000ULL,
    0x1366f78000000ULL,
    0xafb9b54000000ULL,
    0x2918f6a000000ULL,
    0x603cc15000000ULL,
    0xb046949800000ULL,
    0x68a9927c00000ULL,
    0x34a09b6600000ULL,
    0xc250ebb900000ULL,
    0x318631880000ULL,
    0x973c273cc0000ULL,
    0x5c66dc6e0000ULL,
    0x1ee92ae970000ULL,
    0x3580778008000ULL,
    0xb640754084000ULL,
    0xe12096a0aa000ULL,
    0x410715037000ULL,
    0x6a38fc9855800ULL,
    0xd72ca67c06400ULL,
    0x25fed96689200ULL,
    0x8ec528b930100ULL,
    0xcdfe949828380ULL,
    0x7ac5927c142c0ULL,
    0xeffe9b66527e0ULL,
    0x9c5ebb9fb850ULL,
    0xf07ee318db5e8ULL,
    0x4885e73ccb95cULL,
    0xa4de0dc6f3666ULL,
    0x3ad59ae9dfb9bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xd800000000000ULL,
    0xfc00000000000ULL,
    0xf600000000000ULL,
    0xd500000000000ULL,
    0xbf80000000000ULL,
    0x2c40000000000ULL,
    0xeee0000000000ULL,
    0x970000000000ULL,
    0x1908000000000ULL,
    0x2164000000000ULL,
    0xad6a000000000ULL,
    0xd313000000000ULL,
    0x2282800000000ULL,
    0x9707c00000000ULL,
    0x98e0a00000000ULL,
    0x1c70900000000ULL,
    0x8688f80000000ULL,
    0x5d24ac0000000ULL,
    0x9b8a2e0000000ULL,
    0x2663290000000ULL,
    0xcd8ac98000000ULL,
    0x6363394000000ULL,
    0x8a0af16000000ULL,
    0xe323b53000000ULL,
    0x4aea8fe800000ULL,
    0xc353441400000ULL,
    0x1a623a6200000ULL,
    0x1b774b7700000ULL,
    0xe668be6880000ULL,
    0xed54d154c0000ULL,
    0x3302e50220000ULL,
    0x5247d74750000ULL,
    0x1f80f80058000ULL,
    0xbc40ac003c000ULL,
    0x16e02e00d6000ULL,
    0xa570290085000ULL,
    0x3708498067800ULL,
    0x864f940d0400ULL,
    0xe4ea516018e00ULL,
    0x2a532530dc700ULL,
    0x73e277e8a6880ULL,
    0xb237e8140d240ULL,
    0x6f081462438a0ULL,
    0x34646277da630ULL,
    0x32ea77e8bb8a8ULL,
    0xaf53e8147663cULL,
    0x14621462158aaULL,
    0x627762779f639ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0xac00000000000ULL,
    0x6a00000000000ULL,
    0x8500000000000ULL,
    0xfb80000000000ULL,
    0xa8c0000000000ULL,
    0x8420000000000ULL,
    0xae30000000000ULL,
    0x4b08000000000ULL,
    0xe074000000000ULL,
    0x1086000000000ULL,
    0x388f000000000ULL,
    0xfc2e800000000ULL,
    0x320b400000000ULL,
    0x2980e00000000ULL,
    0x91c0100000000ULL,
    0x2da0380000000ULL,
    0x7ff0fc0000000ULL,
    0x6a8320000000ULL,
    0xcf84290000000ULL,
    0x4e2e918000000ULL,
    0x5b0b2dc000000ULL,
    0xd800ffa000000ULL,
    0xec0046f000000ULL,
    0xa00af2800000ULL,
    0xd5001e4400000ULL,
    0xa380038e00000ULL,
    0x4c074fb00000ULL,
    0xee2086a880000ULL,
    0x2b308f8440000ULL,
    0xb0882e2e60000ULL,
    0x48b40b0b50000ULL,
    0x94a68000d8000ULL,
    0x96bf4000ec000ULL,
    0xb726e0000a000ULL,
    0xd27f1000d5000ULL,
    0x3906b800a3800ULL,
    0xa94fbc0004c00ULL,
    0xd18ed200ee200ULL,
    0x4dfb39002b300ULL,
    0x2f282980b0880ULL,
    0x5e4491c048b40ULL,
    0x638e2da094a60ULL,
    0x24fb7ff096bf0ULL,
    0xdea886a8b7268ULL,
    0x23848f84d27f4ULL,
    0x442e2e2e3906eULL,
    0x8e0b0b0ba94f1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0x4400000000000ULL,
    0x5e00000000000ULL,
    0xad00000000000ULL,
    0xef80000000000ULL,
    0x6840000000000ULL,
    0x8460000000000ULL,
    0xfe50000000000ULL,
    0xfd28000000000ULL,
    0x7f4000000000ULL,
    0x2c62000000000ULL,
    0xda4f000000000ULL,
    0x5306800000000ULL,
    0x12dfc00000000ULL,
    0x6f80200000000ULL,
    0xa840300000000ULL,
    0x2460280000000ULL,
    0xae50140000000ULL,
    0x15283a0000000ULL,
    0x43f4110000000ULL,
    0x7262178000000ULL,
    0x774f2b4000000ULL,
    0xbc86bbe000000ULL,
    0x7a9fda1000000ULL,
    0xebe0011800000ULL,
    0x56100f9400000ULL,
    0xd948174a00000ULL,
    0xa9a415fd00000ULL,
    0x394a311880000ULL,
    0x99bb2793c0000ULL,
    0x21648341a0000ULL,
    0x6590eff7f0000ULL,
    0xd306800008000ULL,
    0xd2dfc0000c000ULL,
    0xcf8020000a000ULL,
    0xf840300005000ULL,
    0xcc6028000e800ULL,
    0xea50140004400ULL,
    0x4b283a0005e00ULL,
    0xeef411000ad00ULL,
    0x9de217800ef80ULL,
    0x1f0f2b4006840ULL,
    0x38e6bbe008460ULL,
    0x84cfda100fe50ULL,
    0x16c801180fd28ULL,
    0x51e40f94007f4ULL,
    0xf52a174a02c62ULL,
    0x73eb15fd0da4fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x400000000000ULL,
    0xda00000000000ULL,
    0x900000000000ULL,
    0x2280000000000ULL,
    0xe840000000000ULL,
    0xbc60000000000ULL,
    0xe30000000000ULL,
    0x7b58000000000ULL,
    0x378c000000000ULL,
    0x14c2000000000ULL,
    0x874d000000000ULL,
    0x99d4800000000ULL,
    0xbfb9400000000ULL,
    0x1880200000000ULL,
    0x9140300000000ULL,
    0xe6e0180000000ULL,
    0x52702c0000000ULL,
    0x538060000000ULL,
    0x34bc010000000ULL,
    0x971a368000000ULL,
    0x5181024000000ULL,
    0x13f688a000000ULL,
    0xde847a1000000ULL,
    0x466c8f1800000ULL,
    0x1745738c00000ULL,
    0x91fa26d600000ULL,
    0x73f111e300000ULL,
    0x6ece9b3080000ULL,
    0x5e384cd340000ULL,
    0x1376b6f520000ULL,
    0x4bc45cae50000ULL,
    0x7a8c800008000ULL,
    0x4c3540000c000ULL,
    0xb642200006000ULL,
    0xaf0d30000b000ULL,
    0x45b4980001800ULL,
    0x1896c0000400ULL,
    0x7bd826000da00ULL,
    0xa2cc310000900ULL,
    0x28222e8002280ULL,
    0xdc3d2e400e840ULL,
    0xbe6c8ea00bc60ULL,
    0x63457b1000e30ULL,
    0x33fa399807b58ULL,
    0xcef131cc0378cULL,
    0x8e4e8e76014c2ULL,
    0xbb785bf30874dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x8800000000000ULL,
    0x9c00000000000ULL,
    0x2e00000000000ULL,
    0x500000000000ULL,
    0xab80000000000ULL,
    0x1c40000000000ULL,
    0x6e20000000000ULL,
    0x2510000000000ULL,
    0xfba8000000000ULL,
    0x9404000000000ULL,
    0xf26e000000000ULL,
    0xb07000000000ULL,
    0xfeaa800000000ULL,
    0x3fd1c00000000ULL,
    0xee20200000000ULL,
    0x6510100000000ULL,
    0xdba8080000000ULL,
    0xc404140000000ULL,
    0x7a6e220000000ULL,
    0x9707270000000ULL,
    0xd0aa8b8000000ULL,
    0x3ad1c14000000ULL,
    0x45a00ae000000ULL,
    0x7950171000000ULL,
    0xb588138800000ULL,
    0xe1141d4400000ULL,
    0x81c61cea00000ULL,
    0x303020100000ULL,
    0x22c4b71b80000ULL,
    0x31d6c381c0000ULL,
    0xbb0ab54aa0000ULL,
    0x4681d8e470000ULL,
    0x5ba8080008000ULL,
    0x8404140004000ULL,
    0x5a6e220002000ULL,
    0xc707270005000ULL,
    0x58aa8b8008800ULL,
    0xa6d1c14009c00ULL,
    0x6ba00ae002e00ULL,
    0x7c50171000500ULL,
    0x1e0813880ab80ULL,
    0xfd541d4401c40ULL,
    0xefe61cea06e20ULL,
    0x2613020102510ULL,
    0xd96cb71b8fba8ULL,
    0xa5d2c381c9404ULL,
    0x4964b54aaf26eULL,
    0x4d86d8e470b07ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xc800000000000ULL,
    0x3c00000000000ULL,
    0x3e00000000000ULL,
    0x6700000000000ULL,
    0xf980000000000ULL,
    0xcc40000000000ULL,
    0x6660000000000ULL,
    0xb310000000000ULL,
    0xaba8000000000ULL,
    0x5d24000000000ULL,
    0xc4fe000000000ULL,
    0xb8cf000000000ULL,
    0x66bb800000000ULL,
    0x71a8c00000000ULL,
    0x1060200000000ULL,
    0x2810300000000ULL,
    0x4c28080000000ULL,
    0xa664140000000ULL,
    0x931e320000000ULL,
    0xfb9f0f0000000ULL,
    0x95738f8000000ULL,
    0xf89cd9c000000ULL,
    0x86b61e6000000ULL,
    0x1bb031000000ULL,
    0x880d919800000ULL,
    0xdc13f8c400000ULL,
    0x4e6db8ea00000ULL,
    0xff03e84900000ULL,
    0xdc596bf80000ULL,
    0xce27d3f3c0000ULL,
    0x3f3bbdcee0000ULL,
    0x2de8c47a30000ULL,
    0x9e00080008000ULL,
    0xf70014000c000ULL,
    0x1180320002000ULL,
    0xa0400f0005000ULL,
    0x90600f800c800ULL,
    0xe81019c003c00ULL,
    0x6c283e6003e00ULL,
    0xf664331006700ULL,
    0x5b1e19980f980ULL,
    0xc79f2cc40cc40ULL,
    0xab73aaea06660ULL,
    0x9f9cd7490b310ULL,
    0x7f36113f8aba8ULL,
    0xcdfb1e33c5d24ULL,
    0xee6d91aeec4feULL,
    0x6f03c86a3b8cfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x5800000000000ULL,
    0x4400000000000ULL,
    0x7e00000000000ULL,
    0x6900000000000ULL,
    0x5b80000000000ULL,
    0xdc40000000000ULL,
    0x5a20000000000ULL,
    0x8710000000000ULL,
    0xdad8000000000ULL,
    0x9bec000000000ULL,
    0xbc42000000000ULL,
    0xca0f000000000ULL,
    0x6f7c800000000ULL,
    0xc6d9c00000000ULL,
    0xa1a0200000000ULL,
    0xab50100000000ULL,
    0xf8f8080000000ULL,
    0xe8fc2c0000000ULL,
    0x409a160000000ULL,
    0x7ce3110000000ULL,
    0xf6be9f8000000ULL,
    0xb996da4000000ULL,
    0xcf7cb6e000000ULL,
    0x36d9e71000000ULL,
    0xd9a03e8800000ULL,
    0x5f501dc400000ULL,
    0xdef828b600000ULL,
    0xc5fc1bfb00000ULL,
    0x651a269080000ULL,
    0xc9a339c3c0000ULL,
    0xf71e92bf20000ULL,
    0xe2c6dce670000ULL,
    0x4f84880008000ULL,
    0x2a25ec0004000ULL,
    0xbf3a360002000ULL,
    0xeb301000b000ULL,
    0xdc6978005800ULL,
    0xc92af64004400ULL,
    0xabc6a0e007e00ULL,
    0xa42af61006900ULL,
    0xae46a10805b80ULL,
    0xa16ac7840dc40ULL,
    0xf7e69e5605a20ULL,
    0xbe3afceb08710ULL,
    0x91e98188dad8ULL,
    0xcbc6e407c9becULL,
    0x34049a092bc42ULL,
    0x4665d71d7ca0fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0x7400000000000ULL,
    0xc200000000000ULL,
    0xe700000000000ULL,
    0xb580000000000ULL,
    0xba40000000000ULL,
    0x9b20000000000ULL,
    0xa3d0000000000ULL,
    0x2f18000000000ULL,
    0x8184000000000ULL,
    0xd82a000000000ULL,
    0xcc19000000000ULL,
    0x5e07800000000ULL,
    0xe138c00000000ULL,
    0xd898200000000ULL,
    0x9cc4100000000ULL,
    0x568a280000000ULL,
    0x65892c0000000ULL,
    0xa23f920000000ULL,
    0xb76cdd0000000ULL,
    0xedaa108000000ULL,
    0x365929c000000ULL,
    0x6527856000000ULL,
    0xf2e8c29000000ULL,
    0xbf8014c800000ULL,
    0x694025f400000ULL,
    0x4ca0134600000ULL,
    0x4e9035a100000ULL,
    0x49b8096a80000ULL,
    0xec14009640000ULL,
    0xae1201c9e0000ULL,
    0x94d297a30000ULL,
    0x1cb5908008000ULL,
    0x16e5e9c004000ULL,
    0xc595a5600a000ULL,
    0x1235d2900b000ULL,
    0xff0dbcc804800ULL,
    0x99f1c9f407400ULL,
    0xf407a1460c200ULL,
    0x8238f8a10e700ULL,
    0x471831ea8b580ULL,
    0x58405564ba40ULL,
    0xf22a16a9e9b20ULL,
    0xef1936ea3a3d0ULL,
    0x618794c80af18ULL,
    0xc878e5f40c184ULL,
    0x343833460782aULL,
    0x625425a107c19ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x9800000000000ULL,
    0xb400000000000ULL,
    0x5200000000000ULL,
    0x700000000000ULL,
    0xbf80000000000ULL,
    0x5a40000000000ULL,
    0x3b20000000000ULL,
    0x91d0000000000ULL,
    0xd338000000000ULL,
    0xfdec000000000ULL,
    0x954a000000000ULL,
    0x58f1000000000ULL,
    0xb5df800000000ULL,
    0x91dc00000000ULL,
    0x86b8200000000ULL,
    0xa4ac100000000ULL,
    0x7bea280000000ULL,
    0xd0613c0000000ULL,
    0x2847a60000000ULL,
    0x8c61ed0000000ULL,
    0x166a348000000ULL,
    0xcd2111c000000ULL,
    0xce787e000000ULL,
    0xb7f1ea9000000ULL,
    0x667208c800000ULL,
    0x151d197400000ULL,
    0x1895884e00000ULL,
    0x15ecc2bb00000ULL,
    0xf9678cb280000ULL,
    0x1eb1fdac40000ULL,
    0x10d23f3fe0000ULL,
    0x298d0bf370000ULL,
    0xd70db48008000ULL,
    0x9790d1c004000ULL,
    0xd635a7e00a000ULL,
    0x2d7cfa900f000ULL,
    0x5cffa0c809800ULL,
    0xdfcde5740b400ULL,
    0x4a000e4e05200ULL,
    0xf3003fbb00700ULL,
    0x4d8010328bf80ULL,
    0xad40106c45a40ULL,
    0x1ca03edfe3b20ULL,
    0x7f901c63791d0ULL,
    0xba1820c805338ULL,
    0x6b3c25740bdecULL,
    0xf9f22e4e0354aULL,
    0xff5d2fbb0a8f1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xf800000000000ULL,
    0x4c00000000000ULL,
    0xa600000000000ULL,
    0x8900000000000ULL,
    0x6e80000000000ULL,
    0x1a40000000000ULL,
    0x1760000000000ULL,
    0x4bf0000000000ULL,
    0xa2f8000000000ULL,
    0x7c5c000000000ULL,
    0x7e36000000000ULL,
    0x551b000000000ULL,
    0x4080800000000ULL,
    0x272d400000000ULL,
    0x9398200000000ULL,
    0x7eac300000000ULL,
    0x524e380000000ULL,
    0x43071c0000000ULL,
    0xd1d6be0000000ULL,
    0x75c6530000000ULL,
    0xd7e0898000000ULL,
    0xacdd524000000ULL,
    0xd16003a000000ULL,
    0x72f02a9000000ULL,
    0xd47803d800000ULL,
    0x5a1c1dfc00000ULL,
    0x37563f3e00000ULL,
    0xdbeb2e5700000ULL,
    0x2af8adad80000ULL,
    0xc8317196c0000ULL,
    0x944e2e5820000ULL,
    0x7a072da750000ULL,
    0xa756b18008000ULL,
    0x53864e400c000ULL,
    0x9e80bda00e000ULL,
    0x222d799007000ULL,
    0xbb180a580f800ULL,
    0x9dec0fbc04c00ULL,
    0xd3ae1c9e0a600ULL,
    0x5eb734c708900ULL,
    0xc24e967586e80ULL,
    0xcb6a706ac1a40ULL,
    0x65aeaf6621760ULL,
    0x9fda50f054bf0ULL,
    0xf8b695ad822f8ULL,
    0x4b366d96cbc5cULL,
    0xa598905829e36ULL,
    0x7fc17ea75251bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0x3c00000000000ULL,
    0xde00000000000ULL,
    0xdf00000000000ULL,
    0x2980000000000ULL,
    0x3240000000000ULL,
    0xe920000000000ULL,
    0x6290000000000ULL,
    0x71d8000000000ULL,
    0x5e3c000000000ULL,
    0x9f2e000000000ULL,
    0x9e7000000000ULL,
    0x26b800000000ULL,
    0x5176c00000000ULL,
    0x5ef8200000000ULL,
    0xafac100000000ULL,
    0x8176080000000ULL,
    0xb69b0c0000000ULL,
    0x3be5ae0000000ULL,
    0xeb41cf0000000ULL,
    0x33eb978000000ULL,
    0x2f36e7c000000ULL,
    0xf1d8226000000ULL,
    0x1e3c109000000ULL,
    0xbf2e1c4800000ULL,
    0x39e71ba400000ULL,
    0xba6b85f600000ULL,
    0x6d76ef4f00000ULL,
    0x80f83a2b80000ULL,
    0x70ac2929c0000ULL,
    0xa8f638b2e0000ULL,
    0x84db0c69b0000ULL,
    0xd2c59f8008000ULL,
    0x89d1ebc004000ULL,
    0x42338c6002000ULL,
    0x710adf9003000ULL,
    0x6ef60bc80b800ULL,
    0x17db3c6403c00ULL,
    0xbd4587960de00ULL,
    0x6891efdf0df00ULL,
    0xe493ae6382980ULL,
    0xc2dafe8dc3240ULL,
    0x18e3344ee920ULL,
    0xc6373c26b6290ULL,
    0x9313ba2b8f1d8ULL,
    0x6f9ae929c1e3cULL,
    0xe12e18b2ebf2eULL,
    0xa6e71c69b39e7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0x3c00000000000ULL,
    0x6e00000000000ULL,
    0x1900000000000ULL,
    0x5080000000000ULL,
    0xca40000000000ULL,
    0x7b20000000000ULL,
    0xafd0000000000ULL,
    0x97a8000000000ULL,
    0x4b9c000000000ULL,
    0x55ae000000000ULL,
    0x64ef000000000ULL,
    0xf028800000000ULL,
    0x6852400000000ULL,
    0x6408200000000ULL,
    0x820c100000000ULL,
    0x8f26280000000ULL,
    0x75a3340000000ULL,
    0xf4aebe0000000ULL,
    0xa8614f0000000ULL,
    0x842ebb8000000ULL,
    0xf221564000000ULL,
    0xa70e9c2000000ULL,
    0xb1f1569000000ULL,
    0xa6a6a8c800000ULL,
    0xdf6d40f400000ULL,
    0xcd88886a00000ULL,
    0x68c27fa700000ULL,
    0x16002ccb80000ULL,
    0x650006ebc0000ULL,
    0x9e803b6220000ULL,
    0x3403e3090000ULL,
    0xd3a0138008000ULL,
    0x5990224004000ULL,
    0x828802200a000ULL,
    0xfd4c09900d000ULL,
    0x92863b480f800ULL,
    0xe53322b403c00ULL,
    0xdea6aa4a06e00ULL,
    0xa36d663701900ULL,
    0x388bf8385080ULL,
    0xa1c2505fcca40ULL,
    0xbe800f2827b20ULL,
    0x934007079afd0ULL,
    0x8ba03f83817a8ULL,
    0xb590105fc0b9cULL,
    0x14882f282f5aeULL,
    0xd84c17079b4efULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xc800000000000ULL,
    0xbc00000000000ULL,
    0x4e00000000000ULL,
    0x5700000000000ULL,
    0x8080000000000ULL,
    0xa40000000000ULL,
    0xfd20000000000ULL,
    0x8db0000000000ULL,
    0xffa8000000000ULL,
    0xa684000000000ULL,
    0x110e000000000ULL,
    0x4bdf000000000ULL,
    0x74d7800000000ULL,
    0xb872400000000ULL,
    0x8408200000000ULL,
    0x8a74100000000ULL,
    0xbd06180000000ULL,
    0xedab340000000ULL,
    0x2fd1b20000000ULL,
    0x6ed96f0000000ULL,
    0xad59b38000000ULL,
    0x5ed45c000000ULL,
    0x23ff982000000ULL,
    0x38b6669000000ULL,
    0x8e26354800000ULL,
    0x771b286c00000ULL,
    0x30f9866a00000ULL,
    0x121d676100000ULL,
    0x8977a5e380000ULL,
    0x7f827aa7c0000ULL,
    0xe68029dde0000ULL,
    0x71403e2090000ULL,
    0x9ba02b8008000ULL,
    0xbcf031c004000ULL,
    0x4080a2006000ULL,
    0xca7419900d000ULL,
    0xdd061ec80c800ULL,
    0x3dab19ac0bc00ULL,
    0xe7d18c4a04e00ULL,
    0xd2d97ef105700ULL,
    0xe359bb2b88080ULL,
    0x52ed630bc0a40ULL,
    0xa37fa597efd20ULL,
    0x32f640d198db0ULL,
    0x730610ab87fa8ULL,
    0xfaab12cbce684ULL,
    0xcf518fb7e710eULL,
    0xb499494199bdfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xa800000000000ULL,
    0xd400000000000ULL,
    0xfa00000000000ULL,
    0xf900000000000ULL,
    0x9280000000000ULL,
    0x1940000000000ULL,
    0x42a0000000000ULL,
    0x2150000000000ULL,
    0x8ef8000000000ULL,
    0xa704000000000ULL,
    0x5992000000000ULL,
    0x36f9000000000ULL,
    0x2b2e800000000ULL,
    0xffd0400000000ULL,
    0x5192200000000ULL,
    0x12f9100000000ULL,
    0x592e880000000ULL,
    0x62d06c0000000ULL,
    0x91120a0000000ULL,
    0x26b9250000000ULL,
    0x730eb68000000ULL,
    0xa3c0524000000ULL,
    0xcfca2ea000000ULL,
    0xb9ad235000000ULL,
    0xe6c4a62800000ULL,
    0x136d5a1400000ULL,
    0x338e8d1e00000ULL,
    0xd7804a9100000ULL,
    0xc5ea104c80000ULL,
    0xc8bd07aa40000ULL,
    0x101cafd5a0000ULL,
    0x5879496510000ULL,
    0x5c44a62808000ULL,
    0x9e2d5a1404000ULL,
    0xab2e8d1e02000ULL,
    0xbfd04a910b000ULL,
    0x7192104c8a800ULL,
    0xa2f907aa4d400ULL,
    0xf12eafd5afa00ULL,
    0xb6d049651f900ULL,
    0x6b12262801280ULL,
    0xdfb91a1405940ULL,
    0xe18ead1e062a0ULL,
    0xba805a9109150ULL,
    0x8d6a184c826f8ULL,
    0x98fd2baa47304ULL,
    0x683c85d5aa392ULL,
    0xb4697c651cff9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x5800000000000ULL,
    0x1c00000000000ULL,
    0x7200000000000ULL,
    0x4f00000000000ULL,
    0xa180000000000ULL,
    0x7740000000000ULL,
    0x4da0000000000ULL,
    0xbd30000000000ULL,
    0xaef8000000000ULL,
    0x369c000000000ULL,
    0x8ab6000000000ULL,
    0xa885000000000ULL,
    0xfe1800000000ULL,
    0xea0dc00000000ULL,
    0xf336200000000ULL,
    0x83c5100000000ULL,
    0xd041b80000000ULL,
    0xa83dec0000000ULL,
    0xa44e360000000ULL,
    0xde19170000000ULL,
    0x6557a48000000ULL,
    0xf288ffc000000ULL,
    0xa4d79e6000000ULL,
    0x75c8cad000000ULL,
    0x517797e800000ULL,
    0x64f8c08c00000ULL,
    0xd58f8dde00000ULL,
    0x164eb7700000ULL,
    0x8cb9a34580000ULL,
    0x91a1edad40000ULL,
    0x6f7812a660000ULL,
    0xb1dc023470000ULL,
    0x7f1617e808000ULL,
    0xb9b5008c04000ULL,
    0x8b19adde0e000ULL,
    0x8f91fb770b000ULL,
    0xaa001b4585800ULL,
    0x130001ad41c00ULL,
    0x338024a667200ULL,
    0x8840153474f00ULL,
    0xb420336802180ULL,
    0xd6703f4c03740ULL,
    0x915813be0ada0ULL,
    0xc4ac21a700d30ULL,
    0x85ce34ad8f6f8ULL,
    0xe9592d2142a9cULL,
    0xc8f79f786f8b6ULL,
    0xffb8e9437e785ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x8800000000000ULL,
    0x3400000000000ULL,
    0xa200000000000ULL,
    0x300000000000ULL,
    0x4180000000000ULL,
    0xf740000000000ULL,
    0x3a0000000000ULL,
    0x410000000000ULL,
    0x9a08000000000ULL,
    0x4f14000000000ULL,
    0xfb2000000000ULL,
    0xea55000000000ULL,
    0xd73b800000000ULL,
    0x13a1c00000000ULL,
    0x2c12200000000ULL,
    0xfe45100000000ULL,
    0x6533a80000000ULL,
    0x38b5d40000000ULL,
    0x9a0020000000ULL,
    0x23101d0000000ULL,
    0x5188008000000ULL,
    0xdf5414c000000ULL,
    0x6792326000000ULL,
    0x2e0530d000000ULL,
    0xad13a86800000ULL,
    0xace5c1c400000ULL,
    0xfb8816e200000ULL,
    0xa8543e1500000ULL,
    0x24122b0480000ULL,
    0x8a452f9140000ULL,
    0x6733b14ce0000ULL,
    0x6bb5da2d70000ULL,
    0xc020006808000ULL,
    0xe05015c404000ULL,
    0xf02814e20a000ULL,
    0xd844231505000ULL,
    0xbc1a2b8488800ULL,
    0x96513b5143400ULL,
    0xa101832cea200ULL,
    0x42a0eafd70300ULL,
    0xb6bba8000c180ULL,
    0xf4e1d4000b740ULL,
    0x7b202000a3a0ULL,
    0x9e551d0005410ULL,
    0xd53b808001208ULL,
    0x40a1d4c007b14ULL,
    0xe59212600adb2ULL,
    0x3d0520d00e955ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0xb800000000000ULL,
    0x1c00000000000ULL,
    0x8200000000000ULL,
    0xfb00000000000ULL,
    0xed80000000000ULL,
    0x8740000000000ULL,
    0xffa0000000000ULL,
    0x2430000000000ULL,
    0xde48000000000ULL,
    0x992c000000000ULL,
    0xc6e6000000000ULL,
    0xd2dd000000000ULL,
    0x6493800000000ULL,
    0x59a7c00000000ULL,
    0x146200000000ULL,
    0xaaed100000000ULL,
    0xd8dbb80000000ULL,
    0xeb8bf40000000ULL,
    0x92200e0000000ULL,
    0xe370170000000ULL,
    0xc1e8188000000ULL,
    0x6d1c0ac000000ULL,
    0xa0ae156000000ULL,
    0x57f126d000000ULL,
    0x20759f6800000ULL,
    0x707af7cc00000ULL,
    0x8855acf200000ULL,
    0x740ad79b00000ULL,
    0x263d965180000ULL,
    0x6556d9bb40000ULL,
    0x94b398b6e0000ULL,
    0x91d7d322f0000ULL,
    0x952e276808000ULL,
    0x5cb103cc04000ULL,
    0x5d5a2f20e000ULL,
    0x634ac09b0d000ULL,
    0x819d8ed18b800ULL,
    0x8d66d37b41c00ULL,
    0x70fb8dd6e8200ULL,
    0xeffbf5f2ffb00ULL,
    0x3c48380006d80ULL,
    0xf22c34000c740ULL,
    0x73662e0001fa0ULL,
    0x999d07000f430ULL,
    0xa133a08006648ULL,
    0x9a97fec00852cULL,
    0xb08e1b60044e6ULL,
    0x4f8131d0029ddULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x4800000000000ULL,
    0xac00000000000ULL,
    0x600000000000ULL,
    0x9500000000000ULL,
    0x580000000000ULL,
    0xc940000000000ULL,
    0x3be0000000000ULL,
    0x810000000000ULL,
    0xcc68000000000ULL,
    0xb674000000000ULL,
    0xcd5e000000000ULL,
    0xe1a7000000000ULL,
    0x635c800000000ULL,
    0xa8e1c00000000ULL,
    0x98be200000000ULL,
    0xb7300000000ULL,
    0x44b4a80000000ULL,
    0xfed5c40000000ULL,
    0x2580320000000ULL,
    0x19401b0000000ULL,
    0xd3e0298000000ULL,
    0xb410214000000ULL,
    0x8268136000000ULL,
    0x8f74195000000ULL,
    0xcede0f7800000ULL,
    0xbde7274400000ULL,
    0x5d3cb27a00000ULL,
    0x69b1dfcd00000ULL,
    0x6f361daf80000ULL,
    0xbed30a6dc0000ULL,
    0x458283cd20000ULL,
    0xa906c3a570000ULL,
    0x8b82a77808000ULL,
    0x5006e3440c000ULL,
    0x2802807a0a000ULL,
    0x1c46c4cd01000ULL,
    0x5e62b42f84800ULL,
    0x7116eb2dcac00ULL,
    0xafeab0ad20600ULL,
    0x5a72eaf579500ULL,
    0xab5c800008580ULL,
    0xc4e1c00000940ULL,
    0x3ebe200009be0ULL,
    0x85b7300001810ULL,
    0x934a80008468ULL,
    0x9b95c40001a74ULL,
    0x186032000cb5eULL,
    0x84501b00074a7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0x3400000000000ULL,
    0x1a00000000000ULL,
    0xff00000000000ULL,
    0xf380000000000ULL,
    0x9340000000000ULL,
    0x2da0000000000ULL,
    0x3e70000000000ULL,
    0x3d48000000000ULL,
    0x88cc000000000ULL,
    0x52b2000000000ULL,
    0x8d91000000000ULL,
    0xce35800000000ULL,
    0x750cc00000000ULL,
    0x9492200000000ULL,
    0x84a1100000000ULL,
    0x5cdd980000000ULL,
    0xd8b0f40000000ULL,
    0xeae81e0000000ULL,
    0xd9bc1d0000000ULL,
    0x47a1e8000000ULL,
    0x721d0bc000000ULL,
    0x532782e000000ULL,
    0xdede9d000000ULL,
    0x8e6fade800000ULL,
    0x1521e05c00000ULL,
    0x44dd8bb200000ULL,
    0x7cb0e2e300000ULL,
    0x68e819c480000ULL,
    0xc2bc05f840000ULL,
    0x15fa1c5f60000ULL,
    0x2a5d39b030000ULL,
    0x9707b5e808000ULL,
    0x5fddd45c04000ULL,
    0x6d07b5b206000ULL,
    0x30ddefe30d000ULL,
    0x6879f448f800ULL,
    0x479dfa3843400ULL,
    0xc92780bf61a00ULL,
    0xb2edcd603ff00ULL,
    0x1def868007380ULL,
    0x5661ffc00d340ULL,
    0x917d9ce004da0ULL,
    0x76c0f4d00ee70ULL,
    0x4fa033680c548ULL,
    0xb5702b9c0bcccULL,
    0xb4c82952048b2ULL,
    0x348c1b3307291ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0xfc00000000000ULL,
    0xb200000000000ULL,
    0x5b00000000000ULL,
    0x3f80000000000ULL,
    0x7f40000000000ULL,
    0x89e0000000000ULL,
    0x2270000000000ULL,
    0xb368000000000ULL,
    0xa3a4000000000ULL,
    0xdd36000000000ULL,
    0xfaad000000000ULL,
    0xe1a3800000000ULL,
    0x7e6ec00000000ULL,
    0x7156200000000ULL,
    0xc09d300000000ULL,
    0x36ab980000000ULL,
    0xcbfac40000000ULL,
    0x81682a0000000ULL,
    0x38a40f0000000ULL,
    0x82b6348000000ULL,
    0x95ed12c000000ULL,
    0x404385e000000ULL,
    0xa01ee0d000000ULL,
    0x703e2ef800000ULL,
    0x38392e5c00000ULL,
    0xd41dbb3a00000ULL,
    0x4e17f33900000ULL,
    0xe92bbf3580000ULL,
    0x64baf93740000ULL,
    0x40880032e0000ULL,
    0xf6d415b2b0000ULL,
    0xabde36f808000ULL,
    0x91492a5c0c000ULL,
    0x10f5b13a06000ULL,
    0x7ef3cc3901000ULL,
    0x27fd93b582800ULL,
    0x1b67eff74fc00ULL,
    0x9fc38fd2eb200ULL,
    0xf5eca62b5b00ULL,
    0xb1de34800bf80ULL,
    0xf64912c00bf40ULL,
    0xfd7585e00e9e0ULL,
    0x4ab3e0d003270ULL,
    0xb99daef809b68ULL,
    0xba57ee5c05fa4ULL,
    0x174b9b3a06f36ULL,
    0xd58ac3390a1adULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x800000000000ULL,
    0x4c00000000000ULL,
    0xf600000000000ULL,
    0x7f00000000000ULL,
    0x7680000000000ULL,
    0x1940000000000ULL,
    0x11a0000000000ULL,
    0x7bf0000000000ULL,
    0x8af8000000000ULL,
    0xa754000000000ULL,
    0x42ae000000000ULL,
    0xcb17000000000ULL,
    0xe4a5800000000ULL,
    0x8c12400000000ULL,
    0xd656200000000ULL,
    0x2f43100000000ULL,
    0x4e8b980000000ULL,
    0x5d454c0000000ULL,
    0xabd3a20000000ULL,
    0xf2e1430000000ULL,
    0x8305858000000ULL,
    0xc8e243c000000ULL,
    0x4a2e27a000000ULL,
    0xa157095000000ULL,
    0x1585a3e800000ULL,
    0xa1a25e3c00000ULL,
    0x338e209e00000ULL,
    0xa6a7334500000ULL,
    0x617dace380000ULL,
    0x35f679a9c0000ULL,
    0xf1a0205f60000ULL,
    0xbf0117d90000ULL,
    0xe2f8266808000ULL,
    0xdb541dfc04000ULL,
    0xbcae073e06000ULL,
    0xf8173a1503000ULL,
    0x64258f0b80800ULL,
    0xea526795c4c00ULL,
    0xb17620c16f600ULL,
    0x4df3323897f00ULL,
    0xd5d3928b8f680ULL,
    0x81e16855c5940ULL,
    0x6385a561671a0ULL,
    0x9ea2786894bf0ULL,
    0x250e34e3802f8ULL,
    0x8fe735a9cab54ULL,
    0x78dd825f6d4aeULL,
    0x206527d98417ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0xb800000000000ULL,
    0x7c00000000000ULL,
    0x4a00000000000ULL,
    0xf300000000000ULL,
    0x9080000000000ULL,
    0x8140000000000ULL,
    0x5fa0000000000ULL,
    0xfb90000000000ULL,
    0x5dd8000000000ULL,
    0x8cec000000000ULL,
    0x5b36000000000ULL,
    0xc4b1000000000ULL,
    0xdf33800000000ULL,
    0x5297400000000ULL,
    0x166e200000000ULL,
    0x891d100000000ULL,
    0x7ba5a80000000ULL,
    0x1db65c0000000ULL,
    0x2c858e0000000ULL,
    0x2b664f0000000ULL,
    0x7cfd9a8000000ULL,
    0xa31a70c000000ULL,
    0x1893822000000ULL,
    0xe507735000000ULL,
    0x19b6236800000ULL,
    0xfaf1112400000ULL,
    0x4213a7d600000ULL,
    0xd7477cab00000ULL,
    0x7696398580000ULL,
    0xf0211c5840000ULL,
    0xf86ba9f2e0000ULL,
    0xdc3b49ead0000ULL,
    0x3a7839e808000ULL,
    0x4b7c21e404000ULL,
    0xecee05f60a000ULL,
    0xcb5d1ffb07000ULL,
    0xac85b2ed8b800ULL,
    0x6b66517c47c00ULL,
    0xdcfd8024e4a00ULL,
    0xd31a7a41df300ULL,
    0xa0939aed81080ULL,
    0x99074d7c4c140ULL,
    0x53b62e24effa0ULL,
    0x9f12541d8b90ULL,
    0xd293a86d865d8ULL,
    0x560761bc4b0ecULL,
    0x29362204eb136ULL,
    0xbb11911d47b1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x8800000000000ULL,
    0xd400000000000ULL,
    0xea00000000000ULL,
    0xb700000000000ULL,
    0xf580000000000ULL,
    0xa540000000000ULL,
    0xfea0000000000ULL,
    0x7e90000000000ULL,
    0x3eb8000000000ULL,
    0x9ef4000000000ULL,
    0x2e82000000000ULL,
    0xa6d9000000000ULL,
    0x729d800000000ULL,
    0x98c9c00000000ULL,
    0x2fba200000000ULL,
    0xda6d100000000ULL,
    0x7f3fa80000000ULL,
    0x81c0ec0000000ULL,
    0xff3f820000000ULL,
    0xc1c0e50000000ULL,
    0x5f3fb28000000ULL,
    0x71c0d1c000000ULL,
    0xd73f976000000ULL,
    0xa5c0e05000000ULL,
    0x3d3faf2800000ULL,
    0x12c0fb6400000ULL,
    0xc8bfa24e00000ULL,
    0xb780ea2d00000ULL,
    0x361f99e880000ULL,
    0xc910fb8240000ULL,
    0x8a7814160000ULL,
    0x57e4d3bb70000ULL,
    0x26259da808000ULL,
    0xf13deaa404000ULL,
    0x54b8152e0a000ULL,
    0x69f41a7d0b000ULL,
    0x7b021ec088800ULL,
    0xb3992ce64d400ULL,
    0x43d810f6ea00ULL,
    0x3259cc967b700ULL,
    0xfb021ec087580ULL,
    0xf3992ce64e540ULL,
    0xa43d810f65ea0ULL,
    0x8259cc967ce90ULL,
    0x73021ec0836b8ULL,
    0x27992ce640af4ULL,
    0x4e3d810f66482ULL,
    0x3559cc967a1d9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x800000000000ULL,
    0x3400000000000ULL,
    0x1a00000000000ULL,
    0xd100000000000ULL,
    0xac80000000000ULL,
    0x5740000000000ULL,
    0x43a0000000000ULL,
    0x18d0000000000ULL,
    0xd48000000000ULL,
    0xb2b4000000000ULL,
    0xe462000000000ULL,
    0x5201000000000ULL,
    0xc566800000000ULL,
    0xe6e9400000000ULL,
    0x8e0a200000000ULL,
    0xdb25100000000ULL,
    0x55ec880000000ULL,
    0x9f8c540000000ULL,
    0x6c6a20000000ULL,
    0xbe395d0000000ULL,
    0xa3422e8000000ULL,
    0x3991304000000ULL,
    0xb98ea12000000ULL,
    0xf98d4cd000000ULL,
    0xd9a0346800000ULL,
    0x89d02f7400000ULL,
    0x81c826f200000ULL,
    0xb5f4193d00000ULL,
    0xafc200d080000ULL,
    0x7ed10a6440000ULL,
    0xd22ea463a0000ULL,
    0x855d676350000ULL,
    0xc6e812e808000ULL,
    0xde640b3404000ULL,
    0xd32a05d202000ULL,
    0x61b518ed05000ULL,
    0x8584923880800ULL,
    0xd7a8415043400ULL,
    0x12cc81b1a1a00ULL,
    0xf41c6f8e5d100ULL,
    0x7a2e88d082c80ULL,
    0xa15d5e6441740ULL,
    0xf4e80663a63a0ULL,
    0x6b643a63548d0ULL,
    0x6daa3c6808548ULL,
    0xd3f53b740c6b4ULL,
    0x70a4a4f20de62ULL,
    0x4938543d0d301ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x800000000000ULL,
    0xe400000000000ULL,
    0xe600000000000ULL,
    0x700000000000ULL,
    0x1080000000000ULL,
    0x7d40000000000ULL,
    0x5da0000000000ULL,
    0x8f0000000000ULL,
    0x2118000000000ULL,
    0x3794000000000ULL,
    0xfdfa000000000ULL,
    0xd8ef000000000ULL,
    0xb925800000000ULL,
    0x2be1400000000ULL,
    0xf7c2200000000ULL,
    0xddcb100000000ULL,
    0x48e7980000000ULL,
    0x412a7c0000000ULL,
    0xc7a5a20000000ULL,
    0xf5a1690000000ULL,
    0x3ce2018000000ULL,
    0x5f7b2dc000000ULL,
    0x2cdf9e2000000ULL,
    0xe70e5a5000000ULL,
    0xa0e78ce800000ULL,
    0x152a6afc00000ULL,
    0x49a58de600000ULL,
    0xe6a17f7500000ULL,
    0xc262163680000ULL,
    0xc13b3e57c0000ULL,
    0x87ff92e760000ULL,
    0x95be41e150000ULL,
    0xccdf956808000ULL,
    0x570e7b3c04000ULL,
    0xc8e791c606000ULL,
    0x12a5c250f000ULL,
    0xa7a5835e80800ULL,
    0x5a1456bce400ULL,
    0x34e203216e600ULL,
    0xbb7b1dc450700ULL,
    0xcadf963689080ULL,
    0xe00e7e57c3d40ULL,
    0xb067b2e763da0ULL,
    0x686a51e15f8f0ULL,
    0x14058d680a918ULL,
    0xee51473c09394ULL,
    0xe37a13c607bfaULL,
    0xf6af252502fefULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0xac00000000000ULL,
    0xa200000000000ULL,
    0xcf00000000000ULL,
    0x5780000000000ULL,
    0x2fc0000000000ULL,
    0x63a0000000000ULL,
    0x51b0000000000ULL,
    0x16e8000000000ULL,
    0xd574000000000ULL,
    0xf4e2000000000ULL,
    0xfa13000000000ULL,
    0x3344800000000ULL,
    0x5dc7400000000ULL,
    0xc4c4a00000000ULL,
    0x207700000000ULL,
    0xbf64a80000000ULL,
    0x4fb75c0000000ULL,
    0x338ca60000000ULL,
    0xf9c3770000000ULL,
    0x32ee8e8000000ULL,
    0xe31044c000000ULL,
    0x358a1b6000000ULL,
    0xc0a70f3000000ULL,
    0x8406a38800000ULL,
    0x46646b5c00000ULL,
    0xd9680e3200000ULL,
    0x26b4020100000ULL,
    0x2d42150a80000ULL,
    0x78a30b85c0000ULL,
    0xe82cb75ba0000ULL,
    0xc473683410000ULL,
    0xa606950a88000ULL,
    0x49644b85cc000ULL,
    0xaee8175ba2000ULL,
    0xb97418341b000ULL,
    0x76e23d0a8b800ULL,
    0x85131785c6c00ULL,
    0x5cc4b15ba8200ULL,
    0xde076f3417f00ULL,
    0x564b38a8ef80ULL,
    0x9cb75345c43c0ULL,
    0xfe0caa3bae1a0ULL,
    0xb503600412eb0ULL,
    0xa4ce90028f968ULL,
    0x52607819c96b4ULL,
    0x17420409a1542ULL,
    0x6ba312051d4a3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0x6c00000000000ULL,
    0x7e00000000000ULL,
    0xff00000000000ULL,
    0x1880000000000ULL,
    0xc0c0000000000ULL,
    0x7ca0000000000ULL,
    0x5ab0000000000ULL,
    0xd9b8000000000ULL,
    0xc704000000000ULL,
    0x94f2000000000ULL,
    0x8eed000000000ULL,
    0xebe2800000000ULL,
    0x5676c00000000ULL,
    0xb62a00000000ULL,
    0x3ab6f00000000ULL,
    0x29c2a80000000ULL,
    0x8f06f40000000ULL,
    0x90fab60000000ULL,
    0xe4c2ef0000000ULL,
    0x6a8a98000000ULL,
    0xcf9fd0c000000ULL,
    0x2c722fa000000ULL,
    0x9e2d20f000000ULL,
    0xcf42908800000ULL,
    0x70c6c65c00000ULL,
    0xd4da8ee600000ULL,
    0x6eb2c39d00000ULL,
    0xdbb0bdda80000ULL,
    0x3e2bff2640000ULL,
    0x1f3806a220000ULL,
    0x28c40e7bf0000ULL,
    0xa8d23dda88000ULL,
    0x689d3f264c000ULL,
    0x48faa6a222000ULL,
    0x58c2fe7bf1000ULL,
    0x20a895da8f800ULL,
    0x4c9fcb264ac00ULL,
    0x32f210a225e00ULL,
    0xcded117bfef00ULL,
    0xd562bc5a8e080ULL,
    0x15b6dbe646cc0ULL,
    0x69429f02222a0ULL,
    0x33c6c18bfb5b0ULL,
    0xea5a84d283938ULL,
    0x2d72e9ba4abc4ULL,
    0xb990a7e42b652ULL,
    0x375bed16f3b5dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xd800000000000ULL,
    0xf400000000000ULL,
    0xd200000000000ULL,
    0xab00000000000ULL,
    0x9880000000000ULL,
    0x90c0000000000ULL,
    0xeca0000000000ULL,
    0x82f0000000000ULL,
    0xe7e8000000000ULL,
    0x2a04000000000ULL,
    0xaf3e000000000ULL,
    0x32b7000000000ULL,
    0xfff2800000000ULL,
    0x7e46c00000000ULL,
    0x4d72a00000000ULL,
    0x4186f00000000ULL,
    0x9352880000000ULL,
    0x3cb6fc0000000ULL,
    0xa9abe0000000ULL,
    0x5b82c10000000ULL,
    0xe46c8a8000000ULL,
    0xfa01ebc000000ULL,
    0x27682ca000000ULL,
    0x8ec40ff000000ULL,
    0x319e378800000ULL,
    0x2b471f4c00000ULL,
    0x589aa67200000ULL,
    0x3082d9cd00000ULL,
    0xdcec9bbd80000ULL,
    0x5ac1d860c0000ULL,
    0x13c838c120000ULL,
    0xf834213170000ULL,
    0x4761bbd88000ULL,
    0xaa431860cc000ULL,
    0x6f2498c12a000ULL,
    0x92f5d13173000ULL,
    0xcfbe13bd85800ULL,
    0xa6772460c3400ULL,
    0xb95286c127200ULL,
    0x93b6e03179b00ULL,
    0x381a913d8c080ULL,
    0xa442f3a0ca4c0ULL,
    0x9a4cb46129ea0ULL,
    0xb731dec1719f0ULL,
    0x66a0243582768ULL,
    0x1df03b2cc8ec4ULL,
    0xd6820b32319eULL,
    0x21c439fc72b47ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x800000000000ULL,
    0xc00000000000ULL,
    0x7200000000000ULL,
    0xf900000000000ULL,
    0x4a80000000000ULL,
    0x86c0000000000ULL,
    0x14e0000000000ULL,
    0x7db0000000000ULL,
    0xf28000000000ULL,
    0x8dec000000000ULL,
    0xe70a000000000ULL,
    0x1183000000000ULL,
    0xad57800000000ULL,
    0xecdec00000000ULL,
    0x99b7a00000000ULL,
    0xe16ed00000000ULL,
    0x3e9f880000000ULL,
    0x5082dc0000000ULL,
    0xa3958a0000000ULL,
    0xb401df0000000ULL,
    0x3642168000000ULL,
    0x271f214000000ULL,
    0xf195a42000000ULL,
    0x3d01d0f000000ULL,
    0xd4c2291800000ULL,
    0x9ddf139c00000ULL,
    0x9f75a0d200000ULL,
    0xb5b1efe700000ULL,
    0xe36a0f9080000ULL,
    0x6ff30ac7c0000ULL,
    0x261fa0e560000ULL,
    0x5f42f10070000ULL,
    0x55f5879088000ULL,
    0x7371d6c7c4000ULL,
    0x578a2ae56a000ULL,
    0x22432e0073000ULL,
    0x2137911088800ULL,
    0xdeaef787c4c00ULL,
    0xc0ff8ec56d200ULL,
    0x9bf2fef07c900ULL,
    0xb05db8088c280ULL,
    0x485de41bccac0ULL,
    0xac602e176c6e0ULL,
    0x427011177b4b0ULL,
    0xf14837980cda8ULL,
    0x469c2edc0472cULL,
    0xf4c22ef2021eaULL,
    0xeddf30170a533ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0xe400000000000ULL,
    0x1e00000000000ULL,
    0xd00000000000ULL,
    0x4f80000000000ULL,
    0x3c0000000000ULL,
    0xb9e0000000000ULL,
    0xcad0000000000ULL,
    0xd878000000000ULL,
    0xbc2c000000000ULL,
    0xe27e000000000ULL,
    0x8f41000000000ULL,
    0x90ef800000000ULL,
    0xbb1c400000000ULL,
    0xe68fa00000000ULL,
    0x320c500000000ULL,
    0xe717b80000000ULL,
    0x14f0440000000ULL,
    0xf511b20000000ULL,
    0xc39d7d0000000ULL,
    0x9980358000000ULL,
    0xfac03e4000000ULL,
    0xa060066000000ULL,
    0x70102eb000000ULL,
    0x1818301800000ULL,
    0x9c3c080400000ULL,
    0xd2660c0600000ULL,
    0xf77d1e0f00000ULL,
    0x5c89b31980000ULL,
    0x41617e9f40000ULL,
    0xf58624c260000ULL,
    0x70ad00a850000ULL,
    0xab718b1988000ULL,
    0xae8d7a9f44000ULL,
    0x861836c266000ULL,
    0xd13c2da855000ULL,
    0xfde606998a800ULL,
    0xa4bd00df4a400ULL,
    0xcd6982a267e00ULL,
    0x6fb17e1855d00ULL,
    0x33fe03018e780ULL,
    0xc181369b4a7c0ULL,
    0x68f88c46c7e0ULL,
    0x220c4ea7597d0ULL,
    0xaf17800003ff8ULL,
    0xa0f0400001becULL,
    0xc311a0000259eULL,
    0x2a9d500001891ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0x2c00000000000ULL,
    0xd200000000000ULL,
    0x8d00000000000ULL,
    0x7080000000000ULL,
    0x14c0000000000ULL,
    0xb2e0000000000ULL,
    0x51f0000000000ULL,
    0xf628000000000ULL,
    0xb74000000000ULL,
    0x23c2000000000ULL,
    0x8b7b000000000ULL,
    0x6385800000000ULL,
    0xab51c00000000ULL,
    0xd3e5a00000000ULL,
    0x9361d00000000ULL,
    0xffada80000000ULL,
    0x4125fc0000000ULL,
    0x72a7a60000000ULL,
    0x31daf70000000ULL,
    0x6648128000000ULL,
    0x8344144000000ULL,
    0x378a2ea000000ULL,
    0x753f017000000ULL,
    0x3c8f8a1800000ULL,
    0x56aef90c00000ULL,
    0xb78a199200000ULL,
    0x353f20d100000ULL,
    0x1c8f8de280000ULL,
    0xe6aedd4fc0000ULL,
    0x8f8a2f23e0000ULL,
    0x193f05abb0000ULL,
    0xce8fa5e288000ULL,
    0x6baee14fc4000ULL,
    0xff0a2923e2000ULL,
    0xdff22abbb000ULL,
    0x7c6f9f628b800ULL,
    0x3a5ec90fc6c00ULL,
    0x9220183ef200ULL,
    0x68b04dbb3d00ULL,
    0x5fadaffa8c880ULL,
    0xb125d843c78c0ULL,
    0x6aa790b1e40e0ULL,
    0xaddad27ab6cf0ULL,
    0x8c483a8003ea8ULL,
    0x22442840073b4ULL,
    0x950a28a006322ULL,
    0xecff26700e78bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xd800000000000ULL,
    0xec00000000000ULL,
    0xf200000000000ULL,
    0x6500000000000ULL,
    0x8780000000000ULL,
    0x5c0000000000ULL,
    0x48a0000000000ULL,
    0xcb10000000000ULL,
    0x58f8000000000ULL,
    0xb334000000000ULL,
    0x84d2000000000ULL,
    0xc913000000000ULL,
    0xd5f5800000000ULL,
    0x5094400000000ULL,
    0x470da00000000ULL,
    0xfaa0700000000ULL,
    0xe5fb80000000ULL,
    0xef73640000000ULL,
    0x3e8a0e0000000ULL,
    0xf8371f0000000ULL,
    0x1c5f928000000ULL,
    0x1a73764000000ULL,
    0x10a0b6000000ULL,
    0x41f7133000000ULL,
    0x7eff974800000ULL,
    0x58637ef400000ULL,
    0x2c7233f600000ULL,
    0x9203147900000ULL,
    0x350d81a280000ULL,
    0x5fa0610dc0000ULL,
    0xe9dfb597e0000ULL,
    0xbab340dcd0000ULL,
    0xae2a132288000ULL,
    0xdf27174dcc000ULL,
    0xb6a7bef7e6000ULL,
    0xcc4753ecd5000ULL,
    0x258046a85800ULL,
    0x8d2429b9c2c00ULL,
    0xe3aa2d01e9200ULL,
    0xc3e73795d3500ULL,
    0x3387bdc80df80ULL,
    0xdb976cb4029c0ULL,
    0xbf8036960daa0ULL,
    0x79c018490fe10ULL,
    0x2a0046a80778ULL,
    0x121029b9c5af4ULL,
    0xf5782d01e3e72ULL,
    0x3ff43795d6703ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x5800000000000ULL,
    0xc400000000000ULL,
    0x6600000000000ULL,
    0x3b00000000000ULL,
    0x3980000000000ULL,
    0xd7c0000000000ULL,
    0x10a0000000000ULL,
    0xbb70000000000ULL,
    0xf9f8000000000ULL,
    0x77f4000000000ULL,
    0x80a6000000000ULL,
    0xe30d000000000ULL,
    0x3db4800000000ULL,
    0x11c6400000000ULL,
    0xbbcca00000000ULL,
    0xdaf2700000000ULL,
    0xea4a880000000ULL,
    0x14f540000000ULL,
    0xa61e0000000ULL,
    0x230d250000000ULL,
    0x9db4a78000000ULL,
    0x81c65bc000000ULL,
    0xe3cca1e000000ULL,
    0x1ef27a3000000ULL,
    0x8c4a9bc800000ULL,
    0x3a4f41ec00000ULL,
    0x39262a3600000ULL,
    0xf4cd23d100000ULL,
    0x8d14bdff80000ULL,
    0x3ab6502240000ULL,
    0x1a34b0daa0000ULL,
    0x69065b7fd0000ULL,
    0xcec9a7f88000ULL,
    0xd9424be24c000ULL,
    0x492b13aaa000ULL,
    0xe50b514fd9000ULL,
    0x36d809b78d800ULL,
    0xe0441e0e40400ULL,
    0xf07e250cac600ULL,
    0x6849279edab00ULL,
    0xc4a9bc80e180ULL,
    0xfa4f41ec0d3c0ULL,
    0x99262a360d6a0ULL,
    0x64cd23d101070ULL,
    0xd514bdff89878ULL,
    0xfeb6502246434ULL,
    0x7c34b0daaf606ULL,
    0x52065b7fd637dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0xb400000000000ULL,
    0xfa00000000000ULL,
    0x4700000000000ULL,
    0xd180000000000ULL,
    0x1fc0000000000ULL,
    0xe2e0000000000ULL,
    0x9410000000000ULL,
    0x4a58000000000ULL,
    0xf24000000000ULL,
    0xcd8e000000000ULL,
    0xe9bb000000000ULL,
    0xebe4800000000ULL,
    0xf8a6400000000ULL,
    0xc35ca00000000ULL,
    0x2392500000000ULL,
    0xa48a980000000ULL,
    0xd50d540000000ULL,
    0x3ae0360000000ULL,
    0x7010390000000ULL,
    0xe858288000000ULL,
    0xec2438c000000ULL,
    0x5e0e24e000000ULL,
    0x57b3b3000000ULL,
    0x2284b25800000ULL,
    0x3476733400000ULL,
    0xba64be4e00000ULL,
    0xa766713d00000ULL,
    0xc1bc814d80000ULL,
    0xa78248a3c0000ULL,
    0x56d2ad0ca0000ULL,
    0x6e297e8e50000ULL,
    0xd6e31cd88000ULL,
    0xdeab2463c4000ULL,
    0xd23cbfecae000ULL,
    0xb427cbe51000ULL,
    0x7fb2ab1583800ULL,
    0xb2f96f97cf400ULL,
    0xcc562542a1a00ULL,
    0xee5f36b355700ULL,
    0x4d0a98000e980ULL,
    0x3ecd54000ebc0ULL,
    0xc20036000f8e0ULL,
    0xb30039000c310ULL,
    0xcb802880023d8ULL,
    0x48c038c00a4e4ULL,
    0x8b6024e00d56eULL,
    0x3fd03b3003aabULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xb800000000000ULL,
    0x1400000000000ULL,
    0xd200000000000ULL,
    0x6d00000000000ULL,
    0x2580000000000ULL,
    0x73c0000000000ULL,
    0x54e0000000000ULL,
    0x3850000000000ULL,
    0x5438000000000ULL,
    0xb244000000000ULL,
    0x3d7e000000000ULL,
    0x9dbf000000000ULL,
    0x6795800000000ULL,
    0x86ad400000000ULL,
    0x554da00000000ULL,
    0x71b9500000000ULL,
    0xc18bb80000000ULL,
    0x6982440000000ULL,
    0xa580160000000ULL,
    0x33c0010000000ULL,
    0x34e0028000000ULL,
    0x68500a4000000ULL,
    0xec3813e000000ULL,
    0xa64402b000000ULL,
    0xef7e28d800000ULL,
    0xf0bf09a400000ULL,
    0x4215895600000ULL,
    0xf56d7e7500000ULL,
    0x1adaf6980000ULL,
    0x49e955eac0000ULL,
    0x95b3bbb4e0000ULL,
    0xdbc66e5590000ULL,
    0x98fe15e988000ULL,
    0xae7f1baac4000ULL,
    0x5375be54e6000ULL,
    0xeefd6de595000ULL,
    0xb975bfb183800ULL,
    0xd7fd584ec5400ULL,
    0x2ef584e2eb200ULL,
    0x993d412093d00ULL,
    0xe795800001d80ULL,
    0xc6ad4000027c0ULL,
    0x354da0000e6e0ULL,
    0x21b9500000550ULL,
    0x798bb8000c9b8ULL,
    0x7d8244000d584ULL,
    0x778016000bb9eULL,
    0x5ec001000c8efULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0x2c00000000000ULL,
    0x8600000000000ULL,
    0x7900000000000ULL,
    0xe280000000000ULL,
    0xd8c0000000000ULL,
    0xafe0000000000ULL,
    0xc010000000000ULL,
    0xa028000000000ULL,
    0x1014000000000ULL,
    0xc872000000000ULL,
    0x1449000000000ULL,
    0xaa69800000000ULL,
    0xff0ec00000000ULL,
    0x9ba1a00000000ULL,
    0x3a0ad00000000ULL,
    0x777b980000000ULL,
    0x6f97ec0000000ULL,
    0x6000160000000ULL,
    0xb000270000000ULL,
    0xd800178000000ULL,
    0xdc00294000000ULL,
    0xbe00172000000ULL,
    0x5500237000000ULL,
    0x648032d800000ULL,
    0xa1c0187400000ULL,
    0x4d603b5200000ULL,
    0x18d0023100000ULL,
    0xfc831ee80000ULL,
    0xd004311340000ULL,
    0x685a308ce0000ULL,
    0x45d3ed4f0000ULL,
    0x621bbe6e88000ULL,
    0xeb47f45344000ULL,
    0x31c831acee000ULL,
    0xc5043aa4ff000ULL,
    0xecda1b368b800ULL,
    0x559d056746c00ULL,
    0x177bbddee6600ULL,
    0xdf97cbe5f8900ULL,
    0xb800000005a80ULL,
    0x6c0000000b4c0ULL,
    0x660000000c9e0ULL,
    0x8900000004910ULL,
    0xda80000007aa8ULL,
    0xf4c000000e4d4ULL,
    0x29e000000e192ULL,
    0xb91000000ad59ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0x3400000000000ULL,
    0x3e00000000000ULL,
    0x1b00000000000ULL,
    0xe080000000000ULL,
    0xe2c0000000000ULL,
    0xd3a0000000000ULL,
    0xc650000000000ULL,
    0xa708000000000ULL,
    0xacc000000000ULL,
    0xf7e6000000000ULL,
    0x6001000000000ULL,
    0xf018800000000ULL,
    0xa80ac00000000ULL,
    0x430a00000000ULL,
    0x7656f00000000ULL,
    0x2f7e980000000ULL,
    0xdecbfc0000000ULL,
    0xf9880a0000000ULL,
    0x330c310000000ULL,
    0x24c6258000000ULL,
    0x749107c000000ULL,
    0xccb0a5a000000ULL,
    0x5096f37000000ULL,
    0x6adea34800000ULL,
    0x79bffe400000ULL,
    0xc8003d0a00000ULL,
    0xf400379700000ULL,
    0xde000ad380000ULL,
    0x2b002a2740000ULL,
    0xa88035bda0000ULL,
    0xd6c03b71f0000ULL,
    0xeda0375388000ULL,
    0xdd5011e74c000ULL,
    0x47883a1dae000ULL,
    0xe80c0901f3000ULL,
    0x2446299b8c800ULL,
    0xa65115c34f400ULL,
    0x5710a8b7ade00ULL,
    0xa2c6fce6f2b00ULL,
    0xf3d6a58002880ULL,
    0x1657c7c0016c0ULL,
    0xdf6605a000da0ULL,
    0x76c103700ed50ULL,
    0xfdb8bb4800f88ULL,
    0x455ac3e40dc0cULL,
    0xbb8970a01a46ULL,
    0xaa5af6970bd51ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x9800000000000ULL,
    0x9c00000000000ULL,
    0x4e00000000000ULL,
    0x5900000000000ULL,
    0x780000000000ULL,
    0xddc0000000000ULL,
    0xdea0000000000ULL,
    0x1a30000000000ULL,
    0x2308000000000ULL,
    0x34a4000000000ULL,
    0xa13a000000000ULL,
    0x8bc5000000000ULL,
    0xdb95800000000ULL,
    0x73d0400000000ULL,
    0x57bda00000000ULL,
    0x7584700000000ULL,
    0xfaafa80000000ULL,
    0x38154c0000000ULL,
    0xac280e0000000ULL,
    0xf6542b0000000ULL,
    0x35123d8000000ULL,
    0xd1910d4000000ULL,
    0x1887b46000000ULL,
    0x9741463000000ULL,
    0x9eba05c800000ULL,
    0xfa0517bc00000ULL,
    0xf335b68a00000ULL,
    0x5ce040d500000ULL,
    0xa535912480000ULL,
    0x59e07e5440000ULL,
    0xccb5ade9e0000ULL,
    0x2d20696c50000ULL,
    0x8d9584a488000ULL,
    0x76d07f144c000ULL,
    0x3e3db789e2000ULL,
    0x144745c5f000ULL,
    0xbb8f94ec81800ULL,
    0x63e569e845c00ULL,
    0x1f801b63e6e00ULL,
    0x81c029b95a900ULL,
    0xb0a0158001f80ULL,
    0xb3300140081c0ULL,
    0xbc881a600b0a0ULL,
    0x75641d300b330ULL,
    0x319a10480bc88ULL,
    0xc8f516fc07564ULL,
    0xff1dacea0319aULL,
    0x9ab45de50c8f5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x8400000000000ULL,
    0x9200000000000ULL,
    0x9100000000000ULL,
    0xbd80000000000ULL,
    0x8cc0000000000ULL,
    0x6160000000000ULL,
    0xc5b0000000000ULL,
    0x30d8000000000ULL,
    0x6f6c000000000ULL,
    0x4af6000000000ULL,
    0xa53000000000ULL,
    0x5d2d800000000ULL,
    0x8bc0400000000ULL,
    0x9fdba00000000ULL,
    0x4593500000000ULL,
    0x70f6280000000ULL,
    0x4f53140000000ULL,
    0x5aad8a0000000ULL,
    0x200650000000ULL,
    0xd93b868000000ULL,
    0x19e3554000000ULL,
    0xece23e000000ULL,
    0xf84f137000000ULL,
    0xfc63bd3800000ULL,
    0x2e4f775c00000ULL,
    0x9f5812ee00000ULL,
    0x32ac3ff700000ULL,
    0xb6161d6b80000ULL,
    0x53231a3fc0000ULL,
    0x495b0cee0000ULL,
    0xa51c5338d0000ULL,
    0x77f5a05388000ULL,
    0xb1ac6d63c4000ULL,
    0xdaada220e2000ULL,
    0x42006ccfd1000ULL,
    0xf93bbd3808800ULL,
    0x9e3775c0c400ULL,
    0x6ce12ee0b200ULL,
    0x7c4f3ff708100ULL,
    0x6e639d6b8b580ULL,
    0xbf4f5a3fc08c0ULL,
    0x22d810ceef360ULL,
    0xbe6c0338d54b0ULL,
    0xd776085380d58ULL,
    0x96933963ca3acULL,
    0x344d8820e0b96ULL,
    0xca7059cfddfe3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x7800000000000ULL,
    0xac00000000000ULL,
    0x3a00000000000ULL,
    0xd00000000000ULL,
    0xf180000000000ULL,
    0x6cc0000000000ULL,
    0xf520000000000ULL,
    0x9df0000000000ULL,
    0x76a8000000000ULL,
    0x864000000000ULL,
    0x141a000000000ULL,
    0xb623000000000ULL,
    0xc75f800000000ULL,
    0x8494400000000ULL,
    0x3145a00000000ULL,
    0xa3b7700000000ULL,
    0x659a280000000ULL,
    0x1ae30c0000000ULL,
    0x127f960000000ULL,
    0xe964570000000ULL,
    0x3fedb08000000ULL,
    0x7d3584000000ULL,
    0x4b801ae000000ULL,
    0xa1c0147000000ULL,
    0x24a0172800000ULL,
    0x1302b4c00000ULL,
    0xfb88306200000ULL,
    0x39940d2500000ULL,
    0x58b22a4c80000ULL,
    0xb34737e1c0000ULL,
    0x22c5b5f960000ULL,
    0x5e7770e1d0000ULL,
    0x33a3d6488000ULL,
    0xbad31cadcc000ULL,
    0x2277859b62000ULL,
    0xb1307dc4df000ULL,
    0x63ff97280f800ULL,
    0x45a46b4c06c00ULL,
    0xeacd906201a00ULL,
    0x6a237d250fd00ULL,
    0x4528024c88980ULL,
    0x5a43be1cc0c0ULL,
    0xaba23f96cf20ULL,
    0xba1327e1d90f0ULL,
    0xcd578de480728ULL,
    0xd1c044edca4a4ULL,
    0x9cd79f7b6c13aULL,
    0x8d0069b4ddbd3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0x6c00000000000ULL,
    0xaa00000000000ULL,
    0x8300000000000ULL,
    0xd780000000000ULL,
    0xc0c0000000000ULL,
    0xa160000000000ULL,
    0x30d0000000000ULL,
    0x9928000000000ULL,
    0x8cf4000000000ULL,
    0x9b4a000000000ULL,
    0xfbdb000000000ULL,
    0x8ae8800000000ULL,
    0x1264400000000ULL,
    0x7f42a00000000ULL,
    0x35af500000000ULL,
    0x87e2180000000ULL,
    0x28ef1c0000000ULL,
    0xb5429e0000000ULL,
    0xc6af570000000ULL,
    0x28622c8000000ULL,
    0xb42f2bc000000ULL,
    0x2622a76000000ULL,
    0x197f5cf000000ULL,
    0xccca1bb800000ULL,
    0x7b1b370400000ULL,
    0xcb889c9200000ULL,
    0x12b443c900000ULL,
    0x7e6ab37880000ULL,
    0xd55b66fbc0000ULL,
    0xb6a81eb0a0000ULL,
    0x50340a9bd0000ULL,
    0xe82a30c088000ULL,
    0x140b0dffc4000ULL,
    0xf640bc22ae000ULL,
    0xb1504e52d3000ULL,
    0x38e8b73801800ULL,
    0xbd645cc402c00ULL,
    0xe2c29bf204a00ULL,
    0x466f4f390b300ULL,
    0x690230c084f80ULL,
    0xb4ff0dffcacc0ULL,
    0x270abc22a0b60ULL,
    0xf98b4e52db3d0ULL,
    0xfd8037380cea8ULL,
    0x3c01cc400c34ULL,
    0x96e03bf20da2aULL,
    0xc0101f390fb0bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x2800000000000ULL,
    0x8c00000000000ULL,
    0x2e00000000000ULL,
    0xc300000000000ULL,
    0xae80000000000ULL,
    0x79c0000000000ULL,
    0x9d20000000000ULL,
    0xe5d0000000000ULL,
    0xb68000000000ULL,
    0xd2ec000000000ULL,
    0x1fa2000000000ULL,
    0xe269000000000ULL,
    0x4d32800000000ULL,
    0x3dd8c00000000ULL,
    0xcf30a00000000ULL,
    0x40a1f00000000ULL,
    0xdaca380000000ULL,
    0x3853c0000000ULL,
    0xb410920000000ULL,
    0x1a71ef0000000ULL,
    0x1922218000000ULL,
    0xd7a923c000000ULL,
    0x9e12b82000000ULL,
    0x2b08e2b000000ULL,
    0x42d8a6e800000ULL,
    0x678df40400000ULL,
    0x7648161200000ULL,
    0xc73c010f00000ULL,
    0x5cca3c9280000ULL,
    0x8c853d5140000ULL,
    0x5490b26c20000ULL,
    0x90b1dd5870000ULL,
    0x282227a88000ULL,
    0xc7b935554c000ULL,
    0x265a967e26000ULL,
    0xdf34c35773000ULL,
    0xf89287680a800ULL,
    0x2ec8d7c404c00ULL,
    0xb9f8ae3204e00ULL,
    0xfd5de3bf0f300ULL,
    0xd5a01a7a88680ULL,
    0x231009554f5c0ULL,
    0x5ec8047e2b320ULL,
    0x31fc2c57726d0ULL,
    0x216a26e8025e8ULL,
    0xe395340406b2cULL,
    0x4458b6120e282ULL,
    0x524df10f037b9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x4800000000000ULL,
    0x6c00000000000ULL,
    0x4e00000000000ULL,
    0x3b00000000000ULL,
    0x9480000000000ULL,
    0xc1c0000000000ULL,
    0xbe20000000000ULL,
    0xb350000000000ULL,
    0x9888000000000ULL,
    0xffdc000000000ULL,
    0xcd32000000000ULL,
    0x4bc1000000000ULL,
    0x1772800000000ULL,
    0x7aabc00000000ULL,
    0xeac8a00000000ULL,
    0x12b6f00000000ULL,
    0x5688380000000ULL,
    0x4dc2c0000000ULL,
    0x39b20a0000000ULL,
    0xfa01070000000ULL,
    0xe152818000000ULL,
    0xa5fbd5c000000ULL,
    0x3c4096a000000ULL,
    0xd66aceb000000ULL,
    0xf3a32a800000ULL,
    0x8edd30a400000ULL,
    0x90e083aa00000ULL,
    0x33fad42300000ULL,
    0x931234ee80000ULL,
    0x489108c740000ULL,
    0xa7fa983020000ULL,
    0x9977eeeab0000ULL,
    0x21fa87c688000ULL,
    0xe77eda34c000ULL,
    0x9b7a8d3a26000ULL,
    0x84b7f479b7000ULL,
    0xf9da81800c800ULL,
    0x9a27d5c00ac00ULL,
    0x917296a002e00ULL,
    0xedabceb004b00ULL,
    0x5048b2a80dc80ULL,
    0x9876f0a40adc0ULL,
    0x342823aa0f020ULL,
    0x1a4c242308850ULL,
    0x511a0cee80c08ULL,
    0x8d8d24c743e1cULL,
    0x2068923027312ULL,
    0xd026e9eabf891ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x5800000000000ULL,
    0x4400000000000ULL,
    0x1a00000000000ULL,
    0xf100000000000ULL,
    0x4e80000000000ULL,
    0xf5c0000000000ULL,
    0x3260000000000ULL,
    0x3d10000000000ULL,
    0x28f8000000000ULL,
    0xcaa4000000000ULL,
    0xcfee000000000ULL,
    0x337f000000000ULL,
    0xbbad800000000ULL,
    0xc14bc00000000ULL,
    0xa6bba00000000ULL,
    0x1990d00000000ULL,
    0xa478380000000ULL,
    0xca64340000000ULL,
    0xc90e0e0000000ULL,
    0x9aaf350000000ULL,
    0xb7b5908000000ULL,
    0x873fed4000000ULL,
    0x69cdb52000000ULL,
    0x2c5bd13000000ULL,
    0xb643a73800000ULL,
    0x734c63400000ULL,
    0x299628a600000ULL,
    0x4c1b18ed00000ULL,
    0x2623b14580000ULL,
    0x5f24f736c0000ULL,
    0x6dee3e30e0000ULL,
    0x567f3cbd30000ULL,
    0xd72d86fd88000ULL,
    0x118bdc42c4000ULL,
    0x985ba3b6e6000ULL,
    0x6440f56039000ULL,
    0xea6010800d800ULL,
    0x39102d4000400ULL,
    0x52f8152007a00ULL,
    0xaba4013006100ULL,
    0xd96e1f3801680ULL,
    0x82bf32340b1c0ULL,
    0x93cd86a602860ULL,
    0xd5bfded0cc10ULL,
    0xc0c399c586678ULL,
    0x26f4ee76c3f64ULL,
    0x59f62510efd8eULL,
    0xc40b088d30e6fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0xe400000000000ULL,
    0x4200000000000ULL,
    0xbd00000000000ULL,
    0x6a80000000000ULL,
    0x5c0000000000ULL,
    0x2a20000000000ULL,
    0x8910000000000ULL,
    0xf088000000000ULL,
    0x64dc000000000ULL,
    0x2eb6000000000ULL,
    0x9783000000000ULL,
    0x4f57800000000ULL,
    0x3fe7c00000000ULL,
    0x9b69a00000000ULL,
    0x55b8f00000000ULL,
    0x3208180000000ULL,
    0xb51c0c0000000ULL,
    0x6e960a0000000ULL,
    0xb793050000000ULL,
    0x5f5fa28000000ULL,
    0x7fbd64000000ULL,
    0xb77faa2000000ULL,
    0xf3ebde3000000ULL,
    0xcd77882800000ULL,
    0x62f7ef3400000ULL,
    0x1e19caa00000ULL,
    0x9864ce7300000ULL,
    0xfc3e0d0f80000ULL,
    0x7e5f2697c0000ULL,
    0x2361985860000ULL,
    0xb9a4c12930000ULL,
    0x741e27a788000ULL,
    0xba4f1fe3cc000ULL,
    0x7169aed26e000ULL,
    0x3cb8d16a3f000ULL,
    0x3288228004800ULL,
    0x19dc164002400ULL,
    0xa4360a200a200ULL,
    0x62432e3004d00ULL,
    0xad7790280a280ULL,
    0x52f7e3340e1c0ULL,
    0x29e196aa06820ULL,
    0x8c64cb7303410ULL,
    0x763e2f8f89a08ULL,
    0x275f30d7c611cULL,
    0xbe1927860496ULL,
    0x164ef1931e93ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x4800000000000ULL,
    0xdc00000000000ULL,
    0x9200000000000ULL,
    0x5300000000000ULL,
    0x6c80000000000ULL,
    0x85c0000000000ULL,
    0x3660000000000ULL,
    0xe550000000000ULL,
    0xc9f8000000000ULL,
    0xac6c000000000ULL,
    0x8a6a000000000ULL,
    0x2757000000000ULL,
    0x32e8800000000ULL,
    0xcfbc00000000ULL,
    0xd5faa00000000ULL,
    0x9e00d00000000ULL,
    0x2918180000000ULL,
    0x13fc140000000ULL,
    0x23722a0000000ULL,
    0x74ab330000000ULL,
    0xf19ab68000000ULL,
    0x6850e3c000000ULL,
    0x6c601fa000000ULL,
    0x2a5025b000000ULL,
    0xd7782eb800000ULL,
    0x6aac1c2400000ULL,
    0x988a2de600000ULL,
    0x9bc7254f00000ULL,
    0x5f70b46480000ULL,
    0x16c7f60ec0000ULL,
    0xfae89500a0000ULL,
    0x90fbfb9b30000ULL,
    0xa7faac5c88000ULL,
    0xdd00c9eac4000ULL,
    0xd980746ae000ULL,
    0x4a3c2b6431000ULL,
    0x87122e800c800ULL,
    0xc2fb37c009c00ULL,
    0x54e295a007200ULL,
    0x41fcc6b004300ULL,
    0xd06a003802480ULL,
    0xe8572be4059c0ULL,
    0x2c68b8460a460ULL,
    0xca3be3ff0b650ULL,
    0xc71ab45c8a578ULL,
    0x2290ddeac29acULL,
    0x44802d46abc0aULL,
    0x9c018643c207ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0x6c00000000000ULL,
    0x2e00000000000ULL,
    0x7100000000000ULL,
    0x7c80000000000ULL,
    0xebc0000000000ULL,
    0xd220000000000ULL,
    0x6750000000000ULL,
    0xd1d8000000000ULL,
    0xf164000000000ULL,
    0xbc9a000000000ULL,
    0x8bd1000000000ULL,
    0x267800000000ULL,
    0xff1ac00000000ULL,
    0xbda5a00000000ULL,
    0xdf6ff00000000ULL,
    0xcdf8380000000ULL,
    0xf734040000000ULL,
    0xe9c23e0000000ULL,
    0x2d752f0000000ULL,
    0xdaddad8000000ULL,
    0xe9bc74000000ULL,
    0x3c9a34a000000ULL,
    0x4bd116b000000ULL,
    0x6267b3a800000ULL,
    0x2f1ad72400000ULL,
    0x25a586fe00000ULL,
    0xb36fce8d00000ULL,
    0xe3f828d080000ULL,
    0x863406ed40000ULL,
    0x95420e9f60000ULL,
    0xc6b508c2f0000ULL,
    0x8fdb6f888000ULL,
    0x69cbd6894c000ULL,
    0xed421cc166000ULL,
    0xbab520fffd000ULL,
    0xdefd958001800ULL,
    0xa4cbc3400ac00ULL,
    0x27c20aa004e00ULL,
    0x4c7539b00a100ULL,
    0x5e5d9e280e480ULL,
    0x595bd064087c0ULL,
    0x58ba125e0fc20ULL,
    0x3181283d01650ULL,
    0xe13fa3788ad58ULL,
    0x44bed5c941aa4ULL,
    0x379fb66166ebaULL,
    0xb42ee94ffec81ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0xac00000000000ULL,
    0x600000000000ULL,
    0xfd00000000000ULL,
    0xef80000000000ULL,
    0xf8c0000000000ULL,
    0x8c20000000000ULL,
    0xf630000000000ULL,
    0xe548000000000ULL,
    0x73c4000000000ULL,
    0x46ca000000000ULL,
    0xdd75000000000ULL,
    0x1fcd800000000ULL,
    0xe081400000000ULL,
    0x106fa00000000ULL,
    0x4800700000000ULL,
    0xb420080000000ULL,
    0x9a303c0000000ULL,
    0x4348060000000ULL,
    0xbec4270000000ULL,
    0x114a2f8000000ULL,
    0x89b5144000000ULL,
    0x95edba6000000ULL,
    0xebb1417000000ULL,
    0x1aa7b8e800000ULL,
    0xc30473bc00000ULL,
    0x7eca125a00000ULL,
    0xb1751d7d00000ULL,
    0xb9cdaee080000ULL,
    0x2d814cac40000ULL,
    0x47ef99c1e0000ULL,
    0x1cc06b3d10000ULL,
    0x3e00318888000ULL,
    0x910017504c000ULL,
    0x498037fbea000ULL,
    0x35c0103013000ULL,
    0xdba0080003800ULL,
    0xa2f03c0006c00ULL,
    0x6f6806000a600ULL,
    0x78f427000cd00ULL,
    0x4c022f8005780ULL,
    0x56711440054c0ULL,
    0xd527ba6008a20ULL,
    0xcbc4417000b30ULL,
    0xeaea38e800ac8ULL,
    0xdb4533bc08b04ULL,
    0xe285b25a0caeaULL,
    0xf456d7d02b45ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x1400000000000ULL,
    0x5a00000000000ULL,
    0x7500000000000ULL,
    0x6c80000000000ULL,
    0x87c0000000000ULL,
    0xdc60000000000ULL,
    0xf670000000000ULL,
    0xcb78000000000ULL,
    0x4b84000000000ULL,
    0xd266000000000ULL,
    0x7907000000000ULL,
    0x82c7800000000ULL,
    0xf8e4c00000000ULL,
    0x9db9a00000000ULL,
    0x917d00000000ULL,
    0xcae0080000000ULL,
    0x14b0040000000ULL,
    0x83983e0000000ULL,
    0x7e34110000000ULL,
    0xc77e008000000ULL,
    0xa5f3184000000ULL,
    0xad598da000000ULL,
    0x38a7fcb000000ULL,
    0x7df80c3800000ULL,
    0xf9440c6c00000ULL,
    0xc2862dc600000ULL,
    0x58b73b7d00000ULL,
    0xcddf904780000ULL,
    0x5110c9a0c0000ULL,
    0xdea7a2a860000ULL,
    0x4e94ed38f0000ULL,
    0xf6c194ff88000ULL,
    0x1293d98cc4000ULL,
    0x40863cce6a000ULL,
    0x79b73bf5f5000ULL,
    0x5b5f88000d800ULL,
    0xf3d0c40005400ULL,
    0x36479e000fa00ULL,
    0x2b24c10002500ULL,
    0xbbd9888003480ULL,
    0xda67dc40093c0ULL,
    0x351813a008660ULL,
    0xccf43db008370ULL,
    0xd79e04b80a7f8ULL,
    0x8443102c0cc44ULL,
    0xe2419e6600e06ULL,
    0x9153d6cd08f77ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xe800000000000ULL,
    0x9400000000000ULL,
    0x4200000000000ULL,
    0x7b00000000000ULL,
    0x4980000000000ULL,
    0x3cc0000000000ULL,
    0x9020000000000ULL,
    0x5850000000000ULL,
    0x1c08000000000ULL,
    0xa64c000000000ULL,
    0xd13e000000000ULL,
    0xa6eb000000000ULL,
    0x375c800000000ULL,
    0xd7f9400000000ULL,
    0x81caa00000000ULL,
    0x78ce700000000ULL,
    0x2a00380000000ULL,
    0x2f002c0000000ULL,
    0x6b80220000000ULL,
    0x37c0390000000ULL,
    0x31a02a8000000ULL,
    0xf0903bc000000ULL,
    0xce2802e000000ULL,
    0x851c11f000000ULL,
    0x84b6366800000ULL,
    0x3c67192400000ULL,
    0x7642a30a00000ULL,
    0x29427f8700000ULL,
    0xaa9e134d80000ULL,
    0x97b3029c0000ULL,
    0x7af4a198a0000ULL,
    0xf125404490000ULL,
    0x76dcb7a588000ULL,
    0xcf397ecdcc000ULL,
    0xdbea8272a6000ULL,
    0xbf9e673397000ULL,
    0xd588000006800ULL,
    0x5a8c000005400ULL,
    0x211e000002200ULL,
    0x8ebb000000b00ULL,
    0xc35480000a180ULL,
    0xe5b540000a8c0ULL,
    0x12f4a0000d220ULL,
    0xa525700002350ULL,
    0x54dcb80005588ULL,
    0xc4396c0009a8cULL,
    0x7a6a82000411eULL,
    0x175e49000febbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x2800000000000ULL,
    0x1400000000000ULL,
    0x4a00000000000ULL,
    0xe300000000000ULL,
    0x6f80000000000ULL,
    0x72c0000000000ULL,
    0x7020000000000ULL,
    0xe830000000000ULL,
    0x3408000000000ULL,
    0xba3c000000000ULL,
    0xcb0a000000000ULL,
    0x7b85000000000ULL,
    0x38d2800000000ULL,
    0x9318c00000000ULL,
    0x87abe00000000ULL,
    0x46d4b00000000ULL,
    0xca00080000000ULL,
    0x23000c0000000ULL,
    0x4f80020000000ULL,
    0x82c00f0000000ULL,
    0x5820028000000ULL,
    0xfc30014000000ULL,
    0x7e0804a000000ULL,
    0x593c0e3000000ULL,
    0xa48a06f800000ULL,
    0x945072c00000ULL,
    0x48f2870200000ULL,
    0x7b28ce8300000ULL,
    0xb3a3e34080000ULL,
    0xfce8bba3c0000ULL,
    0x10a04b0a0000ULL,
    0x58850bb850000ULL,
    0x7752818d28000ULL,
    0x11d8c6318c000ULL,
    0xdf8beafabe000ULL,
    0xbae4b52d4b000ULL,
    0xb408000000800ULL,
    0x7a3c000000c00ULL,
    0xeb0a000000200ULL,
    0x8b85000000f00ULL,
    0x10d2800000280ULL,
    0x8718c00000140ULL,
    0xcdabe000004a0ULL,
    0xa5d4b00000e30ULL,
    0xa5800800006f8ULL,
    0x51c00c000072cULL,
    0x3fa0020000702ULL,
    0x6af00f0000e83ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0x9c00000000000ULL,
    0x7e00000000000ULL,
    0xff00000000000ULL,
    0x4380000000000ULL,
    0x79c0000000000ULL,
    0xb820000000000ULL,
    0x1410000000000ULL,
    0x5238000000000ULL,
    0xf914000000000ULL,
    0x88a000000000ULL,
    0xd867000000000ULL,
    0x40ff800000000ULL,
    0x108fc00000000ULL,
    0x7c78e00000000ULL,
    0x6ad2700000000ULL,
    0x5d80080000000ULL,
    0x96c0040000000ULL,
    0x33a00e0000000ULL,
    0xa1d0050000000ULL,
    0xbc18028000000ULL,
    0x8e0409c000000ULL,
    0x673207e000000ULL,
    0xa7b30ff000000ULL,
    0xb3d5843800000ULL,
    0xa538c79c00000ULL,
    0xd69f6b8200000ULL,
    0x9759b14100000ULL,
    0x7b4aed2380000ULL,
    0xdd617b9140000ULL,
    0x26558688a0000ULL,
    0xfff8cc8670000ULL,
    0xb33f688ff8000ULL,
    0x5589bdc8fc000ULL,
    0xfad2eaa78e000ULL,
    0xd5a5749d27000ULL,
    0xbac7800000800ULL,
    0x359bc00000400ULL,
    0xeaf2e00000e00ULL,
    0x1db5700000500ULL,
    0x76ff880000280ULL,
    0x638fc400009c0ULL,
    0x89f8ee00007e0ULL,
    0x2012750000ff0ULL,
    0xf0200a8000438ULL,
    0x98100dc00079cULL,
    0xe43809e000b82ULL,
    0xca140af000141ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x4800000000000ULL,
    0x1c00000000000ULL,
    0xae00000000000ULL,
    0xf900000000000ULL,
    0x6c80000000000ULL,
    0x95c0000000000ULL,
    0x7c20000000000ULL,
    0x3e30000000000ULL,
    0xe108000000000ULL,
    0x489c000000000ULL,
    0x6fd2000000000ULL,
    0x3727000000000ULL,
    0x59b800000000ULL,
    0xe176400000000ULL,
    0xcde7200000000ULL,
    0xb827700000000ULL,
    0x9420080000000ULL,
    0x92300c0000000ULL,
    0x2708020000000ULL,
    0xdd9c070000000ULL,
    0xe552048000000ULL,
    0x47e701c000000ULL,
    0xbb3b8ae000000ULL,
    0xb3864f9000000ULL,
    0x3c4f26c800000ULL,
    0x5b4b795c00000ULL,
    0x66da0fc200000ULL,
    0xd3bb0fe300000ULL,
    0xac498c1080000ULL,
    0x43514389c0000ULL,
    0x42fca27d20000ULL,
    0x299132b270000ULL,
    0xe76722b9b8000ULL,
    0x78e77d8764000ULL,
    0x4280001672000ULL,
    0xacc009de77000ULL,
    0x30a0080000800ULL,
    0xdbf00c0000c00ULL,
    0xd528020000200ULL,
    0x6aac070000700ULL,
    0x20da048000480ULL,
    0x86bb01c0001c0ULL,
    0x6c98ae000ae0ULL,
    0x43914f9000f90ULL,
    0xb45ca6c8006c8ULL,
    0x6761395c0095cULL,
    0xb8cf2fc2007c2ULL,
    0x628b7fe3003e3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0xbc00000000000ULL,
    0xe00000000000ULL,
    0xe100000000000ULL,
    0xb580000000000ULL,
    0x3dc0000000000ULL,
    0x8c20000000000ULL,
    0xd610000000000ULL,
    0x7518000000000ULL,
    0xd7b4000000000ULL,
    0x9ad2000000000ULL,
    0x648f000000000ULL,
    0x5053800000000ULL,
    0x25c0400000000ULL,
    0x3829600000000ULL,
    0x415700000000ULL,
    0x4a20080000000ULL,
    0xcb10040000000ULL,
    0xae98060000000ULL,
    0xdb740d0000000ULL,
    0xeb72048000000ULL,
    0x335f0bc000000ULL,
    0xa76b80e000000ULL,
    0xc5644e1000000ULL,
    0x62636b5800000ULL,
    0x8aee73dc00000ULL,
    0xc8180c200000ULL,
    0x5c4f496100000ULL,
    0xb3fae15180000ULL,
    0x2d15307b40000ULL,
    0x3a9652d20000ULL,
    0x98d57988f0000ULL,
    0x13800be538000ULL,
    0xf0c0054c04000ULL,
    0x1fa00a5a96000ULL,
    0x66d0055d57000ULL,
    0xab8080000800ULL,
    0x6164040000400ULL,
    0xd86a060000600ULL,
    0xb9eb0d0000d00ULL,
    0x8639848000480ULL,
    0x7d2b4bc000bc0ULL,
    0xb90e0e0000e0ULL,
    0x44fe3e1000e10ULL,
    0xcd90e35800b58ULL,
    0x59fe37dc003dcULL,
    0x1610e6c2008c2ULL,
    0x553e346100d61ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0x3400000000000ULL,
    0x1200000000000ULL,
    0x4300000000000ULL,
    0x480000000000ULL,
    0xb840000000000ULL,
    0x4620000000000ULL,
    0x4130000000000ULL,
    0x3fb8000000000ULL,
    0x58f4000000000ULL,
    0x7446000000000ULL,
    0x701d000000000ULL,
    0x680c800000000ULL,
    0x9c1cc00000000ULL,
    0x6e13200000000ULL,
    0xfd05100000000ULL,
    0x6198080000000ULL,
    0xedc40c0000000ULL,
    0xb9fe0e0000000ULL,
    0xbbe90d0000000ULL,
    0x80ca898000000ULL,
    0x6041c34000000ULL,
    0x523fa12000000ULL,
    0x6329d43000000ULL,
    0x34b3284800000ULL,
    0xf075178400000ULL,
    0xea00026200000ULL,
    0x6700051300000ULL,
    0x6e80047b80000ULL,
    0x1f400bcf40000ULL,
    0xc8a00fe460000ULL,
    0x8e700071d0000ULL,
    0x6f1807e8c8000ULL,
    0xe2840675cc000ULL,
    0x95e02cb32000ULL,
    0xd199004751000ULL,
    0x65d2818000800ULL,
    0xf5c5cf4000c00ULL,
    0x4de1af2000e00ULL,
    0x49f0d93000d00ULL,
    0x13c1a1c800980ULL,
    0xfcc0d4c400340ULL,
    0xde79a34200120ULL,
    0x3734d12300430ULL,
    0x36bfac3380048ULL,
    0xcb69dc4b40b84ULL,
    0xa932d8660462ULL,
    0x55451562d0413ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0xec00000000000ULL,
    0xe00000000000ULL,
    0x2900000000000ULL,
    0x9f80000000000ULL,
    0xa940000000000ULL,
    0x5220000000000ULL,
    0x8f30000000000ULL,
    0x32a8000000000ULL,
    0x1cd4000000000ULL,
    0xa846000000000ULL,
    0x89ab000000000ULL,
    0xac5b800000000ULL,
    0x6396400000000ULL,
    0x5f65e00000000ULL,
    0x673f500000000ULL,
    0xd688080000000ULL,
    0xc6e40c0000000ULL,
    0x336e0a0000000ULL,
    0xa93f050000000ULL,
    0x5fbd898000000ULL,
    0x94d4ec000000ULL,
    0x23660e000000ULL,
    0x170d129000000ULL,
    0xdea3e1f800000ULL,
    0x12d4569400000ULL,
    0x8173872200000ULL,
    0x160241f300000ULL,
    0x503e0aa80000ULL,
    0x31a45a0d40000ULL,
    0xd07b8be460000ULL,
    0x55a648cab0000ULL,
    0xca4de9ddb8000ULL,
    0x6eab5b3d64000ULL,
    0xbaee08ac5e000ULL,
    0x57f06d4f5000ULL,
    0x3c1d818000800ULL,
    0x563d42c000c00ULL,
    0x653e6ae000a00ULL,
    0xc1a9179000500ULL,
    0x186de87800980ULL,
    0x219b585400ec0ULL,
    0x284607c2000e0ULL,
    0x49ab036300290ULL,
    0xc5b8952809f8ULL,
    0x3396409940a94ULL,
    0xc765e6c660522ULL,
    0x8b3f5c39b08f3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xf800000000000ULL,
    0xfc00000000000ULL,
    0x1e00000000000ULL,
    0x2b00000000000ULL,
    0x6780000000000ULL,
    0xc540000000000ULL,
    0xab20000000000ULL,
    0x2790000000000ULL,
    0x6568000000000ULL,
    0x9b2c000000000ULL,
    0xdfae000000000ULL,
    0x9957000000000ULL,
    0x852b800000000ULL,
    0xf4a4c00000000ULL,
    0xfecee00000000ULL,
    0x405ad00000000ULL,
    0x5fae080000000ULL,
    0xd957040000000ULL,
    0x252b8a0000000ULL,
    0xc4a4c30000000ULL,
    0x6ceef8000000ULL,
    0xbc5adfc000000ULL,
    0x41ae09e000000ULL,
    0xf25706b000000ULL,
    0x42ab8c7800000ULL,
    0x1e4cf5400000ULL,
    0xadeee53200000ULL,
    0x9bcaddb900000ULL,
    0x24c60fb680000ULL,
    0x697b0f02c0000ULL,
    0x9d058182e0000ULL,
    0x98b3c6c170000ULL,
    0x28c56d60b8000ULL,
    0x6f6e12f34c000ULL,
    0xda08e05a6e000ULL,
    0x2921db076d000ULL,
    0xc2ab8c7800800ULL,
    0x41e4cf5400400ULL,
    0xdeee53200a00ULL,
    0xabcaddb900300ULL,
    0xdcc60fb680f80ULL,
    0x957b0f02c0fc0ULL,
    0x83058182e01e0ULL,
    0xb3b3c6c1702b0ULL,
    0x4f456d60b8678ULL,
    0xaa2e12f34cc54ULL,
    0x7128e05a6eab2ULL,
    0xeb1db076d279ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x6800000000000ULL,
    0x9c00000000000ULL,
    0x600000000000ULL,
    0x2f00000000000ULL,
    0xf880000000000ULL,
    0x2a40000000000ULL,
    0x7f20000000000ULL,
    0x3090000000000ULL,
    0xc678000000000ULL,
    0x8104000000000ULL,
    0xeb8a000000000ULL,
    0xa4df000000000ULL,
    0x8245800000000ULL,
    0x4321c00000000ULL,
    0x46b1200000000ULL,
    0x1157100000000ULL,
    0x8d8a080000000ULL,
    0x5bdf040000000ULL,
    0xf2c58e0000000ULL,
    0x6561c90000000ULL,
    0x5791268000000ULL,
    0x92c719c000000ULL,
    0xb572086000000ULL,
    0xdf9b06f000000ULL,
    0x9eef818800000ULL,
    0xdb6ecba400000ULL,
    0x6c8ca17200000ULL,
    0x6072dac900000ULL,
    0xde31240780000ULL,
    0xeb171ee040000ULL,
    0x7aaa0730a0000ULL,
    0x674f05e9f0000ULL,
    0x5abd875658000ULL,
    0x5765c7fb1c000ULL,
    0x429b2eec92000ULL,
    0x3358123531000ULL,
    0xb097818800800ULL,
    0x866acba400400ULL,
    0x6106a17200e00ULL,
    0x7baddac900900ULL,
    0xccf4a40780680ULL,
    0x1e76dee0409c0ULL,
    0x453b2730a0060ULL,
    0x698815e9f02f0ULL,
    0xe9cf8f5658f88ULL,
    0xa7fec3fb1c2a4ULL,
    0x24f4a0ec927f2ULL,
    0xc276db3531309ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0x4400000000000ULL,
    0xc200000000000ULL,
    0x1300000000000ULL,
    0xcf80000000000ULL,
    0xe240000000000ULL,
    0x7120000000000ULL,
    0x6cb0000000000ULL,
    0xa5c8000000000ULL,
    0xa77c000000000ULL,
    0x77ba000000000ULL,
    0x9e69000000000ULL,
    0xf04800000000ULL,
    0x2182c00000000ULL,
    0x5740e00000000ULL,
    0x1fa5100000000ULL,
    0xfa72080000000ULL,
    0xbd150c0000000ULL,
    0x9abe820000000ULL,
    0xdcebc70000000ULL,
    0x3fc46a8000000ULL,
    0x9867d44000000ULL,
    0x1e12e42000000ULL,
    0xdd001d3000000ULL,
    0xa8486f800000ULL,
    0x24c2c52400000ULL,
    0xa3e0ef9200000ULL,
    0xb655158b00000ULL,
    0x8b1a04fc80000ULL,
    0xc3990307c0000ULL,
    0x346c85a3a0000ULL,
    0x780ec1f290000ULL,
    0x5c12e19a48000ULL,
    0xe001eb72c000ULL,
    0xe5048c1a8e000ULL,
    0xb682c07691000ULL,
    0x7ac0ec7800800ULL,
    0x9ee5116400c00ULL,
    0xecd20bb200200ULL,
    0x77e508bb00700ULL,
    0x8c56820480a80ULL,
    0x427c623c0440ULL,
    0x22366a31a0c20ULL,
    0x4332d47990130ULL,
    0x178c6566c8cf8ULL,
    0xe5bddb0ece24ULL,
    0xf708e9b92e712ULL,
    0xbd991184016cbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x1800000000000ULL,
    0x7c00000000000ULL,
    0x8e00000000000ULL,
    0x6f00000000000ULL,
    0x5280000000000ULL,
    0x1fc0000000000ULL,
    0x5920000000000ULL,
    0x71b0000000000ULL,
    0x2b78000000000ULL,
    0x5de4000000000ULL,
    0x9016000000000ULL,
    0xd817000000000ULL,
    0x9c1f800000000ULL,
    0x9e19c00000000ULL,
    0x770da00000000ULL,
    0x2ebb700000000ULL,
    0x91ee080000000ULL,
    0x36330c0000000ULL,
    0x23298e0000000ULL,
    0x34bec10000000ULL,
    0x4ea218000000ULL,
    0xe186b7c000000ULL,
    0xf355a0e000000ULL,
    0xc1ef7af000000ULL,
    0xe00032800000ULL,
    0xaf000cfc00000ULL,
    0xb2800a1200000ULL,
    0xfc001db00000ULL,
    0x41200bd780000ULL,
    0xdb004ee40000ULL,
    0xa57804c960000ULL,
    0x32e40a8d70000ULL,
    0xc296097bf8000ULL,
    0xc7d70f069c000ULL,
    0xc53f80555a000ULL,
    0xefa9c02ef7000ULL,
    0x5c75a2a800800ULL,
    0x735f7b3c00c00ULL,
    0x1f80af200e00ULL,
    0xee240b2b00100ULL,
    0xbf3600ff80180ULL,
    0xaaa70412407c0ULL,
    0x73e780db608e0ULL,
    0xcf3dca56706f0ULL,
    0x62bba32c78528ULL,
    0xf7dc7c28dc1fcULL,
    0x2d29847c3a592ULL,
    0x9bbec0538771bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x800000000000ULL,
    0xf400000000000ULL,
    0xa600000000000ULL,
    0x7700000000000ULL,
    0x6580000000000ULL,
    0xd3c0000000000ULL,
    0x4520000000000ULL,
    0xe490000000000ULL,
    0xd968000000000ULL,
    0xbf4c000000000ULL,
    0x2872000000000ULL,
    0x5de5000000000ULL,
    0x361d800000000ULL,
    0x8f0bc00000000ULL,
    0x39a2600000000ULL,
    0x31ce700000000ULL,
    0x9c3a080000000ULL,
    0x239040000000ULL,
    0xc9078a0000000ULL,
    0x5ea2cb0000000ULL,
    0xec4de08000000ULL,
    0xb3e0bf4000000ULL,
    0x1525e26000000ULL,
    0xfcacb37000000ULL,
    0x9557e45800000ULL,
    0xe549b23c00000ULL,
    0xd4a66d200000ULL,
    0xe9427e0900000ULL,
    0xf768057680000ULL,
    0x7c4c0cc4c0000ULL,
    0x4bf2043f20000ULL,
    0x49250bd250000ULL,
    0x1ebd8b8bd8000ULL,
    0x4c5bc3b5bc000ULL,
    0x3ea6666a6000ULL,
    0x1d127ced27000ULL,
    0x8a00cd800800ULL,
    0x3350097c00400ULL,
    0x92480eb200a00ULL,
    0x68dc067900b00ULL,
    0x3a9a01ae80080ULL,
    0xb26901b8c0f40ULL,
    0x98cf808d20a60ULL,
    0x92bec6ab50770ULL,
    0xf677eaa558658ULL,
    0x36d9bd4d7cd3cULL,
    0x11a2648b86452ULL,
    0x35ce793677e49ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0xf400000000000ULL,
    0x2600000000000ULL,
    0x6100000000000ULL,
    0x1780000000000ULL,
    0x8c0000000000ULL,
    0xbb20000000000ULL,
    0x4b0000000000ULL,
    0xe858000000000ULL,
    0x5d54000000000ULL,
    0x1cc2000000000ULL,
    0x8d35000000000ULL,
    0x4d95800000000ULL,
    0xdbe6400000000ULL,
    0x3bbee00000000ULL,
    0x32d4b00000000ULL,
    0xb83a080000000ULL,
    0xcc110c0000000ULL,
    0x2a2f860000000ULL,
    0x2b374d0000000ULL,
    0xecb1648000000ULL,
    0xac53ff4000000ULL,
    0xe3536a6000000ULL,
    0xc1d6fa1000000ULL,
    0x489eef7800000ULL,
    0x264b18c00000ULL,
    0x1662013200000ULL,
    0x20450e0b00000ULL,
    0x696d8e6580000ULL,
    0x5ac2418540000ULL,
    0x3c04e65420000ULL,
    0x1205b10f50000ULL,
    0x2735847358000ULL,
    0xa696476964000ULL,
    0x5746e1f46e000ULL,
    0x47f0bb3f0b000ULL,
    0x998003f800800ULL,
    0x8dc002cc00c00ULL,
    0xa2a00d5200600ULL,
    0x4970091b00d00ULL,
    0x2af80d9d80480ULL,
    0xc424034940f40ULL,
    0x7e3a0b0620260ULL,
    0xbd11081450610ULL,
    0x15af81eed8178ULL,
    0x7f748202408cULL,
    0x39916cf24ebb2ULL,
    0x3de3fe2b5b04bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x7800000000000ULL,
    0x7400000000000ULL,
    0x7e00000000000ULL,
    0x5f00000000000ULL,
    0xd080000000000ULL,
    0x7540000000000ULL,
    0x7d20000000000ULL,
    0x2d90000000000ULL,
    0x18f8000000000ULL,
    0x85fc000000000ULL,
    0xd86e000000000ULL,
    0xb895000000000ULL,
    0x496b800000000ULL,
    0xef0dc00000000ULL,
    0x8bb200000000ULL,
    0x9179d00000000ULL,
    0xb36080000000ULL,
    0x7eb9040000000ULL,
    0xc25d8e0000000ULL,
    0xd1b4c70000000ULL,
    0x2ae6a78000000ULL,
    0x30cd174000000ULL,
    0x59d0afe000000ULL,
    0x3a7411f000000ULL,
    0xe58d2b0800000ULL,
    0xb4c0d45400000ULL,
    0x1feb865200000ULL,
    0xf14dc69900000ULL,
    0x3b1b2fef80000ULL,
    0xe6a9ddefc0000ULL,
    0xc66e08eee0000ULL,
    0xd7950a6d50000ULL,
    0x1eb87acb8000ULL,
    0x9e4dc98ddc000ULL,
    0x739b2c5e32000ULL,
    0x97e9d3855d000ULL,
    0xbd4e048800800ULL,
    0xd105071400400ULL,
    0xb79387b200e00ULL,
    0x31f1c06900700ULL,
    0x6552b6780780ULL,
    0x77acdafbc0740ULL,
    0x91fd8f5ce07e0ULL,
    0x9664ca04505f0ULL,
    0x7fbeaccb38d08ULL,
    0x9de113761c754ULL,
    0x9c66a302d27d2ULL,
    0x5e8d19810d2d9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0xc800000000000ULL,
    0x2400000000000ULL,
    0x8e00000000000ULL,
    0x3900000000000ULL,
    0x6a80000000000ULL,
    0x6040000000000ULL,
    0x5aa0000000000ULL,
    0xf870000000000ULL,
    0x96a8000000000ULL,
    0xc254000000000ULL,
    0xe99a000000000ULL,
    0xb5dd000000000ULL,
    0x6d79800000000ULL,
    0xb633400000000ULL,
    0xa533200000000ULL,
    0xb8b3500000000ULL,
    0xab79880000000ULL,
    0x6b334c0000000ULL,
    0x61b3220000000ULL,
    0x71f3590000000ULL,
    0x5359848000000ULL,
    0xd7034e4000000ULL,
    0x23bb2ae000000ULL,
    0x72d75a9000000ULL,
    0x46eb822800000ULL,
    0xc0ca484400000ULL,
    0xfdf8af4a00000ULL,
    0x8949151700000ULL,
    0x18092b4280000ULL,
    0xc1e546140000ULL,
    0x1a2809d3a0000ULL,
    0xef14024ad0000ULL,
    0xbfba079518000ULL,
    0xa0ed0a0274000ULL,
    0x8df1850092000ULL,
    0xf1574e81e5000ULL,
    0x64212b4280800ULL,
    0x6e0a546140c00ULL,
    0x891209d3a0200ULL,
    0x32b9024ad0900ULL,
    0x8c6b879518c80ULL,
    0xf08a4a0274240ULL,
    0x4f58a500928e0ULL,
    0xc5391e81e5390ULL,
    0xc8a1234280ea8ULL,
    0xd34a586140a04ULL,
    0x17320bd3a07aaULL,
    0x3890b4ad0687ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0xbc00000000000ULL,
    0xca00000000000ULL,
    0x3900000000000ULL,
    0x1380000000000ULL,
    0x5540000000000ULL,
    0xbba0000000000ULL,
    0xd170000000000ULL,
    0x6d88000000000ULL,
    0xf244000000000ULL,
    0xbf36000000000ULL,
    0x8ab000000000ULL,
    0x9be4800000000ULL,
    0x5b75400000000ULL,
    0x3498600000000ULL,
    0x91ec100000000ULL,
    0xc264880000000ULL,
    0xf7354c0000000ULL,
    0x3cb86a0000000ULL,
    0xc5dc1d0000000ULL,
    0xec4c878000000ULL,
    0x680147c000000ULL,
    0x240666a000000ULL,
    0x6331e9000000ULL,
    0xdb1e06b800000ULL,
    0x6e9f029400000ULL,
    0x30da8d1a00000ULL,
    0x1dda438700000ULL,
    0x406ae86080000ULL,
    0xfa0251b040000ULL,
    0x713064e960000ULL,
    0x2798120db0000ULL,
    0xb7a8c5ec8000ULL,
    0xcaa4ec714000ULL,
    0x8de2e680e6000ULL,
    0xd846524371000ULL,
    0x3606686080800ULL,
    0x933311b040c00ULL,
    0x5a9e04e960a00ULL,
    0x6edf020db0d00ULL,
    0xaafa845ec8f80ULL,
    0x1cea42c714bc0ULL,
    0x4fc2ec80e6ca0ULL,
    0xb5765f4371390ULL,
    0xbae67e080938ULL,
    0x59471a7040954ULL,
    0xf9800849601baULL,
    0x7c40019db0017ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xf800000000000ULL,
    0xe400000000000ULL,
    0xfa00000000000ULL,
    0xad00000000000ULL,
    0xb680000000000ULL,
    0x89c0000000000ULL,
    0x92a0000000000ULL,
    0x53d0000000000ULL,
    0x6fb8000000000ULL,
    0x2d5c000000000ULL,
    0xfa46000000000ULL,
    0xa1c5000000000ULL,
    0xfea8800000000ULL,
    0xd5d6400000000ULL,
    0xec99200000000ULL,
    0x34d2300000000ULL,
    0x8c08880000000ULL,
    0xf606440000000ULL,
    0x1b21260000000ULL,
    0xcd8e330000000ULL,
    0x744e878000000ULL,
    0x1ec34a4000000ULL,
    0xa909a9a000000ULL,
    0x3c9879d000000ULL,
    0xbcf7ace800000ULL,
    0xf00172dc00000ULL,
    0xd819288a00000ULL,
    0xb41238ed00000ULL,
    0x32288c1380000ULL,
    0xb1164309c0000ULL,
    0xa8b920ae60000ULL,
    0xdec238b150000ULL,
    0x89108a5908000ULL,
    0x6c8a4784a4000ULL,
    0x74df2a8ff2000ULL,
    0xec173d2073000ULL,
    0xc620041380800ULL,
    0xe3100709c0400ULL,
    0x299806ae60600ULL,
    0x8e4c0bb150300ULL,
    0xb3de0dd908f80ULL,
    0x1f890dc4a4e40ULL,
    0xb576832ff2fa0ULL,
    0x2e5f44f073ad0ULL,
    0xa3efa8fb80368ULL,
    0xb78d75d5c0c9cULL,
    0x99672e2460f2aULL,
    0xc84b335c5063dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0xe400000000000ULL,
    0x8600000000000ULL,
    0x9d00000000000ULL,
    0xe180000000000ULL,
    0xb0c0000000000ULL,
    0xeda0000000000ULL,
    0x12f0000000000ULL,
    0x1698000000000ULL,
    0x7e74000000000ULL,
    0x2fc2000000000ULL,
    0xc72d000000000ULL,
    0x56b3800000000ULL,
    0x5e62400000000ULL,
    0xdfe7e00000000ULL,
    0xbf38700000000ULL,
    0xda93880000000ULL,
    0x3c524c0000000ULL,
    0xc4dfee0000000ULL,
    0xc3bc710000000ULL,
    0x8bc98e8000000ULL,
    0x610b424000000ULL,
    0x3bae666000000ULL,
    0xc7f338d000000ULL,
    0xe31de09800000ULL,
    0x3091794c00000ULL,
    0xd37a00ba00000ULL,
    0x566905ff00000ULL,
    0xebc987f180000ULL,
    0xb10b43ab40000ULL,
    0xb3ae62c620000ULL,
    0x33f33acdd0000ULL,
    0xd1dea7ab8000ULL,
    0x49917cdd64000ULL,
    0xb4fa09c05e000ULL,
    0x7ba900a257000ULL,
    0xe7e98ff180800ULL,
    0x133b4fab40c00ULL,
    0x48966cc620e00ULL,
    0x5f773bcdd0100ULL,
    0x3447ecfab8680ULL,
    0xf0c8729d64e40ULL,
    0xcd8b81a05e860ULL,
    0xe2e64972579d0ULL,
    0x6ebde1e980618ULL,
    0xf26174a74070cULL,
    0x4de20a1c200daULL,
    0xdc1d06e2d002fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0x3400000000000ULL,
    0xd200000000000ULL,
    0x5900000000000ULL,
    0xd680000000000ULL,
    0xf140000000000ULL,
    0x9aa0000000000ULL,
    0x8f50000000000ULL,
    0xada8000000000ULL,
    0x96cc000000000ULL,
    0xa942000000000ULL,
    0x46a1000000000ULL,
    0x4946800000000ULL,
    0x56af400000000ULL,
    0xb167200000000ULL,
    0xbaa5100000000ULL,
    0xff66880000000ULL,
    0x5bf440000000ULL,
    0xa2ef220000000ULL,
    0x7b79170000000ULL,
    0x1fac828000000ULL,
    0x9fc2474000000ULL,
    0xa7e3af2000000ULL,
    0x2beb529000000ULL,
    0x35e72fe800000ULL,
    0x52e5185400000ULL,
    0x93468e8a00000ULL,
    0xbaf4e6500000ULL,
    0x3de72f3280000ULL,
    0x56e51238c0000ULL,
    0xc9468c9e20000ULL,
    0x16af4e4f10000ULL,
    0x91672c06e8000ULL,
    0xcaa5118234000ULL,
    0x576687c052000ULL,
    0x31bf4f6141000ULL,
    0x70ef273280800ULL,
    0x22791638c0400ULL,
    0xc92c8e9e20200ULL,
    0x6e82494f10700ULL,
    0x3d43a686e8a80ULL,
    0xa4bb52c234340ULL,
    0x984f2ae052d20ULL,
    0xc4291af141590ULL,
    0x3a04825a80568ULL,
    0x4d0e4d2cc0b14ULL,
    0x74a1ad3420baaULL,
    0x4a52ba10ff5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x8800000000000ULL,
    0xcc00000000000ULL,
    0x5a00000000000ULL,
    0x7700000000000ULL,
    0x4e80000000000ULL,
    0x2340000000000ULL,
    0xd4a0000000000ULL,
    0xb450000000000ULL,
    0xaa08000000000ULL,
    0x8f34000000000ULL,
    0x3a8a000000000ULL,
    0xad57000000000ULL,
    0xbd94800000000ULL,
    0x1bfec00000000ULL,
    0xeacd200000000ULL,
    0x4141100000000ULL,
    0x379c880000000ULL,
    0x44cac40000000ULL,
    0xf8472a0000000ULL,
    0xb016190000000ULL,
    0x5808008000000ULL,
    0xe43408c000000ULL,
    0x60a0fa000000ULL,
    0xa5170e7000000ULL,
    0xf5b48c6800000ULL,
    0x37eecef400000ULL,
    0x80e528ea00000ULL,
    0x2e651c3500000ULL,
    0x3d3e864880000ULL,
    0xf1b9cec740000ULL,
    0x1f1a4e2a0000ULL,
    0x3ddbd89070000ULL,
    0x9fd3a1f9c8000ULL,
    0x9ce8d18cac000ULL,
    0x5c452aa472000ULL,
    0x163514b161000ULL,
    0x6d3686c880800ULL,
    0x998dc60740400ULL,
    0xfdfbab42a0a00ULL,
    0x7fccd6e070900ULL,
    0xace72d91c8880ULL,
    0x44461f78accc0ULL,
    0x5200024e725a0ULL,
    0xfb00088461770ULL,
    0xb480008000ce8ULL,
    0xc44008c000634ULL,
    0x12200fa00074aULL,
    0x5b100e7000245ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0xb400000000000ULL,
    0xb600000000000ULL,
    0x900000000000ULL,
    0x4080000000000ULL,
    0xb940000000000ULL,
    0x3ea0000000000ULL,
    0x5470000000000ULL,
    0x3018000000000ULL,
    0x482c000000000ULL,
    0x2422000000000ULL,
    0xae31000000000ULL,
    0xd537800000000ULL,
    0x42af400000000ULL,
    0x67da00000000ULL,
    0x770c100000000ULL,
    0xadaf880000000ULL,
    0xb7c34c0000000ULL,
    0x22ffae0000000ULL,
    0x404d170000000ULL,
    0xd6000e8000000ULL,
    0xb900074000000ULL,
    0xc880056000000ULL,
    0x7d40079000000ULL,
    0xe0a0028800000ULL,
    0xe97000d400000ULL,
    0xc698088a00000ULL,
    0xf86c05d700000ULL,
    0x5a02070980000ULL,
    0x43010f16c0000ULL,
    0xdb8f81a820000ULL,
    0x5ef34fa410000ULL,
    0x1247ae52f8000ULL,
    0x911110a834000ULL,
    0x5cba0a25fa000ULL,
    0x5b5d0193d1000ULL,
    0xf1b5898980800ULL,
    0x75ee4856c0c00ULL,
    0x7dd224c820e00ULL,
    0x79cf583410700ULL,
    0x47d024daf8680ULL,
    0x8ace5c7c34b40ULL,
    0x145facaffab60ULL,
    0x103d1344d1090ULL,
    0xd818000000c08ULL,
    0x3c2c000000794ULL,
    0x7222000000deaULL,
    0xd731000000247ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0xe400000000000ULL,
    0x8600000000000ULL,
    0x8d00000000000ULL,
    0x4b80000000000ULL,
    0x8ec0000000000ULL,
    0x79a0000000000ULL,
    0x1df0000000000ULL,
    0xab18000000000ULL,
    0xf6b4000000000ULL,
    0x8d56000000000ULL,
    0xbb5d000000000ULL,
    0xbe5f800000000ULL,
    0x59f2400000000ULL,
    0x1d1f600000000ULL,
    0x33a2f00000000ULL,
    0x8ae7880000000ULL,
    0xefb64c0000000ULL,
    0x28d16e0000000ULL,
    0x48bf10000000ULL,
    0xfe4e038000000ULL,
    0x79e9024000000ULL,
    0xed09866000000ULL,
    0x9baf49d000000ULL,
    0xd6c0ef3800000ULL,
    0x8d90b6ac00000ULL,
    0x23d8e7fa00000ULL,
    0xc224b50f00000ULL,
    0x3b0ee80980000ULL,
    0x8eb9ba8740000ULL,
    0x897162cf60000ULL,
    0x2d7bf92ad0000ULL,
    0x8b5604b478000ULL,
    0xf65d016464000ULL,
    0x15df805c96000ULL,
    0xc73247f3ff000ULL,
    0xdcbf638980800ULL,
    0xca52f4c740c00ULL,
    0xa7ff8aaf60e00ULL,
    0x940241fad0100ULL,
    0xee07680c78b80ULL,
    0x3116f58864e40ULL,
    0x39b181c696860ULL,
    0x3deb4b2cff8d0ULL,
    0x5b0eecb800cb8ULL,
    0x5eb9b4ec004ecULL,
    0xd171619a0099aULL,
    0xd97bfcdf000dfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x8800000000000ULL,
    0xbc00000000000ULL,
    0xba00000000000ULL,
    0x6b00000000000ULL,
    0xbb80000000000ULL,
    0x2740000000000ULL,
    0x30a0000000000ULL,
    0x6cd0000000000ULL,
    0xfff8000000000ULL,
    0x505c000000000ULL,
    0xa10a000000000ULL,
    0x788b000000000ULL,
    0xf0f8800000000ULL,
    0x95dbc00000000ULL,
    0x37d600000000ULL,
    0x2eba900000000ULL,
    0x59f2880000000ULL,
    0x1150c40000000ULL,
    0x2985e60000000ULL,
    0x60615b0000000ULL,
    0x690fe08000000ULL,
    0xa4aa5fc000000ULL,
    0xfad765a000000ULL,
    0x76e199b000000ULL,
    0x4f20d3800000ULL,
    0xb3d706b400000ULL,
    0x272862a00000ULL,
    0xc610c0bd00000ULL,
    0xf125e76780000ULL,
    0xb15a01c0000ULL,
    0xa4f7e102a0000ULL,
    0x23f65181b0000ULL,
    0x5a5d664208000ULL,
    0x422a97e17c000ULL,
    0x7f2a8052f6000ULL,
    0x6d9cc26bd9000ULL,
    0xce57ea5f80800ULL,
    0xd4265cb5c0400ULL,
    0xf6256728a0600ULL,
    0x3936913cb0b00ULL,
    0xdc80812588880ULL,
    0xaec7cde0bcbc0ULL,
    0xc0d7615056ba0ULL,
    0x5de193ea696b0ULL,
    0xdf720c1d883b8ULL,
    0x24970b54bc674ULL,
    0xbad2877a5650aULL,
    0x16c0c35769dcdULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x5400000000000ULL,
    0x5600000000000ULL,
    0x3300000000000ULL,
    0x5480000000000ULL,
    0xe4c0000000000ULL,
    0x17a0000000000ULL,
    0x1870000000000ULL,
    0xcf78000000000ULL,
    0x5c4000000000ULL,
    0xbe1e000000000ULL,
    0xaf29000000000ULL,
    0x6e8f800000000ULL,
    0x85dbc00000000ULL,
    0x7e23a00000000ULL,
    0xcf05700000000ULL,
    0x3e91880000000ULL,
    0xddf2cc0000000ULL,
    0x2a2c260000000ULL,
    0x991eb50000000ULL,
    0xd922d8000000ULL,
    0x8947b94000000ULL,
    0xcee5ab6000000ULL,
    0x8e987a3000000ULL,
    0x15f80ec800000ULL,
    0x4604020c00000ULL,
    0xcb3e079a00000ULL,
    0x309902f700000ULL,
    0xbad7895f80000ULL,
    0x28afc86040000ULL,
    0x4ee5aab3e0000ULL,
    0x4e98760990000ULL,
    0x75f8062d78000ULL,
    0x16040bcafc000ULL,
    0x933e078e5a000ULL,
    0x649902d987000ULL,
    0xecd7879780800ULL,
    0x1bafca6c40c00ULL,
    0x1a65ad29e0600ULL,
    0xaa5874fe90500ULL,
    0x62580f72f8580ULL,
    0xe7403aabc540ULL,
    0x5c460d3dba560ULL,
    0x615d04d017330ULL,
    0x52c989baf8d48ULL,
    0xb486cda6bc24cULL,
    0x74ea2ca7ba77aULL,
    0x2f83b32717487ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x7800000000000ULL,
    0xdc00000000000ULL,
    0xca00000000000ULL,
    0x4300000000000ULL,
    0xe380000000000ULL,
    0x9c40000000000ULL,
    0xb8a0000000000ULL,
    0x73d0000000000ULL,
    0x6c8000000000ULL,
    0x1c7c000000000ULL,
    0xf886000000000ULL,
    0xd3c3000000000ULL,
    0x36e8800000000ULL,
    0x6445c00000000ULL,
    0x24bb600000000ULL,
    0x19c6500000000ULL,
    0x75ee880000000ULL,
    0x87c6c40000000ULL,
    0xb8f3ea0000000ULL,
    0xa153930000000ULL,
    0x61def8000000ULL,
    0x813c99c000000ULL,
    0xa4bb6ea000000ULL,
    0x59c6533000000ULL,
    0xd5ee8bb800000ULL,
    0xb7c6c30400000ULL,
    0xc0f3eaaa00000ULL,
    0x7d539dcd00000ULL,
    0xcc1ded7480000ULL,
    0xc23c95f3c0000ULL,
    0x473b649a60000ULL,
    0xc58650f530000ULL,
    0x6d4e833008000ULL,
    0xc416c3ba9c000ULL,
    0xc63be70556000ULL,
    0x612f90aa95000ULL,
    0x349be6cc80800ULL,
    0x11ff96f7c0400ULL,
    0x71d3ee3060a00ULL,
    0xa1c39d3830300ULL,
    0x49f5e64488780ULL,
    0xddd092495cdc0ULL,
    0xb3d5699f36ca0ULL,
    0xe6e9535fa5430ULL,
    0x8c680a7c88638ULL,
    0xb0ac0c8d5cdc4ULL,
    0x77ce07953618aULL,
    0x20ff0ea2a543dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0x5c00000000000ULL,
    0x3a00000000000ULL,
    0x2f00000000000ULL,
    0xac80000000000ULL,
    0x94c0000000000ULL,
    0x5fa0000000000ULL,
    0xc270000000000ULL,
    0x4448000000000ULL,
    0xa174000000000ULL,
    0x1afa000000000ULL,
    0xe68b000000000ULL,
    0x43f0800000000ULL,
    0x9732c00000000ULL,
    0xa8a4a00000000ULL,
    0x5add700000000ULL,
    0x86aa880000000ULL,
    0x73c9cc0000000ULL,
    0xf1c2a0000000ULL,
    0xfc9bb90000000ULL,
    0x3cf4288000000ULL,
    0x939fb9c000000ULL,
    0xf04621a000000ULL,
    0x3760b7f000000ULL,
    0x37cca84800000ULL,
    0xa119798c00000ULL,
    0x15b884da00000ULL,
    0x1546ce1700000ULL,
    0x8cdea7ac80000ULL,
    0xcb967d6b40000ULL,
    0x47a05bda0000ULL,
    0xc14b0033b0000ULL,
    0x2ad086e988000ULL,
    0x7e82cdff6c000ULL,
    0x17cca0d36a000ULL,
    0xf119717927000ULL,
    0xbdb887e480800ULL,
    0xd946c8e740c00ULL,
    0xbedeab67a0a00ULL,
    0xb8967724b0900ULL,
    0x92fa09c508080ULL,
    0x7a8b09542c5c0ULL,
    0xd9f084ceca3a0ULL,
    0x2832c6ba972f0ULL,
    0xc24a945082c8ULL,
    0x921d7c942c54cULL,
    0xe30a8f6ecaffaULL,
    0x9eb9c84a97527ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x2800000000000ULL,
    0xc400000000000ULL,
    0x3a00000000000ULL,
    0x9b00000000000ULL,
    0xa180000000000ULL,
    0x9340000000000ULL,
    0xa0a0000000000ULL,
    0xf9f0000000000ULL,
    0x2a58000000000ULL,
    0x560c000000000ULL,
    0xa502000000000ULL,
    0xe095000000000ULL,
    0xd9d8800000000ULL,
    0xba7dc00000000ULL,
    0xe07e00000000ULL,
    0x4904900000000ULL,
    0x1e82880000000ULL,
    0x78e4cc0000000ULL,
    0x80dd6e0000000ULL,
    0x3cec570000000ULL,
    0x7addea8000000ULL,
    0x47dd904000000ULL,
    0xab7805a000000ULL,
    0xfcbc02b000000ULL,
    0xcffa069800000ULL,
    0x3f69027400000ULL,
    0x7e828b2a00000ULL,
    0xc8e4ca6f00000ULL,
    0x48dd6b1d80000ULL,
    0x88ec55e4c0000ULL,
    0x68dde24220000ULL,
    0x18dd94a250000ULL,
    0x30f8033208000ULL,
    0xf4fc0a581c000ULL,
    0xceda0495de000ULL,
    0x55d904b9d9000ULL,
    0xf47a870580800ULL,
    0x6718c7d0c0c00ULL,
    0xc7876cc820e00ULL,
    0x3e755c7d50700ULL,
    0x14076eb788280ULL,
    0x42355dc8dcc40ULL,
    0xe7276dfdfe3a0ULL,
    0x7855a74899b0ULL,
    0xde5f6f2a08218ULL,
    0x6439586c1c534ULL,
    0x6a256a1fde40aULL,
    0x23105c66d989fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x2800000000000ULL,
    0xa400000000000ULL,
    0xfe00000000000ULL,
    0x3d00000000000ULL,
    0x8280000000000ULL,
    0xb340000000000ULL,
    0x5a0000000000ULL,
    0x42f0000000000ULL,
    0x4178000000000ULL,
    0xa28c000000000ULL,
    0x6362000000000ULL,
    0x3d8d000000000ULL,
    0xbed9800000000ULL,
    0x3354400000000ULL,
    0xc5ba200000000ULL,
    0x22fe100000000ULL,
    0x3163880000000ULL,
    0x8aa54c0000000ULL,
    0xc779a60000000ULL,
    0xc3ab570000000ULL,
    0x83e22a8000000ULL,
    0xb1c2164000000ULL,
    0x76d981e000000ULL,
    0x275448d000000ULL,
    0x73ba2ca800000ULL,
    0xcbfe167400000ULL,
    0x65e3853a00000ULL,
    0xa0e541bf00000ULL,
    0xbe59a5df80000ULL,
    0xf1b55ccc0000ULL,
    0x45ba24c420000ULL,
    0xe2fe1a83d0000ULL,
    0x51638e4018000ULL,
    0xfaa545e284000ULL,
    0xef79a2d202000ULL,
    0x67ab5dabf1000ULL,
    0x7de22bf780800ULL,
    0x8cc219f8c0c00ULL,
    0xf459861e20600ULL,
    0x941444ecd0700ULL,
    0x761a2db798280ULL,
    0x890e101a44a40ULL,
    0x249b82cc22fe0ULL,
    0x2694e47213d0ULL,
    0xdd3bacc018028ULL,
    0x32965fa284734ULL,
    0xfb63a5320265aULL,
    0xd1aa527bf132fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0x5c00000000000ULL,
    0xea00000000000ULL,
    0x4300000000000ULL,
    0x4080000000000ULL,
    0xfc40000000000ULL,
    0x9da0000000000ULL,
    0x19f0000000000ULL,
    0x9058000000000ULL,
    0xdf8c000000000ULL,
    0x96ea000000000ULL,
    0xc2d3000000000ULL,
    0xb4d4800000000ULL,
    0xfdcdc00000000ULL,
    0x8e49a00000000ULL,
    0xba83500000000ULL,
    0x8746880000000ULL,
    0xe922cc0000000ULL,
    0x238f2e0000000ULL,
    0x2ce1970000000ULL,
    0x99e9a08000000ULL,
    0x507359c000000ULL,
    0x3f9e88a000000ULL,
    0xe6eecf3000000ULL,
    0x4ac5228800000ULL,
    0xe8c29d0400000ULL,
    0x17e527fa00000ULL,
    0xcd729c6f00000ULL,
    0xfa1d25ad80000ULL,
    0x7b0e9d0cc0000ULL,
    0x74af2a3ca0000ULL,
    0x3a519cb630000ULL,
    0xbc91abb2c8000ULL,
    0x464f594b1c000ULL,
    0xc68c8ca7ba000ULL,
    0xfd41cf49c5000ULL,
    0x5223afa580800ULL,
    0xb71055c8c0c00ULL,
    0x66b20b66a0e00ULL,
    0xad5f08e930700ULL,
    0x4a3e841748880ULL,
    0x131ecc83dc5c0ULL,
    0x589d27c11aea0ULL,
    0x584e97a0f5430ULL,
    0xa38f23b2c8c08ULL,
    0xece1954b1c3c4ULL,
    0x79e9a2a7ba7daULL,
    0x20735849c569fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0x9c00000000000ULL,
    0x2a00000000000ULL,
    0xcf00000000000ULL,
    0x3d80000000000ULL,
    0x4c40000000000ULL,
    0xca0000000000ULL,
    0x70d0000000000ULL,
    0xb478000000000ULL,
    0x6084000000000ULL,
    0xb2de000000000ULL,
    0x6f6b000000000ULL,
    0xb318800000000ULL,
    0x878ac00000000ULL,
    0xbb4ee00000000ULL,
    0x7d28500000000ULL,
    0x4e9e880000000ULL,
    0xebf5c40000000ULL,
    0x7d0660000000ULL,
    0x85dd9d0000000ULL,
    0x3eceed8000000ULL,
    0x3d685dc000000ULL,
    0x503e8ca000000ULL,
    0x1825c5f000000ULL,
    0xfc28605800000ULL,
    0xfa19900400000ULL,
    0x9730efea00000ULL,
    0xa193563d00000ULL,
    0x667e043f80000ULL,
    0xc3bb073c40000ULL,
    0x4d608bbfe0000ULL,
    0xf80ec1ffb0000ULL,
    0x6c10e79c08000ULL,
    0xc2035c4dec000ULL,
    0xdb2602a68e000ULL,
    0xd3af0b1875000ULL,
    0x356681e780800ULL,
    0xd431cef840400ULL,
    0xce2e6ef5e0600ULL,
    0xc9269f32b0d00ULL,
    0x508e667b88580ULL,
    0x7af692b5ac9c0ULL,
    0xcb766c536e2a0ULL,
    0xd532942ac5cf0ULL,
    0x3a88679c08bd8ULL,
    0xd5c99c4dec0c4ULL,
    0x26c8e2a68e6caULL,
    0xc1575b1875a0dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0xc00000000000ULL,
    0xbe00000000000ULL,
    0x2f00000000000ULL,
    0x7380000000000ULL,
    0x1240000000000ULL,
    0x3a60000000000ULL,
    0x1670000000000ULL,
    0x5868000000000ULL,
    0xf4c000000000ULL,
    0x98f6000000000ULL,
    0x38b5000000000ULL,
    0x93fa800000000ULL,
    0xd901400000000ULL,
    0xe0b5a00000000ULL,
    0xc7dd300000000ULL,
    0x7335a80000000ULL,
    0x59d340000000ULL,
    0xc155ae0000000ULL,
    0x8fed3d0000000ULL,
    0xcf3da08000000ULL,
    0xa3a13dc000000ULL,
    0x9a4bab6000000ULL,
    0xa6543f3000000ULL,
    0x40512c5800000ULL,
    0x7b657e1400000ULL,
    0xc2ec8ffe00000ULL,
    0xa5844f7300000ULL,
    0x7147227880000ULL,
    0x97e07b87c0000ULL,
    0xbb1e05f7e0000ULL,
    0xf9b9050c90000ULL,
    0x76c8c4848000ULL,
    0x44c4455c84000ULL,
    0xea7212312000ULL,
    0x20d0721157000ULL,
    0xfc960c4848800ULL,
    0xd1c5055c84400ULL,
    0x3012812312600ULL,
    0x580d421157900ULL,
    0x1423a44848680ULL,
    0xca18315c844c0ULL,
    0x75272f2312de0ULL,
    0xee907f1157bf0ULL,
    0xf0f604c8481b8ULL,
    0x74b50c9c845e4ULL,
    0x4dfa844312e46ULL,
    0x6601402157a97ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0x7400000000000ULL,
    0xea00000000000ULL,
    0x1500000000000ULL,
    0xdc80000000000ULL,
    0x40c0000000000ULL,
    0xe8e0000000000ULL,
    0xccd0000000000ULL,
    0xed8000000000ULL,
    0xfffc000000000ULL,
    0xc17e000000000ULL,
    0xe0a3000000000ULL,
    0x3ef9800000000ULL,
    0xa7ecc00000000ULL,
    0x9d63e00000000ULL,
    0xee9fd00000000ULL,
    0xc9e3e80000000ULL,
    0x1a5fdc0000000ULL,
    0xeb03ea0000000ULL,
    0x538fd50000000ULL,
    0x315bea8000000ULL,
    0x98b3d24000000ULL,
    0xf2c5e42000000ULL,
    0xa1c0d31000000ULL,
    0x1e6469e800000ULL,
    0xb910171c00000ULL,
    0xaa99876600000ULL,
    0x7bfccbd100000ULL,
    0x535bef8b80000ULL,
    0x39b3d82ec0000ULL,
    0xe445e99c60000ULL,
    0x6400d324f0000ULL,
    0x220468f3f8000ULL,
    0x4100171a3c000ULL,
    0xa6a18f05c6000ULL,
    0x5dd0cfe3c1000ULL,
    0x407de8f3f8800ULL,
    0x662cd71a3cc00ULL,
    0xf3226f05c6200ULL,
    0xef9f1fe3c1900ULL,
    0x8f4600f3f8880ULL,
    0xf78f0b1a3cb40ULL,
    0x335f8505c6ca0ULL,
    0x49b3cae3c1850ULL,
    0x5c646a73f8548ULL,
    0x8810195a3cf4cULL,
    0xb4198125c622eULL,
    0xca3cc9f3c149dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xa800000000000ULL,
    0x9c00000000000ULL,
    0xfa00000000000ULL,
    0xd300000000000ULL,
    0xdb80000000000ULL,
    0xb4c0000000000ULL,
    0x88e0000000000ULL,
    0x62d0000000000ULL,
    0x39d8000000000ULL,
    0xae6c000000000ULL,
    0x4886000000000ULL,
    0x8f41000000000ULL,
    0xcc2f800000000ULL,
    0x9203400000000ULL,
    0x6f39a00000000ULL,
    0xf182900000000ULL,
    0xcfd9a80000000ULL,
    0xcf529c0000000ULL,
    0x2c01aa0000000ULL,
    0x623e910000000ULL,
    0x1707a08000000ULL,
    0xc5bf98c000000ULL,
    0xa9c82f2000000ULL,
    0xe66cd5f000000ULL,
    0x24a9829800000ULL,
    0xd424ebc00000ULL,
    0x2b16221600000ULL,
    0x2f81d49100000ULL,
    0xf2e00b8b80000ULL,
    0x71d00f77c0000ULL,
    0xc2580f83e0000ULL,
    0xcaac0f43d0000ULL,
    0x6866046118000ULL,
    0x71910f93e4000ULL,
    0xff78a0a82000ULL,
    0xef6f4f37cd000ULL,
    0xfc3fac6118800ULL,
    0xca039393e4c00ULL,
    0x8b16200a82200ULL,
    0x3f81de37cdd00ULL,
    0x7ae00ce118280ULL,
    0x3dd00b53e45c0ULL,
    0x90580f2a82da0ULL,
    0x85ac0bc7cd030ULL,
    0x49e60e7918f38ULL,
    0x165105efe4e8cULL,
    0x5c978d3c8252eULL,
    0x397f4f56cd61dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x8800000000000ULL,
    0xfc00000000000ULL,
    0xe600000000000ULL,
    0x5900000000000ULL,
    0x1880000000000ULL,
    0xc140000000000ULL,
    0x9560000000000ULL,
    0x1f70000000000ULL,
    0xf848000000000ULL,
    0xfdc4000000000ULL,
    0xb626000000000ULL,
    0xa139000000000ULL,
    0x8c90800000000ULL,
    0xeb7ec00000000ULL,
    0xa254e00000000ULL,
    0xa2d6b00000000ULL,
    0xc7b4e80000000ULL,
    0xf0e6b40000000ULL,
    0x249ce60000000ULL,
    0x8752b70000000ULL,
    0x1c72ee8000000ULL,
    0xbfefb8c000000ULL,
    0x4d2460e000000ULL,
    0x72987d5000000ULL,
    0x1648016800000ULL,
    0x18c4014400000ULL,
    0xa8a6003e00000ULL,
    0x97904b300000ULL,
    0x897081ba80000ULL,
    0xc94ec86f40000ULL,
    0x297ce258e0000ULL,
    0x1962bdbcd0000ULL,
    0x915aecf1e8000ULL,
    0x6d5bb29b3c000ULL,
    0x8b4a695ca6000ULL,
    0xd26577a257000ULL,
    0xcafe84f1e8800ULL,
    0xb83c69b3c400ULL,
    0x9ee26f5ca6e00ULL,
    0x819170a257300ULL,
    0x79f08271e8080ULL,
    0x840eca5b3cbc0ULL,
    0x321ce9bca6060ULL,
    0x9312baf257690ULL,
    0x1f92ed99e8108ULL,
    0xf4dfb3df3c7d4ULL,
    0x568c6962a6936ULL,
    0xf46c731157767ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0x8400000000000ULL,
    0xca00000000000ULL,
    0x6b00000000000ULL,
    0x1f80000000000ULL,
    0x1240000000000ULL,
    0x9860000000000ULL,
    0xd350000000000ULL,
    0xfce8000000000ULL,
    0x669c000000000ULL,
    0x22ea000000000ULL,
    0xbf9b000000000ULL,
    0xe241800000000ULL,
    0xf047400000000ULL,
    0xdf6aa00000000ULL,
    0xead5300000000ULL,
    0x4382a80000000ULL,
    0x9c493c0000000ULL,
    0xd968a60000000ULL,
    0x77d2310000000ULL,
    0xa929238000000ULL,
    0x6895794000000ULL,
    0xa3c38f2000000ULL,
    0xfb004ff000000ULL,
    0x67a126d800000ULL,
    0xa65972d400000ULL,
    0xda41895e00000ULL,
    0x64474ee100000ULL,
    0xad6aa51080000ULL,
    0xd5d531c8c0000ULL,
    0xce02a81e20000ULL,
    0x610935c170000ULL,
    0x9488a0e238000ULL,
    0xddc2381104000ULL,
    0xd2212c4a92000ULL,
    0xcf19745d57000ULL,
    0xe5a1836238800ULL,
    0xf157415104c00ULL,
    0x5be2a36a92e00ULL,
    0x8f193bad57d00ULL,
    0xc580a5ba38d80ULL,
    0xc14e338504440ULL,
    0xd3c32a34922a0ULL,
    0x530e754c57bb0ULL,
    0x8b8200aab8c78ULL,
    0x6047024dc4564ULL,
    0xa74b822ab2b26ULL,
    0x5ecc408d27685ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x6800000000000ULL,
    0xe400000000000ULL,
    0x7e00000000000ULL,
    0x8700000000000ULL,
    0xda80000000000ULL,
    0x2c40000000000ULL,
    0x6260000000000ULL,
    0x3d70000000000ULL,
    0x9bd8000000000ULL,
    0x158c000000000ULL,
    0x2f6000000000ULL,
    0xf83f000000000ULL,
    0x6c38800000000ULL,
    0x6a35400000000ULL,
    0xc125e00000000ULL,
    0xd199f00000000ULL,
    0x6cfde80000000ULL,
    0xb715f40000000ULL,
    0x28be20000000ULL,
    0x506af10000000ULL,
    0xc053648000000ULL,
    0x486fbf4000000ULL,
    0x5c4e836000000ULL,
    0x1a4a473000000ULL,
    0x117d66c800000ULL,
    0x51dcb1f400000ULL,
    0x848002ee00000ULL,
    0xbb40032300000ULL,
    0x70e00fd380000ULL,
    0xa5300d3bc0000ULL,
    0xefb8049ce0000ULL,
    0x4bfc018830000ULL,
    0x3dae0e9768000ULL,
    0x46f303df64000ULL,
    0xd62e8eeb36000ULL,
    0x833a47a5fb000ULL,
    0x54a56a1768800ULL,
    0x9350bc9f64400ULL,
    0x34f60d8b36a00ULL,
    0x8b3f0095fb500ULL,
    0xb88cdf68e80ULL,
    0x75754d6b64a40ULL,
    0x6fc5ef6536de0ULL,
    0xa3a9f3b6fbd70ULL,
    0x31c5eb0ce8328ULL,
    0x34a9f450a4884ULL,
    0x2345e9f9d6bc6ULL,
    0xace9f33ecbea7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xd800000000000ULL,
    0xfc00000000000ULL,
    0xba00000000000ULL,
    0x2f00000000000ULL,
    0xd080000000000ULL,
    0x78c0000000000ULL,
    0xaae0000000000ULL,
    0x81f0000000000ULL,
    0x768000000000ULL,
    0x42b4000000000ULL,
    0x33e2000000000ULL,
    0x5c6d000000000ULL,
    0x1c0b800000000ULL,
    0x8a24400000000ULL,
    0x972a600000000ULL,
    0x5ca5500000000ULL,
    0x1ac2680000000ULL,
    0x79d1540000000ULL,
    0xeb406e0000000ULL,
    0x508c530000000ULL,
    0x38c3e38000000ULL,
    0xcaec1cc000000ULL,
    0xf1e3882000000ULL,
    0xdf504e3000000ULL,
    0xbea86d2800000ULL,
    0x89f85dbc00000ULL,
    0x7341e98600000ULL,
    0xccb116a300000ULL,
    0xf2e0027080000ULL,
    0x3df00a4840000ULL,
    0xdd680f6ea0000ULL,
    0x1db406be90000ULL,
    0x3b620e0618000ULL,
    0xd8ad0c60d4000ULL,
    0xceb8d52be000ULL,
    0x24d449f981000ULL,
    0x40c26d8618800ULL,
    0x66d150a0d4400ULL,
    0x83c06572be600ULL,
    0xa44c57c981700ULL,
    0xf023e8ae18580ULL,
    0x981c191cd4bc0ULL,
    0x9c0b82f4beda0ULL,
    0xca24426a815f0ULL,
    0xf72a615e98888ULL,
    0x2ca55b9494c4cULL,
    0xc2c26bba1e70eULL,
    0x85d159e411defULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0x7400000000000ULL,
    0x200000000000ULL,
    0xbf00000000000ULL,
    0x1880000000000ULL,
    0x5c0000000000ULL,
    0x5be0000000000ULL,
    0x6af0000000000ULL,
    0xdb68000000000ULL,
    0x6b2c000000000ULL,
    0x2aa2000000000ULL,
    0x72ef000000000ULL,
    0x9f57800000000ULL,
    0xa12c400000000ULL,
    0x49bc600000000ULL,
    0xa474100000000ULL,
    0x53b4680000000ULL,
    0x5f68140000000ULL,
    0x811e6e0000000ULL,
    0x399b150000000ULL,
    0xc63e58000000ULL,
    0x9f84524000000ULL,
    0x294205a000000ULL,
    0x3c1f09b000000ULL,
    0x9e3f8c2800000ULL,
    0x11004dec00000ULL,
    0xc19e6f9600000ULL,
    0x185b1a4300000ULL,
    0x8d83e1a080000ULL,
    0x2e745fb1c0000ULL,
    0x50aa052aa0000ULL,
    0x99f3026f30000ULL,
    0xf5fd8ed7d8000ULL,
    0xb3df4b61f4000ULL,
    0xc6c1e9d21e000ULL,
    0xbd6b5d79b5000ULL,
    0x6e158357d8800ULL,
    0xf9334d21f4400ULL,
    0x6d83e2721e600ULL,
    0x7e7451c9b5100ULL,
    0x88aa02ffd8380ULL,
    0xfdf3068df4340ULL,
    0x4ffd86441e620ULL,
    0x78df473ab5af0ULL,
    0xdc41e2f758208ULL,
    0x7ab52903431cULL,
    0x2d758758be39eULL,
    0x960343a685c5fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0xdc00000000000ULL,
    0x1200000000000ULL,
    0x6500000000000ULL,
    0x6d80000000000ULL,
    0x97c0000000000ULL,
    0xc1e0000000000ULL,
    0x72f0000000000ULL,
    0xec68000000000ULL,
    0x8514000000000ULL,
    0xdd86000000000ULL,
    0x3fe1000000000ULL,
    0xeddd800000000ULL,
    0xa8e8c00000000ULL,
    0x5559a00000000ULL,
    0xfab7500000000ULL,
    0x2f51a80000000ULL,
    0x9393540000000ULL,
    0x8dfae0000000ULL,
    0x85565b0000000ULL,
    0xa28c228000000ULL,
    0xcb7b96c000000ULL,
    0x958603a000000ULL,
    0xa3e1009000000ULL,
    0x9fdd8d7800000ULL,
    0x3de8cdec00000ULL,
    0xf0d9a76600000ULL,
    0xb17755c300000ULL,
    0xfcb1ad2080000ULL,
    0x8463545240000ULL,
    0x8937a7d8e0000ULL,
    0x97825efc50000ULL,
    0xbeea235d38000ULL,
    0x866a90cedc000ULL,
    0x94338b36a2000ULL,
    0x8e1dce1aa9000ULL,
    0x170229dd38800ULL,
    0xf8be920edc400ULL,
    0x32558696a2600ULL,
    0x8a0cc58aa9f00ULL,
    0xa137ae2538480ULL,
    0xfb825d22dc9c0ULL,
    0x4ea2c50a2720ULL,
    0xcf6a9bd9a9950ULL,
    0x23b384fdb8258ULL,
    0xa0ddc65c9c0bcULL,
    0xa962214e42b3eULL,
    0x788e9b76f9e7fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0x4c00000000000ULL,
    0x6600000000000ULL,
    0xfb00000000000ULL,
    0xc480000000000ULL,
    0x4cc0000000000ULL,
    0x47e0000000000ULL,
    0x7b50000000000ULL,
    0x5b98000000000ULL,
    0xa64000000000ULL,
    0xb23a000000000ULL,
    0x6131000000000ULL,
    0x418c800000000ULL,
    0x4f68400000000ULL,
    0x91b8600000000ULL,
    0xc77fd00000000ULL,
    0xc582680000000ULL,
    0x9d4edc0000000ULL,
    0x608eea0000000ULL,
    0xeee69b0000000ULL,
    0x8ed6898000000ULL,
    0x1ec94fc000000ULL,
    0x76cce7e000000ULL,
    0x72e39c7000000ULL,
    0x60f801a800000ULL,
    0xb1f403bc00000ULL,
    0xe8420c5600000ULL,
    0x9b050bc900000ULL,
    0x54ae8e0f80000ULL,
    0x24fd471f40000ULL,
    0x43eeec8420000ULL,
    0x69769eb050000ULL,
    0x8aae8ecae8000ULL,
    0x53fd4a8fd4000ULL,
    0xc16ee05eee000ULL,
    0xaeb69e2769000ULL,
    0x31ce8f62e8800ULL,
    0x286d4933d4c00ULL,
    0xbb16ec08ee200ULL,
    0x248295ee69700ULL,
    0x1cec816d68b80ULL,
    0xff84e2c948c0ULL,
    0xf40608cce460ULL,
    0x718bdb5e398b0ULL,
    0x974067a7807c8ULL,
    0x8d8bd8a340c0cULL,
    0xe9406ad22001eULL,
    0x4a8bde7950f05ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0xdc00000000000ULL,
    0x2200000000000ULL,
    0x4b00000000000ULL,
    0xa880000000000ULL,
    0xc0000000000ULL,
    0xbfe0000000000ULL,
    0x1950000000000ULL,
    0xd888000000000ULL,
    0x68e4000000000ULL,
    0x33c6000000000ULL,
    0xc359000000000ULL,
    0x8785800000000ULL,
    0x3e49400000000ULL,
    0x5a36600000000ULL,
    0x9f01d00000000ULL,
    0x3690680000000ULL,
    0x59c8dc0000000ULL,
    0x947dee0000000ULL,
    0x6d359f0000000ULL,
    0x4585828000000ULL,
    0x854942c000000ULL,
    0x5ab668a000000ULL,
    0x73c1da7000000ULL,
    0x63706c2800000ULL,
    0xd798d57c00000ULL,
    0xc675ed5600000ULL,
    0x4e119a2900000ULL,
    0x61238e7e80000ULL,
    0x5f8045d740000ULL,
    0xba5bedeae0000ULL,
    0x3c3c915ed0000ULL,
    0xd2080ae4b8000ULL,
    0xe3240c2344000ULL,
    0x44a60bb15e000ULL,
    0xeac9000819000ULL,
    0x28ed8eccb8800ULL,
    0x93fd455f44c00ULL,
    0x937868e75e600ULL,
    0x7fbcd52119300ULL,
    0x2a53ea3238480ULL,
    0xa4189e48041c0ULL,
    0xf62e03adbe420ULL,
    0xd52d010fc97b0ULL,
    0xf1ab8e7e80e08ULL,
    0x2b6445d7401ccULL,
    0xcb9dedeae0fdeULL,
    0x8465915ed0625ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xc800000000000ULL,
    0x4400000000000ULL,
    0x8200000000000ULL,
    0x2d00000000000ULL,
    0xef80000000000ULL,
    0x6740000000000ULL,
    0x4860000000000ULL,
    0xcaf0000000000ULL,
    0x6238000000000ULL,
    0x9d2c000000000ULL,
    0xd78e000000000ULL,
    0x3b57000000000ULL,
    0x4663800000000ULL,
    0x21cac00000000ULL,
    0x22a8200000000ULL,
    0x38c5700000000ULL,
    0x171e280000000ULL,
    0xdebe740000000ULL,
    0xa6f3aa0000000ULL,
    0x1423b90000000ULL,
    0xa38068000000ULL,
    0x492c0d4000000ULL,
    0xbd8e06a000000ULL,
    0x82570b9000000ULL,
    0xe3e38a5800000ULL,
    0x2f8ac0e400000ULL,
    0x748225e00000ULL,
    0xb875780b00000ULL,
    0xd2c6245d80000ULL,
    0xee227709c0000ULL,
    0x5b25a5dd60000ULL,
    0x78a8bfc8b0000ULL,
    0xf9ed89bd58000ULL,
    0xce9dc53b1c000ULL,
    0xecbad945a000ULL,
    0xa00fb9b58b000ULL,
    0x9036056558800ULL,
    0xe83b089f1c400ULL,
    0x940d896a5a200ULL,
    0x4a2dca2e8bd00ULL,
    0x6913ab60d8480ULL,
    0x6d93bf72dc040ULL,
    0x4a600ee93aa20ULL,
    0xa7f00ded3bfd0ULL,
    0xadb8068000a78ULL,
    0x2a6c0d4000634ULL,
    0x57ee06a000ea6ULL,
    0xb5a70b900037fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xd800000000000ULL,
    0xc400000000000ULL,
    0xc600000000000ULL,
    0xa100000000000ULL,
    0x6a80000000000ULL,
    0xb5c0000000000ULL,
    0xcae0000000000ULL,
    0x6570000000000ULL,
    0xec08000000000ULL,
    0x9a34000000000ULL,
    0x4312000000000ULL,
    0x7d9b000000000ULL,
    0xa676800000000ULL,
    0xd188c00000000ULL,
    0xdc53600000000ULL,
    0xe287900000000ULL,
    0x79c9680000000ULL,
    0x80e8940000000ULL,
    0x6e4de60000000ULL,
    0xfd8b530000000ULL,
    0xe6600b8000000ULL,
    0x31b00f4000000ULL,
    0xac680fe000000ULL,
    0x3a84015000000ULL,
    0xbdfa074800000ULL,
    0x46df0d0c00000ULL,
    0xcf6c866600000ULL,
    0x9727c71b00000ULL,
    0x53b7e4c680000ULL,
    0xfb5454a840000ULL,
    0xc90c8e5fa0000ULL,
    0xd697cc2df0000ULL,
    0x27dfee96c8000ULL,
    0x5d054627c000ULL,
    0xb2f686137e000ULL,
    0x3148cca945000ULL,
    0x8233625ec8800ULL,
    0x2737962e7c400ULL,
    0x2ba16f957ee00ULL,
    0xaf6c9ae245700ULL,
    0xa737e9d048580ULL,
    0x6b945b8a3c840ULL,
    0x4f6c81acde260ULL,
    0xd727c2d4b5d10ULL,
    0xb3b7e80000328ULL,
    0x8b5454000031cULL,
    0x110c860000eceULL,
    0x1297c30000b47ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0x1400000000000ULL,
    0x5600000000000ULL,
    0x2900000000000ULL,
    0x7d80000000000ULL,
    0xd640000000000ULL,
    0xe160000000000ULL,
    0x69d0000000000ULL,
    0x8838000000000ULL,
    0x7c1c000000000ULL,
    0x5a02000000000ULL,
    0x7b2f000000000ULL,
    0xbabe800000000ULL,
    0x96dac00000000ULL,
    0x1ca0a00000000ULL,
    0x77d0b00000000ULL,
    0x7d1aa80000000ULL,
    0x4ba3bc0000000ULL,
    0x4f46220000000ULL,
    0x34c67d0000000ULL,
    0x1380098000000ULL,
    0x2b400c4000000ULL,
    0x6ae004e000000ULL,
    0x869002d000000ULL,
    0xacd8093800000ULL,
    0xd78c0eb400000ULL,
    0x655a0cae00000ULL,
    0x47e3096900000ULL,
    0x1504814d80000ULL,
    0x47a9cc38c0000ULL,
    0x1d7c2335a0000ULL,
    0xf3f576ee30000ULL,
    0x533c8e8848000ULL,
    0xd6b5c51e9c000ULL,
    0x74fe25c1c2000ULL,
    0x739a78a257000ULL,
    0x9b620e3048800ULL,
    0xc2ff07ea9cc00ULL,
    0x2a868d8fc2a00ULL,
    0xeec6c31b57100ULL,
    0xa8a2ae45c8380ULL,
    0x31ffb9665cd40ULL,
    0xec24201462f60ULL,
    0x2239719c67390ULL,
    0xcf06800000458ULL,
    0xfc86c00000024ULL,
    0x7c2a00000176ULL,
    0x752fb0000050dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x5800000000000ULL,
    0xf400000000000ULL,
    0x4e00000000000ULL,
    0x5700000000000ULL,
    0xbf80000000000ULL,
    0xd0c0000000000ULL,
    0xf5e0000000000ULL,
    0x4f50000000000ULL,
    0xf508000000000ULL,
    0xf2b4000000000ULL,
    0x6652000000000ULL,
    0xe597000000000ULL,
    0x79c8800000000ULL,
    0x2543400000000ULL,
    0xa425a00000000ULL,
    0xc630100000000ULL,
    0xeb1fa80000000ULL,
    0x2d831c0000000ULL,
    0x65ed2e0000000ULL,
    0xa7735b0000000ULL,
    0x393a0b8000000ULL,
    0x38b3044000000ULL,
    0x2772876000000ULL,
    0xf9304d3000000ULL,
    0x58b72a9800000ULL,
    0x5750573c00000ULL,
    0xa120884600000ULL,
    0xaca7408900000ULL,
    0x197fadf680000ULL,
    0xf61311d240000ULL,
    0x13052beba0000ULL,
    0xc99756c730000ULL,
    0x3e0044928000ULL,
    0x5c500d5604000ULL,
    0x3c8803c3f2000ULL,
    0xf174052245000ULL,
    0x3a320d5128800ULL,
    0xd907022a04c00ULL,
    0x88a082e5f2600ULL,
    0x1f67439b45700ULL,
    0x7d1fa9bfa8d80ULL,
    0x4e831c8444340ULL,
    0xf46d2828522e0ULL,
    0x50b353e575270ULL,
    0x2b5a091800678ULL,
    0x53230f7c00e4cULL,
    0x699a812600dbeULL,
    0x13d446b900685ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x6800000000000ULL,
    0x1c00000000000ULL,
    0x3a00000000000ULL,
    0x700000000000ULL,
    0xfc80000000000ULL,
    0xe6c0000000000ULL,
    0xd260000000000ULL,
    0xf8b0000000000ULL,
    0xe8c8000000000ULL,
    0x9344000000000ULL,
    0xc916000000000ULL,
    0x9d95000000000ULL,
    0x476d800000000ULL,
    0x3f09400000000ULL,
    0x58b3e00000000ULL,
    0x38d7100000000ULL,
    0x6b5e680000000ULL,
    0x4d1e540000000ULL,
    0xd38d8e0000000ULL,
    0x66794b0000000ULL,
    0xfe9be08000000ULL,
    0x45e31ec000000ULL,
    0xf0e06d2000000ULL,
    0xa77f55b000000ULL,
    0x8f3e026800000ULL,
    0x50a1051c00000ULL,
    0xd4d3826e00000ULL,
    0x39684f2700000ULL,
    0x5600668a80000ULL,
    0x150f570f40000ULL,
    0x87960675e0000ULL,
    0xd0550af110000ULL,
    0x5b8d8c09b8000ULL,
    0xca7944cec4000ULL,
    0xcc9be4d766000ULL,
    0xaee31702a5000ULL,
    0x5e606a8338800ULL,
    0x5abf53c184400ULL,
    0x9bde02a286600ULL,
    0x49d10df3b5f00ULL,
    0x12fb8e8a80e80ULL,
    0xb45c430f405c0ULL,
    0xa5be6875e05a0ULL,
    0xe36e51f110f70ULL,
    0xe1258489b8948ULL,
    0xe18d4e0ec4facULL,
    0x8d45e7f766e86ULL,
    0x503219b2a5ffbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x7c00000000000ULL,
    0x5e00000000000ULL,
    0x9d00000000000ULL,
    0xde80000000000ULL,
    0x7d40000000000ULL,
    0xfbe0000000000ULL,
    0x290000000000ULL,
    0x1348000000000ULL,
    0xbedc000000000ULL,
    0x981e000000000ULL,
    0x641d000000000ULL,
    0xfa34800000000ULL,
    0x873ec00000000ULL,
    0x4985e00000000ULL,
    0x1ce0700000000ULL,
    0x9b31680000000ULL,
    0xc79ebc0000000ULL,
    0x49d4860000000ULL,
    0x29aecd0000000ULL,
    0xcccdec8000000ULL,
    0x533c76c000000ULL,
    0xabaf6f6000000ULL,
    0x3fc3b21000000ULL,
    0xc8800e0800000ULL,
    0x4c40030400000ULL,
    0xb3600ed600000ULL,
    0x8ed0013d00000ULL,
    0x402809ea80000ULL,
    0x200c05d4c0000ULL,
    0xf03600bd60000ULL,
    0x381103a810000ULL,
    0x54028474a8000ULL,
    0x222fc9cf3c000ULL,
    0xc30765b196000ULL,
    0x438fb2792b000ULL,
    0xa3d60d9e28800ULL,
    0x86810c1bfcc00ULL,
    0xf94a850cf6e00ULL,
    0x11f3c1d13b100ULL,
    0xad9961ea80a80ULL,
    0x26d2b9d4c0bc0ULL,
    0xfc0286bd60be0ULL,
    0x9e2fcea8108d0ULL,
    0x7d0768f4a8f68ULL,
    0xce8fbf0f3c014ULL,
    0x55560ad196a5eULL,
    0x87c100692b9f9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0xac00000000000ULL,
    0x1200000000000ULL,
    0x9300000000000ULL,
    0x3980000000000ULL,
    0x9740000000000ULL,
    0x82e0000000000ULL,
    0x47b0000000000ULL,
    0x3648000000000ULL,
    0x3864000000000ULL,
    0x916e000000000ULL,
    0x87c1000000000ULL,
    0x431b800000000ULL,
    0x61b6400000000ULL,
    0x2347a00000000ULL,
    0x44fa300000000ULL,
    0x62bc280000000ULL,
    0x22fc740000000ULL,
    0x17b38e0000000ULL,
    0x2e62430000000ULL,
    0xec61a68000000ULL,
    0x275f3dc000000ULL,
    0xaac9a9a000000ULL,
    0xfb8b37f000000ULL,
    0x5c6facb800000ULL,
    0xf6e334400000ULL,
    0x16fa253600000ULL,
    0x91a974cf00000ULL,
    0x6b6e04ea80000ULL,
    0x88c1070d40000ULL,
    0x9b8e4a60000ULL,
    0xb9f6467e50000ULL,
    0x8227a73158000ULL,
    0xab0a309874000ULL,
    0xfd9427f742000ULL,
    0x596871ed87000ULL,
    0xbf58bdbd8800ULL,
    0x4137439534400ULL,
    0x8abc27bd22600ULL,
    0xbefc7493d7700ULL,
    0x6db38a6a80880ULL,
    0x61624ecd40ec0ULL,
    0xcfe1a9ea60720ULL,
    0x8f1f328e50e30ULL,
    0x3a9ad0958318ULL,
    0xb87b3e1c743b4ULL,
    0xd147ab614290eULL,
    0xe7fa32d287d4bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x2800000000000ULL,
    0x400000000000ULL,
    0xee00000000000ULL,
    0xd300000000000ULL,
    0xe380000000000ULL,
    0xfa40000000000ULL,
    0x65e0000000000ULL,
    0x9d90000000000ULL,
    0xa168000000000ULL,
    0x324c000000000ULL,
    0x71ee000000000ULL,
    0x3b83000000000ULL,
    0xa666800000000ULL,
    0x17c9400000000ULL,
    0x5ca3200000000ULL,
    0x53def00000000ULL,
    0x12a5a80000000ULL,
    0x30c7bc0000000ULL,
    0x90e8e0000000ULL,
    0xe685410000000ULL,
    0x86cd248000000ULL,
    0x461dfd4000000ULL,
    0x1723246000000ULL,
    0x6d9ef17000000ULL,
    0xf945aed800000ULL,
    0xae57b39400000ULL,
    0x63e684e600000ULL,
    0x2a89473d00000ULL,
    0x7cc3262880000ULL,
    0x330efa8dc0000ULL,
    0xf3ada75060000ULL,
    0xb25bbe48f0000ULL,
    0xb1e885fe88000ULL,
    0x5b9a4889a4000ULL,
    0x764da95cda000ULL,
    0x3fcbbc58bb000ULL,
    0x58808bd608800ULL,
    0xbdd64e0464c00ULL,
    0xc1a3a00cba600ULL,
    0xd348b3104bd00ULL,
    0xf36602a880a80ULL,
    0x835f07cdc0c40ULL,
    0x1b608330608e0ULL,
    0xe7464f38f0030ULL,
    0x254bab2688cb8ULL,
    0x1c44bb1da4fe4ULL,
    0xc2e80dbada8beULL,
    0x80c0b65bb4e9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x6800000000000ULL,
    0x4c00000000000ULL,
    0x600000000000ULL,
    0xf300000000000ULL,
    0x2580000000000ULL,
    0xd2c0000000000ULL,
    0x5160000000000ULL,
    0xb0000000000ULL,
    0x2c48000000000ULL,
    0x8a2c000000000ULL,
    0xd51e000000000ULL,
    0xe691000000000ULL,
    0xcf46800000000ULL,
    0xa787400000000ULL,
    0x1bf6a00000000ULL,
    0xd9c4500000000ULL,
    0x70f8280000000ULL,
    0xa86f140000000ULL,
    0xb0108e0000000ULL,
    0x783a410000000ULL,
    0x442e208000000ULL,
    0x1a1211c000000ULL,
    0x9d280ee000000ULL,
    0x9a9c0ff000000ULL,
    0xf156043800000ULL,
    0x70bd071c00000ULL,
    0x745881ce00000ULL,
    0xfe1649e700000ULL,
    0xf73027b280000ULL,
    0x5f831799c0000ULL,
    0x1fee85cd60000ULL,
    0xa3db4fe7d0000ULL,
    0xbdc0a2b388000ULL,
    0x5ac9511a64000ULL,
    0xd68a10f82000ULL,
    0xe955f86f1000ULL,
    0xc35ea58108800ULL,
    0xc1985743a4400ULL,
    0xb8ce2a22e2600ULL,
    0x94621f9121500ULL,
    0xee00030a80e80ULL,
    0xff000145c00c0ULL,
    0x43800ae360660ULL,
    0x71c009f0d0a30ULL,
    0x1ce00139084d8ULL,
    0x9e70019fa49ecULL,
    0x7b28050ce2576ULL,
    0x799c098621f3bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0x4c00000000000ULL,
    0xb600000000000ULL,
    0xf900000000000ULL,
    0xb280000000000ULL,
    0x9340000000000ULL,
    0x87e0000000000ULL,
    0x5590000000000ULL,
    0xc4c8000000000ULL,
    0x8804000000000ULL,
    0x8c2e000000000ULL,
    0x5629000000000ULL,
    0x8922800000000ULL,
    0xba91c00000000ULL,
    0xdf41e00000000ULL,
    0x31c5f00000000ULL,
    0xacab680000000ULL,
    0x76503c0000000ULL,
    0x1b44860000000ULL,
    0xbfccb0000000ULL,
    0x3ad6e8000000ULL,
    0x4ded33c000000ULL,
    0x328003e000000ULL,
    0x5340075000000ULL,
    0x67e00e4800000ULL,
    0x259001a400000ULL,
    0xccc803d600000ULL,
    0xc40408ad00000ULL,
    0x3a2e0f5280000ULL,
    0xaf29024940000ULL,
    0x3ba28fa660000ULL,
    0x29d1c4d6d0000ULL,
    0x58a1e82ec8000ULL,
    0x6455fb928c000ULL,
    0x6863612ab6000ULL,
    0xfe543c3503000ULL,
    0x976a89fc48800ULL,
    0x5dd5ca1bccc00ULL,
    0x8a8fed6cd6e00ULL,
    0xf77cffb3d3700ULL,
    0xedc1ef9a80880ULL,
    0x6285f02d408c0ULL,
    0xcb4b6f9060560ULL,
    0x53c03b2bd0890ULL,
    0xd78c813448ba8ULL,
    0xcff8c47fccdf4ULL,
    0x39836b5ad631eULL,
    0xe2c43b4ed3ac9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0x6400000000000ULL,
    0xda00000000000ULL,
    0xc500000000000ULL,
    0x8b80000000000ULL,
    0x66c0000000000ULL,
    0x9f60000000000ULL,
    0x88b0000000000ULL,
    0x6e58000000000ULL,
    0xb10c000000000ULL,
    0x19b6000000000ULL,
    0xe7d3000000000ULL,
    0x5eeb800000000ULL,
    0x2375c00000000ULL,
    0x1eb2a00000000ULL,
    0x8571d00000000ULL,
    0x2d81280000000ULL,
    0x55c8140000000ULL,
    0x4fe58a0000000ULL,
    0xeadacd0000000ULL,
    0xd157228000000ULL,
    0x2fab1f4000000ULL,
    0xdcd6052000000ULL,
    0xea630e1000000ULL,
    0x9b33871800000ULL,
    0x64b9c33c00000ULL,
    0x9064a1ce00000ULL,
    0x8e12d8a700000ULL,
    0xc732aab380000ULL,
    0x2b1dbcbc0000ULL,
    0xc36125c6e0000ULL,
    0xeeb81e01f0000ULL,
    0x3d5d8703d8000ULL,
    0xd1a6cf816c000ULL,
    0xe3d92dc012000ULL,
    0x34c4126081000ULL,
    0x7e538f3058800ULL,
    0xf909cb0aac400ULL,
    0x5dbcad26f2200ULL,
    0xadded27171900ULL,
    0x93e4a72b80880ULL,
    0xccd2d3b7c0240ULL,
    0xa252ab28e0fa0ULL,
    0xdf01d5b6f0550ULL,
    0x2eb9202858838ULL,
    0x5d741c36ac02cULL,
    0x618b86e8f2456ULL,
    0x7bc5c7d6714dbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0xe400000000000ULL,
    0x7600000000000ULL,
    0x8700000000000ULL,
    0x5f80000000000ULL,
    0x12c0000000000ULL,
    0xf360000000000ULL,
    0xfa90000000000ULL,
    0xe478000000000ULL,
    0xd14000000000ULL,
    0x7aa2000000000ULL,
    0x245f000000000ULL,
    0x2d38800000000ULL,
    0x8aa2400000000ULL,
    0x9c47200000000ULL,
    0xc912d00000000ULL,
    0xfc87a80000000ULL,
    0x1b649c0000000ULL,
    0x96828a0000000ULL,
    0xee79430000000ULL,
    0xe825a98000000ULL,
    0x6c3b914000000ULL,
    0xa3a04e000000ULL,
    0xe51b0a3000000ULL,
    0x1682809800000ULL,
    0x2e79465c00000ULL,
    0xc825a14e00000ULL,
    0x9c3b90c500000ULL,
    0xb23a0e1180000ULL,
    0x11b0b0840000ULL,
    0x60828e15a0000ULL,
    0xa9794cb8b0000ULL,
    0x97a5a1cfa8000ULL,
    0x8efb9b06d4000ULL,
    0x415a0a307a000ULL,
    0xfb8b009a49000ULL,
    0x84fa865e28800ULL,
    0xa46d414e94c00ULL,
    0xed07a0c5da200ULL,
    0xaaa49612f9f00ULL,
    0x6c628f0980380ULL,
    0x7129401440240ULL,
    0x18bda1bba0560ULL,
    0x6d7f954db0770ULL,
    0x118006c628e78ULL,
    0xb1c0071294f6cULL,
    0xfae0018bda856ULL,
    0x9f5006d7f97d9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0x5400000000000ULL,
    0x4e00000000000ULL,
    0xfd00000000000ULL,
    0xbf80000000000ULL,
    0x5740000000000ULL,
    0x32e0000000000ULL,
    0x36b0000000000ULL,
    0x16f8000000000ULL,
    0xd0b4000000000ULL,
    0x1fd6000000000ULL,
    0xd127000000000ULL,
    0x5daa800000000ULL,
    0x87ec00000000ULL,
    0x7259a00000000ULL,
    0xd163b00000000ULL,
    0x2beb280000000ULL,
    0x8f19740000000ULL,
    0x189c820000000ULL,
    0xfbe9cb0000000ULL,
    0xd70b248000000ULL,
    0x4a97a4000000ULL,
    0x11e4826000000ULL,
    0x8c1dce9000000ULL,
    0x123d251800000ULL,
    0xb73e75a400000ULL,
    0x14b60e5600000ULL,
    0xa9d7075f00000ULL,
    0xc0328ca180000ULL,
    0xe03aceb040000ULL,
    0x5017a5eae0000ULL,
    0x1800b02d30000ULL,
    0xbc0fabe9c8000ULL,
    0x1a04b52e9c000ULL,
    0xb321aa68b2000ULL,
    0x4297beee97000ULL,
    0xe8dd2bc848800ULL,
    0x658e75dedc400ULL,
    0x44e0fe252a00ULL,
    0x20630b53a7f00ULL,
    0xc66481b980680ULL,
    0xcf5dcf1440140ULL,
    0xcedd29bce0ee0ULL,
    0x8c8e7c72300d0ULL,
    0x55ce03c848578ULL,
    0x7a2301dedc034ULL,
    0xa3048de2527ceULL,
    0xfaadc053a7cbbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0xb400000000000ULL,
    0x7a00000000000ULL,
    0x9700000000000ULL,
    0x6380000000000ULL,
    0xf540000000000ULL,
    0x70e0000000000ULL,
    0xd490000000000ULL,
    0x66e8000000000ULL,
    0x2dac000000000ULL,
    0xd842000000000ULL,
    0xe473000000000ULL,
    0x1277800000000ULL,
    0x7b55400000000ULL,
    0x3dcc200000000ULL,
    0x1001300000000ULL,
    0x3833a80000000ULL,
    0x84287c0000000ULL,
    0xb2358e0000000ULL,
    0xbb26490000000ULL,
    0xadbbaf8000000ULL,
    0x18547e4000000ULL,
    0x847f862000000ULL,
    0x42694e3000000ULL,
    0xe366279800000ULL,
    0x89de332400000ULL,
    0x6a0620b600000ULL,
    0xaf0e355d00000ULL,
    0xe78e28c080000ULL,
    0x47723fe3c0000ULL,
    0xcbc42a52a0000ULL,
    0x793d3189f0000ULL,
    0x7e99ab8d58000ULL,
    0xa9f77f8b64000ULL,
    0x9a00028d3a000ULL,
    0x700000b87000ULL,
    0x9b8004cdd8800ULL,
    0x11400228a4c00ULL,
    0x92e000ff9a600ULL,
    0xf79006b277500ULL,
    0x7f680f5880180ULL,
    0x4fec0cc7c0740ULL,
    0xcb220ae4a01a0ULL,
    0xc5a304d4f0c70ULL,
    0x47f834dd8fb8ULL,
    0x82694068a4414ULL,
    0x836628df9a0aeULL,
    0xd9de318277439ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0xf400000000000ULL,
    0xd600000000000ULL,
    0x8d00000000000ULL,
    0x4680000000000ULL,
    0x40000000000ULL,
    0x3be0000000000ULL,
    0x2510000000000ULL,
    0x298000000000ULL,
    0xee5c000000000ULL,
    0xdac6000000000ULL,
    0x9bf000000000ULL,
    0xadd0800000000ULL,
    0xc82ac00000000ULL,
    0x3424e00000000ULL,
    0x7616b00000000ULL,
    0x5d32680000000ULL,
    0xe837c0000000ULL,
    0xf446020000000ULL,
    0xedff010000000ULL,
    0xa8308e8000000ULL,
    0x443ac24000000ULL,
    0xee3ce9e000000ULL,
    0xe10ab79000000ULL,
    0x2c94610800000ULL,
    0xaf6c74d400000ULL,
    0x266e85d600000ULL,
    0xeed9c35500000ULL,
    0x7faa6d1780000ULL,
    0xf0df7ef4c0000ULL,
    0xc6800c65e0000ULL,
    0xc04005ee30000ULL,
    0x9be0047968000ULL,
    0xf51009cd5c000ULL,
    0x4a980ac946000ULL,
    0x1a5c06f6c7000ULL,
    0xcc60866e8800ULL,
    0x84bf03ed9cc00ULL,
    0xeb50837aa6a00ULL,
    0xc86ac04df7d00ULL,
    0xfc4e10800c80ULL,
    0x5306b4d400340ULL,
    0x5faa65d600760ULL,
    0xe0df7355005d0ULL,
    0x2e800517800e8ULL,
    0xe44002f4c0f44ULL,
    0x5e00e65e0edeULL,
    0x8c1004ee30a81ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x6400000000000ULL,
    0xaa00000000000ULL,
    0x6900000000000ULL,
    0x9a80000000000ULL,
    0xf4c0000000000ULL,
    0x9f60000000000ULL,
    0x7510000000000ULL,
    0x24b8000000000ULL,
    0xcfe4000000000ULL,
    0xb8fa000000000ULL,
    0x697f000000000ULL,
    0x422b800000000ULL,
    0x1524c00000000ULL,
    0x74aee00000000ULL,
    0x87d8900000000ULL,
    0xb4ff680000000ULL,
    0x7f435c0000000ULL,
    0xe51a0a0000000ULL,
    0x4caf070000000ULL,
    0x73f38f8000000ULL,
    0x76d0cd4000000ULL,
    0xaa6ce72000000ULL,
    0xb18390d000000ULL,
    0x7b4eeb0800000ULL,
    0x1f0895dc00000ULL,
    0x6da76a5e00000ULL,
    0xe5775f1d00000ULL,
    0x94380c3d80000ULL,
    0x92240aef40000ULL,
    0x9d1a0ce420000ULL,
    0x58af01b9b0000ULL,
    0x21f3812518000ULL,
    0xcbd0c859bc000ULL,
    0x42ece99656000ULL,
    0x484397a2c5000ULL,
    0xd4aee61098800ULL,
    0xf7d8976afcc00ULL,
    0x4cff6f2c76200ULL,
    0xab43590675b00ULL,
    0x971a030800580ULL,
    0x41af09dc00a40ULL,
    0x4373805e008a0ULL,
    0xeb10c81d00d90ULL,
    0xaf8cebbd80428ULL,
    0x30539baf4090cULL,
    0xc096e1c420356ULL,
    0xa5fc9669b01c1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x9800000000000ULL,
    0x1c00000000000ULL,
    0xe600000000000ULL,
    0xdf00000000000ULL,
    0xae80000000000ULL,
    0xac0000000000ULL,
    0x7f60000000000ULL,
    0xe110000000000ULL,
    0xcda8000000000ULL,
    0x9274000000000ULL,
    0x82aa000000000ULL,
    0xb4d1000000000ULL,
    0xdc46800000000ULL,
    0x19b2400000000ULL,
    0x8052600000000ULL,
    0x5fba100000000ULL,
    0x8f5ee80000000ULL,
    0xd9095c0000000ULL,
    0x1820e0000000ULL,
    0x5c65070000000ULL,
    0xd98c8f8000000ULL,
    0xe0734ac000000ULL,
    0xefbce7e000000ULL,
    0x177c5c3000000ULL,
    0xc5268c8800000ULL,
    0xe7a2415c00000ULL,
    0x837a6b1e00000ULL,
    0x770e15bd00000ULL,
    0xea94eaac80000ULL,
    0x90c85af640000ULL,
    0xf66c849020000ULL,
    0x8a3489a50000ULL,
    0x75f4e086c8000ULL,
    0x1d850ba34000ULL,
    0xc3c48dd74e000ULL,
    0x36d74a4185000ULL,
    0x895eeea248800ULL,
    0x7609571074c00ULL,
    0x57020c596e600ULL,
    0xfaa50066d5b00ULL,
    0xd8ec830800180ULL,
    0xc2634b9c00dc0ULL,
    0x6a94ecfe00860ULL,
    0x50c8598d006f0ULL,
    0x966c8e2480368ULL,
    0xb8a347aa4016cULL,
    0xedf4e18e20996ULL,
    0x1dd85a27503e1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0xa800000000000ULL,
    0x9c00000000000ULL,
    0xf600000000000ULL,
    0x1500000000000ULL,
    0x8980000000000ULL,
    0xe7c0000000000ULL,
    0x7560000000000ULL,
    0x9a30000000000ULL,
    0x9b28000000000ULL,
    0x489c000000000ULL,
    0x944e000000000ULL,
    0xd9ad000000000ULL,
    0xefca800000000ULL,
    0x9960c00000000ULL,
    0x5433200000000ULL,
    0xba19900000000ULL,
    0xab1fa80000000ULL,
    0x9088540000000ULL,
    0x3048060000000ULL,
    0x1bac0d0000000ULL,
    0x90e6048000000ULL,
    0xf3f100c000000ULL,
    0x2f6485e000000ULL,
    0xa13dc89000000ULL,
    0xa3b1a7f800000ULL,
    0xe4d55f2c00000ULL,
    0x79ca87ce00000ULL,
    0x5c60c3df00000ULL,
    0x95b320e480000ULL,
    0x51d9942ac0000ULL,
    0x807fa27660000ULL,
    0x83b854d310000ULL,
    0xd4e00ed848000ULL,
    0xa1f0085cdc000ULL,
    0xf84807a79a000ULL,
    0x57ac0c8b95000ULL,
    0x2ee60d44c8800ULL,
    0xeaf1039a1c400ULL,
    0xf8e487fffae00ULL,
    0xcffdc31785900ULL,
    0xa951ac8000280ULL,
    0x8c2554c000dc0ULL,
    0x1e0283e000160ULL,
    0x690cc59000850ULL,
    0xefb5237800218ULL,
    0x5ad89fec007bcULL,
    0x60d3222e00836ULL,
    0x8be99b4f008f3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x7800000000000ULL,
    0xa400000000000ULL,
    0x5e00000000000ULL,
    0xf300000000000ULL,
    0x9780000000000ULL,
    0x6d40000000000ULL,
    0x4be0000000000ULL,
    0x1830000000000ULL,
    0x3428000000000ULL,
    0xc60c000000000ULL,
    0x872a000000000ULL,
    0xb1b7000000000ULL,
    0x3a7e800000000ULL,
    0x824b400000000ULL,
    0x8669200000000ULL,
    0xe876b00000000ULL,
    0xb375a80000000ULL,
    0xf8f6f40000000ULL,
    0x6f80060000000ULL,
    0x8940090000000ULL,
    0xf5e0098000000ULL,
    0x3b30074000000ULL,
    0xdba8026000000ULL,
    0xf4c057000000ULL,
    0x92ca0c9800000ULL,
    0x5a8709e400000ULL,
    0x99d68dc600000ULL,
    0x2907475700000ULL,
    0x4aa327fc80000ULL,
    0x41f1bde3c0000ULL,
    0xbd23233020000ULL,
    0xbcb1b37bb0000ULL,
    0x6ec32dd548000ULL,
    0xd081ba3fc4000ULL,
    0x7ceb2a417a000ULL,
    0x418db5e3df000ULL,
    0x32412731c8800ULL,
    0x6e7abd7804400ULL,
    0xd45fa0d75ae00ULL,
    0x9941fdbf6fd00ULL,
    0x2dfe800000f80ULL,
    0xaf0b400000e40ULL,
    0x2d89200000be0ULL,
    0x2046b00000230ULL,
    0xff5da80000ef8ULL,
    0x9afaf40000c94ULL,
    0xb6aa06000015eULL,
    0xcbf7090000eb3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0xc00000000000ULL,
    0x1e00000000000ULL,
    0x6300000000000ULL,
    0xcc80000000000ULL,
    0xb3c0000000000ULL,
    0xc660000000000ULL,
    0xb030000000000ULL,
    0x5818000000000ULL,
    0x5c14000000000ULL,
    0x5612000000000ULL,
    0x570d000000000ULL,
    0xde84800000000ULL,
    0xcedf400000000ULL,
    0x69cbe00000000ULL,
    0xcf6fd00000000ULL,
    0x2da5680000000ULL,
    0x2a59940000000ULL,
    0xbe000a0000000ULL,
    0x5300030000000ULL,
    0xd480018000000ULL,
    0xcfc007c000000ULL,
    0xe060026000000ULL,
    0xdf3006f000000ULL,
    0x8a980d2800000ULL,
    0x8cd40d0c00000ULL,
    0x5cf200ae00000ULL,
    0x54fd003f00000ULL,
    0x40fc89e780000ULL,
    0x22fb4ec240000ULL,
    0x67c1e0e0a0000ULL,
    0xc476d0b190000ULL,
    0xa533e08968000ULL,
    0xb38bdd9d24000ULL,
    0x94f6174f6000ULL,
    0x52b0931b09000ULL,
    0x90ee8dc6e8800ULL,
    0x2af64d9364400ULL,
    0x73c5695a56200ULL,
    0xa669966599700ULL,
    0xe018000000b80ULL,
    0x10140000004c0ULL,
    0x68120000003e0ULL,
    0x440d000000130ULL,
    0x2a04800000f48ULL,
    0x711f400000bfcULL,
    0xb1abe00000d86ULL,
    0x1c5fd00000d33ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xd800000000000ULL,
    0xc400000000000ULL,
    0xe200000000000ULL,
    0x7b00000000000ULL,
    0x3880000000000ULL,
    0x10c0000000000ULL,
    0x3aa0000000000ULL,
    0xe010000000000ULL,
    0x3018000000000ULL,
    0xb80c000000000ULL,
    0x341e000000000ULL,
    0x3a1b000000000ULL,
    0xbf18800000000ULL,
    0xda9c400000000ULL,
    0x6bcf600000000ULL,
    0x227100000000ULL,
    0xf0d2180000000ULL,
    0xabf540000000ULL,
    0x5800020000000ULL,
    0x400030000000ULL,
    0x8200018000000ULL,
    0x8b0003c000000ULL,
    0xe080036000000ULL,
    0xd4c0031000000ULL,
    0xd8a0038800000ULL,
    0x9b1001ec00000ULL,
    0x89800e200000ULL,
    0xa8cc004300000ULL,
    0xebe00ea80000ULL,
    0xda0b038040000ULL,
    0x8f0080c060000ULL,
    0x629042e030000ULL,
    0x5fd160d078000ULL,
    0x383c10e86c000ULL,
    0x4fca9afc62000ULL,
    0xd023176a71000ULL,
    0x33cf63af3d800ULL,
    0x62713089c400ULL,
    0x72d21a4348600ULL,
    0x81bf57eafd500ULL,
    0xb880000000080ULL,
    0xd0c00000000c0ULL,
    0x5aa0000000060ULL,
    0x10100000000f0ULL,
    0xe8180000000d8ULL,
    0x7c0c0000000c4ULL,
    0xd61e0000000e2ULL,
    0x411b00000007bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x5c00000000000ULL,
    0xae00000000000ULL,
    0xa900000000000ULL,
    0x3b80000000000ULL,
    0x78c0000000000ULL,
    0x18e0000000000ULL,
    0xb810000000000ULL,
    0xcc18000000000ULL,
    0xe61c000000000ULL,
    0x5d16000000000ULL,
    0xd183000000000ULL,
    0x23cb800000000ULL,
    0x2475c00000000ULL,
    0x5245200000000ULL,
    0x97bf700000000ULL,
    0x3e23180000000ULL,
    0x31fd1c0000000ULL,
    0x4380020000000ULL,
    0x54c0030000000ULL,
    0x4ee0038000000ULL,
    0xfd1002c000000ULL,
    0x4198006000000ULL,
    0x6bdc017000000ULL,
    0xd07602b800000ULL,
    0xb85302a400000ULL,
    0xccb380ee00000ULL,
    0x2b9c1e300000ULL,
    0xdbab206380000ULL,
    0x183072e040000ULL,
    0x8ce69b3060000ULL,
    0x2217df9870000ULL,
    0x6f18a37458000ULL,
    0x3689b0460c000ULL,
    0x14dbb0f2e000ULL,
    0x7f27ae51d7000ULL,
    0x6e7e3b2914800ULL,
    0x995e6c2efdc00ULL,
    0xe3351bc08c600ULL,
    0x207e1fa3f4700ULL,
    0x804b800000080ULL,
    0xc0b5c000000c0ULL,
    0x4a52000000e0ULL,
    0x36af7000000b0ULL,
    0xd1bb180000018ULL,
    0xf3211c000005cULL,
    0xa8760200000aeULL,
    0x94530300000a9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0x4400000000000ULL,
    0xda00000000000ULL,
    0xe500000000000ULL,
    0x5980000000000ULL,
    0x3540000000000ULL,
    0x5ca0000000000ULL,
    0x7410000000000ULL,
    0x208000000000ULL,
    0xf114000000000ULL,
    0x4b9e000000000ULL,
    0xac47000000000ULL,
    0x9b28800000000ULL,
    0x7e4b400000000ULL,
    0xe234a00000000ULL,
    0xe9cf300000000ULL,
    0x2070a80000000ULL,
    0x75e8940000000ULL,
    0xa6a0020000000ULL,
    0x2110010000000ULL,
    0xc388028000000ULL,
    0x705403c000000ULL,
    0xf53e00e000000ULL,
    0x7957011000000ULL,
    0x1aa0836800000ULL,
    0x5f1f439400000ULL,
    0xac8aa16600000ULL,
    0x4d830d500000ULL,
    0xe5f0297280000ULL,
    0x8ea7d5d040000ULL,
    0xd02a20820000ULL,
    0x958c32c450000ULL,
    0x334e2bae78000ULL,
    0xd7b0d5711c000ULL,
    0x2a82208ca2000ULL,
    0xcfc371e92d000ULL,
    0x1b6c8a60d2800ULL,
    0xb36ce7f33cc00ULL,
    0xdf640b07c2a00ULL,
    0x2977a512a2500ULL,
    0x9a78a80000080ULL,
    0x80fc940000040ULL,
    0x973e0200000a0ULL,
    0x98570100000f0ULL,
    0x3920828000038ULL,
    0x7f5f43c000044ULL,
    0x91aaa0e0000daULL,
    0x1883110000e5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0x4400000000000ULL,
    0x3200000000000ULL,
    0x8700000000000ULL,
    0xc980000000000ULL,
    0x8b40000000000ULL,
    0xb320000000000ULL,
    0xb410000000000ULL,
    0xfa08000000000ULL,
    0x3b14000000000ULL,
    0x6f9a000000000ULL,
    0x664b000000000ULL,
    0xb9a8800000000ULL,
    0xc4d6400000000ULL,
    0x4578e00000000ULL,
    0xf5bd300000000ULL,
    0xaad3680000000ULL,
    0x1461640000000ULL,
    0x5920020000000ULL,
    0x3710010000000ULL,
    0xa188028000000ULL,
    0xe754034000000ULL,
    0x4d3a016000000ULL,
    0x1d1b011000000ULL,
    0xc28080c800000ULL,
    0xccd2421c00000ULL,
    0x196ae32600000ULL,
    0x23a2322d00000ULL,
    0xcfc1eacc80000ULL,
    0x2ec26d040000ULL,
    0x5ff861e820000ULL,
    0x3d6f71ec50000ULL,
    0x21b98b3e68000ULL,
    0x60c356d92c000ULL,
    0x761e986a2000ULL,
    0xfabc270359000ULL,
    0x7f50615de3800ULL,
    0xe92b738af4c00ULL,
    0x5f0b8a6d4da00ULL,
    0xcd8c542c85900ULL,
    0x195b680000080ULL,
    0xe435640000040ULL,
    0x259a0200000a0ULL,
    0x354b0100000d0ULL,
    0xba28828000058ULL,
    0x5c96434000044ULL,
    0x55d8e16000032ULL,
    0x9ed311000087ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0xf400000000000ULL,
    0x8600000000000ULL,
    0xd700000000000ULL,
    0x5e80000000000ULL,
    0xec40000000000ULL,
    0x9b60000000000ULL,
    0xe610000000000ULL,
    0x2718000000000ULL,
    0x6684000000000ULL,
    0x42000000000ULL,
    0xe161000000000ULL,
    0x430e800000000ULL,
    0x2888c00000000ULL,
    0x35ee00000000ULL,
    0x29e9d00000000ULL,
    0x704c880000000ULL,
    0x1975ec0000000ULL,
    0x8f18020000000ULL,
    0x4284030000000ULL,
    0xae42008000000ULL,
    0xd261004000000ULL,
    0x938e802000000ULL,
    0xe7c8c3d000000ULL,
    0x40bee21800000ULL,
    0xf4b9d35c00000ULL,
    0x92b4897a00000ULL,
    0x75a1efb100000ULL,
    0x3322006d80000ULL,
    0x2371009840000ULL,
    0xca16801c60000ULL,
    0x7d0cc1da10000ULL,
    0xd39ce02108000ULL,
    0x7c8d05584000ULL,
    0x70a20b143a000ULL,
    0xecad2ffe23000ULL,
    0x6ebee3777b800ULL,
    0x7b9d016a7400ULL,
    0x62348b2cb2200ULL,
    0xaae1efbd97b00ULL,
    0x78c2008000080ULL,
    0xa210040000c0ULL,
    0xaeee802000020ULL,
    0xc6d8c3d000010ULL,
    0x3126e21800008ULL,
    0x8a7dd35c000f4ULL,
    0x8f96897a00086ULL,
    0xa5d0efb1000d7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x8800000000000ULL,
    0x8400000000000ULL,
    0x1200000000000ULL,
    0xdd00000000000ULL,
    0xd680000000000ULL,
    0x3640000000000ULL,
    0xd1e0000000000ULL,
    0x5610000000000ULL,
    0x6f08000000000ULL,
    0x3b9c000000000ULL,
    0xb8da000000000ULL,
    0xebb1000000000ULL,
    0x11e0800000000ULL,
    0xf61a400000000ULL,
    0x5f0fa00000000ULL,
    0x639cd00000000ULL,
    0xb4cdc80000000ULL,
    0x7dbbbc0000000ULL,
    0xdee8020000000ULL,
    0xfd8c010000000ULL,
    0xbfd2038000000ULL,
    0x842d034000000ULL,
    0x333a822000000ULL,
    0x44ab421000000ULL,
    0x8a6f204800000ULL,
    0x7ec6937400000ULL,
    0xeca26b5a00000ULL,
    0x7e776cd900000ULL,
    0xd4cdc94780000ULL,
    0xedbbbc5840000ULL,
    0xb6e8003c20000ULL,
    0xa98c02ae70000ULL,
    0x25d2034368000ULL,
    0xdd2d02fec4000ULL,
    0xf7ba822f82000ULL,
    0xafeb42bc69000ULL,
    0x8d0f226e3e800ULL,
    0x1e96922373400ULL,
    0x524a6aceb7200ULL,
    0x13fb6c77aef00ULL,
    0x31fcb8000080ULL,
    0x3d96bf4000040ULL,
    0x1fd28020000e0ULL,
    0xb4274310000d0ULL,
    0x6b3d23c800088ULL,
    0x48ab903400084ULL,
    0x1c78e97a00012ULL,
    0xb1cc2ec9000ddULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0xc00000000000ULL,
    0x1e00000000000ULL,
    0x5d00000000000ULL,
    0x2e80000000000ULL,
    0xd840000000000ULL,
    0x8320000000000ULL,
    0xf210000000000ULL,
    0xb308000000000ULL,
    0xcb8c000000000ULL,
    0x4ad6000000000ULL,
    0x9d7b000000000ULL,
    0x2031800000000ULL,
    0x719bc00000000ULL,
    0xfdcfa00000000ULL,
    0x2cffd00000000ULL,
    0xfde6080000000ULL,
    0x8d72e40000000ULL,
    0x6828020000000ULL,
    0x759c010000000ULL,
    0x87de018000000ULL,
    0xbbf702c000000ULL,
    0x9c67836000000ULL,
    0x38a0c03000000ULL,
    0x40de207800000ULL,
    0xf274117400000ULL,
    0x9da1a8ba00000ULL,
    0xb241376100000ULL,
    0x5c38080c80000ULL,
    0x9785e6c840000ULL,
    0x7ccf814c20000ULL,
    0x747cc0ee30000ULL,
    0xaca023cb58000ULL,
    0x32d31085ec000ULL,
    0x16e2b98c6000ULL,
    0x363df4826f000ULL,
    0x489829353e800ULL,
    0x1956f7a6ff400ULL,
    0xbba1aac118200ULL,
    0x1341355c8b900ULL,
    0xd4b8098000080ULL,
    0xaec5e6c000040ULL,
    0x176f816000060ULL,
    0xf2cc130000b0ULL,
    0xac0821f8000d8ULL,
    0x8e0f13b40000cULL,
    0x55102bda0001eULL,
    0x4a9af7510005dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xc800000000000ULL,
    0xa400000000000ULL,
    0xf200000000000ULL,
    0x2300000000000ULL,
    0xba80000000000ULL,
    0x5ec0000000000ULL,
    0xc160000000000ULL,
    0x1e10000000000ULL,
    0xb518000000000ULL,
    0x8b9c000000000ULL,
    0xb74e000000000ULL,
    0xed39000000000ULL,
    0x25a4800000000ULL,
    0x9876400000000ULL,
    0x390600000000ULL,
    0x335d500000000ULL,
    0x8f20d80000000ULL,
    0xbeabac0000000ULL,
    0x4ef8020000000ULL,
    0xb4c030000000ULL,
    0x2336038000000ULL,
    0x8b501c000000ULL,
    0xeff2832000000ULL,
    0x5ad3429000000ULL,
    0x637ae3c800000ULL,
    0x6512108c00000ULL,
    0x13943aea00000ULL,
    0x4b40bd7b00000ULL,
    0x328b90580000ULL,
    0x98aaff7840000ULL,
    0x57f6db5460000ULL,
    0x36ceacee70000ULL,
    0x3572807d38000ULL,
    0xb41343e4e4000ULL,
    0x8a1ae3fe92000ULL,
    0xaf0211bdd9000ULL,
    0x9c8c3a0c41800ULL,
    0x47dcbfaa75400ULL,
    0xfce6ba1b03600ULL,
    0x853fc75eeb00ULL,
    0x9b25b8000080ULL,
    0xee68edc0000c0ULL,
    0x429ae120000e0ULL,
    0x12c2119000070ULL,
    0x76c3848000c8ULL,
    0x770cbd4c000a4ULL,
    0x409ebbca000f2ULL,
    0x39dffceb00023ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x5800000000000ULL,
    0x8c00000000000ULL,
    0x5a00000000000ULL,
    0xcd00000000000ULL,
    0x7880000000000ULL,
    0x240000000000ULL,
    0x760000000000ULL,
    0x5e10000000000ULL,
    0xe318000000000ULL,
    0x938c000000000ULL,
    0x25d2000000000ULL,
    0x64ab000000000ULL,
    0xf5a1800000000ULL,
    0x2f23400000000ULL,
    0x9c7da00000000ULL,
    0x7199100000000ULL,
    0x34cd480000000ULL,
    0x7e246c0000000ULL,
    0x26f8020000000ULL,
    0x52dc030000000ULL,
    0x812a018000000ULL,
    0xb777024000000ULL,
    0x360b816000000ULL,
    0xc714423000000ULL,
    0x8d96216800000ULL,
    0x66dd533400000ULL,
    0x72369e200000ULL,
    0x18653c0900000ULL,
    0x9f916a1d80000ULL,
    0xbfde3e7840000ULL,
    0xc9a8e80c60000ULL,
    0x1d317f0e30000ULL,
    0x75674a7748000ULL,
    0x17136ee2ac000ULL,
    0x7593805e86000ULL,
    0x1ad840f88d000ULL,
    0x9524221bf6800ULL,
    0x166528b64400ULL,
    0x311ae924b5200ULL,
    0x2a8a7d8dd1b00ULL,
    0x7b5ec98000080ULL,
    0x39fc2e40000c0ULL,
    0x935c236000060ULL,
    0xddfa513000090ULL,
    0xed50e8e800058ULL,
    0xeed7d740008cULL,
    0xd6cd4a820005aULL,
    0x6f246d39000cdULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x1800000000000ULL,
    0xcc00000000000ULL,
    0xe00000000000ULL,
    0xb100000000000ULL,
    0x1a80000000000ULL,
    0xf3c0000000000ULL,
    0xb20000000000ULL,
    0x6110000000000ULL,
    0xb298000000000ULL,
    0x7c4000000000ULL,
    0xb92e000000000ULL,
    0xc613000000000ULL,
    0xd501800000000ULL,
    0xe085c00000000ULL,
    0xf0d8200000000ULL,
    0xb6a0500000000ULL,
    0xf557f80000000ULL,
    0x5ef8a40000000ULL,
    0x2f2e020000000ULL,
    0x7b13030000000ULL,
    0xe181808000000ULL,
    0xd245c1c000000ULL,
    0xf978206000000ULL,
    0xe870533000000ULL,
    0x42eff83800000ULL,
    0x892ca6c400000ULL,
    0x3e18026a00000ULL,
    0x490400cf00000ULL,
    0x868e00ac80000ULL,
    0x95c3004440000ULL,
    0x6e3982aa60000ULL,
    0xb991c32f10000ULL,
    0xfe4e22dcb8000ULL,
    0xa76751dc4c000ULL,
    0x31607b3e06000ULL,
    0x8c6a644d17000ULL,
    0xb8f9a1efe0800ULL,
    0x8a35905ec1400ULL,
    0x8397d99f3fe00ULL,
    0xdd5cf4a4f2900ULL,
    0x6af7f83800080ULL,
    0xbd28a6c4000c0ULL,
    0xac16026a00020ULL,
    0x9e0700cf00070ULL,
    0xf91780ac80018ULL,
    0xbe82c044400ccULL,
    0x29cfa2aa6000eULL,
    0x7822932f100b1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xd800000000000ULL,
    0x400000000000ULL,
    0xe200000000000ULL,
    0xe700000000000ULL,
    0x9580000000000ULL,
    0xbdc0000000000ULL,
    0x6ba0000000000ULL,
    0xdf10000000000ULL,
    0x8188000000000ULL,
    0xe7cc000000000ULL,
    0xd8aa000000000ULL,
    0x708b000000000ULL,
    0xdf48800000000ULL,
    0xfbf0400000000ULL,
    0x5d66e00000000ULL,
    0xf31b00000000ULL,
    0x5a43380000000ULL,
    0xc97f340000000ULL,
    0x152a020000000ULL,
    0x894b010000000ULL,
    0x36e8818000000ULL,
    0x93e0414000000ULL,
    0x916ee36000000ULL,
    0x513db01000000ULL,
    0xb493b8800000ULL,
    0x81e4379c00000ULL,
    0xde6a805600000ULL,
    0x28b743f700000ULL,
    0xd884602e80000ULL,
    0x334af23c40000ULL,
    0x95ed596620000ULL,
    0x847ec78f30000ULL,
    0x9bafdaeaa8000ULL,
    0x7715875e2c000ULL,
    0x6d89baab22000ULL,
    0x89d87558c1000ULL,
    0x1a6603b1b800ULL,
    0xe40df31086c00ULL,
    0xf20fda072ce00ULL,
    0x5f058676ccd00ULL,
    0xc181bb8800080ULL,
    0x87d4779c00040ULL,
    0x88ac605600060ULL,
    0xa896f3f700050ULL,
    0xdb4f582e800d8ULL,
    0x19f9c63c40004ULL,
    0xba6d5b66200e2ULL,
    0x9abec68f300e7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0x7400000000000ULL,
    0xa00000000000ULL,
    0xff00000000000ULL,
    0x4580000000000ULL,
    0xdec0000000000ULL,
    0xfde0000000000ULL,
    0xe710000000000ULL,
    0xa998000000000ULL,
    0xf8cc000000000ULL,
    0x44fe000000000ULL,
    0xeb89000000000ULL,
    0xb3d6800000000ULL,
    0x6b6d400000000ULL,
    0x3a51e00000000ULL,
    0xc3a9b00000000ULL,
    0x26a9580000000ULL,
    0x6428fc0000000ULL,
    0x237e020000000ULL,
    0x8e49030000000ULL,
    0xa9b6818000000ULL,
    0x29bd43c000000ULL,
    0xe9a9e32000000ULL,
    0x89b5b1d000000ULL,
    0x79af582800000ULL,
    0xb1bdfffc00000ULL,
    0xc5ae831600000ULL,
    0xcfb1407b00000ULL,
    0x30b7e27780000ULL,
    0x752cb05c40000ULL,
    0xabe1d98660000ULL,
    0x561cbe3330000ULL,
    0xb101633bf8000ULL,
    0x1891f35224000ULL,
    0xe0483a595a000ULL,
    0xa4a90d16b5000ULL,
    0x4f2e3a3ec7800ULL,
    0xfcec0e42e6c00ULL,
    0x9786b994c5600ULL,
    0xadc84d4f93f00ULL,
    0x6e61d82800080ULL,
    0x48dcbffc000c0ULL,
    0x2ce1631600060ULL,
    0xf81f07b000f0ULL,
    0x81d03a77800c8ULL,
    0x28650c5c40074ULL,
    0x1d03b866000aULL,
    0xe8650d33300ffULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0x2400000000000ULL,
    0xda00000000000ULL,
    0xc100000000000ULL,
    0xd180000000000ULL,
    0x17c0000000000ULL,
    0x220000000000ULL,
    0x7510000000000ULL,
    0xa398000000000ULL,
    0xd2c4000000000ULL,
    0x19a2000000000ULL,
    0xcbdd000000000ULL,
    0x543c800000000ULL,
    0x6a0f400000000ULL,
    0x7902200000000ULL,
    0xbd93300000000ULL,
    0xf9c4780000000ULL,
    0x7136040000000ULL,
    0x4182020000000ULL,
    0xbfcd030000000ULL,
    0x624808000000ULL,
    0xbf0b404000000ULL,
    0xa8021a000000ULL,
    0x275e309000000ULL,
    0xd460fb6800000ULL,
    0x8fd470400000ULL,
    0xf0a2214600000ULL,
    0xde43335f00000ULL,
    0xa9fc788880000ULL,
    0xd122059440000ULL,
    0x9198012e60000ULL,
    0xf7c400db10000ULL,
    0x3222038e88000ULL,
    0xd1d006b74000ULL,
    0xef9c83b6f2000ULL,
    0x2cdf41673d000ULL,
    0x2ba228488800ULL,
    0xdb4730260cc00ULL,
    0x927e7a2f71e00ULL,
    0x7fef06d0c8100ULL,
    0xe3c836800080ULL,
    0x6b0f4304000c0ULL,
    0x8882234600020ULL,
    0xba53305f00010ULL,
    0x93e4780880068ULL,
    0x202605d440024ULL,
    0x381a008e600daULL,
    0xac09004b100c1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x1800000000000ULL,
    0x1c00000000000ULL,
    0x3a00000000000ULL,
    0x2900000000000ULL,
    0x4380000000000ULL,
    0xd1c0000000000ULL,
    0x6c60000000000ULL,
    0xad10000000000ULL,
    0x2588000000000ULL,
    0x22dc000000000ULL,
    0x96f2000000000ULL,
    0x2753000000000ULL,
    0x842b800000000ULL,
    0xd9ab400000000ULL,
    0x376f200000000ULL,
    0x638f700000000ULL,
    0xe1dfb80000000ULL,
    0x467cc0000000ULL,
    0x3912020000000ULL,
    0x1b83010000000ULL,
    0x2dc3838000000ULL,
    0xc667424000000ULL,
    0x9c15206000000ULL,
    0x7a00707000000ULL,
    0xc90638e800000ULL,
    0xd39f8ca400000ULL,
    0xc9d6a30e00000ULL,
    0x7067324700000ULL,
    0x97131a3180000ULL,
    0xc9ffcf440000ULL,
    0x61509af620000ULL,
    0x4738bdfb70000ULL,
    0x4b25bb33c8000ULL,
    0x2928cf794c000ULL,
    0xfc2b82feae000ULL,
    0x15ab4311ad000ULL,
    0xf56f22643c800ULL,
    0xc68f71ae7dc00ULL,
    0x805fba975ee00ULL,
    0xe0a7cf09ef300ULL,
    0x2cf200e800080ULL,
    0x4e5300a400040ULL,
    0x27ab810e000e0ULL,
    0x986b434700090ULL,
    0x430f21b180018ULL,
    0xd29f72b44001cULL,
    0xfe57b8962003aULL,
    0xfbbcc8b70029ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0xc800000000000ULL,
    0xc400000000000ULL,
    0xf600000000000ULL,
    0xd900000000000ULL,
    0x3880000000000ULL,
    0xe340000000000ULL,
    0x1ae0000000000ULL,
    0x9f10000000000ULL,
    0x1988000000000ULL,
    0xa7dc000000000ULL,
    0x93b2000000000ULL,
    0xa6e9000000000ULL,
    0x5500800000000ULL,
    0x4a8ac00000000ULL,
    0x2c55200000000ULL,
    0x6b6c100000000ULL,
    0x8ccde80000000ULL,
    0x242d1c0000000ULL,
    0xd43a020000000ULL,
    0xcc35010000000ULL,
    0x2032838000000ULL,
    0x8223c24000000ULL,
    0x6535a32000000ULL,
    0x40b6d31000000ULL,
    0x6d70cbd800000ULL,
    0x4dcd0f6400000ULL,
    0xf0adeae200000ULL,
    0x957d1e8d00000ULL,
    0x31d201eb80000ULL,
    0x9ab9013c40000ULL,
    0xb66880c620000ULL,
    0xe246c3cf70000ULL,
    0x6e6f22b6c8000ULL,
    0xae5912efa4000ULL,
    0xbc7f6a6e02000ULL,
    0x114eddc32b000ULL,
    0x95efa238d4800ULL,
    0xe93d15cf0400ULL,
    0x9a4a49be97a00ULL,
    0x5272ce3384700ULL,
    0x244a485800080ULL,
    0xf72cd2400040ULL,
    0xaca49c2000e0ULL,
    0xa532cd9d00090ULL,
    0xe0aa4a33800c8ULL,
    0x1d62ce58400c4ULL,
    0x15c24a24200f6ULL,
    0xfcaecd42700d9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0x3c00000000000ULL,
    0x5600000000000ULL,
    0x8500000000000ULL,
    0x6c80000000000ULL,
    0x51c0000000000ULL,
    0x70a0000000000ULL,
    0xeb10000000000ULL,
    0x9588000000000ULL,
    0x8b5c000000000ULL,
    0x9466000000000ULL,
    0x4f27000000000ULL,
    0x135f800000000ULL,
    0x387ec00000000ULL,
    0x713aa00000000ULL,
    0x2244900000000ULL,
    0x86eeb80000000ULL,
    0xcfe9540000000ULL,
    0xed6e020000000ULL,
    0xd5bb010000000ULL,
    0x1799838000000ULL,
    0xac49c0c000000ULL,
    0x4fed22e000000ULL,
    0xad6650f000000ULL,
    0x35b2195800000ULL,
    0x278ac61400000ULL,
    0x145f3bb200000ULL,
    0x73ec944700000ULL,
    0xfb6d204280000ULL,
    0xb0a6526c40000ULL,
    0x4b121b3620000ULL,
    0x459ac61d70000ULL,
    0x3573be998000ULL,
    0x107096d89c000ULL,
    0x252b21277e000ULL,
    0xc0515272fb000ULL,
    0xd1e598d46a800ULL,
    0x4c68065252400ULL,
    0x32398371ae00ULL,
    0x1d5f055a95500ULL,
    0xb1741ad800080ULL,
    0xf3bdc6d400040ULL,
    0xca88b952000e0ULL,
    0xccce54b700030ULL,
    0xf031811a800b8ULL,
    0x64c5c0784003cULL,
    0xe423228420056ULL,
    0x66cd535a70085ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x6800000000000ULL,
    0xd400000000000ULL,
    0x7e00000000000ULL,
    0x7b00000000000ULL,
    0xee80000000000ULL,
    0xb1c0000000000ULL,
    0xad60000000000ULL,
    0x5110000000000ULL,
    0xab88000000000ULL,
    0x444c000000000ULL,
    0xc226000000000ULL,
    0x25bd000000000ULL,
    0x83e2800000000ULL,
    0xc0dbc00000000ULL,
    0x56ed600000000ULL,
    0x4d4ad00000000ULL,
    0xd5afb80000000ULL,
    0xcbf1ec0000000ULL,
    0x44ce020000000ULL,
    0x70e1010000000ULL,
    0x8a4c818000000ULL,
    0x912ac0c000000ULL,
    0x7f29e1a000000ULL,
    0x7c2c135000000ULL,
    0x7ea059f800000ULL,
    0x3d60fdec00000ULL,
    0x290cd9ba00000ULL,
    0x479a3fc700000ULL,
    0xb64d393580000ULL,
    0x7b2a2c8440000ULL,
    0x1a23608e20000ULL,
    0xd9abd38130000ULL,
    0x49e33ad098000ULL,
    0xf5db2eeaf4000ULL,
    0xab67e3ed8a000ULL,
    0xc60d12786f000ULL,
    0xb70cd82c35800ULL,
    0x4c9a3f9a6b400ULL,
    0x50cd3b571ee00ULL,
    0x2eea2fedb7b00ULL,
    0xa143607800080ULL,
    0x27bbd12c00040ULL,
    0x72eb3a1a00060ULL,
    0x7b572d9700030ULL,
    0x2aa1e14d80068ULL,
    0x36011a8400d4ULL,
    0x320658942007eULL,
    0x991dff163007bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0x7400000000000ULL,
    0x9600000000000ULL,
    0x7700000000000ULL,
    0x5a80000000000ULL,
    0xef40000000000ULL,
    0x1ee0000000000ULL,
    0x3510000000000ULL,
    0x6b88000000000ULL,
    0x6acc000000000ULL,
    0x173e000000000ULL,
    0x1eb7000000000ULL,
    0xd576800000000ULL,
    0x8fc6c00000000ULL,
    0xd4b4e00000000ULL,
    0xc86e500000000ULL,
    0x1852680000000ULL,
    0x47c9c0000000ULL,
    0xba56020000000ULL,
    0x856b010000000ULL,
    0x67c0818000000ULL,
    0x8bdc3c000000ULL,
    0xe27c60e000000ULL,
    0xb55f91d000000ULL,
    0xcbf00a5800000ULL,
    0xee840ddc00000ULL,
    0x59588b6a00000ULL,
    0xf9e5cebd00000ULL,
    0xa792ebfb80000ULL,
    0xc8c15e1440000ULL,
    0x962a62ce20000ULL,
    0xc33492bb30000ULL,
    0x58b08b64f8000ULL,
    0xa79cfb6dc000ULL,
    0x6944ea87da000ULL,
    0xe1ea5c8e1b000ULL,
    0x438ae39b53800ULL,
    0x96d95354f9400ULL,
    0xad24e9bee9a00ULL,
    0x7bba5cc382700ULL,
    0x56e2e3d800080ULL,
    0x3905521c00040ULL,
    0xe992e98a00060ULL,
    0x7bc15e6d000f0ULL,
    0x2aa602380038ULL,
    0xdf74900840074ULL,
    0xb2d0894420096ULL,
    0xd329cdd630077ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x6400000000000ULL,
    0xda00000000000ULL,
    0x9700000000000ULL,
    0x8580000000000ULL,
    0x18c0000000000ULL,
    0xb6a0000000000ULL,
    0x5910000000000ULL,
    0x7088000000000ULL,
    0xfe44000000000ULL,
    0x99f6000000000ULL,
    0xdc6b000000000ULL,
    0xa4b4800000000ULL,
    0x8207400000000ULL,
    0xb308e00000000ULL,
    0x7f99b00000000ULL,
    0x3fc2980000000ULL,
    0xeb3a140000000ULL,
    0x25de020000000ULL,
    0x1c3f010000000ULL,
    0x304a808000000ULL,
    0x6ce842c000000ULL,
    0x3aea636000000ULL,
    0x8be5f19000000ULL,
    0x776fb6800000ULL,
    0x6720e65c00000ULL,
    0x3c2781600000ULL,
    0x1527a56300000ULL,
    0x48ca985a80000ULL,
    0xbebe16a440000ULL,
    0x7508002220000ULL,
    0xa68401a910000ULL,
    0xf5602efd8000ULL,
    0x357b027dac000ULL,
    0xc3c808cd2000ULL,
    0x184341a71d000ULL,
    0xf0fee168a3800ULL,
    0x34f2b2a526c00ULL,
    0x1ef619f1aa600ULL,
    0x71fd5552b8500ULL,
    0x2076e3e800080ULL,
    0x3ab6b09c00040ULL,
    0x7f00197600020ULL,
    0x799655f3000b0ULL,
    0x86c263b2800d8ULL,
    0x4bb1f23840064ULL,
    0x9388fb54200daULL,
    0x89cfe55a10097ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x2800000000000ULL,
    0x1400000000000ULL,
    0x6a00000000000ULL,
    0x9100000000000ULL,
    0x580000000000ULL,
    0xb540000000000ULL,
    0xbba0000000000ULL,
    0xa710000000000ULL,
    0xda98000000000ULL,
    0x83dc000000000ULL,
    0x7e7e000000000ULL,
    0xee75000000000ULL,
    0xf66a800000000ULL,
    0x2a79400000000ULL,
    0xa478200000000ULL,
    0x776fb00000000ULL,
    0xf7fb280000000ULL,
    0x2d37b40000000ULL,
    0xb2c6020000000ULL,
    0xabf9030000000ULL,
    0x632c838000000ULL,
    0x81c043c000000ULL,
    0xdb74a0a000000ULL,
    0x91fff05000000ULL,
    0xa3789a800000ULL,
    0xa844464400000ULL,
    0xc8378a1600000ULL,
    0xed4445d500000ULL,
    0x47b78b6e80000ULL,
    0x3904465c40000ULL,
    0xd1978a4a60000ULL,
    0x3f54469f70000ULL,
    0xdaaf8971f8000ULL,
    0x8a98456dd4000ULL,
    0x7bc98b47aa000ULL,
    0x527144a8e5000ULL,
    0x487d08e160800ULL,
    0x316d0740fec00ULL,
    0xc0f7aac50ca00ULL,
    0xefa7f426eed00ULL,
    0xed0aa22800080ULL,
    0x5b8af184000c0ULL,
    0x5e5d08b6000e0ULL,
    0xf73d0685000f0ULL,
    0x2bcfa94680028ULL,
    0xaa6bf7d840014ULL,
    0x646ca2fc6006aULL,
    0x9763f01a70091ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x4800000000000ULL,
    0xac00000000000ULL,
    0x8e00000000000ULL,
    0xe500000000000ULL,
    0x4e80000000000ULL,
    0x97c0000000000ULL,
    0xe560000000000ULL,
    0x3e90000000000ULL,
    0xfd8000000000ULL,
    0xe17c000000000ULL,
    0xc92000000000ULL,
    0x2cd1000000000ULL,
    0xe6e9800000000ULL,
    0x5bc3c00000000ULL,
    0xbb6da00000000ULL,
    0x7390500000000ULL,
    0xfd59380000000ULL,
    0xb0b50c0000000ULL,
    0xa0e9820000000ULL,
    0xd2c3c30000000ULL,
    0x9beda38000000ULL,
    0x1150504000000ULL,
    0x1eb9392000000ULL,
    0xb5e50eb000000ULL,
    0xc451803800000ULL,
    0xe82fc09400000ULL,
    0xd627a2ba00000ULL,
    0x4b3d521f00000ULL,
    0x11a2bab580000ULL,
    0xfc67ce4a40000ULL,
    0x960da00760000ULL,
    0x21005311f0000ULL,
    0x9c813a8848000ULL,
    0xa4c90eac44000ULL,
    0xaafb832e26000ULL,
    0xc5d2c0254f000ULL,
    0x1664236ad6800ULL,
    0x9103919fb1400ULL,
    0xe48c9bdd2ce00ULL,
    0xb0c95e9e90300ULL,
    0xd0fabab580080ULL,
    0x4adbce4a400c0ULL,
    0x9fffa007600e0ULL,
    0x23415311f0010ULL,
    0x3db0ba8848048ULL,
    0xb276ceac440acULL,
    0x9304232e2608eULL,
    0x7f9390254f0e5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xc800000000000ULL,
    0x400000000000ULL,
    0xa00000000000ULL,
    0x1b00000000000ULL,
    0xeb80000000000ULL,
    0x5740000000000ULL,
    0xfe0000000000ULL,
    0x6390000000000ULL,
    0xf358000000000ULL,
    0xb5ec000000000ULL,
    0x608a000000000ULL,
    0xd4c1000000000ULL,
    0xecac800000000ULL,
    0x7e6b400000000ULL,
    0x47d2600000000ULL,
    0xa329f00000000ULL,
    0x932da80000000ULL,
    0x4b299c0000000ULL,
    0xe72c820000000ULL,
    0x392b430000000ULL,
    0xe032618000000ULL,
    0x14b9f34000000ULL,
    0xa275ab2000000ULL,
    0xe1c59c1000000ULL,
    0x6626822800000ULL,
    0xa1aa436c00000ULL,
    0xe8fee22e00000ULL,
    0x5e02b21d00000ULL,
    0x191fcb1f80000ULL,
    0x94906d9e40000ULL,
    0x66d929e560000ULL,
    0x8baeddbbb0000ULL,
    0x83f4e1ac28000ULL,
    0xcd83b24e04000ULL,
    0x5253492d32000ULL,
    0x5d6b2f27ed000ULL,
    0x10534bda29800ULL,
    0x826b2f2717c00ULL,
    0x91d34bc89ea00ULL,
    0x1e2b2f0ea2700ULL,
    0xbdb34b1f80080ULL,
    0x2efb2d9e400c0ULL,
    0x4b0b49e560060ULL,
    0xe3872dbbb00d0ULL,
    0x335949ac280c8ULL,
    0xd5ea2e4e04004ULL,
    0xb09fcb2d3200aULL,
    0x1cd06c27ed01bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0x9400000000000ULL,
    0x1200000000000ULL,
    0xdd00000000000ULL,
    0x8680000000000ULL,
    0xe040000000000ULL,
    0xd2a0000000000ULL,
    0xb290000000000ULL,
    0xa248000000000ULL,
    0x97a4000000000ULL,
    0xb012000000000ULL,
    0xe81f000000000ULL,
    0xc16800000000ULL,
    0x9600400000000ULL,
    0x770ca00000000ULL,
    0xcf82500000000ULL,
    0x74de480000000ULL,
    0xefe1f40000000ULL,
    0xe6b6820000000ULL,
    0xf090410000000ULL,
    0xe744a08000000ULL,
    0x9526504000000ULL,
    0xfa4c4ae000000ULL,
    0x73bef65000000ULL,
    0x2a00024800000ULL,
    0x900027400000ULL,
    0xb480029a00000ULL,
    0x2d4003c100000ULL,
    0xec2001aa80000ULL,
    0xc6d0009a40000ULL,
    0x62e802c120000ULL,
    0xf834012a90000ULL,
    0x94da00da48000ULL,
    0x9ffb00217c000ULL,
    0x6ea4837ada000ULL,
    0xcc8f409241000ULL,
    0xd952235512800ULL,
    0xce26116099400ULL,
    0xb3c0eb1331200ULL,
    0xc87ca41ffbd00ULL,
    0x9e7e49aa80080ULL,
    0x8971f49a40040ULL,
    0x76fe80c120020ULL,
    0xaa34402a90010ULL,
    0x69d6a05a480b8ULL,
    0x97950617c094ULL,
    0x36fac99ada012ULL,
    0x8a2eb6c2410ddULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xf800000000000ULL,
    0x7400000000000ULL,
    0x6600000000000ULL,
    0x1300000000000ULL,
    0x9b80000000000ULL,
    0x8a40000000000ULL,
    0xf620000000000ULL,
    0x5f90000000000ULL,
    0x5448000000000ULL,
    0xf12c000000000ULL,
    0x9216000000000ULL,
    0x3517000000000ULL,
    0xe882800000000ULL,
    0x21cac00000000ULL,
    0x847d600000000ULL,
    0xddadf00000000ULL,
    0x6dc5880000000ULL,
    0xb678a40000000ULL,
    0xf8a2820000000ULL,
    0x2d5ac10000000ULL,
    0x2bb5618000000ULL,
    0x96c1f0c000000ULL,
    0xf1f38be000000ULL,
    0xa8ffa5d000000ULL,
    0x2268039800000ULL,
    0xeebc014c00000ULL,
    0xa65e03ee00000ULL,
    0xf43b02e900000ULL,
    0x8294803880000ULL,
    0x60ddc0ae40000ULL,
    0xaffe0c920000ULL,
    0xef673388b0000ULL,
    0x7238e82658000ULL,
    0xe19556fd5c000ULL,
    0x63470a7a8a000ULL,
    0xc4b264f96b000ULL,
    0x875fe0c0d5800ULL,
    0x4ab7317207c00ULL,
    0x4850e99f4e200ULL,
    0xb29571dbe900ULL,
    0x3b19083880080ULL,
    0x678964ae40040ULL,
    0x4b62c920060ULL,
    0xc72af288b0030ULL,
    0x490f09a6580f8ULL,
    0x229e663d5c074ULL,
    0x70c9e19a8a066ULL,
    0xa2e031296b013ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0xdc00000000000ULL,
    0x2200000000000ULL,
    0x2700000000000ULL,
    0xed80000000000ULL,
    0xa7c0000000000ULL,
    0xbf20000000000ULL,
    0x990000000000ULL,
    0x51d8000000000ULL,
    0xd63c000000000ULL,
    0x991a000000000ULL,
    0xc89d000000000ULL,
    0xdd47800000000ULL,
    0x8d6e400000000ULL,
    0xe571e00000000ULL,
    0x196e300000000ULL,
    0x8b64b80000000ULL,
    0xd46d040000000ULL,
    0xcde7820000000ULL,
    0x943e430000000ULL,
    0x2e09e38000000ULL,
    0xed02314000000ULL,
    0xf686b8a000000ULL,
    0x385c077000000ULL,
    0x17e2028800000ULL,
    0x8731039c00000ULL,
    0xdd85803600000ULL,
    0x1fcf43df00000ULL,
    0xab2c625c80000ULL,
    0xa79d735640000ULL,
    0x7cd2d9cf60000ULL,
    0xc0ad77c4f0000ULL,
    0xf152dbd268000ULL,
    0xf76d76bd74000ULL,
    0x8672d8099e000ULL,
    0x72fd7453f9000ULL,
    0xddaadbf2a7800ULL,
    0x5fc1760d48c00ULL,
    0x8b30d9e1fae00ULL,
    0x179c75dfc0100ULL,
    0x4d75a5c80080ULL,
    0x34a23756400c0ULL,
    0xf5ebbcf600e0ULL,
    0xf26004c4f0050ULL,
    0x4cf8005268028ULL,
    0x38ac03fd740dcULL,
    0xc54202a99e022ULL,
    0xe9610023f9027ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x8c00000000000ULL,
    0x5e00000000000ULL,
    0x1b00000000000ULL,
    0xda80000000000ULL,
    0x62c0000000000ULL,
    0xcc60000000000ULL,
    0x2e90000000000ULL,
    0xe8c8000000000ULL,
    0xed7c000000000ULL,
    0x2d12000000000ULL,
    0x1d89000000000ULL,
    0xe4d800000000ULL,
    0x29a9c00000000ULL,
    0x6e72600000000ULL,
    0x3392d00000000ULL,
    0x4d53980000000ULL,
    0x9724ec0000000ULL,
    0xcead820000000ULL,
    0xb2f9c10000000ULL,
    0x2e5a638000000ULL,
    0x99bed04000000ULL,
    0x9669982000000ULL,
    0xa781ee3000000ULL,
    0x975a037800000ULL,
    0x5e35016c00000ULL,
    0x513f80ea00000ULL,
    0x11b0c1cb00000ULL,
    0x5a77e31180000ULL,
    0x1987128a40000ULL,
    0x9c53fadb20000ULL,
    0x8caf3fd9f0000ULL,
    0xbffb99de48000ULL,
    0x83c8ecfd24000ULL,
    0x6ff78288b6000ULL,
    0x2bccc15ce7000ULL,
    0x13e5e1bae9800ULL,
    0x6dce110bbb400ULL,
    0x8cfe78d906600ULL,
    0x6556fc76b7b00ULL,
    0xab21fb1180080ULL,
    0x68b63e8a40040ULL,
    0x3dfe18db200e0ULL,
    0x2edd2ed9f0010ULL,
    0x3277e25e48008ULL,
    0xc58710bd2408cULL,
    0x2a53f8a8b605eULL,
    0xbaf3e6ce701bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x6800000000000ULL,
    0x5c00000000000ULL,
    0x5600000000000ULL,
    0x8700000000000ULL,
    0xa380000000000ULL,
    0x22c0000000000ULL,
    0x30a0000000000ULL,
    0x9590000000000ULL,
    0xd5c8000000000ULL,
    0x1b3c000000000ULL,
    0xdb56000000000ULL,
    0xdb65000000000ULL,
    0x55a7800000000ULL,
    0xfb14c00000000ULL,
    0x6589e00000000ULL,
    0x1dc5f00000000ULL,
    0x9725980000000ULL,
    0xd540740000000ULL,
    0x686f820000000ULL,
    0xfc28c10000000ULL,
    0x8dfe38000000ULL,
    0x71a0f0c000000ULL,
    0x90219a000000ULL,
    0x5094b57000000ULL,
    0x6b46635800000ULL,
    0xf37d331c00000ULL,
    0xe9b2790e00000ULL,
    0x9d1c844b00000ULL,
    0x8a9b9b6280000ULL,
    0xe249762640000ULL,
    0xe3f6038f20000ULL,
    0x62f503b0f0000ULL,
    0x5e6f83c358000ULL,
    0xb28c25694000ULL,
    0x235fe16c1e000ULL,
    0x3f60f2d613000ULL,
    0x7a2191707800ULL,
    0x1e04b68ce7c00ULL,
    0x4b0e617d4e600ULL,
    0x4d8133e5d5d00ULL,
    0xa1c478e280080ULL,
    0xf12986e640040ULL,
    0x3a541a2f200e0ULL,
    0x97f1b6c0f0030ULL,
    0x88e1e09b58068ULL,
    0xbf69f14a9405cULL,
    0x47bb98621e056ULL,
    0xfe19769d13087ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0x6400000000000ULL,
    0xb600000000000ULL,
    0x1900000000000ULL,
    0xd680000000000ULL,
    0x36c0000000000ULL,
    0x4020000000000ULL,
    0xd090000000000ULL,
    0xa7d8000000000ULL,
    0x12bc000000000ULL,
    0x6056000000000ULL,
    0xb6f1000000000ULL,
    0x70a0800000000ULL,
    0x4758c00000000ULL,
    0xfd6e200000000ULL,
    0x16fc500000000ULL,
    0x60b7180000000ULL,
    0xdf5aa40000000ULL,
    0xe178820000000ULL,
    0x8ce4c30000000ULL,
    0xabb8238000000ULL,
    0xa6cd50c000000ULL,
    0x1837992000000ULL,
    0x2c92659000000ULL,
    0xdcea0d800000ULL,
    0x91a4936400000ULL,
    0x7dd938da00000ULL,
    0xf9a6f41b00000ULL,
    0xc9cf9a2080000ULL,
    0x37be65d240000ULL,
    0xfcc0a1c760000ULL,
    0x332993eef0000ULL,
    0x650fb87b58000ULL,
    0xbc9f3750c4000ULL,
    0x55d9393a02000ULL,
    0x6da6f4ab23000ULL,
    0xd7cf9968d8800ULL,
    0x7abe676e01400ULL,
    0xd440a2f904600ULL,
    0x78e9906fee900ULL,
    0x45afb9a080080ULL,
    0x43cf3512400c0ULL,
    0x64a138e7600e0ULL,
    0x994af67ef0030ULL,
    0x506198a358048ULL,
    0xe636434c4064ULL,
    0x636e21e0020b6ULL,
    0x9bfc50b023019ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x6800000000000ULL,
    0x7400000000000ULL,
    0xee00000000000ULL,
    0xc900000000000ULL,
    0x8a80000000000ULL,
    0x2740000000000ULL,
    0xff60000000000ULL,
    0x8c90000000000ULL,
    0xda48000000000ULL,
    0xfbec000000000ULL,
    0x32da000000000ULL,
    0xc7a5000000000ULL,
    0x2432800000000ULL,
    0xf9efc00000000ULL,
    0x79c0200000000ULL,
    0x4631d00000000ULL,
    0xe2e2280000000ULL,
    0x90574c0000000ULL,
    0xf4fa820000000ULL,
    0x6543c10000000ULL,
    0xd47a218000000ULL,
    0x5d04d14000000ULL,
    0x7498a9a000000ULL,
    0xe6548dd000000ULL,
    0x51e0a1b800000ULL,
    0x2dd7122400000ULL,
    0x982a8baa00000ULL,
    0x13fc5ddd00000ULL,
    0x6c2085d80000ULL,
    0x49b69ee240000ULL,
    0xbd30ab5120000ULL,
    0x1b688c8bb0000ULL,
    0x2a92a0c168000ULL,
    0x574e121394000ULL,
    0xa76a08f54a000ULL,
    0xc08a9e61ff000ULL,
    0x2842a8bc20800ULL,
    0xa8f18d9e77400ULL,
    0x9f5222af60a00ULL,
    0xa378d09489300ULL,
    0x768aa9dd80080ULL,
    0xad5d8fa240040ULL,
    0xd06822f120060ULL,
    0xeb0dd15bb0050ULL,
    0xf190297968068ULL,
    0x9ece4c3794074ULL,
    0x25ba015f4a0eeULL,
    0x7f3502bcff0c9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xe800000000000ULL,
    0xe400000000000ULL,
    0x1a00000000000ULL,
    0x1100000000000ULL,
    0x5380000000000ULL,
    0xb8c0000000000ULL,
    0x80e0000000000ULL,
    0xd190000000000ULL,
    0xc5d8000000000ULL,
    0xbd74000000000ULL,
    0xba56000000000ULL,
    0xba5000000000ULL,
    0x79b8800000000ULL,
    0x4cad400000000ULL,
    0xa523200000000ULL,
    0x3cecf00000000ULL,
    0x793580000000ULL,
    0x8ad9bc0000000ULL,
    0xd5e0820000000ULL,
    0xa819430000000ULL,
    0x8415228000000ULL,
    0x8a19f0c000000ULL,
    0xc913dba000000ULL,
    0x5f90ff9000000ULL,
    0x46cda26800000ULL,
    0x8be4b34400000ULL,
    0x9308fbce00000ULL,
    0x2e980e2300000ULL,
    0x8550fba380000ULL,
    0xeb2c0fd640000ULL,
    0x1fe6fbff60000ULL,
    0x1190d71d0000ULL,
    0x4b86788758000ULL,
    0x14c04e9d94000ULL,
    0x4ef358ad62000ULL,
    0x3289be60f5000ULL,
    0x63588385ec800ULL,
    0x4c3d40f163c00ULL,
    0x937b20f295600ULL,
    0x958f0b4b2f00ULL,
    0xd525592380080ULL,
    0xb4ecbf16400c0ULL,
    0x7380005f600a0ULL,
    0x48c002e1d0030ULL,
    0xc8e002ef580e8ULL,
    0x59001d9940e4ULL,
    0x37d801636201aULL,
    0x48740343f5011ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0x5400000000000ULL,
    0xbe00000000000ULL,
    0x1b00000000000ULL,
    0x7f80000000000ULL,
    0xa9c0000000000ULL,
    0xbf60000000000ULL,
    0x3d90000000000ULL,
    0xf8d8000000000ULL,
    0x55ec000000000ULL,
    0x48da000000000ULL,
    0x2de5000000000ULL,
    0x94d6800000000ULL,
    0x97e5c00000000ULL,
    0xd9c2600000000ULL,
    0xa777700000000ULL,
    0xb183f80000000ULL,
    0xaacc0c0000000ULL,
    0x4cee820000000ULL,
    0xc659c30000000ULL,
    0x5ea0618000000ULL,
    0xe6ee714000000ULL,
    0xc3577ba000000ULL,
    0xa20cd5000000ULL,
    0x88a060f800000ULL,
    0x69ee736c00000ULL,
    0x62d7787e00000ULL,
    0xe8e0cfe700000ULL,
    0xa040635d80000ULL,
    0xa9be72a640000ULL,
    0x9b6f789b60000ULL,
    0x9b9ccc7bb0000ULL,
    0x6fc2617d68000ULL,
    0x7877704094000ULL,
    0xf803fbd6da000ULL,
    0x1c0c0f85d7000ULL,
    0xda0e80da69800ULL,
    0x1d09c33d6dc00ULL,
    0xd8986260e7e00ULL,
    0x3c5273c6e4300ULL,
    0xb3b57add80080ULL,
    0xe679cfe6400c0ULL,
    0x1314e13b60060ULL,
    0xbb92b22bb0050ULL,
    0x9fc19805680e8ULL,
    0xa07b7e6c94054ULL,
    0x140d7a08da0beULL,
    0x1e05ce32d701bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x4800000000000ULL,
    0x1400000000000ULL,
    0x2a00000000000ULL,
    0x4700000000000ULL,
    0x1e80000000000ULL,
    0xf040000000000ULL,
    0x9620000000000ULL,
    0x890000000000ULL,
    0xb958000000000ULL,
    0x4dbc000000000ULL,
    0x454e000000000ULL,
    0xa3b1000000000ULL,
    0x704e800000000ULL,
    0x5623400000000ULL,
    0xe88de00000000ULL,
    0x4947500000000ULL,
    0x5ae480000000ULL,
    0x5156a40000000ULL,
    0x89b6820000000ULL,
    0x374f430000000ULL,
    0x48bbe38000000ULL,
    0x18da53c000000ULL,
    0xdf76c92000000ULL,
    0xd38e45000000ULL,
    0xe81be2a800000ULL,
    0xc40a521c00000ULL,
    0x720ecbfa00000ULL,
    0xeb14e40100000ULL,
    0x688de17880000ULL,
    0x8947537240000ULL,
    0xe5ae49cd60000ULL,
    0xa156a7eaf0000ULL,
    0xc1b6824f38000ULL,
    0x234f421fc4000ULL,
    0x62bbe0b1ba000ULL,
    0x5fda53a6cd000ULL,
    0xc1f6c81d57800ULL,
    0xfd78e782ed400ULL,
    0x7e3be1f301200ULL,
    0xcc9a51f5de900ULL,
    0xcb56caf880080ULL,
    0xa6a8e4b2400c0ULL,
    0x2dc3e2ed600e0ULL,
    0x2af650baf00f0ULL,
    0x95e0cb6738048ULL,
    0xf775e7c3c4014ULL,
    0x293b606bba02aULL,
    0x6a0810f7cd047ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x7800000000000ULL,
    0xe400000000000ULL,
    0x4200000000000ULL,
    0x5d00000000000ULL,
    0xe580000000000ULL,
    0xd8c0000000000ULL,
    0x1be0000000000ULL,
    0xb90000000000ULL,
    0xebd8000000000ULL,
    0xd74000000000ULL,
    0x65ca000000000ULL,
    0x4e77000000000ULL,
    0xdb58800000000ULL,
    0xbc2a400000000ULL,
    0xd8e0a00000000ULL,
    0x7512300000000ULL,
    0xb990580000000ULL,
    0xdeda5c0000000ULL,
    0xd4e0820000000ULL,
    0xcb0e430000000ULL,
    0x3292a28000000ULL,
    0x8c41334000000ULL,
    0x28bad9e000000ULL,
    0x4ba31f9000000ULL,
    0xe52aa30800000ULL,
    0xcf65327400000ULL,
    0xf8c8d91600000ULL,
    0xbf01c2300000ULL,
    0xd380238f80000ULL,
    0xdfdc70be40000ULL,
    0x3762f82760000ULL,
    0xdccb6e01d0000ULL,
    0xe9e259e128000ULL,
    0x5e895ecadc000ULL,
    0x424a028ae2000ULL,
    0xbb703eae9000ULL,
    0x8538803ae2800ULL,
    0xbf7a425298c00ULL,
    0x50d8a3f6e9600ULL,
    0x97f63088f5700ULL,
    0x7582590f80080ULL,
    0xc0d95ffe400c0ULL,
    0x8ff203c7600a0ULL,
    0xe1930291d00d0ULL,
    0x2aca806928078ULL,
    0x4ee943fedc0e4ULL,
    0xa212207ce2042ULL,
    0xed1f7359e905dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0xac00000000000ULL,
    0xfe00000000000ULL,
    0xbf00000000000ULL,
    0x2180000000000ULL,
    0x2c40000000000ULL,
    0x8620000000000ULL,
    0x5e90000000000ULL,
    0x6dc8000000000ULL,
    0x7a64000000000ULL,
    0xa0ae000000000ULL,
    0x1749000000000ULL,
    0x45b9800000000ULL,
    0x9bd5c00000000ULL,
    0x2970e00000000ULL,
    0x5f34b00000000ULL,
    0x7409c80000000ULL,
    0x6a05e40000000ULL,
    0x4517820000000ULL,
    0x3c9cc10000000ULL,
    0xc4c9608000000ULL,
    0x98e173c000000ULL,
    0x2b792a2000000ULL,
    0x263156b000000ULL,
    0xee9e49f800000ULL,
    0xc5d927fc00000ULL,
    0x267ee20600000ULL,
    0xd6adb27100000ULL,
    0x45848b880000ULL,
    0x9a24240a40000ULL,
    0x881626f20000ULL,
    0x8ec570a590000ULL,
    0x2df7297cb8000ULL,
    0x9fe855d024000ULL,
    0x4eefcba866000ULL,
    0x8868e41417000ULL,
    0x51a0027263800ULL,
    0x21d001d302c00ULL,
    0x146800c3bf200ULL,
    0x47b401dda3900ULL,
    0xe2c6014080080ULL,
    0xb3fd03f640040ULL,
    0xf0ff806920020ULL,
    0x1768c2d4900f0ULL,
    0x802f61c438088ULL,
    0x858c71da640acULL,
    0x3e4ea9c7460feULL,
    0xe73d94b1870bfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0x1c00000000000ULL,
    0x2a00000000000ULL,
    0xc700000000000ULL,
    0x2580000000000ULL,
    0x4940000000000ULL,
    0xc960000000000ULL,
    0xf290000000000ULL,
    0x44c8000000000ULL,
    0xf424000000000ULL,
    0x45ee000000000ULL,
    0xff4f000000000ULL,
    0x646f800000000ULL,
    0x300f400000000ULL,
    0xd819e00000000ULL,
    0xcc14300000000ULL,
    0x6211e80000000ULL,
    0x63138c0000000ULL,
    0x1381820000000ULL,
    0xa440410000000ULL,
    0x2bf6608000000ULL,
    0x9e5b73c000000ULL,
    0xc4e80ae000000ULL,
    0xcfd7bc7000000ULL,
    0xf3b86aa800000ULL,
    0x4fa7ce1c00000ULL,
    0xd5b1e21600000ULL,
    0x8aa033e500000ULL,
    0x4337e94580000ULL,
    0x98788c7a40000ULL,
    0x8a00015b20000ULL,
    0x770001bc90000ULL,
    0xbd8001a9b8000ULL,
    0xa54002043c000ULL,
    0x5b6000c23e000ULL,
    0x2990025f7d000ULL,
    0x4b4801fec7800ULL,
    0x7a64003680c00ULL,
    0xa90e011adfa00ULL,
    0x449f0084e2300ULL,
    0xe9c783ed80080ULL,
    0x36bb426640040ULL,
    0xd93fe34d20020ULL,
    0xc77f3259900f0ULL,
    0x439068ec380b8ULL,
    0xac53ce7e7c01cULL,
    0xaff7e1991e02aULL,
    0x585b33e3ed0c7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x7800000000000ULL,
    0x9c00000000000ULL,
    0x2a00000000000ULL,
    0x9500000000000ULL,
    0x2080000000000ULL,
    0xb940000000000ULL,
    0x27e0000000000ULL,
    0xb990000000000ULL,
    0xebd8000000000ULL,
    0xa7bc000000000ULL,
    0x3cea000000000ULL,
    0x9a07000000000ULL,
    0xfd17800000000ULL,
    0x2483400000000ULL,
    0xdf47a00000000ULL,
    0xe0f8900000000ULL,
    0x9011e80000000ULL,
    0x580b1c0000000ULL,
    0xac1d820000000ULL,
    0x8214430000000ULL,
    0x7108238000000ULL,
    0x9687d34000000ULL,
    0x65c49e000000ULL,
    0x92648e7000000ULL,
    0x2043eaa800000ULL,
    0x75601d5400000ULL,
    0x39d8010200000ULL,
    0x6ebc02a500000ULL,
    0xd66a02ff80000ULL,
    0x664703d640000ULL,
    0x827782e760000ULL,
    0xb85342baf0000ULL,
    0x397fa0d9a8000ULL,
    0x6bd492d91c000ULL,
    0x67a3ebe9de000ULL,
    0xdcf01c914d000ULL,
    0x4a00024dfe800ULL,
    0x850000fb52400ULL,
    0xb880029c8fa00ULL,
    0xf54002d6c0700ULL,
    0x75e0005780080ULL,
    0xb0900282400c0ULL,
    0xe15801e5600e0ULL,
    0x8bfc031ff00d0ULL,
    0x3b8a01a628078ULL,
    0x9ad7024f5c09cULL,
    0x312f80eebe02aULL,
    0x3aaf405bbd095ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xb800000000000ULL,
    0x3400000000000ULL,
    0xca00000000000ULL,
    0x5900000000000ULL,
    0x8a80000000000ULL,
    0x81c0000000000ULL,
    0x58e0000000000ULL,
    0x3790000000000ULL,
    0x958000000000ULL,
    0xec3c000000000ULL,
    0xabe6000000000ULL,
    0x941f000000000ULL,
    0x5a02800000000ULL,
    0xb113400000000ULL,
    0xb68a200000000ULL,
    0xc7c0d00000000ULL,
    0xffeff80000000ULL,
    0x2e097c0000000ULL,
    0x5b04820000000ULL,
    0xbf9c430000000ULL,
    0x4550a38000000ULL,
    0xf22f92c000000ULL,
    0x88e3dae000000ULL,
    0xff86acd000000ULL,
    0x6551f92800000ULL,
    0xa22a7e6400000ULL,
    0x80e003aa00000ULL,
    0x739003c700000ULL,
    0x9b58000380000ULL,
    0x313c02ce40000ULL,
    0x536601ed60000ULL,
    0x78df0204f0000ULL,
    0x426283ad98000ULL,
    0x5e4341337c000ULL,
    0x6db222218a000ULL,
    0x9dacd01d0d000ULL,
    0x5b1fb9cc8800ULL,
    0x61ba7c71b3400ULL,
    0xa3b8037547e00ULL,
    0x76ac0258a9f00ULL,
    0x23e012b80080ULL,
    0x10e300aa400c0ULL,
    0x9b848047600e0ULL,
    0xa75c42c3f00b0ULL,
    0x7730a02e180b8ULL,
    0xf47f913d3c034ULL,
    0x615bd92cea0caULL,
    0x102aaec9fd059ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x9800000000000ULL,
    0x6c00000000000ULL,
    0xaa00000000000ULL,
    0xcb00000000000ULL,
    0xdc80000000000ULL,
    0x1840000000000ULL,
    0xece0000000000ULL,
    0x9390000000000ULL,
    0x5ad8000000000ULL,
    0x19bc000000000ULL,
    0x32ee000000000ULL,
    0x5e9b000000000ULL,
    0x749800000000ULL,
    0xe677400000000ULL,
    0x82c4600000000ULL,
    0x55bd100000000ULL,
    0x88f7c80000000ULL,
    0xfd87bc0000000ULL,
    0x2fc7820000000ULL,
    0x383c430000000ULL,
    0xf35e38000000ULL,
    0xd1a653c000000ULL,
    0x16e5aa6000000ULL,
    0x108dadb000000ULL,
    0x624fc8a800000ULL,
    0xafebbc2c00000ULL,
    0x4b1182f200000ULL,
    0x1c8b40a100000ULL,
    0xf84a625380000ULL,
    0x1cf6103e40000ULL,
    0xb8649a360000ULL,
    0x36dcfefaf0000ULL,
    0xb3b5e311b8000ULL,
    0xf9e651376c000ULL,
    0x8205ab5ca6000ULL,
    0x1f1dae769d000ULL,
    0xa97c8b3f1800ULL,
    0x1157bdce44400ULL,
    0xf7f80ab07200ULL,
    0x9150410682f00ULL,
    0xcf63e2fb80080ULL,
    0x71515012400c0ULL,
    0x3f7a2951600e0ULL,
    0xe94ded5bf00f0ULL,
    0x53742ac238098ULL,
    0x4346eec92c06cULL,
    0x9865aa9fc60aaULL,
    0x9fcdae3c6d0cbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xe800000000000ULL,
    0x6400000000000ULL,
    0x2a00000000000ULL,
    0x9b00000000000ULL,
    0x580000000000ULL,
    0x2fc0000000000ULL,
    0x18a0000000000ULL,
    0xd490000000000ULL,
    0x6158000000000ULL,
    0xcae4000000000ULL,
    0x5ff6000000000ULL,
    0xb75000000000ULL,
    0xaa20800000000ULL,
    0x40cf400000000ULL,
    0x2f34600000000ULL,
    0xcc43300000000ULL,
    0x26f380000000ULL,
    0x68b0b40000000ULL,
    0xdc8e820000000ULL,
    0x155e430000000ULL,
    0x38e2e08000000ULL,
    0x48f970c000000ULL,
    0x40fbdba000000ULL,
    0x34fcc59000000ULL,
    0xc6f5daa800000ULL,
    0xd1fdc56c00000ULL,
    0x9a7b5a9600000ULL,
    0x4a3877f00000ULL,
    0x8299b94280000ULL,
    0x7c5af50240000ULL,
    0x2a62600d60000ULL,
    0x2ca6331790000ULL,
    0xc697bae1d8000ULL,
    0x665bf7aed4000ULL,
    0x596ce3d402000ULL,
    0x4d3872127d000ULL,
    0xc355596531800ULL,
    0xe5f2855bdcc00ULL,
    0x886f392784e00ULL,
    0x3b0b7d9c6d00ULL,
    0x110e836a80080ULL,
    0x6e9e40ae400c0ULL,
    0xe242e13b60020ULL,
    0x636971f890030ULL,
    0xe23d90b580e8ULL,
    0x4ad8c7c094064ULL,
    0x8423d94f6202aULL,
    0x21d8c67aed09bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0xd800000000000ULL,
    0x2c00000000000ULL,
    0xa200000000000ULL,
    0x6500000000000ULL,
    0xe480000000000ULL,
    0xa8c0000000000ULL,
    0xd2e0000000000ULL,
    0x2b90000000000ULL,
    0x9d48000000000ULL,
    0xf0bc000000000ULL,
    0xc6ba000000000ULL,
    0x3da3000000000ULL,
    0x2621800000000ULL,
    0x737a400000000ULL,
    0x97d9a00000000ULL,
    0x8f7c100000000ULL,
    0xadd9d80000000ULL,
    0x267e3c0000000ULL,
    0xbb53820000000ULL,
    0x33a5410000000ULL,
    0xa122238000000ULL,
    0x12f5514000000ULL,
    0x8b89fb6000000ULL,
    0x2d546cb000000ULL,
    0x78a1f88800000ULL,
    0x32b86c9400000ULL,
    0xb3b3fa1200000ULL,
    0xe1376ee300000ULL,
    0xf2e079ab80000ULL,
    0xdb922c5e40000ULL,
    0xf5425a1d20000ULL,
    0x54a77ea6f0000ULL,
    0x90aba0e0e8000ULL,
    0xd6a313318c000ULL,
    0x5a25a2906000ULL,
    0x5a377fa4a9000ULL,
    0x963a31bc6800ULL,
    0xdedf13d640400ULL,
    0xc9f85a092f600ULL,
    0x60047d74c4f00ULL,
    0x100a22a380080ULL,
    0x3819518a40040ULL,
    0x7c1bfb6f200e0ULL,
    0x7a1b6cf5f0050ULL,
    0x491279c3680d8ULL,
    0x468d2ffbcc02cULL,
    0xcdd9d826260a2ULL,
    0x367e3ee159065ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xc800000000000ULL,
    0x1400000000000ULL,
    0x8200000000000ULL,
    0xc100000000000ULL,
    0x1480000000000ULL,
    0x8dc0000000000ULL,
    0x61a0000000000ULL,
    0xe390000000000ULL,
    0x4a58000000000ULL,
    0xd1e4000000000ULL,
    0x5f7a000000000ULL,
    0xfb1000000000ULL,
    0xcc8e800000000ULL,
    0xb1c6400000000ULL,
    0x87af200000000ULL,
    0x3883100000000ULL,
    0x43c2780000000ULL,
    0xdea2d40000000ULL,
    0xf00c820000000ULL,
    0x5803430000000ULL,
    0xfc03a08000000ULL,
    0xc610514000000ULL,
    0x8b19db2000000ULL,
    0xc196845000000ULL,
    0x1b4f580800000ULL,
    0x2d74c40400000ULL,
    0x96ba7ad200000ULL,
    0x2416d47700000ULL,
    0xfa0e802680000ULL,
    0x6d06419e40000ULL,
    0x1a8f208160000ULL,
    0x12d3115390000ULL,
    0x223a7807e8000ULL,
    0x39d6d55dc4000ULL,
    0x73ae82eeba000ULL,
    0xca96437a59000ULL,
    0x1ad722635c800ULL,
    0x1637135cdc400ULL,
    0xebc07bfd01e00ULL,
    0x7aa7d7dd9f500ULL,
    0xca0000ae80080ULL,
    0x150000da400c0ULL,
    0xb680037360020ULL,
    0x1cc0027490050ULL,
    0xbd2000a9680c8ULL,
    0x7a50018784014ULL,
    0xa9f8039dda082ULL,
    0xf374020ec90c1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x3c00000000000ULL,
    0xe200000000000ULL,
    0x4d00000000000ULL,
    0xcd80000000000ULL,
    0xadc0000000000ULL,
    0x8ae0000000000ULL,
    0x6690000000000ULL,
    0x4358000000000ULL,
    0x6da4000000000ULL,
    0xf53a000000000ULL,
    0x4963000000000ULL,
    0xfc4b800000000ULL,
    0x552a400000000ULL,
    0xd97ca00000000ULL,
    0x8457300000000ULL,
    0x3939780000000ULL,
    0x637d3c0000000ULL,
    0xf549820000000ULL,
    0x16bd430000000ULL,
    0x83b5208000000ULL,
    0xb22a714000000ULL,
    0xddec596000000ULL,
    0x4a074cf000000ULL,
    0xb91dd98800000ULL,
    0x6b8e0e3400000ULL,
    0x6ecaf9b600000ULL,
    0xb0637cf700000ULL,
    0x77d721cb80000ULL,
    0x4bed732a40000ULL,
    0x611ddb6560000ULL,
    0x978e0d3290000ULL,
    0xaccaf80ae8000ULL,
    0xad637d568c000ULL,
    0xe25723e4ae000ULL,
    0xda2d710de9000ULL,
    0x9fdda9512800ULL,
    0xbc1e0d3a8cc00ULL,
    0x2212f8feede00ULL,
    0x6d077e40a8f00ULL,
    0x9d8d20c380080ULL,
    0xf5de705e400c0ULL,
    0xb6ee59b360020ULL,
    0x84904e3590050ULL,
    0xe5458c968058ULL,
    0xa0334d08cc03cULL,
    0x58ffda57ce0e2ULL,
    0xc3890f387904dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x5800000000000ULL,
    0xd400000000000ULL,
    0xee00000000000ULL,
    0xab00000000000ULL,
    0x1b80000000000ULL,
    0x4a40000000000ULL,
    0x3fa0000000000ULL,
    0xca50000000000ULL,
    0x7fa8000000000ULL,
    0xaa5c000000000ULL,
    0xefba000000000ULL,
    0xf257000000000ULL,
    0x3ba0800000000ULL,
    0x1c4ac00000000ULL,
    0x90b5e00000000ULL,
    0x7c9b00000000ULL,
    0xdafca80000000ULL,
    0x386e440000000ULL,
    0x10b5e20000000ULL,
    0x47c9b10000000ULL,
    0xbafca98000000ULL,
    0xa86e464000000ULL,
    0x48b5e36000000ULL,
    0x93c9b25000000ULL,
    0x54fcaa3800000ULL,
    0x36e44ec00000ULL,
    0x5335e30e00000ULL,
    0xd989b37900000ULL,
    0x6b5caac680000ULL,
    0xc93e47c540000ULL,
    0x2c9de2f0a0000ULL,
    0x73d5b1d070000ULL,
    0x84e6a97868000ULL,
    0x3b69440c1c000ULL,
    0x173d621e22000ULL,
    0x6f9f71a15b000ULL,
    0x14534b3abf800ULL,
    0x3ca0f4133ac00ULL,
    0xcdc1c975d0a00ULL,
    0x57f13540e2100ULL,
    0x4e6a97868080ULL,
    0x7b69440c1c040ULL,
    0x773d621e22060ULL,
    0xff9f71a15b090ULL,
    0x4c534b3abf858ULL,
    0xe8a0f4133acd4ULL,
    0x23c1c975d0aeeULL,
    0xfcf13540e21abULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x5800000000000ULL,
    0x9c00000000000ULL,
    0xe200000000000ULL,
    0x8b00000000000ULL,
    0x5280000000000ULL,
    0x56c0000000000ULL,
    0x4a60000000000ULL,
    0xb6d0000000000ULL,
    0x3a68000000000ULL,
    0xded4000000000ULL,
    0xee6a000000000ULL,
    0xf8df000000000ULL,
    0x1b79800000000ULL,
    0xc343400000000ULL,
    0x9428e00000000ULL,
    0x8d69100000000ULL,
    0x3e52380000000ULL,
    0x4bb85c0000000ULL,
    0x6c28e20000000ULL,
    0x4169110000000ULL,
    0xa452388000000ULL,
    0x4cb85c4000000ULL,
    0x84a8e36000000ULL,
    0xa9137000000ULL,
    0x5eb23b0800000ULL,
    0x27a85e6c00000ULL,
    0xa620e22a00000ULL,
    0x3e6d122b00000ULL,
    0xc0d03a2180000ULL,
    0xb7735cb740000ULL,
    0x695b62c3a0000ULL,
    0xdb25515050000ULL,
    0xa1eb599828000ULL,
    0x1860f543c000ULL,
    0xd8583aae46000ULL,
    0xdeb75e5d5d000ULL,
    0x67b961c88b800ULL,
    0x863e506198400ULL,
    0x2e70d8d70ee00ULL,
    0x98de4e33bc700ULL,
    0x2b6b599828080ULL,
    0x8b460f543c040ULL,
    0x50383aae46020ULL,
    0xf3675e5d5d010ULL,
    0x575161c88b858ULL,
    0x922a50619849cULL,
    0x687ad8d70eee2ULL,
    0x5dd14e33bc78bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0x4400000000000ULL,
    0xca00000000000ULL,
    0xd00000000000ULL,
    0x2880000000000ULL,
    0x5140000000000ULL,
    0xd8a0000000000ULL,
    0x5950000000000ULL,
    0xbcb8000000000ULL,
    0xc34c000000000ULL,
    0x49ae000000000ULL,
    0xe7c5000000000ULL,
    0x96e6800000000ULL,
    0xf87c400000000ULL,
    0xea37200000000ULL,
    0x3d01500000000ULL,
    0xa091080000000ULL,
    0xf544440000000ULL,
    0xa2b7220000000ULL,
    0x1c41530000000ULL,
    0xd0310b8000000ULL,
    0x581446c000000ULL,
    0x9c0f232000000ULL,
    0x960d521000000ULL,
    0x7b1f08a800000ULL,
    0xe39146f400000ULL,
    0xfac9a38200000ULL,
    0x6661135500000ULL,
    0xf5302bca80000ULL,
    0x448c179140000ULL,
    0xaf4ea970e0000ULL,
    0xb7ac545830000ULL,
    0x88cf88ec38000ULL,
    0x4774070e54000ULL,
    0x3ae00ab7a000ULL,
    0x2ac50179c1000ULL,
    0x5e668064e4800ULL,
    0x193c41ea51400ULL,
    0x7a9720813e200ULL,
    0x20515358d0100ULL,
    0xd62908ec38080ULL,
    0x3b08470e540c0ULL,
    0xc39920ab7a0e0ULL,
    0xaac45179c10b0ULL,
    0x9e778864e4848ULL,
    0xf93805ea51444ULL,
    0xca8002813e2caULL,
    0x68400058d010dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0xac00000000000ULL,
    0xe600000000000ULL,
    0x900000000000ULL,
    0x3880000000000ULL,
    0xe2c0000000000ULL,
    0x50a0000000000ULL,
    0xead0000000000ULL,
    0x2cb8000000000ULL,
    0x84d4000000000ULL,
    0x99ba000000000ULL,
    0xf241000000000ULL,
    0xde6f800000000ULL,
    0x7c7dc00000000ULL,
    0xdf7ea00000000ULL,
    0x38e2d00000000ULL,
    0x53a6f80000000ULL,
    0xdd4ac40000000ULL,
    0xcffea20000000ULL,
    0xb622d30000000ULL,
    0xc506f88000000ULL,
    0x2e9ac44000000ULL,
    0x73c6a0a000000ULL,
    0x7c36d1b000000ULL,
    0xea1cfb1800000ULL,
    0x3f0bc46400000ULL,
    0xb991204200000ULL,
    0x665f123b00000ULL,
    0xfc785a5a80000ULL,
    0x1f7817cf40000ULL,
    0x18e058f0e0000ULL,
    0x43bc142850000ULL,
    0x75425a3c68000ULL,
    0x63f9170644000ULL,
    0x502fdb095e000ULL,
    0xcc11d599a7000ULL,
    0x1604fbe192800ULL,
    0x910fc555cf400ULL,
    0x2c9321dfc5e00ULL,
    0xca12c88c100ULL,
    0x13adda3c68080ULL,
    0x3d44d706440c0ULL,
    0xfff17b095e020ULL,
    0xe230599a7010ULL,
    0xc11a03e1928a8ULL,
    0x64910155cf4acULL,
    0x9cd783dfc5ee6ULL,
    0x4da9c1c88c109ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0x6400000000000ULL,
    0xea00000000000ULL,
    0xc500000000000ULL,
    0x4f80000000000ULL,
    0x75c0000000000ULL,
    0x3b60000000000ULL,
    0xa1d0000000000ULL,
    0xd968000000000ULL,
    0xe0cc000000000ULL,
    0x4cee000000000ULL,
    0x1805000000000ULL,
    0x5c02800000000ULL,
    0x2608400000000ULL,
    0x7b12200000000ULL,
    0x289db00000000ULL,
    0x9b5a980000000ULL,
    0xeb33dc0000000ULL,
    0x2a72220000000ULL,
    0xc4db10000000ULL,
    0xedb29b8000000ULL,
    0x4e3fdcc000000ULL,
    0x15fc232000000ULL,
    0xd198b09000000ULL,
    0x82d8182800000ULL,
    0x4dfb9fd400000ULL,
    0x6d80021e00000ULL,
    0x94c0014700000ULL,
    0x7ee000c580000ULL,
    0x2110015340000ULL,
    0xe588037ba0000ULL,
    0x50dc03c430000ULL,
    0x44e6027638000ULL,
    0x9c1901f354000ULL,
    0x8604832baa000ULL,
    0xab0142cc11000ULL,
    0x509ea03270800ULL,
    0xb740f24522c00ULL,
    0x65223878c0600ULL,
    0x56a2cb7de700ULL,
    0x86d49a7638080ULL,
    0xd7e6ddf354040ULL,
    0x98a12baa0e0ULL,
    0x8f49f3cc11030ULL,
    0xa92ebbb270848ULL,
    0xbb776e8522c64ULL,
    0xe1cc3b58c06eaULL,
    0x396f2c27de7c5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x5800000000000ULL,
    0x5400000000000ULL,
    0x7200000000000ULL,
    0xcf00000000000ULL,
    0xf280000000000ULL,
    0x9940000000000ULL,
    0xcfa0000000000ULL,
    0xfb50000000000ULL,
    0x98b8000000000ULL,
    0x7dd4000000000ULL,
    0xe3ea000000000ULL,
    0x257f000000000ULL,
    0xbe30800000000ULL,
    0x5689400000000ULL,
    0xc35d600000000ULL,
    0xbcad100000000ULL,
    0xe7c9480000000ULL,
    0xb0f0640000000ULL,
    0xa9e5620000000ULL,
    0xce79130000000ULL,
    0xd6a3488000000ULL,
    0x9ccf664000000ULL,
    0x8075e36000000ULL,
    0x37a0525000000ULL,
    0xff46294800000ULL,
    0x92b6757c00000ULL,
    0x76d6a8aa00000ULL,
    0x3b6f343500000ULL,
    0x2733c87680000ULL,
    0xf116259140000ULL,
    0xfb908248e0000ULL,
    0x66d9438250000ULL,
    0xa365609928000ULL,
    0x53391354bc000ULL,
    0x13034bf822000ULL,
    0x6c9f64a475000ULL,
    0xc04de3be5d800ULL,
    0x483453d308400ULL,
    0x538c2b7187200ULL,
    0x1ad975a6f4900ULL,
    0x6d7e281928080ULL,
    0x72227514bc0c0ULL,
    0x509ca89822020ULL,
    0xee4036f475090ULL,
    0xd93b4af65d858ULL,
    0xd80b66af08454ULL,
    0x9407e3db87272ULL,
    0x521b5193f49cfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xd800000000000ULL,
    0xdc00000000000ULL,
    0x6200000000000ULL,
    0xdf00000000000ULL,
    0x2c80000000000ULL,
    0x1540000000000ULL,
    0xbe60000000000ULL,
    0x2b50000000000ULL,
    0x4368000000000ULL,
    0x38cc000000000ULL,
    0xaa6000000000ULL,
    0x4bf7000000000ULL,
    0x438d800000000ULL,
    0x41c3400000000ULL,
    0xe732600000000ULL,
    0x5f2cd00000000ULL,
    0xb321c80000000ULL,
    0xe9219c0000000ULL,
    0x9a3a620000000ULL,
    0xcb0d10000000ULL,
    0x1aefcb8000000ULL,
    0xea1a9dc000000ULL,
    0xb11e16000000ULL,
    0xda84927000000ULL,
    0xdc502a0800000ULL,
    0x2bf50ebc00000ULL,
    0x738249d200000ULL,
    0x79c9de2500000ULL,
    0x4b2b82f180000ULL,
    0xe534431140000ULL,
    0xb03fe15fa0000ULL,
    0xa7af920630000ULL,
    0x5073a83b18000ULL,
    0x9e5d4d8e9c000ULL,
    0x24f3ab3996000ULL,
    0x171d4fcd3d000ULL,
    0x1893a87dd1800ULL,
    0x934d4c6b2f400ULL,
    0xaf7baad691200ULL,
    0x62c14d5dfb700ULL,
    0x79bdabbb18080ULL,
    0xdd664c4e9c040ULL,
    0x55d82859960e0ULL,
    0xb1290cbd3d070ULL,
    0x62c49f5d18d8ULL,
    0x8ea2df172f4dcULL,
    0xb5e8026491262ULL,
    0x1e8c0108fb7dfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0xa400000000000ULL,
    0x5a00000000000ULL,
    0x3300000000000ULL,
    0xdf80000000000ULL,
    0x4ac0000000000ULL,
    0xa960000000000ULL,
    0xa4d0000000000ULL,
    0x2868000000000ULL,
    0xd444000000000ULL,
    0x732e000000000ULL,
    0xc6f7000000000ULL,
    0x360a800000000ULL,
    0x8504400000000ULL,
    0x1a80e00000000ULL,
    0xf045b00000000ULL,
    0x6928b80000000ULL,
    0x55e11c0000000ULL,
    0xd988e20000000ULL,
    0x57d1b10000000ULL,
    0x17eeba8000000ULL,
    0xe921cc000000ULL,
    0x724c606000000ULL,
    0x5e32f39000000ULL,
    0xe06cdbe800000ULL,
    0x3847ec0c00000ULL,
    0x85223b1e00000ULL,
    0xa3e55ebb00000ULL,
    0xbc88034d80000ULL,
    0xdd54039f40000ULL,
    0x4fa6033fa0000ULL,
    0x5ba3002a10000ULL,
    0xd9ac806138000ULL,
    0xeea743d49c000ULL,
    0x5b2c626f8a000ULL,
    0xbae2f0a201000ULL,
    0x6804d87d3b800ULL,
    0xdc03ef628ac00ULL,
    0x6e0c3978a8e00ULL,
    0xc1125f81c5700ULL,
    0xd08282e138080ULL,
    0x6b5043149c040ULL,
    0x8aa6e00f8a0a0ULL,
    0xe126b23201030ULL,
    0x19e439153b898ULL,
    0x1f965fae8aca4ULL,
    0xaacc8206a8e5aULL,
    0x397742aac5733ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x800000000000ULL,
    0x8400000000000ULL,
    0xea00000000000ULL,
    0x8f00000000000ULL,
    0x9c80000000000ULL,
    0x2ec0000000000ULL,
    0x8620000000000ULL,
    0x18d0000000000ULL,
    0xaf38000000000ULL,
    0x4554000000000ULL,
    0xa862000000000ULL,
    0xf635000000000ULL,
    0xc2800000000ULL,
    0x320400000000ULL,
    0xfb4f600000000ULL,
    0xc56dd00000000ULL,
    0x61b1b80000000ULL,
    0x589c940000000ULL,
    0x24d7620000000ULL,
    0xf929d30000000ULL,
    0x5c4bb88000000ULL,
    0xded974000000ULL,
    0xad6fe22000000ULL,
    0xd5b8911000000ULL,
    0x4a9c5b2800000ULL,
    0xf7c1057c00000ULL,
    0x3b1b85200000ULL,
    0x939c95ab00000ULL,
    0x7257633080000ULL,
    0x88e9d21f40000ULL,
    0x4eebb86ee0000ULL,
    0xbffd94fe50000ULL,
    0x6e77e13108000ULL,
    0x73c909794000ULL,
    0xd1465be5ea000ULL,
    0x6a6004ded1000ULL,
    0x2d313b0635800ULL,
    0x7e59d50523400ULL,
    0x26e28009ace00ULL,
    0xbf04264e3500ULL,
    0x7c7761b108080ULL,
    0xd439d3d7940c0ULL,
    0x2bd3bbc5ea020ULL,
    0xa5a996ced10d0ULL,
    0x5295e0ae35808ULL,
    0x5bc9933923484ULL,
    0xbda4d87baceeaULL,
    0xfe9045dfe358fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x2800000000000ULL,
    0x2c00000000000ULL,
    0x8200000000000ULL,
    0x6100000000000ULL,
    0xbc80000000000ULL,
    0x5e40000000000ULL,
    0xbba0000000000ULL,
    0x8850000000000ULL,
    0x8ca8000000000ULL,
    0x43cc000000000ULL,
    0xf966000000000ULL,
    0xfda9000000000ULL,
    0xe753800000000ULL,
    0x4321400000000ULL,
    0x980ba00000000ULL,
    0xa404d00000000ULL,
    0x3e1b680000000ULL,
    0xbb1b240000000ULL,
    0xd983a20000000ULL,
    0x4cd8d10000000ULL,
    0x6f56b8000000ULL,
    0xee6e25c000000ULL,
    0xe63e22a000000ULL,
    0x2a8c91b000000ULL,
    0x8943498800000ULL,
    0x3eb44400000ULL,
    0xd593685200000ULL,
    0x1ec724c900000ULL,
    0xdfeda16680000ULL,
    0x26edd36540000ULL,
    0x3e68ebe0a0000ULL,
    0x5e2a650630000ULL,
    0x5e8001a318000ULL,
    0xf4001e3e4000ULL,
    0xcf2002d5ee000ULL,
    0x8a10013eb5000ULL,
    0x9d08013936800ULL,
    0x869c028ef7400ULL,
    0x4b4e01b103a00ULL,
    0x8125014a99900ULL,
    0x1915822318080ULL,
    0x68984023e4040ULL,
    0x48502275ee0e0ULL,
    0x2cb9918eb5070ULL,
    0xd3decb3136828ULL,
    0xa17af70af742cULL,
    0xf9ad494303a82ULL,
    0x494bb53399961ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0x8c00000000000ULL,
    0xbe00000000000ULL,
    0x4100000000000ULL,
    0x3180000000000ULL,
    0x5c0000000000ULL,
    0x9ae0000000000ULL,
    0xe4d0000000000ULL,
    0x7b78000000000ULL,
    0x6914000000000ULL,
    0x7d9a000000000ULL,
    0x9bd1000000000ULL,
    0xcbf3800000000ULL,
    0x7d52c00000000ULL,
    0xf2a1a00000000ULL,
    0x4de5f00000000ULL,
    0xd842980000000ULL,
    0xd13c6c0000000ULL,
    0xa73ba20000000ULL,
    0x9a34f30000000ULL,
    0x8db1188000000ULL,
    0xfd6eac4000000ULL,
    0xcc1a00a000000ULL,
    0x5e11013000000ULL,
    0x7113827800000ULL,
    0x8982c14400000ULL,
    0x21d9a26600000ULL,
    0xa8f1f22700000ULL,
    0x1bd8989380000ULL,
    0xbed6e9740000ULL,
    0x5d48232be0000ULL,
    0xe2a632b350000ULL,
    0xe5f0bb1de8000ULL,
    0x545b5dbc04000ULL,
    0x6f209a622e000ULL,
    0xe6396d611b000ULL,
    0xabb22244ee800ULL,
    0x8867311cd3c00ULL,
    0x679b3aa8c4600ULL,
    0x28cd9cd6bab00ULL,
    0x25633965e8080ULL,
    0x18199cf8040c0ULL,
    0xf41938042e020ULL,
    0xba189f461b010ULL,
    0x6312bad76e8a8ULL,
    0x669e5f8b93c8cULL,
    0xf9491983246beULL,
    0x10baae65eab41ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x1800000000000ULL,
    0x9400000000000ULL,
    0xbe00000000000ULL,
    0xa700000000000ULL,
    0x3980000000000ULL,
    0x6a40000000000ULL,
    0x6c60000000000ULL,
    0x5550000000000ULL,
    0x81e8000000000ULL,
    0x2104000000000ULL,
    0xda9a000000000ULL,
    0x55d7000000000ULL,
    0xa520800000000ULL,
    0xa6a4c00000000ULL,
    0x8dfe600000000ULL,
    0x4b1cf00000000ULL,
    0x6397a80000000ULL,
    0x5b4fbc0000000ULL,
    0xeee4620000000ULL,
    0x348bf10000000ULL,
    0xad72a8000000ULL,
    0x38bb7e4000000ULL,
    0xfaf2026000000ULL,
    0xca93035000000ULL,
    0xdda807800000ULL,
    0x9123c0dc00000ULL,
    0x88b6e08600000ULL,
    0x32fc33f900000ULL,
    0xe693cb4980000ULL,
    0xb7d44fc940000ULL,
    0x903b48e1a0000ULL,
    0x82248c2d10000ULL,
    0xdf372a5be8000ULL,
    0x47ab7f421c000ULL,
    0xb77a007322000ULL,
    0x2ec7030e83000ULL,
    0x4ea8814591800ULL,
    0x71f0c2f72fc00ULL,
    0x490c6164dca00ULL,
    0x668ff292adf00ULL,
    0xf7cd2a23e8080ULL,
    0x302c7f9e1c040ULL,
    0x123280f5220a0ULL,
    0xc727c0f783090ULL,
    0xd3ace20c11818ULL,
    0x96b313e6fc94ULL,
    0x89d34b857cabeULL,
    0x77208fbfbdfa7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x800000000000ULL,
    0x9400000000000ULL,
    0x3600000000000ULL,
    0xbd00000000000ULL,
    0x2880000000000ULL,
    0xa9c0000000000ULL,
    0x60a0000000000ULL,
    0x2cd0000000000ULL,
    0xc438000000000ULL,
    0x4f1c000000000ULL,
    0x6b82000000000ULL,
    0xc84d000000000ULL,
    0x7e8800000000ULL,
    0x50b7c00000000ULL,
    0x34cd200000000ULL,
    0x3837d00000000ULL,
    0x3d04980000000ULL,
    0xe88a040000000ULL,
    0xc9cf220000000ULL,
    0xb0bad30000000ULL,
    0x24cc198000000ULL,
    0x502dc74000000ULL,
    0x791a022000000ULL,
    0xd681015000000ULL,
    0xe0d2815800000ULL,
    0xae26c1b400000ULL,
    0x3007a08200000ULL,
    0x181d10f700000ULL,
    0xfc19395a80000ULL,
    0x7216164740000ULL,
    0x830499b2e0000ULL,
    0x18a049b70000ULL,
    0xb74f23ac88000ULL,
    0x747ad3d274000ULL,
    0x64ec1aaf42000ULL,
    0x413dc56eaf000ULL,
    0xeb8203853c800ULL,
    0x84d0065eb400ULL,
    0x67e88111b0600ULL,
    0x80b7c1f3f7100ULL,
    0x3ccd22f488080ULL,
    0xac37d266740c0ULL,
    0xb049a2d42060ULL,
    0x558a0599af0d0ULL,
    0xe14f22dfbc808ULL,
    0x197ad222ab494ULL,
    0x446c1aa350636ULL,
    0x7cfdc668871bdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x800000000000ULL,
    0xf400000000000ULL,
    0xc600000000000ULL,
    0xe900000000000ULL,
    0xd680000000000ULL,
    0xf240000000000ULL,
    0x620000000000ULL,
    0x4f50000000000ULL,
    0x86a8000000000ULL,
    0x8c1c000000000ULL,
    0x1a1e000000000ULL,
    0xb0d000000000ULL,
    0x4188800000000ULL,
    0x31d9c00000000ULL,
    0x10e3a00000000ULL,
    0x9432100000000ULL,
    0xf05b680000000ULL,
    0x613e140000000ULL,
    0xe4dda20000000ULL,
    0xd46f110000000ULL,
    0x997be98000000ULL,
    0x71fbd6c000000ULL,
    0xf6a0022000000ULL,
    0xa41002d000000ULL,
    0x3e08029800000ULL,
    0x750c016400000ULL,
    0x5496037a00000ULL,
    0xd541001900000ULL,
    0xcdbe830080000ULL,
    0xad88c29940000ULL,
    0x9bd52140a0000ULL,
    0x13f6d1f970000ULL,
    0x21b84870f8000ULL,
    0x799c51174000ULL,
    0x98d3691c82000ULL,
    0xa67217f717000ULL,
    0xd66ba3abf6800ULL,
    0xfe7e11bcfc400ULL,
    0xda6d6b674fa00ULL,
    0xa46f14539f500ULL,
    0xb16b22e8f8080ULL,
    0x55ebd07574040ULL,
    0x88b8c86682060ULL,
    0xb10c06ee170b0ULL,
    0xda86c92b76808ULL,
    0xa85105e5bc4f4ULL,
    0x6d264807efac6ULL,
    0xbed4c77aef5e9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xc800000000000ULL,
    0xfc00000000000ULL,
    0xc600000000000ULL,
    0xa500000000000ULL,
    0x7a80000000000ULL,
    0xc640000000000ULL,
    0x8ee0000000000ULL,
    0x2f50000000000ULL,
    0x9a68000000000ULL,
    0xe014000000000ULL,
    0x1006000000000ULL,
    0x981d000000000ULL,
    0x4411800000000ULL,
    0xf201c00000000ULL,
    0x9f0b200000000ULL,
    0x1997900000000ULL,
    0x19cc680000000ULL,
    0x323e8c0000000ULL,
    0x67ed220000000ULL,
    0x3bda910000000ULL,
    0x5535e88000000ULL,
    0x6a6b4dc000000ULL,
    0x6800012000000ULL,
    0xcc0002f000000ULL,
    0x2e00039800000ULL,
    0x2900035400000ULL,
    0x748002ca00000ULL,
    0x9f4000e900000ULL,
    0x3260012380000ULL,
    0x4c10022940000ULL,
    0x6e080383a0000ULL,
    0x904009950000ULL,
    0x48e027b98000ULL,
    0x575902dd34000ULL,
    0xce7f8379e6000ULL,
    0x8a08c04857000ULL,
    0xcb1ca23c34800ULL,
    0x738b52062a400ULL,
    0xc2d6c97777a00ULL,
    0xd9a8df00fd300ULL,
    0xe12a69e398080ULL,
    0x10738d8934040ULL,
    0x2b14a3b3e6020ULL,
    0x638f51a157070ULL,
    0x5ad8cb9fb48c8ULL,
    0x9db1ddef6a4fcULL,
    0x1335ebd4d7ac6ULL,
    0x8f6b4d69ad3a5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x5800000000000ULL,
    0xac00000000000ULL,
    0x6e00000000000ULL,
    0x4100000000000ULL,
    0x2380000000000ULL,
    0x46c0000000000ULL,
    0x2220000000000ULL,
    0x33d0000000000ULL,
    0x23a8000000000ULL,
    0x4a04000000000ULL,
    0x8b1e000000000ULL,
    0xe89f000000000ULL,
    0xe43800000000ULL,
    0x9c66c00000000ULL,
    0xf7ada00000000ULL,
    0x7811b00000000ULL,
    0x5c1cf80000000ULL,
    0x9602340000000ULL,
    0x5d13a20000000ULL,
    0x159eb10000000ULL,
    0xabd77a8000000ULL,
    0x6fb0f6c000000ULL,
    0x3408036000000ULL,
    0x221403b000000ULL,
    0x3f16033800000ULL,
    0x8a8b03c400000ULL,
    0x915581ee00000ULL,
    0xa6edc3ab00000ULL,
    0x3ef8213080000ULL,
    0x72fc71cb40000ULL,
    0xce4d800a0000ULL,
    0xa5fe443310000ULL,
    0x727778a4f8000ULL,
    0xf6a0f66d3c000ULL,
    0xfb800037ae000ULL,
    0xaac000998b000ULL,
    0xec200212ce800ULL,
    0xc2d002323ac00ULL,
    0x582801917de00ULL,
    0xa0c4039d93d00ULL,
    0xc73e039cf8080ULL,
    0x9a4f01a93c040ULL,
    0xe6b83d9ae0a0ULL,
    0x90a2c2328b0b0ULL,
    0x5e93a1a24e858ULL,
    0xa35eb1397acacULL,
    0x71f778f1dde6eULL,
    0x4060f51e83d41ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x4800000000000ULL,
    0xc00000000000ULL,
    0xca00000000000ULL,
    0xfb00000000000ULL,
    0xb180000000000ULL,
    0x5ec0000000000ULL,
    0xc960000000000ULL,
    0xc1d0000000000ULL,
    0xaee8000000000ULL,
    0x4604000000000ULL,
    0x7102000000000ULL,
    0xea9d000000000ULL,
    0xbf4b800000000ULL,
    0xdfb0400000000ULL,
    0x4aee00000000ULL,
    0xa537700000000ULL,
    0x13f8b80000000ULL,
    0x86891c0000000ULL,
    0xc54ce20000000ULL,
    0x9cba710000000ULL,
    0xa13b3a8000000ULL,
    0x75ed5d4000000ULL,
    0xe788032000000ULL,
    0xc7d4013000000ULL,
    0x7fea01a800000ULL,
    0xfc9902ac00000ULL,
    0x864983e600000ULL,
    0x392d414b00000ULL,
    0x71e5600d80000ULL,
    0x818730eb40000ULL,
    0xa6d6587da0000ULL,
    0x7d7e6c6310000ULL,
    0x1fd458e188000ULL,
    0xdbe36dad34000ULL,
    0xca9fd8c68e000ULL,
    0xaf532e26d1000ULL,
    0x37b13876b3800ULL,
    0x58a45d4ea9c00ULL,
    0x2729803acce00ULL,
    0xe4fd42f8e5700ULL,
    0xfd0d614988080ULL,
    0x6083330134040ULL,
    0xe45459208e0a0ULL,
    0x3e236e6dd1050ULL,
    0x127fdafb33848ULL,
    0x60432ce5e9c0cULL,
    0x183939676cecaULL,
    0xd3705eabf57fbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0xfc00000000000ULL,
    0xaa00000000000ULL,
    0x6300000000000ULL,
    0x1880000000000ULL,
    0xbb40000000000ULL,
    0xa6e0000000000ULL,
    0x7a50000000000ULL,
    0x9168000000000ULL,
    0xfb94000000000ULL,
    0xe3d6000000000ULL,
    0x3db1000000000ULL,
    0x2cb1800000000ULL,
    0x4338400000000ULL,
    0x96efe00000000ULL,
    0xb254500000000ULL,
    0xbd7c080000000ULL,
    0x9974c0000000ULL,
    0xd4d1e20000000ULL,
    0x7331510000000ULL,
    0x5efb888000000ULL,
    0x9e4e0fc000000ULL,
    0x4f6780a000000ULL,
    0x3e8942f000000ULL,
    0x9a5e622800000ULL,
    0x16c124c00000ULL,
    0x8393eac200000ULL,
    0x47c31d1d00000ULL,
    0xc3adea3380000ULL,
    0x19a61d6540000ULL,
    0x92aa68a7a0000ULL,
    0x563f5fc350000ULL,
    0xb77c0934d8000ULL,
    0xda974c2f84000ULL,
    0x4451e07f66000ULL,
    0xc471525eb1000ULL,
    0xfa1b893ee7800ULL,
    0x7b1e0d6e95400ULL,
    0x6c8f82b6b6200ULL,
    0x1d5d4152fc300ULL,
    0xc7e8639cd8080ULL,
    0xfdcd11a384040ULL,
    0x98aa6a1d66020ULL,
    0x853f5db3b10f0ULL,
    0x27fc0925678a8ULL,
    0x6dd74e47d54fcULL,
    0xe0b1e2d3162aaULL,
    0x2121528cac363ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xd800000000000ULL,
    0x2400000000000ULL,
    0xf200000000000ULL,
    0x2b00000000000ULL,
    0x7f80000000000ULL,
    0xb2c0000000000ULL,
    0x2420000000000ULL,
    0x3d0000000000ULL,
    0xbca8000000000ULL,
    0xb084000000000ULL,
    0xdf52000000000ULL,
    0xcdef000000000ULL,
    0x506e800000000ULL,
    0xf1bd400000000ULL,
    0xfe0de00000000ULL,
    0xb517b00000000ULL,
    0xfa86380000000ULL,
    0xe04ef40000000ULL,
    0xe877e20000000ULL,
    0xe5bcb10000000ULL,
    0xa41aba8000000ULL,
    0xb20cb74000000ULL,
    0x8b1c816000000ULL,
    0xaf82419000000ULL,
    0x6acb614800000ULL,
    0x2ef3ec00000ULL,
    0xf1d9da9e00000ULL,
    0x97b6465b00000ULL,
    0xcf1f595880000ULL,
    0x6d8f05a340000ULL,
    0xe9c0bb8ca0000ULL,
    0x53b7b64910000ULL,
    0x4d08000dc8000ULL,
    0x4e9401e8fc000ULL,
    0x6a5a019b1a000ULL,
    0x377b0278e5000ULL,
    0xb03481d37f800ULL,
    0x19c64314e2c00ULL,
    0x1bb9600d22e00ULL,
    0x1111f22f8ed00ULL,
    0x489f5bc5c8080ULL,
    0x6b4f0544fc040ULL,
    0x47e0ba651a0a0ULL,
    0x8f67b5b3e50d0ULL,
    0xa42001c3ff8d8ULL,
    0x43d0015ba2c24ULL,
    0x1ca8031f82ef2ULL,
    0x6084033d9ed2bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xd800000000000ULL,
    0x4c00000000000ULL,
    0x2200000000000ULL,
    0xdf00000000000ULL,
    0x2680000000000ULL,
    0x8ec0000000000ULL,
    0xb0e0000000000ULL,
    0xefd0000000000ULL,
    0x7378000000000ULL,
    0xf494000000000ULL,
    0x19d2000000000ULL,
    0x227f000000000ULL,
    0x5f13800000000ULL,
    0xe687400000000ULL,
    0xaed6600000000ULL,
    0xe0ea900000000ULL,
    0x37c1380000000ULL,
    0x3f634c0000000ULL,
    0xd69c620000000ULL,
    0xc6d1930000000ULL,
    0x4f8b88000000ULL,
    0xd1cf0d4000000ULL,
    0x566b816000000ULL,
    0x4113423000000ULL,
    0x9384600800000ULL,
    0xc355923c00000ULL,
    0x26b2bbfa00000ULL,
    0xf4f40f0b00000ULL,
    0x99d2004b80000ULL,
    0xe27f03c340000ULL,
    0x7f1381d7e0000ULL,
    0xb68740a950000ULL,
    0x76d661c4c8000ULL,
    0xacea9206bc000ULL,
    0x15c13839ae000ULL,
    0xe0634f344d000ULL,
    0xf01c604611800ULL,
    0x4811923156400ULL,
    0xb418bbe0cae00ULL,
    0x3e1f0d98d0300ULL,
    0x2513814cc8080ULL,
    0xb587417abc0c0ULL,
    0x8a5660a3ae020ULL,
    0xe12a910f4d050ULL,
    0x79a13885918d8ULL,
    0x12734e8e1644cULL,
    0x370460ad2ae22ULL,
    0x295900a803dfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x6800000000000ULL,
    0xec00000000000ULL,
    0xfa00000000000ULL,
    0xfb00000000000ULL,
    0x6180000000000ULL,
    0x89c0000000000ULL,
    0xd920000000000ULL,
    0x34d0000000000ULL,
    0x1a8000000000ULL,
    0x3384000000000ULL,
    0x7ece000000000ULL,
    0xd2b9000000000ULL,
    0x1e0b800000000ULL,
    0x3d1a400000000ULL,
    0x9896e00000000ULL,
    0xa755700000000ULL,
    0x4f63580000000ULL,
    0x893a940000000ULL,
    0x8cd0e20000000ULL,
    0x15b8710000000ULL,
    0x7d8eda8000000ULL,
    0xfbddd4c000000ULL,
    0x5e2383a000000ULL,
    0xf75e42b000000ULL,
    0xf778e16800000ULL,
    0x9d3c732c00000ULL,
    0xc2c0d82600000ULL,
    0x90a4d59700000ULL,
    0xf108028c80000ULL,
    0x1294023f40000ULL,
    0x94460300a0000ULL,
    0xf2ed022910000ULL,
    0x92ed833fb8000ULL,
    0x42e74029a4000ULL,
    0xbaf363968e000ULL,
    0xee633d679000ULL,
    0x70f63a1f63800ULL,
    0x9de1a7f0b1c00ULL,
    0xfd63582123600ULL,
    0xee3a9703c3500ULL,
    0xdf50e0d7b8080ULL,
    0xbb7873c5a4040ULL,
    0x572eda108e0a0ULL,
    0x51cdd5f179030ULL,
    0x1d2b837be3868ULL,
    0x82ca4223f1cecULL,
    0x30bee0a7836faULL,
    0xc111720dd35fbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x6800000000000ULL,
    0x6c00000000000ULL,
    0x6200000000000ULL,
    0x6900000000000ULL,
    0xc280000000000ULL,
    0xf2c0000000000ULL,
    0x5e60000000000ULL,
    0x7dd0000000000ULL,
    0x3bf8000000000ULL,
    0x269c000000000ULL,
    0xccc6000000000ULL,
    0x5d61000000000ULL,
    0x4a43800000000ULL,
    0x1b9400000000ULL,
    0x8f3ca00000000ULL,
    0x4fe2100000000ULL,
    0x809a780000000ULL,
    0xcbd41c0000000ULL,
    0x64e2a20000000ULL,
    0x7b0f130000000ULL,
    0x387f98000000ULL,
    0x3c405fc000000ULL,
    0x3ea383a000000ULL,
    0x22a942b000000ULL,
    0xe8a4a00800000ULL,
    0x8dae126400000ULL,
    0xdd247aaa00000ULL,
    0x2ee91e7b00000ULL,
    0xde0722f180000ULL,
    0x3307535340000ULL,
    0x3f80da65e0000ULL,
    0xa6470c9170000ULL,
    0xa3a359ea98000ULL,
    0x862e4e32c4000ULL,
    0x7d67fbceee000ULL,
    0x1a505df895000ULL,
    0xf9bb8025ea800ULL,
    0x7b2542f10c400ULL,
    0x29faa3fae7e00ULL,
    0xe783110af5700ULL,
    0x259f86298080ULL,
    0x3dad5f96c40c0ULL,
    0x153e00c4ee060ULL,
    0xd2fd0233950f0ULL,
    0x2405835c6a868ULL,
    0x5e1840064c46cULL,
    0xf31f229507e62ULL,
    0x5f8b525085769ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0xac00000000000ULL,
    0x4600000000000ULL,
    0x6b00000000000ULL,
    0xdf80000000000ULL,
    0xbec0000000000ULL,
    0x1fa0000000000ULL,
    0x37d0000000000ULL,
    0x9938000000000ULL,
    0x8784000000000ULL,
    0x82c6000000000ULL,
    0xb1a3000000000ULL,
    0xb8db800000000ULL,
    0x14a7c00000000ULL,
    0xb85ce00000000ULL,
    0x6ff5300000000ULL,
    0x8c29b80000000ULL,
    0x5f1b840000000ULL,
    0x582e20000000ULL,
    0x8bc2330000000ULL,
    0xf72c3a8000000ULL,
    0xa88b47c000000ULL,
    0xff5b80e000000ULL,
    0x3667c1b000000ULL,
    0xf9fce39800000ULL,
    0x6f25326c00000ULL,
    0x3491b99e00000ULL,
    0xa15f844b00000ULL,
    0x164e36680000ULL,
    0xd871327340000ULL,
    0x16efbb9ae0000ULL,
    0xb2b8872510000ULL,
    0x4359629598000ULL,
    0x5865f3a9cc000ULL,
    0xd6f0db1f8e000ULL,
    0x12be74208f000ULL,
    0xb3523b746b800ULL,
    0xe06c443258c00ULL,
    0x7ae6013548e00ULL,
    0x54b3009db1100ULL,
    0xd843838d98080ULL,
    0x3ff3c205cc0c0ULL,
    0xc422e0618e0a0ULL,
    0x4b1232db8f0f0ULL,
    0xef94390aeb8b8ULL,
    0xa6cf47ed18cacULL,
    0x43bd81d1a8e46ULL,
    0xc9d4c143a116bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xa800000000000ULL,
    0xc00000000000ULL,
    0x2200000000000ULL,
    0x2700000000000ULL,
    0xf880000000000ULL,
    0x65c0000000000ULL,
    0x4c60000000000ULL,
    0xdd50000000000ULL,
    0x49b8000000000ULL,
    0x8124000000000ULL,
    0x8cfe000000000ULL,
    0x6c89000000000ULL,
    0x23db800000000ULL,
    0xe573400000000ULL,
    0x90d6600000000ULL,
    0x5bf7d00000000ULL,
    0x5c09980000000ULL,
    0x5a183c0000000ULL,
    0x930de20000000ULL,
    0xce84930000000ULL,
    0xc4dffa8000000ULL,
    0xbdefeec000000ULL,
    0x450478a000000ULL,
    0xbf9caf3000000ULL,
    0x8d52180800000ULL,
    0x31ab7d5c00000ULL,
    0x353b81c200000ULL,
    0xbae3406700000ULL,
    0xcd8e619980000ULL,
    0xd243d11940000ULL,
    0x7d2f98ece0000ULL,
    0x66e53f3f90000ULL,
    0x5790626878000ULL,
    0xe15ad0cc64000ULL,
    0x3ac19fa0e000ULL,
    0xa227db31d000ULL,
    0xe66000c7c1800ULL,
    0x8650029c2b400ULL,
    0x9b3800e250600ULL,
    0x7fe4031719f00ULL,
    0xb21e03f1f8080ULL,
    0xff1901d5240c0ULL,
    0xfc838116ee0a0ULL,
    0xfbc7428c8d0b0ULL,
    0xe17062afb98a8ULL,
    0xecad2504f40cULL,
    0xf6f419185e622ULL,
    0x8f967ea404f27ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x7800000000000ULL,
    0x7c00000000000ULL,
    0xe200000000000ULL,
    0x1100000000000ULL,
    0xb80000000000ULL,
    0x7ec0000000000ULL,
    0xf960000000000ULL,
    0x2550000000000ULL,
    0xbfa8000000000ULL,
    0x402c000000000ULL,
    0xa8e2000000000ULL,
    0x968b000000000ULL,
    0x9f41800000000ULL,
    0xa2b5400000000ULL,
    0xe1a8a00000000ULL,
    0x1325b00000000ULL,
    0x827a380000000ULL,
    0x4bde3c0000000ULL,
    0x1ce9220000000ULL,
    0xe090f10000000ULL,
    0x88529b8000000ULL,
    0xf63b8f4000000ULL,
    0x1ff31be000000ULL,
    0xf21eccf000000ULL,
    0xc913ba0800000ULL,
    0x47877e0400000ULL,
    0x34c3804e00000ULL,
    0xec6e424b00000ULL,
    0xb0c1220d80000ULL,
    0x327cf26140000ULL,
    0xa3d098b8a0000ULL,
    0xf8e08d4fb0000ULL,
    0xae9a98e008000ULL,
    0x3478e706c000ULL,
    0x90b91ac826000ULL,
    0x88b9cca425000ULL,
    0x64b0385e0a800ULL,
    0x1eb93c334ac00ULL,
    0xa3aaa3c1c6e00ULL,
    0x323eb30b31f00ULL,
    0x21f3baed88080ULL,
    0x31177c112c040ULL,
    0x7b8b8270860e0ULL,
    0x36d241eb950d0ULL,
    0x2d6b20be02878ULL,
    0xc34bf24326c7cULL,
    0x30bb1909e0ee2ULL,
    0xb8a2cfaf14f11ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x1800000000000ULL,
    0x3c00000000000ULL,
    0x3e00000000000ULL,
    0x9900000000000ULL,
    0xd380000000000ULL,
    0x13c0000000000ULL,
    0x3ee0000000000ULL,
    0x5850000000000ULL,
    0x5128000000000ULL,
    0xb8ac000000000ULL,
    0x6066000000000ULL,
    0x2207000000000ULL,
    0xf70d800000000ULL,
    0x4286c00000000ULL,
    0x1449a00000000ULL,
    0x3739300000000ULL,
    0xfda1180000000ULL,
    0x7df14c0000000ULL,
    0xb0c4220000000ULL,
    0x267ff10000000ULL,
    0x3708bb8000000ULL,
    0xe2987fc000000ULL,
    0x44d3a6000000ULL,
    0xdf22bdf000000ULL,
    0xd9aa997800000ULL,
    0x7fe08ca400000ULL,
    0x17c802ae00000ULL,
    0x6cfc027f00000ULL,
    0xf74e03e380000ULL,
    0xcfab003540000ULL,
    0x62eb8292a0000ULL,
    0xd641c039b0000ULL,
    0x302422cc18000ULL,
    0xa72ff3c25c000ULL,
    0x55a0b92d16000ULL,
    0xb9f47ec6eb000ULL,
    0x42cb3a6f9e800ULL,
    0x9975bed708c00ULL,
    0x418f196f8a600ULL,
    0x1cca4c5772300ULL,
    0xb067a12f98080ULL,
    0x6a0233f71c040ULL,
    0xc3029bbfb60e0ULL,
    0xa88c8eff5b0f0ULL,
    0x974e00a386818ULL,
    0x7fab011554c3cULL,
    0x9aeb82429c63eULL,
    0x1a41c39199399ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0x6400000000000ULL,
    0x2e00000000000ULL,
    0xab00000000000ULL,
    0xfc80000000000ULL,
    0xe6c0000000000ULL,
    0xf8a0000000000ULL,
    0x2650000000000ULL,
    0xf468000000000ULL,
    0x2bec000000000ULL,
    0x8daa000000000ULL,
    0x39d7000000000ULL,
    0xf22a800000000ULL,
    0xef88c00000000ULL,
    0x4e50e00000000ULL,
    0xd879d00000000ULL,
    0x19e2f80000000ULL,
    0xfcbb240000000ULL,
    0xd85a620000000ULL,
    0x2761110000000ULL,
    0x237a1b8000000ULL,
    0xdd7ef64000000ULL,
    0xe7a986000000ULL,
    0x6e1349000000ULL,
    0x5620f93800000ULL,
    0x618025ec00000ULL,
    0x955ae01200000ULL,
    0x2cfed14b00000ULL,
    0x3207aba80000ULL,
    0x4e1fe5e540000ULL,
    0x7b0080fba0000ULL,
    0x849fc108b0000ULL,
    0x12da611e28000ULL,
    0x4ea113091c000ULL,
    0xe95a1a698a000ULL,
    0x26eef583d3000ULL,
    0x66329984cb800ULL,
    0x899d341c4b400ULL,
    0xf942f83a29e00ULL,
    0xfeeb24df23900ULL,
    0xe23263a4a8080ULL,
    0x378d12ec5c040ULL,
    0xca5018922a0e0ULL,
    0x6669f58b63090ULL,
    0x2af01b1ae3898ULL,
    0x6439f55557464ULL,
    0x10981833a3e2eULL,
    0x74d5f4ccf09abULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xa800000000000ULL,
    0x6c00000000000ULL,
    0xf200000000000ULL,
    0xdf00000000000ULL,
    0x4280000000000ULL,
    0x15c0000000000ULL,
    0x5720000000000ULL,
    0x6350000000000ULL,
    0x4e8000000000ULL,
    0xed64000000000ULL,
    0x5c2a000000000ULL,
    0x57d9000000000ULL,
    0xf023800000000ULL,
    0x45dd400000000ULL,
    0x1f39600000000ULL,
    0x3f4a900000000ULL,
    0xcef0980000000ULL,
    0xf676140000000ULL,
    0x80bae20000000ULL,
    0x6f07d10000000ULL,
    0x3a81fa8000000ULL,
    0x71c8864000000ULL,
    0x592878a000000ULL,
    0xe65cc4b000000ULL,
    0xf57a9b4800000ULL,
    0x483f143c00000ULL,
    0xc1d1632a00000ULL,
    0x212e92a700000ULL,
    0x825a9ab480000ULL,
    0xfb6f150140000ULL,
    0xcd396271a0000ULL,
    0x304a902e90000ULL,
    0x8470986e28000ULL,
    0x1fb615b924000ULL,
    0x8d9ae325ae000ULL,
    0xbf57d3c8a5000ULL,
    0x8ee9f90b6d800ULL,
    0x566c86a69e400ULL,
    0x10a2783244600ULL,
    0xc715c4c359500ULL,
    0x56911adaa8080ULL,
    0x83d654b864040ULL,
    0x862a03540e0a0ULL,
    0xa4d902e635090ULL,
    0xe0a383e5458a8ULL,
    0x1f1d415fba46cULL,
    0xa29961b7ea6f2ULL,
    0x25da92bbfc5dfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x4800000000000ULL,
    0x4c00000000000ULL,
    0xd600000000000ULL,
    0x6300000000000ULL,
    0x9f80000000000ULL,
    0xbf40000000000ULL,
    0xb6a0000000000ULL,
    0x76d0000000000ULL,
    0x2ae8000000000ULL,
    0xffe4000000000ULL,
    0xb372000000000ULL,
    0x2ba5000000000ULL,
    0x764f800000000ULL,
    0x6425c00000000ULL,
    0xc102e00000000ULL,
    0x3e8bb00000000ULL,
    0x31d5c80000000ULL,
    0x3f70040000000ULL,
    0x1da5620000000ULL,
    0xe54a710000000ULL,
    0xe3a52a8000000ULL,
    0x7a5eb54000000ULL,
    0x123f2b2000000ULL,
    0xf21fb43000000ULL,
    0xe902abd800000ULL,
    0xc29f75cc00000ULL,
    0x5fcfcbde00000ULL,
    0x871078d00000ULL,
    0x5038e02280000ULL,
    0xa31ab12740000ULL,
    0x7f804a2da0000ULL,
    0x4f54c4fe90000ULL,
    0xaeba021148000ULL,
    0x72d10234d4000ULL,
    0xb0f5800e1e000ULL,
    0x4af4c18547000ULL,
    0x4ff760e6e3800ULL,
    0xb7f71bd6ac00ULL,
    0x7fa2aafa81200ULL,
    0xa44f766b13100ULL,
    0x9d27c8b3c8080ULL,
    0xeb95065394040ULL,
    0x7d4ae103be0a0ULL,
    0xa7bfb14bd7050ULL,
    0x404fc92fab848ULL,
    0xf7310645bec4cULL,
    0x4698e12a9f2d6ULL,
    0x85cab06354163ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0x6400000000000ULL,
    0x600000000000ULL,
    0x6500000000000ULL,
    0x5580000000000ULL,
    0xb640000000000ULL,
    0x1760000000000ULL,
    0x15d0000000000ULL,
    0x7c28000000000ULL,
    0x633c000000000ULL,
    0x71be000000000ULL,
    0x9167000000000ULL,
    0x30de800000000ULL,
    0x49b1c00000000ULL,
    0xe569200000000ULL,
    0xfecbf00000000ULL,
    0xe0a1280000000ULL,
    0x4af2bc0000000ULL,
    0x4f9fa20000000ULL,
    0xc146310000000ULL,
    0xd9f6098000000ULL,
    0xc71e4cc000000ULL,
    0x1680086000000ULL,
    0xd6d54c9000000ULL,
    0x5ca88b9800000ULL,
    0x48ef8c5400000ULL,
    0x1c8928b600000ULL,
    0x69cebe8900000ULL,
    0x5e21a1a580000ULL,
    0x6021319340000ULL,
    0x71288b5ea0000ULL,
    0xeaaf8f91f0000ULL,
    0xf5e928b578000ULL,
    0x4d1ebdcfdc000ULL,
    0xe989a3a05a000ULL,
    0xb45d327077000ULL,
    0x447689487c800ULL,
    0xbd588fec03c00ULL,
    0xfbffab5226a00ULL,
    0xc4037e9721f00ULL,
    0x16168090f8080ULL,
    0xad1dc09c9c040ULL,
    0x999f209efa060ULL,
    0x4c40f17187030ULL,
    0x1069aa6504898ULL,
    0x23587e77dfc64ULL,
    0xfaf600447ca06ULL,
    0x978b026e56f65ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0x5c00000000000ULL,
    0xaa00000000000ULL,
    0x5d00000000000ULL,
    0x9180000000000ULL,
    0x3440000000000ULL,
    0x4a0000000000ULL,
    0xf7d0000000000ULL,
    0x19e8000000000ULL,
    0xf4fc000000000ULL,
    0x3d7a000000000ULL,
    0x4d39000000000ULL,
    0xc79d800000000ULL,
    0xf35e400000000ULL,
    0xd023200000000ULL,
    0x7613f00000000ULL,
    0xb716e80000000ULL,
    0xac8bc40000000ULL,
    0xb5d6a20000000ULL,
    0x58f1b10000000ULL,
    0xaf6fc98000000ULL,
    0x4431344000000ULL,
    0xb015cba000000ULL,
    0x5808347000000ULL,
    0x4408492800000ULL,
    0x8e16743400000ULL,
    0xc30b686600000ULL,
    0x3a9584e100000ULL,
    0x52d5809a80000ULL,
    0xfc72429b40000ULL,
    0x56b120a9a0000ULL,
    0xded6f346f0000ULL,
    0x1e7169a968000ULL,
    0x27ac863ea4000ULL,
    0x9d48018556000ULL,
    0x432c0364c9000ULL,
    0x4492000644800ULL,
    0xa9c500711bc00ULL,
    0x92e780e2e5a00ULL,
    0xe267413702100ULL,
    0xbdbea0b3e8080ULL,
    0xd84db0e5e4040ULL,
    0xf6b5ca8cf6060ULL,
    0xeed8345239010ULL,
    0x66048872c868ULL,
    0x3aa767bbfc5cULL,
    0x3516b01b3aaaULL,
    0xe83c87b2cb15dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0x6400000000000ULL,
    0x9a00000000000ULL,
    0xfd00000000000ULL,
    0xa680000000000ULL,
    0xecc0000000000ULL,
    0x8620000000000ULL,
    0x850000000000ULL,
    0xabe8000000000ULL,
    0xcef4000000000ULL,
    0x7466000000000ULL,
    0x8ea5000000000ULL,
    0xfb8e800000000ULL,
    0xfa58400000000ULL,
    0x82f0200000000ULL,
    0x1a7c100000000ULL,
    0xf1bcf80000000ULL,
    0xfc18740000000ULL,
    0x8616a20000000ULL,
    0xeb10510000000ULL,
    0x558ad88000000ULL,
    0xe55167c000000ULL,
    0x556cd92000000ULL,
    0x5e34649000000ULL,
    0x344258e800000ULL,
    0x4dfc253400000ULL,
    0xf5fa7b3a00000ULL,
    0x19e434e300000ULL,
    0xc7e881d080000ULL,
    0x50fd408540000ULL,
    0xc37ea1fda0000ULL,
    0xed24512cd0000ULL,
    0x1dccd81b18000ULL,
    0x6ea467ccd4000ULL,
    0x6b8a59ab1a000ULL,
    0xe25825b4f1000ULL,
    0xdef47b7778800ULL,
    0x2c75343af4400ULL,
    0xf2a000ee71e00ULL,
    0x3d90026904d00ULL,
    0x3148014b98080ULL,
    0x2764008994040ULL,
    0x372e0176ba020ULL,
    0xc0c10008210f0ULL,
    0x38208384608c8ULL,
    0x4f5942c220464ULL,
    0x70a07f6be9aULL,
    0x4cb5523ef5dfdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x8800000000000ULL,
    0x7400000000000ULL,
    0x6e00000000000ULL,
    0x2f00000000000ULL,
    0x7f80000000000ULL,
    0x4c40000000000ULL,
    0xfea0000000000ULL,
    0xc9d0000000000ULL,
    0x23e8000000000ULL,
    0x8ffc000000000ULL,
    0x3dea000000000ULL,
    0xf8e5000000000ULL,
    0xde68800000000ULL,
    0x3ebac00000000ULL,
    0xe9dc600000000ULL,
    0xd3ebb00000000ULL,
    0x97f0280000000ULL,
    0xc1ed840000000ULL,
    0xe2fce20000000ULL,
    0x9f7d710000000ULL,
    0x6e2e498000000ULL,
    0xda1f364000000ULL,
    0x610e482000000ULL,
    0xa08f34d000000ULL,
    0x2bc64a3800000ULL,
    0x4ee337fc00000ULL,
    0x2d64485e00000ULL,
    0xab2a37a100000ULL,
    0xfc8ecbe280000ULL,
    0x81c9f50b40000ULL,
    0x77f02969a0000ULL,
    0x11ed8522f0000ULL,
    0xafce2eb28000ULL,
    0x7b7d73d9d4000ULL,
    0x882e4bea82000ULL,
    0x811f37ff5b000ULL,
    0x708e4843f9800ULL,
    0xc3cf34e98ac00ULL,
    0xaae64962eaa00ULL,
    0xcb7337cb79100ULL,
    0xf02c4889a8080ULL,
    0xed06349294040ULL,
    0xe28cc8a322060ULL,
    0xf6d0f70dab090ULL,
    0x9472a910d1888ULL,
    0xd7b2468c5ec74ULL,
    0x3d4802f668a6eULL,
    0x962c03452212fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0xfc00000000000ULL,
    0x1200000000000ULL,
    0x6700000000000ULL,
    0xf480000000000ULL,
    0xeac0000000000ULL,
    0xbe60000000000ULL,
    0x4050000000000ULL,
    0xf9b8000000000ULL,
    0x89bc000000000ULL,
    0x31ba000000000ULL,
    0x45b9000000000ULL,
    0x3ba9800000000ULL,
    0xa6a9400000000ULL,
    0xc938600000000ULL,
    0xc568500000000ULL,
    0xf6c7780000000ULL,
    0xfc637c0000000ULL,
    0xaf49e20000000ULL,
    0x612d130000000ULL,
    0x997d198000000ULL,
    0x14ce2e4000000ULL,
    0xe37d19a000000ULL,
    0x4fce2cf000000ULL,
    0x65fd1bc800000ULL,
    0x520e2cdc00000ULL,
    0xc71d1bf200000ULL,
    0x49e2d1b00000ULL,
    0x92c51a9180000ULL,
    0xaa722e2d40000ULL,
    0xae47185ce0000ULL,
    0x8cb72ea1f0000ULL,
    0x1a34998568000ULL,
    0x2ff76c90a4000ULL,
    0x111d7a4bc6000ULL,
    0xc58a7d3a15000ULL,
    0x1f58614769800ULL,
    0xb93851a3f5400ULL,
    0x7d7f7b5633e00ULL,
    0x82df7cdd7cf00ULL,
    0x8273e294e8080ULL,
    0x325410fde40c0ULL,
    0xeb499b726060ULL,
    0x95376c6be5090ULL,
    0x277d7a8a018e8ULL,
    0xe9da7faf514fcULL,
    0x1ce0634ff5e12ULL,
    0xab84500c69f67ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0x7c00000000000ULL,
    0xb200000000000ULL,
    0xaf00000000000ULL,
    0xb880000000000ULL,
    0x54c0000000000ULL,
    0x4a60000000000ULL,
    0x8a50000000000ULL,
    0x77b8000000000ULL,
    0xc1ac000000000ULL,
    0x80b6000000000ULL,
    0x3533000000000ULL,
    0xbe65800000000ULL,
    0x5c5b400000000ULL,
    0x6a7600000000ULL,
    0xac3bd00000000ULL,
    0x87f7b80000000ULL,
    0x5d017c0000000ULL,
    0x379ae20000000ULL,
    0xfc5c930000000ULL,
    0xd6bedb8000000ULL,
    0x8435afc000000ULL,
    0x33e6dae000000ULL,
    0xab09aef000000ULL,
    0x5688db4800000ULL,
    0x59c6ae7c00000ULL,
    0x95e35b8200000ULL,
    0x2202ec6300000ULL,
    0xa717ba8180000ULL,
    0x1c917da540000ULL,
    0x6ac2e094e0000ULL,
    0x6f6091d9b0000ULL,
    0xe1d0d86158000ULL,
    0x29faaf228c000ULL,
    0x700d58c4f6000ULL,
    0xf80ded419d000ULL,
    0x9c1c3b4525800ULL,
    0x42053c64d3400ULL,
    0x970e0191f0e00ULL,
    0xc49f019d54f00ULL,
    0xe6d38160d8080ULL,
    0xe5684147cc0c0ULL,
    0x32c2e0b0160e0ULL,
    0x236091682d0f0ULL,
    0x8bd0dbec7d838ULL,
    0xafaaefa5f47cULL,
    0x428d5a3706eb2ULL,
    0x7fcded4fc9fafULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0xc800000000000ULL,
    0x1400000000000ULL,
    0x8e00000000000ULL,
    0xcf00000000000ULL,
    0x7d80000000000ULL,
    0x5bc0000000000ULL,
    0x5320000000000ULL,
    0x4950000000000ULL,
    0x8578000000000ULL,
    0xb9ac000000000ULL,
    0x9312000000000ULL,
    0x279d000000000ULL,
    0xfac4800000000ULL,
    0x31b2c00000000ULL,
    0xa712600000000ULL,
    0x9984f00000000ULL,
    0x2ddf580000000ULL,
    0x9026340000000ULL,
    0x66c4e20000000ULL,
    0x8bab330000000ULL,
    0xd609bb8000000ULL,
    0x3310074000000ULL,
    0xd789d92000000ULL,
    0x2c9f75000000ULL,
    0x3db203b800000ULL,
    0xf50d037c00000ULL,
    0xcc9c815600000ULL,
    0x114ec27f00000ULL,
    0x7978605480000ULL,
    0x13b5f24940000ULL,
    0xca09d9dbe0000ULL,
    0x4909f4b5b0000ULL,
    0x469200f6c8000ULL,
    0x785d03d434000ULL,
    0xfe4833272000ULL,
    0x73e2c0453b000ULL,
    0x19ea61e5e1800ULL,
    0xa0e8f04ed7c00ULL,
    0x1e6d5b2827600ULL,
    0x6a2b352427d00ULL,
    0x37d8631a48080ULL,
    0x1125f2e1740c0ULL,
    0x3451dbbf920e0ULL,
    0x7df5f68f8b0d0ULL,
    0x16f80147a98c8ULL,
    0x3d6c01d3a3c14ULL,
    0x95b201c1b568eULL,
    0xf10d01d4acdcfULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xa800000000000ULL,
    0x6400000000000ULL,
    0x9600000000000ULL,
    0x2300000000000ULL,
    0xf80000000000ULL,
    0x5040000000000ULL,
    0xc860000000000ULL,
    0xcd0000000000ULL,
    0x9fb8000000000ULL,
    0xa0fc000000000ULL,
    0x550e000000000ULL,
    0xdc81000000000ULL,
    0x97c6800000000ULL,
    0xcc21c00000000ULL,
    0xfaade00000000ULL,
    0xd477b00000000ULL,
    0xa6d8a80000000ULL,
    0x86ae9c0000000ULL,
    0x4e65620000000ULL,
    0x17d7730000000ULL,
    0xc33c98000000ULL,
    0x9ab8ecc000000ULL,
    0xe47028a000000ULL,
    0xede5e9000000ULL,
    0xe2b601d800000ULL,
    0xd87d034c00000ULL,
    0x34c8831e00000ULL,
    0x3a0c01100000ULL,
    0xcaeb63d980000ULL,
    0x2c16712f40000ULL,
    0x2154818e0000ULL,
    0x7d092f4ef0000ULL,
    0x7885c84bb8000ULL,
    0x61c5ed0044000ULL,
    0xdf38a8807a000ULL,
    0x5d3e9e4037000ULL,
    0xe03d61e06f800ULL,
    0xf8bb70f02ac00ULL,
    0xa965c8e840a00ULL,
    0xde55ef8439700ULL,
    0xe760aa4a38080ULL,
    0x91529f63040c0ULL,
    0x4aeb63869a060ULL,
    0xec16711fc7030ULL,
    0x62154a72578a8ULL,
    0x4d092cdf2ec64ULL,
    0xd085c870daa96ULL,
    0x5c5ee8afe723ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x5800000000000ULL,
    0x400000000000ULL,
    0x8200000000000ULL,
    0x500000000000ULL,
    0x580000000000ULL,
    0x5240000000000ULL,
    0x42e0000000000ULL,
    0xcad0000000000ULL,
    0x7128000000000ULL,
    0xe874000000000ULL,
    0x340a000000000ULL,
    0xa07000000000ULL,
    0x7916800000000ULL,
    0x1389400000000ULL,
    0x8d4b200000000ULL,
    0x4670700000000ULL,
    0x1f0aa80000000ULL,
    0x64804c0000000ULL,
    0x65d7a20000000ULL,
    0xddbe310000000ULL,
    0xd1b7088000000ULL,
    0x6fa97e4000000ULL,
    0x5cbe2b6000000ULL,
    0x163a0d1000000ULL,
    0x18e0008800000ULL,
    0x8bd0035400000ULL,
    0xd6a801f600000ULL,
    0x2f34031900000ULL,
    0x2b6a026380000ULL,
    0x9697032f40000ULL,
    0xc8de83daa0000ULL,
    0x342d42fcd0000ULL,
    0xcde921cda8000ULL,
    0xf643724a5c000ULL,
    0x10f6292b7a000ULL,
    0xb7de0cc4b5000ULL,
    0xe0a20121a4800ULL,
    0x603303900dc00ULL,
    0x77fc81c858a00ULL,
    0x475e43b438300ULL,
    0x5f75a32628080ULL,
    0xbc8d32311c040ULL,
    0x21cb8b07da020ULL,
    0x7fb73d2165090ULL,
    0x44ab8a8f8c858ULL,
    0x32273ef511c04ULL,
    0xae3893982a82ULL,
    0xd6c33c8c5d305ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0x3400000000000ULL,
    0x2200000000000ULL,
    0x2b00000000000ULL,
    0xbd80000000000ULL,
    0x7940000000000ULL,
    0x2fe0000000000ULL,
    0x57d0000000000ULL,
    0xaf28000000000ULL,
    0x1b64000000000ULL,
    0xf81e000000000ULL,
    0xe40f000000000ULL,
    0x2a14800000000ULL,
    0xb711400000000ULL,
    0x3390e00000000ULL,
    0x4455700000000ULL,
    0x9b6e080000000ULL,
    0xb8072c0000000ULL,
    0x441a620000000ULL,
    0x1a0b310000000ULL,
    0x2f0a6a8000000ULL,
    0x7931cc000000ULL,
    0x664c886000000ULL,
    0xb07d6dd000000ULL,
    0x580000800000ULL,
    0x3d40016c00000ULL,
    0x35e0021600000ULL,
    0x78d001f500000ULL,
    0xa8a8025780000ULL,
    0x7d24012340000ULL,
    0x487e00c2a0000ULL,
    0xe19f012490000ULL,
    0x175c81a9f8000ULL,
    0x82e5432a7c000ULL,
    0x4b46e2ab72000ULL,
    0xecee73ee95000ULL,
    0xe64c89929b800ULL,
    0xf07d6eecf9c00ULL,
    0xa580032592200ULL,
    0xd4002ac65b00ULL,
    0xade003f678080ULL,
    0x4cd003653c040ULL,
    0x8aa8007fd20a0ULL,
    0x5624033f05030ULL,
    0xf5fe026ce3898ULL,
    0x98df00e5c5c34ULL,
    0x38bc834c40222ULL,
    0xd535416660b2bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x6800000000000ULL,
    0x2400000000000ULL,
    0xaa00000000000ULL,
    0x9300000000000ULL,
    0xf480000000000ULL,
    0xc4c0000000000ULL,
    0x8da0000000000ULL,
    0xb750000000000ULL,
    0xcff8000000000ULL,
    0x5124000000000ULL,
    0x3792000000000ULL,
    0x9841000000000ULL,
    0xcd6a800000000ULL,
    0x40f2400000000ULL,
    0x83b3e00000000ULL,
    0x4e53500000000ULL,
    0x869b80000000ULL,
    0xb16e440000000ULL,
    0x16eb620000000ULL,
    0x86b0130000000ULL,
    0x1fc8da8000000ULL,
    0x4d2b574000000ULL,
    0x11833ba000000ULL,
    0x255c079000000ULL,
    0xd0f8822800000ULL,
    0x8bb3420c00000ULL,
    0x1a5960f200000ULL,
    0x1a6110c300000ULL,
    0x6e7a593e80000ULL,
    0x6c6d140140000ULL,
    0x7b7ad845e0000ULL,
    0xf5fa571b90000ULL,
    0xca31b9bac8000ULL,
    0x971a47ef44000ULL,
    0x4e81631cca000ULL,
    0x9fd5124a19000ULL,
    0x8d30586767800ULL,
    0xb1981780d9400ULL,
    0xf55a591044e00ULL,
    0xb8fd176874100ULL,
    0xafa2daac48080ULL,
    0xb04e55e2040c0ULL,
    0x897bb9ab2a0a0ULL,
    0x9aef4692890d0ULL,
    0xa8a1e2632f868ULL,
    0xf6d2532edd424ULL,
    0x42a339e96eeaaULL,
    0x5cc05a9fd193ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0xcc00000000000ULL,
    0x4a00000000000ULL,
    0x8100000000000ULL,
    0x6180000000000ULL,
    0x4a40000000000ULL,
    0xdee0000000000ULL,
    0x2ed0000000000ULL,
    0x8338000000000ULL,
    0x6974000000000ULL,
    0x678a000000000ULL,
    0x4159000000000ULL,
    0x1e7f800000000ULL,
    0x151e400000000ULL,
    0x3f94a00000000ULL,
    0x5558f00000000ULL,
    0x8072280000000ULL,
    0x2a12cc0000000ULL,
    0x7101220000000ULL,
    0x798fb30000000ULL,
    0xbe41088000000ULL,
    0x70f07c4000000ULL,
    0x29d5aaa000000ULL,
    0x29a88c3000000ULL,
    0xc3a783a800000ULL,
    0x92ba414400000ULL,
    0xfb26a1a600000ULL,
    0x6d75f25900000ULL,
    0xd187aaf380000ULL,
    0xb2558d8f40000ULL,
    0x1aea00a2e0000ULL,
    0xb8c90188d0000ULL,
    0x602783e3a8000ULL,
    0x45fa40d724000ULL,
    0x646a1ae9e000ULL,
    0x54e5f27ae9000ULL,
    0x8fdfaa449a800ULL,
    0xf2b18ff1d7c00ULL,
    0xb3800601ea00ULL,
    0x757403d056300ULL,
    0x258a02b828080ULL,
    0x1c59001c640c0ULL,
    0x1dff82aa7e020ULL,
    0x125e42ab39010ULL,
    0xcaf4a3d4b2828ULL,
    0xb0c8f2e9b3cccULL,
    0xbc2a29cc60a4aULL,
    0x27f6cf126f381ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x9800000000000ULL,
    0xe400000000000ULL,
    0x7600000000000ULL,
    0xb500000000000ULL,
    0xe380000000000ULL,
    0xae40000000000ULL,
    0xfe20000000000ULL,
    0x94d0000000000ULL,
    0x9e8000000000ULL,
    0x1bbc000000000ULL,
    0x5786000000000ULL,
    0x9057000000000ULL,
    0xa72e800000000ULL,
    0x8d53c00000000ULL,
    0x18b1200000000ULL,
    0x2112300000000ULL,
    0x4d9ba80000000ULL,
    0x9f4a140000000ULL,
    0x9bb9a20000000ULL,
    0x1786f10000000ULL,
    0xf04c098000000ULL,
    0x5727e7c000000ULL,
    0x155d286000000ULL,
    0xfca5d69000000ULL,
    0x570e825800000ULL,
    0xf883c01400000ULL,
    0x7cd9206e00000ULL,
    0x35ee32e900000ULL,
    0xe9bda84080000ULL,
    0x648d171740000ULL,
    0x5edf2071a0000ULL,
    0xef93103f0000ULL,
    0xab3328a898000ULL,
    0xc74ed56b1c000ULL,
    0x5fa601239a000ULL,
    0xf1870298ff000ULL,
    0x2d4681437c800ULL,
    0xc8afc3bfe4c00ULL,
    0x2917231acca00ULL,
    0xc19533203b500ULL,
    0x955d2ab018080ULL,
    0xbca5d6685c040ULL,
    0x370e833c3a060ULL,
    0x883c0720f0f0ULL,
    0xe4d9202b64898ULL,
    0xd1ee3203b8ce4ULL,
    0x9fbda828f6a76ULL,
    0xd18d172b345b5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xd800000000000ULL,
    0x2400000000000ULL,
    0x8200000000000ULL,
    0x9100000000000ULL,
    0xa280000000000ULL,
    0x2f40000000000ULL,
    0x1920000000000ULL,
    0xed0000000000ULL,
    0x1df8000000000ULL,
    0x5e2c000000000ULL,
    0xb74a000000000ULL,
    0x1d3f000000000ULL,
    0x7cca800000000ULL,
    0x44f1400000000ULL,
    0xa1a00000000ULL,
    0x3e0a900000000ULL,
    0x1718880000000ULL,
    0x419fd40000000ULL,
    0xd4d9220000000ULL,
    0x68e8d30000000ULL,
    0x9eb9ab8000000ULL,
    0x191b044000000ULL,
    0x7e8a896000000ULL,
    0x795cd79000000ULL,
    0xc221a38800000ULL,
    0xc14a910400000ULL,
    0x36388a6a00000ULL,
    0x7b4fd46d00000ULL,
    0x9321230c80000ULL,
    0x83c4d1ef40000ULL,
    0x973ab75e0000ULL,
    0xba6407c1b0000ULL,
    0xb9e00953a8000ULL,
    0x1c3d9762bc000ULL,
    0xc65802684a000ULL,
    0xafbc015435000ULL,
    0x8b9200c24e800ULL,
    0x79c3019966400ULL,
    0x8c78827ee0200ULL,
    0xb2e2428e06500ULL,
    0xeba1225728080ULL,
    0xd984d3c9fc0c0ULL,
    0xd0d3aa17aa0e0ULL,
    0x1af4056885010ULL,
    0xc7b80915668d8ULL,
    0x478196109a424ULL,
    0xf7ca01094a282ULL,
    0x737f007683591ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xf800000000000ULL,
    0x5c00000000000ULL,
    0x2600000000000ULL,
    0x500000000000ULL,
    0xf280000000000ULL,
    0x9140000000000ULL,
    0x4360000000000ULL,
    0x8d0000000000ULL,
    0x61b8000000000ULL,
    0x8274000000000ULL,
    0x9056000000000ULL,
    0xfbe3000000000ULL,
    0x6281800000000ULL,
    0x795dc00000000ULL,
    0x3770200000000ULL,
    0x7adf100000000ULL,
    0xe6b1480000000ULL,
    0xfe91c0000000ULL,
    0xd09fa20000000ULL,
    0xde55d30000000ULL,
    0x4af6e88000000ULL,
    0xc218cfc000000ULL,
    0x9f0749e000000ULL,
    0x19a1e7000000ULL,
    0x6ec6221800000ULL,
    0x84ac10d400000ULL,
    0xa0e8c8aa00000ULL,
    0x910dcf500000ULL,
    0x1c81807580000ULL,
    0x105dc24740000ULL,
    0x3bf02254e0000ULL,
    0x429f1398d0000ULL,
    0x89514b66d8000ULL,
    0xcf791cf9cc000ULL,
    0x26c7a39966000ULL,
    0xc0b1d0dbe7000ULL,
    0xaf8e85078800ULL,
    0x221fcd8820400ULL,
    0x4f08c8bc1b200ULL,
    0x980dc661f700ULL,
    0xcad9818b58080ULL,
    0xfeb9c1aa8c0c0ULL,
    0x83fe208786020ULL,
    0xfe9811c6370f0ULL,
    0x7f5ec95b208f8ULL,
    0xc263dfe2ac45cULL,
    0x705803db9d226ULL,
    0x2be400d028705ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0x3c00000000000ULL,
    0x8200000000000ULL,
    0x5300000000000ULL,
    0x5c80000000000ULL,
    0xd3c0000000000ULL,
    0x67e0000000000ULL,
    0xb650000000000ULL,
    0x4fa8000000000ULL,
    0x58ec000000000ULL,
    0x70c2000000000ULL,
    0x137f000000000ULL,
    0x3191800000000ULL,
    0x1e49400000000ULL,
    0x5bbae00000000ULL,
    0x8ef6500000000ULL,
    0x15d3580000000ULL,
    0xfafe8c0000000ULL,
    0x53c1620000000ULL,
    0x27ec110000000ULL,
    0x565a3b8000000ULL,
    0x9fae9f4000000ULL,
    0xf15b6000000ULL,
    0x4cd18df000000ULL,
    0x9178e38800000ULL,
    0x6289530c00000ULL,
    0x42c2db9200000ULL,
    0x8877ccff00000ULL,
    0xe91b80f780000ULL,
    0xa38a426540000ULL,
    0xb54161c4a0000ULL,
    0xb2c1020b0000ULL,
    0x573a384e88000ULL,
    0x453e9d2bbc000ULL,
    0xae39586f66000ULL,
    0x1ead8f8fd5000ULL,
    0x1772e16143800ULL,
    0x1f8a521a95400ULL,
    0x77595a2d83600ULL,
    0xb83d8eb663300ULL,
    0xdbbae0b108080ULL,
    0xcef65302fc040ULL,
    0xf5d35959c60e0ULL,
    0x2afe8ea0650d0ULL,
    0xbc162504b858ULL,
    0x1bec12586943cULL,
    0xd45a3a1445682ULL,
    0xccae9ce606353ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0xdc00000000000ULL,
    0x9200000000000ULL,
    0xf00000000000ULL,
    0x9780000000000ULL,
    0x7fc0000000000ULL,
    0xee20000000000ULL,
    0x4950000000000ULL,
    0x8b78000000000ULL,
    0xa824000000000ULL,
    0x5046000000000ULL,
    0x8deb000000000ULL,
    0x9571800000000ULL,
    0xbd33400000000ULL,
    0x8cc1600000000ULL,
    0x5bb9300000000ULL,
    0xa191580000000ULL,
    0x5ec4540000000ULL,
    0x34aee20000000ULL,
    0xe615730000000ULL,
    0x691fba8000000ULL,
    0x3e8125c000000ULL,
    0xe1495ae000000ULL,
    0x7f70547000000ULL,
    0xe30e2c800000ULL,
    0x594a72fc00000ULL,
    0x6370383e00000ULL,
    0x3c2d674f00000ULL,
    0x2647bb1080000ULL,
    0xccf5256940000ULL,
    0x9ff75abbe0000ULL,
    0x5a7f55a390000ULL,
    0x8aa760c798000ULL,
    0x8302335dec000ULL,
    0x8d98dbe8a6000ULL,
    0x34d317fd1d000ULL,
    0x47a983a9fd800ULL,
    0x9387404ad8c00ULL,
    0x21df60bc1b600ULL,
    0x9b26309e30500ULL,
    0x45dedb9f18080ULL,
    0x15381508ac0c0ULL,
    0x78d8018d460a0ULL,
    0xfdb402618d070ULL,
    0xa89e0036e5838ULL,
    0xb05f014274cdcULL,
    0x9def83315d692ULL,
    0x7d6c41ffbd50fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x3800000000000ULL,
    0x5400000000000ULL,
    0x5a00000000000ULL,
    0xf00000000000ULL,
    0xc680000000000ULL,
    0x6040000000000ULL,
    0x4aa0000000000ULL,
    0x17d0000000000ULL,
    0xd178000000000ULL,
    0x38a4000000000ULL,
    0xb4da000000000ULL,
    0xb9eb000000000ULL,
    0x9dfc800000000ULL,
    0xff6400000000ULL,
    0x5cf0600000000ULL,
    0x3c77100000000ULL,
    0x652fa80000000ULL,
    0xe813e40000000ULL,
    0xc0ee20000000ULL,
    0x3e1e530000000ULL,
    0xfd014a8000000ULL,
    0xa59db64000000ULL,
    0xa8d7aae000000ULL,
    0x7ff7e65000000ULL,
    0x94f4e1e800000ULL,
    0x6065517c00000ULL,
    0xc325c97a00000ULL,
    0xd11ff69100000ULL,
    0xcb85caa280000ULL,
    0xcdcff63340000ULL,
    0x1e7dc9b7e0000ULL,
    0x5e2bf41f90000ULL,
    0x4487c883e8000ULL,
    0xcb50f729ec000ULL,
    0xde23491092000ULL,
    0x8492b4ae09000ULL,
    0x6b512af749800ULL,
    0x4e3aa759e0400ULL,
    0xbc800208a4a00ULL,
    0x3f40008a7a900ULL,
    0x1420014968080ULL,
    0xb3900226ac0c0ULL,
    0xf9d8013d720a0ULL,
    0x7474037099090ULL,
    0xf92203be21838ULL,
    0xee0f017f4c454ULL,
    0xa5068135d6a5aULL,
    0xc18d40fae390fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0x3400000000000ULL,
    0x9200000000000ULL,
    0xdd00000000000ULL,
    0xff80000000000ULL,
    0xdcc0000000000ULL,
    0x9e20000000000ULL,
    0xee30000000000ULL,
    0x7628000000000ULL,
    0x9234000000000ULL,
    0x8832000000000ULL,
    0xd925000000000ULL,
    0x80a3800000000ULL,
    0x1371c00000000ULL,
    0xafca600000000ULL,
    0x62a5900000000ULL,
    0x5676080000000ULL,
    0xb441cc0000000ULL,
    0xa4760a0000000ULL,
    0x9941cf0000000ULL,
    0x3f6098000000ULL,
    0x4181cfc000000ULL,
    0xb7d60b6000000ULL,
    0x46b1cf1000000ULL,
    0xac7e092800000ULL,
    0xd545cc6400000ULL,
    0x45ec0ad600000ULL,
    0x3e90cf1700000ULL,
    0x2d4788ae80000ULL,
    0x51e50cafc0000ULL,
    0x7c97e97620000ULL,
    0x78519ee710000ULL,
    0x22706356e8000ULL,
    0x644918384000ULL,
    0xa96f8b5466000ULL,
    0xeed10ece43000ULL,
    0x5325e86b4f800ULL,
    0x79b49c84d5400ULL,
    0x16f3e05297a00ULL,
    0x12055245d2700ULL,
    0x1d0de86b4f880ULL,
    0x1f809c84d54c0ULL,
    0xecc1e05297ae0ULL,
    0x26205245d2730ULL,
    0xda2e686b4f838ULL,
    0xe4315c84d54f4ULL,
    0x4f2b805297a72ULL,
    0x77b5c245d27edULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x800000000000ULL,
    0xcc00000000000ULL,
    0x9200000000000ULL,
    0xb900000000000ULL,
    0x1080000000000ULL,
    0xf3c0000000000ULL,
    0xb420000000000ULL,
    0xe430000000000ULL,
    0x1c38000000000ULL,
    0xb834000000000ULL,
    0x5622000000000ULL,
    0x7523000000000ULL,
    0x10ba800000000ULL,
    0x61e8c00000000ULL,
    0x9f1fe00000000ULL,
    0xdf8df00000000ULL,
    0xd45b880000000ULL,
    0xc4610c0000000ULL,
    0xce5b8a0000000ULL,
    0xf1610d0000000ULL,
    0x2cdb8b8000000ULL,
    0xba10fc000000ULL,
    0x807b8ba000000ULL,
    0xd0510cf000000ULL,
    0xba6389e800000ULL,
    0x35550e1400000ULL,
    0xe0f989aa00000ULL,
    0xb820ddb00000ULL,
    0x12410b7a80000ULL,
    0xfb79ce4bc0000ULL,
    0x81dc6b0a60000ULL,
    0xfd28fcab10000ULL,
    0x9cba0252e8000ULL,
    0x93e7017f9c000ULL,
    0x9600801002000ULL,
    0xc70fc1f83f000ULL,
    0xeb9f63ec7d800ULL,
    0xe242304608c00ULL,
    0x9364e89d13a00ULL,
    0xfdd33c678cf00ULL,
    0x672763ec7d880ULL,
    0xe9b6304608c40ULL,
    0x1166e89d13a60ULL,
    0xdcc03c678cfb0ULL,
    0x63a5e3ec7d888ULL,
    0xfc6af04608c8cULL,
    0x4a5b089d13af2ULL,
    0xcf6ecc678cf09ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0xec00000000000ULL,
    0x2a00000000000ULL,
    0x2700000000000ULL,
    0xa980000000000ULL,
    0x940000000000ULL,
    0x11e0000000000ULL,
    0x99f0000000000ULL,
    0x2de8000000000ULL,
    0x8be4000000000ULL,
    0x16f6000000000ULL,
    0x65000000000ULL,
    0x4cb8800000000ULL,
    0x7e1dc00000000ULL,
    0xd109200000000ULL,
    0xcc84100000000ULL,
    0x63c9380000000ULL,
    0xef39040000000ULL,
    0x60493a0000000ULL,
    0x179070000000ULL,
    0xb8293b8000000ULL,
    0x1c9054000000ULL,
    0x1c2139e000000ULL,
    0xffdd06f000000ULL,
    0xd3f394800000ULL,
    0x535c066c00000ULL,
    0xfef1bbee00000ULL,
    0x2464c64900000ULL,
    0x72a01ba980000ULL,
    0xf0d142ec0000ULL,
    0xed88031e20000ULL,
    0xa754020150000ULL,
    0xf8fe0145f8000ULL,
    0x49710100c4000ULL,
    0x6c2681f71a000ULL,
    0x37dcc2b8b3000ULL,
    0x1927a0d33e800ULL,
    0x54cd23aa3400ULL,
    0x6bf099f41a600ULL,
    0x46f4d51a47500ULL,
    0xf87720d33e880ULL,
    0x30b5123aa34c0ULL,
    0xcc0fb9f41a660ULL,
    0x1a15c51a47590ULL,
    0x4f0698d33e818ULL,
    0x4d91d63aa342cULL,
    0x574fa3f41a64aULL,
    0xf0e8d21a475b7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xf800000000000ULL,
    0x8400000000000ULL,
    0x2600000000000ULL,
    0xa700000000000ULL,
    0x3a80000000000ULL,
    0x9ec0000000000ULL,
    0x4fe0000000000ULL,
    0x87f0000000000ULL,
    0x4bf8000000000ULL,
    0xe1e4000000000ULL,
    0xce6000000000ULL,
    0xcb79000000000ULL,
    0x6a29800000000ULL,
    0x8d400000000ULL,
    0xcbc9a00000000ULL,
    0xec66f00000000ULL,
    0x10b9280000000ULL,
    0x7e5ad40000000ULL,
    0x94392a0000000ULL,
    0x939ad50000000ULL,
    0xdf59298000000ULL,
    0xb9aad54000000ULL,
    0x3fc12a6000000ULL,
    0x627ed75000000ULL,
    0x2bbf2af800000ULL,
    0xf6d3d5cc00000ULL,
    0x73e8aa1200000ULL,
    0xc5f397b700000ULL,
    0xdaf68b2d80000ULL,
    0x847525a8c0000ULL,
    0x2cb1820260000ULL,
    0x3c59422f50000ULL,
    0x537a031f8000ULL,
    0xf60bf002b4000ULL,
    0xff0eaa195e000ULL,
    0x4e8a944081000ULL,
    0x50df08d678800ULL,
    0x94f864e11ac00ULL,
    0x1f78238c9c200ULL,
    0x543fb3c471900ULL,
    0x338e88d678880ULL,
    0x2f5124e11ac40ULL,
    0x51b7838c9c2e0ULL,
    0x43d043c471910ULL,
    0xc06620d678878ULL,
    0xaaa2b0e11acc4ULL,
    0x6b41098c9c2c6ULL,
    0xd7a566c4719b7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x2800000000000ULL,
    0x3400000000000ULL,
    0x7600000000000ULL,
    0x5300000000000ULL,
    0xb180000000000ULL,
    0x5740000000000ULL,
    0xf5e0000000000ULL,
    0x4bf0000000000ULL,
    0x7cf8000000000ULL,
    0x837c000000000ULL,
    0xab3a000000000ULL,
    0x855f000000000ULL,
    0x88e9800000000ULL,
    0x957f400000000ULL,
    0xe839200000000ULL,
    0x2ccd100000000ULL,
    0xc3b4f80000000ULL,
    0x2291c40000000ULL,
    0x86ccfa0000000ULL,
    0xb2adc50000000ULL,
    0xa616fa8000000ULL,
    0x2b02c5c000000ULL,
    0xbd877a2000000ULL,
    0xd41851000000ULL,
    0xcce45bf800000ULL,
    0xeb63945c00000ULL,
    0xbf21213e00000ULL,
    0xc341110100000ULL,
    0xf3f6f8e980000ULL,
    0x10f2c52ec0000ULL,
    0xc97f7b9a60000ULL,
    0x8a3d86e330000ULL,
    0x39de591688000ULL,
    0x93c94e64c000ULL,
    0xf048a24d2e000ULL,
    0x527e532fb1000ULL,
    0x5fafd8f3ca800ULL,
    0x208fd68d85400ULL,
    0x83d3806c99600ULL,
    0x602043b502500ULL,
    0x68d0a273ca880ULL,
    0xbdb2534d85440ULL,
    0x758dda4c99620ULL,
    0x695cd6a502530ULL,
    0x82f8018bca8a8ULL,
    0x947c031185474ULL,
    0x64ba017299656ULL,
    0x851f02a402563ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0xc00000000000ULL,
    0x6a00000000000ULL,
    0x8500000000000ULL,
    0x3080000000000ULL,
    0xd7c0000000000ULL,
    0x7ae0000000000ULL,
    0x9cf0000000000ULL,
    0x33f8000000000ULL,
    0xa67c000000000ULL,
    0x1136000000000ULL,
    0x2405000000000ULL,
    0xce14800000000ULL,
    0xb11c00000000ULL,
    0x1b85600000000ULL,
    0x9c49700000000ULL,
    0x7ea0880000000ULL,
    0xee5e540000000ULL,
    0xb7b88a0000000ULL,
    0x94d2550000000ULL,
    0xb5768a8000000ULL,
    0x46ab544000000ULL,
    0xf25408e000000ULL,
    0x65bf947000000ULL,
    0x4dc5e94800000ULL,
    0x77e7266400000ULL,
    0x1860018a00000ULL,
    0xd230013b00000ULL,
    0xab98026180000ULL,
    0x344c0248c0000ULL,
    0x9aae022e60000ULL,
    0x4049019130000ULL,
    0xccba800ab8000ULL,
    0x4758c13124000ULL,
    0xbd3fe29aea000ULL,
    0x5e11b30963000ULL,
    0xf31f6a36ff800ULL,
    0x678fe62746c00ULL,
    0x3e47e027fda00ULL,
    0x6fadb2edff900ULL,
    0xb8c968b6ff880ULL,
    0x8f7ae66746c40ULL,
    0x5bab62c7fda20ULL,
    0xcec0739dff950ULL,
    0xd87a0b7eff818ULL,
    0xb236944346c4cULL,
    0xdb9f6badfda4aULL,
    0xfc4fe6d6ff9d5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0xa400000000000ULL,
    0xb600000000000ULL,
    0x6900000000000ULL,
    0x5680000000000ULL,
    0x1840000000000ULL,
    0x4e20000000000ULL,
    0xcc30000000000ULL,
    0xbb28000000000ULL,
    0xd0ac000000000ULL,
    0xb076000000000ULL,
    0x631f000000000ULL,
    0xa59b800000000ULL,
    0x15d5400000000ULL,
    0xffc600000000ULL,
    0xeddcb00000000ULL,
    0xd3feb80000000ULL,
    0x47cffc0000000ULL,
    0x30f6ba0000000ULL,
    0x2253ff0000000ULL,
    0x528ba8000000ULL,
    0xfda0fc4000000ULL,
    0xa0e53ba000000ULL,
    0x8a5abed000000ULL,
    0x512ad97800000ULL,
    0xd3bf4f7400000ULL,
    0x25fe002200000ULL,
    0xcec3031500000ULL,
    0x9665831a80000ULL,
    0xa2164325c0000ULL,
    0xa719e17620000ULL,
    0x238af32770000ULL,
    0xc4c75a97f8000ULL,
    0x65750c3b0c000ULL,
    0xaf99e0d996000ULL,
    0xe6caf2c859000ULL,
    0x2675b3c67800ULL,
    0xac050d5e2bc00ULL,
    0xf211e2eb1d600ULL,
    0x2f16f1a1d4300ULL,
    0xa79959bc67880ULL,
    0xa2c60d1e2bcc0ULL,
    0x4474634b1d620ULL,
    0x5d00b071d43d0ULL,
    0x4880ba44678c8ULL,
    0x254cfe2a2bc64ULL,
    0x36b338c91d696ULL,
    0x5175bdb4d43b9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x3800000000000ULL,
    0xc400000000000ULL,
    0xe600000000000ULL,
    0x8500000000000ULL,
    0x2380000000000ULL,
    0x39c0000000000ULL,
    0x84e0000000000ULL,
    0x5ef0000000000ULL,
    0x19f8000000000ULL,
    0xc96c000000000ULL,
    0x5e2a000000000ULL,
    0x3c8d000000000ULL,
    0xfd45800000000ULL,
    0x5db1400000000ULL,
    0x7545e00000000ULL,
    0xe1bf900000000ULL,
    0xf753a80000000ULL,
    0xb2b1b40000000ULL,
    0x69cbaa0000000ULL,
    0x6cedb50000000ULL,
    0x32f9a88000000ULL,
    0x33fcb5c000000ULL,
    0x566e286000000ULL,
    0xdaacf6d000000ULL,
    0x45c44bf800000ULL,
    0xe6ef24c400000ULL,
    0x9df7837600000ULL,
    0x8f60402300000ULL,
    0xeb32636580000ULL,
    0x271fd058c0000ULL,
    0x81c98260000ULL,
    0x3f5e67bd70000ULL,
    0xaeb2621ac8000ULL,
    0xdbdfd15f44000ULL,
    0x7e1caf7de000ULL,
    0x686e67fd81000ULL,
    0xbaa63ccc9800ULL,
    0x8843d24c7f400ULL,
    0xa633c9fa07200ULL,
    0x188f653979900ULL,
    0x8b45e34cc9880ULL,
    0xd0bf938c7f440ULL,
    0xaad3ab9a072a0ULL,
    0xfa71b6e979930ULL,
    0x10aba834c98b8ULL,
    0x4addb6887f484ULL,
    0x6a61aa8c07246ULL,
    0x18a0b51a799b5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x8800000000000ULL,
    0xac00000000000ULL,
    0x5e00000000000ULL,
    0x2d00000000000ULL,
    0xf880000000000ULL,
    0x88c0000000000ULL,
    0x5ea0000000000ULL,
    0x13b0000000000ULL,
    0xfb38000000000ULL,
    0x1be4000000000ULL,
    0xb94e000000000ULL,
    0x7cef000000000ULL,
    0x6c2800000000ULL,
    0xbbad400000000ULL,
    0xa726e00000000ULL,
    0x7df6f00000000ULL,
    0xe055680000000ULL,
    0xfe6bfc0000000ULL,
    0x511b6a0000000ULL,
    0x6e84fd0000000ULL,
    0xe9d9e98000000ULL,
    0xa829bc4000000ULL,
    0x3e7f0ba000000ULL,
    0xf11f4ef000000ULL,
    0xde8a62d800000ULL,
    0x31c4b24400000ULL,
    0x8c290b3a00000ULL,
    0xcc644d6700000ULL,
    0x821ee3c080000ULL,
    0xb12f169c0000ULL,
    0x419b6b8c60000ULL,
    0x5a44fff650000ULL,
    0x8179ebb158000ULL,
    0x6a99bc41ec000ULL,
    0xebc7089052000ULL,
    0xe33b4fc859000ULL,
    0x9fe463cc49800ULL,
    0xfb5bb25642c00ULL,
    0xd7f38a417c200ULL,
    0xf75d0e19bd300ULL,
    0x39ee011449880ULL,
    0x25f001242c40ULL,
    0xe57a817b7c2e0ULL,
    0x7889437ebd350ULL,
    0xc8c8e2d4c9808ULL,
    0xbea9f17b82cecULL,
    0x43afeaf71c2beULL,
    0x7322bc88ed37dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0xf400000000000ULL,
    0xf600000000000ULL,
    0xbf00000000000ULL,
    0x7c80000000000ULL,
    0x7c0000000000ULL,
    0x4ee0000000000ULL,
    0x81f0000000000ULL,
    0x9568000000000ULL,
    0xc6bc000000000ULL,
    0x8e42000000000ULL,
    0x18a7000000000ULL,
    0x551800000000ULL,
    0x3233400000000ULL,
    0x4d86a00000000ULL,
    0xa85e300000000ULL,
    0x3fa7480000000ULL,
    0x85c1d40000000ULL,
    0xc7e54a0000000ULL,
    0xe266d70000000ULL,
    0x1e34c88000000ULL,
    0x279594c000000ULL,
    0xb5526a2000000ULL,
    0xfa3ba71000000ULL,
    0xe99d21f800000ULL,
    0x64671ec00000ULL,
    0xdcba6a0a00000ULL,
    0xfb47a6f300000ULL,
    0x893f23b180000ULL,
    0x6f117034c0000ULL,
    0xe483ebc420000ULL,
    0xfbc8e43e30000ULL,
    0xbcfb800528000ULL,
    0x60e841b0ac000ULL,
    0xa2f5201a6e000ULL,
    0x4bfa708b61000ULL,
    0x78786a1df4800ULL,
    0xab20a6ded9c00ULL,
    0x160ea30749a00ULL,
    0x2f1233e7ee900ULL,
    0x848d4be5f4880ULL,
    0xabdad732d9cc0ULL,
    0xe4f6c90d49aa0ULL,
    0x3cf29514ee9f0ULL,
    0xa0e3e85474828ULL,
    0x2f8e70619c34ULL,
    0xbbf382c969a56ULL,
    0xd064412ade94fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x8800000000000ULL,
    0x1c00000000000ULL,
    0x1600000000000ULL,
    0xb900000000000ULL,
    0x3f80000000000ULL,
    0x2a40000000000ULL,
    0xf3e0000000000ULL,
    0x62f0000000000ULL,
    0xf168000000000ULL,
    0x3534000000000ULL,
    0x7bc6000000000ULL,
    0xa4a7000000000ULL,
    0xcc8c800000000ULL,
    0x9ada400000000ULL,
    0x2f3b600000000ULL,
    0x9cd3900000000ULL,
    0xde3e580000000ULL,
    0x5f44640000000ULL,
    0x72785a0000000ULL,
    0x9da3670000000ULL,
    0x3314d98000000ULL,
    0x5089254000000ULL,
    0x4cc7bba000000ULL,
    0x762eb53000000ULL,
    0x335fe3f800000ULL,
    0x7c7dd3d400000ULL,
    0xb0a3390600000ULL,
    0x680f47d00000ULL,
    0xd5c2834980000ULL,
    0x39b940b6c0000ULL,
    0x119e32c20000ULL,
    0x1b9ad15210000ULL,
    0xd84fbb3b38000ULL,
    0x58eab7148c000ULL,
    0x6671e08f0a000ULL,
    0x57aed042e5000ULL,
    0x7c09bada67800ULL,
    0x460db4b76b400ULL,
    0xd11d62f6bee00ULL,
    0xb384910c6ad00ULL,
    0xb45ad92267880ULL,
    0x56ea27636b4c0ULL,
    0x4b653bf0beee0ULL,
    0x6227f5716ad90ULL,
    0xf94e026be7808ULL,
    0x336303d5ab4dcULL,
    0xa62282dc9eef6ULL,
    0x9b4943237ad29ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0x2c00000000000ULL,
    0x1200000000000ULL,
    0x3100000000000ULL,
    0xe80000000000ULL,
    0x1f40000000000ULL,
    0x62e0000000000ULL,
    0x1bf0000000000ULL,
    0x8178000000000ULL,
    0xd02c000000000ULL,
    0x19d6000000000ULL,
    0x49ad000000000ULL,
    0xe710800000000ULL,
    0x699ec00000000ULL,
    0x36dee00000000ULL,
    0xf43a300000000ULL,
    0x3fdd580000000ULL,
    0xe6a5c40000000ULL,
    0x1a8b5a0000000ULL,
    0x1148c50000000ULL,
    0x69fbd88000000ULL,
    0x8066064000000ULL,
    0xf6bd39e000000ULL,
    0xa28036f000000ULL,
    0x4d4e61a800000ULL,
    0xf3e4f23400000ULL,
    0xc563bb9200000ULL,
    0xc62ff74900000ULL,
    0x9ece029980000ULL,
    0x10310166c0000ULL,
    0xf9de817c60000ULL,
    0x39afc0d670000ULL,
    0x6f0062b338000ULL,
    0x1d95f2c4c4000ULL,
    0x8dd383d7a000ULL,
    0xd730376bbf000ULL,
    0x56631f81800ULL,
    0xf778f16d97c00ULL,
    0x672db8d494e00ULL,
    0x685ef7e570d00ULL,
    0xf37082b781880ULL,
    0xd12ec35997c40ULL,
    0x3f46e34694ea0ULL,
    0xf2e630ac70dd0ULL,
    0xe3f3582e018d8ULL,
    0x7d64c63f57c6cULL,
    0x9a2dd83af4eb2ULL,
    0x4cb057a00de1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xf800000000000ULL,
    0xa400000000000ULL,
    0x1a00000000000ULL,
    0xa700000000000ULL,
    0x8880000000000ULL,
    0x6b40000000000ULL,
    0x720000000000ULL,
    0x830000000000ULL,
    0x6cb8000000000ULL,
    0x61fc000000000ULL,
    0x8fc2000000000ULL,
    0x4a6d000000000ULL,
    0xc601800000000ULL,
    0x591ec00000000ULL,
    0x1598200000000ULL,
    0xb4cc300000000ULL,
    0x1cf4d80000000ULL,
    0xc04cfc0000000ULL,
    0x79b6da0000000ULL,
    0xa261fd0000000ULL,
    0xa175a8000000ULL,
    0xcf0f3ec000000ULL,
    0x497796000000ULL,
    0xdd4f0c5000000ULL,
    0xe639a10800000ULL,
    0x99a2f2cc00000ULL,
    0x3274f92a00000ULL,
    0x220cce6100000ULL,
    0x631803b680000ULL,
    0x428c0081c0000ULL,
    0xc45a036460000ULL,
    0xd3a1015630000ULL,
    0xdd7b825368000ULL,
    0xd68fc3b384000ULL,
    0x465ba0e16e000ULL,
    0xbff376ff000ULL,
    0x47ed7ae18e800ULL,
    0x46de0eb40fc00ULL,
    0x77fa231e3de00ULL,
    0xded130bf0c300ULL,
    0x3ed5be98e880ULL,
    0xccde3c780fc40ULL,
    0xf8f4fa343de20ULL,
    0xfa4ccede0c3f0ULL,
    0x2eb8005f0e878ULL,
    0xd2fc00f9cfce4ULL,
    0xc54203505de3aULL,
    0xd22d02883c357ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xd800000000000ULL,
    0x1c00000000000ULL,
    0x4600000000000ULL,
    0xd100000000000ULL,
    0x2080000000000ULL,
    0x4240000000000ULL,
    0xae20000000000ULL,
    0xfb30000000000ULL,
    0xe1a8000000000ULL,
    0xa4fc000000000ULL,
    0x135a000000000ULL,
    0x4eb1000000000ULL,
    0x197a800000000ULL,
    0x7f9e400000000ULL,
    0x87c6600000000ULL,
    0xeee7700000000ULL,
    0xac53b80000000ULL,
    0xdb233c0000000ULL,
    0x11a9ba0000000ULL,
    0xdce23f0000000ULL,
    0x3f5b388000000ULL,
    0xd0b07fc000000ULL,
    0xd46f5be000000ULL,
    0x191a0fb000000ULL,
    0x149c62f800000ULL,
    0x605670f400000ULL,
    0x1529387a00000ULL,
    0x94bd7efd00000ULL,
    0x4e6fda4280000ULL,
    0x2e054ed1c0000ULL,
    0xd508802420000ULL,
    0xda9340f230000ULL,
    0xe546e01148000ULL,
    0x87b831fcf4000ULL,
    0x85e75b6ea2000ULL,
    0x4bd60e8f8d000ULL,
    0x20ee622b3b800ULL,
    0x295b70a1d0c00ULL,
    0x49a9b83c55600ULL,
    0xe23e366c300ULL,
    0xd95b38d33b880ULL,
    0x31b07c55d0cc0ULL,
    0x2cef5a46556a0ULL,
    0x475a0fcb6c330ULL,
    0xfcbc6211bb858ULL,
    0x4a66714410cdcULL,
    0xd4013b82756e6ULL,
    0x72017f895c3e1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0x4c00000000000ULL,
    0x6a00000000000ULL,
    0x8f00000000000ULL,
    0x4380000000000ULL,
    0x240000000000ULL,
    0x7fa0000000000ULL,
    0xf4b0000000000ULL,
    0x5928000000000ULL,
    0xfa6c000000000ULL,
    0x4756000000000ULL,
    0xa339000000000ULL,
    0x6d78800000000ULL,
    0xb8c9c00000000ULL,
    0xe376200000000ULL,
    0x29d3500000000ULL,
    0x29e9980000000ULL,
    0x5d14ac0000000ULL,
    0x609f9a0000000ULL,
    0xdbddaf0000000ULL,
    0xdaef1a8000000ULL,
    0x7c886c4000000ULL,
    0x89c7396000000ULL,
    0x39fe3d7000000ULL,
    0x850820c800000ULL,
    0xfc86534c00000ULL,
    0x49c71bc600000ULL,
    0x19e46c4500000ULL,
    0x551138b880000ULL,
    0x4873fd7c0000ULL,
    0x5d0a2bc20000ULL,
    0x73ff924e70000ULL,
    0xda193be978000ULL,
    0x471b3cce94000ULL,
    0x78ea07a9a000ULL,
    0xc5a9058b3000ULL,
    0x2ebfba67c2800ULL,
    0x1e37fe543e400ULL,
    0xfdfe029244e00ULL,
    0x4b1500571cf00ULL,
    0x8d8e82afc2880ULL,
    0x7340c1183e4c0ULL,
    0x4526a35444e20ULL,
    0xa87693121cfd0ULL,
    0xa449b89742878ULL,
    0x5abefe8ffe48cULL,
    0x182e828864e4aULL,
    0xc8f0c32c6cf5fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0x1400000000000ULL,
    0xe600000000000ULL,
    0x9900000000000ULL,
    0x4780000000000ULL,
    0xef40000000000ULL,
    0x4de0000000000ULL,
    0x12f0000000000ULL,
    0xbc78000000000ULL,
    0xdca4000000000ULL,
    0x421e000000000ULL,
    0x1711000000000ULL,
    0xd29b800000000ULL,
    0x8adb400000000ULL,
    0x25b2600000000ULL,
    0x9595900000000ULL,
    0xe052f80000000ULL,
    0x5b73840000000ULL,
    0x8634fa0000000ULL,
    0x14c6850000000ULL,
    0xe8b1798000000ULL,
    0xd40cc64000000ULL,
    0x46189a6000000ULL,
    0xa902161000000ULL,
    0x6f9801f800000ULL,
    0x354007400000ULL,
    0xbfe602e600000ULL,
    0x6df502c900000ULL,
    0x62e5805180000ULL,
    0x747a41c2c0000ULL,
    0xe0b1e14060000ULL,
    0x481ad2e050000ULL,
    0x7c069bd018000ULL,
    0xea1314d814000ULL,
    0xbb03838476000ULL,
    0xc08f404e79000ULL,
    0x25d463a53f800ULL,
    0x3209013ef400ULL,
    0x29577ab594600ULL,
    0xa4f9c51cb1500ULL,
    0x9d7d1bdd3f880ULL,
    0x6f385627ef440ULL,
    0x9b49e233946e0ULL,
    0x3bfed1c5b15d0ULL,
    0x93f89a74bf878ULL,
    0x3ff217912f454ULL,
    0x2de00195f4606ULL,
    0x82f001ece1549ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xc800000000000ULL,
    0x4c00000000000ULL,
    0x3a00000000000ULL,
    0x1f00000000000ULL,
    0x9480000000000ULL,
    0xebc0000000000ULL,
    0x3760000000000ULL,
    0x8270000000000ULL,
    0x31f8000000000ULL,
    0x6aac000000000ULL,
    0x701a000000000ULL,
    0x2813000000000ULL,
    0x5c12800000000ULL,
    0x4201c00000000ULL,
    0x9b14200000000ULL,
    0xe29fb00000000ULL,
    0xcecb480000000ULL,
    0xbceae40000000ULL,
    0xfd294a0000000ULL,
    0xed55e50000000ULL,
    0xdfa1c88000000ULL,
    0xc38727c000000ULL,
    0x73476ba000000ULL,
    0x46a956f000000ULL,
    0x1a00034800000ULL,
    0xef00028c00000ULL,
    0xfc80031a00000ULL,
    0x17c0002300000ULL,
    0xc560014780000ULL,
    0xd17000eac0000ULL,
    0x9f7802a060000ULL,
    0x9e6c007070000ULL,
    0xd3fa038808000ULL,
    0x41a300ac3c000ULL,
    0x5a8a80aa42000ULL,
    0xaaddc20b3b000ULL,
    0xdaf6221b92800ULL,
    0xa020b38885c00ULL,
    0xe2c3c9e75fa00ULL,
    0xd6f824d19e500ULL,
    0x3a2fe9d392880ULL,
    0x4dcb96c485c40ULL,
    0x8a7ea15d5faa0ULL,
    0x9df272029e5b0ULL,
    0x40a56bdc12848ULL,
    0x171654a245c0cULL,
    0x388880e73fa9aULL,
    0xc1d2c251ee5afULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xa800000000000ULL,
    0x2c00000000000ULL,
    0xfe00000000000ULL,
    0x5100000000000ULL,
    0xef80000000000ULL,
    0x640000000000ULL,
    0x5760000000000ULL,
    0x1070000000000ULL,
    0x22e8000000000ULL,
    0x524000000000ULL,
    0xfd0e000000000ULL,
    0xd18b000000000ULL,
    0x3756800000000ULL,
    0xe8e5400000000ULL,
    0xbe35600000000ULL,
    0x5985900000000ULL,
    0xeb5cd80000000ULL,
    0x8efbf40000000ULL,
    0x3b3ada0000000ULL,
    0xcc14f70000000ULL,
    0x6e025b8000000ULL,
    0x990ab64000000ULL,
    0x9389b92000000ULL,
    0x504e66f000000ULL,
    0x2a6e02d800000ULL,
    0x1fb03b400000ULL,
    0x75be836600000ULL,
    0xbdc140ad00000ULL,
    0xeb3b63bb80000ULL,
    0xa40e91acc0000ULL,
    0x220a599020000ULL,
    0x371eb50850000ULL,
    0x6a8fbb9c18000ULL,
    0x93d1640a7c000ULL,
    0xd23e817f42000ULL,
    0x78142f0e9000ULL,
    0x8a5b627a17800ULL,
    0x997e90673f400ULL,
    0xb9625b24c4e00ULL,
    0x497ab64c40900ULL,
    0xd161bb2217880ULL,
    0x56a65933f4c0ULL,
    0x7f600162c4e60ULL,
    0xfc70001140950ULL,
    0xbce8024197828ULL,
    0x424038bff4ecULL,
    0xba8e0194e4e9eULL,
    0xfbcb02b410901ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xc800000000000ULL,
    0x7c00000000000ULL,
    0xc200000000000ULL,
    0x1300000000000ULL,
    0x6180000000000ULL,
    0xc340000000000ULL,
    0x5820000000000ULL,
    0xa130000000000ULL,
    0x5fb8000000000ULL,
    0x9774000000000ULL,
    0xd302000000000ULL,
    0x418b000000000ULL,
    0x134c800000000ULL,
    0x2038c00000000ULL,
    0x153ca00000000ULL,
    0xe1a1900000000ULL,
    0x467a780000000ULL,
    0xa19f3c0000000ULL,
    0xe3407a0000000ULL,
    0x88203d0000000ULL,
    0xd92efb8000000ULL,
    0xeba3ffc000000ULL,
    0x2966d8a000000ULL,
    0x20eae3000000ULL,
    0x330203a800000ULL,
    0xb18b027c00000ULL,
    0xbb4c802e00000ULL,
    0xec38c07100000ULL,
    0x1f3ca0ce80000ULL,
    0x8ea19035c0000ULL,
    0xe5fa7a9060000ULL,
    0x71df3f9810000ULL,
    0xdae078d468000ULL,
    0xea503fd23c000ULL,
    0xdeb6f89f5a000ULL,
    0xdde7fe9fdf000ULL,
    0xa5dcda8b28800ULL,
    0xd4f1ae6d99400ULL,
    0xf34c81e421600ULL,
    0xd038c2fa35b00ULL,
    0xbd3ca2a328880ULL,
    0x2da193d199440ULL,
    0x4c7a796a21660ULL,
    0xce9f3cbb35bb0ULL,
    0x40c079c5a8848ULL,
    0x58603d985943cULL,
    0xe08ef9d4416a2ULL,
    0x89d3fe5225ba3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0x2c00000000000ULL,
    0x9600000000000ULL,
    0xcf00000000000ULL,
    0x6280000000000ULL,
    0xdd40000000000ULL,
    0xfea0000000000ULL,
    0x4bb0000000000ULL,
    0x9038000000000ULL,
    0x10ec000000000ULL,
    0xf986000000000ULL,
    0x45c9000000000ULL,
    0xdae4800000000ULL,
    0x4892400000000ULL,
    0x9c59a00000000ULL,
    0x8f36f00000000ULL,
    0xca66d80000000ULL,
    0xa8d6cc0000000ULL,
    0x7d78da0000000ULL,
    0xa643cd0000000ULL,
    0x9622588000000ULL,
    0xc7f48c4000000ULL,
    0x47197b6000000ULL,
    0xa6993cf000000ULL,
    0x3f42813800000ULL,
    0xdbab43cc00000ULL,
    0x882522b200000ULL,
    0xccf8b1b900000ULL,
    0xc7817bc880000ULL,
    0x5ec53dd7c0000ULL,
    0x27c806860000ULL,
    0xccce402470000ULL,
    0xcf67a15678000ULL,
    0x7b53f24f54000ULL,
    0x79a45ad7ea000ULL,
    0xd3d8de81d000ULL,
    0x5f7df8e40c800ULL,
    0x634b7c7606c00ULL,
    0xa5bb20df77e00ULL,
    0x332db39fedf00ULL,
    0x447bf95c0c880ULL,
    0xbbc27ffa06c40ULL,
    0x21ffa30d77ea0ULL,
    0x600ff3d6edf50ULL,
    0xb01a5b2c8c878ULL,
    0x8188da1c6c6cULL,
    0x841f7ab717e36ULL,
    0x42103ebb9df9fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0x5c00000000000ULL,
    0x3a00000000000ULL,
    0xe700000000000ULL,
    0x9980000000000ULL,
    0x440000000000ULL,
    0xbb20000000000ULL,
    0xa30000000000ULL,
    0x46b8000000000ULL,
    0x407c000000000ULL,
    0xec8a000000000ULL,
    0xf6d3000000000ULL,
    0xc4f0800000000ULL,
    0xf643c00000000ULL,
    0x4031200000000ULL,
    0x89b7100000000ULL,
    0x15e6180000000ULL,
    0x2acd3c0000000ULL,
    0xbef41a0000000ULL,
    0x31523d0000000ULL,
    0x69b69a8000000ULL,
    0xc5feffc000000ULL,
    0xf2dd3ba000000ULL,
    0x1ae92eb000000ULL,
    0x1f42834800000ULL,
    0xe8acc12c00000ULL,
    0x816ba32e00000ULL,
    0x8817d03d00000ULL,
    0x3c1fb94280000ULL,
    0xaa05eed5c0000ULL,
    0x1f09237860000ULL,
    0x8d8b13a430000ULL,
    0x624c1aa248000ULL,
    0x662e3c237c000ULL,
    0x74bc98778a000ULL,
    0xdb6dfd5b73000ULL,
    0xff0dbad3ce800ULL,
    0x5d9aef796f400ULL,
    0xba4ba030b6e00ULL,
    0xc227d02eabb00ULL,
    0x5aa7bb1bce880ULL,
    0x5a79ed956f440ULL,
    0xbb8322beb6e20ULL,
    0x275813a3abbb0ULL,
    0x9cbc9b914e8c8ULL,
    0x776dfdacaf41cULL,
    0xad0db948d6e1aULL,
    0x569aee8a9bb57ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x5800000000000ULL,
    0xac00000000000ULL,
    0xde00000000000ULL,
    0xe900000000000ULL,
    0x9980000000000ULL,
    0x82c0000000000ULL,
    0x9d60000000000ULL,
    0x3e70000000000ULL,
    0x6cf8000000000ULL,
    0x8ac000000000ULL,
    0x239a000000000ULL,
    0xb1c1000000000ULL,
    0x67fc800000000ULL,
    0xe23400000000ULL,
    0x85c0600000000ULL,
    0xf5fc500000000ULL,
    0x2937880000000ULL,
    0x1d48e40000000ULL,
    0x1ab58a0000000ULL,
    0x4495e50000000ULL,
    0x894b088000000ULL,
    0x38aba7c000000ULL,
    0x3b95e9e000000ULL,
    0xbdd9b57000000ULL,
    0x9e4829800000ULL,
    0xbf3f42d400000ULL,
    0xb04262fe00000ULL,
    0xa92151df00000ULL,
    0x5d490a0b80000ULL,
    0xbab6a7e6c0000ULL,
    0xf48b68d860000ULL,
    0xd157f47470000ULL,
    0x94a261ae08000ULL,
    0xe59153d774000ULL,
    0x54d10937fa000ULL,
    0x906aa654f9000ULL,
    0x3de96aad7b800ULL,
    0x2d3af51ec8400ULL,
    0x9744e23c45a00ULL,
    0x31b313325bd00ULL,
    0xb20de8b57b880ULL,
    0xd705b40ac8440ULL,
    0x6086832245aa0ULL,
    0xf352429d5bdb0ULL,
    0xeba4e0a6fb8d8ULL,
    0xd10312f8084ecULL,
    0x6595eae425a7eULL,
    0x14d9b7462bd59ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x800000000000ULL,
    0x5400000000000ULL,
    0xd600000000000ULL,
    0xe500000000000ULL,
    0xab80000000000ULL,
    0xe940000000000ULL,
    0x1b20000000000ULL,
    0x70b0000000000ULL,
    0x39f8000000000ULL,
    0x92c4000000000ULL,
    0xba7e000000000ULL,
    0xdf87000000000ULL,
    0xcf49800000000ULL,
    0x462d400000000ULL,
    0xd736e00000000ULL,
    0x5aaa100000000ULL,
    0xeefa780000000ULL,
    0x525ffc0000000ULL,
    0x41b3fa0000000ULL,
    0x472bd0000000ULL,
    0x7e85198000000ULL,
    0x5ad8ac4000000ULL,
    0x4e7f61a000000ULL,
    0xb987511000000ULL,
    0x724c98f800000ULL,
    0xb1b5ef8400000ULL,
    0xbc6983d600000ULL,
    0x729d406100000ULL,
    0xd0cee21a80000ULL,
    0x296e10b3c0000ULL,
    0x21047a0560000ULL,
    0xd598fd7cd0000ULL,
    0xe85a7b3a98000ULL,
    0x3eaffe63cc000ULL,
    0x20ebfa1d3e000ULL,
    0xb46bfc8b9000ULL,
    0x38231b4485800ULL,
    0xd62bafdec1400ULL,
    0x7f30e1c5f4600ULL,
    0xbea911f472b00ULL,
    0x60edf95e05880ULL,
    0xeb45bf6d01440ULL,
    0x68349bc0946e0ULL,
    0xde31ec88a2b50ULL,
    0x2b3782649d888ULL,
    0x68aa410ecd414ULL,
    0x85ff61ddaa636ULL,
    0x40c753401bbb5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x5400000000000ULL,
    0xfa00000000000ULL,
    0x8300000000000ULL,
    0xca80000000000ULL,
    0x95c0000000000ULL,
    0xb7a0000000000ULL,
    0x2d30000000000ULL,
    0x70f8000000000ULL,
    0x1b5c000000000ULL,
    0x307a000000000ULL,
    0x7581000000000ULL,
    0x8957800000000ULL,
    0x9f72c00000000ULL,
    0x8103a00000000ULL,
    0xcd9d700000000ULL,
    0x2d55080000000ULL,
    0x5d745c0000000ULL,
    0xe6028a0000000ULL,
    0xe5069d0000000ULL,
    0x6f812a8000000ULL,
    0xda5befc000000ULL,
    0xddf422e000000ULL,
    0xe8dfb29000000ULL,
    0xcc2eab0800000ULL,
    0x2d752d9c00000ULL,
    0x9e0d82a200000ULL,
    0x2103c00b00000ULL,
    0x3d8c229c80000ULL,
    0x1543b22fc0000ULL,
    0xb974a85760000ULL,
    0x4042fdeb0000ULL,
    0x3202023488000ULL,
    0x5f1d0203b4000ULL,
    0xcc8d808d56000ULL,
    0xa0c3c351bf000ULL,
    0x502c225e38800ULL,
    0xb73b1797ac00ULL,
    0x1b0caabfe4a00ULL,
    0xde982e5f1fb00ULL,
    0x4fd800c2b8880ULL,
    0x84ac0356bac40ULL,
    0xffa202e884a20ULL,
    0xb12d0181afbb0ULL,
    0x56f582f630898ULL,
    0x9e5fc1550ec14ULL,
    0xcff62265d2adaULL,
    0x7c2b2d010b33ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x4800000000000ULL,
    0x9c00000000000ULL,
    0x4200000000000ULL,
    0x5100000000000ULL,
    0xc280000000000ULL,
    0x25c0000000000ULL,
    0x6560000000000ULL,
    0x9ff0000000000ULL,
    0xfe28000000000ULL,
    0xad5c000000000ULL,
    0xeda6000000000ULL,
    0x8a07000000000ULL,
    0xd08800000000ULL,
    0x2086c00000000ULL,
    0x24dbe00000000ULL,
    0xeff3300000000ULL,
    0x2635480000000ULL,
    0xd950040000000ULL,
    0x63bdca0000000ULL,
    0xd116c70000000ULL,
    0x286288000000ULL,
    0x85d5f64000000ULL,
    0x357b61a000000ULL,
    0xd7e9f03000000ULL,
    0x6228aaa800000ULL,
    0xef54367400000ULL,
    0xbca8012200000ULL,
    0x489c00a300000ULL,
    0x28c6011780000ULL,
    0x45f702ecc0000ULL,
    0xbb20804720000ULL,
    0x11dac22db0000ULL,
    0x8b7de0d3b8000ULL,
    0x34f433a6ac000ULL,
    0xe9bdcb701a000ULL,
    0xdc16c48877000ULL,
    0x2206288455800ULL,
    0xa115f46a0bc00ULL,
    0xda9b634738a00ULL,
    0xf1d9f0ade7d00ULL,
    0xbb60a993d5880ULL,
    0x8cf83686cbcc0ULL,
    0x6da6030018aa0ULL,
    0x4a07028057d50ULL,
    0xad0881406d8c8ULL,
    0x7086c12067c5cULL,
    0x6cdbe27002ae2ULL,
    0x73f3310820d01ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0x400000000000ULL,
    0x7600000000000ULL,
    0xdf00000000000ULL,
    0x8480000000000ULL,
    0x8440000000000ULL,
    0xfca0000000000ULL,
    0x5c30000000000ULL,
    0x7e78000000000ULL,
    0xf5c4000000000ULL,
    0x617a000000000ULL,
    0xd145000000000ULL,
    0x7521800000000ULL,
    0x65e5400000000ULL,
    0x1503a00000000ULL,
    0x6996b00000000ULL,
    0x49c9780000000ULL,
    0x53784c0000000ULL,
    0x9848fa0000000ULL,
    0x5ead0d0000000ULL,
    0x5d33598000000ULL,
    0x69ffbcc000000ULL,
    0xbf00216000000ULL,
    0xb482f0d000000ULL,
    0x5c495ab800000ULL,
    0x88babeac00000ULL,
    0x1221a12a00000ULL,
    0xa567b07d00000ULL,
    0x74afbb880000ULL,
    0x3a2c0dddc0000ULL,
    0xa968daf960000ULL,
    0xad5ffda6d0000ULL,
    0xe72203d688000ULL,
    0xbcf1025ec4000ULL,
    0x768380da8e000ULL,
    0x6d5442c491000ULL,
    0x472022cfe0800ULL,
    0x2cf2f1681bc00ULL,
    0x3e915a944d600ULL,
    0x514ebfc63ef00ULL,
    0x3523a17760880ULL,
    0x85e6b0b5dbc40ULL,
    0x65117a6d2d6e0ULL,
    0x518c4f60eef70ULL,
    0x4dcafb21e88b8ULL,
    0x256c0e2b1fc44ULL,
    0x4748dbd7a3696ULL,
    0xda2ffd747ffafULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0x5c00000000000ULL,
    0xaa00000000000ULL,
    0xb500000000000ULL,
    0x1780000000000ULL,
    0x4240000000000ULL,
    0x620000000000ULL,
    0xddb0000000000ULL,
    0x8de8000000000ULL,
    0xacc000000000ULL,
    0x6fa000000000ULL,
    0x9e5b000000000ULL,
    0x6c3c800000000ULL,
    0x48a3c00000000ULL,
    0xaa66600000000ULL,
    0x84100000000ULL,
    0x5ccd980000000ULL,
    0xe9f43c0000000ULL,
    0x54d11a0000000ULL,
    0x55e7ff0000000ULL,
    0xeedf7a8000000ULL,
    0x98efefc000000ULL,
    0xed48e3a000000ULL,
    0x2cb0d2b000000ULL,
    0xf46d7b0800000ULL,
    0x5388ef6400000ULL,
    0x2c4661d600000ULL,
    0x1d3411ad00000ULL,
    0xf1259a6e80000ULL,
    0xd3383c6bc0000ULL,
    0x1a2b195120000ULL,
    0x97bcfc24f0000ULL,
    0x28e3fa1cc8000ULL,
    0x654c2c309c000ULL,
    0x50ae8062ba000ULL,
    0x6e74c0c9d3000ULL,
    0xae80e01203800ULL,
    0x67ccd34b33400ULL,
    0xf57f7b8bfde00ULL,
    0x421fec017fb00ULL,
    0x1900e27c83880ULL,
    0xd58cd320f34c0ULL,
    0x9b5f78dadde20ULL,
    0xf3afef258fb30ULL,
    0x76e8e2e04b8c8ULL,
    0x3640d0d06f49cULL,
    0x20257b1867e8aULL,
    0x9ab4ed5c5cb85ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x9800000000000ULL,
    0x5c00000000000ULL,
    0x1200000000000ULL,
    0x700000000000ULL,
    0x4e80000000000ULL,
    0x4ac0000000000ULL,
    0xf7e0000000000ULL,
    0x2b70000000000ULL,
    0xa6a8000000000ULL,
    0xbfdc000000000ULL,
    0x6e7a000000000ULL,
    0xa721000000000ULL,
    0x6f86800000000ULL,
    0xb556c00000000ULL,
    0xaaab200000000ULL,
    0xe5de700000000ULL,
    0xbd60e80000000ULL,
    0x8fb3b40000000ULL,
    0xf44e6a0000000ULL,
    0x4539770000000ULL,
    0x909f488000000ULL,
    0xb7c605c000000ULL,
    0xda7922e000000ULL,
    0xd12370b000000ULL,
    0x969c68a800000ULL,
    0x76c477ac00000ULL,
    0xf5e3c91200000ULL,
    0x471c74700000ULL,
    0x2c34822d80000ULL,
    0xbb1bc05ac0000ULL,
    0x8c9fa09f20000ULL,
    0xc5c5b0c9b0000ULL,
    0xcd7f4ad4c8000ULL,
    0xb7b605a234000ULL,
    0x185121ef52000ULL,
    0xcf3f7081af000ULL,
    0xcb866ac8de800ULL,
    0xeb55749866c00ULL,
    0x97ad48c415200ULL,
    0x684b075e5c100ULL,
    0xf72da0655e880ULL,
    0x2788b102a6cc0ULL,
    0x614bcabb352a0ULL,
    0xccadc727ec1b0ULL,
    0x34ce821996818ULL,
    0xbafac30c92c9cULL,
    0x9ef92246672b2ULL,
    0xe370e1431b7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x7800000000000ULL,
    0x1400000000000ULL,
    0x2a00000000000ULL,
    0x6900000000000ULL,
    0xe80000000000ULL,
    0x54c0000000000ULL,
    0x8220000000000ULL,
    0xc6b0000000000ULL,
    0x4b68000000000ULL,
    0xcfcc000000000ULL,
    0x61b6000000000ULL,
    0xeef1000000000ULL,
    0xb08f800000000ULL,
    0xd7d4c00000000ULL,
    0xc5a4e00000000ULL,
    0xccfcb00000000ULL,
    0xe583e80000000ULL,
    0x9f40cc0000000ULL,
    0xc6646a0000000ULL,
    0x3580f0000000ULL,
    0x10768a8000000ULL,
    0xc55be4000000ULL,
    0x19fae36000000ULL,
    0x6d01b21000000ULL,
    0xbc9a69c800000ULL,
    0x59d50cb400000ULL,
    0x8ea70b7200000ULL,
    0x977c7fa700000ULL,
    0xf9c7801a80000ULL,
    0x1ea8c1adc0000ULL,
    0xcf7ae27f20000ULL,
    0xbdc1b066f0000ULL,
    0x4cba6bebf8000ULL,
    0xb2650e6a34000ULL,
    0x994f08fb46000ULL,
    0x71707c2ca7000ULL,
    0x3ed180e0f5800ULL,
    0xb29c3dea5c00ULL,
    0xb83d600782200ULL,
    0x37a9732452f00ULL,
    0x21e08a7a75880ULL,
    0x9914bc3365cc0ULL,
    0xe69d6118a2220ULL,
    0x18d97152a2f50ULL,
    0xec2888598d8f8ULL,
    0xfda8beed51cd4ULL,
    0x38e36291e420aULL,
    0xbf9472d905f39ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x8400000000000ULL,
    0x8200000000000ULL,
    0x5500000000000ULL,
    0x3780000000000ULL,
    0xb040000000000ULL,
    0x3be0000000000ULL,
    0x670000000000ULL,
    0x3738000000000ULL,
    0xb15c000000000ULL,
    0xc66e000000000ULL,
    0x5733000000000ULL,
    0x814f800000000ULL,
    0xfe69c00000000ULL,
    0xfb3c600000000ULL,
    0x8743100000000ULL,
    0x2963d80000000ULL,
    0x99a1440000000ULL,
    0x945a0000000ULL,
    0xa2d4850000000ULL,
    0xa4263a8000000ULL,
    0x31d4954000000ULL,
    0x24b2622000000ULL,
    0xd300135000000ULL,
    0xa0945a2800000ULL,
    0xf2d4870400000ULL,
    0xac26387600000ULL,
    0xa5d4948500000ULL,
    0x8eb260b980000ULL,
    0x20012ccc0000ULL,
    0x151458cd60000ULL,
    0x1794864db0000ULL,
    0xa0463b82d8000ULL,
    0x13e495c47c000ULL,
    0x826a619666000ULL,
    0xb52c12b51b000ULL,
    0xe4425ae1f7800ULL,
    0xf1fb8430a7400ULL,
    0xe767b9d720e00ULL,
    0xbabe571a9e500ULL,
    0xf81980d877880ULL,
    0xcc06c3bc67440ULL,
    0x361de33a40e20ULL,
    0xef19d2072e510ULL,
    0xce903972af8a8ULL,
    0x81cb957c1b4c4ULL,
    0x5cabe0da26ea2ULL,
    0x5f06d13735545ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x4800000000000ULL,
    0x8c00000000000ULL,
    0x9a00000000000ULL,
    0x5700000000000ULL,
    0x4b80000000000ULL,
    0x5040000000000ULL,
    0xca0000000000ULL,
    0x6530000000000ULL,
    0xe678000000000ULL,
    0x975c000000000ULL,
    0x1f22000000000ULL,
    0xa173000000000ULL,
    0xc4ce800000000ULL,
    0xfb67c00000000ULL,
    0xf3d2200000000ULL,
    0xe0ffd00000000ULL,
    0x9b85180000000ULL,
    0x28462c0000000ULL,
    0xe8b39a0000000ULL,
    0x33ded0000000ULL,
    0x6363ba8000000ULL,
    0x7c13cc000000ULL,
    0x9ef023a000000ULL,
    0xaa8cd2f000000ULL,
    0xe6cb9bc800000ULL,
    0x2861eeac00000ULL,
    0x8e41b86600000ULL,
    0xdb23f2d00000ULL,
    0xc3bea1f480000ULL,
    0xdaab1149c0000ULL,
    0xc839bb2560000ULL,
    0xaaee3d78b0000ULL,
    0xb49ca29fe8000ULL,
    0x87d813207c000ULL,
    0xdef73b3052000ULL,
    0x8a89fee823000ULL,
    0x96ce829c7a800ULL,
    0x6067c10e6c400ULL,
    0x25223f106e00ULL,
    0x97bfd39ac8f00ULL,
    0x94a519e8fa880ULL,
    0x91362c87ac440ULL,
    0x986b9b7466e20ULL,
    0xa651ec1278f70ULL,
    0xd1b9b8bf128c8ULL,
    0x61ae3d0bd04ccULL,
    0x49bca22234ebaULL,
    0x95a810d75bf27ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0x400000000000ULL,
    0x6600000000000ULL,
    0xc500000000000ULL,
    0x5780000000000ULL,
    0xcac0000000000ULL,
    0x99a0000000000ULL,
    0x6c30000000000ULL,
    0xe5f8000000000ULL,
    0x3ecc000000000ULL,
    0x37b2000000000ULL,
    0xd3b000000000ULL,
    0xa467800000000ULL,
    0xb911400000000ULL,
    0xfd86e00000000ULL,
    0xcdd5900000000ULL,
    0x1d34880000000ULL,
    0xbc6ddc0000000ULL,
    0xe50b0a0000000ULL,
    0x67809d0000000ULL,
    0xe2c7e88000000ULL,
    0xeda20cc000000ULL,
    0x8226e2a000000ULL,
    0x24e590d000000ULL,
    0xf4c893800000ULL,
    0x3861dec400000ULL,
    0xc31908e600000ULL,
    0x28b9d2f00000ULL,
    0xc5586a2080000ULL,
    0xaf7f4d4fc0000ULL,
    0x1f92020f60000ULL,
    0x2ecb02b0f0000ULL,
    0x2fbf8017a8000ULL,
    0x512d43bb1c000ULL,
    0x3e6ce33eb6000ULL,
    0x9612920499000ULL,
    0xd190b19cd800ULL,
    0xf38b9c683f400ULL,
    0xdcd86bbc37a00ULL,
    0xd4bf4c4254300ULL,
    0x3fb201b94d880ULL,
    0x493b01e7ff440ULL,
    0x6267811357aa0ULL,
    0xc114322a4370ULL,
    0x2206e016e5808ULL,
    0x3159058e3444ULL,
    0xe294886be1ac6ULL,
    0x155ddcd93d3b5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0x9400000000000ULL,
    0xaa00000000000ULL,
    0xb900000000000ULL,
    0x7780000000000ULL,
    0x87c0000000000ULL,
    0x5b60000000000ULL,
    0xc2f0000000000ULL,
    0xbe28000000000ULL,
    0x97c4000000000ULL,
    0xd36a000000000ULL,
    0x8ef5000000000ULL,
    0x7033800000000ULL,
    0x3cd4c00000000ULL,
    0x89fae00000000ULL,
    0xd4b9900000000ULL,
    0x159aa80000000ULL,
    0xd2d1a40000000ULL,
    0x72e12a0000000ULL,
    0xa631670000000ULL,
    0x93d9cb8000000ULL,
    0x3179f4c000000ULL,
    0x1bfae16000000ULL,
    0x39b9929000000ULL,
    0xa81aa9c800000ULL,
    0x1c11a77400000ULL,
    0xe6012b9600000ULL,
    0x770165ab00000ULL,
    0xdc91c99b80000ULL,
    0xdd4df730c0000ULL,
    0x138e32b20000ULL,
    0xa74891dbd0000ULL,
    0x5023291088000ULL,
    0x6cc0641b04000ULL,
    0xa1e04be3c6000ULL,
    0x8a8353c97000ULL,
    0x939b81310d800ULL,
    0x55d0c2aaa1400ULL,
    0x1670e29a6f200ULL,
    0xef7c923123d00ULL,
    0x38e12b2a8d880ULL,
    0x2f31655a614c0ULL,
    0x3c59cad14f260ULL,
    0xd2b9f67af3df0ULL,
    0x529ae07205838ULL,
    0xd64992f565454ULL,
    0xcbb2abc4892caULL,
    0xb515a47d64d49ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0xdc00000000000ULL,
    0xde00000000000ULL,
    0x6d00000000000ULL,
    0x9080000000000ULL,
    0x6dc0000000000ULL,
    0x1e60000000000ULL,
    0x34f0000000000ULL,
    0xf628000000000ULL,
    0xebcc000000000ULL,
    0x3f72000000000ULL,
    0x127d000000000ULL,
    0x62fa800000000ULL,
    0x5f2b400000000ULL,
    0x94fe00000000ULL,
    0xd9b1700000000ULL,
    0x2d94880000000ULL,
    0x754d240000000ULL,
    0xb7a60a0000000ULL,
    0x589a670000000ULL,
    0x49d3ea8000000ULL,
    0x3c6a16c000000ULL,
    0xf7e7e26000000ULL,
    0x33bd71b000000ULL,
    0xca868b1800000ULL,
    0x92c0270400000ULL,
    0x55f48bda00000ULL,
    0x30bd247300000ULL,
    0x2f0e0bc380000ULL,
    0xc3966610c0000ULL,
    0xc041eb8320000ULL,
    0xab27177bf0000ULL,
    0xa3556244e8000ULL,
    0xdeaa333104000ULL,
    0x9a136bf482000ULL,
    0xdf00542969000ULL,
    0x3b8882f09d800ULL,
    0x3c5642f35cc00ULL,
    0x5356303a7a00ULL,
    0xf65a3170ac500ULL,
    0x923b6bb31d880ULL,
    0x29cc56239ccc0ULL,
    0xac7a80e087a20ULL,
    0x9feb40bb5c570ULL,
    0xa7afe06ff58b8ULL,
    0xf08170d698c1cULL,
    0xfddc88ae05afeULL,
    0x767126513551dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x8800000000000ULL,
    0xd400000000000ULL,
    0x4600000000000ULL,
    0xd900000000000ULL,
    0x8180000000000ULL,
    0xd540000000000ULL,
    0xd060000000000ULL,
    0xc0f0000000000ULL,
    0xe038000000000ULL,
    0xf304000000000ULL,
    0x282000000000ULL,
    0xffdb000000000ULL,
    0xabbd800000000ULL,
    0x1553400000000ULL,
    0x7061600000000ULL,
    0x90efd00000000ULL,
    0xd83fb80000000ULL,
    0xaf05940000000ULL,
    0x90803a0000000ULL,
    0x60cdd50000000ULL,
    0xf33cd98000000ULL,
    0x418147c000000ULL,
    0x755a03a000000ULL,
    0x806f009000000ULL,
    0xf8e780b800000ULL,
    0xbc3c42f400000ULL,
    0x6106e13e00000ULL,
    0x9d93936100000ULL,
    0xa759585f80000ULL,
    0xff660732c0000ULL,
    0x106160c760000ULL,
    0x60efd29ad0000ULL,
    0xb03fbb4b68000ULL,
    0xcb0595f0bc000ULL,
    0x5e8039041e000ULL,
    0x6dcdd63631000ULL,
    0x34bcdbfd7b800ULL,
    0x4dc1454d9e400ULL,
    0x24ba01a98d600ULL,
    0x95df036fe4100ULL,
    0xc8bf831afb880ULL,
    0x8fc8408b5e440ULL,
    0x83bce050ed6e0ULL,
    0x914c9294341b0ULL,
    0xe66d80e13808ULL,
    0x15ee464922494ULL,
    0xcbbd8393936a6ULL,
    0xe5534338d5169ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0xfc00000000000ULL,
    0xae00000000000ULL,
    0x1100000000000ULL,
    0x2480000000000ULL,
    0x8d40000000000ULL,
    0x8860000000000ULL,
    0x45f0000000000ULL,
    0x6828000000000ULL,
    0x870c000000000ULL,
    0x918e000000000ULL,
    0xd3d7000000000ULL,
    0xd62c800000000ULL,
    0x5e1b400000000ULL,
    0x6910600000000ULL,
    0x6097700000000ULL,
    0x4745180000000ULL,
    0xc365340000000ULL,
    0x22679a0000000ULL,
    0x7ee9770000000ULL,
    0xb7bb7a8000000ULL,
    0xf5d5454000000ULL,
    0xfb2602a000000ULL,
    0xf49b02b000000ULL,
    0x6542821800000ULL,
    0xc47c41f400000ULL,
    0x73f4e20a00000ULL,
    0x8d30318100000ULL,
    0xf193780b80000ULL,
    0xa3d94766c0000ULL,
    0x6e28011320000ULL,
    0x3a0c023ef0000ULL,
    0x330e014718000ULL,
    0x23970204ac000ULL,
    0xdccc809e2a000ULL,
    0x7bab43db01000ULL,
    0x3d86202cb800ULL,
    0x3e2b73f16cc00ULL,
    0x12031b73e7e00ULL,
    0x5f0e35c2a4100ULL,
    0x858d18114b880ULL,
    0x31d93563accc0ULL,
    0xf121986ac7e20ULL,
    0x9f82767d54190ULL,
    0x92d1f95dd3888ULL,
    0xdaa50401c0c3cULL,
    0xbf5ce3e7cde8eULL,
    0x477c3098a5181ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0xa400000000000ULL,
    0x200000000000ULL,
    0x3900000000000ULL,
    0x9380000000000ULL,
    0x41c0000000000ULL,
    0x80a0000000000ULL,
    0x3e30000000000ULL,
    0xf678000000000ULL,
    0xca14000000000ULL,
    0xd502000000000ULL,
    0x3d81000000000ULL,
    0xb6d9800000000ULL,
    0xd43fc00000000ULL,
    0x7363e00000000ULL,
    0x9f8d100000000ULL,
    0x9fd9680000000ULL,
    0x4fa05c0000000ULL,
    0xfea2ea0000000ULL,
    0xe12e9d0000000ULL,
    0x3ee08b8000000ULL,
    0x11484dc000000ULL,
    0xdff802e000000ULL,
    0x66d4035000000ULL,
    0xfc2200e800000ULL,
    0xef7102b400000ULL,
    0x981832600000ULL,
    0xfcdbc37300000ULL,
    0x4139e04480000ULL,
    0x2ee8111bc0000ULL,
    0x195aeb9560000ULL,
    0x13fa9ed790000ULL,
    0x98c2888f68000ULL,
    0x63394e2694000ULL,
    0x47f981d68e000ULL,
    0xe2cfc14eab000ULL,
    0xae3be0a2e1800ULL,
    0xbe691388cf400ULL,
    0x660368018c200ULL,
    0x1b055e644a300ULL,
    0xfa81680e61880ULL,
    0xba445c270f440ULL,
    0x3d78eab2ec260ULL,
    0x188b9ec0da330ULL,
    0xf3430b45898d8ULL,
    0x96e28eda5b4e4ULL,
    0xcd40621102262ULL,
    0x29e7d209e1309ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xd800000000000ULL,
    0xb400000000000ULL,
    0xda00000000000ULL,
    0xa100000000000ULL,
    0xcd80000000000ULL,
    0x6940000000000ULL,
    0xae60000000000ULL,
    0x7ef0000000000ULL,
    0x8c38000000000ULL,
    0x340c000000000ULL,
    0x9a1e000000000ULL,
    0x10d000000000ULL,
    0xdd8f800000000ULL,
    0xb156c00000000ULL,
    0x1a79e00000000ULL,
    0xa4e6d00000000ULL,
    0x2d38b80000000ULL,
    0xf98d540000000ULL,
    0xf3493a0000000ULL,
    0xaf66950000000ULL,
    0xa367588000000ULL,
    0x3d6a854000000ULL,
    0x2e6003e000000ULL,
    0x3ef0039000000ULL,
    0x2c38028800000ULL,
    0x240c001400000ULL,
    0x421e033e00000ULL,
    0xb50d01f100000ULL,
    0x78f82e780000ULL,
    0x1056c2dac0000ULL,
    0xd7f9e3bf60000ULL,
    0xcda6d08ef0000ULL,
    0x8358bae118000ULL,
    0x877d552fc4000ULL,
    0x7f71392ea6000ULL,
    0x9b6a97115f000ULL,
    0x397959f7a1800ULL,
    0x3c678512f4400ULL,
    0xf3ef804b3b600ULL,
    0x8fa6c020f5100ULL,
    0x3641e19821880ULL,
    0x80ead3dc34440ULL,
    0x6f26baca5b6a0ULL,
    0x4c80545f05110ULL,
    0xf4c6b91eb9858ULL,
    0xbf305569304f4ULL,
    0x749eb9bb9d67aULL,
    0xf0cc5550aa1b1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xc800000000000ULL,
    0x8c00000000000ULL,
    0x5e00000000000ULL,
    0xbf00000000000ULL,
    0xa980000000000ULL,
    0x6840000000000ULL,
    0xd7e0000000000ULL,
    0x1570000000000ULL,
    0xbea8000000000ULL,
    0x41c000000000ULL,
    0xb21a000000000ULL,
    0xf11b000000000ULL,
    0x6e9e800000000ULL,
    0x85c3400000000ULL,
    0x6dada00000000ULL,
    0x2399100000000ULL,
    0xbd57d80000000ULL,
    0x7b76c40000000ULL,
    0x9b35a0000000ULL,
    0x819e870000000ULL,
    0x3448788000000ULL,
    0x51e8d5c000000ULL,
    0x5e6003a000000ULL,
    0xd3003f000000ULL,
    0x14800d800000ULL,
    0x2d6c020c00000ULL,
    0x9ab200fe00000ULL,
    0xc607016d00000ULL,
    0x2b04808180000ULL,
    0xa3984308c0000ULL,
    0x7d53230320000ULL,
    0xdb6a51e4b0000ULL,
    0xb9b27b6d48000ULL,
    0x4983d781dc000ULL,
    0xb8568388b2000ULL,
    0xfef42c311000ULL,
    0xe17fa044a4800ULL,
    0xa4ae139d05400ULL,
    0x691b59d993e00ULL,
    0xfa828544b2500ULL,
    0x8fd2781d24880ULL,
    0x78b3d699c54c0ULL,
    0x2f1e8024b3ea0ULL,
    0x118342cd025b0ULL,
    0x8c4da371ec848ULL,
    0xb5e913d0d944cULL,
    0x3c7fd90f21efeULL,
    0x242ac51aa350fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0x400000000000ULL,
    0xba00000000000ULL,
    0xfb00000000000ULL,
    0xaf80000000000ULL,
    0x3640000000000ULL,
    0xdce0000000000ULL,
    0x4270000000000ULL,
    0xc8b8000000000ULL,
    0x298c000000000ULL,
    0xcb42000000000ULL,
    0xce7f000000000ULL,
    0xc6bb800000000ULL,
    0x609ac00000000ULL,
    0x6bd4a00000000ULL,
    0x45b7d00000000ULL,
    0xd317580000000ULL,
    0xeb95240000000ULL,
    0x2c56da0000000ULL,
    0xd7fce50000000ULL,
    0x45fbf88000000ULL,
    0xfaeef6c000000ULL,
    0x4f63822000000ULL,
    0x7226c2d000000ULL,
    0xa94ea2c800000ULL,
    0xd974d03c00000ULL,
    0x7736daf600000ULL,
    0xe7cce72500000ULL,
    0x4ba3f92580000ULL,
    0x9a12f73cc0000ULL,
    0x4b19826f60000ULL,
    0xa795c3b6f0000ULL,
    0xc257235c68000ULL,
    0xcee111460c000ULL,
    0xbd61fb5d06000ULL,
    0xdd2df5e1a7000ULL,
    0xe4c203ceb4800ULL,
    0xb83f019848400ULL,
    0xba5b82b461e00ULL,
    0xd2eac12a20d00ULL,
    0xb6ca0a334880ULL,
    0x683bd05888440ULL,
    0xa255580d01ea0ULL,
    0xdeea2769d0df0ULL,
    0x456d5812dc828ULL,
    0x8126261e44444ULL,
    0xf2cf5bc967e1aULL,
    0xfd29261b87d0bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x1c00000000000ULL,
    0x9600000000000ULL,
    0x9b00000000000ULL,
    0x9c80000000000ULL,
    0x34c0000000000ULL,
    0x3d60000000000ULL,
    0x94f0000000000ULL,
    0xaba8000000000ULL,
    0x6084000000000ULL,
    0xd2c2000000000ULL,
    0xde79000000000ULL,
    0xa466800000000ULL,
    0x117a400000000ULL,
    0xdae6200000000ULL,
    0x4ca0300000000ULL,
    0xda17680000000ULL,
    0x6510240000000ULL,
    0x639bea0000000ULL,
    0x9257670000000ULL,
    0x1cb94b8000000ULL,
    0xb20415c000000ULL,
    0x10683e000000ULL,
    0x598a41b000000ULL,
    0x874e23b800000ULL,
    0x72430dc00000ULL,
    0x8c55684a00000ULL,
    0x93a926cf00000ULL,
    0x6c9d6adf80000ULL,
    0x8cdd27ecc0000ULL,
    0xf1776ba920000ULL,
    0xaae026c2d0000ULL,
    0x34b3eaf028000ULL,
    0x7613651834000ULL,
    0xeb1b4acc32000ULL,
    0xe48d15621d000ULL,
    0x98c8017b0a800ULL,
    0xb37401018dc00ULL,
    0x13ea03addf200ULL,
    0xa13d02bfe9500ULL,
    0xcf44839c8a880ULL,
    0x733343f14dcc0ULL,
    0x7e48a3aeff260ULL,
    0x32ae7302395b0ULL,
    0xe51b480b22898ULL,
    0xa38d15d9b9cdcULL,
    0xf2480081ed2f6ULL,
    0xacb4016df452bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x2400000000000ULL,
    0xca00000000000ULL,
    0xa100000000000ULL,
    0xe280000000000ULL,
    0xe440000000000ULL,
    0x8560000000000ULL,
    0x44f0000000000ULL,
    0x3f28000000000ULL,
    0x6284000000000ULL,
    0x245e000000000ULL,
    0xe56d000000000ULL,
    0x14f5800000000ULL,
    0x673ec00000000ULL,
    0x468de00000000ULL,
    0xee4f300000000ULL,
    0x4479f80000000ULL,
    0xf665f40000000ULL,
    0x837a7a0000000ULL,
    0xc3f2370000000ULL,
    0xaabc1b8000000ULL,
    0x7b5ec64000000ULL,
    0x94f582e000000ULL,
    0xa73ec2d000000ULL,
    0x268de3c800000ULL,
    0xbe4f335400000ULL,
    0x1c79f9c200000ULL,
    0xd265f58500000ULL,
    0x497a78b780000ULL,
    0x62f23706c0000ULL,
    0x483c1ae320000ULL,
    0x9f1ec548d0000ULL,
    0x1195809858000ULL,
    0xe3cec1dc64000ULL,
    0x19a5e3f60e000ULL,
    0xdccb30d75f000ULL,
    0x3827f81a99800ULL,
    0x3708f43573c00ULL,
    0x5d8ff8ef86600ULL,
    0x5ccf6fa90100ULL,
    0xeb1f8e519880ULL,
    0x7151f627b3cc0ULL,
    0x55ec792ea6660ULL,
    0x15ab34e740150ULL,
    0x9adf9b02c18d8ULL,
    0x1f3907a917ce4ULL,
    0x929be3f9886aaULL,
    0x4c5633fdcf1f1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0xb400000000000ULL,
    0xfe00000000000ULL,
    0x8d00000000000ULL,
    0x5180000000000ULL,
    0xb5c0000000000ULL,
    0xc7a0000000000ULL,
    0xd130000000000ULL,
    0x678000000000ULL,
    0xc894000000000ULL,
    0x2a46000000000ULL,
    0xf761000000000ULL,
    0x9b0b800000000ULL,
    0x508e400000000ULL,
    0xce54e00000000ULL,
    0x1166500000000ULL,
    0xb212280000000ULL,
    0x871afc0000000ULL,
    0x4287aa0000000ULL,
    0x7951bd0000000ULL,
    0x5be6cb8000000ULL,
    0x174cad4000000ULL,
    0xeeed826000000ULL,
    0x92df439000000ULL,
    0xb527639800000ULL,
    0xa07c10a400000ULL,
    0x8180cb5e00000ULL,
    0xedddae3300000ULL,
    0x3be03a080000ULL,
    0x473502a7c0000ULL,
    0x476d83c160000ULL,
    0xd31f42f190000ULL,
    0xec87624e78000ULL,
    0xec4c116b14000ULL,
    0xae78c924d6000ULL,
    0x2489af89ed000ULL,
    0x105802ba65800ULL,
    0xec64010d24400ULL,
    0x8b9e0103b5200ULL,
    0xfec5001f0bb00ULL,
    0xcf358102e5880ULL,
    0xdb7b424ee4440ULL,
    0x811961fcd5260ULL,
    0x3b89124d9bb10ULL,
    0xb6cd48f41d8f8ULL,
    0x7332ec66304f4ULL,
    0xf96162276329eULL,
    0x3e1d1296e6b9dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0x8400000000000ULL,
    0x7e00000000000ULL,
    0x4900000000000ULL,
    0x2480000000000ULL,
    0x5540000000000ULL,
    0xd360000000000ULL,
    0x64f0000000000ULL,
    0xf228000000000ULL,
    0x184000000000ULL,
    0x3bda000000000ULL,
    0xd923000000000ULL,
    0x2e0f800000000ULL,
    0x9106400000000ULL,
    0xb89a600000000ULL,
    0xd74c700000000ULL,
    0x6866180000000ULL,
    0xf374d40000000ULL,
    0xb4fb9a0000000ULL,
    0xea25970000000ULL,
    0xfd9c7b8000000ULL,
    0xc9c8a6c000000ULL,
    0x6a3583a000000ULL,
    0x3d9540d000000ULL,
    0xa9dde05800000ULL,
    0x1a3e32f400000ULL,
    0x358e794a00000ULL,
    0x2ddfa66100000ULL,
    0x6428002780000ULL,
    0x7c8402e2c0000ULL,
    0x95a019720000ULL,
    0x31630100d0000ULL,
    0xafef804a48000ULL,
    0x6db640e15c000ULL,
    0xc3526267f6000ULL,
    0xae78700285000ULL,
    0x56741a273f800ULL,
    0x1a63d728a4c00ULL,
    0x80661ab64fe00ULL,
    0xc774d4473b900ULL,
    0xa2fb99d8bf880ULL,
    0x572595fe64cc0ULL,
    0xaf1c7bcb6fe60ULL,
    0x5188a4f6eb970ULL,
    0xe3d5826d77888ULL,
    0x452543c9f8c44ULL,
    0xac15e2d1b9e1eULL,
    0x2a0a3345be939ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x7800000000000ULL,
    0x9c00000000000ULL,
    0x6a00000000000ULL,
    0xeb00000000000ULL,
    0x2680000000000ULL,
    0x18c0000000000ULL,
    0x76a0000000000ULL,
    0x1530000000000ULL,
    0x2668000000000ULL,
    0x379c000000000ULL,
    0xcd56000000000ULL,
    0xa4e1000000000ULL,
    0xb956800000000ULL,
    0x4ae9c00000000ULL,
    0xd447a00000000ULL,
    0x4576f00000000ULL,
    0xe511a80000000ULL,
    0xfb989c0000000ULL,
    0x1f592a0000000ULL,
    0x73fc5f0000000ULL,
    0x25d6088000000ULL,
    0xc12e6c4000000ULL,
    0x1868816000000ULL,
    0x2294c23000000ULL,
    0x5ec722c800000ULL,
    0xffbe329c00000ULL,
    0xeca088d200000ULL,
    0x9637adbf00000ULL,
    0x44e722e880000ULL,
    0x94e309bc0000ULL,
    0x42e88b5920000ULL,
    0xe05bafa9b0000ULL,
    0xcb79219678000ULL,
    0xf803336934000ULL,
    0x5c000a61a2000ULL,
    0xca0f6f8a53000ULL,
    0x3b1e01fb1c800ULL,
    0x5e8d03bef8c00ULL,
    0x84c880d282200ULL,
    0x1ca4c368deb00ULL,
    0xfe2f215b9c880ULL,
    0xe231f938cc0ULL,
    0x2f568839a22a0ULL,
    0xbbe6ac4e6ebd0ULL,
    0xb1d9a06d648f8ULL,
    0x9f3bf0d7ccc5cULL,
    0x7d792ab3202caULL,
    0x190c5ce28db3bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0x4400000000000ULL,
    0xc200000000000ULL,
    0xd00000000000ULL,
    0x6b80000000000ULL,
    0x4b40000000000ULL,
    0xf0e0000000000ULL,
    0x9bb0000000000ULL,
    0xfb58000000000ULL,
    0x38fc000000000ULL,
    0x57be000000000ULL,
    0x555d000000000ULL,
    0x4beb800000000ULL,
    0xb72ec00000000ULL,
    0xba84e00000000ULL,
    0x96c7500000000ULL,
    0xfcbad80000000ULL,
    0x61db940000000ULL,
    0x3e3e3a0000000ULL,
    0xf31cc50000000ULL,
    0xa084e08000000ULL,
    0x4fc7524000000ULL,
    0xd3ad96000000ULL,
    0xb39b965000000ULL,
    0x9f5e3ae800000ULL,
    0x6aecc42400000ULL,
    0x2bce22600000ULL,
    0xaacb535900000ULL,
    0x3abcdb0d80000ULL,
    0xeca9513c0000ULL,
    0x88b3bac6e0000ULL,
    0x2bd306a930000ULL,
    0x5f35809598000ULL,
    0xe683c2ef84000ULL,
    0x60d7637cd6000ULL,
    0x1ba5929a0f000ULL,
    0xbb583b035d800ULL,
    0x98fdc72e96400ULL,
    0x87b1630d60e00ULL,
    0xad449273f7100ULL,
    0xfedbb96c5880ULL,
    0x753e05c112440ULL,
    0xb7860071b6ea0ULL,
    0xfd5100e9f81d0ULL,
    0xb7ed809598078ULL,
    0x913fc2ef84004ULL,
    0xa589637cd6062ULL,
    0x848929a0f0ddULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x9400000000000ULL,
    0xb600000000000ULL,
    0x5d00000000000ULL,
    0xdc80000000000ULL,
    0x3ec0000000000ULL,
    0x8d60000000000ULL,
    0xcbb0000000000ULL,
    0x8ad8000000000ULL,
    0x8b74000000000ULL,
    0x6eae000000000ULL,
    0xda5b000000000ULL,
    0x9fa7800000000ULL,
    0x9cca400000000ULL,
    0x2676600000000ULL,
    0xaa2fb00000000ULL,
    0x7882080000000ULL,
    0x30de140000000ULL,
    0xbc746a0000000ULL,
    0x2d31a50000000ULL,
    0xc316618000000ULL,
    0xa59fb14000000ULL,
    0xac5a0ba000000ULL,
    0x62aa161000000ULL,
    0xb05a68f800000ULL,
    0xaaa62400000ULL,
    0xbb51e02a00000ULL,
    0x9125f0cf00000ULL,
    0x51146ae780000ULL,
    0xb681a5c5c0000ULL,
    0xa1ce60e6e0000ULL,
    0xaaebb32710000ULL,
    0x7cf409bbd8000ULL,
    0x71f1148bbc000ULL,
    0x457de8a3a6000ULL,
    0xffa0e59f85000ULL,
    0xccc781a1a7800ULL,
    0xce7a41ec87c00ULL,
    0x2e2e623809a00ULL,
    0xc69bb1c47a900ULL,
    0xf9cc081a7f880ULL,
    0xd6f515673bc40ULL,
    0x4eebea9bafae0ULL,
    0x92ffe45bff910ULL,
    0x50f601bbd8088ULL,
    0xabef008bbc0d4ULL,
    0x226982a3a6056ULL,
    0x5421409f8504dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0xc800000000000ULL,
    0x1400000000000ULL,
    0xd200000000000ULL,
    0x4500000000000ULL,
    0x2880000000000ULL,
    0xebc0000000000ULL,
    0xb660000000000ULL,
    0xc9b0000000000ULL,
    0x6dc8000000000ULL,
    0x164000000000ULL,
    0xd43e000000000ULL,
    0xf683000000000ULL,
    0x80df800000000ULL,
    0x6df6c00000000ULL,
    0xc5e9e00000000ULL,
    0x81fe300000000ULL,
    0x7bf4a80000000ULL,
    0xbaff540000000ULL,
    0x887d4a0000000ULL,
    0x32b1670000000ULL,
    0x7e41e18000000ULL,
    0x192a314000000ULL,
    0xea02a8a000000ULL,
    0xc918561000000ULL,
    0xc69cca6800000ULL,
    0xf8c4a64400000ULL,
    0xc1f7826a00000ULL,
    0x5be2c3fb00000ULL,
    0xcaffe2db80000ULL,
    0xd0693299c0000ULL,
    0xeebd2b06a0000ULL,
    0xb85e966750000ULL,
    0x8e3d2b0dd8000ULL,
    0x879e96649c000ULL,
    0xa5d2aa806000ULL,
    0x9b2e97a417000ULL,
    0x871528da79800ULL,
    0x658a958373c00ULL,
    0x374b29f7ad200ULL,
    0x1fb9973799900ULL,
    0xf2dcabd7a1880ULL,
    0x98eb57e7efcc0ULL,
    0x556b495fab2e0ULL,
    0x262667938e990ULL,
    0xc388628dd8048ULL,
    0xf04bf3249c0d4ULL,
    0x5a22c80806032ULL,
    0xb587a6b4170d5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xf800000000000ULL,
    0xac00000000000ULL,
    0x1a00000000000ULL,
    0xd100000000000ULL,
    0xd80000000000ULL,
    0xfe40000000000ULL,
    0x3ea0000000000ULL,
    0xccf0000000000ULL,
    0x7848000000000ULL,
    0xcda4000000000ULL,
    0x5c7a000000000ULL,
    0x2181000000000ULL,
    0x244e800000000ULL,
    0xfbcc00000000ULL,
    0x7160600000000ULL,
    0x7e19100000000ULL,
    0x5f0cb80000000ULL,
    0x8a8b7c0000000ULL,
    0x88ccda0000000ULL,
    0xe4626f0000000ULL,
    0xad88618000000ULL,
    0x6e4d11c000000ULL,
    0x96beb86000000ULL,
    0x28ee7c7000000ULL,
    0x3658598800000ULL,
    0xaaafaef400000ULL,
    0x9aee81de00000ULL,
    0x34cc17d00000ULL,
    0xe92860ac80000ULL,
    0x3bd10bac0000ULL,
    0xfb76ba93a0000ULL,
    0x70a7df150000ULL,
    0xb6825acec8000ULL,
    0x3adeac0d94000ULL,
    0xd16800ac52000ULL,
    0xee1401b237000ULL,
    0xf712002f1b800ULL,
    0x6e9501b3c7400ULL,
    0xc6dc83a17b600ULL,
    0x8369c196ddb00ULL,
    0x6b1ce2e1d3880ULL,
    0x4c80d1be534c0ULL,
    0x5bd8590d296e0ULL,
    0x24efaf24eabb0ULL,
    0xbc4e834ec8078ULL,
    0xd3bcc1cd9406cULL,
    0x736062cc520faULL,
    0xb31912c237061ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x800000000000ULL,
    0x8400000000000ULL,
    0x3e00000000000ULL,
    0xe900000000000ULL,
    0x5080000000000ULL,
    0x7ec0000000000ULL,
    0x3060000000000ULL,
    0xfab0000000000ULL,
    0x5cd8000000000ULL,
    0xdb74000000000ULL,
    0x1132000000000ULL,
    0xd187000000000ULL,
    0x1a45800000000ULL,
    0x98b5400000000ULL,
    0x57d5e00000000ULL,
    0xc0eeb00000000ULL,
    0x947e880000000ULL,
    0x94b5340000000ULL,
    0xadcb6a0000000ULL,
    0xf7eb850000000ULL,
    0xddede18000000ULL,
    0xb2eab2c000000ULL,
    0x67748ba000000ULL,
    0x33635d000000ULL,
    0x9284ead800000ULL,
    0x65ddc4b400000ULL,
    0xd3f7803a00000ULL,
    0xf3f2409f00000ULL,
    0xa3f0622380000ULL,
    0xbbebf3c1c0000ULL,
    0xc7f3696ae0000ULL,
    0x75ef853310000ULL,
    0x26e7e28da8000ULL,
    0xa169b374cc000ULL,
    0x663b0a2e5e000ULL,
    0x780074f509000ULL,
    0xcc1e8aa4a1800ULL,
    0x5a0536f67fc00ULL,
    0x27136b4105a00ULL,
    0xb19f841ea2100ULL,
    0xaa5fe02909880ULL,
    0x70adb182b3c40ULL,
    0x23d10b6f5bae0ULL,
    0xf6f375ebab1f0ULL,
    0xf9690b0da8088ULL,
    0xfa3775b4cc0c4ULL,
    0x3a030b8e5e0deULL,
    0x9704742509019ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0xd400000000000ULL,
    0x4200000000000ULL,
    0xb100000000000ULL,
    0xfb80000000000ULL,
    0x9cc0000000000ULL,
    0xf0e0000000000ULL,
    0xd030000000000ULL,
    0x16c8000000000ULL,
    0x5f4000000000ULL,
    0x21b6000000000ULL,
    0xbf07000000000ULL,
    0x648b800000000ULL,
    0x854c00000000ULL,
    0xe0b1600000000ULL,
    0x9c86d00000000ULL,
    0x3446280000000ULL,
    0x16bfe40000000ULL,
    0x3f974a0000000ULL,
    0x56c9370000000ULL,
    0xa5f9638000000ULL,
    0x11b2d24000000ULL,
    0xc710292000000ULL,
    0x9888e61000000ULL,
    0x9e54cba800000ULL,
    0x13a9f69400000ULL,
    0xd61e016600000ULL,
    0x530302f700000ULL,
    0x7a95810d80000ULL,
    0x1f57c323c0000ULL,
    0x9024e230a0000ULL,
    0xb6d112c310000ULL,
    0x35e2ca3bf8000ULL,
    0x59aef65ccc000ULL,
    0x4315833976000ULL,
    0xf297c0ee8f000ULL,
    0xfb44e0084b800ULL,
    0xaa21104458400ULL,
    0xfbcacb2e25200ULL,
    0x586af79328d00ULL,
    0x366b8233b3880ULL,
    0xf964c218944c0ULL,
    0x45f9621753260ULL,
    0x81b2d07da7d50ULL,
    0x8f1029bbf80a8ULL,
    0x1c88e41ccc014ULL,
    0xf454ca1976022ULL,
    0x76a9f6fe8f0e1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0xf400000000000ULL,
    0xee00000000000ULL,
    0x1f00000000000ULL,
    0xe480000000000ULL,
    0x3940000000000ULL,
    0xea60000000000ULL,
    0x2f30000000000ULL,
    0xa758000000000ULL,
    0x6d74000000000ULL,
    0x6fba000000000ULL,
    0xf809000000000ULL,
    0x540d800000000ULL,
    0xbe09400000000ULL,
    0x970f200000000ULL,
    0x2886f00000000ULL,
    0x2358f80000000ULL,
    0x1b64440000000ULL,
    0xd4b7da0000000ULL,
    0x7a92b50000000ULL,
    0xbe57218000000ULL,
    0xaaf2f3c000000ULL,
    0x7062fb6000000ULL,
    0x9e2d451000000ULL,
    0xbcda595800000ULL,
    0xabf6ac00000ULL,
    0x848001aa00000ULL,
    0xc940035900000ULL,
    0x3260035b80000ULL,
    0x6b300149c0000ULL,
    0x7158026ce0000ULL,
    0x867402a510000ULL,
    0x653a030988000ULL,
    0xde4902ccf4000ULL,
    0x5aed80d55e000ULL,
    0xa8794141c1000ULL,
    0xda372238ea800ULL,
    0x6ac2f3532ec00ULL,
    0xebbaf9f2d7600ULL,
    0x8e19469a7ed00ULL,
    0xef00593162880ULL,
    0x3c92f59fdac40ULL,
    0x7d558327896e0ULL,
    0x3c7d42dbbfdb0ULL,
    0xc4352289880b8ULL,
    0xadcff10cf40b4ULL,
    0x4b357bb55e00eULL,
    0x615d0451c10afULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x9800000000000ULL,
    0x1400000000000ULL,
    0x6a00000000000ULL,
    0xc900000000000ULL,
    0x4e80000000000ULL,
    0xdd40000000000ULL,
    0x8420000000000ULL,
    0x5370000000000ULL,
    0xe58000000000ULL,
    0xdbbc000000000ULL,
    0x74ba000000000ULL,
    0xcd35000000000ULL,
    0x5dfd800000000ULL,
    0x731dc00000000ULL,
    0xaf85600000000ULL,
    0xbfcb700000000ULL,
    0xd770b80000000ULL,
    0xec5ecc0000000ULL,
    0xceadda0000000ULL,
    0x2c29bd0000000ULL,
    0x3f67608000000ULL,
    0x204272c000000ULL,
    0x10b738e000000ULL,
    0x7f360f9000000ULL,
    0x20f539c800000ULL,
    0xa78f0e7400000ULL,
    0x23cab81200000ULL,
    0x616bce9100000ULL,
    0x335059ca80000ULL,
    0xaf347ca8c0000ULL,
    0x8e20161e0000ULL,
    0x8b89039730000ULL,
    0xadc7819988000ULL,
    0x5a68c39b24000ULL,
    0xa0d8e0479e000ULL,
    0x56e6b3b463000ULL,
    0x988dd97203800ULL,
    0x1259bd416ac00ULL,
    0x2dbf6122df600ULL,
    0xbbe722cb2f00ULL,
    0x5cad386b8b880ULL,
    0xe1330c1a4ec40ULL,
    0xd3f0b985416a0ULL,
    0x481ece08d1ff0ULL,
    0x3c0dd8d188018ULL,
    0x4619bf2f24054ULL,
    0x471f60b59e0caULL,
    0x758e72b563039ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0x8c00000000000ULL,
    0x6600000000000ULL,
    0x7100000000000ULL,
    0x7c80000000000ULL,
    0x6a40000000000ULL,
    0x7e20000000000ULL,
    0xa270000000000ULL,
    0x8548000000000ULL,
    0xa7a4000000000ULL,
    0xb6be000000000ULL,
    0x9a3b000000000ULL,
    0x1864800000000ULL,
    0x7a5b400000000ULL,
    0xb635e00000000ULL,
    0xce6f300000000ULL,
    0x7344980000000ULL,
    0xdeb4ec0000000ULL,
    0x46397a0000000ULL,
    0x967fdf0000000ULL,
    0x1743e18000000ULL,
    0xa4b0314000000ULL,
    0x593e1ba000000ULL,
    0x17e4ac7000000ULL,
    0x67801bb800000ULL,
    0xc1dfacf400000ULL,
    0xf9e499ea00000ULL,
    0x5a84ed2d00000ULL,
    0x3b5179aa80000ULL,
    0x72abdc50c0000ULL,
    0x5035e3d5a0000ULL,
    0x7f6f32a350000ULL,
    0xefc49b05d8000ULL,
    0x24f4eceb7c000ULL,
    0x30197909ea000ULL,
    0xb80fdc5541000ULL,
    0xf40be0bee5800ULL,
    0x721433ea81c00ULL,
    0x930018709de00ULL,
    0xe79fafe5d3700ULL,
    0x1c49a3b3d880ULL,
    0x19f4ee41fdcc0ULL,
    0xca997ad977ee0ULL,
    0x334fdfc092790ULL,
    0xfeabe13dd8088ULL,
    0x3624315f7c04cULL,
    0xe681b43ea086ULL,
    0x934bad08410e1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0xd400000000000ULL,
    0x4a00000000000ULL,
    0x9b00000000000ULL,
    0x3b80000000000ULL,
    0x5640000000000ULL,
    0x4ee0000000000ULL,
    0xdab0000000000ULL,
    0x6b48000000000ULL,
    0xc86c000000000ULL,
    0xca6a000000000ULL,
    0x9573000000000ULL,
    0x9cf3800000000ULL,
    0x5da1c00000000ULL,
    0x1edd200000000ULL,
    0x9739900000000ULL,
    0x5801780000000ULL,
    0xbc00b40000000ULL,
    0x3e145a0000000ULL,
    0xa115270000000ULL,
    0x89f208000000ULL,
    0xb9d692c000000ULL,
    0x52b0f82000000ULL,
    0xf4e769000000ULL,
    0x8a78f98800000ULL,
    0xf562763c00000ULL,
    0x4cf2f94600000ULL,
    0x85a175f500000ULL,
    0x62c97bf580000ULL,
    0x92cb7a3c0000ULL,
    0x891e5b9ea0000ULL,
    0x1c9627b770000ULL,
    0x53c4a16288000ULL,
    0xb9ab53d17c000ULL,
    0x9ccfd987e6000ULL,
    0x828e5ec8b000ULL,
    0xf18003f81a800ULL,
    0xd40000411400ULL,
    0xd56001e279600ULL,
    0xfcf0004758900ULL,
    0x8da8021a92880ULL,
    0xc6dc01156d4c0ULL,
    0xeb2200459f6a0ULL,
    0xc61f033bd3970ULL,
    0x6d19806a88028ULL,
    0x9e92c32d7c014ULL,
    0xcccea2e1e60eaULL,
    0x102851898b0ebULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x7800000000000ULL,
    0x7c00000000000ULL,
    0xda00000000000ULL,
    0x4500000000000ULL,
    0xfb80000000000ULL,
    0x94c0000000000ULL,
    0xeae0000000000ULL,
    0x4c30000000000ULL,
    0xf7c8000000000ULL,
    0x4e6c000000000ULL,
    0xc66a000000000ULL,
    0x269000000000ULL,
    0x47c800000000ULL,
    0xeb66c00000000ULL,
    0x2deee00000000ULL,
    0x3ea5500000000ULL,
    0x9a84280000000ULL,
    0x795ea40000000ULL,
    0x3922ca0000000ULL,
    0x1457f70000000ULL,
    0x76ace08000000ULL,
    0xfe9052c000000ULL,
    0xf5aab6000000ULL,
    0xae3d643000000ULL,
    0xaedaa88800000ULL,
    0xbffd67e400000ULL,
    0x1fbaa80600000ULL,
    0x170d658700000ULL,
    0x7a92a92580000ULL,
    0x69516453c0000ULL,
    0x9130aa7ca0000ULL,
    0x6054662d70000ULL,
    0xa8a6292088000ULL,
    0x1d9ba54714000ULL,
    0x6bd44945da000ULL,
    0x846836e3ff000ULL,
    0x2b768134e9800ULL,
    0x8dffc3297e400ULL,
    0x4eba6256a3200ULL,
    0xe29f92683bd00ULL,
    0x548c89461880ULL,
    0xe33ef4ae6a4c0ULL,
    0x51506073792a0ULL,
    0x8d3690bbc4d70ULL,
    0x6a544928880f8ULL,
    0xe5a83463140bcULL,
    0xe2168023da07aULL,
    0x590fc054ff035ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xc800000000000ULL,
    0x8c00000000000ULL,
    0x8200000000000ULL,
    0x8300000000000ULL,
    0xc980000000000ULL,
    0xb040000000000ULL,
    0x3ee0000000000ULL,
    0xc8b0000000000ULL,
    0xbb48000000000ULL,
    0xdb74000000000ULL,
    0x8afe000000000ULL,
    0xdeb3000000000ULL,
    0x9e4c800000000ULL,
    0x8ff7c00000000ULL,
    0x3a27200000000ULL,
    0x968bd00000000ULL,
    0x93c3f80000000ULL,
    0xf1ae740000000ULL,
    0x65ccda0000000ULL,
    0x4a1a70000000ULL,
    0x9959238000000ULL,
    0x2878d34000000ULL,
    0x9b6f7aa000000ULL,
    0x2ae9b67000000ULL,
    0xaea3fb2800000ULL,
    0x465e773c00000ULL,
    0xcbe4daae00000ULL,
    0x3425a78d00000ULL,
    0x978f237d80000ULL,
    0xd94fd093c0000ULL,
    0x8875fabea0000ULL,
    0xeb69743310000ULL,
    0xf2fe58c8d8000ULL,
    0xeaa565ea1c000ULL,
    0x485282074a000ULL,
    0xcaf4c1ead3000ULL,
    0x7ea3a1010e800ULL,
    0xee48109be0400ULL,
    0x57fadbb2cb600ULL,
    0x7e26a6954a900ULL,
    0x988ba249d6880ULL,
    0x92cc1231fc4c0ULL,
    0xbb2cd91581660ULL,
    0x1c11a60f99910ULL,
    0x8a1123e0d8048ULL,
    0x6f0cd2d61c04cULL,
    0x5b9178a94a0e2ULL,
    0xfb5ab667d3093ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x9800000000000ULL,
    0x8400000000000ULL,
    0x1600000000000ULL,
    0x9900000000000ULL,
    0x1680000000000ULL,
    0x9e40000000000ULL,
    0x3e60000000000ULL,
    0x9a30000000000ULL,
    0x1748000000000ULL,
    0xe4000000000ULL,
    0xf862000000000ULL,
    0xcb35000000000ULL,
    0x8dd7800000000ULL,
    0xe4b8400000000ULL,
    0x5500e00000000ULL,
    0x4c84700000000ULL,
    0x9d4c580000000ULL,
    0xcbfb640000000ULL,
    0x81e4ba0000000ULL,
    0xaaeb170000000ULL,
    0xa362e18000000ULL,
    0x5ab172c000000ULL,
    0xf01bd9e000000ULL,
    0x580327d000000ULL,
    0x6404583800000ULL,
    0x661f677400000ULL,
    0x106b98200000ULL,
    0x929e16dd00000ULL,
    0x8855634380000ULL,
    0xa77930c1c0000ULL,
    0x8cb3391ca0000ULL,
    0x891357df50000ULL,
    0x3e8203dea8000ULL,
    0x6245022244000ULL,
    0xdc7f82ad56000ULL,
    0x8d2c430bf5000ULL,
    0x1ccae1fdfd800ULL,
    0x9e2570a2a0c00ULL,
    0xc151d83c1ae00ULL,
    0x79e2253e78500ULL,
    0x9ef9dba355880ULL,
    0xcd762440e4cc0ULL,
    0x37b3d9714cee0ULL,
    0x689726e58d570ULL,
    0xdb4e5be6a8018ULL,
    0x5afe655644044ULL,
    0xfb7b3b2f560f6ULL,
    0x3eb755d6f50e9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0xc00000000000ULL,
    0x3a00000000000ULL,
    0x5d00000000000ULL,
    0xf380000000000ULL,
    0x840000000000ULL,
    0x57e0000000000ULL,
    0x81b0000000000ULL,
    0x7548000000000ULL,
    0x5474000000000ULL,
    0xe1f6000000000ULL,
    0x26a9000000000ULL,
    0xbbce800000000ULL,
    0xffb8c00000000ULL,
    0xe651e00000000ULL,
    0xfced500000000ULL,
    0x752eb80000000ULL,
    0x8384940000000ULL,
    0xa0575a0000000ULL,
    0x33edc70000000ULL,
    0xbfa7e38000000ULL,
    0x4644524000000ULL,
    0xcce038e000000ULL,
    0x7d3c557000000ULL,
    0xd786bb8800000ULL,
    0x9640964400000ULL,
    0x54e958a600000ULL,
    0x1130c71500000ULL,
    0xbd9f607180000ULL,
    0x93559057c0000ULL,
    0xab7f5882a0000ULL,
    0x2369c75310000ULL,
    0xb779e214f8000ULL,
    0xe16953ae74000ULL,
    0xd670ba9162000ULL,
    0x20e9957787000ULL,
    0xd727d952dd800ULL,
    0xb288044b46400ULL,
    0x39ce822885600ULL,
    0x3eb8c10413100ULL,
    0x17d1e0c625880ULL,
    0xf5ad51a5324c0ULL,
    0xb34eba59e7660ULL,
    0x5b74960394150ULL,
    0x4b7f599cf80d8ULL,
    0xb369c5ea740ccULL,
    0x8f79e2376205aULL,
    0xbd6952628700dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0xd400000000000ULL,
    0xd600000000000ULL,
    0xb300000000000ULL,
    0xc180000000000ULL,
    0xe7c0000000000ULL,
    0xdba0000000000ULL,
    0xf70000000000ULL,
    0x5558000000000ULL,
    0xdde4000000000ULL,
    0xa302000000000ULL,
    0xd983000000000ULL,
    0xfbc3800000000ULL,
    0x29be400000000ULL,
    0xc269a00000000ULL,
    0xf3ccd00000000ULL,
    0x2dba080000000ULL,
    0x2c73ec0000000ULL,
    0xccd1aa0000000ULL,
    0x463c3d0000000ULL,
    0x7aa8238000000ULL,
    0xb3f192c000000ULL,
    0xdc102b2000000ULL,
    0xd2017c9000000ULL,
    0x5d0203f800000ULL,
    0xfe83009c00000ULL,
    0x8c4381de00000ULL,
    0x8d7e429300000ULL,
    0x7049a2c880000ULL,
    0xcf7cd0f2c0000ULL,
    0x75420963e0000ULL,
    0x4de7ed8650000ULL,
    0xfb0baadf68000ULL,
    0xa59b3e8e9c000ULL,
    0xf9c9a18d86000ULL,
    0x4cbcd13d35000ULL,
    0xb0e20a93c8800ULL,
    0xd597ec8e5a400ULL,
    0x11d3aacb46a00ULL,
    0xf8bf3c84f0f00ULL,
    0x96eba3b4a0880ULL,
    0xce8fd29cc6440ULL,
    0x4598a98c0a60ULL,
    0xc97daf2ac5ff0ULL,
    0xfe400befe8028ULL,
    0x64eee05c094ULL,
    0xb6c82930660b6ULL,
    0xcf257e2865043ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0xb400000000000ULL,
    0x6600000000000ULL,
    0x9300000000000ULL,
    0x5780000000000ULL,
    0x4f40000000000ULL,
    0x2d20000000000ULL,
    0x9970000000000ULL,
    0x77d8000000000ULL,
    0xd7e4000000000ULL,
    0xc612000000000ULL,
    0xc313000000000ULL,
    0xef8f800000000ULL,
    0x2354400000000ULL,
    0x572ba00000000ULL,
    0xd87e100000000ULL,
    0xd557d80000000ULL,
    0x5c3a2c0000000ULL,
    0xf3ee7a0000000ULL,
    0x38173d0000000ULL,
    0x2c16238000000ULL,
    0x1a0950c000000ULL,
    0x310bfb2000000ULL,
    0xa847f1000000ULL,
    0x3fd8013800000ULL,
    0x53e4019c00000ULL,
    0x6812014600000ULL,
    0x941302b100000ULL,
    0x760f834a80000ULL,
    0x4b144188c0000ULL,
    0x4b8ba0f3e0000ULL,
    0x9d4e137650000ULL,
    0xd82fdb9928000ULL,
    0x5dee2d6edc000ULL,
    0x6f047a12f6000ULL,
    0xb5903ce4dd000ULL,
    0x7253a0bdf0800ULL,
    0x2daa107339400ULL,
    0x4fbdda31b9e00ULL,
    0x32bd2e0d5cf00ULL,
    0x522bfa1cd8880ULL,
    0xf4f47c81e5440ULL,
    0x198001654fe60ULL,
    0x2840005881f70ULL,
    0x7ca001eba8028ULL,
    0x3530017a1c0f4ULL,
    0xa57801a716006ULL,
    0xb5d400238d0e3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0x3400000000000ULL,
    0x1e00000000000ULL,
    0xf300000000000ULL,
    0x8680000000000ULL,
    0xc6c0000000000ULL,
    0xf5a0000000000ULL,
    0x6870000000000ULL,
    0x758000000000ULL,
    0xebec000000000ULL,
    0x881a000000000ULL,
    0x6c01000000000ULL,
    0x7a05800000000ULL,
    0xa50b400000000ULL,
    0x799aa00000000ULL,
    0x6a48300000000ULL,
    0xde6bc80000000ULL,
    0xe8480c0000000ULL,
    0x2f6b6a0000000ULL,
    0xdfc13d0000000ULL,
    0xfe25228000000ULL,
    0x8b3270c000000ULL,
    0xfaace86000000ULL,
    0xbce77d1000000ULL,
    0x3880029800000ULL,
    0x5c0021c00000ULL,
    0x6b2002e200000ULL,
    0xeab0011700000ULL,
    0xd4f8012c80000ULL,
    0x449c026ac0000ULL,
    0x17c203b3e0000ULL,
    0xb22d01c270000ULL,
    0x813f802708000ULL,
    0x67ba4104b4000ULL,
    0xf167201efe000ULL,
    0x4cdf7285e9000ULL,
    0x28b368131c800ULL,
    0x5ed3c92bdc00ULL,
    0x31f22bfada00ULL,
    0xfe8372b804f00ULL,
    0xd2d16aac14880ULL,
    0x9bb03f8a09c40ULL,
    0xa378a04353a20ULL,
    0xb5d5312aedf70ULL,
    0x332c4b93880b8ULL,
    0x8eae4d7274074ULL,
    0x82ee4b4f1e03eULL,
    0xbb834f5099083ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0x8400000000000ULL,
    0x7200000000000ULL,
    0x7300000000000ULL,
    0xdd80000000000ULL,
    0x8640000000000ULL,
    0x4e60000000000ULL,
    0xff30000000000ULL,
    0x3ed8000000000ULL,
    0x4a4000000000ULL,
    0x6196000000000ULL,
    0x51000000000ULL,
    0xe76f800000000ULL,
    0x1da4c00000000ULL,
    0xb0b200000000ULL,
    0x4981500000000ULL,
    0xc43380000000ULL,
    0xe96e640000000ULL,
    0xc8be1a0000000ULL,
    0x4f8e350000000ULL,
    0xe54aa38000000ULL,
    0x6be0904000000ULL,
    0xad699be000000ULL,
    0x9abef75000000ULL,
    0xc8f81a800000ULL,
    0xf0d4c2dc00000ULL,
    0xf1b3233e00000ULL,
    0x1515529500000ULL,
    0x648d382f80000ULL,
    0x9cdb67f5c0000ULL,
    0xafa79a0ae0000ULL,
    0x580bf72250000ULL,
    0xa416018b38000ULL,
    0x4211024ad4000ULL,
    0xbb0f80c266000ULL,
    0xc194c0db17000ULL,
    0x705320e2f2800ULL,
    0x4f65529e06400ULL,
    0x51b53b2578600ULL,
    0x650f6557f8d00ULL,
    0xcc8998c1ca880ULL,
    0xd0cef608d2440ULL,
    0xc1b782d91e660ULL,
    0xdd00c219efd50ULL,
    0x789d238cb8018ULL,
    0x6ad05323140c4ULL,
    0xaeacb81686012ULL,
    0xf68aa73c47023ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x1800000000000ULL,
    0x9c00000000000ULL,
    0x3200000000000ULL,
    0x3d00000000000ULL,
    0x5c80000000000ULL,
    0xb340000000000ULL,
    0x8320000000000ULL,
    0x5970000000000ULL,
    0x90c8000000000ULL,
    0x19f4000000000ULL,
    0xd18e000000000ULL,
    0xc7c9000000000ULL,
    0x447c800000000ULL,
    0xfc57c00000000ULL,
    0x42bae00000000ULL,
    0x46a3b00000000ULL,
    0x28a1b80000000ULL,
    0x47a6ac0000000ULL,
    0xb6355a0000000ULL,
    0x41fc1f0000000ULL,
    0xed80638000000ULL,
    0x5c9714000000ULL,
    0xf169d9e000000ULL,
    0x24dbdc3000000ULL,
    0x5ff282a800000ULL,
    0xca9ec18400000ULL,
    0x1046603a00000ULL,
    0x38b4707900000ULL,
    0x9fbb5b1e80000ULL,
    0x4a351dd8c0000ULL,
    0xe3fce287a0000ULL,
    0xc89eb2f610000ULL,
    0xc55339f718000ULL,
    0x70386ff5f4000ULL,
    0x9af33a3b4a000ULL,
    0x5a086cfbbb000ULL,
    0xe91b391039800ULL,
    0x8a8c6f3801c00ULL,
    0xb05d387c55600ULL,
    0xc8b16ee630700ULL,
    0x17afba4f21880ULL,
    0xce2fad49f5cc0ULL,
    0x4de9d87d1f660ULL,
    0xc79bdd648b790ULL,
    0xa4d283c198098ULL,
    0x9feec2e93405cULL,
    0xaa8e6166ea052ULL,
    0x80407244ab0adULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x3800000000000ULL,
    0x9400000000000ULL,
    0xd600000000000ULL,
    0x9900000000000ULL,
    0x380000000000ULL,
    0x4340000000000ULL,
    0xc5a0000000000ULL,
    0x59f0000000000ULL,
    0x2ec8000000000ULL,
    0x4be4000000000ULL,
    0xa9ca000000000ULL,
    0x9d79000000000ULL,
    0x9b19800000000ULL,
    0x6c96400000000ULL,
    0x9d2e00000000ULL,
    0x8d67d00000000ULL,
    0x7305d80000000ULL,
    0xf0897c0000000ULL,
    0x73d53a0000000ULL,
    0x5673af0000000ULL,
    0x3f83618000000ULL,
    0x295593c000000ULL,
    0xf6bd3b6000000ULL,
    0x8967ae9000000ULL,
    0x8d0161b800000ULL,
    0x1588933400000ULL,
    0x3a4eb8d600000ULL,
    0xf638efa900000ULL,
    0x22a201f880000ULL,
    0x7f6d013ac0000ULL,
    0xc41b8375a0000ULL,
    0x1e0b42ec50000ULL,
    0xd50162f208000ULL,
    0x7188900774000ULL,
    0x344eb885ce000ULL,
    0xcb38ecc47d000ULL,
    0xcf2200fe0d800ULL,
    0x312d00a556400ULL,
    0xd43b83daf4e00ULL,
    0x9dbb42259eb00ULL,
    0x3de9623405880ULL,
    0x20dc9056224c0ULL,
    0x76ecb8e93aee0ULL,
    0x4455efd8e3b30ULL,
    0xd339828a880b8ULL,
    0x8b2642fdb4054ULL,
    0xef3ae0906e036ULL,
    0xe133d0b82d0a9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x3800000000000ULL,
    0xa400000000000ULL,
    0xb600000000000ULL,
    0x2900000000000ULL,
    0x4880000000000ULL,
    0x7a40000000000ULL,
    0x9ba0000000000ULL,
    0x16f0000000000ULL,
    0xd8000000000ULL,
    0xeef4000000000ULL,
    0x4ca000000000ULL,
    0x88e9000000000ULL,
    0x25df800000000ULL,
    0x5c6c400000000ULL,
    0x4d98e00000000ULL,
    0x58d6b00000000ULL,
    0x3aeb980000000ULL,
    0x6adb9c0000000ULL,
    0x35e17a0000000ULL,
    0xe3502d0000000ULL,
    0xab3f618000000ULL,
    0x28bef1c000000ULL,
    0x1d617b6000000ULL,
    0xe9102e5000000ULL,
    0xe89f603800000ULL,
    0xaa4ef33400000ULL,
    0x93b9787a00000ULL,
    0x8ae42d8d00000ULL,
    0x12d561ac80000ULL,
    0x71e7f222c0000ULL,
    0x6546fb35e0000ULL,
    0xba386f8410000ULL,
    0xc43583d248000ULL,
    0xd13543e174000ULL,
    0x5bbf63b2d6000ULL,
    0xb6fef36dd5000ULL,
    0xd0c178607d800ULL,
    0xe6e02cd03bc00ULL,
    0x98c7637865e00ULL,
    0x9afaf31450b00ULL,
    0xbad37a0a35880ULL,
    0x3dfd2dc54fc40ULL,
    0x7f52e3d0b3ee0ULL,
    0xb93fb3a485b30ULL,
    0xb7b41bfec80b8ULL,
    0x7cf7dc03b40e4ULL,
    0xdbd999e736056ULL,
    0x9769fb9c5019ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x3800000000000ULL,
    0x1c00000000000ULL,
    0xd200000000000ULL,
    0x5100000000000ULL,
    0xc980000000000ULL,
    0x65c0000000000ULL,
    0x2360000000000ULL,
    0x21b0000000000ULL,
    0xb458000000000ULL,
    0x8abc000000000ULL,
    0x48d6000000000ULL,
    0x48fd000000000ULL,
    0x1d60800000000ULL,
    0xeab3c00000000ULL,
    0x38db600000000ULL,
    0x40fe900000000ULL,
    0xa97ec80000000ULL,
    0x1cac740000000ULL,
    0xa7cbaa0000000ULL,
    0xa63e50000000ULL,
    0x543be08000000ULL,
    0x938d534000000ULL,
    0xc0c5aa6000000ULL,
    0xbce2e63000000ULL,
    0x4b6d61a800000ULL,
    0xe5b3903400000ULL,
    0x4a464aee00000ULL,
    0xa1a3b69300000ULL,
    0xf446c94b80000ULL,
    0x2aa07461c0000ULL,
    0xd8c5a994e0000ULL,
    0x70e2e4a830000ULL,
    0x16d60b438000ULL,
    0x38b391ae04000ULL,
    0x69c648f35a000ULL,
    0x8963b57bfb000ULL,
    0xcca6c8c98f800ULL,
    0x3fd07520c5400ULL,
    0x867dab064ea00ULL,
    0xbe2ee7477f900ULL,
    0xde836155b7880ULL,
    0xdb4293fac1440ULL,
    0x8828c97b14aa0ULL,
    0xa191779f84990ULL,
    0xa1cb29ffb80b8ULL,
    0xdd6025cfc405cULL,
    0xab88167ba072ULL,
    0x8cfc1d3cb0c1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0xd800000000000ULL,
    0xdc00000000000ULL,
    0x3a00000000000ULL,
    0x4d00000000000ULL,
    0x6080000000000ULL,
    0xf0c0000000000ULL,
    0x76a0000000000ULL,
    0xe570000000000ULL,
    0xc058000000000ULL,
    0xfe74000000000ULL,
    0x3fde000000000ULL,
    0x5721000000000ULL,
    0x772c800000000ULL,
    0x4736c00000000ULL,
    0xef31e00000000ULL,
    0x7b28300000000ULL,
    0x4538480000000ULL,
    0xee2fec0000000ULL,
    0xf9afaa0000000ULL,
    0x24e2dd0000000ULL,
    0xc21d618000000ULL,
    0xa11ef34000000ULL,
    0xf289a8e000000ULL,
    0x29c7dd3000000ULL,
    0x2837e18800000ULL,
    0xbebd334400000ULL,
    0xa16acaea00000ULL,
    0xc6482cb700000ULL,
    0x196ac93880000ULL,
    0xca482e26c0000ULL,
    0x1b6acbb3e0000ULL,
    0xcb482e1810000ULL,
    0x99eac89c18000ULL,
    0xaa882d5654000ULL,
    0xb5cacaf14a000ULL,
    0xf2382ed19f000ULL,
    0x63b2caeb15800ULL,
    0x19fc2dceabc00ULL,
    0xea94c807dea00ULL,
    0x55d92e5a5b700ULL,
    0x623e4a7f0d880ULL,
    0xbbbaec9cffc40ULL,
    0xc5fd2afc94ae0ULL,
    0xd0851c0cc4790ULL,
    0x18d801a498058ULL,
    0x2b403709409cULL,
    0x4b7e0142aa0daULL,
    0xb35100c98f0ddULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0xcc00000000000ULL,
    0xd600000000000ULL,
    0x3f00000000000ULL,
    0xc980000000000ULL,
    0xe540000000000ULL,
    0x65e0000000000ULL,
    0xb930000000000ULL,
    0x1198000000000ULL,
    0x915c000000000ULL,
    0x17f2000000000ULL,
    0xac2f000000000ULL,
    0xf918800000000ULL,
    0x4e9f400000000ULL,
    0x88cf200000000ULL,
    0xaaa1d00000000ULL,
    0xfdd5180000000ULL,
    0x923b540000000ULL,
    0x7a02ba0000000ULL,
    0x3905c50000000ULL,
    0xae98808000000ULL,
    0x18df41c000000ULL,
    0x52af21a000000ULL,
    0xf9d1d3f000000ULL,
    0x882d187800000ULL,
    0x931757cc00000ULL,
    0xcf88ba7e00000ULL,
    0x8246c66900000ULL,
    0x986a023180000ULL,
    0x8e7300b1c0000ULL,
    0x516a8071e0000ULL,
    0x8f04351b0000ULL,
    0x15b7a061a8000ULL,
    0x744e9339cc000ULL,
    0xd7623bc5aa000ULL,
    0x2ff684e3c1000ULL,
    0x483da176de800ULL,
    0x730d93e13a400ULL,
    0x5f90ba8588e00ULL,
    0x7a5ac483da100ULL,
    0x9c780326f6880ULL,
    0x946c006936440ULL,
    0xb86a0131c2ea0ULL,
    0xfe730331ab130ULL,
    0x396a823180048ULL,
    0xf4f040b1c008cULL,
    0xbb7a071e0076ULL,
    0x874e9351b000fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x3800000000000ULL,
    0xf400000000000ULL,
    0x2600000000000ULL,
    0x3300000000000ULL,
    0xe480000000000ULL,
    0x95c0000000000ULL,
    0x42e0000000000ULL,
    0xbbb0000000000ULL,
    0xe098000000000ULL,
    0x2bd4000000000ULL,
    0x55fa000000000ULL,
    0x7131000000000ULL,
    0x8a41800000000ULL,
    0x5fae400000000ULL,
    0x6e88a00000000ULL,
    0x4d3f00000000ULL,
    0x6b6ba80000000ULL,
    0x3df6a40000000ULL,
    0xcd228a0000000ULL,
    0x804b150000000ULL,
    0x8ea1818000000ULL,
    0xa71e41c000000ULL,
    0xb290a36000000ULL,
    0x7ec7f31000000ULL,
    0x6271a87800000ULL,
    0x3077a71c00000ULL,
    0x657b090a00000ULL,
    0x52f1579b00000ULL,
    0x13b3231980000ULL,
    0xfc8cb179c0000ULL,
    0xf1c289e9e0000ULL,
    0x8cfb155190000ULL,
    0x74b980ad88000ULL,
    0xee0a437794000ULL,
    0x5f0aa3f4ee000ULL,
    0xe686f2e17d000ULL,
    0x88c829eaa4800ULL,
    0x597de60436c00ULL,
    0x18f1a8066c200ULL,
    0x22b7a76905500ULL,
    0x51b0a5eac880ULL,
    0x7f81540a62c40ULL,
    0x6d4b221b622e0ULL,
    0xd28b0d9e8530ULL,
    0x20408b19800b8ULL,
    0x5eae1579c00b4ULL,
    0xaf1a03e9e00c6ULL,
    0x7e81005190003ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0xcc00000000000ULL,
    0xd200000000000ULL,
    0x9900000000000ULL,
    0x7780000000000ULL,
    0x35c0000000000ULL,
    0xe6e0000000000ULL,
    0x50b0000000000ULL,
    0x5d88000000000ULL,
    0x40c4000000000ULL,
    0xb37e000000000ULL,
    0x5465000000000ULL,
    0xb8e6800000000ULL,
    0x3bb7c00000000ULL,
    0x431f600000000ULL,
    0xaa8cd00000000ULL,
    0x1458c80000000ULL,
    0xfc2ca40000000ULL,
    0xfec12a0000000ULL,
    0xc867b70000000ULL,
    0xf2ee818000000ULL,
    0x9eb3c34000000ULL,
    0x4e81616000000ULL,
    0x7259d37000000ULL,
    0x9b3649a800000ULL,
    0xd25f655400000ULL,
    0xab20489600000ULL,
    0x9a4e65b300000ULL,
    0xcf30cbc580000ULL,
    0x3c58a6d5c0000ULL,
    0x4837298da0000ULL,
    0xc6b431d0000ULL,
    0x93760393d8000ULL,
    0xa461024684000ULL,
    0x10f883d862000ULL,
    0x4fa2c07c4b000ULL,
    0x5d11e00267800ULL,
    0xe18f12852c400ULL,
    0xfad1a926bea00ULL,
    0xbe7177a851d00ULL,
    0x2de960543f880ULL,
    0x7e2dd216684c0ULL,
    0xffc049737cae0ULL,
    0x83fe64e5cad10ULL,
    0xbd38ca4580038ULL,
    0x955ca595c000cULL,
    0x77a928eda0032ULL,
    0x5113b741d0089ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xd800000000000ULL,
    0x5c00000000000ULL,
    0x9e00000000000ULL,
    0x3100000000000ULL,
    0xf780000000000ULL,
    0xb40000000000ULL,
    0xf1e0000000000ULL,
    0x2930000000000ULL,
    0x6198000000000ULL,
    0xfe44000000000ULL,
    0xa466000000000ULL,
    0xdd61000000000ULL,
    0x8ee8800000000ULL,
    0x77ac400000000ULL,
    0xa14be00000000ULL,
    0xcaedd00000000ULL,
    0x95b9180000000ULL,
    0x3e4e140000000ULL,
    0x847a7a0000000ULL,
    0xed7f850000000ULL,
    0x6f0838000000ULL,
    0xf3a8404000000ULL,
    0x634de2e000000ULL,
    0x65fcd03000000ULL,
    0x5329991800000ULL,
    0xc29655b400000ULL,
    0x7ecf9ba600000ULL,
    0x35b755e900000ULL,
    0x4e471b9980000ULL,
    0x6c6b14c9c0000ULL,
    0x3974faa1e0000ULL,
    0x1cf2c66dd0000ULL,
    0xb3e18ff8000ULL,
    0x3bd9d17894000ULL,
    0xa8271bec3a000ULL,
    0x5e1b167235000ULL,
    0x110cfbbf0d800ULL,
    0xc786c6c8c6400ULL,
    0x834de0b42be00ULL,
    0x75fcd0265f100ULL,
    0xeb2999a975880ULL,
    0xce96577992440ULL,
    0x38cf9bf9f1e60ULL,
    0x58b75539ba150ULL,
    0x27c7181980058ULL,
    0x562b1489c001cULL,
    0x3f14f841e00feULL,
    0x3e82c65dd0061ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x5800000000000ULL,
    0xbc00000000000ULL,
    0x6e00000000000ULL,
    0x8500000000000ULL,
    0xf780000000000ULL,
    0xf9c0000000000ULL,
    0xb6a0000000000ULL,
    0x9ef0000000000ULL,
    0x4098000000000ULL,
    0x354c000000000ULL,
    0xf5f2000000000ULL,
    0x720d000000000ULL,
    0x1b14800000000ULL,
    0xa9e400000000ULL,
    0x24b600000000ULL,
    0x7965500000000ULL,
    0x1146480000000ULL,
    0x47ec2c0000000ULL,
    0x901aa0000000ULL,
    0x819b3d0000000ULL,
    0xe0de828000000ULL,
    0x9f2f43c000000ULL,
    0xba35e1e000000ULL,
    0xbdba103000000ULL,
    0x3c6ba8d800000ULL,
    0x86da3ce400000ULL,
    0x2e38006600000ULL,
    0x57bc00f300000ULL,
    0xfb6a018480000ULL,
    0x7241035cc0000ULL,
    0x416682b8e0000ULL,
    0x3d53405ef0000ULL,
    0xc1ffe36da8000ULL,
    0x680b130904000ULL,
    0xa41528659a000ULL,
    0xf2057e054d000ULL,
    0x5b15e0bfff800ULL,
    0x2a8a12b42c400ULL,
    0xb253a9ce54a00ULL,
    0x21663eef15f00ULL,
    0xad5200d6d7880ULL,
    0x29fd0121e8440ULL,
    0x8c0c82f32ea20ULL,
    0x76124084a8fb0ULL,
    0x191963dc800d8ULL,
    0x29985078c00fcULL,
    0x24cac93ee004eULL,
    0xfd3e6d9df0035ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xd800000000000ULL,
    0xcc00000000000ULL,
    0xba00000000000ULL,
    0x1100000000000ULL,
    0xb180000000000ULL,
    0xbec0000000000ULL,
    0xd160000000000ULL,
    0x9830000000000ULL,
    0xa498000000000ULL,
    0x1144000000000ULL,
    0xb0b2000000000ULL,
    0x6fd5000000000ULL,
    0x40ea800000000ULL,
    0xb6f2c00000000ULL,
    0x5df3600000000ULL,
    0x9d7b100000000ULL,
    0x6232680000000ULL,
    0xd58bb40000000ULL,
    0x50d38a0000000ULL,
    0xd676650000000ULL,
    0x72b8838000000ULL,
    0x62d7c2c000000ULL,
    0x361e2e000000ULL,
    0xfd3dd0f000000ULL,
    0x930b098800000ULL,
    0xdc91a77400000ULL,
    0xd41622e00000ULL,
    0x42ae13bf00000ULL,
    0x9ad8e80380000ULL,
    0x5f79775bc0000ULL,
    0x6f20e8b7e0000ULL,
    0x960d7415d0000ULL,
    0x1b0ae938a8000ULL,
    0x189c777e44000ULL,
    0x335269e762000ULL,
    0xdbbb46f9f000ULL,
    0x944b8939e7800ULL,
    0x3732677ae7400ULL,
    0x1a0a82a364200ULL,
    0xc102c229d2900ULL,
    0xf98b6382cf880ULL,
    0x5acf119f63440ULL,
    0x7f786b13e6260ULL,
    0xff2ab7a39d9f0ULL,
    0xbe13098b80058ULL,
    0xf15a42fc008cULL,
    0x6e936099e00daULL,
    0x984b12aad00e1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x800000000000ULL,
    0xc00000000000ULL,
    0x9e00000000000ULL,
    0x7700000000000ULL,
    0xf980000000000ULL,
    0x440000000000ULL,
    0x7c60000000000ULL,
    0x83b0000000000ULL,
    0xea88000000000ULL,
    0xbfdc000000000ULL,
    0x8922000000000ULL,
    0x4557000000000ULL,
    0xef8800000000ULL,
    0xc866400000000ULL,
    0xd9b3200000000ULL,
    0x2f9cd00000000ULL,
    0x9f44380000000ULL,
    0xbffc40000000ULL,
    0xc8e79a0000000ULL,
    0x6b69570000000ULL,
    0xaa3a828000000ULL,
    0xf6c143c000000ULL,
    0xf7a3a0a000000ULL,
    0x909690f000000ULL,
    0x4add1a5800000ULL,
    0xa1a815ec00000ULL,
    0xcb99239e00000ULL,
    0x1d57d2cd00000ULL,
    0xdafeba9780000ULL,
    0x427e86dfc0000ULL,
    0x4a4387ba0000ULL,
    0x3b0fc7c1b0000ULL,
    0x878f9976a8000ULL,
    0x634555f62c000ULL,
    0xbdf083596a000ULL,
    0x83fa41d5c5000ULL,
    0x4f12074ae800ULL,
    0xd57bd0bd2a400ULL,
    0xed34bb0ffc600ULL,
    0x7458793fc500ULL,
    0xfff6b91586880ULL,
    0x72e28454c64c0ULL,
    0xbe663a8d36620ULL,
    0xb2a8c67789530ULL,
    0x701f18cf80088ULL,
    0xe80f1733c00ccULL,
    0x1c09a1e5a00beULL,
    0xa61d920cb0047ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0xac00000000000ULL,
    0x5a00000000000ULL,
    0x1100000000000ULL,
    0x5380000000000ULL,
    0xf340000000000ULL,
    0x63e0000000000ULL,
    0xed30000000000ULL,
    0xca98000000000ULL,
    0x4cc000000000ULL,
    0x2ea6000000000ULL,
    0x1dd000000000ULL,
    0xeb32800000000ULL,
    0x898ac00000000ULL,
    0xa243600000000ULL,
    0x1063500000000ULL,
    0xe73980000000ULL,
    0xd16e540000000ULL,
    0x45fa7a0000000ULL,
    0xbe3bc50000000ULL,
    0x1414828000000ULL,
    0x9617c14000000ULL,
    0x7b11e16000000ULL,
    0x2a9992f000000ULL,
    0x74c8f88800000ULL,
    0x66b106f400000ULL,
    0xc5d7e22600000ULL,
    0x6534928900000ULL,
    0x6e82784180000ULL,
    0xbac7c439c0000ULL,
    0xa1aa8145e0000ULL,
    0xcd46c017f0000ULL,
    0xace56278f8000ULL,
    0x1be501044000ULL,
    0x9d411bb852000ULL,
    0xf4e4975c5f000ULL,
    0xbdb91bc247800ULL,
    0xbf58949766400ULL,
    0x49e71a54a3e00ULL,
    0xb439959a04100ULL,
    0x5d0b9b7b3f880ULL,
    0x799255fee2440ULL,
    0xaa447bc911e20ULL,
    0xf46ac521ab110ULL,
    0x906000c9800f8ULL,
    0x4e7002cdc00ecULL,
    0xf1780363e007aULL,
    0x55fc029ef0001ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xd800000000000ULL,
    0xc400000000000ULL,
    0xb200000000000ULL,
    0x1d00000000000ULL,
    0xf180000000000ULL,
    0xe440000000000ULL,
    0xfce0000000000ULL,
    0x3430000000000ULL,
    0xbd18000000000ULL,
    0x818c000000000ULL,
    0x1c42000000000ULL,
    0x8fd000000000ULL,
    0x5e33800000000ULL,
    0x6416c00000000ULL,
    0xc219e00000000ULL,
    0xe51f900000000ULL,
    0x59fb80000000ULL,
    0x8e40940000000ULL,
    0x25f7da0000000ULL,
    0x77b4c50000000ULL,
    0x4442028000000ULL,
    0x8cfd01c000000ULL,
    0xcc3383e000000ULL,
    0x4916c3d000000ULL,
    0xeb99e1a800000ULL,
    0xc55f936400000ULL,
    0x4b7fb90e00000ULL,
    0xa77097e500000ULL,
    0x696fda3580000ULL,
    0x1278c641c0000ULL,
    0xa4e00387e0000ULL,
    0xb0300316f0000ULL,
    0x2f18016568000ULL,
    0xac8c01f5c4000ULL,
    0x35c200a1c6000ULL,
    0x28bd02d7af000ULL,
    0x10d3837ea9800ULL,
    0x4d26c1e125400ULL,
    0x8e81e32b99600ULL,
    0x80d392ecbc100ULL,
    0xe53db98641880ULL,
    0xb28d953121440ULL,
    0xc6dc5903bf620ULL,
    0x922e04c8e3130ULL,
    0x9a19e1a800058ULL,
    0x611f936400084ULL,
    0x979fb90e00092ULL,
    0xa34097e50002dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x2800000000000ULL,
    0x2c00000000000ULL,
    0x9600000000000ULL,
    0x700000000000ULL,
    0x1480000000000ULL,
    0xd4c0000000000ULL,
    0x4c60000000000ULL,
    0x2330000000000ULL,
    0xc318000000000ULL,
    0x8e84000000000ULL,
    0x15ca000000000ULL,
    0x47e3000000000ULL,
    0xe770800000000ULL,
    0xe1af400000000ULL,
    0x70c6200000000ULL,
    0xe660700000000ULL,
    0xca27080000000ULL,
    0xe48d940000000ULL,
    0xdcdbaa0000000ULL,
    0x3061a50000000ULL,
    0xad2a038000000ULL,
    0xc01301c000000ULL,
    0x2008812000000ULL,
    0x501b407000000ULL,
    0x181422f800000ULL,
    0x40770ac00000ULL,
    0xba1d8a0a00000ULL,
    0x9101d74f00000ULL,
    0x138d0b6380000ULL,
    0xc05e96dfc0000ULL,
    0x98b329bde0000ULL,
    0x6f4ae776d0000ULL,
    0xe026207b48000ULL,
    0x4d9072559c000ULL,
    0x9b5f08b2ea000ULL,
    0x523996f531000ULL,
    0xa089aaf4da800ULL,
    0x6c6a6e03cc00ULL,
    0x9170825004a00ULL,
    0x96af420877900ULL,
    0x2c46231412880ULL,
    0x2ea072c660c40ULL,
    0x384709550ea60ULL,
    0xecbd94c496930ULL,
    0x9d43aaf8000a8ULL,
    0x6d25a4ac0006cULL,
    0xe000000a000f6ULL,
    0x7000024f00037ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x3800000000000ULL,
    0xe400000000000ULL,
    0x4e00000000000ULL,
    0x7b00000000000ULL,
    0x8080000000000ULL,
    0x46c0000000000ULL,
    0x3f60000000000ULL,
    0xda30000000000ULL,
    0x1708000000000ULL,
    0xd29c000000000ULL,
    0xb7c2000000000ULL,
    0xbae1000000000ULL,
    0x5f6b800000000ULL,
    0xa3d400000000ULL,
    0x3f0aa00000000ULL,
    0x3e85700000000ULL,
    0x25ddc80000000ULL,
    0x6bf2b40000000ULL,
    0xeafeea0000000ULL,
    0xb76b870000000ULL,
    0xc622028000000ULL,
    0x9d1103c000000ULL,
    0xd783806000000ULL,
    0x7451435000000ULL,
    0x98a0a1d800000ULL,
    0x68c8727c00000ULL,
    0x947c4b3a00000ULL,
    0x72b2f4f700000ULL,
    0xbdddc8ff80000ULL,
    0x7ff2b573c0000ULL,
    0xbcfeea21a0000ULL,
    0x186b87e2b0000ULL,
    0x30a200e328000ULL,
    0x44d101f1f4000ULL,
    0x266382faa6000ULL,
    0x93a1407f71000ULL,
    0x3048a25b84800ULL,
    0x26a473b5e0c00ULL,
    0xbd64bacdda00ULL,
    0xc0fff6ca1f100ULL,
    0x427c4b9f2c880ULL,
    0x1db2f54bd4cc0ULL,
    0x6b5dc84ddba20ULL,
    0x9632b5a0de130ULL,
    0x751ee9d8000b8ULL,
    0x1b9b867c00024ULL,
    0xd64a013a0006eULL,
    0x71bd03f70004bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x5c00000000000ULL,
    0xe200000000000ULL,
    0xe500000000000ULL,
    0x5d80000000000ULL,
    0x4ac0000000000ULL,
    0x10a0000000000ULL,
    0xcaf0000000000ULL,
    0xe908000000000ULL,
    0x1794000000000ULL,
    0x4bda000000000ULL,
    0x432d000000000ULL,
    0x5f20800000000ULL,
    0x1d3fc00000000ULL,
    0x2824e00000000ULL,
    0x1da5100000000ULL,
    0xd36be80000000ULL,
    0x7dc9cc0000000ULL,
    0xb0358a0000000ULL,
    0xe1be1f0000000ULL,
    0xe17a038000000ULL,
    0xf0dd01c000000ULL,
    0x69a882e000000ULL,
    0x156bc3b000000ULL,
    0xf6dee0e800000ULL,
    0x82b812e400000ULL,
    0x97e36afe00000ULL,
    0x58920ebf00000ULL,
    0x67436b3480000ULL,
    0xe2620d00c0000ULL,
    0x364b6866a0000ULL,
    0x19f60eb590000ULL,
    0x47916a6b48000ULL,
    0xe3db0de2e4000ULL,
    0xa731eb3bea000ULL,
    0x5124ce5c4b000ULL,
    0xc2350bc211800ULL,
    0xccb1de0d6b400ULL,
    0xe8f6e1b1bc200ULL,
    0x6c1c10a573700ULL,
    0xfa116a75d9880ULL,
    0xd91b0c0b4f4c0ULL,
    0xf91e812f6260ULL,
    0x77d4cef3a87b0ULL,
    0x113d08e800058ULL,
    0x6225dee40009cULL,
    0x1cace0fe00082ULL,
    0x80f111bf00055ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0xfc00000000000ULL,
    0xca00000000000ULL,
    0x4d00000000000ULL,
    0xe180000000000ULL,
    0x2a40000000000ULL,
    0x2a20000000000ULL,
    0xf0f0000000000ULL,
    0x2b18000000000ULL,
    0x6e84000000000ULL,
    0x1ad6000000000ULL,
    0xfbf3000000000ULL,
    0xf586800000000ULL,
    0xec4ec00000000ULL,
    0x553ce00000000ULL,
    0x387f900000000ULL,
    0x9ec8780000000ULL,
    0x15fb8c0000000ULL,
    0xa6841a0000000ULL,
    0xa6c9dd0000000ULL,
    0x51ee038000000ULL,
    0x6887024000000ULL,
    0x45c880a000000ULL,
    0x8379c0b000000ULL,
    0xd84c620800000ULL,
    0x233252c400000ULL,
    0xdf6a18ae00000ULL,
    0xe24edd9d00000ULL,
    0x9626812e80000ULL,
    0x5afec26ac0000ULL,
    0xb604e184e0000ULL,
    0xc70b9139d0000ULL,
    0xcc8679e738000ULL,
    0x1bcc8da5dc000ULL,
    0x48749a9542000ULL,
    0x26c41c6af7000ULL,
    0x11ec6284e9800ULL,
    0x88251b9c5400ULL,
    0x95d21b2752600ULL,
    0xcb7adfc5d0700ULL,
    0x2448804551880ULL,
    0xe939c0b2d9440ULL,
    0x926c6298f0660ULL,
    0x3c2520bf77d0ULL,
    0xbc721988000c8ULL,
    0x70cadc84000bcULL,
    0x46f0820e000aaULL,
    0xec0dc02d0009dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x1800000000000ULL,
    0xec00000000000ULL,
    0x5a00000000000ULL,
    0xc300000000000ULL,
    0x7280000000000ULL,
    0x67c0000000000ULL,
    0x3320000000000ULL,
    0x5970000000000ULL,
    0xcd08000000000ULL,
    0x1784000000000ULL,
    0x445a000000000ULL,
    0x2165000000000ULL,
    0xd10e800000000ULL,
    0xb58ec00000000ULL,
    0x5b4f200000000ULL,
    0x11f0300000000ULL,
    0x99d1a80000000ULL,
    0xae2f1c0000000ULL,
    0xa6ea0a0000000ULL,
    0x9d44ef0000000ULL,
    0xb0f2018000000ULL,
    0x145103c000000ULL,
    0xc97c83e000000ULL,
    0xc51fc37000000ULL,
    0x3393a10800000ULL,
    0xda5ff0bc00000ULL,
    0x4c6a08a200000ULL,
    0xd684ee9300000ULL,
    0x39d2030680000ULL,
    0xbe2103fac0000ULL,
    0x6ef48078a0000ULL,
    0x595bc1fbd0000ULL,
    0x1ee9a24548000ULL,
    0x614af36b84000ULL,
    0x22ec8b3d52000ULL,
    0x134e2d5faf000ULL,
    0x15e721e306800ULL,
    0xd7c4300efbc00ULL,
    0xeb23a846fa200ULL,
    0x557e1f5abcb00ULL,
    0xa71689a8ce880ULL,
    0xcc9b2c23bfcc0ULL,
    0xdac1a1a1082e0ULL,
    0x1cbef26dc3b30ULL,
    0x213e888800098ULL,
    0x9e6f2f7c0002cULL,
    0xf193a142000baULL,
    0x355ff2e3000f3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0x4c00000000000ULL,
    0x8e00000000000ULL,
    0x5d00000000000ULL,
    0x1580000000000ULL,
    0xd640000000000ULL,
    0x2e20000000000ULL,
    0x84f0000000000ULL,
    0x5b08000000000ULL,
    0xc8c000000000ULL,
    0xc9d2000000000ULL,
    0xf6d000000000ULL,
    0x7852800000000ULL,
    0x8320400000000ULL,
    0x697e600000000ULL,
    0xf147100000000ULL,
    0xf4bf780000000ULL,
    0x5c348c0000000ULL,
    0xcfe19a0000000ULL,
    0xff8edf0000000ULL,
    0x715a008000000ULL,
    0x34a103c000000ULL,
    0xfc2083e000000ULL,
    0xfffd41f000000ULL,
    0xa784e35800000ULL,
    0x3d5b504400000ULL,
    0xbabb1a6e00000ULL,
    0xa1229e2d00000ULL,
    0xea7662ee80000ULL,
    0x71cb124ac0000ULL,
    0x136d7b54a0000ULL,
    0x3e598ee1f0000ULL,
    0xfa33192b68000ULL,
    0xe6ee9e3f84000ULL,
    0xb804607e02000ULL,
    0x1c16108535000ULL,
    0x4617fbf2b3800ULL,
    0x7905cf609d400ULL,
    0x8f977b9784600ULL,
    0x49488fe20e700ULL,
    0xe8bb9b6f5b880ULL,
    0x1a2fdf51d94c0ULL,
    0xb6fa80d3266a0ULL,
    0x701c40abcb730ULL,
    0x380463d8000d8ULL,
    0xdc1613840008cULL,
    0xe617f98e0002eULL,
    0x4905cfdd0006dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0x8400000000000ULL,
    0xe600000000000ULL,
    0x1300000000000ULL,
    0x7880000000000ULL,
    0xafc0000000000ULL,
    0x1ee0000000000ULL,
    0xccb0000000000ULL,
    0xde08000000000ULL,
    0x2704000000000ULL,
    0x768e000000000ULL,
    0x30c1000000000ULL,
    0x9c63800000000ULL,
    0x127c400000000ULL,
    0x4d76a00000000ULL,
    0x2ff9700000000ULL,
    0x613ce80000000ULL,
    0x6a43240000000ULL,
    0xe6afca0000000ULL,
    0x1703170000000ULL,
    0x5e88018000000ULL,
    0x5cc4014000000ULL,
    0xf66e01e000000ULL,
    0xfb71005000000ULL,
    0x44eb81f800000ULL,
    0xdb8425c00000ULL,
    0xbb98a27a00000ULL,
    0x6f4872f300000ULL,
    0x4537699980000ULL,
    0x3c4b658dc0000ULL,
    0x1dbf6903a0000ULL,
    0xe38f64aed0000ULL,
    0xb516b2218000ULL,
    0x333e651f14000ULL,
    0xb75ae84bb6000ULL,
    0xe13627daf5000ULL,
    0xaa4a4b3c54800ULL,
    0x6ba57ea14c00ULL,
    0x8713212b29200ULL,
    0xc68030b5e9500ULL,
    0xd8c7c9ffcc880ULL,
    0x10771464c0cc0ULL,
    0xe86e03f93f2e0ULL,
    0x3c710162cc590ULL,
    0xa26b818000018ULL,
    0xa578414000044ULL,
    0xa3f8a1e000006ULL,
    0x9b38705000083ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x2800000000000ULL,
    0x8c00000000000ULL,
    0x2a00000000000ULL,
    0x9b00000000000ULL,
    0xe780000000000ULL,
    0xfb40000000000ULL,
    0x8620000000000ULL,
    0x43f0000000000ULL,
    0xae08000000000ULL,
    0x4d0c000000000ULL,
    0xde82000000000ULL,
    0xbbd3000000000ULL,
    0x98f4800000000ULL,
    0x2984c00000000ULL,
    0xa64e200000000ULL,
    0xc0bb900000000ULL,
    0x5c25980000000ULL,
    0x90e2ac0000000ULL,
    0xd5953a0000000ULL,
    0x442ff0000000ULL,
    0x67a8008000000ULL,
    0xb9bc01c000000ULL,
    0x7caa022000000ULL,
    0x9e2f00f000000ULL,
    0x27fe800800000ULL,
    0xa81bc05c00000ULL,
    0x4c18a33600000ULL,
    0x8a1c518100000ULL,
    0x2b17398680000ULL,
    0xcf91fee2c0000ULL,
    0x775c82a0a0000ULL,
    0xac38c33bf0000ULL,
    0xd8e4234228000ULL,
    0x4994911b7c000ULL,
    0xb65b19b9da000ULL,
    0x58b96c795f000ULL,
    0xf82d9a52ea800ULL,
    0x36eeaec8fd400ULL,
    0x649738f7aee00ULL,
    0x78d1ff6c64f00ULL,
    0x7b7c801e42880ULL,
    0xc4c8c0ad414c0ULL,
    0xb96c22f8d4ea0ULL,
    0x73d8915fcbfb0ULL,
    0xc4f91880000a8ULL,
    0x3b9a6dc00004cULL,
    0x295118200008aULL,
    0xa9266ff00002bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0x3c00000000000ULL,
    0xda00000000000ULL,
    0x5300000000000ULL,
    0x2e80000000000ULL,
    0x1140000000000ULL,
    0x94a0000000000ULL,
    0x7e70000000000ULL,
    0xac08000000000ULL,
    0x204000000000ULL,
    0x771a000000000ULL,
    0x309f000000000ULL,
    0x6c5e800000000ULL,
    0xf3cc00000000ULL,
    0xc8a1200000000ULL,
    0x5467b00000000ULL,
    0xd713f80000000ULL,
    0x2092bc0000000ULL,
    0x745e5a0000000ULL,
    0xcb22cf0000000ULL,
    0xe6a0018000000ULL,
    0xe17003c000000ULL,
    0x708800a000000ULL,
    0x4c44003000000ULL,
    0xdf3a004800000ULL,
    0x30af01bc00000ULL,
    0xa07683d200000ULL,
    0x3108c10900000ULL,
    0xa99322e880000ULL,
    0x9ccb1bcc0000ULL,
    0xf4ff78e2a0000ULL,
    0x63457df1d0000ULL,
    0xbb3f96c48000ULL,
    0xa2e2bfca6c000ULL,
    0xe25659ed12000ULL,
    0xaa26cf3e8f000ULL,
    0x773a0033fe800ULL,
    0xfcaf025d6dc00ULL,
    0x527683b6cb600ULL,
    0x6e08c0afd4300ULL,
    0x9513227f36880ULL,
    0x778cb257c1cc0ULL,
    0x94df7bcb796e0ULL,
    0x5f757c598b330ULL,
    0x1d9bf80000048ULL,
    0xcfd6bc00000fcULL,
    0xade45a000003aULL,
    0xe6cdcf0000063ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x2800000000000ULL,
    0x5400000000000ULL,
    0xba00000000000ULL,
    0x8500000000000ULL,
    0x4380000000000ULL,
    0x29c0000000000ULL,
    0x70e0000000000ULL,
    0x81b0000000000ULL,
    0x4c08000000000ULL,
    0xa60c000000000ULL,
    0xbb1a000000000ULL,
    0x248b000000000ULL,
    0x934f800000000ULL,
    0xf4bdc00000000ULL,
    0xe797e00000000ULL,
    0x7bc0d00000000ULL,
    0x1f9480000000ULL,
    0x835240000000ULL,
    0xc8d32a0000000ULL,
    0xc17f370000000ULL,
    0xa960008000000ULL,
    0x9d7002c000000ULL,
    0xf768022000000ULL,
    0x2a7c009000000ULL,
    0x15f2024800000ULL,
    0xd237034400000ULL,
    0x9ddd83e600000ULL,
    0xdafac2b300000ULL,
    0xfca260cd80000ULL,
    0x38612a1c0000ULL,
    0x49c928f3a0000ULL,
    0xa0f4349ef0000ULL,
    0xd9af81dc48000ULL,
    0x300dc10a1c000ULL,
    0x481fe22156000ULL,
    0x840cd280db000ULL,
    0xe203497361800ULL,
    0xf90e276df4400ULL,
    0xad94ab71baa00ULL,
    0x16cef4dbd7d00ULL,
    0xb66de2aaa9880ULL,
    0xebfbd142284c0ULL,
    0x153eca654caa0ULL,
    0x5744e7e6fcd70ULL,
    0x76bec800000a8ULL,
    0xce84e40000094ULL,
    0x8e5eca000001aULL,
    0x6b34e700000f5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x1800000000000ULL,
    0xec00000000000ULL,
    0xf200000000000ULL,
    0xa100000000000ULL,
    0x8e80000000000ULL,
    0xcbc0000000000ULL,
    0xdd20000000000ULL,
    0xe790000000000ULL,
    0x2a08000000000ULL,
    0xed04000000000ULL,
    0xec8e000000000ULL,
    0x2c7000000000ULL,
    0xa7a1800000000ULL,
    0x325ec00000000ULL,
    0xa427200000000ULL,
    0x251e100000000ULL,
    0x83cee80000000ULL,
    0xf92fbc0000000ULL,
    0x7192520000000ULL,
    0xfd0fb90000000ULL,
    0xc488008000000ULL,
    0x16c4004000000ULL,
    0xc9ae00e000000ULL,
    0x7957007000000ULL,
    0x67a9801800000ULL,
    0x925ac0ec00000ULL,
    0x342920f200000ULL,
    0x4d1910a100000ULL,
    0x77cf688e80000ULL,
    0xe7217ccbc0000ULL,
    0x229d72dd20000ULL,
    0xd285a9e790000ULL,
    0x81c0e8aa08000ULL,
    0x28bcad04000ULL,
    0xf313d20c8e000ULL,
    0xb4c17972c7000ULL,
    0xa0a7203fa1800ULL,
    0x93de109e5ec00ULL,
    0xda6ee8b627200ULL,
    0xe87fbcf41e100ULL,
    0xe23a52154ee80ULL,
    0x715bb9deefbc0ULL,
    0xa3ae00deb2520ULL,
    0x745700fb9fb90ULL,
    0xfb29808000008ULL,
    0x889ac04000004ULL,
    0x7f8920e00000eULL,
    0x8d49107000007ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x1800000000000ULL,
    0xbc00000000000ULL,
    0xce00000000000ULL,
    0xab00000000000ULL,
    0xbe80000000000ULL,
    0x9840000000000ULL,
    0x1720000000000ULL,
    0x1750000000000ULL,
    0xe208000000000ULL,
    0xbd0c000000000ULL,
    0x8982000000000ULL,
    0xf8c1000000000ULL,
    0x3c61800000000ULL,
    0x2fbc00000000ULL,
    0xa314e00000000ULL,
    0xe32eb00000000ULL,
    0x9d55e80000000ULL,
    0xb02840000000ULL,
    0x6e8df20000000ULL,
    0xa049350000000ULL,
    0xbb28008000000ULL,
    0xc15c00c000000ULL,
    0xf50a002000000ULL,
    0xcd8d001000000ULL,
    0xbac3801800000ULL,
    0x516ac0bc00000ULL,
    0xb37d60ce00000ULL,
    0xf7d970ab00000ULL,
    0x94308be80000ULL,
    0x88ad349840000ULL,
    0xd8999a1720000ULL,
    0xbee0711750000ULL,
    0x94b9126208000ULL,
    0x3f37857d0c000ULL,
    0x5af5e82982000ULL,
    0xff128428c1000ULL,
    0x1d25f20461800ULL,
    0x3e5535aefbc00ULL,
    0x118200f514e00ULL,
    0x84c100342eb00ULL,
    0xd26180cdd5e80ULL,
    0xb9fbc02842840ULL,
    0x594e0df2df20ULL,
    0xc76eb09359350ULL,
    0x4475e88000008ULL,
    0xb75284c00000cULL,
    0x3205f22000002ULL,
    0x8505351000001ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0x8400000000000ULL,
    0xd600000000000ULL,
    0x8300000000000ULL,
    0xdc80000000000ULL,
    0x6f40000000000ULL,
    0x5620000000000ULL,
    0x80b0000000000ULL,
    0x5a08000000000ULL,
    0xc90c000000000ULL,
    0x9d86000000000ULL,
    0xaecf000000000ULL,
    0x5aeb800000000ULL,
    0x3758400000000ULL,
    0xb6d5600000000ULL,
    0x1a1c300000000ULL,
    0x3dffc80000000ULL,
    0x992bf40000000ULL,
    0x763be20000000ULL,
    0x844acb0000000ULL,
    0xe6a8008000000ULL,
    0x95fc00c000000ULL,
    0x952e006000000ULL,
    0xfc3300f000000ULL,
    0xa54580b800000ULL,
    0xd72b408400000ULL,
    0x2130e0d600000ULL,
    0xa6c7708300000ULL,
    0xc6e728dc80000ULL,
    0xf550846f40000ULL,
    0xabd2ca5620000ULL,
    0x79994f80b0000ULL,
    0xdc374ada08000ULL,
    0xf5420f090c000ULL,
    0xff2faa7d86000ULL,
    0xed397f9ecf000ULL,
    0x4cc68202eb800ULL,
    0x17eafb8358400ULL,
    0x7cd9c838d5600ULL,
    0x9b14f42d1c300ULL,
    0x9c7862ef7fc80ULL,
    0x65ee8b016bf40ULL,
    0x79d360129be20ULL,
    0xec9330acbacb0ULL,
    0xe3b4488000008ULL,
    0xb683b4c00000cULL,
    0x7e46826000006ULL,
    0xbfaafbf00000fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0xd400000000000ULL,
    0xca00000000000ULL,
    0xa700000000000ULL,
    0xd980000000000ULL,
    0xc8c0000000000ULL,
    0x9ce0000000000ULL,
    0xec50000000000ULL,
    0x7608000000000ULL,
    0xf10c000000000ULL,
    0x3886000000000ULL,
    0x184b000000000ULL,
    0xd8a3800000000ULL,
    0x12fd400000000ULL,
    0xddf4a00000000ULL,
    0x987e700000000ULL,
    0x16bf980000000ULL,
    0x83558c0000000ULL,
    0xe3834e0000000ULL,
    0xf7c1050000000ULL,
    0x4168008000000ULL,
    0x669c00c000000ULL,
    0xa1ee006000000ULL,
    0xdad700b000000ULL,
    0xeb4d803800000ULL,
    0xb2a40d400000ULL,
    0x1d3920ca00000ULL,
    0x289430a700000ULL,
    0x84e6b8d980000ULL,
    0x2851bcc8c0000ULL,
    0x540df69ce0000ULL,
    0xa0cb9ec50000ULL,
    0xc70bf6f608000ULL,
    0x6987b9310c000ULL,
    0xf0c876d886000ULL,
    0x48eaf9684b000ULL,
    0x2654d600a3800ULL,
    0xd10889b6fd400ULL,
    0x28854ecff4a00ULL,
    0xf04a059b7e700ULL,
    0x84ab80dd3f980ULL,
    0x34f14048958c0ULL,
    0x64f2a034e34e0ULL,
    0x2cf5701051050ULL,
    0xa0fc188000008ULL,
    0xef8ccc00000cULL,
    0x5bffee6000006ULL,
    0xf17375b00000bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x7800000000000ULL,
    0xec00000000000ULL,
    0x5200000000000ULL,
    0x1300000000000ULL,
    0x4380000000000ULL,
    0xd0c0000000000ULL,
    0x9de0000000000ULL,
    0x1f90000000000ULL,
    0x5b08000000000ULL,
    0x9784000000000ULL,
    0xceca000000000ULL,
    0x6ceb000000000ULL,
    0x3717800000000ULL,
    0x2446c00000000ULL,
    0x88a1200000000ULL,
    0x5fbb300000000ULL,
    0x6bf7380000000ULL,
    0x2b568c0000000ULL,
    0xe0e11e0000000ULL,
    0x3511d90000000ULL,
    0x5f4a008000000ULL,
    0xef2b004000000ULL,
    0x497780a000000ULL,
    0x5b16c0b000000ULL,
    0x3649207800000ULL,
    0x3baf30ec00000ULL,
    0xac35385200000ULL,
    0xc3398c1300000ULL,
    0x5abc9e4380000ULL,
    0xad7c19d0c0000ULL,
    0x7d1ca01de0000ULL,
    0x8b46f05f90000ULL,
    0xf12998fb08000ULL,
    0xb87f7c2784000ULL,
    0x739506b6ca000ULL,
    0x49036580eb000ULL,
    0x2489a6e517800ULL,
    0x3d45957746c00ULL,
    0xc4203eeb21200ULL,
    0x46fae97f7b300ULL,
    0x69d538ae17380ULL,
    0xc0a98c28c68c0ULL,
    0x8bb49e31e91e0ULL,
    0x75f819ed95d90ULL,
    0xda56a07800008ULL,
    0xc86df0ec00004ULL,
    0x4a5e18520000aULL,
    0x4069bc130000bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xa800000000000ULL,
    0xc400000000000ULL,
    0x2e00000000000ULL,
    0xdb00000000000ULL,
    0x3f80000000000ULL,
    0xf040000000000ULL,
    0x8520000000000ULL,
    0x1810000000000ULL,
    0xdd08000000000ULL,
    0x6084000000000ULL,
    0x1ce000000000ULL,
    0x1e6b000000000ULL,
    0xaba800000000ULL,
    0xf154400000000ULL,
    0x16aee00000000ULL,
    0xa257b00000000ULL,
    0xfd2ef80000000ULL,
    0x841a840000000ULL,
    0x7f08920000000ULL,
    0xe180210000000ULL,
    0x6346008000000ULL,
    0xceaf004000000ULL,
    0x6e5480e000000ULL,
    0x472f40b000000ULL,
    0x691c60a800000ULL,
    0xf787f0c400000ULL,
    0xc44e182e00000ULL,
    0xe32634db00000ULL,
    0xb71cea3f80000ULL,
    0x648ee5f040000ULL,
    0x8fc0720520000ULL,
    0x9568915810000ULL,
    0x2d3478bd08000ULL,
    0x6d1ec49084000ULL,
    0x798e7249ce000ULL,
    0x4f43916a6b000ULL,
    0xc4aef80cba800ULL,
    0x2b5a84ae54400ULL,
    0xba892e72ee00ULL,
    0x62d0213917b00ULL,
    0xacee006f8ef80ULL,
    0x827b00e84a840ULL,
    0x19b280e920920ULL,
    0xfad040f214210ULL,
    0x80e0e02800008ULL,
    0x887cb08400004ULL,
    0x5cb478ce0000eULL,
    0xb65ec46b0000bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0xbc00000000000ULL,
    0xea00000000000ULL,
    0xd900000000000ULL,
    0x1680000000000ULL,
    0xb6c0000000000ULL,
    0xcce0000000000ULL,
    0x2dd0000000000ULL,
    0x3f08000000000ULL,
    0x7d8c000000000ULL,
    0x954a000000000ULL,
    0x4eaf000000000ULL,
    0xaf3800000000ULL,
    0x9eb3c00000000ULL,
    0xea9aa00000000ULL,
    0xd4a3900000000ULL,
    0x2bf8680000000ULL,
    0x9435ec0000000ULL,
    0x465f8e0000000ULL,
    0xf94a3d0000000ULL,
    0xaca2008000000ULL,
    0xf7f300c000000ULL,
    0x2e3180a000000ULL,
    0x5750c0f000000ULL,
    0x6bc3203800000ULL,
    0x4c6f50bc00000ULL,
    0x81948ea00000ULL,
    0xcc69bcd900000ULL,
    0xc817461680000ULL,
    0x6c6341b6c0000ULL,
    0x381e664ce0000ULL,
    0x546311edd0000ULL,
    0x8414ae1f08000ULL,
    0xbe696d4d8c000ULL,
    0x5d11488d4a000ULL,
    0xa8e5bcc2af000ULL,
    0xebdd4678f3800ULL,
    0x640c410bb3c00ULL,
    0xc60de6ae1aa00ULL,
    0x5b00d1c763900ULL,
    0xbb860e3b98680ULL,
    0xce46fde625ec0ULL,
    0xf52320bb378e0ULL,
    0xc4bf5093d63d0ULL,
    0x6b9148b800008ULL,
    0x2e25bc7c0000cULL,
    0xbf3d464a0000aULL,
    0x5dc41290000fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0xe400000000000ULL,
    0xb600000000000ULL,
    0x5700000000000ULL,
    0x2880000000000ULL,
    0x8c0000000000ULL,
    0x9ea0000000000ULL,
    0x3ef0000000000ULL,
    0x888000000000ULL,
    0x18c4000000000ULL,
    0x16ae000000000ULL,
    0xb2fb000000000ULL,
    0x8285800000000ULL,
    0x1dca400000000ULL,
    0xdf25600000000ULL,
    0xc5be700000000ULL,
    0x3c67080000000ULL,
    0xb55acc0000000ULL,
    0x77f48a0000000ULL,
    0xeb019f0000000ULL,
    0x3a85808000000ULL,
    0x9ca404000000ULL,
    0x512560e000000ULL,
    0xc6be70b000000ULL,
    0x7ae708d800000ULL,
    0xe9acce400000ULL,
    0x77d48ab600000ULL,
    0x8a319f5700000ULL,
    0x842d80a880000ULL,
    0x273e4048c0000ULL,
    0xd1a3607ea0000ULL,
    0x5271708ef0000ULL,
    0xe64488d088000ULL,
    0xb96f8cfcc4000ULL,
    0x3cda6aa0ae000ULL,
    0xe0beafe5fb000ULL,
    0xe5ea682a05800ULL,
    0x4a10bc550a400ULL,
    0x451582a185600ULL,
    0xc994534b4e700ULL,
    0x97528a6cef080ULL,
    0xeefe9f099ecc0ULL,
    0x208e00b75a8a0ULL,
    0xc4cb00fefa9f0ULL,
    0xf4ad80a880008ULL,
    0x8bfe4048c0004ULL,
    0x1903607ea000eULL,
    0x8b81708ef000bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x7800000000000ULL,
    0x9400000000000ULL,
    0x6a00000000000ULL,
    0xdb00000000000ULL,
    0x6080000000000ULL,
    0xea40000000000ULL,
    0xed60000000000ULL,
    0xe950000000000ULL,
    0x888000000000ULL,
    0x9644000000000ULL,
    0xbb66000000000ULL,
    0x45b000000000ULL,
    0x350f800000000ULL,
    0x398d400000000ULL,
    0x6dc0a00000000ULL,
    0x51a6b00000000ULL,
    0xb371880000000ULL,
    0x493fe40000000ULL,
    0x4c54760000000ULL,
    0x9909250000000ULL,
    0xa78f808000000ULL,
    0x6ccd404000000ULL,
    0x9220a06000000ULL,
    0xadb6b0b000000ULL,
    0x2419887800000ULL,
    0x936be49400000ULL,
    0x185a766a00000ULL,
    0xd30625db00000ULL,
    0xac8e00e080000ULL,
    0xc44f00aa40000ULL,
    0x9461808d60000ULL,
    0xbed2405950000ULL,
    0x7c49207088000ULL,
    0x2060f00244000ULL,
    0x4d6a8d166000ULL,
    0x6f4014df5b000ULL,
    0xaceb5e558f800ULL,
    0x109f71d3cd400ULL,
    0x332bfe00a0a00ULL,
    0xfe39c1f8f6b00ULL,
    0x92da765bf9880ULL,
    0xa246252f7be40ULL,
    0x416e006f32760ULL,
    0x775f00f952250ULL,
    0x98980e080008ULL,
    0x55c640aa40004ULL,
    0xa5a7208d60006ULL,
    0x697ff0595000bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x6800000000000ULL,
    0x2c00000000000ULL,
    0x3600000000000ULL,
    0x1300000000000ULL,
    0x6380000000000ULL,
    0x6240000000000ULL,
    0x78a0000000000ULL,
    0x8930000000000ULL,
    0xf88000000000ULL,
    0x344c000000000ULL,
    0xfbaa000000000ULL,
    0xb2b3000000000ULL,
    0x29ce800000000ULL,
    0x56eec00000000ULL,
    0x5799600000000ULL,
    0xcdba300000000ULL,
    0x1c44b80000000ULL,
    0xb7a2e40000000ULL,
    0x14bdea0000000ULL,
    0x62c4230000000ULL,
    0x716e808000000ULL,
    0x2fdec0c000000ULL,
    0x901160a000000ULL,
    0xe5f6303000000ULL,
    0xb9eeb86800000ULL,
    0x3a11e42c00000ULL,
    0x68f36a3600000ULL,
    0x456ae31300000ULL,
    0x3dd7e0e380000ULL,
    0x914f0a240000ULL,
    0xfb7dd8d8a0000ULL,
    0xef28d4b930000ULL,
    0x5971526788000ULL,
    0xde2ac7184c000ULL,
    0xcbf96a4daa000ULL,
    0x8ee9e361b3000ULL,
    0xd391606a4e800ULL,
    0x77b630c4aec00ULL,
    0x94eb8e739600ULL,
    0xaf21e4588a300ULL,
    0x397b6a4dccb80ULL,
    0x4e26e3bceee40ULL,
    0x93fde03a97ea0ULL,
    0xcae7f06137230ULL,
    0xc99358e380008ULL,
    0x52b614a24000cULL,
    0x79c032d8a000aULL,
    0xaeecf7b930003ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x2800000000000ULL,
    0x400000000000ULL,
    0x4a00000000000ULL,
    0xad00000000000ULL,
    0xce80000000000ULL,
    0xffc0000000000ULL,
    0x4520000000000ULL,
    0x1990000000000ULL,
    0x9288000000000ULL,
    0x79cc000000000ULL,
    0xb622000000000ULL,
    0x2813000000000ULL,
    0x22ca800000000ULL,
    0x43ac400000000ULL,
    0xf256a00000000ULL,
    0x85a1d00000000ULL,
    0xe152680000000ULL,
    0xa425bc0000000ULL,
    0x491ff20000000ULL,
    0xb243c90000000ULL,
    0x436a808000000ULL,
    0xf8fc40c000000ULL,
    0xe37ea02000000ULL,
    0x2e3dd03000000ULL,
    0xe2d8682800000ULL,
    0x456abc0400000ULL,
    0xcbff724a00000ULL,
    0xf2f089ad00000ULL,
    0xae74a04e80000ULL,
    0xf0b2d03fc0000ULL,
    0x518e86520000ULL,
    0x2c49fc2990000ULL,
    0x9c6952ba88000ULL,
    0x8772197dcc000ULL,
    0xb430e87c22000ULL,
    0x77d5fc4513000ULL,
    0x97e352cc4a800ULL,
    0x523d198c6c400ULL,
    0x54d0681f76a00ULL,
    0x9e66bc5831d00ULL,
    0xfe7d7219da680ULL,
    0x8b38940e9bc0ULL,
    0x39162099bdf20ULL,
    0x7a4290a190c90ULL,
    0x5764484e80008ULL,
    0xaaf72c3fc000cULL,
    0x6273ba6520002ULL,
    0xaeb8e52990003ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x7800000000000ULL,
    0x7400000000000ULL,
    0xf600000000000ULL,
    0x5f00000000000ULL,
    0x7f80000000000ULL,
    0x5d40000000000ULL,
    0xe0e0000000000ULL,
    0xf010000000000ULL,
    0x9988000000000ULL,
    0xea4c000000000ULL,
    0xc362000000000ULL,
    0x2757000000000ULL,
    0xdc6f800000000ULL,
    0xb8db400000000ULL,
    0xd125600000000ULL,
    0x503ef00000000ULL,
    0x2d3a780000000ULL,
    0x4bb1940000000ULL,
    0x6e70ee0000000ULL,
    0xa85cb10000000ULL,
    0x6be7808000000ULL,
    0xc99740c000000ULL,
    0xc3c7602000000ULL,
    0x7129f07000000ULL,
    0xe035f87800000ULL,
    0x753ad47400000ULL,
    0x4fbd8ef600000ULL,
    0xe07e415f00000ULL,
    0x8357f8ff80000ULL,
    0xe26dd49d40000ULL,
    0xcbd20ec0e0000ULL,
    0x5ca5018010000ULL,
    0xdc72986188000ULL,
    0x9953245e4c000ULL,
    0x6f68761562000ULL,
    0x1554950857000ULL,
    0x2d62765bef800ULL,
    0x9c5f95519b400ULL,
    0x7de7f667c5600ULL,
    0xc69fd54f2ef00ULL,
    0xb44f161332780ULL,
    0x206d6538bd940ULL,
    0x82df8e63f2ee0ULL,
    0x2c2941e41bb10ULL,
    0xf6b8787f80008ULL,
    0x28f6945d4000cULL,
    0xfd976ee0e0002ULL,
    0xd5cbf1f010007ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xa800000000000ULL,
    0xe400000000000ULL,
    0x6600000000000ULL,
    0x1100000000000ULL,
    0xff80000000000ULL,
    0x30c0000000000ULL,
    0x90a0000000000ULL,
    0x7270000000000ULL,
    0xd188000000000ULL,
    0x75cc000000000ULL,
    0xd12a000000000ULL,
    0xcfbd000000000ULL,
    0x94a2800000000ULL,
    0xa472400000000ULL,
    0x1884600000000ULL,
    0x1640100000000ULL,
    0xcbef780000000ULL,
    0xcc104c0000000ULL,
    0x6e5dea0000000ULL,
    0xabbc770000000ULL,
    0x32aa808000000ULL,
    0x157e40c000000ULL,
    0x370e60a000000ULL,
    0x8e8d10d000000ULL,
    0xbf45f8a800000ULL,
    0xd86e0ce400000ULL,
    0xaed38a6600000ULL,
    0x21f1671100000ULL,
    0xd34f787f80000ULL,
    0x4a604cf0c0000ULL,
    0xd1d5ea30a0000ULL,
    0xfb7077a270000ULL,
    0xd20080f988000ULL,
    0x1f034051cc000ULL,
    0xaa8ce0172a000ULL,
    0x794f500ebd000ULL,
    0x196998c322800ULL,
    0xf9521c70b2400ULL,
    0xf53ef2ee24600ULL,
    0x25e02b7530100ULL,
    0x29121265e7780ULL,
    0xffdf7b491c4c0ULL,
    0xbe738a0fd7ea0ULL,
    0x9381670671770ULL,
    0xa2c778ff80008ULL,
    0xefac4c30c000cULL,
    0xa8ffea90a000aULL,
    0xd0cd77727000dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0x4c00000000000ULL,
    0xe00000000000ULL,
    0x4900000000000ULL,
    0x9380000000000ULL,
    0x81c0000000000ULL,
    0x4960000000000ULL,
    0xdc30000000000ULL,
    0x5188000000000ULL,
    0x86c4000000000ULL,
    0xb3e2000000000ULL,
    0x3e7f000000000ULL,
    0x7126800000000ULL,
    0x5f90c00000000ULL,
    0x305ae00000000ULL,
    0xa037900000000ULL,
    0xd789b80000000ULL,
    0x53c5dc0000000ULL,
    0x3667f60000000ULL,
    0x52bc130000000ULL,
    0xe9ce808000000ULL,
    0x4564c04000000ULL,
    0xf230e02000000ULL,
    0xe88c90f000000ULL,
    0xfd4d38e800000ULL,
    0x7e2a1c4c00000ULL,
    0x791b960e00000ULL,
    0xe41b434900000ULL,
    0x9d9dd81380000ULL,
    0x37568cc1c0000ULL,
    0x5abeae6960000ULL,
    0x35c55f2c30000ULL,
    0x736c4e3988000ULL,
    0xef36cf8ac4000ULL,
    0xd107f61de2000ULL,
    0x778c13c77f000ULL,
    0xe3c680aaa6800ULL,
    0xfe60c02250c00ULL,
    0xeeb2e0bf3ae00ULL,
    0xfc3908907900ULL,
    0x4063b8f381b80ULL,
    0x6fbedc51c1dc0ULL,
    0x2043765165f60ULL,
    0x27a3d37833130ULL,
    0x23dae09380008ULL,
    0x61f79081c0004ULL,
    0xbee9b84960002ULL,
    0x7ff5dcdc3000fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x6800000000000ULL,
    0x8c00000000000ULL,
    0xe600000000000ULL,
    0xbf00000000000ULL,
    0x3d80000000000ULL,
    0x85c0000000000ULL,
    0x59a0000000000ULL,
    0x67d0000000000ULL,
    0x4788000000000ULL,
    0x54c4000000000ULL,
    0x272a000000000ULL,
    0x819d000000000ULL,
    0xc0ee800000000ULL,
    0x3d3cc00000000ULL,
    0xdbbc600000000ULL,
    0x49faf00000000ULL,
    0x6c93580000000ULL,
    0x6b9c0000000ULL,
    0xf07e7a0000000ULL,
    0xb35ccd0000000ULL,
    0x18c6808000000ULL,
    0x2128c04000000ULL,
    0x4e9e60a000000ULL,
    0x4563f0d000000ULL,
    0x5cf7d86800000ULL,
    0xe81a5c8c00000ULL,
    0x77249ae600000ULL,
    0xa99efdbf00000ULL,
    0xece3b8bd80000ULL,
    0xb34acc5c0000ULL,
    0xcb5c2f9a0000ULL,
    0xf87c61b7d0000ULL,
    0xf5142af88000ULL,
    0xe6cda198c4000ULL,
    0xaa2ba2e12a000ULL,
    0x711f91ae9d000ULL,
    0x15a69ab56e800ULL,
    0x61d7fda4fcc00ULL,
    0x888f382c1c600ULL,
    0xd1416c8d2af00ULL,
    0x46e522389b580ULL,
    0xb23351326f9c0ULL,
    0xfe32fabd747a0ULL,
    0xf8390da3d1cd0ULL,
    0x373e603d80008ULL,
    0xb2b3f085c0004ULL,
    0xd37fd859a000aULL,
    0xe0de5c67d000dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xb800000000000ULL,
    0x5400000000000ULL,
    0x5e00000000000ULL,
    0xfb00000000000ULL,
    0xbd80000000000ULL,
    0xe640000000000ULL,
    0x8be0000000000ULL,
    0x1290000000000ULL,
    0xa688000000000ULL,
    0xcbcc000000000ULL,
    0xb5a6000000000ULL,
    0x9d75000000000ULL,
    0x5213800000000ULL,
    0xc241400000000ULL,
    0x9defe00000000ULL,
    0x359cb00000000ULL,
    0xf90d580000000ULL,
    0x648ca40000000ULL,
    0x72ce1e0000000ULL,
    0x712e790000000ULL,
    0xdfb5808000000ULL,
    0xcb3440c000000ULL,
    0xf1fc606000000ULL,
    0x5cddf05000000ULL,
    0x6162b8b800000ULL,
    0xe350145400000ULL,
    0x5e23465e00000ULL,
    0xfc32ddfb00000ULL,
    0xb6739e3d80000ULL,
    0x9796392640000ULL,
    0x100fe06be0000ULL,
    0x180cb08290000ULL,
    0x6405587e88000ULL,
    0xb600a4cfcc000ULL,
    0x17081e53a6000ULL,
    0xb78b793275000ULL,
    0x434e003193800ULL,
    0xcd69001f01400ULL,
    0x495d80cb8fe00ULL,
    0xcb2840914cb00ULL,
    0x52b2606c65580ULL,
    0x1b4f0e9d0a40ULL,
    0xf03f383fe01e0ULL,
    0x2c7854dc97790ULL,
    0xea91268580008ULL,
    0x52862d724000cULL,
    0xa5cca635e0006ULL,
    0xa6ae6d7990005ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0xd400000000000ULL,
    0xf200000000000ULL,
    0xc900000000000ULL,
    0x4380000000000ULL,
    0x3040000000000ULL,
    0x4160000000000ULL,
    0xb7f0000000000ULL,
    0xf688000000000ULL,
    0xf5cc000000000ULL,
    0xba26000000000ULL,
    0x419000000000ULL,
    0xeb6800000000ULL,
    0x5469400000000ULL,
    0x8275200000000ULL,
    0x75cb900000000ULL,
    0x7a2fb80000000ULL,
    0x6416c40000000ULL,
    0x9eb7760000000ULL,
    0xbc624f0000000ULL,
    0x5670808000000ULL,
    0x87c040c000000ULL,
    0xb32ba06000000ULL,
    0x279ed09000000ULL,
    0xaef498e800000ULL,
    0xfd0854d400000ULL,
    0xe1884ef200000ULL,
    0x7144cbc900000ULL,
    0x46e456c380000ULL,
    0x9db0dff040000ULL,
    0xaae9b8a160000ULL,
    0xf3bfc4e7f0000ULL,
    0xb5e9f67e88000ULL,
    0xf3370fb1cc000ULL,
    0x332ba02026000ULL,
    0xe79ed0d919000ULL,
    0xcef498df36800ULL,
    0x6d08543d29400ULL,
    0x9884e6895200ULL,
    0xa544cb267b900ULL,
    0xb4e456bfc7b80ULL,
    0x54b0df2f2ac40ULL,
    0xe969b87199760ULL,
    0xc3ffc42df74f0ULL,
    0xf489f62b80008ULL,
    0x44c70f244000cULL,
    0xc5a3a05360006ULL,
    0x1252d02ef0009ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xe800000000000ULL,
    0xb400000000000ULL,
    0xfa00000000000ULL,
    0x900000000000ULL,
    0x3180000000000ULL,
    0xd8c0000000000ULL,
    0x3520000000000ULL,
    0x2590000000000ULL,
    0xda88000000000ULL,
    0xfc44000000000ULL,
    0xb262000000000ULL,
    0x4b77000000000ULL,
    0x1ab6800000000ULL,
    0x5e17400000000ULL,
    0x8341a00000000ULL,
    0x1ce1900000000ULL,
    0x2d38980000000ULL,
    0xf1d74c0000000ULL,
    0xbce5320000000ULL,
    0x1d30290000000ULL,
    0x39dc808000000ULL,
    0x78e4404000000ULL,
    0xf35202000000ULL,
    0x84d1d07000000ULL,
    0xb367b8e800000ULL,
    0xdef59cb400000ULL,
    0x80760afa00000ULL,
    0x4e35f50900000ULL,
    0x3155aab180000ULL,
    0x59a36598c0000ULL,
    0x26dbb29520000ULL,
    0x8663691590000ULL,
    0xf17f209288000ULL,
    0x33b2d07844000ULL,
    0x1f9b388062000ULL,
    0xb381dc8677000ULL,
    0x9dcb2a3936800ULL,
    0xf2a0253bd7400ULL,
    0x225012fde1a00ULL,
    0x7121f9a8b1900ULL,
    0x779b385310980ULL,
    0x4781dcc0c34c0ULL,
    0x47cb2a292f320ULL,
    0x8ba0254b93290ULL,
    0xfbd012d980008ULL,
    0x1de1f96cc0004ULL,
    0xb8bb384f20002ULL,
    0x6b11dc6c90007ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x6800000000000ULL,
    0x3c00000000000ULL,
    0x8600000000000ULL,
    0xb500000000000ULL,
    0x3680000000000ULL,
    0x4940000000000ULL,
    0xa320000000000ULL,
    0xc070000000000ULL,
    0x9d88000000000ULL,
    0xeecc000000000ULL,
    0xbee2000000000ULL,
    0xdd19000000000ULL,
    0xaf5e800000000ULL,
    0xc3f7c00000000ULL,
    0x224e600000000ULL,
    0x24ac500000000ULL,
    0x4dbee80000000ULL,
    0xe8e0d40000000ULL,
    0xe01b120000000ULL,
    0x15d6f70000000ULL,
    0xf4b4808000000ULL,
    0x6062c0c000000ULL,
    0x6852e02000000ULL,
    0xaa72909000000ULL,
    0x2686086800000ULL,
    0xe147443c00000ULL,
    0xbf219a8600000ULL,
    0xd67f73b500000ULL,
    0x4085fab680000ULL,
    0xe446238940000ULL,
    0x71a7920320000ULL,
    0xcb38379070000ULL,
    0x59a4605588000ULL,
    0x17395082cc000ULL,
    0x6fa26870e2000ULL,
    0x5a3e14c419000ULL,
    0xd23f277de800ULL,
    0xa9716703b7c00ULL,
    0x9d0e08b1ee600ULL,
    0xea8b44d89c500ULL,
    0x7f439a6596e80ULL,
    0xee26731f1cd40ULL,
    0xa2fb7a8851120ULL,
    0x6ec1e31a73f70ULL,
    0x7ee1f25e80008ULL,
    0xfd1867754000cULL,
    0x3f5888a520002ULL,
    0xabf084b570009ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0x8c00000000000ULL,
    0x1e00000000000ULL,
    0xd100000000000ULL,
    0x9880000000000ULL,
    0x15c0000000000ULL,
    0x5fe0000000000ULL,
    0xaeb0000000000ULL,
    0x65c8000000000ULL,
    0x27ec000000000ULL,
    0x6ab6000000000ULL,
    0x67c1000000000ULL,
    0xc0ee800000000ULL,
    0xaf3ec00000000ULL,
    0xf480e00000000ULL,
    0x5bc3900000000ULL,
    0xc6e7480000000ULL,
    0x2a39bc0000000ULL,
    0x46026e0000000ULL,
    0x2507a30000000ULL,
    0x4280e08000000ULL,
    0xc6c390c000000ULL,
    0x2067486000000ULL,
    0x7ef9bc9000000ULL,
    0xa9626e2800000ULL,
    0x1277a38c00000ULL,
    0x66a8e09e00000ULL,
    0x9e9f901100000ULL,
    0xb79948f880000ULL,
    0x2b14bc85c0000ULL,
    0x5cdaee77e0000ULL,
    0x74386322b0000ULL,
    0x370e80fbc8000ULL,
    0x4d8ec036ec000ULL,
    0xef48e09236000ULL,
    0x3d2f90e201000ULL,
    0x1cd148b70e800ULL,
    0xd438bc8d8ec00ULL,
    0xc70cee8f48e00ULL,
    0xf58963ad2f900ULL,
    0x4b4800b4d1480ULL,
    0xaf2c009838bc0ULL,
    0xd3d600b90cee0ULL,
    0x9db100b489630ULL,
    0x4a4680fbc8008ULL,
    0xbfa2c036ec00cULL,
    0xba1ee09236006ULL,
    0x645e90e201009ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0xe400000000000ULL,
    0xe200000000000ULL,
    0x7b00000000000ULL,
    0x5980000000000ULL,
    0xa0c0000000000ULL,
    0x12a0000000000ULL,
    0x20b0000000000ULL,
    0xbcc8000000000ULL,
    0x4ac000000000ULL,
    0xb1b6000000000ULL,
    0x1243000000000ULL,
    0x23e1800000000ULL,
    0x7ed8400000000ULL,
    0xe0dd200000000ULL,
    0x3dde300000000ULL,
    0x2d59d80000000ULL,
    0xc71d2c0000000ULL,
    0x10f21a0000000ULL,
    0x842d530000000ULL,
    0x93fd208000000ULL,
    0x29ae30c000000ULL,
    0xd931d86000000ULL,
    0xcc012cb000000ULL,
    0x7e0c1ad800000ULL,
    0xad0253e400000ULL,
    0xa88aa06200000ULL,
    0xbe4570bb00000ULL,
    0xede5783980000ULL,
    0xbdb5c10c0000ULL,
    0xac56e2caa0000ULL,
    0x619e4fc4b0000ULL,
    0xbbbee2dec8000ULL,
    0x95424fbfac000ULL,
    0x1c60e20836000ULL,
    0xf71d4fc283000ULL,
    0x8ff620941800ULL,
    0x2a0fca68400ULL,
    0xc1f5c28615200ULL,
    0x8aaf7fd672300ULL,
    0x64b0ba1f6fd80ULL,
    0x8ec4239a9e2c0ULL,
    0x17ae58a2339a0ULL,
    0xd4366c9585130ULL,
    0xb486badec8008ULL,
    0xa84723bfac00cULL,
    0x7cefd80836006ULL,
    0xa55e2cc28300bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x5800000000000ULL,
    0x4400000000000ULL,
    0x8600000000000ULL,
    0xdf00000000000ULL,
    0x1f80000000000ULL,
    0xa440000000000ULL,
    0x29e0000000000ULL,
    0xced0000000000ULL,
    0xa848000000000ULL,
    0x3e4000000000ULL,
    0xebd6000000000ULL,
    0x5ac7000000000ULL,
    0xb921800000000ULL,
    0xe2f2400000000ULL,
    0x803f600000000ULL,
    0x8d14700000000ULL,
    0xb1efb80000000ULL,
    0xaadf240000000ULL,
    0xbe4bee0000000ULL,
    0x74e7d50000000ULL,
    0xe85f608000000ULL,
    0x3c84704000000ULL,
    0xc9c7b86000000ULL,
    0xecab24f000000ULL,
    0x93b5ee5800000ULL,
    0x354d54400000ULL,
    0xbd00e00600000ULL,
    0x9685309f00000ULL,
    0xacc7587f80000ULL,
    0x7e2e145440000ULL,
    0xd972b671e0000ULL,
    0x527ac18ad0000ULL,
    0x23f256ae48000ULL,
    0x24bff19ce4000ULL,
    0x20d50e1456000ULL,
    0x4b41e54e87000ULL,
    0x4e6fb828c1800ULL,
    0xbe9f24d822400ULL,
    0xafabee9677600ULL,
    0xe37d5e5f0700ULL,
    0x9e17601bb9b80ULL,
    0xa460708f58240ULL,
    0xbb91b8d70a6e0ULL,
    0xcd2c24d385950ULL,
    0x1cf46eae48008ULL,
    0x8b36959ce4004ULL,
    0xbc97801456006ULL,
    0xd6a5404e8700fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0xc800000000000ULL,
    0x7400000000000ULL,
    0xda00000000000ULL,
    0x9f00000000000ULL,
    0x1380000000000ULL,
    0x8040000000000ULL,
    0x4760000000000ULL,
    0x6e10000000000ULL,
    0xce48000000000ULL,
    0xf264000000000ULL,
    0x9a92000000000ULL,
    0x7181000000000ULL,
    0x5348800000000ULL,
    0x82e5400000000ULL,
    0x8754a00000000ULL,
    0x63ad700000000ULL,
    0x3eb2780000000ULL,
    0xc57fa40000000ULL,
    0xa658860000000ULL,
    0xbd25590000000ULL,
    0x667ca08000000ULL,
    0x5bd9704000000ULL,
    0x5868782000000ULL,
    0x3d9aa49000000ULL,
    0x6e0206c800000ULL,
    0x2501197400000ULL,
    0x3c80805a00000ULL,
    0xcbc140df00000ULL,
    0x7b26a03380000ULL,
    0x877c701040000ULL,
    0xe552f88f60000ULL,
    0xb0aee41a10000ULL,
    0xfb36261448000ULL,
    0x2c3d296d64000ULL,
    0xbbc580912000ULL,
    0x71f694b1c1000ULL,
    0x3996de3428800ULL,
    0x8c02cd7cf5400ULL,
    0xb60afe011ca00ULL,
    0xd90afda5c9700ULL,
    0xb28c26de20780ULL,
    0x1ec829fbfea40ULL,
    0x3faed88e90060ULL,
    0x60b6d41b80190ULL,
    0xf878fe9448008ULL,
    0x6dbfd2d64004ULL,
    0x48eca62912002ULL,
    0x90596921c1009ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0xfc00000000000ULL,
    0x9600000000000ULL,
    0x1300000000000ULL,
    0xe380000000000ULL,
    0x9640000000000ULL,
    0xf60000000000ULL,
    0x2b30000000000ULL,
    0x1448000000000ULL,
    0x5664000000000ULL,
    0x3db6000000000ULL,
    0x4a89000000000ULL,
    0xf8c6800000000ULL,
    0xe5a9c00000000ULL,
    0xff98600000000ULL,
    0x615bb00000000ULL,
    0xeffdf80000000ULL,
    0x47ea040000000ULL,
    0xcaf2c60000000ULL,
    0x87610b0000000ULL,
    0xe730608000000ULL,
    0xda4fb04000000ULL,
    0x8163f86000000ULL,
    0x9c37041000000ULL,
    0xa5ca462800000ULL,
    0x9125cbfc00000ULL,
    0xa858801600000ULL,
    0x9174c05300000ULL,
    0x8520e08380000ULL,
    0xe25f708640000ULL,
    0x6475182760000ULL,
    0x4da174d730000ULL,
    0x4399de8248000ULL,
    0x975d7f4564000ULL,
    0xecf13e5e36000ULL,
    0x8c660f9cc9000ULL,
    0xa0b22617a6800ULL,
    0x1e0e7b9e99c00ULL,
    0xdf0d7823d0600ULL,
    0x2d8ac49b3fb00ULL,
    0x414c260c4bf80ULL,
    0xaee37bf263040ULL,
    0xc47df8efb4460ULL,
    0x7daa045b88cb0ULL,
    0x1b92c60248008ULL,
    0x53510b0564004ULL,
    0xaef8603e36006ULL,
    0xf56bb08cc9001ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xa800000000000ULL,
    0x1400000000000ULL,
    0x5a00000000000ULL,
    0xa900000000000ULL,
    0xbc80000000000ULL,
    0x8040000000000ULL,
    0xf3e0000000000ULL,
    0xa050000000000ULL,
    0xb648000000000ULL,
    0xece4000000000ULL,
    0x43d2000000000ULL,
    0xf58b000000000ULL,
    0x6cce800000000ULL,
    0xcba3400000000ULL,
    0xdfb6a00000000ULL,
    0xf018100000000ULL,
    0xbda6880000000ULL,
    0xe0bba40000000ULL,
    0x239cae0000000ULL,
    0x56644d0000000ULL,
    0x749ea08000000ULL,
    0xf1ec104000000ULL,
    0xd55c882000000ULL,
    0x54c4a43000000ULL,
    0x7a82ea800000ULL,
    0x19b80d1400000ULL,
    0x271c80da00000ULL,
    0x5a2840e900000ULL,
    0x6178209c80000ULL,
    0xb6bb50b040000ULL,
    0x2c90285be0000ULL,
    0x2de3b4b450000ULL,
    0x8b5a26ec48000ULL,
    0x3fcfe945e4000ULL,
    0xae2a0e7f52000ULL,
    0x6b7c5d35cb000ULL,
    0xa7b8283f2e800ULL,
    0x1c17b41bf3400ULL,
    0x4ba026e1fea00ULL,
    0x9fb0e938fc100ULL,
    0xd01e8e8c74880ULL,
    0x8da01de830a40ULL,
    0x48ba0809d22e0ULL,
    0x3793e4c4870d0ULL,
    0xc648e6c48008ULL,
    0xdd9f1d05e4004ULL,
    0x4d6e885f52002ULL,
    0x551fa405cb003ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x8400000000000ULL,
    0xee00000000000ULL,
    0x4500000000000ULL,
    0x7e80000000000ULL,
    0xe2c0000000000ULL,
    0x64a0000000000ULL,
    0x5a90000000000ULL,
    0xbc8000000000ULL,
    0x2824000000000ULL,
    0xf756000000000ULL,
    0xae3000000000ULL,
    0xb3f5800000000ULL,
    0x8076400000000ULL,
    0x1039e00000000ULL,
    0x345fd00000000ULL,
    0x8d6ea80000000ULL,
    0x95b84c0000000ULL,
    0x51eda0000000ULL,
    0x8982610000000ULL,
    0x934fe08000000ULL,
    0x83ecd04000000ULL,
    0x4f73286000000ULL,
    0x35ba0cb000000ULL,
    0x95193a1800000ULL,
    0x418ab18400000ULL,
    0xbf4ac86e00000ULL,
    0xf1e5dc0500000ULL,
    0x6077921e80000ULL,
    0xe032fd52c0000ULL,
    0x4c5412fca0000ULL,
    0xb967bd9e90000ULL,
    0x63b87205c8000ULL,
    0xc41e2d9d24000ULL,
    0x19073a71d6000ULL,
    0x348db19c23000ULL,
    0x99c948c155800ULL,
    0xf7209cebe6400ULL,
    0x5ad3f2f371e00ULL,
    0xe7af6d8fbbd00ULL,
    0xbd1d5a1618a80ULL,
    0x9d8721a60b4c0ULL,
    0xb54b80a9035a0ULL,
    0xeae1402c80210ULL,
    0x43f2601dc8008ULL,
    0xf87e901924004ULL,
    0x243cc81fd6006ULL,
    0xc256dc992300bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0x3c00000000000ULL,
    0x2200000000000ULL,
    0x700000000000ULL,
    0xf980000000000ULL,
    0xf40000000000ULL,
    0x55e0000000000ULL,
    0x74b0000000000ULL,
    0xaa48000000000ULL,
    0xeb64000000000ULL,
    0x227a000000000ULL,
    0x20e5000000000ULL,
    0x323b800000000ULL,
    0xc081c00000000ULL,
    0x43cb200000000ULL,
    0xfc25f00000000ULL,
    0x3b91580000000ULL,
    0x5e55540000000ULL,
    0x5b7aee0000000ULL,
    0xcc6cf30000000ULL,
    0x4bf9208000000ULL,
    0x77a4f04000000ULL,
    0x8bd0d8a000000ULL,
    0x6e3194d000000ULL,
    0xd28a4ef800000ULL,
    0xccc8c33c00000ULL,
    0x11a358a200000ULL,
    0xd2d4544700000ULL,
    0x12bb6e5980000ULL,
    0xf34833df40000ULL,
    0x97e9802de0000ULL,
    0x3b0c008b0000ULL,
    0x7bc2a02848000ULL,
    0x2025303c64000ULL,
    0x699bf823fa000ULL,
    0x71546413a5000ULL,
    0x66fb1645db800ULL,
    0xdd2d97b331c00ULL,
    0x3b11b69003200ULL,
    0xfd9ca75801f00ULL,
    0xd7584e6c0b580ULL,
    0xdff9c39a00540ULL,
    0xd1aad81b096e0ULL,
    0x32d4944b89330ULL,
    0x62b1ce5048008ULL,
    0xdb49034064004ULL,
    0x53e87821fa00aULL,
    0x1db1a484a500dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x8800000000000ULL,
    0x8400000000000ULL,
    0x9200000000000ULL,
    0x4d00000000000ULL,
    0x9580000000000ULL,
    0x3cc0000000000ULL,
    0x5020000000000ULL,
    0x9290000000000ULL,
    0x13c8000000000ULL,
    0x20ac000000000ULL,
    0xefde000000000ULL,
    0xd523000000000ULL,
    0xa314800000000ULL,
    0xed0e400000000ULL,
    0x58e200000000ULL,
    0xe4c2500000000ULL,
    0xec2a180000000ULL,
    0xc926c0000000ULL,
    0x48c1920000000ULL,
    0x6a25510000000ULL,
    0xb98208000000ULL,
    0x2c4d50c000000ULL,
    0x5d6098e000000ULL,
    0x3c7f2cb000000ULL,
    0xa47b328800000ULL,
    0x3879418400000ULL,
    0xf674181200000ULL,
    0x95716c8d00000ULL,
    0xd3f5127580000ULL,
    0x21bb118cc0000ULL,
    0x7de005820000ULL,
    0x212300d690000ULL,
    0x59148061c8000ULL,
    0x940e40ddac000ULL,
    0x8a0e20f25e000ULL,
    0x1102506de3000ULL,
    0xbb8a186134800ULL,
    0xefc26c329e400ULL,
    0x9ea99283c6200ULL,
    0xe4d951f8ae500ULL,
    0xa7ae20d3d4180ULL,
    0x4b52508b216c0ULL,
    0x26218181d120ULL,
    0x24fe6c1787110ULL,
    0xed3f9269c8008ULL,
    0xe4965199ac00cULL,
    0xbcc4a0005e00eULL,
    0x902f1050e300bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0x8c00000000000ULL,
    0x3e00000000000ULL,
    0xeb00000000000ULL,
    0xb380000000000ULL,
    0xa1c0000000000ULL,
    0x2f60000000000ULL,
    0x4890000000000ULL,
    0x8248000000000ULL,
    0xd6a4000000000ULL,
    0xf3fe000000000ULL,
    0x80d3000000000ULL,
    0x45e8800000000ULL,
    0xcfd2c00000000ULL,
    0xb46ee00000000ULL,
    0x1315300000000ULL,
    0xff81780000000ULL,
    0x3fce3c0000000ULL,
    0x9465260000000ULL,
    0x312c10000000ULL,
    0xe786608000000ULL,
    0x4bc7f04000000ULL,
    0x6e6f98e000000ULL,
    0x5a1b0cb000000ULL,
    0x81045e4800000ULL,
    0xb28cfd8c00000ULL,
    0x534b46be00000ULL,
    0x9c2131ab00000ULL,
    0x64bff8d380000ULL,
    0xaefbfc51c0000ULL,
    0xf455468760000ULL,
    0x39a2317490000ULL,
    0x527f787448000ULL,
    0xa41d3cf1a4000ULL,
    0xca0da69e7e000ULL,
    0x5100017a13000ULL,
    0xa88801108800ULL,
    0x7742c0ea82c00ULL,
    0x9e26e0c746e00ULL,
    0xf9b130d621300ULL,
    0x7a7f7875b7780ULL,
    0xd81d3c44793c0ULL,
    0x5c0da63313a60ULL,
    0x860001ef83010ULL,
    0xcf0880a7c8008ULL,
    0xb182c0a064004ULL,
    0x3cc6e0191e00eULL,
    0xfbe1300e8300bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xf800000000000ULL,
    0x6400000000000ULL,
    0xa00000000000ULL,
    0x4300000000000ULL,
    0x1080000000000ULL,
    0xb40000000000ULL,
    0x36a0000000000ULL,
    0xfe30000000000ULL,
    0xf3c8000000000ULL,
    0x91ec000000000ULL,
    0x3e96000000000ULL,
    0xd0f7000000000ULL,
    0xc3ab800000000ULL,
    0x93bc400000000ULL,
    0xc901a00000000ULL,
    0x9388b00000000ULL,
    0x7bcec80000000ULL,
    0xcde9540000000ULL,
    0x30987a0000000ULL,
    0x69f49b0000000ULL,
    0x682a208000000ULL,
    0xec74f0c000000ULL,
    0xfeef686000000ULL,
    0x1811e4f000000ULL,
    0x66beb2f800000ULL,
    0xa481cf6400000ULL,
    0xa94c5a8a00000ULL,
    0x79ab6b8300000ULL,
    0x88b0c8f080000ULL,
    0x2d82543b40000ULL,
    0x9acdfaaea0000ULL,
    0x9263db6a30000ULL,
    0x4d5e0081c8000ULL,
    0x811b0076ec000ULL,
    0x9d3d804416000ULL,
    0xb34b4068b7000ULL,
    0xf2aa201d8b800ULL,
    0x6434f002cc400ULL,
    0xb8cf680669a00ULL,
    0x1d61e4bf54b00ULL,
    0x5bd6b2a610c80ULL,
    0xaf5dcf87b2540ULL,
    0x6e125a7b05fa0ULL,
    0x7bb06b148fdb0ULL,
    0x650d487148008ULL,
    0x6589144dac00cULL,
    0xa6c7daeab6006ULL,
    0x6c672b028700fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0x4400000000000ULL,
    0x4a00000000000ULL,
    0x5700000000000ULL,
    0xa680000000000ULL,
    0x3f40000000000ULL,
    0xbda0000000000ULL,
    0xb690000000000ULL,
    0x5dc8000000000ULL,
    0x88e4000000000ULL,
    0x3c36000000000ULL,
    0xfdd3000000000ULL,
    0xd26f800000000ULL,
    0x4d76400000000ULL,
    0x3d71a00000000ULL,
    0x2571f00000000ULL,
    0xb177a80000000ULL,
    0x7372140000000ULL,
    0x58788a0000000ULL,
    0xf0f2b10000000ULL,
    0xd2be208000000ULL,
    0x9e97b04000000ULL,
    0xb1ce086000000ULL,
    0x6ee7e4b000000ULL,
    0xed39223800000ULL,
    0x3a53a54400000ULL,
    0x12a92aca00000ULL,
    0x7413411700000ULL,
    0xf881884680000ULL,
    0xea41a4cf40000ULL,
    0x502082e5a0000ULL,
    0x9156554290000ULL,
    0xfa2082afc8000ULL,
    0x365655dbe4000ULL,
    0x4a08230b6000ULL,
    0xfd16556593000ULL,
    0xcb0082114f800ULL,
    0x58865570a6400ULL,
    0x7a48824f19a00ULL,
    0xb82255f805f00ULL,
    0x5d5e82e409a80ULL,
    0xcc2155da05140ULL,
    0x6f59023f010a0ULL,
    0xbf23152a87f10ULL,
    0x33d6a2e948008ULL,
    0xef65e514a4004ULL,
    0x6cf88ad516006ULL,
    0x2cb2b1270300bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0x3c00000000000ULL,
    0xd600000000000ULL,
    0xbd00000000000ULL,
    0x4c80000000000ULL,
    0xc040000000000ULL,
    0xbe20000000000ULL,
    0xd630000000000ULL,
    0xf8c8000000000ULL,
    0xec64000000000ULL,
    0xdf1e000000000ULL,
    0xf977000000000ULL,
    0x1d6f800000000ULL,
    0x3d99c00000000ULL,
    0x5834600000000ULL,
    0xc9cb500000000ULL,
    0xaeee880000000ULL,
    0x6e53a40000000ULL,
    0xe5d9d20000000ULL,
    0x8a1e3b0000000ULL,
    0xe1fbe08000000ULL,
    0x1f22904000000ULL,
    0x1cb2e8e000000ULL,
    0xad8cf4f000000ULL,
    0xeac15a3800000ULL,
    0x1b6e9f3c00000ULL,
    0xe89bb25600000ULL,
    0xb66bfd00000ULL,
    0x2b8ce82c80000ULL,
    0x7fcbf47040000ULL,
    0xa3e6da6620000ULL,
    0xfad35f1a30000ULL,
    0xe991d296c8000ULL,
    0xda3a3b2d64000ULL,
    0xb6c5e0259e000ULL,
    0x7d65903437000ULL,
    0x8d956837cf800ULL,
    0x803134e7e9c00ULL,
    0x5cb3af0dc600ULL,
    0x40e2cf729f500ULL,
    0xef52ba93b8880ULL,
    0x7f580f4600a40ULL,
    0xf75f5ab508520ULL,
    0x13599fe880fb0ULL,
    0x95432ba48008ULL,
    0x5a5fab5d24004ULL,
    0x97d08843be00eULL,
    0xcd14a42e0700fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x2800000000000ULL,
    0xf400000000000ULL,
    0xe200000000000ULL,
    0x2500000000000ULL,
    0x9280000000000ULL,
    0x6240000000000ULL,
    0xe20000000000ULL,
    0x2b0000000000ULL,
    0x1cc8000000000ULL,
    0xe26c000000000ULL,
    0x679a000000000ULL,
    0x47f5000000000ULL,
    0x1566800000000ULL,
    0x4819400000000ULL,
    0x1b31200000000ULL,
    0x9709d00000000ULL,
    0x5f82e80000000ULL,
    0xa4c0440000000ULL,
    0x5e63120000000ULL,
    0x196930000000ULL,
    0x78f7a08000000ULL,
    0x7ee090c000000ULL,
    0xaedbc8a000000ULL,
    0xb05594d000000ULL,
    0x2c13fa2800000ULL,
    0xb13fd7f400000ULL,
    0x3600326200000ULL,
    0xd70643e500000ULL,
    0x3f89c8b280000ULL,
    0xd4cc947240000ULL,
    0xa66f7a8620000ULL,
    0xdd939726b0000ULL,
    0x6ef79256c8000ULL,
    0xb9e6d3f36c000ULL,
    0x195200371a000ULL,
    0x40990010b5000ULL,
    0x407c8001c6800ULL,
    0xbdac400ce9400ULL,
    0x2877a0c3d9200ULL,
    0x29a090a6d5d00ULL,
    0xba7bc85450e80ULL,
    0xf4a5944619440ULL,
    0xf4fbfab03f920ULL,
    0x80e3d76e8ad30ULL,
    0x3dd232e448008ULL,
    0x35df43812c00cULL,
    0xd1d548b13a00aULL,
    0xbbd0d4360500dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0xac00000000000ULL,
    0xd600000000000ULL,
    0x1f00000000000ULL,
    0x5380000000000ULL,
    0x24c0000000000ULL,
    0x25a0000000000ULL,
    0x3fd0000000000ULL,
    0xf148000000000ULL,
    0x4664000000000ULL,
    0x2df2000000000ULL,
    0x6057000000000ULL,
    0xb382800000000ULL,
    0xb4ccc00000000ULL,
    0x1da0600000000ULL,
    0x8bd0700000000ULL,
    0x6349780000000ULL,
    0x2361ec0000000ULL,
    0xb77cca0000000ULL,
    0x810350000000ULL,
    0xe16ae08000000ULL,
    0x8a78b04000000ULL,
    0xf69b182000000ULL,
    0x3269cf000000ULL,
    0xf99732e800000ULL,
    0x28ad19ac00000ULL,
    0x495e4a5600000ULL,
    0x6f0cf55f00000ULL,
    0xfb8280f380000ULL,
    0xa8ccc094c0000ULL,
    0x3a060eda0000ULL,
    0xc8d07063d0000ULL,
    0xec978cf48000ULL,
    0xb4a1ecf564000ULL,
    0x175ccaa872000ULL,
    0xc00355b97000ULL,
    0x6602e045a2800ULL,
    0xd70cb0efdcc00ULL,
    0xf8118e948600ULL,
    0x1ac59c0264700ULL,
    0x96afb257f3780ULL,
    0xba52d9a952ec0ULL,
    0xca8eaa7f0c4a0ULL,
    0x304745838bf50ULL,
    0x36e918bcc8008ULL,
    0xb4b19c21a4004ULL,
    0xca35b265d2002ULL,
    0x6871d9c84700fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x1800000000000ULL,
    0x400000000000ULL,
    0x4600000000000ULL,
    0x5d00000000000ULL,
    0x6180000000000ULL,
    0x2d40000000000ULL,
    0xfd60000000000ULL,
    0x5d30000000000ULL,
    0x96c8000000000ULL,
    0x8f24000000000ULL,
    0xfad2000000000ULL,
    0xa35000000000ULL,
    0xb04d800000000ULL,
    0xfce6400000000ULL,
    0x807b600000000ULL,
    0xa3a1500000000ULL,
    0xce13d80000000ULL,
    0x2e1cf40000000ULL,
    0x9e1f660000000ULL,
    0x76199b0000000ULL,
    0xba1ee08000000ULL,
    0xe013104000000ULL,
    0xff12b82000000ULL,
    0x859ca4d000000ULL,
    0x945b3e1800000ULL,
    0x25f22f0400000ULL,
    0xa8e566c600000ULL,
    0x9e789b1d00000ULL,
    0xdaa960c180000ULL,
    0x399450bd40000ULL,
    0x465e58c560000ULL,
    0x6fab48930000ULL,
    0x406406c8c8000ULL,
    0x8cb8cbd624000ULL,
    0x538d385d52000ULL,
    0xbe4fe43a75000ULL,
    0xfdedde0cad800ULL,
    0x83f53f5c96400ULL,
    0x45edde73d3600ULL,
    0x17f53f35b5500ULL,
    0x3bedde4409d80ULL,
    0x9ef53f660df40ULL,
    0x46dde0d00e60ULL,
    0xeab53f398adb0ULL,
    0xde8dde8948008ULL,
    0xc7c53f2b64004ULL,
    0xd4a5deb832002ULL,
    0x38913f634500dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x6800000000000ULL,
    0x8400000000000ULL,
    0x9a00000000000ULL,
    0xc700000000000ULL,
    0x4c80000000000ULL,
    0x3740000000000ULL,
    0xd460000000000ULL,
    0xe570000000000ULL,
    0x64c8000000000ULL,
    0x83ac000000000ULL,
    0x7856000000000ULL,
    0x38db000000000ULL,
    0xe192800000000ULL,
    0x60fa400000000ULL,
    0xba08a00000000ULL,
    0x5703f00000000ULL,
    0x748f080000000ULL,
    0xeb47140000000ULL,
    0xa266560000000ULL,
    0x3c770f0000000ULL,
    0x754c208000000ULL,
    0x3f62b0c000000ULL,
    0xd7f5286000000ULL,
    0x3e8ea43000000ULL,
    0xb449fe6800000ULL,
    0x62efeb8400000ULL,
    0x253b7e1a00000ULL,
    0x9425ab0700000ULL,
    0xd59bdeac80000ULL,
    0x52fa5bc740000ULL,
    0x990ad65c60000ULL,
    0xb18a4f9170000ULL,
    0xecc80076c8000ULL,
    0xf7ac0030ac000ULL,
    0xea5600a6d6000ULL,
    0x4bdb007c9b000ULL,
    0x5f12808b72800ULL,
    0x14ba40f1ca400ULL,
    0xb8e8a0dc20a00ULL,
    0x4233f0c19ff00ULL,
    0x88a708f0f1080ULL,
    0xbadb148200140ULL,
    0x6a98568b0ad60ULL,
    0x62700f028a4f0ULL,
    0x8840a0b248008ULL,
    0xe4eff073ec00cULL,
    0xf43908e0b6006ULL,
    0x31ac14eaeb003ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x3800000000000ULL,
    0xc400000000000ULL,
    0xf600000000000ULL,
    0x9b00000000000ULL,
    0xf780000000000ULL,
    0xaa40000000000ULL,
    0x6660000000000ULL,
    0x3290000000000ULL,
    0x42c8000000000ULL,
    0xadac000000000ULL,
    0x15ba000000000ULL,
    0x9b5000000000ULL,
    0x8bb7800000000ULL,
    0xaeb2400000000ULL,
    0xee3c600000000ULL,
    0xb9f9300000000ULL,
    0xdcd3b80000000ULL,
    0x12a9040000000ULL,
    0xa434760000000ULL,
    0x70fa410000000ULL,
    0x3651e08000000ULL,
    0x3c6e70c000000ULL,
    0x639058a000000ULL,
    0xbc4e74d000000ULL,
    0x4d61ae3800000ULL,
    0x8d1f75c400000ULL,
    0xc481ae7600000ULL,
    0xc1cf755b00000ULL,
    0x8e29aed780000ULL,
    0xd1f375ba40000ULL,
    0xe0dbae7e60000ULL,
    0x80aa75e690000ULL,
    0x19362eacc8000ULL,
    0x247d35e2ac000ULL,
    0xaf95ce0c3a000ULL,
    0x64a45ecf5000ULL,
    0xac6816f457800ULL,
    0x3b93717962400ULL,
    0x84a58d314600ULL,
    0x536b74c385300ULL,
    0xea1e2e7441b80ULL,
    0x10135a160040ULL,
    0x687cea711f60ULL,
    0x84c3457d81010ULL,
    0xbead964348008ULL,
    0x6e38319cec00cULL,
    0x79fbb8045a00aULL,
    0x7cd504516500dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xd800000000000ULL,
    0xf400000000000ULL,
    0xea00000000000ULL,
    0xbd00000000000ULL,
    0x5b80000000000ULL,
    0xddc0000000000ULL,
    0x4d20000000000ULL,
    0xfd50000000000ULL,
    0xbf48000000000ULL,
    0x396c000000000ULL,
    0x84ba000000000ULL,
    0x4eb7000000000ULL,
    0xc3b9800000000ULL,
    0xe031400000000ULL,
    0x39ffa00000000ULL,
    0x46d7500000000ULL,
    0xf286780000000ULL,
    0xfc48bc0000000ULL,
    0xa5e6620000000ULL,
    0x37f2dd0000000ULL,
    0x59dc208000000ULL,
    0x680110c000000ULL,
    0x4c0858a000000ULL,
    0x4e02acf000000ULL,
    0x7f05bad800000ULL,
    0xca8a31f400000ULL,
    0x984dba6a00000ULL,
    0xc7e6317d00000ULL,
    0x56f7ba7b80000ULL,
    0xc45131edc0000ULL,
    0x16ce3ab520000ULL,
    0xda0713950000ULL,
    0x88119aad48000ULL,
    0xb2721406c000ULL,
    0x9e5fe2cd3a000ULL,
    0x13c39dea77000ULL,
    0xf22380c719800ULL,
    0x97d640b9a1400ULL,
    0xd70e200217a00ULL,
    0x268a10e62b500ULL,
    0x2643d8edd4780ULL,
    0x60e8ece203bc0ULL,
    0x68799aa10de20ULL,
    0xb61b213d889d0ULL,
    0x6c2de20ec8008ULL,
    0xd89d59ac00cULL,
    0x918000121a00aULL,
    0x50c000ae2700fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xf800000000000ULL,
    0xec00000000000ULL,
    0xfa00000000000ULL,
    0xeb00000000000ULL,
    0xc280000000000ULL,
    0x5c40000000000ULL,
    0x37e0000000000ULL,
    0x40f0000000000ULL,
    0xd7e8000000000ULL,
    0x50f4000000000ULL,
    0x7fea000000000ULL,
    0x34f3000000000ULL,
    0x91e5800000000ULL,
    0xc9fdc00000000ULL,
    0x426a200000000ULL,
    0xbc33700000000ULL,
    0xeb46080000000ULL,
    0x9766b40000000ULL,
    0x4bbd760000000ULL,
    0x508ebb0000000ULL,
    0xb346088000000ULL,
    0x4b66b44000000ULL,
    0x69bd762000000ULL,
    0x278ebb7000000ULL,
    0x73c6087800000ULL,
    0x1026b4ac00000ULL,
    0x66dd76da00000ULL,
    0xd03ebb9b00000ULL,
    0x514e08ba80000ULL,
    0x5c62b4f040000ULL,
    0xf93f76ede0000ULL,
    0xf4c9bbdbf0000ULL,
    0x68a9886d68000ULL,
    0xf19874a0b4000ULL,
    0x555ad6920a000ULL,
    0xb5f40bef03000ULL,
    0x506020fc8d800ULL,
    0x1330706949c00ULL,
    0xb7cb88d060200ULL,
    0xce2f745330700ULL,
    0x43dd5697cb880ULL,
    0x9fbecbbe2f740ULL,
    0x268d803bdd560ULL,
    0xf249c033becb0ULL,
    0xeae020fc8d808ULL,
    0xe370706949c04ULL,
    0x5a2b88d060202ULL,
    0x15df745330707ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x9c00000000000ULL,
    0xe600000000000ULL,
    0x5900000000000ULL,
    0x7c80000000000ULL,
    0x5e40000000000ULL,
    0x35e0000000000ULL,
    0xea30000000000ULL,
    0x1e8000000000ULL,
    0x503c000000000ULL,
    0x1ee2000000000ULL,
    0x65bd000000000ULL,
    0x1420800000000ULL,
    0x9214c00000000ULL,
    0x2dfee00000000ULL,
    0x20c9500000000ULL,
    0xad280000000ULL,
    0x3052b40000000ULL,
    0x4e11760000000ULL,
    0xabf1970000000ULL,
    0xc9cd288000000ULL,
    0x4422b4c000000ULL,
    0xda1976a000000ULL,
    0x1fd971000000ULL,
    0xfec728a800000ULL,
    0xeda3b45c00000ULL,
    0x36dbf64600000ULL,
    0xaf54574900000ULL,
    0xbb9948d480000ULL,
    0x13e240240000ULL,
    0x2e683e73e0000ULL,
    0x55ffb3a330000ULL,
    0xf4cd16d568000ULL,
    0xcaa107527c000ULL,
    0xb75660ed02000ULL,
    0xdf9190068d000ULL,
    0xf339c8e148800ULL,
    0x1d6ae41068c00ULL,
    0xbe76dec8fce00ULL,
    0x1306e36a44500ULL,
    0x3b883e0fe5a80ULL,
    0x6fcfb3353a740ULL,
    0x7d2516146d960ULL,
    0x169d078af5c70ULL,
    0x67b460e148808ULL,
    0x7f2c901068c0cULL,
    0x7d9948c8fce0aULL,
    0x883e246a44501ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x3800000000000ULL,
    0xbc00000000000ULL,
    0x7600000000000ULL,
    0x9d00000000000ULL,
    0x2680000000000ULL,
    0x6f40000000000ULL,
    0x5760000000000ULL,
    0xd6f0000000000ULL,
    0x3368000000000ULL,
    0x2cf4000000000ULL,
    0x906a000000000ULL,
    0x4371000000000ULL,
    0x2ba9800000000ULL,
    0x2c5ac00000000ULL,
    0x67bee00000000ULL,
    0x1d0b100000000ULL,
    0x6680880000000ULL,
    0x4f43e40000000ULL,
    0x766fe0000000ULL,
    0xeefa0b0000000ULL,
    0x8f60888000000ULL,
    0x5af3e44000000ULL,
    0xd6efe2000000ULL,
    0x65fe0b5000000ULL,
    0x44e288b800000ULL,
    0x7b36e4fc00000ULL,
    0xb14d7e5600000ULL,
    0x2e65cbcd00000ULL,
    0x4a7de89e80000ULL,
    0xdf23349340000ULL,
    0x4411160160000ULL,
    0xc5583f1bf0000ULL,
    0xa3301eade8000ULL,
    0x3d451bbfb4000ULL,
    0x106a00110a000ULL,
    0x371001881000ULL,
    0xba9802641800ULL,
    0x7c5ac083eec00ULL,
    0x5fbee0eeb4e00ULL,
    0xa10b10a98a100ULL,
    0x1080882ec1080ULL,
    0xd243e4bdad240ULL,
    0x21e6fe81521e0ULL,
    0x81ba0b09301b0ULL,
    0xd800882641808ULL,
    0x8c03e483eec04ULL,
    0x3e06feeeb4e02ULL,
    0x490a0ba98a105ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x5400000000000ULL,
    0x2600000000000ULL,
    0x4b00000000000ULL,
    0x580000000000ULL,
    0x95c0000000000ULL,
    0xc4e0000000000ULL,
    0x2df0000000000ULL,
    0x12e8000000000ULL,
    0x4ef4000000000ULL,
    0x2b6e000000000ULL,
    0xa131000000000ULL,
    0x8e8b800000000ULL,
    0xb044400000000ULL,
    0x6121e00000000ULL,
    0xe114f00000000ULL,
    0x331fb80000000ULL,
    0x2e14ac0000000ULL,
    0x894760000000ULL,
    0xc4d0330000000ULL,
    0xbff7b88000000ULL,
    0x6fe0ac4000000ULL,
    0x387a766000000ULL,
    0xbf21335000000ULL,
    0x8e1c38d800000ULL,
    0x7894ec1400000ULL,
    0xacd3964600000ULL,
    0xe3f1c31b00000ULL,
    0x45e580dd80000ULL,
    0x1754081c0000ULL,
    0xd7aa6082e0000ULL,
    0x5550b036f0000ULL,
    0x2c3e584f68000ULL,
    0xd0005c8f34000ULL,
    0x180bcec98e000ULL,
    0x34049fc7c1000ULL,
    0x7603ce19e3800ULL,
    0x13009f2b70400ULL,
    0x5185ce6eafe00ULL,
    0xb3c59f7dd5f00ULL,
    0x8fe04e177c380ULL,
    0x2870df94a4ec0ULL,
    0x872a2edcdb960ULL,
    0x8a106f8bf5c30ULL,
    0x69c7699e3808ULL,
    0xb3d4336b70404ULL,
    0xc071b80eafe06ULL,
    0x9b25ac2dd5f05ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0x3400000000000ULL,
    0x5200000000000ULL,
    0x4500000000000ULL,
    0x6380000000000ULL,
    0x23c0000000000ULL,
    0xdc20000000000ULL,
    0xc8f0000000000ULL,
    0x628000000000ULL,
    0xd1fc000000000ULL,
    0x3ba2000000000ULL,
    0xe13f000000000ULL,
    0xb0b800000000ULL,
    0x188c400000000ULL,
    0xc346a00000000ULL,
    0xbb64100000000ULL,
    0x899a980000000ULL,
    0x66bd2c0000000ULL,
    0xb2cada0000000ULL,
    0xbdace30000000ULL,
    0xa632988000000ULL,
    0x25812cc000000ULL,
    0xc4c8daa000000ULL,
    0xc2a3e33000000ULL,
    0x8cb1181800000ULL,
    0x33c16cf400000ULL,
    0xb4247af200000ULL,
    0xc4f4f37500000ULL,
    0x502a007b80000ULL,
    0x5ef300d7c0000ULL,
    0x2921802e20000ULL,
    0xf37f40bdf0000ULL,
    0xb1e720fda8000ULL,
    0x6fdb50c63c000ULL,
    0x2e5db8b582000ULL,
    0xb0967c6ccf000ULL,
    0xab3f626ea3800ULL,
    0xba069feab0400ULL,
    0x890ffaa4c4a00ULL,
    0x9588b352ab100ULL,
    0x9cc4a0a4b9180ULL,
    0x56ab105fcd6c0ULL,
    0xeeb918722e7a0ULL,
    0xeecd6ce3f7f30ULL,
    0xe3ae7aeea3808ULL,
    0xb537f32ab040cULL,
    0xc9038004c4a0aULL,
    0xf5804062ab103ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x9800000000000ULL,
    0xac00000000000ULL,
    0x3200000000000ULL,
    0x2900000000000ULL,
    0x1480000000000ULL,
    0x2c0000000000ULL,
    0xcde0000000000ULL,
    0xadf0000000000ULL,
    0xb3e8000000000ULL,
    0xa6fc000000000ULL,
    0xc662000000000ULL,
    0xf4bb000000000ULL,
    0xf4b800000000ULL,
    0x23a1c00000000ULL,
    0x1ed0a00000000ULL,
    0x9f7500000000ULL,
    0xa5efe80000000ULL,
    0x59fe7c0000000ULL,
    0x8debb60000000ULL,
    0xcdfae30000000ULL,
    0x63e7e88000000ULL,
    0x4ef27cc000000ULL,
    0xf261b6a000000ULL,
    0x6abde37000000ULL,
    0x144e681800000ULL,
    0x1e28bc6c00000ULL,
    0x89a969200000ULL,
    0xc6db735900000ULL,
    0xc5f9200c80000ULL,
    0x47ed906ec0000ULL,
    0x98f4c85fe0000ULL,
    0xad68ecf4f0000ULL,
    0x5134fe3f68000ULL,
    0xb503cf083c000ULL,
    0xde8bb6b982000ULL,
    0x57cae3b04b000ULL,
    0x36fe88823800ULL,
    0xb23e7c379dc00ULL,
    0xf48bb62d52a00ULL,
    0x12cae38cbc500ULL,
    0x85efe8334c680ULL,
    0xe9fe7c99a3bc0ULL,
    0xb5ebb6d3d9160ULL,
    0x11fae36b76b30ULL,
    0xc9e7e80823808ULL,
    0xcbf27cf79dc0cULL,
    0xd4e1b68d52a0aULL,
    0x417de3fcbc507ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0xb400000000000ULL,
    0xde00000000000ULL,
    0x6100000000000ULL,
    0xca80000000000ULL,
    0x63c0000000000ULL,
    0x3920000000000ULL,
    0xc930000000000ULL,
    0x6b28000000000ULL,
    0x923c000000000ULL,
    0xa6aa000000000ULL,
    0x7e75000000000ULL,
    0x70c3800000000ULL,
    0x38ae400000000ULL,
    0xff76600000000ULL,
    0xa4c500000000ULL,
    0x5364c80000000ULL,
    0xea596c0000000ULL,
    0xa973da0000000ULL,
    0x87413f0000000ULL,
    0xd3ecc88000000ULL,
    0xa6956cc000000ULL,
    0xa3d1da2000000ULL,
    0x53383f9000000ULL,
    0x3c2d481800000ULL,
    0xa5b22c7400000ULL,
    0x6dee3afe00000ULL,
    0xb79f2ff100000ULL,
    0x415460d280000ULL,
    0x8cf55017c0000ULL,
    0xf70548c720000ULL,
    0x678e2c3830000ULL,
    0x73443a39a8000ULL,
    0xedea2f45fc000ULL,
    0x7797e0c18a000ULL,
    0x615b101645000ULL,
    0x1cf328f16b800ULL,
    0x6f027c5952400ULL,
    0xd380f2f8fc600ULL,
    0xad43430909500ULL,
    0x8cec3a168f480ULL,
    0xbd162fe1cb2c0ULL,
    0x29de09a2fba0ULL,
    0x25de10c0b86f0ULL,
    0xa638a8716b808ULL,
    0xb8a03c995240cULL,
    0x3f7492d8fc602ULL,
    0x2a46139909509ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0x3c00000000000ULL,
    0x8200000000000ULL,
    0x7900000000000ULL,
    0x1d80000000000ULL,
    0xf640000000000ULL,
    0x67a0000000000ULL,
    0x8d30000000000ULL,
    0x82a8000000000ULL,
    0x22b4000000000ULL,
    0xc5ee000000000ULL,
    0xcc95000000000ULL,
    0xb895800000000ULL,
    0xee9ec00000000ULL,
    0xf199a00000000ULL,
    0x3b1f500000000ULL,
    0x93d9780000000ULL,
    0x5abbb40000000ULL,
    0xd9ecc20000000ULL,
    0x3e93470000000ULL,
    0x6997788000000ULL,
    0x771eb44000000ULL,
    0xb9d1426000000ULL,
    0xa7b9871000000ULL,
    0x7a60d83800000ULL,
    0x33d4e47c00000ULL,
    0x6abdbae200000ULL,
    0x11ecf36900000ULL,
    0xaa9dba2580000ULL,
    0x6f9cf38a40000ULL,
    0xb015ba05a0000ULL,
    0x5f58f3a430000ULL,
    0x3573bac728000ULL,
    0xf609f3b8f4000ULL,
    0x2f003a784e000ULL,
    0x2863354a5000ULL,
    0x3cca1afdbd800ULL,
    0xcf66a32f6ac00ULL,
    0x4451429457a00ULL,
    0x1f98799fa500ULL,
    0xc5c0d809c4f80ULL,
    0x92e4e4f8e1740ULL,
    0xd215ba4f13620ULL,
    0x7658f3c5dd170ULL,
    0xf0f3bac5bd808ULL,
    0x2c49f3536ac04ULL,
    0x72a03a7657a06ULL,
    0xcab633f0fa501ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0x3400000000000ULL,
    0xd600000000000ULL,
    0xe500000000000ULL,
    0x2580000000000ULL,
    0x89c0000000000ULL,
    0xa260000000000ULL,
    0x93b0000000000ULL,
    0xe368000000000ULL,
    0xd834000000000ULL,
    0x53ae000000000ULL,
    0xcdd3000000000ULL,
    0xf2d6800000000ULL,
    0x7858400000000ULL,
    0xc31fe00000000ULL,
    0x630100000000ULL,
    0xc2aab80000000ULL,
    0x5e580c0000000ULL,
    0xee17de0000000ULL,
    0x77b2570000000ULL,
    0xed64b88000000ULL,
    0x913b0c4000000ULL,
    0x6c295e6000000ULL,
    0xcb1e177000000ULL,
    0x7235580800000ULL,
    0x74a81c7400000ULL,
    0xcb5d66b600000ULL,
    0x439a5b9500000ULL,
    0xca7b662d80000ULL,
    0x990d5bfdc0000ULL,
    0xe78be69460000ULL,
    0xaac21b46b0000ULL,
    0x9ae486aee8000ULL,
    0x83fd4b55f4000ULL,
    0x2ac15e4fce000ULL,
    0xdaea17bf63000ULL,
    0xe3fb588a3e800ULL,
    0x5acb1cc8ac400ULL,
    0x52e3e62951e00ULL,
    0xd7f61b7093100ULL,
    0x8cca868af4380ULL,
    0xb7ee4b75444c0ULL,
    0xf277de2c2e3e0ULL,
    0x50257ab15470ULL,
    0x158cb8823e808ULL,
    0x61cf0cbcac404ULL,
    0xe6675e9f51e06ULL,
    0xcdbd17e593107ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xa800000000000ULL,
    0xe400000000000ULL,
    0x2e00000000000ULL,
    0xc700000000000ULL,
    0xfe80000000000ULL,
    0x22c0000000000ULL,
    0xfae0000000000ULL,
    0xf230000000000ULL,
    0x7be8000000000ULL,
    0xcfb4000000000ULL,
    0x79a2000000000ULL,
    0x2891000000000ULL,
    0xc3f8800000000ULL,
    0x4987400000000ULL,
    0x6446600000000ULL,
    0x342d300000000ULL,
    0x92d2880000000ULL,
    0xc8d29c0000000ULL,
    0x69d0e60000000ULL,
    0x44559f0000000ULL,
    0x4e18888000000ULL,
    0x1d379c4000000ULL,
    0x216a66a000000ULL,
    0x6373df5000000ULL,
    0x144e682800000ULL,
    0x8c29eca400000ULL,
    0x7edc8e8e00000ULL,
    0x52dd739700000ULL,
    0x28d486d680000ULL,
    0x99d9af86c0000ULL,
    0xbc5a80f4e0000ULL,
    0x216402530000ULL,
    0xd73ee08d68000ULL,
    0xc86a705974000ULL,
    0x5af4e80542000ULL,
    0xc80facf9a1000ULL,
    0x540a6e6890800ULL,
    0x76030363f3400ULL,
    0xdb026e9984600ULL,
    0x9c87038c4c300ULL,
    0x2fc86ef022080ULL,
    0xed6203acd1dc0ULL,
    0xe972ee87dc860ULL,
    0x5d4443635daf0ULL,
    0xe5a40ec090808ULL,
    0xa9a3387f3404ULL,
    0x6ef2e6b78460aULL,
    0xe049f4b4c305ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x9800000000000ULL,
    0xd400000000000ULL,
    0x1e00000000000ULL,
    0xe500000000000ULL,
    0x6b80000000000ULL,
    0xb940000000000ULL,
    0x56e0000000000ULL,
    0x3370000000000ULL,
    0xed68000000000ULL,
    0x423c000000000ULL,
    0x678a000000000ULL,
    0x9347000000000ULL,
    0x3deb800000000ULL,
    0x95f6400000000ULL,
    0x73a0600000000ULL,
    0x679f100000000ULL,
    0xf91d580000000ULL,
    0xc3dac40000000ULL,
    0x407c560000000ULL,
    0x5fe8e30000000ULL,
    0xe2f6d88000000ULL,
    0x172c84c000000ULL,
    0xfe5c362000000ULL,
    0x37f3b000000ULL,
    0xa08b801800000ULL,
    0x6fc6401400000ULL,
    0x7028603e00000ULL,
    0x72d3105500000ULL,
    0xf5ff58f380000ULL,
    0x23a1c46d40000ULL,
    0x6f9dd648e0000ULL,
    0x519a3d670000ULL,
    0x91dd3806e8000ULL,
    0x6f75d43b7c000ULL,
    0xcf690e116a000ULL,
    0xd53e271037000ULL,
    0x93088ec883800ULL,
    0x228f67c3ca400ULL,
    0x88c36eaa2a600ULL,
    0x3ca63761d8100ULL,
    0xa71e569776d80ULL,
    0xc6d3e34b6c840ULL,
    0xbbf758433c360ULL,
    0xceadc44a07f30ULL,
    0xf81fd63b03808ULL,
    0xee52a3ae8a40cULL,
    0xe834b8e2ca602ULL,
    0xcc8894b7a810bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xe800000000000ULL,
    0xc00000000000ULL,
    0x2200000000000ULL,
    0xa300000000000ULL,
    0xd080000000000ULL,
    0x8bc0000000000ULL,
    0x44a0000000000ULL,
    0x13f0000000000ULL,
    0xc28000000000ULL,
    0xac3c000000000ULL,
    0x6e86000000000ULL,
    0x22cd000000000ULL,
    0xdb20800000000ULL,
    0x7ab1c00000000ULL,
    0x8844a00000000ULL,
    0x27e6f00000000ULL,
    0xd891280000000ULL,
    0x657b8c0000000ULL,
    0xdae3c20000000ULL,
    0xb11b430000000ULL,
    0x9931a88000000ULL,
    0xfb0a4cc000000ULL,
    0x48762e000000ULL,
    0xbdcdb31000000ULL,
    0x31a8806800000ULL,
    0xa67dc0cc00000ULL,
    0xda6aa0c200000ULL,
    0x32d7f0b300000ULL,
    0xc197a83880000ULL,
    0x22f74c87c0000ULL,
    0xa7afe266a0000ULL,
    0xe37073b0f0000ULL,
    0x77e2205ca8000ULL,
    0x109a30e7fc000ULL,
    0x997508ca26000ULL,
    0xecbc213d000ULL,
    0xf6164a3f08800ULL,
    0x67b63fda8dc00ULL,
    0xf1cb4244c2a00ULL,
    0x33a683662bf00ULL,
    0xf57b08b331a80ULL,
    0xf2edbc440a4c0ULL,
    0x5d18ca1e07620ULL,
    0xab36ff990db30ULL,
    0xb009620788808ULL,
    0xd80cb35d4dc0cULL,
    0x1406002262a0eULL,
    0xd60d00d6dbf01ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x8800000000000ULL,
    0x2400000000000ULL,
    0x5e00000000000ULL,
    0x8500000000000ULL,
    0xb680000000000ULL,
    0x1c0000000000ULL,
    0xf220000000000ULL,
    0x42f0000000000ULL,
    0x6aa8000000000ULL,
    0x5e3c000000000ULL,
    0xa20a000000000ULL,
    0xb701000000000ULL,
    0x698a800000000ULL,
    0xbc4f400000000ULL,
    0x486d600000000ULL,
    0x239a100000000ULL,
    0x85b6880000000ULL,
    0x14c4c0000000ULL,
    0x62ea4a0000000ULL,
    0x205b330000000ULL,
    0xd094088000000ULL,
    0x823f0cc000000ULL,
    0x400d2a2000000ULL,
    0xe00023d000000ULL,
    0xf008000800000ULL,
    0x580c00e400000ULL,
    0xac02007e00000ULL,
    0x7a0d005500000ULL,
    0xdb08803e80000ULL,
    0x33824025c0000ULL,
    0xb745e0ac20000ULL,
    0xf3e850c7f0000ULL,
    0xb0db68dc28000ULL,
    0x28581c5ffc000ULL,
    0x349b22502a000ULL,
    0xfc322ff5f1000ULL,
    0x150daa8322800ULL,
    0xde8e632273400ULL,
    0xd5cfe04a67600ULL,
    0xf42950849b100ULL,
    0x6bf1e8443c080ULL,
    0xa6275c49030c0ULL,
    0x84fe42fc872a0ULL,
    0x63a43fa2c1230ULL,
    0x42b9223da2808ULL,
    0xf0cf2fc7b340cULL,
    0x52ad2ac647602ULL,
    0xc23023936b10dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x9800000000000ULL,
    0x9400000000000ULL,
    0x6e00000000000ULL,
    0xa300000000000ULL,
    0x9680000000000ULL,
    0x3e40000000000ULL,
    0x5620000000000ULL,
    0x9170000000000ULL,
    0xbba8000000000ULL,
    0xcdbc000000000ULL,
    0x6dc2000000000ULL,
    0x41eb000000000ULL,
    0xdd93800000000ULL,
    0xf3b6400000000ULL,
    0xe6cb600000000ULL,
    0x7b65700000000ULL,
    0x69d2880000000ULL,
    0xf09f540000000ULL,
    0xd638ca0000000ULL,
    0x60b530000000ULL,
    0x6f03088000000ULL,
    0xcc8214c000000ULL,
    0x83402aa000000ULL,
    0xfba8637000000ULL,
    0xadb2e01800000ULL,
    0xbdc4305400000ULL,
    0xa9e868ce00000ULL,
    0xd19764d300000ULL,
    0x9baa20e80000ULL,
    0x2bcb37aa40000ULL,
    0x4ee82ab820000ULL,
    0xc11463f270000ULL,
    0x98f0e00d28000ULL,
    0x116f3043fc000ULL,
    0x2cdbe803e2000ULL,
    0x191124349b000ULL,
    0x6cf9c2903b800ULL,
    0xaf6247090a400ULL,
    0x67d0a2f389600ULL,
    0x839c3767ce700ULL,
    0xa8b9aa54e1080ULL,
    0x3449231c19140ULL,
    0xc32800657baa0ULL,
    0x90fc0005a2230ULL,
    0xd620086bb808ULL,
    0x9edb00f74a40cULL,
    0xa81b8085a960aULL,
    0x3b7a4046be707ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0x5c00000000000ULL,
    0x7a00000000000ULL,
    0xcd00000000000ULL,
    0x8080000000000ULL,
    0x5e40000000000ULL,
    0x4ce0000000000ULL,
    0x5b10000000000ULL,
    0x5b68000000000ULL,
    0x58dc000000000ULL,
    0x3142000000000ULL,
    0x9d61000000000ULL,
    0x5fdf800000000ULL,
    0x54c0c00000000ULL,
    0x97ae200000000ULL,
    0x3373100000000ULL,
    0xa83aa80000000ULL,
    0xa058b40000000ULL,
    0x2307060000000ULL,
    0x538cf50000000ULL,
    0x45cf288000000ULL,
    0xad2574c000000ULL,
    0x34a6a000000ULL,
    0xb45e25d000000ULL,
    0xfd0420d800000ULL,
    0x688e109c00000ULL,
    0x2a4728da00000ULL,
    0xe2e9741d00000ULL,
    0xe81ea6d880000ULL,
    0x30e3250240000ULL,
    0x3119a0b6e0000ULL,
    0x6e6fd05610000ULL,
    0x5456887be8000ULL,
    0x4d0aa4d69c000ULL,
    0x40822e25a2000ULL,
    0xfe44819a71000ULL,
    0x9cee0efeb7800ULL,
    0x31691011cc00ULL,
    0x76b26066c200ULL,
    0x22dee5e052100ULL,
    0xfc4a00c305280ULL,
    0x1ded00e388740ULL,
    0x195806dc1260ULL,
    0x182dc07923e50ULL,
    0xccbba07e37808ULL,
    0x681ed05f5cc0cULL,
    0xf0e108ca8c20aULL,
    0x9116647b4210dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0x9400000000000ULL,
    0x1a00000000000ULL,
    0x4b00000000000ULL,
    0xa580000000000ULL,
    0x7bc0000000000ULL,
    0x4860000000000ULL,
    0xd270000000000ULL,
    0xe4e8000000000ULL,
    0xa33c000000000ULL,
    0x60c2000000000ULL,
    0x35ed000000000ULL,
    0x8db1800000000ULL,
    0x1e80400000000ULL,
    0x1646200000000ULL,
    0xafa7f00000000ULL,
    0xbc1af80000000ULL,
    0x4b908c0000000ULL,
    0x27539e0000000ULL,
    0x4cb53b0000000ULL,
    0x8801788000000ULL,
    0xfc01ccc000000ULL,
    0x96063ea000000ULL,
    0xd50f8b1000000ULL,
    0x4c84203800000ULL,
    0x14af05400000ULL,
    0x8c2b78ba00000ULL,
    0xaad0cc5b00000ULL,
    0xdb75be1d80000ULL,
    0xee62cbefc0000ULL,
    0x6f7380d260000ULL,
    0x246d405970000ULL,
    0x3c77a0e168000ULL,
    0x5e7b0c8fc000ULL,
    0xe5bcd890a2000ULL,
    0x92877c739d000ULL,
    0x8841667359800ULL,
    0x46a9b7f6bc400ULL,
    0xc698e65304200ULL,
    0x8fd5f7218af00ULL,
    0x5ffcc659cb780ULL,
    0x732f075760cc0ULL,
    0xc55fbe4dfdbe0ULL,
    0xf3b3cb042ecb0ULL,
    0x878000d6d9808ULL,
    0x64c0008d7c40cULL,
    0x57e0009b6420aULL,
    0xf2b00033faf01ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0x5c00000000000ULL,
    0x6200000000000ULL,
    0x8100000000000ULL,
    0xed80000000000ULL,
    0x87c0000000000ULL,
    0x6e60000000000ULL,
    0x9e30000000000ULL,
    0x58e8000000000ULL,
    0xf174000000000ULL,
    0x4aca000000000ULL,
    0xa9ed000000000ULL,
    0xe4f2800000000ULL,
    0x2904c00000000ULL,
    0x6182a00000000ULL,
    0xcdcfd00000000ULL,
    0x2361f80000000ULL,
    0x19b0ec0000000ULL,
    0x225fe0000000ULL,
    0x1093bf0000000ULL,
    0xd231788000000ULL,
    0x72ed2c4000000ULL,
    0xc7fde2000000ULL,
    0x5545af9000000ULL,
    0xa728a08800000ULL,
    0x3b12d01c00000ULL,
    0x52fb784200000ULL,
    0xd6002c1100000ULL,
    0x4f0d5ee580000ULL,
    0xb6816fdbc0000ULL,
    0x2f4a008c60000ULL,
    0x322d005f30000ULL,
    0xc892809568000ULL,
    0xa634c0e6b4000ULL,
    0xdceaa02caa000ULL,
    0xe77bd06bdd000ULL,
    0x65cbf8de1a800ULL,
    0xaf6dec5970c00ULL,
    0x53bf7ec6c8a00ULL,
    0x4f237fe3e2d00ULL,
    0x9719d829f3780ULL,
    0x88fffcee842c0ULL,
    0x9304a61b4f5e0ULL,
    0x948583bc286f0ULL,
    0x8e45fe339a808ULL,
    0x4fa3bfdeb0c04ULL,
    0x47597828a8a02ULL,
    0x94592c3dd2d09ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xb800000000000ULL,
    0xec00000000000ULL,
    0x6600000000000ULL,
    0x2300000000000ULL,
    0xab80000000000ULL,
    0xadc0000000000ULL,
    0x1b60000000000ULL,
    0x1710000000000ULL,
    0x9e8000000000ULL,
    0xc5c000000000ULL,
    0xabca000000000ULL,
    0x8869000000000ULL,
    0x7499800000000ULL,
    0xa023c00000000ULL,
    0x2531e00000000ULL,
    0x15d6f00000000ULL,
    0x6f0dd80000000ULL,
    0x5d8d6c0000000ULL,
    0x16c2ce0000000ULL,
    0xce34d0000000ULL,
    0x64d6588000000ULL,
    0xdd8bacc000000ULL,
    0xd6c8ae2000000ULL,
    0x2ce37d5000000ULL,
    0x34d9e03800000ULL,
    0x658af02c00000ULL,
    0x3ac7d84600000ULL,
    0x4ae46c7300000ULL,
    0x17db4e1380000ULL,
    0xce008d41c0000ULL,
    0x9707b8fd60000ULL,
    0x518d5cf410000ULL,
    0xcd768268000ULL,
    0xc7e211f19c000ULL,
    0x9b592e88aa000ULL,
    0xfa4cbdb379000ULL,
    0x88aa00bb71800ULL,
    0xb379001f7fc00ULL,
    0x3b7180bd7be00ULL,
    0xdf7fc08c7ff00ULL,
    0x9d7be0def4580ULL,
    0xdc7ff025beac0ULL,
    0x66f458f79b2e0ULL,
    0xc9beac9ba9bd0ULL,
    0x919b2e10f1808ULL,
    0xb8a9bdb2bfc0cULL,
    0xbb7180261be02ULL,
    0x1f7fc05b6ff05ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xd800000000000ULL,
    0x4400000000000ULL,
    0x7e00000000000ULL,
    0x6100000000000ULL,
    0x2380000000000ULL,
    0x7ec0000000000ULL,
    0xf120000000000ULL,
    0x2650000000000ULL,
    0x1fe8000000000ULL,
    0xcf7c000000000ULL,
    0x1dbe000000000ULL,
    0xbe95000000000ULL,
    0x6bcf800000000ULL,
    0xc4af400000000ULL,
    0xd59d600000000ULL,
    0x3f46d00000000ULL,
    0x14e0180000000ULL,
    0x8bf45c0000000ULL,
    0x15f5da0000000ULL,
    0xa4f8210000000ULL,
    0x3f7d788000000ULL,
    0x15b28cc000000ULL,
    0x4295c26000000ULL,
    0x41cc7d1000000ULL,
    0x3a8a25800000ULL,
    0xd31aad8400000ULL,
    0x1c00ba9e00000ULL,
    0xfa02f1b100000ULL,
    0x7f03601b80000ULL,
    0x5283d0eac0000ULL,
    0x8547983720000ULL,
    0xcbe71c1350000ULL,
    0xa976ba9a68000ULL,
    0x58bbf194bc000ULL,
    0xf31ae0311e000ULL,
    0xac05904705000ULL,
    0x520b78c687800ULL,
    0xf30b8c4343400ULL,
    0xb08c42feeb600ULL,
    0xde4a3d2cffd00ULL,
    0xf764428379980ULL,
    0x95363d9fb21c0ULL,
    0xf4da4275993a0ULL,
    0x5aa33d4f44b10ULL,
    0x6495c25cef808ULL,
    0xa4cc7dd7ff40cULL,
    0x3e28a2cff5606ULL,
    0xdcdaad6bfad01ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0x7c00000000000ULL,
    0x1600000000000ULL,
    0x6d00000000000ULL,
    0x7b80000000000ULL,
    0x6640000000000ULL,
    0xe320000000000ULL,
    0x87d0000000000ULL,
    0xf368000000000ULL,
    0xf9f4000000000ULL,
    0x1732000000000ULL,
    0xa8dd000000000ULL,
    0xcfec800000000ULL,
    0x3f38c00000000ULL,
    0x94d8e00000000ULL,
    0x79e6900000000ULL,
    0x4235980000000ULL,
    0x4755140000000ULL,
    0x63a9ba0000000ULL,
    0xb71c310000000ULL,
    0xad85788000000ULL,
    0xeb47844000000ULL,
    0x28ae22a000000ULL,
    0x5994251000000ULL,
    0xc440422800000ULL,
    0x1423753c00000ULL,
    0x9f53ba3600000ULL,
    0x47a5313d00000ULL,
    0xc513f8f380000ULL,
    0x1286444a40000ULL,
    0x3fc042fd20000ULL,
    0x32637586d0000ULL,
    0xdc73bab6e8000ULL,
    0xd07531ceb4000ULL,
    0x9e7bf8b992000ULL,
    0x977244744d000ULL,
    0x3ef242ac24800ULL,
    0xf7be754b5cc00ULL,
    0x681f3aada2e00ULL,
    0x890df1fe1f900ULL,
    0xe983182403180ULL,
    0x6944d47204d40ULL,
    0x8fafdabf0fda0ULL,
    0x491f61d48f610ULL,
    0x1c84809acc808ULL,
    0x96ccc0c5e8c04ULL,
    0x8beae0b430e0aULL,
    0xbd3b909a52901ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0x2400000000000ULL,
    0xe00000000000ULL,
    0xd100000000000ULL,
    0x9680000000000ULL,
    0x5c0000000000ULL,
    0x1260000000000ULL,
    0x69b0000000000ULL,
    0x41a8000000000ULL,
    0x3ed4000000000ULL,
    0x689e000000000ULL,
    0x7fb3000000000ULL,
    0x64a1800000000ULL,
    0x8e5f400000000ULL,
    0xd05a600000000ULL,
    0x6957d00000000ULL,
    0xf3da480000000ULL,
    0x941aec0000000ULL,
    0xc57b3e0000000ULL,
    0xc14e7f0000000ULL,
    0xcca8288000000ULL,
    0xa2593c4000000ULL,
    0x225f766000000ULL,
    0x625793f000000ULL,
    0x25a963800000ULL,
    0xf25c036400000ULL,
    0x4a533eee00000ULL,
    0x6e5a7f6100000ULL,
    0x605628ce80000ULL,
    0xb15a3c91c0000ULL,
    0x27d6f6c460000ULL,
    0x221cd36cb0000ULL,
    0x307ef6e128000ULL,
    0x59c8d38e14000ULL,
    0x1860f6027e000ULL,
    0x26bbd372c3000ULL,
    0x4e2176f9e9800ULL,
    0x31949348fb400ULL,
    0x553316bd0c600ULL,
    0xdb67430480d00ULL,
    0xb3f5e7ecdc80ULL,
    0x626aaf1be2ac0ULL,
    0x91b3e0cbfede0ULL,
    0x5ac907085ef0ULL,
    0xc0d62898c1808ULL,
    0x19a3c86ef404ULL,
    0xcd36f6df72606ULL,
    0x6f6cd38643d0fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0x4c00000000000ULL,
    0x4a00000000000ULL,
    0xe500000000000ULL,
    0x7c80000000000ULL,
    0x4540000000000ULL,
    0x8b60000000000ULL,
    0x9050000000000ULL,
    0x7c28000000000ULL,
    0xda34000000000ULL,
    0x76f6000000000ULL,
    0x1c59000000000ULL,
    0x9620800000000ULL,
    0x8f3bc00000000ULL,
    0x7275200000000ULL,
    0x3d11100000000ULL,
    0x1b47280000000ULL,
    0xb065640000000ULL,
    0x97dc0e0000000ULL,
    0xdeef890000000ULL,
    0xa39a088000000ULL,
    0xb700744000000ULL,
    0xed8d26e000000ULL,
    0xf7c3ed5000000ULL,
    0x7d2e86a800000ULL,
    0xe0b03d0c00000ULL,
    0x7cbc0e2a00000ULL,
    0x5ebf89f500000ULL,
    0x17b2083480000ULL,
    0x7134741940000ULL,
    0xf97b260960000ULL,
    0x429aed6950000ULL,
    0xdd8e06e2a8000ULL,
    0xcfcbfd7674000ULL,
    0xf9292eab16000ULL,
    0xb6be993c49000ULL,
    0xfbbd2055e8800ULL,
    0x8b3510dc1fc00ULL,
    0x64792871cb200ULL,
    0x5a1864922c100ULL,
    0x9ec28e9939a80ULL,
    0xeba9491573a40ULL,
    0xf0f1a8389fae0ULL,
    0xf357a4008c590ULL,
    0xefa1ae3740808ULL,
    0xe6f159ea6bc04ULL,
    0x945e003add20eULL,
    0x6a2d00fe65105ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0xfc00000000000ULL,
    0xa00000000000ULL,
    0xf900000000000ULL,
    0x8e80000000000ULL,
    0x6bc0000000000ULL,
    0x2ae0000000000ULL,
    0x9a30000000000ULL,
    0xe428000000000ULL,
    0x2454000000000ULL,
    0xe4da000000000ULL,
    0xc819000000000ULL,
    0x137c800000000ULL,
    0xe704c00000000ULL,
    0xe181200000000ULL,
    0x6645d00000000ULL,
    0x4ea0080000000ULL,
    0x8194cc0000000ULL,
    0xc932b60000000ULL,
    0x83a4c70000000ULL,
    0xc91b288000000ULL,
    0x9f81c4000000ULL,
    0x5ac63e2000000ULL,
    0x6860cb5000000ULL,
    0xa3f2bea800000ULL,
    0xb3c00bbc00000ULL,
    0xeee19eaa00000ULL,
    0x6438dbe900000ULL,
    0xbb2f160680000ULL,
    0x71d5d787c0000ULL,
    0x49200a8e0000ULL,
    0xdbd008f30000ULL,
    0x316e8068a8000ULL,
    0x3d79c01a94000ULL,
    0xd00fa062ba000ULL,
    0x680c103ce9000ULL,
    0xdc07a85934800ULL,
    0x5a0cdccba0c00ULL,
    0xd10f1e4513200ULL,
    0x72811b7bf8d00ULL,
    0x61c0b6f7ce880ULL,
    0xd3e9c750ed0c0ULL,
    0x14bda89b3d160ULL,
    0x8fe5dcfea8d70ULL,
    0xebb9e999c808ULL,
    0x7ee1db6d34c04ULL,
    0x2c33968da9202ULL,
    0x372117ae11d05ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0xb400000000000ULL,
    0x4600000000000ULL,
    0x4b00000000000ULL,
    0xfc80000000000ULL,
    0x8640000000000ULL,
    0x3a0000000000ULL,
    0xb670000000000ULL,
    0x52e8000000000ULL,
    0x7354000000000ULL,
    0xc3da000000000ULL,
    0xd797000000000ULL,
    0x493d800000000ULL,
    0xc0ce400000000ULL,
    0x64eae00000000ULL,
    0x6055700000000ULL,
    0x1b50680000000ULL,
    0x9fdc740000000ULL,
    0x8d9a620000000ULL,
    0x78391b0000000ULL,
    0xbd40888000000ULL,
    0xa72e044000000ULL,
    0x143f8a2000000ULL,
    0x9f4f2fb000000ULL,
    0x2220ab800000ULL,
    0x9fb16ff400000ULL,
    0x2f80eae600000ULL,
    0x3ec01fbb00000ULL,
    0xdbe2826480000ULL,
    0xfadf6bc240000ULL,
    0x661f60dda0000ULL,
    0x58ff30b970000ULL,
    0x9a888f068000ULL,
    0x6f7a04ba14000ULL,
    0x33658a42fa000ULL,
    0xca982f18a7000ULL,
    0x36bf8a22f5800ULL,
    0x160f2f88aa400ULL,
    0xa3020aaaf8e00ULL,
    0xe0816f04a6700ULL,
    0xfc48ea58ffe80ULL,
    0x82a41f09a1340ULL,
    0xf3f082ef7f020ULL,
    0x252c6b73612b0ULL,
    0x4130e06a9d808ULL,
    0xcc270c6be404ULL,
    0xb6ede80e02e02ULL,
    0xdd5234a70170bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x5800000000000ULL,
    0x9c00000000000ULL,
    0x1e00000000000ULL,
    0x3f00000000000ULL,
    0x8380000000000ULL,
    0x240000000000ULL,
    0x18e0000000000ULL,
    0x1bd0000000000ULL,
    0xe128000000000ULL,
    0xad7c000000000ULL,
    0x412000000000ULL,
    0x48b000000000ULL,
    0x8dc3800000000ULL,
    0xcacc00000000ULL,
    0x83b600000000ULL,
    0xa377b00000000ULL,
    0xf314180000000ULL,
    0x130c340000000ULL,
    0xa583460000000ULL,
    0xf149210000000ULL,
    0x2d6cf88000000ULL,
    0xe29744c000000ULL,
    0xf84c3ea000000ULL,
    0x95e2a5f000000ULL,
    0x4153a6d800000ULL,
    0x33ee515c00000ULL,
    0x7251803e00000ULL,
    0xa667c00f00000ULL,
    0x7b18e07b80000ULL,
    0xe70b706e40000ULL,
    0xdf87785ee0000ULL,
    0xbc4784b8d0000ULL,
    0xd7e55e7ca8000ULL,
    0xc05e15903c000ULL,
    0x7f643e1f72000ULL,
    0xab9ea5dd1b000ULL,
    0x20c1a6540b800ULL,
    0x6625518a00c00ULL,
    0x22f2009501600ULL,
    0x105b00a680b00ULL,
    0x176b809cc5980ULL,
    0xcf90c0482bf40ULL,
    0x52c960e5fba60ULL,
    0x1f2cb0ffd2510ULL,
    0x27f985323808ULL,
    0x8f9cf4747cc0cULL,
    0x32ca26d49360aULL,
    0x4f2591c34bb0fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xd800000000000ULL,
    0xfc00000000000ULL,
    0x6a00000000000ULL,
    0xab00000000000ULL,
    0x7180000000000ULL,
    0xfc0000000000ULL,
    0x8320000000000ULL,
    0x33b0000000000ULL,
    0x9568000000000ULL,
    0x5b5c000000000ULL,
    0xd3fe000000000ULL,
    0xc87000000000ULL,
    0x1f47800000000ULL,
    0x66ac00000000ULL,
    0x26d8200000000ULL,
    0xe30f00000000ULL,
    0xe8aa780000000ULL,
    0xef76ec0000000ULL,
    0xcbc89a0000000ULL,
    0xe5265f0000000ULL,
    0x4ab5d88000000ULL,
    0x83ecdcc000000ULL,
    0x971ac26000000ULL,
    0x810433000000ULL,
    0xa79f3a5800000ULL,
    0x1d506f3c00000ULL,
    0x3af1808a00000ULL,
    0x9201c05b00000ULL,
    0xc709a0c980000ULL,
    0x938130c3c0000ULL,
    0xb0cbd83120000ULL,
    0xbcabdc64b0000ULL,
    0x17d428ee8000ULL,
    0xa6ca83ff9c000ULL,
    0xbdaf1aa15e000ULL,
    0xbbfc9ff0f7000ULL,
    0xe885f8e90f800ULL,
    0xe9402c9e86c00ULL,
    0xf76ebad84e200ULL,
    0x2451af95ebf00ULL,
    0x8c78201613f80ULL,
    0xe40f0729b2c0ULL,
    0xf4e27889d73a0ULL,
    0x1c9aec79bc6f0ULL,
    0x24de9aae67808ULL,
    0x413d5fa2dac0cULL,
    0x6f2c584830206ULL,
    0x11b11c01acf03ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x3800000000000ULL,
    0x2c00000000000ULL,
    0xe600000000000ULL,
    0xb700000000000ULL,
    0x5980000000000ULL,
    0xea40000000000ULL,
    0xdf20000000000ULL,
    0x17d0000000000ULL,
    0x4ae8000000000ULL,
    0xb5b4000000000ULL,
    0x6b9e000000000ULL,
    0xeac1000000000ULL,
    0x5fe1800000000ULL,
    0xe939c00000000ULL,
    0xc558e00000000ULL,
    0x726300000000ULL,
    0xabdb380000000ULL,
    0xf4e8b40000000ULL,
    0xfeb99a0000000ULL,
    0xec15010000000ULL,
    0x9b8a588000000ULL,
    0x3f43444000000ULL,
    0xa3a4426000000ULL,
    0x91a85d000000ULL,
    0x4f097ab800000ULL,
    0x5587316c00000ULL,
    0xbc4f600600000ULL,
    0x802af02700000ULL,
    0x5a5c580180000ULL,
    0x6aa6441640000ULL,
    0x3b93c20120000ULL,
    0x92c6458cd0000ULL,
    0x13e61a7568000ULL,
    0xdf3dc1a8f4000ULL,
    0x4a5b380d3e000ULL,
    0x72a8b48751000ULL,
    0xa7999a9229800ULL,
    0x9cc501b75dc00ULL,
    0xb0e258ba2ee00ULL,
    0x4cb7448353300ULL,
    0xf11a424024b80ULL,
    0x430b857a50740ULL,
    0x380fa5aa0fa0ULL,
    0xe34af1939af10ULL,
    0xcda98066c1808ULL,
    0x7a1dc049e9c04ULL,
    0xe48ee0d630e06ULL,
    0xc2c33058d230dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x6800000000000ULL,
    0x3c00000000000ULL,
    0xd600000000000ULL,
    0x5100000000000ULL,
    0x1a80000000000ULL,
    0x49c0000000000ULL,
    0xdf20000000000ULL,
    0xead0000000000ULL,
    0xfa68000000000ULL,
    0x5d3c000000000ULL,
    0x9d1a000000000ULL,
    0x884f000000000ULL,
    0x5b68800000000ULL,
    0xcfbac00000000ULL,
    0x78d8e00000000ULL,
    0x5964500000000ULL,
    0xb4b1080000000ULL,
    0xc55dec0000000ULL,
    0xeb20ca0000000ULL,
    0x50d0d90000000ULL,
    0x8569688000000ULL,
    0xf2bf7cc000000ULL,
    0x4c53222000000ULL,
    0xa5a665b000000ULL,
    0xf3102ae800000ULL,
    0xdd4889fc00000ULL,
    0x53e2607600000ULL,
    0xe57d902100000ULL,
    0x6e73685280000ULL,
    0x5bf07cc5c0000ULL,
    0x45bba26120000ULL,
    0xafdca587d0000ULL,
    0xeae8cab6e8000ULL,
    0x3fcd985fc000ULL,
    0xd1bb68f8ba000ULL,
    0x65dc7c5b5f000ULL,
    0xdde9a23620800ULL,
    0x207fa5f456c00ULL,
    0xbef24a21aae00ULL,
    0x2535192117500ULL,
    0xb911081e43880ULL,
    0x2a4decea682c0ULL,
    0x5068ca4530aa0ULL,
    0x3a3cd9e91e490ULL,
    0x469b685248808ULL,
    0x30c7c746ac0cULL,
    0x9981a29830e02ULL,
    0x1043a54d9850bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0x1400000000000ULL,
    0xda00000000000ULL,
    0xf300000000000ULL,
    0x180000000000ULL,
    0x7e40000000000ULL,
    0x2860000000000ULL,
    0x55d0000000000ULL,
    0x21a8000000000ULL,
    0x2efc000000000ULL,
    0x4c9e000000000ULL,
    0x1249000000000ULL,
    0x966f800000000ULL,
    0x94da400000000ULL,
    0x4f2d200000000ULL,
    0xc737f00000000ULL,
    0x27b2f80000000ULL,
    0x477c740000000ULL,
    0x6eda2e0000000ULL,
    0xcc2ce10000000ULL,
    0x2eb8588000000ULL,
    0xc5fdc4c000000ULL,
    0xf913f66000000ULL,
    0x60265d000000ULL,
    0xc5090ed800000ULL,
    0x8c8211d400000ULL,
    0xdecd203a00000ULL,
    0xe8a7f0e300000ULL,
    0xcc7af83980000ULL,
    0xb5074ba40000ULL,
    0x806c2eaa60000ULL,
    0x69d9e1b2d0000ULL,
    0x7a9d87a28000ULL,
    0x9bfe8463bc000ULL,
    0x281956057e000ULL,
    0x5083d5e9d9000ULL,
    0xe8c0d647a7800ULL,
    0x65ac953bf6400ULL,
    0x6cfc76181b200ULL,
    0xcb98251882f00ULL,
    0x19c42e44c3780ULL,
    0x4325e1bbaf340ULL,
    0xa937d8fdf88e0ULL,
    0xbeb7843d11510ULL,
    0x3df6d6840f808ULL,
    0x5d1995220a40cULL,
    0x540df6d705206ULL,
    0x7a0b65938bf0dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xa800000000000ULL,
    0xcc00000000000ULL,
    0x2200000000000ULL,
    0xb100000000000ULL,
    0xbe80000000000ULL,
    0xea40000000000ULL,
    0x9de0000000000ULL,
    0xf310000000000ULL,
    0x4828000000000ULL,
    0x523c000000000ULL,
    0x609e000000000ULL,
    0xe6ef000000000ULL,
    0x9090800000000ULL,
    0xeee9c00000000ULL,
    0x8c9fa00000000ULL,
    0xd4e3500000000ULL,
    0x5992080000000ULL,
    0x8463140000000ULL,
    0x20d2e60000000ULL,
    0x160cdd0000000ULL,
    0x8703288000000ULL,
    0x298684c000000ULL,
    0xbbcfce6000000ULL,
    0xf22559b000000ULL,
    0x473c662800000ULL,
    0x501a1d0c00000ULL,
    0x2fa408c200000ULL,
    0xaaf014c100000ULL,
    0x853c667680000ULL,
    0x911a1d9640000ULL,
    0x59240897e0000ULL,
    0x3cb0144e10000ULL,
    0x92dc66b4a8000ULL,
    0x1f0a1db97c000ULL,
    0x8d8c086bfe000ULL,
    0x35cc14f3bf000ULL,
    0xd122660758800ULL,
    0xe0b51d4ec5c00ULL,
    0xc8d48832a9a00ULL,
    0x7a09d48670500ULL,
    0x750bc6767c880ULL,
    0x80854d7e75d40ULL,
    0x614800e275c60ULL,
    0xf66c00987a4d0ULL,
    0x49d600ed70808ULL,
    0xac83006df9c0cULL,
    0x3346800cb7a06ULL,
    0x8f6ac0fadf50bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xd800000000000ULL,
    0x6400000000000ULL,
    0xaa00000000000ULL,
    0xc100000000000ULL,
    0x4780000000000ULL,
    0x8f40000000000ULL,
    0x94a0000000000ULL,
    0xbb30000000000ULL,
    0x4968000000000ULL,
    0xc55c000000000ULL,
    0x34b2000000000ULL,
    0xba2d000000000ULL,
    0x3ffb800000000ULL,
    0x748d400000000ULL,
    0xadc2200000000ULL,
    0x566dd00000000ULL,
    0xf1d1980000000ULL,
    0xf97bc40000000ULL,
    0xdc43620000000ULL,
    0x862b9f0000000ULL,
    0x31fa388000000ULL,
    0xbf8654c000000ULL,
    0xbb435aa000000ULL,
    0x46accb9000000ULL,
    0x8e38e25800000ULL,
    0x7ce6dfa400000ULL,
    0xef18188a00000ULL,
    0x4d9b849100000ULL,
    0x4f5ac23f80000ULL,
    0xa5bb0f7b40000ULL,
    0x85a18066a0000ULL,
    0x44bc40de30000ULL,
    0x922ba004e8000ULL,
    0xb3fd901b1c000ULL,
    0xf280383f92000ULL,
    0x7ac754ea5d000ULL,
    0x1ee2da4833800ULL,
    0x52108bcbe1400ULL,
    0x6413421798200ULL,
    0x3b1b4f665cd00ULL,
    0x6f98204e38180ULL,
    0x925cd0dceb840ULL,
    0xbc38187f12c20ULL,
    0xb9eb8495970f0ULL,
    0x3292c22b5b808ULL,
    0x4bd70f0fbd40cULL,
    0x207b80c4aa20aULL,
    0x5fcd40c331d09ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x6800000000000ULL,
    0x2400000000000ULL,
    0xa200000000000ULL,
    0x1500000000000ULL,
    0x4f80000000000ULL,
    0xee40000000000ULL,
    0xb60000000000ULL,
    0x6d70000000000ULL,
    0xb3a8000000000ULL,
    0x21dc000000000ULL,
    0x9cf2000000000ULL,
    0x76ef000000000ULL,
    0x8e30800000000ULL,
    0x1b4b400000000ULL,
    0x94eba00000000ULL,
    0xfb3f900000000ULL,
    0x44cf980000000ULL,
    0xa2af140000000ULL,
    0xbc5d5e0000000ULL,
    0xafb5eb0000000ULL,
    0xa606b88000000ULL,
    0xc704c4c000000ULL,
    0x9281e6a000000ULL,
    0x35c22fb000000ULL,
    0x2f25dee800000ULL,
    0x5712abe400000ULL,
    0x3357188200000ULL,
    0x6938546500000ULL,
    0xf9c4fe8780000ULL,
    0x69257b7a40000ULL,
    0x4019a04160000ULL,
    0xd9d0909c70000ULL,
    0xa0ff18fe28000ULL,
    0x38e4546a9c000ULL,
    0xad36feb012000ULL,
    0x8bca7bd1df000ULL,
    0x4292014f8800ULL,
    0xf39bd082e7400ULL,
    0xd994b85431a00ULL,
    0x389bc4f24c900ULL,
    0xad1966356d180ULL,
    0xaa556f867b540ULL,
    0xbc7e372e7e0ULL,
    0x10823bbb1d3b0ULL,
    0x50c8008550808ULL,
    0xa8ac00763b40cULL,
    0x2d5a002743a0aULL,
    0xf23300dae390bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xd800000000000ULL,
    0xf400000000000ULL,
    0x1600000000000ULL,
    0xfd00000000000ULL,
    0xc380000000000ULL,
    0x5340000000000ULL,
    0x8a20000000000ULL,
    0x27b0000000000ULL,
    0x5be8000000000ULL,
    0x1754000000000ULL,
    0x323e000000000ULL,
    0xd5af000000000ULL,
    0xeff800000000ULL,
    0xaec2400000000ULL,
    0x286be00000000ULL,
    0xdc1b100000000ULL,
    0x2c19180000000ULL,
    0xd41a040000000ULL,
    0x701e0a0000000ULL,
    0x8e13170000000ULL,
    0x5f13788000000ULL,
    0x7e9c544000000ULL,
    0xc6db726000000ULL,
    0x7274433000000ULL,
    0xc5098a5800000ULL,
    0x478557b400000ULL,
    0xfd4698f600000ULL,
    0xb328448d00000ULL,
    0x2a3dea7b80000ULL,
    0x1ac079740000ULL,
    0x48fc60c420000ULL,
    0xbbcd506eb0000ULL,
    0xc7ecf8ee68000ULL,
    0x6d5e148914000ULL,
    0x4d3092a39e000ULL,
    0xcd2f53555f000ULL,
    0xbb30924937800ULL,
    0x402f532326400ULL,
    0xc0b092623de00ULL,
    0xd76f533da0100ULL,
    0x849092a2f8980ULL,
    0xf9df530cc7440ULL,
    0xaf89223626a0ULL,
    0x40cb53929e470ULL,
    0x71669284df808ULL,
    0xe194538972404ULL,
    0xae5112f383e06ULL,
    0x7fb2138b4f103ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x4800000000000ULL,
    0xe400000000000ULL,
    0xbe00000000000ULL,
    0x5100000000000ULL,
    0xd180000000000ULL,
    0xb7c0000000000ULL,
    0x1f60000000000ULL,
    0x4750000000000ULL,
    0x1628000000000ULL,
    0xaf7c000000000ULL,
    0x485e000000000ULL,
    0xa6ab000000000ULL,
    0xe13e800000000ULL,
    0xf4ff400000000ULL,
    0x894600000000ULL,
    0x54dd00000000ULL,
    0x222db80000000ULL,
    0x9780c0000000ULL,
    0x25543e0000000ULL,
    0x9527310000000ULL,
    0x15f1588000000ULL,
    0x311d9cc000000ULL,
    0xc68d666000000ULL,
    0x3b46adf000000ULL,
    0xb3223ec800000ULL,
    0xb8f0312400000ULL,
    0x6291d85e00000ULL,
    0xc249dc6100000ULL,
    0xb6a786f980000ULL,
    0x99343da3c0000ULL,
    0x38fbe66960000ULL,
    0xa295ed3250000ULL,
    0xa2405e99a8000ULL,
    0x46aae179bc000ULL,
    0xd13ce02ebe000ULL,
    0xdcf590823b000ULL,
    0x1c91587e76800ULL,
    0xf34d9c59d3400ULL,
    0x9725667162600ULL,
    0x66faad0e5ad00ULL,
    0xc39c3e7bad380ULL,
    0x5bcb313abc4c0ULL,
    0xe56758f436de0ULL,
    0x385a9c0b79a10ULL,
    0x2ea5e6565e808ULL,
    0x653eed67af40cULL,
    0xbafede08bc606ULL,
    0x1195a12f31d0fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xf800000000000ULL,
    0x9c00000000000ULL,
    0x7600000000000ULL,
    0xd00000000000ULL,
    0x9f80000000000ULL,
    0xc9c0000000000ULL,
    0x9b20000000000ULL,
    0x9810000000000ULL,
    0x7618000000000ULL,
    0x971c000000000ULL,
    0xb696000000000ULL,
    0x865f000000000ULL,
    0x90f0800000000ULL,
    0x2629400000000ULL,
    0x2f96200000000ULL,
    0x7bdaf00000000ULL,
    0x763d080000000ULL,
    0xad82940000000ULL,
    0xfec5a60000000ULL,
    0x27ae670000000ULL,
    0xcf5da68000000ULL,
    0x3572674000000ULL,
    0xcceba62000000ULL,
    0x2a3d67d000000ULL,
    0x3b8326d800000ULL,
    0xc3c8274c00000ULL,
    0xd02306ae00000ULL,
    0x629dd74100000ULL,
    0x44568e3180000ULL,
    0x2ffa0388c0000ULL,
    0xceab08aaa0000ULL,
    0xfadd9410d0000ULL,
    0x27b5265cb8000ULL,
    0x554727c7cc000ULL,
    0x1c6b86ca2e000ULL,
    0x8378979193000ULL,
    0x21eeae02de800ULL,
    0x85b3f3bbba400ULL,
    0x9a4880a348a00ULL,
    0x3ce5405160b00ULL,
    0xa23820bcf5a80ULL,
    0x6f89f07822240ULL,
    0x41c388e6900e0ULL,
    0xcf28d49e5c430ULL,
    0x1a15063cf5a88ULL,
    0x6912d73822244ULL,
    0xef9e0ec6900eaULL,
    0x9bdf434e5c439ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xd800000000000ULL,
    0x3400000000000ULL,
    0xd600000000000ULL,
    0x8d00000000000ULL,
    0xd380000000000ULL,
    0xea40000000000ULL,
    0xfd60000000000ULL,
    0xca10000000000ULL,
    0x7418000000000ULL,
    0x8514000000000ULL,
    0xc92000000000ULL,
    0xd9d7000000000ULL,
    0x743a800000000ULL,
    0x3ee9c00000000ULL,
    0x585ca00000000ULL,
    0x8d70700000000ULL,
    0x1107480000000ULL,
    0x1982ec0000000ULL,
    0x6d49ba0000000ULL,
    0xc9eddb0000000ULL,
    0x44d1ba8000000ULL,
    0xdfb9dbc000000ULL,
    0xd0a3bae000000ULL,
    0xfb3edb9000000ULL,
    0x26613a3800000ULL,
    0x54931ba400000ULL,
    0x2dd79aee00000ULL,
    0xc2306b2900000ULL,
    0xe3e0523d80000ULL,
    0x53d847c340000ULL,
    0x533dc8c0e0000ULL,
    0x3a6b2c0950000ULL,
    0x5e951a34f8000ULL,
    0xcaddab4c44000ULL,
    0xa6b6f2d86a000ULL,
    0xb52b370593000ULL,
    0xcc72001450800ULL,
    0xd087005f7ac00ULL,
    0x40c280c20c200ULL,
    0x72adc0ab0ab00ULL,
    0x36a0b68b680ULL,
    0x48e37085c85c0ULL,
    0xe557c82d22d20ULL,
    0xd6f82c5875870ULL,
    0x17459a368b688ULL,
    0x6e76b45c85ccULL,
    0xfc5ad2cd22d26ULL,
    0xe37187c875875ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x4800000000000ULL,
    0x6400000000000ULL,
    0xfa00000000000ULL,
    0x1700000000000ULL,
    0xdd80000000000ULL,
    0xd0c0000000000ULL,
    0x9ba0000000000ULL,
    0x34f0000000000ULL,
    0xebf8000000000ULL,
    0x9274000000000ULL,
    0xd8b2000000000ULL,
    0xc41b000000000ULL,
    0x656f800000000ULL,
    0x3a51c00000000ULL,
    0xc98a600000000ULL,
    0x32c9100000000ULL,
    0xc0a7c80000000ULL,
    0xe771440000000ULL,
    0x9e3b3e0000000ULL,
    0xa759110000000ULL,
    0x1b093e8000000ULL,
    0x738211c000000ULL,
    0x85c6bee000000ULL,
    0xed23d15000000ULL,
    0xefb4dea800000ULL,
    0x299ec13400000ULL,
    0xda1165200000ULL,
    0x1df4852300000ULL,
    0x2b75a88f80000ULL,
    0x503c54f3c0000ULL,
    0x6256f69420000ULL,
    0x2587550730000ULL,
    0xcccf809fd8000ULL,
    0xbda1c0c544000ULL,
    0x65f260ef6a000ULL,
    0xf77d10355f000ULL,
    0x1635c85805800ULL,
    0xa35a44ec0ec00ULL,
    0x710cbefe0fe00ULL,
    0x2c8cd17d07d00ULL,
    0x3c495e0288280ULL,
    0xc7e401a946940ULL,
    0x631cf6516b160ULL,
    0xc0e855e85b850ULL,
    0x369200aa88288ULL,
    0x142b009d4694cULL,
    0xc73780036b166ULL,
    0x7fd5c0cb5b859ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0x4400000000000ULL,
    0xa600000000000ULL,
    0x6b00000000000ULL,
    0xcd80000000000ULL,
    0x4740000000000ULL,
    0x69e0000000000ULL,
    0x6230000000000ULL,
    0xa138000000000ULL,
    0x78bc000000000ULL,
    0x1f2000000000ULL,
    0x3f1d000000000ULL,
    0xfaa5800000000ULL,
    0x95d9c00000000ULL,
    0x840fa00000000ULL,
    0x60b100000000ULL,
    0x7b02c80000000ULL,
    0xb58f3c0000000ULL,
    0x8b4a620000000ULL,
    0x8beae10000000ULL,
    0xaf38628000000ULL,
    0x7b7e14000000ULL,
    0xf27de26000000ULL,
    0x2f5e21b000000ULL,
    0x34ca42e800000ULL,
    0x39a931f400000ULL,
    0x4c5a8a4e00000ULL,
    0xfd4b0d9f00000ULL,
    0x38ed688380000ULL,
    0xbeb42cd840000ULL,
    0xdaf0aa6a60000ULL,
    0x9a99ddfa70000ULL,
    0x9e000ab58000ULL,
    0xd2300032cc000ULL,
    0xc9380042aa000ULL,
    0xccbc00f9d1000ULL,
    0x2ff200760f800ULL,
    0x101d00b308c00ULL,
    0x9125801180200ULL,
    0xb999c0dd43d00ULL,
    0x206fa068e2e80ULL,
    0x237b1026bcec0ULL,
    0xb3dac8e6f08a0ULL,
    0xaf033cf09a0d0ULL,
    0x2b806280e2e88ULL,
    0xcc4be1d2bcec4ULL,
    0x546fe2a8f08aeULL,
    0xad73216f9a0dfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xe800000000000ULL,
    0xf400000000000ULL,
    0x4a00000000000ULL,
    0x5100000000000ULL,
    0xae80000000000ULL,
    0xb5c0000000000ULL,
    0xb5a0000000000ULL,
    0x7a50000000000ULL,
    0x6758000000000ULL,
    0xdfdc000000000ULL,
    0xe51a000000000ULL,
    0x3537000000000ULL,
    0xed29800000000ULL,
    0xd09ec00000000ULL,
    0x3f76600000000ULL,
    0x23c1700000000ULL,
    0x7aa8980000000ULL,
    0x7fd4440000000ULL,
    0x1510de0000000ULL,
    0x6d359b0000000ULL,
    0x212ade8000000ULL,
    0x86929b4000000ULL,
    0xd07b5ee000000ULL,
    0x96405b9000000ULL,
    0x30ef3e0800000ULL,
    0xd13a2b6400000ULL,
    0x6f2c264200000ULL,
    0xc59baf3500000ULL,
    0xe3f918ec80000ULL,
    0xdb068480c0000ULL,
    0xdf84bed920000ULL,
    0xab4febba90000ULL,
    0x3869c65e78000ULL,
    0xd3f31ff54c000ULL,
    0xa30e603362000ULL,
    0xa38d70e47b000ULL,
    0xc54a98fc4b800ULL,
    0xcb6f44d1e5c00ULL,
    0xd87b5e47bde00ULL,
    0xf2405bc6e4b00ULL,
    0xf2ef3e4e35780ULL,
    0xa43a2bc2a0f40ULL,
    0x63ac26a3dda60ULL,
    0xd55baf8b196f0ULL,
    0xb2d9184635788ULL,
    0x459684a6a0f44ULL,
    0xa3fcbee1dda66ULL,
    0xbb03ebbe196fdULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xf800000000000ULL,
    0x3400000000000ULL,
    0x7600000000000ULL,
    0x4d00000000000ULL,
    0x8880000000000ULL,
    0x93c0000000000ULL,
    0x9e20000000000ULL,
    0x5fd0000000000ULL,
    0x40d8000000000ULL,
    0x9b54000000000ULL,
    0x511a000000000ULL,
    0xde71000000000ULL,
    0x7b4e800000000ULL,
    0x7de5c00000000ULL,
    0xd9fea00000000ULL,
    0x9b0c700000000ULL,
    0x558bf80000000ULL,
    0xb34f440000000ULL,
    0x71e0e60000000ULL,
    0x7bf6fb0000000ULL,
    0x1002e68000000ULL,
    0x6803fbc000000ULL,
    0x9c0e666000000ULL,
    0xa033b7000000ULL,
    0xf704469800000ULL,
    0x878b8b4400000ULL,
    0x20479eee00000ULL,
    0xc8697f0900000ULL,
    0xdab0206680000ULL,
    0x12e9b09ac0000ULL,
    0x1a755878a0000ULL,
    0xd543340510000ULL,
    0x54eb1ed878000ULL,
    0x6f79bf2e44000ULL,
    0x89c200f162000ULL,
    0x125000435000ULL,
    0x4454809c2c800ULL,
    0x6a94c044d0c00ULL,
    0xa43020d552200ULL,
    0xc29b0c81cb00ULL,
    0xecd55890f9d80ULL,
    0xa953345d83f40ULL,
    0x72131e5f413e0ULL,
    0x9ffdbf63e10f0ULL,
    0xee000088f9d88ULL,
    0x90000d983f4cULL,
    0xe68000d1413eeULL,
    0x5ac0001ae10fbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0x4400000000000ULL,
    0x3200000000000ULL,
    0x1900000000000ULL,
    0x5d80000000000ULL,
    0x5240000000000ULL,
    0x2b60000000000ULL,
    0x5c10000000000ULL,
    0xd998000000000ULL,
    0x7dc000000000ULL,
    0xcab2000000000ULL,
    0xa5a5000000000ULL,
    0x14bb800000000ULL,
    0x6aa3c00000000ULL,
    0xf232e00000000ULL,
    0x6466700000000ULL,
    0xfa90280000000ULL,
    0x3752cc0000000ULL,
    0x7f7c9a0000000ULL,
    0xf1802b0000000ULL,
    0x64471a8000000ULL,
    0xe063eb4000000ULL,
    0x6895fa6000000ULL,
    0xbe559b3000000ULL,
    0x3afdd20800000ULL,
    0x8fcb577400000ULL,
    0x392b483a00000ULL,
    0x97727c6d00000ULL,
    0xf585d2e780000ULL,
    0xb647577f40000ULL,
    0x896148ace0000ULL,
    0x5d1b7c1350000ULL,
    0xa814527d78000ULL,
    0x239d97608c000ULL,
    0xcada288dca000ULL,
    0xbd3bcca829000ULL,
    0xc2ed1afef1800ULL,
    0x145aebfdcac00ULL,
    0x4ffc7a4023600ULL,
    0x44435bfafcb00ULL,
    0x306cb2afcb480ULL,
    0x9092e7e9227c0ULL,
    0xe25b80ef7dd20ULL,
    0x24f3c0e98b570ULL,
    0xe0cae0484b488ULL,
    0x4faa7096627c4ULL,
    0x81ba28439dd2eULL,
    0xd12bccfadb577ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x2800000000000ULL,
    0xcc00000000000ULL,
    0x4a00000000000ULL,
    0x6300000000000ULL,
    0x6e80000000000ULL,
    0x2640000000000ULL,
    0x3ee0000000000ULL,
    0xa170000000000ULL,
    0x4bf8000000000ULL,
    0x5bb4000000000ULL,
    0x845a000000000ULL,
    0x94ad000000000ULL,
    0xf597800000000ULL,
    0x198f400000000ULL,
    0x26c1e00000000ULL,
    0x1da2d00000000ULL,
    0xa81fb80000000ULL,
    0xe94c9c0000000ULL,
    0xaa61920000000ULL,
    0x2c37d50000000ULL,
    0xa796128000000ULL,
    0xbe8895c000000ULL,
    0x5e4ff26000000ULL,
    0x4aee45b000000ULL,
    0x7f724a4800000ULL,
    0x86fbd97c00000ULL,
    0xd03e580200000ULL,
    0xe59e4c1f00000ULL,
    0x81862aec80000ULL,
    0x22cf49f940000ULL,
    0xebad80b260000ULL,
    0xa91240e830000ULL,
    0x28ce60b198000ULL,
    0xa8a990cf84000ULL,
    0x579c58b7c2000ULL,
    0x56874c8429000ULL,
    0x724baa4ed5800ULL,
    0x70ed0914e6400ULL,
    0x347be0f274600ULL,
    0x247fd02d74900ULL,
    0xbc703821f3d80ULL,
    0xb877dca8bc0c0ULL,
    0x4e7a72b2d04a0ULL,
    0x4f780556e2d90ULL,
    0x8efe2acd73d88ULL,
    0x8c3b4951fc0ccULL,
    0xf7978000b04aeULL,
    0x68f40bed2d97ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0x6c00000000000ULL,
    0x1600000000000ULL,
    0x4500000000000ULL,
    0x5a80000000000ULL,
    0x77c0000000000ULL,
    0x93e0000000000ULL,
    0xc9d0000000000ULL,
    0x558000000000ULL,
    0xf794000000000ULL,
    0x1ef6000000000ULL,
    0x70eb000000000ULL,
    0x7e5a800000000ULL,
    0x2c18400000000ULL,
    0x21b7200000000ULL,
    0xe3cc700000000ULL,
    0xd9e5580000000ULL,
    0xf2d1640000000ULL,
    0x3ed5fa0000000ULL,
    0x38dd570000000ULL,
    0xc5d77a8000000ULL,
    0x235117c000000ULL,
    0xda965aa000000ULL,
    0x6076671000000ULL,
    0xe529823800000ULL,
    0xd2bf437c00000ULL,
    0xec4b582e00000ULL,
    0x4cae643900000ULL,
    0xaaf97af480000ULL,
    0xeaee178ec0000ULL,
    0xdd5adac760000ULL,
    0x7b95275710000ULL,
    0xf8fc227a38000ULL,
    0x7def731c84000ULL,
    0xd0d720eace000ULL,
    0x21dc70456f000ULL,
    0xe15d585814800ULL,
    0x3595649bb7400ULL,
    0xf1fbfa90c3a00ULL,
    0xe16257166b300ULL,
    0x7a1bfac79ef80ULL,
    0x84b25776fe540ULL,
    0x4943fad4e5020ULL,
    0xe626575c5c030ULL,
    0x9535fab31ef88ULL,
    0x8d0d57383e54cULL,
    0x6e8f7ab385022ULL,
    0x2dc5171b4c03dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xb800000000000ULL,
    0x5400000000000ULL,
    0xce00000000000ULL,
    0x9300000000000ULL,
    0xa180000000000ULL,
    0xe340000000000ULL,
    0x7660000000000ULL,
    0x13b0000000000ULL,
    0xc838000000000ULL,
    0x6674000000000ULL,
    0xfa92000000000ULL,
    0x2fef000000000ULL,
    0x4fc800000000ULL,
    0xc05dc00000000ULL,
    0x8ecb200000000ULL,
    0x7f21100000000ULL,
    0x30d1880000000ULL,
    0xd50f7c0000000ULL,
    0xce88ba0000000ULL,
    0xcdd10000000ULL,
    0x4c2c3a8000000ULL,
    0x215411c000000ULL,
    0x5e4d1ae000000ULL,
    0x54ee011000000ULL,
    0x8972125800000ULL,
    0xd913bd4400000ULL,
    0x75ad089600000ULL,
    0xe612bcd700000ULL,
    0xee239ab780000ULL,
    0xe85cc1f440000ULL,
    0xc2c5b221e0000ULL,
    0x552f6df7f0000ULL,
    0x5d7a031d8000ULL,
    0xb8cd01584000ULL,
    0x1642a83d4a000ULL,
    0xc8ea6c3d6b000ULL,
    0x1b73323636800ULL,
    0xb819ad5d76c00ULL,
    0x652a00571da00ULL,
    0xaddb0046a7d00ULL,
    0x878e80f794280ULL,
    0x5c42c0656cac0ULL,
    0x2defa07236920ULL,
    0x7df8d04b7a7d0ULL,
    0x34d0a8c014288ULL,
    0x63056c512caccULL,
    0x698fb2b3d6926ULL,
    0xbf446dac8a7ddULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0x400000000000ULL,
    0x3e00000000000ULL,
    0x3b00000000000ULL,
    0x2e80000000000ULL,
    0x8040000000000ULL,
    0x5a0000000000ULL,
    0x8df0000000000ULL,
    0x5178000000000ULL,
    0x8834000000000ULL,
    0xb216000000000ULL,
    0x1c25000000000ULL,
    0x6136800000000ULL,
    0xf592c00000000ULL,
    0x1bef200000000ULL,
    0x4355900000000ULL,
    0x6a0bf80000000ULL,
    0x6d0e3c0000000ULL,
    0xb980c60000000ULL,
    0x1cc2c90000000ULL,
    0x3cee468000000ULL,
    0xa7d409c000000ULL,
    0x774f66a000000ULL,
    0x492099f000000ULL,
    0x1cbc1e1800000ULL,
    0xf75d65f400000ULL,
    0x2c0b782600000ULL,
    0x7209fccf00000ULL,
    0x9901668880000ULL,
    0x1f81998f40000ULL,
    0x13c49e2d20000ULL,
    0x946ea5f2b0000ULL,
    0x189cd8e458000ULL,
    0x626fac4e84000ULL,
    0x7f9d3e504e000ULL,
    0x66e9f5ada1000ULL,
    0x72d80081f8800ULL,
    0xcac4001373c00ULL,
    0x6bee00a937a00ULL,
    0x1b5100a996500ULL,
    0x7e008031e4580ULL,
    0xdb07c0ce5c6c0ULL,
    0x3e81a0c38a9e0ULL,
    0x84350f9cfa50ULL,
    0xb9aad83964588ULL,
    0xb7faac811c6ccULL,
    0x5473be4eaa9e2ULL,
    0x9dbf35fb7fa53ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x3800000000000ULL,
    0xac00000000000ULL,
    0xda00000000000ULL,
    0x3b00000000000ULL,
    0x7380000000000ULL,
    0xa40000000000ULL,
    0xbd60000000000ULL,
    0x9510000000000ULL,
    0x8d98000000000ULL,
    0xac54000000000ULL,
    0x77e000000000ULL,
    0xdd0f000000000ULL,
    0x6a84800000000ULL,
    0x4ec2400000000ULL,
    0xea5e00000000ULL,
    0x2137500000000ULL,
    0x7a6be80000000ULL,
    0x64958c0000000ULL,
    0xa0ddfa0000000ULL,
    0x90b01b0000000ULL,
    0x37277a8000000ULL,
    0xe87d5bc000000ULL,
    0xb8861a2000000ULL,
    0x31c84bd000000ULL,
    0x9328121800000ULL,
    0x767a977c00000ULL,
    0x6d8600c200000ULL,
    0x1f4b004700000ULL,
    0x33e2803180000ULL,
    0x25d9408d40000ULL,
    0xe63f602ce0000ULL,
    0x8bea100850000ULL,
    0x49d2889978000ULL,
    0x9c349c0804000ULL,
    0xa0edf24406000ULL,
    0x25dc7ee0b000ULL,
    0x3a75e8dd02800ULL,
    0xc78a8c6a89400ULL,
    0xac417a4ec7600ULL,
    0xc4665b0eae100ULL,
    0x619c9aa134880ULL,
    0x16550bba6f9c0ULL,
    0x8c71724497720ULL,
    0x868b8770d0870ULL,
    0xf4ce0888b4888ULL,
    0x85a2dc4b2f9ccULL,
    0x7ab612aa7772aULL,
    0xe425973f80871ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0xe400000000000ULL,
    0x6200000000000ULL,
    0xdf00000000000ULL,
    0x8d80000000000ULL,
    0xfd40000000000ULL,
    0x160000000000ULL,
    0x2bd0000000000ULL,
    0xe558000000000ULL,
    0xd394000000000ULL,
    0xb2b6000000000ULL,
    0x9a83000000000ULL,
    0xdcc4800000000ULL,
    0xf226c00000000ULL,
    0xdbb2e00000000ULL,
    0xe202100000000ULL,
    0x1f01480000000ULL,
    0xad8edc0000000ULL,
    0xad4fea0000000ULL,
    0xf96a270000000ULL,
    0xcfdd6a8000000ULL,
    0x875fe7c000000ULL,
    0xc930aa000000ULL,
    0x3f3f379000000ULL,
    0x67cea25800000ULL,
    0xdda4fb7400000ULL,
    0xd9f2803a00000ULL,
    0x3ee5c0ab00000ULL,
    0x3196603780000ULL,
    0xadb4d09640000ULL,
    0x370ba816e0000ULL,
    0x7188cced90000ULL,
    0xb40a20bb8000ULL,
    0x1463fbda04000ULL,
    0x6558005b0e000ULL,
    0x1394005f87000ULL,
    0x92b600aa4a800ULL,
    0xca8300c0e1c00ULL,
    0x24c480a898600ULL,
    0x1626c0bd33d00ULL,
    0xb9b2e0c8c1280ULL,
    0x3d0210f8290c0ULL,
    0x92814838b8c20ULL,
    0x50cedcb9802b0ULL,
    0xac2feaa741288ULL,
    0xd2ba271a690ccULL,
    0x2a856a1458c22ULL,
    0x54cbe7ff102b5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x9800000000000ULL,
    0xd400000000000ULL,
    0x6e00000000000ULL,
    0x1f00000000000ULL,
    0xf80000000000ULL,
    0x4ac0000000000ULL,
    0x3de0000000000ULL,
    0x2b50000000000ULL,
    0x47d8000000000ULL,
    0x709c000000000ULL,
    0x9ab6000000000ULL,
    0x3687000000000ULL,
    0xf646800000000ULL,
    0x72afc00000000ULL,
    0xa573200000000ULL,
    0xdebd00000000ULL,
    0xe35ea80000000ULL,
    0xabd1440000000ULL,
    0xda9b7a0000000ULL,
    0x73babf0000000ULL,
    0xf20bfa8000000ULL,
    0xdd027f4000000ULL,
    0x1a865a2000000ULL,
    0xbc4a6f5000000ULL,
    0x2ba5d2b800000ULL,
    0xe9fbfb8400000ULL,
    0x6aa880d600000ULL,
    0x3174c09b00000ULL,
    0xc3e3a05980000ULL,
    0xec531091c0000ULL,
    0x3c5308c460000ULL,
    0x445954aa90000ULL,
    0x2058f21bb8000ULL,
    0xc65b2b0e0c000ULL,
    0x955ea86f0e000ULL,
    0x20d144678b000ULL,
    0x1b1b7ab6c8800ULL,
    0x367abf0fe4c00ULL,
    0x586bfa965ba00ULL,
    0x68927fad5f100ULL,
    0xebe5a64dd080ULL,
    0xf8866fad12540ULL,
    0xf94bd2dd70720ULL,
    0xe520fb69efeb0ULL,
    0x3bb800855d088ULL,
    0x5e0c00b8d2544ULL,
    0xd70e00cf1072aULL,
    0xe38b00587feb1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0x8400000000000ULL,
    0x9a00000000000ULL,
    0xfd00000000000ULL,
    0xd580000000000ULL,
    0xc5c0000000000ULL,
    0x5b60000000000ULL,
    0x3fb0000000000ULL,
    0x9d38000000000ULL,
    0x9874000000000ULL,
    0x3752000000000ULL,
    0x9c4b000000000ULL,
    0xb0a1800000000ULL,
    0xdbd5c00000000ULL,
    0xb602600000000ULL,
    0x2300b00000000ULL,
    0xd285680000000ULL,
    0x7d4e740000000ULL,
    0x6329e20000000ULL,
    0x251f090000000ULL,
    0x63e2628000000ULL,
    0xc7f5c9c000000ULL,
    0xe79382e000000ULL,
    0xf6abb95000000ULL,
    0x40d50ac800000ULL,
    0xc880bdd400000ULL,
    0x4043e05200000ULL,
    0xd6a5702900000ULL,
    0x70df080780000ULL,
    0x108ac42cc0000ULL,
    0x7c468a3ce0000ULL,
    0xe0ae7d8370000ULL,
    0x93d80009d8000ULL,
    0xa204005f04000ULL,
    0x910a00448a000ULL,
    0xab8f006e4f000ULL,
    0x32cb8069ab800ULL,
    0x5beac0645ac00ULL,
    0xabf1e036c9e00ULL,
    0x999e7001ea700ULL,
    0x1a68836f4880ULL,
    0x405b04dc10040ULL,
    0x5cceeaec6f6a0ULL,
    0x44e1cd9f340d0ULL,
    0x4f76e87974888ULL,
    0x7d0b4e4d004cULL,
    0xd00a02628f6a6ULL,
    0x880a7965440d9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x800000000000ULL,
    0x2400000000000ULL,
    0xee00000000000ULL,
    0xcb00000000000ULL,
    0x1a80000000000ULL,
    0xdb40000000000ULL,
    0xd560000000000ULL,
    0xdab0000000000ULL,
    0xf538000000000ULL,
    0x67f4000000000ULL,
    0xf5a000000000ULL,
    0xa9c1000000000ULL,
    0x5a29800000000ULL,
    0x9d7c00000000ULL,
    0x708f200000000ULL,
    0x6e4c900000000ULL,
    0x5cefb80000000ULL,
    0x87754c0000000ULL,
    0x39157a0000000ULL,
    0x692aa10000000ULL,
    0x1f5efa8000000ULL,
    0x31c861c000000ULL,
    0x26225a6000000ULL,
    0x7bd2317000000ULL,
    0x798b426800000ULL,
    0x75cc2d5400000ULL,
    0xb826a08600000ULL,
    0x58db509f00000ULL,
    0xf700981c80000ULL,
    0x4889dc8440000ULL,
    0x8242c229e0000ULL,
    0x76ebedeef0000ULL,
    0x12718054d8000ULL,
    0xe093c06d04000ULL,
    0x8ced20d582000ULL,
    0xff79907fc5000ULL,
    0xf51c387d2b800ULL,
    0x13238c3952c00ULL,
    0x3253da3ec4a00ULL,
    0xc441f162ae500ULL,
    0x9e662e393180ULL,
    0xbef5bd526f1c0ULL,
    0x6cda98fa3c620ULL,
    0x8108dc2374bd0ULL,
    0xff8b421713188ULL,
    0xeacc2d422f1ccULL,
    0xa4a6a035dc62eULL,
    0xdc9b502284bdbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0x9c00000000000ULL,
    0x6e00000000000ULL,
    0x9f00000000000ULL,
    0xd380000000000ULL,
    0x1c40000000000ULL,
    0xcea0000000000ULL,
    0xd7f0000000000ULL,
    0xc978000000000ULL,
    0xa7bc000000000ULL,
    0x41da000000000ULL,
    0x134d000000000ULL,
    0x4523800000000ULL,
    0x5fbe400000000ULL,
    0x85d6a00000000ULL,
    0xe94c500000000ULL,
    0x5829e80000000ULL,
    0x1532ec0000000ULL,
    0x4910660000000ULL,
    0x862e290000000ULL,
    0xc231e68000000ULL,
    0x1691694000000ULL,
    0xec66c6e000000ULL,
    0x6792793000000ULL,
    0xe0e00e8800000ULL,
    0xc85f85ac00000ULL,
    0x420ca0e600000ULL,
    0xb901503300000ULL,
    0xc08a68b580000ULL,
    0xb9ccac6f40000ULL,
    0xb9e6c61b20000ULL,
    0x38d279c8b0000ULL,
    0xf3c00e3a58000ULL,
    0xecef85b30c000ULL,
    0xfe54a07582000ULL,
    0x490d504f41000ULL,
    0xe888680b21800ULL,
    0x45cdacd0bf400ULL,
    0xa7e7464e57200ULL,
    0xcfdd390103100ULL,
    0xbc4f2ee486c80ULL,
    0x9ea09573cdfc0ULL,
    0xaff0682cecae0ULL,
    0x4d71acde5ed50ULL,
    0x3dbd465906c88ULL,
    0x2cd039f08dfc4ULL,
    0x31ccae31ccae6ULL,
    0x15eed515eed57ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xa800000000000ULL,
    0xbc00000000000ULL,
    0x6600000000000ULL,
    0x3b00000000000ULL,
    0x1b80000000000ULL,
    0x66c0000000000ULL,
    0x7460000000000ULL,
    0xfb0000000000ULL,
    0x5178000000000ULL,
    0x111c000000000ULL,
    0xf4aa000000000ULL,
    0xd7000000000ULL,
    0x2541800000000ULL,
    0x76a0400000000ULL,
    0x39d1200000000ULL,
    0x47c3100000000ULL,
    0x12eae80000000ULL,
    0x79f9a40000000ULL,
    0xa05bf20000000ULL,
    0x1804e10000000ULL,
    0x40ad28000000ULL,
    0x1207f14000000ULL,
    0xb1003ae000000ULL,
    0xee8e551000000ULL,
    0xfa43c84800000ULL,
    0x6f26b4ac00000ULL,
    0x261b1aae00000ULL,
    0x312a45d700000ULL,
    0x2910a05580000ULL,
    0xc0a350a1c0000ULL,
    0xeadbc869e0000ULL,
    0x804ab40270000ULL,
    0x42291a9698000ULL,
    0x9e9145c46c000ULL,
    0x8632037b2000ULL,
    0x49b810657b000ULL,
    0x5a7968fb13800ULL,
    0xf292e451ab400ULL,
    0x866152345aa00ULL,
    0xaeb0b18204500ULL,
    0x77f09ab902480ULL,
    0x75d052286f40ULL,
    0x558000844a3a0ULL,
    0xa1c0005029550ULL,
    0x69e0002f9a488ULL,
    0x27000e6eaf44ULL,
    0x969800b3f83a6ULL,
    0xc46c003552555ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xf800000000000ULL,
    0x6400000000000ULL,
    0xa200000000000ULL,
    0x7300000000000ULL,
    0xca80000000000ULL,
    0x3bc0000000000ULL,
    0x5ca0000000000ULL,
    0x1c70000000000ULL,
    0x33b8000000000ULL,
    0x1514000000000ULL,
    0xce6a000000000ULL,
    0x5d51000000000ULL,
    0x9c8a800000000ULL,
    0x2c6c00000000ULL,
    0xa92de00000000ULL,
    0xe338500000000ULL,
    0x19d6380000000ULL,
    0x2a4a640000000ULL,
    0xc1e2fe0000000ULL,
    0xa51d810000000ULL,
    0xf66f1e8000000ULL,
    0xd955d1c000000ULL,
    0xe81266000000ULL,
    0x89cbb5f000000ULL,
    0x7a9d89800000ULL,
    0x7af7349400000ULL,
    0x367446ba00000ULL,
    0xfcb0252700000ULL,
    0xc992801080000ULL,
    0xeca2c0ecc0000ULL,
    0x247fe0d420000ULL,
    0xb7bd5064b0000ULL,
    0x8716b85d98000ULL,
    0x456da456a4000ULL,
    0xf3dd9e0372000ULL,
    0x547112735000ULL,
    0x2d66c62bd8800ULL,
    0x7fd2e5b143c00ULL,
    0xeb4d60c76d600ULL,
    0x506f9050df900ULL,
    0xf8515807c9580ULL,
    0xaf04f48aa0f40ULL,
    0x2c8126e573260ULL,
    0x3acbb5263eb50ULL,
    0x2d29d85a51588ULL,
    0x713734dc04f4cULL,
    0x92d446e60126eULL,
    0x84c025010bb53ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xf800000000000ULL,
    0xc00000000000ULL,
    0xca00000000000ULL,
    0x5300000000000ULL,
    0xef80000000000ULL,
    0x7ac0000000000ULL,
    0xde60000000000ULL,
    0x8b10000000000ULL,
    0xc5d8000000000ULL,
    0x3dbc000000000ULL,
    0x9fa6000000000ULL,
    0xeff000000000ULL,
    0x520c800000000ULL,
    0xaf0a400000000ULL,
    0x6d89e00000000ULL,
    0xcdc8500000000ULL,
    0xfe8e80000000ULL,
    0x64dca40000000ULL,
    0x6d30120000000ULL,
    0x70ebcb0000000ULL,
    0x1159f28000000ULL,
    0x1cf39b4000000ULL,
    0x5d091a2000000ULL,
    0xd2833f5000000ULL,
    0x584708d800000ULL,
    0xce2bf45c00000ULL,
    0xa1347a9200000ULL,
    0x5aed2f4f00000ULL,
    0xf258005d80000ULL,
    0x1b7c0065c0000ULL,
    0xd3c6005be0000ULL,
    0xcaef00b2d0000ULL,
    0x4a54800c38000ULL,
    0xb77640c06c000ULL,
    0x9cfe04e1e000ULL,
    0x61e750eb53000ULL,
    0xa9dc68e7f2800ULL,
    0x7bae4e689400ULL,
    0x84a7f27e43600ULL,
    0x5709be72d100ULL,
    0x16c39a75b5880ULL,
    0xe4667f7ba2b40ULL,
    0x901a68b0f71a0ULL,
    0xce55e487003f0ULL,
    0x797372798d888ULL,
    0xd4c6dbbbceb44ULL,
    0xf36c7afee91aaULL,
    0x51912f6c533f1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0x9400000000000ULL,
    0xd600000000000ULL,
    0xa100000000000ULL,
    0x6580000000000ULL,
    0x7dc0000000000ULL,
    0x5fe0000000000ULL,
    0xd6b0000000000ULL,
    0x5978000000000ULL,
    0xf994000000000ULL,
    0x1fa2000000000ULL,
    0xb899000000000ULL,
    0x8a29800000000ULL,
    0xed52c00000000ULL,
    0x71cea00000000ULL,
    0xa5e1300000000ULL,
    0x1db3a80000000ULL,
    0xeff9d40000000ULL,
    0xded99a0000000ULL,
    0x6d0e190000000ULL,
    0xff8f3a8000000ULL,
    0x86cb29c000000ULL,
    0xa16692e000000ULL,
    0x65fffdf000000ULL,
    0xfdd488a800000ULL,
    0x1f8a246400000ULL,
    0x76c492fe00000ULL,
    0x8966fd0500000ULL,
    0xc1fd087b80000ULL,
    0x63d8e488c0000ULL,
    0x2a8a328c60000ULL,
    0xc547cd3a70000ULL,
    0x55aea0ab18000ULL,
    0xfb913006e4000ULL,
    0xc8aba8af3a000ULL,
    0x2c1dd406bd000ULL,
    0xea639ae173800ULL,
    0x137319459fc00ULL,
    0xba9cbaeda5200ULL,
    0x5d24e9479af00ULL,
    0x79dbb23aac880ULL,
    0x11810dd31e240ULL,
    0x5bc2005ae6920ULL,
    0xd6e9002d3ffd0ULL,
    0x17318011b4888ULL,
    0xbab6c015fa24cULL,
    0x9374a015dc926ULL,
    0x7a9c30db82fd3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0x6c00000000000ULL,
    0x7a00000000000ULL,
    0x7700000000000ULL,
    0x7480000000000ULL,
    0x9c0000000000ULL,
    0xabe0000000000ULL,
    0xff90000000000ULL,
    0x5158000000000ULL,
    0xa634000000000ULL,
    0x4c6a000000000ULL,
    0x54db000000000ULL,
    0x1e7f800000000ULL,
    0x61cb400000000ULL,
    0x97e3e00000000ULL,
    0x6d91100000000ULL,
    0xda56180000000ULL,
    0xa0b2e40000000ULL,
    0x5ea60a0000000ULL,
    0xf1be0b0000000ULL,
    0x9f2fea8000000ULL,
    0xeff41bc000000ULL,
    0xc786726000000ULL,
    0xdb4dbf5000000ULL,
    0xd723986800000ULL,
    0xa3f2a43c00000ULL,
    0xcd826a9200000ULL,
    0x34405b8b00000ULL,
    0xc7a8128680000ULL,
    0xbc38efd2c0000ULL,
    0x6b63e0c560000ULL,
    0xc85110d150000ULL,
    0xebb6186638000ULL,
    0xb822e4ac64000ULL,
    0x737e0ac4d2000ULL,
    0x324a0b167f000ULL,
    0x2a5ea8dcd800ULL,
    0x33bf1b2de4400ULL,
    0xfc21f27a96600ULL,
    0x1572fffed1500ULL,
    0xa74a78c172780ULL,
    0x6528b4c94cb40ULL,
    0x58f3f29c21f20ULL,
    0x530dff4572ff0ULL,
    0x4287f8cf4a788ULL,
    0x74ccf45928b4cULL,
    0x306592caf3f2eULL,
    0xe6dcafd80dff9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0x7c00000000000ULL,
    0x5a00000000000ULL,
    0x3300000000000ULL,
    0xb880000000000ULL,
    0x92c0000000000ULL,
    0x2fe0000000000ULL,
    0xafd0000000000ULL,
    0xdc18000000000ULL,
    0x447c000000000ULL,
    0x3be6000000000ULL,
    0xf9d7000000000ULL,
    0x8d1b800000000ULL,
    0x2bf2400000000ULL,
    0x3fa4e00000000ULL,
    0x51f6500000000ULL,
    0xdca6980000000ULL,
    0x417bd40000000ULL,
    0xa260ba0000000ULL,
    0x3c9d7f0000000ULL,
    0x42ba5a8000000ULL,
    0x2002f4000000ULL,
    0xa701422000000ULL,
    0xae8ebbd000000ULL,
    0x63c6982800000ULL,
    0xd06bd4ac00000ULL,
    0xa398baf200000ULL,
    0x8317fdf00000ULL,
    0x4fc45a6a80000ULL,
    0x626b2f9dc0000ULL,
    0xdc9cc2ed60000ULL,
    0x72bbfbde10000ULL,
    0x9a01f86378000ULL,
    0xd303c4d56c000ULL,
    0x8884c23a1e000ULL,
    0xac7fbcd7b000ULL,
    0x5be7f80065800ULL,
    0x89d4c40b99400ULL,
    0xb51f42e439600ULL,
    0xcff5bb1dc3100ULL,
    0x11a318ad61f80ULL,
    0x44f2947e13c40ULL,
    0xd21da737cc20ULL,
    0x58326f9d6bfb0ULL,
    0xa7c5a26619f88ULL,
    0x2e68eb477fc44ULL,
    0x1e98009b62c2aULL,
    0x35bc005f10fb9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0x1c00000000000ULL,
    0x6200000000000ULL,
    0x7700000000000ULL,
    0xf380000000000ULL,
    0x93c0000000000ULL,
    0xcb60000000000ULL,
    0x350000000000ULL,
    0x1998000000000ULL,
    0xf474000000000ULL,
    0x6a000000000ULL,
    0xe2d7000000000ULL,
    0xb55c800000000ULL,
    0x789f400000000ULL,
    0x16f2600000000ULL,
    0x392f900000000ULL,
    0xa4f6e80000000ULL,
    0x762c340000000ULL,
    0xab7bd20000000ULL,
    0xd7ecaf0000000ULL,
    0xef1bb28000000ULL,
    0x3b303fc000000ULL,
    0x743da6000000ULL,
    0x73204b9000000ULL,
    0xf7fce89800000ULL,
    0x1bab348c00000ULL,
    0x7dbf527a00000ULL,
    0x6007ef3b00000ULL,
    0x9003d2e980000ULL,
    0x1808af38c0000ULL,
    0x4c09b23ae0000ULL,
    0x9a033f7790000ULL,
    0x6b0d5ab978000ULL,
    0x918c0be8e4000ULL,
    0xe4c0082892000ULL,
    0x38e8e4eef3000ULL,
    0x90955a252e800ULL,
    0xd2f80bc6fc400ULL,
    0xf72a088124e00ULL,
    0x19ffe498f7d00ULL,
    0x16a9daa420080ULL,
    0xb5374b7478e40ULL,
    0x9a4068406d5a0ULL,
    0xa3a474c2dc0b0ULL,
    0x41b5320558088ULL,
    0xb20c7fd09ce4cULL,
    0x4f073a72ff5aeULL,
    0xf879b872f0b5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x4800000000000ULL,
    0xbc00000000000ULL,
    0x200000000000ULL,
    0xdf00000000000ULL,
    0xb880000000000ULL,
    0xb1c0000000000ULL,
    0xc860000000000ULL,
    0x77d0000000000ULL,
    0x8f18000000000ULL,
    0xcefc000000000ULL,
    0xc562000000000ULL,
    0xf85b000000000ULL,
    0xca53800000000ULL,
    0xed52400000000ULL,
    0x51dd600000000ULL,
    0x615100000000ULL,
    0xe771d80000000ULL,
    0x4921240000000ULL,
    0x1278720000000ULL,
    0xc2a5770000000ULL,
    0xb13f128000000ULL,
    0xc807674000000ULL,
    0xfc074a6000000ULL,
    0xe203035000000ULL,
    0xcf0bd82800000ULL,
    0xf08624ec00000ULL,
    0xdc9f2aa00000ULL,
    0xca6c377300000ULL,
    0xa8d1f27280000ULL,
    0x37903792c0000ULL,
    0x7f33f212e0000ULL,
    0xd0b374910000ULL,
    0x8f8072d7f8000ULL,
    0x454977e4ec000ULL,
    0x23a512a81a000ULL,
    0x94b0677277000ULL,
    0xfe4eca12a9800ULL,
    0x2d26435935400ULL,
    0xa47f38040ce00ULL,
    0x43a674e60c500ULL,
    0xc4b4ca2907380ULL,
    0x564143d98a740ULL,
    0x812eb8544eca0ULL,
    0xee7f345e26430ULL,
    0x20a22a56ff388ULL,
    0x7e3a139166744ULL,
    0x3880003654caeULL,
    0xf1c0000f51431ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xa800000000000ULL,
    0x9400000000000ULL,
    0xde00000000000ULL,
    0x1100000000000ULL,
    0xf180000000000ULL,
    0xb8c0000000000ULL,
    0x9520000000000ULL,
    0xad70000000000ULL,
    0x9438000000000ULL,
    0x71d4000000000ULL,
    0x4d8e000000000ULL,
    0x32c5000000000ULL,
    0xfa23800000000ULL,
    0x7df0c00000000ULL,
    0x7576200000000ULL,
    0xc831b00000000ULL,
    0xabdde80000000ULL,
    0x1a88840000000ULL,
    0xee42460000000ULL,
    0xf962c90000000ULL,
    0x9f17e68000000ULL,
    0xada3b9c000000ULL,
    0xb83c2e2000000ULL,
    0x63da8df000000ULL,
    0x1e83808800000ULL,
    0xa840c06400000ULL,
    0xd46e20d600000ULL,
    0x2495b0b500000ULL,
    0xda6be88780000ULL,
    0xcd99843dc0000ULL,
    0x87efc6baa0000ULL,
    0xa7570904b0000ULL,
    0xe1c2467098000ULL,
    0xa0a2c9a464000ULL,
    0xf3b7e66c96000ULL,
    0x1c13b91e61000ULL,
    0xcf242eab95800ULL,
    0x3a7e8d5ae1c00ULL,
    0xe8b58094dba00ULL,
    0x4291c0f204700ULL,
    0x763a00308480ULL,
    0x7e1070a289f40ULL,
    0x54264822498e0ULL,
    0xa4fcf4bb6bfd0ULL,
    0x80f20ef410488ULL,
    0x36ff3d3b2df4cULL,
    0x13f068f47f8eaULL,
    0xc7d44a1bafd3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0x2c00000000000ULL,
    0x6600000000000ULL,
    0x7500000000000ULL,
    0xe380000000000ULL,
    0x2fc0000000000ULL,
    0xbaa0000000000ULL,
    0xeed0000000000ULL,
    0xea98000000000ULL,
    0xf3f4000000000ULL,
    0xd182000000000ULL,
    0x58cd000000000ULL,
    0x629800000000ULL,
    0x419d400000000ULL,
    0x6174200000000ULL,
    0x47c7f00000000ULL,
    0x26a5080000000ULL,
    0x60da940000000ULL,
    0xc390ee0000000ULL,
    0xfe79fb0000000ULL,
    0xa74d4e8000000ULL,
    0x67e34bc000000ULL,
    0xb23c66e000000ULL,
    0xf22e2fb000000ULL,
    0x5391806800000ULL,
    0xc679409c00000ULL,
    0x634e208e00000ULL,
    0x5deef02900000ULL,
    0x2936880d80000ULL,
    0x48aed476c0000ULL,
    0xf9d74e3f20000ULL,
    0x261a4bb410000ULL,
    0xd4b7e633b8000ULL,
    0x766e6ff2e4000ULL,
    0x8af42061ba000ULL,
    0x8407f0f5e9000ULL,
    0x9a0508d533800ULL,
    0x8b0a94b6a4400ULL,
    0x4288ee58dfa00ULL,
    0xe4dfb0797b00ULL,
    0xaa6f4e4478a80ULL,
    0xa4fe4bfc40240ULL,
    0xbd0de63d61c60ULL,
    0x6f876fa8749f0ULL,
    0x59c7a07a40a88ULL,
    0x37a3b0786424cULL,
    0xa95aa863fbc66ULL,
    0x8f5d24e98d9f7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xb800000000000ULL,
    0x9c00000000000ULL,
    0xc600000000000ULL,
    0xb500000000000ULL,
    0xab80000000000ULL,
    0x7940000000000ULL,
    0xc60000000000ULL,
    0x78b0000000000ULL,
    0xe278000000000ULL,
    0xc35c000000000ULL,
    0x6586000000000ULL,
    0x3847000000000ULL,
    0xe5e8800000000ULL,
    0x7efb400000000ULL,
    0x8e96200000000ULL,
    0x4022f00000000ULL,
    0x9f53880000000ULL,
    0x438d7c0000000ULL,
    0xbd4b6a0000000ULL,
    0x6668990000000ULL,
    0x23b5ca8000000ULL,
    0xd8f1294000000ULL,
    0x4b90622000000ULL,
    0x63aea5d000000ULL,
    0x5210809800000ULL,
    0xade7404c00000ULL,
    0x2af020de00000ULL,
    0x5c95f0b900000ULL,
    0x872308d580000ULL,
    0x63da3c5040000ULL,
    0x7e434a61e0000ULL,
    0x10e169b4f0000ULL,
    0x7570c2c598000ULL,
    0x67dc1582ac000ULL,
    0xf443a8ab9e000ULL,
    0x7be88c53ab000ULL,
    0x67f0627a16800ULL,
    0xcb1ea589e0400ULL,
    0xa86880f0f8a00ULL,
    0x62bb406f9eb00ULL,
    0x31762039ad280ULL,
    0x4dd2f02114cc0ULL,
    0xf4b88b36ec20ULL,
    0xd1617cf837150ULL,
    0x57356a29b5288ULL,
    0x513399f3f8cc4ULL,
    0x43b4a7910c2aULL,
    0x9fbd691f6c159ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xe800000000000ULL,
    0xc400000000000ULL,
    0x3e00000000000ULL,
    0x4700000000000ULL,
    0xf480000000000ULL,
    0x83c0000000000ULL,
    0xdc60000000000ULL,
    0xb5d0000000000ULL,
    0x7498000000000ULL,
    0x3834000000000ULL,
    0x2306000000000ULL,
    0x7a89000000000ULL,
    0xccb800000000ULL,
    0x9ce5c00000000ULL,
    0xe011200000000ULL,
    0x2bf3d00000000ULL,
    0x476cd80000000ULL,
    0x935d040000000ULL,
    0x6a57b20000000ULL,
    0x19d3270000000ULL,
    0x4e95128000000ULL,
    0x613137c000000ULL,
    0xe08eeaa000000ULL,
    0xe5c6e3b000000ULL,
    0x9766004800000ULL,
    0xcb59007400000ULL,
    0x665380f600000ULL,
    0x93d1c0f300000ULL,
    0xdf97202280000ULL,
    0x16bad000c0000ULL,
    0xa9c75816e0000ULL,
    0xfd68c47110000ULL,
    0xa5e925c78000ULL,
    0x89d4f70e24000ULL,
    0xf69fca0bfe000ULL,
    0x3d3533376d000ULL,
    0xf28ad8fb55800ULL,
    0x58c4046e58c00ULL,
    0x1ae43207dca00ULL,
    0xfb12e7799f100ULL,
    0xcd7a32fdb6780ULL,
    0x79afe7674b140ULL,
    0xc2b7b2a7aa4a0ULL,
    0xefc327d5bdf30ULL,
    0xc66d12034e788ULL,
    0x9cd537a9af14cULL,
    0x5f10ea1ab44a2ULL,
    0x637be323c0f37ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x2800000000000ULL,
    0x1400000000000ULL,
    0x9200000000000ULL,
    0xbd00000000000ULL,
    0xa080000000000ULL,
    0xe40000000000ULL,
    0x94e0000000000ULL,
    0xe650000000000ULL,
    0x5d98000000000ULL,
    0x33bc000000000ULL,
    0x7d46000000000ULL,
    0x5f65000000000ULL,
    0x5f93800000000ULL,
    0x56b0c00000000ULL,
    0x41c6e00000000ULL,
    0x672db00000000ULL,
    0xcc79f80000000ULL,
    0xb960bc0000000ULL,
    0x2094620000000ULL,
    0x733daf0000000ULL,
    0x6307028000000ULL,
    0x5385df4000000ULL,
    0x85cb9a2000000ULL,
    0x8d2813f000000ULL,
    0x5d78e00800000ULL,
    0x7e4b0e400000ULL,
    0x3dd4781a00000ULL,
    0x7c597c1900000ULL,
    0x49f021a80000ULL,
    0x8939dfa740000ULL,
    0x6a0d9a2660000ULL,
    0xd10d131510000ULL,
    0x1e8b6049f8000ULL,
    0xad44702bac000ULL,
    0x276a983c3e000ULL,
    0x7398cc2e89000ULL,
    0x48b8fac54d800ULL,
    0x52c0639369c00ULL,
    0xfcac78d193600ULL,
    0x4db57c9db8700ULL,
    0x3e4102464c980ULL,
    0xfce0dfd0edcc0ULL,
    0x52581acc537a0ULL,
    0x7f98d32c9ca30ULL,
    0xa6be001d34988ULL,
    0xc9c900b801cc4ULL,
    0xc32d80cc0d7aaULL,
    0x7679c00e05a3bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x6800000000000ULL,
    0xb400000000000ULL,
    0xfa00000000000ULL,
    0xcf00000000000ULL,
    0xb980000000000ULL,
    0x67c0000000000ULL,
    0x2760000000000ULL,
    0x3d70000000000ULL,
    0xf838000000000ULL,
    0xeb1c000000000ULL,
    0x61c6000000000ULL,
    0x1661000000000ULL,
    0x79f3800000000ULL,
    0xad7ac00000000ULL,
    0x803e600000000ULL,
    0x671e100000000ULL,
    0xb7c1c80000000ULL,
    0xff66940000000ULL,
    0x4177520000000ULL,
    0x4632bd0000000ULL,
    0xb61cb28000000ULL,
    0xa3476d4000000ULL,
    0x32289a2000000ULL,
    0xf65929b000000ULL,
    0xee6604800000ULL,
    0x22b2100400000ULL,
    0x535fc83200000ULL,
    0x6b948b00000ULL,
    0x70fad22b80000ULL,
    0xc5f57d1cc0000ULL,
    0xf37752e4e0000ULL,
    0x8d32bdd5b0000ULL,
    0x3d9cb246d8000ULL,
    0x4f876d01ac000ULL,
    0x3ec89af69e000ULL,
    0xd7e929c40d000ULL,
    0x923e60520d800ULL,
    0x5c1e109b07c00ULL,
    0x5441c8138be00ULL,
    0xa7a69430c5d00ULL,
    0xb7975242ec280ULL,
    0xa882bd14b2440ULL,
    0x9344b26a50fa0ULL,
    0xba2b6de0e6390ULL,
    0x12569a67b4288ULL,
    0x6ce4290dde444ULL,
    0x31b3e04a2efaaULL,
    0xa4d9d07a5b39fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xc800000000000ULL,
    0xec00000000000ULL,
    0x4200000000000ULL,
    0x1f00000000000ULL,
    0x7780000000000ULL,
    0xe5c0000000000ULL,
    0xe860000000000ULL,
    0xa950000000000ULL,
    0x7018000000000ULL,
    0x9c34000000000ULL,
    0x71c2000000000ULL,
    0x1e6f000000000ULL,
    0xb05f800000000ULL,
    0xe69f400000000ULL,
    0x9b72600000000ULL,
    0xf8e9100000000ULL,
    0x8f1da80000000ULL,
    0x7bbf540000000ULL,
    0x3c04e20000000ULL,
    0x4a062f0000000ULL,
    0x930b028000000ULL,
    0x658f7fc000000ULL,
    0x32c34ae000000ULL,
    0x73e27b9000000ULL,
    0xe92602800000ULL,
    0x8779107c00000ULL,
    0x42e5a8ea00000ULL,
    0xe41b54a300000ULL,
    0x6a3ee2fd80000ULL,
    0x68cd2f16c0000ULL,
    0x88ee825de0000ULL,
    0xb71b3f9390000ULL,
    0x6fb4aa0ff8000ULL,
    0xa0f2b40a4000ULL,
    0x330028c1ba000ULL,
    0x558f14570b000ULL,
    0xaac9025b85800ULL,
    0x57e07fc7c4400ULL,
    0xa09cca276fe00ULL,
    0xda7d3bd6d9500ULL,
    0x2a600099d0480ULL,
    0x765000a659040ULL,
    0x6798006f932a0ULL,
    0x29f400a5f46b0ULL,
    0x51a20043a8488ULL,
    0x5b3f008c3d04cULL,
    0x82478019c92a6ULL,
    0x65ab40c26f6b5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xd800000000000ULL,
    0xf400000000000ULL,
    0x2e00000000000ULL,
    0x7300000000000ULL,
    0x1e80000000000ULL,
    0x6740000000000ULL,
    0x9f60000000000ULL,
    0xde70000000000ULL,
    0x7ab8000000000ULL,
    0x2f14000000000ULL,
    0xb7ce000000000ULL,
    0x3b2b000000000ULL,
    0xf79a800000000ULL,
    0x3f87c00000000ULL,
    0xc8cc200000000ULL,
    0xefa0900000000ULL,
    0xb5d9380000000ULL,
    0xe9652c0000000ULL,
    0x997d2a0000000ULL,
    0xea32850000000ULL,
    0xeb5d8a8000000ULL,
    0xee2ad5c000000ULL,
    0x761c122000000ULL,
    0x3c43a91000000ULL,
    0xbdeea0f800000ULL,
    0x1b3350e400000ULL,
    0x9cdb185600000ULL,
    0x6aeebc5700000ULL,
    0x33be92e880000ULL,
    0xdc9069e040000ULL,
    0x150c802fe0000ULL,
    0x2188c00a30000ULL,
    0xd3c0a05b58000ULL,
    0xad28504624000ULL,
    0x99988a16000ULL,
    0x70d7c3e4f000ULL,
    0xf084b214ec800ULL,
    0xb44ff958b8c00ULL,
    0x51e1381618a00ULL,
    0x61312c4c4c500ULL,
    0x11d32ab5ef980ULL,
    0xdf698537327c0ULL,
    0xbe7f0ac6d0320ULL,
    0xab915f7e3390ULL,
    0x271e327e37988ULL,
    0x9bc839b5567ccULL,
    0xe12d18152632aULL,
    0xaa91bc849c39dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x2800000000000ULL,
    0xec00000000000ULL,
    0xca00000000000ULL,
    0x5100000000000ULL,
    0x1780000000000ULL,
    0xf1c0000000000ULL,
    0x6520000000000ULL,
    0xc750000000000ULL,
    0xd218000000000ULL,
    0xcc74000000000ULL,
    0x524a000000000ULL,
    0x5aed000000000ULL,
    0x3573800000000ULL,
    0x69cb400000000ULL,
    0x812ce00000000ULL,
    0xf159700000000ULL,
    0x911fc80000000ULL,
    0x4efdf40000000ULL,
    0x920cf60000000ULL,
    0x55049b0000000ULL,
    0xd181968000000ULL,
    0x9acfabc000000ULL,
    0xbab3e6000000ULL,
    0xcd1d6f3000000ULL,
    0x8cff604800000ULL,
    0x3f0230dc00000ULL,
    0x508b280200000ULL,
    0xb540844d00000ULL,
    0xce613ef580000ULL,
    0xa6306f4cc0000ULL,
    0x3cace038a0000ULL,
    0x619970a790000ULL,
    0x2bbfc840b8000ULL,
    0x646df40ae4000ULL,
    0xc734f6cd72000ULL,
    0xe3209bbdc9000ULL,
    0x8c5396ff21800ULL,
    0xac96ab6e52400ULL,
    0x1e32be9195600ULL,
    0xc8ab2f03bf300ULL,
    0x8f98000860a80ULL,
    0xacb400cd3fc40ULL,
    0xc0ea005227de0ULL,
    0x9c7d006bd41f0ULL,
    0xaa4b807558a88ULL,
    0x8eef40971bc4cULL,
    0xcb7ee0c5f5deeULL,
    0x36c0700c8d1ffULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xf800000000000ULL,
    0x3c00000000000ULL,
    0x4e00000000000ULL,
    0x1900000000000ULL,
    0x5680000000000ULL,
    0x7e40000000000ULL,
    0x6520000000000ULL,
    0xa750000000000ULL,
    0xe278000000000ULL,
    0xb52c000000000ULL,
    0xf5e000000000ULL,
    0x3679000000000ULL,
    0x8f28800000000ULL,
    0x9c52400000000ULL,
    0xbfe200000000ULL,
    0xf0edf00000000ULL,
    0xc8b6b80000000ULL,
    0xe14cac0000000ULL,
    0xaea4460000000ULL,
    0x5d9e9f0000000ULL,
    0x1e12fe8000000ULL,
    0xc0d2334000000ULL,
    0x5eb6b82000000ULL,
    0x344cacf000000ULL,
    0xee24465800000ULL,
    0xb6de9f8c00000ULL,
    0x9bb2fe3600000ULL,
    0x3cc2336500000ULL,
    0xc16eb83880000ULL,
    0x4170ac9740000ULL,
    0x30a2466ba0000ULL,
    0xec9b9f5510000ULL,
    0x9c9c7eb158000ULL,
    0x849573777c000ULL,
    0xa89e18d5a6000ULL,
    0xae9a1c1415000ULL,
    0x739c5eebd6800ULL,
    0x571183ff3b400ULL,
    0x7e5820358ea00ULL,
    0x54f8f01bc3b00ULL,
    0xdb603896ee980ULL,
    0x2277ec45b45c0ULL,
    0x552ae62dc47e0ULL,
    0x1f5d2ff3e9730ULL,
    0x7e7c667d38188ULL,
    0x4b266fba8f1c4ULL,
    0xee52c6184adeaULL,
    0x5cf5dfe82ac3bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0x3400000000000ULL,
    0xbe00000000000ULL,
    0x5900000000000ULL,
    0xff80000000000ULL,
    0x1fc0000000000ULL,
    0xae60000000000ULL,
    0xb770000000000ULL,
    0x118000000000ULL,
    0xc26c000000000ULL,
    0xdd7e000000000ULL,
    0xd61b000000000ULL,
    0x6cec800000000ULL,
    0x9134c00000000ULL,
    0xedb3a00000000ULL,
    0x92f2f00000000ULL,
    0x81d6a80000000ULL,
    0xce89a40000000ULL,
    0x3c4bea0000000ULL,
    0x93a2390000000ULL,
    0xc21d428000000ULL,
    0x2eb9d4000000ULL,
    0xf036a82000000ULL,
    0xb639a4d000000ULL,
    0xab33ea3800000ULL,
    0x42be39a400000ULL,
    0x387b42a600000ULL,
    0x7b9c9d2d00000ULL,
    0x24286180000ULL,
    0xb7d66496c0000ULL,
    0x7b8cca69e0000ULL,
    0xe9c8090cb0000ULL,
    0xfb664a0978000ULL,
    0x52fbc9581c000ULL,
    0x61df6abde6000ULL,
    0xfe8af982b7000ULL,
    0x3448e25872800ULL,
    0x3fae6d0b9fc00ULL,
    0x4812802827200ULL,
    0xe5efc0cbda300ULL,
    0x56bf20c983080ULL,
    0x567630aacc540ULL,
    0x1a9d087befc20ULL,
    0x5ba754dfb45d0ULL,
    0x8e1b4291f1888ULL,
    0xb8ec9da153944ULL,
    0x1f3c2853c8e2aULL,
    0xbcba64146e6d9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0xb400000000000ULL,
    0x1200000000000ULL,
    0x3900000000000ULL,
    0x2d80000000000ULL,
    0x4bc0000000000ULL,
    0xa1e0000000000ULL,
    0xc8f0000000000ULL,
    0xf118000000000ULL,
    0x3dec000000000ULL,
    0xd6fe000000000ULL,
    0xce17000000000ULL,
    0x6b66800000000ULL,
    0x71bac00000000ULL,
    0xbdbd600000000ULL,
    0x5bb4300000000ULL,
    0xd0b3880000000ULL,
    0x943f040000000ULL,
    0xa70920000000ULL,
    0x5450190000000ULL,
    0x2f431a8000000ULL,
    0x2faf1d4000000ULL,
    0xced3882000000ULL,
    0x6e0f041000000ULL,
    0xd708925800000ULL,
    0xba8c19e400000ULL,
    0x51451a6a00000ULL,
    0xa0a41dcd00000ULL,
    0x9053081f80000ULL,
    0xd54ec462c0000ULL,
    0xfaad72d460000ULL,
    0x1555e96730000ULL,
    0xf6cd723af8000ULL,
    0xd665e9381c000ULL,
    0x635723866000ULL,
    0x7379e9913b000ULL,
    0xd9d372e9f8800ULL,
    0x3482e9989dc00ULL,
    0x764bf24c23e00ULL,
    0xb22f294212f00ULL,
    0x8d1012ad68e80ULL,
    0xd3e1d9eab0340ULL,
    0x41fefa213b9a0ULL,
    0xd49aed61f2dd0ULL,
    0x6a25e04490688ULL,
    0x2919f0722df44ULL,
    0x99e8686d187aaULL,
    0x9cf1f423e02d5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0x3c00000000000ULL,
    0xe200000000000ULL,
    0xbb00000000000ULL,
    0x6c80000000000ULL,
    0x7a40000000000ULL,
    0xa0a0000000000ULL,
    0x70b0000000000ULL,
    0xcc18000000000ULL,
    0xcea4000000000ULL,
    0x71ba000000000ULL,
    0x479d000000000ULL,
    0xa6e800000000ULL,
    0x8850400000000ULL,
    0xc700e00000000ULL,
    0x2e83100000000ULL,
    0x9141380000000ULL,
    0x642f8c0000000ULL,
    0xaef1be0000000ULL,
    0xb2bbb90000000ULL,
    0xe710868000000ULL,
    0x6a2435c000000ULL,
    0x9ff938e000000ULL,
    0xa13b8cf000000ULL,
    0x1fd3bef800000ULL,
    0x71c2b90c00000ULL,
    0xa46406fa00000ULL,
    0xe959754700000ULL,
    0x7c8f586e80000ULL,
    0x724cdc3140000ULL,
    0x54a8663420000ULL,
    0x6b32506f0000ULL,
    0x311a0096b8000ULL,
    0xc72d00f914000ULL,
    0x3e76805322000ULL,
    0x4af4407879000ULL,
    0xccbae02ff4800ULL,
    0xee1e10f93d400ULL,
    0x15afb843d6600ULL,
    0x2d3fcca3c7500ULL,
    0xa5d15e8763d80ULL,
    0x96c8a9b9d59c0ULL,
    0x1ae9be64c4060ULL,
    0xb01fb9a9e9750ULL,
    0x8caa86a897588ULL,
    0x9ab93540e8dccULL,
    0x8317b82712666ULL,
    0xd42bcc0a2e253ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0x7400000000000ULL,
    0x7200000000000ULL,
    0x8b00000000000ULL,
    0x7f80000000000ULL,
    0x3140000000000ULL,
    0x4320000000000ULL,
    0x8870000000000ULL,
    0x4958000000000ULL,
    0x722c000000000ULL,
    0x38f2000000000ULL,
    0x6d9f000000000ULL,
    0x874b800000000ULL,
    0xe223c00000000ULL,
    0x10f7600000000ULL,
    0x7990100000000ULL,
    0xe542880000000ULL,
    0x1268c0000000ULL,
    0x1b7fb60000000ULL,
    0x3adc7d0000000ULL,
    0x2d653e8000000ULL,
    0xf6d6f14000000ULL,
    0x636888e000000ULL,
    0xbd58c5000000ULL,
    0x9de6360800000ULL,
    0xd210bd6400000ULL,
    0x9601de9a00000ULL,
    0xd10921bf00000ULL,
    0xe08f60ed80000ULL,
    0x2ccc10ea40000ULL,
    0x91688834a0000ULL,
    0xc0d58cdd30000ULL,
    0x8266361078000ULL,
    0xf350bd055c000ULL,
    0xbd21defc2a000ULL,
    0x2d7921e5f3000ULL,
    0xdbd7606319800ULL,
    0xd5e010668cc00ULL,
    0xd61a8895c4e00ULL,
    0x9c0a8c05efd00ULL,
    0x460db69e1fe80ULL,
    0x99037d98059c0ULL,
    0xe48ebecc04be0ULL,
    0x26c5314e06310ULL,
    0x667e87d06688ULL,
    0xb9599cbe895c4ULL,
    0x4a2ebeb9c05e6ULL,
    0x44f5311be9e11ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x5800000000000ULL,
    0x1c00000000000ULL,
    0x6a00000000000ULL,
    0xff00000000000ULL,
    0xe980000000000ULL,
    0x240000000000ULL,
    0x5ea0000000000ULL,
    0xc5f0000000000ULL,
    0x458000000000ULL,
    0x3824000000000ULL,
    0xb63e000000000ULL,
    0x64b5000000000ULL,
    0x397e800000000ULL,
    0xd092400000000ULL,
    0xad4e600000000ULL,
    0xdf2dd00000000ULL,
    0xa3b8280000000ULL,
    0xdcfb5c0000000ULL,
    0x30dcde0000000ULL,
    0xb8e9bd0000000ULL,
    0xdd5af68000000ULL,
    0x2ca7e1c000000ULL,
    0xc6f8a8a000000ULL,
    0x57dc1c3000000ULL,
    0x6d6c3e7800000ULL,
    0x45162dec00000ULL,
    0xd50cbeb200000ULL,
    0xf6816d2300000ULL,
    0x3bc45e2380000ULL,
    0xf4eafdcd40000ULL,
    0xdf54164f20000ULL,
    0xb7ad71ebb0000ULL,
    0xa97060c8f8000ULL,
    0x9898d02ed4000ULL,
    0x3946a849e6000ULL,
    0x1291c4dd1000ULL,
    0x72b2be0a60800ULL,
    0xbc746d9097400ULL,
    0x4e1adecd48e00ULL,
    0xf288bdcf2b900ULL,
    0xadc276abb0480ULL,
    0xb1e4a1a8f78c0ULL,
    0x61d648bedc760ULL,
    0x18668c01e1a10ULL,
    0x839cf6d9d0c88ULL,
    0x96c6e1d460cccULL,
    0xee6028c194962ULL,
    0xd69f5cedca31fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0x3400000000000ULL,
    0x1200000000000ULL,
    0xbd00000000000ULL,
    0xb380000000000ULL,
    0x50c0000000000ULL,
    0x620000000000ULL,
    0x14d0000000000ULL,
    0xd7f8000000000ULL,
    0xbca4000000000ULL,
    0x259a000000000ULL,
    0x5c11000000000ULL,
    0xedd1800000000ULL,
    0x1e7bc00000000ULL,
    0x56f600000000ULL,
    0x7235700000000ULL,
    0xd08280000000ULL,
    0xbb85d40000000ULL,
    0xf4c05e0000000ULL,
    0x4c2ca70000000ULL,
    0x35d2768000000ULL,
    0x427873c000000ULL,
    0x4363a8e000000ULL,
    0x2d3f143000000ULL,
    0xab86bec800000ULL,
    0x6cc617c400000ULL,
    0xb02f3e3a00000ULL,
    0xe3d9d74900000ULL,
    0x457a5e4180000ULL,
    0x79eda7ddc0000ULL,
    0xdc7bf6fda0000ULL,
    0x4067b34010000ULL,
    0xcdb6c88bd8000ULL,
    0x23cb641174000ULL,
    0x8ea7169be2000ULL,
    0xc89c03c975000ULL,
    0xd79a0047eb800ULL,
    0xd111004f7ac00ULL,
    0x965180f8e6e00ULL,
    0x8abbc0d9fab00ULL,
    0xb94f60e7a5480ULL,
    0xefe5705915a40ULL,
    0x7b7028f25bf60ULL,
    0xeae1d4a0b7b30ULL,
    0x64fa5e684ec88ULL,
    0x542da7d26f64cULL,
    0x9dbf630bd166ULL,
    0xf477b3304d03fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0x9400000000000ULL,
    0x2a00000000000ULL,
    0x7700000000000ULL,
    0xc880000000000ULL,
    0x2740000000000ULL,
    0xcf60000000000ULL,
    0x42d0000000000ULL,
    0xeb8000000000ULL,
    0xdee4000000000ULL,
    0x5812000000000ULL,
    0xa51d000000000ULL,
    0x8a9c800000000ULL,
    0xd50c00000000ULL,
    0xf97ae00000000ULL,
    0xf2c7500000000ULL,
    0x77a8380000000ULL,
    0x32739c0000000ULL,
    0x2c42920000000ULL,
    0xf1e7430000000ULL,
    0x7498aa8000000ULL,
    0x7059dfc000000ULL,
    0x16feb86000000ULL,
    0x950a5c7000000ULL,
    0x338ef2d800000ULL,
    0x1c9d32400000ULL,
    0x3d24f29200000ULL,
    0x9e30d32300000ULL,
    0x1daa728280000ULL,
    0x657d132040000ULL,
    0xb4cc125fe0000ULL,
    0x5eaa838190000ULL,
    0x17fecab3d8000ULL,
    0x8c8e4f0f34000ULL,
    0x454c60ec2a000ULL,
    0xf46e90cfb9000ULL,
    0x845c58276e800ULL,
    0x4cf90c9eddc00ULL,
    0x3a0cca68be600ULL,
    0xdf034fa7e3900ULL,
    0x3488e07598d80ULL,
    0x114a5069ddcc0ULL,
    0x3e6cb8603c2a0ULL,
    0x43575c60ad1f0ULL,
    0xbc72728af6588ULL,
    0x894913d3000ccULL,
    0x3a66129a824aeULL,
    0x15383e44e8fbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0x8400000000000ULL,
    0xe600000000000ULL,
    0xe500000000000ULL,
    0x4a80000000000ULL,
    0x25c0000000000ULL,
    0xfd60000000000ULL,
    0x1170000000000ULL,
    0xb518000000000ULL,
    0xe4ec000000000ULL,
    0x233e000000000ULL,
    0x8b35000000000ULL,
    0x3f33800000000ULL,
    0xf136c00000000ULL,
    0xe037200000000ULL,
    0xc4b9b00000000ULL,
    0x807b380000000ULL,
    0xd199540000000ULL,
    0x2c7a0000000ULL,
    0x2d3990000000ULL,
    0xd289428000000ULL,
    0xb9cfcd4000000ULL,
    0xc76eb82000000ULL,
    0x6e76943000000ULL,
    0xf096da2800000ULL,
    0xca9e9f400000ULL,
    0xb216daee00000ULL,
    0xed69e92100000ULL,
    0x976da8c80000ULL,
    0x6919e9f0c0000ULL,
    0x7eeeda1fe0000ULL,
    0x2c35e980b0000ULL,
    0x46b0da0678000ULL,
    0x5370e9a49c000ULL,
    0x861b5a92a6000ULL,
    0x636a291b19000ULL,
    0xb8727ae5ed800ULL,
    0x3d96991fb3c00ULL,
    0xb222c293fca00ULL,
    0xf9d50db4d3700ULL,
    0x7107980f8a180ULL,
    0xb48a241f49e40ULL,
    0x1cc66250a2c20ULL,
    0x2dea7d68150d0ULL,
    0x3bb380c267988ULL,
    0x85f6c0f4fa244ULL,
    0x99d7202d5e62aULL,
    0xe109b0fdc67d7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x2800000000000ULL,
    0xc400000000000ULL,
    0x9a00000000000ULL,
    0x9b00000000000ULL,
    0xbf80000000000ULL,
    0x6040000000000ULL,
    0xf820000000000ULL,
    0x3b10000000000ULL,
    0xe238000000000ULL,
    0x4dac000000000ULL,
    0xf85e000000000ULL,
    0x199d000000000ULL,
    0x90f1800000000ULL,
    0xf48c00000000ULL,
    0x15a4e00000000ULL,
    0x5457900000000ULL,
    0xe79b880000000ULL,
    0x61f69c0000000ULL,
    0x3c3d60000000ULL,
    0xe66db0000000ULL,
    0x563e5e8000000ULL,
    0xbfa1474000000ULL,
    0x752082000000ULL,
    0xcc125cb000000ULL,
    0x43b9368800000ULL,
    0x8cec4b3400000ULL,
    0xd4f4563200000ULL,
    0xd54f1b1f00000ULL,
    0x2ead3e0580000ULL,
    0x1bdf174b40000ULL,
    0xafdae04fa0000ULL,
    0x5dda902f50000ULL,
    0xa2d2080818000ULL,
    0x77525cd9bc000ULL,
    0xa4193617e6000ULL,
    0x27bc4b6b71000ULL,
    0xe6ec56350f800ULL,
    0x67f31b9685c00ULL,
    0xaecb3e00cd600ULL,
    0xd4ee17fde3500ULL,
    0x78f5601879680ULL,
    0x2b4f502e800c0ULL,
    0xdfa7687ccfde0ULL,
    0x175d0c9be9870ULL,
    0xb41e5e2576e88ULL,
    0x5fb147cc05cc4ULL,
    0xfaea086e02beaULL,
    0x11fe5cc90ad7fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x3800000000000ULL,
    0xdc00000000000ULL,
    0xaa00000000000ULL,
    0x2900000000000ULL,
    0xb280000000000ULL,
    0x83c0000000000ULL,
    0xd660000000000ULL,
    0x6530000000000ULL,
    0x8058000000000ULL,
    0x86e4000000000ULL,
    0xbf2000000000ULL,
    0x42b5000000000ULL,
    0xeb14800000000ULL,
    0x2244400000000ULL,
    0xe626600000000ULL,
    0xb616b00000000ULL,
    0xceccf80000000ULL,
    0xd2ebf40000000ULL,
    0x85feaa0000000ULL,
    0xfdb32d0000000ULL,
    0xea98528000000ULL,
    0x209d9c000000ULL,
    0xdd00786000000ULL,
    0x2c8bb4f000000ULL,
    0xd4cacad800000ULL,
    0x3e09dec00000ULL,
    0xb782a1200000ULL,
    0xe4726d3500000ULL,
    0x2df2b27880000ULL,
    0x9bf295ac0000ULL,
    0x7498e03ce0000ULL,
    0x5503f0caf0000ULL,
    0x88c182438000ULL,
    0x42cc0426d4000ULL,
    0xb0e0b2ab2a000ULL,
    0xa8fa29b291000ULL,
    0x93460ae06800ULL,
    0xd253b06f01400ULL,
    0x93e078098ae00ULL,
    0x237bb4c446f00ULL,
    0xd072ca4d20980ULL,
    0x53f49d199c440ULL,
    0x6eb22a018cd20ULL,
    0x99136d204d990ULL,
    0xe74c32bb26188ULL,
    0xc6aa695a9d04cULL,
    0xc0d8007a0632eULL,
    0x2400210b693ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0xb400000000000ULL,
    0xe600000000000ULL,
    0x1500000000000ULL,
    0x6380000000000ULL,
    0xf1c0000000000ULL,
    0xc5a0000000000ULL,
    0xedf0000000000ULL,
    0x6b58000000000ULL,
    0x682c000000000ULL,
    0xfd32000000000ULL,
    0x4b7f000000000ULL,
    0x7117800000000ULL,
    0x984bc00000000ULL,
    0x8b66200000000ULL,
    0xc4d2b00000000ULL,
    0x43efa80000000ULL,
    0xf92840000000ULL,
    0x8e09360000000ULL,
    0x810e8d0000000ULL,
    0xf58c9e8000000ULL,
    0x5ccf094000000ULL,
    0xba2028e000000ULL,
    0x4e35445000000ULL,
    0x5dfd16c800000ULL,
    0xf3533da400000ULL,
    0x42cb6ce00000ULL,
    0x173a4de100000ULL,
    0xa47d3e6580000ULL,
    0x559679b4c0000ULL,
    0xf909a0ee20000ULL,
    0xc98570f830000ULL,
    0x7ec38880f8000ULL,
    0xf1233434dc000ULL,
    0x24bb1ebbea000ULL,
    0x9db4c97393000ULL,
    0x343e084c05800ULL,
    0x5afbf49a04c00ULL,
    0x3d8be5709a00ULL,
    0x9362b93885700ULL,
    0xe8d800b343880ULL,
    0xc9ec00f1e3340ULL,
    0xf09200ac9b1e0ULL,
    0x28f003084c90ULL,
    0x544f80b746088ULL,
    0x5167c08fe7f44ULL,
    0xf3d420d592be6ULL,
    0x6b6db0b901b91ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0x9c00000000000ULL,
    0xea00000000000ULL,
    0xbf00000000000ULL,
    0x5480000000000ULL,
    0x1240000000000ULL,
    0x33a0000000000ULL,
    0xd590000000000ULL,
    0x4738000000000ULL,
    0x2f2c000000000ULL,
    0x6bd2000000000ULL,
    0x5699000000000ULL,
    0xa9b1800000000ULL,
    0x3a69400000000ULL,
    0x48f2600000000ULL,
    0x4f46d00000000ULL,
    0x142ef80000000ULL,
    0xc15f8c0000000ULL,
    0x9dd18e0000000ULL,
    0x4398fb0000000ULL,
    0x2235768000000ULL,
    0xca2774000000ULL,
    0xc11f78e000000ULL,
    0x3576cc3000000ULL,
    0x6c83eec800000ULL,
    0x164e2bec00000ULL,
    0x8da38ec200000ULL,
    0x6c91fb6300000ULL,
    0xeebcf65e80000ULL,
    0x4ae7379d40000ULL,
    0xf4bf182f20000ULL,
    0x1de91c6bd0000ULL,
    0x5c3c965698000ULL,
    0x95a8e7a9bc000ULL,
    0x789860ba6a000ULL,
    0x88b3d008f5000ULL,
    0xc7ed782f43800ULL,
    0x2b3fcc6420400ULL,
    0x2d2a6e695be00ULL,
    0x28db6b01d3900ULL,
    0x981bee2996980ULL,
    0xacf22bdd3c5c0ULL,
    0xc1498eb824f60ULL,
    0x6524fbe35b370ULL,
    0x84df764ed5188ULL,
    0xba1777151c1c4ULL,
    0xfffcf8f37f166ULL,
    0xd7c68cb188a77ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x800000000000ULL,
    0x7c00000000000ULL,
    0x7200000000000ULL,
    0xc300000000000ULL,
    0x7380000000000ULL,
    0x5040000000000ULL,
    0x2de0000000000ULL,
    0x4030000000000ULL,
    0xfe58000000000ULL,
    0x5e2c000000000ULL,
    0x7596000000000ULL,
    0xcf87000000000ULL,
    0x4249800000000ULL,
    0x1ee5400000000ULL,
    0xebbde00000000ULL,
    0xda13900000000ULL,
    0x7dc0c80000000ULL,
    0x84a31c0000000ULL,
    0x8154ea0000000ULL,
    0x3fa4650000000ULL,
    0x16dda28000000ULL,
    0x49e2394000000ULL,
    0x2634a8a000000ULL,
    0xc355cc9000000ULL,
    0x44a9c22800000ULL,
    0xe154e9ac00000ULL,
    0xcfa980fa00000ULL,
    0xced540ff00000ULL,
    0x3de5e02180000ULL,
    0x283f904340000ULL,
    0x7256c85660000ULL,
    0xf4241c6c70000ULL,
    0xc29d6aa1b8000ULL,
    0xb20125dd1c000ULL,
    0xa30042784e000ULL,
    0x8381a981eb000ULL,
    0x884c60ba3f800ULL,
    0x59ead00152400ULL,
    0x4e33287fac600ULL,
    0x4f5b8c36dad00ULL,
    0xeeaba299eb280ULL,
    0x5655392e378c0ULL,
    0xb225283f5da20ULL,
    0x2f9c8c76a2390ULL,
    0xa082220254a88ULL,
    0x4bc0796c25cc4ULL,
    0xf1a0c81691c22ULL,
    0xc7d31c2c08e9dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0x7400000000000ULL,
    0x5e00000000000ULL,
    0xa100000000000ULL,
    0x2380000000000ULL,
    0x68c0000000000ULL,
    0xd720000000000ULL,
    0x1b50000000000ULL,
    0xdbf8000000000ULL,
    0xd664000000000ULL,
    0xabb2000000000ULL,
    0x2d05000000000ULL,
    0x198a800000000ULL,
    0x1fccc00000000ULL,
    0x49a1a00000000ULL,
    0x221a700000000ULL,
    0xef17680000000ULL,
    0xc695540000000ULL,
    0x815bee0000000ULL,
    0xbcf7ff0000000ULL,
    0xc0e6068000000ULL,
    0x8efe6bc000000ULL,
    0x6be4486000000ULL,
    0x8277e4f000000ULL,
    0xe9a726b800000ULL,
    0x3219db4400000ULL,
    0x6712008600000ULL,
    0xda95001500000ULL,
    0xf352809d80000ULL,
    0x37f8c0f9c0000ULL,
    0x1c6ba0aca0000ULL,
    0x64bb700790000ULL,
    0xf78fe852d8000ULL,
    0x26cc946c34000ULL,
    0xfe28ce53ca000ULL,
    0x24d54f93a1000ULL,
    0xc13acee518800ULL,
    0x8a404fe999c00ULL,
    0x6c684eebd3200ULL,
    0x1cb88f1bb2b00ULL,
    0x5383ee3504c80ULL,
    0x10c3ff4d8a240ULL,
    0x732c0611c6060ULL,
    0x2d5f6b40ae6b0ULL,
    0x56fcc84d9c488ULL,
    0xdfee245dd3e4cULL,
    0x3c740656b526eULL,
    0x78ab6b5c8cdb3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0x2400000000000ULL,
    0x5600000000000ULL,
    0x4100000000000ULL,
    0x4e80000000000ULL,
    0x40000000000ULL,
    0xe760000000000ULL,
    0x3370000000000ULL,
    0xc98000000000ULL,
    0xdeac000000000ULL,
    0xabda000000000ULL,
    0x2e45000000000ULL,
    0x1262800000000ULL,
    0xa3f9c00000000ULL,
    0x1dc200000000ULL,
    0x5d48f00000000ULL,
    0xabe8b80000000ULL,
    0x9c3c8c0000000ULL,
    0xf074e20000000ULL,
    0xad12690000000ULL,
    0x2566da8000000ULL,
    0x3c7b254000000ULL,
    0xc714186000000ULL,
    0xf664bc7000000ULL,
    0x55f8fab800000ULL,
    0x90dad51400000ULL,
    0x7bc4208e00000ULL,
    0xd7a4f02500000ULL,
    0x952b8f880000ULL,
    0xd4098c7140000ULL,
    0xae0e62f1e0000ULL,
    0xb507a91730000ULL,
    0x7080fa3df8000ULL,
    0x3d46d5ecdc000ULL,
    0xdbe62089c2000ULL,
    0x243df080a9000ULL,
    0xe47238e6d8800ULL,
    0xa3194caaccc00ULL,
    0x406ac21126a00ULL,
    0x24f399451d300ULL,
    0x8656e2996e980ULL,
    0x5f8b69ee717c0ULL,
    0x66c65a801eda0ULL,
    0x7b2be5d1e7250ULL,
    0x9610b80736188ULL,
    0x70e08c75fdbc4ULL,
    0xf9b6e200d87aeULL,
    0x6dbb69f3ca153ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xa800000000000ULL,
    0x3c00000000000ULL,
    0x3e00000000000ULL,
    0x6500000000000ULL,
    0x9280000000000ULL,
    0x4ac0000000000ULL,
    0xfe0000000000ULL,
    0x5290000000000ULL,
    0xd9f8000000000ULL,
    0xd8ac000000000ULL,
    0x8236000000000ULL,
    0x77cf000000000ULL,
    0x196b800000000ULL,
    0x925b400000000ULL,
    0xe11d200000000ULL,
    0xebb3300000000ULL,
    0xda82b80000000ULL,
    0x46cd640000000ULL,
    0xe9ef820000000ULL,
    0xf390bf0000000ULL,
    0xe17eba8000000ULL,
    0xf56a9b4000000ULL,
    0x445a18a000000ULL,
    0x9816141000000ULL,
    0x17359a8800000ULL,
    0x5d46ab6c00000ULL,
    0xe7ab201600000ULL,
    0xcfbc301900000ULL,
    0x4089388c80000ULL,
    0xf9c6247fc0000ULL,
    0xe46aa23560000ULL,
    0xe4df8f2450000ULL,
    0x51d2026818000ULL,
    0x2b54ffaf3c000ULL,
    0x7d901a694e000ULL,
    0x1c7eebf5a3000ULL,
    0x83eb809cbd800ULL,
    0xf49b409b04400ULL,
    0x58fd20978ea00ULL,
    0xb02330a844700ULL,
    0x7fab8bd29980ULL,
    0x8da164c971540ULL,
    0xc8b9828966ba0ULL,
    0xf90fbf5a569b0ULL,
    0xbc8d3a2d14188ULL,
    0xa7cddb6db5144ULL,
    0xf169388b881a2ULL,
    0x8e5624c642eb5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xd800000000000ULL,
    0x400000000000ULL,
    0x1600000000000ULL,
    0xe500000000000ULL,
    0x7880000000000ULL,
    0xb040000000000ULL,
    0x560000000000ULL,
    0x5c30000000000ULL,
    0x3fd8000000000ULL,
    0x4aac000000000ULL,
    0xea9a000000000ULL,
    0x7747000000000ULL,
    0xe6e8800000000ULL,
    0x5df3c00000000ULL,
    0x2878200000000ULL,
    0x8cbeb00000000ULL,
    0x39b980000000ULL,
    0x65ceec0000000ULL,
    0x51a3820000000ULL,
    0x1b10dd0000000ULL,
    0x85889a8000000ULL,
    0x6cc1f14000000ULL,
    0x5329386000000ULL,
    0x74589c5000000ULL,
    0xc76aba3800000ULL,
    0xd738411400000ULL,
    0x565a204e00000ULL,
    0x5c65b0a100000ULL,
    0x66b1188e80000ULL,
    0x44962c4540000ULL,
    0xc64922a5e0000ULL,
    0xb06aade870000ULL,
    0x1cbba2acb8000ULL,
    0x9b9e6db39c000ULL,
    0x81cb02cdc2000ULL,
    0x57a31dddab000ULL,
    0x2610bab112800ULL,
    0xf90f412284c00ULL,
    0xca8aa05f48a00ULL,
    0xb34a702ae1700ULL,
    0x50eb3897f9b80ULL,
    0x48f39cdf775c0ULL,
    0x98f83a07309a0ULL,
    0xe0fc812e5df10ULL,
    0x14f280286b388ULL,
    0xcaf4c0f8b39c4ULL,
    0xf3f0a09d983aeULL,
    0x997d70bccc811ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x3c00000000000ULL,
    0x9200000000000ULL,
    0xb700000000000ULL,
    0x6080000000000ULL,
    0xdc0000000000ULL,
    0xa820000000000ULL,
    0x310000000000ULL,
    0x2db8000000000ULL,
    0xea64000000000ULL,
    0xe272000000000ULL,
    0xfd47000000000ULL,
    0x4def800000000ULL,
    0x773f400000000ULL,
    0xaca3e00000000ULL,
    0x58dcd00000000ULL,
    0x5498b80000000ULL,
    0x80faf40000000ULL,
    0xaf89ae0000000ULL,
    0x594c7b0000000ULL,
    0xfbe6968000000ULL,
    0xb63dcfc000000ULL,
    0xed26d86000000ULL,
    0x49e64d000000ULL,
    0x88fd76c800000ULL,
    0x63851f2c00000ULL,
    0xf34c603a00000ULL,
    0x58e3904b00000ULL,
    0x78bb581280000ULL,
    0xc5e624aac0000ULL,
    0x7b3116e0a0000ULL,
    0xe6a68f32d0000ULL,
    0xebd7389798000ULL,
    0xb215b49e74000ULL,
    0x94324e4f4a000ULL,
    0x4224ab0ae3000ULL,
    0x80142eafbd800ULL,
    0x13343b7568400ULL,
    0xfaaaf63ef4600ULL,
    0x89de5fa287900ULL,
    0x3d1d8032c9580ULL,
    0xe0b840c4a1240ULL,
    0xe1ec6044de960ULL,
    0xd33903699cf0ULL,
    0xc7a3580ff4d88ULL,
    0xba5224db0964cULL,
    0xc65b16fa8af6eULL,
    0x74558f76ce5f1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0xc00000000000ULL,
    0x6e00000000000ULL,
    0x1500000000000ULL,
    0x5480000000000ULL,
    0xa5c0000000000ULL,
    0xd160000000000ULL,
    0x890000000000ULL,
    0x7878000000000ULL,
    0x71a4000000000ULL,
    0x857a000000000ULL,
    0x112d000000000ULL,
    0x12b6800000000ULL,
    0x1347400000000ULL,
    0xcfa9200000000ULL,
    0xf878300000000ULL,
    0xb1afd80000000ULL,
    0xe57c740000000ULL,
    0x2128da0000000ULL,
    0x8ab1370000000ULL,
    0x1f4b828000000ULL,
    0xa1a703c000000ULL,
    0xed7cf8e000000ULL,
    0xe52944f000000ULL,
    0x40b182f800000ULL,
    0xf04a033c00000ULL,
    0x822a787600000ULL,
    0x673e04d900000ULL,
    0xd000a25a80000ULL,
    0x68063380c0000ULL,
    0xf407a09de0000ULL,
    0x520b706150000ULL,
    0xe304f8a718000ULL,
    0x4d8d445c34000ULL,
    0x9f4b823182000ULL,
    0x61a703c949000ULL,
    0x8d7cf8a8ac800ULL,
    0xd52944dffa400ULL,
    0xd8b18291e7a00ULL,
    0xfc4a030f5b700ULL,
    0xec2a78321cf80ULL,
    0x723e04c8b9440ULL,
    0x8480a2f449820ULL,
    0xcdc633282e030ULL,
    0x2567a03830788ULL,
    0x5a9b70ab8304cULL,
    0x9b7cf88e4e226ULL,
    0x3c29449f25733ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x3800000000000ULL,
    0x5400000000000ULL,
    0x8a00000000000ULL,
    0x4300000000000ULL,
    0xba80000000000ULL,
    0xf140000000000ULL,
    0x9f20000000000ULL,
    0x8e70000000000ULL,
    0xb8d8000000000ULL,
    0x56c000000000ULL,
    0xa7d6000000000ULL,
    0xe9e7000000000ULL,
    0x39a800000000ULL,
    0xb94cc00000000ULL,
    0x6324e00000000ULL,
    0x487d100000000ULL,
    0xf5dfd80000000ULL,
    0x4ee39c0000000ULL,
    0xfb11960000000ULL,
    0xb7051d0000000ULL,
    0xa082ce8000000ULL,
    0xaa4d414000000ULL,
    0xa1ad38a000000ULL,
    0xcd398c9000000ULL,
    0x30f4ce1800000ULL,
    0xbb9a418400000ULL,
    0xad4fb83200000ULL,
    0xc9294c5700000ULL,
    0xdb7e2e1080000ULL,
    0x775c516240000ULL,
    0xeba4e09da0000ULL,
    0xee3d106b30000ULL,
    0x7a7fd88df8000ULL,
    0xa2d39c181c000ULL,
    0x5e69969d8e000ULL,
    0x99591d49cb000ULL,
    0xaaacce316c800ULL,
    0x8bb641dddbc00ULL,
    0x7b9b8c2e6600ULL,
    0x69be4c951dd00ULL,
    0x68bcaeb60d380ULL,
    0x2d3c91e5098c0ULL,
    0x80f600878cce0ULL,
    0x73970012c6410ULL,
    0x1142808fe1b88ULL,
    0x2f20c0de924c4ULL,
    0x4672e0eacaae2ULL,
    0x4da10bbeb91dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0x5c00000000000ULL,
    0x2200000000000ULL,
    0x2b00000000000ULL,
    0xaa80000000000ULL,
    0xd340000000000ULL,
    0x3a20000000000ULL,
    0xd1b0000000000ULL,
    0xf0d8000000000ULL,
    0x774c000000000ULL,
    0x3c2a000000000ULL,
    0xbcb9000000000ULL,
    0xd755800000000ULL,
    0xf383400000000ULL,
    0x2acde00000000ULL,
    0x27e9900000000ULL,
    0x9ed5980000000ULL,
    0xe640dc0000000ULL,
    0xa9ad760000000ULL,
    0xcb70290000000ULL,
    0x9f350e8000000ULL,
    0xc599654000000ULL,
    0x506de06000000ULL,
    0x6219903000000ULL,
    0x94ad984800000ULL,
    0x94fcdc2c00000ULL,
    0x7fff768a00000ULL,
    0x7575297700000ULL,
    0x36328e0880000ULL,
    0xd41f25b840000ULL,
    0xf1a780f0a0000ULL,
    0x7f764032f0000ULL,
    0x1326082f8000ULL,
    0xbc93d08afc000ULL,
    0x79edf846f2000ULL,
    0x47da0cbcf5000ULL,
    0x5fcd0e63ff800ULL,
    0x546565b77a400ULL,
    0xf41fe0ed38600ULL,
    0x21ac90969ad00ULL,
    0x4772189ee0780ULL,
    0x85369c17554c0ULL,
    0x6a9f16538aee0ULL,
    0x2ce3f9bac5f50ULL,
    0xe458f67fe7f88ULL,
    0xd03692ad30c4ULL,
    0x1780eef8408eeULL,
    0xccccf590aa257ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xc800000000000ULL,
    0x1400000000000ULL,
    0xa200000000000ULL,
    0x6500000000000ULL,
    0x7480000000000ULL,
    0x22c0000000000ULL,
    0x2ce0000000000ULL,
    0x8090000000000ULL,
    0x96b8000000000ULL,
    0x26c4000000000ULL,
    0x26e6000000000ULL,
    0x19f000000000ULL,
    0xd831800000000ULL,
    0xbd09c00000000ULL,
    0xc88d600000000ULL,
    0x64c3700000000ULL,
    0x73e8980000000ULL,
    0x4d18a40000000ULL,
    0x76f3820000000ULL,
    0xefe2b70000000ULL,
    0x9b167a8000000ULL,
    0x41f963c000000ULL,
    0x666d602000000ULL,
    0x1153709000000ULL,
    0xf9d0986800000ULL,
    0xd1ca44400000ULL,
    0x16f5826a00000ULL,
    0x1fedb77100000ULL,
    0x31ffa5680000ULL,
    0x9df4a387c0000ULL,
    0xd066007860000ULL,
    0xd65f003250000ULL,
    0xe85180d258000ULL,
    0x5b59c0e254000ULL,
    0x18d560da5e000ULL,
    0xb39770565b000ULL,
    0x1536982857800ULL,
    0x2d83a4fb56c00ULL,
    0xd0440248dce00ULL,
    0xf124777b9ab00ULL,
    0x31f29a813df80ULL,
    0x3e67d34f8fd40ULL,
    0x6d569895451a0ULL,
    0x1fd3a415a1130ULL,
    0x21c027b32788ULL,
    0x137077568d14cULL,
    0xebac9a07c7faaULL,
    0x683cd33860a35ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x7800000000000ULL,
    0x9c00000000000ULL,
    0x6e00000000000ULL,
    0x300000000000ULL,
    0xeb80000000000ULL,
    0xbf40000000000ULL,
    0x8ee0000000000ULL,
    0xf650000000000ULL,
    0x4bf8000000000ULL,
    0xb54c000000000ULL,
    0x3fe2000000000ULL,
    0x10dd000000000ULL,
    0xac34800000000ULL,
    0x1762400000000ULL,
    0x641d200000000ULL,
    0x4d93d00000000ULL,
    0xa7de080000000ULL,
    0x3db10c0000000ULL,
    0x412d3a0000000ULL,
    0x407fdb0000000ULL,
    0xfa0e128000000ULL,
    0xa90d074000000ULL,
    0xca8520e000000ULL,
    0xf1cfd0f000000ULL,
    0x3d24081800000ULL,
    0xde700c2c00000ULL,
    0xe103ba1600000ULL,
    0x6e8c9b9f00000ULL,
    0x63c5b20580000ULL,
    0xe02197fc40000ULL,
    0x4ef2888560000ULL,
    0x29cf4cb910000ULL,
    0x312a1add18000ULL,
    0x187d0b6f1c000ULL,
    0xb6069ae21a000ULL,
    0xf034b7a91000ULL,
    0x2d81baf656800ULL,
    0xe8419b4bff400ULL,
    0xef69323549a00ULL,
    0xb81fd77fe1900ULL,
    0x4395a8f0db280ULL,
    0x14dd9c5c3edc0ULL,
    0xae3a920f69320ULL,
    0x626f47481fd70ULL,
    0xa09800db95a88ULL,
    0xf5c0078dd9c4ULL,
    0x97a00d83a926ULL,
    0xc081004d6f47bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x4800000000000ULL,
    0xe400000000000ULL,
    0xa200000000000ULL,
    0x4f00000000000ULL,
    0x6b80000000000ULL,
    0x6c40000000000ULL,
    0x29a0000000000ULL,
    0xf0f0000000000ULL,
    0x7718000000000ULL,
    0x514c000000000ULL,
    0x152a000000000ULL,
    0x4b31000000000ULL,
    0xccf3800000000ULL,
    0xc916c00000000ULL,
    0xf046600000000ULL,
    0x87abd00000000ULL,
    0xe9f7c80000000ULL,
    0x499dbc0000000ULL,
    0xb588be0000000ULL,
    0x7d47050000000ULL,
    0x1321168000000ULL,
    0x463d694000000ULL,
    0x1874606000000ULL,
    0x86d6d0b000000ULL,
    0x51ae48a800000ULL,
    0xfcfa7c1400000ULL,
    0x211d5eea00000ULL,
    0x44a15ab00000ULL,
    0x9da8be4980000ULL,
    0xaf7056340000ULL,
    0x6419162220000ULL,
    0x34c1692cb0000ULL,
    0x70666076b8000ULL,
    0xe41bd0f5bc000ULL,
    0x74cfc86832000ULL,
    0x9061bc417d000ULL,
    0x141abe5859800ULL,
    0x3cca054567c00ULL,
    0x7460965c95e00ULL,
    0xb616a9dd0d100ULL,
    0x73cb804c89a80ULL,
    0x1feac0b3ca6c0ULL,
    0xda5460bfed760ULL,
    0x5a66d0ca57b90ULL,
    0xef1648e268288ULL,
    0xad467c4311ac4ULL,
    0xb2f5eeb4a96eULL,
    0xfa3715e627a9fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0xb400000000000ULL,
    0xa600000000000ULL,
    0x1f00000000000ULL,
    0x6880000000000ULL,
    0x5ac0000000000ULL,
    0x57a0000000000ULL,
    0x5170000000000ULL,
    0xe218000000000ULL,
    0xfcc000000000ULL,
    0xd22e000000000ULL,
    0x2a3f000000000ULL,
    0x50f4800000000ULL,
    0x175ec00000000ULL,
    0x7e27200000000ULL,
    0x7830900000000ULL,
    0x49f9f80000000ULL,
    0x30d8e40000000ULL,
    0xd46e760000000ULL,
    0xc15bbd0000000ULL,
    0xe928ae8000000ULL,
    0x9cbfc94000000ULL,
    0x5131202000000ULL,
    0x4673901000000ULL,
    0x469b783800000ULL,
    0xb70524e400000ULL,
    0xf48bd63e00000ULL,
    0x20c6edab00000ULL,
    0x92acf64e80000ULL,
    0xacf67d05c0000ULL,
    0xdd550e1f20000ULL,
    0xd32e991bb0000ULL,
    0xf9bb780db8000ULL,
    0xfcb524fabc000ULL,
    0xe133d6ae36000ULL,
    0x2e7aeddef3000ULL,
    0x3a9af6545a800ULL,
    0x3d057d8ca1c00ULL,
    0xf98f8e97f3a00ULL,
    0xf14f596bde500ULL,
    0xbfe8d852e6d80ULL,
    0xc91b74e894740ULL,
    0x814d0e64018e0ULL,
    0x37e2997e00590ULL,
    0x4515788b04588ULL,
    0xc34a24de89b44ULL,
    0x16e7567dc42eaULL,
    0x86942ddb2d095ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x800000000000ULL,
    0x3400000000000ULL,
    0x6e00000000000ULL,
    0xa300000000000ULL,
    0xd980000000000ULL,
    0xc9c0000000000ULL,
    0x23e0000000000ULL,
    0x7ef0000000000ULL,
    0x6ed8000000000ULL,
    0x8ec4000000000ULL,
    0x5c62000000000ULL,
    0xc03b000000000ULL,
    0xaabd800000000ULL,
    0x92f7c00000000ULL,
    0xfcdfa00000000ULL,
    0x97c3d00000000ULL,
    0x18e0280000000ULL,
    0x7b79c40000000ULL,
    0xad16820000000ULL,
    0x682ec10000000ULL,
    0x6c110a8000000ULL,
    0xd0a0d5c000000ULL,
    0x6d5da06000000ULL,
    0x4a08d09000000ULL,
    0xe505a8e800000ULL,
    0xfe8a046400000ULL,
    0x264b226600000ULL,
    0x7526119700000ULL,
    0xde94a23780000ULL,
    0x14ead1aac0000ULL,
    0xb976829a60000ULL,
    0xbc1ec12730000ULL,
    0x18a90a2538000ULL,
    0xb954d55434000ULL,
    0x7407a034ba000ULL,
    0x4e07d049ff000ULL,
    0x130228a95f800ULL,
    0x8182c45c0cc00ULL,
    0xf5cb02ca02200ULL,
    0x79e9012504100ULL,
    0xb3f6aa9e87880ULL,
    0x145705b64e140ULL,
    0x9e87881d2caa0ULL,
    0xb64e147a98050ULL,
    0x9d2caa92e0088ULL,
    0xba98057e76d4cULL,
    0x72e00883948aeULL,
    0x2e76d48663155ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x1800000000000ULL,
    0xc00000000000ULL,
    0x2200000000000ULL,
    0xc500000000000ULL,
    0x1380000000000ULL,
    0x2a40000000000ULL,
    0x3d20000000000ULL,
    0x6e90000000000ULL,
    0xff8000000000ULL,
    0x1744000000000ULL,
    0x92aa000000000ULL,
    0x1ed1000000000ULL,
    0x8bde800000000ULL,
    0x5055400000000ULL,
    0x7e10e00000000ULL,
    0x53b700000000ULL,
    0x98a9e80000000ULL,
    0x6fdfdc0000000ULL,
    0xa65b160000000ULL,
    0xf5190b0000000ULL,
    0x81ba1e8000000ULL,
    0x7c69a7c000000ULL,
    0x673ae0e000000ULL,
    0xfdaa70b000000ULL,
    0x6c57687800000ULL,
    0xe41a9c7c00000ULL,
    0xdc33f63a00000ULL,
    0xc1267bc900000ULL,
    0x9499f6b180000ULL,
    0xa6f77b2f40000ULL,
    0x56c776cea0000ULL,
    0x65e23bf4d0000ULL,
    0x3c7796cad8000ULL,
    0xed094bc5d4000ULL,
    0xa7867e4752000ULL,
    0x1442977095000ULL,
    0xb22f68d0f4800ULL,
    0xac1e9c1dc4400ULL,
    0x1839f6816e600ULL,
    0xe7277ba8be300ULL,
    0x179f767de1080ULL,
    0x46763b3070ac0ULL,
    0x8405964f00fe0ULL,
    0x860c4ba283d70ULL,
    0x930afee7cd888ULL,
    0x8882d72860eccULL,
    0xc6cb08693c9e6ULL,
    0xcde1accaa8e77ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0xac00000000000ULL,
    0xae00000000000ULL,
    0xa100000000000ULL,
    0x9080000000000ULL,
    0x54c0000000000ULL,
    0x9a20000000000ULL,
    0x9310000000000ULL,
    0xcf8000000000ULL,
    0xd4c000000000ULL,
    0xb262000000000ULL,
    0x93f3000000000ULL,
    0xf4c3800000000ULL,
    0xea20400000000ULL,
    0x4b16200000000ULL,
    0x18fdf00000000ULL,
    0xe741980000000ULL,
    0x1164840000000ULL,
    0xc750e0000000ULL,
    0x9189030000000ULL,
    0xb440b68000000ULL,
    0x16e3774000000ULL,
    0x1d37a0e000000ULL,
    0x75eeb01000000ULL,
    0xa2b4388800000ULL,
    0x20a934fc00000ULL,
    0x44dab64600000ULL,
    0x395c770d00000ULL,
    0x431620be80000ULL,
    0xa4fdf0b5c0000ULL,
    0xc141986aa0000ULL,
    0x4c648497d0000ULL,
    0xdaf50e7ed8000ULL,
    0xc84903325c000ULL,
    0x10e0b6909a000ULL,
    0x7033777fbf000ULL,
    0x1b6fa03621800ULL,
    0xbf72b03d13400ULL,
    0x860e38adf5a00ULL,
    0x2d06349dcdb00ULL,
    0xe8336e6afb80ULL,
    0x4dc33709d5740ULL,
    0x4ea180e7d6960ULL,
    0xc5d340a6de870ULL,
    0x99d5a02656388ULL,
    0xafddb0fa9a344ULL,
    0x7ad7b89cb9366ULL,
    0x50597449ac375ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0xa800000000000ULL,
    0x2400000000000ULL,
    0x9a00000000000ULL,
    0x100000000000ULL,
    0xe280000000000ULL,
    0x29c0000000000ULL,
    0xcd60000000000ULL,
    0x8b30000000000ULL,
    0x9098000000000ULL,
    0xf44000000000ULL,
    0x2eaa000000000ULL,
    0x8f5f000000000ULL,
    0x4d2b800000000ULL,
    0x8414c00000000ULL,
    0xc985e00000000ULL,
    0x524cb00000000ULL,
    0x7a22b80000000ULL,
    0xe997f40000000ULL,
    0xc1c56a0000000ULL,
    0x496a350000000ULL,
    0xe130328000000ULL,
    0xa99a71c000000ULL,
    0x61ce60e000000ULL,
    0xb968705000000ULL,
    0xd93f58c800000ULL,
    0x259f44b400000ULL,
    0xdfcdd23200000ULL,
    0x2262c12500000ULL,
    0x3abed8f880000ULL,
    0xeed484e8c0000ULL,
    0x3b63b24fe0000ULL,
    0x643ab132f0000ULL,
    0x21198075f8000ULL,
    0x710fc06074000ULL,
    0x1a84604432000ULL,
    0xc5c770111b000ULL,
    0xe36cd82901800ULL,
    0xb83f84068bc00ULL,
    0x571a32d3ce600ULL,
    0x5e05715c68700ULL,
    0xcb05e041bf580ULL,
    0x2b8cb05d5f440ULL,
    0x7f42b8382dd20ULL,
    0xd6a7f43492c10ULL,
    0x635d6a5546d88ULL,
    0x632e354fa084cULL,
    0xb71a327dd1b26ULL,
    0xe05710ce1b19ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0x1c00000000000ULL,
    0x8600000000000ULL,
    0x2100000000000ULL,
    0x6080000000000ULL,
    0x3c40000000000ULL,
    0x80e0000000000ULL,
    0x9230000000000ULL,
    0xb618000000000ULL,
    0xc48c000000000ULL,
    0x64a000000000ULL,
    0xd7ef000000000ULL,
    0xbbb6800000000ULL,
    0x8ede400000000ULL,
    0xe225a00000000ULL,
    0xb390700000000ULL,
    0xb243180000000ULL,
    0x95e74c0000000ULL,
    0xf0b38a0000000ULL,
    0x21518d0000000ULL,
    0xfbe3b28000000ULL,
    0xf5b8f14000000ULL,
    0x3bd6806000000ULL,
    0x70ae405000000ULL,
    0x4cdda09800000ULL,
    0xe92c700c00000ULL,
    0xfc1118fe00000ULL,
    0xbb844c3d00000ULL,
    0xabcf0ae680000ULL,
    0x6520cd1d40000ULL,
    0x4210926060000ULL,
    0x6686c1ee70000ULL,
    0x5d483856f8000ULL,
    0x657c06bc000ULL,
    0xbe7f322852000ULL,
    0x4ef9b11f63000ULL,
    0x4abda0c3fc800ULL,
    0x365c702431400ULL,
    0x3269185f13200ULL,
    0xdd784c300e300ULL,
    0x657d0ac806b80ULL,
    0xd973cd94073c0ULL,
    0xaf74127208920ULL,
    0xe67b81830ac10ULL,
    0xe2f118bb82388ULL,
    0x44b44cabca7c4ULL,
    0x63570a6529b2eULL,
    0xb0eccd4217f11ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x800000000000ULL,
    0xfc00000000000ULL,
    0x5a00000000000ULL,
    0x4100000000000ULL,
    0x8480000000000ULL,
    0xa540000000000ULL,
    0x6a20000000000ULL,
    0x7bb0000000000ULL,
    0xc858000000000ULL,
    0x6884000000000ULL,
    0x874e000000000ULL,
    0xff29000000000ULL,
    0xe93b800000000ULL,
    0x7e17400000000ULL,
    0x3b2e600000000ULL,
    0x873fb00000000ULL,
    0xb91d580000000ULL,
    0x14ad9c0000000ULL,
    0xeffdb60000000ULL,
    0xefa790000000ULL,
    0xfa750e8000000ULL,
    0x773f15c000000ULL,
    0x511b80a000000ULL,
    0x78a7407000000ULL,
    0xdf6602800000ULL,
    0xbbfbb04c00000ULL,
    0xd8f3585200000ULL,
    0xc9749cbd00000ULL,
    0xfebe36de80000ULL,
    0x22d939e440000ULL,
    0xacd6e6ea0000ULL,
    0xc2eda51ef0000ULL,
    0xec53580278000ULL,
    0x96849c6334000ULL,
    0xf846366716000ULL,
    0x24ad39dbad000ULL,
    0x27fb6ebc75800ULL,
    0xd2f0a5fc3e400ULL,
    0x107ed8ac95e00ULL,
    0x3e3edc6d68f00ULL,
    0x299dd64493380ULL,
    0x87ecc901622c0ULL,
    0x26d3d62698ee0ULL,
    0x44c5c97463e50ULL,
    0xb5e856241eb88ULL,
    0xcbd2897a286ccULL,
    0x2463683bb0e2ULL,
    0x15ad39dc5615bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0xc400000000000ULL,
    0x600000000000ULL,
    0xb00000000000ULL,
    0x4f80000000000ULL,
    0x3540000000000ULL,
    0x5260000000000ULL,
    0xb170000000000ULL,
    0x3dd8000000000ULL,
    0xdd84000000000ULL,
    0x604e000000000ULL,
    0xf2ef000000000ULL,
    0x1db4800000000ULL,
    0xc9fdc00000000ULL,
    0xb518200000000ULL,
    0xdfa6d00000000ULL,
    0x4bd7480000000ULL,
    0x5e892c0000000ULL,
    0x53c2b20000000ULL,
    0xd5a9910000000ULL,
    0xdad95a8000000ULL,
    0xf80badc000000ULL,
    0xf40c80a000000ULL,
    0x6e09c01000000ULL,
    0x870e20b800000ULL,
    0x158dd01400000ULL,
    0xfc4dc89e00000ULL,
    0x10ebeccf00000ULL,
    0xc0b6124980000ULL,
    0x1576813e40000ULL,
    0xbd8329de0000ULL,
    0xbe8b514430000ULL,
    0xa3cdfacfb8000ULL,
    0x9da4bd7cf4000ULL,
    0x86d5e86596000ULL,
    0x3a0d3cfb6b000ULL,
    0xf9015ac3fa800ULL,
    0x2a8fad2412c00ULL,
    0xfdc280f92ca00ULL,
    0x72a6c02c1b100ULL,
    0x1f5aa0c52f680ULL,
    0x9c4010de1ffc0ULL,
    0x20ede8a02dfa0ULL,
    0xa8b93c1694bd0ULL,
    0x99775a80ede88ULL,
    0x51d4adb8b93ccULL,
    0x778000a1775a2ULL,
    0xe1400085d4addULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0x7400000000000ULL,
    0x3600000000000ULL,
    0xe300000000000ULL,
    0x7e80000000000ULL,
    0x7040000000000ULL,
    0xaf60000000000ULL,
    0xd630000000000ULL,
    0xfd98000000000ULL,
    0x9584000000000ULL,
    0x1ac6000000000ULL,
    0xf921000000000ULL,
    0x625f800000000ULL,
    0x692bc00000000ULL,
    0x5a58200000000ULL,
    0xf52f500000000ULL,
    0xa85ed80000000ULL,
    0x68266c0000000ULL,
    0x41ddb20000000ULL,
    0x50e2750000000ULL,
    0x7dfcca8000000ULL,
    0x573489c000000ULL,
    0x2619802000000ULL,
    0x904ac07000000ULL,
    0x7f67a0e800000ULL,
    0x8e3490c400000ULL,
    0x719ef87e00000ULL,
    0x9f8d3c9700000ULL,
    0xbbc56a4880000ULL,
    0x52a5199340000ULL,
    0x8f9ef851e0000ULL,
    0xc88d3c6670000ULL,
    0x53456af2f8000ULL,
    0x71e519f3b4000ULL,
    0x167ef82f5e000ULL,
    0x1afd3cd8a5000ULL,
    0x37bd6aee99800ULL,
    0xd15119c30ac00ULL,
    0x8fa0f88e87a00ULL,
    0x6183c5844900ULL,
    0xe044eacb66f80ULL,
    0x976bd918393c0ULL,
    0x4a3f58a29b6a0ULL,
    0xf98ac6900190ULL,
    0x884121f87788ULL,
    0xf343e57bc7fccULL,
    0xc1e3b272a2caaULL,
    0x5e7775ff9189bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x3800000000000ULL,
    0xc00000000000ULL,
    0x9a00000000000ULL,
    0x8f00000000000ULL,
    0x8a80000000000ULL,
    0xc2c0000000000ULL,
    0xbb60000000000ULL,
    0x1db0000000000ULL,
    0xc998000000000ULL,
    0x530c000000000ULL,
    0x882000000000ULL,
    0x31c7000000000ULL,
    0x83e2800000000ULL,
    0x6473400000000ULL,
    0x6e71600000000ULL,
    0x997d500000000ULL,
    0x7ffa580000000ULL,
    0x373f840000000ULL,
    0x3b5f1a0000000ULL,
    0xa0622d0000000ULL,
    0x3134a28000000ULL,
    0x5654b94000000ULL,
    0xc9e200e000000ULL,
    0xf377005000000ULL,
    0x98fa805800000ULL,
    0xe9bf401c00000ULL,
    0x7f9360a200000ULL,
    0x360a508300000ULL,
    0x2500d81080000ULL,
    0x4d80c44dc0000ULL,
    0x6c4c7a31e0000ULL,
    0xd7a87ddf70000ULL,
    0xbfd47af2f8000ULL,
    0x4ba47d0ebc000ULL,
    0x5dd67a211a000ULL,
    0xa8a37d32cb000ULL,
    0x5d54fa5360800ULL,
    0xdd603d09b4400ULL,
    0x60bd9aaf93e00ULL,
    0x98116d2e0e100ULL,
    0x9d45c2d90b380ULL,
    0x5c29e9bf82d40ULL,
    0x9e98583745420ULL,
    0x8588849b2da90ULL,
    0x8459a3013b88ULL,
    0x19ad6de94a944ULL,
    0xb6dfc22a2ca26ULL,
    0xec22e9db98b91ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x5c00000000000ULL,
    0xae00000000000ULL,
    0x8300000000000ULL,
    0xc480000000000ULL,
    0xd440000000000ULL,
    0x1820000000000ULL,
    0xc750000000000ULL,
    0xd8b8000000000ULL,
    0xdd0c000000000ULL,
    0xf86000000000ULL,
    0x34c7000000000ULL,
    0x6e60800000000ULL,
    0xaa70400000000ULL,
    0x467200000000ULL,
    0xbb76d00000000ULL,
    0x9dee780000000ULL,
    0x20b00c0000000ULL,
    0x9103460000000ULL,
    0xd989370000000ULL,
    0xbbcc9e8000000ULL,
    0x5ce8ab4000000ULL,
    0xa13e00a000000ULL,
    0x76cb001000000ULL,
    0x2b66807800000ULL,
    0x99f7400c00000ULL,
    0xeea7a0f600000ULL,
    0x5e1690df00000ULL,
    0xf711586a80000ULL,
    0x29adc5740000ULL,
    0x1f533edca0000ULL,
    0xc4b23b1310000ULL,
    0x5309584098000ULL,
    0xdc86dc5a5c000ULL,
    0xa84d3ef73e000ULL,
    0xe6293bb9cb000ULL,
    0x1c57d8b9e6800ULL,
    0x403d9c82b7400ULL,
    0xa74c9ee407a00ULL,
    0x94a8abc206900ULL,
    0x371e008509580ULL,
    0x629b005386dc0ULL,
    0x6f5e809acd3e0ULL,
    0xccbb40ed693b0ULL,
    0x5701a06ef7d88ULL,
    0x2e8190d02d9c4ULL,
    0x8549d8a3549e2ULL,
    0xa1a69c5ab4ab5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0xc00000000000ULL,
    0xb200000000000ULL,
    0x900000000000ULL,
    0xe080000000000ULL,
    0x3cc0000000000ULL,
    0xbba0000000000ULL,
    0x6cb0000000000ULL,
    0x8658000000000ULL,
    0x5b04000000000ULL,
    0x3986000000000ULL,
    0xd441000000000ULL,
    0x236a800000000ULL,
    0x7111400000000ULL,
    0x5de3e00000000ULL,
    0x86d7b00000000ULL,
    0xb7cc980000000ULL,
    0x8a2ac40000000ULL,
    0x1cf3f60000000ULL,
    0x33eb10000000ULL,
    0x9d100e8000000ULL,
    0x3fe385c000000ULL,
    0x87de002000000ULL,
    0xf345007000000ULL,
    0x10ec80b800000ULL,
    0x105040bc00000ULL,
    0x340960aa00000ULL,
    0xce06f00500000ULL,
    0x30f785280000ULL,
    0x558d7435c0000ULL,
    0x76476e5b20000ULL,
    0x8260755070000ULL,
    0x859df8bdf8000ULL,
    0xdf2834f7b4000ULL,
    0x867a8e1fde000ULL,
    0x32f2c53f45000ULL,
    0xd03de082ec800ULL,
    0xc092b06950400ULL,
    0xeda0186c89600ULL,
    0xa3ba844ec6f00ULL,
    0xc1da9612af780ULL,
    0x9448413c3d740ULL,
    0x436776229f6e0ULL,
    0x611af12ca4750ULL,
    0xf5e76ec73bf88ULL,
    0x92d0752b1934cULL,
    0x9c5f8a0e80eaULL,
    0x312c34085785bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0x7400000000000ULL,
    0xe200000000000ULL,
    0xf00000000000ULL,
    0x2080000000000ULL,
    0x98c0000000000ULL,
    0x10e0000000000ULL,
    0xe670000000000ULL,
    0xb3d8000000000ULL,
    0xda04000000000ULL,
    0x730e000000000ULL,
    0x68b000000000ULL,
    0xdc7800000000ULL,
    0xa369c00000000ULL,
    0xc839600000000ULL,
    0xd6f6d00000000ULL,
    0x316f80000000ULL,
    0x9eed840000000ULL,
    0xe77f820000000ULL,
    0x5252e90000000ULL,
    0x83419a8000000ULL,
    0xaaaf7dc000000ULL,
    0xd15800a000000ULL,
    0x1dc4005000000ULL,
    0x8b6e00c800000ULL,
    0x9c3b00e400000ULL,
    0xa4ff800a00000ULL,
    0xe41dc07b00000ULL,
    0xca6f60c280000ULL,
    0x9db9d097c0000ULL,
    0x4dbf783060000ULL,
    0x45bf447eb0000ULL,
    0x81b962a338000ULL,
    0x1bb9f93c74000ULL,
    0x88b80240d6000ULL,
    0x3e3b291c8f000ULL,
    0x4bf8fadec9800ULL,
    0x7499adf5e2c00ULL,
    0x2aaef88dfee00ULL,
    0x115984519f100ULL,
    0x3dc982e12f980ULL,
    0x1b6de9a31b540ULL,
    0x74301a4ee97a0ULL,
    0xd0f9bdef7f6d0ULL,
    0x610e0165e188ULL,
    0xc56410d94d94cULL,
    0xbd301819a19a2ULL,
    0xd57694f7df7d9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0xa800000000000ULL,
    0x4c00000000000ULL,
    0x4e00000000000ULL,
    0x9700000000000ULL,
    0xc580000000000ULL,
    0x7d40000000000ULL,
    0xa760000000000ULL,
    0x3bf0000000000ULL,
    0x5528000000000ULL,
    0xf804000000000ULL,
    0x8402000000000ULL,
    0xb201000000000ULL,
    0xe104800000000ULL,
    0xb685400000000ULL,
    0xbac2600000000ULL,
    0x322700000000ULL,
    0xce14b80000000ULL,
    0xd61e2c0000000ULL,
    0x770fea0000000ULL,
    0xe0933b0000000ULL,
    0x58dadf8000000ULL,
    0xfe2b294000000ULL,
    0x2b80002000000ULL,
    0x3a40001000000ULL,
    0xeae0000800000ULL,
    0x9ab0002400000ULL,
    0x1448002a00000ULL,
    0x18f4001300000ULL,
    0x5aaa001380000ULL,
    0xa0450025c0000ULL,
    0x7e6803160000ULL,
    0x4234401f50000ULL,
    0xa98ee029d8000ULL,
    0x7653300efc000ULL,
    0xa5fcd8154a000ULL,
    0x9f395c3e01000ULL,
    0xdc1dd22100800ULL,
    0x3209572c80400ULL,
    0x7413d5b841200ULL,
    0xab1f226da1500ULL,
    0x78c078eb0980ULL,
    0x11577550c89c0ULL,
    0x787b521b852e0ULL,
    0x967d1701878b0ULL,
    0xd17d35bfc3fa8ULL,
    0x9cfc124f24cecULL,
    0x3db8df8fb6b7eULL,
    0x7cda29594aca5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x3800000000000ULL,
    0x8c00000000000ULL,
    0x5a00000000000ULL,
    0x2100000000000ULL,
    0x3b80000000000ULL,
    0x3840000000000ULL,
    0xcbe0000000000ULL,
    0x69f0000000000ULL,
    0x2d98000000000ULL,
    0x2c04000000000ULL,
    0x6a06000000000ULL,
    0xe907000000000ULL,
    0x9f80800000000ULL,
    0xd641c00000000ULL,
    0x3ce4600000000ULL,
    0x2972d00000000ULL,
    0xf59080000000ULL,
    0xe425dc0000000ULL,
    0xf057c20000000ULL,
    0x66895f0000000ULL,
    0xf7efcf8000000ULL,
    0xbddaacc000000ULL,
    0x93e0002000000ULL,
    0x35f0003000000ULL,
    0xaf98003800000ULL,
    0x9104000400000ULL,
    0x3386000e00000ULL,
    0x7c47002300000ULL,
    0x35e0801680000ULL,
    0xa6f1c00840000ULL,
    0xe11c600ee0000ULL,
    0x54c6d00e10000ULL,
    0x83270832f8000ULL,
    0x48d6dc1a7c000ULL,
    0x2849420b66000ULL,
    0x75cb9f0b01000ULL,
    0x3e8d2f9a81800ULL,
    0xabeebcfa41c00ULL,
    0x3fdde807e0200ULL,
    0x2ee6cc0590700ULL,
    0x6c72aa3739180ULL,
    0x3ada530e5cb40ULL,
    0x3b67058dd6420ULL,
    0x43312fda09770ULL,
    0xa1b88d8a95f08ULL,
    0x485233e1e257cULL,
    0x2a8bafab1bf3eULL,
    0x4de87cd566ab3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0xdc00000000000ULL,
    0x4a00000000000ULL,
    0x4f00000000000ULL,
    0x5380000000000ULL,
    0xe2c0000000000ULL,
    0x8660000000000ULL,
    0x10f0000000000ULL,
    0x7ed8000000000ULL,
    0xb404000000000ULL,
    0xe602000000000ULL,
    0xed01000000000ULL,
    0xf081800000000ULL,
    0x3346400000000ULL,
    0xf7a6e00000000ULL,
    0x8f12500000000ULL,
    0xc3ea780000000ULL,
    0x1f3e9c0000000ULL,
    0xa031160000000ULL,
    0xf3bf330000000ULL,
    0xb974078000000ULL,
    0x5b1e36c000000ULL,
    0x9de0002000000ULL,
    0x6e30001000000ULL,
    0x92b8000800000ULL,
    0xdbf4000c00000ULL,
    0x35a003200000ULL,
    0x67c5003700000ULL,
    0xdae3801280000ULL,
    0x81b74013c0000ULL,
    0x2a7f6014e0000ULL,
    0xea901038b0000ULL,
    0x542e982198000ULL,
    0x6dddcc043c000ULL,
    0xed82ee1fb6000ULL,
    0x6bc3ef2d01000ULL,
    0x8e1f1b980800ULL,
    0xcab255fb40400ULL,
    0x17ffff9c20600ULL,
    0x1956eadcd1900ULL,
    0x58cff615e9b80ULL,
    0xc969633fc4940ULL,
    0xeb7c7f8afa9e0ULL,
    0xf811aafccfa70ULL,
    0xe46896288c458ULL,
    0x7bfd73082fcccULL,
    0xdb50e7a03d01eULL,
    0xebcd66f1b78dbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x2800000000000ULL,
    0x3400000000000ULL,
    0x6a00000000000ULL,
    0xa500000000000ULL,
    0xda80000000000ULL,
    0x10c0000000000ULL,
    0xd520000000000ULL,
    0xc690000000000ULL,
    0xfcf8000000000ULL,
    0x204000000000ULL,
    0xf106000000000ULL,
    0xc085000000000ULL,
    0x4dc6800000000ULL,
    0x13a1400000000ULL,
    0x8851a00000000ULL,
    0xe6db500000000ULL,
    0xbb11280000000ULL,
    0xc7b8d40000000ULL,
    0x761860000000ULL,
    0xaf76290000000ULL,
    0xe94a748000000ULL,
    0x3b6cc7c000000ULL,
    0x678002000000ULL,
    0x2c4003000000ULL,
    0xac26002800000ULL,
    0xe215003400000ULL,
    0xf33e800a00000ULL,
    0x80a5400d00000ULL,
    0xc9d7a01a80000ULL,
    0x939e502940000ULL,
    0xf977a836a0000ULL,
    0x249940430000ULL,
    0xa6e8263548000ULL,
    0x8d397931a4000ULL,
    0x5fa55cbf3e000ULL,
    0x3e5513c081000ULL,
    0xbdd9061c41800ULL,
    0x7e96690021400ULL,
    0x80fb54bb71a00ULL,
    0x2c03d7f0e8500ULL,
    0xc606080814680ULL,
    0xa302c404b6d40ULL,
    0xd9810e1c444a0ULL,
    0x1945ad2cae350ULL,
    0xe462dabd78618ULL,
    0x16f63ae2ed8a4ULL,
    0x800df2b59a9d2ULL,
    0x270feee63f31fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xd800000000000ULL,
    0xe400000000000ULL,
    0xbe00000000000ULL,
    0x3700000000000ULL,
    0xdd80000000000ULL,
    0x3d40000000000ULL,
    0xf320000000000ULL,
    0x6bd0000000000ULL,
    0x3c48000000000ULL,
    0xba04000000000ULL,
    0x9906000000000ULL,
    0x5283000000000ULL,
    0xd4c6800000000ULL,
    0xa866c00000000ULL,
    0x4bf7200000000ULL,
    0x341df00000000ULL,
    0x8c0db80000000ULL,
    0xda4ec0000000ULL,
    0x9d14ea0000000ULL,
    0x492a190000000ULL,
    0xad771e8000000ULL,
    0x46de024000000ULL,
    0x9768002000000ULL,
    0xf5d4003000000ULL,
    0x7b4e001800000ULL,
    0xf87003400000ULL,
    0x4840803600000ULL,
    0x23a5c03900000ULL,
    0xd211a02f80000ULL,
    0xc0ab300dc0000ULL,
    0x2632983760000ULL,
    0xbefd1c0f50000ULL,
    0x7b3f523cc8000ULL,
    0x7dddf51af4000ULL,
    0xd8ed748f12000ULL,
    0x1d96db6e81000ULL,
    0xe8ee3e8641800ULL,
    0xd594f264a0c00ULL,
    0xb4ed380d31a00ULL,
    0x5f912c2e19b00ULL,
    0xe5ed4a3cfdc80ULL,
    0x51152900077c0ULL,
    0xbb2b86ba836e0ULL,
    0xa8741e77a93b0ULL,
    0xfe5fd21fa53a8ULL,
    0xc2a83520da864ULL,
    0x8b34d4b8f5c7aULL,
    0x8a79eb7013809ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0x8400000000000ULL,
    0xda00000000000ULL,
    0x2300000000000ULL,
    0x3180000000000ULL,
    0xcec0000000000ULL,
    0x2aa0000000000ULL,
    0x9610000000000ULL,
    0x4b58000000000ULL,
    0xae04000000000ULL,
    0xa106000000000ULL,
    0xe81000000000ULL,
    0x1944800000000ULL,
    0xb161400000000ULL,
    0xf034200000000ULL,
    0xdb0ed00000000ULL,
    0x13bd180000000ULL,
    0x4cf38c0000000ULL,
    0x96a9760000000ULL,
    0x4e28d50000000ULL,
    0x7a6b708000000ULL,
    0xb8c3ac000000ULL,
    0x33f8002000000ULL,
    0xcf14003000000ULL,
    0x9de000800000ULL,
    0x5945002400000ULL,
    0x5162800a00000ULL,
    0x4030402100000ULL,
    0x6308a03680000ULL,
    0xbfbb9008c0000ULL,
    0x12f7380c60000ULL,
    0x6fa85c33b0000ULL,
    0x5caeee0aa8000ULL,
    0x852f192584000ULL,
    0xefeca692d6000ULL,
    0x8f4e7feb81000ULL,
    0x1258c88841800ULL,
    0xec8526f3a0400ULL,
    0x5642ce2e51200ULL,
    0xfee5c93858500ULL,
    0x57f7bebe0d080ULL,
    0xcb2cf3f3c3b40ULL,
    0xfeed3e986f460ULL,
    0x39c8b3cafce30ULL,
    0xa71b9eb74a5d8ULL,
    0x3e623fcfa354ULL,
    0x5f7626ba52dc2ULL,
    0x376a3fe1d70ebULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x3800000000000ULL,
    0x8400000000000ULL,
    0x5600000000000ULL,
    0x2700000000000ULL,
    0x2b80000000000ULL,
    0xc7c0000000000ULL,
    0x4f60000000000ULL,
    0xd990000000000ULL,
    0xdfb8000000000ULL,
    0x7704000000000ULL,
    0xf382000000000ULL,
    0x73c5000000000ULL,
    0xb164800000000ULL,
    0x4291c00000000ULL,
    0x263c200000000ULL,
    0xc1c6b00000000ULL,
    0xb063380000000ULL,
    0x46145c0000000ULL,
    0xe67abe0000000ULL,
    0xa367bb0000000ULL,
    0xd390ec8000000ULL,
    0x1ab94dc000000ULL,
    0x8582002000000ULL,
    0x84c5001000000ULL,
    0x2e4802800000ULL,
    0x9151c02400000ULL,
    0x75c200e00000ULL,
    0xbb56b02100000ULL,
    0x125b381580000ULL,
    0xd1d05c09c0000ULL,
    0x7118be0ae0000ULL,
    0xcef2bb31f0000ULL,
    0xf22c6c93d8000ULL,
    0xf6bc8df664000ULL,
    0x8f842017ee000ULL,
    0x41c2b00dc1000ULL,
    0xf0613814e0800ULL,
    0xe6115c38f1400ULL,
    0x767e3e2259200ULL,
    0x9b667b31a4700ULL,
    0x5794cc9c0f080ULL,
    0x4cbbfdf9b1ac0ULL,
    0xa2833806f8ce0ULL,
    0xaf445c3075170ULL,
    0xc522be0246af8ULL,
    0xde33bb3abdeecULL,
    0xdecaec8d0a3b2ULL,
    0x64e84dfa6f537ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xf800000000000ULL,
    0x3c00000000000ULL,
    0x5a00000000000ULL,
    0x1b00000000000ULL,
    0xa880000000000ULL,
    0x1c40000000000ULL,
    0x7b60000000000ULL,
    0x8110000000000ULL,
    0x4f08000000000ULL,
    0xab04000000000ULL,
    0x8086000000000ULL,
    0xf841000000000ULL,
    0xd60800000000ULL,
    0x3817c00000000ULL,
    0xc089e00000000ULL,
    0x45c6d00000000ULL,
    0xfca6d80000000ULL,
    0xaab4440000000ULL,
    0xdf38620000000ULL,
    0xa7781b0000000ULL,
    0x6a1fe88000000ULL,
    0x720ba84000000ULL,
    0xa286002000000ULL,
    0x1f41003000000ULL,
    0xdfe0800800000ULL,
    0x2f57c00400000ULL,
    0xeb69e03e00000ULL,
    0xe496d00f00000ULL,
    0x92ced81680000ULL,
    0x9ba04406c0000ULL,
    0xb836622a20000ULL,
    0xe87d1b0710000ULL,
    0x9c99689ed8000ULL,
    0x334d686044000ULL,
    0x20676033c2000ULL,
    0xc994101ac1000ULL,
    0x6349b82821800ULL,
    0x3864543a10400ULL,
    0xc597da3d58200ULL,
    0xd14d4f0105f00ULL,
    0xe760b2a6a2780ULL,
    0xb152757b1b40ULL,
    0x8c0952b509b60ULL,
    0x2f83f75dbd110ULL,
    0x3ec78a8116188ULL,
    0xa423b37d9a06cULL,
    0x7ef1e89f45fa2ULL,
    0x705ea84d43ea1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0x3c00000000000ULL,
    0xd200000000000ULL,
    0x7b00000000000ULL,
    0xef80000000000ULL,
    0x4840000000000ULL,
    0xc920000000000ULL,
    0x8810000000000ULL,
    0x90a8000000000ULL,
    0xa304000000000ULL,
    0x8b82000000000ULL,
    0xbe45000000000ULL,
    0xe424800000000ULL,
    0xda91400000000ULL,
    0xa269e00000000ULL,
    0x64e2900000000ULL,
    0x1671d80000000ULL,
    0xc81a7c0000000ULL,
    0xdefac20000000ULL,
    0x2389090000000ULL,
    0x2c16a08000000ULL,
    0x86a9954000000ULL,
    0xbe02002000000ULL,
    0x6105001000000ULL,
    0x9884802800000ULL,
    0xcdc1402400000ULL,
    0xee61e00a00000ULL,
    0x40b6900f00000ULL,
    0xf9fbd83480000ULL,
    0x6e0b7c1ec0000ULL,
    0x754423be0000ULL,
    0x2409491210000ULL,
    0xb851c0b248000ULL,
    0xad8b456204000ULL,
    0xf514b8042a000ULL,
    0xa2dac38c1000ULL,
    0x9dc0fa0ae0800ULL,
    0x2664e50b91400ULL,
    0x4cb13ab309200ULL,
    0x93ffa079a4500ULL,
    0x10d82bc1a780ULL,
    0x6d60c57f8a40ULL,
    0xc54f78b67c760ULL,
    0xe5f7e954169f0ULL,
    0x825ac207f6b08ULL,
    0xe4d90911e6424ULL,
    0xe81ea0b9afa82ULL,
    0xefd957cab655ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xf800000000000ULL,
    0xbc00000000000ULL,
    0xc200000000000ULL,
    0x5700000000000ULL,
    0xc080000000000ULL,
    0x30c0000000000ULL,
    0x8220000000000ULL,
    0x28b0000000000ULL,
    0x6638000000000ULL,
    0xcb04000000000ULL,
    0x7286000000000ULL,
    0x3fc5000000000ULL,
    0x4ea5800000000ULL,
    0x2277c00000000ULL,
    0xf1de00000000ULL,
    0xe132100000000ULL,
    0x737cb80000000ULL,
    0xb667040000000ULL,
    0x2452060000000ULL,
    0xf9aed10000000ULL,
    0xa20d258000000ULL,
    0xd7f8e1c000000ULL,
    0x88a6002000000ULL,
    0x6b75003000000ULL,
    0x4a9d802800000ULL,
    0xe73c02c00000ULL,
    0x451be03e00000ULL,
    0x5237102f00000ULL,
    0x7df9383080000ULL,
    0xeba0c415c0000ULL,
    0x8df7e63020000ULL,
    0xe358c10c30000ULL,
    0x21d79da088000ULL,
    0x76eae5ca2c000ULL,
    0x846986398e000ULL,
    0x7ba81102c1000ULL,
    0x950b45b4a1800ULL,
    0x77c31e3f1400ULL,
    0xf064d80da9600ULL,
    0xad52d4179df00ULL,
    0x1c2b5e1b47780ULL,
    0xfd4fc5018c840ULL,
    0x199d9b92ff2e0ULL,
    0xd0f034cea9c10ULL,
    0xf0daa3b91c818ULL,
    0xf91f0d187b44ULL,
    0xe10ec5892d496ULL,
    0x417bf1d70f387ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0xc00000000000ULL,
    0xca00000000000ULL,
    0x3700000000000ULL,
    0x2b80000000000ULL,
    0x9d40000000000ULL,
    0xd2e0000000000ULL,
    0x56b0000000000ULL,
    0x55c8000000000ULL,
    0xeb04000000000ULL,
    0x3982000000000ULL,
    0x3e43000000000ULL,
    0xc763800000000ULL,
    0xb2f4400000000ULL,
    0x5da8600000000ULL,
    0xf672500000000ULL,
    0x3f6bb80000000ULL,
    0xe7d65c0000000ULL,
    0xb23d6a0000000ULL,
    0xb62dd70000000ULL,
    0x4b37558000000ULL,
    0xfd89be4000000ULL,
    0x4962002000000ULL,
    0x63f3001000000ULL,
    0x9b2b801800000ULL,
    0x8fb0401c00000ULL,
    0xdf4a602200000ULL,
    0x38c1500300000ULL,
    0x9ea0383280000ULL,
    0x42d61c0dc0000ULL,
    0x7abf0a0ae0000ULL,
    0x5ee8872750000ULL,
    0xd956db4b8000ULL,
    0x2b5ca255ac000ULL,
    0xdde8a3572000ULL,
    0x441fc72ac1000ULL,
    0x4c3e8d9660800ULL,
    0xef2ab25390c00ULL,
    0x81b55233d8e00ULL,
    0x4e4fcb3fbd100ULL,
    0x9e425fbdea180ULL,
    0x9766396c5c940ULL,
    0x2af6ed873aee0ULL,
    0xa9a8e24da5970ULL,
    0xb876ea12b75a8ULL,
    0xce6d9739e775cULL,
    0x315535b75fd56ULL,
    0x33fcee7df36f9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x3800000000000ULL,
    0x5400000000000ULL,
    0x1200000000000ULL,
    0xed00000000000ULL,
    0x9c80000000000ULL,
    0x8940000000000ULL,
    0x8a60000000000ULL,
    0x750000000000ULL,
    0xc138000000000ULL,
    0x9104000000000ULL,
    0x8286000000000ULL,
    0xd245000000000ULL,
    0x81e2800000000ULL,
    0x9391c00000000ULL,
    0x189aa00000000ULL,
    0x6a74900000000ULL,
    0xbf09680000000ULL,
    0x90ede40000000ULL,
    0x507cca0000000ULL,
    0xdf62930000000ULL,
    0xfd31a8000000ULL,
    0xfa78d9c000000ULL,
    0xa666002000000ULL,
    0x2155003000000ULL,
    0xce3a802800000ULL,
    0x8885c01400000ULL,
    0xfb44a00e00000ULL,
    0x9765901500000ULL,
    0xf3d5e80480000ULL,
    0x247d243b40000ULL,
    0x5d62ea2720000ULL,
    0x7ad2c32250000ULL,
    0x62fa52a298000ULL,
    0x5246dc1d4000ULL,
    0x123702104e000ULL,
    0x47efe71441000ULL,
    0x2df8b888a1800ULL,
    0x7ca6aee091400ULL,
    0x9bf5508e78a00ULL,
    0x6acf8ac1e4700ULL,
    0x33c9ba8aa6a80ULL,
    0x8d4c49c5dd240ULL,
    0x210f680ee25a0ULL,
    0xbe8e4376b790ULL,
    0xfbfe4a1407328ULL,
    0x1ba3530c4ca4cULL,
    0xee71ba9e1ac6aULL,
    0x550849d6cf367ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x4800000000000ULL,
    0x1400000000000ULL,
    0x6e00000000000ULL,
    0xc900000000000ULL,
    0xf380000000000ULL,
    0xd4c0000000000ULL,
    0x89e0000000000ULL,
    0xbb10000000000ULL,
    0x7eb8000000000ULL,
    0xe504000000000ULL,
    0x1182000000000ULL,
    0x4fc3000000000ULL,
    0x5964800000000ULL,
    0x7352400000000ULL,
    0xd18a00000000ULL,
    0x9eb7700000000ULL,
    0x400c480000000ULL,
    0x15089c0000000ULL,
    0xbc8c260000000ULL,
    0x3f4b0f0000000ULL,
    0x542cf88000000ULL,
    0x947b45c000000ULL,
    0xcde2002000000ULL,
    0xbd13001000000ULL,
    0x53bc801800000ULL,
    0xa086402400000ULL,
    0x5042a01200000ULL,
    0x4fa0700500000ULL,
    0x91b2c81b80000ULL,
    0xbe8ddc3240000ULL,
    0x744a063ce0000ULL,
    0xaca93f3530000ULL,
    0xd83e90a278000ULL,
    0x9845e9eec4000ULL,
    0x1ba64e3fae000ULL,
    0x9fb1a32941000ULL,
    0xe78ab69c60800ULL,
    0xcfcae6f7f0c00ULL,
    0x6c68b6a459200ULL,
    0x3fd9e6c9d4900ULL,
    0xea5436a0c6280ULL,
    0x969fa6e1eddc0ULL,
    0xae769686e3120ULL,
    0x7fefd6c172270ULL,
    0x3b1c5ebcdb098ULL,
    0x4bb60ae256c3cULL,
    0xa98c5887c53e2ULL,
    0xf6c835ea2fd17ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xf800000000000ULL,
    0x4c00000000000ULL,
    0x7600000000000ULL,
    0x700000000000ULL,
    0xb580000000000ULL,
    0x5f40000000000ULL,
    0x91e0000000000ULL,
    0x8090000000000ULL,
    0xb298000000000ULL,
    0xb104000000000ULL,
    0x5282000000000ULL,
    0xdac5000000000ULL,
    0xa6a4800000000ULL,
    0xa577c00000000ULL,
    0x80a600000000ULL,
    0x729fb00000000ULL,
    0x5106380000000ULL,
    0x6282ac0000000ULL,
    0xb2c37a0000000ULL,
    0x12a7cf0000000ULL,
    0x9f71a48000000ULL,
    0x790c44c000000ULL,
    0xc01a002000000ULL,
    0xbbc1001000000ULL,
    0xac26802800000ULL,
    0xa3b2c02400000ULL,
    0x20aee03e00000ULL,
    0x9ce8701300000ULL,
    0x9a8c581d80000ULL,
    0x485d1c01c0000ULL,
    0xc7a5422d60000ULL,
    0xaff56317d0000ULL,
    0xecadea478000ULL,
    0x5a3f8be024000ULL,
    0xbf71a48ca6000ULL,
    0xa90c44fc41000ULL,
    0x981a001ca0800ULL,
    0x67c10002b1400ULL,
    0x2226803fa9200ULL,
    0xe8b2c01e5df00ULL,
    0xe32ee02182980ULL,
    0xc4a8700e67ec0ULL,
    0xbeec5824a18e0ULL,
    0x978d1c0eb0ab0ULL,
    0xe4dd4225a8de8ULL,
    0x9e6163335df3cULL,
    0xeed0deaf02692ULL,
    0x31fe8bd226113ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0xfc00000000000ULL,
    0x2600000000000ULL,
    0xe300000000000ULL,
    0xc180000000000ULL,
    0xcec0000000000ULL,
    0x3ce0000000000ULL,
    0x4190000000000ULL,
    0xe978000000000ULL,
    0xb504000000000ULL,
    0x5a86000000000ULL,
    0x9343000000000ULL,
    0x2427800000000ULL,
    0x2675c00000000ULL,
    0x956fe00000000ULL,
    0xb4bd300000000ULL,
    0x3065180000000ULL,
    0x49d30c0000000ULL,
    0x90daf60000000ULL,
    0x8bb7a70000000ULL,
    0xa80e2c8000000ULL,
    0x5bee9bc000000ULL,
    0x15fe002000000ULL,
    0x547003000000ULL,
    0xdf21801800000ULL,
    0x8bf6c03c00000ULL,
    0x35a8602e00000ULL,
    0x2f58f03f00000ULL,
    0x6a72f80980000ULL,
    0xab6a3c38c0000ULL,
    0x3bb9ee3060000ULL,
    0x9fe7ab33b0000ULL,
    0x20135a8f38000ULL,
    0xb7bcfcd064000ULL,
    0xc1e7cc9a5e000ULL,
    0x5f10abdd41000ULL,
    0xa03c982ea1800ULL,
    0x5421cc28d0c00ULL,
    0x5e74963f09e00ULL,
    0x96c570a9d700ULL,
    0x62bb5482dbf80ULL,
    0x6b6167eaef4c0ULL,
    0x74500e15f9460ULL,
    0x78199b2904c30ULL,
    0x5451c2836ebd8ULL,
    0x281a30fd39e9cULL,
    0x7c52da91658b2ULL,
    0x9c1a3ce4dea6fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xb800000000000ULL,
    0x6400000000000ULL,
    0x7a00000000000ULL,
    0xf500000000000ULL,
    0xb580000000000ULL,
    0xd7c0000000000ULL,
    0x7860000000000ULL,
    0xdff0000000000ULL,
    0xca58000000000ULL,
    0x6b04000000000ULL,
    0xfa86000000000ULL,
    0xb741000000000ULL,
    0x4a22800000000ULL,
    0xc855c00000000ULL,
    0x9cb200000000ULL,
    0x4afd00000000ULL,
    0xaedda80000000ULL,
    0x9346ac0000000ULL,
    0xd0233e0000000ULL,
    0x4d55830000000ULL,
    0x544a9f8000000ULL,
    0xf6f62c000000ULL,
    0xc8be002000000ULL,
    0xc3b5003000000ULL,
    0x5afc800800000ULL,
    0x4410c01400000ULL,
    0x16fa02e00000ULL,
    0x1fbb101900000ULL,
    0x9734081e80000ULL,
    0xaabcbc3d40000ULL,
    0xc2b5b62d60000ULL,
    0xd7cff35f0000ULL,
    0x52d4099e18000ULL,
    0xe8c4df7fc000ULL,
    0x868fa19296000ULL,
    0xea8be1eac1000ULL,
    0x3c8e1f96a1800ULL,
    0x3f8ba2c9d0400ULL,
    0xd90fa01488a00ULL,
    0x504b100f15700ULL,
    0xc56c083af2c80ULL,
    0xf5b8bc316bf40ULL,
    0xfa33b636576a0ULL,
    0x2b3dff3561ab0ULL,
    0xd7768999f0cf8ULL,
    0xe4198dec1960cULL,
    0x42a481b4fca7eULL,
    0xe61431eb16d8bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x6c00000000000ULL,
    0xa200000000000ULL,
    0x5500000000000ULL,
    0x8980000000000ULL,
    0xc40000000000ULL,
    0x1d60000000000ULL,
    0xcdb0000000000ULL,
    0x2558000000000ULL,
    0x7304000000000ULL,
    0xf282000000000ULL,
    0xc2c5000000000ULL,
    0xf5a0800000000ULL,
    0x1111400000000ULL,
    0x37cb600000000ULL,
    0x7189100000000ULL,
    0xb5eca80000000ULL,
    0xe3df4c0000000ULL,
    0xa7c3e20000000ULL,
    0x4262b0000000ULL,
    0xf951cd8000000ULL,
    0xf4ae9ac000000ULL,
    0x633a002000000ULL,
    0x2031001000000ULL,
    0xb71a802800000ULL,
    0x1160400400000ULL,
    0x9fb1e00a00000ULL,
    0xe859501b00000ULL,
    0xae85482880000ULL,
    0x18c21c1540000ULL,
    0x44a4aa2260000ULL,
    0x4691370310000ULL,
    0xe48de78758000ULL,
    0xdc6aedf36c000ULL,
    0x5f9e87a956000ULL,
    0x82a7fdccc1000ULL,
    0x8d902fb4a0800ULL,
    0x120db1e4b1400ULL,
    0x70ab4dbf68200ULL,
    0x4d3edacb44500ULL,
    0x6733e02772d80ULL,
    0x539c500622440ULL,
    0xd0a5c8059b2a0ULL,
    0x40935c31a7d30ULL,
    0xcf8fca2e48f88ULL,
    0x9ae8273f358acULL,
    0xcb594fbaba736ULL,
    0xd401a1c4d6a6bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0xb800000000000ULL,
    0xd400000000000ULL,
    0xaa00000000000ULL,
    0x6b00000000000ULL,
    0xff80000000000ULL,
    0xc2c0000000000ULL,
    0x9e20000000000ULL,
    0x4450000000000ULL,
    0xf598000000000ULL,
    0xaf84000000000ULL,
    0x4ac2000000000ULL,
    0x5225000000000ULL,
    0x4a53800000000ULL,
    0x8c9dc00000000ULL,
    0xef02a00000000ULL,
    0xdd87500000000ULL,
    0x65c6580000000ULL,
    0x6fa47c0000000ULL,
    0xff93d60000000ULL,
    0x2b3a510000000ULL,
    0x7c93728000000ULL,
    0x88bef4c000000ULL,
    0xd853802000000ULL,
    0x739dc01000000ULL,
    0x1a82a02800000ULL,
    0x447501c00000ULL,
    0xbc66582e00000ULL,
    0x3d347c3500000ULL,
    0x3e2bd62a80000ULL,
    0xabee511ac0000ULL,
    0x3c4972bfe0000ULL,
    0xb7dff4f0b0000ULL,
    0x46e2000788000ULL,
    0xe975000114000ULL,
    0x4a4b801566000ULL,
    0xfad9c037e1000ULL,
    0x7c60a03cb0800ULL,
    0xdd32502189400ULL,
    0xee2dd83814e00ULL,
    0x63edbc39e7700ULL,
    0x504b760420a80ULL,
    0xc9dc0107d1d40ULL,
    0x87e4aabef9960ULL,
    0x7df248cafd1f0ULL,
    0x7709762282f58ULL,
    0xa63901312f944ULL,
    0xa6172aa5945caULL,
    0x6cff88eaa6fd3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x4800000000000ULL,
    0x5400000000000ULL,
    0x9600000000000ULL,
    0xdb00000000000ULL,
    0x1d80000000000ULL,
    0x7940000000000ULL,
    0xd960000000000ULL,
    0xf050000000000ULL,
    0x22a8000000000ULL,
    0xb584000000000ULL,
    0xfd46000000000ULL,
    0x6763000000000ULL,
    0x6f50800000000ULL,
    0xe12a400000000ULL,
    0x43c6a00000000ULL,
    0xafa2b00000000ULL,
    0x3575d80000000ULL,
    0x89186c0000000ULL,
    0xddbd8a0000000ULL,
    0x454e6b0000000ULL,
    0x8d12328000000ULL,
    0x85ca4d4000000ULL,
    0xbad0802000000ULL,
    0xc6a403000000ULL,
    0x6ca6a01800000ULL,
    0x94f2b00400000ULL,
    0x425dd81200000ULL,
    0x11dc6c1500000ULL,
    0x6f9b8a2580000ULL,
    0x97d6b36c0000ULL,
    0xdd6ab28760000ULL,
    0xa8240d5e50000ULL,
    0xdd30201658000ULL,
    0x34fbf00c14000ULL,
    0x142bf810aa000ULL,
    0x49449c2961000ULL,
    0x2160f22d51800ULL,
    0x9c53b70cd8c00ULL,
    0xb8ace0be54200ULL,
    0xe4850a4e8a900ULL,
    0xf9c118b791a80ULL,
    0x4ea29645b8ac0ULL,
    0x9f16a8305760ULL,
    0xacdb616a521b0ULL,
    0x311b2a15c5628ULL,
    0x11bcdb1d329acULL,
    0xaf4feab3950caULL,
    0x8416215f6a535ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0xa400000000000ULL,
    0x2600000000000ULL,
    0x8f00000000000ULL,
    0x9280000000000ULL,
    0x61c0000000000ULL,
    0xfce0000000000ULL,
    0x6ff0000000000ULL,
    0x8978000000000ULL,
    0xa684000000000ULL,
    0xefc6000000000ULL,
    0x77e5000000000ULL,
    0xcb70800000000ULL,
    0xfbb400000000ULL,
    0x6ce1200000000ULL,
    0xc7f7300000000ULL,
    0x8d79780000000ULL,
    0x9080140000000ULL,
    0x8c64e0000000ULL,
    0x4167c70000000ULL,
    0x8cb2cf8000000ULL,
    0x7c5b73c000000ULL,
    0x9190802000000ULL,
    0x2f4b403000000ULL,
    0xd719202800000ULL,
    0x10b3300400000ULL,
    0xf65f781a00000ULL,
    0x2c95142900000ULL,
    0x6ccece0980000ULL,
    0x67588723c0000ULL,
    0x9d15efa4a0000ULL,
    0xad8943d870000ULL,
    0x2b79781f38000ULL,
    0xdf80142bfc000ULL,
    0x3a464e0a5e000ULL,
    0x30a7c72da1000ULL,
    0x1852cfa1f1800ULL,
    0xb7ab73f4f9400ULL,
    0x3ee8801b5c200ULL,
    0x6cf40102ed00ULL,
    0xaa5f201798480ULL,
    0x696302d8dcc0ULL,
    0xc1cff806665e0ULL,
    0x4cde5416dc050ULL,
    0x8957ee29ef938ULL,
    0x62bb71a38f1cULL,
    0xffaa97bc7d33eULL,
    0x4aec57da9f9cfULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0x3400000000000ULL,
    0x2a00000000000ULL,
    0xc300000000000ULL,
    0x6b80000000000ULL,
    0x6740000000000ULL,
    0x4160000000000ULL,
    0x8330000000000ULL,
    0x4928000000000ULL,
    0xf84000000000ULL,
    0x3546000000000ULL,
    0xe67000000000ULL,
    0xaeb7800000000ULL,
    0x5b6d400000000ULL,
    0xf865a00000000ULL,
    0x53b7500000000ULL,
    0xc9e9180000000ULL,
    0x6fa0dc0000000ULL,
    0x82107a0000000ULL,
    0x72fcab0000000ULL,
    0xddec98000000ULL,
    0xef0b114000000ULL,
    0xdd7802000000ULL,
    0x1f5d403000000ULL,
    0xb8cda03800000ULL,
    0x3c73503c00000ULL,
    0x544f182a00000ULL,
    0x72b7dc0d00000ULL,
    0x656ffa0a80000ULL,
    0x165eb30c0000ULL,
    0xa335699ae0000ULL,
    0x592f4159d0000ULL,
    0x5787183058000ULL,
    0xa943dc10cc000ULL,
    0x1061fa2a4a000ULL,
    0x47b6eb3fe1000ULL,
    0xf3ece9a751800ULL,
    0xf4a1014e99c00ULL,
    0x759338012de00ULL,
    0xbbecc161b500ULL,
    0xa5bac21cf9680ULL,
    0xc4bc27313dd40ULL,
    0x48382b8822460ULL,
    0x36fe267624370ULL,
    0x4fda93984e1e8ULL,
    0xf80aaa5f9e2acULL,
    0xbc52718cc6326ULL,
    0x131c9d6d8b045ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xe800000000000ULL,
    0xcc00000000000ULL,
    0x5e00000000000ULL,
    0x3900000000000ULL,
    0x4380000000000ULL,
    0x8fc0000000000ULL,
    0x75e0000000000ULL,
    0x6890000000000ULL,
    0x10e8000000000ULL,
    0x9784000000000ULL,
    0x95c2000000000ULL,
    0xcee1000000000ULL,
    0x9411800000000ULL,
    0xc1af400000000ULL,
    0xbc22600000000ULL,
    0x6070f00000000ULL,
    0xb9f8c80000000ULL,
    0xb7af9c0000000ULL,
    0x29253e0000000ULL,
    0x6df6cf0000000ULL,
    0xd739348000000ULL,
    0xa5cb8f4000000ULL,
    0x5c71802000000ULL,
    0x6fff401000000ULL,
    0x52aa600800000ULL,
    0xeca4f00c00000ULL,
    0xff32c83a00000ULL,
    0x735a9c3300000ULL,
    0x259ebe1780000ULL,
    0x43fc8f0e40000ULL,
    0x5ca8d490e0000ULL,
    0x2da53f63f0000ULL,
    0xa8b2a83d78000ULL,
    0x869b6c0a24000ULL,
    0xfb7ff60c3a000ULL,
    0xfe85329e1000ULL,
    0xb9058a9f70800ULL,
    0x3860040b8400ULL,
    0xafc0d49284600ULL,
    0x45e13f6e2bd00ULL,
    0x8090a817e8980ULL,
    0xdcea6c27ec3c0ULL,
    0xc986760106320ULL,
    0xacc31308cfe70ULL,
    0x8d65eaa3f34f8ULL,
    0x1bd7f043dcb3cULL,
    0xb4499cb7decd2ULL,
    0xd4b1e3677aa3dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x4800000000000ULL,
    0x3400000000000ULL,
    0xe200000000000ULL,
    0xaf00000000000ULL,
    0xdf80000000000ULL,
    0x1cc0000000000ULL,
    0x5220000000000ULL,
    0xfed0000000000ULL,
    0xbaa8000000000ULL,
    0x7d84000000000ULL,
    0xd3c6000000000ULL,
    0x3da5000000000ULL,
    0xba10800000000ULL,
    0x948a400000000ULL,
    0x5555a00000000ULL,
    0x2469100000000ULL,
    0x30a4780000000ULL,
    0xaa907c0000000ULL,
    0xe7cfa60000000ULL,
    0x43b7b10000000ULL,
    0xf49ca68000000ULL,
    0xa2d80d4000000ULL,
    0x20b8802000000ULL,
    0x390e403000000ULL,
    0x6e93a02800000ULL,
    0x3dcc100400000ULL,
    0x20b4f81200000ULL,
    0xa51a3c0d00000ULL,
    0x8f1a063880000ULL,
    0xd41ea12bc0000ULL,
    0x4998deb7e0000ULL,
    0xea58714730000ULL,
    0x2fff263488000ULL,
    0xf9edf10fb4000ULL,
    0xf3610686aa000ULL,
    0xdf351d5b61000ULL,
    0x69daf806f1800ULL,
    0x353b3c3269400ULL,
    0xecc863e04200ULL,
    0x5931e10ae2900ULL,
    0xcddfeb0b5680ULL,
    0xc1bb21432a440ULL,
    0x778efe00211e0ULL,
    0xc4d49d0ed01f0ULL,
    0xa9aad888b9e98ULL,
    0xd402d04cbcec4ULL,
    0x9201f8af5ea9aULL,
    0x570080656b435ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x8800000000000ULL,
    0xa400000000000ULL,
    0x4a00000000000ULL,
    0x5f00000000000ULL,
    0x9480000000000ULL,
    0x90c0000000000ULL,
    0x71e0000000000ULL,
    0x89f0000000000ULL,
    0x4a98000000000ULL,
    0x9a84000000000ULL,
    0x55c2000000000ULL,
    0xa261000000000ULL,
    0x11b5800000000ULL,
    0xcdbc400000000ULL,
    0xd391200000000ULL,
    0x1728500000000ULL,
    0x293ff80000000ULL,
    0xab57240000000ULL,
    0x52cbc60000000ULL,
    0xd2ce2f0000000ULL,
    0x92c9df8000000ULL,
    0xb2ceccc000000ULL,
    0x2cd802000000ULL,
    0x8ac8401000000ULL,
    0x2ecb200800000ULL,
    0x64cd502c00000ULL,
    0x3bc8782200000ULL,
    0xaf4a642900000ULL,
    0x3f8f661280000ULL,
    0x4e6a3f17c0000ULL,
    0xc79f07a520000ULL,
    0x8d05b8e430000ULL,
    0x1783be3c78000ULL,
    0x42444b327c000ULL,
    0xe026b99aa6000ULL,
    0xf194f3caa1000ULL,
    0x3c2a879770800ULL,
    0xefb9f8d198400ULL,
    0xf8929e3eed600ULL,
    0xd1ac1b18af100ULL,
    0x7af941bbc4480ULL,
    0x2833d7e4fa140ULL,
    0xfaf94186b7fe0ULL,
    0x6833d7e669c90ULL,
    0xdaf94199b4f18ULL,
    0xd833d7dde28bcULL,
    0x52f941829af7eULL,
    0x7c33d7d767f33ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0x2c00000000000ULL,
    0x2e00000000000ULL,
    0xe700000000000ULL,
    0x2e80000000000ULL,
    0xa6c0000000000ULL,
    0x58e0000000000ULL,
    0x61b0000000000ULL,
    0xf8c8000000000ULL,
    0x5c84000000000ULL,
    0xb7c6000000000ULL,
    0x2d65000000000ULL,
    0x4ff1800000000ULL,
    0x75edc00000000ULL,
    0x4955600000000ULL,
    0x417f700000000ULL,
    0x184e380000000ULL,
    0x4d42f40000000ULL,
    0xe9a3f60000000ULL,
    0x4e17270000000ULL,
    0x55de3d8000000ULL,
    0xadd85e4000000ULL,
    0xe1d9802000000ULL,
    0x5fd9c03000000ULL,
    0x30db602800000ULL,
    0x8a5e700c00000ULL,
    0x2e99b82e00000ULL,
    0xbf7a340b00000ULL,
    0x174f160b80000ULL,
    0x67c19739c0000ULL,
    0xc563658ba0000ULL,
    0x4bf0da69b0000ULL,
    0xffedce3638000ULL,
    0x1455d3286c000ULL,
    0x8afdcb9632000ULL,
    0x590f795b21000ULL,
    0x7ae7bda3f1800ULL,
    0x58b19e7059400ULL,
    0x294ae0107c600ULL,
    0xc8c3b018bb700ULL,
    0xdfe4d81ff5580ULL,
    0xdf31443eefdc0ULL,
    0xd60f2e35ab8e0ULL,
    0x90626339fcbd0ULL,
    0xd47713847afd8ULL,
    0x1caf3d54d4c9cULL,
    0xdaf713851e0f6ULL,
    0x4a6f3d61f3579ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x8800000000000ULL,
    0xec00000000000ULL,
    0x1a00000000000ULL,
    0xb500000000000ULL,
    0x4680000000000ULL,
    0xf7c0000000000ULL,
    0xe320000000000ULL,
    0x1c90000000000ULL,
    0x3578000000000ULL,
    0x8884000000000ULL,
    0x7cc2000000000ULL,
    0x6a7000000000ULL,
    0x32d1800000000ULL,
    0x5a1c400000000ULL,
    0x2a73600000000ULL,
    0x5bcad00000000ULL,
    0xf1eea80000000ULL,
    0xa9fdb40000000ULL,
    0x6a42fe0000000ULL,
    0x964f90000000ULL,
    0x55f2f48000000ULL,
    0x808ea3c000000ULL,
    0x3809802000000ULL,
    0xccc8401000000ULL,
    0x2669603800000ULL,
    0xeb9d00c00000ULL,
    0xe1a5282200000ULL,
    0x8d52f43b00000ULL,
    0x215a1e0680000ULL,
    0x3391692d40000ULL,
    0x86fcbc91a0000ULL,
    0xd9c287fdf0000ULL,
    0x58273618c8000ULL,
    0x71179d1724000ULL,
    0xf73ca2b55e000ULL,
    0xfde0eeee21000ULL,
    0x6b300a9d30800ULL,
    0x672a5aeaa9c00ULL,
    0x325b749234600ULL,
    0x2216e3e7c7100ULL,
    0x6b8e0213cd80ULL,
    0x4da5900c02b40ULL,
    0x7756481833aa0ULL,
    0xa45824371b6d0ULL,
    0xfd14b622eebf8ULL,
    0x9d3cdd0bc83e4ULL,
    0x20e642a5a43d2ULL,
    0xf1b27ef61e68fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0xec00000000000ULL,
    0xa200000000000ULL,
    0xcd00000000000ULL,
    0x7a80000000000ULL,
    0x8640000000000ULL,
    0x8320000000000ULL,
    0x3e50000000000ULL,
    0x38b8000000000ULL,
    0x6784000000000ULL,
    0xc4c6000000000ULL,
    0x8963000000000ULL,
    0x4f72800000000ULL,
    0x33efc00000000ULL,
    0xc9bb600000000ULL,
    0x8707100000000ULL,
    0x383680000000ULL,
    0x82c2540000000ULL,
    0xc261720000000ULL,
    0x1ef4b90000000ULL,
    0xb42b828000000ULL,
    0x355ebdc000000ULL,
    0xa9b4802000000ULL,
    0xb78cc03000000ULL,
    0x9c49e01800000ULL,
    0x62a8d01400000ULL,
    0xb118083e00000ULL,
    0xd795443b00000ULL,
    0x5b5a1a2880000ULL,
    0x36b2ed3340000ULL,
    0xc80cf09ea0000ULL,
    0x248904e190000ULL,
    0xd1cd8280c8000ULL,
    0x8f6dbdff94000ULL,
    0xc4fe00362e000ULL,
    0x35a7003de1000ULL,
    0xea14801731800ULL,
    0xbe9cc00d58c00ULL,
    0x6751e0055ca00ULL,
    0xcc3cd004bbf00ULL,
    0xeec608044ed80ULL,
    0xa062443311c40ULL,
    0xb3f69a3e88da0ULL,
    0x9eaa2d0eb4950ULL,
    0x4b1b109e7e5c8ULL,
    0xc696d4d1c82e4ULL,
    0x2bd90ab21560aULL,
    0x69f039c6ee6f7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x8800000000000ULL,
    0x3400000000000ULL,
    0x1200000000000ULL,
    0x4f00000000000ULL,
    0x880000000000ULL,
    0x2e40000000000ULL,
    0xa0000000000ULL,
    0x8a30000000000ULL,
    0xc3b8000000000ULL,
    0x6f84000000000ULL,
    0x62c6000000000ULL,
    0x4e7000000000ULL,
    0x2992800000000ULL,
    0xef0c400000000ULL,
    0xeb7da00000000ULL,
    0x5862900000000ULL,
    0x85d3780000000ULL,
    0xc8adc40000000ULL,
    0xdcbb20000000ULL,
    0x2f9fe50000000ULL,
    0x91f0618000000ULL,
    0x3adf75c000000ULL,
    0xabd4802000000ULL,
    0x61ab403000000ULL,
    0xb84f203800000ULL,
    0x165ed01400000ULL,
    0x3f96d82200000ULL,
    0xaa0b540d00000ULL,
    0xf0feca0480000ULL,
    0x8a52113c0000ULL,
    0x7e31538220000ULL,
    0x31b8d0cb90000ULL,
    0x708741a028000ULL,
    0xe245a5d28c000ULL,
    0x1ea45828ee000ULL,
    0x3b37143fe1000ULL,
    0x2a3b6a02b1800ULL,
    0x2043b11839c00ULL,
    0x19a42bace4a00ULL,
    0xe7b214e503100ULL,
    0x467e739c7f680ULL,
    0xe7e600fe48a40ULL,
    0xaf1199bb7cde0ULL,
    0x184ef1ff37710ULL,
    0x665a9209b4ec8ULL,
    0x792351b8af94ULL,
    0x460a3992a3986ULL,
    0x5efb61deaf1d7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x3800000000000ULL,
    0xd400000000000ULL,
    0x7a00000000000ULL,
    0x9100000000000ULL,
    0x680000000000ULL,
    0x4740000000000ULL,
    0x9660000000000ULL,
    0x3b30000000000ULL,
    0x5ba8000000000ULL,
    0x1f84000000000ULL,
    0xbdc6000000000ULL,
    0xa727000000000ULL,
    0x4256800000000ULL,
    0xb519c00000000ULL,
    0x40eaa00000000ULL,
    0xfe1d00000000ULL,
    0x61f5880000000ULL,
    0x1509b40000000ULL,
    0xcf117a0000000ULL,
    0x9a7dd30000000ULL,
    0x715fa98000000ULL,
    0x6d09054000000ULL,
    0x3b10802000000ULL,
    0xd07ec03000000ULL,
    0x85c203800000ULL,
    0x8788103400000ULL,
    0xd257280e00000ULL,
    0xad1c643500000ULL,
    0xa4eaf21e80000ULL,
    0x9de7672440000ULL,
    0x1cf65381a0000ULL,
    0xbd8e1651d0000ULL,
    0x6355098598000ULL,
    0x9b98d57ecc000ULL,
    0xbad080eea000ULL,
    0xe7837403e1000ULL,
    0x89c35a1971800ULL,
    0xd26c328c9c00ULL,
    0xeb50018015a00ULL,
    0x679fa17c06700ULL,
    0x7da8522f1aa80ULL,
    0x882b70668740ULL,
    0x5c45db8445620ULL,
    0x9e0a26e5e6d0ULL,
    0x46f2f3aeb65e8ULL,
    0x1c8cc66ab274cULL,
    0x8dd001a34c6a6ULL,
    0x30dfa1552a815ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xf800000000000ULL,
    0x4c00000000000ULL,
    0x8a00000000000ULL,
    0xd900000000000ULL,
    0x1780000000000ULL,
    0x6e40000000000ULL,
    0xeca0000000000ULL,
    0x2a10000000000ULL,
    0x8b18000000000ULL,
    0x6a84000000000ULL,
    0xd7c2000000000ULL,
    0xb5e7000000000ULL,
    0x631800000000ULL,
    0x3fcfc00000000ULL,
    0x6bfe600000000ULL,
    0x5132500000000ULL,
    0x8f4bc80000000ULL,
    0xed3a3c0000000ULL,
    0xa753320000000ULL,
    0x113fc50000000ULL,
    0x9555e08000000ULL,
    0x643a80c000000ULL,
    0x38d3802000000ULL,
    0x2b78c01000000ULL,
    0x8ff7e03800000ULL,
    0xe529900c00000ULL,
    0x314fa83e00000ULL,
    0xb23b6c1300000ULL,
    0x3d37a2280000ULL,
    0x81f9393640000ULL,
    0xf8333285e0000ULL,
    0x80cbd5db90000ULL,
    0xff78489b28000ULL,
    0x39f6ecda84000ULL,
    0xae297a3ac6000ULL,
    0x83ca3906a1000ULL,
    0xb9f8b2b3f0800ULL,
    0x143715f279c00ULL,
    0xdacda8bd0c600ULL,
    0xee787cfab3f00ULL,
    0x9a7752259f980ULL,
    0x66e95259c940ULL,
    0x3c6da89b1af20ULL,
    0x5d687cf55a8f0ULL,
    0xe6ef520cbacc8ULL,
    0x32aa95212af14ULL,
    0xff0fa8ad03f82ULL,
    0x8e9f7ccf06603ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xc800000000000ULL,
    0x2400000000000ULL,
    0x2e00000000000ULL,
    0xe100000000000ULL,
    0xc080000000000ULL,
    0x9d40000000000ULL,
    0x4b60000000000ULL,
    0xea30000000000ULL,
    0x8488000000000ULL,
    0x8f84000000000ULL,
    0xfcc2000000000ULL,
    0x36a1000000000ULL,
    0x2c10800000000ULL,
    0xedde400000000ULL,
    0xc53d200000000ULL,
    0xd9cf700000000ULL,
    0xa4e0080000000ULL,
    0x26f5840000000ULL,
    0x6a2b2a0000000ULL,
    0x4f91bb0000000ULL,
    0x1b1b418000000ULL,
    0x3c9e7c4000000ULL,
    0xd45a802000000ULL,
    0x14fb401000000ULL,
    0x35efa00800000ULL,
    0x12b0300400000ULL,
    0x85cda83200000ULL,
    0x36e4b40900000ULL,
    0x25f6020b80000ULL,
    0x51ab4f3840000ULL,
    0x155063b020000ULL,
    0xc8ba436750000ULL,
    0xee0aebb2d8000ULL,
    0x8dc4876a8c000ULL,
    0x7e26618922000ULL,
    0xb5510c77e1000ULL,
    0x98ba882530800ULL,
    0x60ec410a8400ULL,
    0xb9c48a3a84200ULL,
    0x98218b0e37900ULL,
    0x7056e9b8ef480ULL,
    0x763ac86063dc0ULL,
    0x7a4c822040020ULL,
    0x32200f3c21610ULL,
    0xef57c39150ca8ULL,
    0xbfbe7369d96ecULL,
    0x138d43980c506ULL,
    0x2053342e2df1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0x9400000000000ULL,
    0x3e00000000000ULL,
    0xed00000000000ULL,
    0xd780000000000ULL,
    0x13c0000000000ULL,
    0xbe20000000000ULL,
    0x1cb0000000000ULL,
    0x4ed8000000000ULL,
    0x6484000000000ULL,
    0x5946000000000ULL,
    0xe261000000000ULL,
    0x9d54800000000ULL,
    0x6e49400000000ULL,
    0x71e8a00000000ULL,
    0x779bf00000000ULL,
    0xf6e4680000000ULL,
    0x5c153c0000000ULL,
    0xa02c5e0000000ULL,
    0x6eb8910000000ULL,
    0x5ed6f58000000ULL,
    0x838f4ec000000ULL,
    0x2ca802000000ULL,
    0x2cac403000000ULL,
    0x83fa200800000ULL,
    0x12b3b02400000ULL,
    0xdbd8480a00000ULL,
    0x2f078c2500000ULL,
    0x7080960f80000ULL,
    0xa7465d3b40000ULL,
    0x2f66c3b5e0000ULL,
    0xdad6e3c4f0000ULL,
    0x558e2b8f88000ULL,
    0x5bce9ff72c000ULL,
    0x552cd5bbb6000ULL,
    0x553cfecd21000ULL,
    0xef12c83451800ULL,
    0xeaabcc0998400ULL,
    0x32fab62ad5200ULL,
    0xdf35ed21d2500ULL,
    0xa31e8bac1a280ULL,
    0x17a16fc756fc0ULL,
    0x276bd88511a0ULL,
    0x977cc2df994f0ULL,
    0xd3f41609d5178ULL,
    0x15bf1d1553244ULL,
    0x305663bfda3d6ULL,
    0x59c913c77693bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0xac00000000000ULL,
    0xd200000000000ULL,
    0xb300000000000ULL,
    0x280000000000ULL,
    0x3840000000000ULL,
    0x63e0000000000ULL,
    0x6f10000000000ULL,
    0x98e8000000000ULL,
    0x3384000000000ULL,
    0x31c6000000000ULL,
    0x5d27000000000ULL,
    0x7ab1800000000ULL,
    0xbd1a400000000ULL,
    0xd9f9600000000ULL,
    0x726c900000000ULL,
    0xb7c0980000000ULL,
    0x1820940000000ULL,
    0x5535020000000ULL,
    0xd6ddbf0000000ULL,
    0x7bdec88000000ULL,
    0xe858af4000000ULL,
    0xc99f802000000ULL,
    0xcb9403000000ULL,
    0xa20ee03800000ULL,
    0x2911d00c00000ULL,
    0x3de8781200000ULL,
    0x2c06442b00000ULL,
    0x1204fa3480000ULL,
    0x5305bb2cc0000ULL,
    0x32855280a0000ULL,
    0x7046844e10000ULL,
    0xcfe3cab8f8000ULL,
    0xbd11106bc4000ULL,
    0x2bef48be3a000ULL,
    0x3102ef70e1000ULL,
    0x986e00671800ULL,
    0x3ec5d00049c00ULL,
    0x15a678002c600ULL,
    0x25f5442486900ULL,
    0xea7b7a105e580ULL,
    0x43acfb154b240ULL,
    0xeae3b28128260ULL,
    0x6293547f1c250ULL,
    0xe82db28b2f408ULL,
    0xf20545c826fcULL,
    0x9b232b444322ULL,
    0x5f9914467aebdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xe800000000000ULL,
    0xb400000000000ULL,
    0x3a00000000000ULL,
    0x9f00000000000ULL,
    0x8580000000000ULL,
    0x3dc0000000000ULL,
    0x38e0000000000ULL,
    0x34f0000000000ULL,
    0x7a28000000000ULL,
    0x7884000000000ULL,
    0x4b42000000000ULL,
    0x96a3000000000ULL,
    0xd8d0800000000ULL,
    0xabf400000000ULL,
    0x9f19a00000000ULL,
    0x6c4fd00000000ULL,
    0x8e31f80000000ULL,
    0xcb4dac0000000ULL,
    0x77b12e0000000ULL,
    0x889a70000000ULL,
    0x1255978000000ULL,
    0x7af9d94000000ULL,
    0x5e3a802000000ULL,
    0xa258401000000ULL,
    0x3eb201800000ULL,
    0xc263900400000ULL,
    0xdcb0d83a00000ULL,
    0xf7093c2d00000ULL,
    0xd093760e80000ULL,
    0xd79cdb27c0000ULL,
    0xbf0fc1a160000ULL,
    0x5495924f70000ULL,
    0x759d19ae38000ULL,
    0x6c0fae5d3c000ULL,
    0xb716efa68a000ULL,
    0x5958354a21000ULL,
    0x746a8e10d0800ULL,
    0xa4a5771ca8c00ULL,
    0xa3d4ef9ab4200ULL,
    0x3d3b354c6fd00ULL,
    0x99da0e1226680ULL,
    0x32a370ea3f40ULL,
    0xac054f9a547e0ULL,
    0x3600e5535f6b0ULL,
    0x9901f6363e4b8ULL,
    0x84809b234f69cULL,
    0xf546e19077e5eULL,
    0xaba5024bcbb65ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xc800000000000ULL,
    0x5400000000000ULL,
    0xf200000000000ULL,
    0x3900000000000ULL,
    0xe980000000000ULL,
    0x44c0000000000ULL,
    0x59e0000000000ULL,
    0x5f10000000000ULL,
    0xe4b8000000000ULL,
    0xb684000000000ULL,
    0x4642000000000ULL,
    0xdda3000000000ULL,
    0x13b3800000000ULL,
    0x3a8e400000000ULL,
    0xf2cea00000000ULL,
    0x8069900000000ULL,
    0x195cc80000000ULL,
    0x9391cc0000000ULL,
    0xaffee60000000ULL,
    0xa2af0000000ULL,
    0xe030088000000ULL,
    0x34e9dc000000ULL,
    0x8a9802000000ULL,
    0x3eb9401000000ULL,
    0xeb87201800000ULL,
    0xf5c0d01c00000ULL,
    0x8463e83200000ULL,
    0x99d51c1500000ULL,
    0x4c5f0e3c80000ULL,
    0x5414b30e40000ULL,
    0x143c86ba60000ULL,
    0xbc46ed130000ULL,
    0xe363a6b678000ULL,
    0x4750bec7c4000ULL,
    0xe1a4e812e000ULL,
    0xa3b2a2e1a1000ULL,
    0xd28cc09b90800ULL,
    0xb6cf51ee68c00ULL,
    0xca6f66326ce00ULL,
    0xbc5fef19e3900ULL,
    0xdc15289053a80ULL,
    0x203d4df66a640ULL,
    0x89c1e804cf320ULL,
    0x12661c2950730ULL,
    0xfad48e1a49b98ULL,
    0xb8def3393dabcULL,
    0xc350268eaa822ULL,
    0x641efec9ee677ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0x8c00000000000ULL,
    0xee00000000000ULL,
    0x7b00000000000ULL,
    0x2880000000000ULL,
    0xf4c0000000000ULL,
    0x9460000000000ULL,
    0xe270000000000ULL,
    0x8628000000000ULL,
    0xe984000000000ULL,
    0x546000000000ULL,
    0xfd21000000000ULL,
    0xe151800000000ULL,
    0x747ac00000000ULL,
    0x6178600000000ULL,
    0xf2fd700000000ULL,
    0xeb8d80000000ULL,
    0x5e1bc40000000ULL,
    0x700fe60000000ULL,
    0x9850830000000ULL,
    0x19f8438000000ULL,
    0x8e3ab94000000ULL,
    0x5edf802000000ULL,
    0xb66fc03000000ULL,
    0xb727e00800000ULL,
    0x4052b00c00000ULL,
    0x55ff381600000ULL,
    0x4039742300000ULL,
    0x15d8de3b80000ULL,
    0xc6edf71ec0000ULL,
    0xcfe69d8a20000ULL,
    0x3a364e7d30000ULL,
    0xcc889d8518000ULL,
    0xee934e489c000ULL,
    0x89f1d898a000ULL,
    0x57c88e4661000ULL,
    0xd0b6fdbf51800ULL,
    0x5d4f3e6048400ULL,
    0x517645bdd4600ULL,
    0x8aa98a5cdeb00ULL,
    0x7417b97fe180ULL,
    0xb822cd404f5c0ULL,
    0xfad6de2916360ULL,
    0x4c38f713eaf10ULL,
    0x3bd91db731f98ULL,
    0x9de98e681920cULL,
    0xd7677d841d90eULL,
    0x96f5fe6ecbee5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x7800000000000ULL,
    0x6400000000000ULL,
    0xf200000000000ULL,
    0xeb00000000000ULL,
    0xc80000000000ULL,
    0x79c0000000000ULL,
    0x6d20000000000ULL,
    0x6fb0000000000ULL,
    0x3dc8000000000ULL,
    0xb984000000000ULL,
    0x1042000000000ULL,
    0x3561000000000ULL,
    0x66d3800000000ULL,
    0xbd1bc00000000ULL,
    0x659f200000000ULL,
    0xf659900000000ULL,
    0x4478580000000ULL,
    0xdd4ce40000000ULL,
    0x6e448e0000000ULL,
    0x2063090000000ULL,
    0x3f568d8000000ULL,
    0xbd5a464000000ULL,
    0x78f9802000000ULL,
    0xfc8ec01000000ULL,
    0x1766a00800000ULL,
    0xc5d7501c00000ULL,
    0x8d9ef81e00000ULL,
    0xfa5bb41900000ULL,
    0xfa7a763c80000ULL,
    0xa848bd3ac0000ULL,
    0x67c4fb8320000ULL,
    0x2826fb5e70000ULL,
    0x3e377bbb48000ULL,
    0x218d3b4bec000ULL,
    0x32e05ba772000ULL,
    0x390ab6261000ULL,
    0x81fa03b210800ULL,
    0xb0d4f5858400ULL,
    0x4a250d9334e00ULL,
    0xbd31864086f00ULL,
    0x610ea010c7c80ULL,
    0xd523500c26640ULL,
    0x6bb4f823f6160ULL,
    0x9fceb40d0f390ULL,
    0x5a83f62e0b238ULL,
    0xc67d0025c24ULL,
    0xdaa25ba05f236ULL,
    0x12f1ab7333d19ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x1c00000000000ULL,
    0x2200000000000ULL,
    0x8700000000000ULL,
    0xac80000000000ULL,
    0x5840000000000ULL,
    0x2fe0000000000ULL,
    0xfab0000000000ULL,
    0x5068000000000ULL,
    0x7044000000000ULL,
    0x6be2000000000ULL,
    0xe4b1000000000ULL,
    0x456d800000000ULL,
    0x43c4c00000000ULL,
    0x8322e00000000ULL,
    0xb110100000000ULL,
    0x1739b80000000ULL,
    0x4599a40000000ULL,
    0xfb4c220000000ULL,
    0x85576f0000000ULL,
    0x62dd6d8000000ULL,
    0xb42fa74000000ULL,
    0x8da2e02000000ULL,
    0x2e50101000000ULL,
    0xb459b80800000ULL,
    0x5769a42c00000ULL,
    0x9cc4220600000ULL,
    0x13a36f0700000ULL,
    0x7b576d8880000ULL,
    0xa7daa761c0000ULL,
    0xfad600b20000ULL,
    0xd165d00610000ULL,
    0x5df6d803f8000ULL,
    0x5f0d7412ac000ULL,
    0x58b77a121a000ULL,
    0x976edb1b11000ULL,
    0xfcc0f79278800ULL,
    0x83a56c58ec400ULL,
    0xd351af9a7b600ULL,
    0xa3d9d856e1300ULL,
    0x31abb58330b80ULL,
    0x7462d36ee8040ULL,
    0x76759a3fd46e0ULL,
    0xabcecb3677690ULL,
    0x2f114fa2ab888ULL,
    0x4238c852b99bcULL,
    0x561f8d8c4c3b6ULL,
    0xa38fb75d2ad9dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xd800000000000ULL,
    0x3400000000000ULL,
    0xc200000000000ULL,
    0x7b00000000000ULL,
    0x2080000000000ULL,
    0xa8c0000000000ULL,
    0x76e0000000000ULL,
    0x5050000000000ULL,
    0xccf8000000000ULL,
    0xfcc4000000000ULL,
    0x84e6000000000ULL,
    0x1357000000000ULL,
    0x287f800000000ULL,
    0x4e02c00000000ULL,
    0xbd07a00000000ULL,
    0xa181100000000ULL,
    0x3744580000000ULL,
    0xe4a7c40000000ULL,
    0xc174e60000000ULL,
    0x8cda70000000ULL,
    0xb3ad5a8000000ULL,
    0xe2bd63c000000ULL,
    0x93e7a02000000ULL,
    0x5d1103000000ULL,
    0xd9bc583800000ULL,
    0x9363c43c00000ULL,
    0xbd12e63600000ULL,
    0x875aa70d00000ULL,
    0x2f32dab080000ULL,
    0x87efa3dec0000ULL,
    0xc298002820000ULL,
    0xf054001a30000ULL,
    0x1cfe0025b8000ULL,
    0x34c3002814000ULL,
    0x98e180053e000ULL,
    0x3d51c03231000ULL,
    0xa57e2011b9800ULL,
    0xd784d01a15c00ULL,
    0x4e4478023fe00ULL,
    0x5f201439b0b00ULL,
    0xb9311e32f9e80ULL,
    0xeec733c74440ULL,
    0xa11a64beef160ULL,
    0xbc11c0c618f10ULL,
    0xd8dfbcb164b98ULL,
    0x5d77c4f6e6a9cULL,
    0xe6cd7abef4b6aULL,
    0xdeaab3e72fe8fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x3800000000000ULL,
    0x400000000000ULL,
    0x8e00000000000ULL,
    0x1300000000000ULL,
    0xa280000000000ULL,
    0xbf40000000000ULL,
    0x2a20000000000ULL,
    0x4ad0000000000ULL,
    0x5448000000000ULL,
    0x2344000000000ULL,
    0xb026000000000ULL,
    0x3fd7000000000ULL,
    0xe9c9800000000ULL,
    0x8c85c00000000ULL,
    0xbc46200000000ULL,
    0x60a3700000000ULL,
    0xf991180000000ULL,
    0xcc6cd40000000ULL,
    0xf091da0000000ULL,
    0xdbeb210000000ULL,
    0x1251ce8000000ULL,
    0x6e08364000000ULL,
    0x6ce6202000000ULL,
    0x6133703000000ULL,
    0xd1f9183800000ULL,
    0x82f8d40c00000ULL,
    0x7fda0e00000ULL,
    0x6f38210100000ULL,
    0x4d1e4ea380000ULL,
    0x3bcaf644c0000ULL,
    0xe5818008a0000ULL,
    0x5bc1c01fd0000ULL,
    0x5a60203288000ULL,
    0x7874701eb4000ULL,
    0x4d8981b12000ULL,
    0xe8a91409d1000ULL,
    0x4a77fa0f89800ULL,
    0x5dd8510b35c00ULL,
    0x3728d692d2600ULL,
    0x74b0e24cf1700ULL,
    0x5239fa0599880ULL,
    0xec9b513a9cdc0ULL,
    0xce0f569376460ULL,
    0x7ce62277ca350ULL,
    0x89305a3e2df68ULL,
    0xddfae1080f084ULL,
    0x30f9eea56613aULL,
    0x99784670637d9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xf800000000000ULL,
    0x9400000000000ULL,
    0x1600000000000ULL,
    0x4700000000000ULL,
    0x5480000000000ULL,
    0x46c0000000000ULL,
    0xa60000000000ULL,
    0xed90000000000ULL,
    0xdd58000000000ULL,
    0x1ac4000000000ULL,
    0xa066000000000ULL,
    0x7095000000000ULL,
    0x5cd8800000000ULL,
    0xa183c00000000ULL,
    0x6b42a00000000ULL,
    0x9325b00000000ULL,
    0x67b2b80000000ULL,
    0xd56d640000000ULL,
    0xa7ea960000000ULL,
    0x8e2ae30000000ULL,
    0x1cca548000000ULL,
    0x259bcec000000ULL,
    0x5ba2a02000000ULL,
    0x2b75b03000000ULL,
    0x520ab82800000ULL,
    0x33f9640400000ULL,
    0x7c34963e00000ULL,
    0xdb2be32500000ULL,
    0x214cd48580000ULL,
    0x44d90ed1c0000ULL,
    0x4586803520000ULL,
    0x3542c021b0000ULL,
    0xa824202a98000ULL,
    0x4937703f64000ULL,
    0x56ae980956000ULL,
    0xa80e1423b1000ULL,
    0x36fa0e2d99800ULL,
    0xd9b5f70de5400ULL,
    0x9e6eda8216200ULL,
    0xc168f9f9d0f00ULL,
    0x71ee5ab848a80ULL,
    0x692f39efad6c0ULL,
    0x584afaa6baae0ULL,
    0x9b5f89c7ea590ULL,
    0xc5c0c2b7e3258ULL,
    0xd0e12deeaff8cULL,
    0xc850f4a384b52ULL,
    0x7dba7ee1c603bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0x7400000000000ULL,
    0xa600000000000ULL,
    0x4100000000000ULL,
    0x8980000000000ULL,
    0xa9c0000000000ULL,
    0xdaa0000000000ULL,
    0xfa70000000000ULL,
    0xf228000000000ULL,
    0xcfc4000000000ULL,
    0xbba2000000000ULL,
    0xe3f3000000000ULL,
    0x23ef800000000ULL,
    0xe960400000000ULL,
    0x93d1a00000000ULL,
    0xf6de300000000ULL,
    0x24a9880000000ULL,
    0x72860c0000000ULL,
    0x347ee0000000ULL,
    0xf5e7650000000ULL,
    0xc5961b8000000ULL,
    0xdcbc3d4000000ULL,
    0x81f9a02000000ULL,
    0x891a301000000ULL,
    0x770b881800000ULL,
    0x15750c3c00000ULL,
    0xea86e2200000ULL,
    0x2987251d00000ULL,
    0x79c7bba980000ULL,
    0xc2a20d5040000ULL,
    0xf670280260000ULL,
    0xa82c3c3a70000ULL,
    0x5cc4662ea8000ULL,
    0xd52669029c000ULL,
    0x82b4759e8a000ULL,
    0xd90c186ef1000ULL,
    0x60739ba768800ULL,
    0x412b7d78bcc00ULL,
    0x314580129be00ULL,
    0xa2e7403c28100ULL,
    0xe51420285c680ULL,
    0x78f970222b8c0ULL,
    0xdc9da83e20620ULL,
    0x9ccf7c2210830ULL,
    0x50d24625d97b8ULL,
    0xc05c193fb5194ULL,
    0x29ee5dad5666eULL,
    0x226724599b1f5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0x2c00000000000ULL,
    0x5200000000000ULL,
    0x5f00000000000ULL,
    0x7680000000000ULL,
    0x5a40000000000ULL,
    0xe360000000000ULL,
    0xadf0000000000ULL,
    0x4a78000000000ULL,
    0x7044000000000ULL,
    0x1862000000000ULL,
    0xd573000000000ULL,
    0xa538800000000ULL,
    0x3ea1400000000ULL,
    0xe753600000000ULL,
    0x2b29900000000ULL,
    0xb26e780000000ULL,
    0xed88f40000000ULL,
    0xb13eb20000000ULL,
    0xa8a00b0000000ULL,
    0xa653578000000ULL,
    0x40ae47c000000ULL,
    0x69ab602000000ULL,
    0xe2d901000000ULL,
    0xf7ec781800000ULL,
    0xf34bf40400000ULL,
    0x471e322a00000ULL,
    0x38b54b0b00000ULL,
    0x379a379480000ULL,
    0xcbf0d7d7c0000ULL,
    0xf37f983da0000ULL,
    0xffc7240690000ULL,
    0xada12a20d8000ULL,
    0x93d72f2f7c000ULL,
    0xe66a7db89e000ULL,
    0x1b8d68d711000ULL,
    0xe03b1db298800ULL,
    0x6b27f8f29cc00ULL,
    0x5195e58cee200ULL,
    0xf10a4ccd38500ULL,
    0x707a37930cd80ULL,
    0x2340d7feb6640ULL,
    0x4ce79838859e0ULL,
    0x6833242fb33d0ULL,
    0xb4db2a11772c8ULL,
    0xb7102f1124c2cULL,
    0x4348fd9b22f5eULL,
    0xdf1b28feafc1fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xe800000000000ULL,
    0xe400000000000ULL,
    0xde00000000000ULL,
    0xbb00000000000ULL,
    0xe880000000000ULL,
    0x2440000000000ULL,
    0x4ba0000000000ULL,
    0x7c30000000000ULL,
    0xfd68000000000ULL,
    0xc244000000000ULL,
    0x8ca6000000000ULL,
    0xa6b7000000000ULL,
    0xe82d800000000ULL,
    0xec63400000000ULL,
    0x6351200000000ULL,
    0x2b9f00000000ULL,
    0x82bc580000000ULL,
    0x42ba040000000ULL,
    0xa2bb020000000ULL,
    0x12bc2d0000000ULL,
    0xfab9f98000000ULL,
    0x1eb84f4000000ULL,
    0xc0b9202000000ULL,
    0x7bbdf03000000ULL,
    0x933a583800000ULL,
    0xb77d042c00000ULL,
    0xfcde823a00000ULL,
    0x80eb6d3900000ULL,
    0x7d86d9b780000ULL,
    0xbfc2bf6ec0000ULL,
    0x3366f81a20000ULL,
    0x95d7b43910000ULL,
    0x7dfbfa2ae8000ULL,
    0x919c99330c000ULL,
    0xf2cf83855a000ULL,
    0xf077964991000ULL,
    0x72cf83b4a9800ULL,
    0x307796776dc00ULL,
    0x92cf83982b600ULL,
    0x8077966e08d00ULL,
    0x7acf83a83c480ULL,
    0x6477967aa27c0ULL,
    0xa4cf838a75160ULL,
    0xdf77966bff810ULL,
    0x4c4f83a427408ULL,
    0xfb37966fd2cb4ULL,
    0x7ef83b9ed1e6ULL,
    0x870796616ac3dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0x3c00000000000ULL,
    0x6e00000000000ULL,
    0x5300000000000ULL,
    0xb280000000000ULL,
    0xfa40000000000ULL,
    0x5060000000000ULL,
    0x5ed0000000000ULL,
    0xcc08000000000ULL,
    0x8044000000000ULL,
    0x8962000000000ULL,
    0x8d53000000000ULL,
    0x89c9800000000ULL,
    0x53e0c00000000ULL,
    0xf313e00000000ULL,
    0xdba8700000000ULL,
    0x9837180000000ULL,
    0x7a9b540000000ULL,
    0x4129320000000ULL,
    0x8676730000000ULL,
    0x18fdae8000000ULL,
    0x12fc944000000ULL,
    0x33fbe02000000ULL,
    0xec7c701000000ULL,
    0x5fbd181800000ULL,
    0xe31c540c00000ULL,
    0x45eab22600000ULL,
    0xba51b30f00000ULL,
    0x494dce9b80000ULL,
    0xc4a32454c0000ULL,
    0xeaf4980ca0000ULL,
    0x69bc942e90000ULL,
    0xac19520c18000ULL,
    0xc969c31bb4000ULL,
    0xbb12d69502000ULL,
    0xfac706f11000ULL,
    0x3237aa19d8800ULL,
    0x4b9de72794c00ULL,
    0x56af7c96d2600ULL,
    0xa1b6976668300ULL,
    0x6e5b56aedcf80ULL,
    0xb50cb07e5e1c0ULL,
    0x43c44a368fc60ULL,
    0x4a5973677d50ULL,
    0xcaf064a0b2cc8ULL,
    0x39b9c37f595ccULL,
    0x41864add50baULL,
    0x6d6dc34c33151ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0x8c00000000000ULL,
    0x9200000000000ULL,
    0x1700000000000ULL,
    0x2680000000000ULL,
    0x9ac0000000000ULL,
    0xbe60000000000ULL,
    0x6bd0000000000ULL,
    0xfe88000000000ULL,
    0x5cc4000000000ULL,
    0xef66000000000ULL,
    0xdc51000000000ULL,
    0xf3ca800000000ULL,
    0xbfe3400000000ULL,
    0x8f92600000000ULL,
    0xce2d900000000ULL,
    0x2f76380000000ULL,
    0x9b38740000000ULL,
    0x519db60000000ULL,
    0x7929e30000000ULL,
    0x99f7a68000000ULL,
    0xc9ff204000000ULL,
    0x33fa602000000ULL,
    0x68f9903000000ULL,
    0xfc78380800000ULL,
    0x21bd741400000ULL,
    0x5159363a00000ULL,
    0x2c4fa32300000ULL,
    0xfea146a480000ULL,
    0xdeb4f045c0000ULL,
    0xe9dab829a0000ULL,
    0xf38a3416b0000ULL,
    0x5145562798000ULL,
    0x5927330ef4000ULL,
    0xdbf3fe85a2000ULL,
    0xf6fac47431000ULL,
    0xb979ee3f59800ULL,
    0x303c0702d4400ULL,
    0xbd1c28bd52a00ULL,
    0xe0eeb75d48d00ULL,
    0xa71070967c980ULL,
    0xd9ef537a7f640ULL,
    0x5c95feb8ff8e0ULL,
    0x66abc4503f1d0ULL,
    0xf8b36e341eed8ULL,
    0xc8df4708ee38cULL,
    0xfc0e48a39749aULL,
    0x38032775b3181ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0x5c00000000000ULL,
    0xe00000000000ULL,
    0xc500000000000ULL,
    0xe080000000000ULL,
    0x5440000000000ULL,
    0x59a0000000000ULL,
    0xb690000000000ULL,
    0x9d8000000000ULL,
    0x4144000000000ULL,
    0x6122000000000ULL,
    0x1ed5000000000ULL,
    0xe7a800000000ULL,
    0xaad0c00000000ULL,
    0xd478e00000000ULL,
    0x5d1700000000ULL,
    0xf3fea80000000ULL,
    0xe212c40000000ULL,
    0x8798c20000000ULL,
    0xb7e77d0000000ULL,
    0xf031fc8000000ULL,
    0xf0cefac000000ULL,
    0x3b5ae02000000ULL,
    0x9e04701000000ULL,
    0xbd04282800000ULL,
    0x4c82041400000ULL,
    0x9240222600000ULL,
    0x58a60d1700000ULL,
    0x417548380000ULL,
    0x96983ef140000ULL,
    0x3d60221820000ULL,
    0x63760d0510000ULL,
    0x1aef54be68000ULL,
    0xa00c3ef9a4000ULL,
    0x74ba220476000ULL,
    0x82370d1751000ULL,
    0x2bcfd4b3c8800ULL,
    0x26ddfee2f5400ULL,
    0x26c2423dbea00ULL,
    0x26e7bd38a4300ULL,
    0x3ab11c88f6380ULL,
    0xc38b8ac9905c0ULL,
    0x817e4800a9aa0ULL,
    0x9d57b43ac5b10ULL,
    0xbbc6a0446b08ULL,
    0x7db4b922a89f4ULL,
    0x3009bea3c4df2ULL,
    0xcb847c4c68ebULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x3800000000000ULL,
    0xfc00000000000ULL,
    0xea00000000000ULL,
    0x6100000000000ULL,
    0x3080000000000ULL,
    0xcb40000000000ULL,
    0xf0a0000000000ULL,
    0xb010000000000ULL,
    0x3158000000000ULL,
    0x7244000000000ULL,
    0xcc26000000000ULL,
    0x9955000000000ULL,
    0xf4fc800000000ULL,
    0xdcd5c00000000ULL,
    0x4139e00000000ULL,
    0xa036500000000ULL,
    0x1209880000000ULL,
    0x7fb9440000000ULL,
    0x4c773a0000000ULL,
    0xc12a150000000ULL,
    0x396ce88000000ULL,
    0xa4481ec000000ULL,
    0xa59fe02000000ULL,
    0x5e23503000000ULL,
    0x6455082800000ULL,
    0x1e7c842400000ULL,
    0xde96da0e00000ULL,
    0x4518453f00000ULL,
    0xcd6360ba80000ULL,
    0x58f45ad840000ULL,
    0xec6c5a2c20000ULL,
    0x4ac88502d0000ULL,
    0x11de009428000ULL,
    0x4103cac804000ULL,
    0x6082322256000ULL,
    0x5346911391000ULL,
    0x9ca232a189800ULL,
    0x62145bda15400ULL,
    0xac5a809f1f200ULL,
    0xa8c20acae5700ULL,
    0x6665d21ee6780ULL,
    0x5971c10849940ULL,
    0x8fa93aa2f4620ULL,
    0x1d29dfeaaf510ULL,
    0x836eda88bc4e8ULL,
    0x5d4e8fed5bc54ULL,
    0xf91dd281921a2ULL,
    0x47670bfc2677bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xf800000000000ULL,
    0x1c00000000000ULL,
    0x3200000000000ULL,
    0xeb00000000000ULL,
    0x7380000000000ULL,
    0x2140000000000ULL,
    0x7de0000000000ULL,
    0x55b0000000000ULL,
    0x3088000000000ULL,
    0x4244000000000ULL,
    0x8a62000000000ULL,
    0xcaf1000000000ULL,
    0xdc6d800000000ULL,
    0x6b73c00000000ULL,
    0xc62ae00000000ULL,
    0xd14900000000ULL,
    0xe6d8d80000000ULL,
    0x96795c0000000ULL,
    0x342a6a0000000ULL,
    0x8612bf0000000ULL,
    0x55b958000000ULL,
    0xff3cc84000000ULL,
    0xadc8e02000000ULL,
    0xfda5901000000ULL,
    0xecd5580800000ULL,
    0x25fa9c2c00000ULL,
    0x75688a3e00000ULL,
    0x6bf22f0700000ULL,
    0x18e94d8c80000ULL,
    0x2bb0947ac0000ULL,
    0xc18d0a3ce0000ULL,
    0xaec5ef1850000ULL,
    0xbea1ad9778000ULL,
    0xab5504796c000ULL,
    0x42b8521222000ULL,
    0x858f730791000ULL,
    0x70c127a618800ULL,
    0xbfa32b647c400ULL,
    0x9fd31fb5fb600ULL,
    0xda7ee7458cf00ULL,
    0x2e29ad8a72b80ULL,
    0x19110450e9240ULL,
    0x10da523f74360ULL,
    0xe37e73065f570ULL,
    0x66aca78a6a1a8ULL,
    0x23d0eb5794efcULL,
    0x1879ffb40f856ULL,
    0x1d2a775012c21ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0x1c00000000000ULL,
    0x5e00000000000ULL,
    0xab00000000000ULL,
    0x8580000000000ULL,
    0x3cc0000000000ULL,
    0xa660000000000ULL,
    0x2370000000000ULL,
    0xbc88000000000ULL,
    0xd3c4000000000ULL,
    0xc1e2000000000ULL,
    0x5ab7000000000ULL,
    0xfc6a800000000ULL,
    0xcdf0c00000000ULL,
    0xa5cae00000000ULL,
    0xf961f00000000ULL,
    0x8cf5d80000000ULL,
    0x414fec0000000ULL,
    0xd127860000000ULL,
    0x2ad1030000000ULL,
    0x101ba38000000ULL,
    0xe6fdb84000000ULL,
    0x27c8e02000000ULL,
    0xac66f01000000ULL,
    0x1277583800000ULL,
    0x300b2c1400000ULL,
    0xcf87662600000ULL,
    0xedc3f30700000ULL,
    0x7ae6fb9780000ULL,
    0xa735946ac0000ULL,
    0x8caf060160000ULL,
    0xad96c31f30000ULL,
    0x31bbc39198000ULL,
    0x9e6c885cdc000ULL,
    0xc8f7d82922000ULL,
    0xa348ec23f1000ULL,
    0x9425061ff8800ULL,
    0xcc55c3286dc00ULL,
    0x2d5b43b87aa00ULL,
    0x3e5f487b4c300ULL,
    0x57ddb8376ab80ULL,
    0x1d1adc0c447c0ULL,
    0xc47abe157f760ULL,
    0xb8881f3f92fb0ULL,
    0xd1c37d8ba9618ULL,
    0xd4e39753c580cULL,
    0x243625a3be48eULL,
    0xdd2fbb44f25e1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x3800000000000ULL,
    0xc400000000000ULL,
    0x6e00000000000ULL,
    0x5900000000000ULL,
    0x3680000000000ULL,
    0xb7c0000000000ULL,
    0xb5e0000000000ULL,
    0xc790000000000ULL,
    0xc2f8000000000ULL,
    0x62c4000000000ULL,
    0xe162000000000ULL,
    0xb55000000000ULL,
    0x9a98800000000ULL,
    0xef15c00000000ULL,
    0xe43c200000000ULL,
    0xada2700000000ULL,
    0x1330480000000ULL,
    0xb74d740000000ULL,
    0xfa4f1e0000000ULL,
    0x7ac91f0000000ULL,
    0xe009d48000000ULL,
    0x256f53c000000ULL,
    0xf03e202000000ULL,
    0x1ba7701000000ULL,
    0x3e30c82800000ULL,
    0xc7ccb40400000ULL,
    0xe8893e0e00000ULL,
    0x53aa6f3100000ULL,
    0xa75b1c9b80000ULL,
    0x3df6e7d640000ULL,
    0xd42f9e2da0000ULL,
    0x3a18df3df0000ULL,
    0x25d7f48578000ULL,
    0xb95823f5e4000ULL,
    0xacf6e81ebe000ULL,
    0xaeafc439b1000ULL,
    0xfdbf60bd8800ULL,
    0x5b33db1095400ULL,
    0xbb4aa2a506200ULL,
    0xd84948e735700ULL,
    0xa1c8a2aff7080ULL,
    0x1d8c48dccc9c0ULL,
    0x572822b1d2120ULL,
    0xf39d88cc925d0ULL,
    0xb19602adb3478ULL,
    0xffaf8c88307cULL,
    0xa246cab39a552ULL,
    0xdba64cd96fa4fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0xf400000000000ULL,
    0x1200000000000ULL,
    0x6700000000000ULL,
    0x180000000000ULL,
    0xe840000000000ULL,
    0xb920000000000ULL,
    0xb6f0000000000ULL,
    0x5c68000000000ULL,
    0x4d44000000000ULL,
    0xe7a2000000000ULL,
    0x8331000000000ULL,
    0x9b0b800000000ULL,
    0x3f94400000000ULL,
    0x9fbda00000000ULL,
    0x69dd900000000ULL,
    0x78eb80000000ULL,
    0x55574c0000000ULL,
    0x8558620000000ULL,
    0x654b990000000ULL,
    0xbbb4ef8000000ULL,
    0x9bce9f4000000ULL,
    0xbe77a02000000ULL,
    0x74a8901000000ULL,
    0x48a7380800000ULL,
    0x16b20c1c00000ULL,
    0x314e420200000ULL,
    0x99b2493d00000ULL,
    0xd4cff78480000ULL,
    0x3bf04359c0000ULL,
    0x86eb7a2060000ULL,
    0x6281452a10000ULL,
    0xb3c235a648000ULL,
    0xe2624a71bc000ULL,
    0x6c532db51a000ULL,
    0x6dd9967e51000ULL,
    0x5d8e779568800ULL,
    0xe65103750c400ULL,
    0xe6df5a0ca2e00ULL,
    0xe2099504f5100ULL,
    0xf712ad8727680ULL,
    0xd778d64f0b640ULL,
    0x3c3a57b019ae0ULL,
    0x8e99d348d4d30ULL,
    0x1b2fc21a16988ULL,
    0x61e3093bf2a64ULL,
    0xfb13d7bb1ddbeULL,
    0x797c934cebb7dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x8800000000000ULL,
    0x5400000000000ULL,
    0x9600000000000ULL,
    0x3100000000000ULL,
    0x7f80000000000ULL,
    0xfec0000000000ULL,
    0x9120000000000ULL,
    0x8550000000000ULL,
    0x24c8000000000ULL,
    0xbdc4000000000ULL,
    0xf1a2000000000ULL,
    0x9915000000000ULL,
    0xaeae800000000ULL,
    0x2c70400000000ULL,
    0x4d98a00000000ULL,
    0xa30db00000000ULL,
    0xe761080000000ULL,
    0x6c30bc0000000ULL,
    0x45fed60000000ULL,
    0x35bef90000000ULL,
    0xc5dd428000000ULL,
    0x5f6fca4000000ULL,
    0x37d2a02000000ULL,
    0xf78cb01000000ULL,
    0x5d25882800000ULL,
    0x2f51fc3400000ULL,
    0xa7caf62200000ULL,
    0x3d46091500000ULL,
    0x9de26aa580000ULL,
    0x4b76864c40000ULL,
    0xe31fde3fe0000ULL,
    0x524e452fb0000ULL,
    0xd083948c48000ULL,
    0xa441335554000ULL,
    0x2667e28b32000ULL,
    0xbb77a6a71000ULL,
    0xc73d2811e8800ULL,
    0xb69c4c0e05400ULL,
    0x6f8bfe3e4ba00ULL,
    0x6126b515ac100ULL,
    0x9d54bcb0ae280ULL,
    0x58cc7f50d76c0ULL,
    0xafc09caa8a420ULL,
    0xdea48f578d2f0ULL,
    0x8397b49ebf358ULL,
    0xb9e9c351cefe4ULL,
    0x7290ca9da6f0aULL,
    0x266a3663e2e29ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x6800000000000ULL,
    0x3c00000000000ULL,
    0x5200000000000ULL,
    0xf700000000000ULL,
    0x6980000000000ULL,
    0x6640000000000ULL,
    0x9160000000000ULL,
    0x20b0000000000ULL,
    0x5258000000000ULL,
    0xf744000000000ULL,
    0xb1e6000000000ULL,
    0xf271000000000ULL,
    0x5ff9800000000ULL,
    0x4f57400000000ULL,
    0xad2fe00000000ULL,
    0x2a3f900000000ULL,
    0x9374380000000ULL,
    0x377f0c0000000ULL,
    0x9196520000000ULL,
    0x9a8adb0000000ULL,
    0x59291d8000000ULL,
    0xb4392ec000000ULL,
    0x7e71e02000000ULL,
    0xa5fa903000000ULL,
    0xa453b80800000ULL,
    0xa6ad4c0c00000ULL,
    0xd37e321a00000ULL,
    0x57970b0f00000ULL,
    0x238d459480000ULL,
    0xa5aff2fdc0000ULL,
    0x24fdea3a60000ULL,
    0xdcd3972990000ULL,
    0x20692fac58000ULL,
    0xb0db25c42c000ULL,
    0x4a8325ae96000ULL,
    0xa1c322c2d1000ULL,
    0x72203210f9800ULL,
    0x1f520b3d5c400ULL,
    0x152ac59f9e600ULL,
    0x6e3db2c945d00ULL,
    0xa5758a2193f80ULL,
    0xc67e470063e40ULL,
    0x611577b6ab0e0ULL,
    0x3049f9d85ec30ULL,
    0x30892f8c24148ULL,
    0xa2b25f482f6cULL,
    0xfbbb258dfa276ULL,
    0xb13722cae69bbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0xec00000000000ULL,
    0xb600000000000ULL,
    0x9500000000000ULL,
    0xef80000000000ULL,
    0x61c0000000000ULL,
    0x5520000000000ULL,
    0x9070000000000ULL,
    0xc348000000000ULL,
    0x36c4000000000ULL,
    0xfda6000000000ULL,
    0x4137000000000ULL,
    0xe32b800000000ULL,
    0xacd1400000000ULL,
    0x2ff9600000000ULL,
    0x332eb00000000ULL,
    0x54d5280000000ULL,
    0xfbfb3c0000000ULL,
    0xb12cee0000000ULL,
    0x33d3d90000000ULL,
    0xdb7b0b8000000ULL,
    0x1c6b364000000ULL,
    0x7db7602000000ULL,
    0x906db03000000ULL,
    0x7bb0a83800000ULL,
    0x4d697c1c00000ULL,
    0x8300e2a00000ULL,
    0x32af293b00000ULL,
    0x2412c3ad80000ULL,
    0xfb5cfa6540000ULL,
    0x139a461be0000ULL,
    0x773da52870000ULL,
    0x1f8885ad48000ULL,
    0xa0615f781c000ULL,
    0x5a12c3bad2000ULL,
    0x325cfa46b1000ULL,
    0x21a460ae9800ULL,
    0x1ffda5190dc00ULL,
    0xbb2885912ae00ULL,
    0x28d15f5444500ULL,
    0x95fac3b936580ULL,
    0x6028fa4697ac0ULL,
    0x8654462b874a0ULL,
    0x99bea5227fcf0ULL,
    0x33cd05af8abb8ULL,
    0x63831f47d5364ULL,
    0x67c6238c0e22eULL,
    0x88240a6bc39d9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0xd400000000000ULL,
    0x2a00000000000ULL,
    0xd300000000000ULL,
    0x6280000000000ULL,
    0xcbc0000000000ULL,
    0x59e0000000000ULL,
    0x6450000000000ULL,
    0x7f8000000000ULL,
    0x8944000000000ULL,
    0x6222000000000ULL,
    0x45b5000000000ULL,
    0x7ad800000000ULL,
    0x7cbac00000000ULL,
    0xec60a00000000ULL,
    0x6f16500000000ULL,
    0x5ade180000000ULL,
    0x4074540000000ULL,
    0x3f4c3e0000000ULL,
    0xdd6c3f0000000ULL,
    0xd8df6a8000000ULL,
    0x7f7073c000000ULL,
    0xb3cd202000000ULL,
    0x87ac901000000ULL,
    0x3cbeb82800000ULL,
    0x4c62042c00000ULL,
    0xdf12263600000ULL,
    0x82d86b3500000ULL,
    0x9473548a80000ULL,
    0x154c4cf4c0000ULL,
    0xe6a4ab8a0000ULL,
    0xba58e3e2f0000ULL,
    0xb4b1981e78000ULL,
    0xea2b942514000ULL,
    0xe3f91e1ffe000ULL,
    0x3b44af3b51000ULL,
    0xc523d2a408800ULL,
    0xbd3777d0ad400ULL,
    0xc76a8633cb600ULL,
    0x93da3b091eb00ULL,
    0x69f74c9dc0280ULL,
    0xe20918dc21940ULL,
    0xd549f49731860ULL,
    0xee6b1cce58150ULL,
    0xaa5bd294258f8ULL,
    0xdcb377cca74fcULL,
    0xe6288609f43aaULL,
    0x1dff3b2a6fecfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0x8400000000000ULL,
    0x7600000000000ULL,
    0x300000000000ULL,
    0xa180000000000ULL,
    0x40000000000ULL,
    0xca0000000000ULL,
    0x46f0000000000ULL,
    0xfd48000000000ULL,
    0x85c4000000000ULL,
    0x6ae2000000000ULL,
    0x5153000000000ULL,
    0xe639800000000ULL,
    0xc2ce400000000ULL,
    0x1282200000000ULL,
    0xa9c2b00000000ULL,
    0xe8e2980000000ULL,
    0xc56cc0000000ULL,
    0xf0be620000000ULL,
    0x5d0bf50000000ULL,
    0x99679f8000000ULL,
    0x3413fe4000000ULL,
    0x501ba02000000ULL,
    0xd9fcf01000000ULL,
    0x5928b81800000ULL,
    0x97507c0c00000ULL,
    0xed3efa1200000ULL,
    0x874e392100000ULL,
    0x54c07d9d80000ULL,
    0xee664b40c0000ULL,
    0x8b961f8860000ULL,
    0x8759be5010000ULL,
    0x7adb803b28000ULL,
    0xd09d400dbc000ULL,
    0x353ba03552000ULL,
    0x5b4cf00c71000ULL,
    0xbec0b81538800ULL,
    0x67647c3594c00ULL,
    0x9314fa0c6e600ULL,
    0xd799393063900ULL,
    0xa2bbfdafe8880ULL,
    0x380b0b7bdcac0ULL,
    0x23e5bfa642a60ULL,
    0x69d14e63d8b30ULL,
    0xe6f93821c5188ULL,
    0x62aa3c3e673d4ULL,
    0x3616da0f6f07eULL,
    0x4d1b8904e3af9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xe800000000000ULL,
    0x1400000000000ULL,
    0x2200000000000ULL,
    0x8d00000000000ULL,
    0x1980000000000ULL,
    0xf7c0000000000ULL,
    0x6920000000000ULL,
    0x1470000000000ULL,
    0x7898000000000ULL,
    0xaa44000000000ULL,
    0xf4e6000000000ULL,
    0x9453000000000ULL,
    0xaf68800000000ULL,
    0x541b400000000ULL,
    0x1802a00000000ULL,
    0x2c04100000000ULL,
    0xbe07e80000000ULL,
    0xab070c0000000ULL,
    0x5e805e0000000ULL,
    0x7742f90000000ULL,
    0xa565db8000000ULL,
    0x79760c000000ULL,
    0x1c4a202000000ULL,
    0x316f503000000ULL,
    0x4f1d481800000ULL,
    0xde871c0400000ULL,
    0xb741b63a00000ULL,
    0xc566f50500000ULL,
    0x1795058880000ULL,
    0xf44ad9e340000ULL,
    0x256b5ba660000ULL,
    0x6d1f20cdf0000ULL,
    0x5380002248000ULL,
    0xaec000311c000ULL,
    0x32a0003c26000ULL,
    0x7eb0002b91000ULL,
    0xe038000fb9800ULL,
    0x5df4000354c00ULL,
    0xc75e00253a200ULL,
    0xa767000bb6d00ULL,
    0x3a96801a28a80ULL,
    0x9dcc4033ed040ULL,
    0x2aac202beffa0ULL,
    0xf83c50054cc30ULL,
    0x71f5c834bf978ULL,
    0x795c5c12557e4ULL,
    0xc63163fbad6eULL,
    0x6412e531f7c83ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x4c00000000000ULL,
    0x2e00000000000ULL,
    0x5900000000000ULL,
    0x5780000000000ULL,
    0x2b40000000000ULL,
    0x14a0000000000ULL,
    0xb410000000000ULL,
    0x3ac8000000000ULL,
    0x32c4000000000ULL,
    0xd6e6000000000ULL,
    0xdf31000000000ULL,
    0xe19a800000000ULL,
    0x26aec00000000ULL,
    0x6b30600000000ULL,
    0x139e700000000ULL,
    0x29ad480000000ULL,
    0x79b0fc0000000ULL,
    0x195efa0000000ULL,
    0x34c350000000ULL,
    0xa084588000000ULL,
    0xe5c6024000000ULL,
    0x4862e02000000ULL,
    0xdbf4b03000000ULL,
    0xc27b280800000ULL,
    0xf01f8c1400000ULL,
    0xf0e9321600000ULL,
    0x6212091300000ULL,
    0xbfcac28b80000ULL,
    0x3344475640000ULL,
    0xb8a3f0b5e0000ULL,
    0xea164e7ad0000ULL,
    0x6bc9322d28000ULL,
    0x7142090904000ULL,
    0x9fa2c290b2000ULL,
    0xbc90474bb1000ULL,
    0x5b0df08839800ULL,
    0x60a34e428c400ULL,
    0x6615b20e06a00ULL,
    0x65cdc9313bb00ULL,
    0x7840229404180ULL,
    0x9024f764339c0ULL,
    0xdbd6d8bc71520ULL,
    0x61acc24a993f0ULL,
    0x8db4800b3c3e8ULL,
    0xb5bc0033e4d4ULL,
    0x7c4ce02b36362ULL,
    0xba01b027c2709ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x6800000000000ULL,
    0xcc00000000000ULL,
    0xae00000000000ULL,
    0x2700000000000ULL,
    0xf780000000000ULL,
    0x98c0000000000ULL,
    0xa7a0000000000ULL,
    0x5fb0000000000ULL,
    0xdba8000000000ULL,
    0x944000000000ULL,
    0x6462000000000ULL,
    0xe991000000000ULL,
    0x7df800000000ULL,
    0x44cf400000000ULL,
    0x4150600000000ULL,
    0xa87e700000000ULL,
    0x77bb80000000ULL,
    0xacfd7c0000000ULL,
    0x223ee60000000ULL,
    0xf69ead0000000ULL,
    0x54ad758000000ULL,
    0x52c5f94000000ULL,
    0x82a7e02000000ULL,
    0xf135301000000ULL,
    0xbe9d80800000ULL,
    0x65a20c3c00000ULL,
    0x66b2de1a00000ULL,
    0x4328913300000ULL,
    0x5a01f3ab80000ULL,
    0x9d042449c0000ULL,
    0xfa86ad9de0000ULL,
    0xda46f57630000ULL,
    0x29e2be01e8000ULL,
    0x7c56e13bec000ULL,
    0xe2fa4ba4ea000ULL,
    0x9539584d51000ULL,
    0xb9184b8898800ULL,
    0x98685850a4400ULL,
    0xff67cbbf97e00ULL,
    0x98171852c3d00ULL,
    0x541fabb65c180ULL,
    0x4aed6844c39c0ULL,
    0xc92613875cee0ULL,
    0xeff11468425f0ULL,
    0x674f75831d398ULL,
    0x9394f97562eb4ULL,
    0x6ad860004c3d6ULL,
    0xd64a7018faee5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x6800000000000ULL,
    0xdc00000000000ULL,
    0x9600000000000ULL,
    0xb900000000000ULL,
    0x1c80000000000ULL,
    0xf40000000000ULL,
    0xefa0000000000ULL,
    0x8030000000000ULL,
    0xba68000000000ULL,
    0xddc4000000000ULL,
    0xade2000000000ULL,
    0x5913000000000ULL,
    0xaa1d800000000ULL,
    0x458f400000000ULL,
    0x3fd0e00000000ULL,
    0x387db00000000ULL,
    0xf85f480000000ULL,
    0x64a8240000000ULL,
    0x4ba0da0000000ULL,
    0x3231ad0000000ULL,
    0xf168b98000000ULL,
    0xea44df4000000ULL,
    0x2525602000000ULL,
    0x5676f01000000ULL,
    0xf94da81800000ULL,
    0x3ab6942c00000ULL,
    0x242a121a00000ULL,
    0x92e2c93700000ULL,
    0x49283a580000ULL,
    0xdddfc26e40000ULL,
    0x376d11a720000ULL,
    0xcb464b53d0000ULL,
    0xeda57203e8000ULL,
    0xa333391c0c000ULL,
    0x51e8abac9a000ULL,
    0xc301166c71000ULL,
    0x1b8063b4f8800ULL,
    0x16c1725f04c00ULL,
    0xda67d9b027600ULL,
    0xb1d52f40f3d00ULL,
    0x157f48113c380ULL,
    0x7ed82426c36c0ULL,
    0x9ce8da0ce5d20ULL,
    0xb585ad3b17090ULL,
    0xebc2b9afaab68ULL,
    0x84e3df7729ab4ULL,
    0xfd92e004edce6ULL,
    0xa15eb0202a27dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x3800000000000ULL,
    0xd400000000000ULL,
    0x1600000000000ULL,
    0xd900000000000ULL,
    0x6a80000000000ULL,
    0xe940000000000ULL,
    0x7320000000000ULL,
    0xe170000000000ULL,
    0xa1b8000000000ULL,
    0x71c4000000000ULL,
    0xdd62000000000ULL,
    0x3dd5000000000ULL,
    0x690a800000000ULL,
    0x691dc00000000ULL,
    0x4f70a00000000ULL,
    0xecbfb00000000ULL,
    0xad45480000000ULL,
    0xbd22140000000ULL,
    0x1c752a0000000ULL,
    0xb53f090000000ULL,
    0xad84b38000000ULL,
    0x6c431c000000ULL,
    0xfae2202000000ULL,
    0x816701000000ULL,
    0x7a6fe82800000ULL,
    0xa9cca41400000ULL,
    0x32f8e20e00000ULL,
    0x59e5dd3500000ULL,
    0x3993b98580000ULL,
    0x96ad48f640000ULL,
    0xd0297bbaa0000ULL,
    0x1b6ee5ea50000ULL,
    0x574d2a14c8000ULL,
    0x6dbb093c5c000ULL,
    0xa3c6b38e6e000ULL,
    0x8a6131fd71000ULL,
    0xa50a01cd8800ULL,
    0xc4cfb01c35400ULL,
    0xfe7d482d62a00ULL,
    0xa1a6140257700ULL,
    0x9c372a1bb4280ULL,
    0x649a090c23ec0ULL,
    0x19b6339a77520ULL,
    0x2e5df1fd25850ULL,
    0x715080228bca8ULL,
    0xd14cc0141b824ULL,
    0x6cb82032b30ceULL,
    0xed477038df3c7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x8800000000000ULL,
    0xd400000000000ULL,
    0xce00000000000ULL,
    0xfb00000000000ULL,
    0xb880000000000ULL,
    0xc340000000000ULL,
    0xa4e0000000000ULL,
    0xe850000000000ULL,
    0x8ab8000000000ULL,
    0x69c4000000000ULL,
    0x46a6000000000ULL,
    0xa133000000000ULL,
    0xb22a800000000ULL,
    0xb458400000000ULL,
    0x7494a00000000ULL,
    0xef9b700000000ULL,
    0xe332580000000ULL,
    0x1b2d040000000ULL,
    0x8ddbfa0000000ULL,
    0x6a55770000000ULL,
    0x43bfaa8000000ULL,
    0x4569c000000ULL,
    0xd066202000000ULL,
    0xd917303000000ULL,
    0x9f58f81800000ULL,
    0x8411741400000ULL,
    0x98dd222200000ULL,
    0xd9d7333500000ULL,
    0x807c70b380000ULL,
    0x6be02efec0000ULL,
    0xed5f28e20000ULL,
    0x5afc6dc0d0000ULL,
    0xf1a3da1138000ULL,
    0x2bb5471e14000ULL,
    0xa06bd298ae000ULL,
    0x453f5dfb71000ULL,
    0x4005222029800ULL,
    0xa00333138cc00ULL,
    0x300270a92aa00ULL,
    0xd8072ef706100ULL,
    0x5c0172983d280ULL,
    0x1a032df422dc0ULL,
    0x3503fa38da960ULL,
    0x4381773c6e410ULL,
    0x7bc1aa84517e8ULL,
    0x67a269f9789dcULL,
    0x4cb2a03574caaULL,
    0x62e870049f8a7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x800000000000ULL,
    0x9c00000000000ULL,
    0x3200000000000ULL,
    0xeb00000000000ULL,
    0xa080000000000ULL,
    0xc140000000000ULL,
    0x1f60000000000ULL,
    0xc270000000000ULL,
    0xc7e8000000000ULL,
    0x37c4000000000ULL,
    0x8326000000000ULL,
    0x9095000000000ULL,
    0xf15e800000000ULL,
    0x5f8c400000000ULL,
    0x7f36e00000000ULL,
    0x5a8a900000000ULL,
    0x66b0d80000000ULL,
    0xa04dc40000000ULL,
    0xc013aa0000000ULL,
    0x281d0b0000000ULL,
    0x74e99b8000000ULL,
    0xc344034000000ULL,
    0x4c60602000000ULL,
    0x86f2d03000000ULL,
    0x70a8382800000ULL,
    0x65a6543400000ULL,
    0xa4d3f20200000ULL,
    0xefbd8f2700000ULL,
    0x533c518c80000ULL,
    0x607ed87ac0000ULL,
    0x341f438820000ULL,
    0x86ecc74050000ULL,
    0x88454a0fd8000ULL,
    0x3ce79b349c000ULL,
    0x4fb1439bfa000ULL,
    0xf3cdc75ef1000ULL,
    0x95d5ca0e49800ULL,
    0x883adb09e5400ULL,
    0x78ff23b0f7a00ULL,
    0x115e176973100ULL,
    0xef8d721235b80ULL,
    0x731cf016ea40ULL,
    0x1e8ab18b0e360ULL,
    0xc0b4485bbe710ULL,
    0xe54f9bab176a8ULL,
    0xb99103448302cULL,
    0xa2dee002fc46eULL,
    0xa4e900d4b50dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xf800000000000ULL,
    0x1400000000000ULL,
    0x3200000000000ULL,
    0x1d00000000000ULL,
    0xa280000000000ULL,
    0x6040000000000ULL,
    0xa420000000000ULL,
    0x7f0000000000ULL,
    0x9308000000000ULL,
    0xffc4000000000ULL,
    0xf6e2000000000ULL,
    0xeb97000000000ULL,
    0xecd9800000000ULL,
    0xb53bc00000000ULL,
    0xa12ea00000000ULL,
    0x6f30900000000ULL,
    0x2e6c680000000ULL,
    0xd490540000000ULL,
    0x3b5fe20000000ULL,
    0x4bf8d10000000ULL,
    0x8b4e678000000ULL,
    0xfe6344000000ULL,
    0x2315202000000ULL,
    0x559c501000000ULL,
    0x499b483800000ULL,
    0x379b040c00000ULL,
    0xdc9d2a3e00000ULL,
    0x9918150500000ULL,
    0xaadded8c80000ULL,
    0xea3eb14740000ULL,
    0x6aca588a0000ULL,
    0x89f6b54810000ULL,
    0x200a0fb108000ULL,
    0x7e42605dfc000ULL,
    0x3f20c202c2000ULL,
    0x5a778126f1000ULL,
    0x84ceafb738800ULL,
    0x7521f074a5c00ULL,
    0x13772a0116600ULL,
    0xf44b15271ef00ULL,
    0x38666d9d63a80ULL,
    0xedd2714960240ULL,
    0x87fb85b0f11a0ULL,
    0x3d4de54639150ULL,
    0x8e0c7aa25788ULL,
    0x4096a44d56f44ULL,
    0x4959480e3f79eULL,
    0xb6fc042ab3bd1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x7800000000000ULL,
    0x6c00000000000ULL,
    0x1e00000000000ULL,
    0x5500000000000ULL,
    0x7780000000000ULL,
    0x3cc0000000000ULL,
    0xcc20000000000ULL,
    0xd910000000000ULL,
    0x51e8000000000ULL,
    0xce44000000000ULL,
    0x4f62000000000ULL,
    0x9df5000000000ULL,
    0xc6df800000000ULL,
    0x75bfc00000000ULL,
    0x9e4d600000000ULL,
    0x4993f00000000ULL,
    0xb1aa280000000ULL,
    0x4c26fc0000000ULL,
    0x9914c60000000ULL,
    0xf1eb310000000ULL,
    0x3e44708000000ULL,
    0x3762ab4000000ULL,
    0xf1f0e02000000ULL,
    0xd8d9301000000ULL,
    0x20b8c82800000ULL,
    0xe9cacc3c00000ULL,
    0x75538e1e00000ULL,
    0x7d8e3d1b00000ULL,
    0x95329e8780000ULL,
    0xc8fb665540000ULL,
    0x3faa56bde0000ULL,
    0x7121aa5f30000ULL,
    0xaa91d8bb08000ULL,
    0x372b975a44000ULL,
    0xad6146027a000ULL,
    0xbef5f10491000ULL,
    0xa05c90a258800ULL,
    0xc4ff9b553d400ULL,
    0x31aa2815d7e00ULL,
    0xc26fc0c1ff00ULL,
    0x3914c626fb580ULL,
    0x1eb310250fc0ULL,
    0x46447088f88a0ULL,
    0x5b62ab42ecbf0ULL,
    0xeff0e0356fb18ULL,
    0x8dd9301b928c4ULL,
    0x5738c80c647c2ULL,
    0xd50acc376b1adULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0x4c00000000000ULL,
    0xe200000000000ULL,
    0xd00000000000ULL,
    0x4480000000000ULL,
    0x2940000000000ULL,
    0x320000000000ULL,
    0xb8d0000000000ULL,
    0xebe8000000000ULL,
    0xc8c4000000000ULL,
    0xd2e6000000000ULL,
    0x58b3000000000ULL,
    0x2119800000000ULL,
    0x757e400000000ULL,
    0x9988600000000ULL,
    0xe736100000000ULL,
    0x37dce80000000ULL,
    0xa09ee40000000ULL,
    0x94baea0000000ULL,
    0x31ee290000000ULL,
    0x61c05e8000000ULL,
    0x6864c34000000ULL,
    0xe6f7e02000000ULL,
    0x6fbb503000000ULL,
    0xe6d081800000ULL,
    0xb706b40c00000ULL,
    0xdd86621200000ULL,
    0xdbc2dd1300000ULL,
    0xf1605cb880000ULL,
    0x14770e4340000ULL,
    0x9dfcd4b120000ULL,
    0xa2cbfa7a50000ULL,
    0xc114d6b8c8000ULL,
    0xb78c375234000ULL,
    0xc831e210fa000ULL,
    0x1e5b9d1d31000ULL,
    0xfd5fbc8639800ULL,
    0xf6d85e4a6cc00ULL,
    0xc61fdc93e6600ULL,
    0xd0fa4e774f900ULL,
    0x464d34b70a180ULL,
    0xb853aa62e9840ULL,
    0xcca85e8ce53a0ULL,
    0x74e0c35172b90ULL,
    0x5bb1e039253a8ULL,
    0xda98502a124a4ULL,
    0x6dbc883dd5f7aULL,
    0xf36cf4354b60dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0x5400000000000ULL,
    0x600000000000ULL,
    0x6100000000000ULL,
    0xd080000000000ULL,
    0x2dc0000000000ULL,
    0x96a0000000000ULL,
    0x59d0000000000ULL,
    0x7b8000000000ULL,
    0x8844000000000ULL,
    0x8de6000000000ULL,
    0x93b1000000000ULL,
    0x7049800000000ULL,
    0xf7eec00000000ULL,
    0xc6b8a00000000ULL,
    0xa8c3300000000ULL,
    0xd825880000000ULL,
    0x6114c40000000ULL,
    0x77988e0000000ULL,
    0xc550e50000000ULL,
    0x987d568000000ULL,
    0x69e081c000000ULL,
    0xdb7202000000ULL,
    0x654cf03000000ULL,
    0x116ca80800000ULL,
    0xd27d340c00000ULL,
    0xbae3a61600000ULL,
    0xaa36111500000ULL,
    0x9109d08180000ULL,
    0x768a60d840000ULL,
    0x62ca589420000ULL,
    0x152fa4fb70000ULL,
    0xcd9b568da8000ULL,
    0x6e5181ea74000ULL,
    0x5bfea03fee000ULL,
    0xc3a2300b11000ULL,
    0x5f54083cf9800ULL,
    0x37e0425ac400ULL,
    0xf2662e3fb2600ULL,
    0xf3f2d53bcbb00ULL,
    0x31a95ea1a6280ULL,
    0x165e85ef74cc0ULL,
    0xe1f10e066f620ULL,
    0xb6ae251750310ULL,
    0xb7ddf692d9a38ULL,
    0x74b7b1eddd794ULL,
    0x1cca82e9ab5aULL,
    0x2aad34047ed07ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x8800000000000ULL,
    0x5400000000000ULL,
    0xfa00000000000ULL,
    0x3700000000000ULL,
    0x2180000000000ULL,
    0x60c0000000000ULL,
    0x3020000000000ULL,
    0x9130000000000ULL,
    0x1bb8000000000ULL,
    0x9244000000000ULL,
    0x4362000000000ULL,
    0x42d7000000000ULL,
    0x5129800000000ULL,
    0xc708400000000ULL,
    0x8f3ca00000000ULL,
    0x9806d00000000ULL,
    0x4c05380000000ULL,
    0xf603cc0000000ULL,
    0x2104660000000ULL,
    0x3080b10000000ULL,
    0xd840418000000ULL,
    0xbc67b9c000000ULL,
    0xd757202000000ULL,
    0xfbe9901000000ULL,
    0x4828183800000ULL,
    0x6b895c0c00000ULL,
    0x197ffe2200000ULL,
    0xd962ad1500000ULL,
    0x5d09fbe80000ULL,
    0x18a884cdc0000ULL,
    0xc3cda7a860000ULL,
    0xcd1f48c830000ULL,
    0x6a33c19408000ULL,
    0x8c3cf9f84c000ULL,
    0x3380003cee000ULL,
    0x73c0002d91000ULL,
    0x83a0003458800ULL,
    0xa2f0000475c00ULL,
    0x78180000aa600ULL,
    0xb4003132100ULL,
    0xb37a001927280ULL,
    0x16630017bdb40ULL,
    0x18538011874e0ULL,
    0xe66b40256df30ULL,
    0xb6ef20127f998ULL,
    0x1ead901c48ec4ULL,
    0xcaca18346cf06ULL,
    0x799e5c2aff3e7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0x2c00000000000ULL,
    0xce00000000000ULL,
    0xdf00000000000ULL,
    0x7480000000000ULL,
    0xa040000000000ULL,
    0x10a0000000000ULL,
    0x19f0000000000ULL,
    0x2888000000000ULL,
    0x4fc4000000000ULL,
    0xe666000000000ULL,
    0x1815000000000ULL,
    0xda5a800000000ULL,
    0x427b400000000ULL,
    0x1dcb600000000ULL,
    0x76e1700000000ULL,
    0xca57780000000ULL,
    0x93fb640000000ULL,
    0x8c0da20000000ULL,
    0xc782d50000000ULL,
    0x7ac2278000000ULL,
    0x9fe5e84000000ULL,
    0xc5d7e02000000ULL,
    0x553f303000000ULL,
    0x95ee982800000ULL,
    0xa155541400000ULL,
    0x7d7fba3a00000ULL,
    0x5949c10b00000ULL,
    0xcea47db380000ULL,
    0x4ef21977c0000ULL,
    0x800985bd20000ULL,
    0x99823d5810000ULL,
    0xedc7478c28000ULL,
    0x976598627c000ULL,
    0x4394183822000ULL,
    0xb69e143cf1000ULL,
    0x369cda1819800ULL,
    0xf69cb12ec5400ULL,
    0x569d05a236a00ULL,
    0x6987d444ed00ULL,
    0xee98a78dfad80ULL,
    0xc29ea874145c0ULL,
    0xc9c8029bfde0ULL,
    0xd39e403d63d90ULL,
    0xa719e00690e88ULL,
    0x75e303d68f54ULL,
    0x17fa18121da9eULL,
    0xe0f141e53ea1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x800000000000ULL,
    0x2400000000000ULL,
    0xa00000000000ULL,
    0xbd00000000000ULL,
    0x7580000000000ULL,
    0xfbc0000000000ULL,
    0xaa60000000000ULL,
    0xf910000000000ULL,
    0x1bc8000000000ULL,
    0x3f44000000000ULL,
    0x1a26000000000ULL,
    0xc7b3000000000ULL,
    0x1c3d800000000ULL,
    0x75dc400000000ULL,
    0xfc0b200000000ULL,
    0x7321500000000ULL,
    0x1035680000000ULL,
    0xaefd6c0000000ULL,
    0x183f3e0000000ULL,
    0x6fde630000000ULL,
    0x990a508000000ULL,
    0x9aa24a4000000ULL,
    0xcdf0a02000000ULL,
    0x979e103000000ULL,
    0x23abc81800000ULL,
    0x47547c2c00000ULL,
    0x826f760200000ULL,
    0x28355f0900000ULL,
    0x42fb868280000ULL,
    0x563e056f40000ULL,
    0x68db6ebd60000ULL,
    0x598b294ef0000ULL,
    0x30e1709298000ULL,
    0x96531a6244000ULL,
    0x79edc83cf2000ULL,
    0xb4f77c1ad1000ULL,
    0x981af61e09800ULL,
    0xd6d1f3bacc00ULL,
    0xcb6a69aef600ULL,
    0x82bc5545c7100ULL,
    0x9b86aa7ac80ULL,
    0xfb2e05433c540ULL,
    0x11136e8d075a0ULL,
    0x4fcf2978da5b0ULL,
    0x5d477096ec4f8ULL,
    0x33201a4e8e58cULL,
    0xb0304819b6742ULL,
    0x7efb3c3fa2429ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0xb800000000000ULL,
    0x4c00000000000ULL,
    0x2600000000000ULL,
    0xb700000000000ULL,
    0x8680000000000ULL,
    0x64c0000000000ULL,
    0xcd60000000000ULL,
    0xde70000000000ULL,
    0xa228000000000ULL,
    0x7144000000000ULL,
    0xdd26000000000ULL,
    0xdad7000000000ULL,
    0x8ebc800000000ULL,
    0xac59c00000000ULL,
    0x4b68600000000ULL,
    0x9064300000000ULL,
    0x9f0380000000ULL,
    0x176a740000000ULL,
    0xce64060000000ULL,
    0x12f6fb0000000ULL,
    0x27ef1b8000000ULL,
    0xa5a52d4000000ULL,
    0x1512e02000000ULL,
    0xbb5af03000000ULL,
    0x7decd83800000ULL,
    0x3ca3842400000ULL,
    0xa0925e2e00000ULL,
    0x1b1bbf1300000ULL,
    0xb149a58980000ULL,
    0x7433626dc0000ULL,
    0xaf8b1d81a0000ULL,
    0x5053d66930000ULL,
    0xec7dfbab58000ULL,
    0xa63fdd639c000ULL,
    0x3b9e381e8a000ULL,
    0xa289741b51000ULL,
    0xdfd68628c9800ULL,
    0x853c3b2c75c00ULL,
    0x871dfba58f200ULL,
    0xf4fdd4ce6700ULL,
    0xff363829a2180ULL,
    0x270d7417750c0ULL,
    0x779086000e0e0ULL,
    0xcd9b3b13279d0ULL,
    0x8d897bb102818ULL,
    0x65521d4ac57ecULL,
    0xeff8581c1766eULL,
    0x97e44355bfb5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xe800000000000ULL,
    0xb400000000000ULL,
    0x2200000000000ULL,
    0x4500000000000ULL,
    0xed80000000000ULL,
    0x2540000000000ULL,
    0xa160000000000ULL,
    0xa810000000000ULL,
    0xee48000000000ULL,
    0x9bc4000000000ULL,
    0xcea6000000000ULL,
    0x64b7000000000ULL,
    0x1ff9800000000ULL,
    0xb1bb400000000ULL,
    0x65fa00000000ULL,
    0x558c100000000ULL,
    0x8566a80000000ULL,
    0x7210ac0000000ULL,
    0x274d4a0000000ULL,
    0x38437b0000000ULL,
    0xd0e2988000000ULL,
    0xfb51864000000ULL,
    0x1828202000000ULL,
    0x7354503000000ULL,
    0x5c2e883800000ULL,
    0x5954fc0c00000ULL,
    0x9d2bc23a00000ULL,
    0x7ed3872d00000ULL,
    0x496f5a8880000ULL,
    0x1035015140000ULL,
    0x813efa9b60000ULL,
    0xb29a117950000ULL,
    0x97afd2b058000ULL,
    0x3312fd5604000ULL,
    0xc0cab89992000ULL,
    0xc05d677f1000ULL,
    0x8e06a81929800ULL,
    0xdb00ac196dc00ULL,
    0xe854a361e600ULL,
    0x47c77b157ed00ULL,
    0x98a498a02fe80ULL,
    0xe3b6864a77040ULL,
    0x8779a02173aa0ULL,
    0x917b1038612b0ULL,
    0x367f282ad0d28ULL,
    0x7efbec1e981ecULL,
    0x863aea29a5462ULL,
    0xea1b6b2a62719ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xd800000000000ULL,
    0x9400000000000ULL,
    0x5200000000000ULL,
    0x4900000000000ULL,
    0x7e80000000000ULL,
    0xfa40000000000ULL,
    0x35e0000000000ULL,
    0x4570000000000ULL,
    0xf678000000000ULL,
    0x5bc4000000000ULL,
    0x8226000000000ULL,
    0x1157000000000ULL,
    0xf429800000000ULL,
    0xe46ac00000000ULL,
    0xfc8ea00000000ULL,
    0xad7f900000000ULL,
    0x4c46c80000000ULL,
    0xd6e5b40000000ULL,
    0x36f3320000000ULL,
    0xc8b8df0000000ULL,
    0x9164938000000ULL,
    0x4a363fc000000ULL,
    0x4319202000000ULL,
    0x5bf6503000000ULL,
    0xfc3fe83800000ULL,
    0x5623e40c00000ULL,
    0x6354da3600000ULL,
    0x6d2f3b2500000ULL,
    0x72ee499480000ULL,
    0x4aca04d240000ULL,
    0x5e98e9bfa0000ULL,
    0x123194ce90000ULL,
    0x1718219578000ULL,
    0xe9f320ed5c000ULL,
    0x853a93939e000ULL,
    0xf0a53fcff1000ULL,
    0xd16a01a09800ULL,
    0xacb900f15c00ULL,
    0x7e98c8382a600ULL,
    0xc236b43ccab00ULL,
    0xff1cb20f7ba80ULL,
    0xa5f51f33d3e40ULL,
    0x433bb3a8d7b20ULL,
    0xeba76fdec46d0ULL,
    0x3a974822734c8ULL,
    0x8e0b743f46f7cULL,
    0xb13b92313cc4eULL,
    0xb2a04f246ffffULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x1800000000000ULL,
    0x1c00000000000ULL,
    0xae00000000000ULL,
    0x8f00000000000ULL,
    0x9080000000000ULL,
    0x7bc0000000000ULL,
    0x4ee0000000000ULL,
    0xbc30000000000ULL,
    0xc038000000000ULL,
    0x1e44000000000ULL,
    0xc0a2000000000ULL,
    0xc497000000000ULL,
    0xa8aa800000000ULL,
    0x50ecc00000000ULL,
    0x1b4ee00000000ULL,
    0xfd5c700000000ULL,
    0xf37f80000000ULL,
    0xebbc40000000ULL,
    0x42827e0000000ULL,
    0x4ac7a70000000ULL,
    0xa963c98000000ULL,
    0x1b71edc000000ULL,
    0x431e602000000ULL,
    0x7f93b01000000ULL,
    0x8229983800000ULL,
    0xee28741400000ULL,
    0xa82be60600000ULL,
    0x932fd30700000ULL,
    0xf9a82fab80000ULL,
    0xa76e3ee3c0000ULL,
    0x440e4f8420000ULL,
    0x32798ecef0000ULL,
    0xd8e5d78bb8000ULL,
    0xdf36faeb0c000ULL,
    0x56bcb1ae0e000ULL,
    0xbe81e9c491000ULL,
    0xb4c07e05a8800ULL,
    0x3e60a711e5c00ULL,
    0x97f1499b8aa00ULL,
    0x96d92dfa0b300ULL,
    0xbe72803ccbb80ULL,
    0xae98c03a6b1c0ULL,
    0x55d4e017dbfe0ULL,
    0xf88f700203f10ULL,
    0xefbf78211e1f8ULL,
    0xfd000412f929cULL,
    0x11861e386cd26ULL,
    0xd44717289f8b7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x2800000000000ULL,
    0xb400000000000ULL,
    0x2200000000000ULL,
    0xa700000000000ULL,
    0xb80000000000ULL,
    0xf3c0000000000ULL,
    0x6420000000000ULL,
    0x7b70000000000ULL,
    0xd588000000000ULL,
    0x7424000000000ULL,
    0x3372000000000ULL,
    0x9189000000000ULL,
    0xae22800000000ULL,
    0x873400000000ULL,
    0xc0ca00000000ULL,
    0xd8e3900000000ULL,
    0xc0d6780000000ULL,
    0x8f3cfc0000000ULL,
    0x9a8c0e0000000ULL,
    0xaba1590000000ULL,
    0x12b5278000000ULL,
    0xaaafa24000000ULL,
    0x42d6782000000ULL,
    0xb83cfc1000000ULL,
    0x990c0e0800000ULL,
    0x3c61593400000ULL,
    0x7c95278a00000ULL,
    0xc2dfa26d00000ULL,
    0xbede782880000ULL,
    0x98d8fc39c0000ULL,
    0xc5de0e0ae0000ULL,
    0x25585908f0000ULL,
    0x631fa79308000ULL,
    0xc5f8e273dc000ULL,
    0x5428d81de2000ULL,
    0xa5966c24c9000ULL,
    0x9858f6063c800ULL,
    0x339ee52c92400ULL,
    0x5bbd89b880a00ULL,
    0xbec96b71c0d00ULL,
    0x8a47279ee1280ULL,
    0x58e6a252f1e40ULL,
    0x80d4f816091e0ULL,
    0xaf3fbc1f5d7f0ULL,
    0x4a88ae1623a38ULL,
    0x83a6c92f28864ULL,
    0xa6b15fb04c61eULL,
    0x88aa5e455a0c9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0x5400000000000ULL,
    0x4e00000000000ULL,
    0x1700000000000ULL,
    0x6a80000000000ULL,
    0x6b40000000000ULL,
    0xa2a0000000000ULL,
    0x80b0000000000ULL,
    0x628000000000ULL,
    0xdaa4000000000ULL,
    0x34b6000000000ULL,
    0xd829000000000ULL,
    0x45a1800000000ULL,
    0x3237c00000000ULL,
    0xa96ba00000000ULL,
    0xbe03f00000000ULL,
    0xcf07780000000ULL,
    0xae80f40000000ULL,
    0xad44aa0000000ULL,
    0xd9a46d0000000ULL,
    0xf035f18000000ULL,
    0x34691b4000000ULL,
    0x587782000000ULL,
    0xb5c0f43000000ULL,
    0x17e4aa0800000ULL,
    0xbd146d2c00000ULL,
    0x801df18e00000ULL,
    0xadcd1b5500000ULL,
    0x15b1783380000ULL,
    0x11a9f435c0000ULL,
    0x9a652a12a0000ULL,
    0x64d3ad36d0000ULL,
    0x8dfe51a6a8000ULL,
    0x49daeb752c000ULL,
    0xe82800320a000ULL,
    0xbda4000369000ULL,
    0x4636001f8d800ULL,
    0x57690000da400ULL,
    0x91018037c0600ULL,
    0xf187c039a1f00ULL,
    0x8bc3a03850e80ULL,
    0x18e7f01ce9fc0ULL,
    0x339178244c5e0ULL,
    0x9d59f4077a7d0ULL,
    0x4c6d2a12914a8ULL,
    0xb187ad1ac8eb4ULL,
    0x6bc05197dd946ULL,
    0x88e7eb5433badULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x1800000000000ULL,
    0xe400000000000ULL,
    0xf200000000000ULL,
    0x8500000000000ULL,
    0xe80000000000ULL,
    0xb1c0000000000ULL,
    0xf520000000000ULL,
    0x9c10000000000ULL,
    0x5e8000000000ULL,
    0x3d24000000000ULL,
    0x3016000000000ULL,
    0x9beb000000000ULL,
    0x4623800000000ULL,
    0xb596c00000000ULL,
    0xb7ac200000000ULL,
    0x7b44100000000ULL,
    0xe6e2e80000000ULL,
    0x6434540000000ULL,
    0x9b7d9e0000000ULL,
    0x870f410000000ULL,
    0x6b11b48000000ULL,
    0x4e6f314000000ULL,
    0xe262e82000000ULL,
    0x84f4543000000ULL,
    0xeadd9e1800000ULL,
    0x3bdf413c00000ULL,
    0x7f59b48600000ULL,
    0x3f9b317900000ULL,
    0xde3ce81c80000ULL,
    0x8aeb541140000ULL,
    0x62a01e1ba0000ULL,
    0x556811070000ULL,
    0x4e0814bb48000ULL,
    0xcb96e15e04000ULL,
    0xfcafa01dfa000ULL,
    0x86c2d01e09000ULL,
    0xdea6c817a5800ULL,
    0xb35444368ac00ULL,
    0xa909762ac0e00ULL,
    0xa810153361b00ULL,
    0xbfefaa9011080ULL,
    0xd426b070d8040ULL,
    0xc0977c961d3a0ULL,
    0xa12b756387d50ULL,
    0x2e839e321f878ULL,
    0xe1c04127a2604ULL,
    0x7d2434a855652ULL,
    0x9012f14703c85ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0xdc00000000000ULL,
    0xe600000000000ULL,
    0x8900000000000ULL,
    0x7480000000000ULL,
    0xa840000000000ULL,
    0x7b20000000000ULL,
    0xb210000000000ULL,
    0xb8b8000000000ULL,
    0x4724000000000ULL,
    0x4412000000000ULL,
    0x9bb000000000ULL,
    0xbfa6800000000ULL,
    0x5250c00000000ULL,
    0x279de00000000ULL,
    0x9f31b00000000ULL,
    0xcbac880000000ULL,
    0x6f1a440000000ULL,
    0xc176f20000000ULL,
    0x18c510000000ULL,
    0x258d548000000ULL,
    0x6f89f7c000000ULL,
    0x688c882000000ULL,
    0x410a441000000ULL,
    0xffcef21800000ULL,
    0x9fa8511400000ULL,
    0x4d1f549600000ULL,
    0x1272f7f700000ULL,
    0x4a0a081980000ULL,
    0x284a843240000ULL,
    0x146b120520000ULL,
    0xeffde13e10000ULL,
    0xb981dc88c8000ULL,
    0xc6c3b3db84000ULL,
    0x8c627a37ae000ULL,
    0x3cb2152389000ULL,
    0x5269a69424800ULL,
    0x16fea6fc7ec00ULL,
    0xa5075c8721a00ULL,
    0xba8373df10300ULL,
    0x95479a2649780ULL,
    0x3da7a510456c0ULL,
    0xf1572eb0cfa20ULL,
    0x41fe2d0b8510ULL,
    0x46f72e8efc1c8ULL,
    0xd24fe2fd33244ULL,
    0x5b6f2eb20a2d2ULL,
    0x227be2e1b711fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0xb400000000000ULL,
    0x200000000000ULL,
    0xf100000000000ULL,
    0xa480000000000ULL,
    0x50c0000000000ULL,
    0x73a0000000000ULL,
    0xb050000000000ULL,
    0x6818000000000ULL,
    0xb7a4000000000ULL,
    0x1a52000000000ULL,
    0xcd1f000000000ULL,
    0xe123800000000ULL,
    0x5395400000000ULL,
    0xae3aa00000000ULL,
    0x3b3900000000ULL,
    0xb92ac80000000ULL,
    0xd4b840000000ULL,
    0x213b160000000ULL,
    0xb032550000000ULL,
    0x466f868000000ULL,
    0xd7aa16c000000ULL,
    0x488ac82000000ULL,
    0xc1b841000000ULL,
    0xda3163800000ULL,
    0xa756553c00000ULL,
    0xc79d86ba00000ULL,
    0x1ee516ed00000ULL,
    0xc3b1482080000ULL,
    0x192ac42c40000ULL,
    0x1d4bb61120000ULL,
    0x393ac52830000ULL,
    0xec34cea6e8000ULL,
    0xf06bd2c114000ULL,
    0x24a8fe3a86000ULL,
    0x1d0f0101a9000ULL,
    0xf85cf897b4800ULL,
    0x2ec457db77c00ULL,
    0x64a690bea0e00ULL,
    0x1fd743c5f1500ULL,
    0xc15ecea908a80ULL,
    0x6e40d2fd45e40ULL,
    0xce617e23fe320ULL,
    0x3171412525210ULL,
    0x8b8fd8ae6e258ULL,
    0x119987f9bdc54ULL,
    0x5de77893b349aULL,
    0x9e3517cd9f61bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0x7c00000000000ULL,
    0xa200000000000ULL,
    0x6900000000000ULL,
    0x6380000000000ULL,
    0x2440000000000ULL,
    0x1860000000000ULL,
    0x5550000000000ULL,
    0x9ff8000000000ULL,
    0x5c64000000000ULL,
    0xd352000000000ULL,
    0xe0fd000000000ULL,
    0xf8e7800000000ULL,
    0x3791400000000ULL,
    0xaa5ee00000000ULL,
    0x5012900000000ULL,
    0xa81a080000000ULL,
    0x4771fc0000000ULL,
    0x31cdb20000000ULL,
    0x4049cb0000000ULL,
    0x850fd68000000ULL,
    0x52e88dc000000ULL,
    0x99fa082000000ULL,
    0x6361fc1000000ULL,
    0xd7d5b22800000ULL,
    0x743dcb1c00000ULL,
    0xda45d69a00000ULL,
    0xe3618ddf00000ULL,
    0x97d7880880000ULL,
    0xd439bc0a40000ULL,
    0xaa46d230e0000ULL,
    0x8b671b1510000ULL,
    0xebd4be9c18000ULL,
    0x763ea1ca54000ULL,
    0xc341522f7e000ULL,
    0xe8e65b1d59000ULL,
    0xcf925e8434800ULL,
    0x6e5831ed2f400ULL,
    0x96115a0221e00ULL,
    0x771ea717b0500ULL,
    0x93f26c8de9b80ULL,
    0xbd09bac55da40ULL,
    0x76edecbc32020ULL,
    0x8ffcfaeff33f0ULL,
    0xa4610c94d28c8ULL,
    0x17536adee222cULL,
    0x26fc84a6ca4daULL,
    0x27e3d6cdb7877ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0xdc00000000000ULL,
    0x1600000000000ULL,
    0x1f00000000000ULL,
    0x3a80000000000ULL,
    0x55c0000000000ULL,
    0x9e20000000000ULL,
    0x72d0000000000ULL,
    0x4c98000000000ULL,
    0x1024000000000ULL,
    0x71d6000000000ULL,
    0x801f000000000ULL,
    0xeae1800000000ULL,
    0xcd76c00000000ULL,
    0x7b09e00000000ULL,
    0x2e59300000000ULL,
    0xb082380000000ULL,
    0x7cc6340000000ULL,
    0xeba09e0000000ULL,
    0xb591490000000ULL,
    0x437c628000000ULL,
    0x36d29ac000000ULL,
    0x469a382000000ULL,
    0xf922343000000ULL,
    0xe4569e3800000ULL,
    0xf75e492c00000ULL,
    0xfd05e28600000ULL,
    0x37805af700000ULL,
    0x5a45d82580000ULL,
    0x48640437c0000ULL,
    0x84b52636a0000ULL,
    0x132ebd3970000ULL,
    0xf38c9ca188000ULL,
    0xde9823ebb4000ULL,
    0xe52382b6a6000ULL,
    0x1254aaf3c9000ULL,
    0x5859800ad5800ULL,
    0xdf82c02977c00ULL,
    0xbe47e02330600ULL,
    0xd2623034e9b00ULL,
    0x25b5b80e64780ULL,
    0xf2aff43f5f4c0ULL,
    0x56c8fe3b750e0ULL,
    0x1c7eb92d864d0ULL,
    0x2c57baa978478ULL,
    0x435d9ed7fde24ULL,
    0xcf009e1ab360aULL,
    0x5281492e9feabULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xe800000000000ULL,
    0xe400000000000ULL,
    0x7a00000000000ULL,
    0xb300000000000ULL,
    0xd680000000000ULL,
    0x93c0000000000ULL,
    0x1620000000000ULL,
    0x7990000000000ULL,
    0x4378000000000ULL,
    0x2424000000000ULL,
    0x7e92000000000ULL,
    0x17fb000000000ULL,
    0x78e7800000000ULL,
    0xc835400000000ULL,
    0xd0ac200000000ULL,
    0x323c500000000ULL,
    0x63c4d80000000ULL,
    0xbe20940000000ULL,
    0xfd97ce0000000ULL,
    0x4979e90000000ULL,
    0x7f21188000000ULL,
    0x4c1075c000000ULL,
    0xfe3cd82000000ULL,
    0xddc4941000000ULL,
    0x6725ce1800000ULL,
    0x12e91c00000ULL,
    0x3e98ba00000ULL,
    0x64c135f900000ULL,
    0xeaa2f83e80000ULL,
    0xa153c43cc0000ULL,
    0xffde962da0000ULL,
    0xb8733d38f0000ULL,
    0x6cf76bf88000ULL,
    0x55cb8ce764000ULL,
    0xb34ab8ae5e000ULL,
    0xa88965f5c9000ULL,
    0x2aac201204800ULL,
    0xc13c502d0ec00ULL,
    0xd544d839b1e00ULL,
    0x5de0940969500ULL,
    0x3b7ce0075080ULL,
    0xd4e9e91046140ULL,
    0x465918ac75b60ULL,
    0xdb3475e246e50ULL,
    0x562ed83174138ULL,
    0x59ff940ac72a4ULL,
    0x9e24e26b54e2ULL,
    0xb1b7a914e6097ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xb800000000000ULL,
    0xc400000000000ULL,
    0xca00000000000ULL,
    0x8900000000000ULL,
    0xa980000000000ULL,
    0x50c0000000000ULL,
    0x8160000000000ULL,
    0x6d50000000000ULL,
    0xf1c8000000000ULL,
    0xf64000000000ULL,
    0xee56000000000ULL,
    0x714d000000000ULL,
    0x4621800000000ULL,
    0x87f3c00000000ULL,
    0x597b200000000ULL,
    0x10bfd00000000ULL,
    0xf05b880000000ULL,
    0x11cc6c0000000ULL,
    0xdf66560000000ULL,
    0x4653030000000ULL,
    0xbd4d468000000ULL,
    0xf021384000000ULL,
    0xf3882000000ULL,
    0xb3f86c3000000ULL,
    0x60f8562800000ULL,
    0x887a032c00000ULL,
    0xad3ac6ae00000ULL,
    0xc29ff87100000ULL,
    0xd5a9281280000ULL,
    0xadb47c1240000ULL,
    0x6058fe0260000ULL,
    0xd9c9bf3830000ULL,
    0x367188e58000ULL,
    0xf850976a54000ULL,
    0x464a38aef2000ULL,
    0x14a2475199000ULL,
    0x33b03099f5800ULL,
    0xeb5deb5463400ULL,
    0x254d4697d0600ULL,
    0x44213857a8f00ULL,
    0xd2f3883eacc80ULL,
    0x4ef86c08b6f40ULL,
    0xbb78561963620ULL,
    0x95ba0333505b0ULL,
    0x4fdac6b0e9f58ULL,
    0x760ff86c4c30cULL,
    0xc81281dc799aULL,
    0x9f407c26daba1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xa800000000000ULL,
    0x8c00000000000ULL,
    0x6e00000000000ULL,
    0x4f00000000000ULL,
    0x980000000000ULL,
    0x4340000000000ULL,
    0x83e0000000000ULL,
    0xa5b0000000000ULL,
    0xdd58000000000ULL,
    0x11e4000000000ULL,
    0x3cb6000000000ULL,
    0x9fd9000000000ULL,
    0x6922800000000ULL,
    0x7513400000000ULL,
    0x730d600000000ULL,
    0x3509f00000000ULL,
    0x360d380000000ULL,
    0x718f2c0000000ULL,
    0xadccea0000000ULL,
    0x8faea70000000ULL,
    0xe559c18000000ULL,
    0xd5e320c000000ULL,
    0x2eb5382000000ULL,
    0xc6db2c3000000ULL,
    0xba2ea0800000ULL,
    0xddd3a73400000ULL,
    0x92ad41aa00000ULL,
    0x55d960e300000ULL,
    0x422d83b80000ULL,
    0x7d959c23c0000ULL,
    0xa6ccb20a60000ULL,
    0x54287b24d0000ULL,
    0xdf18138af8000ULL,
    0xe482abca6c000ULL,
    0x8bc0138cd6000ULL,
    0x7626abe7b9000ULL,
    0x149613a54d800ULL,
    0x9c4fabf326400ULL,
    0x8ec93b8b0a00ULL,
    0x74b8ebd328d00ULL,
    0x2957f3b215580ULL,
    0x91681bfdfb7c0ULL,
    0x7ff84b994ece0ULL,
    0xd6b477df858b0ULL,
    0x22d9c18a239a8ULL,
    0xc9a320f3d379cULL,
    0x2cd53829db286ULL,
    0x7c2b2c313fb43ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0x2400000000000ULL,
    0xbe00000000000ULL,
    0xa700000000000ULL,
    0xda80000000000ULL,
    0x5a40000000000ULL,
    0x9420000000000ULL,
    0xff30000000000ULL,
    0x12d8000000000ULL,
    0xd624000000000ULL,
    0x9a36000000000ULL,
    0x6d5b000000000ULL,
    0x93e1800000000ULL,
    0xfbd4400000000ULL,
    0xbb8a200000000ULL,
    0x73ec700000000ULL,
    0x73fd780000000ULL,
    0xe96f40000000ULL,
    0x712fa20000000ULL,
    0xb899590000000ULL,
    0xaa05cd8000000ULL,
    0x9101d4c000000ULL,
    0x3985782000000ULL,
    0x8ec2f43000000ULL,
    0x2161a21800000ULL,
    0x9596592c00000ULL,
    0x9aa4d9200000ULL,
    0xf7da94c900000ULL,
    0x69a0d80f80000ULL,
    0xbff5c419c0000ULL,
    0xbcb97a2ea0000ULL,
    0xd379d3a90000ULL,
    0x3fdd37b708000ULL,
    0x8da209f6cc000ULL,
    0x61f26f8b36000ULL,
    0xabbc8dec49000ULL,
    0x9fb6b5a82d800ULL,
    0x419820d1c6c00ULL,
    0xa7855a2bf0600ULL,
    0x39c0ed1439100ULL,
    0x63e1cf8fd4880ULL,
    0x13d0bde5b21c0ULL,
    0x4f8fed9152de0ULL,
    0x35eda4dea37d0ULL,
    0xb8f8003c9b888ULL,
    0x4e14000c4f464ULL,
    0x326e00197dfb6ULL,
    0x513f00016e693ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xf800000000000ULL,
    0xf400000000000ULL,
    0x9a00000000000ULL,
    0x4100000000000ULL,
    0xf480000000000ULL,
    0xa440000000000ULL,
    0x5660000000000ULL,
    0x9770000000000ULL,
    0x8198000000000ULL,
    0xd764000000000ULL,
    0x3f2000000000ULL,
    0x75d9000000000ULL,
    0x907800000000ULL,
    0x9885c00000000ULL,
    0x9a46a00000000ULL,
    0x567500000000ULL,
    0x2ef5c80000000ULL,
    0xbf58040000000ULL,
    0xfe45f20000000ULL,
    0x7762bb0000000ULL,
    0x33f21f8000000ULL,
    0xaddaaec000000ULL,
    0x8d07c82000000ULL,
    0xfa81041000000ULL,
    0x2f42720800000ULL,
    0x6be77b1c00000ULL,
    0xcbb4bfbe00000ULL,
    0x1dbdfefd00000ULL,
    0xcd72000680000ULL,
    0xa099000040000ULL,
    0x73e7803520000ULL,
    0x2fb5c03510000ULL,
    0x2fbea02b98000ULL,
    0xf0735018dc000ULL,
    0xc21fc806e6000ULL,
    0xf8a5042599000ULL,
    0x5650723ddc800ULL,
    0xa90e7b3466400ULL,
    0x774b3f97d9e00ULL,
    0x702c3ec3fd700ULL,
    0xbbdea026f7a80ULL,
    0xe203502480d40ULL,
    0xf507c80341f20ULL,
    0x4e81042ea0410ULL,
    0x95427203d09c8ULL,
    0x5ae77b06f9decULL,
    0xc734bf8ced2feULL,
    0x4dfdfefa6f7fbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0x9400000000000ULL,
    0xae00000000000ULL,
    0xbf00000000000ULL,
    0x9b80000000000ULL,
    0x7140000000000ULL,
    0xfba0000000000ULL,
    0x33d0000000000ULL,
    0x51d8000000000ULL,
    0xdca4000000000ULL,
    0xf456000000000ULL,
    0xca9d000000000ULL,
    0xee07800000000ULL,
    0xdf03400000000ULL,
    0x4b81a00000000ULL,
    0xa946f00000000ULL,
    0xc7a0b80000000ULL,
    0x9d07c0000000ULL,
    0x40defa0000000ULL,
    0xf826250000000ULL,
    0x1e93428000000ULL,
    0x407c04c000000ULL,
    0x2676b82000000ULL,
    0xbd0d7c3000000ULL,
    0xc6f97a2800000ULL,
    0x81b5651c00000ULL,
    0xf96ae2aa00000ULL,
    0x2d4ef4e500000ULL,
    0x71d8000b80000ULL,
    0x6ca4001fc0000ULL,
    0xfc56000ee0000ULL,
    0x2e9d000050000ULL,
    0xe8078014e8000ULL,
    0xf4034029f4000ULL,
    0x7e01a03ff6000ULL,
    0x6706f018e9000ULL,
    0xa780b81bf5800ULL,
    0x4b407c2ef7400ULL,
    0xeaa6fa0569e00ULL,
    0x1752253b34d00ULL,
    0xbb1d428696680ULL,
    0x564504dd78bc0ULL,
    0x3c27382cfdae0ULL,
    0xa8933c00d35f0ULL,
    0x637f5a03b65e8ULL,
    0xf7f0d515c9594ULL,
    0x754bfabf44b8aULL,
    0x8dd878df4ebd3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x8800000000000ULL,
    0xbc00000000000ULL,
    0x4600000000000ULL,
    0x6500000000000ULL,
    0x280000000000ULL,
    0xcb40000000000ULL,
    0x90a0000000000ULL,
    0x4ff0000000000ULL,
    0xe628000000000ULL,
    0xcda4000000000ULL,
    0xa972000000000ULL,
    0xff6d000000000ULL,
    0x5200800000000ULL,
    0x4f06400000000ULL,
    0x980e00000000ULL,
    0x80c6b00000000ULL,
    0x9767680000000ULL,
    0xcd15f40000000ULL,
    0x7dfc6a0000000ULL,
    0xca3bad0000000ULL,
    0x10dc6b8000000ULL,
    0x1f0a6b4000000ULL,
    0x7295682000000ULL,
    0x6038f41000000ULL,
    0x5bdcea2800000ULL,
    0xf48ded2400000ULL,
    0xe5548ba200000ULL,
    0x6ad8db6f00000ULL,
    0x7c08003180000ULL,
    0xb514000940000ULL,
    0x39fa0028a0000ULL,
    0x88390016d0000ULL,
    0x97da800628000ULL,
    0xa8f403cfc000ULL,
    0x245260280a000ULL,
    0x1a5df02a29000ULL,
    0x284f082afc800ULL,
    0x431040d0b400ULL,
    0xdac9e230a8200ULL,
    0x3775e9003d900ULL,
    0x4e6fe9bbea380ULL,
    0x9a86725358ac0ULL,
    0xbf4789af855a0ULL,
    0xcaa2825c9e3d0ULL,
    0xfcf20183ff3a8ULL,
    0x9acc6554f7b4ULL,
    0xb86183ac5722eULL,
    0xb490df7cf336dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0x8c00000000000ULL,
    0xe200000000000ULL,
    0xc700000000000ULL,
    0xb880000000000ULL,
    0xb440000000000ULL,
    0x70a0000000000ULL,
    0x5bf0000000000ULL,
    0xf3c8000000000ULL,
    0x43a4000000000ULL,
    0xfd72000000000ULL,
    0x9a8d000000000ULL,
    0xf085800000000ULL,
    0x1040c00000000ULL,
    0xe6a1600000000ULL,
    0xc2f6900000000ULL,
    0x364cf80000000ULL,
    0x465a40000000ULL,
    0x6310b20000000ULL,
    0xc298bd0000000ULL,
    0x909a9b8000000ULL,
    0x4f9efc4000000ULL,
    0xdb1ef82000000ULL,
    0xbd58a41000000ULL,
    0x52fd322800000ULL,
    0x3d8c7d0c00000ULL,
    0x9801fb9600000ULL,
    0x6c016c6300000ULL,
    0x7205801880000ULL,
    0xaf00c021c0000ULL,
    0x6c81600620000ULL,
    0xda46902110000ULL,
    0x55a4f80a28000ULL,
    0x2471a435fc000ULL,
    0xff0ab20472000ULL,
    0x8741bd2129000ULL,
    0xd6251b917c800ULL,
    0x32b73c4bb3400ULL,
    0x40e8180009600ULL,
    0xae93f402ec300ULL,
    0x815d2a2d5a580ULL,
    0xa8fb892c54a40ULL,
    0xd68ed1a4cfbe0ULL,
    0xf287e544ba290ULL,
    0x4746d1bc654c8ULL,
    0x3623e555761f4ULL,
    0xa2b4d1bbaefeeULL,
    0x28eee5524a1b1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x4800000000000ULL,
    0xf400000000000ULL,
    0xa600000000000ULL,
    0x9d00000000000ULL,
    0xc380000000000ULL,
    0xb8c0000000000ULL,
    0xf4a0000000000ULL,
    0x84b0000000000ULL,
    0x87a8000000000ULL,
    0xe5a4000000000ULL,
    0xed36000000000ULL,
    0xc869000000000ULL,
    0x586800000000ULL,
    0x95c4400000000ULL,
    0xf26a00000000ULL,
    0xd077b00000000ULL,
    0x690ea80000000ULL,
    0xae15bc0000000ULL,
    0x921ef60000000ULL,
    0xcb8e4d0000000ULL,
    0x6752b98000000ULL,
    0x957cbb4000000ULL,
    0xfd18a82000000ULL,
    0xa30cbc3000000ULL,
    0xe910760800000ULL,
    0xbe9e0d1400000ULL,
    0xdbca199200000ULL,
    0x3db60b7d00000ULL,
    0x6a2e800980000ULL,
    0x1460401740000ULL,
    0x2c10a038e0000ULL,
    0x211eb03a30000ULL,
    0x4108282f28000ULL,
    0xea11fc1c2c000ULL,
    0xc1856086a000ULL,
    0xba89fd1e29000ULL,
    0xbed4118bad800ULL,
    0xe2bd075c2a400ULL,
    0xf1385e1c49a00ULL,
    0xc1fff1345d100ULL,
    0xe15a4faa23a80ULL,
    0x93ebf65974ec0ULL,
    0xc14491930e2a0ULL,
    0x136047549f2f0ULL,
    0xe0907e24661d8ULL,
    0x41580114d2834ULL,
    0x3ec47b0fd066ULL,
    0xa944fa5672c2dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x9800000000000ULL,
    0x4c00000000000ULL,
    0xde00000000000ULL,
    0xe100000000000ULL,
    0xbf80000000000ULL,
    0x36c0000000000ULL,
    0x7ca0000000000ULL,
    0x2cd0000000000ULL,
    0xea88000000000ULL,
    0xc3a4000000000ULL,
    0xb256000000000ULL,
    0xc3cd000000000ULL,
    0x39c1800000000ULL,
    0x7a22c00000000ULL,
    0x4997600000000ULL,
    0xceeb700000000ULL,
    0xead7c80000000ULL,
    0x87899c0000000ULL,
    0x226460000000ULL,
    0xd5976d0000000ULL,
    0x98ed5a8000000ULL,
    0x6fd5024000000ULL,
    0x3209c82000000ULL,
    0x47e09c3000000ULL,
    0x29b1c62800000ULL,
    0xdc78ad2c00000ULL,
    0x70bbbaa600000ULL,
    0x971cb25300000ULL,
    0x4f49601780000ULL,
    0xef82700840000ULL,
    0x7ec04807e0000ULL,
    0xb8a65c21b0000ULL,
    0x96d0a63928000ULL,
    0x18edd1834000ULL,
    0xf25f28d22000ULL,
    0x413ee48a9000ULL,
    0xd72e462375800ULL,
    0xe0f36d0d43400ULL,
    0xf61b5ab958600ULL,
    0x30c80279bcb00ULL,
    0x7940482ec7d80ULL,
    0xb2665c2053dc0ULL,
    0xc70a609a0720ULL,
    0x305edd0551270ULL,
    0x1c2df28719f18ULL,
    0x5c77ee755d6b4ULL,
    0x7858463cf68eaULL,
    0xd82e6d2b3b9c9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x6800000000000ULL,
    0xfc00000000000ULL,
    0x4600000000000ULL,
    0x7f00000000000ULL,
    0x6880000000000ULL,
    0xcfc0000000000ULL,
    0x3ea0000000000ULL,
    0xbc90000000000ULL,
    0x6d98000000000ULL,
    0x67a4000000000ULL,
    0xfb16000000000ULL,
    0x32d9000000000ULL,
    0xb2c1800000000ULL,
    0xab25400000000ULL,
    0x26d6e00000000ULL,
    0xe97fb00000000ULL,
    0xefd4b80000000ULL,
    0x76faa40000000ULL,
    0x24914e0000000ULL,
    0x19998d0000000ULL,
    0x4da0e08000000ULL,
    0x1a1172c000000ULL,
    0xb15ab82000000ULL,
    0xaf87a43000000ULL,
    0x6346ce0800000ULL,
    0xb365cd2c00000ULL,
    0x9f37809a00000ULL,
    0x148b82ff00000ULL,
    0x6878e03180000ULL,
    0x3c52b02fc0000ULL,
    0x93bb381220000ULL,
    0xc8f2e41ff0000ULL,
    0x5e282e15a8000ULL,
    0xfeee7d1024000ULL,
    0xb0cd388ae6000ULL,
    0x779c26c629000ULL,
    0x2ea62e04e5800ULL,
    0xa4937d0f46400ULL,
    0xd99ab8ab18600ULL,
    0x6da066e9ed500ULL,
    0xaa114e28d3b80ULL,
    0xd9598d2cb6ec0ULL,
    0x5380e09cb0ae0ULL,
    0x254172e2c8e90ULL,
    0xcc62b805b4338ULL,
    0xf7b3a4105f334ULL,
    0xdb48ce0cf5802ULL,
    0x56d8cd203fb0bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0x1c00000000000ULL,
    0x1600000000000ULL,
    0xa300000000000ULL,
    0x2880000000000ULL,
    0x8cc0000000000ULL,
    0xb4a0000000000ULL,
    0xc8f0000000000ULL,
    0xc748000000000ULL,
    0xd1a4000000000ULL,
    0x8b76000000000ULL,
    0xf0f000000000ULL,
    0x87c0800000000ULL,
    0xa021400000000ULL,
    0x2237e00000000ULL,
    0xa8ec300000000ULL,
    0xc5d6580000000ULL,
    0xcaffa40000000ULL,
    0xdf0fd60000000ULL,
    0x4fc63d0000000ULL,
    0xcc21438000000ULL,
    0x4c313c4000000ULL,
    0xffe8582000000ULL,
    0xe754a43000000ULL,
    0xf339563800000ULL,
    0xe0287d2400000ULL,
    0x237623ba00000ULL,
    0x330c4c4700000ULL,
    0xe1c1e02580000ULL,
    0x7b233018c0000ULL,
    0xfeb6d83220000ULL,
    0x2e2ee40730000ULL,
    0xc470361728000ULL,
    0x898e0d353c000ULL,
    0xbc011bb452000ULL,
    0xa601985ca9000ULL,
    0x3b070e08fd800ULL,
    0x4c83d910f3c00ULL,
    0x6ec7f5b4d8200ULL,
    0x1da1717e34500ULL,
    0x557623a35ff80ULL,
    0xc00c4c69520c0ULL,
    0xc141e00e28160ULL,
    0x7be3300dbc290ULL,
    0xb416d81e13d58ULL,
    0x59dee43bc9df4ULL,
    0x3db836292da8eULL,
    0x77ea0d24cb431ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xc800000000000ULL,
    0xc00000000000ULL,
    0x4e00000000000ULL,
    0xc300000000000ULL,
    0xf580000000000ULL,
    0xb40000000000ULL,
    0x68e0000000000ULL,
    0xfb50000000000ULL,
    0x9918000000000ULL,
    0xa9e4000000000ULL,
    0xa3d6000000000ULL,
    0xf4df000000000ULL,
    0x47c1800000000ULL,
    0xff20400000000ULL,
    0xbc77600000000ULL,
    0x16ff00000000ULL,
    0x5a8c580000000ULL,
    0xbc5dcc0000000ULL,
    0x72072a0000000ULL,
    0x8501df0000000ULL,
    0xda84b68000000ULL,
    0xc732c000000ULL,
    0x68a2582000000ULL,
    0x6a36cc3000000ULL,
    0x2708aa3800000ULL,
    0x9b1a9f2c00000ULL,
    0x4e456b200000ULL,
    0xc55782c300000ULL,
    0x7218e03380000ULL,
    0xe064b000c0000ULL,
    0x2e94b80560000ULL,
    0x53397c2ed0000ULL,
    0x713922838000ULL,
    0xae78a33dd4000ULL,
    0x88f724b5c6000ULL,
    0x52af91dab9000ULL,
    0x5fad7cb595800ULL,
    0xa92d5dcfe7c00ULL,
    0xb7ebd693c8600ULL,
    0x644cc2dd1c100ULL,
    0x3ff800235bd80ULL,
    0x51b4003522fc0ULL,
    0x2f4e0027d6960ULL,
    0xe67b000de0b30ULL,
    0x44f7801191aa8ULL,
    0xfcaf402f9867cULL,
    0x2caee02d64f5aULL,
    0x94abb00cae30bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x6800000000000ULL,
    0x1400000000000ULL,
    0x8600000000000ULL,
    0x8f00000000000ULL,
    0x3a80000000000ULL,
    0x7e40000000000ULL,
    0xa020000000000ULL,
    0xe5f0000000000ULL,
    0xb1f8000000000ULL,
    0xeaa4000000000ULL,
    0x33b6000000000ULL,
    0xe5df000000000ULL,
    0x7953800000000ULL,
    0x654d400000000ULL,
    0x21fba00000000ULL,
    0xb2a1b00000000ULL,
    0xafb6380000000ULL,
    0x87d9f40000000ULL,
    0x1857020000000ULL,
    0xc4cc990000000ULL,
    0xe3396b8000000ULL,
    0xe3c3b5c000000ULL,
    0xd0e5b82000000ULL,
    0xad94b43000000ULL,
    0xe32ca23800000ULL,
    0xf82d293c00000ULL,
    0x84af539a00000ULL,
    0x95ea41c500000ULL,
    0xff4aba0180000ULL,
    0xcfc2d13c0000ULL,
    0x923c9b6a0000ULL,
    0x80719ce390000ULL,
    0x8d396b9208000ULL,
    0xb8c3b5cc7c000ULL,
    0x8c65b815fe000ULL,
    0xacd4b41569000ULL,
    0x118ca2204d800ULL,
    0x779d291fe7c00ULL,
    0x1377538ddce00ULL,
    0x15be41c6ef500ULL,
    0x4784ba2b20e80ULL,
    0x7dc72d1a516c0ULL,
    0xe3e649b9a80e0ULL,
    0xe513dcc26dbd0ULL,
    0x64694ba637208ULL,
    0x858b45deb5764ULL,
    0x319e2029a332eULL,
    0x7c73f03846417ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x4800000000000ULL,
    0x4c00000000000ULL,
    0x2a00000000000ULL,
    0xdd00000000000ULL,
    0x4980000000000ULL,
    0xbec0000000000ULL,
    0x5560000000000ULL,
    0x8a30000000000ULL,
    0x1268000000000ULL,
    0xf4e4000000000ULL,
    0xc8f6000000000ULL,
    0x550f000000000ULL,
    0x37d0800000000ULL,
    0xbd1c400000000ULL,
    0xa429600000000ULL,
    0x14c7d00000000ULL,
    0x4865a80000000ULL,
    0x23b3ac0000000ULL,
    0x3caa660000000ULL,
    0xe980630000000ULL,
    0xec32d8000000ULL,
    0x6d65dd4000000ULL,
    0x1e35282000000ULL,
    0x3c6fec3000000ULL,
    0x4fe3063800000ULL,
    0x7677b32400000ULL,
    0x7f4e859200000ULL,
    0x95f2715300000ULL,
    0xdc894e2a80000ULL,
    0x69108f0740000ULL,
    0x7878abaa60000ULL,
    0x661a2e4bb0000ULL,
    0x4aaccda758000ULL,
    0x96814d418c000ULL,
    0x364160161a000ULL,
    0x2023d01e79000ULL,
    0x6093a80a5d800ULL,
    0xe6bcac0df3c00ULL,
    0x437ae6002c200ULL,
    0x189c23298b100ULL,
    0x80ea4db570580ULL,
    0xa4a20d60f8f40ULL,
    0x1fd080071cea0ULL,
    0xa11c4020932b0ULL,
    0x2629602b1cb98ULL,
    0x15c7d02e9208cULL,
    0x63e5a80e1d136ULL,
    0xc73ac2512435ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xa800000000000ULL,
    0xbc00000000000ULL,
    0x6a00000000000ULL,
    0x6d00000000000ULL,
    0x9580000000000ULL,
    0xc9c0000000000ULL,
    0x9020000000000ULL,
    0x4b90000000000ULL,
    0xf8c8000000000ULL,
    0x1da4000000000ULL,
    0x8652000000000ULL,
    0x76eb000000000ULL,
    0xbd36800000000ULL,
    0x441f400000000ULL,
    0xe60ae00000000ULL,
    0x9d07d00000000ULL,
    0x4d87280000000ULL,
    0xedc0cc0000000ULL,
    0xbe22de0000000ULL,
    0x5897490000000ULL,
    0xd648408000000ULL,
    0x46e5b04000000ULL,
    0x2731a82000000ULL,
    0xf11f8c1000000ULL,
    0x57883e1800000ULL,
    0x7ac0991400000ULL,
    0xcea768aa00000ULL,
    0x88d17c6f00000ULL,
    0x1da9763a80000ULL,
    0xe457c50b40000ULL,
    0xd7ecfebd60000ULL,
    0xdab5696670000ULL,
    0xa4d8a0ae08000ULL,
    0x9dad606de4000ULL,
    0xa452001cb2000ULL,
    0xb7eb001829000ULL,
    0x8ab68036f4800ULL,
    0xcdf4014cac00ULL,
    0x21aae03bc5a00ULL,
    0xce57d037a3d00ULL,
    0xdaef283850b80ULL,
    0x1f34cc0918f40ULL,
    0xc518de339d4a0ULL,
    0xb18849065ef30ULL,
    0x85c4c09aff178ULL,
    0x2225f052af024ULL,
    0x297c833b6282ULL,
    0x43481c3c2b581ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0x7c00000000000ULL,
    0x5a00000000000ULL,
    0x2b00000000000ULL,
    0xfc80000000000ULL,
    0xa1c0000000000ULL,
    0x8ce0000000000ULL,
    0xa790000000000ULL,
    0x99a8000000000ULL,
    0x8c64000000000ULL,
    0x9c56000000000ULL,
    0xde4f000000000ULL,
    0x6771800000000ULL,
    0xec3c400000000ULL,
    0xa2c8e00000000ULL,
    0x6b5500000000ULL,
    0x80da180000000ULL,
    0xb55a840000000ULL,
    0xd71a1e0000000ULL,
    0x70b99f0000000ULL,
    0x730b688000000ULL,
    0x22508b4000000ULL,
    0xeb4b982000000ULL,
    0x3ef6c43000000ULL,
    0xc7afe3800000ULL,
    0x4a68cf2c00000ULL,
    0x2787709200000ULL,
    0x35450f5f00000ULL,
    0x120063680000ULL,
    0x89731b3ac0000ULL,
    0x2139768720000ULL,
    0xcf4d144470000ULL,
    0xc0f6f09138000ULL,
    0x19794f46e4000ULL,
    0x43e8e628ea000ULL,
    0x3fc64b35d9000ULL,
    0xe9e36eb235800ULL,
    0x617906ce3c00ULL,
    0x4decee9e64600ULL,
    0x42c0d0772b100ULL,
    0xcc638ebf78380ULL,
    0xbc56c05c04540ULL,
    0x8e48f6913b060ULL,
    0x9f71545851610ULL,
    0xd83e10b5c8e78ULL,
    0x84cc1f641c77cULL,
    0x77b2fe16af622ULL,
    0x575ccf3c03b6dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xf800000000000ULL,
    0xa400000000000ULL,
    0x7e00000000000ULL,
    0x7100000000000ULL,
    0xc780000000000ULL,
    0xb5c0000000000ULL,
    0x5420000000000ULL,
    0xa1b0000000000ULL,
    0xa458000000000ULL,
    0xe7a4000000000ULL,
    0x2272000000000ULL,
    0x7d7d000000000ULL,
    0xab93800000000ULL,
    0xc4edc00000000ULL,
    0xdc7c200000000ULL,
    0x2412700000000ULL,
    0x8d2b480000000ULL,
    0xa25d9c0000000ULL,
    0x72a01e0000000ULL,
    0x3bf4c90000000ULL,
    0x49bf618000000ULL,
    0xc034b4c000000ULL,
    0x7498c82000000ULL,
    0x52005c1000000ULL,
    0x13043e2800000ULL,
    0xcc82b93c00000ULL,
    0x8d4629be00000ULL,
    0xaf6428e900000ULL,
    0x23d3563f80000ULL,
    0x9a0d550c40000ULL,
    0xc76d7f99e0000ULL,
    0x7bd7dd170000ULL,
    0xa934298b08000ULL,
    0x471928d16c000ULL,
    0x91c0d63e96000ULL,
    0x6a209509a9000ULL,
    0x70b15faf7c800ULL,
    0x93df0de72f400ULL,
    0xaa6761be6ce00ULL,
    0xd250b4fc17700ULL,
    0xa2cac83869080ULL,
    0x7ecd5c2ddd9c0ULL,
    0xe4cfbe193e520ULL,
    0x4bcb7927d4270ULL,
    0xd4809abd2e78ULL,
    0x870b58db03424ULL,
    0xc2eb9e0a9a506ULL,
    0x497d093e8ff13ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x8800000000000ULL,
    0xd400000000000ULL,
    0x4600000000000ULL,
    0x6300000000000ULL,
    0x1c80000000000ULL,
    0x3cc0000000000ULL,
    0xb4e0000000000ULL,
    0x77f0000000000ULL,
    0x5878000000000ULL,
    0xa464000000000ULL,
    0x1132000000000ULL,
    0x3d99000000000ULL,
    0x3a12800000000ULL,
    0x820e400000000ULL,
    0xd95ba00000000ULL,
    0x92f2b00000000ULL,
    0x47fc580000000ULL,
    0x7425c40000000ULL,
    0xc113160000000ULL,
    0xa28d5f0000000ULL,
    0x879d4b8000000ULL,
    0x5b104dc000000ULL,
    0x138ed82000000ULL,
    0x9e1b841000000ULL,
    0xc850b60800000ULL,
    0xf2bef3400000ULL,
    0xb5cb13a200000ULL,
    0x153889f500000ULL,
    0xe7c54e3180000ULL,
    0x44659b08c0000ULL,
    0xa136dd8f20000ULL,
    0x459e52fb30000ULL,
    0xb610b3af38000ULL,
    0x980a39f8fc000ULL,
    0x2859162f9e000ULL,
    0xab705f15d9000ULL,
    0x4bdcba96c800ULL,
    0xe0870dc156400ULL,
    0x3ec7f8303ca00ULL,
    0x39e77400bf900ULL,
    0xc774e3ee8e80ULL,
    0xb63c9b2e55ac0ULL,
    0x9b445d9dab960ULL,
    0x88a012f5a3310ULL,
    0x4dd313b466658ULL,
    0x6e6c89f405c7cULL,
    0x7c6f4e2ac33aeULL,
    0x19689b3f37ff7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xb800000000000ULL,
    0xcc00000000000ULL,
    0x1a00000000000ULL,
    0x900000000000ULL,
    0x1880000000000ULL,
    0x7040000000000ULL,
    0x7ca0000000000ULL,
    0xc010000000000ULL,
    0x89b8000000000ULL,
    0x8624000000000ULL,
    0x5556000000000ULL,
    0xa79f000000000ULL,
    0x3771800000000ULL,
    0xcccbc00000000ULL,
    0x47ed600000000ULL,
    0x843f500000000ULL,
    0x3ce0880000000ULL,
    0xc4b2240000000ULL,
    0x65ac120000000ULL,
    0xa5998d0000000ULL,
    0x6277b48000000ULL,
    0x264837c000000ULL,
    0xfaa9082000000ULL,
    0x8e1de43000000ULL,
    0x53b7723800000ULL,
    0x3a29dd2c00000ULL,
    0x425ebcae00000ULL,
    0x6d15d3f300000ULL,
    0x8f3e7a2680000ULL,
    0x7164393240000ULL,
    0x5e71cebe20000ULL,
    0xa4480ef010000ULL,
    0x6faec69128000ULL,
    0x849aeaf304000ULL,
    0x5ef034bcee000ULL,
    0x880cf7ffc9000ULL,
    0xe10de82575800ULL,
    0x898d741af7c00ULL,
    0xa1cc9a0274600ULL,
    0xa16ba91e76f00ULL,
    0xc37ba69d35580ULL,
    0x8fc1baedd6d40ULL,
    0xb66bca5e5a20ULL,
    0x2771d3c69f490ULL,
    0xe4c87a37d1648ULL,
    0x63eb390bc9934ULL,
    0x5a384eb775352ULL,
    0x5be7cecff719fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0xb800000000000ULL,
    0x8400000000000ULL,
    0xb200000000000ULL,
    0x2900000000000ULL,
    0xcf80000000000ULL,
    0xd40000000000ULL,
    0x5720000000000ULL,
    0x4590000000000ULL,
    0x2e88000000000ULL,
    0xb2a4000000000ULL,
    0x55d6000000000ULL,
    0x1c2d000000000ULL,
    0x2777800000000ULL,
    0xe9fbc00000000ULL,
    0x4b5d200000000ULL,
    0x768c100000000ULL,
    0x26a3880000000ULL,
    0x8fd2dc0000000ULL,
    0x792eba0000000ULL,
    0x66f5910000000ULL,
    0xfbbe208000000ULL,
    0x48f8e64000000ULL,
    0xd8dc082000000ULL,
    0x9dcd1c3000000ULL,
    0x22859a2800000ULL,
    0x10c4811c00000ULL,
    0xa36228ae00000ULL,
    0x1535fa6100000ULL,
    0x5a59920c80000ULL,
    0xfd099d3a40000ULL,
    0x39e7b29be0000ULL,
    0x81f17b5f50000ULL,
    0x4b3bba9bc8000ULL,
    0xc13c674064000ULL,
    0xac3e20af22000ULL,
    0x71b8e64ae9000ULL,
    0x25fc080095800ULL,
    0x55d1c095b400ULL,
    0xc98d9a3695e00ULL,
    0x220811c5af00ULL,
    0xdc1428a815480ULL,
    0x68c8fa591a040ULL,
    0xcb06123675620ULL,
    0xee865d0c8bf70ULL,
    0x5ec492921c4e8ULL,
    0x1c646b415e944ULL,
    0x31b1b2a927402ULL,
    0x99c7b645b7d9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0x3400000000000ULL,
    0x3a00000000000ULL,
    0xd00000000000ULL,
    0x7f80000000000ULL,
    0xbb40000000000ULL,
    0x9a0000000000ULL,
    0xd870000000000ULL,
    0x3708000000000ULL,
    0xb824000000000ULL,
    0x7432000000000ULL,
    0x5c2b000000000ULL,
    0xdc97800000000ULL,
    0xcdda400000000ULL,
    0x3bdea00000000ULL,
    0x70dc500000000ULL,
    0x6c5d280000000ULL,
    0xef9cdc0000000ULL,
    0x27faca0000000ULL,
    0x6ef450000000ULL,
    0x61738f8000000ULL,
    0x328eba4000000ULL,
    0x6e62a82000000ULL,
    0x72129c1000000ULL,
    0x371e6a1800000ULL,
    0xd63c151c00000ULL,
    0x978b278200000ULL,
    0x75e3264d00000ULL,
    0xeb51422e80000ULL,
    0x7fbbc91340000ULL,
    0x77ce6d87e0000ULL,
    0x18c22372d0000ULL,
    0x8ce66da068000ULL,
    0x8ed6236b1c000ULL,
    0xd9fc6dbb42000ULL,
    0x79e9236149000ULL,
    0x47f1edb8ec800ULL,
    0xbc4c6378dac00ULL,
    0xd982cd81cde00ULL,
    0xe84573472a900ULL,
    0x5924c5a855a80ULL,
    0x5db4bf5eae140ULL,
    0xfc6a07af13ca0ULL,
    0xd1b1367661f70ULL,
    0xda68ca1a11528ULL,
    0xc2b4450648414ULL,
    0xeaec0f8f6dcbeULL,
    0x3770fa631b7a9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0x3c00000000000ULL,
    0xba00000000000ULL,
    0xbf00000000000ULL,
    0x9a80000000000ULL,
    0xd940000000000ULL,
    0x5c60000000000ULL,
    0x4110000000000ULL,
    0x5f68000000000ULL,
    0xa8e4000000000ULL,
    0x6156000000000ULL,
    0xfc8b000000000ULL,
    0x6537800000000ULL,
    0x499d400000000ULL,
    0xac5ee00000000ULL,
    0xfc7c500000000ULL,
    0xbc8fb80000000ULL,
    0xc537b40000000ULL,
    0x599eda0000000ULL,
    0xb4593b0000000ULL,
    0xa8782c8000000ULL,
    0x3a8fb94000000ULL,
    0xc030382000000ULL,
    0x7c1ef43000000ULL,
    0xf79e3a1800000ULL,
    0x2d5a6b1c00000ULL,
    0x27fe149a00000ULL,
    0xde4a4d4f00000ULL,
    0x8b91820e80000ULL,
    0x3e2ddf1fc0000ULL,
    0xb080cebea0000ULL,
    0xbe43766a50000ULL,
    0xf2e1aead18000ULL,
    0x6e56666f44000ULL,
    0x6e0ef6815a000ULL,
    0xf0728269f9000ULL,
    0xc7fe149cf5800ULL,
    0x6e4a4d6a72c00ULL,
    0x83918202d5e00ULL,
    0x722ddf0ee3500ULL,
    0x6280ceb6edb80ULL,
    0x3d43765fb6140ULL,
    0xd261ae924e6e0ULL,
    0x8166678bb2d0ULL,
    0xa8eef6bfc8568ULL,
    0x682282709c1ecULL,
    0xc4f6149d3e332ULL,
    0x87be4d61b33a5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x2800000000000ULL,
    0x4c00000000000ULL,
    0xda00000000000ULL,
    0xd300000000000ULL,
    0xad80000000000ULL,
    0xd1c0000000000ULL,
    0x6fa0000000000ULL,
    0xb830000000000ULL,
    0xf1b8000000000ULL,
    0x9c24000000000ULL,
    0x9cf6000000000ULL,
    0x629d000000000ULL,
    0x3f52800000000ULL,
    0xc0af400000000ULL,
    0xbdeb600000000ULL,
    0x210a500000000ULL,
    0xd8d8d80000000ULL,
    0x60368c0000000ULL,
    0x85bc9e0000000ULL,
    0x8221450000000ULL,
    0x9f5dd8000000ULL,
    0xee1f6bc000000ULL,
    0xdc12582000000ULL,
    0x94dcc3000000ULL,
    0x68b9fe2800000ULL,
    0xaaa2153400000ULL,
    0xdcb1858a00000ULL,
    0xfeffa7d300000ULL,
    0x5fc1261680000ULL,
    0xb2a49904c0000ULL,
    0x8b51b8360000ULL,
    0x30fae2c070000ULL,
    0xe2c2fbb1e8000ULL,
    0x7226f2cd0c000ULL,
    0x31f5c382ee000ULL,
    0x2a186ed7c9000ULL,
    0x9a10658e5d800ULL,
    0x284eb7fbd7400ULL,
    0x5a3c9e20bca00ULL,
    0xce1453de7d00ULL,
    0xb1d5ddacf4d80ULL,
    0x84ef6bc8fb940ULL,
    0xc78a583d03b60ULL,
    0xb099cc0d16e30ULL,
    0xb057fe3cdd878ULL,
    0x3f2b152d16814ULL,
    0x232d058cdc2f6ULL,
    0x1129e7c5170efULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xd800000000000ULL,
    0xc00000000000ULL,
    0xee00000000000ULL,
    0x9d00000000000ULL,
    0xf280000000000ULL,
    0x18c0000000000ULL,
    0xd4e0000000000ULL,
    0xf370000000000ULL,
    0x11e8000000000ULL,
    0xdb64000000000ULL,
    0x2936000000000ULL,
    0xf5c9000000000ULL,
    0xf8f2800000000ULL,
    0xf1a8c00000000ULL,
    0x4045600000000ULL,
    0x8b27f00000000ULL,
    0x5a13280000000ULL,
    0x73d9740000000ULL,
    0xbd2d760000000ULL,
    0xdd802f0000000ULL,
    0x81407f8000000ULL,
    0x8fa3214000000ULL,
    0xf3d7a82000000ULL,
    0x6bb8b43000000ULL,
    0x139a960800000ULL,
    0xe60f1f3400000ULL,
    0x5f9637b600000ULL,
    0xfe9da54300000ULL,
    0xdc89f61b80000ULL,
    0xb351ef1740000ULL,
    0x187f9fb4a0000ULL,
    0x84f8117230000ULL,
    0xc13fe02338000ULL,
    0xc75b300fdc000ULL,
    0xfce84837fa000ULL,
    0xe1e3842599000ULL,
    0xc5f2de00ed800ULL,
    0x132c9b3842400ULL,
    0xa084e9b084a00ULL,
    0x83c13e67f6300ULL,
    0x9f651f88cb580ULL,
    0xf334d16220fc0ULL,
    0x7acc8001f14a0ULL,
    0x7175c019581d0ULL,
    0x52e9e02b0dfd8ULL,
    0x9ce2300ad33bcULL,
    0xc772c8378cc7eULL,
    0x3ef441813745ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x800000000000ULL,
    0x5c00000000000ULL,
    0xf600000000000ULL,
    0x6900000000000ULL,
    0xf80000000000ULL,
    0x59c0000000000ULL,
    0x2260000000000ULL,
    0x6ef0000000000ULL,
    0x7328000000000ULL,
    0xb4e4000000000ULL,
    0x10b6000000000ULL,
    0x48f000000000ULL,
    0xb671800000000ULL,
    0x306e400000000ULL,
    0x66c1e00000000ULL,
    0xd4e4300000000ULL,
    0x60b4080000000ULL,
    0xec8f1c0000000ULL,
    0x5a713e0000000ULL,
    0xce6ebb0000000ULL,
    0x53c7bb8000000ULL,
    0x2d619f4000000ULL,
    0x5073882000000ULL,
    0xc16e5c3000000ULL,
    0x6d415e3800000ULL,
    0x9f24cb2c00000ULL,
    0x89d2538200000ULL,
    0x4efab35700000ULL,
    0xe01ebe1d80000ULL,
    0x67abfb2a40000ULL,
    0x59a1dbbbe0000ULL,
    0x4f94ef7a70000ULL,
    0x6d5fe02a98000ULL,
    0x488f303cbc000ULL,
    0xd87388194a000ULL,
    0x5d6e5c1b79000ULL,
    0x7b415e05cd800ULL,
    0x4624cb0053c00ULL,
    0x8e52539884600ULL,
    0x4b3ab34de7900ULL,
    0x347ebe269a780ULL,
    0x605bfb3e700c0ULL,
    0x2509dbac98820ULL,
    0xa2b0ef4dbc070ULL,
    0x5f89e005ca2f8ULL,
    0x22f03035b93ecULL,
    0x1d2a080cec16eULL,
    0xd9e41c0902abdULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xd800000000000ULL,
    0x2c00000000000ULL,
    0xfa00000000000ULL,
    0x300000000000ULL,
    0xce80000000000ULL,
    0x31c0000000000ULL,
    0xff60000000000ULL,
    0xb4f0000000000ULL,
    0x5c08000000000ULL,
    0xae4000000000ULL,
    0xf7b6000000000ULL,
    0xd0ab000000000ULL,
    0x3e76800000000ULL,
    0xf3c8c00000000ULL,
    0x9886600000000ULL,
    0x8c3500000000ULL,
    0x52e5d80000000ULL,
    0x1bb5940000000ULL,
    0x4aa89e0000000ULL,
    0x6d72830000000ULL,
    0xe54a038000000ULL,
    0x8544764000000ULL,
    0xda5582000000ULL,
    0xe516543000000ULL,
    0x89387e1800000ULL,
    0x7189131400000ULL,
    0x65a1bbb600000ULL,
    0x8116b24b00000ULL,
    0xe73e1e1e80000ULL,
    0xf48a4330c0000ULL,
    0x8a2463aba0000ULL,
    0x5153265870000ULL,
    0x1d1e8029d8000ULL,
    0x8ddcc0163c000ULL,
    0x5338603182000ULL,
    0x828c501679000ULL,
    0x432558384d800ULL,
    0x14d654335ac00ULL,
    0x16587e0ec5a00ULL,
    0x957913110e300ULL,
    0xe1a9bb8283980ULL,
    0xa7f2b24cf9d40ULL,
    0xea881e368cf60ULL,
    0x2721432ffba50ULL,
    0x7ad2e39235878ULL,
    0x935be6711790cULL,
    0x7af8e0349e98eULL,
    0x31ef902c4b099ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x7800000000000ULL,
    0xc400000000000ULL,
    0xf200000000000ULL,
    0x9f00000000000ULL,
    0x7480000000000ULL,
    0x87c0000000000ULL,
    0x1820000000000ULL,
    0xfa70000000000ULL,
    0xa08000000000ULL,
    0x3ba4000000000ULL,
    0x7536000000000ULL,
    0xc3eb000000000ULL,
    0x84f5800000000ULL,
    0x16cdc00000000ULL,
    0xc501200000000ULL,
    0xa780100000000ULL,
    0xdd44380000000ULL,
    0x6e4040000000ULL,
    0x83d06e0000000ULL,
    0x583a190000000ULL,
    0x28cfe78000000ULL,
    0xb402964000000ULL,
    0x2a07b82000000ULL,
    0xb02c43000000ULL,
    0xce84ce1800000ULL,
    0xa4c7c90c00000ULL,
    0x5aa2ff9e00000ULL,
    0x10b2827100000ULL,
    0xf9adee1c80000ULL,
    0x3293d917c0000ULL,
    0xe0d8c78520000ULL,
    0x1a19866df0000ULL,
    0x1be003808000ULL,
    0xe48f003f9c000ULL,
    0xf63802602000ULL,
    0x796c02529000ULL,
    0x5f5ca01e6d800ULL,
    0x3359d0200ac00ULL,
    0x8d5b181bb5600ULL,
    0x3c5b141cef700ULL,
    0xd1dfd62ee2480ULL,
    0x379cdd06f9040ULL,
    0x87fd29ac148e0ULL,
    0x64ee5f4fdfc10ULL,
    0x9470c7a34b7b8ULL,
    0x330d866d54f64ULL,
    0xa20001e3e31eULL,
    0x957000081a619ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x4800000000000ULL,
    0xbc00000000000ULL,
    0xae00000000000ULL,
    0x100000000000ULL,
    0x1980000000000ULL,
    0xfcc0000000000ULL,
    0xc760000000000ULL,
    0x8bf0000000000ULL,
    0xca18000000000ULL,
    0x67e4000000000ULL,
    0xeab2000000000ULL,
    0x9bbb000000000ULL,
    0x2076800000000ULL,
    0xe9d8400000000ULL,
    0xc002e00000000ULL,
    0x2005f00000000ULL,
    0x3001480000000ULL,
    0x1807ac0000000ULL,
    0xf400d60000000ULL,
    0x1203d30000000ULL,
    0xaf01e38000000ULL,
    0x1881bec000000ULL,
    0xe545c82000000ULL,
    0x3ba4ec1000000ULL,
    0x4c94b61800000ULL,
    0x41ee631400000ULL,
    0xadfa4b9200000ULL,
    0x8d57e2ef00000ULL,
    0x710e560b80000ULL,
    0xbbcf931040000ULL,
    0xc9a9039e60000ULL,
    0x29db4eeb30000ULL,
    0xe0000003d8000ULL,
    0x1000001dfc000ULL,
    0x2800000106000ULL,
    0xec00000db9000ULL,
    0xe600002ecc800ULL,
    0xbd000036dec00ULL,
    0xb780001245a00ULL,
    0xfdc00018ca100ULL,
    0xdee00004e6b80ULL,
    0x7730002ec87c0ULL,
    0xd78000774d20ULL,
    0xec140002132b0ULL,
    0x20aa00079b958ULL,
    0xfc5f00270fe4cULL,
    0xcac480216c40eULL,
    0x72634028cfd3bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0x8400000000000ULL,
    0xae00000000000ULL,
    0x8900000000000ULL,
    0xea80000000000ULL,
    0xbb40000000000ULL,
    0xb0a0000000000ULL,
    0x3250000000000ULL,
    0xd508000000000ULL,
    0x6724000000000ULL,
    0x7596000000000ULL,
    0xabef000000000ULL,
    0xbd50800000000ULL,
    0x828e400000000ULL,
    0xe0e7200000000ULL,
    0xeb77f00000000ULL,
    0xe19b080000000ULL,
    0x424bf40000000ULL,
    0x4d82ea0000000ULL,
    0x18c4ed0000000ULL,
    0x160368000000ULL,
    0xa9b3fa4000000ULL,
    0x5ffd882000000ULL,
    0x47ab43000000ULL,
    0x69bd4a3800000ULL,
    0x30595d2400000ULL,
    0x89aa1e8200000ULL,
    0x2f70fe6100000ULL,
    0x6f9cea0b80000ULL,
    0xbb4fed1240000ULL,
    0x3f06b682a0000ULL,
    0x2f82ba4ad0000ULL,
    0x9bc2280e28000ULL,
    0xbce7041d94000ULL,
    0xe9776226c2000ULL,
    0x329a591f89000ULL,
    0x17cb7ca5c5800ULL,
    0x1945a7652bc00ULL,
    0xd3a71688fc200ULL,
    0x1fd40a7e77900ULL,
    0x9dce802fdbc80ULL,
    0x8e454019c4fc0ULL,
    0xc821a0332b420ULL,
    0x5216b033fd3d0ULL,
    0xb42ca820f69a8ULL,
    0xafb2443c1bab4ULL,
    0xe2fec2342445aULL,
    0x38f8e936dac29ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0xb400000000000ULL,
    0xbe00000000000ULL,
    0xe300000000000ULL,
    0xe80000000000ULL,
    0xa940000000000ULL,
    0x9f60000000000ULL,
    0xd430000000000ULL,
    0x8b18000000000ULL,
    0x2ee4000000000ULL,
    0xf1f6000000000ULL,
    0xec39000000000ULL,
    0x5232800000000ULL,
    0xcc1ac00000000ULL,
    0x2660a00000000ULL,
    0xdfb5700000000ULL,
    0x5bdad80000000ULL,
    0x7847540000000ULL,
    0x68e27a0000000ULL,
    0x96f7430000000ULL,
    0x34bda58000000ULL,
    0xe476cac000000ULL,
    0xcffe582000000ULL,
    0xb214943000000ULL,
    0x98485a0800000ULL,
    0xe5ccf33400000ULL,
    0x3a09dda600000ULL,
    0x4829eeed00000ULL,
    0xf57a7a0f80000ULL,
    0x65534308c0000ULL,
    0xc42ba58ba0000ULL,
    0xef7fcade50000ULL,
    0x8054d821d8000ULL,
    0x4daa54280c000ULL,
    0x6ebefa0546000ULL,
    0xe170830779000ULL,
    0xb6798599dd800ULL,
    0x7bd07afc5e400ULL,
    0xd56ea03cd4a00ULL,
    0x1418700ecab00ULL,
    0x72665820fe280ULL,
    0x91b09422045c0ULL,
    0xf0de5a2b53360ULL,
    0x5ac5f33d13950ULL,
    0xcba35d83f23e8ULL,
    0x54972ee5c260cULL,
    0xd0cda1b6ac16ULL,
    0xc8af332faeaebULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x3800000000000ULL,
    0xfc00000000000ULL,
    0x4200000000000ULL,
    0x4700000000000ULL,
    0xdd80000000000ULL,
    0x8c40000000000ULL,
    0xa3e0000000000ULL,
    0xbfd0000000000ULL,
    0xd8f8000000000ULL,
    0x1b64000000000ULL,
    0xb912000000000ULL,
    0xf2df000000000ULL,
    0x2716800000000ULL,
    0x7dbc00000000ULL,
    0x3594e00000000ULL,
    0xc21a900000000ULL,
    0x9c30f80000000ULL,
    0x27298c0000000ULL,
    0x239cb20000000ULL,
    0xf274470000000ULL,
    0x73cdc28000000ULL,
    0x29cefdc000000ULL,
    0x62cc782000000ULL,
    0x75494c1000000ULL,
    0x2a0cd23800000ULL,
    0xd26a171400000ULL,
    0x18ff5a8e00000ULL,
    0xbb6621ff00000ULL,
    0x914d23080000ULL,
    0x9ade1701c0000ULL,
    0xe3155a8f60000ULL,
    0xb9dd21f710000ULL,
    0x30905206f8000ULL,
    0x589ad700f4000ULL,
    0xcdf73a9ebe000ULL,
    0x88c71c319000ULL,
    0x3fac4a3724800ULL,
    0x955dcb30a7c00ULL,
    0xb0517089bda00ULL,
    0x8bbabaebc2f00ULL,
    0x2901baa23b380ULL,
    0xa087b1fa4fa40ULL,
    0xac0aa2930be0ULL,
    0xe0235b0f49a30ULL,
    0xef7388b67c8c8ULL,
    0xe54c36f3a2e1cULL,
    0xb20b88a9aac8aULL,
    0x9e6836d776db7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0x9400000000000ULL,
    0x2e00000000000ULL,
    0x4500000000000ULL,
    0x680000000000ULL,
    0x67c0000000000ULL,
    0xb8e0000000000ULL,
    0x4ad0000000000ULL,
    0x1ae8000000000ULL,
    0xe564000000000ULL,
    0x9692000000000ULL,
    0xef49000000000ULL,
    0x5292800000000ULL,
    0xc948400000000ULL,
    0xa391a00000000ULL,
    0x31cdf00000000ULL,
    0xc955680000000ULL,
    0x1ba8140000000ULL,
    0xca450e0000000ULL,
    0xfca7cf0000000ULL,
    0x6370528000000ULL,
    0x171be14000000ULL,
    0xd1bde82000000ULL,
    0x99cd541000000ULL,
    0xed542e0800000ULL,
    0x8dab7f3400000ULL,
    0x83461a9200000ULL,
    0x826456500000ULL,
    0xfbb42e2b80000ULL,
    0xc27b7f0140000ULL,
    0xbf2e1a89a0000ULL,
    0x5a82456df0000ULL,
    0x9dc62e1c38000ULL,
    0xf3e27f27b4000ULL,
    0xd9549a853a000ULL,
    0x33ae055c19000ULL,
    0xae458e1604800ULL,
    0x4aa68f1722400ULL,
    0xfa7372b91ca00ULL,
    0xab9e5145a6100ULL,
    0xdd79201dfe680ULL,
    0x62a8b019da7c0ULL,
    0xc6c4483a49da0ULL,
    0x4864a40acc450ULL,
    0x941346332fe38ULL,
    0xce8a6b08d2e3cULL,
    0xeb7194b024fcaULL,
    0xe319ca7193c45ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xd800000000000ULL,
    0x5400000000000ULL,
    0x1a00000000000ULL,
    0xe900000000000ULL,
    0x3480000000000ULL,
    0xc240000000000ULL,
    0x2b20000000000ULL,
    0x98b0000000000ULL,
    0x2288000000000ULL,
    0x14a4000000000ULL,
    0xa376000000000ULL,
    0x8d6b000000000ULL,
    0xd072800000000ULL,
    0x90e8c00000000ULL,
    0x46b5a00000000ULL,
    0xa988500000000ULL,
    0x2d26880000000ULL,
    0x47b3840000000ULL,
    0x10b020000000ULL,
    0x1966310000000ULL,
    0x1b95718000000ULL,
    0x8639ee4000000ULL,
    0x702a082000000ULL,
    0xc8d4443000000ULL,
    0xbd9a221800000ULL,
    0x58dda13400000ULL,
    0x747cd9b600000ULL,
    0x930efa5500000ULL,
    0xbc64222680000ULL,
    0x112a10a40000ULL,
    0x6778599520000ULL,
    0x5e8d3a4490000ULL,
    0xf2a3021cc8000ULL,
    0x6c7231032c000ULL,
    0x16eb719622000ULL,
    0x59b6ee4b69000ULL,
    0xea0e8833fd800ULL,
    0xf0e78436cac00ULL,
    0x2755022054a00ULL,
    0x5e59310c56300ULL,
    0xa339f1a22f680ULL,
    0xaaae2e4adb140ULL,
    0x4993280fdc220ULL,
    0x433bd41eda210ULL,
    0xbaad8a29dc608ULL,
    0xf195b523dabc4ULL,
    0xc73e73835db46ULL,
    0x78afdf5b9aaf9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xb800000000000ULL,
    0x3400000000000ULL,
    0x2a00000000000ULL,
    0x9300000000000ULL,
    0x6e80000000000ULL,
    0x8940000000000ULL,
    0x920000000000ULL,
    0xaa90000000000ULL,
    0xc438000000000ULL,
    0x9aa4000000000ULL,
    0xec52000000000ULL,
    0xdddf000000000ULL,
    0xfcd1800000000ULL,
    0x1d9fc00000000ULL,
    0x5072a00000000ULL,
    0xd4ced00000000ULL,
    0x862a580000000ULL,
    0x4ad9540000000ULL,
    0x4051da0000000ULL,
    0xd3ddf10000000ULL,
    0xadd4308000000ULL,
    0xdc1d13c000000ULL,
    0x11b1d82000000ULL,
    0xd9ad941000000ULL,
    0xf218fa3800000ULL,
    0x50b7e12c00000ULL,
    0xf02f48ae00000ULL,
    0x8fda57cd00000ULL,
    0xfbd17a2a80000ULL,
    0x91c2134c0000ULL,
    0x4237e8a3a0000ULL,
    0x3f6f87ce50000ULL,
    0xbbf8a20c48000ULL,
    0x3ec5b537a4000ULL,
    0x8de512838e000ULL,
    0x1c7366ee69000ULL,
    0x6acc6aaeb4800ULL,
    0x6f2b22c827c00ULL,
    0xbf5d48affc600ULL,
    0x2b9557f503f00ULL,
    0x4db8fa08b2a80ULL,
    0xb767e100cab40ULL,
    0x4f37488456160ULL,
    0xa8ee57caa5950ULL,
    0x2f3b7a010e168ULL,
    0x6027210ea98c4ULL,
    0x1f1468af15242ULL,
    0x4fff47f57730fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xc800000000000ULL,
    0x6c00000000000ULL,
    0x1e00000000000ULL,
    0x2d00000000000ULL,
    0x7180000000000ULL,
    0x9c0000000000ULL,
    0x960000000000ULL,
    0x4a70000000000ULL,
    0x6f58000000000ULL,
    0xc0a4000000000ULL,
    0xe312000000000ULL,
    0xb52f000000000ULL,
    0x17ff800000000ULL,
    0x87b0400000000ULL,
    0x243e600000000ULL,
    0x91d0700000000ULL,
    0xcccea80000000ULL,
    0xdbcdac0000000ULL,
    0xb54e5e0000000ULL,
    0x1e0e930000000ULL,
    0x3229578000000ULL,
    0xc17c88c000000ULL,
    0x88f0c82000000ULL,
    0x7a1ddc1000000ULL,
    0x5180f63800000ULL,
    0xd9c33f1c00000ULL,
    0x516709b200000ULL,
    0x9e721bdb00000ULL,
    0xd5599fa780000ULL,
    0x9fa154db40000ULL,
    0xa1903e0460000ULL,
    0xe06ee30e70000ULL,
    0x66dfff8858000ULL,
    0xcd6524d59c000ULL,
    0x874960e56000ULL,
    0x74584f3069000ULL,
    0x8024219b24800ULL,
    0x4d50f7f87bc00ULL,
    0xf48e2189c7e00ULL,
    0x3b6bf7fa00100ULL,
    0x865ba18f01980ULL,
    0xf320b7c1811c0ULL,
    0xf1d041a6412a0ULL,
    0xfccb87fee1ab0ULL,
    0xf3cd09bfb0f78ULL,
    0xa9491bc5f874cULL,
    0xc80c1faa4c2deULL,
    0x732a14dbde2e3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0xc400000000000ULL,
    0xc600000000000ULL,
    0xa700000000000ULL,
    0x7180000000000ULL,
    0x340000000000ULL,
    0x2c20000000000ULL,
    0x4170000000000ULL,
    0xfc58000000000ULL,
    0xf64000000000ULL,
    0x5d52000000000ULL,
    0x752d000000000ULL,
    0x473f800000000ULL,
    0xfc35400000000ULL,
    0x4b79200000000ULL,
    0x8594b00000000ULL,
    0x1f4ff80000000ULL,
    0xeaefec0000000ULL,
    0x8cdd8a0000000ULL,
    0x4aa2290000000ULL,
    0xbbb61f8000000ULL,
    0xc73f84c000000ULL,
    0xbc36d82000000ULL,
    0xeb7b5c1000000ULL,
    0xf592722800000ULL,
    0x774dc51c00000ULL,
    0x2eeb959a00000ULL,
    0x4addadf100000ULL,
    0xeda0c79180000ULL,
    0xca34d8f9c0000ULL,
    0xc47caa1460000ULL,
    0x9012990cd0000ULL,
    0xaa0be7b908000ULL,
    0x9cd68fd5c000ULL,
    0x782cd23496000ULL,
    0x73b8350b19000ULL,
    0x3ff74d92b4800ULL,
    0xaa9ff1e85b400ULL,
    0x360735bca7e00ULL,
    0x8f055dce81500ULL,
    0x15839f9f40480ULL,
    0xb543c4d7202c0ULL,
    0xe3227821f17e0ULL,
    0xf4f7ac19f9bb0ULL,
    0x391b2a2d24028ULL,
    0x8443d924729a4ULL,
    0x6da0c7a50a2feULL,
    0x8a34d8f46e9d3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0x2400000000000ULL,
    0x3600000000000ULL,
    0x700000000000ULL,
    0xe680000000000ULL,
    0xb840000000000ULL,
    0xc2e0000000000ULL,
    0xcab0000000000ULL,
    0x1a8000000000ULL,
    0x2a4000000000ULL,
    0x6c52000000000ULL,
    0x9d1b000000000ULL,
    0xd40a800000000ULL,
    0x7071c00000000ULL,
    0x6d08200000000ULL,
    0xbdf0300000000ULL,
    0x69cb780000000ULL,
    0xfa56d40000000ULL,
    0x6a19d20000000ULL,
    0x7a8f5f0000000ULL,
    0xc430f98000000ULL,
    0x45ed8b4000000ULL,
    0x6243582000000ULL,
    0xbfe6e41000000ULL,
    0xa132aa1800000ULL,
    0x9a698b3400000ULL,
    0x57812bbe00000ULL,
    0x19c6d44900000ULL,
    0xfc21a1ad80000ULL,
    0x60106f51c0000ULL,
    0xf1fb7201a0000ULL,
    0xedbeaf0a10000ULL,
    0x595ba196b8000ULL,
    0xf4ef6f4fac000ULL,
    0xc3c3f213ea000ULL,
    0x81246f1869000ULL,
    0xb9101b734800ULL,
    0x6a3a9f7c96c00ULL,
    0xb89aaa221aa00ULL,
    0x2ccd8b19a0700ULL,
    0x95d32bbe10080ULL,
    0x77ddd468b90c0ULL,
    0xab21a6ac5e0ULL,
    0x8b21af6e6a750ULL,
    0x8e935231a8548ULL,
    0x25be9f329567cULL,
    0x1558d98086966ULL,
    0x7eedbb71a21edULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0xa400000000000ULL,
    0xc200000000000ULL,
    0xa500000000000ULL,
    0x5580000000000ULL,
    0xcd40000000000ULL,
    0x4960000000000ULL,
    0x8590000000000ULL,
    0x47a8000000000ULL,
    0x9024000000000ULL,
    0x76f6000000000ULL,
    0x4b39000000000ULL,
    0xf40c800000000ULL,
    0xe497400000000ULL,
    0x202a200000000ULL,
    0x6063900000000ULL,
    0x7616e80000000ULL,
    0xcdebcc0000000ULL,
    0x39c1da0000000ULL,
    0xc3a2330000000ULL,
    0x2cb7888000000ULL,
    0x6ad86b4000000ULL,
    0xf55cc82000000ULL,
    0xe9185c3000000ULL,
    0xbfff320800000ULL,
    0xe62dff0400000ULL,
    0xd76052aa00000ULL,
    0xe93586900000ULL,
    0xe12fc09080000ULL,
    0xd7e3775940000ULL,
    0xe57da3d60000ULL,
    0x790b330750000ULL,
    0xd51308b058000ULL,
    0xf6b2b4c64000ULL,
    0x9c00e82b6a000ULL,
    0xe02cc1449000ULL,
    0xe3055a305d800ULL,
    0x2281730cde400ULL,
    0x95c3a8903b200ULL,
    0x75a6fb4211d00ULL,
    0xc3b0a01b38880ULL,
    0x2e5dd00435e40ULL,
    0x798480cb23a0ULL,
    0x6c3b1c32edb30ULL,
    0xe48b121696d68ULL,
    0x4c536f13671ccULL,
    0x1c0c3a9f4e4a2ULL,
    0xa096d476c22edULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x8800000000000ULL,
    0xa400000000000ULL,
    0x3a00000000000ULL,
    0xbb00000000000ULL,
    0x2b80000000000ULL,
    0x95c0000000000ULL,
    0xb760000000000ULL,
    0x5730000000000ULL,
    0xb958000000000ULL,
    0xdea4000000000ULL,
    0xa656000000000ULL,
    0xd369000000000ULL,
    0x5179800000000ULL,
    0x6bb6400000000ULL,
    0x591a200000000ULL,
    0x5d01500000000ULL,
    0x6682180000000ULL,
    0xb413c0000000ULL,
    0xb025de0000000ULL,
    0xc915f30000000ULL,
    0x794b1d8000000ULL,
    0x936de8c000000ULL,
    0xb178382000000ULL,
    0xfbb06c3000000ULL,
    0x611fc60800000ULL,
    0x7100cf2c00000ULL,
    0xf880c3a200000ULL,
    0x8a451be900000ULL,
    0x20a4a5ae80000ULL,
    0x7756c4dec0000ULL,
    0x5bea5e22e0000ULL,
    0x733ab33970000ULL,
    0x5f10bd87d8000ULL,
    0x9c4ef8d0cc000ULL,
    0x19ee0022d6000ULL,
    0x43d000069000ULL,
    0x7a978025f5800ULL,
    0xb08b40136a400ULL,
    0x120da01166600ULL,
    0x734a100351900ULL,
    0x606fb81b48880ULL,
    0x1efb2c1be5540ULL,
    0xe472661483060ULL,
    0xe57adf24d30f0ULL,
    0xa9b77bbc1a978ULL,
    0xee1a37c9beaccULL,
    0x5880c3921c2f6ULL,
    0xfa451beedfbe3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x4800000000000ULL,
    0xc400000000000ULL,
    0xb200000000000ULL,
    0xcd00000000000ULL,
    0xcc80000000000ULL,
    0x7140000000000ULL,
    0x7a0000000000ULL,
    0x7830000000000ULL,
    0x2748000000000ULL,
    0x94e4000000000ULL,
    0x9a92000000000ULL,
    0xe7f9000000000ULL,
    0xa8e8800000000ULL,
    0x48d4400000000ULL,
    0xbfdd200000000ULL,
    0x661c100000000ULL,
    0x72fea80000000ULL,
    0xa86d040000000ULL,
    0x6f93ba0000000ULL,
    0x577e050000000ULL,
    0x57ae2d8000000ULL,
    0x2c77a44000000ULL,
    0xf86b882000000ULL,
    0x4795141000000ULL,
    0x237f120800000ULL,
    0x3daa012400000ULL,
    0x6d75179200000ULL,
    0x42ede17100000ULL,
    0x49d0858c80000ULL,
    0x255aa06340000ULL,
    0xf858321b20000ULL,
    0x3cdb112850000ULL,
    0xa9993f9be8000ULL,
    0x8c39a54b0c000ULL,
    0x710c9f9752000ULL,
    0x5ec1f57779000ULL,
    0x967179104800ULL,
    0x5954e162ee400ULL,
    0xa1805a2f2200ULL,
    0xf4fee05569100ULL,
    0x2f6d12314d480ULL,
    0x72130114f2040ULL,
    0xd4bd97b6692a0ULL,
    0x9a49a14ccc010ULL,
    0xe065a5bb324e8ULL,
    0xd7d2b06e08d14ULL,
    0x4e5c9a2fbce36ULL,
    0x23db15208afd1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0xac00000000000ULL,
    0xb200000000000ULL,
    0x8f00000000000ULL,
    0xd780000000000ULL,
    0x14c0000000000ULL,
    0x9b60000000000ULL,
    0x7030000000000ULL,
    0x7078000000000ULL,
    0x35a4000000000ULL,
    0xe856000000000ULL,
    0xb5cf000000000ULL,
    0xa61d800000000ULL,
    0x7510400000000ULL,
    0x276c600000000ULL,
    0xc5cf100000000ULL,
    0xce1db80000000ULL,
    0x9149c0000000ULL,
    0xed68d60000000ULL,
    0xaecdf30000000ULL,
    0x79de58000000ULL,
    0x20d229c000000ULL,
    0x2e89d82000000ULL,
    0x1dbf8c3000000ULL,
    0xf8436e3800000ULL,
    0xfe266f0c00000ULL,
    0xc690b39200000ULL,
    0xedab9aeb00000ULL,
    0x83ae5d8c80000ULL,
    0xd6adb5d3c0000ULL,
    0xf22a8e2de0000ULL,
    0xfb6d3f3930000ULL,
    0x1fcf6b8cd8000ULL,
    0x3d1b16fb0c000ULL,
    0x1490b3829e000ULL,
    0x92ab9af5a9000ULL,
    0xfc2e5dbb75800ULL,
    0x5e6db5f783c00ULL,
    0x934a8e30bf600ULL,
    0xa85d3f1378100ULL,
    0xa376bb59d180ULL,
    0x937f16e4d6c40ULL,
    0xb026b3beecee0ULL,
    0x43949adfafe70ULL,
    0xb12bdd8f70d58ULL,
    0x6ee9f5dd78accULL,
    0x2c08ee249d016ULL,
    0xedf92f27565e7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x1800000000000ULL,
    0x6400000000000ULL,
    0x7600000000000ULL,
    0xdb00000000000ULL,
    0x1580000000000ULL,
    0x37c0000000000ULL,
    0x63a0000000000ULL,
    0xa5b0000000000ULL,
    0xbbf8000000000ULL,
    0x6a64000000000ULL,
    0xa116000000000ULL,
    0xb9cd000000000ULL,
    0x9b5a800000000ULL,
    0x7e52c00000000ULL,
    0xfb28200000000ULL,
    0xbb0c300000000ULL,
    0x927a980000000ULL,
    0x2fa44c0000000ULL,
    0x1fb4ae0000000ULL,
    0x1afd350000000ULL,
    0x7ee2398000000ULL,
    0x52529dc000000ULL,
    0x312ab82000000ULL,
    0xd20c7c3000000ULL,
    0xfaf8362800000ULL,
    0xcee4793400000ULL,
    0x3a54178600000ULL,
    0x9d2968d900000ULL,
    0xd80ea1bd80000ULL,
    0x33fbd1c6c0000ULL,
    0x7664960d60000ULL,
    0xc3138909f0000ULL,
    0xdcca2fb6e8000ULL,
    0x29ded4c46c000ULL,
    0x4e1237957e000ULL,
    0x24c58c559000ULL,
    0xb398b9b8a5800ULL,
    0xae705dd143400ULL,
    0xaa5a98355ea00ULL,
    0x12d44c2608b00ULL,
    0x1c6cae3b5c080ULL,
    0x8de93506f60c0ULL,
    0x7c2c39bcc5260ULL,
    0xac8b9ddd33530ULL,
    0xb0be3829760b8ULL,
    0x3687bc1504bd4ULL,
    0xf644961cd2266ULL,
    0x4a63892fc61b7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0xdc00000000000ULL,
    0x8200000000000ULL,
    0xcf00000000000ULL,
    0x6a80000000000ULL,
    0x4dc0000000000ULL,
    0xa0e0000000000ULL,
    0x4f50000000000ULL,
    0xca18000000000ULL,
    0xc324000000000ULL,
    0xdab2000000000ULL,
    0x5ccd000000000ULL,
    0x5c7b800000000ULL,
    0x3b7c00000000ULL,
    0xdf49e00000000ULL,
    0x4a3a900000000ULL,
    0x3916380000000ULL,
    0xe0bb740000000ULL,
    0x94d61e0000000ULL,
    0x105b2f0000000ULL,
    0x1381768000000ULL,
    0x3e401ac000000ULL,
    0x8ea7d82000000ULL,
    0xa9f5e41000000ULL,
    0x77ea262800000ULL,
    0xd2c95b3c00000ULL,
    0x997ee88e00000ULL,
    0xe231f5f700000ULL,
    0x560cce8080000ULL,
    0xd21caee3c0000ULL,
    0xaf202612a0000ULL,
    0xc0b05b3f70000ULL,
    0xbfcf688e38000ULL,
    0x8cff35d8d4000ULL,
    0x5d74aebc06000ULL,
    0x9728fee409000ULL,
    0x87ae7e048c800ULL,
    0x396f7f1b83400ULL,
    0x210f2e8386e00ULL,
    0x249f3ecb49f00ULL,
    0xc0e79e23ec780ULL,
    0x5f55ef2553a40ULL,
    0xa21916984f0e0ULL,
    0xd7244ac3649d0ULL,
    0xbcb180181fe78ULL,
    0xcdcec0346c7bcULL,
    0x7bf8603612c5aULL,
    0xebf450312e52bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xa800000000000ULL,
    0xe400000000000ULL,
    0x3e00000000000ULL,
    0xb700000000000ULL,
    0xaf80000000000ULL,
    0xd640000000000ULL,
    0x71e0000000000ULL,
    0x6930000000000ULL,
    0xdf98000000000ULL,
    0x75a4000000000ULL,
    0xe5d6000000000ULL,
    0x682f000000000ULL,
    0x98f9800000000ULL,
    0xea57400000000ULL,
    0x2e6e200000000ULL,
    0x111f700000000ULL,
    0x7f67780000000ULL,
    0x83f61c0000000ULL,
    0x9bae20000000ULL,
    0xbc33070000000ULL,
    0x251fed8000000ULL,
    0x9960e2c000000ULL,
    0x98f1582000000ULL,
    0xcc3d6c3000000ULL,
    0x4b739a3800000ULL,
    0xa87e1b2c00000ULL,
    0xb7928faa00000ULL,
    0x57cfa5f900000ULL,
    0xecf15af80000ULL,
    0xce4abeddc0000ULL,
    0x1b8a1a33e0000ULL,
    0xd6295b2990000ULL,
    0xeffcaf8e78000ULL,
    0xa5d0d5cf4c000ULL,
    0x48286db266000ULL,
    0xc8fca2f9a9000ULL,
    0xf250f80515800ULL,
    0x626a5c1e5bc00ULL,
    0xcb1b4203a6600ULL,
    0xf664370049d00ULL,
    0x9b77358585880ULL,
    0x707aced7a2dc0ULL,
    0x1b94e2152a5e0ULL,
    0x3dc8070fcf470ULL,
    0x2fc86d92bd588ULL,
    0x32cca2cc4ed1cULL,
    0x5c48f8057c136ULL,
    0xc68e5c16af34bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x400000000000ULL,
    0x2a00000000000ULL,
    0x2f00000000000ULL,
    0xc180000000000ULL,
    0x5fc0000000000ULL,
    0xc720000000000ULL,
    0x730000000000ULL,
    0x76e8000000000ULL,
    0x9ee4000000000ULL,
    0x7916000000000ULL,
    0xa15d000000000ULL,
    0x1b49800000000ULL,
    0x4d94c00000000ULL,
    0xa81b200000000ULL,
    0x562dd00000000ULL,
    0x9040380000000ULL,
    0xede3ec0000000ULL,
    0x56930e0000000ULL,
    0x5b9d510000000ULL,
    0x82edad8000000ULL,
    0xce4354000000ULL,
    0x4213182000000ULL,
    0x22da3c3000000ULL,
    0xb78d362800000ULL,
    0x6537bd2c00000ULL,
    0xf5e923b600000ULL,
    0x964a44100000ULL,
    0x97d215aa80000ULL,
    0x7afe197bc0000ULL,
    0xa93ab63860000ULL,
    0x481a7d0bf0000ULL,
    0x862d83afc8000ULL,
    0x5840b46ccc000ULL,
    0x81e08d813a000ULL,
    0xa094e55d79000ULL,
    0x5a9aa02ca5800ULL,
    0x466d102867400ULL,
    0xbda518197a600ULL,
    0x1b773c2859300ULL,
    0xbd0cb63af4c80ULL,
    0x1777d393e740ULL,
    0x8a0c039d0f8e0ULL,
    0x64f07468a6bb0ULL,
    0x4fcdadac9b238ULL,
    0x54d4354be9244ULL,
    0x8d7b18281dc36ULL,
    0x57fe3c3fa2495ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0x5c00000000000ULL,
    0xae00000000000ULL,
    0xad00000000000ULL,
    0x5a80000000000ULL,
    0x9ac0000000000ULL,
    0xcea0000000000ULL,
    0xf5b0000000000ULL,
    0xbdc8000000000ULL,
    0x9964000000000ULL,
    0x9196000000000ULL,
    0x8ab9000000000ULL,
    0xce0c800000000ULL,
    0x6741400000000ULL,
    0xf9e5e00000000ULL,
    0x8454f00000000ULL,
    0x559ea80000000ULL,
    0x1aff740000000ULL,
    0x1269460000000ULL,
    0x30556d0000000ULL,
    0x679f918000000ULL,
    0x39fd2c4000000ULL,
    0xf5ed482000000ULL,
    0x9812843000000ULL,
    0x6ffb6e0800000ULL,
    0xaceb590400000ULL,
    0x1093379a00000ULL,
    0xe63cb15700000ULL,
    0xfdcc718b80000ULL,
    0x7960dc5b40000ULL,
    0xa197603ea0000ULL,
    0xf2b8b012b0000ULL,
    0xfa09c821a8000ULL,
    0x9547c42e6c000ULL,
    0xfae08e1ef2000ULL,
    0x73d2a91a19000ULL,
    0x95df1f9945800ULL,
    0x4e9f856f5e400ULL,
    0x297a57b68b200ULL,
    0x782901720c500ULL,
    0x4337398aa3780ULL,
    0x310b584e203c0ULL,
    0xeec08e2c702a0ULL,
    0xdca2a921489d0ULL,
    0xc6b71faafcf18ULL,
    0x324b854f2b4b4ULL,
    0x6d2457933dbc6ULL,
    0x37f4015c4a271ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x8800000000000ULL,
    0xac00000000000ULL,
    0x4a00000000000ULL,
    0x3d00000000000ULL,
    0x9980000000000ULL,
    0x9b40000000000ULL,
    0xdf20000000000ULL,
    0x370000000000ULL,
    0x398000000000ULL,
    0x3164000000000ULL,
    0x49d2000000000ULL,
    0x61af000000000ULL,
    0x8de800000000ULL,
    0xc642400000000ULL,
    0x56a0600000000ULL,
    0xf037d00000000ULL,
    0x20bf280000000ULL,
    0xf011ec0000000ULL,
    0xdb49ea0000000ULL,
    0x834e410000000ULL,
    0xe749bf8000000ULL,
    0xe14d4ac000000ULL,
    0xc64d482000000ULL,
    0xeac93c1000000ULL,
    0x4408423800000ULL,
    0x4a6ded1400000ULL,
    0xab3835a200000ULL,
    0x3250dbeb00000ULL,
    0x9be9dfb280000ULL,
    0x3c7a9adf40000ULL,
    0x1772603e60000ULL,
    0x7d98d022d0000ULL,
    0x8261a82dc8000ULL,
    0x5b53ac3fdc000ULL,
    0x9c698a3066000ULL,
    0x4439912819000ULL,
    0x52d697a694800ULL,
    0x2f2ca6f1fbc00ULL,
    0x871ca2339fa00ULL,
    0xc3a37d079c900ULL,
    0x35b3fd9a86180ULL,
    0xc9ffa7e488f40ULL,
    0x6633fd98bd4a0ULL,
    0x2fbfa7d336bb0ULL,
    0xc093fdac115a8ULL,
    0xe78fa7c268c04ULL,
    0x942bfdb22dd7eULL,
    0x799ba7fc1e36bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0xf400000000000ULL,
    0xf600000000000ULL,
    0xf00000000000ULL,
    0x7680000000000ULL,
    0xa7c0000000000ULL,
    0xb820000000000ULL,
    0x1d30000000000ULL,
    0x3b98000000000ULL,
    0xa8e4000000000ULL,
    0x2f96000000000ULL,
    0x636f000000000ULL,
    0x225b800000000ULL,
    0x2bc4400000000ULL,
    0x5226a00000000ULL,
    0x37300000000ULL,
    0x8c19b80000000ULL,
    0x4aa6540000000ULL,
    0x44f32e0000000ULL,
    0x98bd090000000ULL,
    0xc550a58000000ULL,
    0x74cf16c000000ULL,
    0xf129182000000ULL,
    0x413e643000000ULL,
    0xc411163800000ULL,
    0xb32f1d3c00000ULL,
    0x983d2bb200000ULL,
    0xad912ffd00000ULL,
    0x5a6e059d80000ULL,
    0xbbdc26f3c0000ULL,
    0xa86a005a0000ULL,
    0x85c73025f0000ULL,
    0x5121b82408000ULL,
    0x5cb254064c000ULL,
    0xd6dd2e0166000ULL,
    0xb5060914f9000ULL,
    0x738525abc5800ULL,
    0x3c4056f2ebc00ULL,
    0x6062381b3ee00ULL,
    0xb35214144d100ULL,
    0xbbcd8e0c67a80ULL,
    0x67ae391178cc0ULL,
    0x16ff1d9405ee0ULL,
    0xb43642dd4b550ULL,
    0x5a19b60fcfab8ULL,
    0x55a32d07c5924ULL,
    0xa711388eb616ULL,
    0x3783bc23f19bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0xe400000000000ULL,
    0x9600000000000ULL,
    0xed00000000000ULL,
    0x2a80000000000ULL,
    0xaec0000000000ULL,
    0x13e0000000000ULL,
    0xdf90000000000ULL,
    0x9518000000000ULL,
    0xdc24000000000ULL,
    0xd4f2000000000ULL,
    0x7f4f000000000ULL,
    0xcd5e800000000ULL,
    0xbb01400000000ULL,
    0x6782200000000ULL,
    0x3441300000000ULL,
    0x522a80000000ULL,
    0xc071340000000ULL,
    0x388ee60000000ULL,
    0x323f470000000ULL,
    0xcf56c08000000ULL,
    0x2ffd8ac000000ULL,
    0xf32882000000ULL,
    0xba2f041000000ULL,
    0x960ace3800000ULL,
    0x1afb331400000ULL,
    0x39b006ba00000ULL,
    0x6ee8fdf900000ULL,
    0x4aea608580000ULL,
    0x7cedfaeb40000ULL,
    0x21ea0012a0000ULL,
    0xb36b002fb0000ULL,
    0x11ac8006f8000ULL,
    0x704e401ae4000ULL,
    0xd4dca03ac6000ULL,
    0x8640703549000ULL,
    0xde20881a1c800ULL,
    0xb7f0040f23c00ULL,
    0x4cc4e3d0fa00ULL,
    0x831e733294500ULL,
    0x712026b85e880ULL,
    0x1e76cdcdfd4c0ULL,
    0x818e48b1322a0ULL,
    0x36b98ec8c68d0ULL,
    0x8094c61048998ULL,
    0x649a77399c41cULL,
    0x566689b62982ULL,
    0xfb53bec52f7ebULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xc800000000000ULL,
    0x4c00000000000ULL,
    0x6600000000000ULL,
    0x100000000000ULL,
    0x4980000000000ULL,
    0x7c0000000000ULL,
    0x6fe0000000000ULL,
    0xa390000000000ULL,
    0x1388000000000ULL,
    0xd24000000000ULL,
    0x7ff2000000000ULL,
    0x88d9000000000ULL,
    0xdc9800000000ULL,
    0x7d00400000000ULL,
    0xa781600000000ULL,
    0x6ac7b00000000ULL,
    0xf063c80000000ULL,
    0x6d566c0000000ULL,
    0x79e96e0000000ULL,
    0xcf74470000000ULL,
    0x29d588000000ULL,
    0x6fefee4000000ULL,
    0x6670a82000000ULL,
    0x9718dc1000000ULL,
    0x762b260800000ULL,
    0x8c966b2c00000ULL,
    0x570f56b200000ULL,
    0x561195300000ULL,
    0xa6d5b8b980000ULL,
    0xc5281e5040000ULL,
    0xf612003a60000ULL,
    0x6a49003df0000ULL,
    0x77c18021f8000ULL,
    0xc7e44017e4000ULL,
    0x7f93602f62000ULL,
    0xd8eb01009000ULL,
    0x882248141c800ULL,
    0x1c722c2386400ULL,
    0xe81a0e0aea600ULL,
    0x2aaaf73654100ULL,
    0x90d710b6fa580ULL,
    0xdc29c2711cec0ULL,
    0xab90a60706720ULL,
    0x3f8f2b2a2adb0ULL,
    0xfb27b68af4bb8ULL,
    0x6f6e962ab41cULL,
    0x455f10b8356e2ULL,
    0x200dc25d0a079ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0x3c00000000000ULL,
    0xe600000000000ULL,
    0xaf00000000000ULL,
    0xc580000000000ULL,
    0x61c0000000000ULL,
    0x8ea0000000000ULL,
    0x7230000000000ULL,
    0x24e8000000000ULL,
    0xa464000000000ULL,
    0x9b16000000000ULL,
    0x4a1f000000000ULL,
    0xeeae800000000ULL,
    0x4805400000000ULL,
    0x4c00e00000000ULL,
    0xbe02b00000000ULL,
    0x2b00b80000000ULL,
    0xf7808c0000000ULL,
    0x14c5de0000000ULL,
    0x2263d0000000ULL,
    0x7970c58000000ULL,
    0xe0c054000000ULL,
    0x3936582000000ULL,
    0x436d3c3000000ULL,
    0xb8a3e63800000ULL,
    0xf537f11400000ULL,
    0x3d6bfbba00000ULL,
    0x73a3884f00000ULL,
    0x52b6a59980000ULL,
    0xc1aff55bc0000ULL,
    0x4d80002960000ULL,
    0xcdc0003c70000ULL,
    0x60a00021a8000ULL,
    0xb13000078c000ULL,
    0xef68000aba000ULL,
    0x56a4000dd9000ULL,
    0x3636002e25800ULL,
    0xf6ef002137c00ULL,
    0x8166802963a00ULL,
    0xff914016fd500ULL,
    0x7d5ee03992380ULL,
    0x2249b01a15ac0ULL,
    0x7a5038253fae0ULL,
    0x51fecc067ee30ULL,
    0x2d7dbe185f578ULL,
    0xbe3ecd000f1f4ULL,
    0xf0de1db6f7a96ULL,
    0xf8b797e02fd5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0x1c00000000000ULL,
    0x1600000000000ULL,
    0xc100000000000ULL,
    0x2c80000000000ULL,
    0x3cc0000000000ULL,
    0x4620000000000ULL,
    0xecb0000000000ULL,
    0x8e38000000000ULL,
    0xede4000000000ULL,
    0x2716000000000ULL,
    0xa2cd000000000ULL,
    0xe13f800000000ULL,
    0xf466400000000ULL,
    0xed53e00000000ULL,
    0xf72e300000000ULL,
    0x322dc80000000ULL,
    0x8cafc40000000ULL,
    0x236eb60000000ULL,
    0x7ac8090000000ULL,
    0xc538c98000000ULL,
    0xd664d3c000000ULL,
    0x5650282000000ULL,
    0x84a8f43000000ULL,
    0xff6afe2800000ULL,
    0xcccc8d1c00000ULL,
    0x743a1fa200000ULL,
    0x72e4aac700000ULL,
    0x7696c9a580000ULL,
    0xd48dd3c040000ULL,
    0xd2d9a80320000ULL,
    0x6e73b42330000ULL,
    0xa51e9e1b88000ULL,
    0x13d0fd202c000ULL,
    0x38ea37840e000ULL,
    0xbb8c5ecc39000ULL,
    0xcb5c37a765800ULL,
    0xa4315ee7c3400ULL,
    0xf0fbb7aae7e00ULL,
    0xc0c31ee505900ULL,
    0x402657a352f80ULL,
    0x75b42ed69e8c0ULL,
    0x46ba1f8940f20ULL,
    0x5324aaeba1b10ULL,
    0xaa36c99df1cd8ULL,
    0xb5fdd3df68d24ULL,
    0xbe41a8077d2a6ULL,
    0x4fe7b41a1728fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0xe400000000000ULL,
    0xe00000000000ULL,
    0x7d00000000000ULL,
    0xc980000000000ULL,
    0xecc0000000000ULL,
    0x9da0000000000ULL,
    0x1990000000000ULL,
    0xfc98000000000ULL,
    0x1664000000000ULL,
    0xe6b6000000000ULL,
    0xc949000000000ULL,
    0xd419800000000ULL,
    0x3923c00000000ULL,
    0x74d0200000000ULL,
    0x71fcf00000000ULL,
    0xc853a80000000ULL,
    0x48beac0000000ULL,
    0x7335360000000ULL,
    0xf8cf50000000ULL,
    0x1ab9a88000000ULL,
    0x24366ec000000ULL,
    0x950d882000000ULL,
    0xf2ff5c3000000ULL,
    0xa4d11e0800000ULL,
    0x19fc992c00000ULL,
    0x84533e8e00000ULL,
    0x2abcabf900000ULL,
    0xdc3628a380000ULL,
    0x510caeef40000ULL,
    0x4cfc281a60000ULL,
    0xe1d46c2730000ULL,
    0x347c962168000ULL,
    0x6693c51364000ULL,
    0xca1a20b2a6000ULL,
    0xc2432e3d9000ULL,
    0x415316284d800ULL,
    0xc739052622400ULL,
    0x1ef380a80e600ULL,
    0x1ab02df1cf00ULL,
    0xa4e89e2b7a080ULL,
    0x158f5916823c0ULL,
    0xb1bb1ebf1f6a0ULL,
    0xe4b45bd6a4eb0ULL,
    0x364b808c06ad8ULL,
    0x229f02cdc98d4ULL,
    0x3669e2274022ULL,
    0x6332591e1e57bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xa800000000000ULL,
    0xfc00000000000ULL,
    0xe600000000000ULL,
    0xe900000000000ULL,
    0x7380000000000ULL,
    0xe640000000000ULL,
    0xda60000000000ULL,
    0xe330000000000ULL,
    0x4eb8000000000ULL,
    0x4724000000000ULL,
    0x29d2000000000ULL,
    0x3f4d000000000ULL,
    0x833a800000000ULL,
    0x8c63400000000ULL,
    0x9230e00000000ULL,
    0x193cb00000000ULL,
    0xc360880000000ULL,
    0xa8b33c0000000ULL,
    0x5cfae20000000ULL,
    0x57442b0000000ULL,
    0x6de48d8000000ULL,
    0x17743bc000000ULL,
    0xb7da682000000ULL,
    0x70968c1000000ULL,
    0xe7aaea2800000ULL,
    0xd1cd573400000ULL,
    0x9cfe0faa00000ULL,
    0xb746e0ff00000ULL,
    0x1de60d9980000ULL,
    0x6f737bea40000ULL,
    0xe3d88814e0000ULL,
    0x6a973c1d90000ULL,
    0xe8a8e23498000ULL,
    0x4b492b33cc000ULL,
    0x93e0d802e000ULL,
    0x8b677bd489000ULL,
    0x24b2880f14800ULL,
    0xc2fe3c1c03400ULL,
    0xea406222b6a00ULL,
    0x4676b1644d00ULL,
    0xfe346dabba380ULL,
    0xf7388bd04a2c0ULL,
    0x662e03082a20ULL,
    0x9531b01e768f0ULL,
    0xafba080964988ULL,
    0x18a07c348a9acULL,
    0x81920203a3bb6ULL,
    0xf02c9b2746f2fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x6c00000000000ULL,
    0x1e00000000000ULL,
    0x5b00000000000ULL,
    0xfc80000000000ULL,
    0x93c0000000000ULL,
    0xb560000000000ULL,
    0xf290000000000ULL,
    0x5da8000000000ULL,
    0x224000000000ULL,
    0x5832000000000ULL,
    0x145d000000000ULL,
    0x5e19800000000ULL,
    0x8f3ec00000000ULL,
    0xaf88600000000ULL,
    0xe216700000000ULL,
    0x9068980000000ULL,
    0x4c40c40000000ULL,
    0xe6204e0000000ULL,
    0xc231230000000ULL,
    0x855cb88000000ULL,
    0x3b9f7f4000000ULL,
    0xbd79782000000ULL,
    0x42a8741000000ULL,
    0xd4a0b62800000ULL,
    0x42f7972c00000ULL,
    0xa8bc6e8600000ULL,
    0x80ca985b00000ULL,
    0x64b78ea780000ULL,
    0x679b2846c0000ULL,
    0xfb7cf6b720000ULL,
    0xd5ae5c58f0000ULL,
    0x8625c0a358000ULL,
    0x92370b5ba4000ULL,
    0x8d59ce1eea000ULL,
    0xff9fe33149000ULL,
    0xd77cd880ac800ULL,
    0x6bad0f4027400ULL,
    0x6d23e0247e600ULL,
    0x76b5b0265bb00ULL,
    0x7299780170180ULL,
    0x54f8741a989c0ULL,
    0x7ee8b63104a60ULL,
    0xca839715da710ULL,
    0xfcc66e85b0f38ULL,
    0x9be3987cb9b8cULL,
    0x94540e8375d62ULL,
    0x574ce864c3b3dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xa800000000000ULL,
    0x2400000000000ULL,
    0xda00000000000ULL,
    0x9900000000000ULL,
    0x9e80000000000ULL,
    0x8540000000000ULL,
    0x4a20000000000ULL,
    0xf7f0000000000ULL,
    0x5768000000000ULL,
    0x45e4000000000ULL,
    0xa92000000000ULL,
    0x37bb000000000ULL,
    0x11f9800000000ULL,
    0xac5b400000000ULL,
    0x886c200000000ULL,
    0x8466500000000ULL,
    0xe55080000000ULL,
    0x195e140000000ULL,
    0x78ef1a0000000ULL,
    0x3a22290000000ULL,
    0x7ff1738000000ULL,
    0xa36efd4000000ULL,
    0x87e0a82000000ULL,
    0x1f93041000000ULL,
    0x573e321800000ULL,
    0xd7be6d2c00000ULL,
    0xe1f961aa00000ULL,
    0x6459c04900000ULL,
    0x1c6f419680000ULL,
    0xf660907640000ULL,
    0xb351c99fa0000ULL,
    0x5ddec45d50000ULL,
    0x64ab738088000ULL,
    0xee81fd48fc000ULL,
    0xd4328315a000ULL,
    0xbe27440239000ULL,
    0x35f1922184800ULL,
    0x426c7d2ffec00ULL,
    0x2563c9add6600ULL,
    0xccd5c458fad00ULL,
    0x7a1af39469080ULL,
    0xfdcebd5a8c940ULL,
    0x34550819c3c20ULL,
    0x705e143a6c450ULL,
    0x2e6f1a2db3268ULL,
    0x2b622921f59a4ULL,
    0x47d173b267b4eULL,
    0xe99efd51033b5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0xe400000000000ULL,
    0xf200000000000ULL,
    0xf500000000000ULL,
    0x6d80000000000ULL,
    0xd8c0000000000ULL,
    0x78e0000000000ULL,
    0x3170000000000ULL,
    0xddb8000000000ULL,
    0xf1a4000000000ULL,
    0x6f56000000000ULL,
    0xc72b000000000ULL,
    0xefea800000000ULL,
    0xaf08400000000ULL,
    0x827e200000000ULL,
    0xd1c0100000000ULL,
    0x8b66680000000ULL,
    0x6ab38c0000000ULL,
    0x1fdab60000000ULL,
    0xc8929f0000000ULL,
    0x9b4bff8000000ULL,
    0x55dc7fc000000ULL,
    0xe192c82000000ULL,
    0x18cbdc3000000ULL,
    0xf61afe1800000ULL,
    0x5ff5033400000ULL,
    0x7ff9219200000ULL,
    0xe1026cf900000ULL,
    0xa787019c80000ULL,
    0x31c27ccd40000ULL,
    0x9b6169a360000ULL,
    0x42b1f0f230000ULL,
    0x2bdbdfb438000ULL,
    0x72936ff15c000ULL,
    0x8a482021ee000ULL,
    0xca5b103c29000ULL,
    0xcc54e83635800ULL,
    0xdafcc37bac00ULL,
    0x1faa961022a00ULL,
    0xfaad8f15ee100ULL,
    0xbf2917ae29880ULL,
    0x53e8b3cf35040ULL,
    0x910ade2b3a1a0ULL,
    0xb17a131d63230ULL,
    0xf343c9968f8d8ULL,
    0x2f22a0ec1977cULL,
    0x551117a30c57eULL,
    0x7b8cb3ce66cbfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0x5400000000000ULL,
    0xc600000000000ULL,
    0x2300000000000ULL,
    0xbb80000000000ULL,
    0xefc0000000000ULL,
    0x2920000000000ULL,
    0xff70000000000ULL,
    0xe018000000000ULL,
    0xcb64000000000ULL,
    0x292000000000ULL,
    0x81c9000000000ULL,
    0xd1c8800000000ULL,
    0x59c9400000000ULL,
    0xbdcda00000000ULL,
    0x3cdb00000000ULL,
    0x9ccdd80000000ULL,
    0xb54bbc0000000ULL,
    0xbf8f2e0000000ULL,
    0xe28010000000ULL,
    0xa51b6f8000000ULL,
    0x83e4eac000000ULL,
    0x9ed0f82000000ULL,
    0x832b4c1000000ULL,
    0x299d560800000ULL,
    0x31a70d2400000ULL,
    0xeb3119ba00000ULL,
    0x6fe17d500000ULL,
    0x8d31199180000ULL,
    0xf5fe17d8c0000ULL,
    0xfeb11986e0000ULL,
    0xde3e17cff0000ULL,
    0xf991199848000ULL,
    0x564e17dedc000ULL,
    0x6409199b86000ULL,
    0x51ea17db19000ULL,
    0xf43b198dc4800ULL,
    0xc09317f242400ULL,
    0xeccb99b35a200ULL,
    0xad4e57fb9e500ULL,
    0xb38c39905d680ULL,
    0xe42ee7dfc66c0ULL,
    0xfc1b6198b9f60ULL,
    0x89651bd215af0ULL,
    0x2f916fa41b6b8ULL,
    0xdd49eaeabf104ULL,
    0xab8a780dcd13eULL,
    0xe82b0c093f46bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x5800000000000ULL,
    0x4400000000000ULL,
    0x2200000000000ULL,
    0x6100000000000ULL,
    0x2c80000000000ULL,
    0x21c0000000000ULL,
    0x4ee0000000000ULL,
    0xee10000000000ULL,
    0x65a8000000000ULL,
    0xd9a4000000000ULL,
    0x5432000000000ULL,
    0x23d9000000000ULL,
    0x9ed8800000000ULL,
    0xb45cc00000000ULL,
    0x669d200000000ULL,
    0xadfa900000000ULL,
    0x52aa480000000ULL,
    0x3e22040000000ULL,
    0x4777e0000000ULL,
    0x967d0f0000000ULL,
    0x4ec6c8000000ULL,
    0x7887bf4000000ULL,
    0x1bc7e82000000ULL,
    0x4be0541000000ULL,
    0x7092160800000ULL,
    0x7d6c9b2400000ULL,
    0xffc1da9600000ULL,
    0xb9e0745100000ULL,
    0x6993daa880000ULL,
    0x85e9744840000ULL,
    0xa4035aa320000ULL,
    0xd201b47c70000ULL,
    0x89047aadb8000ULL,
    0xa086245e84000ULL,
    0x1fc4b2afea000ULL,
    0x49e1e04b29000ULL,
    0x81966ca8ac800ULL,
    0x9eabf7dc6400ULL,
    0x9a056809ae200ULL,
    0xd5019423a3300ULL,
    0xe685b60ed5480ULL,
    0xccc3cb03a3a40ULL,
    0x5c66329ed4120ULL,
    0xc8d0204ba3c10ULL,
    0x44c9cc9ad57f8ULL,
    0x5bf1ef4da333cULL,
    0x72388003d4f32ULL,
    0x7b4cc021233bdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0xcc00000000000ULL,
    0xb600000000000ULL,
    0xf900000000000ULL,
    0xe580000000000ULL,
    0xc840000000000ULL,
    0x5ea0000000000ULL,
    0x9310000000000ULL,
    0x3f88000000000ULL,
    0x3d64000000000ULL,
    0x28f2000000000ULL,
    0xe5b9000000000ULL,
    0x2abb800000000ULL,
    0x763b400000000ULL,
    0x7bf9600000000ULL,
    0x1d1d300000000ULL,
    0x78aa080000000ULL,
    0x18348c0000000ULL,
    0xacdc520000000ULL,
    0x40492d0000000ULL,
    0x68836c8000000ULL,
    0x6fc2ca4000000ULL,
    0x6de0e82000000ULL,
    0xf336fc1000000ULL,
    0xaa5d3a0800000ULL,
    0xf289913c00000ULL,
    0xbae6b6aa00000ULL,
    0xebb02b7300000ULL,
    0x2d9cb68d80000ULL,
    0x736d2b6e40000ULL,
    0xfd5369160000ULL,
    0x10ef6b5e10000ULL,
    0xf697d695a8000ULL,
    0xb7c91b7bc4000ULL,
    0x43c4bea062000ULL,
    0x4be0a76e19000ULL,
    0x7232e49cdc800ULL,
    0xdbdd066a3e400ULL,
    0xe8cdda0be6e00ULL,
    0x6746e12a1ad00ULL,
    0xd2265eb054580ULL,
    0xc6d6d76f8a4c0ULL,
    0x3d698c92bc020ULL,
    0x22d0ba7bae630ULL,
    0x769800f0f748ULL,
    0x4dd2403f3fdb4ULL,
    0xebeae00766e32ULL,
    0xc812702eda6e9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xb800000000000ULL,
    0x7400000000000ULL,
    0x9a00000000000ULL,
    0x3500000000000ULL,
    0xc780000000000ULL,
    0xbfc0000000000ULL,
    0xe320000000000ULL,
    0x87b0000000000ULL,
    0x32e8000000000ULL,
    0xa564000000000ULL,
    0xac56000000000ULL,
    0xe0fd000000000ULL,
    0xda78800000000ULL,
    0x9e3fc00000000ULL,
    0xa4dca00000000ULL,
    0x334f500000000ULL,
    0x2297e80000000ULL,
    0x165fdc0000000ULL,
    0x2b0ace0000000ULL,
    0x4e72810000000ULL,
    0x5aca198000000ULL,
    0x52510d4000000ULL,
    0x37fbc82000000ULL,
    0xfcfb4c3000000ULL,
    0xec7f862800000ULL,
    0x753b0d2400000ULL,
    0xe459bfae00000ULL,
    0xb20e905d00000ULL,
    0x57f73f8680000ULL,
    0x620c507d40000ULL,
    0x8ff31fb9e0000ULL,
    0x260cc07bf0000ULL,
    0x3df0579ec8000ULL,
    0xdf084c68ec000ULL,
    0x1473718c3a000ULL,
    0xcfcc116d19000ULL,
    0x5d5261a75800ULL,
    0x30395d138f400ULL,
    0x6bded79f36200ULL,
    0xf1ca8c5dd3f00ULL,
    0x72d7518425280ULL,
    0x86bc814bd6d40ULL,
    0xa19e6e08a27a0ULL,
    0x5429d103adb70ULL,
    0x3fc3f194db138ULL,
    0x2327d15269104ULL,
    0x27b7863efc0e6ULL,
    0xa2ef0d2f02275ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0x1400000000000ULL,
    0x5600000000000ULL,
    0x7900000000000ULL,
    0x880000000000ULL,
    0xd6c0000000000ULL,
    0xb8e0000000000ULL,
    0xec90000000000ULL,
    0xcf88000000000ULL,
    0xca4000000000ULL,
    0x99b2000000000ULL,
    0x8c7b000000000ULL,
    0x30fe800000000ULL,
    0xc039400000000ULL,
    0xd9dda00000000ULL,
    0x91cd300000000ULL,
    0x2283080000000ULL,
    0xadc4240000000ULL,
    0xff62660000000ULL,
    0x1bd1df0000000ULL,
    0x4129b88000000ULL,
    0x9a15564000000ULL,
    0x5448282000000ULL,
    0xdac4541000000ULL,
    0x62e6ce1800000ULL,
    0xf97cb1400000ULL,
    0x440c56ba00000ULL,
    0x1e6ed4500000ULL,
    0xc412d6b580000ULL,
    0x294fad4e40000ULL,
    0xc4776ba20000ULL,
    0x89269d71b0000ULL,
    0x1f67eac38000ULL,
    0x1359b97a24000ULL,
    0x280a98a462000ULL,
    0x8be1265c69000ULL,
    0xcf168033cc800ULL,
    0xb6cd4009eec00ULL,
    0x5707a017a7a00ULL,
    0xad823031da500ULL,
    0xac4788060d680ULL,
    0xf92264328e4c0ULL,
    0xd9f34624b6420ULL,
    0xbf5eaf1b92c90ULL,
    0x8209908a71b98ULL,
    0xb0e5025859e7cULL,
    0xe894e631b4262ULL,
    0x118c9f39eb899ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0xf400000000000ULL,
    0x5a00000000000ULL,
    0x6d00000000000ULL,
    0xb980000000000ULL,
    0xd040000000000ULL,
    0xb60000000000ULL,
    0x990000000000ULL,
    0x8e48000000000ULL,
    0x4a4000000000ULL,
    0x69b2000000000ULL,
    0x963b000000000ULL,
    0xadbf800000000ULL,
    0xd4ffc00000000ULL,
    0x641aa00000000ULL,
    0x54cc500000000ULL,
    0xa64280000000ULL,
    0x4e112c0000000ULL,
    0x410ab20000000ULL,
    0x6740030000000ULL,
    0x9fe0588000000ULL,
    0x52244000000ULL,
    0x2569082000000ULL,
    0x3e56bc1000000ULL,
    0x1a6e3a1800000ULL,
    0x46d27f1c00000ULL,
    0xadab428e00000ULL,
    0x92f3cb7d00000ULL,
    0x8fdcc2b680000ULL,
    0x26e80b4b40000ULL,
    0xb314629660000ULL,
    0xf08f5b7810000ULL,
    0x1b07cab4d8000ULL,
    0x4a85b77364000ULL,
    0x6c5d88312000ULL,
    0x8ba2e46b69000ULL,
    0x93628148c800ULL,
    0xc5fa2c3bdec00ULL,
    0x6b9d320757e00ULL,
    0x578bc3390bf00ULL,
    0xc780f88eaca80ULL,
    0x8f4174442e140ULL,
    0x3e0a0215f8a0ULL,
    0x8653500a178b0ULL,
    0xae69a8366bcc8ULL,
    0x7cd5ec223c30cULL,
    0xb0af9220b7de2ULL,
    0x1373933bdbad1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0xdc00000000000ULL,
    0xa200000000000ULL,
    0x7b00000000000ULL,
    0xc280000000000ULL,
    0x22c0000000000ULL,
    0xf1a0000000000ULL,
    0x37b0000000000ULL,
    0x37e8000000000ULL,
    0xaae4000000000ULL,
    0x652000000000ULL,
    0x33b000000000ULL,
    0xd319800000000ULL,
    0x48efc00000000ULL,
    0xe663e00000000ULL,
    0xa591900000000ULL,
    0x271c980000000ULL,
    0x9ee8b40000000ULL,
    0xb65260000000ULL,
    0xea16a50000000ULL,
    0xfa5fe18000000ULL,
    0x8a0d254000000ULL,
    0xf434f82000000ULL,
    0x80ade41000000ULL,
    0x4103de1800000ULL,
    0xf580412c00000ULL,
    0xaa45bf8e00000ULL,
    0xfc62a47700000ULL,
    0x4292a78880000ULL,
    0x279ed04ec0000ULL,
    0x772de188a0000ULL,
    0x462574b0000ULL,
    0x2365780a68000ULL,
    0x9e162406ec000ULL,
    0x6c5a3e337a000ULL,
    0x70ed12f79000ULL,
    0xbb2a797b4800ULL,
    0x65eed061bec00ULL,
    0x89e5e1b08e600ULL,
    0xa8d22559a7f00ULL,
    0xbbff7800aaf80ULL,
    0x85b9243a81640ULL,
    0x1fd9be2fc1a60ULL,
    0x884e111b21ed0ULL,
    0x3c52c79cf1a98ULL,
    0x343f806b09994ULL,
    0x5b99198a7de06ULL,
    0x452bc15f621d5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0x4400000000000ULL,
    0xba00000000000ULL,
    0xa500000000000ULL,
    0xfb80000000000ULL,
    0xb340000000000ULL,
    0xc220000000000ULL,
    0x5c30000000000ULL,
    0x32d8000000000ULL,
    0x67e4000000000ULL,
    0x52d6000000000ULL,
    0x458f000000000ULL,
    0x9e2b800000000ULL,
    0x685c400000000ULL,
    0xb527200000000ULL,
    0x96b3500000000ULL,
    0x181be80000000ULL,
    0xcf803c0000000ULL,
    0x1422a0000000ULL,
    0xc323c90000000ULL,
    0xedb2458000000ULL,
    0xec98b8c000000ULL,
    0x1a41482000000ULL,
    0x7a02c3000000ULL,
    0x1075623800000ULL,
    0x5e7ce53c00000ULL,
    0xbb14a7b200000ULL,
    0x556c1dd100000ULL,
    0x67ccf8e80000ULL,
    0x71461d940000ULL,
    0x836fc5a6e0000ULL,
    0xad7bf8e0d0000ULL,
    0x2b95e81a88000ULL,
    0x2aab3c0a0c000ULL,
    0x9d9faa2836000ULL,
    0xa3c0891db9000ULL,
    0x7de6e5b6d5800ULL,
    0x27d3a8c5f3c00ULL,
    0x660b802762e00ULL,
    0x516c40388b100ULL,
    0x9c7f201797c80ULL,
    0xb217503b89d40ULL,
    0x40ede81c8d7a0ULL,
    0x923f3c3d76cf0ULL,
    0x17b1aa1159ea8ULL,
    0x699b890b05224ULL,
    0xf1c3659bfa396ULL,
    0x8ce4e8f92e2a3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x6800000000000ULL,
    0xfc00000000000ULL,
    0x1a00000000000ULL,
    0xe100000000000ULL,
    0xb280000000000ULL,
    0x8ec0000000000ULL,
    0x420000000000ULL,
    0xfe90000000000ULL,
    0x7698000000000ULL,
    0xa564000000000ULL,
    0x9cf2000000000ULL,
    0xc3e9000000000ULL,
    0x4bca800000000ULL,
    0x575d400000000ULL,
    0x74c0e00000000ULL,
    0x5525500000000ULL,
    0xd413c80000000ULL,
    0xbc5fb40000000ULL,
    0x2f44060000000ULL,
    0x6562190000000ULL,
    0xfcf4b88000000ULL,
    0x33ebdec000000ULL,
    0xf3cba82000000ULL,
    0xc35ea41000000ULL,
    0x92c5ae0800000ULL,
    0xae21bd3400000ULL,
    0x8791969a00000ULL,
    0x801a23ff00000ULL,
    0xa5a25ea680000ULL,
    0x9fd597e840000ULL,
    0x74fe5884a0000ULL,
    0xe0138ec7b0000ULL,
    0xca58e03308000ULL,
    0x9c415024a4000ULL,
    0x1ae1c80926000ULL,
    0xb2b6b41a19000ULL,
    0xa40e86171c800ULL,
    0xa1ff59140a400ULL,
    0x249458b9daa00ULL,
    0xf79e8ef183500ULL,
    0xe7e0600d3e380ULL,
    0xaa35103004540ULL,
    0x53cba80696720ULL,
    0x535ea43d10ad0ULL,
    0xdac5ae39b9218ULL,
    0x8221bd2a2c964ULL,
    0xf591969143362ULL,
    0x9d1a23c4deb3bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x800000000000ULL,
    0x7c00000000000ULL,
    0xd600000000000ULL,
    0x9100000000000ULL,
    0xcb80000000000ULL,
    0xf040000000000ULL,
    0x3ea0000000000ULL,
    0xbab0000000000ULL,
    0x338000000000ULL,
    0xd064000000000ULL,
    0x5dd2000000000ULL,
    0x2c6b000000000ULL,
    0x6bc9800000000ULL,
    0x96fe400000000ULL,
    0xdb86e00000000ULL,
    0x6842300000000ULL,
    0x9aa2480000000ULL,
    0xa8b07c0000000ULL,
    0x303a920000000ULL,
    0x20e42d0000000ULL,
    0x2113b18000000ULL,
    0xb80e4fc000000ULL,
    0xd41f282000000ULL,
    0xe1970c1000000ULL,
    0x8ccfba1800000ULL,
    0x5c7c212c00000ULL,
    0x38478b8200000ULL,
    0x62a72edf00000ULL,
    0xbcb7c39580000ULL,
    0x2a3c52f440000ULL,
    0x6fe4d18ae0000ULL,
    0x7963fc010000ULL,
    0x55c98035a8000ULL,
    0x8bfe400dac000ULL,
    0xae06e02f4e000ULL,
    0xc502302359000ULL,
    0xb182481214800ULL,
    0xf407c0c4ac00ULL,
    0x102292123a600ULL,
    0x2b702d1343900ULL,
    0x8ad9b18b67b80ULL,
    0xef14ffcb58c0ULL,
    0xdf9ca81eba120ULL,
    0x31d64c2b83df0ULL,
    0x626ada1946448ULL,
    0x5ecf513e855b4ULL,
    0x4f78a3b123b46ULL,
    0x18c022c5b64ffULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0xfc00000000000ULL,
    0x6200000000000ULL,
    0xc100000000000ULL,
    0x5680000000000ULL,
    0x4640000000000ULL,
    0x320000000000ULL,
    0x9a10000000000ULL,
    0xb938000000000ULL,
    0x90e4000000000ULL,
    0xf8f2000000000ULL,
    0x78cf000000000ULL,
    0xa2ac800000000ULL,
    0x6e1bc00000000ULL,
    0x6cf6e00000000ULL,
    0xfece900000000ULL,
    0xda8480000000ULL,
    0x339e140000000ULL,
    0xf932020000000ULL,
    0x732f610000000ULL,
    0x71dd3c8000000ULL,
    0x8417e7c000000ULL,
    0x5a38282000000ULL,
    0xe760441000000ULL,
    0xf8322a3800000ULL,
    0x85ab250400000ULL,
    0xc79d16ae00000ULL,
    0x2f33c2ff00000ULL,
    0x8429beb880000ULL,
    0xc05846e040000ULL,
    0xcbd5748da0000ULL,
    0xead9f3c590000ULL,
    0xaf922a36c8000ULL,
    0xc8fb250d84000ULL,
    0x730516a0ce000ULL,
    0x8f87c2ef79000ULL,
    0x1cc3bebd1c800ULL,
    0x8f6346d0e3c00ULL,
    0x1c33f4bd43200ULL,
    0xeba933c8d2f00ULL,
    0xcc9a4a161bb80ULL,
    0xfcb175079ea40ULL,
    0xaef3e9d50920ULL,
    0x263886db29450ULL,
    0xc5651485b5208ULL,
    0xd937a3f397e84ULL,
    0xc32a020435572ULL,
    0x39db6108576dfULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x400000000000ULL,
    0xaa00000000000ULL,
    0x4d00000000000ULL,
    0x8880000000000ULL,
    0xad40000000000ULL,
    0x2ba0000000000ULL,
    0x4d70000000000ULL,
    0x8cb8000000000ULL,
    0x3964000000000ULL,
    0xa612000000000ULL,
    0x3baf000000000ULL,
    0xcc4c800000000ULL,
    0x341e400000000ULL,
    0x4391200000000ULL,
    0xae68d00000000ULL,
    0xb2a9a80000000ULL,
    0x4ecaa40000000ULL,
    0x24595a0000000ULL,
    0xf8b7c50000000ULL,
    0x425cd78000000ULL,
    0xbbb4c3c000000ULL,
    0xddde082000000ULL,
    0xcb77341000000ULL,
    0xdfbf523800000ULL,
    0xaee4f10400000ULL,
    0xd2d1858200000ULL,
    0x93cf32c100000ULL,
    0xe4db0d8a80000ULL,
    0xb1f246c340000ULL,
    0xd3ff7fba20000ULL,
    0xb3c567ff50000ULL,
    0xdde1d210e8000ULL,
    0xf555b1065c000ULL,
    0x1f0c25b3ae000ULL,
    0x50b9a2d819000ULL,
    0x3f63858324800ULL,
    0xb51032e7fbc00ULL,
    0x4c2f8da95b200ULL,
    0xa88806f24b900ULL,
    0x947c5f8102480ULL,
    0x6f02b7c8cf340ULL,
    0x8184fa1c68ea0ULL,
    0x6fc155061c690ULL,
    0xdbe45f820ff68ULL,
    0xe656b7ce88a14ULL,
    0x688efa3aacddeULL,
    0x347a553cd754fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0xf400000000000ULL,
    0x1200000000000ULL,
    0x7900000000000ULL,
    0xc80000000000ULL,
    0x5940000000000ULL,
    0xb460000000000ULL,
    0x1950000000000ULL,
    0xf498000000000ULL,
    0x62a4000000000ULL,
    0x1df2000000000ULL,
    0xc26d000000000ULL,
    0x7549800000000ULL,
    0x8d7fc00000000ULL,
    0x9a34a00000000ULL,
    0xb64a100000000ULL,
    0xc4fe880000000ULL,
    0xb1f5840000000ULL,
    0x746b3a0000000ULL,
    0x364ddb0000000ULL,
    0x84f9b68000000ULL,
    0x11f3a2c000000ULL,
    0xc469a82000000ULL,
    0xe49541000000ULL,
    0x70fa922800000ULL,
    0x3f08f2c00000ULL,
    0xbd69248e00000ULL,
    0x2ca2dfd00000ULL,
    0x29bb0ca480000ULL,
    0xb791b9ce40000ULL,
    0xa43cbe8b20000ULL,
    0xf654e6ea50000ULL,
    0x4b1fb22b18000ULL,
    0xaa615f0754000ULL,
    0x66510c9fa6000ULL,
    0x8318b9d7e9000ULL,
    0xc6673eaedc800ULL,
    0x305626d58b400ULL,
    0xd01a923fea600ULL,
    0x47e08f111bf00ULL,
    0x7791249393280ULL,
    0x443e2df43f840ULL,
    0xe6510c8cfd220ULL,
    0xc318b9f15b210ULL,
    0x66673eb5b22e8ULL,
    0x805626f2eec6cULL,
    0xe81a9235a5a5aULL,
    0xb3e08f2b2edcbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0xd400000000000ULL,
    0x1600000000000ULL,
    0xcf00000000000ULL,
    0x4880000000000ULL,
    0x34c0000000000ULL,
    0xd4e0000000000ULL,
    0x4390000000000ULL,
    0x4228000000000ULL,
    0x55a4000000000ULL,
    0x1a32000000000ULL,
    0xda1f000000000ULL,
    0xb6ba800000000ULL,
    0xfb0cc00000000ULL,
    0x557a00000000ULL,
    0x87cb300000000ULL,
    0xfdb1380000000ULL,
    0x365aa40000000ULL,
    0xee9e760000000ULL,
    0x687e4f0000000ULL,
    0x4aed508000000ULL,
    0x8b46ab4000000ULL,
    0xeca6182000000ULL,
    0xdb6541000000ULL,
    0x1e586e3800000ULL,
    0x52981b3400000ULL,
    0x227d3e9600000ULL,
    0x1feab07500000ULL,
    0xcec126a580000ULL,
    0x49e7e463c0000ULL,
    0x6a11c88a20000ULL,
    0xf56c3f6930000ULL,
    0xb301d63b38000ULL,
    0xa2817f01e4000ULL,
    0x51c668bb0a000ULL,
    0x19670f67a9000ULL,
    0x6ad0ee272c800ULL,
    0xe78bdb2d77c00ULL,
    0x67901eb736a00ULL,
    0x7c2d4064d7300ULL,
    0x26a7be80c7e80ULL,
    0x18b6704c4fcc0ULL,
    0xbbde8681d2ce0ULL,
    0x27d8d4695c690ULL,
    0x1ddaf09c2fbd8ULL,
    0x20dd9b5b0263cULL,
    0x395f2006749c2ULL,
    0x618f0394396dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x5c00000000000ULL,
    0x2600000000000ULL,
    0x900000000000ULL,
    0x9480000000000ULL,
    0x45c0000000000ULL,
    0xb920000000000ULL,
    0x8e10000000000ULL,
    0xc798000000000ULL,
    0x4364000000000ULL,
    0xd572000000000ULL,
    0x6eeb000000000ULL,
    0xdb89800000000ULL,
    0x6ff8c00000000ULL,
    0x1197e00000000ULL,
    0xf5ddb00000000ULL,
    0xc04080000000ULL,
    0xe07040000000ULL,
    0xed043e0000000ULL,
    0x5e83d10000000ULL,
    0xb2c76c8000000ULL,
    0x78a21ec000000ULL,
    0x7950682000000ULL,
    0x327d741000000ULL,
    0xe0d4561800000ULL,
    0xd93ea52c00000ULL,
    0x53333ab600000ULL,
    0xc38cbbd700000ULL,
    0x13fb52a980000ULL,
    0xe795cfd240000ULL,
    0x94dd049d20000ULL,
    0x1c806aed70000ULL,
    0x31c7be2048000ULL,
    0x7b24110884000ULL,
    0x4d130c9666000ULL,
    0xa4186ee999000ULL,
    0xc720000ffc800ULL,
    0x9b10000f8ac00ULL,
    0x1518002c0a600ULL,
    0xbfa400074a300ULL,
    0x20d20008ebf80ULL,
    0xf93b00085a6c0ULL,
    0x8331802073820ULL,
    0xab8cc02926010ULL,
    0x97fde02c39ef8ULL,
    0x9d92b0164d044ULL,
    0xbbdf880522c32ULL,
    0x8104c4009f1bbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x7800000000000ULL,
    0xb400000000000ULL,
    0x600000000000ULL,
    0x2900000000000ULL,
    0x9880000000000ULL,
    0x15c0000000000ULL,
    0xe2e0000000000ULL,
    0xd850000000000ULL,
    0xa728000000000ULL,
    0xf854000000000ULL,
    0x572e000000000ULL,
    0xb057000000000ULL,
    0xcb2c800000000ULL,
    0x7a54c00000000ULL,
    0x5029200000000ULL,
    0xcdd4f00000000ULL,
    0xf468680000000ULL,
    0xa270340000000ULL,
    0xdb18c60000000ULL,
    0x3fef230000000ULL,
    0x5c32048000000ULL,
    0x37b81a4000000ULL,
    0x4318c62000000ULL,
    0x1bef233000000ULL,
    0x4232049800000ULL,
    0xfab81a5400000ULL,
    0xa598c63e00000ULL,
    0x932f231d00000ULL,
    0x3e52049980000ULL,
    0x1e281a5e40000ULL,
    0x78d0c61820000ULL,
    0xa6eb231870000ULL,
    0x2cb404a138000ULL,
    0x8e7b1a6854000ULL,
    0x43fa4631ea000ULL,
    0x94bce32665000ULL,
    0xe09fa4b4f3800ULL,
    0x89ac2a4441c00ULL,
    0x2c978e0321200ULL,
    0x814ce738f0300ULL,
    0x9fc62aa0f9c80ULL,
    0xd9e7cd7734fc0ULL,
    0x5fd524be3b3a0ULL,
    0x2b6bea506c3d0ULL,
    0x5bf42e363ff98ULL,
    0xca5fd708cf34cULL,
    0x2cde29137bb2ULL,
    0xfd84c979823b9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0x9400000000000ULL,
    0x9e00000000000ULL,
    0x7d00000000000ULL,
    0x8180000000000ULL,
    0x8c40000000000ULL,
    0x87e0000000000ULL,
    0xffb0000000000ULL,
    0x18c8000000000ULL,
    0xe7b4000000000ULL,
    0x4ce000000000ULL,
    0x8db7000000000ULL,
    0x37c8800000000ULL,
    0x3935400000000ULL,
    0xae0c200000000ULL,
    0xac91b00000000ULL,
    0xab5fc80000000ULL,
    0xca6dbc0000000ULL,
    0xd867aa0000000ULL,
    0xaef3830000000ULL,
    0xaea9578000000ULL,
    0xc44c54000000ULL,
    0x47e7aa2000000ULL,
    0x9fb3833000000ULL,
    0xc8c9579800000ULL,
    0xafb4c57400000ULL,
    0x90cfaa3200000ULL,
    0x13b7831500000ULL,
    0x4acf57bf80000ULL,
    0xb8b7c56b40000ULL,
    0x22492a1260000ULL,
    0x2b75c33610000ULL,
    0x54ebf79e78000ULL,
    0xd2a33554ac000ULL,
    0x3fd2c21452000ULL,
    0xaa3dcf0ffd000ULL,
    0x231d959f4b800ULL,
    0x3b8a0a77c1c00ULL,
    0x7ed4bf99a0200ULL,
    0x31bfc941b0500ULL,
    0x645f4814c8880ULL,
    0x4ecfc2ce5ac0ULL,
    0x5aa58a0b77d20ULL,
    0xcbd533372b3f0ULL,
    0xe43e1f88d1628ULL,
    0x161c3966594ccULL,
    0x2e0c202d5d87eULL,
    0x6c91b00a7a0e5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0xf800000000000ULL,
    0xdc00000000000ULL,
    0xe00000000000ULL,
    0x9100000000000ULL,
    0xad80000000000ULL,
    0xe840000000000ULL,
    0x49a0000000000ULL,
    0xadf0000000000ULL,
    0xb888000000000ULL,
    0xc5f4000000000ULL,
    0x8c8e000000000ULL,
    0x3ff1000000000ULL,
    0x478d800000000ULL,
    0x2976c00000000ULL,
    0x414b600000000ULL,
    0xba16b00000000ULL,
    0x71dfe80000000ULL,
    0xa30bdc0000000ULL,
    0xb4aa0000000ULL,
    0xb6ab910000000ULL,
    0x6342c58000000ULL,
    0x3f20d54000000ULL,
    0x7b34aa2000000ULL,
    0xa3eb913000000ULL,
    0xd162c5a800000ULL,
    0x4790d55c00000ULL,
    0xd19caa1e00000ULL,
    0x6eaf910700000ULL,
    0xf44c5ab80000ULL,
    0x6925d57840000ULL,
    0x46372a3560000ULL,
    0xf86c513d10000ULL,
    0x742425b9e8000ULL,
    0xadb0a5533c000ULL,
    0x528221b42000ULL,
    0x3202fd0c6d000ULL,
    0x8f07879acb800ULL,
    0x1481985cc0400ULL,
    0x11c7458aa1600ULL,
    0x8be2154630b00ULL,
    0xa751ca2a99580ULL,
    0x27fc210245ec0ULL,
    0xa9b8ad9ed69a0ULL,
    0x4519c942f2470ULL,
    0xbf6d603cb4728ULL,
    0x24a3b004ef084ULL,
    0xbe74683b862b6ULL,
    0x99c81c1e7a725ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x1800000000000ULL,
    0x6400000000000ULL,
    0x200000000000ULL,
    0xb500000000000ULL,
    0x680000000000ULL,
    0x3cc0000000000ULL,
    0x51a0000000000ULL,
    0x8bb0000000000ULL,
    0xe8a8000000000ULL,
    0x2fb4000000000ULL,
    0x4aaa000000000ULL,
    0x8ab3000000000ULL,
    0xa42d800000000ULL,
    0xca73c00000000ULL,
    0x938ea00000000ULL,
    0xf6c3d00000000ULL,
    0xc8a3080000000ULL,
    0xe337e40000000ULL,
    0xef68ee0000000ULL,
    0xb396690000000ULL,
    0x2859338000000ULL,
    0xbefeec4000000ULL,
    0xdec8ee2000000ULL,
    0x8826691000000ULL,
    0x38f133b800000ULL,
    0x54aec7c00000ULL,
    0x8e62ee2600000ULL,
    0xd395690900000ULL,
    0x985cb3b880000ULL,
    0x46f92c5140000ULL,
    0x4acc4e27a0000ULL,
    0x9226b90630000ULL,
    0xe9f7bbace8000ULL,
    0x1cac873ac000ULL,
    0x7a6a01d8a000ULL,
    0x84b7d00ddd000ULL,
    0x2f29083e42800ULL,
    0xfff4e41100c00ULL,
    0xeecd6e1481600ULL,
    0x3021a92f41f00ULL,
    0x4cf593a2a1280ULL,
    0xef4a3c50b0340ULL,
    0x47646600a9a20ULL,
    0xb3154d1e8c090ULL,
    0x1398fda1fb138ULL,
    0xa4db956d15ae4ULL,
    0xfb39f58d1feeeULL,
    0x17eb716703b21ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x8800000000000ULL,
    0xd400000000000ULL,
    0x8200000000000ULL,
    0x700000000000ULL,
    0x1080000000000ULL,
    0x91c0000000000ULL,
    0xaa60000000000ULL,
    0x5210000000000ULL,
    0x4878000000000ULL,
    0x8e14000000000ULL,
    0xde7e000000000ULL,
    0xab11000000000ULL,
    0x99fc800000000ULL,
    0xa255400000000ULL,
    0x765a200000000ULL,
    0xd821500000000ULL,
    0x6b32180000000ULL,
    0xecdd40000000ULL,
    0xaa9c960000000ULL,
    0xd1c7870000000ULL,
    0xca63868000000ULL,
    0xa21404c000000ULL,
    0x907c962000000ULL,
    0xd217873000000ULL,
    0x887b86a800000ULL,
    0x2e1004d400000ULL,
    0x8e7a960200000ULL,
    0x2312870500000ULL,
    0x4df9068880000ULL,
    0x205444d5c0000ULL,
    0x715c360620000ULL,
    0xc8a6972170000ULL,
    0xfaf13ea218000ULL,
    0xa4a8c0c144000ULL,
    0xf88ab8143e000ULL,
    0x99b8c402f5000ULL,
    0x44702e1587800ULL,
    0x7c6a432b80400ULL,
    0x3b69289241200ULL,
    0x4bee07da60500ULL,
    0x2a2d1e8011080ULL,
    0x584c90f988140ULL,
    0x565a20228da60ULL,
    0x48215008d3250ULL,
    0x43321820362d8ULL,
    0x8acdd41d39f5cULL,
    0xa09c961eb547aULL,
    0x2c78714e6243ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xd800000000000ULL,
    0xbc00000000000ULL,
    0xc200000000000ULL,
    0xc900000000000ULL,
    0x5d80000000000ULL,
    0x5040000000000ULL,
    0x7e0000000000ULL,
    0x5eb0000000000ULL,
    0xf768000000000ULL,
    0xab4000000000ULL,
    0xf16e000000000ULL,
    0xadb7000000000ULL,
    0xfed800000000ULL,
    0xd71c00000000ULL,
    0x7bc8600000000ULL,
    0xf67d00000000ULL,
    0x4176280000000ULL,
    0xa1c93c0000000ULL,
    0x1a66aa0000000ULL,
    0xaef4030000000ULL,
    0xe088df8000000ULL,
    0xfc04f84000000ULL,
    0x6206aa2000000ULL,
    0xd904033000000ULL,
    0xf580df9800000ULL,
    0x3440f85c00000ULL,
    0x79e0aa1600000ULL,
    0x55b7031f00000ULL,
    0x63eb5fa880000ULL,
    0x772386e40000ULL,
    0xa6cb4a0160000ULL,
    0xf4e6130b10000ULL,
    0xa63097a978000ULL,
    0xf0a91479ec000ULL,
    0x8015a83cba000ULL,
    0x53bffc09bd000ULL,
    0xe32b4a1523800ULL,
    0xa356131281c00ULL,
    0x6cd8979f41600ULL,
    0xda5d147ae1700ULL,
    0xae9ba833d1980ULL,
    0x1cb8fc3d58340ULL,
    0xd9aeca211cea0ULL,
    0x6d93d321933f0ULL,
    0xbbfef793c8328ULL,
    0x28cdc47ba534cULL,
    0xe7e00006deddeULL,
    0xeeb00014c2011ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xb800000000000ULL,
    0x2400000000000ULL,
    0x600000000000ULL,
    0xa500000000000ULL,
    0x1e80000000000ULL,
    0x4c0000000000ULL,
    0xdfa0000000000ULL,
    0xec70000000000ULL,
    0x5ee8000000000ULL,
    0xc074000000000ULL,
    0xd4ea000000000ULL,
    0x9f77000000000ULL,
    0x1d68800000000ULL,
    0x9e32c00000000ULL,
    0x5989a00000000ULL,
    0xb1a2f00000000ULL,
    0x9574880000000ULL,
    0x26e040000000ULL,
    0x37b2ae0000000ULL,
    0x8c9f90000000ULL,
    0x4d444d8000000ULL,
    0x9ee44e4000000ULL,
    0xc892ae2000000ULL,
    0x2179f91000000ULL,
    0xb48c4d9800000ULL,
    0x43204e5400000ULL,
    0x3bb0ae0e00000ULL,
    0xb2caf91900000ULL,
    0xfa46cd9980000ULL,
    0xcb618e7d40000ULL,
    0xebd38e09a0000ULL,
    0xc699c91830000ULL,
    0x209965ae68000ULL,
    0x9598ba465c000ULL,
    0x531c881e1a000ULL,
    0x23da04282d000ULL,
    0x4278ae1b52800ULL,
    0x2f0ef92181c00ULL,
    0x6964cdb940200ULL,
    0x8d28e5fa1b00ULL,
    0x1d190e3530e80ULL,
    0x6ad80909e97c0ULL,
    0xf6fa45aa1d020ULL,
    0x44cb8a423a310ULL,
    0x974320175c438ULL,
    0x79e3302b9b024ULL,
    0x2517a83f6c116ULL,
    0x263d340df3229ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x7800000000000ULL,
    0xac00000000000ULL,
    0xfa00000000000ULL,
    0xbf00000000000ULL,
    0x2180000000000ULL,
    0x4440000000000ULL,
    0xc1e0000000000ULL,
    0x8b30000000000ULL,
    0xa428000000000ULL,
    0xd134000000000ULL,
    0xb2e000000000ULL,
    0xa8b5000000000ULL,
    0x3368800000000ULL,
    0xeb56c00000000ULL,
    0xab5de00000000ULL,
    0x94f9100000000ULL,
    0x812c180000000ULL,
    0x7fb61c0000000ULL,
    0xe6eb3a0000000ULL,
    0x2910930000000ULL,
    0x57bda38000000ULL,
    0x2d4e524000000ULL,
    0xbac33a2000000ULL,
    0x9424933000000ULL,
    0x8693a38800000ULL,
    0xeafb527400000ULL,
    0xd02bba3e00000ULL,
    0x9732531b00000ULL,
    0x162e43b680000ULL,
    0x4a32425bc0000ULL,
    0xd4afa23660000ULL,
    0x7df04f0a10000ULL,
    0x3a0b7986f8000ULL,
    0x40a7d1790c000ULL,
    0x1452819f6a000ULL,
    0x6adcdd7e5d000ULL,
    0x20bba3a433800ULL,
    0xe8cf526321400ULL,
    0x2085ba1bb0200ULL,
    0x14c7533088b00ULL,
    0xbd26c3b0e4f80ULL,
    0xc214825d1f040ULL,
    0xe03a420d7b260ULL,
    0xc0d5f34a5370ULL,
    0x35a1619f3e368ULL,
    0x2d0cd6d4b20cULL,
    0xb1f3b966c4eeULL,
    0x8b1f8e72faa39ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xc800000000000ULL,
    0xd400000000000ULL,
    0xd600000000000ULL,
    0x2100000000000ULL,
    0x9580000000000ULL,
    0x3e40000000000ULL,
    0xb860000000000ULL,
    0x8b90000000000ULL,
    0x1c58000000000ULL,
    0x6994000000000ULL,
    0x5b5a000000000ULL,
    0xe515000000000ULL,
    0x3c9c800000000ULL,
    0xa933400000000ULL,
    0x2c6a200000000ULL,
    0x5c7df00000000ULL,
    0xd367280000000ULL,
    0x7d175c0000000ULL,
    0x6099da0000000ULL,
    0x8b331f0000000ULL,
    0xb6e468000000ULL,
    0x20fc9dc000000ULL,
    0xaca1da2000000ULL,
    0x2d371f1000000ULL,
    0x726c468800000ULL,
    0x897d9df400000ULL,
    0x40e75a1200000ULL,
    0xaa515f2500000ULL,
    0x997ae6bd80000ULL,
    0xe8e32dfc40000ULL,
    0x8e52523760000ULL,
    0x577ff32a90000ULL,
    0xd5e6149398000ULL,
    0x19d66edea4000ULL,
    0x9e3b4eb076000ULL,
    0xd90231f0f5000ULL,
    0x398528254e800ULL,
    0x4465c37e1400ULL,
    0x83675a1751200ULL,
    0xf5115f3eb9d00ULL,
    0x949ae6b454080ULL,
    0x8d332df1fe3c0ULL,
    0xe26a520c08ea0ULL,
    0x617bf338bc070ULL,
    0x44e41486927e8ULL,
    0xb4576ed5e2fbcULL,
    0x6c7dce822b7baULL,
    0xab6471e0b7207ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0x9400000000000ULL,
    0xaa00000000000ULL,
    0x1100000000000ULL,
    0x6780000000000ULL,
    0xa0c0000000000ULL,
    0x5320000000000ULL,
    0xb750000000000ULL,
    0xa278000000000ULL,
    0x1154000000000ULL,
    0x5d7e000000000ULL,
    0xd5d1000000000ULL,
    0xf3e800000000ULL,
    0x53b4c00000000ULL,
    0xce8a200000000ULL,
    0xd839900000000ULL,
    0x7b36a80000000ULL,
    0x2c8ac0000000ULL,
    0xf558ae0000000ULL,
    0x7403b50000000ULL,
    0xda02548000000ULL,
    0x1900c6c000000ULL,
    0x5b80ae2000000ULL,
    0x26c7b53000000ULL,
    0x7c2454a800000ULL,
    0x6bd5c6c400000ULL,
    0x74382e0e00000ULL,
    0x8536751500000ULL,
    0x19cef48280000ULL,
    0x93d996c040000ULL,
    0xb422617e0000ULL,
    0xbde7893d30000ULL,
    0x5774d29648000ULL,
    0x5faa1fed94000ULL,
    0xf268f4bf7e000ULL,
    0x90cc96f965000ULL,
    0xb05aa62117800ULL,
    0x19864928e0400ULL,
    0x7bc672b4b1a00ULL,
    0x95a24fd988300ULL,
    0xd814fc94b5080ULL,
    0x7d986affee240ULL,
    0xcb20000efc0a0ULL,
    0xf35000087a1b0ULL,
    0x1078001283238ULL,
    0x8454003a9ec94ULL,
    0x28fe0013549f2ULL,
    0xf011001ade2abULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xc800000000000ULL,
    0xe400000000000ULL,
    0x7600000000000ULL,
    0x2900000000000ULL,
    0x780000000000ULL,
    0xbbc0000000000ULL,
    0x5de0000000000ULL,
    0x6d50000000000ULL,
    0xdc68000000000ULL,
    0x2454000000000ULL,
    0xebea000000000ULL,
    0xb793000000000ULL,
    0x2209800000000ULL,
    0x64c5400000000ULL,
    0x3366a00000000ULL,
    0x3112f00000000ULL,
    0x4a4de80000000ULL,
    0xdce1cc0000000ULL,
    0xbed1b60000000ULL,
    0x39ae630000000ULL,
    0xc4b07c8000000ULL,
    0x3f38704000000ULL,
    0x1d3bb62000000ULL,
    0x2a3d631000000ULL,
    0x70b9fcb800000ULL,
    0x2fd305c00000ULL,
    0xe1dd161200000ULL,
    0x44ef932900000ULL,
    0x111414a580000ULL,
    0x9a4cfc5640000ULL,
    0x84e4a013e0000ULL,
    0xe2d5f007f0000ULL,
    0x63ae6832f8000ULL,
    0x7fb78c0d14000ULL,
    0x67be9604fa000ULL,
    0x8879d31ee5000ULL,
    0xcb9b34b002800ULL,
    0xfbcb4c7cf0c00ULL,
    0xee27e81e78600ULL,
    0x74b2cc2ed4500ULL,
    0x573836195b280ULL,
    0x693b2326f47c0ULL,
    0xe43edc9f0b1a0ULL,
    0x75be805e1c230ULL,
    0x977c5e2417458ULL,
    0x51faf358be4cULL,
    0x3789ca92dd052ULL,
    0xc0021353373f1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0x6400000000000ULL,
    0xfe00000000000ULL,
    0x5d00000000000ULL,
    0x8c80000000000ULL,
    0x9940000000000ULL,
    0x6d60000000000ULL,
    0x8af0000000000ULL,
    0xc648000000000ULL,
    0xe3f4000000000ULL,
    0x7cce000000000ULL,
    0x33b7000000000ULL,
    0x1b2e800000000ULL,
    0x7104400000000ULL,
    0xa685a00000000ULL,
    0x2247b00000000ULL,
    0xa0e1480000000ULL,
    0x6d30d40000000ULL,
    0xf56d020000000ULL,
    0x2be4ff0000000ULL,
    0x78b4758000000ULL,
    0x6eaa3d4000000ULL,
    0xfac3022000000ULL,
    0x1fa3ff3000000ULL,
    0x2152f59800000ULL,
    0x111a7d4400000ULL,
    0xdbe8a23a00000ULL,
    0xbda34f2900000ULL,
    0x2e553da780000ULL,
    0x2a9ae95340000ULL,
    0x152e001920000ULL,
    0x9407000f50000ULL,
    0x4606803cd8000ULL,
    0x31004031fc000ULL,
    0x683a008b2000ULL,
    0x5244b007ad000ULL,
    0xd8e1c83beb800ULL,
    0x6133943911c00ULL,
    0x6f6e221479a00ULL,
    0x88e30f02ec100ULL,
    0xa9369dadcae80ULL,
    0x7b6e5966c02c0ULL,
    0xee7c83f61f20ULL,
    0xf830940ff0250ULL,
    0x6deea213c9a88ULL,
    0x34a04f0a8513cULL,
    0x44d5bdaadeef6ULL,
    0xf2d9a945a7aa5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xe800000000000ULL,
    0xdc00000000000ULL,
    0x9e00000000000ULL,
    0xf900000000000ULL,
    0x1580000000000ULL,
    0x4f40000000000ULL,
    0x42a0000000000ULL,
    0x3e30000000000ULL,
    0x3e38000000000ULL,
    0xb34000000000ULL,
    0x2dba000000000ULL,
    0x2971000000000ULL,
    0x89b800000000ULL,
    0xd706400000000ULL,
    0x4485600000000ULL,
    0x26c6b00000000ULL,
    0xa3e0a80000000ULL,
    0x6d931c0000000ULL,
    0xc98b520000000ULL,
    0xe44a490000000ULL,
    0x9d2a438000000ULL,
    0x2f7848c000000ULL,
    0x5091522000000ULL,
    0x960b491000000ULL,
    0x8009c3a800000ULL,
    0x750a08cc00000ULL,
    0x868e321a00000ULL,
    0x34ccf92700000ULL,
    0xc9eaeb8f80000ULL,
    0x39b54f240000ULL,
    0xd82001f60000ULL,
    0x9b450034d0000ULL,
    0x90a1801f28000ULL,
    0x8137403dcc000ULL,
    0xe6bee030ee000ULL,
    0x13f0f0261d000ULL,
    0x475dc83c46800ULL,
    0xb961ac3b90400ULL,
    0x5251fa08c8e00ULL,
    0xefe85524dc900ULL,
    0x1e9a918ae7d80ULL,
    0x220441cc61ec0ULL,
    0xb706718550ca0ULL,
    0x9481b1fc68570ULL,
    0x5ec239a02d0c8ULL,
    0xa7e35de6ff7e4ULL,
    0xc79723bf945aeULL,
    0x728af8ce7b453ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0xc400000000000ULL,
    0x9e00000000000ULL,
    0x1f00000000000ULL,
    0x6080000000000ULL,
    0x3a40000000000ULL,
    0xb3a0000000000ULL,
    0xa370000000000ULL,
    0x1998000000000ULL,
    0x2074000000000ULL,
    0xfb1e000000000ULL,
    0x4737000000000ULL,
    0x1539800000000ULL,
    0x7383400000000ULL,
    0xa0c7a00000000ULL,
    0x98e3b00000000ULL,
    0x1751580000000ULL,
    0xac2bb40000000ULL,
    0xea0b0a0000000ULL,
    0x4ab8690000000ULL,
    0x7945318000000ULL,
    0x3124d5c000000ULL,
    0x34350a2000000ULL,
    0x7fbf693000000ULL,
    0xd0c4b19800000ULL,
    0xd0e395fc00000ULL,
    0xb354aa0200000ULL,
    0xc22fd90100000ULL,
    0x7d0a69bf80000ULL,
    0xee3861fbc0000ULL,
    0xdd07801a20000ULL,
    0x9d84400f90000ULL,
    0xf7c6201368000ULL,
    0x5c64f0131c000ULL,
    0x4310f83c46000ULL,
    0x888b04378d000ULL,
    0xedfdd235af800ULL,
    0xf7679d3ed1c00ULL,
    0xf5901b9b08600ULL,
    0x9cc4cda6dd00ULL,
    0x50dec39a1e680ULL,
    0xf957b8ef292c0ULL,
    0xfb2de9befc360ULL,
    0x2e8c21f3f73d0ULL,
    0x1ef9a03674aa8ULL,
    0x5de4b0185b364ULL,
    0x36d0d8034b7a6ULL,
    0x89ecf43163087ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0xc00000000000ULL,
    0xf600000000000ULL,
    0x9100000000000ULL,
    0xbc80000000000ULL,
    0x240000000000ULL,
    0x38e0000000000ULL,
    0x4a70000000000ULL,
    0x7b88000000000ULL,
    0xfd74000000000ULL,
    0x2e0a000000000ULL,
    0x27b1000000000ULL,
    0x1ead800000000ULL,
    0xfa27400000000ULL,
    0xed5e00000000ULL,
    0x341cf00000000ULL,
    0x760f680000000ULL,
    0xf3b2140000000ULL,
    0x7cabfa0000000ULL,
    0xa925130000000ULL,
    0x1155098000000ULL,
    0xe1dcdf4000000ULL,
    0x9729fa2000000ULL,
    0xaee0131000000ULL,
    0xcb7289a800000ULL,
    0x5f0a9f7c00000ULL,
    0xcb319a1200000ULL,
    0x24eba31300000ULL,
    0x6c0019580000ULL,
    0xbea07b5840000ULL,
    0x2897083d20000ULL,
    0xa6f9a41390000ULL,
    0x6379721bb8000ULL,
    0x61bff70adc000ULL,
    0xc51b9b83c2000ULL,
    0xda8ad87ccd000ULL,
    0x69f289983a800ULL,
    0x704a9f6f30400ULL,
    0xd1519a3e69600ULL,
    0x1dba32d44d00ULL,
    0xc728019c0ff80ULL,
    0x96e47b69777c0ULL,
    0xf75082b4aba0ULL,
    0xa50ca40a78550ULL,
    0xac36f22dbd068ULL,
    0x96db7367238cULL,
    0xb801fb8d65f86ULL,
    0x8404686b1e62dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x7800000000000ULL,
    0x4c00000000000ULL,
    0x5200000000000ULL,
    0xe700000000000ULL,
    0x2880000000000ULL,
    0x6240000000000ULL,
    0x60000000000ULL,
    0x130000000000ULL,
    0x4b48000000000ULL,
    0xec34000000000ULL,
    0x38ce000000000ULL,
    0x2cf1000000000ULL,
    0xc1ef800000000ULL,
    0x2f22c00000000ULL,
    0x42d5e00000000ULL,
    0xf2bc500000000ULL,
    0x1258d80000000ULL,
    0x852a140000000ULL,
    0x70014a0000000ULL,
    0x1801570000000ULL,
    0xdc03638000000ULL,
    0x1a033d4000000ULL,
    0xd3074a2000000ULL,
    0x3684573000000ULL,
    0xd742e3a800000ULL,
    0xcfe0fd4c00000ULL,
    0x4bf52a3e00000ULL,
    0x296ec72300000ULL,
    0xed61dbbc80000ULL,
    0x72b7b975c0000ULL,
    0x8b8b383420000ULL,
    0x1513443b90000ULL,
    0x3b18123c98000ULL,
    0xafc883358c000ULL,
    0x1c704986f2000ULL,
    0x7fa8fa709d000ULL,
    0x354711baab800ULL,
    0x90e02e42b0400ULL,
    0x8f73bba089e00ULL,
    0x292d295455b00ULL,
    0x720000089ef80ULL,
    0x17000018df540ULL,
    0xf0800004bfd60ULL,
    0x1e4000374f350ULL,
    0x2a60000226aa8ULL,
    0xaa300003c301cULL,
    0x31c8001d550eeULL,
    0x6974000d1efa5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xd800000000000ULL,
    0xac00000000000ULL,
    0xe200000000000ULL,
    0xb300000000000ULL,
    0x380000000000ULL,
    0xd540000000000ULL,
    0x7660000000000ULL,
    0xaab0000000000ULL,
    0x948000000000ULL,
    0x93b4000000000ULL,
    0xfdca000000000ULL,
    0xa377000000000ULL,
    0x6bea800000000ULL,
    0x5921c00000000ULL,
    0xe3d7e00000000ULL,
    0x327ad00000000ULL,
    0x34bc780000000ULL,
    0x6f9fcc0000000ULL,
    0x6f4b520000000ULL,
    0xe6b13f0000000ULL,
    0xbb49278000000ULL,
    0x98b7c54000000ULL,
    0x4249522000000ULL,
    0x4c323f1000000ULL,
    0x289a79800000ULL,
    0x1211054400000ULL,
    0x8cdc321600000ULL,
    0xd42d2f3b00000ULL,
    0xb6803fa080000ULL,
    0xb3c71968c0000ULL,
    0x3da39816e0000ULL,
    0x4b111c0e50000ULL,
    0x685d2a3d18000ULL,
    0x3ce9f3026c000ULL,
    0x8ca0f5b4b2000ULL,
    0xab933a7abd000ULL,
    0x69d95ba6a800ULL,
    0x23c82a7eb1c00ULL,
    0x6a760da048a00ULL,
    0x176d3653f5700ULL,
    0x7de127941f780ULL,
    0x2bf3c571ef740ULL,
    0x412b522307be0ULL,
    0xa0013f3a82830ULL,
    0x3001278a35ac8ULL,
    0xa803c56b7f3bcULL,
    0x6403520afff7eULL,
    0x96053f23be725ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xe800000000000ULL,
    0x9c00000000000ULL,
    0x6a00000000000ULL,
    0xbf00000000000ULL,
    0xdd80000000000ULL,
    0xac40000000000ULL,
    0x4220000000000ULL,
    0xd010000000000ULL,
    0xe838000000000ULL,
    0x5914000000000ULL,
    0x80ba000000000ULL,
    0x7d5000000000ULL,
    0xeb5f800000000ULL,
    0x4da2400000000ULL,
    0x6753600000000ULL,
    0xf99d100000000ULL,
    0x3440980000000ULL,
    0x5620fc0000000ULL,
    0xb6157a0000000ULL,
    0x653a2d0000000ULL,
    0x4f955a8000000ULL,
    0x77d5cc000000ULL,
    0x3cb77a2000000ULL,
    0x1beb2d1000000ULL,
    0x448da8800000ULL,
    0x381e1cec00000ULL,
    0xb2019a1a00000ULL,
    0x8b017d3700000ULL,
    0xb84a29280000ULL,
    0xc941b0c3c0000ULL,
    0xc8a7182d60000ULL,
    0x3dd6bc1c10000ULL,
    0x6c5c1a0208000ULL,
    0x74223d37c4000ULL,
    0x651242b76e000ULL,
    0x1abbe0da55000ULL,
    0x70d3600a26800ULL,
    0x1add100a31400ULL,
    0x63e0983fb9e00ULL,
    0x670fc023d900ULL,
    0x9e0d7a1372580ULL,
    0xcf3e2d2c96040ULL,
    0x90975a9749c60ULL,
    0x4afc5cff65af0ULL,
    0xc8f2fa031f068ULL,
    0x2dcc6d1dcc8f4ULL,
    0x225c3abcaa90aULL,
    0x5254cdf3bf83ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x1800000000000ULL,
    0x7400000000000ULL,
    0x2600000000000ULL,
    0x4300000000000ULL,
    0xdc80000000000ULL,
    0x1c40000000000ULL,
    0x76a0000000000ULL,
    0xad10000000000ULL,
    0x8248000000000ULL,
    0x6c14000000000ULL,
    0xe3ce000000000ULL,
    0x7bd3000000000ULL,
    0x23ac800000000ULL,
    0xe723c00000000ULL,
    0xd2d7200000000ULL,
    0x9e2ef00000000ULL,
    0xeae7380000000ULL,
    0x1bb1140000000ULL,
    0xcf5e5a0000000ULL,
    0xbe5ae10000000ULL,
    0x97dd528000000ULL,
    0xec1fe34000000ULL,
    0x7e785a2000000ULL,
    0x878de13000000ULL,
    0xe977d2b800000ULL,
    0x50bb235400000ULL,
    0x26dfa2600000ULL,
    0x5c43d12d00000ULL,
    0x56a34ab180000ULL,
    0x1d130744c0000ULL,
    0xca4db81120000ULL,
    0x15d42a10000ULL,
    0xb1cbfa2c28000ULL,
    0x1ed4d12f84000ULL,
    0xbc29ca91b2000ULL,
    0x27e7c74115000ULL,
    0xb830180cdb800ULL,
    0x459fe41570c00ULL,
    0xc5b9622759200ULL,
    0xf5ebf531ddf00ULL,
    0x40830897ee480ULL,
    0x2645024f3b7c0ULL,
    0xcfa508bb40ee0ULL,
    0x28920244e1b50ULL,
    0x4b8f8890b1de8ULL,
    0xcb76c275f9c44ULL,
    0x9dba28940dbaaULL,
    0xa1e8f256674ddULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x5800000000000ULL,
    0xdc00000000000ULL,
    0xaa00000000000ULL,
    0xa900000000000ULL,
    0x6b80000000000ULL,
    0xcbc0000000000ULL,
    0x6460000000000ULL,
    0x9290000000000ULL,
    0x7bf8000000000ULL,
    0xe594000000000ULL,
    0xf77e000000000ULL,
    0x8ad1000000000ULL,
    0xd85b800000000ULL,
    0xf163c00000000ULL,
    0x8315600000000ULL,
    0x513a900000000ULL,
    0xde72280000000ULL,
    0x802acc0000000ULL,
    0x4a4cf60000000ULL,
    0xc7daaf0000000ULL,
    0x74a2428000000ULL,
    0x2873c0c000000ULL,
    0x432af62000000ULL,
    0x28cfaf3000000ULL,
    0xd79fc2a800000ULL,
    0x830500ec00000ULL,
    0x282163600000ULL,
    0x43ff0700000ULL,
    0x1fa50a8280000ULL,
    0xaef39cc640000ULL,
    0x356c282ce0000ULL,
    0x346bcc35f0000ULL,
    0xbbef761b98000ULL,
    0x162d6f22e4000ULL,
    0x994922921e000ULL,
    0x4d5850fc95000ULL,
    0xe0e35e2e47800ULL,
    0xa9d6a31c50400ULL,
    0x6ade549a08e00ULL,
    0xa9213fdbcdf00ULL,
    0x40b47cb202d80ULL,
    0x55cff3d55ee40ULL,
    0x421e8abdf46a0ULL,
    0xdfc05cd377430ULL,
    0x3a61480b69e58ULL,
    0xb5955c1efc4fcULL,
    0x7f7b5e113afaaULL,
    0x1ed2a303aab33ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0xb400000000000ULL,
    0xde00000000000ULL,
    0x500000000000ULL,
    0xd180000000000ULL,
    0x5dc0000000000ULL,
    0x25e0000000000ULL,
    0x6bd0000000000ULL,
    0x8c48000000000ULL,
    0xaa54000000000ULL,
    0xb98a000000000ULL,
    0xdbb3000000000ULL,
    0x9c5e800000000ULL,
    0x7afa400000000ULL,
    0x538f200000000ULL,
    0x40b6b00000000ULL,
    0xa8dc080000000ULL,
    0x66bebc0000000ULL,
    0x3aca60000000ULL,
    0xba84b30000000ULL,
    0x9143488000000ULL,
    0x45a7d94000000ULL,
    0xc1f2262000000ULL,
    0x857ef31000000ULL,
    0xf34c68b800000ULL,
    0xc8d1696400000ULL,
    0x64ce2e2a00000ULL,
    0x3c104f3d00000ULL,
    0xa2a8ce8f80000ULL,
    0xdd01da6540000ULL,
    0x9d87669e60000ULL,
    0xffc4966a70000ULL,
    0xdae468a6f8000ULL,
    0x4955696fb4000ULL,
    0xb10c2e0572000ULL,
    0xddf74f24e5000ULL,
    0xaf7c4ea29a800ULL,
    0xc8489a6458c00ULL,
    0x8c56c6ade5a00ULL,
    0xc088665f1b900ULL,
    0xf03740a819480ULL,
    0x329d655e056c0ULL,
    0x3f9c80012aa20ULL,
    0x6a1d4039003f0ULL,
    0xb1dba01580618ULL,
    0x9d3ff0304140cULL,
    0x656da82de0f02ULL,
    0xdbe24c3131095ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0xac00000000000ULL,
    0x9e00000000000ULL,
    0x1300000000000ULL,
    0xb80000000000ULL,
    0x6840000000000ULL,
    0x2fa0000000000ULL,
    0x21d0000000000ULL,
    0x99d8000000000ULL,
    0x3a54000000000ULL,
    0x899e000000000ULL,
    0x1f1000000000ULL,
    0x8249800000000ULL,
    0x512e400000000ULL,
    0x9298e00000000ULL,
    0x3674b00000000ULL,
    0xcc09780000000ULL,
    0xd98d6c0000000ULL,
    0x62cc7a0000000ULL,
    0xe6e8f10000000ULL,
    0x5f7bc48000000ULL,
    0x10647c000000ULL,
    0xb685fa2000000ULL,
    0x38c6b13000000ULL,
    0x106324a800000ULL,
    0x1032f7fc00000ULL,
    0x28ac821a00000ULL,
    0x17dbdd1b00000ULL,
    0x51575e8f80000ULL,
    0x961e06f8c0000ULL,
    0x43b14698e0000ULL,
    0x64e89ac110000ULL,
    0x5a7d24a468000ULL,
    0xe583f7c0b4000ULL,
    0x5345021696000ULL,
    0x28259d3385000ULL,
    0x2797be9c0f800ULL,
    0x4d7eb6dbc8400ULL,
    0xbc063eb984600ULL,
    0xe604f6df0e900ULL,
    0x700dea049b80ULL,
    0x218146d7456c0ULL,
    0xa140268eee3e0ULL,
    0x5226ad9d9cb0ULL,
    0xcf14bca66ca68ULL,
    0xc6be2bd17a504ULL,
    0x57e600173f4f2ULL,
    0x76750015501afULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x1800000000000ULL,
    0x7c00000000000ULL,
    0xea00000000000ULL,
    0x2700000000000ULL,
    0x3a80000000000ULL,
    0x4c40000000000ULL,
    0x9960000000000ULL,
    0x4c70000000000ULL,
    0x478000000000ULL,
    0xeef4000000000ULL,
    0x743a000000000ULL,
    0xfd97000000000ULL,
    0xef4d800000000ULL,
    0xdb6fc00000000ULL,
    0x31fa600000000ULL,
    0xdc32900000000ULL,
    0xa71c580000000ULL,
    0xd86c40000000ULL,
    0xdec5fa0000000ULL,
    0xd1226f0000000ULL,
    0xb312098000000ULL,
    0x9d0b94c000000ULL,
    0x8b087a2000000ULL,
    0x60daf1000000ULL,
    0x7b88699800000ULL,
    0xfd4904fc00000ULL,
    0x306c222600000ULL,
    0x997f6b0f00000ULL,
    0x3b7793a280000ULL,
    0xf6fc6bf5c0000ULL,
    0x56b3aba8a0000ULL,
    0x935b3fdc10000ULL,
    0x18e589a4d8000ULL,
    0x60b354f6dc000ULL,
    0x8e5f9a31be000ULL,
    0x8d60ff1bad000ULL,
    0x8276519fd6800ULL,
    0xd57950c6b9c00ULL,
    0xe9778028ed600ULL,
    0x3df8c018b6f00ULL,
    0x6e37e01b08180ULL,
    0x9c1d503da5640ULL,
    0x2d06383df2760ULL,
    0xad8454110b410ULL,
    0xec1a23b07668ULL,
    0x5920ab1d40ffcULL,
    0x2715f3ac60746ULL,
    0x130afbc530643ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0xbc00000000000ULL,
    0xae00000000000ULL,
    0xef00000000000ULL,
    0x5a80000000000ULL,
    0x5840000000000ULL,
    0xc320000000000ULL,
    0xcf10000000000ULL,
    0x3c08000000000ULL,
    0xed94000000000ULL,
    0x104a000000000ULL,
    0xcb1000000000ULL,
    0x265e800000000ULL,
    0x1138400000000ULL,
    0xdb0f600000000ULL,
    0xdb17300000000ULL,
    0x6e09180000000ULL,
    0xc93640000000ULL,
    0x35cf9a0000000ULL,
    0x26741d0000000ULL,
    0x13dd28000000ULL,
    0xb30f654000000ULL,
    0xc7111a2000000ULL,
    0x500c5d1000000ULL,
    0xcb92b2a800000ULL,
    0xd348556400000ULL,
    0xd030022a00000ULL,
    0x2d1b393f00000ULL,
    0x269f288380000ULL,
    0xa359485fc0000ULL,
    0x7eb950bca0000ULL,
    0x14cd1c6910000ULL,
    0xe3f7d29348000ULL,
    0x1bfe657c04000ULL,
    0x2c6f9a1ba2000ULL,
    0xf1241d3675000ULL,
    0x5e15d2bd5a800ULL,
    0x18b654028400ULL,
    0xc3531a11b5a00ULL,
    0xd295d2dfb100ULL,
    0x5386329739580ULL,
    0x21c1154ffd8c0ULL,
    0x77e1e2117fe60ULL,
    0xbf744926dbc90ULL,
    0x50b95083e8be8ULL,
    0xbbcd1c6b158b4ULL,
    0x1977d2a5ea12aULL,
    0xd3be657af0505ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x1800000000000ULL,
    0x3c00000000000ULL,
    0x6a00000000000ULL,
    0xdd00000000000ULL,
    0xdc80000000000ULL,
    0xf6c0000000000ULL,
    0xf9a0000000000ULL,
    0x9e10000000000ULL,
    0xfbd8000000000ULL,
    0x2e94000000000ULL,
    0xdf1e000000000ULL,
    0x4637000000000ULL,
    0xff8e800000000ULL,
    0xb22fc00000000ULL,
    0xa23b600000000ULL,
    0xb2e2900000000ULL,
    0xe7f2880000000ULL,
    0xd7aef40000000ULL,
    0x7e7cfe0000000ULL,
    0x7a81590000000ULL,
    0x89c21e8000000ULL,
    0xbc20cfc000000ULL,
    0xf2527e2000000ULL,
    0x5bbe993000000ULL,
    0x74a17e9800000ULL,
    0xfa965fc400000ULL,
    0x411ef62600000ULL,
    0xb5376d3f00000ULL,
    0xb80b008280000ULL,
    0x376cc6f340000ULL,
    0x90d9889120000ULL,
    0x951232c2b0000ULL,
    0x505d769ce8000ULL,
    0xb9d76be4c4000ULL,
    0x1cf96837d6000ULL,
    0xcf44a40d15000ULL,
    0x5fe3960d2f800ULL,
    0x9376fd0a49c00ULL,
    0x756908aab5a00ULL,
    0xd9daf2e2def00ULL,
    0x779096ac81580ULL,
    0x259e3be041640ULL,
    0xff68009a1020ULL,
    0x23abc03cf14d0ULL,
    0x507d601e48678ULL,
    0xe1819004b4324ULL,
    0xfa420831de05aULL,
    0x7766341400befULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x7800000000000ULL,
    0xe400000000000ULL,
    0xde00000000000ULL,
    0x8b00000000000ULL,
    0x4480000000000ULL,
    0xdcc0000000000ULL,
    0xeb20000000000ULL,
    0xb070000000000ULL,
    0xb0e8000000000ULL,
    0x8f4000000000ULL,
    0xa62a000000000ULL,
    0x2ed3000000000ULL,
    0x8ddf800000000ULL,
    0xa078c00000000ULL,
    0xddcca00000000ULL,
    0xc784300000000ULL,
    0xa443f80000000ULL,
    0x5de5940000000ULL,
    0xc653de0000000ULL,
    0x531f6d0000000ULL,
    0x1a5aed8000000ULL,
    0xcc3c524000000ULL,
    0x1c2c5e2000000ULL,
    0x3bd7ad1000000ULL,
    0xa25e4db800000ULL,
    0x883c626c00000ULL,
    0x922da63e00000ULL,
    0x78d5392900000ULL,
    0x7ad8138f80000ULL,
    0x6ef8cf4ec0000ULL,
    0x2c0c6baf20000ULL,
    0x7219b5e30000ULL,
    0x5271159548000ULL,
    0x51edc672dc000ULL,
    0xd175803b1a000ULL,
    0x296bc0300d000ULL,
    0x67b32002c2800ULL,
    0x5f4cf01068c00ULL,
    0x84475837ede00ULL,
    0x4de5a406d3300ULL,
    0xee52260291a80ULL,
    0x7f1df903b9cc0ULL,
    0x585cb38db51e0ULL,
    0x7d38ff6676550ULL,
    0xdad93999f5f8ULL,
    0x28930f50b3174ULL,
    0xd13f4b9141256ULL,
    0xfad6b79e1819ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xa800000000000ULL,
    0x2c00000000000ULL,
    0xa00000000000ULL,
    0xd00000000000ULL,
    0xd80000000000ULL,
    0xf340000000000ULL,
    0x4960000000000ULL,
    0x5330000000000ULL,
    0x2ae8000000000ULL,
    0x6ab4000000000ULL,
    0x67ae000000000ULL,
    0x50d1000000000ULL,
    0xea1c800000000ULL,
    0x7fc400000000ULL,
    0x1c09e00000000ULL,
    0x83c2100000000ULL,
    0x8fa0880000000ULL,
    0xcd12fc0000000ULL,
    0x46b9520000000ULL,
    0xbc6a570000000ULL,
    0xb765b8000000ULL,
    0xa908484000000ULL,
    0x6a45d22000000ULL,
    0xaae6173000000ULL,
    0x7b77bba800000ULL,
    0x910e585400000ULL,
    0x4e435a0a00000ULL,
    0xdce1eb3b00000ULL,
    0xd47469aa80000ULL,
    0xbd8d4f5740000ULL,
    0xba86618960000ULL,
    0x6bc2f347d0000ULL,
    0xc3a25398d8000ULL,
    0x3712f4738c000ULL,
    0xb3bb602bda000ULL,
    0x35ef50097d000ULL,
    0xde35e80b33800ULL,
    0xe76cac1cb8400ULL,
    0x39f03a1bdd200ULL,
    0x7ecabb2fc2100ULL,
    0xabe7818d51f80ULL,
    0xecf4e35f98c40ULL,
    0x674cdba2ed020ULL,
    0x76a108700aaf0ULL,
    0xde96b21fa5ac8ULL,
    0x96fd4731bf51cULL,
    0x838c538ae314eULL,
    0x983f44960fd1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xe800000000000ULL,
    0xcc00000000000ULL,
    0xba00000000000ULL,
    0x1d00000000000ULL,
    0x6c80000000000ULL,
    0x3740000000000ULL,
    0x6e0000000000ULL,
    0x9ad0000000000ULL,
    0x3dd8000000000ULL,
    0xcc54000000000ULL,
    0x579a000000000ULL,
    0x4633000000000ULL,
    0x4a0f800000000ULL,
    0x950c400000000ULL,
    0xd08de00000000ULL,
    0x454b900000000ULL,
    0x47ea080000000ULL,
    0x345d740000000ULL,
    0xa394f20000000ULL,
    0xa83f230000000ULL,
    0xe907f48000000ULL,
    0x28125c000000ULL,
    0x5443722000000ULL,
    0xb167631000000ULL,
    0x8e1014b800000ULL,
    0xb1f9b5ec00000ULL,
    0xb1a6fa1a00000ULL,
    0xdc36572300000ULL,
    0x4709069680000ULL,
    0x418d06eb40000ULL,
    0x73cb068120000ULL,
    0xddaa06eed0000ULL,
    0xb63e86b738000ULL,
    0xc20546ddf4000ULL,
    0x2904e6b656000ULL,
    0xa28696f1c5000ULL,
    0x4410e98de800ULL,
    0xe96772ff78c00ULL,
    0xaa12748a55e00ULL,
    0xc7fe65d3c6100ULL,
    0x16a11237ddf80ULL,
    0xadb0b323fa240ULL,
    0x1ccffcba17620ULL,
    0x702b51e8654d0ULL,
    0xeffa00284ec48ULL,
    0x7aa30016e0eccULL,
    0x47b7802fb01f2ULL,
    0x59c8400a49c47ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x1800000000000ULL,
    0x2c00000000000ULL,
    0x6200000000000ULL,
    0x4b00000000000ULL,
    0xbd80000000000ULL,
    0x5cc0000000000ULL,
    0xb420000000000ULL,
    0x3190000000000ULL,
    0xee88000000000ULL,
    0x7e14000000000ULL,
    0x414e000000000ULL,
    0xebb7000000000ULL,
    0x8618800000000ULL,
    0xce1fc00000000ULL,
    0x3a1de00000000ULL,
    0x141ad00000000ULL,
    0xed1cb80000000ULL,
    0x399bc0000000ULL,
    0xcede9e0000000ULL,
    0x443d7d0000000ULL,
    0x8a8c028000000ULL,
    0xe810514000000ULL,
    0x244e1e2000000ULL,
    0xaf36bd3000000ULL,
    0x345fe29800000ULL,
    0xb77d817400000ULL,
    0x816a262600000ULL,
    0x3420c13b00000ULL,
    0xf1949c8080000ULL,
    0x8e8e2c66c0000ULL,
    0xae149c8960000ULL,
    0x594e2c6c30000ULL,
    0xc7b49c8d88000ULL,
    0xe41e2c5aa4000ULL,
    0x851c9caac2000ULL,
    0x879a2c47b5000ULL,
    0x48da9c9bdb800ULL,
    0x59392c6b49c00ULL,
    0x320c1c93c4200ULL,
    0x2051ec66f2f00ULL,
    0x3a297cba3cf80ULL,
    0xcbc4fc797f740ULL,
    0x3a024858b0e0ULL,
    0xa253906a709f0ULL,
    0x612c02b2295f8ULL,
    0xe405177f52b4ULL,
    0xa3661e377b8eaULL,
    0x6c72bd2a994b5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0xb400000000000ULL,
    0xb200000000000ULL,
    0xf900000000000ULL,
    0x2d80000000000ULL,
    0xa240000000000ULL,
    0xd3e0000000000ULL,
    0x3930000000000ULL,
    0xaf98000000000ULL,
    0x9eb4000000000ULL,
    0xd0da000000000ULL,
    0x6ad5000000000ULL,
    0xd6ac800000000ULL,
    0x1129c00000000ULL,
    0x9e6d200000000ULL,
    0x220e500000000ULL,
    0x87be80000000ULL,
    0x7581bc0000000ULL,
    0x76463a0000000ULL,
    0x91e4830000000ULL,
    0x8831638000000ULL,
    0xae19b7c000000ULL,
    0x3af2ba2000000ULL,
    0x4839431000000ULL,
    0x8766438800000ULL,
    0xf6f2e7f400000ULL,
    0xfe3dd20600000ULL,
    0xa4653f3d00000ULL,
    0xde7759a480000ULL,
    0xeffd34ca40000ULL,
    0x2ac3d98d60000ULL,
    0x5220f4d590000ULL,
    0xf94f9b078000ULL,
    0x47cba4d40c000ULL,
    0x54db91ae86000ULL,
    0xf0d7d8c63d000ULL,
    0xf3aa8ba24e800ULL,
    0x72a80be3b9400ULL,
    0x5b2c8037ad200ULL,
    0x2369c03c37700ULL,
    0xf58d202eb5c80ULL,
    0x7f3e5013aad40ULL,
    0x8de3e8294bda0ULL,
    0xa635bc0b1b1f0ULL,
    0x391c3a36c2468ULL,
    0xa07183097ef4cULL,
    0xa0fde3b5f102eULL,
    0x244077e96832fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0x7400000000000ULL,
    0x200000000000ULL,
    0x4d00000000000ULL,
    0x9c80000000000ULL,
    0xdfc0000000000ULL,
    0x40e0000000000ULL,
    0x3030000000000ULL,
    0x938000000000ULL,
    0x82b4000000000ULL,
    0x7dfe000000000ULL,
    0x79d5000000000ULL,
    0x7f8d800000000ULL,
    0x3d4cc00000000ULL,
    0x692a200000000ULL,
    0xeadad00000000ULL,
    0xa501480000000ULL,
    0x6081b40000000ULL,
    0xb1c4760000000ULL,
    0xbe7130000000ULL,
    0xdbb7678000000ULL,
    0x737f72c000000ULL,
    0x8311f62000000ULL,
    0x9fefd33000000ULL,
    0x9cbb478800000ULL,
    0x1bf4a2dc00000ULL,
    0x8bdb3e2e00000ULL,
    0xdf83a72d00000ULL,
    0x9846918880000ULL,
    0x1ca0a1cf40000ULL,
    0x3692b18920000ULL,
    0x76af71daf0000ULL,
    0x529e79b8b8000ULL,
    0xc9a205f34c000ULL,
    0x6e102f836e000ULL,
    0xb36fc6e65d000ULL,
    0x5b7e0029c7800ULL,
    0x5f15003039400ULL,
    0x81ed803c0d600ULL,
    0xefbcc00a4e300ULL,
    0x8472203cad080ULL,
    0xf39ed0117ff40ULL,
    0x9327480375320ULL,
    0xe6d0b41f625d0ULL,
    0x260ff61c12158ULL,
    0x120ad33d6b30cULL,
    0xf00ec78d476beULL,
    0xed0c62f978c1bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x2800000000000ULL,
    0xc00000000000ULL,
    0x200000000000ULL,
    0x3900000000000ULL,
    0xa580000000000ULL,
    0x1b40000000000ULL,
    0x3ca0000000000ULL,
    0xb250000000000ULL,
    0xd578000000000ULL,
    0x61d4000000000ULL,
    0xe13e000000000ULL,
    0xc7f5000000000ULL,
    0x2aa9800000000ULL,
    0x83ec400000000ULL,
    0x4e4de00000000ULL,
    0x859a500000000ULL,
    0x91a5280000000ULL,
    0xd1d4fc0000000ULL,
    0x293eb20000000ULL,
    0x1bf6790000000ULL,
    0xf0abb08000000ULL,
    0x9eeedac000000ULL,
    0xe5cf322000000ULL,
    0xa5de393000000ULL,
    0x3180508800000ULL,
    0xdd458afc00000ULL,
    0xdba59a2a00000ULL,
    0xf4d7853300000ULL,
    0x76bc828880000ULL,
    0xedb4e3f240000ULL,
    0x4f89628360000ULL,
    0xfab3f5d0000ULL,
    0x92924aa7a8000ULL,
    0xf19b4feed4000ULL,
    0x87a578be3e000ULL,
    0xeed176d1a5000ULL,
    0x4bbb2815e7800ULL,
    0xb631fc1c29400ULL,
    0x47cf321414600ULL,
    0xacde390f1e100ULL,
    0x9c00508714f80ULL,
    0x3a058af79fd40ULL,
    0xcd059a1d552a0ULL,
    0x73878518ff2f0ULL,
    0x44482ba85548ULL,
    0xae20e3d8574a4ULL,
    0x379762bad0462ULL,
    0x6e1fb3c529d9bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0xc00000000000ULL,
    0x4600000000000ULL,
    0x6300000000000ULL,
    0x7a80000000000ULL,
    0xf2c0000000000ULL,
    0x68e0000000000ULL,
    0x9cd0000000000ULL,
    0x5798000000000ULL,
    0x554000000000ULL,
    0x9fda000000000ULL,
    0xbf75000000000ULL,
    0x9bed800000000ULL,
    0xfc39400000000ULL,
    0xc521e00000000ULL,
    0x20b5700000000ULL,
    0x868df80000000ULL,
    0x7eae240000000ULL,
    0x139eae0000000ULL,
    0x2f55c30000000ULL,
    0x6aded88000000ULL,
    0xdef7f7c000000ULL,
    0x97a92e2000000ULL,
    0x2c19831000000ULL,
    0x1412b88800000ULL,
    0xf27bc7fc00000ULL,
    0x5c05360200000ULL,
    0x7e02d71300000ULL,
    0xc701ee9980000ULL,
    0xc08020e4c0000ULL,
    0x5fc5409ca0000ULL,
    0x7d60e3efb0000ULL,
    0x52961823b8000ULL,
    0xaebe5413f4000ULL,
    0x8be6d62146000ULL,
    0xa656a702e5000ULL,
    0x855b968e4e800ULL,
    0x9db244d329400ULL,
    0x8b0f8e9c3d600ULL,
    0xe4ed10ca2b500ULL,
    0xd8bed8ba26f80ULL,
    0xe7f7d0741c0ULL,
    0xd12e02861e0ULL,
    0x499d8329c4d90ULL,
    0x1250b8bc3f538ULL,
    0x275ac7cab06ccULL,
    0x64b2b63739282ULL,
    0xac8e970fb564fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0x9400000000000ULL,
    0x7200000000000ULL,
    0xfb00000000000ULL,
    0xa280000000000ULL,
    0x4f40000000000ULL,
    0x4220000000000ULL,
    0x5210000000000ULL,
    0xc758000000000ULL,
    0x394000000000ULL,
    0x1e9e000000000ULL,
    0x2cf7000000000ULL,
    0x8daf800000000ULL,
    0xc73d400000000ULL,
    0xe027200000000ULL,
    0x1114d00000000ULL,
    0xd9d9f80000000ULL,
    0x9ad1440000000ULL,
    0x9dbd420000000ULL,
    0xb367350000000ULL,
    0xd533228000000ULL,
    0x32cfa7c000000ULL,
    0x4c0cc22000000ULL,
    0xe7ed753000000ULL,
    0xd29b829800000ULL,
    0x12f637cc00000ULL,
    0x88aa1a3200000ULL,
    0xbce11500000ULL,
    0x58e1388480000ULL,
    0xa3f746f2c0000ULL,
    0xad2bfa9aa0000ULL,
    0x5b7933c6d0000ULL,
    0xa481f83408000ULL,
    0x3645441644000ULL,
    0x33a3421376000ULL,
    0x1bd0353a35000ULL,
    0xf03ca299af800ULL,
    0x7ca2e7c479c00ULL,
    0x5b53e2069de00ULL,
    0x5c7da52c3a500ULL,
    0x70047a9f86480ULL,
    0xe80473c42cf40ULL,
    0xc406d806439e0ULL,
    0xa0194161a010ULL,
    0x2702ba2d97188ULL,
    0x4485711184394ULL,
    0xc647e09d56d6aULL,
    0x1ba6d2e62528fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0xec00000000000ULL,
    0x200000000000ULL,
    0x7700000000000ULL,
    0x2f80000000000ULL,
    0x3c0000000000ULL,
    0x6360000000000ULL,
    0x9470000000000ULL,
    0xc9f8000000000ULL,
    0x14f4000000000ULL,
    0x51ba000000000ULL,
    0x8255000000000ULL,
    0x7ad800000000ULL,
    0xa15b400000000ULL,
    0x6fe6e00000000ULL,
    0xde35500000000ULL,
    0x105a580000000ULL,
    0x4964ac0000000ULL,
    0x9f70860000000ULL,
    0xfc7c170000000ULL,
    0xb434958000000ULL,
    0x3b5bd0c000000ULL,
    0x8ce7062000000ULL,
    0xf7b2571000000ULL,
    0x7a9ff58800000ULL,
    0xb885c0fc00000ULL,
    0xbc43be1200000ULL,
    0x78a7ab2b00000ULL,
    0x23172b8880000ULL,
    0x7b8c7be1c0000ULL,
    0x44082d99e0000ULL,
    0x1fca2cebf0000ULL,
    0x88add83058000ULL,
    0xcadaec14dc000ULL,
    0x5223e6039e000ULL,
    0xf0560702cd000ULL,
    0xf8aa2dbe36800ULL,
    0x42db2ce349400ULL,
    0x9e225818f5600ULL,
    0x250ac205bd00ULL,
    0x47aa8614af380ULL,
    0x8159170ef4140ULL,
    0x9fe1159d5bf60ULL,
    0x163490e12efb0ULL,
    0xfc5be63ab5198ULL,
    0x4b620720fa11cULL,
    0xe8702dbf7e536ULL,
    0xd3fe2ce8bc4f3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x800000000000ULL,
    0xc400000000000ULL,
    0xea00000000000ULL,
    0x1500000000000ULL,
    0x1780000000000ULL,
    0x2cc0000000000ULL,
    0x6be0000000000ULL,
    0x550000000000ULL,
    0x3428000000000ULL,
    0xf3d4000000000ULL,
    0x5d6e000000000ULL,
    0x1df5000000000ULL,
    0x525c800000000ULL,
    0x334d400000000ULL,
    0xd2c6a00000000ULL,
    0x5ce0100000000ULL,
    0xe3d5080000000ULL,
    0xb56d2c0000000ULL,
    0xa9f3ae0000000ULL,
    0xe05c930000000ULL,
    0xea4bb48000000ULL,
    0xeb426a4000000ULL,
    0x8f212e2000000ULL,
    0x8ab4d33000000ULL,
    0x8b79948800000ULL,
    0xdafb3a5400000ULL,
    0x7d3c862200000ULL,
    0x865cef0100000ULL,
    0x314bb2b280000ULL,
    0x73c3c55140000ULL,
    0xf9643ca7e0000ULL,
    0x1613064a30000ULL,
    0xe70ea00878000ULL,
    0x7f64102014000ULL,
    0xbd130802ea000ULL,
    0x278c2c12c5000ULL,
    0xc5212e3523800ULL,
    0xfb4d33269400ULL,
    0xb4f994a6fd200ULL,
    0x623b3a7e56500ULL,
    0xf4dc861cf2280ULL,
    0x520cef2a21440ULL,
    0xf8e3b2b390620ULL,
    0xb9d7c548291b0ULL,
    0xd86a3ca69cc38ULL,
    0x22760655a760cULL,
    0xea9a20132b0f2ULL,
    0xbaad503d65c19ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x4800000000000ULL,
    0x9400000000000ULL,
    0x2600000000000ULL,
    0xdd00000000000ULL,
    0xe380000000000ULL,
    0xf340000000000ULL,
    0x4760000000000ULL,
    0x5930000000000ULL,
    0x1988000000000ULL,
    0x1bb4000000000ULL,
    0xab4e000000000ULL,
    0xcc93000000000ULL,
    0x679f800000000ULL,
    0x66e9400000000ULL,
    0xbe87200000000ULL,
    0xd0c2700000000ULL,
    0x5424c80000000ULL,
    0x2e55ec0000000ULL,
    0x8bf920000000ULL,
    0x963cb70000000ULL,
    0x96f88b8000000ULL,
    0xbade034000000ULL,
    0x488e122000000ULL,
    0xf236f73000000ULL,
    0x9f082bb800000ULL,
    0x3771334c00000ULL,
    0x9d6bfa3200000ULL,
    0x61c66b1500000ULL,
    0x8da2f1b180000ULL,
    0x5212287b40000ULL,
    0x675c438ae0000ULL,
    0x4acbef69d0000ULL,
    0x4cd1800058000ULL,
    0x6a7a401d0c000ULL,
    0x3918a01482000ULL,
    0x862b30133d000ULL,
    0xa2a3e8208b800ULL,
    0x6a979c3728c00ULL,
    0x7a9b5a2ee5e00ULL,
    0x65695b14c7500ULL,
    0x7dc719ac4a480ULL,
    0xdfa2b46d885c0ULL,
    0x991699bb54d20ULL,
    0x11d8f4728e2b0ULL,
    0xce0e39993fac8ULL,
    0xdef3c4587671cULL,
    0xa92dd1898370eULL,
    0x9a245856bc6bdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0x2400000000000ULL,
    0x8600000000000ULL,
    0xd500000000000ULL,
    0x3f80000000000ULL,
    0xddc0000000000ULL,
    0xa260000000000ULL,
    0xb610000000000ULL,
    0xe8d8000000000ULL,
    0xa094000000000ULL,
    0x909a000000000ULL,
    0x35000000000ULL,
    0x3c6d800000000ULL,
    0x67b9400000000ULL,
    0xc500a00000000ULL,
    0x2783700000000ULL,
    0xa1c1080000000ULL,
    0x78660c0000000ULL,
    0x4915260000000ULL,
    0xa059ff0000000ULL,
    0x11d5c68000000ULL,
    0x5bf01c000000ULL,
    0xf602a62000000ULL,
    0x1d05bf1000000ULL,
    0x3b80e68800000ULL,
    0xabc131fc00000ULL,
    0xff610e0200000ULL,
    0xad91c31900000ULL,
    0xb31b48a980000ULL,
    0xd7f682c940000ULL,
    0x190ece8de0000ULL,
    0x352c0dee70000ULL,
    0x3e9a002118000ULL,
    0x41350034c4000ULL,
    0x2ded801fd6000ULL,
    0xbb79402a55000ULL,
    0x56e0a00f3e800ULL,
    0xbd537011c9400ULL,
    0x52f9081b4d600ULL,
    0x66220c33fb500ULL,
    0xacb726321ea80ULL,
    0x6b28ff13199c0ULL,
    0xf79a46b5c5220ULL,
    0x74b741c256d30ULL,
    0x772d861d15e18ULL,
    0xfd9e8f19dee3cULL,
    0xf3b6ce8f38d3aULL,
    0x93a80dfa95537ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xd800000000000ULL,
    0x6c00000000000ULL,
    0x5200000000000ULL,
    0x2b00000000000ULL,
    0x5580000000000ULL,
    0x9bc0000000000ULL,
    0x5660000000000ULL,
    0x1870000000000ULL,
    0x98a8000000000ULL,
    0x74f4000000000ULL,
    0x9dee000000000ULL,
    0x1cd5000000000ULL,
    0x4038800000000ULL,
    0x7e6bc00000000ULL,
    0x8a17e00000000ULL,
    0x76db500000000ULL,
    0x159b80000000ULL,
    0x799a7c0000000ULL,
    0xee7b260000000ULL,
    0x83cc6f0000000ULL,
    0xec04958000000ULL,
    0x9201a24000000ULL,
    0xb05a62000000ULL,
    0x8586af3000000ULL,
    0x43c5f58800000ULL,
    0x3a64327400000ULL,
    0x4a737e1600000ULL,
    0xb3ac432b00000ULL,
    0x21708b9c80000ULL,
    0x629717ec0000ULL,
    0x4ab5758360000ULL,
    0x584bf24df0000ULL,
    0xe6c29e2918000ULL,
    0xfee61308dc000ULL,
    0xeb37b38d4a000ULL,
    0x1d89cd54cd000ULL,
    0x39a733b063800ULL,
    0x90160d60e9400ULL,
    0x9ded39fc4200ULL,
    0x9ad85d6a97f00ULL,
    0x935feb9b06780ULL,
    0x7299e1656f940ULL,
    0x6bfb2da96a4e0ULL,
    0xc00ade4f1d6f0ULL,
    0xd6600007aab18ULL,
    0xd87000147d8fcULL,
    0xb8a800265aeb6ULL,
    0xa4f4002765479ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x8800000000000ULL,
    0x4400000000000ULL,
    0xf600000000000ULL,
    0xef00000000000ULL,
    0x4780000000000ULL,
    0x38c0000000000ULL,
    0x8fe0000000000ULL,
    0x2070000000000ULL,
    0xcc78000000000ULL,
    0x96f4000000000ULL,
    0x283e000000000ULL,
    0xe855000000000ULL,
    0x4c68800000000ULL,
    0xf639400000000ULL,
    0x5356a00000000ULL,
    0x55ecf00000000ULL,
    0xb5fed80000000ULL,
    0xe5354c0000000ULL,
    0xe65fde0000000ULL,
    0x47e0530000000ULL,
    0x8474cd8000000ULL,
    0xca7838c000000ULL,
    0x21f15e2000000ULL,
    0xa3b8133000000ULL,
    0x6294ed8800000ULL,
    0xda8888f400000ULL,
    0x7ec9a60200000ULL,
    0xe06cef2100000ULL,
    0x743d4bb580000ULL,
    0x325467cfc0000ULL,
    0xe16ced93e0000ULL,
    0xf0bc88ef30000ULL,
    0xf17a63678000ULL,
    0x1749ef37dc000ULL,
    0x3badcb88fe000ULL,
    0x2e5927ce8d000ULL,
    0xe3e44d9677800ULL,
    0x827578e8c9400ULL,
    0x7d79fe2464200ULL,
    0xaa71e329c3500ULL,
    0x297ab586c2280ULL,
    0xf47084cc427c0ULL,
    0x527ed81103960ULL,
    0xcdf54c33a2030ULL,
    0xc1bfde3033df8ULL,
    0xf390531adb68cULL,
    0x360ccd856f256ULL,
    0xf78c38db644d3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x3800000000000ULL,
    0xfc00000000000ULL,
    0xc200000000000ULL,
    0x2700000000000ULL,
    0x3280000000000ULL,
    0x8ac0000000000ULL,
    0x8420000000000ULL,
    0x17d0000000000ULL,
    0xf698000000000ULL,
    0x4454000000000ULL,
    0x77de000000000ULL,
    0x2d33000000000ULL,
    0x136c800000000ULL,
    0x341ac00000000ULL,
    0x8697600000000ULL,
    0x87fed00000000ULL,
    0x6ce7580000000ULL,
    0xc4f2c40000000ULL,
    0x5bcc4e0000000ULL,
    0xac0ac50000000ULL,
    0xb1a8008000000ULL,
    0x5439d9c000000ULL,
    0x1f46ce2000000ULL,
    0xfc67053000000ULL,
    0xf935e0b800000ULL,
    0x356ac9d400000ULL,
    0x9d1c762e00000ULL,
    0x3916d10f00000ULL,
    0xdcbb168880000ULL,
    0x6e85d8ddc0000ULL,
    0x38c480a2a0000ULL,
    0x2b2319edb0000ULL,
    0xb151ae0988000ULL,
    0x7a59d52834000ULL,
    0xd972b88706000ULL,
    0xb7880dd8a5000ULL,
    0x44e838227f800ULL,
    0x6b58140878c00ULL,
    0x5af5163d5d200ULL,
    0xcecb012363b00ULL,
    0x3188ce9ffa580ULL,
    0x9de9dcf2f7740ULL,
    0x5cd9ae93ccf60ULL,
    0x95b00cd11b010ULL,
    0xcaac7683a74b8ULL,
    0xd4bb08d614c54ULL,
    0x3a83d89977762ULL,
    0xdec1ddfc8cb77ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x1800000000000ULL,
    0x6c00000000000ULL,
    0xaa00000000000ULL,
    0x8100000000000ULL,
    0x5080000000000ULL,
    0xdac0000000000ULL,
    0xf320000000000ULL,
    0xdd50000000000ULL,
    0x15b8000000000ULL,
    0xf2d4000000000ULL,
    0x8fe000000000ULL,
    0x18b1000000000ULL,
    0x51ce800000000ULL,
    0x1b39c00000000ULL,
    0xaf11e00000000ULL,
    0x205e900000000ULL,
    0xdea6e80000000ULL,
    0x2511940000000ULL,
    0x715c7e0000000ULL,
    0x3627ad0000000ULL,
    0x83d7a48000000ULL,
    0x307fd8c000000ULL,
    0x674fe2000000ULL,
    0x6ceb6d3000000ULL,
    0x996ec4a800000ULL,
    0xad2d88c400000ULL,
    0x73cb762600000ULL,
    0x2e38a92b00000ULL,
    0x9995328280000ULL,
    0xc19de1e440000ULL,
    0xe40124b220000ULL,
    0x1e0218ddb0000ULL,
    0xe7031e1e48000ULL,
    0x6b80fd2314000ULL,
    0x1340acbf4e000ULL,
    0x1560dcd505000ULL,
    0x5eb6683277800ULL,
    0xbac9543a38400ULL,
    0x6abb1e07bda00ULL,
    0x3554fd288b700ULL,
    0x11beac8713f80ULL,
    0x9cd1dcd4dfe40ULL,
    0x47f8e810fc1a0ULL,
    0x173094142b150ULL,
    0x9c8afe0ce2e78ULL,
    0x495a6d0db70f4ULL,
    0x8a2044bbd8732ULL,
    0x91d448e9eda33ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xa800000000000ULL,
    0xec00000000000ULL,
    0x3200000000000ULL,
    0xb700000000000ULL,
    0x1180000000000ULL,
    0x640000000000ULL,
    0x7b60000000000ULL,
    0x5b30000000000ULL,
    0x5a8000000000ULL,
    0x1d74000000000ULL,
    0xdece000000000ULL,
    0x7647000000000ULL,
    0x2364800000000ULL,
    0x2f36400000000ULL,
    0x23ade00000000ULL,
    0xdc70d00000000ULL,
    0xa64cd80000000ULL,
    0xe4861c0000000ULL,
    0xf8c10a0000000ULL,
    0x18a5170000000ULL,
    0x170b8000000ULL,
    0xe4ff264000000ULL,
    0x3b6cea2000000ULL,
    0x54d5c73000000ULL,
    0x6e5bd39800000ULL,
    0xbc793a5400000ULL,
    0x59ade00a00000ULL,
    0x1770d00b00000ULL,
    0x4dccd81480000ULL,
    0xe9c61c39c0000ULL,
    0x8210a2e60000ULL,
    0x1ed5173a90000ULL,
    0x5d5f0b9258000ULL,
    0x13fb267b0c000ULL,
    0x8aeaea250a000ULL,
    0x6296c736cd000ULL,
    0xed3953b16b800ULL,
    0xa34c7a5f5dc00ULL,
    0x82028003b3200ULL,
    0x8f05400790900ULL,
    0x3587602bd8f80ULL,
    0x884190134df40ULL,
    0x5665b80f2a160ULL,
    0x11b08c087c170ULL,
    0x20683224833a8ULL,
    0xd757db0a3971cULL,
    0xef1cd9ac9cf4eULL,
    0x2e9f2d46d34e9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x800000000000ULL,
    0x3c00000000000ULL,
    0xfe00000000000ULL,
    0xed00000000000ULL,
    0xae80000000000ULL,
    0x5440000000000ULL,
    0xba60000000000ULL,
    0x94f0000000000ULL,
    0xf418000000000ULL,
    0xf0b4000000000ULL,
    0xf67e000000000ULL,
    0x2041000000000ULL,
    0x1860800000000ULL,
    0xf7f5400000000ULL,
    0x6f9f600000000ULL,
    0x3e73b00000000ULL,
    0xda58880000000ULL,
    0x1dd3840000000ULL,
    0x3b0d4a0000000ULL,
    0x999e70000000ULL,
    0x8773658000000ULL,
    0xc6dec3c000000ULL,
    0x82922a2000000ULL,
    0xf6ea573000000ULL,
    0xa5abeda800000ULL,
    0x8e4d47f400000ULL,
    0x5b7f602200000ULL,
    0xeec3b03f00000ULL,
    0x3c20881780000ULL,
    0x9597840f40000ULL,
    0xcf6b4a29a0000ULL,
    0x86ce71a10000ULL,
    0x39ede59118000ULL,
    0xa82a83de7c000ULL,
    0xe18dca36a6000ULL,
    0xffdca7193d000ULL,
    0x5e1405bb07800ULL,
    0xc9a973d92c400ULL,
    0xb84ca2191e200ULL,
    0x2a7cd33ed0500ULL,
    0x6e4027b1f8580ULL,
    0x8d64e0c8ccac0ULL,
    0xbd75e5b92e020ULL,
    0xf1de83d099b10ULL,
    0xab13ca0c3c8a8ULL,
    0xf32da7178695cULL,
    0xfe0c85abecfb6ULL,
    0x8b1833e4fe51fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0x1400000000000ULL,
    0xda00000000000ULL,
    0x3300000000000ULL,
    0xcc80000000000ULL,
    0xb740000000000ULL,
    0xae60000000000ULL,
    0xc870000000000ULL,
    0xbd08000000000ULL,
    0xf734000000000ULL,
    0x6f6e000000000ULL,
    0xa941000000000ULL,
    0x7f64800000000ULL,
    0x83f2c00000000ULL,
    0x4cca200000000ULL,
    0x2515100000000ULL,
    0xd1ff380000000ULL,
    0xb179b40000000ULL,
    0xa83ca20000000ULL,
    0x1f5bd70000000ULL,
    0x80aed98000000ULL,
    0xaa640b4000000ULL,
    0x8a76822000000ULL,
    0x6a0ec73000000ULL,
    0x49b1e1a800000ULL,
    0x72dbf5400000ULL,
    0x5a2201e00000ULL,
    0xe951103500000ULL,
    0x6f19381e80000ULL,
    0x84cb418c0000ULL,
    0xee56220d20000ULL,
    0xe9981728d0000ULL,
    0xd408799d18000ULL,
    0xc8b7db7edc000ULL,
    0xb4ad9a3c62000ULL,
    0xe06363201d000ULL,
    0xe176fb9843800ULL,
    0xe28d1c4c4c400ULL,
    0xdcf27bae9b200ULL,
    0x8e4fdc6831b00ULL,
    0xdb505bb669080ULL,
    0xd01ecc4bd5040ULL,
    0xfac963ae86ee0ULL,
    0x1812784072dd0ULL,
    0x547f419225a08ULL,
    0xa9ba6f514fb1cULL,
    0x859f3808b6786ULL,
    0xda09b43b1a6fdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0x8c00000000000ULL,
    0xe200000000000ULL,
    0x1900000000000ULL,
    0x8380000000000ULL,
    0x6140000000000ULL,
    0x2560000000000ULL,
    0x9d90000000000ULL,
    0x37b8000000000ULL,
    0xf0d4000000000ULL,
    0xb0da000000000ULL,
    0xd441000000000ULL,
    0x74e1800000000ULL,
    0x2dd6400000000ULL,
    0x6d5fe00000000ULL,
    0x8600500000000ULL,
    0x3706a80000000ULL,
    0xd8860c0000000ULL,
    0x8bc0420000000ULL,
    0xafa6770000000ULL,
    0x55b3868000000ULL,
    0x6d4f86c000000ULL,
    0x43ffa22000000ULL,
    0xf436271000000ULL,
    0xf50d2ea800000ULL,
    0x351d8adc00000ULL,
    0x10e5e03a00000ULL,
    0x3d1503300000ULL,
    0x365f281080000ULL,
    0x6c844c1a40000ULL,
    0xbdc5a23ae0000ULL,
    0x10a7273b50000ULL,
    0xd134aeb1d8000ULL,
    0x1c8fcae124000ULL,
    0x1158000d0e000ULL,
    0x4c04003465000ULL,
    0x202000d6e800ULL,
    0xc905000e74400ULL,
    0x9b83800ad6600ULL,
    0x8543401440900ULL,
    0x4b64600fe1780ULL,
    0x66971012d0540ULL,
    0xad38c83019ca0ULL,
    0x12101c2504130ULL,
    0xf4f90a30ffe88ULL,
    0x6cb02b2d0d89cULL,
    0xdecd6c95434baULL,
    0xeabbfdd14762bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x7800000000000ULL,
    0xbc00000000000ULL,
    0xc200000000000ULL,
    0x4900000000000ULL,
    0x1b80000000000ULL,
    0x12c0000000000ULL,
    0xe20000000000ULL,
    0xa870000000000ULL,
    0x1ff8000000000ULL,
    0x3eb4000000000ULL,
    0x4fde000000000ULL,
    0x4dc3000000000ULL,
    0xcaa3800000000ULL,
    0xbe34c00000000ULL,
    0xe798600000000ULL,
    0xdea1d00000000ULL,
    0xc831a80000000ULL,
    0x4899cc0000000ULL,
    0x42235e0000000ULL,
    0x32756d0000000ULL,
    0xbaff358000000ULL,
    0x2f371ec000000ULL,
    0xd01b3e2000000ULL,
    0x2664bd3000000ULL,
    0xfb169db800000ULL,
    0xfd6ad2ec00000ULL,
    0x781e603e00000ULL,
    0x9266d01f00000ULL,
    0x1d14280880000ULL,
    0x7a6a0c3e40000ULL,
    0x909ebe18e0000ULL,
    0x6e277d2bb0000ULL,
    0xd8737db308000ULL,
    0x87fcc2f85c000ULL,
    0x32b428211e000ULL,
    0xf5da0c3b1d000ULL,
    0xb8c6be287f800ULL,
    0x13237d156cc00ULL,
    0xe5f57dab56e00ULL,
    0xf23bc2cf20300ULL,
    0x6411a83a91980ULL,
    0xd9e9cc0698b40ULL,
    0xde5b5e3e448a0ULL,
    0x12016d151b430ULL,
    0x410135b4e6cf8ULL,
    0x3f841efe292f4ULL,
    0xdcc0be0bcda76ULL,
    0xfd247d3406b4bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x1800000000000ULL,
    0x3400000000000ULL,
    0x8a00000000000ULL,
    0xd300000000000ULL,
    0x9080000000000ULL,
    0x78c0000000000ULL,
    0xa6e0000000000ULL,
    0x3130000000000ULL,
    0x1528000000000ULL,
    0xc5f4000000000ULL,
    0xddce000000000ULL,
    0x95c5000000000ULL,
    0xaf66800000000ULL,
    0x2a71c00000000ULL,
    0x4b89200000000ULL,
    0x32e0900000000ULL,
    0x6b31380000000ULL,
    0xfe29d40000000ULL,
    0x7171de0000000ULL,
    0x370f730000000ULL,
    0xd426ef8000000ULL,
    0x84d74e4000000ULL,
    0x9498fe2000000ULL,
    0xb81fe33000000ULL,
    0xa6dfd78800000ULL,
    0x4d3a9a4400000ULL,
    0xa58f202600000ULL,
    0x93e1903d00000ULL,
    0x2cb1b82a80000ULL,
    0x94691430c0000ULL,
    0x50507e2220000ULL,
    0x8e5b231330000ULL,
    0xd1f8778b38000ULL,
    0xb16fca788c000ULL,
    0x55d138016a000ULL,
    0x3b19d41f4d000ULL,
    0xce59de16cb800ULL,
    0x31fb732d3d400ULL,
    0x8168efa893a00ULL,
    0x5dd24e76e1700ULL,
    0x171e7e2711c80ULL,
    0x705e232d09640ULL,
    0x68fef79d35ee0ULL,
    0xc2ee0a6f26050ULL,
    0xb5901801063f8ULL,
    0xc93d443037f8cULL,
    0xe78ee6000ff5eULL,
    0x4ce3a71f02a69ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x8800000000000ULL,
    0x1c00000000000ULL,
    0x4a00000000000ULL,
    0x4d00000000000ULL,
    0xfc80000000000ULL,
    0x2240000000000ULL,
    0x8b20000000000ULL,
    0x1590000000000ULL,
    0x5cf8000000000ULL,
    0xadd4000000000ULL,
    0x62da000000000ULL,
    0xb0c7000000000ULL,
    0x7a63800000000ULL,
    0x1937400000000ULL,
    0xa0a9600000000ULL,
    0x61ca100000000ULL,
    0x937c880000000ULL,
    0x7810b40000000ULL,
    0xe83bfa0000000ULL,
    0x19b17d0000000ULL,
    0xb8ee9e8000000ULL,
    0x4fef4ec000000ULL,
    0x566a9a2000000ULL,
    0xc4af6d1000000ULL,
    0x9fc8169800000ULL,
    0xd878facc00000ULL,
    0x3b92e00200000ULL,
    0xeff9501700000ULL,
    0x1a57e80a80000ULL,
    0x319a41f40000ULL,
    0x3c26f21d20000ULL,
    0x6c15890f90000ULL,
    0xfe3d84b048000ULL,
    0x7eb063d624000ULL,
    0x39686c881e000ULL,
    0x192ac7f3e5000ULL,
    0xe30f1e827e800ULL,
    0x101b0ef555c00ULL,
    0x9ba27a13a6e00ULL,
    0xd5d13d3638d00ULL,
    0xd6dc7e981cd80ULL,
    0xb6c61eec03440ULL,
    0xc5657215e7c20ULL,
    0x3cb2c91499fd0ULL,
    0x286ce4854ca68ULL,
    0xcfae73ccaadb4ULL,
    0xbc4ee4823285aULL,
    0xefbd73fc7efebULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x1800000000000ULL,
    0x2c00000000000ULL,
    0xbe00000000000ULL,
    0xc700000000000ULL,
    0xe280000000000ULL,
    0x1b40000000000ULL,
    0x3be0000000000ULL,
    0xe8d0000000000ULL,
    0x778000000000ULL,
    0x3194000000000ULL,
    0x1d9a000000000ULL,
    0xf0c5000000000ULL,
    0x8d24800000000ULL,
    0x3ef1c00000000ULL,
    0x8d0fe00000000ULL,
    0x84d9300000000ULL,
    0xca21580000000ULL,
    0x9c71e40000000ULL,
    0xb649620000000ULL,
    0x6f3c8b0000000ULL,
    0x3af03c8000000ULL,
    0xb70e74c000000ULL,
    0x39de822000000ULL,
    0xb5a1bb1000000ULL,
    0x28b3648800000ULL,
    0x216e90f400000ULL,
    0x3cc9602600000ULL,
    0xd079f01b00000ULL,
    0xab10382780000ULL,
    0xdadc1405c0000ULL,
    0xbd235a3ea0000ULL,
    0x96f59f0dd0000ULL,
    0x490fe6a178000ULL,
    0xdede2bcbf4000ULL,
    0x872484997e000ULL,
    0x2bf3a0dab5000ULL,
    0x368a38219e800ULL,
    0x6a19142205400ULL,
    0x1007da2c97200ULL,
    0x78045f3cd9700ULL,
    0xdc00068da5780ULL,
    0x76071be7c70c0ULL,
    0xf305dca6e1760ULL,
    0x708244cfb0090ULL,
    0x62435a1089a08ULL,
    0x1e659f221d6ecULL,
    0x1117e6bd12592ULL,
    0x27da2bc38ee43ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x9800000000000ULL,
    0xd400000000000ULL,
    0x600000000000ULL,
    0x9100000000000ULL,
    0xb380000000000ULL,
    0xd7c0000000000ULL,
    0x2b60000000000ULL,
    0xef90000000000ULL,
    0x5f18000000000ULL,
    0x3a54000000000ULL,
    0x2b7a000000000ULL,
    0xab43000000000ULL,
    0xaa25800000000ULL,
    0xb6b3c00000000ULL,
    0xb529200000000ULL,
    0x5e38f00000000ULL,
    0xa464280000000ULL,
    0x5910ac0000000ULL,
    0xdd5ab60000000ULL,
    0xa7f3870000000ULL,
    0xc409a28000000ULL,
    0x180e33c000000ULL,
    0xe20b962000000ULL,
    0xd90f771000000ULL,
    0xd98f8ab800000ULL,
    0xb2c99ffc00000ULL,
    0xb8eea00600000ULL,
    0x1e5c302500000ULL,
    0xc372883980000ULL,
    0xd1489c1840000ULL,
    0xa72a3e0ae0000ULL,
    0x693c1b00f0000ULL,
    0x96e41c8b58000ULL,
    0x5556e8dfa4000ULL,
    0x6df92a9b26000ULL,
    0xb101afeb65000ULL,
    0xe386281806800ULL,
    0x5fc7ac3d34c00ULL,
    0x976536234f600ULL,
    0xa593472a39f00ULL,
    0x1c1d028014c80ULL,
    0x1ed103ec1efc0ULL,
    0xde3c9e0ef06a0ULL,
    0xe4642b3a58db0ULL,
    0xb914948424e58ULL,
    0x2d5974cc66ddcULL,
    0x3ff494830562aULL,
    0x100974c036a7fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0xfc00000000000ULL,
    0xfe00000000000ULL,
    0x8100000000000ULL,
    0x3e80000000000ULL,
    0x30c0000000000ULL,
    0x1360000000000ULL,
    0x8a30000000000ULL,
    0xfec8000000000ULL,
    0xf0f4000000000ULL,
    0x2eaa000000000ULL,
    0xdb43000000000ULL,
    0xaf25800000000ULL,
    0xfe91400000000ULL,
    0xd418600000000ULL,
    0xbc4db00000000ULL,
    0x7e35e80000000ULL,
    0x5ccb040000000ULL,
    0x9ff10e0000000ULL,
    0xa92aaf0000000ULL,
    0xe107778000000ULL,
    0x8e82fd4000000ULL,
    0x18c16e2000000ULL,
    0x1f631f1000000ULL,
    0xbc309fb800000ULL,
    0x83cef97c00000ULL,
    0x3077e01200000ULL,
    0x9f6ff02f00000ULL,
    0xf6a0080780000ULL,
    0x15d3f41c40000ULL,
    0x133e863da0000ULL,
    0xaedb1b3330000ULL,
    0x6c2c11bb58000ULL,
    0x55811642cc000ULL,
    0xdd42f79012000ULL,
    0xba23bd600d000ULL,
    0xd2110e1772800ULL,
    0xa3daaf385cc00ULL,
    0x24af779e7b600ULL,
    0xf846fd4099500ULL,
    0xfea36e132c980ULL,
    0x49d41f1683ac0ULL,
    0xfd3f1fbc641a0ULL,
    0x17dcb978e6910ULL,
    0x66aa000b825b8ULL,
    0x67430004e5c7cULL,
    0xb125801b26a7eULL,
    0x8f914036637e5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0xc400000000000ULL,
    0xf200000000000ULL,
    0x7300000000000ULL,
    0x2480000000000ULL,
    0x1dc0000000000ULL,
    0xe760000000000ULL,
    0x2d70000000000ULL,
    0x178000000000ULL,
    0x5eb4000000000ULL,
    0x4b1a000000000ULL,
    0x2243000000000ULL,
    0x6a24800000000ULL,
    0x5dd5400000000ULL,
    0xe468a00000000ULL,
    0x2839d00000000ULL,
    0x4c13b80000000ULL,
    0x750bb40000000ULL,
    0x644ef60000000ULL,
    0xcaecdf0000000ULL,
    0x5afce58000000ULL,
    0x11f100c000000ULL,
    0xd0be562000000ULL,
    0x6fd10f1000000ULL,
    0x376d5db800000ULL,
    0x3cbdb4f400000ULL,
    0xc9d6203200000ULL,
    0x1e6f902100000ULL,
    0x7f3f980480000ULL,
    0x4a972428c0000ULL,
    0x53cdee1b20000ULL,
    0x63aabb3670000ULL,
    0x85bab8558000ULL,
    0xcfe56bd79c000ULL,
    0x46b0c5a97e000ULL,
    0x871d90c0dd000ULL,
    0xf4454e331e800ULL,
    0x3b236b27ccc00ULL,
    0x425013b0d7200ULL,
    0x192adfc5d8500ULL,
    0x209c33a55ca80ULL,
    0xf5064fff5eb40ULL,
    0x7d872ba42dce0ULL,
    0x26442bf587bd0ULL,
    0x3822659af1958ULL,
    0x1ed740e19943cULL,
    0x58ea760abdd76ULL,
    0xb9f99f0c0fad3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x6800000000000ULL,
    0x2c00000000000ULL,
    0x8e00000000000ULL,
    0x700000000000ULL,
    0x1680000000000ULL,
    0x93c0000000000ULL,
    0x10a0000000000ULL,
    0x9670000000000ULL,
    0xadd8000000000ULL,
    0xffb4000000000ULL,
    0x607e000000000ULL,
    0xd247000000000ULL,
    0x3de0800000000ULL,
    0x5910400000000ULL,
    0x750fe00000000ULL,
    0xab1b300000000ULL,
    0x893580000000ULL,
    0x5049c40000000ULL,
    0x987d260000000ULL,
    0x3647f10000000ULL,
    0x2fe70d8000000ULL,
    0x68157bc000000ULL,
    0x208ac62000000ULL,
    0x8c58c13000000ULL,
    0x2f2559800000ULL,
    0xc71fbff400000ULL,
    0x2691603a00000ULL,
    0xe74c703b00000ULL,
    0x36fc383b80000ULL,
    0xe182b435c0000ULL,
    0x9d419e1fa0000ULL,
    0x7765052ff0000ULL,
    0x9cd173a7a8000ULL,
    0xf6af4ee45c000ULL,
    0xe1eeed8ed6000ULL,
    0x938d4beb1d000ULL,
    0x20df1e2437800ULL,
    0xec3245150dc00ULL,
    0xdc3e13a60e200ULL,
    0xf8a43ed6a9100ULL,
    0x7a72559cdc780ULL,
    0x43dfbfe4970c0ULL,
    0x28b160137cf60ULL,
    0x1efc702366610ULL,
    0x6d843838d4b18ULL,
    0xa346b4193b304ULL,
    0xc8679e1f03156ULL,
    0xce56051ea62ffULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xb800000000000ULL,
    0x7400000000000ULL,
    0xae00000000000ULL,
    0x5900000000000ULL,
    0xfa80000000000ULL,
    0x11c0000000000ULL,
    0xdbe0000000000ULL,
    0x45f0000000000ULL,
    0x1578000000000ULL,
    0xd934000000000ULL,
    0xa1a000000000ULL,
    0x4c05000000000ULL,
    0x9a04800000000ULL,
    0xd704c00000000ULL,
    0x7385200000000ULL,
    0x5340b00000000ULL,
    0xbe21680000000ULL,
    0x3012a40000000ULL,
    0x98a460000000ULL,
    0x36cd0b0000000ULL,
    0xc2eca18000000ULL,
    0x9df974c000000ULL,
    0x93f5662000000ULL,
    0x5878bb1000000ULL,
    0x7db1498800000ULL,
    0x2adb10f400000ULL,
    0xa160000e00000ULL,
    0x1430000d00000ULL,
    0xee98002380000ULL,
    0x4cc4002240000ULL,
    0xa7620010a0000ULL,
    0xe131001970000ULL,
    0x3e1e801d78000ULL,
    0xc201c0073c000ULL,
    0x1301a03bfe000ULL,
    0x958470323d000ULL,
    0x164448347e800ULL,
    0x26a214027d400ULL,
    0xa2d32e23df200ULL,
    0xdfebaf038c300ULL,
    0xc17ce79e67c80ULL,
    0xe7317fc7d16c0ULL,
    0xcb1d47a9097a0ULL,
    0x12850fe9c5990ULL,
    0x9dc10fb683d98ULL,
    0x21e31bfd636ecULL,
    0x62f0219933726ULL,
    0xef9b4d27af43ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0x1c00000000000ULL,
    0x1a00000000000ULL,
    0xb700000000000ULL,
    0x1a80000000000ULL,
    0xedc0000000000ULL,
    0xeaa0000000000ULL,
    0xf970000000000ULL,
    0xd398000000000ULL,
    0x4fb4000000000ULL,
    0x11ba000000000ULL,
    0x1003000000000ULL,
    0x5801800000000ULL,
    0x9404400000000ULL,
    0x7603600000000ULL,
    0xc503900000000ULL,
    0xb185580000000ULL,
    0xed40040000000ULL,
    0xb062d60000000ULL,
    0x957810000000ULL,
    0xc728a58000000ULL,
    0x768dc9c000000ULL,
    0xa77bb62000000ULL,
    0xd227111000000ULL,
    0x7b47db800000ULL,
    0xddbd8ddc00000ULL,
    0xf200003a00000ULL,
    0xeb00001700000ULL,
    0xe080003e80000ULL,
    0x2ac00031c0000ULL,
    0x9820001ca0000ULL,
    0x8b0003c70000ULL,
    0x2338003c28000ULL,
    0x1c400139c000ULL,
    0xd8a2003246000ULL,
    0xb27700289d000ULL,
    0xa31b803ec6800ULL,
    0x7d77403a5cc00ULL,
    0xfd9ae022e6600ULL,
    0x1eb3d036ec100ULL,
    0xd63c3819ee580ULL,
    0x3840943540240ULL,
    0x59e60e3a61360ULL,
    0x7013c51e51110ULL,
    0x14913b1986d8ULL,
    0xbad9d8f315044ULL,
    0xd1d64bbdab7f6ULL,
    0x49eadcf6f3737ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0x400000000000ULL,
    0x3600000000000ULL,
    0x2f00000000000ULL,
    0x1e80000000000ULL,
    0x540000000000ULL,
    0x72e0000000000ULL,
    0x3c10000000000ULL,
    0x2218000000000ULL,
    0xaa54000000000ULL,
    0xbc7e000000000ULL,
    0x6a03000000000ULL,
    0x6d03800000000ULL,
    0xdf86c00000000ULL,
    0x80c5a00000000ULL,
    0x6720700000000ULL,
    0x60b1580000000ULL,
    0x4469440000000ULL,
    0x9e1d520000000ULL,
    0x5851630000000ULL,
    0x4578ca8000000ULL,
    0xeb8407c000000ULL,
    0x4ec6f22000000ULL,
    0xac22133000000ULL,
    0xf83212b800000ULL,
    0x562f83ec00000ULL,
    0xf678002e00000ULL,
    0x5704003100000ULL,
    0x3a86003580000ULL,
    0x63470027c0000ULL,
    0xd5e58009a0000ULL,
    0x1e91c00050000ULL,
    0x1558201138000ULL,
    0xc1b5b004c4000ULL,
    0xb1ef780f26000ULL,
    0x53dbf42bc5000ULL,
    0xb0f7aa33a7800ULL,
    0x2d0b571584c00ULL,
    0xf4cf4093c6e00ULL,
    0x826ee0dd74b00ULL,
    0xa918caaf2ee80ULL,
    0x12d407f348dc0ULL,
    0xfe3ef21b6cb60ULL,
    0x8a66133afbe10ULL,
    0x5e5412b1b63c8ULL,
    0x927883e2dc44cULL,
    0x1105802af264aULL,
    0xcd81c0113b50fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x800000000000ULL,
    0x6c00000000000ULL,
    0x200000000000ULL,
    0x500000000000ULL,
    0x1f80000000000ULL,
    0x2740000000000ULL,
    0xa220000000000ULL,
    0x4190000000000ULL,
    0xb828000000000ULL,
    0x41d4000000000ULL,
    0x908e000000000ULL,
    0x9081000000000ULL,
    0x91c4800000000ULL,
    0x38e5400000000ULL,
    0x91f1e00000000ULL,
    0x4c18500000000ULL,
    0xf32f480000000ULL,
    0xc952ec0000000ULL,
    0x54a120000000ULL,
    0x36666d0000000ULL,
    0x5f37f68000000ULL,
    0xa7ef04000000ULL,
    0x341df22000000ULL,
    0x72b3d3000000ULL,
    0xf523ea800000ULL,
    0x3e485c5400000ULL,
    0x26e2802200000ULL,
    0xef0402b00000ULL,
    0x929b602880000ULL,
    0xf5ec101540000ULL,
    0x43322805e0000ULL,
    0x907bfc12d0000ULL,
    0xf51aba2808000ULL,
    0x26add11124000ULL,
    0x27162ca9ea000ULL,
    0xeaef3159a5000ULL,
    0x5db1f6acab800ULL,
    0x36bbf04444400ULL,
    0x2f7f72027b200ULL,
    0xbb9b7d3a4c500ULL,
    0x88695eaa5ff80ULL,
    0x91744c7826540ULL,
    0x75d8a83a3af20ULL,
    0xf7cfbc162deb0ULL,
    0x52a7da0a4efc8ULL,
    0x2054c1020f8f4ULL,
    0xd8ce84866e2faULL,
    0xb4208d727f071ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0x8400000000000ULL,
    0x2200000000000ULL,
    0x8500000000000ULL,
    0xd480000000000ULL,
    0xc740000000000ULL,
    0xd720000000000ULL,
    0x990000000000ULL,
    0x3b8000000000ULL,
    0x51d4000000000ULL,
    0xb11e000000000ULL,
    0x8183000000000ULL,
    0xabc2800000000ULL,
    0xe463c00000000ULL,
    0xf4b0a00000000ULL,
    0x932dd00000000ULL,
    0x20ec080000000ULL,
    0x808db40000000ULL,
    0xb63f120000000ULL,
    0x30119d0000000ULL,
    0x5c7f368000000ULL,
    0x9fb1c4c000000ULL,
    0xdca9b22000000ULL,
    0x602b4d3000000ULL,
    0xeb6fbeb800000ULL,
    0xe24cb0e400000ULL,
    0x58dc800600000ULL,
    0xb0e0c01100000ULL,
    0xf3f2203080000ULL,
    0xa40e100540000ULL,
    0xb97ca81320000ULL,
    0x1b306410d0000ULL,
    0x63eb1a3d48000ULL,
    0xa308292324000ULL,
    0x58fe24b5ce000ULL,
    0xe0f359e5a5000ULL,
    0xff8c04b98f800ULL,
    0x43bd49d204c00ULL,
    0x71d0acb21ea00ULL,
    0xc11d2df96df00ULL,
    0x8983b6af6ba80ULL,
    0xb7c104f1abb40ULL,
    0x426392144ba20ULL,
    0x53b15d0d7b9d0ULL,
    0xc2ad16b303ec8ULL,
    0x332fd4fd5f134ULL,
    0x90ed1a114ccfaULL,
    0x688f29043b1c3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0x9400000000000ULL,
    0x200000000000ULL,
    0xd700000000000ULL,
    0xfe80000000000ULL,
    0xfac0000000000ULL,
    0x7a0000000000ULL,
    0x9310000000000ULL,
    0xeab8000000000ULL,
    0xe4d4000000000ULL,
    0x989e000000000ULL,
    0xf587000000000ULL,
    0x4245800000000ULL,
    0x5060400000000ULL,
    0x3137200000000ULL,
    0x83eb500000000ULL,
    0xb20e180000000ULL,
    0xc37de40000000ULL,
    0xc7f06e0000000ULL,
    0xba49490000000ULL,
    0x1c8e8000000ULL,
    0x724108c000000ULL,
    0x98614e2000000ULL,
    0xb531193000000ULL,
    0x99e9169800000ULL,
    0x990bacdc00000ULL,
    0xabfb803a00000ULL,
    0xe837401500000ULL,
    0x946aa01880000ULL,
    0x974f1029c0000ULL,
    0x659f3825a0000ULL,
    0xe805b41bb0000ULL,
    0x5405f60168000ULL,
    0x6203ed1104000ULL,
    0xa70640850e000ULL,
    0x968751c785000ULL,
    0x6ec2f887cf800ULL,
    0x5a5a5e9a5c00ULL,
    0x4415ae923f600ULL,
    0x143d58c12d100ULL,
    0x1e12d6086a480ULL,
    0x9f38bd149b140ULL,
    0x669058be92e60ULL,
    0xa8feb5e807690ULL,
    0xb4b496b891d38ULL,
    0xa9afecd018864ULL,
    0x766aa02a4ca5aULL,
    0xf04f102ffb3b3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xe800000000000ULL,
    0x4c00000000000ULL,
    0x3e00000000000ULL,
    0x4f00000000000ULL,
    0x6380000000000ULL,
    0x8ac0000000000ULL,
    0x9720000000000ULL,
    0x6430000000000ULL,
    0xbe48000000000ULL,
    0x3f4000000000ULL,
    0x73ea000000000ULL,
    0x7d83000000000ULL,
    0x75c2800000000ULL,
    0xeca0400000000ULL,
    0x5af2e00000000ULL,
    0xb36eb00000000ULL,
    0x5ac4180000000ULL,
    0xdf20ec0000000ULL,
    0xd8312e0000000ULL,
    0x784f250000000ULL,
    0xe8f4378000000ULL,
    0x62692b4000000ULL,
    0x8641ce2000000ULL,
    0xce66951000000ULL,
    0x61d0afb800000ULL,
    0xf95e874400000ULL,
    0x438a801a00000ULL,
    0xf354400300000ULL,
    0x1f18e03780000ULL,
    0xddedb017c0000ULL,
    0x9a869822e0000ULL,
    0xba40ac31b0000ULL,
    0x4863ce2a48000ULL,
    0x6ad1950acc000ULL,
    0xf8d82fb772000ULL,
    0x504dc7624d000ULL,
    0x4f2e01932800ULL,
    0xac6eb0166cc00ULL,
    0x3144180ae2a00ULL,
    0x9e0ec3fd5100ULL,
    0x99112e3046380ULL,
    0x1f7f2520bb6c0ULL,
    0xb3c378441a60ULL,
    0xa45d2b5f202b0ULL,
    0x10bce1850f38ULL,
    0x5d15950078654ULL,
    0x3d7a2fb1440beULL,
    0x723ac7725e71dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x7c00000000000ULL,
    0xde00000000000ULL,
    0x7f00000000000ULL,
    0xa680000000000ULL,
    0x6b40000000000ULL,
    0xe8e0000000000ULL,
    0xf550000000000ULL,
    0xa918000000000ULL,
    0x2b14000000000ULL,
    0xda7e000000000ULL,
    0x8285000000000ULL,
    0x1146800000000ULL,
    0x2de0400000000ULL,
    0xb6d3600000000ULL,
    0x91dfb00000000ULL,
    0xd334180000000ULL,
    0xab0dc40000000ULL,
    0x84eda20000000ULL,
    0x4338f30000000ULL,
    0x366b08000000ULL,
    0x71966fc000000ULL,
    0xaf38c22000000ULL,
    0x3566433000000ULL,
    0x5292288800000ULL,
    0xc7baebc400000ULL,
    0x926802a00000ULL,
    0x60f0402f00000ULL,
    0x87ab603f80000ULL,
    0x37dbb01bc0000ULL,
    0x18321823a0000ULL,
    0x9f8cc405d0000ULL,
    0x76ad220db8000ULL,
    0x2659b32294000ULL,
    0x99f55083e6000ULL,
    0xfa289fd015000ULL,
    0x401f3a2ca7800ULL,
    0x7e95772df5400ULL,
    0xd1b872a617a00ULL,
    0x3a212cc5bd100ULL,
    0x20726ab12b580ULL,
    0xc7e9e8ebd6ac0ULL,
    0xdbb948af1ca60ULL,
    0x27215bca7b610ULL,
    0x7f498072fb08ULL,
    0x652c840569d0cULL,
    0xd69e423bacaa2ULL,
    0x2dd6030f32eafULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0x1400000000000ULL,
    0xf600000000000ULL,
    0xbf00000000000ULL,
    0x1780000000000ULL,
    0xf840000000000ULL,
    0x21a0000000000ULL,
    0xfd30000000000ULL,
    0x41f8000000000ULL,
    0x92f4000000000ULL,
    0xd81a000000000ULL,
    0xae63000000000ULL,
    0x34d0800000000ULL,
    0x556e400000000ULL,
    0x6a3e200000000ULL,
    0xd510f00000000ULL,
    0x4f88580000000ULL,
    0xa32e8c0000000ULL,
    0x889ffa0000000ULL,
    0x3da1510000000ULL,
    0xef30bb8000000ULL,
    0x90fd8ac000000ULL,
    0x56715a2000000ULL,
    0xfddfe11000000ULL,
    0x2a86c3b800000ULL,
    0xb6c3f6d400000ULL,
    0xb966f82a00000ULL,
    0x77503c1500000ULL,
    0xbb29820580000ULL,
    0xa49f2d3bc0000ULL,
    0x17a7198fe0000ULL,
    0xe23257fb10000ULL,
    0xe67e1b95e8000ULL,
    0x12b33ac08c000ULL,
    0xc5bf220d9e000ULL,
    0xb1559d1ead000ULL,
    0x3c2b618c6e800ULL,
    0x6f1f2bc5d4c00ULL,
    0x5de1398aca200ULL,
    0xa296a7cb26900ULL,
    0xa7cc438ce9080ULL,
    0x4a8eb6cb0cfc0ULL,
    0x31a858015e360ULL,
    0xa25e8c3dcc330ULL,
    0xd6c7fa2f3e768ULL,
    0xa965511f1d684ULL,
    0xbf52bbbc36b8eULL,
    0xff2a8ad70101bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x800000000000ULL,
    0x2400000000000ULL,
    0xae00000000000ULL,
    0x500000000000ULL,
    0x1780000000000ULL,
    0x73c0000000000ULL,
    0x6c20000000000ULL,
    0x7150000000000ULL,
    0xe5e8000000000ULL,
    0x3d14000000000ULL,
    0x6e0a000000000ULL,
    0x2661000000000ULL,
    0x2bb5800000000ULL,
    0x4f1d400000000ULL,
    0x2a6aa00000000ULL,
    0xd2d3300000000ULL,
    0x202e080000000ULL,
    0x28334c0000000ULL,
    0x7ade760000000ULL,
    0x374bbd0000000ULL,
    0xe205c48000000ULL,
    0x9f06564000000ULL,
    0xe481562000000ULL,
    0xbd45cd1000000ULL,
    0x8a616ca800000ULL,
    0xe1b62a7c00000ULL,
    0x4419282200000ULL,
    0xc8e93c1900000ULL,
    0xbe90de0380000ULL,
    0x1bcac13d40000ULL,
    0x9b403a87e0000ULL,
    0xeb63e755f0000ULL,
    0xf830449088000ULL,
    0xc2db164d14000ULL,
    0xbb4bf6349a000ULL,
    0x9806fd3fb5000ULL,
    0x3c0764aa8a800ULL,
    0xd2016660cc400ULL,
    0x77055e3a17600ULL,
    0x90878104c2500ULL,
    0xeb429a9778280ULL,
    0xa364d75c9c8c0ULL,
    0x7c344cbe0ee20ULL,
    0x9cd95a406f830ULL,
    0xb6480015df358ULL,
    0xab84000737234ULL,
    0xe1c2003d92732ULL,
    0xbb250002e0069ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0xfc00000000000ULL,
    0xf600000000000ULL,
    0x5300000000000ULL,
    0xff80000000000ULL,
    0xc240000000000ULL,
    0x7e0000000000ULL,
    0x5050000000000ULL,
    0x8dc8000000000ULL,
    0xd194000000000ULL,
    0x5e6e000000000ULL,
    0xc525000000000ULL,
    0xf477800000000ULL,
    0x1f3cc00000000ULL,
    0x766a600000000ULL,
    0x7120700000000ULL,
    0x6672780000000ULL,
    0x7638cc0000000ULL,
    0x94ecca0000000ULL,
    0xe3e7e30000000ULL,
    0xaa53c08000000ULL,
    0xb0c9f94000000ULL,
    0xb1112a2000000ULL,
    0x2dab533000000ULL,
    0xaf83d88800000ULL,
    0xaa45454c00000ULL,
    0x53e1980e00000ULL,
    0xd2517c0f00000ULL,
    0x6ccb523580000ULL,
    0x77179f18c0000ULL,
    0xc6a91291e0000ULL,
    0xac03a64f90000ULL,
    0x9e03d89c78000ULL,
    0x7054570d4000ULL,
    0x7d81983492000ULL,
    0x23417c38f5000ULL,
    0xa163523063800ULL,
    0xc8939f165d400ULL,
    0xe4ef128defe00ULL,
    0xbbe2a6586a300ULL,
    0x465258a861180ULL,
    0xcec88579515c0ULL,
    0xa6127832997e0ULL,
    0xd828cc02c5030ULL,
    0xc8c4ca3c2baa8ULL,
    0x123e337c0e4cULL,
    0x3e75c09461ee2ULL,
    0x9a38f95f500d5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xd800000000000ULL,
    0x6400000000000ULL,
    0x3e00000000000ULL,
    0x4100000000000ULL,
    0xdc80000000000ULL,
    0x8440000000000ULL,
    0xe2e0000000000ULL,
    0x7650000000000ULL,
    0xa18000000000ULL,
    0xaa94000000000ULL,
    0xe2be000000000ULL,
    0xc727000000000ULL,
    0x5675800000000ULL,
    0x5edc00000000ULL,
    0xd9bca00000000ULL,
    0xb4a5300000000ULL,
    0x87b4280000000ULL,
    0x73cf940000000ULL,
    0x4b4f6a0000000ULL,
    0x310e130000000ULL,
    0xf2ec408000000ULL,
    0x283e3fc000000ULL,
    0x7e664a2000000ULL,
    0x5216e33000000ULL,
    0xf8fcc89800000ULL,
    0x74c09bdc00000ULL,
    0x54a3081600000ULL,
    0x37b0642900000ULL,
    0xcbca621780000ULL,
    0x5f4d770c40000ULL,
    0xd70da28120000ULL,
    0xd7e988f810000ULL,
    0xcabac89738000ULL,
    0xbb239bfdd4000ULL,
    0x6c70880da6000ULL,
    0xaeea427b5000ULL,
    0x9c3d422e17800ULL,
    0x886287295dc00ULL,
    0xdf10aaae9b600ULL,
    0xa67eece29e700ULL,
    0xeb852a9660a80ULL,
    0x15c32cddb08c0ULL,
    0x58218a83e86a0ULL,
    0x7bf21ccd0c950ULL,
    0xe52ba29082f28ULL,
    0xab1a88e35bc8cULL,
    0xc61148a6ff9a2ULL,
    0xdef95be8d0aafULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x4c00000000000ULL,
    0xe00000000000ULL,
    0x500000000000ULL,
    0xc680000000000ULL,
    0xcac0000000000ULL,
    0x5e60000000000ULL,
    0xb430000000000ULL,
    0x83c8000000000ULL,
    0x474000000000ULL,
    0x316e000000000ULL,
    0x4f23000000000ULL,
    0x8a14800000000ULL,
    0x1d5dc00000000ULL,
    0x60ece00000000ULL,
    0xd060b00000000ULL,
    0x7133480000000ULL,
    0xa54bc40000000ULL,
    0x9eb23e0000000ULL,
    0x370a670000000ULL,
    0xb712e78000000ULL,
    0x7da354000000ULL,
    0x1c2a5e2000000ULL,
    0x9707173000000ULL,
    0x55854fb800000ULL,
    0xa545415400000ULL,
    0xc25283600000ULL,
    0x7d95b42300000ULL,
    0xe319163b80000ULL,
    0xf74cd31540000ULL,
    0xedb771a7a0000ULL,
    0x88f2661b0000ULL,
    0xbd57cfb418000ULL,
    0x827f815c4c000ULL,
    0x72fb480952000ULL,
    0x613fc417ad000ULL,
    0x4fdc3e35c3800ULL,
    0x28296739c4c00ULL,
    0x65066797f7200ULL,
    0x5687f5540a700ULL,
    0x72c6be0640b80ULL,
    0x4267a72420ec0ULL,
    0xe23607a0f1f20ULL,
    0xcace8565b9810ULL,
    0xccf7163efdb78ULL,
    0xfeafd336cbb5cULL,
    0xd7c3f1a6a1ebeULL,
    0xf4e2e66f31bc5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x4800000000000ULL,
    0x5400000000000ULL,
    0x1a00000000000ULL,
    0x700000000000ULL,
    0x8780000000000ULL,
    0x86c0000000000ULL,
    0xdd20000000000ULL,
    0xcab0000000000ULL,
    0xa668000000000ULL,
    0xb7f4000000000ULL,
    0x838e000000000ULL,
    0x565000000000ULL,
    0xa3d0800000000ULL,
    0x13bb400000000ULL,
    0x8548200000000ULL,
    0x1840900000000ULL,
    0x4965980000000ULL,
    0x65d36c0000000ULL,
    0x7ab8ce0000000ULL,
    0xcfce350000000ULL,
    0x5602138000000ULL,
    0xc104324000000ULL,
    0xee806e2000000ULL,
    0xcc45e53000000ULL,
    0x9367ab8800000ULL,
    0x42d3ce7400000ULL,
    0x2d3b383200000ULL,
    0x109bc2500000ULL,
    0xdf23f60e80000ULL,
    0x11b78935c0000ULL,
    0x4fe9e5b3e0000ULL,
    0xfc37bb74b0000ULL,
    0x962f8b91c8000ULL,
    0x9a935e436c000ULL,
    0x445ea0207a000ULL,
    0xb4dad0084d000ULL,
    0xed9b380dab800ULL,
    0x8a79bc12f5400ULL,
    0x3ebf615ee200ULL,
    0x3a33893dd3d00ULL,
    0xff2fe5a6d1880ULL,
    0xd016bb6739640ULL,
    0xa190b8c05460ULL,
    0x3cb91e7e860b0ULL,
    0x66c800092e3b8ULL,
    0x3c84002eb2a94ULL,
    0x5f460014a062eULL,
    0x2ee1002491979ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xf800000000000ULL,
    0x3400000000000ULL,
    0x5e00000000000ULL,
    0x8300000000000ULL,
    0x5a80000000000ULL,
    0x7240000000000ULL,
    0xc2e0000000000ULL,
    0xa670000000000ULL,
    0xcfe8000000000ULL,
    0x10b4000000000ULL,
    0x5c4a000000000ULL,
    0xdea3000000000ULL,
    0xaf92800000000ULL,
    0xf18c00000000ULL,
    0x2f1d200000000ULL,
    0x9f1b300000000ULL,
    0x971cb80000000ULL,
    0x4b19a40000000ULL,
    0xd919ca0000000ULL,
    0x301b230000000ULL,
    0xb79bb18000000ULL,
    0x1c59604000000ULL,
    0xf67e6a2000000ULL,
    0xe0acd31000000ULL,
    0x4bd029b800000ULL,
    0x32f8f44400000ULL,
    0xb1e9981e00000ULL,
    0x23b6941d00000ULL,
    0xecf722f80000ULL,
    0x70e18724c0000ULL,
    0xff70fba8a0000ULL,
    0x406a835190000ULL,
    0x6770fb8738000ULL,
    0x246a83595c000ULL,
    0x2170fbbd5a000ULL,
    0x836a834cbd000ULL,
    0xddf0fb81aa800ULL,
    0x462a834734c00ULL,
    0x1b90fba99ea00ULL,
    0x111a837e2b300ULL,
    0x4c18fbbbf5c80ULL,
    0xd59e83753e0c0ULL,
    0x1d5afb99bb8e0ULL,
    0xbdf9835dcc590ULL,
    0x216a7b99e33a8ULL,
    0x7cf643632104cULL,
    0xfdafdbbb51426ULL,
    0x3256b36a99f11ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xe800000000000ULL,
    0xdc00000000000ULL,
    0x3200000000000ULL,
    0xcb00000000000ULL,
    0xd180000000000ULL,
    0x16c0000000000ULL,
    0x36a0000000000ULL,
    0x4870000000000ULL,
    0xb1d8000000000ULL,
    0x4d34000000000ULL,
    0x72be000000000ULL,
    0x5e67000000000ULL,
    0xdfd5800000000ULL,
    0xb72c400000000ULL,
    0xa5a9600000000ULL,
    0xeecd00000000ULL,
    0x9489b80000000ULL,
    0x8d5d9c0000000ULL,
    0x7f732e0000000ULL,
    0x8258190000000ULL,
    0x68f6458000000ULL,
    0xd19fd3c000000ULL,
    0x6d7ce2000000ULL,
    0x7dac893000000ULL,
    0x9ae89d9800000ULL,
    0x6a899fdc00000ULL,
    0xec58d81a00000ULL,
    0x51f54c0700000ULL,
    0xab1c961480000ULL,
    0xb196852ec0000ULL,
    0xb5ceeb8e60000ULL,
    0x383c8af2b0000ULL,
    0xa1256ba128000ULL,
    0xdbb7cad0dc000ULL,
    0xbff98ba016000ULL,
    0x91075afafd000ULL,
    0x681539307800ULL,
    0x954216de85c00ULL,
    0x9b65c59503600ULL,
    0x475093d746100ULL,
    0x336d2e0f25d80ULL,
    0xf84f193852f40ULL,
    0xbcfbc5859f0e0ULL,
    0x4c8793c0f0770ULL,
    0xda40ae3588938ULL,
    0x2ce759120c324ULL,
    0x494a5a28e7f6ULL,
    0xc54843ea8853fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xe800000000000ULL,
    0x4c00000000000ULL,
    0x8e00000000000ULL,
    0xd900000000000ULL,
    0xa880000000000ULL,
    0x5240000000000ULL,
    0x3ea0000000000ULL,
    0x3ab0000000000ULL,
    0x44e8000000000ULL,
    0x3274000000000ULL,
    0x70a000000000ULL,
    0x83e7000000000ULL,
    0x9194800000000ULL,
    0x269c400000000ULL,
    0x9d78e00000000ULL,
    0xa8ed300000000ULL,
    0x4c71a80000000ULL,
    0x860eb40000000ULL,
    0x5f629a0000000ULL,
    0xe9d5110000000ULL,
    0x339178000000ULL,
    0x58498a4000000ULL,
    0xcfc6fa2000000ULL,
    0xba60611000000ULL,
    0xd7525f9800000ULL,
    0x9c790e7400000ULL,
    0x346b481a00000ULL,
    0x5430840300000ULL,
    0x73adb23b80000ULL,
    0x7250e50240000ULL,
    0xc2ffed9020000ULL,
    0x7b29eb4790000ULL,
    0xd014a58c28000ULL,
    0xa4596f68ec000ULL,
    0x359917831a000ULL,
    0xbef98a7c0d000ULL,
    0x8d2efa0c6a800ULL,
    0xcd14611955c00ULL,
    0x1ed85fb4df200ULL,
    0xd8de0e547a100ULL,
    0xbddfc8353cb80ULL,
    0xc35cc42b128c0ULL,
    0x3c1d5229714a0ULL,
    0x8039d53058bd0ULL,
    0xf3cc45be349e8ULL,
    0x76045f55ef084ULL,
    0x5d00bf8f02afeULL,
    0xda833e73190f9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x6c00000000000ULL,
    0x2200000000000ULL,
    0xc900000000000ULL,
    0x3080000000000ULL,
    0x4740000000000ULL,
    0x9ea0000000000ULL,
    0xfcb0000000000ULL,
    0x44f8000000000ULL,
    0x7d74000000000ULL,
    0x221a000000000ULL,
    0xd0e7000000000ULL,
    0x7292800000000ULL,
    0x8f0b400000000ULL,
    0xbf2fe00000000ULL,
    0xc7df500000000ULL,
    0xf800280000000ULL,
    0x3401140000000ULL,
    0x2607b20000000ULL,
    0x9706d10000000ULL,
    0xd387ef8000000ULL,
    0xd2c2eac000000ULL,
    0xcb62d22000000ULL,
    0xec56c11000000ULL,
    0x166a279800000ULL,
    0x827faec400000ULL,
    0x8535c82200000ULL,
    0x7339440b00000ULL,
    0xc4951a1080000ULL,
    0x500c853640000ULL,
    0x60afbd8e20000ULL,
    0x271b6bcad0000ULL,
    0xf265158f28000ULL,
    0x84d53fdd6c000ULL,
    0x552f47a51e000ULL,
    0x42dfbeda8d000ULL,
    0x9a8000152e800ULL,
    0xa240000415c00ULL,
    0xec20001f1aa00ULL,
    0x62f00011dfd00ULL,
    0xe2d80021ed780ULL,
    0xaa84003d5e140ULL,
    0xda420001acaa0ULL,
    0x982300327e950ULL,
    0x24f0801cfc148ULL,
    0x65d8402597604ULL,
    0x71076034f171eULL,
    0x6483102df83fbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x1800000000000ULL,
    0xa400000000000ULL,
    0x1a00000000000ULL,
    0xc700000000000ULL,
    0x4680000000000ULL,
    0x2040000000000ULL,
    0xbd60000000000ULL,
    0xc390000000000ULL,
    0xba88000000000ULL,
    0xc254000000000ULL,
    0x312a000000000ULL,
    0xa4e5000000000ULL,
    0x4152800000000ULL,
    0x1da9c00000000ULL,
    0x6ba1a00000000ULL,
    0x16b2100000000ULL,
    0x587a180000000ULL,
    0x154e640000000ULL,
    0x8cf1fa0000000ULL,
    0x7219ff0000000ULL,
    0xa85a4e8000000ULL,
    0xa238474000000ULL,
    0xf28da2000000ULL,
    0x39e72f1000000ULL,
    0xe0d3768800000ULL,
    0x6b6df34400000ULL,
    0xee82982600000ULL,
    0xcc42a43900000ULL,
    0x3b62da0e80000ULL,
    0xaa922f35c0000ULL,
    0x3909f697a0000ULL,
    0xc790336110000ULL,
    0xf0893809d8000ULL,
    0x3d55b41124000ULL,
    0xc3aa423702000ULL,
    0x86a58b3c85000ULL,
    0x9f31aca512800ULL,
    0x82bfdc70dd400ULL,
    0x366b6e9876a00ULL,
    0x2f0297563f700ULL,
    0x8a83e22f02e80ULL,
    0xb6439b3085c40ULL,
    0xcc61b4a712260ULL,
    0xe414b85fdde90ULL,
    0xa54814b7f6168ULL,
    0xc4f2a8617ffbcULL,
    0xee180c8f6295aULL,
    0x659cc6934c8dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x1800000000000ULL,
    0x4400000000000ULL,
    0xee00000000000ULL,
    0x2900000000000ULL,
    0x8780000000000ULL,
    0xe940000000000ULL,
    0x6da0000000000ULL,
    0x2bd0000000000ULL,
    0x44a8000000000ULL,
    0x5414000000000ULL,
    0x33ce000000000ULL,
    0xfd27000000000ULL,
    0xd412800000000ULL,
    0xf3cbc00000000ULL,
    0x9d22a00000000ULL,
    0x4411b00000000ULL,
    0xebc8680000000ULL,
    0xd923cc0000000ULL,
    0xaa10c20000000ULL,
    0xc2ce710000000ULL,
    0x5ea1e48000000ULL,
    0x4356c04000000ULL,
    0xaf6ee22000000ULL,
    0x7573013000000ULL,
    0x7f9ac9800000ULL,
    0xfb7f7c6400000ULL,
    0x46bce82600000ULL,
    0xfadb0c2100000ULL,
    0x2f6ee22380000ULL,
    0xb573012e40000ULL,
    0x67f9aca7e0000ULL,
    0x6b7f7c6b50000ULL,
    0x5ebce800e8000ULL,
    0xbedb0c30b4000ULL,
    0xc16ee228ca000ULL,
    0x9c73010b55000ULL,
    0xe079ac919b800ULL,
    0x823f7c64bdc00ULL,
    0x331ce83faea00ULL,
    0x950b0c13b7f00ULL,
    0x85c6e2125b280ULL,
    0xc86701301dac0ULL,
    0xd3b7ac929eba0ULL,
    0x7f187c514e030ULL,
    0xe70e680766988ULL,
    0x66c0cc0ef2f04ULL,
    0x18e4421849e32ULL,
    0x8c76b10784831ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x7800000000000ULL,
    0xbc00000000000ULL,
    0x600000000000ULL,
    0xc300000000000ULL,
    0x3980000000000ULL,
    0xcd40000000000ULL,
    0x5aa0000000000ULL,
    0x34b0000000000ULL,
    0xbdb8000000000ULL,
    0xcb74000000000ULL,
    0xb7da000000000ULL,
    0x8225000000000ULL,
    0xa971800000000ULL,
    0xfedac00000000ULL,
    0x26a2600000000ULL,
    0x52b1f00000000ULL,
    0x2ebaf80000000ULL,
    0xfaf67c0000000ULL,
    0xbe99b20000000ULL,
    0x6283f90000000ULL,
    0x58c67f8000000ULL,
    0xb9e3e0c000000ULL,
    0x1910522000000ULL,
    0x728dc91000000ULL,
    0xc28f678800000ULL,
    0xda8eacdc00000ULL,
    0x3689783e00000ULL,
    0x388dbc3f00000ULL,
    0x3f88520980000ULL,
    0xbc09c92cc0000ULL,
    0xb44d679060000ULL,
    0x146facdc50000ULL,
    0xd41af83728000ULL,
    0xfe467c2dec000ULL,
    0x5b21b2290e000ULL,
    0x65f7f91d8d000ULL,
    0x911c7fa55e800ULL,
    0x44c6e0c2a5400ULL,
    0x8fe1d224b2600ULL,
    0x8217092dabb00ULL,
    0x870d07a23e180ULL,
    0x71cf5cdb753c0ULL,
    0xff2b80095ade0ULL,
    0x3dbfc033272f0ULL,
    0x8b73e000e0f48ULL,
    0x97db301211824ULL,
    0xf220982009a1eULL,
    0xd1738c2c1c273ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0x8400000000000ULL,
    0xaa00000000000ULL,
    0x4100000000000ULL,
    0x9480000000000ULL,
    0x8ec0000000000ULL,
    0x8c60000000000ULL,
    0xab70000000000ULL,
    0xb2f8000000000ULL,
    0xa434000000000ULL,
    0x36de000000000ULL,
    0xe0e1000000000ULL,
    0xe9b1800000000ULL,
    0xb099400000000ULL,
    0x1447a00000000ULL,
    0xd9a7100000000ULL,
    0xda93280000000ULL,
    0x77ce740000000ULL,
    0x366e3e0000000ULL,
    0x70fee70000000ULL,
    0xf1337d8000000ULL,
    0xd05f7cc000000ULL,
    0xd3261e2000000ULL,
    0x5751b73000000ULL,
    0xfa2e75a800000ULL,
    0x555b58fc00000ULL,
    0x8da2a80200000ULL,
    0xc897341100000ULL,
    0x9ac99e0280000ULL,
    0x7ce9f72c40000ULL,
    0x9d38558720000ULL,
    0x2cd508c2b0000ULL,
    0xcb6e202998000ULL,
    0x227a500a9c000ULL,
    0x40f288219e000ULL,
    0x5bbc6436bd000ULL,
    0x2f12960caf800ULL,
    0x5d88d333e4400ULL,
    0x1ccb63bc52600ULL,
    0x23efcbd56b500ULL,
    0xcab9eb8d86680ULL,
    0x393afdde1840ULL,
    0x7f4b7da7d0aa0ULL,
    0x4eab7cd709cd0ULL,
    0xeb981e0e34678ULL,
    0x29c0b72e5b7dcULL,
    0x17e7f5a7df956ULL,
    0xa3618dc9c023ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x3c00000000000ULL,
    0x4a00000000000ULL,
    0xeb00000000000ULL,
    0xcf80000000000ULL,
    0x4940000000000ULL,
    0xb3e0000000000ULL,
    0xab30000000000ULL,
    0x8258000000000ULL,
    0xe2f4000000000ULL,
    0x217a000000000ULL,
    0x261000000000ULL,
    0xf777800000000ULL,
    0xab3bc00000000ULL,
    0x3a00600000000ULL,
    0xd306900000000ULL,
    0xe383b80000000ULL,
    0x6b41ac0000000ULL,
    0xbce3d20000000ULL,
    0x12b2030000000ULL,
    0x6a1d438000000ULL,
    0x7592d5c000000ULL,
    0xc8e322000000ULL,
    0x7a9e531000000ULL,
    0xd511ba800000ULL,
    0xa32a29ec00000ULL,
    0x58cc381600000ULL,
    0xe7fe6c1f00000ULL,
    0x6ba1b23a80000ULL,
    0xa4519316c0000ULL,
    0x33ab7b85e0000ULL,
    0x600db9dd50000ULL,
    0xe858001e78000ULL,
    0xf9f400000c000ULL,
    0x96fa003b76000ULL,
    0xc7210006ed000ULL,
    0xd6978012a6800ULL,
    0xd70bc02554400ULL,
    0x3dd8600fcbe00ULL,
    0x93b29028b3f00ULL,
    0xbe99b8353e980ULL,
    0x8b50ac2609e40ULL,
    0x7a2c5229c50e0ULL,
    0xf04dc31ad29b0ULL,
    0xf33f23812eec8ULL,
    0x460145c8d024cULL,
    0x39000a0a393eeULL,
    0xb8853f172c9e7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0xdc00000000000ULL,
    0x8e00000000000ULL,
    0x2300000000000ULL,
    0x9980000000000ULL,
    0x2c0000000000ULL,
    0x3220000000000ULL,
    0x7590000000000ULL,
    0x1d28000000000ULL,
    0xbbd4000000000ULL,
    0x834e000000000ULL,
    0xd523000000000ULL,
    0x615800000000ULL,
    0x86ef400000000ULL,
    0x1171600000000ULL,
    0xc89a300000000ULL,
    0x9e6cf80000000ULL,
    0xef313c0000000ULL,
    0x7ef84e0000000ULL,
    0xbc9b3d0000000ULL,
    0x9c6e0a8000000ULL,
    0xda33c04000000ULL,
    0x607aae2000000ULL,
    0x7dd94d3000000ULL,
    0xdd8812b800000ULL,
    0xa9048c5c00000ULL,
    0xb082f82200000ULL,
    0x72423c0700000ULL,
    0xa065ce1b80000ULL,
    0xa5f07d14c0000ULL,
    0xb0d96aa460000ULL,
    0xd70ef077b0000ULL,
    0xda45d60f08000ULL,
    0x2c603125a4000ULL,
    0xb3f2bc992a000ULL,
    0x37ddc17245000ULL,
    0x148aea8e5b800ULL,
    0xa986b06f2cc00ULL,
    0x2ac7363d4f600ULL,
    0x7e22412b8ed00ULL,
    0x8394a4b8efd80ULL,
    0xea2a8d6cde4c0ULL,
    0x7052bcab165e0ULL,
    0x2c8dc148d39f0ULL,
    0x5d82ea81484b8ULL,
    0xe8c2b05785434ULL,
    0xab213610bb3caULL,
    0xed15413edc7f1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xa800000000000ULL,
    0x9400000000000ULL,
    0x600000000000ULL,
    0xb100000000000ULL,
    0x3b80000000000ULL,
    0x54c0000000000ULL,
    0x45e0000000000ULL,
    0xf970000000000ULL,
    0x6b68000000000ULL,
    0x1534000000000ULL,
    0x86ce000000000ULL,
    0xffe5000000000ULL,
    0x1a73800000000ULL,
    0xc7ec400000000ULL,
    0x7b75200000000ULL,
    0x46b700000000ULL,
    0x83b4280000000ULL,
    0xcb8a6c0000000ULL,
    0xf1c76e0000000ULL,
    0xb864d30000000ULL,
    0x3cb11d8000000ULL,
    0xa50ddc4000000ULL,
    0x1087ce2000000ULL,
    0x3c42e33000000ULL,
    0x15a5958800000ULL,
    0x46d1806c00000ULL,
    0x16bc280a00000ULL,
    0xb30e6c1500000ULL,
    0x79816e0980000ULL,
    0x1bc5d30040000ULL,
    0x63649da4e0000ULL,
    0x1c309c7030000ULL,
    0xf14f6e30f8000ULL,
    0xb120d3221c000ULL,
    0xcc971d9c3a000ULL,
    0x1a1cdc7c7d000ULL,
    0x5a5a4e3acb800ULL,
    0x3dfba314a5400ULL,
    0x5c2b3595c6e00ULL,
    0xd812b074f6100ULL,
    0x385b203b8ec80ULL,
    0xc2fe703ed39c0ULL,
    0xd2afa83e89ea0ULL,
    0xa9522c3d058b0ULL,
    0xa47c4e1c96938ULL,
    0xd7eaa31a7ee8cULL,
    0xa376b59e8acd6ULL,
    0x186bf07644ac1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x8800000000000ULL,
    0x4400000000000ULL,
    0xc200000000000ULL,
    0x7700000000000ULL,
    0x9680000000000ULL,
    0x2a40000000000ULL,
    0x13a0000000000ULL,
    0xcc30000000000ULL,
    0x46a8000000000ULL,
    0x79f4000000000ULL,
    0xc0ce000000000ULL,
    0x3a7000000000ULL,
    0x8432800000000ULL,
    0x62af400000000ULL,
    0x2bf2a00000000ULL,
    0x3fcc500000000ULL,
    0xd126980000000ULL,
    0x6c70240000000ULL,
    0x60daa0000000ULL,
    0x7146890000000ULL,
    0x5323638000000ULL,
    0xbb70304000000ULL,
    0x608b8a2000000ULL,
    0x4306993000000ULL,
    0x8c83db9800000ULL,
    0xf140046400000ULL,
    0x9326980200000ULL,
    0xdb70242100000ULL,
    0xf08daa2880000ULL,
    0xcb068939c0000ULL,
    0xc8836387a0000ULL,
    0x3340305b90000ULL,
    0xe4238a1468000ULL,
    0x4df2991ecc000ULL,
    0xdacddbac0a000ULL,
    0xd8a70470ed000ULL,
    0x4b41836db800ULL,
    0x75ef6412e5c00ULL,
    0x9dd70a1826a00ULL,
    0x8d3ed93f16d00ULL,
    0xd96bfbbd6f280ULL,
    0x5c9714548ad40ULL,
    0x661ca0012d060ULL,
    0x5e1b501d7bd90ULL,
    0xa21c182775c28ULL,
    0x5c1b64234fb64ULL,
    0xb5190a365afeeULL,
    0x5a99d91725c51ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0xc00000000000ULL,
    0xde00000000000ULL,
    0x6500000000000ULL,
    0xb180000000000ULL,
    0xd940000000000ULL,
    0x2ee0000000000ULL,
    0xa310000000000ULL,
    0xe688000000000ULL,
    0x3df4000000000ULL,
    0x9d9e000000000ULL,
    0xcf7f000000000ULL,
    0xa6e800000000ULL,
    0xde6400000000ULL,
    0xc391e00000000ULL,
    0xd44e300000000ULL,
    0xed54880000000ULL,
    0xf2ea5c0000000ULL,
    0x8b24720000000ULL,
    0x4933a30000000ULL,
    0x9c39368000000ULL,
    0x4b8de54000000ULL,
    0xb070fa2000000ULL,
    0x82d9ff3000000ULL,
    0xf09d449800000ULL,
    0x22fe466400000ULL,
    0x8529cc9a00000ULL,
    0xda041a7300000ULL,
    0xe705be8f80000ULL,
    0x8283b94d40000ULL,
    0x91c2882e60000ULL,
    0x70215c2150000ULL,
    0x3db4f21e38000ULL,
    0x4fae31684000ULL,
    0xc42e569842000ULL,
    0xd98195636d000ULL,
    0x154392173f800ULL,
    0x90e693040bc00ULL,
    0x56153e84e1a00ULL,
    0xbf0af97690900ULL,
    0xe8b5e83f99f80ULL,
    0x6d7d2c12754c0ULL,
    0x96f9a088b020ULL,
    0x5d618f3e21c70ULL,
    0xe9502ca1b1448ULL,
    0x70ee2a62a934cULL,
    0xb82736b21cb7aULL,
    0x1b2e559b7f25ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x3800000000000ULL,
    0x5400000000000ULL,
    0xf600000000000ULL,
    0x6300000000000ULL,
    0xab80000000000ULL,
    0xad40000000000ULL,
    0xf0e0000000000ULL,
    0x6450000000000ULL,
    0x99d8000000000ULL,
    0x40b4000000000ULL,
    0xcb8e000000000ULL,
    0x1a69000000000ULL,
    0xb0be800000000ULL,
    0x44a6c00000000ULL,
    0xe35200000000ULL,
    0x664ff00000000ULL,
    0xb048f80000000ULL,
    0x834b8c0000000ULL,
    0xb0cc620000000ULL,
    0xd98df30000000ULL,
    0xe76a4c8000000ULL,
    0xb43b9fc000000ULL,
    0x70649a2000000ULL,
    0xfe967f3000000ULL,
    0x6e7e2ea800000ULL,
    0xbd026ce400000ULL,
    0x6480d6ae00000ULL,
    0x4c4e0e500000ULL,
    0x58a434b580000ULL,
    0x6432d3ccc0000ULL,
    0xaf4bd82ce0000ULL,
    0x72c97c2a50000ULL,
    0x7c8c1a27b8000ULL,
    0x87edbf30d4000ULL,
    0xdefb8ebf16000ULL,
    0x8fc25cf6bd000ULL,
    0xff238eb9bb800ULL,
    0x5b765cdc1e400ULL,
    0x62ad8e9481a00ULL,
    0xb21f5ce740b00ULL,
    0x41930ead20c80ULL,
    0xff99cd730bc0ULL,
    0x49462eaf281e0ULL,
    0x6ee66cfe8c130ULL,
    0xcb56d6a012708ULL,
    0x59e0cad370cULL,
    0x40f4b49732452ULL,
    0xa7ed13eee248fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x6800000000000ULL,
    0x7c00000000000ULL,
    0xc600000000000ULL,
    0x1100000000000ULL,
    0x2580000000000ULL,
    0x6040000000000ULL,
    0xc060000000000ULL,
    0x6110000000000ULL,
    0xdd28000000000ULL,
    0x4d74000000000ULL,
    0x323a000000000ULL,
    0xcd5f000000000ULL,
    0xa4cc800000000ULL,
    0x9a26400000000ULL,
    0x3670600000000ULL,
    0xe8bcf00000000ULL,
    0x859ea80000000ULL,
    0xb769bc0000000ULL,
    0x22135a0000000ULL,
    0x13aa170000000ULL,
    0xbfb76e8000000ULL,
    0xf69ae44000000ULL,
    0x41edf22000000ULL,
    0xc4d3ab1000000ULL,
    0x6d0c349800000ULL,
    0xe404f35400000ULL,
    0xa2009cba00000ULL,
    0xf3064f4f00000ULL,
    0xb685468980000ULL,
    0x86c5184040000ULL,
    0x2ea6c82b60000ULL,
    0x33b14c0310000ULL,
    0x289ff20398000ULL,
    0x74e8ab1704000ULL,
    0x6352b495aa000ULL,
    0xce49b3500d000ULL,
    0xaae67ca476800ULL,
    0x51d3ff6743c00ULL,
    0xba8f0e9f01200ULL,
    0x1f42147180900ULL,
    0xa8e15a2440980ULL,
    0xd2d1171961fc0ULL,
    0x5409ee80100a0ULL,
    0xdd87a464193f0ULL,
    0x54431222456e8ULL,
    0x4a621b384be9cULL,
    0x8e147cbadd49aULL,
    0xfda8ff5cce7a1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0x8c00000000000ULL,
    0xd600000000000ULL,
    0xb00000000000ULL,
    0xcb80000000000ULL,
    0xfec0000000000ULL,
    0xa4a0000000000ULL,
    0xe090000000000ULL,
    0xe118000000000ULL,
    0x7834000000000ULL,
    0xbf8e000000000ULL,
    0x862f000000000ULL,
    0x638800000000ULL,
    0xe266c00000000ULL,
    0x6833e00000000ULL,
    0xe788f00000000ULL,
    0x3a2d780000000ULL,
    0x78394c0000000ULL,
    0x9d647e0000000ULL,
    0xf9b2110000000ULL,
    0xc448b68000000ULL,
    0x5e0ea7c000000ULL,
    0xade9062000000ULL,
    0x261b5d3000000ULL,
    0xc5b4c89800000ULL,
    0x7a48b6d400000ULL,
    0x410fb09e00000ULL,
    0x6c6afad300000ULL,
    0xfddd4e8d80000ULL,
    0x2d912be6c0000ULL,
    0x799e9834e0000ULL,
    0xf171bc38b0000ULL,
    0x9be9063aa8000ULL,
    0xbd1b5d0de4000ULL,
    0x9634c88126000ULL,
    0x5888b6c07d000ULL,
    0xcbafb0a1ab800ULL,
    0xbfafad4dfc00ULL,
    0x1454e9a00200ULL,
    0xa0652bc500b00ULL,
    0xa930982281780ULL,
    0x690ebc3940fc0ULL,
    0xd869860b20fe0ULL,
    0xc7d99d2cd1230ULL,
    0xa09128aa58478ULL,
    0x411b46c0ecc84ULL,
    0x483448aa53abaULL,
    0x178a76fda276fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x8800000000000ULL,
    0x8400000000000ULL,
    0xc600000000000ULL,
    0xf500000000000ULL,
    0xc880000000000ULL,
    0xef40000000000ULL,
    0x13a0000000000ULL,
    0x79d0000000000ULL,
    0x3118000000000ULL,
    0x9674000000000ULL,
    0xf2ce000000000ULL,
    0x286f000000000ULL,
    0x433c800000000ULL,
    0xc265400000000ULL,
    0xe837a00000000ULL,
    0xffeff00000000ULL,
    0xa3fd480000000ULL,
    0xfd02140000000ULL,
    0x8c87c20000000ULL,
    0x4942c10000000ULL,
    0x36a7988000000ULL,
    0x39504bc000000ULL,
    0x5a5a8a2000000ULL,
    0x43d0d53000000ULL,
    0x7e185a9800000ULL,
    0xd1f68af400000ULL,
    0x3a0b128200000ULL,
    0x230b9ed100000ULL,
    0xb988508980000ULL,
    0x8dc81ff940000ULL,
    0xf7ee682820000ULL,
    0xe7fca41ed0000ULL,
    0x5b032a0f68000ULL,
    0xa980253634000ULL,
    0x9c192ade6000ULL,
    0x5de5dec20d000ULL,
    0xecf5709bfb800ULL,
    0xd689afe2ffc00ULL,
    0xba4e001241200ULL,
    0x72f0034a0500ULL,
    0x309c803410680ULL,
    0x6bb5401b093c0ULL,
    0x512fa01b45f20ULL,
    0xed9bf03c1f150ULL,
    0x97334836f0b88ULL,
    0x206d1411b9dc4ULL,
    0x73b42395c7c2ULL,
    0x6467811bd23bfULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x3c00000000000ULL,
    0x1600000000000ULL,
    0x6d00000000000ULL,
    0xd680000000000ULL,
    0x4cc0000000000ULL,
    0xeee0000000000ULL,
    0xa8d0000000000ULL,
    0x1ae8000000000ULL,
    0xd034000000000ULL,
    0x1f3e000000000ULL,
    0x3c5d000000000ULL,
    0x33cf800000000ULL,
    0x1585c00000000ULL,
    0x9b43600000000ULL,
    0x2aa2f00000000ULL,
    0x33f1c80000000ULL,
    0x7859240000000ULL,
    0x69cc5e0000000ULL,
    0x96801b0000000ULL,
    0xacc5fc8000000ULL,
    0x7ee7f24000000ULL,
    0xc0d5962000000ULL,
    0xfeed3f3000000ULL,
    0xfa37a28800000ULL,
    0x643ae96c00000ULL,
    0x87dfea9600000ULL,
    0xa98f0d7f00000ULL,
    0xb7a154ad80000ULL,
    0xdd75264740000ULL,
    0x9899802ba0000ULL,
    0xf92cc00030000ULL,
    0xb752e00038000ULL,
    0x4aaa301274000ULL,
    0x991528009a000ULL,
    0x8a8a14337d000ULL,
    0xf020f60c57800ULL,
    0x7136cf2d53400ULL,
    0xe7bfea8451e00ULL,
    0xb19f0d4868700ULL,
    0x75a954bc1d580ULL,
    0x7891266b86fc0ULL,
    0x93cf80297a120ULL,
    0x6585c03c6d790ULL,
    0x634360099faf8ULL,
    0xa6a2f0067e8acULL,
    0xfdf1c83e6e352ULL,
    0x2959241a26799ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x6800000000000ULL,
    0x8c00000000000ULL,
    0x6e00000000000ULL,
    0xff00000000000ULL,
    0x9b80000000000ULL,
    0xef40000000000ULL,
    0xe160000000000ULL,
    0xa710000000000ULL,
    0xa258000000000ULL,
    0xd874000000000ULL,
    0x24a000000000ULL,
    0xa5a9000000000ULL,
    0x778800000000ULL,
    0xbf80400000000ULL,
    0xed45e00000000ULL,
    0x63b00000000ULL,
    0x7b96580000000ULL,
    0x32988c0000000ULL,
    0x3453920000000ULL,
    0xd53d5f0000000ULL,
    0xc460668000000ULL,
    0x9953fc000000ULL,
    0x6b9dca2000000ULL,
    0xcd1d31000000ULL,
    0xa7f9f4a800000ULL,
    0xb94160f400000ULL,
    0xba652cba00000ULL,
    0x7e94acf300000ULL,
    0xc19de9380000ULL,
    0x359703dbc0000ULL,
    0x5d988034e0000ULL,
    0x67d0403cd0000ULL,
    0x667de011d8000ULL,
    0x2307b00104000ULL,
    0xdd84580ff6000ULL,
    0x3c458c110d000ULL,
    0xc4e11205aa800ULL,
    0xdf541f04be400ULL,
    0x54bd06b7f0200ULL,
    0xfe26cfcbe9100ULL,
    0xbe76720b0df80ULL,
    0x414eef26abac0ULL,
    0x482e3e9f3f360ULL,
    0x2339b3c831730ULL,
    0x4f64580d093c8ULL,
    0xb8158c00ddebcULL,
    0x49d91226f2d7aULL,
    0x8f301f2afb90fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x9800000000000ULL,
    0xc00000000000ULL,
    0xb200000000000ULL,
    0x2500000000000ULL,
    0x6980000000000ULL,
    0xc3c0000000000ULL,
    0xa960000000000ULL,
    0xdef0000000000ULL,
    0xdef8000000000ULL,
    0xd194000000000ULL,
    0xdb0a000000000ULL,
    0xd1ed000000000ULL,
    0x35db800000000ULL,
    0x91c3c00000000ULL,
    0x5c66e00000000ULL,
    0xbf75d00000000ULL,
    0x3938080000000ULL,
    0x5ef67c0000000ULL,
    0x9ef8c60000000ULL,
    0xf197bf0000000ULL,
    0x6b0d0d8000000ULL,
    0x49ef2ec000000ULL,
    0x39d8ce2000000ULL,
    0x23c5c31000000ULL,
    0x7967cb8800000ULL,
    0xd6f191ec00000ULL,
    0xfafc438600000ULL,
    0xf7902dd300000ULL,
    0x4008658480000ULL,
    0x2f6f42f540000ULL,
    0xba9ee01460000ULL,
    0x92e1d00ff0000ULL,
    0xe8320828d8000ULL,
    0x161b7c11fc000ULL,
    0xe8a346275e000ULL,
    0x8a947f0ed5000ULL,
    0x458bedaa7a800ULL,
    0xceaafefa77400ULL,
    0x1ef8c62af0e00ULL,
    0xb197bf0759f00ULL,
    0x4b0d0db43d380ULL,
    0xf9ef2ef47f340ULL,
    0xa1d8ce0444620ULL,
    0x2fc5c320d32f0ULL,
    0xcb67cbaf93e98ULL,
    0xf3f191e6f32bcULL,
    0x937c439282196ULL,
    0x34502ded9bd0bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xa800000000000ULL,
    0xc400000000000ULL,
    0x9e00000000000ULL,
    0x1100000000000ULL,
    0xfe80000000000ULL,
    0x52c0000000000ULL,
    0x16e0000000000ULL,
    0xef70000000000ULL,
    0x1ab8000000000ULL,
    0x7894000000000ULL,
    0x434a000000000ULL,
    0x4e9000000000ULL,
    0x4b3c800000000ULL,
    0xd5d6400000000ULL,
    0xf1eba00000000ULL,
    0x3bba300000000ULL,
    0x9e17680000000ULL,
    0x8d8f640000000ULL,
    0xb00b4e0000000ULL,
    0x474afb0000000ULL,
    0x7aeaf18000000ULL,
    0xaa3e50c000000ULL,
    0xd356262000000ULL,
    0xcf2c9f1000000ULL,
    0x775d3fa800000ULL,
    0xfe62ebd400000ULL,
    0x78b7778a00000ULL,
    0x64d8ffe100000ULL,
    0x402471af80000ULL,
    0x879510c040000ULL,
    0x14cb061da0000ULL,
    0x64a9ef21b0000ULL,
    0x51d77a038000ULL,
    0xb41ffda9c000ULL,
    0x2fa0f1948e000ULL,
    0x6bd750efd5000ULL,
    0xf0eaa6054a800ULL,
    0xcd3adf2e16400ULL,
    0x58d69fac79200ULL,
    0xad68dbe13c900ULL,
    0x8a781f823e680ULL,
    0x3d739bcf6dcc0ULL,
    0x41bd3fb3167a0ULL,
    0x112ebddf9090ULL,
    0x6a0f778a7d9b8ULL,
    0x884cfff11fe2cULL,
    0x356e71911cae6ULL,
    0x567c10e6ce513ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0x4c00000000000ULL,
    0xaa00000000000ULL,
    0x6d00000000000ULL,
    0x3480000000000ULL,
    0x82c0000000000ULL,
    0x69a0000000000ULL,
    0xd390000000000ULL,
    0xe9a8000000000ULL,
    0xe734000000000ULL,
    0x76be000000000ULL,
    0xc85b000000000ULL,
    0x4e28800000000ULL,
    0x4272400000000ULL,
    0x7a59e00000000ULL,
    0xf72b900000000ULL,
    0xc8f6c80000000ULL,
    0xcb98d40000000ULL,
    0xc909ce0000000ULL,
    0x2623110000000ULL,
    0x3ad4668000000ULL,
    0x4f4b024000000ULL,
    0xb541062000000ULL,
    0xc4e0c53000000ULL,
    0xc37528b800000ULL,
    0x94da535400000ULL,
    0x40ec809a00000ULL,
    0x61d0576300000ULL,
    0x94cae6b280000ULL,
    0xc206427f40000ULL,
    0xe106662f20000ULL,
    0x7e861517b0000ULL,
    0xbfc480b2e8000ULL,
    0xb5245768a4000ULL,
    0x1d54e687ca000ULL,
    0x2a0d42613d000ULL,
    0x59a6e63867800ULL,
    0xab90551902c00ULL,
    0xadab60bca8200ULL,
    0x5130c77a05900ULL,
    0x59bcae93bbf80ULL,
    0xddd8d67ab5240ULL,
    0x52684839d2120ULL,
    0xc415943651650ULL,
    0xf4eeae37b9ab8ULL,
    0x4fd5c1059c704ULL,
    0xafcdce8116f3aULL,
    0x898106490b819ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xc800000000000ULL,
    0x4400000000000ULL,
    0xfa00000000000ULL,
    0x8300000000000ULL,
    0x1a80000000000ULL,
    0xbd40000000000ULL,
    0x1fa0000000000ULL,
    0xbad0000000000ULL,
    0x5bd8000000000ULL,
    0x4e74000000000ULL,
    0x8f8a000000000ULL,
    0xbaed000000000ULL,
    0x695d800000000ULL,
    0xa731400000000ULL,
    0xf22aa00000000ULL,
    0x7f3f100000000ULL,
    0x5e05f80000000ULL,
    0x6900640000000ULL,
    0x186820000000ULL,
    0x5bc6210000000ULL,
    0xd4e0908000000ULL,
    0x9872edc000000ULL,
    0x82897a2000000ULL,
    0x316b451000000ULL,
    0x799b928800000ULL,
    0x2d558cdc00000ULL,
    0x2d1b4a9200000ULL,
    0x6112b8c100000ULL,
    0xcf3d109680000ULL,
    0xf603adecc0000ULL,
    0x7d03da1ca0000ULL,
    0x4384553250000ULL,
    0x54c66a8368000ULL,
    0x7061e8c374000ULL,
    0x5c37c8bcd6000ULL,
    0x4a999dd0d000ULL,
    0x2c7800342a800ULL,
    0x80a400039f400ULL,
    0xc6520037e9600ULL,
    0x43990036b5500ULL,
    0xce578002f6280ULL,
    0x679c403d9d840ULL,
    0x64572000e29e0ULL,
    0x5c9e50013a090ULL,
    0xf2d7580416e88ULL,
    0x5fdb740fad5c4ULL,
    0x94717a157a922ULL,
    0x7c8f453776367ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0x5c00000000000ULL,
    0xd200000000000ULL,
    0x2700000000000ULL,
    0xaf80000000000ULL,
    0xa940000000000ULL,
    0x1a60000000000ULL,
    0x1910000000000ULL,
    0xfd38000000000ULL,
    0x3874000000000ULL,
    0x39ae000000000ULL,
    0xbb0f000000000ULL,
    0xbc3e800000000ULL,
    0xf6f5c00000000ULL,
    0x6e6b600000000ULL,
    0x9ea9d00000000ULL,
    0xd489980000000ULL,
    0x757ccc0000000ULL,
    0x7c95820000000ULL,
    0xef7f470000000ULL,
    0x3f94668000000ULL,
    0x3efd3cc000000ULL,
    0x6bd21a2000000ULL,
    0x681c8b3000000ULL,
    0xfa07649800000ULL,
    0xd303bbe400000ULL,
    0x49831c8600000ULL,
    0x84767e700000ULL,
    0x84e2668c80000ULL,
    0xb6d63cddc0000ULL,
    0x149a9a35e0000ULL,
    0x1ac24b2950000ULL,
    0xc124848c18000ULL,
    0xd474abfc84000ULL,
    0x43a964a62e000ULL,
    0xa80cbbca8d000ULL,
    0x95bd9caf93800ULL,
    0x6eb2a7df17c00ULL,
    0x728906a339a00ULL,
    0x747fecdcb4700ULL,
    0x12130206a7580ULL,
    0x48be870bf0b40ULL,
    0x1231068a68460ULL,
    0x924becd0ac830ULL,
    0x665d023a23988ULL,
    0x8fe1870ddf15cULL,
    0x35786b0e407aULL,
    0x3eda2cc73e783ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x2c00000000000ULL,
    0x2200000000000ULL,
    0x300000000000ULL,
    0xca80000000000ULL,
    0x7540000000000ULL,
    0xc420000000000ULL,
    0xb310000000000ULL,
    0x61e8000000000ULL,
    0xa834000000000ULL,
    0x627e000000000ULL,
    0xb99000000000ULL,
    0xf2ea800000000ULL,
    0x8ab6400000000ULL,
    0x9b3ae00000000ULL,
    0xfdb9d00000000ULL,
    0x6afab80000000ULL,
    0xdd8640000000ULL,
    0x644c420000000ULL,
    0x58e7bd0000000ULL,
    0x18f3de8000000ULL,
    0xeb9c604000000ULL,
    0x82e8fa2000000ULL,
    0x12b6d93000000ULL,
    0x8f3d1ca800000ULL,
    0xdbb99d4400000ULL,
    0x67ffc4aa00000ULL,
    0xe65a697b00000ULL,
    0xd88dde8080000ULL,
    0x23056074c0000ULL,
    0x1a827a30a0000ULL,
    0xfd40992250000ULL,
    0xf827fc9b88000ULL,
    0xb9104d6304000ULL,
    0x4eed7c885a000ULL,
    0x40b60d4c9d000ULL,
    0x143f9c9bb7800ULL,
    0x53bdd47b2400ULL,
    0x34bb24a568a00ULL,
    0x2f7ab94634900ULL,
    0x801de69da3380ULL,
    0xfe2b444481340ULL,
    0xc954d81ec18e0ULL,
    0x9d4ef42ba0c90ULL,
    0xe5669a32d0888ULL,
    0x31304937484b4ULL,
    0x6fffc49fa441aULL,
    0x1a5a69740be51ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x7800000000000ULL,
    0x1400000000000ULL,
    0xea00000000000ULL,
    0xb300000000000ULL,
    0x4580000000000ULL,
    0x2440000000000ULL,
    0x3ba0000000000ULL,
    0x4710000000000ULL,
    0x4f48000000000ULL,
    0x93b4000000000ULL,
    0x5bda000000000ULL,
    0xadbf000000000ULL,
    0x2b48800000000ULL,
    0xf1b6c00000000ULL,
    0x34df200000000ULL,
    0xbe3d100000000ULL,
    0x3a0e780000000ULL,
    0xc2931c0000000ULL,
    0x4b0cca0000000ULL,
    0x1813790000000ULL,
    0xb4cb4e8000000ULL,
    0x8ef3994000000ULL,
    0x46f8b22000000ULL,
    0xf36f651000000ULL,
    0x6fe7049800000ULL,
    0xf4b2207400000ULL,
    0xd45edcbe00000ULL,
    0x8afaec5500000ULL,
    0xfd6b4e8280000ULL,
    0xfee39948c0000ULL,
    0xbe30b23760000ULL,
    0x339b652810000ULL,
    0xd81d04b268000ULL,
    0x9d5d204c04000ULL,
    0x247e5c8632000ULL,
    0x38a82c543d000ULL,
    0xa3066e8bfe800ULL,
    0x1d85897b7bc00ULL,
    0x80444a2088200ULL,
    0x79a5b92154b00ULL,
    0x98146eba7b480ULL,
    0xf4ce896c09840ULL,
    0xeef6ca14953e0ULL,
    0x96fc79339bb70ULL,
    0x8b6bcebed85a8ULL,
    0x7be1597e9dea4ULL,
    0x1eb592160ea5aULL,
    0x67597529c3d55ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0x5400000000000ULL,
    0x1200000000000ULL,
    0x5d00000000000ULL,
    0x6980000000000ULL,
    0xf0c0000000000ULL,
    0x96e0000000000ULL,
    0xd0f0000000000ULL,
    0x9978000000000ULL,
    0x6914000000000ULL,
    0xad0a000000000ULL,
    0xa12b000000000ULL,
    0x663a800000000ULL,
    0x7030400000000ULL,
    0x879b200000000ULL,
    0x9de6500000000ULL,
    0x9e77880000000ULL,
    0xe93f3c0000000ULL,
    0xe4b7de0000000ULL,
    0x6edec30000000ULL,
    0x43c1058000000ULL,
    0xa462a6c000000ULL,
    0xe6b2562000000ULL,
    0x5bdeff1000000ULL,
    0x9e465bb800000ULL,
    0xd6a725e400000ULL,
    0x552f3aa00000ULL,
    0xb6aa49c500000ULL,
    0xe5f8a59c80000ULL,
    0x7450b6e340000ULL,
    0x112cfe0860000ULL,
    0x7e38931d30000ULL,
    0x7c368d9338000ULL,
    0x799d9ad27c000ULL,
    0xaee58812be000ULL,
    0xecf03c3435000ULL,
    0x3f7f5e281a800ULL,
    0xb611831306c00ULL,
    0xb18aa58a08a00ULL,
    0x6c6fb6db45100ULL,
    0x121c7e2dc2480ULL,
    0x6b23d3260a540ULL,
    0xa7972d916fc20ULL,
    0x9d4b8acc7c1f0ULL,
    0xcc892025bf1f8ULL,
    0x35e95039b4f4cULL,
    0xbadf0815dbcb6ULL,
    0x11c07c0127e2bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x6800000000000ULL,
    0xac00000000000ULL,
    0x2200000000000ULL,
    0x3500000000000ULL,
    0x2880000000000ULL,
    0xc640000000000ULL,
    0xc460000000000ULL,
    0x9db0000000000ULL,
    0x7318000000000ULL,
    0xa6d4000000000ULL,
    0x1f2a000000000ULL,
    0x390d000000000ULL,
    0x1eda800000000ULL,
    0x3f74400000000ULL,
    0xa33ae00000000ULL,
    0x1203d00000000ULL,
    0x8d01080000000ULL,
    0xcc87740000000ULL,
    0xd8476a0000000ULL,
    0xbb649f0000000ULL,
    0x2c37538000000ULL,
    0xbfdbddc000000ULL,
    0x91f4622000000ULL,
    0x6e7aeb1000000ULL,
    0x11e0b98800000ULL,
    0xf7602e400000ULL,
    0x1b3b51ba00000ULL,
    0xf602a6fb00000ULL,
    0x930533a080000ULL,
    0xb3854df940000ULL,
    0x69c70a3820000ULL,
    0x77a30f2e90000ULL,
    0x1b14bbab98000ULL,
    0xce8b79e52c000ULL,
    0xb918002466000ULL,
    0x7fd4002e65000ULL,
    0x35aa003c72800ULL,
    0x5a4d0031ff400ULL,
    0x9a3a803a48a00ULL,
    0xc884401b94100ULL,
    0x3642e00a5a380ULL,
    0x1c67d0095ab40ULL,
    0xc9b30835da620ULL,
    0x951e742a9a8d0ULL,
    0x1dd7ea2fbb428ULL,
    0x20addf0cab83cULL,
    0xe2cf339bf38eeULL,
    0x34784dfe3fce7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xc800000000000ULL,
    0x5400000000000ULL,
    0xa00000000000ULL,
    0x3b00000000000ULL,
    0xf480000000000ULL,
    0xf6c0000000000ULL,
    0xf2a0000000000ULL,
    0xab0000000000ULL,
    0xbbc8000000000ULL,
    0x2114000000000ULL,
    0x5afe000000000ULL,
    0xea9f000000000ULL,
    0x88c800000000ULL,
    0x2af7400000000ULL,
    0xc0a9200000000ULL,
    0x8f87900000000ULL,
    0xa245b80000000ULL,
    0xb460140000000ULL,
    0xe0121e0000000ULL,
    0x2d7ff90000000ULL,
    0xc4df3f8000000ULL,
    0x4ae8094000000ULL,
    0x7fe1a62000000ULL,
    0xe054ed3000000ULL,
    0x261fa19800000ULL,
    0x124fb07400000ULL,
    0xfe53399200000ULL,
    0xb718346500000ULL,
    0x6dcc9fba80000ULL,
    0xe413d94ac0000ULL,
    0x2f7fbe3720000ULL,
    0xcbdb292cb0000ULL,
    0x646da79428000ULL,
    0x7a208d6d6c000ULL,
    0xb272800352000ULL,
    0xd068401e35000ULL,
    0x6025a00db7800ULL,
    0x2170d03f7bc00ULL,
    0xa0ec981f59200ULL,
    0x54e78428e4d00ULL,
    0xbcd7a6364fc80ULL,
    0x54dfed1d6f240ULL,
    0x22ed21afdfce0ULL,
    0x9be7f068f7150ULL,
    0xf25699801a6f8ULL,
    0x8118e46a080a4ULL,
    0xb8c807bd5dc9eULL,
    0x39905d5abb335ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0x9c00000000000ULL,
    0xb200000000000ULL,
    0xab00000000000ULL,
    0xc980000000000ULL,
    0x74c0000000000ULL,
    0xdce0000000000ULL,
    0xd5f0000000000ULL,
    0x5738000000000ULL,
    0xc14000000000ULL,
    0x924e000000000ULL,
    0x2d6f000000000ULL,
    0x8ff9800000000ULL,
    0xcaf1400000000ULL,
    0x60ba600000000ULL,
    0x79d1500000000ULL,
    0xc82c780000000ULL,
    0x63d87c0000000ULL,
    0xf5e60e0000000ULL,
    0x1f75230000000ULL,
    0x667a7d8000000ULL,
    0x6e33e4c000000ULL,
    0xa45c762000000ULL,
    0x48265f3000000ULL,
    0xb913f39800000ULL,
    0xeacc87dc00000ULL,
    0xb72beba200000ULL,
    0x245aabd700000ULL,
    0x88201d9480000ULL,
    0xd912b4c6c0000ULL,
    0x9ac80e0860000ULL,
    0xbf2a231630000ULL,
    0xb85bfd81b8000ULL,
    0x3a26a4e4bc000ULL,
    0x721016292e000ULL,
    0x534c0f23f5000ULL,
    0xcbe80b954b800ULL,
    0x64babbd5d7c00ULL,
    0xefd6059168600ULL,
    0x252b98cef5500ULL,
    0x5f5a780acb180ULL,
    0x59a37c2596940ULL,
    0x49d18e08c8fe0ULL,
    0x602b6304248f0ULL,
    0xefd99dbb82fb8ULL,
    0x3fe3f4d8f28ccULL,
    0x20726e3b6b796ULL,
    0x81fb732b47e23ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0xc400000000000ULL,
    0xb600000000000ULL,
    0x7d00000000000ULL,
    0x5180000000000ULL,
    0xfd40000000000ULL,
    0xb20000000000ULL,
    0xb210000000000ULL,
    0xd408000000000ULL,
    0x1634000000000ULL,
    0x4c9a000000000ULL,
    0x7bff000000000ULL,
    0x314a800000000ULL,
    0x9911c00000000ULL,
    0x288ba00000000ULL,
    0x42f2f00000000ULL,
    0xeeff880000000ULL,
    0xeccadc0000000ULL,
    0xee55120000000ULL,
    0xbca94d0000000ULL,
    0xf266be8000000ULL,
    0x2f346f4000000ULL,
    0xeb189a2000000ULL,
    0x9bb8911000000ULL,
    0xfbeb2c9800000ULL,
    0x4e47e26400000ULL,
    0x33a5048600000ULL,
    0x50d0ce6100000ULL,
    0xb6d1e9580000ULL,
    0xd0869f6b40000ULL,
    0x6ec7120a60000ULL,
    0x55624d3a50000ULL,
    0x59b63e9148000ULL,
    0x20daaf66c4000ULL,
    0x3b59ba0ae2000ULL,
    0x791ba1049d000ULL,
    0x80bf04b00e800ULL,
    0x2f6fce766bc00ULL,
    0xb6879ea9f8a00ULL,
    0xebc75f511d700ULL,
    0x50e4b23b4e680ULL,
    0x8af4bd0c0a7c0ULL,
    0x92fbb6bba9c20ULL,
    0x6ecb736c54070ULL,
    0x2d5428078b8c8ULL,
    0xcc282c15e95f4ULL,
    0x2ca29a15b405aULL,
    0x125791079a60dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x4800000000000ULL,
    0x2c00000000000ULL,
    0x1e00000000000ULL,
    0xf100000000000ULL,
    0x1580000000000ULL,
    0x4a40000000000ULL,
    0xd6e0000000000ULL,
    0x6050000000000ULL,
    0x38e8000000000ULL,
    0x25b4000000000ULL,
    0xa3a000000000ULL,
    0xaf99000000000ULL,
    0x60aa800000000ULL,
    0xcc51400000000ULL,
    0x66eae00000000ULL,
    0x74b7300000000ULL,
    0x8fba280000000ULL,
    0xadda7c0000000ULL,
    0x9a4eca0000000ULL,
    0xb200230000000ULL,
    0xaf04948000000ULL,
    0x4486be4000000ULL,
    0xcfc6e22000000ULL,
    0xd4a75f1000000ULL,
    0x9ab2dea800000ULL,
    0x46badd6400000ULL,
    0xec5816b200000ULL,
    0x3a0e915b00000ULL,
    0xefe6748f80000ULL,
    0x19d58e4840000ULL,
    0xccaeca1f60000ULL,
    0x9250232d90000ULL,
    0x37ec948838000ULL,
    0xf132be4b54000ULL,
    0x8dfce23eda000ULL,
    0x573e5f3cbd000ULL,
    0xe4185e9dd6800ULL,
    0x7beb9d7922400ULL,
    0x9f32f694c8a00ULL,
    0x4f9a16bfd500ULL,
    0xb6bc5caf36380ULL,
    0xd45ff26ff28c0ULL,
    0x6e08002810aa0ULL,
    0x5e40017798f0ULL,
    0x92d2002db52a8ULL,
    0x1a2d00220bd4cULL,
    0x2290801765b72ULL,
    0x4fc84038d3ac9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0x5c00000000000ULL,
    0xce00000000000ULL,
    0xef00000000000ULL,
    0x5f80000000000ULL,
    0x840000000000ULL,
    0x8260000000000ULL,
    0xe410000000000ULL,
    0xcd98000000000ULL,
    0xa9f4000000000ULL,
    0xf9ce000000000ULL,
    0xb20b000000000ULL,
    0xd62d800000000ULL,
    0x4f5a400000000ULL,
    0x2150600000000ULL,
    0xf97fb00000000ULL,
    0xb926580000000ULL,
    0x37f0ac0000000ULL,
    0x2ecbea0000000ULL,
    0x498e870000000ULL,
    0x5469d68000000ULL,
    0x583859c000000ULL,
    0xe7c0322000000ULL,
    0x42246b3000000ULL,
    0x2a725cb800000ULL,
    0xff896efc00000ULL,
    0x3f6fbc9a00000ULL,
    0x9dbc9ee700000ULL,
    0x628184ab80000ULL,
    0x56c782f7c0000ULL,
    0x2aa236b5e0000ULL,
    0x29b2a9d910000ULL,
    0x17ad8a3118000ULL,
    0x901e3719c4000ULL,
    0x58b40e9506000ULL,
    0x9f2db5f8ad000ULL,
    0x63d838208b800ULL,
    0x58941c0b56c00ULL,
    0x6c58323a15600ULL,
    0x68d06b057f900ULL,
    0x4a3c5c8839980ULL,
    0x6c26ed4b42c0ULL,
    0x12a23ca5cf760ULL,
    0x8db6deea117b0ULL,
    0x9da9e4b498028ULL,
    0x51c32f7051dcULL,
    0x7a326eba671baULL,
    0xe9ed05f8fd817ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0xa800000000000ULL,
    0xcc00000000000ULL,
    0xf600000000000ULL,
    0x3700000000000ULL,
    0x6f80000000000ULL,
    0xa140000000000ULL,
    0xb6e0000000000ULL,
    0xfb50000000000ULL,
    0xb818000000000ULL,
    0x9a34000000000ULL,
    0x460e000000000ULL,
    0xcecf000000000ULL,
    0x766e800000000ULL,
    0xc05c400000000ULL,
    0x2053e00000000ULL,
    0x119b700000000ULL,
    0x5471180000000ULL,
    0x6b6d6c0000000ULL,
    0x6df220000000ULL,
    0x7590230000000ULL,
    0xfbfe48000000ULL,
    0x278259c000000ULL,
    0x3d40ba2000000ULL,
    0x88e10f3000000ULL,
    0x9053269800000ULL,
    0x89990ae400000ULL,
    0xc076468a00000ULL,
    0x696a3ac300000ULL,
    0xa3dabe8580000ULL,
    0x171326d9c0000ULL,
    0x7a0489e0000ULL,
    0x68a229cf50000ULL,
    0xbe31220238000ULL,
    0xdc0f231414000ULL,
    0xffc964ba66000ULL,
    0x16ea19d41d000ULL,
    0x6a9d5a105b800ULL,
    0xdf57f2bf7c00ULL,
    0x9d2cbea045a00ULL,
    0x4fb826e9de100ULL,
    0x8782848bc9780ULL,
    0xcd4569c4dc1c0ULL,
    0xb0e2421f3a660ULL,
    0xf4571318e68b0ULL,
    0xb39d1c825d108ULL,
    0x17445c37b04cULL,
    0x31ee8008c7672ULL,
    0x6d1c4024edc17ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x1800000000000ULL,
    0xa400000000000ULL,
    0x3600000000000ULL,
    0xeb00000000000ULL,
    0xc380000000000ULL,
    0x4bc0000000000ULL,
    0xfca0000000000ULL,
    0xc730000000000ULL,
    0xedb8000000000ULL,
    0xac14000000000ULL,
    0xf4a000000000ULL,
    0x5209000000000ULL,
    0xe1ed800000000ULL,
    0x83bc00000000ULL,
    0x84d5a00000000ULL,
    0x5c6df00000000ULL,
    0x24f8380000000ULL,
    0x91f7cc0000000ULL,
    0xd99e560000000ULL,
    0x3360b10000000ULL,
    0x3d95d38000000ULL,
    0x798b72c000000ULL,
    0x562bee2000000ULL,
    0xc69cbd1000000ULL,
    0xeee625a800000ULL,
    0x815233dc00000ULL,
    0xa4ac05a600000ULL,
    0x1fd903f900000ULL,
    0x5b061d8580000ULL,
    0x3b81fff6c0000ULL,
    0x3fc053bee0000ULL,
    0xa2a4b2f7f0000ULL,
    0x90344e3ca8000ULL,
    0xbc384d2e0c000ULL,
    0x3ad39da80e000ULL,
    0xdb6e3ff635000ULL,
    0x1d7ff3a79a800ULL,
    0xbb3042f47e400ULL,
    0x9fb9f6095d600ULL,
    0xa110413cf7f00ULL,
    0x4fca6bb841e80ULL,
    0x664e7efea03c0ULL,
    0xc48d981d51ee0ULL,
    0xeeaa3c2bf8430ULL,
    0x86de6e02f51d8ULL,
    0x95837d127ae04ULL,
    0x30c185950ff2eULL,
    0xf722c3f1b4e3bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0xc400000000000ULL,
    0x6200000000000ULL,
    0x1d00000000000ULL,
    0xbc80000000000ULL,
    0x3ec0000000000ULL,
    0xebe0000000000ULL,
    0x3f90000000000ULL,
    0x28a8000000000ULL,
    0xdcf4000000000ULL,
    0x93fa000000000ULL,
    0x5ebd000000000ULL,
    0xd21a800000000ULL,
    0x482e400000000ULL,
    0xbc35a00000000ULL,
    0xcb1dd00000000ULL,
    0xf6ae080000000ULL,
    0x2ff3140000000ULL,
    0x247e6e0000000ULL,
    0x1bfee30000000ULL,
    0x4abdd68000000ULL,
    0xb81a804000000ULL,
    0x12ae62000000ULL,
    0x4ab3b71000000ULL,
    0x2cde188800000ULL,
    0xc3cdb36400000ULL,
    0x33a338aa00000ULL,
    0x5bb7236100000ULL,
    0xc65810b080000ULL,
    0x1a0aa77340000ULL,
    0x5c7568d20000ULL,
    0x6864c04ab0000ULL,
    0x80d7460078000ULL,
    0xd10a670da4000ULL,
    0x1e4210bf8a000ULL,
    0xab27a76acd000ULL,
    0xf775d683a6800ULL,
    0xa8be8071bb400ULL,
    0xe518e61bf4a00ULL,
    0x9daab71622900ULL,
    0xe476989601e80ULL,
    0xef3af34701340ULL,
    0x385e18af81a20ULL,
    0x790db350c1150ULL,
    0x9a4338a3e0d38ULL,
    0xe927237a511ccULL,
    0x7a7010a7a9bfaULL,
    0x3c3ea757cdc11ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x6800000000000ULL,
    0xa400000000000ULL,
    0x3600000000000ULL,
    0x2f00000000000ULL,
    0x6f80000000000ULL,
    0x5c0000000000ULL,
    0x40a0000000000ULL,
    0xc9b0000000000ULL,
    0xd258000000000ULL,
    0x8a94000000000ULL,
    0x9c2a000000000ULL,
    0x396f000000000ULL,
    0x638f800000000ULL,
    0xe75e400000000ULL,
    0x2811a00000000ULL,
    0xe76a700000000ULL,
    0xa889980000000ULL,
    0xdedb2c0000000ULL,
    0x32d3860000000ULL,
    0xa04c110000000ULL,
    0xc0f8178000000ULL,
    0x7a275dc000000ULL,
    0x5ef59e2000000ULL,
    0x81b97d1000000ULL,
    0x76c2319800000ULL,
    0xc5253ccc00000ULL,
    0x497611ba00000ULL,
    0x187e0cf900000ULL,
    0xfc61a9b580000ULL,
    0xd99110d7c0000ULL,
    0x49aa17b9e0000ULL,
    0xe1ac5df470000ULL,
    0x65a81e3fa8000ULL,
    0x3ac3d0cac000ULL,
    0x74ae11a0f6000ULL,
    0xd72a0ccd15000ULL,
    0x40eba99b42800ULL,
    0x194e10cf87c00ULL,
    0x907d97923de00ULL,
    0x28661dffae900ULL,
    0xe793be3050e80ULL,
    0xf2a94d13385c0ULL,
    0xd02809bef5060ULL,
    0xeb6f60fe339b0ULL,
    0x1a898fa1afe98ULL,
    0x97d871f750c84ULL,
    0x2a541811b96beULL,
    0x60e6c03b5d07ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xf800000000000ULL,
    0x6400000000000ULL,
    0x5e00000000000ULL,
    0x900000000000ULL,
    0xff80000000000ULL,
    0x9bc0000000000ULL,
    0xffe0000000000ULL,
    0x4ef0000000000ULL,
    0x2f38000000000ULL,
    0x294000000000ULL,
    0x980a000000000ULL,
    0x114b000000000ULL,
    0x94ee800000000ULL,
    0x573ec00000000ULL,
    0x2692a00000000ULL,
    0x2608b00000000ULL,
    0x84f280000000ULL,
    0x936f6c0000000ULL,
    0xa8fb860000000ULL,
    0x87701b0000000ULL,
    0x61f9ad8000000ULL,
    0xd8f28cc000000ULL,
    0xa3a2e2000000ULL,
    0xcf11b71000000ULL,
    0xd8c88bb800000ULL,
    0xda2e27c400000ULL,
    0x8d5eab9e00000ULL,
    0xb4a357c900000ULL,
    0xf855a38f80000ULL,
    0x446e4bd640000ULL,
    0x1e79ad99e0000ULL,
    0x3328cebf0000ULL,
    0x15da2e0e78000ULL,
    0x91e1b71cfc000ULL,
    0xff08ba5ae000ULL,
    0xbcba27f915000ULL,
    0x4b54abaf9a800ULL,
    0xace857fa5ec00ULL,
    0x933b23a16da00ULL,
    0x88908bf666b00ULL,
    0xc70b0d8270280ULL,
    0x6bca3cf539ec0ULL,
    0x32ad06049cea0ULL,
    0x1adb291fab0ULL,
    0x3f010d850d698ULL,
    0x2a813cc0d65acULL,
    0xbe43861968296ULL,
    0x23241b34f4243ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x7800000000000ULL,
    0x5c00000000000ULL,
    0x6a00000000000ULL,
    0x100000000000ULL,
    0x8280000000000ULL,
    0x94c0000000000ULL,
    0xaa0000000000ULL,
    0xf10000000000ULL,
    0xdc38000000000ULL,
    0xd34000000000ULL,
    0xf6ea000000000ULL,
    0x412f000000000ULL,
    0xcc09800000000ULL,
    0x735ec00000000ULL,
    0xd200600000000ULL,
    0x7d06100000000ULL,
    0x7885a80000000ULL,
    0x1dc1e40000000ULL,
    0xac26de0000000ULL,
    0xadd5f90000000ULL,
    0xbd9b928000000ULL,
    0x81a20ec000000ULL,
    0x3c92f62000000ULL,
    0xd27edd1000000ULL,
    0x3f572c9800000ULL,
    0xe15ee7fc00000ULL,
    0xcf054cbe00000ULL,
    0xf583f7c700000ULL,
    0x1d4364a280000ULL,
    0xede3d3ec40000ULL,
    0x6bf45a86a0000ULL,
    0x114afade30000ULL,
    0xbeb8003e28000ULL,
    0x29f4003884000ULL,
    0xe44a000b2e000ULL,
    0xe23f000d3d000ULL,
    0x231801b32800ULL,
    0x236ac021ffc00ULL,
    0xcc6a603c84600ULL,
    0xa9e910392eb00ULL,
    0x3cac281c3c980ULL,
    0xf54f242ab3440ULL,
    0xa8bebe253f8a0ULL,
    0xd2f7e90b65090ULL,
    0xefcc3aa63f4f8ULL,
    0xd078eafbe58a4ULL,
    0xaa57a82b7f76aULL,
    0x4ddae424c50ebULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x8800000000000ULL,
    0x1400000000000ULL,
    0xa200000000000ULL,
    0x4900000000000ULL,
    0x4a80000000000ULL,
    0xc340000000000ULL,
    0xfe60000000000ULL,
    0xc5b0000000000ULL,
    0x84b8000000000ULL,
    0xab54000000000ULL,
    0xf74e000000000ULL,
    0xfd0f000000000ULL,
    0x89ef800000000ULL,
    0xef1d400000000ULL,
    0x25c4200000000ULL,
    0xc321d00000000ULL,
    0xa4d0280000000ULL,
    0x508e240000000ULL,
    0x6729a20000000ULL,
    0x3fb85f0000000ULL,
    0x50d43b8000000ULL,
    0x828d50c000000ULL,
    0xc62e0a2000000ULL,
    0xd13f3b3000000ULL,
    0xb997b99800000ULL,
    0x21ebdfec00000ULL,
    0xeb1d998200000ULL,
    0xdfc50ff500000ULL,
    0xa622319080000ULL,
    0xd0566bce40000ULL,
    0x6ccfb388a0000ULL,
    0x38cfe4e9d0000ULL,
    0x3acba00d18000ULL,
    0xa3cc903a2c000ULL,
    0xd14c08210e000ULL,
    0x8e0bf42145000ULL,
    0xc66f8a1a6b800ULL,
    0xe8dd7b29bfc00ULL,
    0x6fe41986ede00ULL,
    0x4d734fc2ee500ULL,
    0x871f9198b4880ULL,
    0x41c1fbd4e2b40ULL,
    0x89223b902a6a0ULL,
    0x49d650d71e190ULL,
    0x300f8a343d008ULL,
    0xf96d7b0877f3cULL,
    0x295c19b9586ceULL,
    0x1f274ffb8c393ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xe800000000000ULL,
    0xec00000000000ULL,
    0x6200000000000ULL,
    0x4f00000000000ULL,
    0xc680000000000ULL,
    0xe140000000000ULL,
    0x8a20000000000ULL,
    0x88f0000000000ULL,
    0xc428000000000ULL,
    0xea54000000000ULL,
    0xd09e000000000ULL,
    0x55db000000000ULL,
    0xe1f9800000000ULL,
    0x4c0a400000000ULL,
    0x9da1e00000000ULL,
    0xda36d00000000ULL,
    0xc2ced80000000ULL,
    0x3142840000000ULL,
    0x3220420000000ULL,
    0x1cf36d0000000ULL,
    0xd22f918000000ULL,
    0xc352764000000ULL,
    0xd71f1a2000000ULL,
    0x5f1fa93000000ULL,
    0x31833b800000ULL,
    0x6918cb5c00000ULL,
    0xba19d39a00000ULL,
    0xf69e1b4b00000ULL,
    0xb4df0b8080000ULL,
    0x9a789f7fc0000ULL,
    0xbcc94993a0000ULL,
    0x3444f26f50000ULL,
    0xb7a1581808000ULL,
    0x2937c426fc000ULL,
    0x1e4e221a2a000ULL,
    0xeb00fd3605000ULL,
    0x188729858f800ULL,
    0xb4476267dac00ULL,
    0x77a1e022dc600ULL,
    0xc936d0103b900ULL,
    0x6e4ed8096df80ULL,
    0x302841f02740ULL,
    0xf480421568560ULL,
    0xd6436d298df10ULL,
    0x38a7918e33888ULL,
    0xfb6765e903f4ULL,
    0x8f091a3ea9ca6ULL,
    0x8920a9042dd89ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x6800000000000ULL,
    0x6c00000000000ULL,
    0x3200000000000ULL,
    0xad00000000000ULL,
    0x3580000000000ULL,
    0x1240000000000ULL,
    0x9e20000000000ULL,
    0x2490000000000ULL,
    0xa418000000000ULL,
    0x5934000000000ULL,
    0x11ce000000000ULL,
    0xc18f000000000ULL,
    0x3aaf800000000ULL,
    0x97ba400000000ULL,
    0xcbe7e00000000ULL,
    0xa5f2500000000ULL,
    0xfb2bc80000000ULL,
    0x13f99c0000000ULL,
    0x76c64a0000000ULL,
    0xf3e7c50000000ULL,
    0xd1f2628000000ULL,
    0xad2b47c000000ULL,
    0x50fa022000000ULL,
    0xb440193000000ULL,
    0x1525c89800000ULL,
    0x5a15d2ec00000ULL,
    0xa8da28ba00000ULL,
    0x7bd382eb00000ULL,
    0x7fbfe0b480000ULL,
    0x67e51ef740000ULL,
    0xf7f62aaf60000ULL,
    0xe6289bc390000ULL,
    0x4e7ba80908000ULL,
    0x8858c2564000ULL,
    0x5fc4622ae6000ULL,
    0x5863090e9d000ULL,
    0x3cb060b81b800ULL,
    0x1b8f5ecd97c00ULL,
    0x3ba9caa1c5e00ULL,
    0xf03ecbe557900ULL,
    0xc4a6600d64780ULL,
    0x6657101ee6540ULL,
    0xa1fba8389c920ULL,
    0x1bc58c391a370ULL,
    0xa664621a160a8ULL,
    0x73b3092605354ULL,
    0x5b0860b377f2aULL,
    0x186b5efc9572fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0x5c00000000000ULL,
    0x9200000000000ULL,
    0xe500000000000ULL,
    0x7580000000000ULL,
    0xddc0000000000ULL,
    0xb7a0000000000ULL,
    0xfe90000000000ULL,
    0x50f8000000000ULL,
    0xa9b4000000000ULL,
    0xae2a000000000ULL,
    0x2f2b000000000ULL,
    0x2ca9800000000ULL,
    0x226e400000000ULL,
    0xd34c600000000ULL,
    0x469d500000000ULL,
    0x2284080000000ULL,
    0x1d401c0000000ULL,
    0xe7e0760000000ULL,
    0xd0f7890000000ULL,
    0x3fccae8000000ULL,
    0x2cdf20c000000ULL,
    0x55e7fe2000000ULL,
    0x65f2d51000000ULL,
    0xb24938b800000ULL,
    0xb51bb9fc00000ULL,
    0xc443389a00000ULL,
    0xb060b9c700000ULL,
    0xe032b8bc80000ULL,
    0x51eaf9d540000ULL,
    0x758cd8bf60000ULL,
    0xbb8a9cc70000ULL,
    0xd55350ab68000ULL,
    0x7459f5fde4000ULL,
    0xf1a4c68fde000ULL,
    0x65926cdf5d000ULL,
    0x5a7b801d82800ULL,
    0x48f140015ec00ULL,
    0xcbcfe0091c600ULL,
    0x2d8103862900ULL,
    0xd2e1e8256f980ULL,
    0x1d730c1694940ULL,
    0x46081e2837e20ULL,
    0x4d7ac51ff8570ULL,
    0x6870d0893de58ULL,
    0x6b8cb5c472e64ULL,
    0xc4b9268af7d1aULL,
    0xe1d57cdb58cb3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x8800000000000ULL,
    0x3c00000000000ULL,
    0xa600000000000ULL,
    0xad00000000000ULL,
    0x6780000000000ULL,
    0xdbc0000000000ULL,
    0xd4a0000000000ULL,
    0x5e50000000000ULL,
    0x58000000000ULL,
    0x7074000000000ULL,
    0xda4a000000000ULL,
    0x3b4f000000000ULL,
    0xa2ca800000000ULL,
    0x9400000000ULL,
    0xb928600000000ULL,
    0xbfb8f00000000ULL,
    0x9dc1480000000ULL,
    0xa9a62c0000000ULL,
    0xd1d4a60000000ULL,
    0x779ce10000000ULL,
    0x8ad6e88000000ULL,
    0x1518fdc000000ULL,
    0xfa956e2000000ULL,
    0xa47c8d1000000ULL,
    0x6960ae9800000ULL,
    0x7f5ace400000ULL,
    0xc98aae8200000ULL,
    0xb3eaacdf00000ULL,
    0xca982e9180000ULL,
    0x2557ecdf40000ULL,
    0x92da4e83e0000ULL,
    0x18301ccdf0000ULL,
    0xdf298686a8000ULL,
    0x32bb70c7d4000ULL,
    0xa4740aa76000ULL,
    0x6a6461faad000ULL,
    0xb1706029da800ULL,
    0xb3ccf03ff7c00ULL,
    0x818b4817eca00ULL,
    0xafe92c3d3b500ULL,
    0x9c9e263f86980ULL,
    0x9055a13804fc0ULL,
    0x415e88bc0e720ULL,
    0x59f00df2811b0ULL,
    0x8c2608c1118ULL,
    0xa66ea111a1944ULL,
    0xb65e089891d02ULL,
    0x15764de199187ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0x9400000000000ULL,
    0x8200000000000ULL,
    0xa100000000000ULL,
    0x1380000000000ULL,
    0x6940000000000ULL,
    0x9020000000000ULL,
    0x7d70000000000ULL,
    0xa758000000000ULL,
    0x6bd4000000000ULL,
    0x52ea000000000ULL,
    0x8cef000000000ULL,
    0x2bec800000000ULL,
    0x536c400000000ULL,
    0x4cab200000000ULL,
    0xc4ad00000000ULL,
    0xbd5f680000000ULL,
    0x66d6ec0000000ULL,
    0x1f6c320000000ULL,
    0x2ad650000000ULL,
    0xa348458000000ULL,
    0x61db95c000000ULL,
    0x8315da2000000ULL,
    0x4288c91000000ULL,
    0xb63bd79800000ULL,
    0x328460d400000ULL,
    0x3ac3d7aa00000ULL,
    0x996060f500000ULL,
    0xbd51d79880000ULL,
    0xf22b60ec40000ULL,
    0x588f57b6e0000ULL,
    0xbb3c20fb50000ULL,
    0x7f02f7b688000ULL,
    0xb485b0d61c000ULL,
    0x11c23fbfb6000ULL,
    0x8fe1ccc9e5000ULL,
    0x3d9045a6d2800ULL,
    0x164f95ef77c00ULL,
    0xb05fda1145200ULL,
    0x2b57c90262100ULL,
    0x912f57bc2e480ULL,
    0x8a0c20e690740ULL,
    0x91faf7b2a87a0ULL,
    0xae61b0ecec6b0ULL,
    0xcdd03f9b0e648ULL,
    0x49eaccc7e03d4ULL,
    0xe26ec5b1d0a36ULL,
    0x1728d5cbc83a7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xb800000000000ULL,
    0x9400000000000ULL,
    0xa00000000000ULL,
    0xb00000000000ULL,
    0x480000000000ULL,
    0x7cc0000000000ULL,
    0x220000000000ULL,
    0x9d0000000000ULL,
    0xb398000000000ULL,
    0xe074000000000ULL,
    0xd00a000000000ULL,
    0xe20b000000000ULL,
    0x3d0b800000000ULL,
    0xd38cc00000000ULL,
    0x1448200000000ULL,
    0xaae9100000000ULL,
    0x4bfdb80000000ULL,
    0xf045940000000ULL,
    0x12e20e0000000ULL,
    0x75f3dd0000000ULL,
    0xb4f648000000ULL,
    0xe46b27c000000ULL,
    0xdc3e362000000ULL,
    0x46e1891000000ULL,
    0xdff6cab800000ULL,
    0x50492aec00000ULL,
    0xe8eeca8e00000ULL,
    0x8cfd2af500000ULL,
    0xdac4ca9a80000ULL,
    0xd7262afec0000ULL,
    0xec574ab720000ULL,
    0x70deeac630000ULL,
    0x22156ab408000ULL,
    0x333cfad9b4000ULL,
    0x5063529946000ULL,
    0x6c35aefced000ULL,
    0x54e97c812a800ULL,
    0xbaff63fe06c00ULL,
    0xa7c3a0168ce00ULL,
    0x4ea5d00a7a300ULL,
    0xb9159808de880ULL,
    0x9fbc843ca1840ULL,
    0x50a7b628710e0ULL,
    0xf8124932e9950ULL,
    0x703f6a9524eb8ULL,
    0x98e7fadc3e434ULL,
    0x3ef0d2a330df2ULL,
    0xefcd6ed68918fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0xc800000000000ULL,
    0x2400000000000ULL,
    0x6200000000000ULL,
    0x700000000000ULL,
    0x1380000000000ULL,
    0xec40000000000ULL,
    0xa8a0000000000ULL,
    0x6df0000000000ULL,
    0x2d08000000000ULL,
    0x5dd4000000000ULL,
    0xe93e000000000ULL,
    0x793b000000000ULL,
    0x913c800000000ULL,
    0x853b400000000ULL,
    0xff3aa00000000ULL,
    0x143ed00000000ULL,
    0x41bcd80000000ULL,
    0xc8ff6c0000000ULL,
    0x74d8aa0000000ULL,
    0xe6ea110000000ULL,
    0x8f03598000000ULL,
    0x17801b4000000ULL,
    0xbe46f22000000ULL,
    0xb7a53d3000000ULL,
    0x9275d3b800000ULL,
    0x874b9a7400000ULL,
    0x9075d39200000ULL,
    0x904b9a7900000ULL,
    0xabf5d38080000ULL,
    0x880b9a45c0000ULL,
    0xa955d3aee0000ULL,
    0xc6fb9a7610000ULL,
    0xf5ddd398a8000ULL,
    0x706f9a57bc000ULL,
    0xa7c3d3bd22000ULL,
    0x88e49a60a5000ULL,
    0xb35753a607800ULL,
    0x3dfbda72e2c00ULL,
    0x885bf3b345200ULL,
    0xd2a0a5a97d00ULL,
    0xb1e5ab928b280ULL,
    0x39d5266098740ULL,
    0xab3b219965960ULL,
    0x4e3aa776e6cb0ULL,
    0x9abe002bf2228ULL,
    0x857b00386c204ULL,
    0x119c80256be86ULL,
    0x1ccb40060979dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xa800000000000ULL,
    0x3c00000000000ULL,
    0xe200000000000ULL,
    0x5100000000000ULL,
    0x8480000000000ULL,
    0xd140000000000ULL,
    0x29a0000000000ULL,
    0xe290000000000ULL,
    0xbfa8000000000ULL,
    0xb4000000000ULL,
    0x5bfa000000000ULL,
    0x279000000000ULL,
    0x5dbf800000000ULL,
    0x685e400000000ULL,
    0xf5eb600000000ULL,
    0x8090d00000000ULL,
    0xaea8a80000000ULL,
    0x2437140000000ULL,
    0xbabc320000000ULL,
    0x83dd510000000ULL,
    0x832e428000000ULL,
    0x35f5c64000000ULL,
    0xa4591a2000000ULL,
    0x5fe9051000000ULL,
    0x7d9490a800000ULL,
    0x502b074c00000ULL,
    0x3071108a00000ULL,
    0xc99c475f00000ULL,
    0x30df0b080000ULL,
    0xc0a6d74840000ULL,
    0x8214388320000ULL,
    0xb8e8136750000ULL,
    0x96172290e8000ULL,
    0x26e8167fe4000ULL,
    0x851432144a000ULL,
    0xc36951333d000ULL,
    0x78d4428736800ULL,
    0x78cc65b2a400ULL,
    0x51e69a214de00ULL,
    0xbb7451d8e900ULL,
    0x6a7ff0bfa6580ULL,
    0x81bbd77423740ULL,
    0x1a59b8a4994a0ULL,
    0x3ceb53405d150ULL,
    0x9011c28746b48ULL,
    0xa1eb866a93b04ULL,
    0xbe927a1b4002aULL,
    0x8da9d519a0509ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x6800000000000ULL,
    0x6c00000000000ULL,
    0x9200000000000ULL,
    0x6d00000000000ULL,
    0x9480000000000ULL,
    0x51c0000000000ULL,
    0x2560000000000ULL,
    0xcb50000000000ULL,
    0xd968000000000ULL,
    0x1b4000000000ULL,
    0x847e000000000ULL,
    0xe7fd000000000ULL,
    0x73bf800000000ULL,
    0xf998400000000ULL,
    0x9f6de00000000ULL,
    0xaab1500000000ULL,
    0x7bf9c80000000ULL,
    0x69b8940000000ULL,
    0x689ad60000000ULL,
    0x41e9570000000ULL,
    0x9272688000000ULL,
    0x341ff24000000ULL,
    0xc2a9e2000000ULL,
    0x6d10833000000ULL,
    0x4e4cde8800000ULL,
    0xbbc6b56c00000ULL,
    0x3c655eba00000ULL,
    0xe9d7f56b00000ULL,
    0xaba93e8c80000ULL,
    0x2753e54740000ULL,
    0x8b6a969720000ULL,
    0x4cb6615370000ULL,
    0xa0fa689fd8000ULL,
    0xde3bf26e94000ULL,
    0x3adc9e05fa000ULL,
    0xa0c983085d000ULL,
    0x2b055ebbe7800ULL,
    0x3f87f5731b400ULL,
    0xae413e984de00ULL,
    0xab27e568ef100ULL,
    0xd074969066f80ULL,
    0x611b6172da140ULL,
    0xcade88b2c120ULL,
    0x1ad7b27c3e750ULL,
    0x902f7e298fad8ULL,
    0x7715d313b651cULL,
    0xdf4b1685b3f02ULL,
    0x6543214201dd9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0x9c00000000000ULL,
    0xf600000000000ULL,
    0x4900000000000ULL,
    0x2080000000000ULL,
    0x140000000000ULL,
    0x84a0000000000ULL,
    0x88b0000000000ULL,
    0x9df8000000000ULL,
    0xa794000000000ULL,
    0x1f8a000000000ULL,
    0x440f000000000ULL,
    0x66c8800000000ULL,
    0xcca9c00000000ULL,
    0xe07f600000000ULL,
    0x5450f00000000ULL,
    0x96de80000000ULL,
    0x9f9f140000000ULL,
    0x1f45520000000ULL,
    0x39a1f90000000ULL,
    0xea31d38000000ULL,
    0xf7be7cc000000ULL,
    0xb8b23a2000000ULL,
    0x95fc2d1000000ULL,
    0xc391619800000ULL,
    0xa58db5f400000ULL,
    0x630be18600000ULL,
    0x934f75f700000ULL,
    0x1b6e018580000ULL,
    0x2c9d45f640000ULL,
    0x78c6099620000ULL,
    0x1d6061c350000ULL,
    0x21535382a8000ULL,
    0x2fe8bcf36c000ULL,
    0xffddda0cde000ULL,
    0x6f611d38f5000ULL,
    0x4251e9ad6a800ULL,
    0xd06951f23fc00ULL,
    0x4719dbac44200ULL,
    0x520358ddf3700ULL,
    0xd30760340b580ULL,
    0xb784f039cefc0ULL,
    0xbcc7e82c7dda0ULL,
    0x5760141d57750ULL,
    0x2e55d23c482c8ULL,
    0xe6c3907dcf24ULL,
    0xda1cb38c4676eULL,
    0x80858cd6c0d63ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x5800000000000ULL,
    0x7c00000000000ULL,
    0x2a00000000000ULL,
    0x7300000000000ULL,
    0xe280000000000ULL,
    0xd640000000000ULL,
    0x58e0000000000ULL,
    0x1ef0000000000ULL,
    0x1128000000000ULL,
    0x3054000000000ULL,
    0x673a000000000ULL,
    0x178f000000000ULL,
    0x7246800000000ULL,
    0x4ee3c00000000ULL,
    0x27f0600000000ULL,
    0x90ab100000000ULL,
    0x5c96b80000000ULL,
    0x959fa40000000ULL,
    0x7b9b7a0000000ULL,
    0xe98cb0000000ULL,
    0x8d1e918000000ULL,
    0xa6da9c4000000ULL,
    0xb8fda22000000ULL,
    0x27ac7f1000000ULL,
    0xb813539800000ULL,
    0x22ddf34400000ULL,
    0xdef849b600000ULL,
    0x56ae284f00000ULL,
    0x1d90603280000ULL,
    0x181b1008c0000ULL,
    0x755eb836a0000ULL,
    0xab3ba42e90000ULL,
    0x55897a0ab8000ULL,
    0x5543cb047c000ULL,
    0xb26211b66a000ULL,
    0x8cb65c6545000ULL,
    0xfcb421756800ULL,
    0x2fa4af27cfc00ULL,
    0x9b958b96e3a00ULL,
    0x3919475c11f00ULL,
    0xe8dd8ba278980ULL,
    0xfdfd4762dc040ULL,
    0x8c2f8ba0fa8e0ULL,
    0xa7d6477bfcd90ULL,
    0x32fb0b9d2b668ULL,
    0x64ae874aa406cULL,
    0x7297eb8f26126ULL,
    0xc0995778878a1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x7800000000000ULL,
    0x7400000000000ULL,
    0x9e00000000000ULL,
    0x6700000000000ULL,
    0x1780000000000ULL,
    0xb240000000000ULL,
    0xd8e0000000000ULL,
    0xc310000000000ULL,
    0xc588000000000ULL,
    0x91b4000000000ULL,
    0x8a7e000000000ULL,
    0x992b000000000ULL,
    0xb141800000000ULL,
    0x9962c00000000ULL,
    0xfa57200000000ULL,
    0x90ecb00000000ULL,
    0x61e7580000000ULL,
    0x9391ec0000000ULL,
    0xf04d5a0000000ULL,
    0xf312c30000000ULL,
    0x9d8ef68000000ULL,
    0x75b5d94000000ULL,
    0x1c7d222000000ULL,
    0xf22f9f3000000ULL,
    0x4cc4f4b800000ULL,
    0xd226f65c00000ULL,
    0x52368ebe00000ULL,
    0xf63c856d00000ULL,
    0xcec9203f80000ULL,
    0x19d7b035c0000ULL,
    0x7caed823e0000ULL,
    0x3e072c2d90000ULL,
    0x37047a0fb8000ULL,
    0xff85733404000ULL,
    0xce402e9302000ULL,
    0x4ae2f5513d000ULL,
    0x4e11583147800ULL,
    0x2b0eec0a1ec00ULL,
    0x5372da130a600ULL,
    0xf75b032ef1b00ULL,
    0x309856b368480ULL,
    0x6f3ba97b1cec0ULL,
    0xe4d5a1136360ULL,
    0x2412c30d360b0ULL,
    0x120ef69336fe8ULL,
    0xc3f5d94e372ccULL,
    0x229d2209b7a3aULL,
    0x223f9f0076215ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0x2400000000000ULL,
    0xc200000000000ULL,
    0x2100000000000ULL,
    0xeb80000000000ULL,
    0xb040000000000ULL,
    0x2920000000000ULL,
    0x5dd0000000000ULL,
    0x2a28000000000ULL,
    0xcb4000000000ULL,
    0x6ada000000000ULL,
    0xf149000000000ULL,
    0x1944800000000ULL,
    0x56a0400000000ULL,
    0x3796a00000000ULL,
    0xbe0bd00000000ULL,
    0xb4e3680000000ULL,
    0x37b26c0000000ULL,
    0xdc5f9a0000000ULL,
    0x54899d0000000ULL,
    0xff21708000000ULL,
    0xe6d3ba4000000ULL,
    0xdcaa522000000ULL,
    0x970211000000ULL,
    0xdcbd82a800000ULL,
    0xa9384b5400000ULL,
    0xd67cb89a00000ULL,
    0x23de065900000ULL,
    0x7ecca03880000ULL,
    0xcb02d00c40000ULL,
    0xe87e828e0000ULL,
    0x69c22c3110000ULL,
    0x613a00c8000ULL,
    0x53764d1634000ULL,
    0x21b81880ea000ULL,
    0xecb8d6677d000ULL,
    0xf139480a1e800ULL,
    0xea7dfc1736400ULL,
    0x7dd8522e93200ULL,
    0x51cd2115f1100ULL,
    0xe78302a7d9280ULL,
    0xb6410b59fdb40ULL,
    0x3a2698a15f7a0ULL,
    0x2f51967c56cb0ULL,
    0xbdeb680e436e8ULL,
    0x73d66c0718834ULL,
    0xfd2d9a075d2e2ULL,
    0x4c349d11ae819ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0x2400000000000ULL,
    0xde00000000000ULL,
    0xd100000000000ULL,
    0x2380000000000ULL,
    0xbc0000000000ULL,
    0x3d60000000000ULL,
    0x85f0000000000ULL,
    0x3828000000000ULL,
    0x754000000000ULL,
    0x36be000000000ULL,
    0x2f8f000000000ULL,
    0xe442800000000ULL,
    0x6ca0400000000ULL,
    0xc792a00000000ULL,
    0xf75b300000000ULL,
    0x9b3ee80000000ULL,
    0xfd4ccc0000000ULL,
    0xaea1260000000ULL,
    0x8493ff0000000ULL,
    0x9fdbad8000000ULL,
    0xac78944000000ULL,
    0x916d6e2000000ULL,
    0x1a74033000000ULL,
    0xd1ec639800000ULL,
    0xc933a74400000ULL,
    0x3c9e58a00000ULL,
    0xbce0687900000ULL,
    0xdd30200f80000ULL,
    0xc5cb700040000ULL,
    0x91e4483ae0000ULL,
    0xbcb3fc0ff0000ULL,
    0x4d09ce12d8000ULL,
    0xa404331c3c000ULL,
    0x1e060bb16a000ULL,
    0xb1042b7765000ULL,
    0x3386e3aa17800ULL,
    0xa3c7e7586fc00ULL,
    0x1965458042a00ULL,
    0x5bf4584f81100ULL,
    0xe92c482041280ULL,
    0x24d7fc0ae00c0ULL,
    0x3d7fce17f09a0ULL,
    0x12ef3316d8e30ULL,
    0x61b28bb63c418ULL,
    0x548f6b786a23cULL,
    0xc0c0c3a0e5f96ULL,
    0xc1e7974e57da1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0x9c00000000000ULL,
    0x2a00000000000ULL,
    0x8700000000000ULL,
    0xde80000000000ULL,
    0x40c0000000000ULL,
    0x4be0000000000ULL,
    0xee30000000000ULL,
    0xaa48000000000ULL,
    0x7f14000000000ULL,
    0x109e000000000ULL,
    0x19ef000000000ULL,
    0x9f03800000000ULL,
    0xe287c00000000ULL,
    0x3ac6600000000ULL,
    0xa4e2100000000ULL,
    0xf4b2180000000ULL,
    0x5c8b440000000ULL,
    0x99f36e0000000ULL,
    0xa72e930000000ULL,
    0x2de4838000000ULL,
    0xeb328b4000000ULL,
    0x57cf162000000ULL,
    0x6753c73000000ULL,
    0xeebbf59800000ULL,
    0x79b85c4c00000ULL,
    0x6f3b7bb600000ULL,
    0xbbf81f5700000ULL,
    0x2e1e003280000ULL,
    0xa92f001dc0000ULL,
    0xece38039a0000ULL,
    0xa0b7c03b30000ULL,
    0xe28e600e78000ULL,
    0xc0f6103d4c000ULL,
    0x10ac1817b2000ULL,
    0x82a4442e35000ULL,
    0x9390ee217f800ULL,
    0xeb59532dc7c00ULL,
    0xf68ae39faae00ULL,
    0xdef49b5ce8f00ULL,
    0x19ab0e0004180ULL,
    0x5d23830086440ULL,
    0xf8d51b86cbe60ULL,
    0x25fe0f65f8e10ULL,
    0xe71a18390da38ULL,
    0x16af442f52d0cULL,
    0xb7a56e2a643eeULL,
    0x3615932916f7dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x5800000000000ULL,
    0x6400000000000ULL,
    0xe200000000000ULL,
    0xfb00000000000ULL,
    0xc580000000000ULL,
    0x4f40000000000ULL,
    0x27a0000000000ULL,
    0x5b0000000000ULL,
    0x3dd8000000000ULL,
    0x3f54000000000ULL,
    0x8cca000000000ULL,
    0x8ebb000000000ULL,
    0x8f02800000000ULL,
    0x3f87c00000000ULL,
    0x3040a00000000ULL,
    0x9021500000000ULL,
    0xe9f6380000000ULL,
    0xbbfa1c0000000ULL,
    0x97a6b20000000ULL,
    0x6db2c90000000ULL,
    0x71db7b8000000ULL,
    0x715229c000000ULL,
    0xa9c82a2000000ULL,
    0x363d851000000ULL,
    0x1cc1f1b800000ULL,
    0x69e1fce400000ULL,
    0x989763b600000ULL,
    0xc0aaa5c900000ULL,
    0xc96a002080000ULL,
    0x300b000ac0000ULL,
    0x975a801f60000ULL,
    0xdf93c02ed0000ULL,
    0xc32aa02768000ULL,
    0x7f2a5026ac000ULL,
    0xb92cb80696000ULL,
    0x4029dc22c5000ULL,
    0xeeac121bba800ULL,
    0x3c68993112c00ULL,
    0x308fc3a55ea00ULL,
    0xf09ff5c158f00ULL,
    0x33f6380054a80ULL,
    0xf4fa1c1873940ULL,
    0x826b2028fae0ULL,
    0x2df2c926b1b70ULL,
    0x29fb7b9c39248ULL,
    0xa4a229e985164ULL,
    0x94302a329b78eULL,
    0xb899853862597ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x1800000000000ULL,
    0x5400000000000ULL,
    0xce00000000000ULL,
    0x8500000000000ULL,
    0xc280000000000ULL,
    0xebc0000000000ULL,
    0x9a60000000000ULL,
    0x6d50000000000ULL,
    0xdcb8000000000ULL,
    0x1af4000000000ULL,
    0x28e000000000ULL,
    0x439d000000000ULL,
    0xec06800000000ULL,
    0xca07c00000000ULL,
    0x7307200000000ULL,
    0x8382300000000ULL,
    0xff46480000000ULL,
    0xa0a2c40000000ULL,
    0xfbb5060000000ULL,
    0xdf2eef0000000ULL,
    0x9eae208000000ULL,
    0x9eef68c000000ULL,
    0x7cc6e2000000ULL,
    0xd83a1b3000000ULL,
    0xf8336e8800000ULL,
    0xe06e43e400000ULL,
    0x1e09c8a600000ULL,
    0xf5585ce500000ULL,
    0xc0e0001b80000ULL,
    0x1290003540000ULL,
    0xa8d8003ea0000ULL,
    0x62a4000bf0000ULL,
    0x4b6001318000ULL,
    0xe6a9002f14000ULL,
    0xbae880340e000ULL,
    0x61cac01d0d000ULL,
    0x8139a0169b800ULL,
    0xb8b1f03143400ULL,
    0x14af682ab7a00ULL,
    0xdedf42f68f00ULL,
    0x344dce2b4c480ULL,
    0xaf7feb36bacc0ULL,
    0x1492068473b20ULL,
    0x81deb7ca4e010ULL,
    0x8a2286b92d298ULL,
    0x2c7077d5aac7cULL,
    0x774d26acbbd22ULL,
    0x64f887fb73e73ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0x2c00000000000ULL,
    0xa00000000000ULL,
    0xf00000000000ULL,
    0xf880000000000ULL,
    0x7740000000000ULL,
    0x3be0000000000ULL,
    0xc070000000000ULL,
    0xd9c8000000000ULL,
    0x1ed4000000000ULL,
    0x395a000000000ULL,
    0xb1ef000000000ULL,
    0xd005800000000ULL,
    0x5801400000000ULL,
    0x3401e00000000ULL,
    0x5e03900000000ULL,
    0x2101580000000ULL,
    0xd186f40000000ULL,
    0x8ac3920000000ULL,
    0xbb26630000000ULL,
    0x74527d8000000ULL,
    0x551cab4000000ULL,
    0x3c8b2a2000000ULL,
    0x3e37071000000ULL,
    0x4faab79800000ULL,
    0x46633c5c00000ULL,
    0xb745a200000ULL,
    0xde88f5b00000ULL,
    0xe200003a80000ULL,
    0x1300000fc0000ULL,
    0x9a80002420000ULL,
    0x2440000ad0000ULL,
    0xc160000e78000ULL,
    0x94300038dc000ULL,
    0x10a8000ad2000ULL,
    0xa6e40019a5000ULL,
    0x23f2003e8e800ULL,
    0x180b0001b7c00ULL,
    0xb77800c8b600ULL,
    0x374a4021a9500ULL,
    0x496603524f80ULL,
    0xa939d03c4e240ULL,
    0xfc5f383217b60ULL,
    0x666b240e1a2d0ULL,
    0x4fc6aa1bb1fc8ULL,
    0x6ca2472f492ccULL,
    0xeb91579854b16ULL,
    0x61bfac76a673dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x7800000000000ULL,
    0x9c00000000000ULL,
    0xd200000000000ULL,
    0x8300000000000ULL,
    0xed80000000000ULL,
    0x7440000000000ULL,
    0x16a0000000000ULL,
    0x3610000000000ULL,
    0x7728000000000ULL,
    0x9af4000000000ULL,
    0x469a000000000ULL,
    0xd54f000000000ULL,
    0x287800000000ULL,
    0x13c2c00000000ULL,
    0x2de6600000000ULL,
    0x5734d00000000ULL,
    0x527b380000000ULL,
    0x58f91c0000000ULL,
    0xcf3f2a0000000ULL,
    0x9cdec10000000ULL,
    0xe2e9fe8000000ULL,
    0x4217ee4000000ULL,
    0x512a722000000ULL,
    0x7ff70d1000000ULL,
    0x2e1fec9800000ULL,
    0xf98b334c00000ULL,
    0x40e126be00000ULL,
    0xfbb3e27700000ULL,
    0x60bd800c80000ULL,
    0x39dc03cc0000ULL,
    0xadc9e03d60000ULL,
    0x642103610000ULL,
    0x85a7582f28000ULL,
    0xf392cc0a44000ULL,
    0xd76b92122a000ULL,
    0xba511d1b6d000ULL,
    0xbd8ab4af6e800ULL,
    0x1ee2ff65c7c00ULL,
    0x82b734b8c3e00ULL,
    0xda3f3f7fc9b00ULL,
    0xac5ed4bf15180ULL,
    0x22d2f5a23f40ULL,
    0x9a718ca399ae0ULL,
    0xde5be35c9d370ULL,
    0x91281eaa97b28ULL,
    0x5ff1fe6dca344ULL,
    0x7e1f2a0ebd29aULL,
    0xb18ec115a6889ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0xfc00000000000ULL,
    0xf600000000000ULL,
    0x8300000000000ULL,
    0xcf80000000000ULL,
    0x8340000000000ULL,
    0x3920000000000ULL,
    0x9a70000000000ULL,
    0x61f8000000000ULL,
    0x7a14000000000ULL,
    0x8faa000000000ULL,
    0x411b000000000ULL,
    0x2682800000000ULL,
    0xe9c1c00000000ULL,
    0x8ee4600000000ULL,
    0x8397f00000000ULL,
    0x2fef380000000ULL,
    0xc7ba2c0000000ULL,
    0x37b1b20000000ULL,
    0xd51ead0000000ULL,
    0x6481358000000ULL,
    0x18c49cc000000ULL,
    0x5862ea2000000ULL,
    0x2257711000000ULL,
    0x250dbfb800000ULL,
    0xde2f1de400000ULL,
    0xf55aed8600000ULL,
    0xa22780ef00000ULL,
    0xe9f0802580000ULL,
    0xf4bec014c0000ULL,
    0x5034e02de0000ULL,
    0x4259303bd0000ULL,
    0x5fa3d815c8000ULL,
    0x93b71c395c000ULL,
    0x7f186a161e000ULL,
    0x6182b11e95000ULL,
    0x5c43df9e42800ULL,
    0x70a3edd60ac00ULL,
    0x923755bf96a00ULL,
    0xf35c6cd269700ULL,
    0x692552002d980ULL,
    0xa2779d1a263c0ULL,
    0xedfaedbfb1ee0ULL,
    0x841780c0d83b0ULL,
    0x68a8801c75d48ULL,
    0x849ac02012274ULL,
    0xd0c6e007029b6ULL,
    0x9c6630332a503ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0x8c00000000000ULL,
    0xf600000000000ULL,
    0x2d00000000000ULL,
    0xaa80000000000ULL,
    0xc940000000000ULL,
    0x3ca0000000000ULL,
    0xfd10000000000ULL,
    0xdcc8000000000ULL,
    0x75f4000000000ULL,
    0xaffe000000000ULL,
    0xc96b000000000ULL,
    0x7865800000000ULL,
    0x7177400000000ULL,
    0x3fbce00000000ULL,
    0xd4a700000000ULL,
    0xeb31480000000ULL,
    0x289f840000000ULL,
    0x759c120000000ULL,
    0xd71e190000000ULL,
    0x725bd68000000ULL,
    0x48f8194000000ULL,
    0x50efba2000000ULL,
    0xaaa0ed3000000ULL,
    0xe0130cb800000ULL,
    0x5e4ec47c00000ULL,
    0xc0b49e9200000ULL,
    0xad5c9d5300000ULL,
    0x957e282580000ULL,
    0xf82db40740000ULL,
    0xe0c23a20a0000ULL,
    0x663ad2d50000ULL,
    0xf071ec80a8000ULL,
    0x733fb45704000ULL,
    0xd108568012000ULL,
    0x391059546d000ULL,
    0x66c8da1677800ULL,
    0x4ef6dd334ec00ULL,
    0xd87b24ac03600ULL,
    0x62ac706164d00ULL,
    0x18524bca2b80ULL,
    0x46c77074355c0ULL,
    0x3360a4800ab20ULL,
    0xef0305130f10ULL,
    0xf87c4483195c8ULL,
    0x72aa4048dcca4ULL,
    0x39850c862f9faULL,
    0x2c1c45d327b5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xe800000000000ULL,
    0x2400000000000ULL,
    0x5e00000000000ULL,
    0xe300000000000ULL,
    0x1980000000000ULL,
    0xd8c0000000000ULL,
    0xea20000000000ULL,
    0x6390000000000ULL,
    0x46f8000000000ULL,
    0xda74000000000ULL,
    0xb2ca000000000ULL,
    0x59dd000000000ULL,
    0x7764800000000ULL,
    0xa374400000000ULL,
    0x264aa00000000ULL,
    0xe39b300000000ULL,
    0xc007780000000ULL,
    0x60001c0000000ULL,
    0x7000ae0000000ULL,
    0xb805450000000ULL,
    0xcc03768000000ULL,
    0x7a0070c000000ULL,
    0xbd07762000000ULL,
    0xfa83691000000ULL,
    0xc140208800000ULL,
    0x32e169d400000ULL,
    0x89b60e9a00000ULL,
    0x25696cd900000ULL,
    0x9c89583f80000ULL,
    0x68bf6c3cc0000ULL,
    0xeb15763460000ULL,
    0x2eba692b30000ULL,
    0xd416a09708000ULL,
    0x853c29e924000ULL,
    0xc5d22ea05e000ULL,
    0x239b1ce86d000ULL,
    0xa00000385a800ULL,
    0x1000003ca3400ULL,
    0xc8000004ef200ULL,
    0x7400001ba4100ULL,
    0xb60000191e280ULL,
    0xc7000029cc8c0ULL,
    0x478000108a7e0ULL,
    0x3bc000285a570ULL,
    0xf3a00034a3a38ULL,
    0xbb500010ee8d4ULL,
    0xacd80021a5abaULL,
    0xb9e400101fd73ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xf800000000000ULL,
    0x2400000000000ULL,
    0x8e00000000000ULL,
    0x7500000000000ULL,
    0x8e80000000000ULL,
    0xb1c0000000000ULL,
    0x68a0000000000ULL,
    0x59b0000000000ULL,
    0xc418000000000ULL,
    0x2dd4000000000ULL,
    0xf8a000000000ULL,
    0x67bb000000000ULL,
    0xd7e1800000000ULL,
    0xc0d2c00000000ULL,
    0x1509a00000000ULL,
    0x4078300000000ULL,
    0x1e43480000000ULL,
    0x61e4240000000ULL,
    0x31d2d60000000ULL,
    0x58cd90000000ULL,
    0x8cb9d78000000ULL,
    0x2463bbc000000ULL,
    0x23923e2000000ULL,
    0x666bcd1000000ULL,
    0x8a69c9b800000ULL,
    0x786986fc00000ULL,
    0xb7689f9e00000ULL,
    0xcae89fd900000ULL,
    0x5cab683b80000ULL,
    0xa14ed43140000ULL,
    0xda983e25a0000ULL,
    0xd410cd1970000ULL,
    0x5b284987a8000ULL,
    0x640b46c22c000ULL,
    0x10f93f9126000ULL,
    0x3284afea45000ULL,
    0xabc220226a800ULL,
    0x8ba1f00ff2c00ULL,
    0x76336818d6600ULL,
    0x8d5ad4182db00ULL,
    0xa5b23e2a26e80ULL,
    0x3e1bcd16c5cc0ULL,
    0x3ed1c9b1aa320ULL,
    0xd80d86c093790ULL,
    0xafa9fa546bd8ULL,
    0xd1879fcb558a4ULL,
    0x8440e812d2cfeULL,
    0xc2e7143066ebfULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x4c00000000000ULL,
    0x9e00000000000ULL,
    0x1900000000000ULL,
    0x9680000000000ULL,
    0x67c0000000000ULL,
    0xe9e0000000000ULL,
    0x2850000000000ULL,
    0x6598000000000ULL,
    0x2774000000000ULL,
    0x86ae000000000ULL,
    0xd77d000000000ULL,
    0xf6a2800000000ULL,
    0x6776400000000ULL,
    0x66abe00000000ULL,
    0xe779300000000ULL,
    0xcea1e80000000ULL,
    0x375240000000ULL,
    0xb4ae860000000ULL,
    0x607ef30000000ULL,
    0x4127e88000000ULL,
    0xf236cdc000000ULL,
    0x3a8a8e2000000ULL,
    0xa1cfe73000000ULL,
    0xcea068800000ULL,
    0xb0db5ac400000ULL,
    0x9b5000aa00000ULL,
    0xf01ae9e300000ULL,
    0x2d30880f80000ULL,
    0x210e543240000ULL,
    0x9a8a8e27a0000ULL,
    0x71cfe70ef0000ULL,
    0x4ea0697f8000ULL,
    0xecdb5adf54000ULL,
    0x2d5000bb46000ULL,
    0xa51ae9e66d000ULL,
    0x25b0881673800ULL,
    0x5fce54123b400ULL,
    0xe5ea8e1cb6a00ULL,
    0x3e5fe70d14900ULL,
    0x8892068ce7780ULL,
    0xe3ff5ad09c0c0ULL,
    0xce66008d8b5a0ULL,
    0x5513e9c36f990ULL,
    0x55bc083ef1f98ULL,
    0xefc5141ff860cULL,
    0x75e3ee1b55782ULL,
    0xbe50971147b67ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x8800000000000ULL,
    0xcc00000000000ULL,
    0xc600000000000ULL,
    0xb900000000000ULL,
    0xdf80000000000ULL,
    0xac0000000000ULL,
    0x4160000000000ULL,
    0xdb0000000000ULL,
    0xd68000000000ULL,
    0x4314000000000ULL,
    0x303e000000000ULL,
    0x2c0d000000000ULL,
    0x9ea3800000000ULL,
    0xa553400000000ULL,
    0x1b1ee00000000ULL,
    0x2c9ff00000000ULL,
    0x3a5de80000000ULL,
    0x253dac0000000ULL,
    0xe5892e0000000ULL,
    0xb567ff0000000ULL,
    0x4fb2678000000ULL,
    0x9e6ea64000000ULL,
    0xd394262000000ULL,
    0xd478a33000000ULL,
    0x742d218800000ULL,
    0x1eb3b54c00000ULL,
    0x5def8f8200000ULL,
    0x87530a4300000ULL,
    0x581d081980000ULL,
    0x241f5c1240000ULL,
    0xaa1f461de0000ULL,
    0xff1d133db0000ULL,
    0xd69ba983d8000ULL,
    0xed5ba95e2c000ULL,
    0xbfb829a53a000ULL,
    0x7ec8e94c35000ULL,
    0x8bc6c983b7800ULL,
    0x1ae71976df400ULL,
    0x2d7321beaae00ULL,
    0x30eb575fdd00ULL,
    0xa0240fb212380ULL,
    0x24144a71a1bc0ULL,
    0xe2bde829501a0ULL,
    0x5b4dac08e8fb0ULL,
    0x56012e2435138ULL,
    0xc103ff1fb6d3cULL,
    0xbb8467acde5beULL,
    0x30c7a669aa3e9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xe800000000000ULL,
    0x3400000000000ULL,
    0x6600000000000ULL,
    0x5d00000000000ULL,
    0x380000000000ULL,
    0xf8c0000000000ULL,
    0x5ea0000000000ULL,
    0xbc90000000000ULL,
    0xd688000000000ULL,
    0xdbf4000000000ULL,
    0x853e000000000ULL,
    0x822d000000000ULL,
    0xb064800000000ULL,
    0x4730400000000ULL,
    0x959e200000000ULL,
    0x77bef00000000ULL,
    0xd36e080000000ULL,
    0xd1018c0000000ULL,
    0x79879e0000000ULL,
    0xd7c0c10000000ULL,
    0xb6225e8000000ULL,
    0xf2d6574000000ULL,
    0x4769b62000000ULL,
    0xa702bd3000000ULL,
    0xec87488800000ULL,
    0x30435a7400000ULL,
    0xc0e4569a00000ULL,
    0xc573db7d00000ULL,
    0x9e78283180000ULL,
    0xd44b7c1340000ULL,
    0x6a579632e0000ULL,
    0xf12c4d0730000ULL,
    0xcae140b428000ULL,
    0x1276d67564000ULL,
    0x8afdc88cc2000ULL,
    0x180e1a6f8d000ULL,
    0x28b6f69e07800ULL,
    0xe7596b759f400ULL,
    0x3e1e001c13200ULL,
    0x5b7d000115100ULL,
    0x1bcc801a0a080ULL,
    0x8944025d9fc0ULL,
    0x7088201aed220ULL,
    0xa6f7f02177330ULL,
    0x56bc8819d74f8ULL,
    0x92e8cc3f879c4ULL,
    0xdac33e1ede9daULL,
    0x9da37132f34adULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0x6400000000000ULL,
    0x9600000000000ULL,
    0x1100000000000ULL,
    0x4d80000000000ULL,
    0xf540000000000ULL,
    0x99e0000000000ULL,
    0x4cd0000000000ULL,
    0x3e58000000000ULL,
    0x1d74000000000ULL,
    0x2cea000000000ULL,
    0x19bf000000000ULL,
    0xaca1800000000ULL,
    0xf530c00000000ULL,
    0xe28fa00000000ULL,
    0xeb2ef00000000ULL,
    0x6d9c680000000ULL,
    0xc750fc0000000ULL,
    0x321d120000000ULL,
    0x5176b0000000ULL,
    0xaf78c08000000ULL,
    0x6503c7c000000ULL,
    0x5384da2000000ULL,
    0xd846671000000ULL,
    0xd6603a9800000ULL,
    0x69090fc00000ULL,
    0x9f3ca8ae00000ULL,
    0x7f673bc900000ULL,
    0x6f13c81d80000ULL,
    0x9c7e0c2840000ULL,
    0x7817a0560000ULL,
    0x5647971850000ULL,
    0x3365d28df8000ULL,
    0x1514acce74000ULL,
    0x277c1a9976000ULL,
    0x5905a0ee4d000ULL,
    0x5184e08622800ULL,
    0x6746f7c40bc00ULL,
    0xeee49234c6600ULL,
    0x2853ab1c25300ULL,
    0xe29d608f8f680ULL,
    0xe7d237dab97c0ULL,
    0xe8d9322395fa0ULL,
    0x26b65b3ae7cf0ULL,
    0x7aca889794cc8ULL,
    0x5d4d0bd0e656cULL,
    0x9b0a001c94c22ULL,
    0x606f00326672fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xd800000000000ULL,
    0xb400000000000ULL,
    0x8e00000000000ULL,
    0x4900000000000ULL,
    0x7080000000000ULL,
    0xbec0000000000ULL,
    0x2f20000000000ULL,
    0x4730000000000ULL,
    0x5978000000000ULL,
    0x19d4000000000ULL,
    0x34ee000000000ULL,
    0xbbd9000000000ULL,
    0x7ea3800000000ULL,
    0x2575c00000000ULL,
    0x5299200000000ULL,
    0x86c7300000000ULL,
    0xcb27280000000ULL,
    0x8137540000000ULL,
    0x4c7a4e0000000ULL,
    0x8b52750000000ULL,
    0xf92feb8000000ULL,
    0x237b34c000000ULL,
    0xbed2462000000ULL,
    0x9d6f113000000ULL,
    0x9d9f0da800000ULL,
    0x8342d5cc00000ULL,
    0x7765439600000ULL,
    0x8bd0a0dd00000ULL,
    0xffeaa82b80000ULL,
    0x285b942e40000ULL,
    0x9fe0ee0220000ULL,
    0xe110850eb0000ULL,
    0xcec9e39e48000ULL,
    0xce6f50ee8c000ULL,
    0xda192009fe000ULL,
    0x7c07300b85000ULL,
    0x12072832d3800ULL,
    0xb07541c8a400ULL,
    0x33824e343ee00ULL,
    0xd1467533e4700ULL,
    0x1c61ebac43c80ULL,
    0x285234db728c0ULL,
    0xc6a9c61c7aaa0ULL,
    0x583ed11856e50ULL,
    0x8db02dbbd83b8ULL,
    0xe0b8e5d7f4214ULL,
    0xaf77eb987a78eULL,
    0x8d9f34d257d43ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x9c00000000000ULL,
    0x8a00000000000ULL,
    0x7100000000000ULL,
    0xe80000000000ULL,
    0x5cc0000000000ULL,
    0x8360000000000ULL,
    0x3190000000000ULL,
    0x7528000000000ULL,
    0x5134000000000ULL,
    0x865a000000000ULL,
    0x50cf000000000ULL,
    0x9066800000000ULL,
    0x8211400000000ULL,
    0xc56f600000000ULL,
    0x7390900000000ULL,
    0xb82d280000000ULL,
    0xa5b7c40000000ULL,
    0xb39fbe0000000ULL,
    0x4928a70000000ULL,
    0xeb317e8000000ULL,
    0x8f58444000000ULL,
    0xda4ff62000000ULL,
    0xd2a4f31000000ULL,
    0x6677689800000ULL,
    0x17d674400000ULL,
    0x25fad6a200000ULL,
    0x383ec07700000ULL,
    0xcddf281a80000ULL,
    0x248cc40840000ULL,
    0xccc33e19a0000ULL,
    0xcb66e73430000ULL,
    0xcd909e8058000ULL,
    0xef2d947724000ULL,
    0x2837be1c6a000ULL,
    0x14dca70f3d000ULL,
    0x860b7e9aee800ULL,
    0x6207444b67c00ULL,
    0xbd01763b8ba00ULL,
    0xec81b30fad500ULL,
    0xa1c208b707580ULL,
    0xfe2f744dae40ULL,
    0x80517e8a246a0ULL,
    0x72c8447dea610ULL,
    0x4d67f603fd978ULL,
    0x7e90f32a0e5dcULL,
    0x6cad68a176cdaULL,
    0xe072677622c41ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xb800000000000ULL,
    0xf400000000000ULL,
    0x3600000000000ULL,
    0xb700000000000ULL,
    0x9980000000000ULL,
    0x6bc0000000000ULL,
    0xa820000000000ULL,
    0xac50000000000ULL,
    0x7df8000000000ULL,
    0x92b4000000000ULL,
    0xbd0a000000000ULL,
    0x1c5f000000000ULL,
    0xad20800000000ULL,
    0xfcd4c00000000ULL,
    0xc4ba200000000ULL,
    0x62d6f00000000ULL,
    0x1fbdd80000000ULL,
    0x8955dc0000000ULL,
    0x9d7bd60000000ULL,
    0x43f6f50000000ULL,
    0x169d08000000ULL,
    0x7cee80c000000ULL,
    0xb92e2e2000000ULL,
    0xe20ed91000000ULL,
    0x89dd5e9800000ULL,
    0x64e269f400000ULL,
    0x5f4888e00000ULL,
    0xfe6f9ced00000ULL,
    0x196fd83580000ULL,
    0x78eedc09c0000ULL,
    0x8729561060000ULL,
    0xc9093513f0000ULL,
    0x6a59708988000ULL,
    0x7a23b0fbd4000ULL,
    0xb55156149e000ULL,
    0x177d35039d000ULL,
    0x3ef370bb2a800ULL,
    0x85ecb0c2f3c00ULL,
    0x43a9d6103e200ULL,
    0x964df51d8c300ULL,
    0x953b509492080ULL,
    0xb51140e2cf7c0ULL,
    0x7d9e8e2b85d60ULL,
    0x3803e92d87870ULL,
    0xb4022682b05d8ULL,
    0x560185c5a9314ULL,
    0x6702a681858a2ULL,
    0x218545f6866b3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0xec00000000000ULL,
    0xea00000000000ULL,
    0xc500000000000ULL,
    0x9780000000000ULL,
    0xb440000000000ULL,
    0x6da0000000000ULL,
    0x2d30000000000ULL,
    0x3208000000000ULL,
    0x354000000000ULL,
    0x56de000000000ULL,
    0xb9cb000000000ULL,
    0xf6b0800000000ULL,
    0x1c48400000000ULL,
    0x33f6e00000000ULL,
    0x786a900000000ULL,
    0x1183080000000ULL,
    0x5b400c0000000ULL,
    0xdf229a0000000ULL,
    0x5ef4190000000ULL,
    0xc3ee678000000ULL,
    0xafc2174000000ULL,
    0xa3e7f22000000ULL,
    0x4d96c53000000ULL,
    0x34b915b800000ULL,
    0x471c925400000ULL,
    0x112807aa00000ULL,
    0xbfa0c74b00000ULL,
    0xc321a2280000ULL,
    0x138c591540000ULL,
    0xe0108797e0000ULL,
    0xadfc874210000ULL,
    0xb13afa0be8000ULL,
    0x325dc9310c000ULL,
    0xd0b0fabe2000ULL,
    0x59d0cb4885000ULL,
    0x291880013f800ULL,
    0xf22c40032ec00ULL,
    0xb720e028c6200ULL,
    0x92f59029cb100ULL,
    0x99ed883a28380ULL,
    0x12c34c26ad640ULL,
    0xf064fa0d93620ULL,
    0xffd6c905bcd30ULL,
    0x761b8faffb3e8ULL,
    0x38a88b6370b24ULL,
    0xe66016b963eULL,
    0x6512d03604e2dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xd800000000000ULL,
    0x8400000000000ULL,
    0xc600000000000ULL,
    0xed00000000000ULL,
    0xf280000000000ULL,
    0xb1c0000000000ULL,
    0x5320000000000ULL,
    0xc050000000000ULL,
    0x9f88000000000ULL,
    0x4034000000000ULL,
    0xa53e000000000ULL,
    0x68c9000000000ULL,
    0x6cd2800000000ULL,
    0x274dc00000000ULL,
    0x9795a00000000ULL,
    0xf1ae700000000ULL,
    0xae4880000000ULL,
    0x1f73040000000ULL,
    0x65df160000000ULL,
    0x60bac50000000ULL,
    0xe48ee88000000ULL,
    0xe7b7a34000000ULL,
    0xd27cbe2000000ULL,
    0x842a713000000ULL,
    0x4ea0d6a800000ULL,
    0xc810124400000ULL,
    0xd2e9c89600000ULL,
    0x5804135100000ULL,
    0x4405963980000ULL,
    0x6603050f40000ULL,
    0xfd0548a2a0000ULL,
    0x2a80d34970000ULL,
    0x35c2b63348000ULL,
    0x9520b51a54000ULL,
    0x2d546082c2000ULL,
    0x6d0da7433d000ULL,
    0xf1f1281727800ULL,
    0xf61d741756400ULL,
    0xa89b9e319ea00ULL,
    0xf359c1168a700ULL,
    0x6779fea1e8e80ULL,
    0x32a9667b24dc0ULL,
    0x996456b58a020ULL,
    0x6630d26f69f10ULL,
    0x3838e8b4e40d8ULL,
    0xf24aa365eb5d4ULL,
    0x91103e1df8702ULL,
    0xee6eb13d7cf5dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0x2400000000000ULL,
    0xce00000000000ULL,
    0xc900000000000ULL,
    0xb80000000000ULL,
    0xedc0000000000ULL,
    0xfda0000000000ULL,
    0xa50000000000ULL,
    0x4168000000000ULL,
    0xf9b4000000000ULL,
    0x3f5a000000000ULL,
    0xe12f000000000ULL,
    0x54d6800000000ULL,
    0x52a400000000ULL,
    0xbad6a00000000ULL,
    0xbc2fb00000000ULL,
    0xc953280000000ULL,
    0x1ded8c0000000ULL,
    0xdef3160000000ULL,
    0x10ba590000000ULL,
    0x5d1d508000000ULL,
    0xf4b604000000ULL,
    0x72201e2000000ULL,
    0x1992251000000ULL,
    0x6dcbce9800000ULL,
    0xa463054400000ULL,
    0x11f570ba00000ULL,
    0xc63a905900000ULL,
    0x6d5f960b80000ULL,
    0x8a2f192640000ULL,
    0xc557080e0000ULL,
    0x9c6a907670000ULL,
    0x24379636e8000ULL,
    0x479b1929d4000ULL,
    0x950f70b93a000ULL,
    0x904590735d000ULL,
    0xb56116085e800ULL,
    0x66715918afc00ULL,
    0xd9f9d09187a00ULL,
    0xcbfa204bf3900ULL,
    0xc0fa3e09d9280ULL,
    0x8878d50b5d2c0ULL,
    0x7938c6bc5fea0ULL,
    0xc3db794aae330ULL,
    0xf66beea486cd8ULL,
    0x6336f57672ea4ULL,
    0xe518f8b098b62ULL,
    0x634cac773d971ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x2800000000000ULL,
    0x7400000000000ULL,
    0x6e00000000000ULL,
    0x5500000000000ULL,
    0x6e80000000000ULL,
    0xa940000000000ULL,
    0x4c60000000000ULL,
    0xe1d0000000000ULL,
    0x8eb8000000000ULL,
    0xa274000000000ULL,
    0x4e4e000000000ULL,
    0xb779000000000ULL,
    0x5a54800000000ULL,
    0x897a400000000ULL,
    0xe754200000000ULL,
    0x33fab00000000ULL,
    0xf011480000000ULL,
    0x29b240000000ULL,
    0xb44d20000000ULL,
    0x4b609f0000000ULL,
    0xb056008000000ULL,
    0x8a7cd4c000000ULL,
    0x20d53a2000000ULL,
    0x3e3b4b3000000ULL,
    0x637baa800000ULL,
    0xe0addff400000ULL,
    0xcdeea0aa00000ULL,
    0xd78824ed00000ULL,
    0x9f5e521380000ULL,
    0xb563df1140000ULL,
    0xad56a0b9a0000ULL,
    0xe0fc24c350000ULL,
    0x1f90522298000ULL,
    0x7b5adf0034000ULL,
    0x9362208b8e000ULL,
    0xfc5664e38d000ULL,
    0x187c72312b800ULL,
    0xbfd46f12fa400ULL,
    0x43bd6895a3200ULL,
    0xe0f440d3f7900ULL,
    0x50c203848880ULL,
    0x9c1eb00eedec0ULL,
    0x9a0748199af20ULL,
    0xfb06241612190ULL,
    0x9b86521f9f1c8ULL,
    0x17c7df23b51bcULL,
    0xcd20a0b2cf5a2ULL,
    0xd9b124ee2d103ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x3800000000000ULL,
    0xf400000000000ULL,
    0x1200000000000ULL,
    0xe900000000000ULL,
    0x8280000000000ULL,
    0x8c40000000000ULL,
    0x7720000000000ULL,
    0xfb70000000000ULL,
    0x4c58000000000ULL,
    0xbe94000000000ULL,
    0xe24a000000000ULL,
    0x799d000000000ULL,
    0x4e76800000000ULL,
    0x50dac00000000ULL,
    0x1dd2200000000ULL,
    0xfcebd00000000ULL,
    0x5d2c280000000ULL,
    0x8fcda40000000ULL,
    0xfc5d7a0000000ULL,
    0xd690c50000000ULL,
    0xe4d7d8000000ULL,
    0x8f9911c000000ULL,
    0x8d75f22000000ULL,
    0xcf5c711000000ULL,
    0x1160f8800000ULL,
    0xee8ba0c400000ULL,
    0x53fbddae00000ULL,
    0xb4a101ed00000ULL,
    0x79b77a2c80000ULL,
    0x713dc52e40000ULL,
    0xd9c3fda6a0000ULL,
    0x6e7d1da10000ULL,
    0x7195d23748000ULL,
    0xfbcea1299c000ULL,
    0xae5ea7bf36000ULL,
    0x1f95c4f2f5000ULL,
    0x9cc807abfa800ULL,
    0x3bddd4eeab400ULL,
    0xe528fb1c3a00ULL,
    0x262860eb0fb00ULL,
    0xa44d7daa50080ULL,
    0xd29911c4e8b40ULL,
    0x3df5f2030c2a0ULL,
    0xba1c71043e990ULL,
    0xccb60f9ac8768ULL,
    0x6dbba0d15d754ULL,
    0x7a83ddbcd65d6ULL,
    0x184501ebc4997ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x800000000000ULL,
    0x4400000000000ULL,
    0xce00000000000ULL,
    0x4700000000000ULL,
    0x6f80000000000ULL,
    0x3540000000000ULL,
    0x6820000000000ULL,
    0x6d30000000000ULL,
    0xdab8000000000ULL,
    0x11d4000000000ULL,
    0x6dee000000000ULL,
    0xc1ff000000000ULL,
    0x2ef4800000000ULL,
    0x75d400000000ULL,
    0xba03a00000000ULL,
    0x9106b00000000ULL,
    0x483580000000ULL,
    0xc1ac0000000ULL,
    0xdc65a20000000ULL,
    0x6990250000000ULL,
    0x234fcf8000000ULL,
    0xd10a06c000000ULL,
    0x16a9da2000000ULL,
    0x2ede793000000ULL,
    0x4c44959800000ULL,
    0x38a23ff400000ULL,
    0xcbf46fa200000ULL,
    0xd5d8b6e100000ULL,
    0xd1c4820b80000ULL,
    0x78e0d515c0000ULL,
    0xd955b781e0000ULL,
    0x272f5ae850000ULL,
    0xf498000b88000ULL,
    0xbfe4002b8c000ULL,
    0x76d60006ce000ULL,
    0x726b002ce5000ULL,
    0x4cba801893800ULL,
    0xdad2403763c00ULL,
    0xe86f20149b200ULL,
    0xadbff00f2e500ULL,
    0x6656f818b5680ULL,
    0x74ac1c1d1b6c0ULL,
    0xf3dc7a0bee760ULL,
    0xc2c3c92ad4fb0ULL,
    0xb1654da48aa08ULL,
    0x3f15d3fb87f54ULL,
    0xe688ed9c084deULL,
    0xafec63d04c86bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0xd400000000000ULL,
    0x7200000000000ULL,
    0xbb00000000000ULL,
    0xc80000000000ULL,
    0xadc0000000000ULL,
    0x96e0000000000ULL,
    0xff90000000000ULL,
    0x59e8000000000ULL,
    0xa34000000000ULL,
    0x8e5e000000000ULL,
    0xb229000000000ULL,
    0xc5d5800000000ULL,
    0x5e4d400000000ULL,
    0xa680200000000ULL,
    0x9ac4500000000ULL,
    0xa460380000000ULL,
    0x7752d40000000ULL,
    0x168d660000000ULL,
    0x49e3ab0000000ULL,
    0x2917668000000ULL,
    0xcb2bce4000000ULL,
    0xda50fe2000000ULL,
    0x7b0c6f3000000ULL,
    0xd7a418a800000ULL,
    0x54b7e17c00000ULL,
    0xa89f469200000ULL,
    0x404b9e4500000ULL,
    0xb386c61480000ULL,
    0x2b43bb22c0000ULL,
    0x3c22fe9920000ULL,
    0x33f00a6270000ULL,
    0xabd802b38000ULL,
    0xa6b9401424000ULL,
    0xa8be2021da000ULL,
    0xe5bd503b3d000ULL,
    0x183db8370f800ULL,
    0x1e7b940d1e400ULL,
    0x8fdb461e37600ULL,
    0x826afb043a500ULL,
    0x574de9f6d880ULL,
    0x8795a63c6540ULL,
    0xeede383e42ee0ULL,
    0xb1ebd42929a50ULL,
    0xee30e6230c318ULL,
    0x945aeb1dd7aecULL,
    0x2d29469b6a5faULL,
    0x73569e5025129ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x3800000000000ULL,
    0x7c00000000000ULL,
    0xae00000000000ULL,
    0x8100000000000ULL,
    0xf680000000000ULL,
    0x81c0000000000ULL,
    0xc260000000000ULL,
    0x4830000000000ULL,
    0x90e8000000000ULL,
    0x2314000000000ULL,
    0x727a000000000ULL,
    0x6b2b000000000ULL,
    0xca76800000000ULL,
    0x7fccc00000000ULL,
    0xb380600000000ULL,
    0xe546300000000ULL,
    0x8ca1280000000ULL,
    0xedd1a40000000ULL,
    0x979cb60000000ULL,
    0x8c5a4f0000000ULL,
    0x753a8b8000000ULL,
    0x408b924000000ULL,
    0xe4237e2000000ULL,
    0x65151b1000000ULL,
    0x177d75b800000ULL,
    0xbfad494400000ULL,
    0x4cb2ebae00000ULL,
    0x8e29a24f00000ULL,
    0x5ef0563380000ULL,
    0x190bbf3440000ULL,
    0x5265438ba0000ULL,
    0x3034c67b70000ULL,
    0xcec803518000ULL,
    0x9d17c03d4c000ULL,
    0xcb7ee00a1a000ULL,
    0xe1aef00cf5000ULL,
    0xe5b3481fa6800ULL,
    0x3ca89437b6c00ULL,
    0xd311e0dbfa00ULL,
    0xf46c2b257a300ULL,
    0x6dd0dda9e4980ULL,
    0xd79b2d4b2e4c0ULL,
    0x6c58bd9db36a0ULL,
    0x65391d7c41990ULL,
    0x788b9587a0958ULL,
    0x9827b959712fcULL,
    0xcb13a3981870eULL,
    0x967e3643cdc19ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0x4400000000000ULL,
    0x2e00000000000ULL,
    0x1d00000000000ULL,
    0x980000000000ULL,
    0x3fc0000000000ULL,
    0x73e0000000000ULL,
    0x6ff0000000000ULL,
    0x60a8000000000ULL,
    0xdc54000000000ULL,
    0xf27a000000000ULL,
    0xe36d000000000ULL,
    0xddb1800000000ULL,
    0x628e400000000ULL,
    0xd40a00000000ULL,
    0xaca1b00000000ULL,
    0x8c56880000000ULL,
    0xea7c9c0000000ULL,
    0x2f69260000000ULL,
    0x47b71b0000000ULL,
    0xf988bd8000000ULL,
    0x5dc30c4000000ULL,
    0xb4e68e2000000ULL,
    0xdd70771000000ULL,
    0xffedb38800000ULL,
    0x1ff43b7c00000ULL,
    0x88a81d8a00000ULL,
    0xb852bc4100000ULL,
    0x2c78062380000ULL,
    0x5668eb2b40000ULL,
    0x903695a060000ULL,
    0x734a2062f0000ULL,
    0x63a3201d78000ULL,
    0xcad6f01dbc000ULL,
    0xd33da819ca000ULL,
    0x45ce6c03a5000ULL,
    0xb2e68e0806800ULL,
    0xc470773dd7400ULL,
    0xf86db3b83e600ULL,
    0xcd343b730a900ULL,
    0x5ac81d8584a80ULL,
    0xac62bc7b162c0ULL,
    0x1130061c1f420ULL,
    0xf8cceb2d1bf70ULL,
    0xb649585cc018ULL,
    0x73b32078733acULL,
    0x3f88a02339256ULL,
    0x24c5b0385cc81ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x8800000000000ULL,
    0x8400000000000ULL,
    0xbe00000000000ULL,
    0x8100000000000ULL,
    0x5780000000000ULL,
    0x7340000000000ULL,
    0x23e0000000000ULL,
    0xa490000000000ULL,
    0x48a8000000000ULL,
    0xa5b4000000000ULL,
    0x819a000000000ULL,
    0x7fef000000000ULL,
    0x1852800000000ULL,
    0x540d400000000ULL,
    0x78c7a00000000ULL,
    0xc2a7300000000ULL,
    0xcc77680000000ULL,
    0x9cbfec0000000ULL,
    0x44da220000000ULL,
    0x190d8b0000000ULL,
    0x3544f28000000ULL,
    0x7ee1484000000ULL,
    0x21106a2000000ULL,
    0x10ec171000000ULL,
    0x66d4189800000ULL,
    0x4cb1f6400000ULL,
    0x8e63528200000ULL,
    0xc8d6787100000ULL,
    0x4dcf021780000ULL,
    0x3de7fb1440000ULL,
    0x95943aafe0000ULL,
    0x67299459d0000ULL,
    0x4af5202578000ULL,
    0x107a703864000ULL,
    0x6078c8304a000ULL,
    0x387cdc31fd000ULL,
    0x547f4a1f7e800ULL,
    0xfe79670b0fc00ULL,
    0x497e50b346a00ULL,
    0x1bfa83750a500ULL,
    0x813bb8999d680ULL,
    0x50982f4aef0c0ULL,
    0x806e3aa597fa0ULL,
    0x1f169470f36b0ULL,
    0xd1efa00139c08ULL,
    0x5153301f84fecULL,
    0xe78d68159b8eaULL,
    0x2580ec3285a51ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0xb800000000000ULL,
    0x4c00000000000ULL,
    0x7e00000000000ULL,
    0xd100000000000ULL,
    0x7c80000000000ULL,
    0x91c0000000000ULL,
    0xaba0000000000ULL,
    0x7f70000000000ULL,
    0x4588000000000ULL,
    0x2b94000000000ULL,
    0x1f9e000000000ULL,
    0xf49000000000ULL,
    0xdfb1800000000ULL,
    0x82aec00000000ULL,
    0xad20e00000000ULL,
    0x75b3b00000000ULL,
    0xd1ac280000000ULL,
    0x66a5f40000000ULL,
    0x75f2360000000ULL,
    0x514f510000000ULL,
    0xbeb0e18000000ULL,
    0xe62eb74000000ULL,
    0xe77e2000000ULL,
    0x1813d53000000ULL,
    0x33d81fa800000ULL,
    0x21aaa25c00000ULL,
    0x1ea6018e00000ULL,
    0x99f0076300000ULL,
    0x5f4cd61780000ULL,
    0xd7b5e11840000ULL,
    0xd6ad499920000ULL,
    0xef25836b70000ULL,
    0x62b5a83b68000ULL,
    0x302f34089c000ULL,
    0xf5e4d639c2000ULL,
    0x9e91e106d5000ULL,
    0xf91b499c2f800ULL,
    0x1588836b7e400ULL,
    0x23922833e6600ULL,
    0x4b9cf431d2b00ULL,
    0x4d4bb60aedb80ULL,
    0xc8b5912c2bac0ULL,
    0x632e018088ea0ULL,
    0x3e6407760c8d0ULL,
    0x9ed2d62f3ad58ULL,
    0x79fce12731884ULL,
    0xcd9cc980488a6ULL,
    0xb04b4346ec7cdULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0x4c00000000000ULL,
    0x9e00000000000ULL,
    0xf100000000000ULL,
    0xe580000000000ULL,
    0x58c0000000000ULL,
    0xb5e0000000000ULL,
    0xcf10000000000ULL,
    0x71b8000000000ULL,
    0xc5b4000000000ULL,
    0xd48e000000000ULL,
    0xf67f000000000ULL,
    0xdcd0800000000ULL,
    0xa2dc400000000ULL,
    0xe8e7e00000000ULL,
    0x2491300000000ULL,
    0x907fa80000000ULL,
    0x89d73c0000000ULL,
    0xf55efe0000000ULL,
    0x8f206b0000000ULL,
    0xadf3e28000000ULL,
    0xae292cc000000ULL,
    0x8b4e362000000ULL,
    0xbb1e273000000ULL,
    0x83549800000ULL,
    0xb7444bd400000ULL,
    0xd62202aa00000ULL,
    0xc4731ce300000ULL,
    0x8ef1e1f80000ULL,
    0x9faa5b1840000ULL,
    0xb98acaab60000ULL,
    0x65f950d130000ULL,
    0x5911a820f8000ULL,
    0x3cb83c2c84000ULL,
    0xa6367e1a8e000ULL,
    0x49482b0f1d000ULL,
    0x5c1a02ab3b800ULL,
    0x68071ccb6bc00ULL,
    0xec011e01a2200ULL,
    0xae055b265e100ULL,
    0x89024ab214780ULL,
    0x818110c7e68c0ULL,
    0x8ac04821704a0ULL,
    0xdae20c27991f0ULL,
    0xdb975613b4678ULL,
    0xccfc571b7646cULL,
    0x28939cb3196aaULL,
    0xae7a07f874083ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0xa800000000000ULL,
    0xa400000000000ULL,
    0xc200000000000ULL,
    0x7700000000000ULL,
    0x4080000000000ULL,
    0xc040000000000ULL,
    0xc0a0000000000ULL,
    0xda70000000000ULL,
    0xf1e8000000000ULL,
    0x2514000000000ULL,
    0x1ffa000000000ULL,
    0x846b000000000ULL,
    0x92d2800000000ULL,
    0xbf98400000000ULL,
    0xcefaa00000000ULL,
    0xd1e9d00000000ULL,
    0x1512d80000000ULL,
    0xc7f8b40000000ULL,
    0x18696a0000000ULL,
    0x5cd5590000000ULL,
    0xae99a58000000ULL,
    0x3b78804000000ULL,
    0x2629922000000ULL,
    0x55777d1000000ULL,
    0x1d6a37b800000ULL,
    0xf354fd6400000ULL,
    0x5259258a00000ULL,
    0x659fc07900000ULL,
    0x85fbb22880000ULL,
    0x2f6ded29c0000ULL,
    0xfc50cf8220000ULL,
    0xfeddd97d10000ULL,
    0x53d8378aa8000ULL,
    0x585bfd525c000ULL,
    0x6699a584da000ULL,
    0x4f78805095000ULL,
    0xac299215f6800ULL,
    0x16777d3a96c00ULL,
    0x37ea378a66a00ULL,
    0xe014fd40ff100ULL,
    0x107925a29a280ULL,
    0x8afc048f5b40ULL,
    0xf433b22e86960ULL,
    0x1049ed104efd0ULL,
    0xd2e2cf9f62528ULL,
    0x85d2d94e60324ULL,
    0x2f18b7ac70f76ULL,
    0x46bcbd7159b91ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0x7400000000000ULL,
    0x4600000000000ULL,
    0xff00000000000ULL,
    0x4680000000000ULL,
    0xea40000000000ULL,
    0xdc60000000000ULL,
    0x290000000000ULL,
    0x2b88000000000ULL,
    0x9934000000000ULL,
    0xefba000000000ULL,
    0xd10b000000000ULL,
    0x173800000000ULL,
    0x6adf400000000ULL,
    0xda1d200000000ULL,
    0xce3bf00000000ULL,
    0xb1cf180000000ULL,
    0xad57240000000ULL,
    0xc9285a0000000ULL,
    0x44844f0000000ULL,
    0x5b41228000000ULL,
    0xf1e3c54000000ULL,
    0x5853e22000000ULL,
    0xaaacdb1000000ULL,
    0x7540c0b800000ULL,
    0x6ae01e6c00000ULL,
    0x40d2a2b200000ULL,
    0x33ec854d00000ULL,
    0x95a6c20980000ULL,
    0x4f332b03c0000ULL,
    0x48bdd8bba0000ULL,
    0xfb883a4b90000ULL,
    0x2133789498000ULL,
    0x33bc8a7264000ULL,
    0x53094092c2000ULL,
    0xf0745e5f1d000ULL,
    0xa75c029756800ULL,
    0x30d8356f76c00ULL,
    0x71cfa2626e00ULL,
    0x29bbff229ed00ULL,
    0x6e089aadabc80ULL,
    0xa7f45155d13c0ULL,
    0x309b8012f9a60ULL,
    0x4e7b400d54d90ULL,
    0xb8af20014bbe8ULL,
    0xdc44f037a13fcULL,
    0xcb6698099046aULL,
    0x6013641198045ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0xcc00000000000ULL,
    0x7a00000000000ULL,
    0x9d00000000000ULL,
    0x5880000000000ULL,
    0xd3c0000000000ULL,
    0xe4a0000000000ULL,
    0xdd30000000000ULL,
    0x8e8000000000ULL,
    0x53d4000000000ULL,
    0x983e000000000ULL,
    0x786b000000000ULL,
    0x6c12800000000ULL,
    0xf69a400000000ULL,
    0x405de00000000ULL,
    0x987a100000000ULL,
    0xc08cc80000000ULL,
    0xdbc2f40000000ULL,
    0xf8a3660000000ULL,
    0x2f33a90000000ULL,
    0x49ecc38000000ULL,
    0x1957e44000000ULL,
    0x1afece2000000ULL,
    0xbe4a0d3000000ULL,
    0xffe40db800000ULL,
    0x4252e94400000ULL,
    0x1d7e43ba00000ULL,
    0xec0da44300000ULL,
    0x36032e0680000ULL,
    0x27001d1340000ULL,
    0x2580c5b420000ULL,
    0x9b441d43f0000ULL,
    0x5f63259da8000ULL,
    0xf5950d530c000ULL,
    0xafdd6d929a000ULL,
    0xc63db97345000ULL,
    0x936beb8b07800ULL,
    0x3390005aa6c00ULL,
    0xf0dde03c96a00ULL,
    0x47ba103e5f900ULL,
    0xbe2cc81e82f80ULL,
    0x8bf2f40841440ULL,
    0xc0cb6626a0120ULL,
    0x6327a902b0ed0ULL,
    0x4f72c38689018ULL,
    0x210ce4607dfe4ULL,
    0x26844e1972faeULL,
    0xc8c44d2fe8001ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0x3400000000000ULL,
    0x1200000000000ULL,
    0x3700000000000ULL,
    0xc680000000000ULL,
    0x4c0000000000ULL,
    0xbae0000000000ULL,
    0x4e70000000000ULL,
    0xe218000000000ULL,
    0x1dd4000000000ULL,
    0xb2ca000000000ULL,
    0xa9d000000000ULL,
    0xb811800000000ULL,
    0x63ae400000000ULL,
    0x75ab200000000ULL,
    0x98ac500000000ULL,
    0x8d2a580000000ULL,
    0x56f840000000ULL,
    0x104cce0000000ULL,
    0x45e3b0000000ULL,
    0x89f0918000000ULL,
    0x3d5ba9c000000ULL,
    0x1e76362000000ULL,
    0xfa1eaf1000000ULL,
    0xa1d4a78800000ULL,
    0x64cc06fc00000ULL,
    0x8799118a00000ULL,
    0x7d91e9dd00000ULL,
    0xb3ef162c80000ULL,
    0xfc8bff21c0000ULL,
    0xaabd7f93a0000ULL,
    0x2584c2f030000ULL,
    0x40457fa038000ULL,
    0xf120c2c35c000ULL,
    0x5e177fa5a6000ULL,
    0x26a9c2e785000ULL,
    0x542cffb1aa800ULL,
    0x2ea82f10b400ULL,
    0xaf0e5fbeba600ULL,
    0xd7fc92e0c2900ULL,
    0x992527be7e480ULL,
    0xa1246ce09540ULL,
    0xe4aa31b864760ULL,
    0x3b29b9e57b310ULL,
    0x586f4e16639b8ULL,
    0x8dc97b06ce0acULL,
    0x481831b1f0fa6ULL,
    0x26d0b9c3985f7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xc800000000000ULL,
    0x9400000000000ULL,
    0x6a00000000000ULL,
    0xed00000000000ULL,
    0x9f80000000000ULL,
    0x3140000000000ULL,
    0x21a0000000000ULL,
    0x4b50000000000ULL,
    0x2e68000000000ULL,
    0x1334000000000ULL,
    0x1f5a000000000ULL,
    0x356b000000000ULL,
    0x3fb5800000000ULL,
    0xa69b400000000ULL,
    0xef0c200000000ULL,
    0xa780900000000ULL,
    0x2d40880000000ULL,
    0x3fa78c0000000ULL,
    0xbc50720000000ULL,
    0x94e8790000000ULL,
    0x29f4dc8000000ULL,
    0x65bfcc4000000ULL,
    0xb29b5a2000000ULL,
    0xc50b251000000ULL,
    0xaa8786b800000ULL,
    0xc2c0e95c00000ULL,
    0xc6e6dc9200000ULL,
    0x9f0cc7500000ULL,
    0xb5bcda0280000ULL,
    0xea9f653740000ULL,
    0xe90c26ade0000ULL,
    0x9c84397550000ULL,
    0xd1c5f480e8000ULL,
    0xde67905c94000ULL,
    0x4a35802efa000ULL,
    0x3adb403fdd000ULL,
    0xb12c20225e800ULL,
    0xad90902adec00ULL,
    0xea8888249f600ULL,
    0xf3c38c21ffd00ULL,
    0xe762720aef880ULL,
    0x5fb7791767e40ULL,
    0x969b5cace3c20ULL,
    0xc70f8c7591f30ULL,
    0x4382fa38c8748ULL,
    0x8f40f52365ea4ULL,
    0x46a32e8ea29d2ULL,
    0x49d3f56af1f41ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0x4400000000000ULL,
    0x4600000000000ULL,
    0xf700000000000ULL,
    0xa180000000000ULL,
    0xc9c0000000000ULL,
    0x1d20000000000ULL,
    0xe610000000000ULL,
    0x89b8000000000ULL,
    0x4a74000000000ULL,
    0xa1ca000000000ULL,
    0x4ab9000000000ULL,
    0x55f1800000000ULL,
    0x7b0ac00000000ULL,
    0x401ca00000000ULL,
    0x3526700000000ULL,
    0xa17580000000ULL,
    0x63bf9c0000000ULL,
    0xf775860000000ULL,
    0x4d4fbd0000000ULL,
    0xd7fd628000000ULL,
    0x9196e2c000000ULL,
    0x1f7dfe2000000ULL,
    0x5ad1911000000ULL,
    0xc4189ca800000ULL,
    0x932373fc00000ULL,
    0xad17628e00000ULL,
    0x8a3fe2c100000ULL,
    0xc2b47e1980000ULL,
    0x526f5111c0000ULL,
    0x80ee3cae60000ULL,
    0x4eac03df70000ULL,
    0x3d49ba98c8000ULL,
    0x2ffebec544000ULL,
    0x759758138e000ULL,
    0xa97f9c312d000ULL,
    0x15d586015a800ULL,
    0x219fbd345a400ULL,
    0x1ce5628fda600ULL,
    0x4732e2d91bb00ULL,
    0xcdaffe03fba80ULL,
    0x82cc9112cadc0ULL,
    0x53b1cb8a3360ULL,
    0xc734b3e377170ULL,
    0x8da8429380298ULL,
    0x22ce52eac1c34ULL,
    0xf53c0634e126aULL,
    0x7f317d203083bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x4800000000000ULL,
    0xec00000000000ULL,
    0xaa00000000000ULL,
    0x6900000000000ULL,
    0xf680000000000ULL,
    0xbec0000000000ULL,
    0xcce0000000000ULL,
    0x1870000000000ULL,
    0x2d58000000000ULL,
    0x13d4000000000ULL,
    0x228a000000000ULL,
    0x935f000000000ULL,
    0x54d5800000000ULL,
    0xeb0b400000000ULL,
    0x201be00000000ULL,
    0x1d72900000000ULL,
    0x31df680000000ULL,
    0xa411040000000ULL,
    0x68e82e0000000ULL,
    0x7de8bb0000000ULL,
    0x596d318000000ULL,
    0x4aaba3c000000ULL,
    0x46cb262000000ULL,
    0x6b7b6f1000000ULL,
    0xf946179800000ULL,
    0x10a0ccdc00000ULL,
    0x45531b200000ULL,
    0xa6cfa3eb00000ULL,
    0x5b79261280000ULL,
    0xd1406f1640000ULL,
    0x8ca197b7a0000ULL,
    0xe6508cc8b0000ULL,
    0x23c951abb8000ULL,
    0x7fd73f75c000ULL,
    0x6822e2476000ULL,
    0xb6c7bb3d05000ULL,
    0x40e0b18c3a800ULL,
    0xc274e3ea7bc00ULL,
    0xc5ac635db600ULL,
    0x956ff096bd00ULL,
    0x364cff84d2780ULL,
    0x36ba88ce8e640ULL,
    0xba26ffad783a0ULL,
    0x789588f5bd510ULL,
    0xff2b7fade7938ULL,
    0x278ac8f64d5acULL,
    0x8fda9f867eee6ULL,
    0xe31758c460cdfULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x7800000000000ULL,
    0x8400000000000ULL,
    0xa00000000000ULL,
    0xfd00000000000ULL,
    0x4080000000000ULL,
    0x9c40000000000ULL,
    0xbee0000000000ULL,
    0xae90000000000ULL,
    0xbd48000000000ULL,
    0xcfb4000000000ULL,
    0xaefe000000000ULL,
    0xdf4f000000000ULL,
    0xeb6800000000ULL,
    0x7ec00000000ULL,
    0x240fa00000000ULL,
    0x75d5100000000ULL,
    0xf62d080000000ULL,
    0xb1e5f40000000ULL,
    0x57121a0000000ULL,
    0xb8cfb0000000ULL,
    0x8017628000000ULL,
    0x860e874000000ULL,
    0xd4d0322000000ULL,
    0xc8a9df3000000ULL,
    0x32a7509800000ULL,
    0xa877586400000ULL,
    0xf7df62be00000ULL,
    0xe8fa875100000ULL,
    0xe44e323a80000ULL,
    0xb536df2b40000ULL,
    0x47b9d09620000ULL,
    0x4d2d986210000ULL,
    0xca66c29338000ULL,
    0x70d49745e4000ULL,
    0xf2abba05f2000ULL,
    0x47a2eb3fbd000ULL,
    0x84f2ea9227800ULL,
    0x9d9ab37e67c00ULL,
    0x251b083fc7a00ULL,
    0xfd5ef43616b00ULL,
    0x29ba9a074e680ULL,
    0xea2d3b251b840ULL,
    0x6fe642b401620ULL,
    0xb8115776010d0ULL,
    0xa20c9a3d00ce8ULL,
    0x2ed63b0880cacULL,
    0xddaec29840f2aULL,
    0x8e209757a158dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x3800000000000ULL,
    0xbc00000000000ULL,
    0x2a00000000000ULL,
    0xa100000000000ULL,
    0x7580000000000ULL,
    0x14c0000000000ULL,
    0x31a0000000000ULL,
    0xdeb0000000000ULL,
    0xba08000000000ULL,
    0x2154000000000ULL,
    0xd35e000000000ULL,
    0x860d000000000ULL,
    0xcb52800000000ULL,
    0x525ec00000000ULL,
    0xe388600000000ULL,
    0xe792100000000ULL,
    0xdffe280000000ULL,
    0x173a9c0000000ULL,
    0xfc9fbe0000000ULL,
    0x8b28090000000ULL,
    0xd0a2cf8000000ULL,
    0x4b36e34000000ULL,
    0x9ecd762000000ULL,
    0x38f7453000000ULL,
    0x89efb98800000ULL,
    0xaa01a64400000ULL,
    0x6102cfae00000ULL,
    0x5586e35f00000ULL,
    0x4c5762280000ULL,
    0x9a3451c40000ULL,
    0x62b1b99b60000ULL,
    0x900ca66e30000ULL,
    0x80504fa8e8000ULL,
    0xa6d82340ec000ULL,
    0x92cd161162000ULL,
    0xfaf1551125000ULL,
    0x8cef91a3df800ULL,
    0x59863a781f400ULL,
    0xc6c7f1b4bea00ULL,
    0xca42a5c2eb00ULL,
    0x9131d99d57980ULL,
    0x37cab64942c40ULL,
    0xd9706786f4aa0ULL,
    0x332fbf5946570ULL,
    0xaca0283b7b578ULL,
    0x41379c0300364ULL,
    0x2fcd3e3080a1eULL,
    0x7576c90141dbdULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0xa400000000000ULL,
    0x3200000000000ULL,
    0x7500000000000ULL,
    0xbb80000000000ULL,
    0xc2c0000000000ULL,
    0xa220000000000ULL,
    0xe830000000000ULL,
    0xfda8000000000ULL,
    0xd854000000000ULL,
    0xa5fe000000000ULL,
    0xbdab000000000ULL,
    0xf857800000000ULL,
    0xf5f9c00000000ULL,
    0x15aba00000000ULL,
    0x4456d00000000ULL,
    0x63ff480000000ULL,
    0x52afec0000000ULL,
    0x8ad36e0000000ULL,
    0x1abf850000000ULL,
    0x324ebb8000000ULL,
    0xc0c1924000000ULL,
    0xf26062000000ULL,
    0x17b0793000000ULL,
    0xbd68bdb800000ULL,
    0x1771eb6c00000ULL,
    0x524ebba600000ULL,
    0xb0c1925900000ULL,
    0xf726061480000ULL,
    0x3b0790140000ULL,
    0x9768bd90e0000ULL,
    0xc671eb75b0000ULL,
    0xdbcebb8208000ULL,
    0x701927e4c000ULL,
    0xee8606050a000ULL,
    0x29407907e5000ULL,
    0xc8e0bd9117800ULL,
    0xf615eb60d6c00ULL,
    0x8398bb8377e00ULL,
    0x62fe927527700ULL,
    0xb32f86291f680ULL,
    0x6112b9399a740ULL,
    0x251c9db7fdb20ULL,
    0x47bafb49034b0ULL,
    0xf5cc53a4a9d38ULL,
    0x7407ae629d254ULL,
    0x5a03a03d7314eULL,
    0x2902d01d01eb9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xf800000000000ULL,
    0x4c00000000000ULL,
    0xce00000000000ULL,
    0xad00000000000ULL,
    0xed80000000000ULL,
    0x2f40000000000ULL,
    0xfe60000000000ULL,
    0xb690000000000ULL,
    0x2298000000000ULL,
    0x829c000000000ULL,
    0xd29e000000000ULL,
    0x3a9d000000000ULL,
    0xfe9e800000000ULL,
    0x8499400000000ULL,
    0xab9b200000000ULL,
    0x251d500000000ULL,
    0x4ad8380000000ULL,
    0x767f540000000ULL,
    0x11ce2e0000000ULL,
    0x7ba1dd0000000ULL,
    0x6d30698000000ULL,
    0x1fab7d4000000ULL,
    0x753069a000000ULL,
    0x63ab7d5000000ULL,
    0x233069b800000ULL,
    0xf2ab7d4c00000ULL,
    0xf8b0698600000ULL,
    0x3ceb7d5f00000ULL,
    0x255069b580000ULL,
    0x83b7d7440000ULL,
    0x1428698ee0000ULL,
    0x13777d7f90000ULL,
    0x1a4e69b178000ULL,
    0x1de67d5234000ULL,
    0x14d6e9b9de000ULL,
    0x217e3d7293000ULL,
    0x934d498d79800ULL,
    0xba672d7c34400ULL,
    0x8c90d1b2de200ULL,
    0x6d9c695d12100ULL,
    0x7c1de79838e80ULL,
    0x455fb05455440ULL,
    0x4a3b980a8ee60ULL,
    0x572a4409ca910ULL,
    0x43f5b62efd6d8ULL,
    0x508b990722e64ULL,
    0x78c5dfadb177eULL,
    0xde20e45cc8391ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0x1c00000000000ULL,
    0xba00000000000ULL,
    0xff00000000000ULL,
    0x9980000000000ULL,
    0x1c40000000000ULL,
    0xeca0000000000ULL,
    0x8b90000000000ULL,
    0xd838000000000ULL,
    0x703c000000000ULL,
    0x1c3e000000000ULL,
    0x8e39000000000ULL,
    0x5d38800000000ULL,
    0xb6bec00000000ULL,
    0x49fe200000000ULL,
    0x86dbf00000000ULL,
    0xee0c080000000ULL,
    0x4310c40000000ULL,
    0xa4fc260000000ULL,
    0x8c5d430000000ULL,
    0xf9cc1f8000000ULL,
    0x60f45e4000000ULL,
    0x284c1fa000000ULL,
    0x20b45e5000000ULL,
    0x9eec1f9800000ULL,
    0x64245e5c00000ULL,
    0x17541faa00000ULL,
    0x14585e5b00000ULL,
    0x5dca1f8480000ULL,
    0xeef15e64c0000ULL,
    0x414a9fa2e0000ULL,
    0x34339e63d0000ULL,
    0xf82abf99c8000ULL,
    0xb7416e4134000ULL,
    0x932637afc6000ULL,
    0x32d36a5d3b000ULL,
    0x621a31a8c9800ULL,
    0xb66cd97eb5400ULL,
    0x28e2a63f87a00ULL,
    0x233683131af00ULL,
    0xedacbf8df8280ULL,
    0x9046e62ac0c0ULL,
    0x6480b7ae7b2a0ULL,
    0x4ec4aa6e683d0ULL,
    0xff64918d44238ULL,
    0xf675e9467f6dcULL,
    0xc58e8e2fb7246ULL,
    0x1c54b726827a5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x4800000000000ULL,
    0xe400000000000ULL,
    0x9e00000000000ULL,
    0x8f00000000000ULL,
    0x5280000000000ULL,
    0x140000000000ULL,
    0xada0000000000ULL,
    0xc10000000000ULL,
    0x15a8000000000ULL,
    0x2dac000000000ULL,
    0x41aa000000000ULL,
    0x9baf000000000ULL,
    0x7aa8800000000ULL,
    0xef2ac00000000ULL,
    0x58ede00000000ULL,
    0x6a09100000000ULL,
    0x44bd680000000ULL,
    0xf87fc0000000ULL,
    0x36c3f60000000ULL,
    0xf7609b0000000ULL,
    0x2175fb8000000ULL,
    0xd5dad6c000000ULL,
    0x6df5fba000000ULL,
    0x9b9ad6f000000ULL,
    0x32d5fb8800000ULL,
    0x66cad6cc00000ULL,
    0xc2ddfb9a00000ULL,
    0xa376d6f500000ULL,
    0x8dffbbd80000ULL,
    0x9a75d6d6c0000ULL,
    0x615d7ba920000ULL,
    0xefb016d690000ULL,
    0xeeb81b8248000ULL,
    0x6683c6d594000ULL,
    0xe740938722000ULL,
    0x2ea12adeff000ULL,
    0xd4940d9748800ULL,
    0x4de94df814c00ULL,
    0xb98a8009e2a00ULL,
    0x8079c003de700ULL,
    0xade7603fd9d80ULL,
    0xa430d0295c340ULL,
    0xb67a0806f6820ULL,
    0x56e72c16bdcb0ULL,
    0x70b1fe19467f8ULL,
    0x643bb71265edcULL,
    0x9bc605b69b016ULL,
    0x88e261c4d3587ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xe800000000000ULL,
    0xb400000000000ULL,
    0xb600000000000ULL,
    0x3100000000000ULL,
    0x5e80000000000ULL,
    0x7340000000000ULL,
    0x36e0000000000ULL,
    0x5030000000000ULL,
    0xf4d8000000000ULL,
    0xb0dc000000000ULL,
    0x9eda000000000ULL,
    0x43df000000000ULL,
    0xe75f800000000ULL,
    0x4f18c00000000ULL,
    0x147d600000000ULL,
    0xee88d00000000ULL,
    0x3071580000000ULL,
    0xcbbfac0000000ULL,
    0xb42e360000000ULL,
    0xd524810000000ULL,
    0xb897808000000ULL,
    0x6e89e64000000ULL,
    0xf07780a000000ULL,
    0x6bb9e67000000ULL,
    0xa42f808800000ULL,
    0x3d25e67400000ULL,
    0xc9580b200000ULL,
    0xd88ae65900000ULL,
    0xc172009f80000ULL,
    0x353e265540000ULL,
    0xd76d608820000ULL,
    0xbc5f64990000ULL,
    0x5ca1b88598000ULL,
    0x2c519a5d9c000ULL,
    0x71af6eb8ae000ULL,
    0xabe60b71ab000ULL,
    0x94b1561f18800ULL,
    0xec9f51215cc00ULL,
    0x13bb5886cf600ULL,
    0x382d8a429af00ULL,
    0x9f21568bd0380ULL,
    0x9b92774d38c40ULL,
    0x5f0bb815cc6e0ULL,
    0x58b3bc16d72f0ULL,
    0xc6980e0f47e38ULL,
    0x80bcfd32de0f4ULL,
    0xf1af6eb34203aULL,
    0x6be60b7dec76dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0xf400000000000ULL,
    0x5200000000000ULL,
    0xe700000000000ULL,
    0xba80000000000ULL,
    0xadc0000000000ULL,
    0x3160000000000ULL,
    0x3930000000000ULL,
    0xa888000000000ULL,
    0xaa8c000000000ULL,
    0x558e000000000ULL,
    0x7b0d000000000ULL,
    0x54cb800000000ULL,
    0xdaa9400000000ULL,
    0xad1ee00000000ULL,
    0xfa54700000000ULL,
    0x86bf480000000ULL,
    0x50849c0000000ULL,
    0x2ec4f20000000ULL,
    0xe1e2f90000000ULL,
    0x57f4708000000ULL,
    0x296ad4c000000ULL,
    0x2d7c70a000000ULL,
    0x24e6d4d000000ULL,
    0xa27270b800000ULL,
    0x222bd4e400000ULL,
    0x9fd9f0ae00000ULL,
    0x35b294d900000ULL,
    0xc84f10ba80000ULL,
    0x826ae4e0c0000ULL,
    0xa1fe589420000ULL,
    0x42378cbb0000ULL,
    0xea912a9878000ULL,
    0x658c1c5fc000ULL,
    0xb8f3ba325a000ULL,
    0x7fea652f5f000ULL,
    0x46be828739800ULL,
    0x70852de19c400ULL,
    0x9ec3802a0b600ULL,
    0x69e5403336100ULL,
    0xfbf0e02f4cd80ULL,
    0x8f697014a30c0ULL,
    0x987cc814630a0ULL,
    0x7961dc30422b0ULL,
    0xb5341233f2368ULL,
    0xbe8b89278a954ULL,
    0x9788b890772aaULL,
    0xa40b08f12c207ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x2800000000000ULL,
    0xcc00000000000ULL,
    0x8600000000000ULL,
    0x3500000000000ULL,
    0xf180000000000ULL,
    0x17c0000000000ULL,
    0xf2e0000000000ULL,
    0xf70000000000ULL,
    0xb0a8000000000ULL,
    0x26ac000000000ULL,
    0xabaa000000000ULL,
    0xee2d000000000ULL,
    0x2be8800000000ULL,
    0x8e09c00000000ULL,
    0xfffa00000000ULL,
    0xea13900000000ULL,
    0xed1e380000000ULL,
    0x5d63b40000000ULL,
    0xe9b60a0000000ULL,
    0xa9cd1d0000000ULL,
    0x151aa68000000ULL,
    0xc96383c000000ULL,
    0x6bb2a6a000000ULL,
    0x66cf83f000000ULL,
    0x7f98a69800000ULL,
    0xd62283dc00000ULL,
    0xf910269200000ULL,
    0xb99b43ef00000ULL,
    0xc32786b380000ULL,
    0x5894d3e240000ULL,
    0x365bbe8fe0000ULL,
    0xd5c667e7b0000ULL,
    0x1de734b358000ULL,
    0x35f3bae46c000ULL,
    0x37e8b21f72000ULL,
    0x500b692dc7000ULL,
    0x3ef90c9598800ULL,
    0xe1900ee64c400ULL,
    0x61deb80762a00ULL,
    0xa7067419ce300ULL,
    0x8683aa169d480ULL,
    0xc9438d0c4ad40ULL,
    0x3fa41ebe5ac60ULL,
    0xcad5f7c952390ULL,
    0xe1790cb3d7448ULL,
    0xcf500ee8917e4ULL,
    0x4beb82fc9ed2ULL,
    0x3ab6743da59ebULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0xe400000000000ULL,
    0x5a00000000000ULL,
    0xeb00000000000ULL,
    0x4480000000000ULL,
    0xc640000000000ULL,
    0x85a0000000000ULL,
    0x67d0000000000ULL,
    0x30a8000000000ULL,
    0x2eac000000000ULL,
    0xefaa000000000ULL,
    0xd82d000000000ULL,
    0xc6ef800000000ULL,
    0x130ec00000000ULL,
    0xa479e00000000ULL,
    0x683300000000ULL,
    0x7140680000000ULL,
    0xb7244c0000000ULL,
    0xa4957e0000000ULL,
    0x688bd30000000ULL,
    0xc4bdad8000000ULL,
    0x826508c000000ULL,
    0x39b5ada000000ULL,
    0xbf1908f000000ULL,
    0x74b7ad9800000ULL,
    0x169808e400000ULL,
    0x6b722d9200000ULL,
    0x14fbc8dd00000ULL,
    0x92c44d8480000ULL,
    0x4be638e7c0000ULL,
    0xb675c595a0000ULL,
    0x757d44d650000ULL,
    0x3902d3b4c8000ULL,
    0x3b83dbcfa4000ULL,
    0x40c78038e2000ULL,
    0x34e2c0040f000ULL,
    0x30f3e02308800ULL,
    0xc43e300204400ULL,
    0x6e27e82ab3600ULL,
    0xcf168c12c7f00ULL,
    0x4e9e09ad180ULL,
    0x1459e32a673c0ULL,
    0xe89045837d020ULL,
    0x168e84fd6e2f0ULL,
    0x95bcb39b785d8ULL,
    0x7de22be6dcdbcULL,
    0x937408333f36eULL,
    0x18f8bc0bb199fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0xdc00000000000ULL,
    0xce00000000000ULL,
    0xb100000000000ULL,
    0x2880000000000ULL,
    0xd740000000000ULL,
    0x8fe0000000000ULL,
    0x8f0000000000ULL,
    0xef38000000000ULL,
    0x853c000000000ULL,
    0x2e3a000000000ULL,
    0x5bf000000000ULL,
    0xb57e800000000ULL,
    0xbc58c00000000ULL,
    0x188e200000000ULL,
    0x69e0500000000ULL,
    0x35f1d80000000ULL,
    0x51be9c0000000ULL,
    0xc779260000000ULL,
    0x535ad90000000ULL,
    0x7909de8000000ULL,
    0x8221c74000000ULL,
    0x7f51dea000000ULL,
    0xa9adc77000000ULL,
    0xb933de8800000ULL,
    0xc69ec77c00000ULL,
    0x8aaf5eba00000ULL,
    0x2eb5074b00000ULL,
    0x285dfe8980000ULL,
    0xca8e976740000ULL,
    0xb6e68683a0000ULL,
    0x9c779b5290000ULL,
    0x667ed8a058000ULL,
    0x43db4e50ac000ULL,
    0x1a49d81b96000ULL,
    0x2fc29c31e3000ULL,
    0xc0a3263018800ULL,
    0x315d9008cc00ULL,
    0xdcf5eb547200ULL,
    0x10507639af00ULL,
    0x2085fea164a80ULL,
    0xab42977ee2e40ULL,
    0x71e486b798de0ULL,
    0x71f49b4acd430ULL,
    0x1bba58be66978ULL,
    0x9c7c8e5ccbf54ULL,
    0x10d978305ce02ULL,
    0x25ca0c123f849ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0xa400000000000ULL,
    0x8600000000000ULL,
    0x8300000000000ULL,
    0x7e80000000000ULL,
    0x6bc0000000000ULL,
    0x9ea0000000000ULL,
    0xc6d0000000000ULL,
    0x6b28000000000ULL,
    0x512c000000000ULL,
    0x682e000000000ULL,
    0x6fad000000000ULL,
    0x63ed800000000ULL,
    0xe108c00000000ULL,
    0xd3de00000000ULL,
    0x2e21d00000000ULL,
    0xda11c80000000ULL,
    0xc5083c0000000ULL,
    0xcb3c620000000ULL,
    0xcd22170000000ULL,
    0xb496618000000ULL,
    0x6ca98c000000ULL,
    0xf19e61a000000ULL,
    0x8df698d000000ULL,
    0x5cb861b800000ULL,
    0x296798d400000ULL,
    0xf273e19200000ULL,
    0x7cfe58fd00000ULL,
    0xf98581b380000ULL,
    0xa34648ddc0000ULL,
    0xae6229ac20000ULL,
    0x3af664c730000ULL,
    0x4c39e38b88000ULL,
    0x5a45ff684000ULL,
    0x8454481142000ULL,
    0xfaecfc22cf000ULL,
    0xd68f822b49800ULL,
    0xd97ec72924400ULL,
    0xb0c229abb2e00ULL,
    0xbc2664d566700ULL,
    0x4711e3a2fd980ULL,
    0x44885fe7ee040ULL,
    0x447a482df9ea0ULL,
    0x3141fc326c0b0ULL,
    0x3362020a96f28ULL,
    0xbb760722d48ecULL,
    0xc37fc9b99b6aeULL,
    0xf9c7b4f61228fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0x1c00000000000ULL,
    0x9a00000000000ULL,
    0xdd00000000000ULL,
    0x8080000000000ULL,
    0xa5c0000000000ULL,
    0x8e0000000000ULL,
    0xb850000000000ULL,
    0x4d88000000000ULL,
    0xb88c000000000ULL,
    0xe40a000000000ULL,
    0xfbcb000000000ULL,
    0x3e2b800000000ULL,
    0xaefec00000000ULL,
    0x5ab6200000000ULL,
    0x70def00000000ULL,
    0xf182180000000ULL,
    0xc7451c0000000ULL,
    0xe427b20000000ULL,
    0xa633350000000ULL,
    0xbf1bd78000000ULL,
    0xbc64cbc000000ULL,
    0x5b91d7a000000ULL,
    0x226fcbf000000ULL,
    0x4d5a57a800000ULL,
    0x24c10bf400000ULL,
    0xf264778200000ULL,
    0xf093fbf300000ULL,
    0x7dec6fa480000ULL,
    0x111de7c440000ULL,
    0x27605d84a0000ULL,
    0xbc1012ed30000ULL,
    0xca2daa0698000ULL,
    0xc8fa290324000ULL,
    0x2db665b5fa000ULL,
    0x955cfedd07000ULL,
    0x60c18024f8800ULL,
    0xec65c017f5c00ULL,
    0x1395a02972600ULL,
    0xe6c300f4a700ULL,
    0x6f5e383b5fe80ULL,
    0x4dc0ec273dcc0ULL,
    0xf4e62a2b9f6e0ULL,
    0x1254e90fdc8b0ULL,
    0x288845960e828ULL,
    0x8c0e0ed674464ULL,
    0xc7c99828b2776ULL,
    0xb42bdc202a74bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x1800000000000ULL,
    0x5c00000000000ULL,
    0x5a00000000000ULL,
    0x6b00000000000ULL,
    0x9580000000000ULL,
    0xb8c0000000000ULL,
    0x7ea0000000000ULL,
    0x30000000000ULL,
    0x62e8000000000ULL,
    0xd5ec000000000ULL,
    0xda6a000000000ULL,
    0xa9af000000000ULL,
    0x728e800000000ULL,
    0xd27e400000000ULL,
    0x9234a00000000ULL,
    0x1dea700000000ULL,
    0xce6b280000000ULL,
    0x27ac840000000ULL,
    0x378dc20000000ULL,
    0x32ff770000000ULL,
    0xd275568000000ULL,
    0x7f0a71c000000ULL,
    0xf6bf56a000000ULL,
    0x169571f000000ULL,
    0x46d9d68800000ULL,
    0x210731fc00000ULL,
    0x1687768e00000ULL,
    0xc94241eb00000ULL,
    0xf062de9880000ULL,
    0x579085f1c0000ULL,
    0xc05bbcbde0000ULL,
    0xc04582dff0000ULL,
    0xa2e5c22248000ULL,
    0x98d3771ffc000ULL,
    0x13f569af2000ULL,
    0x695571da87000ULL,
    0x4ff9d6a468800ULL,
    0x9ef731ccecc00ULL,
    0xdd4f7696cb200ULL,
    0x935e41e373500ULL,
    0xf9c0dea2c6080ULL,
    0xf82385e9c9cc0ULL,
    0xe1773ca2bcc20ULL,
    0xd688c2d4d2ed0ULL,
    0x847de21597b28ULL,
    0x2b3447365130cULL,
    0x976c5e8618e72ULL,
    0xfd2ec5ece4acbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x1800000000000ULL,
    0x9400000000000ULL,
    0xda00000000000ULL,
    0x2500000000000ULL,
    0x7980000000000ULL,
    0xf5c0000000000ULL,
    0x98a0000000000ULL,
    0x69b0000000000ULL,
    0x78e8000000000ULL,
    0xf1ec000000000ULL,
    0x966e000000000ULL,
    0x24ab000000000ULL,
    0x648b800000000ULL,
    0x437b400000000ULL,
    0xe7f7e00000000ULL,
    0x448d300000000ULL,
    0x137a180000000ULL,
    0x5ff6d40000000ULL,
    0xc088fa0000000ULL,
    0xd17a7f0000000ULL,
    0xeef2528000000ULL,
    0x630a65c000000ULL,
    0x1bc52a000000ULL,
    0xfd165d000000ULL,
    0xff7fd28800000ULL,
    0xe1f625d400000ULL,
    0x978e328e00000ULL,
    0x11fc15f100000ULL,
    0x34b1aab880000ULL,
    0x656a81f840000ULL,
    0x762d30a6e0000ULL,
    0xb7ca8ec530000ULL,
    0x629c9a00c8000ULL,
    0x82a00f1f5c000ULL,
    0xacb1aabef2000ULL,
    0xb16a81f327000ULL,
    0xc2d309369800ULL,
    0x82ca8eee0dc00ULL,
    0x31c9a244b600ULL,
    0xe3600f1fd3100ULL,
    0xee11aa8d36980ULL,
    0xfdda81f2b05c0ULL,
    0xd4530a9081e0ULL,
    0x86e68ec57de90ULL,
    0xdd29a0f62208ULL,
    0xae7b0f373f76cULL,
    0xf2722a9c4cb42ULL,
    0x4f4dc1f5aaefbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0x1c00000000000ULL,
    0xa00000000000ULL,
    0xc100000000000ULL,
    0xc680000000000ULL,
    0xd8c0000000000ULL,
    0xbae0000000000ULL,
    0xd6b0000000000ULL,
    0xe888000000000ULL,
    0xfd8c000000000ULL,
    0x650e000000000ULL,
    0x62cb000000000ULL,
    0x61a9800000000ULL,
    0x865e400000000ULL,
    0x1d76a00000000ULL,
    0x59eaf00000000ULL,
    0x677af80000000ULL,
    0x62a3cc0000000ULL,
    0xc7968a0000000ULL,
    0x3f5c1d0000000ULL,
    0x87f4888000000ULL,
    0x2b2ffcc000000ULL,
    0x4c9a88a000000ULL,
    0x8a94fcd000000ULL,
    0xebdb088800000ULL,
    0x7236bcc400000ULL,
    0xf54ba8b600000ULL,
    0x67eb4cc300000ULL,
    0x87ed0b480000ULL,
    0xd321c0f340000ULL,
    0x42d17a8520000ULL,
    0xa2fe6dc570000ULL,
    0xe9e62a2b98000ULL,
    0x7d31ed30dc000ULL,
    0xf4c9f0b1ba000ULL,
    0xeaa970dfbf000ULL,
    0x21db22a0f9800ULL,
    0x533051c30dc00ULL,
    0xc3cdd82e93e00ULL,
    0x172b7c319a500ULL,
    0x569cd2354e480ULL,
    0x13922100a0ec0ULL,
    0x615f7ab7b0f60ULL,
    0x58f56dca781f0ULL,
    0x92afaa074d548ULL,
    0x7ddfad1133184ULL,
    0xf93750b8aa76aULL,
    0x52cf80e577e77ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x8800000000000ULL,
    0xec00000000000ULL,
    0xc200000000000ULL,
    0xe900000000000ULL,
    0xc280000000000ULL,
    0x9540000000000ULL,
    0x7ea0000000000ULL,
    0x65f0000000000ULL,
    0xf18000000000ULL,
    0xa41c000000000ULL,
    0xcf9e000000000ULL,
    0xb85f000000000ULL,
    0x43bf800000000ULL,
    0xd0ebc00000000ULL,
    0x5604a00000000ULL,
    0x3f06b00000000ULL,
    0xbd83f80000000ULL,
    0x8c4ec0000000ULL,
    0x6660c60000000ULL,
    0x8b97730000000ULL,
    0x688b3c8000000ULL,
    0xe91544000000ULL,
    0x280d3ca000000ULL,
    0x52d2545000000ULL,
    0x842cbca800000ULL,
    0x2a66945400000ULL,
    0x19979c8a00000ULL,
    0x298be46f00000ULL,
    0x3010c4ba80000ULL,
    0xf749b85540000ULL,
    0x2973fa8a20000ULL,
    0xe15a277010000ULL,
    0x5938001588000ULL,
    0x1ac00296c000ULL,
    0x16a600364e000ULL,
    0xf9f300106b000ULL,
    0x7519800da9800ULL,
    0x3918c03a7cc00ULL,
    0xab1d203f46600ULL,
    0xea1e703546300ULL,
    0xd49ed81a47480ULL,
    0xbda9c00c79c0ULL,
    0x707e1e0587b60ULL,
    0x150def27a6a70ULL,
    0x66552283b7878ULL,
    0x7e6cbb5c3f7bcULL,
    0x41401e3a5348aULL,
    0x88a2ef269c2edULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x3800000000000ULL,
    0xa400000000000ULL,
    0xd600000000000ULL,
    0x3500000000000ULL,
    0x3780000000000ULL,
    0x1f40000000000ULL,
    0xbd20000000000ULL,
    0x7810000000000ULL,
    0xeef8000000000ULL,
    0xf5fc000000000ULL,
    0x337e000000000ULL,
    0x5dbf000000000ULL,
    0x42da800000000ULL,
    0x2c6b400000000ULL,
    0xd7c6600000000ULL,
    0x92e4d00000000ULL,
    0x2ef7780000000ULL,
    0x860fc40000000ULL,
    0xcef6be0000000ULL,
    0x560c970000000ULL,
    0x66f5778000000ULL,
    0x7a0e304000000ULL,
    0x2cf377a000000ULL,
    0x3d0d305000000ULL,
    0xf877f7a800000ULL,
    0x20c9707c00000ULL,
    0x6d9317a600000ULL,
    0xfabae05500000ULL,
    0x465c0f9380000ULL,
    0x99aef45840000ULL,
    0x6e27499e60000ULL,
    0x296e75f90000ULL,
    0x423ae03128000ULL,
    0x319c900194000ULL,
    0xfb4d982a96000ULL,
    0x58d3542ceb000ULL,
    0x909b262e49800ULL,
    0x62cfc31784c00ULL,
    0xf69651b0ff200ULL,
    0x7c3df3759e100ULL,
    0x89b26108eb80ULL,
    0x36cfc30c67240ULL,
    0x389651bed3660ULL,
    0x5d3df35634d50ULL,
    0xd11b260426c98ULL,
    0xb88fc308b3f0cULL,
    0x643651b5a5946ULL,
    0xf6df3730f7cdULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xf800000000000ULL,
    0x7400000000000ULL,
    0xea00000000000ULL,
    0x300000000000ULL,
    0xb680000000000ULL,
    0xf2c0000000000ULL,
    0x8e60000000000ULL,
    0x84b0000000000ULL,
    0xb8c8000000000ULL,
    0x19cc000000000ULL,
    0xe04a000000000ULL,
    0x5a09000000000ULL,
    0xffa8800000000ULL,
    0xb8ff400000000ULL,
    0xb4c4e00000000ULL,
    0x8367b00000000ULL,
    0x7331a80000000ULL,
    0x438f1c0000000ULL,
    0xb5ed0a0000000ULL,
    0x8558b90000000ULL,
    0xb7943c8000000ULL,
    0x9d5b2ac000000ULL,
    0x33963ca000000ULL,
    0xef5e2af000000ULL,
    0x7494bcb800000ULL,
    0x4bd86afc00000ULL,
    0xf150dc8600000ULL,
    0x993c9ae100000ULL,
    0x842794bc80000ULL,
    0x191136e1c0000ULL,
    0xeb99b69120000ULL,
    0xcff0d3dd70000ULL,
    0x2cece032b8000ULL,
    0x68dbb03c5c000ULL,
    0xd7d3a83c8a000ULL,
    0xc3fa1c0a2f000ULL,
    0xb6478a1c98800ULL,
    0xfba2f910ad400ULL,
    0xce525c9d72a00ULL,
    0x89badae392900ULL,
    0x8be3f4aac3980ULL,
    0xf8f5c6f38b7c0ULL,
    0xd06efe83aff20ULL,
    0xc91d7fd3d8bb0ULL,
    0x2732c2334cb08ULL,
    0x398a550ea2954ULL,
    0x1eeefeaa1bbfaULL,
    0x8fdd7fdf475ffULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xb800000000000ULL,
    0xc00000000000ULL,
    0x200000000000ULL,
    0xa700000000000ULL,
    0xed80000000000ULL,
    0x9fc0000000000ULL,
    0xe260000000000ULL,
    0xf010000000000ULL,
    0x6048000000000ULL,
    0x694c000000000ULL,
    0xb1ce000000000ULL,
    0xfc8f000000000ULL,
    0xb0a9800000000ULL,
    0x871c400000000ULL,
    0x5e24200000000ULL,
    0x48b0300000000ULL,
    0xc8bf080000000ULL,
    0xe355640000000ULL,
    0x71691a0000000ULL,
    0x387cc90000000ULL,
    0xc8b6a98000000ULL,
    0x88bbfbc000000ULL,
    0xc350a9a000000ULL,
    0xa168fbd000000ULL,
    0x807f29a800000ULL,
    0xc4b7bbe400000ULL,
    0x8abc898600000ULL,
    0x6454cbe700000ULL,
    0x4cee218680000ULL,
    0x1fbddfcec0000ULL,
    0x26d413bde0000ULL,
    0x7aa842e930000ULL,
    0x41aa80578000ULL,
    0x25a5141534000ULL,
    0xae74323d6a000ULL,
    0xda5a9d1f67000ULL,
    0xca073b9919800ULL,
    0x830116d444c00ULL,
    0x7b82ba3b33e00ULL,
    0xe6c3b92683d00ULL,
    0x12e201843ae80ULL,
    0x2951eff96fdc0ULL,
    0xf26d1b8b752a0ULL,
    0x43fe26c14a850ULL,
    0x2e743209376c8ULL,
    0x9a5a9d2d91b74ULL,
    0xea073b96e8c6eULL,
    0x530116c8dc49bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x2800000000000ULL,
    0xbc00000000000ULL,
    0x8200000000000ULL,
    0x7b00000000000ULL,
    0x8980000000000ULL,
    0xc3c0000000000ULL,
    0xbe0000000000ULL,
    0x770000000000ULL,
    0xed38000000000ULL,
    0xf83c000000000ULL,
    0x70be000000000ULL,
    0x45fb000000000ULL,
    0x225a800000000ULL,
    0x800bc00000000ULL,
    0xad26200000000ULL,
    0xc16300000000ULL,
    0x530fe80000000ULL,
    0xd8a2a40000000ULL,
    0xedd33a0000000ULL,
    0x13ed250000000ULL,
    0x7e507e8000000ULL,
    0x7f2dc74000000ULL,
    0x62367ea000000ULL,
    0x729ac75000000ULL,
    0x5e6afe8800000ULL,
    0x7b96075c00000ULL,
    0x89c85e8200000ULL,
    0xb1c0f77300000ULL,
    0xb8e316a280000ULL,
    0xa2f3a36dc0000ULL,
    0xb4ffe480e0000ULL,
    0x1cdd125d30000ULL,
    0xc04fc80218000ULL,
    0x9203941cec000ULL,
    0x2300521956000ULL,
    0xbd834132e3000ULL,
    0xcdc1e4ad79800ULL,
    0xdae6126f9dc00ULL,
    0x49f5480fef200ULL,
    0x257854309f300ULL,
    0x4b1e720c26a80ULL,
    0xf5a971315abc0ULL,
    0x6c700c9885520ULL,
    0x3cbfb65482150ULL,
    0xffcf22b899c8ULL,
    0x755eb130655c4ULL,
    0x93882c85b3832ULL,
    0x81e28663912d9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x6800000000000ULL,
    0x8400000000000ULL,
    0xba00000000000ULL,
    0xed00000000000ULL,
    0xe980000000000ULL,
    0x6bc0000000000ULL,
    0x5820000000000ULL,
    0xfad0000000000ULL,
    0x7e18000000000ULL,
    0x379c000000000ULL,
    0x6c5a000000000ULL,
    0x7c7b000000000ULL,
    0xf2ac800000000ULL,
    0x5eb7c00000000ULL,
    0x2fe00000000ULL,
    0x3ff0300000000ULL,
    0xc54c580000000ULL,
    0x8647940000000ULL,
    0x49e30a0000000ULL,
    0x6ff62b0000000ULL,
    0x9d4dcd8000000ULL,
    0x4a40654000000ULL,
    0x87e14da000000ULL,
    0x50f7a57000000ULL,
    0x1dceada800000ULL,
    0x7207954c00000ULL,
    0x5902f5b200000ULL,
    0x1b80016d00000ULL,
    0xf2c1ff9c80000ULL,
    0x63a62a5640000ULL,
    0xf8143226e0000ULL,
    0x75ba4f0cb0000ULL,
    0x4b8f7f90e8000ULL,
    0xa3e6ea4204000ULL,
    0xdaf552276e000ULL,
    0xb8cabf03e3000ULL,
    0xef82478e78800ULL,
    0xe0c18e71fdc00ULL,
    0x2aa3602e53a00ULL,
    0x5b97f03010300ULL,
    0x9b7bb808b8580ULL,
    0x7e2ba4335c3c0ULL,
    0x88f55229034e0ULL,
    0x11cabf10c9d90ULL,
    0x9c0247bc158c8ULL,
    0x96018e78d753cULL,
    0xf303603d3e7feULL,
    0x4e87f017ba8a9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x9800000000000ULL,
    0x1c00000000000ULL,
    0xca00000000000ULL,
    0xe700000000000ULL,
    0x4d80000000000ULL,
    0xb840000000000ULL,
    0xaca0000000000ULL,
    0xd0f0000000000ULL,
    0x4018000000000ULL,
    0xfd9c000000000ULL,
    0xadda000000000ULL,
    0x357f000000000ULL,
    0xbb8f800000000ULL,
    0x5293400000000ULL,
    0xd38fa00000000ULL,
    0xa693f00000000ULL,
    0x2d8bc80000000ULL,
    0x1f96240000000ULL,
    0xc908a60000000ULL,
    0xdb56630000000ULL,
    0xbd6d148000000ULL,
    0xb5c5dc4000000ULL,
    0x74e294a000000ULL,
    0xcc569c7000000ULL,
    0x18ed348800000ULL,
    0x39856c7400000ULL,
    0x8646fcae00000ULL,
    0xb5a3487300000ULL,
    0x24765a9c80000ULL,
    0x1cd92b4ac0000ULL,
    0x95f94e0fe0000ULL,
    0x94ff724d0000ULL,
    0x1bf65a84c8000ULL,
    0x8f992b60ec000ULL,
    0x86d94e1cce000ULL,
    0x8afff72b8b000ULL,
    0xe8ce5a99b8800ULL,
    0xe1b52b55d4c00ULL,
    0x403b4e2bdb600ULL,
    0xcdacf70bb0100ULL,
    0x5fa3da90d8880ULL,
    0x13756b46c4ec0ULL,
    0x69596e3f727a0ULL,
    0x21bf4711cd650ULL,
    0xf7ea32b9fe538ULL,
    0x3000ff4893fdcULL,
    0x480020091d96aULL,
    0x2400b007364adULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x7800000000000ULL,
    0xc400000000000ULL,
    0x5a00000000000ULL,
    0xa300000000000ULL,
    0xd880000000000ULL,
    0xbf40000000000ULL,
    0xfde0000000000ULL,
    0xc90000000000ULL,
    0x50e8000000000ULL,
    0x146c000000000ULL,
    0xe52e000000000ULL,
    0x39cb000000000ULL,
    0x1adf800000000ULL,
    0xff4400000000ULL,
    0x3e3e600000000ULL,
    0xa8e2b00000000ULL,
    0xd114a80000000ULL,
    0xaa29ec0000000ULL,
    0xe74a560000000ULL,
    0xa89ed90000000ULL,
    0x3b948d8000000ULL,
    0x4a6f62c000000ULL,
    0xfc2b0da000000ULL,
    0x724b22d000000ULL,
    0x951d6d8800000ULL,
    0xf15592f400000ULL,
    0x514fc59600000ULL,
    0xd9b7ec500000ULL,
    0xee14138080000ULL,
    0x3caae7edc0000ULL,
    0x79097e36a0000ULL,
    0x7aff750210000ULL,
    0x72c6bba9d8000ULL,
    0x3f240bd134000ULL,
    0x9bb2a835e2000ULL,
    0xd75eec202f000ULL,
    0xcd33d61aa9800ULL,
    0xa91d992b5dc00ULL,
    0x2f536dbc9e600ULL,
    0x84e92d560d00ULL,
    0xe61845adb1f80ULL,
    0xf1d33ed9487c0ULL,
    0xba0c73ae2cd20ULL,
    0xd27f57d3f6070ULL,
    0x858456149c478ULL,
    0xeec5d9213eb14ULL,
    0x71230dbef0e4eULL,
    0xbab722c328a9fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0xa400000000000ULL,
    0x3e00000000000ULL,
    0x5900000000000ULL,
    0x2e80000000000ULL,
    0x9bc0000000000ULL,
    0xd20000000000ULL,
    0x2e90000000000ULL,
    0x89c8000000000ULL,
    0x394c000000000ULL,
    0x4b8a000000000ULL,
    0x802d000000000ULL,
    0x9979800000000ULL,
    0x1f95400000000ULL,
    0xcb4c600000000ULL,
    0x408f900000000ULL,
    0x15ac580000000ULL,
    0x7f3eac0000000ULL,
    0x58f2320000000ULL,
    0xa0bfbb0000000ULL,
    0x61b01f8000000ULL,
    0x445a69c000000ULL,
    0x89019fa000000ULL,
    0x968329f000000ULL,
    0x4fc7ff9800000ULL,
    0x3b21b9d400000ULL,
    0x4b9227be00000ULL,
    0x3d4a55fd00000ULL,
    0xc58c75b180000ULL,
    0x312a7eeb40000ULL,
    0x1bf8323a20000ULL,
    0x8652bb0db0000ULL,
    0x5699f9968000ULL,
    0x279f29f614000ULL,
    0x7025ffa31a000ULL,
    0xfe10b9ec47000ULL,
    0xab09a7aff8800ULL,
    0x8aee15c14c400ULL,
    0x66db95a026e00ULL,
    0xa0c1aef9e9100ULL,
    0xb0a38a16d5f80ULL,
    0x84d7c7197af40ULL,
    0x5ca815bb56ee0ULL,
    0x9b9eedbe15f0ULL,
    0x87366a1450628ULL,
    0x39d172239b1cULL,
    0x8e202d82ace56ULL,
    0x4714d2e0776bbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x5800000000000ULL,
    0xb400000000000ULL,
    0xfe00000000000ULL,
    0x5500000000000ULL,
    0xc280000000000ULL,
    0x6dc0000000000ULL,
    0x67a0000000000ULL,
    0x1dd0000000000ULL,
    0xf8e8000000000ULL,
    0xf46c000000000ULL,
    0xc4ae000000000ULL,
    0x2d8b000000000ULL,
    0x79b800000000ULL,
    0x6bd1400000000ULL,
    0xa1e8e00000000ULL,
    0xcce9100000000ULL,
    0x4a6cb80000000ULL,
    0x31af2c0000000ULL,
    0xff0e020000000ULL,
    0x325a5f0000000ULL,
    0xb870e18000000ULL,
    0x423e7ec000000ULL,
    0x610361a000000ULL,
    0x7c833ed000000ULL,
    0x98c5818800000ULL,
    0xb5212ed400000ULL,
    0x2812b99e00000ULL,
    0x2b4f42f900000ULL,
    0x17bc5ba180000ULL,
    0x69400dec40000ULL,
    0x1b66021120000ULL,
    0xaef65f3730000ULL,
    0x21fee1a8c8000ULL,
    0xbba57ee044000ULL,
    0x5fd0e19ef2000ULL,
    0x1fee7ec95f000ULL,
    0x39eb6191d9800ULL,
    0x98ef3eeb3dc00ULL,
    0x46b81b9bf600ULL,
    0x2caa2ec989900ULL,
    0xd189399ee5580ULL,
    0x159e02f483d40ULL,
    0x74d4bba536760ULL,
    0xc8691de4ec1f0ULL,
    0x36aaba3d27f68ULL,
    0x828973121588cULL,
    0x2618e3bc2a4eeULL,
    0x7d9321e4e3d77ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x6800000000000ULL,
    0xa400000000000ULL,
    0x200000000000ULL,
    0xcf00000000000ULL,
    0xd980000000000ULL,
    0x9240000000000ULL,
    0x26e0000000000ULL,
    0x6bd0000000000ULL,
    0x5588000000000ULL,
    0x360c000000000ULL,
    0xf74a000000000ULL,
    0x5629000000000ULL,
    0x5ebe800000000ULL,
    0xf555c00000000ULL,
    0x8a4ce00000000ULL,
    0x88aef00000000ULL,
    0x217d880000000ULL,
    0xdbf7c40000000ULL,
    0xc47b960000000ULL,
    0xc976c10000000ULL,
    0x1db9af8000000ULL,
    0x2ad3ebc000000ULL,
    0x450f2fa000000ULL,
    0x20ca2bf000000ULL,
    0xcd69cfb800000ULL,
    0x9cdddbcc00000ULL,
    0x7242c7a200000ULL,
    0x16e3dfe500000ULL,
    0x63d7b1a280000ULL,
    0x18eeed6c0000ULL,
    0x5c0f1614e0000ULL,
    0x9c4a013250000ULL,
    0x8dabcfbd58000ULL,
    0x3f8dbd8a4000ULL,
    0xa3647903a000ULL,
    0x73df1fd927000ULL,
    0xfbc5d1afe8800ULL,
    0x7ca5ded9ad400ULL,
    0x79367e22c7200ULL,
    0xa45f353e38300ULL,
    0x6087d196b4180ULL,
    0xbec0defb438c0ULL,
    0xfe22fe19337a0ULL,
    0x38f3f53fda7d0ULL,
    0x8bfdb196f79f8ULL,
    0x9e37eefff0cd4ULL,
    0x79d9961ba9746ULL,
    0x60c3c13b8d77bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x6800000000000ULL,
    0xc400000000000ULL,
    0x2a00000000000ULL,
    0x1f00000000000ULL,
    0x6280000000000ULL,
    0xc7c0000000000ULL,
    0xb060000000000ULL,
    0xb8f0000000000ULL,
    0xdee8000000000ULL,
    0x126c000000000ULL,
    0x40aa000000000ULL,
    0xdd49000000000ULL,
    0x98f8800000000ULL,
    0x9db3c00000000ULL,
    0x6949e00000000ULL,
    0x7afab00000000ULL,
    0x16b7480000000ULL,
    0x79cadc0000000ULL,
    0xe3c220000000ULL,
    0x2a51c10000000ULL,
    0x33fdf68000000ULL,
    0x1d3269c000000ULL,
    0x258d76a000000ULL,
    0xda1da9f000000ULL,
    0xda8696b800000ULL,
    0x9bc219fc00000ULL,
    0x56635ea200000ULL,
    0x79f205cd00000ULL,
    0xe16e1ca880000ULL,
    0x6c2ab4cac0000ULL,
    0x7f8d423020000ULL,
    0xd18b13b30000ULL,
    0x2c035ebc38000ULL,
    0x2e0205e50c000ULL,
    0x55061cb382000ULL,
    0x4d86b4dd97000ULL,
    0xcd474221a8800ULL,
    0xb3a1b117c5400ULL,
    0x2293deb716a00ULL,
    0x791dc5f669b00ULL,
    0xae05fc8764d80ULL,
    0x950504e2e71c0ULL,
    0x2d808a34710a0ULL,
    0xfd44ad1e59ab0ULL,
    0xdba41c9cdc028ULL,
    0xe693b4f8aadb4ULL,
    0x531dc21d93ff2ULL,
    0xb10771061f013ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x3800000000000ULL,
    0xac00000000000ULL,
    0x1200000000000ULL,
    0x8d00000000000ULL,
    0x4280000000000ULL,
    0x90c0000000000ULL,
    0x8460000000000ULL,
    0x5d0000000000ULL,
    0x1de8000000000ULL,
    0x596c000000000ULL,
    0x3aaa000000000ULL,
    0x434b000000000ULL,
    0x29da800000000ULL,
    0x5a13400000000ULL,
    0xbe0a200000000ULL,
    0xaf8b00000000ULL,
    0x85e3d80000000ULL,
    0x90954c0000000ULL,
    0x7acf8a0000000ULL,
    0x109a090000000ULL,
    0x2f30578000000ULL,
    0x7af880c000000ULL,
    0x5de2d7a000000ULL,
    0x2c97c0f000000ULL,
    0x60caf7a800000ULL,
    0x99870fc00000ULL,
    0xd3b1afa600000ULL,
    0x75397cd700000ULL,
    0x160485a280000ULL,
    0xfb0385f440000ULL,
    0x6987aa3220000ULL,
    0xd145b91070000ULL,
    0xf1230fb338000ULL,
    0x72f58ce104000ULL,
    0x5c1dfd9c42000ULL,
    0xd8f639cb5f000ULL,
    0xbd1b5814e8800ULL,
    0xa8710c3c8dc00ULL,
    0x90dd2a341e200ULL,
    0x5696f91249100ULL,
    0xa9c92f87bca80ULL,
    0xdd1d3cdb873c0ULL,
    0x787625b97c5e0ULL,
    0xb8df75c1e66f0ULL,
    0xf296d20d2dbc8ULL,
    0x2fcc053dafed4ULL,
    0xee1dfdb191296ULL,
    0xa5f639daa80d7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0xcc00000000000ULL,
    0x2e00000000000ULL,
    0xd500000000000ULL,
    0xbe80000000000ULL,
    0xd4c0000000000ULL,
    0x5760000000000ULL,
    0x8d50000000000ULL,
    0x6ad8000000000ULL,
    0x1e5c000000000ULL,
    0xa59a000000000ULL,
    0x8b7d000000000ULL,
    0xff6d800000000ULL,
    0xc996400000000ULL,
    0x85bc200000000ULL,
    0xbf0f500000000ULL,
    0xf140f80000000ULL,
    0x38268c0000000ULL,
    0x28756a0000000ULL,
    0xb82d710000000ULL,
    0xf8b28b8000000ULL,
    0x454fc54000000ULL,
    0xcae70ba000000ULL,
    0xce15857000000ULL,
    0xf1f92b9800000ULL,
    0x61abd57400000ULL,
    0x31f6538600000ULL,
    0x916a194700000ULL,
    0x3c90998d80000ULL,
    0xcb3f787240000ULL,
    0x3c94a2220000ULL,
    0x1222210770000ULL,
    0x57727397f8000ULL,
    0xb9a9495424000ULL,
    0xcdf261954e000ULL,
    0x3768f467b3000ULL,
    0x4d93a01a28800ULL,
    0xefb810326c400ULL,
    0xe00b581073e00ULL,
    0xc29c3649d00ULL,
    0x6564322d3ce80ULL,
    0xee52ed1dfa040ULL,
    0xdd5b39b514d60ULL,
    0x221b687697a70ULL,
    0x3f38121be78c8ULL,
    0xc1cdbd391ffb4ULL,
    0x1923c1a1bbae6ULL,
    0x54f1e44c344a1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xc800000000000ULL,
    0x8c00000000000ULL,
    0x8600000000000ULL,
    0x8d00000000000ULL,
    0x1180000000000ULL,
    0x6640000000000ULL,
    0xb1e0000000000ULL,
    0x16d0000000000ULL,
    0x3368000000000ULL,
    0x63ec000000000ULL,
    0xf22a000000000ULL,
    0x988f000000000ULL,
    0x763b800000000ULL,
    0x71c5c00000000ULL,
    0x9aa5a00000000ULL,
    0x16b7900000000ULL,
    0xd3fef80000000ULL,
    0x2966f40000000ULL,
    0xb11860000000ULL,
    0xdfcb690000000ULL,
    0x5a6f8000000ULL,
    0xe4535cc000000ULL,
    0x15abefa000000ULL,
    0x4bc99cf000000ULL,
    0xa25dcf8800000ULL,
    0x5357ccd400000ULL,
    0x572c97ba00000ULL,
    0xf509a8f700000ULL,
    0x6e78699b80000ULL,
    0x9921f5d440000ULL,
    0x5ff6201fe0000ULL,
    0xc49e500dd0000ULL,
    0x1b71581398000ULL,
    0x375e643864000ULL,
    0x66d4fe3742000ULL,
    0xb685d049f000ULL,
    0x17ee49b9c8800ULL,
    0xa02fa5c1bcc00ULL,
    0xd78f7825c6600ULL,
    0xe0bc340e8db00ULL,
    0xd07a60f0f080ULL,
    0xd1853906b0540ULL,
    0xc643378d88b60ULL,
    0x21e138ce5de90ULL,
    0xded511b016d78ULL,
    0xbf6ec1d114334ULL,
    0xe5e806016acc6ULL,
    0x7f2da910f3247ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x7800000000000ULL,
    0x8400000000000ULL,
    0x3600000000000ULL,
    0xb500000000000ULL,
    0x2e80000000000ULL,
    0x15c0000000000ULL,
    0xdc20000000000ULL,
    0xdb70000000000ULL,
    0x2a48000000000ULL,
    0x5cc000000000ULL,
    0xb08a000000000ULL,
    0x2869000000000ULL,
    0x87bc800000000ULL,
    0xae47400000000ULL,
    0xf767600000000ULL,
    0xba90d00000000ULL,
    0x611e780000000ULL,
    0x3ef68c0000000ULL,
    0x980fe20000000ULL,
    0x30ac230000000ULL,
    0x911bc88000000ULL,
    0xe6f272c000000ULL,
    0xcc0d48a000000ULL,
    0xceac32f000000ULL,
    0xd81ea8b800000ULL,
    0x7a77a2dc00000ULL,
    0x5acdb0a600000ULL,
    0x8908fefd00000ULL,
    0x3828aaab80000ULL,
    0x99d911d040000ULL,
    0x5851e02020000ULL,
    0xc8be901530000ULL,
    0x3fc5983728000ULL,
    0xab211c13ec000ULL,
    0x6ef6fa25ba000ULL,
    0x900a7f3e9f000ULL,
    0xacaa529798800ULL,
    0x9318ddd585400ULL,
    0x1df1621bf7a00ULL,
    0xd38e632354900ULL,
    0xc3ea28958e780ULL,
    0xa4fce2d480a40ULL,
    0x53a0d09cc1e60ULL,
    0xbe312ed0e1070ULL,
    0x7aaa52bbd01e8ULL,
    0x5618ddfdf90fcULL,
    0x2b71622b946caULL,
    0xf24e63076fd37ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0xfc00000000000ULL,
    0xde00000000000ULL,
    0x8900000000000ULL,
    0xa080000000000ULL,
    0xccc0000000000ULL,
    0x5820000000000ULL,
    0x60b0000000000ULL,
    0xe338000000000ULL,
    0x1ebc000000000ULL,
    0xfa000000000ULL,
    0x9b1f000000000ULL,
    0x8e0e800000000ULL,
    0x8bc0c00000000ULL,
    0x49a3200000000ULL,
    0x98f3d00000000ULL,
    0x65da980000000ULL,
    0xb92a1c0000000ULL,
    0x56f43a0000000ULL,
    0xd4d81b0000000ULL,
    0x8dae6e8000000ULL,
    0x88307f4000000ULL,
    0x13faeea000000ULL,
    0x389fbf7000000ULL,
    0xd94f4e8800000ULL,
    0x10a0af7c00000ULL,
    0xa074f6ba00000ULL,
    0xd1a634300000ULL,
    0x3b0ed48d80000ULL,
    0x5547a46140000ULL,
    0x3c612025a0000ULL,
    0xa850d01270000ULL,
    0x35ae1813a8000ULL,
    0xdc35dc3a5c000ULL,
    0xa1f99a0366000ULL,
    0x979b0b31f3000ULL,
    0xc4cfd6b158800ULL,
    0x5ee5b36834c00ULL,
    0x63964c855b200ULL,
    0x1c0eb85884f00ULL,
    0xd4c19a1713e80ULL,
    0x3c270b0188040ULL,
    0x1ab5d68a6d4e0ULL,
    0xb03ab378ee830ULL,
    0xdd38cc989e408ULL,
    0xc7be78543785cULL,
    0xf87aba09eb5b2ULL,
    0x53d8db10cc5a1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0xe400000000000ULL,
    0xc600000000000ULL,
    0x3b00000000000ULL,
    0xcd80000000000ULL,
    0xbd40000000000ULL,
    0xab60000000000ULL,
    0xf710000000000ULL,
    0x7378000000000ULL,
    0xe3fc000000000ULL,
    0x783a000000000ULL,
    0xab9b000000000ULL,
    0xbeab800000000ULL,
    0x7fa2c00000000ULL,
    0xc731e00000000ULL,
    0x2409d00000000ULL,
    0x9215080000000ULL,
    0x99f9e40000000ULL,
    0xb93e4e0000000ULL,
    0x671bb50000000ULL,
    0xef6afd8000000ULL,
    0x95065e4000000ULL,
    0xf2837da000000ULL,
    0x86c39e7000000ULL,
    0x15231da800000ULL,
    0x5df38e7400000ULL,
    0x1b2c75a200000ULL,
    0xc6e17a4d00000ULL,
    0x9a56d39380000ULL,
    0xd01afb43c0000ULL,
    0xe0ef6820e0000ULL,
    0xe542f42c90000ULL,
    0x9762a62a38000ULL,
    0xcd17812154000ULL,
    0x527bbbbee6000ULL,
    0x1f7f0f5dab000ULL,
    0x11fc4e0ae8800ULL,
    0x8d3cb53e4dc00ULL,
    0xa91b7dbcc2600ULL,
    0x206f9e5b65700ULL,
    0xb6811d9c6e180ULL,
    0x90c48e7037040ULL,
    0x2625f5a1b35a0ULL,
    0x6474ba584d7d0ULL,
    0x486eb39bc2c98ULL,
    0xb286eb61e4904ULL,
    0x66c200162ef6eULL,
    0x2527003217c7dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x5800000000000ULL,
    0x5400000000000ULL,
    0x3200000000000ULL,
    0x9300000000000ULL,
    0x7780000000000ULL,
    0x71c0000000000ULL,
    0x58a0000000000ULL,
    0x3b90000000000ULL,
    0xe038000000000ULL,
    0x6bc000000000ULL,
    0xbfa000000000ULL,
    0xa11f000000000ULL,
    0x31ea800000000ULL,
    0x79e0400000000ULL,
    0x8776e00000000ULL,
    0x29c9700000000ULL,
    0x4e34e80000000ULL,
    0x9928d40000000ULL,
    0xe1c69a0000000ULL,
    0x70a47f0000000ULL,
    0x9790738000000ULL,
    0x363b0a4000000ULL,
    0xffb8f3a000000ULL,
    0xbb784a7000000ULL,
    0x955e938800000ULL,
    0x8b8e7a5c00000ULL,
    0x6d561b9e00000ULL,
    0x2d1f9e4900000ULL,
    0x97ea899280000ULL,
    0x78e6456dc0000ULL,
    0xd3f2880f60000ULL,
    0x778ee431b0000ULL,
    0x3356123948000ULL,
    0xc8199b0f54000ULL,
    0xa96ee1a946000ULL,
    0xe1a1d162fb000ULL,
    0xeb10721db8800ULL,
    0xc47fab2fbcc00ULL,
    0x54de699542200ULL,
    0x134c354904d00ULL,
    0x6d76e0377f980ULL,
    0x2ec97016068c0ULL,
    0xabb4e8205aa20ULL,
    0xcbe8d41368b90ULL,
    0x96e69a0f45048ULL,
    0x6ef47f0f5ea6cULL,
    0x1d0873b917986ULL,
    0x98170a49c3a45ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x2800000000000ULL,
    0xd400000000000ULL,
    0xb600000000000ULL,
    0xcd00000000000ULL,
    0xaa80000000000ULL,
    0xfc0000000000ULL,
    0x5a60000000000ULL,
    0x61d0000000000ULL,
    0xdef8000000000ULL,
    0x977c000000000ULL,
    0xa33a000000000ULL,
    0xae19000000000ULL,
    0x396e800000000ULL,
    0x47a7c00000000ULL,
    0xa8b5600000000ULL,
    0x6caed00000000ULL,
    0xaac7b80000000ULL,
    0x84e76c0000000ULL,
    0x4813520000000ULL,
    0x919ba10000000ULL,
    0xa02fdf8000000ULL,
    0x1006a04000000ULL,
    0x18035fa000000ULL,
    0x9c04607000000ULL,
    0x9202bfb800000ULL,
    0x5304704c00000ULL,
    0xb386e7b200000ULL,
    0x13460c7900000ULL,
    0x98a56d9f80000ULL,
    0x9131114a40000ULL,
    0xb0ead83520000ULL,
    0x13e5bc09b0000ULL,
    0x5596ea03b8000ULL,
    0xd3d9cd21c4000ULL,
    0x80dac06000ULL,
    0xddf3c1781b000ULL,
    0x410f602ec8800ULL,
    0xfd77d0169d400ULL,
    0x81c9383513200ULL,
    0x8690ac3834b00ULL,
    0xa05e32279e780ULL,
    0x73497129ef040ULL,
    0xb55267bc37960ULL,
    0xf838cc56e2df0ULL,
    0x999e8dbead428ULL,
    0x24280157eb374ULL,
    0x1e00001a50356ULL,
    0xd900001fe8df5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x5800000000000ULL,
    0x5400000000000ULL,
    0xc600000000000ULL,
    0x4f00000000000ULL,
    0x1f80000000000ULL,
    0x9440000000000ULL,
    0xfae0000000000ULL,
    0xe2d0000000000ULL,
    0xa4b8000000000ULL,
    0xe63c000000000ULL,
    0xa4fe000000000ULL,
    0x7ad9000000000ULL,
    0x192e800000000ULL,
    0x69e4400000000ULL,
    0x6752e00000000ULL,
    0xa5fed00000000ULL,
    0x765fa80000000ULL,
    0x48ead40000000ULL,
    0x7243b60000000ULL,
    0xe5e5410000000ULL,
    0xb554a78000000ULL,
    0xccfe5a4000000ULL,
    0xd6dc27a000000ULL,
    0x9b2f1a5000000ULL,
    0x48e6479800000ULL,
    0x3bd08a6c00000ULL,
    0xbc3b8f8e00000ULL,
    0x91f9ce7900000ULL,
    0x5bf1bf80000ULL,
    0xbfeccb6ac0000ULL,
    0xc9c1a83860000ULL,
    0xefa3d40fd0000ULL,
    0x54b53626d8000ULL,
    0xf7ad012764000ULL,
    0xa92047b7f6000ULL,
    0x98758a62eb000ULL,
    0x9a8b0fa8c9800ULL,
    0x1f548e795d400ULL,
    0xa1ff91a702200ULL,
    0xf85a5b4ce4500ULL,
    0xfbeae02f36980ULL,
    0xa7c2d0158be40ULL,
    0xaca1a81d99720ULL,
    0xd933d43cc5510ULL,
    0xeaed3617c79f8ULL,
    0x341012523014ULL,
    0xf16647bb15366ULL,
    0x50908a651ed7dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x4800000000000ULL,
    0x5400000000000ULL,
    0x5a00000000000ULL,
    0xaf00000000000ULL,
    0xe280000000000ULL,
    0x6cc0000000000ULL,
    0xe7e0000000000ULL,
    0xa590000000000ULL,
    0xd978000000000ULL,
    0xcafc000000000ULL,
    0x59ba000000000ULL,
    0xbb19000000000ULL,
    0x7529800000000ULL,
    0x31e3c00000000ULL,
    0x8c95600000000ULL,
    0xfaf9b00000000ULL,
    0xa1bbc80000000ULL,
    0xd71a5c0000000ULL,
    0x23287a0000000ULL,
    0xd8e5650000000ULL,
    0xcf14d98000000ULL,
    0x81bbca4000000ULL,
    0x671f59a000000ULL,
    0x1b2d0a7000000ULL,
    0xd4e1b9b800000ULL,
    0x89127a7400000ULL,
    0x20bc91aa00000ULL,
    0x709b566100000ULL,
    0x3a6bc3bc80000ULL,
    0xbd421f4ac0000ULL,
    0xa7a3c80420000ULL,
    0xbbb65c11f0000ULL,
    0xc68a7a1dd8000ULL,
    0x7050650894000ULL,
    0x951f59b386000ULL,
    0x302d0a7e2b000ULL,
    0x4461b997e8800ULL,
    0xed27a75ed400ULL,
    0x37dc918422e00ULL,
    0x42cb565655b00ULL,
    0xbc73c38da7b80ULL,
    0x11ee1f66dbdc0ULL,
    0x2201c80ab14a0ULL,
    0x3035c11b94b0ULL,
    0xd481fa2c85548ULL,
    0x95c6a53e0f124ULL,
    0xec61b9a8d6e2eULL,
    0x8ad27a5dc3a0dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x7800000000000ULL,
    0x8400000000000ULL,
    0xf200000000000ULL,
    0x7900000000000ULL,
    0x7d80000000000ULL,
    0x6bc0000000000ULL,
    0x8ce0000000000ULL,
    0xce90000000000ULL,
    0x8338000000000ULL,
    0xe1bc000000000ULL,
    0xb4fe000000000ULL,
    0xcb5d000000000ULL,
    0x196a800000000ULL,
    0x40e1400000000ULL,
    0xe091600000000ULL,
    0x1c3cf00000000ULL,
    0x9f38380000000ULL,
    0x27be540000000ULL,
    0x27fdca0000000ULL,
    0xbbda0d0000000ULL,
    0x8528298000000ULL,
    0x5ac3d84000000ULL,
    0x2d64a9a000000ULL,
    0x4353985000000ULL,
    0xcd949b800000ULL,
    0xd7af286c00000ULL,
    0x578491a600000ULL,
    0x46c1cc4d00000ULL,
    0xeb62839a80000ULL,
    0xd054255340000ULL,
    0x7c583805e0000ULL,
    0x4bee5409b0000ULL,
    0x4da5ca06d8000ULL,
    0x8b360d2a14000ULL,
    0xb40e29be16000ULL,
    0x43b2d86e7b000ULL,
    0x8c4829ad29800ULL,
    0x3b93d86dac400ULL,
    0xb0bca98ff3200ULL,
    0xe57f985fd4100ULL,
    0x3e9f49b6b7780ULL,
    0x128e28632b2c0ULL,
    0x84f0118ec1760ULL,
    0x7aed8c4720b90ULL,
    0xec2163b890048ULL,
    0x6f5954b493a4ULL,
    0x3bef6017dd8eeULL,
    0x75a1f01c8a8c5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0xf800000000000ULL,
    0x3c00000000000ULL,
    0x8e00000000000ULL,
    0x1d00000000000ULL,
    0xac80000000000ULL,
    0x61c0000000000ULL,
    0x90e0000000000ULL,
    0xaf90000000000ULL,
    0xac8000000000ULL,
    0x314c000000000ULL,
    0x330e000000000ULL,
    0x4daf000000000ULL,
    0xca9b800000000ULL,
    0xf664400000000ULL,
    0x2151a00000000ULL,
    0x85afd00000000ULL,
    0xde9a380000000ULL,
    0x9c61dc0000000ULL,
    0x1e50120000000ULL,
    0x422bf50000000ULL,
    0x32d8398000000ULL,
    0x4fc7e7c000000ULL,
    0x5de5b9a000000ULL,
    0xdb10a7d000000ULL,
    0xc70999a800000ULL,
    0xd7a837f400000ULL,
    0xbd9f819600000ULL,
    0x65e17bfb00000ULL,
    0x87118bb580000ULL,
    0xf90fc2fc40000ULL,
    0xa2a9b81ea0000ULL,
    0xd5199c2430000ULL,
    0xb627b21a98000ULL,
    0x84f7251fd4000ULL,
    0xe71f81902a000ULL,
    0x65217bf787000ULL,
    0x15718b82e9800ULL,
    0xba5fc2fbecc00ULL,
    0x6c01b81bcf600ULL,
    0x16059c2535d00ULL,
    0x9101b226bb080ULL,
    0x4a8425032e240ULL,
    0xb8c2019ea5860ULL,
    0x8e663bec92530ULL,
    0x5d55abaea3828ULL,
    0x2bab52f73bae4ULL,
    0x5399a0216e0ceULL,
    0xc8e3d03f0597bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0x7400000000000ULL,
    0xbe00000000000ULL,
    0x9700000000000ULL,
    0xa380000000000ULL,
    0xb5c0000000000ULL,
    0x50e0000000000ULL,
    0x6570000000000ULL,
    0x5cd8000000000ULL,
    0xaa5c000000000ULL,
    0x691a000000000ULL,
    0x3abd000000000ULL,
    0xec6d800000000ULL,
    0x7660400000000ULL,
    0xfe37e00000000ULL,
    0x6b7c100000000ULL,
    0x7a8ba80000000ULL,
    0x1010340000000ULL,
    0xaf6b5a0000000ULL,
    0x5be0fd0000000ULL,
    0x94f6e68000000ULL,
    0x441a984000000ULL,
    0xd83966a000000ULL,
    0x412bd87000000ULL,
    0xeac1069800000ULL,
    0xef66887400000ULL,
    0x42b2ceb600000ULL,
    0x413bec6900000ULL,
    0xfdaa5c9980000ULL,
    0xc086754cc0000ULL,
    0x6844283160000ULL,
    0x92217421b0000ULL,
    0x86133a0558000ULL,
    0x346dad08ec000ULL,
    0xd2652eaa6e000ULL,
    0x1837fc667b000ULL,
    0x9879f4be28800ULL,
    0xdf0a4145d4400ULL,
    0x46d57212b3e00ULL,
    0xe20c89398c500ULL,
    0x1c505c8bde180ULL,
    0x3b4b756323540ULL,
    0x3ef1a836c4f20ULL,
    0xbd1d342fbb590ULL,
    0xb4beda14c8248ULL,
    0x336cbd17a4664ULL,
    0x9e306860b1d2ULL,
    0x39f7886d91805ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x4800000000000ULL,
    0x7400000000000ULL,
    0x9e00000000000ULL,
    0x5b00000000000ULL,
    0x3f80000000000ULL,
    0xe340000000000ULL,
    0xa1a0000000000ULL,
    0x5230000000000ULL,
    0xa618000000000ULL,
    0xb55c000000000ULL,
    0x1cfe000000000ULL,
    0xdac9000000000ULL,
    0x92d2800000000ULL,
    0x348cc00000000ULL,
    0x63f1600000000ULL,
    0xc47c900000000ULL,
    0xac0cc80000000ULL,
    0xf531f40000000ULL,
    0xd39f8e0000000ULL,
    0x31a130000000ULL,
    0xe1da4a8000000ULL,
    0x683f8e4000000ULL,
    0x9c2b2aa000000ULL,
    0xb7431e5000000ULL,
    0xefa7e29800000ULL,
    0xd132ea4c00000ULL,
    0xd5986c8a00000ULL,
    0xf418f95100000ULL,
    0xc5a262d80000ULL,
    0x727b7707c0000ULL,
    0xb30f0c8260000ULL,
    0xfcb1696f10000ULL,
    0x6fda6e3208000ULL,
    0x4b3f43379c000ULL,
    0x7fab62918e000ULL,
    0xce072a4bcb000ULL,
    0xc3038cbb31800ULL,
    0x9384a97ab9400ULL,
    0x31418e1de5200ULL,
    0x58a313188a700ULL,
    0xdcb0cab711780ULL,
    0xbfdf4e5e09540ULL,
    0x933c4aad9c4a0ULL,
    0x33aa8e788e290ULL,
    0x6c07aaba4a298ULL,
    0x7206de4ef1edcULL,
    0x69040295d99b2ULL,
    0xb682ba5d747e5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x6800000000000ULL,
    0x9400000000000ULL,
    0xb200000000000ULL,
    0x4900000000000ULL,
    0x4a80000000000ULL,
    0x7740000000000ULL,
    0xeea0000000000ULL,
    0xd0f0000000000ULL,
    0x1068000000000ULL,
    0xb72c000000000ULL,
    0x618e000000000ULL,
    0x5d7b000000000ULL,
    0xb316800000000ULL,
    0x339c00000000ULL,
    0x4733600000000ULL,
    0xdc8df00000000ULL,
    0xf5f9380000000ULL,
    0xf552ac0000000ULL,
    0xeb1a760000000ULL,
    0xae84f30000000ULL,
    0x4d414c8000000ULL,
    0x23a263c000000ULL,
    0xf0722ca000000ULL,
    0xd22f93d000000ULL,
    0x350b148800000ULL,
    0x3d3d3ffc00000ULL,
    0xa031629200000ULL,
    0xe909ccd900000ULL,
    0xeb382e3e80000ULL,
    0x9337af0b40000ULL,
    0xce8c028c20000ULL,
    0xcff3cc690000ULL,
    0xd7d7963f88000ULL,
    0x85cc30eac000ULL,
    0xf22514a992000ULL,
    0xd4b63ffa67000ULL,
    0x794fe28f71800ULL,
    0x301c0ce679c00ULL,
    0x5d054e2f94200ULL,
    0xb8815f2027b00ULL,
    0x9e43ba81d0f80ULL,
    0x142450d9a8cc0ULL,
    0xcfb680353cb60ULL,
    0x6ac9c00e9b670ULL,
    0xd5db60288bab8ULL,
    0x38e1f02f439bcULL,
    0xa0d7382e4fd8aULL,
    0xa5d9ac328cb33ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x2800000000000ULL,
    0xec00000000000ULL,
    0x6e00000000000ULL,
    0xff00000000000ULL,
    0x6f80000000000ULL,
    0x72c0000000000ULL,
    0xfd60000000000ULL,
    0xed0000000000ULL,
    0x8368000000000ULL,
    0x89ac000000000ULL,
    0xca000000000ULL,
    0x41f000000000ULL,
    0x8a74800000000ULL,
    0xd559c00000000ULL,
    0xdbd4a00000000ULL,
    0x51e8d00000000ULL,
    0xa5eba80000000ULL,
    0x6fe9140000000ULL,
    0xc2ea220000000ULL,
    0x646b190000000ULL,
    0x4229c78000000ULL,
    0x200d48c000000ULL,
    0x307d67a000000ULL,
    0xd02598f000000ULL,
    0xc976cf8800000ULL,
    0x4cdc8cec00000ULL,
    0x3214ed8200000ULL,
    0x310b95d700000ULL,
    0x60ff2a1980000ULL,
    0x6d65dd28c0000ULL,
    0x36d4cda260000ULL,
    0x976a85c470000ULL,
    0x53a8221538000ULL,
    0x55c8192bc4000ULL,
    0x169f47b7e2000ULL,
    0x63788deaf000ULL,
    0xca7f47ae50800ULL,
    0x352788f768c00ULL,
    0x53f747aeada00ULL,
    0x5c9b88c64eb00ULL,
    0x6b3547ad60880ULL,
    0xccf888c9f0840ULL,
    0x6363c79bf8620ULL,
    0x19d248d0a5c10ULL,
    0xa0e9e7af126a8ULL,
    0xe56c58f5d7c74ULL,
    0x2aaa6f82b5bb6ULL,
    0x89485ce85a257ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xd800000000000ULL,
    0x2c00000000000ULL,
    0xbe00000000000ULL,
    0x4100000000000ULL,
    0xa280000000000ULL,
    0xd0c0000000000ULL,
    0x7160000000000ULL,
    0xed90000000000ULL,
    0x7658000000000ULL,
    0x6a9c000000000ULL,
    0xb5fe000000000ULL,
    0x916f000000000ULL,
    0xc1b4800000000ULL,
    0x31ee400000000ULL,
    0x2671200000000ULL,
    0x70bd00000000ULL,
    0x83a5580000000ULL,
    0xdbf20c0000000ULL,
    0x44baa0000000ULL,
    0x7884f10000000ULL,
    0xcbc6458000000ULL,
    0x88e3bb4000000ULL,
    0xe4d765a000000ULL,
    0x7e786b5000000ULL,
    0x572a3da800000ULL,
    0xd216674c00000ULL,
    0x221f979e00000ULL,
    0x863d964700000ULL,
    0x458d523180000ULL,
    0x43606d1740000ULL,
    0x229317b920000ULL,
    0x6ddfd67370000ULL,
    0x44da720d78000ULL,
    0x7258bd0414000ULL,
    0x609ccf8eee000ULL,
    0x46fc9a59b3000ULL,
    0xbcec781211800ULL,
    0x2af5dc2aa8c00ULL,
    0x1ec8f23b5ca00ULL,
    0xd445fd05a3500ULL,
    0xaca76f97b8e80ULL,
    0xd0760a4c75a40ULL,
    0xa20c8027bfbe0ULL,
    0x3b2240243a270ULL,
    0x703720253cd48ULL,
    0xdca8d017f3db4ULL,
    0x3457d81db1e5eULL,
    0x41bf4c1f18659ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x8400000000000ULL,
    0xc600000000000ULL,
    0xaf00000000000ULL,
    0x1980000000000ULL,
    0xe440000000000ULL,
    0x2960000000000ULL,
    0x73b0000000000ULL,
    0x8128000000000ULL,
    0x96c000000000ULL,
    0xfa0e000000000ULL,
    0x44bf000000000ULL,
    0x512800000000ULL,
    0x86b8400000000ULL,
    0x2c12600000000ULL,
    0x103b500000000ULL,
    0x61d4a80000000ULL,
    0x518240000000ULL,
    0xbf07660000000ULL,
    0x31807d0000000ULL,
    0x9844c88000000ULL,
    0xdb6525c000000ULL,
    0xc2b6a8a000000ULL,
    0xb3ae75d000000ULL,
    0x32aa00a800000ULL,
    0x982a51fc00000ULL,
    0x7eb669e00000ULL,
    0x13c92cdd00000ULL,
    0x279b2e2f80000ULL,
    0xacc74936c0000ULL,
    0x2fa36689e0000ULL,
    0x29152cdfd0000ULL,
    0x7cbd2e2bb8000ULL,
    0x5114493f3c000ULL,
    0x8bfe695f2000ULL,
    0x6f126ce067000ULL,
    0x93bdce04f1800ULL,
    0x6897590788c00ULL,
    0x5cf92eac55200ULL,
    0x9e7118e9f6d00ULL,
    0x640e0004e9b80ULL,
    0x2fbf003ec4040ULL,
    0xfa92801f6e920ULL,
    0x7df8402ee50d0ULL,
    0xc4f2602ddecb8ULL,
    0x3cb5032cdd24ULL,
    0xf9ca81ffad9aULL,
    0xd0c42419f24b3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x6800000000000ULL,
    0x9400000000000ULL,
    0x3e00000000000ULL,
    0x6300000000000ULL,
    0x8880000000000ULL,
    0x6640000000000ULL,
    0xe460000000000ULL,
    0xf050000000000ULL,
    0xab98000000000ULL,
    0xe3dc000000000ULL,
    0x6cba000000000ULL,
    0x706f000000000ULL,
    0xbbb5800000000ULL,
    0x230ac00000000ULL,
    0x9d64600000000ULL,
    0x7dd2900000000ULL,
    0x2a58080000000ULL,
    0x7b7c8c0000000ULL,
    0xbfccde0000000ULL,
    0x8e40ed0000000ULL,
    0xb067a78000000ULL,
    0x6e51174000000ULL,
    0x989bc7a000000ULL,
    0x35f877000000ULL,
    0x9ef9cf8800000ULL,
    0xaa0c0b6400000ULL,
    0x28e0919200000ULL,
    0x16264100000ULL,
    0x18fb561d80000ULL,
    0xf509a119c0000ULL,
    0xaa62999fa0000ULL,
    0x6b55aa7050000ULL,
    0x7f1a082eb8000ULL,
    0x7f9f8c2844000ULL,
    0xbddb5e165e000ULL,
    0xffb92d11b3000ULL,
    0xc0ec47b0f0800ULL,
    0x21f6476468c00ULL,
    0x6d6a2f89bde00ULL,
    0x30375b78fa700ULL,
    0x3dcb79925cf80ULL,
    0x2741fa4ecad40ULL,
    0x95e4603314fa0ULL,
    0xdb92902866f70ULL,
    0x6e38083bb74d8ULL,
    0xdb2c8c3bce0c4ULL,
    0x7c54de02ab446ULL,
    0xf99ced01e4399ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xb800000000000ULL,
    0x4c00000000000ULL,
    0xd600000000000ULL,
    0x5500000000000ULL,
    0x8d80000000000ULL,
    0x82c0000000000ULL,
    0x3fa0000000000ULL,
    0x1c30000000000ULL,
    0xd28000000000ULL,
    0x21ec000000000ULL,
    0x274e000000000ULL,
    0x70fd000000000ULL,
    0x6216800000000ULL,
    0x5df400000000ULL,
    0x2663200000000ULL,
    0xe710900000000ULL,
    0x6058b80000000ULL,
    0x40a5940000000ULL,
    0x8ab3a20000000ULL,
    0xb6a4f0000000ULL,
    0xf08568000000ULL,
    0x271c0bc000000ULL,
    0x49c376a000000ULL,
    0x53209bd000000ULL,
    0xe975ceb800000ULL,
    0x43480fcc00000ULL,
    0x62f8ec9600000ULL,
    0x751100df00000ULL,
    0x375d9a2380000ULL,
    0x12209b0a40000ULL,
    0xcef0d4a0e0000ULL,
    0x9a8ad4faf0000ULL,
    0xb9dd981708000ULL,
    0x8640431fc000ULL,
    0x9e139a0242000ULL,
    0x4bdd9b2687000ULL,
    0x2f66548811800ULL,
    0x749594e0f8400ULL,
    0x439eb80874200ULL,
    0x688494177f900ULL,
    0x97432220e4e80ULL,
    0xe1640f1047b40ULL,
    0x3d93f6a130c60ULL,
    0xb01edbe0a9d10ULL,
    0x3b466ebd6dae8ULL,
    0x766dfe77b02cULL,
    0xb093f4a2539b2ULL,
    0x19a44d27f903ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0x2400000000000ULL,
    0x6600000000000ULL,
    0x5100000000000ULL,
    0x2b80000000000ULL,
    0x9ac0000000000ULL,
    0x7a0000000000ULL,
    0xaf30000000000ULL,
    0xa4b8000000000ULL,
    0xc07c000000000ULL,
    0x1ada000000000ULL,
    0x7c6f000000000ULL,
    0x514800000000ULL,
    0xd04bc00000000ULL,
    0xf0e0e00000000ULL,
    0xd354d00000000ULL,
    0xefaed80000000ULL,
    0x536440000000ULL,
    0x77b9720000000ULL,
    0x7cff1f0000000ULL,
    0x8a9a7e8000000ULL,
    0xa28a7f4000000ULL,
    0x13429ea000000ULL,
    0xba62af7000000ULL,
    0x6a16468800000ULL,
    0x86cbeb6c00000ULL,
    0xd3a3b49200000ULL,
    0x6133346500000ULL,
    0x71bb2a0b80000ULL,
    0xfdfe9b3140000ULL,
    0xb919eca160000ULL,
    0x744eb067f0000ULL,
    0x56e2b82888000ULL,
    0x225554203c000ULL,
    0xb42d4a13a6000ULL,
    0xf7f28b3523000ULL,
    0x541954be90800ULL,
    0xb5c8e46b78c00ULL,
    0x7f217216b5a00ULL,
    0x49731f149a300ULL,
    0x93587eb005980ULL,
    0xc1a97f5573040ULL,
    0xc0341ebbc82e0ULL,
    0xf23a6f4c5d950ULL,
    0xe33826a3d6f28ULL,
    0xa8bbfb56ea92cULL,
    0xe2790ca6cdcd2ULL,
    0xfdda605eae3d1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x4800000000000ULL,
    0xb400000000000ULL,
    0x4600000000000ULL,
    0xdb00000000000ULL,
    0xc880000000000ULL,
    0x43c0000000000ULL,
    0x7720000000000ULL,
    0x2890000000000ULL,
    0xd5f8000000000ULL,
    0xc43c000000000ULL,
    0x9e1e000000000ULL,
    0xcd0f000000000ULL,
    0xd7b0800000000ULL,
    0x296ec00000000ULL,
    0x87c5600000000ULL,
    0x3920500000000ULL,
    0x6792080000000ULL,
    0xab788c0000000ULL,
    0x14fef20000000ULL,
    0x95ba9b0000000ULL,
    0xe05b978000000ULL,
    0xae68144000000ULL,
    0xd46f7a000000ULL,
    0x8fe4445000000ULL,
    0x4732ffa800000ULL,
    0x6afc86c00000ULL,
    0x9ae28dba00000ULL,
    0x80b4934100000ULL,
    0x1becfa2b80000ULL,
    0xcd021737c0000ULL,
    0xeb8565b9a0000ULL,
    0x57428f7730000ULL,
    0x36e5600c68000ULL,
    0x8ab0501114000ULL,
    0x5aea080316000ULL,
    0xdc848c211b000ULL,
    0xb5c0f22f11800ULL,
    0xc4259b3598c00ULL,
    0xa41317ab5da00ULL,
    0x983ad474f3700ULL,
    0xdc1d9794c4f80ULL,
    0x380b1446af640ULL,
    0x8030779fb67a0ULL,
    0xac29847baa470ULL,
    0xdc211f87b8468ULL,
    0xe812587bac21cULL,
    0x23b65893b036ULL,
    0x51d8f6ee154dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0xb400000000000ULL,
    0x2200000000000ULL,
    0xdb00000000000ULL,
    0x5480000000000ULL,
    0xd140000000000ULL,
    0xb960000000000ULL,
    0x6450000000000ULL,
    0x5aa8000000000ULL,
    0xd0ec000000000ULL,
    0x7d0e000000000ULL,
    0x2819000000000ULL,
    0x7bd4800000000ULL,
    0x876ac00000000ULL,
    0x14c9600000000ULL,
    0xce3d700000000ULL,
    0x13e1280000000ULL,
    0xe8104c0000000ULL,
    0x564b260000000ULL,
    0xa6795d0000000ULL,
    0xc502378000000ULL,
    0xb985dec000000ULL,
    0x2cc557a000000ULL,
    0x1fa1aed000000ULL,
    0xa4f0ff9800000ULL,
    0xc0db22fc00000ULL,
    0xae72b98a00000ULL,
    0x1c9f0fd100000ULL,
    0x5a91a60280000ULL,
    0x960a9d27c0000ULL,
    0x749fd7b7a0000ULL,
    0xfe926ec390000ULL,
    0xac0d1f81f8000ULL,
    0xe39c92e684000ULL,
    0x7412f1bd52000ULL,
    0x304e33d3bf000ULL,
    0xf77da83891800ULL,
    0x4e868c0279400ULL,
    0x5644461bc4a00ULL,
    0x2be12d3133f00ULL,
    0xb4119f9d8ef80ULL,
    0x904a52c3f8ac0ULL,
    0xa77d919384b20ULL,
    0xb68643e9d3bf0ULL,
    0xaa46003f7ffb8ULL,
    0xbde50024b0a84ULL,
    0x4d128028a9d66ULL,
    0x1fcfc0285d2ffULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0xac00000000000ULL,
    0x4a00000000000ULL,
    0x2100000000000ULL,
    0x9180000000000ULL,
    0x59c0000000000ULL,
    0x8e0000000000ULL,
    0xb530000000000ULL,
    0xdbe8000000000ULL,
    0xcb2c000000000ULL,
    0x5e4e000000000ULL,
    0x68b9000000000ULL,
    0x42b2800000000ULL,
    0xa92bc00000000ULL,
    0x9348a00000000ULL,
    0x533ff00000000ULL,
    0x4a72780000000ULL,
    0x984c340000000ULL,
    0xd3b8da0000000ULL,
    0xca32ad0000000ULL,
    0x156cdc8000000ULL,
    0x8c6f47c000000ULL,
    0x29ea7ca000000ULL,
    0xde29b7d000000ULL,
    0x61ca849800000ULL,
    0x867e43cc00000ULL,
    0x18d2feb200000ULL,
    0x595f1ee700000ULL,
    0xc2025a2080000ULL,
    0x5d056d2f40000ULL,
    0xe3827ca4e0000ULL,
    0xcc5b7e930000ULL,
    0xd76484bed8000ULL,
    0x2bf743c87c000ULL,
    0x29887e9a22000ULL,
    0x9758decdb7000ULL,
    0x4504fa3531800ULL,
    0x47839d34d9400ULL,
    0x7ac284b37d200ULL,
    0x646243d0a3b00ULL,
    0x9f74fea9f7080ULL,
    0x7ca1ef3504c0ULL,
    0xad7eda1da9960ULL,
    0x4857ad3dc4410ULL,
    0xe1185c90ae208ULL,
    0x3b2187d4edaa4ULL,
    0x4cd65cbc2a37aULL,
    0xc75887d5a2cbbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0xd400000000000ULL,
    0x4a00000000000ULL,
    0xf00000000000ULL,
    0x4a80000000000ULL,
    0xed40000000000ULL,
    0x44e0000000000ULL,
    0xd750000000000ULL,
    0xaa78000000000ULL,
    0xe03c000000000ULL,
    0x25a000000000ULL,
    0xc649000000000ULL,
    0xbdd7800000000ULL,
    0x373c400000000ULL,
    0xecd8e00000000ULL,
    0x290fb00000000ULL,
    0xa32480000000ULL,
    0xcced5c0000000ULL,
    0x3ae1560000000ULL,
    0x8256810000000ULL,
    0x77f9938000000ULL,
    0x43f8ac4000000ULL,
    0x19fb73a000000ULL,
    0x8efe1c7000000ULL,
    0xc07ebbb800000ULL,
    0x9f3f004400000ULL,
    0xdf0d8600000ULL,
    0xd70a317100000ULL,
    0x9f36d61480000ULL,
    0x716ac132c0000ULL,
    0x492173a620000ULL,
    0x61f71c7990000ULL,
    0xeb493b8f18000ULL,
    0x7453404844000ULL,
    0xbeffed8386000ULL,
    0x287981714b000ULL,
    0xd33e9e2f90800ULL,
    0x4ede9d3619400ULL,
    0x6a0fa5a0c5600ULL,
    0xeb1dd7346500ULL,
    0x9caa48206b580ULL,
    0xef815c3381bc0ULL,
    0xb8c3561f41ca0ULL,
    0x7b23813c61eb0ULL,
    0x7af4138f711f8ULL,
    0x8bcdec57a90b4ULL,
    0x461413a34c1b6ULL,
    0x489dec469a705ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0xb400000000000ULL,
    0xa200000000000ULL,
    0xdf00000000000ULL,
    0x6080000000000ULL,
    0x6dc0000000000ULL,
    0xffe0000000000ULL,
    0xf270000000000ULL,
    0x18a8000000000ULL,
    0x26c000000000ULL,
    0x10a000000000ULL,
    0x60bb000000000ULL,
    0x86f1800000000ULL,
    0x4769400000000ULL,
    0x9a8ee00000000ULL,
    0x77fcf00000000ULL,
    0x2fd1880000000ULL,
    0x1078cc0000000ULL,
    0x7690c20000000ULL,
    0x505c8d0000000ULL,
    0x6c01768000000ULL,
    0xf6010fc000000ULL,
    0xd0596a000000ULL,
    0x7786fff000000ULL,
    0x21459ea800000ULL,
    0x842773e400000ULL,
    0x7093bc8e00000ULL,
    0x555b0ec900000ULL,
    0x1781422680000ULL,
    0x9145cd3ec0000ULL,
    0x6c27969ea0000ULL,
    0xc91ffd5b0000ULL,
    0xdb5e1e8958000ULL,
    0xde8533cd2c000ULL,
    0x8cc4dca172000ULL,
    0xbe62bef4b7000ULL,
    0xfe342a2fb0800ULL,
    0xbb0af13659c00ULL,
    0xcbbf5c9eace00ULL,
    0x6470fefb33900ULL,
    0xa5ab4a37d7580ULL,
    0x9def412520ac0ULL,
    0x5c48348bf13a0ULL,
    0x6a98c2d4399f0ULL,
    0x3f600036bd8a8ULL,
    0xfb000265b7c4ULL,
    0x5f480032e3372ULL,
    0x141c00379fafbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0xfc00000000000ULL,
    0x4600000000000ULL,
    0x3d00000000000ULL,
    0x7f80000000000ULL,
    0x8cc0000000000ULL,
    0x3960000000000ULL,
    0xbaf0000000000ULL,
    0x68f8000000000ULL,
    0x953c000000000ULL,
    0x8dde000000000ULL,
    0x9ae9000000000ULL,
    0xc2f0800000000ULL,
    0x4fe400000000ULL,
    0x5b3fa00000000ULL,
    0x24d8900000000ULL,
    0x76ff80000000ULL,
    0xc937440000000ULL,
    0x391a020000000ULL,
    0x2f88190000000ULL,
    0x8603168000000ULL,
    0x9d00c54000000ULL,
    0xaf82b6a000000ULL,
    0xe4c1555000000ULL,
    0x9d65ce9800000ULL,
    0xf4515c00000ULL,
    0x13fe6c8e00000ULL,
    0xd7bdd86300000ULL,
    0x7e9a021f80000ULL,
    0x2f48192c40000ULL,
    0x416316a060000ULL,
    0xd6f0c55f70000ULL,
    0xa6fab6b638000ULL,
    0x3c3d556dcc000ULL,
    0x105bce8206000ULL,
    0x912d517b83000ULL,
    0xff16eca6f1800ULL,
    0x704f986d79400ULL,
    0xf9e3a2182da00ULL,
    0xbeb5890036d00ULL,
    0x61da6ebf5a480ULL,
    0x14ecc1548cf40ULL,
    0x8bf194b2e7b60ULL,
    0xa97a9c7132250ULL,
    0x8f8002be8b68ULL,
    0xe53c000a25634ULL,
    0x35de003e23d32ULL,
    0x56e9002aa0821ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x6800000000000ULL,
    0xc400000000000ULL,
    0x5e00000000000ULL,
    0xb900000000000ULL,
    0x7780000000000ULL,
    0x3740000000000ULL,
    0xbfa0000000000ULL,
    0xe330000000000ULL,
    0x3fe8000000000ULL,
    0x7dac000000000ULL,
    0x5f8e000000000ULL,
    0x34fb000000000ULL,
    0x4637800000000ULL,
    0x6a68c00000000ULL,
    0xa1e8e00000000ULL,
    0x24ad100000000ULL,
    0x1809580000000ULL,
    0xfbbd240000000ULL,
    0x5591fe0000000ULL,
    0x135cd30000000ULL,
    0x7902f28000000ULL,
    0x978679c000000ULL,
    0x74412a000000ULL,
    0x47a069d000000ULL,
    0x4f32ca8800000ULL,
    0xa5e98df400000ULL,
    0x9aadd49200000ULL,
    0x910f4ec500000ULL,
    0x743ffe0580000ULL,
    0xced7d32b40000ULL,
    0x36fd72b860000ULL,
    0x7d32b9f690000ULL,
    0x66eaf29f88000ULL,
    0x3a2a79fa5c000ULL,
    0x90ca12a272000ULL,
    0x275b69f037000ULL,
    0x3f054ab811800ULL,
    0xb2814de249c00ULL,
    0x12c534a3fc600ULL,
    0x3be25edb43f00ULL,
    0xa416a6398e580ULL,
    0xe11af72774b40ULL,
    0xe3248cbf9e0e0ULL,
    0xf0f26aee3cfd0ULL,
    0x7f8e003ae3f18ULL,
    0xe4fb000cbe09cULL,
    0x8e378030cd2d2ULL,
    0x3e68c03bfb67bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0x9400000000000ULL,
    0xfe00000000000ULL,
    0xdf00000000000ULL,
    0x1980000000000ULL,
    0x5940000000000ULL,
    0x3c60000000000ULL,
    0xefd0000000000ULL,
    0xd6c8000000000ULL,
    0x7c8c000000000ULL,
    0xdb6a000000000ULL,
    0x8fd000000000ULL,
    0x22d4800000000ULL,
    0x724ec00000000ULL,
    0x6948600000000ULL,
    0xb6ce900000000ULL,
    0x8c8ee80000000ULL,
    0xf369e40000000ULL,
    0xacf9ce0000000ULL,
    0x14d37d0000000ULL,
    0x39490b8000000ULL,
    0x8ecb89c000000ULL,
    0x308b6ba000000ULL,
    0xa96819f000000ULL,
    0x45f9039800000ULL,
    0x46533dfc00000ULL,
    0x878aadaa00000ULL,
    0x34efd0d900000ULL,
    0xfabbce1580000ULL,
    0xc9327d2ec0000ULL,
    0xd3df8bb3e0000ULL,
    0xe6449c890000ULL,
    0x82d58ba4f8000ULL,
    0x624949cf64000ULL,
    0x71490b9b4a000ULL,
    0xdacb89f947000ULL,
    0x2e8b6ba010800ULL,
    0x466819e9b8400ULL,
    0x9479039145a00ULL,
    0x8b133dd4bbf00ULL,
    0x45eaad826fb80ULL,
    0x43fd0d4ac540ULL,
    0x35f3ce1be6020ULL,
    0xecfe7d0ca12d0ULL,
    0x34d58b93b0f18ULL,
    0xe94949e489b24ULL,
    0x76c90bbb8d136ULL,
    0x6c8b89e597a03ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0xf800000000000ULL,
    0x9400000000000ULL,
    0xe200000000000ULL,
    0x1f00000000000ULL,
    0xc180000000000ULL,
    0xd040000000000ULL,
    0x9b60000000000ULL,
    0x6070000000000ULL,
    0xacf8000000000ULL,
    0x41bc000000000ULL,
    0xa45e000000000ULL,
    0xa5e9000000000ULL,
    0x2a35800000000ULL,
    0xfc9e400000000ULL,
    0xe24ce00000000ULL,
    0x57e6f00000000ULL,
    0x46b1880000000ULL,
    0xaa5ec40000000ULL,
    0x8ce8260000000ULL,
    0xf6b1cd0000000ULL,
    0x6259e68000000ULL,
    0x90ec554000000ULL,
    0xe8b306a000000ULL,
    0xf35fa55000000ULL,
    0x38690e9800000ULL,
    0x476217400000ULL,
    0x66f848a600000ULL,
    0xd2bf5c5100000ULL,
    0x83dcc61e80000ULL,
    0x70ab3d16c0000ULL,
    0xfbd66e8ee0000ULL,
    0x6f2b9172d0000ULL,
    0x1a96a0b038000ULL,
    0x2ccc285ecc000ULL,
    0xfe22081d06000ULL,
    0x7995840fa3000ULL,
    0x234f461291800ULL,
    0xa7607d1419400ULL,
    0x4e710ea8fce00ULL,
    0xb5fa217ceed00ULL,
    0x153e4898d7d80ULL,
    0xc69a5c701b6c0ULL,
    0x594f46041dba0ULL,
    0xfc607d313fdd0ULL,
    0x75f10e8b6e338ULL,
    0x7eba217496ee4ULL,
    0x55de48b63aaa2ULL,
    0xfdaa5c6d2c2b1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0xb800000000000ULL,
    0xec00000000000ULL,
    0xee00000000000ULL,
    0xdd00000000000ULL,
    0x2480000000000ULL,
    0x2e40000000000ULL,
    0xfae0000000000ULL,
    0xb450000000000ULL,
    0x5018000000000ULL,
    0x955c000000000ULL,
    0xa3a000000000ULL,
    0x52ab000000000ULL,
    0xcf17800000000ULL,
    0x8a7e400000000ULL,
    0x954b200000000ULL,
    0xefc3500000000ULL,
    0xdc24380000000ULL,
    0xaef25c0000000ULL,
    0x53a88e0000000ULL,
    0xed97c90000000ULL,
    0xd3c338000000ULL,
    0xd92d1f4000000ULL,
    0xfed513a000000ULL,
    0x63594f7000000ULL,
    0x6b3caba800000ULL,
    0x2e535400000ULL,
    0x4050858600000ULL,
    0x2188a6f00000ULL,
    0x3e59ae3d80000ULL,
    0x8fbf991840000ULL,
    0xe6f8b94a0000ULL,
    0x2ab10363d0000ULL,
    0xe4ebd8218000ULL,
    0x8241d65ac4000ULL,
    0xf4e6a0301e000ULL,
    0xd956102093000ULL,
    0x5c98980710800ULL,
    0xef1f4c0879c00ULL,
    0xf2df960e75600ULL,
    0xd5fac53236500ULL,
    0x668a85b2bfa80ULL,
    0x15e38a5442840ULL,
    0x4bd62e3488a60ULL,
    0xf3ddd930fd130ULL,
    0xf77eab97ca058ULL,
    0xe1c95365f5214ULL,
    0x5985059d76696ULL,
    0x5ac1ca679f369ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x3800000000000ULL,
    0x8400000000000ULL,
    0x6a00000000000ULL,
    0xb300000000000ULL,
    0x2180000000000ULL,
    0x2e40000000000ULL,
    0xc0a0000000000ULL,
    0x70f0000000000ULL,
    0x9198000000000ULL,
    0x4edc000000000ULL,
    0xc0fe000000000ULL,
    0xa1cf000000000ULL,
    0xefb1800000000ULL,
    0x3fbc400000000ULL,
    0xffee600000000ULL,
    0xb686300000000ULL,
    0xf5c0280000000ULL,
    0x45e4640000000ULL,
    0xc5d1f60000000ULL,
    0x9d2e030000000ULL,
    0x28e68c8000000ULL,
    0xb554b44000000ULL,
    0x6deeeca000000ULL,
    0xe181845000000ULL,
    0x4e4144a800000ULL,
    0x30a6a06400000ULL,
    0x98f752a600000ULL,
    0x2d9ed34500000ULL,
    0xa0d9963c80000ULL,
    0x19fb3329c0000ULL,
    0x33492494e0000ULL,
    0xe073907250000ULL,
    0xd158fa8cc8000ULL,
    0x4fb9f74a6c000ULL,
    0x57ef802eae000ULL,
    0x2a83400cdb000ULL,
    0xcbc7e02a11800ULL,
    0xa4e6703968c00ULL,
    0xd350480b1de00ULL,
    0xf8ed5425d7d00ULL,
    0x75005e02ae780ULL,
    0x2486273eda5c0ULL,
    0xa2c11ab110720ULL,
    0xfe60874ce8450ULL,
    0xb096483c5d8f8ULL,
    0xc00e540bf705cULL,
    0x9857de259fccaULL,
    0x3d69673902f8dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xf800000000000ULL,
    0x9400000000000ULL,
    0x2e00000000000ULL,
    0xd700000000000ULL,
    0x8680000000000ULL,
    0x5940000000000ULL,
    0x19e0000000000ULL,
    0x2850000000000ULL,
    0x83d8000000000ULL,
    0xd79c000000000ULL,
    0xe1fa000000000ULL,
    0xa96b000000000ULL,
    0x3096800000000ULL,
    0xcdfd400000000ULL,
    0xf36be00000000ULL,
    0x5995900000000ULL,
    0xd47ba80000000ULL,
    0x40afdc0000000ULL,
    0xa331760000000ULL,
    0x1ccce90000000ULL,
    0xbaa2238000000ULL,
    0x28b4d14000000ULL,
    0x6a0bc3a000000ULL,
    0xc386417000000ULL,
    0x1ac4eba800000ULL,
    0xc323dd5c00000ULL,
    0xcb72fd9600000ULL,
    0xf8ace47900000ULL,
    0xd736161d80000ULL,
    0xa2cf390cc0000ULL,
    0x25a4eb9c20000ULL,
    0xc233dd6a90000ULL,
    0x894afdb258000ULL,
    0x2360e47c84000ULL,
    0x63141624ae000ULL,
    0x9f38390063000ULL,
    0x5c486b8950800ULL,
    0x28e59d4af9c00ULL,
    0xd5d79d8755200ULL,
    0xc618344616900ULL,
    0xde3c5e2a57d80ULL,
    0x79ce750537f40ULL,
    0x4920b5ac47320ULL,
    0xa71a84a2f030ULL,
    0x1d28481e23ef8ULL,
    0x96f64c2d31f94ULL,
    0xcd68de3508476ULL,
    0x694351b7d5d1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0x6c00000000000ULL,
    0xd600000000000ULL,
    0x3700000000000ULL,
    0x2d80000000000ULL,
    0x11c0000000000ULL,
    0xd3e0000000000ULL,
    0x7170000000000ULL,
    0x2c8000000000ULL,
    0x1e8c000000000ULL,
    0xacaa000000000ULL,
    0x2639000000000ULL,
    0x7187800000000ULL,
    0x4fc7400000000ULL,
    0x70e2200000000ULL,
    0x36f7900000000ULL,
    0x6a08a80000000ULL,
    0xbbe9040000000ULL,
    0x379cea0000000ULL,
    0xd1d7c50000000ULL,
    0xe01f768000000ULL,
    0x8110ccc000000ULL,
    0x997ad6a000000ULL,
    0x84201cf000000ULL,
    0xad905eb800000ULL,
    0x823e88c400000ULL,
    0x93841c9e00000ULL,
    0xa2c049df00000ULL,
    0x6c67802b80000ULL,
    0xa5b74012c0000ULL,
    0xe9aa2000e0000ULL,
    0xdebb9026b0000ULL,
    0xa0c2a82c18000ULL,
    0x9160040eec000ULL,
    0x41336a0aaa000ULL,
    0xc6ec85124f000ULL,
    0x131f569400800ULL,
    0x9e925cd601400ULL,
    0xfdbffea301600ULL,
    0x274758f180900ULL,
    0x99a11493c1e80ULL,
    0x34559dd461740ULL,
    0xc5dbe22ff1c20ULL,
    0x88f7112638350ULL,
    0xd90e949cbc788ULL,
    0x46edde603844ULL,
    0xe35bc20bba452ULL,
    0x3ab58115c6677ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x1800000000000ULL,
    0x4c00000000000ULL,
    0x4600000000000ULL,
    0xf700000000000ULL,
    0x1880000000000ULL,
    0x8440000000000ULL,
    0xbe60000000000ULL,
    0xcf50000000000ULL,
    0xc038000000000ULL,
    0xa4fc000000000ULL,
    0x62de000000000ULL,
    0x2ded000000000ULL,
    0x6982800000000ULL,
    0x4bc6400000000ULL,
    0xb2a6200000000ULL,
    0x1d77100000000ULL,
    0xe50f280000000ULL,
    0xa1932c0000000ULL,
    0xb71b6e0000000ULL,
    0x449ad0000000ULL,
    0xba76ef8000000ULL,
    0x35886a4000000ULL,
    0xe1d24fa000000ULL,
    0xeb793a5000000ULL,
    0x2e1b47b800000ULL,
    0x9fcd066c00000ULL,
    0xbcb701be00000ULL,
    0x56ab877f00000ULL,
    0x2b64802f80000ULL,
    0x72d74002c0000ULL,
    0xc1faa009a0000ULL,
    0xc75c5033d0000ULL,
    0x602b883e38000ULL,
    0x4c227c3c04000ULL,
    0xbe32660836000ULL,
    0xcbed91063b000ULL,
    0x4e82a98101800ULL,
    0xdb42eb7080400ULL,
    0x92e7ce3d41200ULL,
    0xb114fd1361d00ULL,
    0xd759e78670a80ULL,
    0xc82d564be8140ULL,
    0x782289b13d620ULL,
    0x4434fb75b3df0ULL,
    0x62ec660bcc398ULL,
    0xed00912d1af44ULL,
    0x180298e91226ULL,
    0x5fc4ab6d597edULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xb800000000000ULL,
    0x6400000000000ULL,
    0xd600000000000ULL,
    0x8b00000000000ULL,
    0xf180000000000ULL,
    0x7dc0000000000ULL,
    0x9fa0000000000ULL,
    0xf30000000000ULL,
    0xf638000000000ULL,
    0x4e7c000000000ULL,
    0x121e000000000ULL,
    0x4589000000000ULL,
    0x3702800000000ULL,
    0x7383400000000ULL,
    0x78c2600000000ULL,
    0xf123d00000000ULL,
    0xed74080000000ULL,
    0xfb5e440000000ULL,
    0xf96a8a0000000ULL,
    0x1dd2c70000000ULL,
    0x93ede68000000ULL,
    0x19905b4000000ULL,
    0x4d8d06a000000ULL,
    0x4b00cb5000000ULL,
    0x51836e9800000ULL,
    0xedc15f4c00000ULL,
    0x57a3ecb600000ULL,
    0xd334dc5500000ULL,
    0x443e000380000ULL,
    0x13790037c0000ULL,
    0x689a801fe0000ULL,
    0xc9cf4038b0000ULL,
    0xd564600008000ULL,
    0xe316d0277c000ULL,
    0x81c8881386000ULL,
    0x4964042de3000ULL,
    0xb1126a2281800ULL,
    0xacce571e41400ULL,
    0x8be58e9321200ULL,
    0x5954cf5dd1900ULL,
    0x98a9048af9b80ULL,
    0x9073087255640ULL,
    0x51d862342aba0ULL,
    0x5d2913031df50ULL,
    0x53b58480f7988ULL,
    0x979485c0b04cULL,
    0xd99882044c412ULL,
    0x594983284fc21ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x1800000000000ULL,
    0x3c00000000000ULL,
    0x1600000000000ULL,
    0xc500000000000ULL,
    0x4380000000000ULL,
    0xd1c0000000000ULL,
    0x7b20000000000ULL,
    0x1130000000000ULL,
    0xacb8000000000ULL,
    0xb0fc000000000ULL,
    0x231a000000000ULL,
    0x7c89000000000ULL,
    0x5c7800000000ULL,
    0x3923400000000ULL,
    0x5634a00000000ULL,
    0xc83f100000000ULL,
    0xd5bbb80000000ULL,
    0x257f640000000ULL,
    0x57db0a0000000ULL,
    0x942ee30000000ULL,
    0xdd30828000000ULL,
    0xd2ba1b4000000ULL,
    0xe1fba2a000000ULL,
    0x829a4b7000000ULL,
    0x5a4ebab800000ULL,
    0xd2633f4400000ULL,
    0x695188be00000ULL,
    0xd6edf84b00000ULL,
    0x8094a03b80000ULL,
    0xe1cf103a40000ULL,
    0x2fa3b80b60000ULL,
    0x7c73643e30000ULL,
    0xeed90a0da8000ULL,
    0x61abe30e7c000ULL,
    0x79f502a086000ULL,
    0x221c5b6d43000ULL,
    0x50a828dc0800ULL,
    0x7f031b7321400ULL,
    0x2884228451600ULL,
    0xdb450b7d19900ULL,
    0x37601a8914480ULL,
    0x5ad52f445a540ULL,
    0xaf2db092d4a60ULL,
    0xfb1dc507a8d0ULL,
    0xc2fb0a1f04e48ULL,
    0x6c1ee31e2375cULL,
    0x5c088297f1242ULL,
    0x9a861b5089231ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xc800000000000ULL,
    0xbc00000000000ULL,
    0xbe00000000000ULL,
    0x1500000000000ULL,
    0x3880000000000ULL,
    0x4cc0000000000ULL,
    0x7d20000000000ULL,
    0xbf90000000000ULL,
    0x938000000000ULL,
    0x1b7c000000000ULL,
    0xbb9a000000000ULL,
    0x2baf000000000ULL,
    0xf6c6800000000ULL,
    0xc224c00000000ULL,
    0x17200000000ULL,
    0x2e7fd00000000ULL,
    0x731af80000000ULL,
    0xf6b7c0000000ULL,
    0x7e1fa0000000ULL,
    0xbb2630000000ULL,
    0xa02f778000000ULL,
    0xb3821a4000000ULL,
    0x9146d7a000000ULL,
    0x61650a7000000ULL,
    0x2710f8800000ULL,
    0xbd8ea67c00000ULL,
    0x4d48dba00000ULL,
    0x2f5f795300000ULL,
    0xd28f201580000ULL,
    0x2353d01640000ULL,
    0xec18f83ba0000ULL,
    0x40e87c3570000ULL,
    0x48a57a0ce8000ULL,
    0xefd5a31694000ULL,
    0x62dcd7bca6000ULL,
    0x36ca0a7f4b000ULL,
    0xeab78f9dc0800ULL,
    0x5aaa6640e0c00ULL,
    0xf443ad8651200ULL,
    0xf1e0a96f59f00ULL,
    0x62b5d82b1ce80ULL,
    0x86a8ac2422040ULL,
    0xda41023a14560ULL,
    0x1ce61f1467db0ULL,
    0x2e300db32aa88ULL,
    0xc868b9435083cULL,
    0xc64802bd8d56ULL,
    0x8ef7c02fdd355ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0x4c00000000000ULL,
    0xc600000000000ULL,
    0xd500000000000ULL,
    0xb580000000000ULL,
    0x7040000000000ULL,
    0x5be0000000000ULL,
    0xb230000000000ULL,
    0x4ec8000000000ULL,
    0x790c000000000ULL,
    0xddae000000000ULL,
    0x6afd000000000ULL,
    0x5ac1800000000ULL,
    0xad21400000000ULL,
    0x911200000000ULL,
    0x6adc100000000ULL,
    0x7a52380000000ULL,
    0x89bd540000000ULL,
    0x21a1b60000000ULL,
    0xbad4d90000000ULL,
    0xfc7ea08000000ULL,
    0x62045ac000000ULL,
    0x170600a000000ULL,
    0x12850ad000000ULL,
    0xac318b800000ULL,
    0x15254ef400000ULL,
    0xb517169e00000ULL,
    0xa4dc83e700000ULL,
    0x7b56a02f80000ULL,
    0xb63c501240000ULL,
    0x42e49822e0000ULL,
    0x81b1041e50000ULL,
    0x8b8d2e2c18000ULL,
    0x769dd16dc000ULL,
    0x87dd8e99aa000ULL,
    0x33d087db9f000ULL,
    0x57fa0e3f41800ULL,
    0xeb44cd2460400ULL,
    0xaf6036bc91e00ULL,
    0x9df193e738100ULL,
    0xe66b18362da80ULL,
    0x685d443603140ULL,
    0xa8958e148b260ULL,
    0xb0988d186f410ULL,
    0xc2b0968968cb8ULL,
    0x530cc3f235374ULL,
    0x7ea8002cdf03aULL,
    0x627c00072071fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0xc400000000000ULL,
    0x7e00000000000ULL,
    0x3d00000000000ULL,
    0x4e80000000000ULL,
    0xd640000000000ULL,
    0x3e20000000000ULL,
    0xaf10000000000ULL,
    0xda58000000000ULL,
    0xf79c000000000ULL,
    0x257e000000000ULL,
    0xc40d000000000ULL,
    0xce61800000000ULL,
    0x9034c00000000ULL,
    0xc9c8e00000000ULL,
    0xec84100000000ULL,
    0xc947780000000ULL,
    0x2fa54c0000000ULL,
    0x8d25e0000000ULL,
    0xd3bd9f0000000ULL,
    0x796a6f8000000ULL,
    0xb5d1194000000ULL,
    0xdd3d0fa000000ULL,
    0xcf2cc95000000ULL,
    0x1bf317b800000ULL,
    0x5a29557400000ULL,
    0xd176d1b200000ULL,
    0x9269964500000ULL,
    0x4257982d80000ULL,
    0x5bfd5c0a40000ULL,
    0x8a4b261e20000ULL,
    0xec45d32fd0000ULL,
    0x3d21b189a8000ULL,
    0xbc94466014000ULL,
    0xd29980153e000ULL,
    0x37f8c03cf3000ULL,
    0x704ee03be1800ULL,
    0xcf451036b0400ULL,
    0xdea0f809d9e00ULL,
    0xc508c322d700ULL,
    0xfefd3e21a3d80ULL,
    0xd8cc4f1c88740ULL,
    0x380277a944060ULL,
    0xac05855056270ULL,
    0xda07499e07118ULL,
    0xd305ca5bee00cULL,
    0x5b833e12dbea6ULL,
    0x5cc14f20b4769ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0xa400000000000ULL,
    0xda00000000000ULL,
    0x7700000000000ULL,
    0xf980000000000ULL,
    0xeac0000000000ULL,
    0x6c60000000000ULL,
    0xe5b0000000000ULL,
    0xfab8000000000ULL,
    0xc6fc000000000ULL,
    0x75de000000000ULL,
    0xd8cf000000000ULL,
    0x9ea3800000000ULL,
    0x65d7c00000000ULL,
    0x4108e00000000ULL,
    0x143300000000ULL,
    0x35a2080000000ULL,
    0x1a56840000000ULL,
    0x1ac9720000000ULL,
    0x5a4010000000ULL,
    0xc253f48000000ULL,
    0x96cb1d4000000ULL,
    0xeba694a000000ULL,
    0x6750ed5000000ULL,
    0xbc4ffca800000ULL,
    0x22e2997400000ULL,
    0x96f466aa00000ULL,
    0xcc1f2c5d00000ULL,
    0xd72ae81c80000ULL,
    0xc6d5b400c0000ULL,
    0x9a8b7a02e0000ULL,
    0x8082852a70000ULL,
    0x46428691f8000ULL,
    0x14231c771c000ULL,
    0x7c93600d56000ULL,
    0x98a8f03fa3000ULL,
    0x4514e80ea1800ULL,
    0x126ab42050c00ULL,
    0x9970fa1d69600ULL,
    0x6fd9450e95300ULL,
    0xcfcc66a2b3580ULL,
    0xd7232c4369fc0ULL,
    0x1714e83995da0ULL,
    0x816ab403335d0ULL,
    0x9af0fa2f28068ULL,
    0x62194503b59d4ULL,
    0x522c6697a21baULL,
    0x7c532c7aa1aa1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0x8400000000000ULL,
    0xda00000000000ULL,
    0x7b00000000000ULL,
    0x9080000000000ULL,
    0xc0000000000ULL,
    0x9ee0000000000ULL,
    0xad50000000000ULL,
    0x8548000000000ULL,
    0xba0c000000000ULL,
    0xe6aa000000000ULL,
    0x1adb000000000ULL,
    0x9ea3800000000ULL,
    0x9b74c00000000ULL,
    0x12fee00000000ULL,
    0x5815300000000ULL,
    0xcaec680000000ULL,
    0x52f8ac0000000ULL,
    0xb8130a0000000ULL,
    0xfaefed0000000ULL,
    0x2afcbf8000000ULL,
    0x5412a1c000000ULL,
    0xa4ebdfa000000ULL,
    0x8bf851f000000ULL,
    0xbf92d7a800000ULL,
    0x34ad0df400000ULL,
    0x15d955b200000ULL,
    0x8c247cd500000ULL,
    0x1cb1080480000ULL,
    0x2a995c0bc0000ULL,
    0xd0818200a0000ULL,
    0xe0c2713bf0000ULL,
    0xaee3ddaf18000ULL,
    0xd555e0d4a4000ULL,
    0x694c6a344a000ULL,
    0xe4091d2b27000ULL,
    0x47afb79360800ULL,
    0xf15cfdc751c00ULL,
    0xee3dda2e8600ULL,
    0x555e0f4bcf00ULL,
    0x214c6a256fd80ULL,
    0x70091d322dbc0ULL,
    0xf5afb79e26c20ULL,
    0xe5cfdc7e0970ULL,
    0x4463dd9490808ULL,
    0x7e95e0fe49ac4ULL,
    0x2f2c6a094d4f6ULL,
    0xdd991d2d77343ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x4800000000000ULL,
    0xe400000000000ULL,
    0xa600000000000ULL,
    0xdd00000000000ULL,
    0x2f80000000000ULL,
    0xac0000000000ULL,
    0x94a0000000000ULL,
    0x1c90000000000ULL,
    0xdee8000000000ULL,
    0xbaac000000000ULL,
    0xc14a000000000ULL,
    0xfcbb000000000ULL,
    0x4ee6800000000ULL,
    0x8670c00000000ULL,
    0x7699e00000000ULL,
    0xd531d00000000ULL,
    0x2df8b80000000ULL,
    0x4e80840000000ULL,
    0x5f41720000000ULL,
    0x6960270000000ULL,
    0x88b7cb8000000ULL,
    0xd43d90c000000ULL,
    0xeca2aba000000ULL,
    0x709780f000000ULL,
    0xbced73a800000ULL,
    0x71aa14dc00000ULL,
    0x7bcf59ba00000ULL,
    0x3dfc67e500000ULL,
    0x7685d81380000ULL,
    0xd346941240000ULL,
    0x9b64aa3860000ULL,
    0xbbb6b320f0000ULL,
    0xc2bb619548000ULL,
    0x6fe723cbd4000ULL,
    0x33f3ca10f2000ULL,
    0x1b5ba32c7f000ULL,
    0xb91039a120800ULL,
    0x8b2d77f811c00ULL,
    0x5a8c800379200ULL,
    0x959bc03e3df00ULL,
    0xdbb7602877580ULL,
    0x92bd101e15840ULL,
    0xb7e3582cd3760ULL,
    0x2ff65403eea50ULL,
    0x115d4a1e19aa8ULL,
    0x2617632d4c6ccULL,
    0xdfabd99f7ff86ULL,
    0xa2cba7e5a038fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0xc400000000000ULL,
    0x4200000000000ULL,
    0xdb00000000000ULL,
    0x7280000000000ULL,
    0x4840000000000ULL,
    0x55a0000000000ULL,
    0xd1f0000000000ULL,
    0x8c78000000000ULL,
    0xafbc000000000ULL,
    0x5bde000000000ULL,
    0x549000000000ULL,
    0x7520800000000ULL,
    0x8ab6400000000ULL,
    0xe75a600000000ULL,
    0xd80b700000000ULL,
    0x8706280000000ULL,
    0x54847c0000000ULL,
    0x1439e0000000ULL,
    0x9426e30000000ULL,
    0x7731148000000ULL,
    0x179e334000000ULL,
    0xb2edf4a000000ULL,
    0xafd6035000000ULL,
    0x1c4f3c9800000ULL,
    0x9ca64f5c00000ULL,
    0x50706a8e00000ULL,
    0xcab9e06d00000ULL,
    0x45ea81e80000ULL,
    0x1e8e3c1bc0000ULL,
    0x3547fe2220000ULL,
    0x2e249319d0000ULL,
    0xb837bc8f48000ULL,
    0x4f1c0f61ac000ULL,
    0xbdac0a9a56000ULL,
    0xef7906b43000ULL,
    0x1cfe002421800ULL,
    0xecf90010d1400ULL,
    0x14f88013c9a00ULL,
    0xfa40016cd00ULL,
    0x2afc6017f7380ULL,
    0x6dfe7039d20c0ULL,
    0x9978a81f88b20ULL,
    0x483b3c240c130ULL,
    0xb4197e0cc7d58ULL,
    0x5f2bd302ebebcULL,
    0xceb55cb3fdb0aULL,
    0xe55d3f5c5ef71ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0xac00000000000ULL,
    0x8600000000000ULL,
    0xc100000000000ULL,
    0x1480000000000ULL,
    0xc0c0000000000ULL,
    0x32e0000000000ULL,
    0x510000000000ULL,
    0x7998000000000ULL,
    0x8adc000000000ULL,
    0x7d7a000000000ULL,
    0x16cb000000000ULL,
    0x9c62800000000ULL,
    0xead2400000000ULL,
    0x8af9200000000ULL,
    0x4188100000000ULL,
    0x29c1980000000ULL,
    0xd260fc0000000ULL,
    0xcfd41a0000000ULL,
    0x3c7c7d0000000ULL,
    0xc24abd8000000ULL,
    0x7ca7254000000ULL,
    0xe8331da000000ULL,
    0x17ea757000000ULL,
    0x941325a800000ULL,
    0x251bd97c00000ULL,
    0x6e1d078e00000ULL,
    0xcd9d085700000ULL,
    0x60db182f80000ULL,
    0x1a7ebc2740000ULL,
    0xf34f3a0aa0000ULL,
    0xd0236d2770000ULL,
    0xb4f3a58e18000ULL,
    0x3b0e996a34000ULL,
    0xfc04a7967e000ULL,
    0xce00585383000ULL,
    0xe5032020a0800ULL,
    0x9683102a71c00ULL,
    0xb3c3180498200ULL,
    0xcd62bc3bf5500ULL,
    0x96553a031e680ULL,
    0x5e386d1693540ULL,
    0x696925b4280e0ULL,
    0xf3d0d95b8c6b0ULL,
    0xd27f8795bb888ULL,
    0x174f485025b44ULL,
    0x722238117787eULL,
    0xf7f6ac2b3fed1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x2800000000000ULL,
    0x7400000000000ULL,
    0x6600000000000ULL,
    0xaf00000000000ULL,
    0x9480000000000ULL,
    0x86c0000000000ULL,
    0x20000000000ULL,
    0x8c50000000000ULL,
    0xc938000000000ULL,
    0x5a7c000000000ULL,
    0x631e000000000ULL,
    0x5329000000000ULL,
    0xfba4800000000ULL,
    0x3e11c00000000ULL,
    0x1fd8600000000ULL,
    0x4e08500000000ULL,
    0x7a72a80000000ULL,
    0xd26c4c0000000ULL,
    0xbc05a0000000ULL,
    0x35a32b0000000ULL,
    0xa511118000000ULL,
    0xd5b0c4000000ULL,
    0xd7cbf1a000000ULL,
    0xc6d79c5000000ULL,
    0xacfbb99800000ULL,
    0xa4db407c00000ULL,
    0x4c8dab9200000ULL,
    0x9bb1b76100000ULL,
    0x82c8280b80000ULL,
    0x4f548c0ac0000ULL,
    0x27bcba0ea0000ULL,
    0x81babb3b70000ULL,
    0x8ebbd9b6a8000ULL,
    0x4a3f107464000ULL,
    0x54f903aee6000ULL,
    0x8d8fb6ffb000ULL,
    0x968af217a1800ULL,
    0x96b3673cf1400ULL,
    0xf74f4ba668a00ULL,
    0x8611275fc5300ULL,
    0x53de602016b80ULL,
    0x240d503893240ULL,
    0x5f70283be4120ULL,
    0xd3e88c1a27370ULL,
    0xa482ba0a5b848ULL,
    0xeec3bb3a51bdcULL,
    0x9427599918e0eULL,
    0x5a52d0776c0edULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0x8400000000000ULL,
    0xea00000000000ULL,
    0xb700000000000ULL,
    0xff80000000000ULL,
    0x7a40000000000ULL,
    0xa060000000000ULL,
    0xcd10000000000ULL,
    0xbad8000000000ULL,
    0xa61c000000000ULL,
    0xfeba000000000ULL,
    0xb10d000000000ULL,
    0xbb62800000000ULL,
    0x9c93c00000000ULL,
    0x7d99e00000000ULL,
    0xbefdb00000000ULL,
    0x1e69880000000ULL,
    0x55f4340000000ULL,
    0x9a08460000000ULL,
    0xa2e5550000000ULL,
    0x2fd7a58000000ULL,
    0x277c5f4000000ULL,
    0x6d2ec5a000000ULL,
    0xef132f7000000ULL,
    0xb9de2d9800000ULL,
    0xfb986b4c00000ULL,
    0x47fe83b200000ULL,
    0xacea7a6d00000ULL,
    0x3f33880880000ULL,
    0x93a93400c0000ULL,
    0xdbd2c61760000ULL,
    0xa57a952e50000ULL,
    0xbe2c458778000ULL,
    0xa90ef6114000ULL,
    0xdc9fcda3ce000ULL,
    0x2079db5993000ULL,
    0x1aad0b9ee0800ULL,
    0x1534e6410400ULL,
    0xabb94e25d8200ULL,
    0xad8fa13ca4b00ULL,
    0x6ca4838826580ULL,
    0x2db77a7d9fdc0ULL,
    0x29690806ea3a0ULL,
    0x6a76f413f5d10ULL,
    0x492605deab8ULL,
    0xf28625094bc44ULL,
    0x4ac54da6c5e2eULL,
    0x19a61b7ef7239ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x4800000000000ULL,
    0xc00000000000ULL,
    0xfa00000000000ULL,
    0x9300000000000ULL,
    0xb380000000000ULL,
    0x8940000000000ULL,
    0x2e20000000000ULL,
    0xbe90000000000ULL,
    0x4b28000000000ULL,
    0x74ec000000000ULL,
    0x990a000000000ULL,
    0x87d000000000ULL,
    0xb527800000000ULL,
    0x2111c00000000ULL,
    0x886fa00000000ULL,
    0xd1cb700000000ULL,
    0xd01b680000000ULL,
    0x3c13b40000000ULL,
    0x86e91e0000000ULL,
    0x260fff0000000ULL,
    0xf1fff38000000ULL,
    0xb76266c000000ULL,
    0xf8b5d3a000000ULL,
    0x49b9d6f000000ULL,
    0xddc49b9800000ULL,
    0x3ae1d2e400000ULL,
    0xd4f4cd8a00000ULL,
    0x5d1a29e700000ULL,
    0x9a91683480000ULL,
    0x2d2eb403c0000ULL,
    0xb5ee9e3860000ULL,
    0xe58e3f1190000ULL,
    0x80b8538be8000ULL,
    0x8d4516ea34000ULL,
    0xf824bbab22000ULL,
    0x679762c40f000ULL,
    0x73aa058be0800ULL,
    0xa2fedf6d0400ULL,
    0xc86c9e1849600ULL,
    0xf1cf3f1fc4300ULL,
    0x601dd3af5a880ULL,
    0x2415d6d0d2ec0ULL,
    0xc2ee9bbd76520ULL,
    0xd00cd2c4bd010ULL,
    0x98fb4d9b06958ULL,
    0x97e7e9f785becULL,
    0xc274c8347a896ULL,
    0xeed8c428e2577ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x9800000000000ULL,
    0x1400000000000ULL,
    0x8200000000000ULL,
    0x6100000000000ULL,
    0xf180000000000ULL,
    0x5e40000000000ULL,
    0xbe60000000000ULL,
    0x5b10000000000ULL,
    0x5d98000000000ULL,
    0x7d5c000000000ULL,
    0x8bfa000000000ULL,
    0x6a4f000000000ULL,
    0xe064800000000ULL,
    0x8414400000000ULL,
    0x631ae00000000ULL,
    0x559ff00000000ULL,
    0xe159780000000ULL,
    0x55ff340000000ULL,
    0x754ee60000000ULL,
    0x7ee3c50000000ULL,
    0xdcd6b58000000ULL,
    0x91b9c54000000ULL,
    0x7e2ad5a000000ULL,
    0xf671757000000ULL,
    0xce0fcd8800000ULL,
    0x3b06f16c00000ULL,
    0xe48433ae00000ULL,
    0xfdc2b06900000ULL,
    0x6921780e80000ULL,
    0x98f3343140000ULL,
    0xcecce612e0000ULL,
    0x7da0c516d0000ULL,
    0x3d30359578000ULL,
    0x13ee855c14000ULL,
    0x1756b5841e000ULL,
    0xaf9c57643000ULL,
    0x8bcad5ae60800ULL,
    0xf621756890c00ULL,
    0xc677cd8c19a00ULL,
    0x360af15b05d00ULL,
    0xff06339bc7180ULL,
    0x8e81b040e62c0ULL,
    0x10c7f80897460ULL,
    0xea4741d5ee10ULL,
    0x25b08638a27f8ULL,
    0xe028753eb1d04ULL,
    0x89752d88e8f2eULL,
    0x808901788db11ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x5800000000000ULL,
    0x5400000000000ULL,
    0x5200000000000ULL,
    0x7100000000000ULL,
    0x8a80000000000ULL,
    0x20c0000000000ULL,
    0xb860000000000ULL,
    0x4130000000000ULL,
    0x49f8000000000ULL,
    0x1cbc000000000ULL,
    0xc19e000000000ULL,
    0x558b000000000ULL,
    0x2463800000000ULL,
    0x5f35400000000ULL,
    0x9efde00000000ULL,
    0xa53e700000000ULL,
    0x42df780000000ULL,
    0x13acec0000000ULL,
    0xaff7ca0000000ULL,
    0x9e9e590000000ULL,
    0xb788000000ULL,
    0x1120f34000000ULL,
    0x721318a000000ULL,
    0xe2ecc35000000ULL,
    0xc8d4008800000ULL,
    0x678c1f5400000ULL,
    0xa564d29e00000ULL,
    0x3db29a4100000ULL,
    0xa23f780a80000ULL,
    0x435cec1d40000ULL,
    0x74efca0820000ULL,
    0xf3d2590570000ULL,
    0x680d788e38000ULL,
    0x4d27f3513c000ULL,
    0x8c1698aa46000ULL,
    0x85ee835313000ULL,
    0x395460a6a1800ULL,
    0xe8cc2f5e31c00ULL,
    0xe545ca9b19600ULL,
    0x95254654cc900ULL,
    0x98122a063e180ULL,
    0x77ec29318f0c0ULL,
    0x5852008657c60ULL,
    0x3a4b1f6a7b370ULL,
    0x918152bbd5b48ULL,
    0x7f40da5cf2614ULL,
    0xa827181eb18aaULL,
    0xb490dc19580d9ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0x4400000000000ULL,
    0xca00000000000ULL,
    0x4500000000000ULL,
    0xdd80000000000ULL,
    0xd840000000000ULL,
    0x6620000000000ULL,
    0x58d0000000000ULL,
    0xa378000000000ULL,
    0x555c000000000ULL,
    0xb58a000000000ULL,
    0x32f1000000000ULL,
    0xbdad800000000ULL,
    0x3524400000000ULL,
    0x9e57600000000ULL,
    0x31bcb00000000ULL,
    0x27bad80000000ULL,
    0x3cbb540000000ULL,
    0x7639260000000ULL,
    0x62fccf0000000ULL,
    0xb31f5d8000000ULL,
    0x54aa604000000ULL,
    0xdea585a000000ULL,
    0xed11347000000ULL,
    0x151ca3b800000ULL,
    0x47adfb7400000ULL,
    0xb823fe2600000ULL,
    0xfd79b2500000ULL,
    0xffe7bb480000ULL,
    0xf29aaf4440000ULL,
    0x72e8d83be0000ULL,
    0x5f86540650000ULL,
    0x1146a60468000ULL,
    0x3da58f3564000ULL,
    0x2397bd9836000ULL,
    0xabdf906473000ULL,
    0xc34fbd8eb4800ULL,
    0xfa13906e9f400ULL,
    0x199dbda5b7e00ULL,
    0x906e9056b2500ULL,
    0x17423d9a14380ULL,
    0x5ea7d07eae7c0ULL,
    0x2d12ddbd4e0e0ULL,
    0x751e206f1fe90ULL,
    0x57aae5bd77a78ULL,
    0xc0218455128acULL,
    0x4bd47bada448eULL,
    0xc5fbaf5b966edULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x1800000000000ULL,
    0xdc00000000000ULL,
    0xbe00000000000ULL,
    0x6300000000000ULL,
    0x2580000000000ULL,
    0x30c0000000000ULL,
    0xf120000000000ULL,
    0xf5b0000000000ULL,
    0xb6f8000000000ULL,
    0xafdc000000000ULL,
    0xbe6e000000000ULL,
    0x3a95000000000ULL,
    0x6048800000000ULL,
    0x5ca6400000000ULL,
    0x1175a00000000ULL,
    0x6dd8d00000000ULL,
    0x736d880000000ULL,
    0x3411040000000ULL,
    0x810c520000000ULL,
    0xe744530000000ULL,
    0x14e1fa8000000ULL,
    0xd6119f4000000ULL,
    0x1c0c72a000000ULL,
    0xc1c09b5000000ULL,
    0x31a020b800000ULL,
    0xff4c85c00000ULL,
    0xc499da3e00000ULL,
    0x3c89572b00000ULL,
    0x4b83a8b180000ULL,
    0x3bc0cc63c0000ULL,
    0x70a50800e0000ULL,
    0x97774433f0000ULL,
    0x82d9f202a8000ULL,
    0xc0ec83259c000ULL,
    0xa3d4729e96000ULL,
    0x37ac9b6dab000ULL,
    0xff3620b1ed800ULL,
    0x15bdc850fe400ULL,
    0x993f5a2b57a00ULL,
    0xd57a17224bd00ULL,
    0x501e88919cc80ULL,
    0xacce5c4496e40ULL,
    0x70e52024aa2a0ULL,
    0xa41290156ce50ULL,
    0x890e2829be9e8ULL,
    0x7340d439f621cULL,
    0x5ee75a129a282ULL,
    0xcf16171024a21ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0x5c00000000000ULL,
    0xba00000000000ULL,
    0xd500000000000ULL,
    0x7880000000000ULL,
    0x89c0000000000ULL,
    0xa360000000000ULL,
    0x2490000000000ULL,
    0xf188000000000ULL,
    0x6eec000000000ULL,
    0x207a000000000ULL,
    0x8cf3000000000ULL,
    0x6e9e800000000ULL,
    0xb920400000000ULL,
    0xa5b1a00000000ULL,
    0xf83af00000000ULL,
    0x94d4180000000ULL,
    0xdaa7c0000000ULL,
    0xdbdec20000000ULL,
    0xb584c10000000ULL,
    0x6d45198000000ULL,
    0xb0a217c000000ULL,
    0x62f101a000000ULL,
    0x35986bf000000ULL,
    0x4aa7c3a800000ULL,
    0x57f0aafc00000ULL,
    0x5d18da0e00000ULL,
    0x6b61bd2b00000ULL,
    0x30975b8080000ULL,
    0x9f8996ee40000ULL,
    0xf5e93836a0000ULL,
    0xf3ffcc3030000ULL,
    0x4eb7fa1078000ULL,
    0x53bb0d1214000ULL,
    0x1a92e3ac9a000ULL,
    0x22891ae7ef000ULL,
    0x696be21224800ULL,
    0xac3d7134e3c00ULL,
    0x1ad621b9fb200ULL,
    0x86aedbf8bfd00ULL,
    0xa0587b940d480ULL,
    0xb3c326dd5e6c0ULL,
    0xb664802d1cce0ULL,
    0x7c134001d7330ULL,
    0x484f2002905e8ULL,
    0x558ab00a49d34ULL,
    0x58edb8336c50eULL,
    0xc77c8c070f3abULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0xb800000000000ULL,
    0x3c00000000000ULL,
    0x6a00000000000ULL,
    0xf900000000000ULL,
    0xe980000000000ULL,
    0x81c0000000000ULL,
    0xbe0000000000ULL,
    0x2e50000000000ULL,
    0xcc8000000000ULL,
    0xed2c000000000ULL,
    0xfa7a000000000ULL,
    0x3f33000000000ULL,
    0xe3da800000000ULL,
    0xaa45400000000ULL,
    0x8727a00000000ULL,
    0x236700000000ULL,
    0xbc58f80000000ULL,
    0x2882140000000ULL,
    0x94436a0000000ULL,
    0xc0214d0000000ULL,
    0x6cb76b8000000ULL,
    0x731bc7c000000ULL,
    0x5c2793a000000ULL,
    0x76b5d3f000000ULL,
    0x121ef9a800000ULL,
    0x39a79edc00000ULL,
    0x2573120600000ULL,
    0xdcf9191300000ULL,
    0x947321bc80000ULL,
    0x517abadd40000ULL,
    0xd3b5202ee0000ULL,
    0xc19f302130000ULL,
    0x6965582a18000ULL,
    0xf3176439a4000ULL,
    0x9d691215aa000ULL,
    0x3c5a191faf000ULL,
    0xe881a1a5d4800ULL,
    0xb443fac153c00ULL,
    0x7020801f3a200ULL,
    0xd4b6400e66900ULL,
    0x4f1d202d59c80ULL,
    0x36233002440c0ULL,
    0x8fb7581b1b760ULL,
    0xfb98641477490ULL,
    0xb861920fd18c8ULL,
    0x2e90590b298a4ULL,
    0xf2ae8193bd9e6ULL,
    0x98bfcaec0f77bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0x9400000000000ULL,
    0x5600000000000ULL,
    0xcb00000000000ULL,
    0x7480000000000ULL,
    0xdb40000000000ULL,
    0x9a0000000000ULL,
    0x78d0000000000ULL,
    0x5548000000000ULL,
    0x5aec000000000ULL,
    0xb13e000000000ULL,
    0xccf7000000000ULL,
    0x7f5c800000000ULL,
    0xb2c5c00000000ULL,
    0x2d66600000000ULL,
    0xfeb0500000000ULL,
    0xf7fc80000000ULL,
    0x96d40c0000000ULL,
    0x524d860000000ULL,
    0x246a830000000ULL,
    0x937a898000000ULL,
    0xc4d2c5c000000ULL,
    0xf74d41a000000ULL,
    0x17eac9d000000ULL,
    0x16bec7a800000ULL,
    0x1f374aec00000ULL,
    0x6fb8ce1200000ULL,
    0x5eb04f0900000ULL,
    0xbf7befa780000ULL,
    0x5ed6d6ebc0000ULL,
    0x364ce012a0000ULL,
    0x9a6e903110000ULL,
    0xcc7b2802c8000ULL,
    0xe6569c2624000ULL,
    0xe708ae301a000ULL,
    0x6acb1f1b5f000ULL,
    0xb52ea7aef5800ULL,
    0x43dc1ac972c00ULL,
    0x4d85861beaa00ULL,
    0xbac6832fe7b00ULL,
    0x296489b2a9380ULL,
    0x90b5c5c214a40ULL,
    0x4879c198034a0ULL,
    0xc85309d513670ULL,
    0x800ea78dda8b8ULL,
    0xc44c1afefe77cULL,
    0x8f6d86236511eULL,
    0x37fa830efa62bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x9c00000000000ULL,
    0x5600000000000ULL,
    0xa100000000000ULL,
    0x7f80000000000ULL,
    0x3c0000000000ULL,
    0xb6e0000000000ULL,
    0x3410000000000ULL,
    0xf78000000000ULL,
    0x279c000000000ULL,
    0x668a000000000ULL,
    0x5475000000000ULL,
    0x732c800000000ULL,
    0x6fc0400000000ULL,
    0x18e1a00000000ULL,
    0x5917900000000ULL,
    0x4efed80000000ULL,
    0xc15e840000000ULL,
    0x65e80a0000000ULL,
    0x94a1e50000000ULL,
    0x1435818000000ULL,
    0xbd04000000ULL,
    0xc43359a000000ULL,
    0xa809547000000ULL,
    0x6031539800000ULL,
    0xd20db15c00000ULL,
    0x4f30522e00000ULL,
    0x908a213b00000ULL,
    0x6570ab9b80000ULL,
    0x84ade56340000ULL,
    0x9801a01c60000ULL,
    0xbc07903fb0000ULL,
    0x606d81fd8000ULL,
    0xc9028409b4000ULL,
    0x3b820a0786000ULL,
    0xc9c4e52313000ULL,
    0x41e101a244800ULL,
    0xea979079be400ULL,
    0x7338f9b957a00ULL,
    0x92bbc477fa500ULL,
    0xe47b0bb6e9c80ULL,
    0x6f1f7549acb40ULL,
    0x5bcbf82492fe0ULL,
    0x2ed5540f05550ULL,
    0x2a1d720edf5c8ULL,
    0x7e4df135677c4ULL,
    0x5217d3af63f4eULL,
    0xb678f1413d3c5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0xd400000000000ULL,
    0xa200000000000ULL,
    0x4300000000000ULL,
    0xbb80000000000ULL,
    0xcdc0000000000ULL,
    0xede0000000000ULL,
    0xbf50000000000ULL,
    0xc58000000000ULL,
    0xa7bc000000000ULL,
    0xa1ee000000000ULL,
    0x6131000000000ULL,
    0xa2c8800000000ULL,
    0x7806400000000ULL,
    0x2401e00000000ULL,
    0x5a04700000000ULL,
    0x2704280000000ULL,
    0x182140000000ULL,
    0x5ac70a0000000ULL,
    0xf461cd0000000ULL,
    0x3196e38000000ULL,
    0x5a3f2b4000000ULL,
    0xd52acba000000ULL,
    0x40513f5000000ULL,
    0x79dbc19800000ULL,
    0xf7df27c00000ULL,
    0xbe8ba21e00000ULL,
    0x2725990900000ULL,
    0x4330099680000ULL,
    0xa1ca9649c0000ULL,
    0x2380002060000ULL,
    0x59c00006b0000ULL,
    0xafe0000518000ULL,
    0x4c50002064000ULL,
    0xafd800108e000ULL,
    0xbe7c00004b000ULL,
    0xee0e003895800ULL,
    0x9d61000eb7400ULL,
    0x1510802d3fa00ULL,
    0x127a401cd2d00ULL,
    0x680fe032b1d80ULL,
    0x8465703f18cc0ULL,
    0x8994a82b65520ULL,
    0xde3854210f090ULL,
    0xdf28ea3e0b308ULL,
    0xcf54bd33b5aa4ULL,
    0xb45a4bbba7506ULL,
    0x23bb7f75f6ac9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0x9400000000000ULL,
    0xba00000000000ULL,
    0xe900000000000ULL,
    0x7b80000000000ULL,
    0x7d40000000000ULL,
    0xa520000000000ULL,
    0xccb0000000000ULL,
    0xa908000000000ULL,
    0xba2c000000000ULL,
    0xc19e000000000ULL,
    0x817000000000ULL,
    0xa1fb800000000ULL,
    0x9586400000000ULL,
    0x4e45e00000000ULL,
    0x7a3b00000000ULL,
    0xa276b80000000ULL,
    0x9d6a840000000ULL,
    0xfdbfde0000000ULL,
    0xf726cf0000000ULL,
    0x89b4ee8000000ULL,
    0xdc8c58c000000ULL,
    0x846a56a000000ULL,
    0x8e3adcd000000ULL,
    0x566388a800000ULL,
    0x6a9713d400000ULL,
    0xaf3ae63600000ULL,
    0x91e60b3100000ULL,
    0x61d0d0b880000ULL,
    0xdd1e27db40000ULL,
    0xdd3800e60000ULL,
    0xb31a401010000ULL,
    0x7ed3e01128000ULL,
    0x3198b0123c000ULL,
    0x133803ea000ULL,
    0x7dfbc427f7000ULL,
    0xd381be1ded800ULL,
    0xf1433f25e2c00ULL,
    0xbb27b6acbb600ULL,
    0xf7b56cd6bf500ULL,
    0xd78b3082c0180ULL,
    0x92ea97c5a0bc0ULL,
    0x4afeb81130360ULL,
    0xf306843958dd0ULL,
    0xc281de2204a18ULL,
    0x5ec1cf387e3ecULL,
    0x5c676ead61822ULL,
    0xab9618d79194fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0x2c00000000000ULL,
    0xa00000000000ULL,
    0x6900000000000ULL,
    0x4580000000000ULL,
    0xf5c0000000000ULL,
    0x9ae0000000000ULL,
    0x2130000000000ULL,
    0x7948000000000ULL,
    0xcdac000000000ULL,
    0x3b9a000000000ULL,
    0x5851000000000ULL,
    0xebe800000000ULL,
    0x6285400000000ULL,
    0x1745200000000ULL,
    0x4da3f00000000ULL,
    0xc91f80000000ULL,
    0x5d8f40000000ULL,
    0xf073aa0000000ULL,
    0xe7ef050000000ULL,
    0xb5be388000000ULL,
    0xd205b8c000000ULL,
    0xf507c0a000000ULL,
    0x17814cf000000ULL,
    0xc0c66ab800000ULL,
    0xed6349ec00000ULL,
    0x91f4d23600000ULL,
    0xac2eb12700000ULL,
    0x705ab29480000ULL,
    0x9db44dcd40000ULL,
    0x410c803de0000ULL,
    0xd688401c30000ULL,
    0xd649a02d58000ULL,
    0x5b2bb0337c000ULL,
    0xbad858278a000ULL,
    0x2ef3440d57000ULL,
    0x72abf2348c800ULL,
    0xe51c413773400ULL,
    0xcd15ca827b200ULL,
    0x5e19f9c0ae100ULL,
    0x7d920a0e20680ULL,
    0xbc58b52591ec0ULL,
    0x27b4608308960ULL,
    0x700bfce5d4910ULL,
    0xcf08b29a0f5c8ULL,
    0x11894dc6f1404ULL,
    0x84c8002b78baaULL,
    0xd46c0022edc27ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0xc400000000000ULL,
    0xd600000000000ULL,
    0xe300000000000ULL,
    0xb480000000000ULL,
    0x1940000000000ULL,
    0x2c60000000000ULL,
    0xa590000000000ULL,
    0xe6b8000000000ULL,
    0x49dc000000000ULL,
    0x88ce000000000ULL,
    0x7f33000000000ULL,
    0x768e800000000ULL,
    0xd1d4400000000ULL,
    0xb959600000000ULL,
    0x78d500000000ULL,
    0x1050080000000ULL,
    0x579fcc0000000ULL,
    0xb62cd60000000ULL,
    0x18e1fd0000000ULL,
    0xfcd67d8000000ULL,
    0x6adace4000000ULL,
    0xdc4875a000000ULL,
    0xf676025000000ULL,
    0xf2ea238800000ULL,
    0x2843bf5c00000ULL,
    0x4de53e2e00000ULL,
    0x7b54212d00000ULL,
    0xcf1d43bb80000ULL,
    0x856def45c0000ULL,
    0xf583b61ea0000ULL,
    0x40c3ad1f90000ULL,
    0x6a6f5bbb8000ULL,
    0xc232425bf4000ULL,
    0xd0b439996000ULL,
    0x5612ef5c43000ULL,
    0x4d7b362d05800ULL,
    0x97f8ed001fc00ULL,
    0xd9bf15a51e200ULL,
    0xaf5852429ed00ULL,
    0x48caba95e780ULL,
    0x34d4335afe840ULL,
    0xe6de881eeffa0ULL,
    0xc64b8c1096b70ULL,
    0xaf75b60dc2ef8ULL,
    0x2f6cad1fc4184ULL,
    0x748675aa3f08eULL,
    0xf945024fcf2fdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0xc800000000000ULL,
    0xa400000000000ULL,
    0xca00000000000ULL,
    0x7900000000000ULL,
    0x6880000000000ULL,
    0x4940000000000ULL,
    0xf220000000000ULL,
    0x8d30000000000ULL,
    0xada8000000000ULL,
    0x2e8c000000000ULL,
    0xcf3e000000000ULL,
    0x51d1000000000ULL,
    0x6c7f800000000ULL,
    0x2af7c00000000ULL,
    0xf1cbe00000000ULL,
    0xf21e300000000ULL,
    0x8b64380000000ULL,
    0xa7159c0000000ULL,
    0xac9a660000000ULL,
    0xbd200f0000000ULL,
    0x9ab31a8000000ULL,
    0x2b69a1c000000ULL,
    0xd66922a000000ULL,
    0x84ed3dd000000ULL,
    0xccacc49800000ULL,
    0x6a0af2c400000ULL,
    0x87c3e2a00000ULL,
    0x40f1632d00000ULL,
    0x38cf24b880000ULL,
    0x298c2e340000ULL,
    0xae26063aa0000ULL,
    0x3b35ff3510000ULL,
    0x92aac2ac28000ULL,
    0x10f0dfb5c000ULL,
    0x95fefc9fc2000ULL,
    0x5f326ec3bf000ULL,
    0xf8afd80ead800ULL,
    0xc80bac26db400ULL,
    0x657e5e139a600ULL,
    0x7a75933a3ab00ULL,
    0x64897ca8aa980ULL,
    0xf639aef6423c0ULL,
    0xd95238267e160ULL,
    0x75389c084c1b0ULL,
    0x10d3e618eaef8ULL,
    0xd8facf0b62a8cULL,
    0x95b17a91aeb12ULL,
    0xdced51c705fcbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x3800000000000ULL,
    0x400000000000ULL,
    0x1200000000000ULL,
    0xd00000000000ULL,
    0xe780000000000ULL,
    0xfec0000000000ULL,
    0x7c60000000000ULL,
    0xc570000000000ULL,
    0x358000000000ULL,
    0xba3c000000000ULL,
    0xe4ce000000000ULL,
    0x6f55000000000ULL,
    0xc60a800000000ULL,
    0x3535400000000ULL,
    0x4478600000000ULL,
    0xa8ee700000000ULL,
    0xcd05580000000ULL,
    0xc787040000000ULL,
    0xeec0020000000ULL,
    0x3461f90000000ULL,
    0xf973b68000000ULL,
    0x155a5e4000000ULL,
    0xa538eea000000ULL,
    0xe485a5000000ULL,
    0x76126cb800000ULL,
    0x44ace34c00000ULL,
    0x8c21ba3600000ULL,
    0x8254cd0d00000ULL,
    0x118a0c9280000ULL,
    0x93f2935e40000ULL,
    0x4c1ce21360000ULL,
    0x479fc92df0000ULL,
    0xc75c0eba78000ULL,
    0x883a6a51ac000ULL,
    0xf9cbd4882e000ULL,
    0xc0d5d77163000ULL,
    0x4ce001a7d800ULL,
    0x5f55000b46400ULL,
    0x9e0a802787200ULL,
    0x4135401aa7100ULL,
    0x6e78600c37380ULL,
    0xa1ee700fbf8c0ULL,
    0x3885582e6bee0ULL,
    0x34470404e80d0ULL,
    0x7520020624ce8ULL,
    0xfd1f903fb634ULL,
    0x864bb698e11b2ULL,
    0x6a165e6b3178dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x2800000000000ULL,
    0x8c00000000000ULL,
    0xfe00000000000ULL,
    0x6100000000000ULL,
    0x8280000000000ULL,
    0xc540000000000ULL,
    0x64e0000000000ULL,
    0x3270000000000ULL,
    0xc388000000000ULL,
    0x6a6c000000000ULL,
    0x549a000000000ULL,
    0xbb57000000000ULL,
    0xb5e800000000ULL,
    0x7ef5c00000000ULL,
    0x8fcea00000000ULL,
    0xa00bd00000000ULL,
    0xedaf180000000ULL,
    0x553eac0000000ULL,
    0x5dc7d60000000ULL,
    0xd2a1290000000ULL,
    0x5b12368000000ULL,
    0x4f3fecc000000ULL,
    0x2ec72ea000000ULL,
    0xdf2640f000000ULL,
    0xe5d6788800000ULL,
    0xc81ea9dc00000ULL,
    0xcf10ee0200000ULL,
    0xd53d953f00000ULL,
    0x9dc6589d80000ULL,
    0x72a0b9d740000ULL,
    0xeb11563520000ULL,
    0x6738e93a10000ULL,
    0xa2c696ae18000ULL,
    0x21233cc98c000ULL,
    0x84d6b6a37a000ULL,
    0x4a9d2cf457000ULL,
    0xa570e8b7c800ULL,
    0xb1d850fc92c00ULL,
    0xafb7c085b3200ULL,
    0xb12bd5e5a3b00ULL,
    0x81782010ba880ULL,
    0x33a21032b6440ULL,
    0x1993b8178dce0ULL,
    0x2a7e7c1e7b2f0ULL,
    0xfa244e16d71b8ULL,
    0xc5514521bcd54ULL,
    0xaa5fc08ef31e2ULL,
    0x5c77d5d2826a7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x3400000000000ULL,
    0x5a00000000000ULL,
    0x5b00000000000ULL,
    0xd080000000000ULL,
    0x1940000000000ULL,
    0xf6a0000000000ULL,
    0xe3d0000000000ULL,
    0x41d8000000000ULL,
    0xb07c000000000ULL,
    0xa52e000000000ULL,
    0x70b1000000000ULL,
    0x83ef800000000ULL,
    0x55d3400000000ULL,
    0x34dbe00000000ULL,
    0x81fab00000000ULL,
    0x67ed280000000ULL,
    0x17d54c0000000ULL,
    0xbbdab60000000ULL,
    0x5b786b0000000ULL,
    0x7dad598000000ULL,
    0x5f4914000000ULL,
    0x1b4e71a000000ULL,
    0xb700dd5000000ULL,
    0xfe83479800000ULL,
    0xf847f64400000ULL,
    0x2d23fe0e00000ULL,
    0x7214d70900000ULL,
    0x9a3927b880000ULL,
    0xff0f064fc0000ULL,
    0x5c22b614a0000ULL,
    0x51946b0d90000ULL,
    0x8ffb59a708000ULL,
    0x76e9917c64000ULL,
    0x2457f1affe000ULL,
    0x161e9d4fbb000ULL,
    0xb699278a15800ULL,
    0x7df065a07400ULL,
    0x2d7ab623e6600ULL,
    0xa8a86b2a17900ULL,
    0x84755981ae980ULL,
    0xd188915e123c0ULL,
    0xbc6071b0e0520ULL,
    0xa8b1dd51702f0ULL,
    0xf7ecc794789f8ULL,
    0xefd4b6431da5cULL,
    0x3fd81e1f6351eULL,
    0x97e671c99bd9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x2800000000000ULL,
    0xfc00000000000ULL,
    0x7200000000000ULL,
    0x3100000000000ULL,
    0x6580000000000ULL,
    0x2740000000000ULL,
    0xe60000000000ULL,
    0x5710000000000ULL,
    0xfda8000000000ULL,
    0x40cc000000000ULL,
    0x2b5e000000000ULL,
    0x835000000000ULL,
    0x63da800000000ULL,
    0x8f5c00000000ULL,
    0x33ff200000000ULL,
    0xb802300000000ULL,
    0xe404780000000ULL,
    0xc603740000000ULL,
    0xcf07fe0000000ULL,
    0xe80b50000000ULL,
    0x8fc5998000000ULL,
    0x3ea59ac000000ULL,
    0x4f37e1a000000ULL,
    0xc15feed000000ULL,
    0xcd349fb800000ULL,
    0x985f9bcc00000ULL,
    0x34b4a63200000ULL,
    0xbd1df13300000ULL,
    0x23d01f8e80000ULL,
    0x738f5bef40000ULL,
    0x8b39062fe0000ULL,
    0xc26012a90000ULL,
    0x92f5479e78000ULL,
    0xdefb1fcc54000ULL,
    0xbf80000f92000ULL,
    0xaa40002327000ULL,
    0x79e00032a5800ULL,
    0x31500017ba400ULL,
    0xbe48003e2b200ULL,
    0xcc9c0026d3300ULL,
    0xaa960001c6e80ULL,
    0x2ee90037b4bc0ULL,
    0xd0ac800b02760ULL,
    0x674cc0045e210ULL,
    0x751ba00f70a98ULL,
    0xefd2f00be93c4ULL,
    0x4989d80ead27eULL,
    0x3638842f87e6fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x1800000000000ULL,
    0xac00000000000ULL,
    0x1e00000000000ULL,
    0x9b00000000000ULL,
    0x3f80000000000ULL,
    0x7a40000000000ULL,
    0x91a0000000000ULL,
    0xe750000000000ULL,
    0x9818000000000ULL,
    0x84bc000000000ULL,
    0x136a000000000ULL,
    0x8cb7000000000ULL,
    0x2ee9800000000ULL,
    0x5bf3400000000ULL,
    0x5fc9200000000ULL,
    0x364500000000ULL,
    0x1d35c80000000ULL,
    0xe7a8140000000ULL,
    0xb1d5160000000ULL,
    0x38d8930000000ULL,
    0x3bd8dd8000000ULL,
    0x685efb4000000ULL,
    0xac1f15a000000ULL,
    0xf6bdef7000000ULL,
    0x3669838800000ULL,
    0x78313c6400000ULL,
    0x7329fe0e00000ULL,
    0x1c14d70f00000ULL,
    0x373803a980000ULL,
    0xe72e7c59c0000ULL,
    0x3e12de2e60000ULL,
    0xa3b872350000ULL,
    0xbfae4b8408000ULL,
    0x7dd2285584000ULL,
    0xd6df682b8e000ULL,
    0xe8d8041d6b000ULL,
    0xe3d87e2934800ULL,
    0x645b970916c00ULL,
    0xe21b23a086e00ULL,
    0x75bd2c616e100ULL,
    0xa5ee961b7aa80ULL,
    0x1c70d31a1c040ULL,
    0x798a7d9ae35a0ULL,
    0xc4c2eb4470c10ULL,
    0xd5607dba78ff8ULL,
    0xf235eb76fcf5cULL,
    0xca29fd87f260eULL,
    0x1e96ab58d84f1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xe800000000000ULL,
    0x7400000000000ULL,
    0x8a00000000000ULL,
    0xd700000000000ULL,
    0xcf80000000000ULL,
    0x5bc0000000000ULL,
    0x2e60000000000ULL,
    0x76d0000000000ULL,
    0xbdd8000000000ULL,
    0x9abc000000000ULL,
    0xaeee000000000ULL,
    0x2871000000000ULL,
    0xe4ee800000000ULL,
    0x5f77c00000000ULL,
    0xbb6d600000000ULL,
    0x9cb3b00000000ULL,
    0x908c80000000ULL,
    0x1460d40000000ULL,
    0xe9d03a0000000ULL,
    0x9659c10000000ULL,
    0xd37a838000000ULL,
    0xcb8c474000000ULL,
    0x6f244ba000000ULL,
    0x5ff1935000000ULL,
    0xf32cf19800000ULL,
    0xf712924c00000ULL,
    0x7cbb922200000ULL,
    0x2feba51100000ULL,
    0x2ef4f1a080000ULL,
    0x5dae9274c0000ULL,
    0xda55920b60000ULL,
    0x39aa52e30000ULL,
    0xa81a71a2f8000ULL,
    0xa1d9526284000ULL,
    0x24b8f20d0e000ULL,
    0x13e91520eb000ULL,
    0x40f2b995d5800ULL,
    0x98a98658c7400ULL,
    0x5ed0c81416200ULL,
    0x69dcd4209eb00ULL,
    0x80be3a2d43780ULL,
    0xe1e8c10e995c0ULL,
    0x7bf403b334ca0ULL,
    0x412b8758b6290ULL,
    0x2c112bb44ee48ULL,
    0xd53e2375cae94ULL,
    0xb52a39b1453c6ULL,
    0xe6134679efa49ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x5800000000000ULL,
    0x8400000000000ULL,
    0xb200000000000ULL,
    0x9500000000000ULL,
    0xd880000000000ULL,
    0xb040000000000ULL,
    0xb3a0000000000ULL,
    0x63d0000000000ULL,
    0xfc8000000000ULL,
    0xf76c000000000ULL,
    0xf73e000000000ULL,
    0x1337000000000ULL,
    0x283a800000000ULL,
    0xaab6400000000ULL,
    0x8afd600000000ULL,
    0x8856700000000ULL,
    0x680e980000000ULL,
    0x45083c0000000ULL,
    0x498c4e0000000ULL,
    0x3ce8b0000000ULL,
    0xc96d098000000ULL,
    0x1c39e9c000000ULL,
    0xb591a000000ULL,
    0xbbfad5d000000ULL,
    0xd2d55fa800000ULL,
    0x15491edc00000ULL,
    0xaa29b63e00000ULL,
    0x281bc73d00000ULL,
    0x2aa35fb280000ULL,
    0x85521ec840000ULL,
    0x548d362ca0000ULL,
    0x574a873850000ULL,
    0x72cbfbe48000ULL,
    0x649e2edda4000ULL,
    0x40e0ce2f3a000ULL,
    0x7ff3cb383f000ULL,
    0xd9dce9b655800ULL,
    0xcfc2d9c8a2c00ULL,
    0xe62e9b413200ULL,
    0x98b5d9ceab500ULL,
    0x5ff869bd16780ULL,
    0x70d399edc1cc0ULL,
    0xe84d09b9605e0ULL,
    0xaea9e9f531030ULL,
    0xae5d91a9786d8ULL,
    0xbe06d5f7ddf1cULL,
    0xab035f976757eULL,
    0x33821ef59947bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x6800000000000ULL,
    0x9400000000000ULL,
    0x2600000000000ULL,
    0x7700000000000ULL,
    0x280000000000ULL,
    0x9140000000000ULL,
    0xfea0000000000ULL,
    0x6610000000000ULL,
    0x4198000000000ULL,
    0x8dbc000000000ULL,
    0xc2ea000000000ULL,
    0x21d7000000000ULL,
    0x687b800000000ULL,
    0xd20ac00000000ULL,
    0x9fe5600000000ULL,
    0x80b5d00000000ULL,
    0x6b8ba80000000ULL,
    0x6255c0000000ULL,
    0xe256c60000000ULL,
    0x24bbfb0000000ULL,
    0x8b6e478000000ULL,
    0xa014254000000ULL,
    0x469e6fa000000ULL,
    0x573bb97000000ULL,
    0xbfadc98800000ULL,
    0x575927400000ULL,
    0xcb68261200000ULL,
    0xc014eb1100000ULL,
    0xf6988fbb80000ULL,
    0x2f38a97cc0000ULL,
    0x43a901b320000ULL,
    0xb7721e6c90000ULL,
    0x9a6ec83e88000ULL,
    0xb5908c1414000ULL,
    0x655d6e1d6e000ULL,
    0x40dea73fbb000ULL,
    0xdb18818cf4800ULL,
    0x90ffde4a5ec00ULL,
    0x56482813e2600ULL,
    0xfac39c1608700ULL,
    0x8661a61a55380ULL,
    0x9752b250e040ULL,
    0x616c6f92cb520ULL,
    0xdd10b94acd930ULL,
    0x491c498cd2838ULL,
    0x11f8526e501dcULL,
    0xebcec622a8aa6ULL,
    0xed07fb0c05d09ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x2800000000000ULL,
    0xc400000000000ULL,
    0xce00000000000ULL,
    0xf500000000000ULL,
    0x6b80000000000ULL,
    0x80c0000000000ULL,
    0x10a0000000000ULL,
    0x7d70000000000ULL,
    0x3c68000000000ULL,
    0x1f4c000000000ULL,
    0xfefa000000000ULL,
    0xc035000000000ULL,
    0xa50f800000000ULL,
    0xdb1cc00000000ULL,
    0x1023e00000000ULL,
    0x8b30100000000ULL,
    0xe38e780000000ULL,
    0xec59440000000ULL,
    0x7245920000000ULL,
    0x31e68f0000000ULL,
    0x3617308000000ULL,
    0x1f39b9c000000ULL,
    0x9d16c8a000000ULL,
    0x89bc3df000000ULL,
    0x8250ba9800000ULL,
    0x2fdaa2c400000ULL,
    0xf681f21200000ULL,
    0xf465f3500000ULL,
    0x6e60a88180000ULL,
    0x60d5edf840000ULL,
    0xc69aa28360000ULL,
    0xa9e336dc70000ULL,
    0xca13f83548000ULL,
    0xdd3c84062c000ULL,
    0xb213f223d2000ULL,
    0xf13f5f3dbf000ULL,
    0x5815288d0c800ULL,
    0x1a3c2de4c2400ULL,
    0xee96c29087600ULL,
    0x357fe6ec69700ULL,
    0xb0f660143d980ULL,
    0xada9d01f6a740ULL,
    0x9a6a182c9a860ULL,
    0xe49943292650ULL,
    0x6b7a0a295e828ULL,
    0x4df6db2e3db6cULL,
    0x32295a966a56aULL,
    0x2cafb2eb1becbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x800000000000ULL,
    0xc00000000000ULL,
    0x3200000000000ULL,
    0xc500000000000ULL,
    0x7e80000000000ULL,
    0x60c0000000000ULL,
    0xd260000000000ULL,
    0x8db0000000000ULL,
    0x6e8000000000ULL,
    0x60c000000000ULL,
    0x897a000000000ULL,
    0xb0f3000000000ULL,
    0xd9cb800000000ULL,
    0xc99c00000000ULL,
    0xa106a00000000ULL,
    0x3086b00000000ULL,
    0x2fc2580000000ULL,
    0xfde08c0000000ULL,
    0xa5f6fa0000000ULL,
    0x4f4f910000000ULL,
    0xd05a408000000ULL,
    0xd96515c000000ULL,
    0x7c3398a000000ULL,
    0x69ac59f000000ULL,
    0x6c2bc2a800000ULL,
    0x456978d400000ULL,
    0x6c49da2a00000ULL,
    0xbddfe11700000ULL,
    0x8c27388680000ULL,
    0xdf15e9d640000ULL,
    0xe9381ab120000ULL,
    0x155334da70000ULL,
    0x8d9a000fb8000ULL,
    0x183001e1c000ULL,
    0x1743802082000ULL,
    0xde25c01ddf000ULL,
    0x6a14a019fc800ULL,
    0x2fb9b03c93c00ULL,
    0xe193d806b6600ULL,
    0x69fa4c07e9b00ULL,
    0x7733da1755c80ULL,
    0x982ce118671c0ULL,
    0x36cb8a0e1de0ULL,
    0x2f4c29d290ff0ULL,
    0xa05eba8928008ULL,
    0x616584fd35ab4ULL,
    0xe830582ab7d8aULL,
    0x5faf8c21e8323ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0x2400000000000ULL,
    0x9200000000000ULL,
    0xc900000000000ULL,
    0xca80000000000ULL,
    0xf9c0000000000ULL,
    0x36a0000000000ULL,
    0xf250000000000ULL,
    0xbce8000000000ULL,
    0x4ccc000000000ULL,
    0x315a000000000ULL,
    0x4017000000000ULL,
    0xd0c800000000ULL,
    0x1bfbc00000000ULL,
    0xfc41e00000000ULL,
    0x96e3f00000000ULL,
    0x4ab1780000000ULL,
    0x815e4c0000000ULL,
    0x8817b20000000ULL,
    0x790b9f0000000ULL,
    0xd1fa318000000ULL,
    0xb9465cc000000ULL,
    0xea67c9a000000ULL,
    0xe873d0f000000ULL,
    0xb4799b8800000ULL,
    0x4907bfcc00000ULL,
    0xa80d22200000ULL,
    0x59c4af0500000ULL,
    0xc6a629a680000ULL,
    0x5a5020c740000ULL,
    0x98e8e39c20000ULL,
    0xdec9f3f530000ULL,
    0xf85f601388000ULL,
    0x8a93303ca4000ULL,
    0xf4ce181232000ULL,
    0x2d5d7c24d7000ULL,
    0xe11aa2044800ULL,
    0x2a0ae304e2c00ULL,
    0x6799bb60fa00ULL,
    0xb007bfc9f8300ULL,
    0xc800d22f0dd80ULL,
    0x7404af01a7cc0ULL,
    0xca0629a46d060ULL,
    0x450020d0b69f0ULL,
    0x7c80e3a6d54a8ULL,
    0xa2c5f3e19b48cULL,
    0x35256027429eeULL,
    0xc1143038ffd3fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xf800000000000ULL,
    0xdc00000000000ULL,
    0x3200000000000ULL,
    0x5300000000000ULL,
    0xd280000000000ULL,
    0x62c0000000000ULL,
    0xe1e0000000000ULL,
    0x8510000000000ULL,
    0xf3a8000000000ULL,
    0x44cc000000000ULL,
    0x7d1e000000000ULL,
    0x8655000000000ULL,
    0xf209800000000ULL,
    0xdaf8400000000ULL,
    0x2647a00000000ULL,
    0x7e21b00000000ULL,
    0xb776680000000ULL,
    0x85fadc0000000ULL,
    0x1ec23a0000000ULL,
    0xe3e7fd0000000ULL,
    0x5e10398000000ULL,
    0x152836c000000ULL,
    0x300fd1a000000ULL,
    0x21faaad000000ULL,
    0xb0c24bb800000ULL,
    0xe2e0e7d400000ULL,
    0xcf921a0600000ULL,
    0x5d6b0d2300000ULL,
    0x6f2871aa80000ULL,
    0x770b1ae7c0000ULL,
    0x957c23a620000ULL,
    0x87063beb70000ULL,
    0x9486203858000ULL,
    0x47c5f03934000ULL,
    0x9e67c81632000ULL,
    0xac526c1bc7000ULL,
    0x2d0bd22955800ULL,
    0xa27c613222400ULL,
    0xdb8223a70fe00ULL,
    0x87433bca68500ULL,
    0xdfa7a0068c080ULL,
    0x7731b021573c0ULL,
    0xeede68045c120ULL,
    0x5e36dc3c3fcb0ULL,
    0x7b5c3a3e517c8ULL,
    0x8872fd01a9644ULL,
    0xad79b992acfaeULL,
    0x7b0076df2615fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0x5400000000000ULL,
    0xf200000000000ULL,
    0xb700000000000ULL,
    0x3e80000000000ULL,
    0x56c0000000000ULL,
    0x11e0000000000ULL,
    0xe650000000000ULL,
    0x86e8000000000ULL,
    0xbb8c000000000ULL,
    0xec1e000000000ULL,
    0x9d95000000000ULL,
    0xd289800000000ULL,
    0xe39a400000000ULL,
    0xc0d1e00000000ULL,
    0xb429f00000000ULL,
    0x906f080000000ULL,
    0x41489c0000000ULL,
    0x3bfc9a0000000ULL,
    0x2ec37d0000000ULL,
    0x65e3638000000ULL,
    0x645326c000000ULL,
    0x99edeba000000ULL,
    0x690dfad000000ULL,
    0x1cde91b800000ULL,
    0xc97277d400000ULL,
    0xbd5b7a1600000ULL,
    0xd338d0100000ULL,
    0x3c7beb8a80000ULL,
    0xaf84fafcc0000ULL,
    0x6d4111bd20000ULL,
    0xaea137fd70000ULL,
    0x8ff51a0f58000ULL,
    0x9f993d15e4000ULL,
    0x9ed2839c62000ULL,
    0xc52ad6d3c7000ULL,
    0xfbeae3ac45800ULL,
    0x260966d8d2400ULL,
    0x165c0ba53fe00ULL,
    0x7db40add10d00ULL,
    0xa3b999a249980ULL,
    0x81e6ebdfadac0ULL,
    0x1e51e02fcfda0ULL,
    0xb2e9f03e09cb0ULL,
    0x598f08054d1c8ULL,
    0xe3189c161ec44ULL,
    0xf7149a2561dc6ULL,
    0x764f7d0991e1fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x8800000000000ULL,
    0xb400000000000ULL,
    0x2200000000000ULL,
    0xdf00000000000ULL,
    0x8580000000000ULL,
    0xc6c0000000000ULL,
    0xe2e0000000000ULL,
    0x51d0000000000ULL,
    0x4d8000000000ULL,
    0x51bc000000000ULL,
    0x1aa000000000ULL,
    0xe11000000000ULL,
    0xe6bc800000000ULL,
    0xc02ac00000000ULL,
    0x2d2600000000ULL,
    0x5f5df00000000ULL,
    0xbe7cc80000000ULL,
    0xabcc040000000ULL,
    0x4b84020000000ULL,
    0xf7c53d0000000ULL,
    0xc662878000000ULL,
    0x6b970b4000000ULL,
    0xd07acfa000000ULL,
    0xcacdcf7000000ULL,
    0xd706adb800000ULL,
    0xf184027c00000ULL,
    0x84c4621a00000ULL,
    0xde5cd1100000ULL,
    0x5c50cfb280000ULL,
    0x761ccf56c0000ULL,
    0x915a2dabe0000ULL,
    0x8f7ec25b70000ULL,
    0x8f4e022958000ULL,
    0x71c43d2e04000ULL,
    0x236607a2ee000ULL,
    0x5d11cb50ab000ULL,
    0xbd3aafb364800ULL,
    0x2fed3f695f400ULL,
    0xa8b4e5a913a00ULL,
    0x1acfc65d91f00ULL,
    0xaf04803aa9380ULL,
    0xad86c0109d8c0ULL,
    0x22c0601d8a8a0ULL,
    0x78e0f01bf5cd0ULL,
    0xb2d24837762a8ULL,
    0x175bc43c4e724ULL,
    0x6a78e2247b336ULL,
    0xb9cf0d2c6d7c9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0xec00000000000ULL,
    0xa00000000000ULL,
    0x8500000000000ULL,
    0x780000000000ULL,
    0xb2c0000000000ULL,
    0xbbe0000000000ULL,
    0x990000000000ULL,
    0x5a38000000000ULL,
    0x345c000000000ULL,
    0xde0e000000000ULL,
    0x7657000000000ULL,
    0x6a5c800000000ULL,
    0x550ec00000000ULL,
    0xc2d2a00000000ULL,
    0x2818700000000ULL,
    0xbaa9980000000ULL,
    0x23e1840000000ULL,
    0xfd91a20000000ULL,
    0xe439c50000000ULL,
    0x2f5bbd8000000ULL,
    0x328e764000000ULL,
    0xa016a5a000000ULL,
    0xebfd325000000ULL,
    0xd73ba7a800000ULL,
    0x9fdb876c00000ULL,
    0x46cd020a00000ULL,
    0xb8b6b51700000ULL,
    0xec4ea5a880000ULL,
    0x51f1326640000ULL,
    0x476da78160000ULL,
    0x7a808766f0000ULL,
    0x9147820598000ULL,
    0xcf23752394000ULL,
    0xcd76859396000ULL,
    0x592c8264c3000ULL,
    0x37241f8f75800ULL,
    0xe970b358a6c00ULL,
    0x9f299810faa00ULL,
    0x4821842431700ULL,
    0xe3f1a218b8080ULL,
    0xc669c52b046c0ULL,
    0x4b03bd9c7fee0ULL,
    0xd48276428ecd0ULL,
    0x9240a5a317268ULL,
    0x97a6327682cc4ULL,
    0x853127b01441eULL,
    0x338e477757fddULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xd800000000000ULL,
    0x3400000000000ULL,
    0xee00000000000ULL,
    0x4d00000000000ULL,
    0x7980000000000ULL,
    0xf8c0000000000ULL,
    0xe760000000000ULL,
    0xe870000000000ULL,
    0x92b8000000000ULL,
    0x665c000000000ULL,
    0x5dee000000000ULL,
    0xd0b7000000000ULL,
    0x15db800000000ULL,
    0xbe2b400000000ULL,
    0x751e00000000ULL,
    0x5aeb100000000ULL,
    0x9231380000000ULL,
    0xcd9a940000000ULL,
    0x260c020000000ULL,
    0x6583610000000ULL,
    0x8ac7668000000ULL,
    0x9865694000000ULL,
    0xcef5dea000000ULL,
    0x5cf8bd5000000ULL,
    0x977e3ca800000ULL,
    0x70bbcc5400000ULL,
    0x815de21e00000ULL,
    0x6f68711900000ULL,
    0xe0f65e8580000ULL,
    0x71fffd5a40000ULL,
    0xdef9dcb3e0000ULL,
    0x407bdc7070000ULL,
    0x8a395a3438000ULL,
    0x5d1ea5036c000ULL,
    0xd1483cbd16000ULL,
    0x2320cc54bb000ULL,
    0x25062278d800ULL,
    0x8f683119e6c00ULL,
    0x90f23e9b43600ULL,
    0x99f8ad6b40100ULL,
    0x22fa84a261180ULL,
    0x427d185431d40ULL,
    0x1d3b802ed8d60ULL,
    0x879b402e9db10ULL,
    0x1d09e03f6eb68ULL,
    0x4507102636794ULL,
    0xf5873829ebf72ULL,
    0x12c19433e4371ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0xc400000000000ULL,
    0x5600000000000ULL,
    0xa500000000000ULL,
    0x1880000000000ULL,
    0x9840000000000ULL,
    0x9e20000000000ULL,
    0x4750000000000ULL,
    0xfc8000000000ULL,
    0xa36c000000000ULL,
    0xab7a000000000ULL,
    0xa113000000000ULL,
    0xc8ed800000000ULL,
    0x4eb8400000000ULL,
    0x9ff1600000000ULL,
    0xd9dc500000000ULL,
    0x6006380000000ULL,
    0xb0041c0000000ULL,
    0x8801ce0000000ULL,
    0x6c06e70000000ULL,
    0x6a02518000000ULL,
    0x370579c000000ULL,
    0xeb81e9a000000ULL,
    0x25c525f000000ULL,
    0x1ee347a800000ULL,
    0x413092e400000ULL,
    0xd6b8ae1600000ULL,
    0xebf6b71500000ULL,
    0x7de69a380000ULL,
    0xa90265cc40000ULL,
    0xc285a78da0000ULL,
    0x274782ce50000ULL,
    0x19a2761c28000ULL,
    0x896bb3a84000ULL,
    0x2628ff945a000ULL,
    0x9dccefa1f000ULL,
    0x5800000524800ULL,
    0x540000390bc00ULL,
    0x8e000000b7e00ULL,
    0x3100000861d00ULL,
    0xb6800020b1b80ULL,
    0xf940000b59c40ULL,
    0xd0a0003c3cb60ULL,
    0x7a10001997030ULL,
    0x89680032f1258ULL,
    0x7c7c0027f86acULL,
    0x3a92001fec89eULL,
    0x452f0010ff30bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x5800000000000ULL,
    0xfc00000000000ULL,
    0x2a00000000000ULL,
    0xfd00000000000ULL,
    0x7280000000000ULL,
    0x9240000000000ULL,
    0xc0e0000000000ULL,
    0xe870000000000ULL,
    0x8168000000000ULL,
    0x4d0c000000000ULL,
    0x343e000000000ULL,
    0x4431000000000ULL,
    0xa28b800000000ULL,
    0x44ff400000000ULL,
    0xae95a00000000ULL,
    0xa619700000000ULL,
    0x5fe4d80000000ULL,
    0x63f36c0000000ULL,
    0x87aace0000000ULL,
    0x40ad310000000ULL,
    0x372fae8000000ULL,
    0x1befd3c000000ULL,
    0xeb48f6a000000ULL,
    0x9adfffd000000ULL,
    0x5741989800000ULL,
    0xde66bee400000ULL,
    0xe8376e0e00000ULL,
    0x9088411b00000ULL,
    0xa5fd76a480000ULL,
    0x2611bff440000ULL,
    0x915fb8a020000ULL,
    0x11808ef4d0000ULL,
    0x33c6163e18000ULL,
    0x3b225d05cc000ULL,
    0x975360a2c2000ULL,
    0x103fe2c6cf000ULL,
    0xd232d821ed800ULL,
    0x958e6c2813400ULL,
    0x1b7f4e2bd7600ULL,
    0x1653713ca0900ULL,
    0x8b98eb490080ULL,
    0x5d75e3c839cc0ULL,
    0xf7ef8eae1cbe0ULL,
    0x3948e3e65a570ULL,
    0xabda0eb4436d8ULL,
    0x87c6a3e38e8b4ULL,
    0x15242ea04d0e2ULL,
    0x8c5093ef03f3bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xb800000000000ULL,
    0x6c00000000000ULL,
    0x2200000000000ULL,
    0x8300000000000ULL,
    0x9380000000000ULL,
    0xdfc0000000000ULL,
    0xbaa0000000000ULL,
    0x5830000000000ULL,
    0x6578000000000ULL,
    0xa75c000000000ULL,
    0x172a000000000ULL,
    0x135000000000ULL,
    0x39fc800000000ULL,
    0x991b400000000ULL,
    0xf949200000000ULL,
    0xa422b00000000ULL,
    0x9572e80000000ULL,
    0x409ed40000000ULL,
    0x1895a0000000ULL,
    0x9b02070000000ULL,
    0x6f86828000000ULL,
    0x15c72b4000000ULL,
    0x5da2eaa000000ULL,
    0x3db7bf7000000ULL,
    0x77be109800000ULL,
    0x2f7c485c00000ULL,
    0x805b5a3600000ULL,
    0x92ab070700000ULL,
    0xa3f0029e80000ULL,
    0x9d96b57c0000ULL,
    0x626f4aa260000ULL,
    0x48d24f7c30000ULL,
    0xbd2fd89ac8000ULL,
    0x96352c5d3c000ULL,
    0x47c682516000ULL,
    0x57dc940f2b000ULL,
    0xeb6efa2cbc800ULL,
    0xdc52f72856400ULL,
    0xb76bca908ba00ULL,
    0x86550f43ec900ULL,
    0xb86cf8972fe80ULL,
    0x87d29c7e3f3c0ULL,
    0x5caa001e67d20ULL,
    0xc2f5000b42c90ULL,
    0xf95c803f50248ULL,
    0x9e2b401779d8cULL,
    0x95b1200ab4ec2ULL,
    0x33beb0078a6e1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0x1c00000000000ULL,
    0xc200000000000ULL,
    0xf00000000000ULL,
    0x3880000000000ULL,
    0x6f40000000000ULL,
    0x6b60000000000ULL,
    0x5cb0000000000ULL,
    0x2308000000000ULL,
    0xe3ec000000000ULL,
    0x6a9e000000000ULL,
    0xe2b3000000000ULL,
    0x9e0b800000000ULL,
    0xc6ac00000000ULL,
    0x595d200000000ULL,
    0xd092b00000000ULL,
    0x78dd080000000ULL,
    0x86556c0000000ULL,
    0xc17be60000000ULL,
    0x6541550000000ULL,
    0xa862cc8000000ULL,
    0xee354ac000000ULL,
    0xcf4a44a000000ULL,
    0x9a09e6d000000ULL,
    0x4a6f028800000ULL,
    0x505cc3f400000ULL,
    0x4113e63200000ULL,
    0x161d553300000ULL,
    0x6bf4cca280000ULL,
    0x6c6a4ae0c0000ULL,
    0x295fc4a4a0000ULL,
    0x89026cf10000ULL,
    0x1cd9a2ac78000ULL,
    0xb054b3fb3c000ULL,
    0x107bce2e3a000ULL,
    0x90c6890707000ULL,
    0xf0a422aa3d800ULL,
    0xd29173f7bbc00ULL,
    0x97db6e1147600ULL,
    0x8ed1f90c5d700ULL,
    0xd6398aa70a280ULL,
    0xfa236fe4aedc0ULL,
    0x2ad52029d8ea0ULL,
    0x3eb035ac470ULL,
    0x1b23082803738ULL,
    0xa7566c3c5a624ULL,
    0xacf86605b668aULL,
    0xf987951cd52cfULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x6800000000000ULL,
    0xc400000000000ULL,
    0xc600000000000ULL,
    0x6b00000000000ULL,
    0xa180000000000ULL,
    0x3240000000000ULL,
    0x5520000000000ULL,
    0xbff0000000000ULL,
    0xa508000000000ULL,
    0x86ac000000000ULL,
    0x9a000000000ULL,
    0x8bf3000000000ULL,
    0x7b0b800000000ULL,
    0xf1aac00000000ULL,
    0x5b1be00000000ULL,
    0xb8b0500000000ULL,
    0xe6a8c80000000ULL,
    0x709aa40000000ULL,
    0x53f4260000000ULL,
    0x470a670000000ULL,
    0x9bac0c8000000ULL,
    0x321fdac000000ULL,
    0xb43544a000000ULL,
    0x1e6cbef000000ULL,
    0xb67902a800000ULL,
    0x8b6049d400000ULL,
    0x8d4263200000ULL,
    0x7fa672500000ULL,
    0x11240ca380000ULL,
    0xb9f3dacfc0000ULL,
    0xee0f44a3e0000ULL,
    0xb72fbed750000ULL,
    0x5ada82a4a8000ULL,
    0x1ad689edac000ULL,
    0x2fdc6266a000ULL,
    0x3fa53727c7000ULL,
    0x7c35449fac800ULL,
    0x8a6cbee36bc00ULL,
    0x3879029446600ULL,
    0xb46049e86d700ULL,
    0x754262c8a980ULL,
    0x9aba671d16640ULL,
    0x23840cb7c42a0ULL,
    0x5f43dad9a70d0ULL,
    0xbfa7448c3d438ULL,
    0xbc33bed4a3e4cULL,
    0xaa68828dfa58aULL,
    0xa87989fc0eb67ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x4800000000000ULL,
    0x4c00000000000ULL,
    0x7a00000000000ULL,
    0xa100000000000ULL,
    0x380000000000ULL,
    0x15c0000000000ULL,
    0x4960000000000ULL,
    0x8970000000000ULL,
    0x57d8000000000ULL,
    0x963c000000000ULL,
    0x500e000000000ULL,
    0x7cf7000000000ULL,
    0xf91c800000000ULL,
    0xadd8c00000000ULL,
    0x7739a00000000ULL,
    0x738cf00000000ULL,
    0x9934380000000ULL,
    0xf878040000000ULL,
    0x68af1a0000000ULL,
    0x5ae0890000000ULL,
    0x44b3568000000ULL,
    0xcabe9a4000000ULL,
    0x914deea000000ULL,
    0xd8d55e5000000ULL,
    0x7e49d4a800000ULL,
    0x6456e76c00000ULL,
    0x2f0b9a3a00000ULL,
    0x5874493f00000ULL,
    0x5c5cf68480000ULL,
    0xeff96a4740000ULL,
    0x936b56ac60000ULL,
    0x1c829a6e30000ULL,
    0xe143eea438000ULL,
    0x54225e636c000ULL,
    0xcf5554bd4e000ULL,
    0x858e277da3000ULL,
    0x22323a372d800ULL,
    0x8af8b90faec00ULL,
    0xc6e8cebbd2a00ULL,
    0x2416e53f4f00ULL,
    0xb2a44c89b2c80ULL,
    0xcf121364c4cc0ULL,
    0xf228b8340a460ULL,
    0x8a0c40f680d0ULL,
    0xe16ba28e5208ULL,
    0x21ac7906dbdf4ULL,
    0xa5676e8c20bd2ULL,
    0x43769e52d175dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x7800000000000ULL,
    0x4c00000000000ULL,
    0xbe00000000000ULL,
    0x2d00000000000ULL,
    0x9580000000000ULL,
    0xcf40000000000ULL,
    0x7aa0000000000ULL,
    0xd970000000000ULL,
    0xf38000000000ULL,
    0xeb1c000000000ULL,
    0x39aa000000000ULL,
    0xb0f1000000000ULL,
    0xa67a800000000ULL,
    0x60bb400000000ULL,
    0x435c200000000ULL,
    0x698d500000000ULL,
    0x5040780000000ULL,
    0xdc21740000000ULL,
    0x3eb69e0000000ULL,
    0xa75ad30000000ULL,
    0xb38a5c8000000ULL,
    0x6745334000000ULL,
    0x1ea2a4a000000ULL,
    0xc372077000000ULL,
    0x58389ab800000ULL,
    0xd99fc44c00000ULL,
    0xbc6c1e2600000ULL,
    0x391931f00000ULL,
    0x726e7ca980000ULL,
    0x9694636440000ULL,
    0xcbe8dcb4e0000ULL,
    0x17d2735b90000ULL,
    0xe44c84ac48000ULL,
    0x17225742cc000ULL,
    0x9a30629e06000ULL,
    0xab98f0604b000ULL,
    0xef6e201a8c800ULL,
    0xdb105034e7400ULL,
    0xd8a8f828da200ULL,
    0x5b77341945900ULL,
    0x243a3e3d6b280ULL,
    0x7f9dc333bdc40ULL,
    0x2d6c84badfb60ULL,
    0x501257414ed90ULL,
    0xdc28629d06318ULL,
    0xc7b4f067cafdcULL,
    0xf0dc2035cd0eaULL,
    0xf7cd503386351ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0x1800000000000ULL,
    0x9400000000000ULL,
    0xe00000000000ULL,
    0x1300000000000ULL,
    0x7180000000000ULL,
    0x340000000000ULL,
    0x31a0000000000ULL,
    0x2010000000000ULL,
    0x2748000000000ULL,
    0x726c000000000ULL,
    0xffbe000000000ULL,
    0x2197000000000ULL,
    0xcc0f800000000ULL,
    0x7fc9400000000ULL,
    0xcdaee00000000ULL,
    0x97db100000000ULL,
    0x46e7c80000000ULL,
    0xefb4240000000ULL,
    0xbc5aee0000000ULL,
    0x38a68b0000000ULL,
    0xc961f8000000ULL,
    0x668f73c000000ULL,
    0xa10857a000000ULL,
    0xf54917d000000ULL,
    0x836dd9a800000ULL,
    0xb73accf400000ULL,
    0xc6556e2e00000ULL,
    0x106fcb1100000ULL,
    0x86b8ff8d80000ULL,
    0xe51463c5c0000ULL,
    0xb1cf9fb9e0000ULL,
    0xbead33f110000ULL,
    0x565f37bb88000ULL,
    0xcda047f814000ULL,
    0x521571b7da000ULL,
    0x64bb8c54f000ULL,
    0x2e92817d5800ULL,
    0x1c7f34093ac00ULL,
    0xebf52614de600ULL,
    0xb27eaf0e1c900ULL,
    0x28f2f1aa0fd80ULL,
    0xfbfef8db75540ULL,
    0x2fb1c82b8baa0ULL,
    0xdc5f2401669d0ULL,
    0x88a36e23f1718ULL,
    0x8494cb2818afcULL,
    0xea897fb2dcae6ULL,
    0x3b0a23ca6fe73ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x7800000000000ULL,
    0x3400000000000ULL,
    0x2200000000000ULL,
    0x6f00000000000ULL,
    0x4380000000000ULL,
    0xe840000000000ULL,
    0x8f20000000000ULL,
    0x32b0000000000ULL,
    0x3848000000000ULL,
    0x4dec000000000ULL,
    0x799e000000000ULL,
    0x5a33000000000ULL,
    0x8988800000000ULL,
    0xd18d400000000ULL,
    0x558be00000000ULL,
    0xaf88700000000ULL,
    0x48f480000000ULL,
    0x1d0f340000000ULL,
    0xae48c60000000ULL,
    0x40eed70000000ULL,
    0xf51e268000000ULL,
    0x1f2bac000000ULL,
    0x96efeea000000ULL,
    0x581fced000000ULL,
    0x7d72488800000ULL,
    0xb52b29ec00000ULL,
    0x737e461600000ULL,
    0xd6e0972100000ULL,
    0xf2d546be80000ULL,
    0x2a5b8aeac0000ULL,
    0x495546a660000ULL,
    0xb61b8afcd0000ULL,
    0x447546b3a8000ULL,
    0x1bab8ac17c000ULL,
    0x47bd46ab3a000ULL,
    0x8a078ac4c7000ULL,
    0x930346ad3d800ULL,
    0x8d848af39bc00ULL,
    0x6143c6bc77a00ULL,
    0xf9a5caf144900ULL,
    0xc2762684cf580ULL,
    0x3eaebac5718c0ULL,
    0x7739ee95990a0ULL,
    0xbfc0cec7c5810ULL,
    0xf564c8820e9b8ULL,
    0xa9569f511a4cULL,
    0x8d7d260849ca2ULL,
    0xc7e5a72aec467ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xd800000000000ULL,
    0xe400000000000ULL,
    0x3e00000000000ULL,
    0x1300000000000ULL,
    0x8b80000000000ULL,
    0x19c0000000000ULL,
    0x56a0000000000ULL,
    0x2770000000000ULL,
    0x6a88000000000ULL,
    0xceac000000000ULL,
    0x4e9a000000000ULL,
    0x37f3000000000ULL,
    0x14ce800000000ULL,
    0x93ca400000000ULL,
    0x1e4a600000000ULL,
    0xe08b100000000ULL,
    0xebacc80000000ULL,
    0xfa19740000000ULL,
    0x97b76a0000000ULL,
    0x4aa8b70000000ULL,
    0xa09f488000000ULL,
    0xecf78ec000000ULL,
    0x934f00a000000ULL,
    0x880bbaf000000ULL,
    0x81ee8aa800000ULL,
    0x617d5dfc00000ULL,
    0x63ea1e00000ULL,
    0xf051f70500000ULL,
    0xb13ba8b180000ULL,
    0x6606def1c0000ULL,
    0x3701a8bb60000ULL,
    0x9585decbb0000ULL,
    0x3ac72890c8000ULL,
    0x5239ef76c000ULL,
    0xdab748b3ea000ULL,
    0x22b8ef907000ULL,
    0xfadd008d2c800ULL,
    0xaf94baf64bc00ULL,
    0xe09a0abf57200ULL,
    0xcf41ddd15500ULL,
    0x834f0a37cfb80ULL,
    0x600ca71f40940ULL,
    0xbdef00a4200a0ULL,
    0xbb7bbaef91090ULL,
    0x2d668abf58f08ULL,
    0x68d15de83574cULL,
    0x2379ea2bde1aaULL,
    0x2962f73ad8c37ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0x3400000000000ULL,
    0x8600000000000ULL,
    0x5300000000000ULL,
    0xb80000000000ULL,
    0xc7c0000000000ULL,
    0xd8e0000000000ULL,
    0x80f0000000000ULL,
    0x6918000000000ULL,
    0xab7c000000000ULL,
    0x3ace000000000ULL,
    0x6877000000000ULL,
    0x8d5b800000000ULL,
    0xd85e400000000ULL,
    0xc0dce00000000ULL,
    0x2c99500000000ULL,
    0x23b8880000000ULL,
    0x3aab40000000ULL,
    0xde411a0000000ULL,
    0x9a25750000000ULL,
    0x5891b68000000ULL,
    0x1ad9c4000000ULL,
    0x2344bea000000ULL,
    0xaea2685000000ULL,
    0x36d4c4a800000ULL,
    0xb38b0d5400000ULL,
    0x17d49a3600000ULL,
    0xd0c351900000ULL,
    0xb896d6b780000ULL,
    0x71aa8c5dc0000ULL,
    0x6b40d69d60000ULL,
    0xf2a18c7830000ULL,
    0xfcd556bd58000ULL,
    0x5288cc510c000ULL,
    0xc95236989e000ULL,
    0x924fdc6213000ULL,
    0xac365eb54d800ULL,
    0xee7c385d3ec00ULL,
    0x5a4fcca2c3600ULL,
    0xb033f97ae5500ULL,
    0x47f603e8a580ULL,
    0xab4b103b20040ULL,
    0x16b2680291fa0ULL,
    0xaf38e41b08690ULL,
    0x39ec1210e4dc8ULL,
    0xda6681158b644ULL,
    0x95b7cc8ba1692ULL,
    0xbcbff965d1c75ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x1800000000000ULL,
    0x400000000000ULL,
    0x7a00000000000ULL,
    0x1f00000000000ULL,
    0x1480000000000ULL,
    0x8540000000000ULL,
    0xfd20000000000ULL,
    0xfe30000000000ULL,
    0xd998000000000ULL,
    0x21fc000000000ULL,
    0xe2ea000000000ULL,
    0xa541000000000ULL,
    0x2d26800000000ULL,
    0xf630400000000ULL,
    0xb59b600000000ULL,
    0x47ff300000000ULL,
    0x83e9080000000ULL,
    0xd4c28c0000000ULL,
    0xa3e2f60000000ULL,
    0x9ad3270000000ULL,
    0x33cf528000000ULL,
    0x9d71e44000000ULL,
    0x85bd3aa000000ULL,
    0xce4c587000000ULL,
    0xc5b6c4b800000ULL,
    0xf05df36c00000ULL,
    0x4d9b603e00000ULL,
    0xf3ff302d00000ULL,
    0x81e9080080000ULL,
    0xbfc28c1ac0000ULL,
    0xd562f61da0000ULL,
    0x493272790000ULL,
    0xa06f52a4e8000ULL,
    0xf901e4591c000ULL,
    0xb5853aac0e000ULL,
    0x94c05866a3000ULL,
    0x3e4c48994800ULL,
    0x8ad0f372a3400ULL,
    0x5bcfe03b95200ULL,
    0x81727031a3500ULL,
    0xfbbee81e15f80ULL,
    0xab4cfc1fe39c0ULL,
    0xce361e1b753a0ULL,
    0x619edb3ed37f0ULL,
    0x35ffcc818c278ULL,
    0xf0ef7f5b67a6cULL,
    0xa6419628ff2b2ULL,
    0x47a017127b13dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0x6c00000000000ULL,
    0xa200000000000ULL,
    0x4f00000000000ULL,
    0x3580000000000ULL,
    0x3340000000000ULL,
    0x3a0000000000ULL,
    0x610000000000ULL,
    0x9898000000000ULL,
    0x187c000000000ULL,
    0x75ce000000000ULL,
    0x5743000000000ULL,
    0x8da4800000000ULL,
    0x4b12c00000000ULL,
    0x921ca00000000ULL,
    0x96be300000000ULL,
    0x292a080000000ULL,
    0xf0f5c40000000ULL,
    0x5c2ece0000000ULL,
    0xfe76eb0000000ULL,
    0xccef178000000ULL,
    0xb797d94000000ULL,
    0xf5d9bfa000000ULL,
    0xaedc2d5000000ULL,
    0x9d5d798800000ULL,
    0x8f1f024c00000ULL,
    0x243ca02a00000ULL,
    0xe3ee301700000ULL,
    0x1212082280000ULL,
    0x9e99c414c0000ULL,
    0x3978ce07e0000ULL,
    0xdd49eb0410000ULL,
    0x968597a508000ULL,
    0xe4c6195e94000ULL,
    0xdfe19f8bae000ULL,
    0x4030dd5b4b000ULL,
    0x25cbd1b3bd800ULL,
    0xef44f65e4bc00ULL,
    0xc9a066323ca00ULL,
    0xf5111f160b700ULL,
    0xf71dd1b81d880ULL,
    0x803bf6767bfc0ULL,
    0x8deae62a448a0ULL,
    0x8f10df1f464d0ULL,
    0x6c1df1b4ce198ULL,
    0x53bb06649b47cULL,
    0x4faa4e07d4f46ULL,
    0xdc342b2e0e3d9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0xc400000000000ULL,
    0xb600000000000ULL,
    0xc900000000000ULL,
    0x8d80000000000ULL,
    0xed40000000000ULL,
    0x5160000000000ULL,
    0xeeb0000000000ULL,
    0x9e88000000000ULL,
    0xaac000000000ULL,
    0xd97e000000000ULL,
    0x3347000000000ULL,
    0xd461800000000ULL,
    0xa937400000000ULL,
    0x98c8600000000ULL,
    0x314b500000000ULL,
    0x5a0bd80000000ULL,
    0x2a68940000000ULL,
    0x4558ce0000000ULL,
    0xdc90cd0000000ULL,
    0x3c59f48000000ULL,
    0xf911dac000000ULL,
    0xfd1a4ca000000ULL,
    0x12721ed000000ULL,
    0x30a95aa800000ULL,
    0xba7a47c400000ULL,
    0x25c0600e00000ULL,
    0x9ba7503500000ULL,
    0xc995d80380000ULL,
    0x83df941740000ULL,
    0x4b514e28e0000ULL,
    0xb8fb8d3810000ULL,
    0x407949ab8000ULL,
    0xd6018af2bc000ULL,
    0xb90614b09a000ULL,
    0x4586cad257000ULL,
    0xb14674ada5800ULL,
    0x23619af1d6c00ULL,
    0x91b3acaae5e00ULL,
    0xda0e0ed137100ULL,
    0x6a6ae286f5f80ULL,
    0x655983ed8ec40ULL,
    0x8c937605488e0ULL,
    0xa45f091095210ULL,
    0x3d14e2955edd8ULL,
    0x4b1e83e650c24ULL,
    0xdb72f633d8b8aULL,
    0xbd284912ed60fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x5800000000000ULL,
    0x3400000000000ULL,
    0x9200000000000ULL,
    0xc500000000000ULL,
    0x1280000000000ULL,
    0x36c0000000000ULL,
    0xf160000000000ULL,
    0xd290000000000ULL,
    0xbc08000000000ULL,
    0x2fac000000000ULL,
    0x4e5e000000000ULL,
    0x5dc1000000000ULL,
    0x78e1800000000ULL,
    0xe5d1400000000ULL,
    0xf62fe00000000ULL,
    0x811bf00000000ULL,
    0x53e5180000000ULL,
    0x8c572c0000000ULL,
    0x516c8a0000000ULL,
    0x933ce90000000ULL,
    0xc955c98000000ULL,
    0x3eccfc000000ULL,
    0x45ff31a000000ULL,
    0xa83013d000000ULL,
    0x897ea39800000ULL,
    0xcdf7d6f400000ULL,
    0x1599e00e00000ULL,
    0x226f03900000ULL,
    0x82b2980a80000ULL,
    0x5bbb6c1840000ULL,
    0x1694ea3620000ULL,
    0xa60b5931f0000ULL,
    0x66a8b19c78000ULL,
    0x6adc53c854000ULL,
    0x6406c3b7fa000ULL,
    0x6a0066e2ff000ULL,
    0x81049818cd800ULL,
    0x48866c1c3f400ULL,
    0x9fc36a0cade00ULL,
    0x45e7190cef100ULL,
    0xb350d182a4180ULL,
    0x9aebe3f202ec0ULL,
    0xd97bbbb76ade0ULL,
    0x35f0fad6d7670ULL,
    0x519d6a1dd19c8ULL,
    0x5826192a880d4ULL,
    0x2bb1518fad16eULL,
    0xef3aa3d16ec2bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x5800000000000ULL,
    0x1c00000000000ULL,
    0x9600000000000ULL,
    0xcd00000000000ULL,
    0x380000000000ULL,
    0xccc0000000000ULL,
    0x31e0000000000ULL,
    0x2950000000000ULL,
    0xcab8000000000ULL,
    0xfb9c000000000ULL,
    0x7d2e000000000ULL,
    0x7c1000000000ULL,
    0x1762800000000ULL,
    0x4212400000000ULL,
    0x4d9da00000000ULL,
    0xe02a900000000ULL,
    0x4c44380000000ULL,
    0x6fa7c40000000ULL,
    0xa1f0920000000ULL,
    0x23cab50000000ULL,
    0x7210838000000ULL,
    0xb598c04000000ULL,
    0xec291ba000000ULL,
    0x7245945000000ULL,
    0xe6a5b19800000ULL,
    0x2874e56c00000ULL,
    0xb40ba00e00000ULL,
    0x8d77902b00000ULL,
    0x5388b80b80000ULL,
    0xdbb4840840000ULL,
    0x916fb23360000ULL,
    0x7862650770000ULL,
    0x1e911ba918000ULL,
    0x25d9944a24000ULL,
    0xb58bb19e36000ULL,
    0xeb5e55783000ULL,
    0x6ee920399d800ULL,
    0xd2a5d01943400ULL,
    0xba7518063d200ULL,
    0x130e141493d00ULL,
    0x25f38a3a74480ULL,
    0x9c9a11d3e740ULL,
    0x391789a0cec60ULL,
    0xd31e217317c50ULL,
    0x676fb20492428ULL,
    0x562650275f84ULL,
    0xa5111ba13f0a6ULL,
    0x5199476ce6c5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xd800000000000ULL,
    0x9400000000000ULL,
    0xbe00000000000ULL,
    0x2900000000000ULL,
    0x9980000000000ULL,
    0x2940000000000ULL,
    0x86a0000000000ULL,
    0xd9d0000000000ULL,
    0x1198000000000ULL,
    0x827c000000000ULL,
    0x8f0a000000000ULL,
    0x1441000000000ULL,
    0x6124800000000ULL,
    0xb994400000000ULL,
    0x3ebbe00000000ULL,
    0xaaeb900000000ULL,
    0x8c32180000000ULL,
    0xf1ed740000000ULL,
    0xd6b24e0000000ULL,
    0xb628f30000000ULL,
    0xf9d6d18000000ULL,
    0x819ca74000000ULL,
    0xa7f29a000000ULL,
    0xf30a437000000ULL,
    0xe6477fb800000ULL,
    0x6223c47c00000ULL,
    0xb711e00e00000ULL,
    0xa77a901900000ULL,
    0x9c8e980180000ULL,
    0xfa05342340000ULL,
    0xbf03ae3fe0000ULL,
    0x9c82632510000ULL,
    0xaac049a848000ULL,
    0xe0e5937664000ULL,
    0x7bf687a3ae000ULL,
    0xc689206cbb000ULL,
    0x1503b63e0c800ULL,
    0xcb82172ffb400ULL,
    0x7a448791eda00ULL,
    0x60242062ea500ULL,
    0xc15360da5580ULL,
    0xdfb571e8fb40ULL,
    0x6349e7a50b5e0ULL,
    0x7166f074b4a90ULL,
    0xeeb24e08c6ed8ULL,
    0xd228f3126f85cULL,
    0xffd6d1ba9ba1eULL,
    0xc9ca7613d481ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x3000000000000ULL,
    0xa800000000000ULL,
    0xc400000000000ULL,
    0x7200000000000ULL,
    0xb300000000000ULL,
    0x3680000000000ULL,
    0x9cc0000000000ULL,
    0xdfa0000000000ULL,
    0x1af0000000000ULL,
    0x5be8000000000ULL,
    0xda8c000000000ULL,
    0xc4da000000000ULL,
    0xd747000000000ULL,
    0x88e2800000000ULL,
    0x5211c00000000ULL,
    0xa9fae00000000ULL,
    0x4374f00000000ULL,
    0x2fae280000000ULL,
    0x3ceebc0000000ULL,
    0xc60b4a0000000ULL,
    0x2719ef0000000ULL,
    0xb8626c8000000ULL,
    0x67d0a5c000000ULL,
    0x97dea4a000000ULL,
    0xb1c6e9f000000ULL,
    0x2c21c68800000ULL,
    0xd1b6bafc00000ULL,
    0xad0a602200000ULL,
    0x1d9e300d00000ULL,
    0x2a4481e80000ULL,
    0xe1708c11c0000ULL,
    0x64af023b20000ULL,
    0xf669633af0000ULL,
    0x74cd6ea6c8000ULL,
    0x55b9c6cd4c000ULL,
    0x9113ca0cb2000ULL,
    0x577f2f272f000ULL,
    0x8bb20cb824800ULL,
    0x1a0995c9cec00ULL,
    0xf9186c8ff4200ULL,
    0x1d67a5d6f6b00ULL,
    0xa05424adf0980ULL,
    0x269b29e1488c0ULL,
    0x2126ac0d920ULL,
    0x7b54ae152e30ULL,
    0xfc0ec8197fe08ULL,
    0x1d4c315ce8cULL,
    0xc4e7e211abc3aULL,
    0x1416931f5afdbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x5800000000000ULL,
    0x4c00000000000ULL,
    0x5600000000000ULL,
    0x8900000000000ULL,
    0x4a80000000000ULL,
    0xa9c0000000000ULL,
    0xc0e0000000000ULL,
    0xf50000000000ULL,
    0x2b28000000000ULL,
    0xcc0c000000000ULL,
    0x86be000000000ULL,
    0x5e41000000000ULL,
    0xc9a7800000000ULL,
    0x5177400000000ULL,
    0x631a200000000ULL,
    0x4936100000000ULL,
    0x9bbe980000000ULL,
    0x4ec7ac0000000ULL,
    0x5f63d20000000ULL,
    0x214dd0000000ULL,
    0xd70bd38000000ULL,
    0x473b54c000000ULL,
    0xce076ba000000ULL,
    0x6506e8d000000ULL,
    0xac84219800000ULL,
    0x28c499c400000ULL,
    0x9e63a00e00000ULL,
    0xbc90501700000ULL,
    0x34cb383b80000ULL,
    0xdafc2540000ULL,
    0x4ed16a1120000ULL,
    0xfb68611b30000ULL,
    0x80af19b718000ULL,
    0x2b4e65cbe4000ULL,
    0x4d9aca0852000ULL,
    0x52f4310aa7000ULL,
    0x205a21a35d800ULL,
    0x901599f347400ULL,
    0x5c0c202f0c600ULL,
    0x7ebb1033ce900ULL,
    0xa247182680e80ULL,
    0x97a1ec27c0140ULL,
    0xcc767208e0ce0ULL,
    0x33998d2cd1bf0ULL,
    0x3ff16b8c49da8ULL,
    0x98dbe8e0ed584ULL,
    0xa2d5a1bc9fa66ULL,
    0x1d6ed9cb89097ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x7800000000000ULL,
    0xc00000000000ULL,
    0x8a00000000000ULL,
    0x5100000000000ULL,
    0x4380000000000ULL,
    0x27c0000000000ULL,
    0xaa60000000000ULL,
    0xd490000000000ULL,
    0xc698000000000ULL,
    0xa93c000000000ULL,
    0x92ca000000000ULL,
    0xb547000000000ULL,
    0xe21800000000ULL,
    0xf934400000000ULL,
    0xe869200000000ULL,
    0xc333f00000000ULL,
    0xf168b80000000ULL,
    0x4b5c40000000ULL,
    0x78ad2a0000000ULL,
    0x1d4b10000000ULL,
    0x38be298000000ULL,
    0x780b534000000ULL,
    0xea7b1a000000ULL,
    0xf071677000000ULL,
    0x2448238800000ULL,
    0xef07126400000ULL,
    0xf482a01600000ULL,
    0x4f40b02700000ULL,
    0xb720181480000ULL,
    0x6eb2740340000ULL,
    0x19acb22c60000ULL,
    0xca52851eb0000ULL,
    0x3b7bbb98f8000ULL,
    0x2c6a266894000ULL,
    0xd34b22bde000ULL,
    0xae6e85269b000ULL,
    0xb831bb978c800ULL,
    0xb3ed26663ac00ULL,
    0x18f532289ce00ULL,
    0xf90ac50a53d00ULL,
    0xf5209b97a0a80ULL,
    0x7bb2d663902c0ULL,
    0x542f8a2b28060ULL,
    0x129401355c9d0ULL,
    0x7d9e31b8732c8ULL,
    0x4bb92744f0a14ULL,
    0xde8b03b318eeeULL,
    0x4ce3e24fe5899ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x4800000000000ULL,
    0xf400000000000ULL,
    0x6200000000000ULL,
    0x3500000000000ULL,
    0x9f80000000000ULL,
    0x640000000000ULL,
    0xcde0000000000ULL,
    0x5110000000000ULL,
    0x1338000000000ULL,
    0x469c000000000ULL,
    0x196a000000000ULL,
    0x23c7000000000ULL,
    0x82a3800000000ULL,
    0x3577c00000000ULL,
    0x36d600000000ULL,
    0x7ac2f00000000ULL,
    0x6323180000000ULL,
    0x8033440000000ULL,
    0x4e0de60000000ULL,
    0x7314e10000000ULL,
    0x463d758000000ULL,
    0x691e6e4000000ULL,
    0x472b0da000000ULL,
    0x5223da7000000ULL,
    0x45b7f38800000ULL,
    0x714f7f7400000ULL,
    0xef76e01a00000ULL,
    0xfa69300900000ULL,
    0x8b44782280000ULL,
    0x9e66b43440000ULL,
    0x39557e2d60000ULL,
    0x25dc6531d0000ULL,
    0x630ff38c18000ULL,
    0x30937f5894000ULL,
    0x667ce03056000ULL,
    0xabfe303473000ULL,
    0xbff819ec800ULL,
    0x4edd740912c00ULL,
    0x9f8a1e303c600ULL,
    0x5e5595290b300ULL,
    0x175d6bbea9380ULL,
    0xe6ccfb565c8c0ULL,
    0x7d35e63f5b7e0ULL,
    0xe588e102719d0ULL,
    0xb7577590e8e78ULL,
    0xaed96e723d554ULL,
    0xef888dac0ae2eULL,
    0xa6541a7a2972dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0xf800000000000ULL,
    0xac00000000000ULL,
    0x1200000000000ULL,
    0x6500000000000ULL,
    0x2380000000000ULL,
    0xe5c0000000000ULL,
    0x29e0000000000ULL,
    0xc4b0000000000ULL,
    0xb5f8000000000ULL,
    0x1fdc000000000ULL,
    0x598a000000000ULL,
    0x2447000000000ULL,
    0x3126800000000ULL,
    0x72d7400000000ULL,
    0xce8f200000000ULL,
    0xb2c6300000000ULL,
    0x1f60d80000000ULL,
    0x5af7a40000000ULL,
    0x65d89a0000000ULL,
    0x188cbd0000000ULL,
    0xe1c7248000000ULL,
    0xbfe1554000000ULL,
    0xf7b0dca000000ULL,
    0x57cc17000000ULL,
    0xfa9a9e8800000ULL,
    0x6d2cd85c00000ULL,
    0x7851a03600000ULL,
    0xcacd703700000ULL,
    0xf3e5f81280000ULL,
    0x35b6941e40000ULL,
    0x887ec23260000ULL,
    0x3d1c590b30000ULL,
    0xcee89e8618000ULL,
    0x8ab7d8511c000ULL,
    0x5afd200fe6000ULL,
    0xe55d303fab000ULL,
    0xc5cc581964800ULL,
    0xd967e413cac00ULL,
    0x41f13a0955200ULL,
    0x915dcd3953100ULL,
    0xbc8dcad08e80ULL,
    0xf660c160559c0ULL,
    0x1b709e90d25e0ULL,
    0xbd1bd86fc9b50ULL,
    0xeef202f75d88ULL,
    0x2ab63011034a4ULL,
    0xeaf8d81da1d9aULL,
    0x1d5ba43f91931ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0x1400000000000ULL,
    0x3a00000000000ULL,
    0x6d00000000000ULL,
    0xff80000000000ULL,
    0xec40000000000ULL,
    0x6c60000000000ULL,
    0xe5d0000000000ULL,
    0xd818000000000ULL,
    0xcd3c000000000ULL,
    0x6f0a000000000ULL,
    0xa483000000000ULL,
    0xa0c7800000000ULL,
    0x60a2400000000ULL,
    0x8b73e00000000ULL,
    0x146ab00000000ULL,
    0x33d6180000000ULL,
    0x5b1b440000000ULL,
    0x45bef60000000ULL,
    0x1c97d0000000ULL,
    0xa323ef8000000ULL,
    0xd135ad4000000ULL,
    0xb0c17a000000ULL,
    0x1687597000000ULL,
    0x91c379a800000ULL,
    0xd927207400000ULL,
    0x5c35803600000ULL,
    0xc48d403100000ULL,
    0x92c6601880000ULL,
    0x91a7f01a40000ULL,
    0x12f0782f60000ULL,
    0xd32cb42550000ULL,
    0x9b368e2a78000ULL,
    0xee09c91924000ULL,
    0xb5076192fe000ULL,
    0x3b8364712b000ULL,
    0x8e46f616dc800ULL,
    0x5657d131bc00ULL,
    0x5851ef8df5600ULL,
    0x4d5aad4647500ULL,
    0x64d997af2f980ULL,
    0xb9a194112fc0ULL,
    0xef7d19a0989e0ULL,
    0x646cd064b4ed0ULL,
    0xbbd7f836a6c38ULL,
    0x71ef419be224ULL,
    0x3bd6e36cac06ULL,
    0x3acf7901cde51ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0xc400000000000ULL,
    0xb600000000000ULL,
    0x6f00000000000ULL,
    0x6b80000000000ULL,
    0x740000000000ULL,
    0xeae0000000000ULL,
    0x610000000000ULL,
    0x4e48000000000ULL,
    0xf4ec000000000ULL,
    0x9f9a000000000ULL,
    0x3281000000000ULL,
    0xc3c2800000000ULL,
    0xe620c00000000ULL,
    0x1bb5e00000000ULL,
    0x2abfd00000000ULL,
    0x78b4e80000000ULL,
    0x933b940000000ULL,
    0x36f71e0000000ULL,
    0x55db10000000ULL,
    0x10241d8000000ULL,
    0xd4b2b5c000000ULL,
    0xd13f15a000000ULL,
    0x7f7f1f000000ULL,
    0x35de63b800000ULL,
    0x42e114cc00000ULL,
    0x9210801a00000ULL,
    0xe04dc03d00000ULL,
    0x27ed601780000ULL,
    0xe1e1016c0000ULL,
    0x28c3881560000ULL,
    0x9ba4842b10000ULL,
    0xe3f6160dd8000ULL,
    0xd3d9f51b94000ULL,
    0x35e7ebb3ca000ULL,
    0x459490fd6f000ULL,
    0x1d0c162ecc800ULL,
    0xd008f5001f400ULL,
    0xba8d6b9c84200ULL,
    0x654850f413700ULL,
    0xe96bf613fb580ULL,
    0x97da253e87840ULL,
    0x43e1039db0e20ULL,
    0x4a9204f029210ULL,
    0x868b880d9d858ULL,
    0x5f48842fe67d4ULL,
    0x946c1618a1faeULL,
    0xd558f536f1243ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0xf400000000000ULL,
    0x1600000000000ULL,
    0x6f00000000000ULL,
    0x5180000000000ULL,
    0xc240000000000ULL,
    0x42a0000000000ULL,
    0x4590000000000ULL,
    0xa6d8000000000ULL,
    0xa13c000000000ULL,
    0x8b8a000000000ULL,
    0xf083000000000ULL,
    0xe8c1800000000ULL,
    0x3763400000000ULL,
    0xe275e00000000ULL,
    0x9c6bd00000000ULL,
    0x6230680000000ULL,
    0xb7cf440000000ULL,
    0xa122e60000000ULL,
    0xa6d60b0000000ULL,
    0xefbeb8000000ULL,
    0xb1682f4000000ULL,
    0x5ab463a000000ULL,
    0xf30fbb7000000ULL,
    0x54476da800000ULL,
    0xeda5b44400000ULL,
    0x3413800e00000ULL,
    0xb49c403900000ULL,
    0x7b9e602b80000ULL,
    0x3a1b9012c0000ULL,
    0x405c0837e0000ULL,
    0x26fbd41650000ULL,
    0xed6d6e0148000ULL,
    0xd0b19f3a34000ULL,
    0x3608e5ad7e000ULL,
    0x10c22049bb000ULL,
    0xf3608e3bfc800ULL,
    0x9c764f350bc00ULL,
    0x4f6a8db964e00ULL,
    0x49b2646db7100ULL,
    0xf889e8264f980ULL,
    0xcb00041be2e40ULL,
    0x1f850617a1020ULL,
    0x7942db28f1f50ULL,
    0xd52003ba39db8ULL,
    0x70d72b6ecca7cULL,
    0x41fae58652596ULL,
    0x30ed2052b8581ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x8800000000000ULL,
    0xdc00000000000ULL,
    0xea00000000000ULL,
    0x5b00000000000ULL,
    0xe380000000000ULL,
    0x4840000000000ULL,
    0x45a0000000000ULL,
    0x3910000000000ULL,
    0x8b38000000000ULL,
    0xf1dc000000000ULL,
    0xdcee000000000ULL,
    0xc981000000000ULL,
    0xb341800000000ULL,
    0xd627c00000000ULL,
    0x6956200000000ULL,
    0x9a9df00000000ULL,
    0xfecf680000000ULL,
    0xe6d4940000000ULL,
    0x80da860000000ULL,
    0xc468f30000000ULL,
    0x1247678000000ULL,
    0xa6a01c4000000ULL,
    0xfe902fa000000ULL,
    0xed78785000000ULL,
    0x517c419800000ULL,
    0x6b7fdf7400000ULL,
    0x5879803a00000ULL,
    0xf7fbc00300000ULL,
    0xddb8202080000ULL,
    0x1f1cf005c0000ULL,
    0x2f8ee82060000ULL,
    0xb7f35433d0000ULL,
    0xe00ca61308000ULL,
    0x4db5030a94000ULL,
    0x4aa80f8b46000ULL,
    0x3124886023000ULL,
    0xb0d2a9a49d800ULL,
    0xe1dc8b5053400ULL,
    0x14ed262125e00ULL,
    0xf582c30f9eb00ULL,
    0xc9462f9bbe680ULL,
    0x52578724ecc0ULL,
    0x56d3c1b1b6320ULL,
    0x38d91f5b5ba90ULL,
    0xe06e201fb0f38ULL,
    0x3c41f01ed865cULL,
    0x43a168289ca26ULL,
    0x7015942e531edULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xc800000000000ULL,
    0x4c00000000000ULL,
    0xf600000000000ULL,
    0x2500000000000ULL,
    0x80000000000ULL,
    0x4540000000000ULL,
    0x59a0000000000ULL,
    0xb950000000000ULL,
    0x8ba8000000000ULL,
    0x744c000000000ULL,
    0x993a000000000ULL,
    0xf203000000000ULL,
    0x5f04800000000ULL,
    0xf384c00000000ULL,
    0x10c7a00000000ULL,
    0xd463900000000ULL,
    0xe9b4f80000000ULL,
    0x9d5e9c0000000ULL,
    0x63b7060000000ULL,
    0x665ee30000000ULL,
    0x5a302d8000000ULL,
    0x6d9edac000000ULL,
    0xe7d175a000000ULL,
    0x37ecd6f000000ULL,
    0x436c0ba800000ULL,
    0xb42b69ec00000ULL,
    0x5288001a00000ULL,
    0x485c003f00000ULL,
    0x6b32000780000ULL,
    0x4f1f000640000ULL,
    0x8596800fa0000ULL,
    0x39cbc00b10000ULL,
    0x20f9202bc8000ULL,
    0xf0e4503644000ULL,
    0xa6f7d834a2000ULL,
    0xfff9cc2217000ULL,
    0xc3645e12cc800ULL,
    0x9633ef0fc7c00ULL,
    0x5b9bd39365a00ULL,
    0xe2d2a5dd72f00ULL,
    0x476c5e15be480ULL,
    0xce2fef0579140ULL,
    0xa189d3a19df60ULL,
    0x1ddda5f2ae730ULL,
    0xe6f2de34b1178ULL,
    0x1ff82f28d8fbcULL,
    0x9362f3818cf4eULL,
    0x2e36f5f0e7a97ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x9800000000000ULL,
    0x5c00000000000ULL,
    0xae00000000000ULL,
    0xeb00000000000ULL,
    0x6880000000000ULL,
    0x7240000000000ULL,
    0xff20000000000ULL,
    0xb4d0000000000ULL,
    0x3868000000000ULL,
    0x280c000000000ULL,
    0x9c7a000000000ULL,
    0x7405000000000ULL,
    0xaa00800000000ULL,
    0xf902400000000ULL,
    0x1d87a00000000ULL,
    0x69c3500000000ULL,
    0xb9e0c80000000ULL,
    0x97b16c0000000ULL,
    0x989f8e0000000ULL,
    0xcc332f0000000ULL,
    0xfe5fcd8000000ULL,
    0x3f509cc000000ULL,
    0xf6aaa5a000000ULL,
    0x1f6ba0f000000ULL,
    0x668f639800000ULL,
    0x113ea3fc00000ULL,
    0xb9a0003e00000ULL,
    0xed90002b00000ULL,
    0x4fc8003580000ULL,
    0xde9c0021c0000ULL,
    0xc3320017a0000ULL,
    0xb4d9003150000ULL,
    0xa012800e68000ULL,
    0x4e0b400b64000ULL,
    0x437d202bf2000ULL,
    0x9684103ba7000ULL,
    0xf147e816cc800ULL,
    0xb3a07c3a36400ULL,
    0x490e60384a00ULL,
    0x1a4d132942d00ULL,
    0x435a8b8aff480ULL,
    0x10d7dff931040ULL,
    0x3a6a662399fa0ULL,
    0x150a5331dc1f0ULL,
    0x73fdabb73e398ULL,
    0x58c6cfc09044cULL,
    0x90650e31c822eULL,
    0x74746f3e34b7fULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x2800000000000ULL,
    0xac00000000000ULL,
    0x2e00000000000ULL,
    0x3d00000000000ULL,
    0xaf80000000000ULL,
    0xc5c0000000000ULL,
    0x760000000000ULL,
    0x7d90000000000ULL,
    0x5ae8000000000ULL,
    0x434c000000000ULL,
    0x973a000000000ULL,
    0xbc01000000000ULL,
    0xa607800000000ULL,
    0xa100c00000000ULL,
    0xf986a00000000ULL,
    0x7cc5500000000ULL,
    0x2ae5f80000000ULL,
    0xab55e40000000ULL,
    0xcf0aba0000000ULL,
    0x549f690000000ULL,
    0xf77f98000000ULL,
    0x85bc00c000000ULL,
    0x1646a1a000000ULL,
    0x4a1b4f000000ULL,
    0x8bf463b800000ULL,
    0xf6faf9e400000ULL,
    0xc960003200000ULL,
    0xd090000f00000ULL,
    0xbd68001980000ULL,
    0x7a8c003040000ULL,
    0x965a002a60000ULL,
    0x5091001530000ULL,
    0x7d6f8021b8000ULL,
    0x1a8cc02154000ULL,
    0xc65ca03c82000ULL,
    0x7894503ec7000ULL,
    0xd16a7832ac800ULL,
    0x34892420b7400ULL,
    0xfb5e1a1af5600ULL,
    0xd717391793700ULL,
    0x14af81b22ec80ULL,
    0x33e824d871240ULL,
    0x86cd3bbd59360ULL,
    0x8dfb4deda41d0ULL,
    0x57e0c23f1a788ULL,
    0xa4d64d04c3f74ULL,
    0x3ac9e3a8a62eeULL,
    0x2bfb39e19da37ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x2800000000000ULL,
    0x400000000000ULL,
    0x3600000000000ULL,
    0xe700000000000ULL,
    0xc080000000000ULL,
    0x4040000000000ULL,
    0x9da0000000000ULL,
    0xa8d0000000000ULL,
    0x3878000000000ULL,
    0xf09c000000000ULL,
    0xf46e000000000ULL,
    0x1305000000000ULL,
    0x9e82800000000ULL,
    0xc341c00000000ULL,
    0x1b22e00000000ULL,
    0x2790d00000000ULL,
    0x615ea80000000ULL,
    0x2e0c8c0000000ULL,
    0xb6b53e0000000ULL,
    0x8bc9eb0000000ULL,
    0x12d7298000000ULL,
    0xbd78a84000000ULL,
    0xd91d61a000000ULL,
    0xcfadf45000000ULL,
    0x146277b800000ULL,
    0x3eb0534c00000ULL,
    0xdfce003200000ULL,
    0x1cd5000d00000ULL,
    0x67a801f80000ULL,
    0x39dc024c0000ULL,
    0x5aece037a0000ULL,
    0x9845d028d0000ULL,
    0xf1a4281ac8000ULL,
    0xfad14c03e4000ULL,
    0x9979de1956000ULL,
    0xff1c3b0603000ULL,
    0x70ab01946d800ULL,
    0xf8e5e452d2400ULL,
    0x4cf2bfb4a5200ULL,
    0x9368cf5836300ULL,
    0x9385f62473980ULL,
    0xbec17731f5040ULL,
    0x2ee4dfab3fb20ULL,
    0x9bf0df6fb7670ULL,
    0x1bed3e14b25d8ULL,
    0xa7c5eb08556dcULL,
    0x3d612994ef2feULL,
    0x6531a873ff73dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xd800000000000ULL,
    0x4c00000000000ULL,
    0xae00000000000ULL,
    0xfd00000000000ULL,
    0x4180000000000ULL,
    0xab40000000000ULL,
    0xd9e0000000000ULL,
    0xf8f0000000000ULL,
    0x52d8000000000ULL,
    0x4f7c000000000ULL,
    0x54ea000000000ULL,
    0x4d03000000000ULL,
    0x4987800000000ULL,
    0x1f47400000000ULL,
    0x2be4200000000ULL,
    0x73f1d00000000ULL,
    0xa258f80000000ULL,
    0xbbc940000000ULL,
    0xdb4abe0000000ULL,
    0x2d92990000000ULL,
    0x48ebd28000000ULL,
    0xdb00de4000000ULL,
    0xc8850aa000000ULL,
    0x38c29a7000000ULL,
    0xe9a2cca800000ULL,
    0x994d77400000ULL,
    0xa2ea001e00000ULL,
    0x3c03002700000ULL,
    0x8607801580000ULL,
    0x5907402840000ULL,
    0x6b84200de0000ULL,
    0x6c41d00690000ULL,
    0x8760f80d98000ULL,
    0x4130942bac000ULL,
    0x9cf8be32ae000ULL,
    0x84ad990733000ULL,
    0x8c6652b7f4800ULL,
    0x71b49e7ea3c00ULL,
    0xf83eaab3ed600ULL,
    0x1b080a774e100ULL,
    0x34f414a5a2680ULL,
    0x3cda937c6d640ULL,
    0x927fc62e0ed60ULL,
    0x56a4d3ec2810ULL,
    0x3e42cc813d998ULL,
    0xdc64d77d37f74ULL,
    0x49b200219e652ULL,
    0x843f0009dbb4dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xa800000000000ULL,
    0x4400000000000ULL,
    0x1600000000000ULL,
    0x5100000000000ULL,
    0x5f80000000000ULL,
    0xa040000000000ULL,
    0xe3a0000000000ULL,
    0xbbd0000000000ULL,
    0xe298000000000ULL,
    0x9efc000000000ULL,
    0x16ca000000000ULL,
    0xe425000000000ULL,
    0xd794800000000ULL,
    0x33fc00000000ULL,
    0x7a2ee00000000ULL,
    0x16d5d00000000ULL,
    0xd719180000000ULL,
    0xc5be440000000ULL,
    0xb1efb60000000ULL,
    0xb32830000000ULL,
    0x2a69d18000000ULL,
    0x7d742f4000000ULL,
    0x78caa9a000000ULL,
    0x29207b7000000ULL,
    0x9212e79800000ULL,
    0x10796c5c00000ULL,
    0x290d983200000ULL,
    0x47c1840d00000ULL,
    0x3fe1561780000ULL,
    0xf277532940000ULL,
    0x4248c9b860000ULL,
    0x88e66b6d50000ULL,
    0x74f71f8a88000ULL,
    0xa98bf862a4000ULL,
    0xc505b60fae000ULL,
    0x9187833d5b000ULL,
    0xfd455197fc800ULL,
    0x5e27ef4d42400ULL,
    0xf496498771a00ULL,
    0x4bbcab6b39b00ULL,
    0xcced7fa4bc180ULL,
    0xe6b1e85c23c40ULL,
    0x7d2a4e03a0de0ULL,
    0x3850173ff1150ULL,
    0x78d9ff90f8138ULL,
    0x9e5e28561d45cULL,
    0x59cae36535feULL,
    0x79c72f187a1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x7800000000000ULL,
    0x7c00000000000ULL,
    0x5200000000000ULL,
    0xd500000000000ULL,
    0x1b80000000000ULL,
    0xe8c0000000000ULL,
    0xf420000000000ULL,
    0x310000000000ULL,
    0xbca8000000000ULL,
    0xcbcc000000000ULL,
    0x5c3a000000000ULL,
    0x23a1000000000ULL,
    0x51d7800000000ULL,
    0xf98a400000000ULL,
    0xad5ea00000000ULL,
    0xf750b00000000ULL,
    0xaecad80000000ULL,
    0xafb9640000000ULL,
    0xaf62720000000ULL,
    0xdbf51b0000000ULL,
    0x5980f8000000ULL,
    0xa37324c000000ULL,
    0x9ddbf7a000000ULL,
    0x3b10b0f000000ULL,
    0x60adfdb800000ULL,
    0xa9cc7fe400000ULL,
    0xa13d582600000ULL,
    0x3c23243b00000ULL,
    0x9714d21280000ULL,
    0x8aa9ab3e40000ULL,
    0x60c8d78c60000ULL,
    0xb8bb40d070000ULL,
    0x79e6058f68000ULL,
    0x4b3ebffb4000ULL,
    0x3379522cc2000ULL,
    0xfd42eb2c07000ULL,
    0x6be1f7832c800ULL,
    0x71b1b0f1df400ULL,
    0x18fa7d86d1600ULL,
    0x3d863fde39d00ULL,
    0x9bc3f8024cc80ULL,
    0x5ca394202efc0ULL,
    0x23560a2978fe0ULL,
    0x38cccf19ace50ULL,
    0x34b8a5b39f828ULL,
    0x63e35bf8b033cULL,
    0x85b38a0b48a96ULL,
    0x7efb8f30a496fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x3800000000000ULL,
    0x6c00000000000ULL,
    0x8200000000000ULL,
    0xdf00000000000ULL,
    0x3880000000000ULL,
    0x13c0000000000ULL,
    0xea60000000000ULL,
    0x38f0000000000ULL,
    0xaa28000000000ULL,
    0x630c000000000ULL,
    0x565e000000000ULL,
    0xd0e1000000000ULL,
    0xb432800000000ULL,
    0x98c9400000000ULL,
    0xb839200000000ULL,
    0x2e10b00000000ULL,
    0xe71c880000000ULL,
    0xca47ac0000000ULL,
    0xf423e20000000ULL,
    0x62d3a90000000ULL,
    0x75fefa8000000ULL,
    0x174edc000000ULL,
    0x4469d2a000000ULL,
    0x12ab1d000000ULL,
    0xe58f989800000ULL,
    0x309e04ec00000ULL,
    0x7906081600000ULL,
    0x7182ec3700000ULL,
    0xfa44c21680000ULL,
    0x6c221910c0000ULL,
    0x1ed0f2a0a0000ULL,
    0x3ffa01e830000ULL,
    0x8a7310b438000ULL,
    0x92e9a8ed0c000ULL,
    0x4fedea0632000ULL,
    0xe86d453e0f000ULL,
    0x232c38b385800ULL,
    0x2a8bf4c5c7400ULL,
    0xc01da01c91200ULL,
    0x3ec4f03609100ULL,
    0x75e1280035680ULL,
    0x9fb35c16be3c0ULL,
    0x218aca007dca0ULL,
    0x569cf50eeb970ULL,
    0x9002309a53928ULL,
    0x880518c09e2d4ULL,
    0xb40762088de42ULL,
    0x1e07e912737a3ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xb800000000000ULL,
    0xdc00000000000ULL,
    0x6600000000000ULL,
    0xa700000000000ULL,
    0xf780000000000ULL,
    0x640000000000ULL,
    0xc620000000000ULL,
    0xd770000000000ULL,
    0x18f8000000000ULL,
    0x731c000000000ULL,
    0xc20a000000000ULL,
    0x8ca5000000000ULL,
    0xb7b2800000000ULL,
    0xa019400000000ULL,
    0x878ca00000000ULL,
    0xcfe2100000000ULL,
    0xeb11e80000000ULL,
    0xffad340000000ULL,
    0xd390f60000000ULL,
    0x3a6c9f0000000ULL,
    0x81f6688000000ULL,
    0xfe3bc54000000ULL,
    0x1cf9a0a000000ULL,
    0x791da17000000ULL,
    0x5b0c1e9800000ULL,
    0x70221a6c00000ULL,
    0x9871683600000ULL,
    0x1b7d741b00000ULL,
    0x75cd60f80000ULL,
    0xa12ecf02c0000ULL,
    0xb153a08a60000ULL,
    0x9f88a15f50000ULL,
    0x43e69eb5e8000ULL,
    0xa5175a6d8c000ULL,
    0x4cafc81256000ULL,
    0x261664048b000ULL,
    0x212dbe1134800ULL,
    0x7153bb2132400ULL,
    0x7f8f76acd0200ULL,
    0x33e66e4928100ULL,
    0x1d153e38ed080ULL,
    0x90affb2087940ULL,
    0x40115682a3f20ULL,
    0x862d3e4c59990ULL,
    0x86d07616756f8ULL,
    0x79ccdf2c93eecULL,
    0xf5c248b9615daULL,
    0xca659544d0db9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0xd800000000000ULL,
    0x3c00000000000ULL,
    0xbe00000000000ULL,
    0xc100000000000ULL,
    0xa880000000000ULL,
    0x67c0000000000ULL,
    0x79e0000000000ULL,
    0xc0d0000000000ULL,
    0x1e38000000000ULL,
    0x49c000000000ULL,
    0x8c6e000000000ULL,
    0x6e65000000000ULL,
    0xee92800000000ULL,
    0xb89e400000000ULL,
    0x7268a00000000ULL,
    0xcf64d00000000ULL,
    0x3613180000000ULL,
    0x75e3c0000000ULL,
    0x3789320000000ULL,
    0xb1b4950000000ULL,
    0xe92be78000000ULL,
    0xab42404000000ULL,
    0xdc22dfa000000ULL,
    0xa636ec5000000ULL,
    0xc76855b800000ULL,
    0xde4954c00000ULL,
    0xaad7980e00000ULL,
    0xe5397c0300000ULL,
    0x9f1d120180000ULL,
    0xe42b052340000ULL,
    0x25c2dfb3a0000ULL,
    0x26e6ec66b0000ULL,
    0xb95055bbd8000ULL,
    0x7978954984000ULL,
    0xfeb9980bd6000ULL,
    0xb75c7c24e3000ULL,
    0xcf8f9214ed800ULL,
    0x9db545398a400ULL,
    0xff2a7f86b1200ULL,
    0x8e423c4bd9d00ULL,
    0xf6a34da185080ULL,
    0xbef6a97fd7e40ULL,
    0xd708aa26e24e0ULL,
    0x274e939ed2b0ULL,
    0xaaca75ab0aa28ULL,
    0x5892057c709e4ULL,
    0xcd9a202739f36ULL,
    0x90ea9032947a5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xf000000000000ULL,
    0x9800000000000ULL,
    0x1c00000000000ULL,
    0xba00000000000ULL,
    0x2900000000000ULL,
    0xd880000000000ULL,
    0x89c0000000000ULL,
    0x3220000000000ULL,
    0xc610000000000ULL,
    0x7c18000000000ULL,
    0xeb7c000000000ULL,
    0xe88e000000000ULL,
    0xfa1000000000ULL,
    0xe552800000000ULL,
    0x7afc400000000ULL,
    0x9cfa00000000ULL,
    0xcc47f00000000ULL,
    0xfd61780000000ULL,
    0xe2f2dc0000000ULL,
    0xaba8820000000ULL,
    0xad333b0000000ULL,
    0xba8cdf8000000ULL,
    0xe2a3d94000000ULL,
    0x7bd087a000000ULL,
    0x903ab55000000ULL,
    0x526edd9800000ULL,
    0x4710a26c00000ULL,
    0x809df83e00000ULL,
    0xd4bf9c2b00000ULL,
    0x21ada23080000ULL,
    0xbc348b3140000ULL,
    0x6e0c07bea0000ULL,
    0xb967f56f30000ULL,
    0xe4f3fdb428000ULL,
    0x28ab1249b4000ULL,
    0x34b3203dae000ULL,
    0x6fcab0052b000ULL,
    0xff445837ad800ULL,
    0x1ce56c33b3400ULL,
    0xdbb05a14f1200ULL,
    0x944b172ac8500ULL,
    0xff01a599a4c80ULL,
    0x93837e4676ac0ULL,
    0x8447fa1ed7160ULL,
    0x2960e716e65b0ULL,
    0xccf6ddbc4f968ULL,
    0x4caca257ba85cULL,
    0x62b3f81834c96ULL,
    0x64ce9c056fcf9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x6800000000000ULL,
    0x3c00000000000ULL,
    0x6a00000000000ULL,
    0x4500000000000ULL,
    0xd880000000000ULL,
    0xec0000000000ULL,
    0x8ba0000000000ULL,
    0xd1f0000000000ULL,
    0xc2b8000000000ULL,
    0x755c000000000ULL,
    0x24ce000000000ULL,
    0x84a1000000000ULL,
    0x5c70800000000ULL,
    0x2cf9c00000000ULL,
    0x143b200000000ULL,
    0x409ff00000000ULL,
    0x84e9980000000ULL,
    0x1c133c0000000ULL,
    0xee2f5a0000000ULL,
    0xe9b6cf0000000ULL,
    0xccdcd48000000ULL,
    0x6c8c1f4000000ULL,
    0x42c6eca000000ULL,
    0x29a5135000000ULL,
    0xf8f50e9800000ULL,
    0xa83e107c00000ULL,
    0x6a9f980200000ULL,
    0x21ee3c3300000ULL,
    0x7491da3880000ULL,
    0x88ee0f3240000ULL,
    0x5e1774b6a0000ULL,
    0x772a2f5df0000ULL,
    0xeb3454ae48000ULL,
    0xef19df768c000ULL,
    0x3abccbc66000ULL,
    0xf7f7e3569b000ULL,
    0x25ba169975800ULL,
    0x84d9ec7b03400ULL,
    0x408d623601a00ULL,
    0x10c3c31501300ULL,
    0x88a1369581680ULL,
    0x1e761c4bc1cc0ULL,
    0xb5fcfa07608e0ULL,
    0x16bcff2591430ULL,
    0x63586c9959788ULL,
    0xebcdd37a9580cULL,
    0xa926aebed305aULL,
    0xe2342072f93b1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0x1400000000000ULL,
    0xe600000000000ULL,
    0xc900000000000ULL,
    0x7080000000000ULL,
    0x1b40000000000ULL,
    0xede0000000000ULL,
    0xf390000000000ULL,
    0xc3c8000000000ULL,
    0x83ec000000000ULL,
    0xdfde000000000ULL,
    0x1ae1000000000ULL,
    0x2e14800000000ULL,
    0x160b400000000ULL,
    0x9848e00000000ULL,
    0xef2c500000000ULL,
    0xecf8d80000000ULL,
    0x15527c0000000ULL,
    0x6bec560000000ULL,
    0x3dbd10000000ULL,
    0xb0e5f98000000ULL,
    0x2517f5c000000ULL,
    0xf18941a000000ULL,
    0x2e8e99d000000ULL,
    0x8f0b2f9800000ULL,
    0x20ca64dc00000ULL,
    0xb86ed83600000ULL,
    0xc31f7c1900000ULL,
    0x71c6d62f80000ULL,
    0x5a1913b40000ULL,
    0xdf1998ba0000ULL,
    0x509ce5f190000ULL,
    0x510779bed8000ULL,
    0xf481b5d874000ULL,
    0x9543218faa000ULL,
    0xb8e489c6cf000ULL,
    0xc911978a7d800ULL,
    0xe38e08c3a7400ULL,
    0x318e360300a00ULL,
    0x8e8dc10881900ULL,
    0x5f0941a9c0980ULL,
    0xa8ce99c561840ULL,
    0x146b2f86712e0ULL,
    0x311a64e6695b0ULL,
    0x5ec6d831fd1b8ULL,
    0xbc237c0ae67f4ULL,
    0x6630d63ba1ddeULL,
    0xa63c911990d63ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x8800000000000ULL,
    0x400000000000ULL,
    0x4e00000000000ULL,
    0x8f00000000000ULL,
    0xb480000000000ULL,
    0xa4c0000000000ULL,
    0x8a20000000000ULL,
    0x170000000000ULL,
    0x99d8000000000ULL,
    0x5bc000000000ULL,
    0xc12e000000000ULL,
    0xb927000000000ULL,
    0x37f6800000000ULL,
    0xd818c00000000ULL,
    0x4a1de00000000ULL,
    0xd718100000000ULL,
    0xbe99280000000ULL,
    0x53d92c0000000ULL,
    0x84ba3a0000000ULL,
    0xdaaa330000000ULL,
    0x1965ec8000000ULL,
    0x91147c4000000ULL,
    0x574fa4a000000ULL,
    0x9cb1805000000ULL,
    0xc47f56a800000ULL,
    0xce8d8f7c00000ULL,
    0x8f17280a00000ULL,
    0x804e2c3d00000ULL,
    0x3434ba3980000ULL,
    0x12bef30ec0000ULL,
    0x9ae0c9ca0000ULL,
    0xffe76c4bf0000ULL,
    0xc8d60cbc28000ULL,
    0x24eb6c5aac000ULL,
    0x48800c81de000ULL,
    0x6c06c5403000ULL,
    0xdf208cbeb5800ULL,
    0x4cf3ac527ac00ULL,
    0x8a9dec8240200ULL,
    0xd5d87c47e0f00ULL,
    0x2fb9a49b91580ULL,
    0x902a805a79f40ULL,
    0xbaa7d6a6f4920ULL,
    0xabb24f609a7f0ULL,
    0xbcfc482c51bc8ULL,
    0x4efc2159c3cULL,
    0x74307228c43faULL,
    0x32bfcf0e9318dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0xd800000000000ULL,
    0xe400000000000ULL,
    0xaa00000000000ULL,
    0xe300000000000ULL,
    0x8280000000000ULL,
    0x1dc0000000000ULL,
    0xeb20000000000ULL,
    0xf7d0000000000ULL,
    0xdab8000000000ULL,
    0xb8dc000000000ULL,
    0x54ee000000000ULL,
    0x7927000000000ULL,
    0xd7800000000ULL,
    0x9a3b400000000ULL,
    0xea19600000000ULL,
    0xab4ab00000000ULL,
    0x3e32680000000ULL,
    0xfac9dc0000000ULL,
    0xa974a20000000ULL,
    0x102cc30000000ULL,
    0x6b81b98000000ULL,
    0xac45da4000000ULL,
    0x8c6531a000000ULL,
    0x7531f65000000ULL,
    0xb44c9ba800000ULL,
    0x92b5597400000ULL,
    0x1e0a681e00000ULL,
    0x6bd5dc0d00000ULL,
    0x44baa21480000ULL,
    0x9dbc325c0000ULL,
    0x416e39bc20000ULL,
    0x94629a46b0000ULL,
    0xf1325190e8000ULL,
    0xae4c465244000ULL,
    0x39b17384c6000ULL,
    0xa08bc560b3000ULL,
    0x3811aa075d800ULL,
    0xe698af008ac00ULL,
    0x9f88f3b720600ULL,
    0x497857330100ULL,
    0xda5f4a33a8380ULL,
    0xb9295f04a5bc0ULL,
    0xfa03fb8dd7220ULL,
    0x5b04e954aa0b0ULL,
    0xd681802591ea8ULL,
    0xffc0402dd8ebcULL,
    0x3420e004ec5ceULL,
    0x3b56f000e2f15ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x7800000000000ULL,
    0x1c00000000000ULL,
    0xd600000000000ULL,
    0x1100000000000ULL,
    0x9d80000000000ULL,
    0xc40000000000ULL,
    0xc6e0000000000ULL,
    0x3890000000000ULL,
    0xf368000000000ULL,
    0xb64c000000000ULL,
    0xcf7a000000000ULL,
    0x70e5000000000ULL,
    0x9992800000000ULL,
    0xf6ef400000000ULL,
    0xd60ca00000000ULL,
    0xa79f900000000ULL,
    0x4573d80000000ULL,
    0x2178540000000ULL,
    0x5de7460000000ULL,
    0x4214190000000ULL,
    0xe3ad018000000ULL,
    0xe96930c000000ULL,
    0x5948f9a000000ULL,
    0xb7fdb4f000000ULL,
    0x2322c79800000ULL,
    0xf03769ec00000ULL,
    0x4e1bd80600000ULL,
    0x4b34542b00000ULL,
    0xa49d461380000ULL,
    0x53f1191f40000ULL,
    0x9fbf818ce0000ULL,
    0xfc670c050000ULL,
    0x9fa4598358000ULL,
    0x39f224f974000ULL,
    0x8b91fb202000ULL,
    0x6b433dcca7000ULL,
    0x1a669e38bc800ULL,
    0x41554d148e400ULL,
    0x2dcac7bae0200ULL,
    0xfa3b69d351900ULL,
    0xdf81d80cd9080ULL,
    0x6f41543835740ULL,
    0x6067c619e37e0ULL,
    0x1e52590976210ULL,
    0xcd492187a40f8ULL,
    0xc5fce0d69bf74ULL,
    0x18250195f2e7eULL,
    0xdab530e94f1f7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xe800000000000ULL,
    0xc00000000000ULL,
    0x5e00000000000ULL,
    0xed00000000000ULL,
    0x4e80000000000ULL,
    0xc040000000000ULL,
    0x68e0000000000ULL,
    0xc7f0000000000ULL,
    0x63d8000000000ULL,
    0x58fc000000000ULL,
    0x712a000000000ULL,
    0x21a3000000000ULL,
    0x7394800000000ULL,
    0xf76dc00000000ULL,
    0x8040a00000000ULL,
    0x88e3d00000000ULL,
    0x97f4f80000000ULL,
    0xfbdc6c0000000ULL,
    0xbcf85e0000000ULL,
    0x232d510000000ULL,
    0x92a28a8000000ULL,
    0xd0158e4000000ULL,
    0x79a852a000000ULL,
    0x28e4f27000000ULL,
    0x27f0d4a800000ULL,
    0x33dbdf6c00000ULL,
    0xc0fe581200000ULL,
    0x952cbc2f00000ULL,
    0x73a0262580000ULL,
    0xc090fd2440000ULL,
    0x54e874be20000ULL,
    0xe840f4850000ULL,
    0x2040a03e18000ULL,
    0x38e3d00aac000ULL,
    0x5ff4f8196e000ULL,
    0x87dc6c0bd3000ULL,
    0xaf85e2484800ULL,
    0xc22d512cabc00ULL,
    0x82228a8fd9a00ULL,
    0xfd558e58ccb00ULL,
    0x5fc852931f880ULL,
    0x2f54f2579b440ULL,
    0x2cc8d4ab30360ULL,
    0xacd7df7ce89f0ULL,
    0xd20c582025018ULL,
    0xec73bc2d3bbb4ULL,
    0x711ea617a10b2ULL,
    0x165e3d2e11ecdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xc800000000000ULL,
    0x9c00000000000ULL,
    0x7e00000000000ULL,
    0xf00000000000ULL,
    0x3780000000000ULL,
    0xd640000000000ULL,
    0x6e20000000000ULL,
    0x5e10000000000ULL,
    0xefb8000000000ULL,
    0x475c000000000ULL,
    0xbaae000000000ULL,
    0x6565000000000ULL,
    0x9eb6800000000ULL,
    0x536ac00000000ULL,
    0x6101200000000ULL,
    0x4083500000000ULL,
    0xd5c4680000000ULL,
    0xaa65b40000000ULL,
    0x8931360000000ULL,
    0x15289f0000000ULL,
    0x3720c38000000ULL,
    0x4a94844000000ULL,
    0xd87d0ba000000ULL,
    0x9c3da05000000ULL,
    0xb1ff5b800000ULL,
    0x91891b6c00000ULL,
    0x11f3480a00000ULL,
    0x29cfe40b00000ULL,
    0x8d5de3580000ULL,
    0x745eeb18c0000ULL,
    0x232855a060000ULL,
    0xe4208b5150000ULL,
    0x63160029e8000ULL,
    0xf1390031d4000ULL,
    0xd98803586000ULL,
    0x1f4fc00fc3000ULL,
    0x6e17a03ccd800ULL,
    0x7b990380a400ULL,
    0x4b5d482d68200ULL,
    0xfcaae41f94f00ULL,
    0x3e635e1ca6680ULL,
    0x4b342b2833240ULL,
    0xf429759114f20ULL,
    0x37a3db79e7090ULL,
    0xff526814931f8ULL,
    0x821cb43e249ecULL,
    0xdd09b62c5ee76ULL,
    0x82375f182f2bdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x1800000000000ULL,
    0x8c00000000000ULL,
    0x5200000000000ULL,
    0xd700000000000ULL,
    0x3180000000000ULL,
    0xdfc0000000000ULL,
    0x8020000000000ULL,
    0x5cb0000000000ULL,
    0x9318000000000ULL,
    0xa27c000000000ULL,
    0x44ae000000000ULL,
    0xeae1000000000ULL,
    0x9215800000000ULL,
    0x45ef400000000ULL,
    0xb880200000000ULL,
    0x2b44300000000ULL,
    0x5967480000000ULL,
    0x22d5240000000ULL,
    0x88490a0000000ULL,
    0x61f3130000000ULL,
    0x7b788a8000000ULL,
    0x482fc44000000ULL,
    0xafa462a000000ULL,
    0xa670905000000ULL,
    0x5bf809800000ULL,
    0x188dd75400000ULL,
    0x1c51681e00000ULL,
    0x4a0c143700000ULL,
    0x9c15c22a80000ULL,
    0x78e87712c0000ULL,
    0x2204209ee0000ULL,
    0x9f07a76130000ULL,
    0x45838018e8000ULL,
    0x11c240051c000ULL,
    0x1d23a030ae000ULL,
    0x3636700c43000ULL,
    0x2f5ce813e5800ULL,
    0x2a9f54260b400ULL,
    0xa9bbe205e8e00ULL,
    0xfa8d47369c900ULL,
    0x2356e897eee80ULL,
    0xff8dc356e29c0ULL,
    0x85d2aa1cf4ba0ULL,
    0xf1c9631433950ULL,
    0xca326286ad988ULL,
    0x355d90768781cULL,
    0x899c009dbe322ULL,
    0x563f974b7a3cdULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x7c00000000000ULL,
    0x3200000000000ULL,
    0xf700000000000ULL,
    0x3880000000000ULL,
    0xd240000000000ULL,
    0x31a0000000000ULL,
    0x7210000000000ULL,
    0xd9c8000000000ULL,
    0xc9ac000000000ULL,
    0x32da000000000ULL,
    0x3de7000000000ULL,
    0x4eb3800000000ULL,
    0xb85dc00000000ULL,
    0xd8a7200000000ULL,
    0xe794300000000ULL,
    0x808c80000000ULL,
    0x38cb6c0000000ULL,
    0x202aae0000000ULL,
    0xd11a3b0000000ULL,
    0x3b829f8000000ULL,
    0x64c0abc000000ULL,
    0x86e4f7a000000ULL,
    0x9c3537f000000ULL,
    0x491a318800000ULL,
    0x8f8190f400000ULL,
    0x8ac7e80a00000ULL,
    0x53e35c2b00000ULL,
    0x5bb0662680000ULL,
    0xdfda5726c0000ULL,
    0x5e61b180a0000ULL,
    0x3e7050d650000ULL,
    0xc0bac82ec8000ULL,
    0x99906c25d4000ULL,
    0x150b2e1c3a000ULL,
    0x234cfb1e7f000ULL,
    0x94ec3fb4ac800ULL,
    0xfffe5bfe96c00ULL,
    0xcbb09f85e8600ULL,
    0x17dbabdec5b00ULL,
    0x826577af13a80ULL,
    0x1c73f7d81bbc0ULL,
    0x3fbc91a1ef1a0ULL,
    0xdd1360ee44a70ULL,
    0xf549800cd3518ULL,
    0xe5eac0273a59cULL,
    0xde7ca020febe6ULL,
    0xf475f03e6dbf3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0xa400000000000ULL,
    0x6200000000000ULL,
    0xf300000000000ULL,
    0x7080000000000ULL,
    0xf340000000000ULL,
    0x1e20000000000ULL,
    0xaf0000000000ULL,
    0x2088000000000ULL,
    0xcc6c000000000ULL,
    0xb87e000000000ULL,
    0xf7e5000000000ULL,
    0x893800000000ULL,
    0xb2dd400000000ULL,
    0x4c57600000000ULL,
    0xf538f00000000ULL,
    0xccc6a80000000ULL,
    0x1be78c0000000ULL,
    0x1694460000000ULL,
    0xe7db4f0000000ULL,
    0xfdd4ec8000000ULL,
    0x1df91ac000000ULL,
    0xf520a4a000000ULL,
    0x9e7226d000000ULL,
    0xd1c8aab800000ULL,
    0x714e55c400000ULL,
    0x5a0a481e00000ULL,
    0x802e3c2d00000ULL,
    0xb8db8e2680000ULL,
    0xf3513301c0000ULL,
    0xebbd82a2a0000ULL,
    0xc28499e910000ULL,
    0xd8416e2328000ULL,
    0xaaa18312ac000ULL,
    0xcbb7ca8b8a000ULL,
    0x55aaa5d977000ULL,
    0x121ae027b5800ULL,
    0x9b0b0355e400ULL,
    0xd6aa482499600ULL,
    0xca9e3c0a95100ULL,
    0x96f38e050eb80ULL,
    0x968d3339d02c0ULL,
    0xf56b82bd08920ULL,
    0x57fd99e97c8f0ULL,
    0x2a24ee1183638ULL,
    0xd0f5c31d8a0ccULL,
    0x478d2ab87600aULL,
    0x16ea15d334367ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x7800000000000ULL,
    0x3c00000000000ULL,
    0xd200000000000ULL,
    0x8d00000000000ULL,
    0x2980000000000ULL,
    0x12c0000000000ULL,
    0xd160000000000ULL,
    0xdc30000000000ULL,
    0xa318000000000ULL,
    0x863c000000000ULL,
    0x552e000000000ULL,
    0xd427000000000ULL,
    0x890800000000ULL,
    0x6ec9400000000ULL,
    0x5c55a00000000ULL,
    0x94aeb00000000ULL,
    0x92e6180000000ULL,
    0x1ff2340000000ULL,
    0xd9ff720000000ULL,
    0xcdc9310000000ULL,
    0x4ad3ac8000000ULL,
    0x72ebfa4000000ULL,
    0x70c694a000000ULL,
    0x44653e5000000ULL,
    0x99b4dea800000ULL,
    0x6bdecb4c00000ULL,
    0x9e5b383600000ULL,
    0x4e99c40300000ULL,
    0x3afaca2280000ULL,
    0xfb4eb53040000ULL,
    0xdc9446a0e0000ULL,
    0xe8cebf68f0000ULL,
    0x17576a2ab8000ULL,
    0xd62c0500fc000ULL,
    0xb2a45e9efe000ULL,
    0x96d78b4e33000ULL,
    0xb0ee9837d5800ULL,
    0xd5c7743c8ac00ULL,
    0x59e4d20329a00ULL,
    0x1d70813074900ULL,
    0xd9bd34941ac80ULL,
    0x34ec8e40203c0ULL,
    0x5bc2469150ce0ULL,
    0x56e5bf7d29710ULL,
    0xb1f1ea0f75328ULL,
    0xa6fe45389a6d4ULL,
    0x194f7eb76131aULL,
    0x9977b7931ffdULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x6800000000000ULL,
    0x6c00000000000ULL,
    0x5e00000000000ULL,
    0xc700000000000ULL,
    0xde80000000000ULL,
    0xd240000000000ULL,
    0x77e0000000000ULL,
    0x4850000000000ULL,
    0x4508000000000ULL,
    0x4f2c000000000ULL,
    0x975e000000000ULL,
    0x5f21000000000ULL,
    0x2ff7800000000ULL,
    0x3bbec00000000ULL,
    0xff6a00000000ULL,
    0xcbb8d00000000ULL,
    0xd7f4a80000000ULL,
    0x9fbabc0000000ULL,
    0x8df61e0000000ULL,
    0x6ab8390000000ULL,
    0xca75df8000000ULL,
    0xa17e0dc000000ULL,
    0xb15657a000000ULL,
    0x4c8fa1d000000ULL,
    0xcbebc19800000ULL,
    0xefa34c400000ULL,
    0x5695880200000ULL,
    0x89acac1f00000ULL,
    0xb11c163580000ULL,
    0xc2c6553ec0000ULL,
    0xaaa169ba20000ULL,
    0x3d3188de50000ULL,
    0xb91c161dd8000ULL,
    0xbec6550744000ULL,
    0x7ca169831a000ULL,
    0xc63188e14f000ULL,
    0x519c16136d800ULL,
    0xc786551d17400ULL,
    0x8bc16990e8600ULL,
    0x9b2188ef2cf00ULL,
    0xbdf4162077c80ULL,
    0x12ba551459c40ULL,
    0x2e77698f85620ULL,
    0xc37c88ec3b2b0ULL,
    0x405596099f058ULL,
    0x39099523f5954ULL,
    0x992849b3b25a6ULL,
    0x6c5b98f503623ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0x7c00000000000ULL,
    0x3e00000000000ULL,
    0x5100000000000ULL,
    0x2a80000000000ULL,
    0x840000000000ULL,
    0x6220000000000ULL,
    0xa4f0000000000ULL,
    0xf3b8000000000ULL,
    0x845c000000000ULL,
    0xdd4a000000000ULL,
    0x1ce7000000000ULL,
    0xc094800000000ULL,
    0x7c6f400000000ULL,
    0x7595a00000000ULL,
    0x24e9d00000000ULL,
    0xad1980000000ULL,
    0xb14bac0000000ULL,
    0xae75e0000000ULL,
    0x590670000000ULL,
    0x5cef448000000ULL,
    0xd6d3434000000ULL,
    0x7f4dfca000000ULL,
    0xe3e57f7000000ULL,
    0x53101a8800000ULL,
    0x6aaf244c00000ULL,
    0xe5f0b82600000ULL,
    0xf13d3c1300000ULL,
    0x181b660980000ULL,
    0xb56e1b3740000ULL,
    0x131382ab20000ULL,
    0xaaf884950000ULL,
    0xb5f1663da8000ULL,
    0xb9391b0f6c000ULL,
    0xdc1f0286c6000ULL,
    0xf76cc8763b000ULL,
    0x7c16c61d34800ULL,
    0x712bcb2012c00ULL,
    0x97301a9c19a00ULL,
    0xd35f245175100ULL,
    0x1ac8b81eb2c80ULL,
    0xa0213c3c08640ULL,
    0xbf16610fc2e0ULL,
    0x28391b3b0f4f0ULL,
    0x569f028526d98ULL,
    0xf2cc8580b86cULL,
    0xa636c6176ce0aULL,
    0xa9dbcb12c7e21ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0xa800000000000ULL,
    0xc00000000000ULL,
    0x5e00000000000ULL,
    0x6300000000000ULL,
    0x7580000000000ULL,
    0xebc0000000000ULL,
    0xd860000000000ULL,
    0x9a30000000000ULL,
    0xf698000000000ULL,
    0xd5bc000000000ULL,
    0x5eae000000000ULL,
    0x2e23000000000ULL,
    0x4591800000000ULL,
    0x9acec00000000ULL,
    0xa217a00000000ULL,
    0xac0dd00000000ULL,
    0x1cf0880000000ULL,
    0xd07ec40000000ULL,
    0x7dcbaa0000000ULL,
    0x1592d10000000ULL,
    0x72cdfc8000000ULL,
    0xe115e4000000ULL,
    0x220d54a000000ULL,
    0xd7f38a5000000ULL,
    0xa9fe568800000ULL,
    0xc80f8f6400000ULL,
    0xaef6a82200000ULL,
    0x9d7dd42700000ULL,
    0x134c821580000ULL,
    0x2fd1c52fc0000ULL,
    0x136edea0e0000ULL,
    0x71414b6130000ULL,
    0x5825021cf8000ULL,
    0x6a930514bc000ULL,
    0x114f7ebede000ULL,
    0xfad39b5d13000ULL,
    0xa9ea0a1615800ULL,
    0x6100010b6bc00ULL,
    0xa082f484e9e00ULL,
    0x51425a7494700ULL,
    0xc820deb92a080ULL,
    0x62924b6fc8040ULL,
    0xcd4c821745aa0ULL,
    0xcd1c523e3750ULL,
    0xc6eede920d208ULL,
    0x4a814b62e7714ULL,
    0x28450238ce37aULL,
    0xfca305143b92dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x8800000000000ULL,
    0xd400000000000ULL,
    0xaa00000000000ULL,
    0xdb00000000000ULL,
    0x2480000000000ULL,
    0x20c0000000000ULL,
    0xa760000000000ULL,
    0x36b0000000000ULL,
    0x2538000000000ULL,
    0xa31c000000000ULL,
    0x78e000000000ULL,
    0x4c27000000000ULL,
    0x6f92800000000ULL,
    0xd72cc00000000ULL,
    0x5f77600000000ULL,
    0x61ddb00000000ULL,
    0xcfe8280000000ULL,
    0xe4152c0000000ULL,
    0x969ba0000000ULL,
    0xabd7390000000ULL,
    0x308f188000000ULL,
    0xdea49f4000000ULL,
    0xca54d0a000000ULL,
    0x15cbc35000000ULL,
    0x3e86a2a800000ULL,
    0x33c3a65c00000ULL,
    0x77e3c80a00000ULL,
    0x8c735c2900000ULL,
    0xd15c720080000ULL,
    0x652f650fc0000ULL,
    0xa877ea81a0000ULL,
    0x935c3a4bf0000ULL,
    0x7a28da2a78000ULL,
    0x9ef189235c000ULL,
    0x349bb0a1b6000ULL,
    0x87ca73615b000ULL,
    0xf9808aa375800ULL,
    0x49418a5362c00ULL,
    0x1620f211c9200ULL,
    0x5c94a524c5f00ULL,
    0x37aa0aaefaf80ULL,
    0x6db14a4f859c0ULL,
    0x41b9923e9b720ULL,
    0xa3de150ad5170ULL,
    0x90e8a28493bc8ULL,
    0xf294a651b0e94ULL,
    0x9ea9481e9822aULL,
    0xde339c17cc629ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x8800000000000ULL,
    0xfc00000000000ULL,
    0xe200000000000ULL,
    0x700000000000ULL,
    0x5c80000000000ULL,
    0x8b40000000000ULL,
    0x5da0000000000ULL,
    0xdeb0000000000ULL,
    0x9b68000000000ULL,
    0x400c000000000ULL,
    0xedde000000000ULL,
    0x1763000000000ULL,
    0x67d4800000000ULL,
    0x71bac00000000ULL,
    0x4634200000000ULL,
    0x662eb00000000ULL,
    0xd4a9c80000000ULL,
    0x72e84c0000000ULL,
    0x28c9c60000000ULL,
    0xb3bb330000000ULL,
    0x7133928000000ULL,
    0x12afdbc000000ULL,
    0xd3ecfaa000000ULL,
    0x454ce7d000000ULL,
    0xd7a548800000ULL,
    0x9654e8cc00000ULL,
    0xbd7f682a00000ULL,
    0xfe533c3300000ULL,
    0x917eae3280000ULL,
    0xe4540f22c0000ULL,
    0xe27b3c8da0000ULL,
    0xa6d4d4dc10000ULL,
    0x8c3d4618c8000ULL,
    0xa0f1f304bc000ULL,
    0x854fb28e92000ULL,
    0xed7d6bd97f000ULL,
    0x46533288c5800ULL,
    0x457babcd77c00ULL,
    0x8a5112b4d8a00ULL,
    0x8f7a1bc475700ULL,
    0x1525a8e2fa80ULL,
    0xb9fb97d1eddc0ULL,
    0x71133cb8ba5a0ULL,
    0x5ad8d4dc330f0ULL,
    0x23e3460def0b8ULL,
    0xc092f329cc73cULL,
    0x361b32aaeb672ULL,
    0xeb87abc69be93ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x7800000000000ULL,
    0x5400000000000ULL,
    0x5600000000000ULL,
    0x6f00000000000ULL,
    0x4d80000000000ULL,
    0xac40000000000ULL,
    0x2b20000000000ULL,
    0x1a70000000000ULL,
    0x7068000000000ULL,
    0x2a1c000000000ULL,
    0xfa72000000000ULL,
    0x406d000000000ULL,
    0x3218800000000ULL,
    0x5e76400000000ULL,
    0x6e6c600000000ULL,
    0x918900000000ULL,
    0x45f7880000000ULL,
    0xad2d740000000ULL,
    0x6fbf9e0000000ULL,
    0xf3c3b30000000ULL,
    0xf664fe8000000ULL,
    0x5fd0cec000000ULL,
    0x79db60a000000ULL,
    0x9c137df000000ULL,
    0x97bf9e3800000ULL,
    0x67c3b30c00000ULL,
    0xc064fe8600000ULL,
    0xc0d0cee900000ULL,
    0x4c5b60ab80000ULL,
    0x64537dfec0000ULL,
    0xea9f9e3ee0000ULL,
    0x12b3b33cd0000ULL,
    0xfd8cfe9fa8000ULL,
    0x468ccec48c000ULL,
    0x9d0960bd52000ULL,
    0x3e4e7df2db000ULL,
    0xa8ef1e1c66800ULL,
    0x66d9f3264c400ULL,
    0x69929eadb2a00ULL,
    0xff95ec30ad00ULL,
    0xeae668aa4f380ULL,
    0xcd1549e700b40ULL,
    0xa93ce016807a0ULL,
    0x9c02d00f41390ULL,
    0xda01e82720a58ULL,
    0xfd04e414b161cULL,
    0xfc82962c39e02ULL,
    0xa2c5873c046b7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x7800000000000ULL,
    0x7400000000000ULL,
    0xba00000000000ULL,
    0x500000000000ULL,
    0xb480000000000ULL,
    0xb240000000000ULL,
    0x56e0000000000ULL,
    0xf610000000000ULL,
    0xd48000000000ULL,
    0xdb5c000000000ULL,
    0x8616000000000ULL,
    0x854f000000000ULL,
    0x6759800000000ULL,
    0x2015400000000ULL,
    0x364f600000000ULL,
    0x18dff00000000ULL,
    0x99d0b80000000ULL,
    0x636aec0000000ULL,
    0xbeefc60000000ULL,
    0x862f9d0000000ULL,
    0x158beb8000000ULL,
    0x18fa1cc000000ULL,
    0x53642da000000ULL,
    0xaad581d000000ULL,
    0x5cefc60800000ULL,
    0xa72f9d1400000ULL,
    0xc30bebb600000ULL,
    0xcbba1cd900000ULL,
    0x73042d9080000ULL,
    0x9f8581cc40000ULL,
    0x9c7c60ba0000ULL,
    0x3d239d39d0000ULL,
    0xaa35eb8e98000ULL,
    0xd1f91cc814000ULL,
    0xc9e3ad866a000ULL,
    0x1793c1c713000ULL,
    0xd38f262977800ULL,
    0x5bfa2d2e54c00ULL,
    0xe4e5b3b6cbe00ULL,
    0xf1640e142900ULL,
    0xbfca33922fb80ULL,
    0xca1c00c921ac0ULL,
    0xd774d38290760ULL,
    0xbe9ab0f0b9870ULL,
    0x85350b9f043f8ULL,
    0x587cacc813cc4ULL,
    0x1b25f590f6b36ULL,
    0x59339dea94cc7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x6800000000000ULL,
    0xa400000000000ULL,
    0xe600000000000ULL,
    0xaf00000000000ULL,
    0x9d80000000000ULL,
    0xea40000000000ULL,
    0xf60000000000ULL,
    0xb190000000000ULL,
    0x8b48000000000ULL,
    0xf2dc000000000ULL,
    0x8d92000000000ULL,
    0x6149000000000ULL,
    0xffd8800000000ULL,
    0x2912c00000000ULL,
    0xd18f600000000ULL,
    0x237a900000000ULL,
    0x9ba0880000000ULL,
    0x4b36f40000000ULL,
    0x76fd2e0000000ULL,
    0x85e1a10000000ULL,
    0x5a54038000000ULL,
    0xd46f78c000000ULL,
    0xb9292da000000ULL,
    0xafced9f000000ULL,
    0x821d2e1800000ULL,
    0xb131a12c00000ULL,
    0xa3fc03a200000ULL,
    0xd6378f500000ULL,
    0xd8932d8380000ULL,
    0x29cbd9f2c0000ULL,
    0x9d1fae06e0000ULL,
    0xa4b6613d50000ULL,
    0x9db9e386b8000ULL,
    0x8c0228e3f4000ULL,
    0x620445a28a000ULL,
    0xd9057de213000ULL,
    0x2a85680756800ULL,
    0xbbc7a419b5400ULL,
    0xa722461a2aa00ULL,
    0xf7f30531e2f00ULL,
    0x85cc5891ff80ULL,
    0xe57bdd089140ULL,
    0x9a6a082ddd7a0ULL,
    0x522d341fa6590ULL,
    0xa44ace3b7dc98ULL,
    0xaf59f12b57254ULL,
    0x87d3ebbbb5b36ULL,
    0x1e2f1cef2aa27ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xc800000000000ULL,
    0x1400000000000ULL,
    0xce00000000000ULL,
    0x3500000000000ULL,
    0xd380000000000ULL,
    0xcac0000000000ULL,
    0x5220000000000ULL,
    0x18f0000000000ULL,
    0xe2c8000000000ULL,
    0x623c000000000ULL,
    0xacf6000000000ULL,
    0xbcc9000000000ULL,
    0x4f3a800000000ULL,
    0x1375c00000000ULL,
    0x640fe00000000ULL,
    0xf21c500000000ULL,
    0x2302180000000ULL,
    0xaa80d40000000ULL,
    0xdb409a0000000ULL,
    0xfe0df0000000ULL,
    0xbe53828000000ULL,
    0xc79cbc000000ULL,
    0x309318a000000ULL,
    0xa85914d000000ULL,
    0xa1609a3800000ULL,
    0xf310df3c00000ULL,
    0x3a9b82aa00000ULL,
    0xff45cbe900000ULL,
    0x49e518a180000ULL,
    0xff5014d840000ULL,
    0xa1fa1a3f60000ULL,
    0x7551f03f0000ULL,
    0x3dfc628a68000ULL,
    0xbd559bdd8c000ULL,
    0x76f9008dba000ULL,
    0x93d5c0c6f3000ULL,
    0x7bbe002cef800ULL,
    0xc53500344d400ULL,
    0x46c80329b200ULL,
    0xec8cc03663300ULL,
    0x4e5d602777580ULL,
    0x7065902ea9640ULL,
    0x4693f8052cfe0ULL,
    0x6159841e6a610ULL,
    0x78e602268b808ULL,
    0xbad0cb363b32cULL,
    0x523878bfb39e2ULL,
    0x24f584e60f213ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0x800000000000ULL,
    0x3400000000000ULL,
    0xca00000000000ULL,
    0xd300000000000ULL,
    0xbf80000000000ULL,
    0x5140000000000ULL,
    0xc20000000000ULL,
    0x6a70000000000ULL,
    0x8a68000000000ULL,
    0x341c000000000ULL,
    0x476000000000ULL,
    0xdb6b000000000ULL,
    0x4c9d800000000ULL,
    0x30b6c00000000ULL,
    0x5d0ae00000000ULL,
    0xa14a700000000ULL,
    0x4bee080000000ULL,
    0x3c5d840000000ULL,
    0x44d7060000000ULL,
    0x4fd9e30000000ULL,
    0x9394368000000ULL,
    0x56fd46c000000ULL,
    0x736330a000000ULL,
    0x1754a5d000000ULL,
    0x889f062800000ULL,
    0x82b5e31400000ULL,
    0x820a368a00000ULL,
    0xf8ca46c900000ULL,
    0x3fa8b09080000ULL,
    0xa2f965e9c0000ULL,
    0xd960663560000ULL,
    0xf455533490000ULL,
    0x2f18dea6e8000ULL,
    0xeff6b2c7cc000ULL,
    0x702c3eb112000ULL,
    0x8bbbc2fe5b000ULL,
    0xd941b696e7800ULL,
    0x782786cf4dc00ULL,
    0x8077d094d2e00ULL,
    0x4969d5ed3b700ULL,
    0x839a8e3d77d80ULL,
    0x6135a71e242c0ULL,
    0x1d4a50b15eba0ULL,
    0xf5ef15f5083d0ULL,
    0x5586e2d9ca38ULL,
    0x3853d7011a69cULL,
    0xa91a588547142ULL,
    0xfaf591d13d057ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0xf800000000000ULL,
    0xb400000000000ULL,
    0xa600000000000ULL,
    0x9d00000000000ULL,
    0xfe80000000000ULL,
    0xaf40000000000ULL,
    0x6920000000000ULL,
    0x4510000000000ULL,
    0x71d8000000000ULL,
    0x38cc000000000ULL,
    0x7b12000000000ULL,
    0x8d9000000000ULL,
    0x784f800000000ULL,
    0x6d51400000000ULL,
    0xc17b600000000ULL,
    0xbb1f100000000ULL,
    0x2d28980000000ULL,
    0x8ee47c0000000ULL,
    0x26779e0000000ULL,
    0xf76bab0000000ULL,
    0x95409d8000000ULL,
    0x3e22784000000ULL,
    0x449703a000000ULL,
    0x7319d37000000ULL,
    0xd12f9e1800000ULL,
    0x14e7ab1400000ULL,
    0xc1729d8600000ULL,
    0x7eeb784900000ULL,
    0x4b8083b780000ULL,
    0x3dc4937a40000ULL,
    0xa466fe0e20000ULL,
    0x7f31bb1890000ULL,
    0x1b4d85a3e8000ULL,
    0xad2447394000ULL,
    0xbebe7db1be000ULL,
    0x3c79286537000ULL,
    0x15997b8c92800ULL,
    0x9a6aff5495400ULL,
    0xc3c478033f600ULL,
    0x3d662c1af6100ULL,
    0xafb66607f3380ULL,
    0xf509c704a4940ULL,
    0x77701b91867e0ULL,
    0xdbecef60eb9b0ULL,
    0xe103602e68a18ULL,
    0xa483102ad5e5cULL,
    0xa84298031ef8eULL,
    0x70a17c1766e0dULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x800000000000ULL,
    0x2400000000000ULL,
    0x8a00000000000ULL,
    0x1100000000000ULL,
    0x4080000000000ULL,
    0x4640000000000ULL,
    0x9120000000000ULL,
    0xb810000000000ULL,
    0xf068000000000ULL,
    0x727c000000000ULL,
    0xfb16000000000ULL,
    0xbdeb000000000ULL,
    0xf2be800000000ULL,
    0xe375c00000000ULL,
    0x475a600000000ULL,
    0x8903f00000000ULL,
    0x7c85180000000ULL,
    0xb042ec0000000ULL,
    0x5626c60000000ULL,
    0x4f92230000000ULL,
    0x49aae58000000ULL,
    0x7e9a25c000000ULL,
    0xc5e423a000000ULL,
    0x627406d000000ULL,
    0x1fd8c62800000ULL,
    0xb345230c00000ULL,
    0x3ba2658a00000ULL,
    0xbf54e5d500000ULL,
    0x59c8c38080000ULL,
    0xa02e36dd40000ULL,
    0x3459be1aa0000ULL,
    0xec833f19d0000ULL,
    0xa8413bbe68000ULL,
    0x2a26eaea94000ULL,
    0x99960018d2000ULL,
    0xfeab002fdb000ULL,
    0x811e80387f800ULL,
    0x5825c03a35c00ULL,
    0xe492603402200ULL,
    0x302ff02133b00ULL,
    0x2c5b1817ab380ULL,
    0x9085ec2c468c0ULL,
    0x7e46461c085e0ULL,
    0x9d20e32c25870ULL,
    0x661005ac8ad78ULL,
    0x636b15c65783cULL,
    0x7fbdbb68030eULL,
    0x77d7dada408dbULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x1000000000000ULL,
    0x5800000000000ULL,
    0x7400000000000ULL,
    0x6600000000000ULL,
    0x8f00000000000ULL,
    0x2680000000000ULL,
    0x6c40000000000ULL,
    0xd760000000000ULL,
    0x90d0000000000ULL,
    0xb048000000000ULL,
    0x8e9c000000000ULL,
    0x5dd2000000000ULL,
    0x7c9000000000ULL,
    0xcf5b800000000ULL,
    0xee70400000000ULL,
    0x8f38600000000ULL,
    0xc23100000000ULL,
    0xc533080000000ULL,
    0x92db2c0000000ULL,
    0x2b6460000000ULL,
    0x5b1e510000000ULL,
    0x7d17d18000000ULL,
    0xd1eee2c000000ULL,
    0xf6997a000000ULL,
    0x272cb3f000000ULL,
    0x3e4c461800000ULL,
    0x1d9b513400000ULL,
    0xc15651ae00000ULL,
    0xbe8ba2d900000ULL,
    0xfbb877af80000ULL,
    0xa966e3cec0000ULL,
    0x4bd4ae0820000ULL,
    0xcf2d0cd0000ULL,
    0x25d97f9278000ULL,
    0x8034cfe624000ULL,
    0xe95968364a000ULL,
    0x41713c0953000ULL,
    0x59bece1346800ULL,
    0x28653d1e85400ULL,
    0x3e51f78eda600ULL,
    0x100fa3d84a500ULL,
    0x5bff4e3052f80ULL,
    0x7c007d2cc7d40ULL,
    0x6a001798445a0ULL,
    0x6d05f3dafb4f0ULL,
    0x8787a6169b338ULL,
    0xe9c401012b0f4ULL,
    0x8fa73982629deULL,
    0xc2f69ede4ff8fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x1000000000000ULL,
    0xa800000000000ULL,
    0xac00000000000ULL,
    0x1a00000000000ULL,
    0x4300000000000ULL,
    0x5880000000000ULL,
    0x7e40000000000ULL,
    0xcba0000000000ULL,
    0x4f30000000000ULL,
    0x98a8000000000ULL,
    0xf19c000000000ULL,
    0xd836000000000ULL,
    0xbe2b000000000ULL,
    0x4add800000000ULL,
    0xaa13c00000000ULL,
    0x90db200000000ULL,
    0x8910300000000ULL,
    0xf85d080000000ULL,
    0x4f55fc0000000ULL,
    0x37fe060000000ULL,
    0xb667270000000ULL,
    0xf653188000000ULL,
    0x5c7b1ac000000ULL,
    0x8a51ea000000ULL,
    0x57b03dd000000ULL,
    0xc6e8062800000ULL,
    0x2a3c271400000ULL,
    0x3f0698a200000ULL,
    0x8a84daef00000ULL,
    0xa1403e8c80000ULL,
    0x31270debc0000ULL,
    0x76768e38a0000ULL,
    0xbd8d1b1b50000ULL,
    0x99e83e86c8000ULL,
    0x90bb0de35c000ULL,
    0x26408e1842000ULL,
    0xbfa61b046b000ULL,
    0x6135be8887800ULL,
    0xd5a8cdd8bdc00ULL,
    0xf41bae0a72e00ULL,
    0xbf62b3252300ULL,
    0xa48b6acc3a80ULL,
    0xab8d31f3abfc0ULL,
    0x90eda816260a0ULL,
    0x33d0c06ecb30ULL,
    0xbc85ae2b3a238ULL,
    0xb8412b3c4e86cULL,
    0xaa3369ce0efaULL,
    0x40b5f1fab0f37ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0xfc00000000000ULL,
    0xde00000000000ULL,
    0x2f00000000000ULL,
    0x7480000000000ULL,
    0xc340000000000ULL,
    0x38a0000000000ULL,
    0x68f0000000000ULL,
    0x6448000000000ULL,
    0x76bc000000000ULL,
    0x5ff6000000000ULL,
    0xccc9000000000ULL,
    0xbbfa800000000ULL,
    0x2055400000000ULL,
    0xf4b8200000000ULL,
    0x3ef3900000000ULL,
    0x9f4ec80000000ULL,
    0x983cfc0000000ULL,
    0x21b72e0000000ULL,
    0xa9ef970000000ULL,
    0xb7cd3e8000000ULL,
    0x157e6bc000000ULL,
    0x3e1210a000000ULL,
    0x219dfcd000000ULL,
    0xeac12e3800000ULL,
    0x5666973c00000ULL,
    0x8c97beae00000ULL,
    0x11db2bd300000ULL,
    0xa8e230a180000ULL,
    0xbad26ce4c0000ULL,
    0x80f9e612a0000ULL,
    0xeed36b0710000ULL,
    0x5afa10bd08000ULL,
    0x33d1fcf9ec000ULL,
    0xb77f2e36ba000ULL,
    0x8f13972353000ULL,
    0x1a1b3ebc4f800ULL,
    0xed076bf98d400ULL,
    0x758090bc0b200ULL,
    0x20c4bcfecb100ULL,
    0xc3670e0b6aa80ULL,
    0x95100737fab40ULL,
    0xe71df6a7323a0ULL,
    0xd08797e4fe9b0ULL,
    0xa141beb915598ULL,
    0xa9a22bc9afcacULL,
    0x8370b0bdfc2c2ULL,
    0xb8b2cf232cb3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xe800000000000ULL,
    0x6c00000000000ULL,
    0xee00000000000ULL,
    0x8d00000000000ULL,
    0xf680000000000ULL,
    0xf2c0000000000ULL,
    0x39a0000000000ULL,
    0x4870000000000ULL,
    0x13c8000000000ULL,
    0x7abc000000000ULL,
    0x51f6000000000ULL,
    0xe489000000000ULL,
    0x4259800000000ULL,
    0x9867c00000000ULL,
    0x7255200000000ULL,
    0xbbf9900000000ULL,
    0xf016d80000000ULL,
    0xd19dfc0000000ULL,
    0xf943320000000ULL,
    0x25e1c30000000ULL,
    0xb714128000000ULL,
    0xd81f1fc000000ULL,
    0xc60120a000000ULL,
    0xc107dcd000000ULL,
    0xa884b23800000ULL,
    0x47c3032400000ULL,
    0x4b26b2a200000ULL,
    0x38b44fef00000ULL,
    0x496d58a180000ULL,
    0x494d70e840000ULL,
    0x467df83e20000ULL,
    0x55516c3bf0000ULL,
    0x627a6a11c8000ULL,
    0x2752ff01ac000ULL,
    0x597b808b1a000ULL,
    0x8d08cf4f3000ULL,
    0x373eca36af800ULL,
    0xccb0af287d400ULL,
    0xd36e78af23e00ULL,
    0x1e4de0c597b00ULL,
    0x87faa01c99280ULL,
    0xb617500f54940ULL,
    0x509a78139f7e0ULL,
    0x31c3ac2a545b0ULL,
    0xb220ca011e9a8ULL,
    0x1435af0594bfcULL,
    0x8ca9f8aa7ee02ULL,
    0x616f20c6c4233ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xb800000000000ULL,
    0x1400000000000ULL,
    0x9e00000000000ULL,
    0xd700000000000ULL,
    0x5180000000000ULL,
    0x2540000000000ULL,
    0xe1a0000000000ULL,
    0x95f0000000000ULL,
    0x1e8000000000ULL,
    0x6f1c000000000ULL,
    0xf576000000000ULL,
    0x9629000000000ULL,
    0xf478800000000ULL,
    0x3b64400000000ULL,
    0x1e14600000000ULL,
    0xa038d00000000ULL,
    0x844080000000ULL,
    0x123540000000ULL,
    0x4234f60000000ULL,
    0xfa8bef0000000ULL,
    0x510e568000000ULL,
    0xc549a6c000000ULL,
    0xd66ca0a000000ULL,
    0x595b49d000000ULL,
    0xf352763800000ULL,
    0xad9aaf2c00000ULL,
    0x77b4b6b600000ULL,
    0xb9cc36f900000ULL,
    0x13ac48a980000ULL,
    0x6f388de0c0000ULL,
    0x9dc0e80be0000ULL,
    0x9263c41090000ULL,
    0x28929e1a08000ULL,
    0xcf96b152c000ULL,
    0x28a628b192000ULL,
    0x53755dce7b000ULL,
    0x152a602ec7800ULL,
    0x5bfdd012dd400ULL,
    0x392288344ba00ULL,
    0x1632143b7f500ULL,
    0x48e162c89380ULL,
    0x760e7f1d6c240ULL,
    0x2ccebe9f339a0ULL,
    0xe72a62e3cba10ULL,
    0x26fe3ea6bf9f8ULL,
    0xb1a222cce97ecULL,
    0xfdf45ebe3d902ULL,
    0x3deff2e05bb67ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0x1400000000000ULL,
    0x200000000000ULL,
    0x2700000000000ULL,
    0xcc80000000000ULL,
    0x8c40000000000ULL,
    0x58a0000000000ULL,
    0x33d0000000000ULL,
    0x2a68000000000ULL,
    0x22bc000000000ULL,
    0x2a52000000000ULL,
    0xc5ab000000000ULL,
    0x16df800000000ULL,
    0xb60400000000ULL,
    0x39b7600000000ULL,
    0xd25b300000000ULL,
    0x6b27680000000ULL,
    0x7153c0000000ULL,
    0x8c8fb60000000ULL,
    0x26cd0f0000000ULL,
    0x736ac78000000ULL,
    0x573d6bc000000ULL,
    0x639771a000000ULL,
    0x3ccb64f000000ULL,
    0x806a360800000ULL,
    0x39ba4f0400000ULL,
    0x18d0278e00000ULL,
    0x30ed1bf100000ULL,
    0x2bf8f98680000ULL,
    0x497528fcc0000ULL,
    0x2ebd883ba0000ULL,
    0xfc554c2ed0000ULL,
    0x40a83e2b08000ULL,
    0x2d5f431ee4000ULL,
    0xd3a779ba32000ULL,
    0xf95568c89b000ULL,
    0xbb2ae813ae800ULL,
    0x559e7c2315c00ULL,
    0xc34756022b600ULL,
    0x8267f2856d00ULL,
    0x9192cfb168380ULL,
    0x93cf67f715dc0ULL,
    0x60edafb42a820ULL,
    0x73f857f556430ULL,
    0xfd70c78de9578ULL,
    0xfcba6bdcd4dccULL,
    0x4352f18c0b646ULL,
    0x982c24e7c6c53ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0x9800000000000ULL,
    0xac00000000000ULL,
    0x1200000000000ULL,
    0xb100000000000ULL,
    0xe080000000000ULL,
    0x6fc0000000000ULL,
    0x920000000000ULL,
    0x83b0000000000ULL,
    0xb128000000000ULL,
    0xc19c000000000ULL,
    0x7932000000000ULL,
    0xfbed000000000ULL,
    0x2639800000000ULL,
    0x7e40400000000ULL,
    0x2664a00000000ULL,
    0x5550700000000ULL,
    0x73bf580000000ULL,
    0xc7072c0000000ULL,
    0xb7878a0000000ULL,
    0xe0452f0000000ULL,
    0xf560588000000ULL,
    0x2cd12ec000000ULL,
    0x18fdd2a000000ULL,
    0x87e501f000000ULL,
    0x32160a3800000ULL,
    0x91596f0400000ULL,
    0xbb96f8be00000ULL,
    0x121c5edf00000ULL,
    0x7f730a8280000ULL,
    0x618e6df740000ULL,
    0x996f2004a0000ULL,
    0xc0fd3033b0000ULL,
    0x8be2780468000ULL,
    0x90171c241c000ULL,
    0x685c720c82000ULL,
    0x6f127317cb000ULL,
    0xc3d88a8ea6800ULL,
    0xd5532df92c400ULL,
    0xb3ba000daae00ULL,
    0xa701002a77500ULL,
    0x6783803615480ULL,
    0x7841402e4f0c0ULL,
    0x59672001707e0ULL,
    0x3ed1300109970ULL,
    0xa9f878224cd48ULL,
    0x67661c04fbd0cULL,
    0x5dd7f2288ebcaULL,
    0x987f3322110c7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0xb800000000000ULL,
    0x3400000000000ULL,
    0x7e00000000000ULL,
    0x6d00000000000ULL,
    0x6580000000000ULL,
    0x9440000000000ULL,
    0x67a0000000000ULL,
    0xfb10000000000ULL,
    0xd708000000000ULL,
    0xcf1c000000000ULL,
    0xd892000000000ULL,
    0xda4b000000000ULL,
    0x1339800000000ULL,
    0xcac7400000000ULL,
    0xb767600000000ULL,
    0x8777d00000000ULL,
    0xdcfdd80000000ULL,
    0x3420240000000ULL,
    0xfe53de0000000ULL,
    0x972a070000000ULL,
    0x2f48218000000ULL,
    0x92b92ec000000ULL,
    0x3881ffa000000ULL,
    0x99c429f000000ULL,
    0xdfe25e0800000ULL,
    0xc6b1473400000ULL,
    0x871d418600000ULL,
    0x2495fec900000ULL,
    0x984da79180000ULL,
    0x9c3f4de640000ULL,
    0x3044e00ee0000ULL,
    0x61a7900a50000ULL,
    0x211380688000ULL,
    0xdc8bb412d4000ULL,
    0x8e58e61daa000ULL,
    0xeeb6b31b43000ULL,
    0xcb1b478506800ULL,
    0xce93ddfb05c00ULL,
    0x9b4c581262e00ULL,
    0x2cbb642ff7100ULL,
    0xf586be2d3db80ULL,
    0xec46d71a6f240ULL,
    0xf3a479b7002e0ULL,
    0x95124ae4803d0ULL,
    0x20f4185c1618ULL,
    0x9e9efec7a198cULL,
    0x32d4278c3147eULL,
    0xd0e80dc318ee7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0x1800000000000ULL,
    0x2400000000000ULL,
    0x6a00000000000ULL,
    0xaf00000000000ULL,
    0xc380000000000ULL,
    0x2a40000000000ULL,
    0xa060000000000ULL,
    0xed10000000000ULL,
    0x8b18000000000ULL,
    0x70c000000000ULL,
    0xca96000000000ULL,
    0x2b5f000000000ULL,
    0xd86a800000000ULL,
    0x8c05400000000ULL,
    0xb606e00000000ULL,
    0x3104f00000000ULL,
    0x6e85680000000ULL,
    0x7ac2840000000ULL,
    0x7a53e0000000ULL,
    0xa232790000000ULL,
    0xaae90f8000000ULL,
    0x88c07e4000000ULL,
    0xcca231a000000ULL,
    0xabb1075000000ULL,
    0x5fafbe0800000ULL,
    0xf327390c00000ULL,
    0x2ff7efae00000ULL,
    0xeac88e5500000ULL,
    0x1ab159bc80000ULL,
    0x712c8372c0000ULL,
    0x29e0002260000ULL,
    0x5850002d50000ULL,
    0x50f80036f8000ULL,
    0x945c0024d4000ULL,
    0x93ee00165e000ULL,
    0x4a43001847000ULL,
    0x3064803a03800ULL,
    0x2516402a44c00ULL,
    0xc71a601c47200ULL,
    0x510eb01802900ULL,
    0x2b91081145180ULL,
    0x2ddf3415c76c0ULL,
    0x9ea8b62643620ULL,
    0xc5a70d19e55d0ULL,
    0xd135d99ef7358ULL,
    0xf76ac35feaaf4ULL,
    0xf826010c8146ULL,
    0x3c42b035fceddULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x7000000000000ULL,
    0xe800000000000ULL,
    0xac00000000000ULL,
    0x1200000000000ULL,
    0x500000000000ULL,
    0x6880000000000ULL,
    0x1340000000000ULL,
    0x73e0000000000ULL,
    0xb7d0000000000ULL,
    0xfe88000000000ULL,
    0xda5c000000000ULL,
    0x1756000000000ULL,
    0x71cf000000000ULL,
    0xf3ba800000000ULL,
    0x7982c00000000ULL,
    0x1dc5200000000ULL,
    0x93a3700000000ULL,
    0x67b4780000000ULL,
    0xcd994c0000000ULL,
    0x85f22e0000000ULL,
    0x60f9bd0000000ULL,
    0x4a644d8000000ULL,
    0xa1014c000000ULL,
    0x1d2863a000000ULL,
    0x55eaa9d000000ULL,
    0x76c8ae0800000ULL,
    0xe63b7d0c00000ULL,
    0x96416d9200000ULL,
    0x5b6364f700000ULL,
    0x4941b9e80000ULL,
    0xfd6fe5fa40000ULL,
    0x858c8016a0000ULL,
    0x45ddc00990000ULL,
    0xb997a014d8000ULL,
    0xe1edb01e24000ULL,
    0x60cf581dda000ULL,
    0xfd393c3123000ULL,
    0x99c2d624d7800ULL,
    0xcda1312f74c00ULL,
    0xa0b7c3bfe3200ULL,
    0x381b19eb37700ULL,
    0xba31f638c5e80ULL,
    0xbe5d4131eb6c0ULL,
    0x99513b8cebd60ULL,
    0x2ecc95e16a4f0ULL,
    0x4238f81baab58ULL,
    0xf8448c124b4c4ULL,
    0xd4658e24fa70eULL,
    0x2d140d3d72067ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0xc400000000000ULL,
    0xa00000000000ULL,
    0x8f00000000000ULL,
    0x4180000000000ULL,
    0xc0c0000000000ULL,
    0xd2a0000000000ULL,
    0x39d0000000000ULL,
    0x43f8000000000ULL,
    0xc32c000000000ULL,
    0xc852000000000ULL,
    0x6b3b000000000ULL,
    0x858e800000000ULL,
    0x8381c00000000ULL,
    0xebc1600000000ULL,
    0xc924f00000000ULL,
    0xe17e80000000ULL,
    0x9cdc140000000ULL,
    0xf43e7a0000000ULL,
    0xdc0ebb0000000ULL,
    0x5f462f8000000ULL,
    0x8f635b4000000ULL,
    0x39f255a000000ULL,
    0x4d6ae07000000ULL,
    0x5bb0fa0800000ULL,
    0xe08f7b0c00000ULL,
    0xdd074f8a00000ULL,
    0xb287ab4d00000ULL,
    0x6745bd8080000ULL,
    0xa366f462c0000ULL,
    0xa7f6801ae0000ULL,
    0xb06dc01ff0000ULL,
    0xd933602ec8000ULL,
    0xafcff03344000ULL,
    0x4a616804d6000ULL,
    0x9771d42c7f000ULL,
    0x9c2d1a300a800ULL,
    0x31d14b39f5c00ULL,
    0x47ff47b7bf200ULL,
    0x692e8f446ab00ULL,
    0xb7554f9fc5f80ULL,
    0xa2bcab5c164c0ULL,
    0x814b3d939f220ULL,
    0x5b2734737bf90ULL,
    0x5d17e007fd448ULL,
    0xcb5930171a1bcULL,
    0xdfc8839cded6ULL,
    0x862fe428b2341ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0xd800000000000ULL,
    0x6c00000000000ULL,
    0x3600000000000ULL,
    0x9f00000000000ULL,
    0x5080000000000ULL,
    0xfdc0000000000ULL,
    0xc660000000000ULL,
    0xfd10000000000ULL,
    0x3df8000000000ULL,
    0x106c000000000ULL,
    0x9056000000000ULL,
    0x2659000000000ULL,
    0xfb1a800000000ULL,
    0x38bd400000000ULL,
    0xbfc8a00000000ULL,
    0x2122500000000ULL,
    0x27b7080000000ULL,
    0xad0d5c0000000ULL,
    0x4fc0fa0000000ULL,
    0x9b63350000000ULL,
    0x896778000000ULL,
    0x89bba54000000ULL,
    0x3c4c0da000000ULL,
    0x6a65d05000000ULL,
    0xeb12da3800000ULL,
    0x72fc253c00000ULL,
    0x28e9dfae00000ULL,
    0xd994a97700000ULL,
    0xba3bff9b80000ULL,
    0xaf0bb96cc0000ULL,
    0xcac457a1a0000ULL,
    0x2e4b564b0000ULL,
    0x8ad3a5abb8000ULL,
    0x3a9adc6734000ULL,
    0x96fd283d66000ULL,
    0x5aee4c3b9f000ULL,
    0xa491521ccb800ULL,
    0x9fb939223d400ULL,
    0x734d0584eb200ULL,
    0x52e18c7b4d100ULL,
    0xa2d0a01672880ULL,
    0xee9e503588c40ULL,
    0x7cf90830cc6a0ULL,
    0x2be85c02b2830ULL,
    0x7147a0828dc8ULL,
    0x4fb752a7d8a4ULL,
    0xd7ea57a60aab6ULL,
    0x3911b5689ca01ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0xb800000000000ULL,
    0xe400000000000ULL,
    0xee00000000000ULL,
    0x4300000000000ULL,
    0xf080000000000ULL,
    0x9040000000000ULL,
    0x30a0000000000ULL,
    0xcf70000000000ULL,
    0xa668000000000ULL,
    0xf99c000000000ULL,
    0x1fb2000000000ULL,
    0x7e8f000000000ULL,
    0xc24c800000000ULL,
    0xa8ae400000000ULL,
    0xda79e00000000ULL,
    0x30e2d00000000ULL,
    0xb7d6280000000ULL,
    0x151ffc0000000ULL,
    0x65f3160000000ULL,
    0xf32c8f0000000ULL,
    0x36bf328000000ULL,
    0x3206a6c000000ULL,
    0x8900a4a000000ULL,
    0x7d8469f000000ULL,
    0x13c6762800000ULL,
    0xe8601f3c00000ULL,
    0x8b90faa600000ULL,
    0xb7bb8af500000ULL,
    0xd3859ab580000ULL,
    0xb0c71ad9c0000ULL,
    0x8e252afa0000ULL,
    0x93d636c8d0000ULL,
    0xdb1d6cb608000ULL,
    0x16f645d2cc000ULL,
    0xbbafc81832000ULL,
    0x42fd2c187b000ULL,
    0xeca53e0fd6800ULL,
    0x573732014c00ULL,
    0x2b6c2492f7a00ULL,
    0x7a1a29cb84500ULL,
    0xc77796041f580ULL,
    0x3a6ecf2f78240ULL,
    0xd39cd28d15720ULL,
    0x82b776f3778b0ULL,
    0x75080ca844ef8ULL,
    0x468ad5c8bf8ccULL,
    0xe648002f28d12ULL,
    0x66ac001b5d6e7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x9800000000000ULL,
    0x9c00000000000ULL,
    0xae00000000000ULL,
    0x3d00000000000ULL,
    0xa480000000000ULL,
    0xa440000000000ULL,
    0xdda0000000000ULL,
    0x1f0000000000ULL,
    0x3558000000000ULL,
    0xc02c000000000ULL,
    0xa932000000000ULL,
    0x98bb000000000ULL,
    0x1679800000000ULL,
    0x829e400000000ULL,
    0xc24f200000000ULL,
    0xdea6100000000ULL,
    0xc073d80000000ULL,
    0x9996c0000000ULL,
    0x7c9d60000000ULL,
    0xd065770000000ULL,
    0x9490be8000000ULL,
    0xe948bb4000000ULL,
    0xb20e8a000000ULL,
    0x56b38c7000000ULL,
    0xc17f760800000ULL,
    0x491d273400000ULL,
    0xe18c468e00000ULL,
    0xb6c7c76300000ULL,
    0x18e2e6ad80000ULL,
    0x76d3975840000ULL,
    0xbdec1e8aa0000ULL,
    0x1457eb5250000ULL,
    0x50af90b048000ULL,
    0x7775b07a6c000ULL,
    0x7219d83fbe000ULL,
    0xac0e6c2c37000ULL,
    0xbc02562bba800ULL,
    0x7e00371375c00ULL,
    0x5061ebc1ae00ULL,
    0x2880eb47a5500ULL,
    0x924410ad93a80ULL,
    0x7ca0f07129140ULL,
    0xb777831ddbe0ULL,
    0xac1a3c3e861f0ULL,
    0xb90cae2a02638ULL,
    0xc844b238156cULL,
    0x704590bb41722ULL,
    0xd7a2b06720a71ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x6400000000000ULL,
    0x1e00000000000ULL,
    0xf100000000000ULL,
    0x4180000000000ULL,
    0xcbc0000000000ULL,
    0xdae0000000000ULL,
    0xbd50000000000ULL,
    0xfb38000000000ULL,
    0xefec000000000ULL,
    0x9b16000000000ULL,
    0xc019000000000ULL,
    0x2f5f800000000ULL,
    0x19ffc00000000ULL,
    0x688f600000000ULL,
    0xdf86500000000ULL,
    0x7ac7b80000000ULL,
    0xfb64640000000ULL,
    0x6696420000000ULL,
    0x9dd910000000ULL,
    0x36b9698000000ULL,
    0x7e2f194000000ULL,
    0xdef0aba000000ULL,
    0xf5cd485000000ULL,
    0x1226a23800000ULL,
    0x9d34011400000ULL,
    0x7b29b19200000ULL,
    0xe9712d5d00000ULL,
    0xcb0f51ad80000ULL,
    0x8741bd7540000ULL,
    0x32a009afe0000ULL,
    0x7f70495ab0000ULL,
    0x6e089394d8000ULL,
    0x80c6ec50a4000ULL,
    0x5467803df6000ULL,
    0xb613c025ef000ULL,
    0x9399601deb800ULL,
    0xf9f50184d400ULL,
    0x7d9838064a600ULL,
    0x869ba40e9db00ULL,
    0x1019220e62380ULL,
    0x275bc12b81640ULL,
    0xdfed1ba41b60ULL,
    0x4e8b7d5f61cd0ULL,
    0x6286e982701e8ULL,
    0x4140d94e78cd4ULL,
    0xdfa7cba5f472eULL,
    0xcf718739f861ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xd000000000000ULL,
    0xd800000000000ULL,
    0x3400000000000ULL,
    0x9a00000000000ULL,
    0x5700000000000ULL,
    0xda80000000000ULL,
    0xe40000000000ULL,
    0x4da0000000000ULL,
    0x4070000000000ULL,
    0xf698000000000ULL,
    0x6c000000000ULL,
    0x96b2000000000ULL,
    0x767d000000000ULL,
    0xfd39800000000ULL,
    0x711a400000000ULL,
    0xfbae600000000ULL,
    0x8bd2f00000000ULL,
    0xe7eb280000000ULL,
    0xbd728c0000000ULL,
    0xb31d960000000ULL,
    0x58ac170000000ULL,
    0xab55f28000000ULL,
    0x6eaeb94000000ULL,
    0xf251e4a000000ULL,
    0xc928ee7000000ULL,
    0x7992763800000ULL,
    0xbc48a70400000ULL,
    0xd402baae00000ULL,
    0x8a03c57900000ULL,
    0xef06dab080000ULL,
    0x3e803568c0000ULL,
    0x4c4672a820000ULL,
    0x2ea5f95250000ULL,
    0xf404abc8000ULL,
    0xafdd5e5a8c000ULL,
    0x974ebe064e000ULL,
    0xd8829b0cc7000ULL,
    0xcd4264a62a800ULL,
    0xbd23ae72d4400ULL,
    0x1137960faae00ULL,
    0x8dbd173a95d00ULL,
    0xe05e72a74bf80ULL,
    0xe089f94ce52c0ULL,
    0xbbe604acd3d20ULL,
    0x49d05e7ea05f0ULL,
    0x44ef3e3f914f8ULL,
    0x9df4db396976cULL,
    0x3a5e04871d612ULL,
    0x178c5e5426c79ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xb800000000000ULL,
    0xcc00000000000ULL,
    0xea00000000000ULL,
    0x2500000000000ULL,
    0xa580000000000ULL,
    0xba40000000000ULL,
    0x4ba0000000000ULL,
    0xc6b0000000000ULL,
    0x9348000000000ULL,
    0xaa7c000000000ULL,
    0x2e72000000000ULL,
    0xee2f000000000ULL,
    0xf2ab800000000ULL,
    0x7f69400000000ULL,
    0x4788a00000000ULL,
    0x551cb00000000ULL,
    0x27a6380000000ULL,
    0x3cb6240000000ULL,
    0x2e4b8e0000000ULL,
    0x13f8d70000000ULL,
    0xc6369c8000000ULL,
    0x4c8d49c000000ULL,
    0x7b9e92a000000ULL,
    0x7360def000000ULL,
    0x3d2ae2800000ULL,
    0x79e270400000ULL,
    0x116184a600000ULL,
    0xa2d19dc700000ULL,
    0xec1804b480000ULL,
    0x1027ddca40000ULL,
    0x4ff324bbe0000ULL,
    0xfa6e2de3d0000ULL,
    0xf20fbcbd88000ULL,
    0x775bb9e83c000ULL,
    0x70018a8aba000ULL,
    0xa8000ae573000ULL,
    0x5406383aae800ULL,
    0xf6062401c4c00ULL,
    0x77038e103e600ULL,
    0x4c84d7352d900ULL,
    0xf5c49ca692c80ULL,
    0xd4e249c57e7c0ULL,
    0x289512964d260ULL,
    0xefb99eed02650ULL,
    0x72920e02170d8ULL,
    0x42be971e212ccULL,
    0x5315bc8cf10caULL,
    0xb6f8b9faf8dbbULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0x4400000000000ULL,
    0xc600000000000ULL,
    0xa700000000000ULL,
    0x2b80000000000ULL,
    0xb440000000000ULL,
    0x5c20000000000ULL,
    0x3150000000000ULL,
    0x43a8000000000ULL,
    0xbb7c000000000ULL,
    0x9d96000000000ULL,
    0x9e4f000000000ULL,
    0x214b800000000ULL,
    0xec8400000000ULL,
    0xdc8de00000000ULL,
    0x77ad900000000ULL,
    0xf57e480000000ULL,
    0x4693040000000ULL,
    0x67c9ae0000000ULL,
    0x6c092b0000000ULL,
    0x9a6b5f8000000ULL,
    0xb41981c000000ULL,
    0x7b6171a000000ULL,
    0x8f4ead000000ULL,
    0x1db9ce0800000ULL,
    0x5d73fb0400000ULL,
    0x5afb778e00000ULL,
    0xd4d355c500000ULL,
    0x6ced77b780000ULL,
    0xbedc55e8c0000ULL,
    0xb186f7b360000ULL,
    0xd14415c0d0000ULL,
    0xb6a317b3e8000ULL,
    0x599585f444000ULL,
    0x184b5f9862000ULL,
    0x264981de4b000ULL,
    0x754971a2ef800ULL,
    0xf0c8ead89cc00ULL,
    0x6f8fce2d5f600ULL,
    0x22cfb39a5d00ULL,
    0xa238f78df3980ULL,
    0xf73715fd02740ULL,
    0x45de97979b6a0ULL,
    0x5802c5cc87650ULL,
    0xa4053fbf18998ULL,
    0x360051ccdc5bcULL,
    0x6f04d9847ea46ULL,
    0xf7857ed5559ebULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0xd000000000000ULL,
    0x800000000000ULL,
    0xcc00000000000ULL,
    0xee00000000000ULL,
    0xd300000000000ULL,
    0x5280000000000ULL,
    0x4ac0000000000ULL,
    0xa5e0000000000ULL,
    0xbed0000000000ULL,
    0x6878000000000ULL,
    0x562c000000000ULL,
    0x5192000000000ULL,
    0x47db000000000ULL,
    0x42df800000000ULL,
    0xb75cc00000000ULL,
    0xf519200000000ULL,
    0x3338900000000ULL,
    0x8388280000000ULL,
    0xbc63ac0000000ULL,
    0x294120000000ULL,
    0xd55b310000000ULL,
    0xa81f058000000ULL,
    0xc2bbbf4000000ULL,
    0x43cc97a000000ULL,
    0x97404e7000000ULL,
    0x3ba0b20800000ULL,
    0x3ef4610400000ULL,
    0x17c98daa00000ULL,
    0xdd40434700000ULL,
    0xbaa38d9980000ULL,
    0x89774377c0000ULL,
    0x188e0da720000ULL,
    0x42e0836270000ULL,
    0xd650adb7d8000ULL,
    0x7bb8d34a04000ULL,
    0x804ba582e6000ULL,
    0xc204ef79ff000ULL,
    0x8d021fab5a800ULL,
    0xf987b2550dc00ULL,
    0xc45a8088b600ULL,
    0xc5246c02e5f00ULL,
    0xea32b21bb7a80ULL,
    0xb92f613a56140ULL,
    0xf5160d9786c20ULL,
    0xba1c835d6f0f0ULL,
    0x47baad98d26e8ULL,
    0x764fd34a20d74ULL,
    0x750625bcf1edeULL,
    0x2d832f5998f79ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0xb400000000000ULL,
    0x2200000000000ULL,
    0x8100000000000ULL,
    0x2180000000000ULL,
    0x5640000000000ULL,
    0xc220000000000ULL,
    0xa910000000000ULL,
    0xe648000000000ULL,
    0x2fdc000000000ULL,
    0x5e56000000000ULL,
    0xb5ef000000000ULL,
    0x988e800000000ULL,
    0x3e3ec00000000ULL,
    0xf921e00000000ULL,
    0x1594300000000ULL,
    0xa788d80000000ULL,
    0xe8bc640000000ULL,
    0x7de64e0000000ULL,
    0x39731b0000000ULL,
    0xe77dcd8000000ULL,
    0xa5810cc000000ULL,
    0x7c4303a000000ULL,
    0xa723d7d000000ULL,
    0xb291ae0800000ULL,
    0x6d082b2c00000ULL,
    0x5a7b959200000ULL,
    0x203a8d100000ULL,
    0x3104ad9280000ULL,
    0x6984fcdd40000ULL,
    0x1244bb88e0000ULL,
    0x82543d9d0000ULL,
    0x9c16d80ae8000ULL,
    0xe5cf643ed4000ULL,
    0xf89ece139a000ULL,
    0xbdf2db10f3000ULL,
    0x4abaad94e7800ULL,
    0xbce7fcfe5cc00ULL,
    0xb8f43ba1de200ULL,
    0x413883ed60700ULL,
    0x8fa7b81111d80ULL,
    0x61569412c9bc0ULL,
    0x636f761b25ce0ULL,
    0x1c4b4f2282d50ULL,
    0x12db5b8c3f858ULL,
    0xb9d273d030afcULL,
    0x58ae8009b8b8eULL,
    0xa62ec0257df9fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x9000000000000ULL,
    0xb800000000000ULL,
    0x1c00000000000ULL,
    0x1a00000000000ULL,
    0xd100000000000ULL,
    0x380000000000ULL,
    0x65c0000000000ULL,
    0x27a0000000000ULL,
    0xc050000000000ULL,
    0xe3b8000000000ULL,
    0x516c000000000ULL,
    0x496000000000ULL,
    0x8f9b000000000ULL,
    0xe0f9800000000ULL,
    0xee8f400000000ULL,
    0x71a2200000000ULL,
    0xd357b00000000ULL,
    0x9d3e180000000ULL,
    0x52b340000000ULL,
    0x33f0c20000000ULL,
    0x6defab0000000ULL,
    0x6dd0fb8000000ULL,
    0xa97fe54000000ULL,
    0x824fb9a000000ULL,
    0xf9840e5000000ULL,
    0x44c4e22800000ULL,
    0x2c231b3400000ULL,
    0x919763a600000ULL,
    0x7a1b916300000ULL,
    0x463d5b8880000ULL,
    0x2fac156340000ULL,
    0x69b201ae60000ULL,
    0xa58bca5930000ULL,
    0xe326180f08000ULL,
    0x9517341a64000ULL,
    0x4f5ec21186000ULL,
    0x5ad8ab330f000ULL,
    0xc1f7b91ab800ULL,
    0xc53ba55d8dc00ULL,
    0xe92c199693e00ULL,
    0x21f0fe5561100ULL,
    0x88eeda3db0e80ULL,
    0xd0549f06493c0ULL,
    0x1bb839ad048e0ULL,
    0x6d6c4e5136e10ULL,
    0x8e914222c6568ULL,
    0xe69ceb356f37cULL,
    0xff7cdbbc9b666ULL,
    0x914f555484ef9ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x4800000000000ULL,
    0xec00000000000ULL,
    0xb200000000000ULL,
    0x300000000000ULL,
    0xf380000000000ULL,
    0x70c0000000000ULL,
    0x3f20000000000ULL,
    0x9c30000000000ULL,
    0xec58000000000ULL,
    0x28ec000000000ULL,
    0x9bf6000000000ULL,
    0x76fb000000000ULL,
    0xb91f800000000ULL,
    0x5a0ec00000000ULL,
    0xa62200000000ULL,
    0x9ed3700000000ULL,
    0x65cc480000000ULL,
    0x64c4240000000ULL,
    0x1926aa0000000ULL,
    0xf931970000000ULL,
    0x5add8c8000000ULL,
    0xbea9c14000000ULL,
    0x7a92a6a000000ULL,
    0xe7ad965000000ULL,
    0x76128a2800000ULL,
    0xbee9e70400000ULL,
    0x86f6449a00000ULL,
    0x147f256f00000ULL,
    0x69582c9e80000ULL,
    0xae68716bc0000ULL,
    0x35b2ceb860000ULL,
    0x639dc258f0000ULL,
    0x3e49e80928000ULL,
    0xc80594043c000ULL,
    0xac06c2025e000ULL,
    0x9201c302f7000ULL,
    0x5306ee9f8b800ULL,
    0xbb85b27075c00ULL,
    0x9cc2203712600ULL,
    0x8d237020c1700ULL,
    0x9f34482ce1680ULL,
    0x1fd82414306c0ULL,
    0x5828aa1bc81a0ULL,
    0xa4d697350c150ULL,
    0xeacc0c8016d88ULL,
    0x55400166bb1ccULL,
    0x72e106863c21aULL,
    0x919726715fb99ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0xa800000000000ULL,
    0x400000000000ULL,
    0x2a00000000000ULL,
    0xcb00000000000ULL,
    0x3280000000000ULL,
    0x6840000000000ULL,
    0xada0000000000ULL,
    0xabf0000000000ULL,
    0x2ba8000000000ULL,
    0xfddc000000000ULL,
    0x66b2000000000ULL,
    0xb58f000000000ULL,
    0x97ee800000000ULL,
    0x627cc00000000ULL,
    0x8a46600000000ULL,
    0xf2a5f00000000ULL,
    0x4b70c80000000ULL,
    0x54e92c0000000ULL,
    0xe4fae60000000ULL,
    0x4003070000000ULL,
    0xe001ae8000000ULL,
    0x50020cc000000ULL,
    0xd807c8a000000ULL,
    0xac02cbf000000ULL,
    0x2e06862800000ULL,
    0xe105f72c00000ULL,
    0xf985e6a200000ULL,
    0x5ac4e0dd00000ULL,
    0xc5e7ce8080000ULL,
    0x657fcc3c0000ULL,
    0x805f00ae20000ULL,
    0xd677e7c4d0000ULL,
    0x9b6e6025c8000ULL,
    0xd339f01dec000ULL,
    0x2262c80902000ULL,
    0xf5962c3a4b000ULL,
    0xe83c663f66800ULL,
    0x78e3c73bc4c00ULL,
    0xb9d5ce911f200ULL,
    0x1f98fcf7d0f00ULL,
    0xb011808a48380ULL,
    0xa4fb27dbad4c0ULL,
    0xa0000020632a0ULL,
    0xb000003d7bf70ULL,
    0x8800003f5fab8ULL,
    0x74000012b17acULL,
    0x8200003f786a2ULL,
    0xcf00003e95cefULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x3800000000000ULL,
    0x7c00000000000ULL,
    0x600000000000ULL,
    0xf700000000000ULL,
    0xae80000000000ULL,
    0x7940000000000ULL,
    0x9d60000000000ULL,
    0x2250000000000ULL,
    0xc588000000000ULL,
    0x625c000000000ULL,
    0x8612000000000ULL,
    0xff69000000000ULL,
    0xfccd800000000ULL,
    0xdf3d400000000ULL,
    0x9447200000000ULL,
    0xd2e4500000000ULL,
    0xda97c80000000ULL,
    0x232e6c0000000ULL,
    0x9a28ae0000000ULL,
    0x77a9490000000ULL,
    0xd26a058000000ULL,
    0x534db8c000000ULL,
    0xf7fd2ba000000ULL,
    0x4ae0b1f000000ULL,
    0x56958e1800000ULL,
    0x2d28190400000ULL,
    0xf92a4db600000ULL,
    0x9b2b94eb00000ULL,
    0x262d25af80000ULL,
    0x91a9e8d2c0000ULL,
    0xf56ae3b220000ULL,
    0xc5cedde790000ULL,
    0xf2bd201af8000ULL,
    0xd181500dc4000ULL,
    0x83c04821ba000ULL,
    0x46262c06c3000ULL,
    0xe2300e34c6800ULL,
    0x8059591bdd400ULL,
    0xfb176d93cfe00ULL,
    0xa8eac4e311100ULL,
    0x480d6d87b8a80ULL,
    0xd79fc4fd25440ULL,
    0xdf32edbc0aba0ULL,
    0xf7db84d7abcf0ULL,
    0x7fd04da9fa398ULL,
    0x484e94dea3d94ULL,
    0x477aa5be3682eULL,
    0x88a1a8d355387ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0xd800000000000ULL,
    0x8400000000000ULL,
    0xd600000000000ULL,
    0xc500000000000ULL,
    0x9280000000000ULL,
    0xb640000000000ULL,
    0xdb20000000000ULL,
    0xa0b0000000000ULL,
    0x338000000000ULL,
    0x600c000000000ULL,
    0x79f6000000000ULL,
    0xd9d000000000ULL,
    0xbf78800000000ULL,
    0x6c2b400000000ULL,
    0xb8c3e00000000ULL,
    0x9f64100000000ULL,
    0xf892d80000000ULL,
    0x5c0bdc0000000ULL,
    0x1bf1760000000ULL,
    0x269dd30000000ULL,
    0x1cfcd08000000ULL,
    0x46ef924000000ULL,
    0x94a326a000000ULL,
    0xa074015000000ULL,
    0x285c961800000ULL,
    0x7698c30c00000ULL,
    0x14f8888e00000ULL,
    0x8aee0e7d00000ULL,
    0x6ea730a380000ULL,
    0xef77824040000ULL,
    0x2ddffe8920000ULL,
    0x415edd50d0000ULL,
    0x2e1b603c68000ULL,
    0xd5bf5028bc000ULL,
    0xa049380d86000ULL,
    0xf7d3cc3c6f000ULL,
    0x97adae3993800ULL,
    0x56070f26b4400ULL,
    0x850326964ba00ULL,
    0x7284016c91900ULL,
    0xc644963348d80ULL,
    0x324c3316cd40ULL,
    0x24b688806fce0ULL,
    0xd53f0e5f93b30ULL,
    0xa509b09fb54b8ULL,
    0xeb71c26fcb13cULL,
    0xbbdc9eb3d031aULL,
    0x645d8d5ae8045ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xc800000000000ULL,
    0x8c00000000000ULL,
    0xa00000000000ULL,
    0xfd00000000000ULL,
    0x4480000000000ULL,
    0x3540000000000ULL,
    0xff20000000000ULL,
    0xd3b0000000000ULL,
    0x3888000000000ULL,
    0x88bc000000000ULL,
    0x94f2000000000ULL,
    0x2eab000000000ULL,
    0xf18d800000000ULL,
    0x7238c00000000ULL,
    0x22b5200000000ULL,
    0xb60bb00000000ULL,
    0xe0fc780000000ULL,
    0xbf516c0000000ULL,
    0x5afe0000000ULL,
    0xba200b0000000ULL,
    0x2330388000000ULL,
    0x73cc1fc000000ULL,
    0x549d46a000000ULL,
    0xb0c3d4f000000ULL,
    0x1ce7de0800000ULL,
    0xd7bb1400000ULL,
    0x469e409a00000ULL,
    0x69c673c700000ULL,
    0xee62389080000ULL,
    0x9a971fec40000ULL,
    0x4438c69ba0000ULL,
    0x4db714e610000ULL,
    0xeb88fe14e8000ULL,
    0x873b0b0ebc000ULL,
    0x4a35b8a96a000ULL,
    0x1948dfee83000ULL,
    0x2ada66aabe800ULL,
    0xa46364c895c00ULL,
    0x796263d37e00ULL,
    0x30be1705d8f00ULL,
    0x20f19e8e04280ULL,
    0x50adc8ed9fdc0ULL,
    0xd28e7834c4d60ULL,
    0x85ba6c337f770ULL,
    0x28777e14f4418ULL,
    0xcfe8cb2247a9cULL,
    0x40ad1885c019aULL,
    0xda8bafd261193ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0x7800000000000ULL,
    0x8c00000000000ULL,
    0x8600000000000ULL,
    0x5300000000000ULL,
    0x9380000000000ULL,
    0x2bc0000000000ULL,
    0x68e0000000000ULL,
    0xb0f0000000000ULL,
    0x1a38000000000ULL,
    0x5c4c000000000ULL,
    0x4536000000000ULL,
    0xdddd000000000ULL,
    0xb13d800000000ULL,
    0x3cb400000000ULL,
    0x8f6a00000000ULL,
    0x363bd00000000ULL,
    0xea48e80000000ULL,
    0xbe31240000000ULL,
    0xba5b520000000ULL,
    0x90fd510000000ULL,
    0xbe2fb48000000ULL,
    0x7884ac4000000ULL,
    0x944766a000000ULL,
    0xf523bd5000000ULL,
    0x2315f21800000ULL,
    0xcd4a812400000ULL,
    0x67b15ca600000ULL,
    0xa498885700000ULL,
    0xe899b49f80000ULL,
    0xce99ac67c0000ULL,
    0xad9ae69d60000ULL,
    0x9618fd7a30000ULL,
    0x49db5238d8000ULL,
    0x2b3d5121cc000ULL,
    0x4ecfb49b36000ULL,
    0x9474ac78ef000ULL,
    0x707f66aca3800ULL,
    0x766fbd4d54400ULL,
    0x73a3f23a5ae00ULL,
    0x6857812189900ULL,
    0x2decdcb024c80ULL,
    0x3c63c86ba3240ULL,
    0x92b714aad4120ULL,
    0x141e7c499afd0ULL,
    0x18dc0eb2e80e8ULL,
    0xa9b8d951158d4ULL,
    0x78b8006fa71aULL,
    0x65d64021d91b5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x4400000000000ULL,
    0x6e00000000000ULL,
    0x8d00000000000ULL,
    0x7b80000000000ULL,
    0xf7c0000000000ULL,
    0xea0000000000ULL,
    0xed70000000000ULL,
    0x2628000000000ULL,
    0xbdc000000000ULL,
    0xc8b2000000000ULL,
    0xff8d000000000ULL,
    0x62a800000000ULL,
    0x1bdf400000000ULL,
    0xb3e00000000ULL,
    0x1389d00000000ULL,
    0x342d580000000ULL,
    0xbcdad40000000ULL,
    0x9834f20000000ULL,
    0x12ca730000000ULL,
    0xb6c99a8000000ULL,
    0xa8cb30c000000ULL,
    0x5dcde8a000000ULL,
    0xd24f03f000000ULL,
    0x538f123800000ULL,
    0x942fa31c00000ULL,
    0x6cdec29e00000ULL,
    0x3030e4fd00000ULL,
    0x4ec99abd80000ULL,
    0x9ccb30c240000ULL,
    0x4bcde8bd60000ULL,
    0xab4f03c2b0000ULL,
    0x5e0f122348000ULL,
    0xaaefa30bac000ULL,
    0x77fec28fa2000ULL,
    0xa780e4e7eb000ULL,
    0x1dc19a98c6800ULL,
    0x8da730c2a4400ULL,
    0xabf7e88d6e200ULL,
    0xb26e03c0f8900ULL,
    0xb6bf922b64580ULL,
    0x4561e30b4ea40ULL,
    0xb9d5a2b0a92e0ULL,
    0x505b74f3dc810ULL,
    0x2f75a2b24a4a8ULL,
    0x392b74efd7b9cULL,
    0x75da286fd822ULL,
    0xff774ea9ab4fULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x5800000000000ULL,
    0x9400000000000ULL,
    0xae00000000000ULL,
    0xe900000000000ULL,
    0xe480000000000ULL,
    0xa340000000000ULL,
    0x7060000000000ULL,
    0xb750000000000ULL,
    0x6108000000000ULL,
    0x91dc000000000ULL,
    0x5816000000000ULL,
    0x6b6f000000000ULL,
    0xc189800000000ULL,
    0x249c400000000ULL,
    0xe575600000000ULL,
    0xdebe500000000ULL,
    0x26c1780000000ULL,
    0x1ba48c0000000ULL,
    0xee742e0000000ULL,
    0xa938810000000ULL,
    0x8605e58000000ULL,
    0xd50564c000000ULL,
    0x36864ba000000ULL,
    0x3842a5d000000ULL,
    0x1fe14e0800000ULL,
    0x2396d11400000ULL,
    0x11ac9dbe00000ULL,
    0xf82de8e100000ULL,
    0xb9ec659d80000ULL,
    0xffc924cf40000ULL,
    0xae7b2b9aa0000ULL,
    0x5060f5c690000ULL,
    0xe756363b38000ULL,
    0x990d5d3404000ULL,
    0x15d93391da000ULL,
    0xae1529d2e3000ULL,
    0x166ae02ae7800ULL,
    0x8b0d10193cc00ULL,
    0x6edd9800dfe00ULL,
    0x71969c28b8d00ULL,
    0xcaa8362245380ULL,
    0x37ae5d3afb840ULL,
    0x3d2eb38932860ULL,
    0xd76a69d87e770ULL,
    0x5388001528b58ULL,
    0x1f9c0023fcd34ULL,
    0x9af600313ecaeULL,
    0x127f0012883d7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x1800000000000ULL,
    0x7c00000000000ULL,
    0x8600000000000ULL,
    0xa700000000000ULL,
    0xda80000000000ULL,
    0x240000000000ULL,
    0x5160000000000ULL,
    0x4550000000000ULL,
    0x2398000000000ULL,
    0x594c000000000ULL,
    0x2f16000000000ULL,
    0x56f9000000000ULL,
    0x8618800000000ULL,
    0xdd89400000000ULL,
    0x6a36200000000ULL,
    0xdc9500000000ULL,
    0x5e51480000000ULL,
    0x5f1d4c0000000ULL,
    0x6c0ed60000000ULL,
    0x4cf1f10000000ULL,
    0x75eebd8000000ULL,
    0xee3a84000000ULL,
    0xe096eba000000ULL,
    0x1b3e195000000ULL,
    0xdab8f63800000ULL,
    0xa478a12c00000ULL,
    0xc05ff59e00000ULL,
    0xe6eee46300000ULL,
    0xe6603d8780000ULL,
    0x37d3e866c0000ULL,
    0x25d84baf20000ULL,
    0x22e094550000ULL,
    0x53479e1cf8000ULL,
    0x74e0bd22c4000ULL,
    0x81966ba33e000ULL,
    0x6bb594dc7000ULL,
    0xf278d63203800ULL,
    0xef58f125bd400ULL,
    0xa86e3db13ba00ULL,
    0x626e866d8100ULL,
    0x3bb6cbb715a80ULL,
    0x9b0e497c87040ULL,
    0x5e713e03e25a0ULL,
    0x83acad03cc470ULL,
    0x4d87038692478ULL,
    0xe0c64553e43f4ULL,
    0xbf27c8256e32eULL,
    0x7a310c06bed25ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x3800000000000ULL,
    0xfc00000000000ULL,
    0x8600000000000ULL,
    0x1100000000000ULL,
    0x580000000000ULL,
    0xdcc0000000000ULL,
    0xe7e0000000000ULL,
    0x7950000000000ULL,
    0x9478000000000ULL,
    0x4bac000000000ULL,
    0x596000000000ULL,
    0x4399000000000ULL,
    0x7aff800000000ULL,
    0x45ef400000000ULL,
    0x4a33200000000ULL,
    0x146a500000000ULL,
    0x94f2f80000000ULL,
    0xf88c040000000ULL,
    0x6f23da0000000ULL,
    0x9cb0010000000ULL,
    0xe62aa18000000ULL,
    0x5d524b4000000ULL,
    0xbe78fba000000ULL,
    0x84a80a5000000ULL,
    0xfd10fa3800000ULL,
    0x24da511400000ULL,
    0xacd8599600000ULL,
    0x58de4f7b00000ULL,
    0x6adb218f80000ULL,
    0x29d80b6b40000ULL,
    0x7f5a5b98e0000ULL,
    0xcd181a6770000ULL,
    0x64b8a20e98000ULL,
    0x324a450264000ULL,
    0xe1c5dbab66000ULL,
    0x7c675a73ff000ULL,
    0x3893822a9b800ULL,
    0xd81c151e7d400ULL,
    0x3b39239ec2600ULL,
    0x98e5e68b9900ULL,
    0x3aa1d83fb5280ULL,
    0x18765400ee440ULL,
    0xedc92213b3f60ULL,
    0x9a00052720d50ULL,
    0xc7077b8b50288ULL,
    0x4c874a4249914ULL,
    0x3543da176cf6eULL,
    0x8520013eeb969ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0x8400000000000ULL,
    0xe200000000000ULL,
    0x7500000000000ULL,
    0xf780000000000ULL,
    0x9c0000000000ULL,
    0xe060000000000ULL,
    0xa010000000000ULL,
    0xac78000000000ULL,
    0xa5ac000000000ULL,
    0x59b2000000000ULL,
    0xb409000000000ULL,
    0xcdc7800000000ULL,
    0x2266400000000ULL,
    0xa517e00000000ULL,
    0x53ffb00000000ULL,
    0xb06c780000000ULL,
    0xdfd2340000000ULL,
    0x8319e20000000ULL,
    0xe33b650000000ULL,
    0x798fff8000000ULL,
    0x1505cd4000000ULL,
    0xa781fda000000ULL,
    0x71c1187000000ULL,
    0xf4627a1800000ULL,
    0xda16e11400000ULL,
    0x5d7a659e00000ULL,
    0xb02c9c4500000ULL,
    0x2577e03e80000ULL,
    0xa3efb00c40000ULL,
    0x6414783d60000ULL,
    0x6e7e343b30000ULL,
    0xa0abe223f8000ULL,
    0xa632650b74000ULL,
    0xa1c87fab86000ULL,
    0x4ba38d5c2f000ULL,
    0x15761da092800ULL,
    0x8beea87619400ULL,
    0x816020505600ULL,
    0x78d5199fd00ULL,
    0x2ba987af2a180ULL,
    0x42b2f9700d7c0ULL,
    0xc88d9fadb2660ULL,
    0x20853d4249210ULL,
    0xab45e5a2cc868ULL,
    0x6e26dc4c93c44ULL,
    0x41b2002c187f6ULL,
    0xf009003604461ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x5000000000000ULL,
    0x800000000000ULL,
    0xc00000000000ULL,
    0xce00000000000ULL,
    0x2500000000000ULL,
    0x6480000000000ULL,
    0x37c0000000000ULL,
    0x89a0000000000ULL,
    0x36d0000000000ULL,
    0xca98000000000ULL,
    0x178c000000000ULL,
    0xd2b6000000000ULL,
    0x3aed000000000ULL,
    0xd841800000000ULL,
    0xefe0c00000000ULL,
    0x437200000000ULL,
    0x1e2c900000000ULL,
    0xa463180000000ULL,
    0x10706c0000000ULL,
    0x9fcbea0000000ULL,
    0x35525b0000000ULL,
    0x32dc9f8000000ULL,
    0x86d3c4000000ULL,
    0x6e8055a000000ULL,
    0x70c3f75000000ULL,
    0xb627d21800000ULL,
    0x1892a70400000ULL,
    0x147a6dba00000ULL,
    0x5d3e0b5700000ULL,
    0x4c58a01180000ULL,
    0x90ad501a40000ULL,
    0x84a3b812a0000ULL,
    0xae513c10b0000ULL,
    0x8b5e523948000ULL,
    0xef2e670344000ULL,
    0x4ae34da34e000ULL,
    0x20b39b4117000ULL,
    0xc9ec383f2b800ULL,
    0x7dc0fc16a8400ULL,
    0x2ea6f218d5e00ULL,
    0x5953372cf7700ULL,
    0xecd8f587bba80ULL,
    0xc56ea77a90140ULL,
    0x56046a0f388e0ULL,
    0x81039b146c7f0ULL,
    0x16843f99da5c8ULL,
    0x84c06c6b40bdcULL,
    0xcc23edaa20156ULL,
    0xa792cb5670fddULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0xe800000000000ULL,
    0xc400000000000ULL,
    0x6600000000000ULL,
    0x2100000000000ULL,
    0xdb80000000000ULL,
    0xe540000000000ULL,
    0x6ba0000000000ULL,
    0x5290000000000ULL,
    0x42b8000000000ULL,
    0xc16c000000000ULL,
    0x1676000000000ULL,
    0xb88f000000000ULL,
    0xedc5800000000ULL,
    0x1a61c00000000ULL,
    0xd2f4600000000ULL,
    0x5b49900000000ULL,
    0x52a1780000000ULL,
    0xd513f40000000ULL,
    0xddffc60000000ULL,
    0x51cdff0000000ULL,
    0xf466558000000ULL,
    0x47f68b4000000ULL,
    0x3ecdf3a000000ULL,
    0x2ae2e45000000ULL,
    0x2f32de0800000ULL,
    0xe92b9b3400000ULL,
    0xe456eb9200000ULL,
    0x785b805500000ULL,
    0x465fe00380000ULL,
    0x9b5b502940000ULL,
    0xaade980760000ULL,
    0x5c98a41510000ULL,
    0x43b95e1608000ULL,
    0xcae95b1cf4000ULL,
    0xbb310b93c6000ULL,
    0x872cd06cbf000ULL,
    0x31577803d3800ULL,
    0xbddcf43728c00ULL,
    0xae1a460c04e00ULL,
    0x897c3f181fb00ULL,
    0x5a0a35a562780ULL,
    0x8f031b5111140ULL,
    0xee828b9c08c60ULL,
    0x10c21055f5590ULL,
    0xfbe6982646cf8ULL,
    0x6cb4a42afff2cULL,
    0x106f5e36337ceULL,
    0x55f65b12f8b91ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0x800000000000ULL,
    0x1400000000000ULL,
    0x2200000000000ULL,
    0x5d00000000000ULL,
    0x6c80000000000ULL,
    0x9ac0000000000ULL,
    0x4ca0000000000ULL,
    0x54d0000000000ULL,
    0xbe78000000000ULL,
    0xbc6c000000000ULL,
    0xb6000000000ULL,
    0xeb09000000000ULL,
    0x6246800000000ULL,
    0xb463c00000000ULL,
    0x2573e00000000ULL,
    0xb628300000000ULL,
    0x90d7b80000000ULL,
    0xe47b2c0000000ULL,
    0x9d69a20000000ULL,
    0x6a33db0000000ULL,
    0x6cae78000000ULL,
    0xb65a34000000ULL,
    0x34f0a5a000000ULL,
    0x106e485000000ULL,
    0x7eb5fa3800000ULL,
    0xe00cc71c00000ULL,
    0x41c2fd9a00000ULL,
    0x1724544900000ULL,
    0x8b95602a80000ULL,
    0x169bf00240000ULL,
    0x6bdc580ba0000ULL,
    0xde3f1c3df0000ULL,
    0x5081a0a08000ULL,
    0x7141f73684000ULL,
    0xdbe5c5b436000ULL,
    0x8835b86d6f000ULL,
    0x7bc9a21493800ULL,
    0x37e3db2369400ULL,
    0x9632e79394200ULL,
    0x40c9a34eeeb00ULL,
    0x1c66a59c53580ULL,
    0x2177487a89fc0ULL,
    0xcc2b7a1dc4960ULL,
    0xe1d3072516870ULL,
    0xb6ff1d9ade268ULL,
    0x6ca964613acdcULL,
    0x351258161c296ULL,
    0xf95a1c22ab921ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x5400000000000ULL,
    0xa00000000000ULL,
    0x2900000000000ULL,
    0x780000000000ULL,
    0x2bc0000000000ULL,
    0x6960000000000ULL,
    0xb790000000000ULL,
    0xd608000000000ULL,
    0xd85c000000000ULL,
    0xda32000000000ULL,
    0x987f000000000ULL,
    0x2680800000000ULL,
    0x1043400000000ULL,
    0x24a3600000000ULL,
    0x19f7700000000ULL,
    0x5d19d80000000ULL,
    0x5f17d40000000ULL,
    0x4d4f320000000ULL,
    0x9e7e5d0000000ULL,
    0xb1814b8000000ULL,
    0xa4c592c000000ULL,
    0x89e519a000000ULL,
    0x10d0bff000000ULL,
    0xf62f8a2800000ULL,
    0x74edf91c00000ULL,
    0xda0d218e00000ULL,
    0x665c5bf900000ULL,
    0x6932802480000ULL,
    0x1efc402f40000ULL,
    0x46c3e00b60000ULL,
    0xce4302cb0000ULL,
    0xd952b81db8000ULL,
    0x56eca43214000ULL,
    0x7f0cea2d5a000ULL,
    0x1fda8911b3000ULL,
    0xd1f4f982ee800ULL,
    0x411b8fdaf8c00ULL,
    0x8915b20574a00ULL,
    0x924e1d3feb100ULL,
    0xf6f8abbd0af80ULL,
    0xeac2a2d87a0c0ULL,
    0xc2e5218ee22e0ULL,
    0x52505bd2a7e90ULL,
    0x34688027359a8ULL,
    0xd4f40220ab24ULL,
    0x7e796039fb466ULL,
    0x2184703b232ffULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x7000000000000ULL,
    0x4800000000000ULL,
    0xe400000000000ULL,
    0x6e00000000000ULL,
    0x7f00000000000ULL,
    0x2780000000000ULL,
    0x47c0000000000ULL,
    0x37a0000000000ULL,
    0x89d0000000000ULL,
    0xf9a8000000000ULL,
    0x81bc000000000ULL,
    0x2cb2000000000ULL,
    0xa15d000000000ULL,
    0xb684800000000ULL,
    0xdf47c00000000ULL,
    0x37e6600000000ULL,
    0x89b7d00000000ULL,
    0xfdac80000000ULL,
    0x15c1e40000000ULL,
    0x3aa6720000000ULL,
    0x73537b0000000ULL,
    0x1cef4f8000000ULL,
    0x8f5ffdc000000ULL,
    0x69875da000000ULL,
    0xe8c756f000000ULL,
    0x4820da3800000ULL,
    0x12144f2c00000ULL,
    0xc0d758a00000ULL,
    0xfd6ba2e500000ULL,
    0xe39e802980000ULL,
    0x3fa6c016c0000ULL,
    0xcdd0e00a60000ULL,
    0x87ad101230000ULL,
    0xc6b8281608000ULL,
    0xa731f40a84000ULL,
    0x6c9ada2802000ULL,
    0x90254f1ddb000ULL,
    0xe13f5bca6800ULL,
    0xae0d62d9c8400ULL,
    0x786e600b65a00ULL,
    0x9d1bd00df2b00ULL,
    0xc4e0c80632b80ULL,
    0xf430e429d3040ULL,
    0x118f21723aa0ULL,
    0xa6e5bb364acd0ULL,
    0x1137afbb7e0e8ULL,
    0xf9eedf1b557cULL,
    0x15a575ad8bd76ULL,
    0xccd7a2fc9ed5bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xd000000000000ULL,
    0xb800000000000ULL,
    0x4400000000000ULL,
    0xba00000000000ULL,
    0x8100000000000ULL,
    0x4e80000000000ULL,
    0x21c0000000000ULL,
    0xda60000000000ULL,
    0x7150000000000ULL,
    0x82b8000000000ULL,
    0xac2c000000000ULL,
    0x8ff6000000000ULL,
    0x7d0d000000000ULL,
    0xcf05800000000ULL,
    0x5d81400000000ULL,
    0xf240e00000000ULL,
    0x9b21700000000ULL,
    0x9f4180000000ULL,
    0xa20c5c0000000ULL,
    0x8a82aa0000000ULL,
    0xdbc1d50000000ULL,
    0xbb602d8000000ULL,
    0xefd0114000000ULL,
    0x1b7a67a000000ULL,
    0x324cb45000000ULL,
    0x44a0521800000ULL,
    0x7eb1f92400000ULL,
    0x2dab1f9600000ULL,
    0xf3b0d86500000ULL,
    0x552e002080000ULL,
    0x2571002140000ULL,
    0xd6cb800520000ULL,
    0xfc60401c30000ULL,
    0x9e53600b38000ULL,
    0xaf3d300524000ULL,
    0xd6697838b6000ULL,
    0x38d06c171f000ULL,
    0x32f85210f3800ULL,
    0x120df93c38400ULL,
    0xa2851fa6a4e00ULL,
    0xc7c1d85477100ULL,
    0x9565800fff580ULL,
    0x6cd1400aa30c0ULL,
    0x90f8e02510be0ULL,
    0xe70d7025080b0ULL,
    0x3e02183d1ccc8ULL,
    0x9b015c1813324ULL,
    0xff872a1e6899eULL,
    0x740950b8c2d1ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0xf800000000000ULL,
    0xec00000000000ULL,
    0xc600000000000ULL,
    0x2100000000000ULL,
    0xd480000000000ULL,
    0x5240000000000ULL,
    0x6ee0000000000ULL,
    0x16f0000000000ULL,
    0x3708000000000ULL,
    0xb9bc000000000ULL,
    0x3756000000000ULL,
    0xe59b000000000ULL,
    0x2707800000000ULL,
    0x9587400000000ULL,
    0xf6c1a00000000ULL,
    0x94a3d00000000ULL,
    0x6c10f80000000ULL,
    0xbf8cc0000000ULL,
    0x69b6060000000ULL,
    0x7b6aed0000000ULL,
    0x5408648000000ULL,
    0xfe3f1dc000000ULL,
    0xca97c2a000000ULL,
    0x42ba20d000000ULL,
    0xecd15e2800000ULL,
    0x765af10400000ULL,
    0xb5211ab600000ULL,
    0xa0d67cef00000ULL,
    0xa05e002f80000ULL,
    0x4c27002340000ULL,
    0xc851800ca0000ULL,
    0xcc1c4008d0000ULL,
    0xefc6201098000ULL,
    0xcc24902a2c000ULL,
    0x88515827fa000ULL,
    0xec1b1c2393000ULL,
    0xbfc6fe3537800ULL,
    0x3422210399c00ULL,
    0x6456629fac600ULL,
    0x2a19f0d9bb100ULL,
    0x9ec1a619b3880ULL,
    0xe0a23d3e26240ULL,
    0x36171ca361d60ULL,
    0x44fc91cab0070ULL,
    0x88366496a85f8ULL,
    0xd7a81dd245584ULL,
    0x8fae42ab5ff0aULL,
    0x73aa60c83d883ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x9000000000000ULL,
    0x8800000000000ULL,
    0xd400000000000ULL,
    0x9a00000000000ULL,
    0x8b00000000000ULL,
    0xd980000000000ULL,
    0x60c0000000000ULL,
    0x620000000000ULL,
    0xe410000000000ULL,
    0x3768000000000ULL,
    0xb5bc000000000ULL,
    0xe1f6000000000ULL,
    0x8dd000000000ULL,
    0xf407800000000ULL,
    0xaa02c00000000ULL,
    0x7305e00000000ULL,
    0x1586f00000000ULL,
    0xa6c4c80000000ULL,
    0xc324340000000ULL,
    0x2c911a0000000ULL,
    0x52f6f0000000ULL,
    0xadddd8000000ULL,
    0x63039dc000000ULL,
    0xdd8127a000000ULL,
    0x92c602d000000ULL,
    0xc926321800000ULL,
    0x2f90ab3400000ULL,
    0x8af8f9a00000ULL,
    0xf01a06d100000ULL,
    0xee20002480000ULL,
    0xe0100007c0000ULL,
    0xc5680028e0000ULL,
    0x7abc001ef0000ULL,
    0x2a760015e8000ULL,
    0x371d001434000ULL,
    0xb1a7802ad2000ULL,
    0xa5d2c036ab000ULL,
    0x9bcde003c7800ULL,
    0x24eaf03768400ULL,
    0x767ac804f4600ULL,
    0x9a553411b3f00ULL,
    0xe089a2c1a980ULL,
    0x124caf028e0c0ULL,
    0x6c29bd8d0dca0ULL,
    0xd45aadf79ff10ULL,
    0xfc478fbf94948ULL,
    0xee6606cb0266cULL,
    0x3076000d5249eULL,
    0xfc1d0028ea80bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0xa800000000000ULL,
    0xe400000000000ULL,
    0x5a00000000000ULL,
    0xa100000000000ULL,
    0xa580000000000ULL,
    0xb3c0000000000ULL,
    0x2220000000000ULL,
    0xd970000000000ULL,
    0x8268000000000ULL,
    0x1edc000000000ULL,
    0x9192000000000ULL,
    0xc0bd000000000ULL,
    0x7000800000000ULL,
    0x6804c00000000ULL,
    0x8401600000000ULL,
    0xaa02700000000ULL,
    0x902980000000ULL,
    0x4187040000000ULL,
    0xe9c2620000000ULL,
    0x8321870000000ULL,
    0x7cf2f68000000ULL,
    0x31af46c000000ULL,
    0x3cf9f4a000000ULL,
    0x48e0b1f000000ULL,
    0x42d39a3800000ULL,
    0x6ed9f30c00000ULL,
    0xf9928cb200000ULL,
    0x44bd05c500000ULL,
    0xda00001c80000ULL,
    0x6100002140000ULL,
    0xc5800027e0000ULL,
    0x43c00038b0000ULL,
    0x8a20000be8000ULL,
    0x3d700023ac000ULL,
    0xd868003e92000ULL,
    0xbfdc0019ab000ULL,
    0x3412002d9e800ULL,
    0x737d001b68400ULL,
    0x52208030eca00ULL,
    0xb174c01c72700ULL,
    0x669601d9a780ULL,
    0xb4de703736ac0ULL,
    0x98909807247e0ULL,
    0x813a040ace1d0ULL,
    0x99c2e224301e8ULL,
    0xeb25472aa890cULL,
    0xf8f396844c7d2ULL,
    0x9bad36c6232c7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x9000000000000ULL,
    0x1800000000000ULL,
    0xc00000000000ULL,
    0xd200000000000ULL,
    0xc900000000000ULL,
    0xba80000000000ULL,
    0xdb40000000000ULL,
    0x72e0000000000ULL,
    0x4ef0000000000ULL,
    0xcfc8000000000ULL,
    0x837c000000000ULL,
    0xdcd6000000000ULL,
    0xa619000000000ULL,
    0x85e3800000000ULL,
    0x3f76400000000ULL,
    0x1f0ba00000000ULL,
    0xc5e700000000ULL,
    0xf186180000000ULL,
    0x90c2fc0000000ULL,
    0xef221e0000000ULL,
    0xbd57450000000ULL,
    0x2edc948000000ULL,
    0x66c023c000000ULL,
    0x50232aa000000ULL,
    0xf8d016d000000ULL,
    0xd01a263800000ULL,
    0x7ae4893400000ULL,
    0x1af0b29e00000ULL,
    0x71cdaae700000ULL,
    0x878181280000ULL,
    0xb757fc2140000ULL,
    0xcbdf9e0220000ULL,
    0xfe44052090000ULL,
    0xca62b4b418000ULL,
    0x24b113e66c000ULL,
    0x9bad12bbca000ULL,
    0xdccadad123000ULL,
    0xa4fd801267800ULL,
    0x313400b89400ULL,
    0x84be2014f5600ULL,
    0xd2713012e7d00ULL,
    0xd38e3827c8080ULL,
    0x281acc1d550c0ULL,
    0xa6e7a616b6ee0ULL,
    0xb0f7c90a31630ULL,
    0x24ce9294c9e18ULL,
    0x78fc9ae1d4fe4ULL,
    0xa9162038f74aaULL,
    0xd1bd30169145bULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0x6800000000000ULL,
    0xac00000000000ULL,
    0x1a00000000000ULL,
    0x6300000000000ULL,
    0x2480000000000ULL,
    0x28c0000000000ULL,
    0x2c20000000000ULL,
    0xdf10000000000ULL,
    0xaf08000000000ULL,
    0x8fdc000000000ULL,
    0x3672000000000ULL,
    0xbb7f000000000ULL,
    0xa3a2800000000ULL,
    0x4f52c00000000ULL,
    0xf9ef200000000ULL,
    0x43ee300000000ULL,
    0x10ea880000000ULL,
    0x7c6c5c0000000ULL,
    0x8ac320000000ULL,
    0x568cff0000000ULL,
    0x469c1d8000000ULL,
    0xd710b6c000000ULL,
    0x130d0fa000000ULL,
    0x2ddd79f000000ULL,
    0x61711a2800000ULL,
    0x41ff530c00000ULL,
    0x5e65079200000ULL,
    0x3ef3e5d700000ULL,
    0x49ba083c80000ULL,
    0xcc819c03c0000ULL,
    0x44c19207a0000ULL,
    0x16200f2ef0000ULL,
    0x4c11b59828000ULL,
    0xe38edad6f4000ULL,
    0xb19b5864a000ULL,
    0x52dafcf3000ULL,
    0x76bb5877e800ULL,
    0x282ddacb18c00ULL,
    0xe849359c7c200ULL,
    0xe3bf1aedcf700ULL,
    0x27861589d1680ULL,
    0x1c412adf183c0ULL,
    0xbce49d8a7dea0ULL,
    0xc73176d4cf5b0ULL,
    0xae1aaf98510e8ULL,
    0xf5d289db5898cULL,
    0xe42c32309dc3eULL,
    0xe24cff115f9e7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0x2800000000000ULL,
    0x2400000000000ULL,
    0x8a00000000000ULL,
    0xb100000000000ULL,
    0xe080000000000ULL,
    0xbc0000000000ULL,
    0x61a0000000000ULL,
    0x4650000000000ULL,
    0x84a8000000000ULL,
    0xe43c000000000ULL,
    0x77b2000000000ULL,
    0x3919000000000ULL,
    0x4c22800000000ULL,
    0xb16c00000000ULL,
    0x21cf600000000ULL,
    0x48f00000000ULL,
    0xe30a980000000ULL,
    0x7d6eac0000000ULL,
    0xc89f8a0000000ULL,
    0xb7631b0000000ULL,
    0x9976db8000000ULL,
    0x3a3cf9c000000ULL,
    0x44b431a000000ULL,
    0x9c9e12f000000ULL,
    0x3562f21800000ULL,
    0xdc76873c00000ULL,
    0x48bc29b200000ULL,
    0xf2767ec500000ULL,
    0xb3ba180880000ULL,
    0x43f16c1540000ULL,
    0x68fa6a22a0000ULL,
    0x1912b22b0000ULL,
    0x6b0923aa48000ULL,
    0x496fa5da64000ULL,
    0xba99a3bbc2000ULL,
    0x3a6065f4db000ULL,
    0xfff4439ce6800ULL,
    0x2efe55c9b9400ULL,
    0xfe91bb848c200ULL,
    0x808809e817f00ULL,
    0x3b2ca98ad1780ULL,
    0x1479bee2988c0ULL,
    0x9c57f825fdfe0ULL,
    0xedaf5c0abfd70ULL,
    0xe8bf922925e48ULL,
    0xe277770a62ddcULL,
    0x4bbc31b5eb8c6ULL,
    0x7ff212ec6ec4bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0x5000000000000ULL,
    0x1800000000000ULL,
    0xb400000000000ULL,
    0x200000000000ULL,
    0x7b00000000000ULL,
    0xcc80000000000ULL,
    0xe1c0000000000ULL,
    0x19a0000000000ULL,
    0x8090000000000ULL,
    0xeea8000000000ULL,
    0x6fc000000000ULL,
    0x4e76000000000ULL,
    0x695b000000000ULL,
    0xd7e7800000000ULL,
    0x4070400000000ULL,
    0xb45c600000000ULL,
    0x5a63300000000ULL,
    0xf231080000000ULL,
    0x9ab91c0000000ULL,
    0xed97fa0000000ULL,
    0x4b2edb0000000ULL,
    0x68ba0d8000000ULL,
    0x9e90efc000000ULL,
    0xbaf97a000000ULL,
    0x6f7a04d000000ULL,
    0xe635122800000ULL,
    0x88b8b70400000ULL,
    0xae969f8e00000ULL,
    0x63af18f900000ULL,
    0x937ce82680000ULL,
    0x48316c23c0000ULL,
    0x45bd123ba0000ULL,
    0x1b14b732b0000ULL,
    0x35e89fb348000ULL,
    0xa79818f554000ULL,
    0x30c5681d42000ULL,
    0x32262c2f5b000ULL,
    0x73d0f21317800ULL,
    0x93ccc73799c00ULL,
    0x864a77a60c600ULL,
    0xc00e74c76ed00ULL,
    0xcce9fa03c4f80ULL,
    0x5019db09fbdc0ULL,
    0x3d838db926da0ULL,
    0x9a47afc6104b0ULL,
    0xf2e27792785c8ULL,
    0x49f274f95de1cULL,
    0x6c1ffa1f7665eULL,
    0xf382db3888413ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x9800000000000ULL,
    0xbc00000000000ULL,
    0xae00000000000ULL,
    0x8700000000000ULL,
    0xf080000000000ULL,
    0xc7c0000000000ULL,
    0xca0000000000ULL,
    0xd510000000000ULL,
    0x4608000000000ULL,
    0xa5dc000000000ULL,
    0x44f2000000000ULL,
    0x9d7b000000000ULL,
    0xdbe5800000000ULL,
    0xee76400000000ULL,
    0x37bc200000000ULL,
    0x82c0100000000ULL,
    0x2d23880000000ULL,
    0xe9535c0000000ULL,
    0xdbe84a0000000ULL,
    0x18aaa30000000ULL,
    0x39cea98000000ULL,
    0x4578d6c000000ULL,
    0x7e0c3a000000ULL,
    0x707565f000000ULL,
    0xb8ba620800000ULL,
    0x5643af1400000ULL,
    0xf8e34b8e00000ULL,
    0xccf639cb00000ULL,
    0x797a282d80000ULL,
    0x69e50c3ec0000ULL,
    0x5777e23fa0000ULL,
    0xd839ef3430000ULL,
    0x9856b9908000ULL,
    0x484129ebb4000ULL,
    0xb7e620192a000ULL,
    0xb877100df3000ULL,
    0x3cbc08149e800ULL,
    0xd4421c1bd9c00ULL,
    0x49e3ea02ede00ULL,
    0x777f30687500ULL,
    0xe03c818114680ULL,
    0x2581dad39a940ULL,
    0x7e45218bbb6a0ULL,
    0x8ce78afc4a130ULL,
    0xe6f2898063908ULL,
    0x7c78c6e6a7efcULL,
    0x28634b8ae4b0eULL,
    0x5b3639e5b3997ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0x1800000000000ULL,
    0x6400000000000ULL,
    0x1e00000000000ULL,
    0x1b00000000000ULL,
    0x5a80000000000ULL,
    0xf1c0000000000ULL,
    0x53a0000000000ULL,
    0xa110000000000ULL,
    0xe288000000000ULL,
    0xf5c000000000ULL,
    0x8bf2000000000ULL,
    0xa8ff000000000ULL,
    0xcce3800000000ULL,
    0xc570400000000ULL,
    0x4f3ae00000000ULL,
    0xd841500000000ULL,
    0x94e6e80000000ULL,
    0x41702c0000000ULL,
    0xa13b2a0000000ULL,
    0xb410b0000000ULL,
    0xb260618000000ULL,
    0xcab462c000000ULL,
    0xf79baba000000ULL,
    0xebd739f000000ULL,
    0xfbaca22800000ULL,
    0x678c370400000ULL,
    0x8edd438e00000ULL,
    0xb715ed00000ULL,
    0xda9f882180000ULL,
    0x26513c2fc0000ULL,
    0xe2ef221920000ULL,
    0xbcec770eb0000ULL,
    0x47efa38448000ULL,
    0xed6a45cd34000ULL,
    0xd4ab600bca000ULL,
    0xfb0e101d53000ULL,
    0x201f8804fe800ULL,
    0xc7913c2198c00ULL,
    0x894f22032c600ULL,
    0xa9fc772a27d00ULL,
    0xa367a39fd4580ULL,
    0x9d3645c25b440ULL,
    0x1bd96022ab820ULL,
    0xb9311000039f0ULL,
    0xe5dc0822a7908ULL,
    0x52317c3f9459cULL,
    0x775dc229ba6aeULL,
    0xdff1271cba0e7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0x3000000000000ULL,
    0xf800000000000ULL,
    0x1400000000000ULL,
    0xce00000000000ULL,
    0x3500000000000ULL,
    0x80000000000ULL,
    0xe3c0000000000ULL,
    0x9720000000000ULL,
    0xe710000000000ULL,
    0xd0e8000000000ULL,
    0xc73c000000000ULL,
    0xc976000000000ULL,
    0xf81d000000000ULL,
    0xb262800000000ULL,
    0xeaf4400000000ULL,
    0xaa58600000000ULL,
    0x2003500000000ULL,
    0x9002180000000ULL,
    0x28060c0000000ULL,
    0xdc00ea0000000ULL,
    0x2205cf0000000ULL,
    0xef03258000000ULL,
    0xfb80d1c000000ULL,
    0xd645afa000000ULL,
    0x74674ed000000ULL,
    0x93f0121800000ULL,
    0xa0d8d31c00000ULL,
    0xf0c7b78600000ULL,
    0xdea142c900000ULL,
    0xf650f82d80000ULL,
    0x830d1c1840000ULL,
    0xa08c920ba0000ULL,
    0xf2cd9339b0000ULL,
    0x60ab579be8000ULL,
    0x1a5b52c434000ULL,
    0x9800000272000ULL,
    0x640000154b000ULL,
    0xd600003e47800ULL,
    0x1100003b38400ULL,
    0x368000050d200ULL,
    0xc2c000387e500ULL,
    0x59a00001b4b80ULL,
    0x31d0000e32c40ULL,
    0x4748003ceb1e0ULL,
    0xc3ec0004f6170ULL,
    0x8ebe001a501c8ULL,
    0xd8310001f9e0cULL,
    0xabfc80126cebeULL,
    0xd5d5400fae93bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x1000000000000ULL,
    0x2800000000000ULL,
    0x1400000000000ULL,
    0x6200000000000ULL,
    0x8700000000000ULL,
    0x8c80000000000ULL,
    0xbac0000000000ULL,
    0x5f20000000000ULL,
    0xb9d0000000000ULL,
    0xf8d8000000000ULL,
    0xeacc000000000ULL,
    0x74b6000000000ULL,
    0xdee9000000000ULL,
    0x3fe7800000000ULL,
    0x1df3c00000000ULL,
    0xdb8ae00000000ULL,
    0x5dd7300000000ULL,
    0xe2dc680000000ULL,
    0x31cffc0000000ULL,
    0x9e34220000000ULL,
    0xa92b570000000ULL,
    0x7f44738000000ULL,
    0xf0670e4000000ULL,
    0x41b4b1a000000ULL,
    0xdd6e695000000ULL,
    0x6da52a3800000ULL,
    0xa2105b1400000ULL,
    0x867ed99200000ULL,
    0xe658954100000ULL,
    0xaf0e883280000ULL,
    0xfb14cc34c0000ULL,
    0xbbfe4a23a0000ULL,
    0x551dab0b70000ULL,
    0x516fd1bee8000ULL,
    0x3a39945c4000ULL,
    0xb1422117e000ULL,
    0xc3fb572507000ULL,
    0x91c73983b800ULL,
    0x376b0e67b9400ULL,
    0xcea2b1af1c600ULL,
    0x1497694bb2b00ULL,
    0x97baaa27ad580ULL,
    0x6bff9b1f7a3c0ULL,
    0x5d1a39a258220ULL,
    0x356aa5430c730ULL,
    0x59a360358bf28ULL,
    0xb011f03bf1dacULL,
    0x397f081328ac6ULL,
    0x56de0c30e5da5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x6000000000000ULL,
    0x5000000000000ULL,
    0x9800000000000ULL,
    0xf400000000000ULL,
    0xfa00000000000ULL,
    0xc700000000000ULL,
    0x9580000000000ULL,
    0x80c0000000000ULL,
    0xaaa0000000000ULL,
    0x1710000000000ULL,
    0xdb98000000000ULL,
    0x554c000000000ULL,
    0xd4f2000000000ULL,
    0xd0ed000000000ULL,
    0xca65800000000ULL,
    0x2db0400000000ULL,
    0x3488600000000ULL,
    0x2ad6f00000000ULL,
    0xe3bb280000000ULL,
    0x371a340000000ULL,
    0x750ffa0000000ULL,
    0xa012ab0000000ULL,
    0x2619618000000ULL,
    0x898cf24000000ULL,
    0xb854fba000000ULL,
    0x96f9a97000000ULL,
    0x4d79323800000ULL,
    0x50be2f2400000ULL,
    0x989dd3be00000ULL,
    0x52ce9d6900000ULL,
    0xcf33483880000ULL,
    0x29ccc43cc0000ULL,
    0xcb4d203e0000ULL,
    0x89f05f0000ULL,
    0x5e969ba9c8000ULL,
    0x5d5e5958f4000ULL,
    0xceed9a02ce000ULL,
    0xcf655b1157000ULL,
    0xbb35c9beba800ULL,
    0x13cb865558400ULL,
    0xabb6e1ab0de00ULL,
    0xc58db27793500ULL,
    0x46531b8fdd780ULL,
    0x3fe19414aac0ULL,
    0x23fdfa3b90520ULL,
    0xd3ffab0a78710ULL,
    0x7bfce1bd1c548ULL,
    0x47fcb273aa37cULL,
    0xd1fc9bb2e1beeULL,
    0x18ff5949712a5ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x2800000000000ULL,
    0xa400000000000ULL,
    0x7e00000000000ULL,
    0xe500000000000ULL,
    0x8d80000000000ULL,
    0x8ec0000000000ULL,
    0x1260000000000ULL,
    0x1bf0000000000ULL,
    0x1488000000000ULL,
    0xa6bc000000000ULL,
    0xdad2000000000ULL,
    0xbdf000000000ULL,
    0x17a6800000000ULL,
    0xd914c00000000ULL,
    0xd03d600000000ULL,
    0xac91500000000ULL,
    0x62fb980000000ULL,
    0x6cf3640000000ULL,
    0x360e6a0000000ULL,
    0x46ff8d0000000ULL,
    0xd2f0808000000ULL,
    0xf309bdc000000ULL,
    0xbb798aa000000ULL,
    0x743460f000000ULL,
    0x456e122800000ULL,
    0xde89792c00000ULL,
    0x85b8128200000ULL,
    0x6e5404f500000ULL,
    0x8a9cf83580000ULL,
    0x9c01342040000ULL,
    0x62017234e0000ULL,
    0x47072906f0000ULL,
    0xea850aadf8000ULL,
    0x1443a0cc4c000ULL,
    0x2e27f23e3a000ULL,
    0x91d3e92613000ULL,
    0xfb586a98f6800ULL,
    0xb8e2f0d4e8c00ULL,
    0xefb46a1dc5200ULL,
    0x6aac8d2b4ef00ULL,
    0x6f6c00a344f80ULL,
    0xad8e7dd18e640ULL,
    0x69386a93653e0ULL,
    0x6312f0c85f4d0ULL,
    0xdb3c6a23ccdc8ULL,
    0xbc108d2f7a0a4ULL,
    0x9dbe00a972e2aULL,
    0x2517df946d83ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xb000000000000ULL,
    0x2800000000000ULL,
    0x3c00000000000ULL,
    0x6e00000000000ULL,
    0x700000000000ULL,
    0x8180000000000ULL,
    0xf40000000000ULL,
    0x360000000000ULL,
    0x5950000000000ULL,
    0xaff8000000000ULL,
    0xd06c000000000ULL,
    0x8fb2000000000ULL,
    0x4bef000000000ULL,
    0x85f0800000000ULL,
    0x180ac00000000ULL,
    0xf165a00000000ULL,
    0x4051100000000ULL,
    0x217b080000000ULL,
    0xd2ac540000000ULL,
    0xc594460000000ULL,
    0x2adeb90000000ULL,
    0x9cdd2d8000000ULL,
    0x6fd9374000000ULL,
    0x2c5c4ba000000ULL,
    0xd21c5e7000000ULL,
    0x43ffce2800000ULL,
    0xc6682d1c00000ULL,
    0xcb4cb8200000ULL,
    0x6a9e6300000ULL,
    0xf7b06e3180000ULL,
    0xcfea3d3ec0000ULL,
    0x4ff543b3e0000ULL,
    0x650f0a6e10000ULL,
    0x25e3080ab8000ULL,
    0xa790542a84000ULL,
    0xabde4638a6000ULL,
    0x65db90f8f000ULL,
    0x3f1fada272800ULL,
    0xf7cf76a30c00ULL,
    0x35a96bb9c8a00ULL,
    0xd4178e736d700ULL,
    0xbd99663ebb480ULL,
    0x8bb9693dddf40ULL,
    0x64898595332a0ULL,
    0xeaa7734310910ULL,
    0x84f1858239f38ULL,
    0x428b7372c50b4ULL,
    0x81a3859a07b2eULL,
    0xd374734dbe679ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0x1000000000000ULL,
    0x1800000000000ULL,
    0xd400000000000ULL,
    0x6600000000000ULL,
    0xfd00000000000ULL,
    0x4d80000000000ULL,
    0x8a40000000000ULL,
    0xa1a0000000000ULL,
    0xe970000000000ULL,
    0xcd38000000000ULL,
    0xe78c000000000ULL,
    0x9f56000000000ULL,
    0xa90f000000000ULL,
    0x3191800000000ULL,
    0x96a400000000ULL,
    0x1e07600000000ULL,
    0xd907100000000ULL,
    0x8381180000000ULL,
    0xab43a40000000ULL,
    0x4625ce0000000ULL,
    0x2c30170000000ULL,
    0xba1b1a8000000ULL,
    0x343e984000000ULL,
    0x340834a000000ULL,
    0x8c13df5000000ULL,
    0x2b2d560800000ULL,
    0x63a5f31400000ULL,
    0x9a77b4ae00000ULL,
    0x1ba9f7100000ULL,
    0x9a4db63f80000ULL,
    0xe3b7a31a40000ULL,
    0xe3584ca2e0000ULL,
    0x49186b59d0000ULL,
    0xb8b8001d08000ULL,
    0xe9cc003dcc000ULL,
    0xe0f6002aa6000ULL,
    0x797f0038ff000ULL,
    0xcf298001fb800ULL,
    0x4da6400430c00ULL,
    0xb71601a59e00ULL,
    0xee38102d05500ULL,
    0x9308982bcbb80ULL,
    0x2695e42f68540ULL,
    0xabecae28dc1e0ULL,
    0x52c4070c8eed0ULL,
    0x5a65828e43558ULL,
    0xd8947c5a657ccULL,
    0xc2ed1abf5bed2ULL,
    0xd941984a01a7dULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xa000000000000ULL,
    0x9000000000000ULL,
    0x2800000000000ULL,
    0xa400000000000ULL,
    0x2a00000000000ULL,
    0x9f00000000000ULL,
    0x3280000000000ULL,
    0xc440000000000ULL,
    0xf7a0000000000ULL,
    0xed70000000000ULL,
    0xf068000000000ULL,
    0xffdc000000000ULL,
    0x6e52000000000ULL,
    0x385b000000000ULL,
    0x6795800000000ULL,
    0x93bc00000000ULL,
    0x105600000000ULL,
    0x6f86f00000000ULL,
    0x6dc1e80000000ULL,
    0xbb64cc0000000ULL,
    0x6997a60000000ULL,
    0x7c38470000000ULL,
    0xc84128000000ULL,
    0x94432c000000ULL,
    0x762354a000000ULL,
    0x5b145f000000ULL,
    0x5e8bce0800000ULL,
    0xec8b4b1400000ULL,
    0xaf8cd4a200000ULL,
    0x9b0d85cd00000ULL,
    0xde492e2080000ULL,
    0x866d7b3ec0000ULL,
    0xa6dddcae20000ULL,
    0x7dd479f2d0000ULL,
    0xc39a001b48000ULL,
    0x75f700134c000ULL,
    0x6b2f802372000ULL,
    0x9afcc0076b000ULL,
    0xd762e0212e800ULL,
    0x17963029f1c00ULL,
    0xd139083519e00ULL,
    0xbd05fc10c4300ULL,
    0x59812e291e380ULL,
    0x54c17b17484c0ULL,
    0x98e7dcbd4da20ULL,
    0x2a5379c0724f0ULL,
    0xe25d8004eaf38ULL,
    0xe097c03c6ed2cULL,
    0x27bf602a91352ULL,
    0x6341f02ca8617ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0x7000000000000ULL,
    0x4800000000000ULL,
    0xc00000000000ULL,
    0xbe00000000000ULL,
    0x6900000000000ULL,
    0x4080000000000ULL,
    0x3540000000000ULL,
    0x97e0000000000ULL,
    0x4030000000000ULL,
    0xb218000000000ULL,
    0xa3ec000000000ULL,
    0xb152000000000ULL,
    0xba69000000000ULL,
    0xac90800000000ULL,
    0x84cfc00000000ULL,
    0xfd41200000000ULL,
    0x5be0100000000ULL,
    0x1e33f80000000ULL,
    0xab1aac0000000ULL,
    0xab681e0000000ULL,
    0x88143d0000000ULL,
    0x93884a8000000ULL,
    0x85a5f34000000ULL,
    0x7651f4a000000ULL,
    0x6bee1e7000000ULL,
    0x7d53661800000ULL,
    0xe46d512c00000ULL,
    0xb59374aa00000ULL,
    0x8c48de5f00000ULL,
    0xc402c61d80000ULL,
    0x7202812940000ULL,
    0x3701ac87a0000ULL,
    0x5982624b10000ULL,
    0x3dc12007d8000ULL,
    0xaea0102e5c000ULL,
    0x69d3f81efe000ULL,
    0x9b2aac3eb7000ULL,
    0x51701e02f2800ULL,
    0x27f83d0831400ULL,
    0x9cda4a8d08a00ULL,
    0x56ccf345e5b00ULL,
    0x9a417488d2680ULL,
    0xda61de6260f40ULL,
    0x17f2462970360ULL,
    0xffbd412ea8ef0ULL,
    0x19b88c80f4d98ULL,
    0x84be725e0b544ULL,
    0xde38d800bcdd2ULL,
    0x407fbc124e379ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0x3000000000000ULL,
    0xe800000000000ULL,
    0x3400000000000ULL,
    0x1200000000000ULL,
    0xfd00000000000ULL,
    0xe080000000000ULL,
    0x5040000000000ULL,
    0x2aa0000000000ULL,
    0xb3f0000000000ULL,
    0x558000000000ULL,
    0x4a6c000000000ULL,
    0xebd6000000000ULL,
    0x5be9000000000ULL,
    0xe516800000000ULL,
    0x720cc00000000ULL,
    0x77c1e00000000ULL,
    0x7362b00000000ULL,
    0x6b93380000000ULL,
    0xe948cc0000000ULL,
    0xeae6560000000ULL,
    0x61579f0000000ULL,
    0x3aa85e8000000ULL,
    0xb1355f4000000ULL,
    0x92b968a000000ULL,
    0x8bbcb05000000ULL,
    0xe13bee3800000ULL,
    0xf07f931c00000ULL,
    0x2c59e8a200000ULL,
    0xc8e9704100000ULL,
    0xce948e1e80000ULL,
    0x95cde32240000ULL,
    0xfca530a4a0000ULL,
    0x94f60c6750000ULL,
    0x7cd9602888000ULL,
    0x30ab7025ec000ULL,
    0xb03258177e000ULL,
    0x43fbc0d27000ULL,
    0xecfa0e3983800ULL,
    0xa1d231d71400ULL,
    0xd54ad08d98200ULL,
    0xbce1bc59c5700ULL,
    0x652d83dc3d80ULL,
    0x232a7c16d09c0ULL,
    0xfbf56e2f49360ULL,
    0x215f53108cdf0ULL,
    0xe06e08a24efb8ULL,
    0xfad2c0611fe4cULL,
    0x75693637967a2ULL,
    0x6e55ef117a5c1ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0xa800000000000ULL,
    0x2c00000000000ULL,
    0x9600000000000ULL,
    0x2300000000000ULL,
    0x1580000000000ULL,
    0xf140000000000ULL,
    0x5460000000000ULL,
    0x8a90000000000ULL,
    0xe038000000000ULL,
    0xf06c000000000ULL,
    0x9a72000000000ULL,
    0x6769000000000ULL,
    0x5df6800000000ULL,
    0xef2ec00000000ULL,
    0x1917200000000ULL,
    0xa7e500000000ULL,
    0x658be80000000ULL,
    0x43a7540000000ULL,
    0xe237d60000000ULL,
    0xab8fd10000000ULL,
    0xc4a62b8000000ULL,
    0x1db7e74000000ULL,
    0xf7c85da000000ULL,
    0x2244a67000000ULL,
    0xc9e0be1800000ULL,
    0x7d3451c00000ULL,
    0x3a5a5d9200000ULL,
    0xddfda66700000ULL,
    0xd5ce3e2f80000ULL,
    0xd3418533c0000ULL,
    0xa5677d98e0000ULL,
    0xe616f65890000ULL,
    0x1f9561a78000ULL,
    0xcbcd1125f4000ULL,
    0x8c430bb096000ULL,
    0x3ee0b77a7f000ULL,
    0xb0553591f2800ULL,
    0xfa1d326291400ULL,
    0x518481979200ULL,
    0xe9ec41475f00ULL,
    0xa0dd9e3e56680ULL,
    0x6f38151e1f240ULL,
    0xb3ed35a323b20ULL,
    0xa831326ac8010ULL,
    0x968a480e7caf8ULL,
    0x8e27c4394a914ULL,
    0x77731e0105376ULL,
    0xf5ead5043fd99ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xf000000000000ULL,
    0xf800000000000ULL,
    0x1400000000000ULL,
    0x3a00000000000ULL,
    0x8900000000000ULL,
    0x9b80000000000ULL,
    0xbbc0000000000ULL,
    0x5e0000000000ULL,
    0x5570000000000ULL,
    0xb908000000000ULL,
    0x7e3c000000000ULL,
    0x5c16000000000ULL,
    0x673f000000000ULL,
    0x2f96800000000ULL,
    0x60fa400000000ULL,
    0xc76600000000ULL,
    0x6a8cf00000000ULL,
    0x89fb580000000ULL,
    0x87f1b40000000ULL,
    0x7948120000000ULL,
    0x901f990000000ULL,
    0x481c88000000ULL,
    0x674695c000000ULL,
    0xc6a13aa000000ULL,
    0xf1d3bcd000000ULL,
    0x25dbca0800000ULL,
    0x2676d2c00000ULL,
    0x78b73a9600000ULL,
    0xbbecbcf900000ULL,
    0xf3cd4a1080000ULL,
    0xb45d2d3740000ULL,
    0x28a15a8060000ULL,
    0xa2d04cf0b0000ULL,
    0x675e123998000ULL,
    0x2a20992eac000ULL,
    0x2a974889ba000ULL,
    0xc57cd5d493000ULL,
    0xacb75a81a7800ULL,
    0x61ef4cfdb0c00ULL,
    0x2ac8921b18200ULL,
    0x27dad91aed500ULL,
    0x7f6128965a380ULL,
    0x893025cae2ac0ULL,
    0x812c02a79ece0ULL,
    0x81aef8c08c910ULL,
    0x74e88039eb728ULL,
    0x27494022fa1b4ULL,
    0x9b1ee03af26eaULL,
    0xa205b03c16bf3ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x8800000000000ULL,
    0x400000000000ULL,
    0x5a00000000000ULL,
    0x4500000000000ULL,
    0x1b80000000000ULL,
    0xa240000000000ULL,
    0x6860000000000ULL,
    0x2390000000000ULL,
    0xedf8000000000ULL,
    0xe9ac000000000ULL,
    0xf176000000000ULL,
    0x9da9000000000ULL,
    0x6370800000000ULL,
    0xbca9c00000000ULL,
    0xd2f3600000000ULL,
    0xd3ef700000000ULL,
    0xa513b80000000ULL,
    0x83f540000000ULL,
    0x658c5e0000000ULL,
    0xd985d90000000ULL,
    0xdb45138000000ULL,
    0xc5e5e94000000ULL,
    0x6ad2ada000000ULL,
    0x91a805000000ULL,
    0x32f9663800000ULL,
    0x972a4d2c00000ULL,
    0xd8b2adba00000ULL,
    0x2f8a807d00000ULL,
    0xa481661480000ULL,
    0x2cc64d0040000ULL,
    0xc9a4ad8860000ULL,
    0x95b38045d0000ULL,
    0x7009e61ef8000ULL,
    0x3cc38d3174000ULL,
    0xf1a1cdafe6000ULL,
    0x79b5f05fcf000ULL,
    0xde0ade03c3800ULL,
    0xabc5192491400ULL,
    0xab26f3b019a00ULL,
    0x9a73596de4300ULL,
    0x512a7593fef80ULL,
    0xfbb6a4582aac0ULL,
    0xc70880153db60ULL,
    0xf645c0243a490ULL,
    0x6a65602f65d98ULL,
    0x7a967034bfd34ULL,
    0xd07b3809cb0b6ULL,
    0xd8ea940c2d901ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0xd000000000000ULL,
    0xf800000000000ULL,
    0x1c00000000000ULL,
    0x2e00000000000ULL,
    0xf00000000000ULL,
    0xc380000000000ULL,
    0x2440000000000ULL,
    0xe6a0000000000ULL,
    0x7210000000000ULL,
    0x2bd8000000000ULL,
    0x2c0c000000000ULL,
    0xf432000000000ULL,
    0x580f000000000ULL,
    0xde33800000000ULL,
    0x950f400000000ULL,
    0x14b2200000000ULL,
    0x99c9100000000ULL,
    0x27d3c80000000ULL,
    0x83e040000000ULL,
    0x543c8e0000000ULL,
    0x9a3a0f0000000ULL,
    0x653f098000000ULL,
    0x8ebef44000000ULL,
    0x4efa27a000000ULL,
    0x9a5eab7000000ULL,
    0xc94ec62800000ULL,
    0x2e144b0400000ULL,
    0xe5da27b600000ULL,
    0xd30eab7300000ULL,
    0x1fb6c61580000ULL,
    0x98484b34c0000ULL,
    0xa9027b360000ULL,
    0xc61dab7ed0000ULL,
    0xf3ef461748000ULL,
    0x52040b2294000ULL,
    0x110387b0de000ULL,
    0xc484fb6047000ULL,
    0xd3c52e191a800ULL,
    0x8e05f2310c00ULL,
    0xa9b4e1b0a8600ULL,
    0xbb4ae05804100ULL,
    0xef14e196b6680ULL,
    0xd95ae07622940ULL,
    0x1ccce1abfcfa0ULL,
    0x3956e0693a050ULL,
    0x3efee18761318ULL,
    0x7259e040d1d2cULL,
    0xd4d619048e9eULL,
    0xcc16a06914badULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x6000000000000ULL,
    0xf000000000000ULL,
    0x2800000000000ULL,
    0x2400000000000ULL,
    0x9e00000000000ULL,
    0x4d00000000000ULL,
    0x3480000000000ULL,
    0x63c0000000000ULL,
    0xf1a0000000000ULL,
    0x9c70000000000ULL,
    0x9568000000000ULL,
    0xb45c000000000ULL,
    0x9856000000000ULL,
    0xa459000000000ULL,
    0x50800000000ULL,
    0xc85cc00000000ULL,
    0x4a57600000000ULL,
    0x3358500000000ULL,
    0x17d0d80000000ULL,
    0xfa1f6c0000000ULL,
    0xc8b00a0000000ULL,
    0xa34a5b0000000ULL,
    0xa2687a8000000ULL,
    0x73d99ec000000ULL,
    0x121790a000000ULL,
    0xdabb55d000000ULL,
    0x246523800000ULL,
    0x8ae0f72c00000ULL,
    0x2417909200000ULL,
    0xf3bb55f500000ULL,
    0xc8c6520d80000ULL,
    0x5420f70a40000ULL,
    0xc93790b2a0000ULL,
    0x280b55f7b0000ULL,
    0x320e521b48000ULL,
    0x310cf726ec000ULL,
    0xf08990a6b2000ULL,
    0x5bce55e54b000ULL,
    0x5ba8d2246f800ULL,
    0xc1793705f1400ULL,
    0x2fe670a5e9a00ULL,
    0x1496c5d35d700ULL,
    0x9e796a10fbf80ULL,
    0xac670b1226240ULL,
    0xf0d622b29c6e0ULL,
    0x859f32d31b8f0ULL,
    0xbef67ab5b6148ULL,
    0x4fac9ec8a57dcULL,
    0x5779108dde442ULL,
    0xd6e295c9b8a57ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x3000000000000ULL,
    0x8800000000000ULL,
    0xbc00000000000ULL,
    0x2600000000000ULL,
    0xd00000000000ULL,
    0x3580000000000ULL,
    0x76c0000000000ULL,
    0x5ee0000000000ULL,
    0xb750000000000ULL,
    0x7588000000000ULL,
    0x139c000000000ULL,
    0x1932000000000ULL,
    0xdd9f000000000ULL,
    0x7834800000000ULL,
    0x561bc00000000ULL,
    0x7f4a00000000ULL,
    0x2f7b500000000ULL,
    0x5160180000000ULL,
    0x1a93a40000000ULL,
    0xbbee6a0000000ULL,
    0x6b89490000000ULL,
    0xea99918000000ULL,
    0x96b328c000000ULL,
    0xb05fdba000000ULL,
    0x9656f1f000000ULL,
    0x1e08f22800000ULL,
    0x7c5e2d3c00000ULL,
    0xd857dbaa00000ULL,
    0xbf0af1e300000ULL,
    0xd7daf20b80000ULL,
    0x97912d1c40000ULL,
    0x4e6b5b8ce0000ULL,
    0x3d4d31d2f0000ULL,
    0x847c5218d8000ULL,
    0xa9e57d2f64000ULL,
    0x79d7c38b5a000ULL,
    0xa3c955e673000ULL,
    0xa3c98144e800ULL,
    0x9444643130c00ULL,
    0xd6a0ca0e79a00ULL,
    0xb7f1190075300ULL,
    0x677f09adf2080ULL,
    0xcd644ce80f240ULL,
    0xc9191add1ae0ULL,
    0x3eef28e489fd0ULL,
    0xe20ddb9cacbc8ULL,
    0xba59f1ec974b4ULL,
    0xc554720e5456eULL,
    0x3289ed3d238c7ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xe000000000000ULL,
    0xb000000000000ULL,
    0x1800000000000ULL,
    0x1c00000000000ULL,
    0xf200000000000ULL,
    0xd100000000000ULL,
    0xc680000000000ULL,
    0xdb40000000000ULL,
    0x24a0000000000ULL,
    0x34f0000000000ULL,
    0xdbf8000000000ULL,
    0x24cc000000000ULL,
    0x2fd6000000000ULL,
    0xb6cd000000000ULL,
    0xed6800000000ULL,
    0x884f400000000ULL,
    0x7993200000000ULL,
    0x46e8f00000000ULL,
    0x8062d80000000ULL,
    0xa997dc0000000ULL,
    0xaee8be0000000ULL,
    0x6460170000000ULL,
    0xf793e68000000ULL,
    0x95e8ca4000000ULL,
    0x6fe6f8a000000ULL,
    0x18536d5000000ULL,
    0xbb0ae61800000ULL,
    0xb9358b3c00000ULL,
    0x2c1ef8be00000ULL,
    0x609f6d6b00000ULL,
    0x86dce61a80000ULL,
    0x6ef88b2340000ULL,
    0xfc4878b520000ULL,
    0x2f902d6e90000ULL,
    0x29efc61e88000ULL,
    0xcde07b2cec000ULL,
    0x6152a0bb56000ULL,
    0x798bf15c4f000ULL,
    0x8c71782aab800ULL,
    0x2bbd6c0550400ULL,
    0x43efc605e8200ULL,
    0x40e07b1adc900ULL,
    0xb5d2a09c8f680ULL,
    0xc3cbf14b4aec0ULL,
    0x7651782021fe0ULL,
    0xd80d6c3f108b0ULL,
    0x4eb7c63749398ULL,
    0x81dc7b028d22cULL,
    0x877ca08ce79e2ULL,
    0x8a8af14a574b5ULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0x2000000000000ULL,
    0xf000000000000ULL,
    0xe800000000000ULL,
    0xb400000000000ULL,
    0xa00000000000ULL,
    0x6d00000000000ULL,
    0xa780000000000ULL,
    0x7340000000000ULL,
    0x11e0000000000ULL,
    0x3a30000000000ULL,
    0x7e68000000000ULL,
    0xcd9c000000000ULL,
    0xfed6000000000ULL,
    0xfdb000000000ULL,
    0xb6b2800000000ULL,
    0x1aac00000000ULL,
    0xe3bfe00000000ULL,
    0xcac4700000000ULL,
    0xb5a7780000000ULL,
    0x2955740000000ULL,
    0xd49e360000000ULL,
    0x33569d0000000ULL,
    0xc19b928000000ULL,
    0x18d780c000000ULL,
    0x14dec4a000000ULL,
    0x3234add000000ULL,
    0x5a6f4e2800000ULL,
    0xff98e92c00000ULL,
    0x3fd724b200000ULL,
    0xfe5bddd100000ULL,
    0x16f2b61880000ULL,
    0xd0cb5d2640000ULL,
    0x6048f2a360000ULL,
    0x718e30fb90000ULL,
    0x20aadcb798000ULL,
    0x823869ef5c000ULL,
    0x5285603162000ULL,
    0xc8c2b00aab000ULL,
    0xfca6983bcf800ULL,
    0xbcd6043d41c00ULL,
    0x66ddce2ee1200ULL,
    0xd332293b50700ULL,
    0x5be8c4a1b9580ULL,
    0xb7dfadc62cac0ULL,
    0x5ab5ce27aba60ULL,
    0x77ae29394e810ULL,
    0xc0bec4a600f38ULL,
    0xe244adc7014a4ULL,
    0xb8674e0781112ULL,
    0xe674e935c09b7ULL,
    0x8000000000000ULL,
    0xc000000000000ULL,
    0x2000000000000ULL,
    0x7000000000000ULL,
    0x5800000000000ULL,
    0xe400000000000ULL,
    0x9200000000000ULL,
    0x5900000000000ULL,
    0x6480000000000ULL,
    0x2b40000000000ULL,
    0x6460000000000ULL,
    0xf710000000000ULL,
    0xa2a8000000000ULL,
    0xeb7c000000000ULL,
    0x1e72000000000ULL,
    0x1b3f000000000ULL,
    0x8f96800000000ULL,
    0x4fef400000000ULL,
    0x3c18e00000000ULL,
    0x40e2300000000ULL,
    0x3c53980000000ULL,
    0x96cc4c0000000ULL,
    0x346bae0000000ULL,
    0xdf510000000ULL,
    0x86470f8000000ULL,
    0x5ae12fc000000ULL,
    0x6950c1a000000ULL,
    0xfc4c0ef000000ULL,
    0x402a362800000ULL,
    0x933c1d2c00000ULL,
    0x8392219e00000ULL,
    0x41ed3ee500000ULL,
    0x631d2e1280000ULL,
    0xb760111f40000ULL,
    0xce97efb5a0000ULL,
    0x2f6f1fc090000ULL,
    0xd55959b638000ULL,
    0x558342fe14000ULL,
    0x13c5183d32000ULL,
    0x69230c3d5b000ULL,
    0x70734e2816800ULL,
    0xd43d6113c0c00ULL,
    0x7014978fe1200ULL,
    0x712d63c7b0100ULL,
    0xabbb6f99e9980ULL,
    0x8ed35fd58c5c0ULL,
    0xc68d39afb65e0ULL,
    0x158d32e8505f0ULL,
    0x2c0ae02e59ad8ULL,
    0xa1cd302464174ULL,
    0x9fed18363b906ULL,
    0xd41f0c09e7d3bULL,
    0x8000000000000ULL,
    0x4000000000000ULL,
    0xa000000000000ULL,
    0xb000000000000ULL,
    0x5800000000000ULL,
    0x8400000000000ULL,
    0x4a00000000000ULL,
    0x1d00000000000ULL,
    0x8380000000000ULL,
    0xfec0000000000ULL,
    0x5220000000000ULL,
    0x5ad0000000000ULL,
    0xf3f8000000000ULL,
    0x6d6c000000000ULL,
    0xbf6000000000ULL,
    0x75af000000000ULL,
    0x32d4800000000ULL,
    0x6ffd400000000ULL,
    0x36fe00000000ULL,
    0xecf5f00000000ULL,
    0xb32f780000000ULL,
    0xcb94540000000ULL,
    0x891cfa0000000ULL,
    0x169a6f0000000ULL,
    0xc659ab8000000ULL,
    0xd379c74000000ULL,
    0xff2831a000000ULL,
    0xad94185000000ULL,
    0xa21d820800000ULL,
    0xf61d3b3c00000ULL,
    0xb41fd1be00000ULL,
    0x651de84d00000ULL,
    0xc09cfa2480000ULL,
    0xb55a6f3640000ULL,
    0xb7f9ab9a60000ULL,
    0xc769c754f0000ULL,
    0x6f0318268000ULL,
    0x1e28185804000ULL,
    0x10138232f6000ULL,
    0xf3de3b36af000ULL,
    0xebd51a0e3800ULL,
    0x818fa87480c00ULL,
    0xa3079a3e40a00ULL,
    0x6c82df2660900ULL,
    0xf44133aaf0d80ULL,
    0x8d64636f692c0ULL,
    0x3735b3ac85a60ULL,
    0xb689234cb7a90ULL,
    0x6d8253b0cfc08ULL,
    0x59c0d35a13decULL,
    0x34a32b93e92c6ULL,
    0x1397876ec5561ULL
};

const uint64_t Matrices::vdc_sobol_matrices[][52] =
{
    { // m = 1
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL,
        0x1ULL
    },
    { // m = 2
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL,
        0x2ULL,
        0x3ULL
    },
    { // m = 3
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL,
        0x4ULL,
        0x6ULL,
        0x5ULL,
        0x7ULL
    },
    { // m = 4
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL,
        0x8ULL,
        0xcULL,
        0xaULL,
        0xfULL
    },
    { // m = 5
        0x14ULL,
        0x1eULL,
        0x11ULL,
        0x19ULL,
        0x15ULL,
        0x1fULL,
        0x10ULL,
        0x18ULL,
        0x14ULL,
        0x1eULL,
        0x11ULL,
        0x19ULL,
        0x15ULL,
        0x1fULL,
        0x10ULL,
        0x18ULL,
        0x14ULL,
        0x1eULL,
        0x11ULL,
        0x19ULL,
        0x15ULL,
        0x1fULL,
        0x10ULL,
        0x18ULL,
        0x14ULL,
        0x1eULL,
        0x11ULL,
        0x19ULL,
        0x15ULL,
        0x1fULL,
        0x10ULL,
        0x18ULL,
        0x14ULL,
        0x1eULL,
        0x11ULL,
        0x19ULL,
        0x15ULL,
        0x1fULL,
        0x10ULL,
        0x18ULL,
        0x14ULL,
        0x1eULL
    },
    { // m = 6
        0x22ULL,
        0x33ULL,
        0x2aULL,
        0x3fULL,
        0x20ULL,
        0x30ULL,
        0x28ULL,
        0x3cULL,
        0x22ULL,
        0x33ULL,
        0x2aULL,
        0x3fULL,
        0x20ULL,
        0x30ULL,
        0x28ULL,
        0x3cULL,
        0x22ULL,
        0x33ULL,
        0x2aULL,
        0x3fULL,
        0x20ULL,
        0x30ULL,
        0x28ULL,
        0x3cULL,
        0x22ULL,
        0x33ULL,
        0x2aULL,
        0x3fULL,
        0x20ULL,
        0x30ULL,
        0x28ULL,
        0x3cULL,
        0x22ULL,
        0x33ULL,
        0x2aULL,
        0x3fULL,
        0x20ULL,
        0x30ULL,
        0x28ULL,
        0x3cULL
    },
    { // m = 7
        0x55ULL,
        0x7fULL,
        0x40ULL,
        0x60ULL,
        0x50ULL,
        0x78ULL,
        0x44ULL,
        0x66ULL,
        0x55ULL,
        0x7fULL,
        0x40ULL,
        0x60ULL,
        0x50ULL,
        0x78ULL,
        0x44ULL,
        0x66ULL,
        0x55ULL,
        0x7fULL,
        0x40ULL,
        0x60ULL,
        0x50ULL,
        0x78ULL,
        0x44ULL,
        0x66ULL,
        0x55ULL,
        0x7fULL,
        0x40ULL,
        0x60ULL,
        0x50ULL,
        0x78ULL,
        0x44ULL,
        0x66ULL,
        0x55ULL,
        0x7fULL,
        0x40ULL,
        0x60ULL,
        0x50ULL,
        0x78ULL
    },
    { // m = 8
        0x80ULL,
        0xc0ULL,
        0xa0ULL,
        0xf0ULL,
        0x88ULL,
        0xccULL,
        0xaaULL,
        0xffULL,
        0x80ULL,
        0xc0ULL,
        0xa0ULL,
        0xf0ULL,
        0x88ULL,
        0xccULL,
        0xaaULL,
        0xffULL,
        0x80ULL,
        0xc0ULL,
        0xa0ULL,
        0xf0ULL,
        0x88ULL,
        0xccULL,
        0xaaULL,
        0xffULL,
        0x80ULL,
        0xc0ULL,
        0xa0ULL,
        0xf0ULL,
        0x88ULL,
        0xccULL,
        0xaaULL,
        0xffULL,
        0x80ULL,
        0xc0ULL,
        0xa0ULL,
        0xf0ULL
    },
    { // m = 9
        0x140ULL,
        0x1e0ULL,
        0x110ULL,
        0x198ULL,
        0x154ULL,
        0x1feULL,
        0x101ULL,
        0x181ULL,
        0x141ULL,
        0x1e1ULL,
        0x111ULL,
        0x199ULL,
        0x155ULL,
        0x1ffULL,
        0x100ULL,
        0x180ULL,
        0x140ULL,
        0x1e0ULL,
        0x110ULL,
        0x198ULL,
        0x154ULL,
        0x1feULL,
        0x101ULL,
        0x181ULL,
        0x141ULL,
        0x1e1ULL,
        0x111ULL,
        0x199ULL,
        0x155ULL,
        0x1ffULL,
        0x100ULL,
        0x180ULL,
        0x140ULL,
        0x1e0ULL
    },
    { // m = 10
        0x220ULL,
        0x330ULL,
        0x2a8ULL,
        0x3fcULL,
        0x202ULL,
        0x303ULL,
        0x282ULL,
        0x3c3ULL,
        0x222ULL,
        0x333ULL,
        0x2aaULL,
        0x3ffULL,
        0x200ULL,
        0x300ULL,
        0x280ULL,
        0x3c0ULL,
        0x220ULL,
        0x330ULL,
        0x2a8ULL,
        0x3fcULL,
        0x202ULL,
        0x303ULL,
        0x282ULL,
        0x3c3ULL,
        0x222ULL,
        0x333ULL,
        0x2aaULL,
        0x3ffULL,
        0x200ULL,
        0x300ULL,
        0x280ULL,
        0x3c0ULL
    },
    { // m = 11
        0x550ULL,
        0x7f8ULL,
        0x404ULL,
        0x606ULL,
        0x505ULL,
        0x787ULL,
        0x444ULL,
        0x666ULL,
        0x555ULL,
        0x7ffULL,
        0x400ULL,
        0x600ULL,
        0x500ULL,
        0x780ULL,
        0x440ULL,
        0x660ULL,
        0x550ULL,
        0x7f8ULL,
        0x404ULL,
        0x606ULL,
        0x505ULL,
        0x787ULL,
        0x444ULL,
        0x666ULL,
        0x555ULL,
        0x7ffULL,
        0x400ULL,
        0x600ULL,
        0x500ULL,
        0x780ULL
    },
    { // m = 12
        0x808ULL,
        0xc0cULL,
        0xa0aULL,
        0xf0fULL,
        0x888ULL,
        0xcccULL,
        0xaaaULL,
        0xfffULL,
        0x800ULL,
        0xc00ULL,
        0xa00ULL,
        0xf00ULL,
        0x880ULL,
        0xcc0ULL,
        0xaa0ULL,
        0xff0ULL,
        0x808ULL,
        0xc0cULL,
        0xa0aULL,
        0xf0fULL,
        0x888ULL,
        0xcccULL,
        0xaaaULL,
        0xfffULL,
        0x800ULL,
        0xc00ULL,
        0xa00ULL,
        0xf00ULL
    },
    { // m = 13
        0x1414ULL,
        0x1e1eULL,
        0x1111ULL,
        0x1999ULL,
        0x1555ULL,
        0x1fffULL,
        0x1000ULL,
        0x1800ULL,
        0x1400ULL,
        0x1e00ULL,
        0x1100ULL,
        0x1980ULL,
        0x1540ULL,
        0x1fe0ULL,
        0x1010ULL,
        0x1818ULL,
        0x1414ULL,
        0x1e1eULL,
        0x1111ULL,
        0x1999ULL,
        0x1555ULL,
        0x1fffULL,
        0x1000ULL,
        0x1800ULL,
        0x1400ULL,
        0x1e00ULL
    },
    { // m = 14
        0x2222ULL,
        0x3333ULL,
        0x2aaaULL,
        0x3fffULL,
        0x2000ULL,
        0x3000ULL,
        0x2800ULL,
        0x3c00ULL,
        0x2200ULL,
        0x3300ULL,
        0x2a80ULL,
        0x3fc0ULL,
        0x2020ULL,
        0x3030ULL,
        0x2828ULL,
        0x3c3cULL,
        0x2222ULL,
        0x3333ULL,
        0x2aaaULL,
        0x3fffULL,
        0x2000ULL,
        0x3000ULL,
        0x2800ULL,
        0x3c00ULL
    },
    { // m = 15
        0x5555ULL,
        0x7fffULL,
        0x4000ULL,
        0x6000ULL,
        0x5000ULL,
        0x7800ULL,
        0x4400ULL,
        0x6600ULL,
        0x5500ULL,
        0x7f80ULL,
        0x4040ULL,
        0x6060ULL,
        0x5050ULL,
        0x7878ULL,
        0x4444ULL,
        0x6666ULL,
        0x5555ULL,
        0x7fffULL,
        0x4000ULL,
        0x6000ULL,
        0x5000ULL,
        0x7800ULL
    },
    { // m = 16
        0x8000ULL,
        0xc000ULL,
        0xa000ULL,
        0xf000ULL,
        0x8800ULL,
        0xcc00ULL,
        0xaa00ULL,
        0xff00ULL,
        0x8080ULL,
        0xc0c0ULL,
        0xa0a0ULL,
        0xf0f0ULL,
        0x8888ULL,
        0xccccULL,
        0xaaaaULL,
        0xffffULL,
        0x8000ULL,
        0xc000ULL,
        0xa000ULL,
        0xf000ULL
    },
    { // m = 17
        0x14000ULL,
        0x1e000ULL,
        0x11000ULL,
        0x19800ULL,
        0x15400ULL,
        0x1fe00ULL,
        0x10100ULL,
        0x18180ULL,
        0x14140ULL,
        0x1e1e0ULL,
        0x11110ULL,
        0x19998ULL,
        0x15554ULL,
        0x1fffeULL,
        0x10001ULL,
        0x18001ULL,
        0x14001ULL,
        0x1e001ULL
    },
    { // m = 18
        0x22000ULL,
        0x33000ULL,
        0x2a800ULL,
        0x3fc00ULL,
        0x20200ULL,
        0x30300ULL,
        0x28280ULL,
        0x3c3c0ULL,
        0x22220ULL,
        0x33330ULL,
        0x2aaa8ULL,
        0x3fffcULL,
        0x20002ULL,
        0x30003ULL,
        0x28002ULL,
        0x3c003ULL
    },
    { // m = 19
        0x55000ULL,
        0x7f800ULL,
        0x40400ULL,
        0x60600ULL,
        0x50500ULL,
        0x78780ULL,
        0x44440ULL,
        0x66660ULL,
        0x55550ULL,
        0x7fff8ULL,
        0x40004ULL,
        0x60006ULL,
        0x50005ULL,
        0x78007ULL
    },
    { // m = 20
        0x80800ULL,
        0xc0c00ULL,
        0xa0a00ULL,
        0xf0f00ULL,
        0x88880ULL,
        0xcccc0ULL,
        0xaaaa0ULL,
        0xffff0ULL,
        0x80008ULL,
        0xc000cULL,
        0xa000aULL,
        0xf000fULL
    },
    { // m = 21
        0x141400ULL,
        0x1e1e00ULL,
        0x111100ULL,
        0x199980ULL,
        0x155540ULL,
        0x1fffe0ULL,
        0x100010ULL,
        0x180018ULL,
        0x140014ULL,
        0x1e001eULL
    },
    { // m = 22
        0x222200ULL,
        0x333300ULL,
        0x2aaa80ULL,
        0x3fffc0ULL,
        0x200020ULL,
        0x300030ULL,
        0x280028ULL,
        0x3c003cULL
    },
    { // m = 23
        0x555500ULL,
        0x7fff80ULL,
        0x400040ULL,
        0x600060ULL,
        0x500050ULL,
        0x780078ULL
    },
    { // m = 24
        0x800080ULL,
        0xc000c0ULL,
        0xa000a0ULL,
        0xf000f0ULL
    },
    { // m = 25
        0x1400140ULL,
        0x1e001e0ULL
    }
};

const uint64_t Matrices::vdc_sobol_matrices_inv[][52] =
{
    { // m = 1
        0x2ULL,
        0x3ULL
    },
    { // m = 2
        0xcULL,
        0x4ULL,
        0xaULL,
        0x5ULL
    },
    { // m = 3
        0x28ULL,
        0x30ULL,
        0x10ULL,
        0x3cULL,
        0x22ULL,
        0x11ULL
    },
    { // m = 4
        0xf0ULL,
        0x50ULL,
        0x30ULL,
        0x10ULL,
        0x88ULL,
        0x44ULL,
        0x22ULL,
        0x11ULL
    },
    { // m = 5
        0x220ULL,
        0x3c0ULL,
        0x360ULL,
        0x300ULL,
        0x100ULL,
        0x330ULL,
        0x2a8ULL,
        0x264ULL,
        0x202ULL,
        0x101ULL
    },
    { // m = 6
        0xcc0ULL,
        0x440ULL,
        0xf00ULL,
        0x500ULL,
        0x300ULL,
        0x100ULL,
        0xaa0ULL,
        0x550ULL,
        0x808ULL,
        0x404ULL,
        0x202ULL,
        0x101ULL
    },
    { // m = 7
        0x2a80ULL,
        0x3300ULL,
        0x1100ULL,
        0xf00ULL,
        0x500ULL,
        0x300ULL,
        0x100ULL,
        0x3fc0ULL,
        0x2020ULL,
        0x1010ULL,
        0x808ULL,
        0x404ULL,
        0x202ULL,
        0x101ULL
    },
    { // m = 8
        0xff00ULL,
        0x5500ULL,
        0x3300ULL,
        0x1100ULL,
        0xf00ULL,
        0x500ULL,
        0x300ULL,
        0x100ULL,
        0x8080ULL,
        0x4040ULL,
        0x2020ULL,
        0x1010ULL,
        0x808ULL,
        0x404ULL,
        0x202ULL,
        0x101ULL
    },
    { // m = 9
        0x20200ULL,
        0x3fc00ULL,
        0x35600ULL,
        0x33000ULL,
        0x31200ULL,
        0x30c00ULL,
        0x30600ULL,
        0x30000ULL,
        0x10000ULL,
        0x30300ULL,
        0x28280ULL,
        0x24240ULL,
        0x22220ULL,
        0x21210ULL,
        0x20a08ULL,
        0x20604ULL,
        0x20002ULL,
        0x10001ULL
    },
    { // m = 10
        0xc0c00ULL,
        0x40400ULL,
        0xff000ULL,
        0x55000ULL,
        0xf3c00ULL,
        0x51400ULL,
        0xf0000ULL,
        0x50000ULL,
        0x30000ULL,
        0x10000ULL,
        0xa0a00ULL,
        0x50500ULL,
        0x88880ULL,
        0x44440ULL,
        0x82820ULL,
        0x41410ULL,
        0x80008ULL,
        0x40004ULL,
        0x20002ULL,
        0x10001ULL
    },
    { // m = 11
        0x282800ULL,
        0x303000ULL,
        0x101000ULL,
        0xff000ULL,
        0x2d7800ULL,
        0x330000ULL,
        0x110000ULL,
        0xf0000ULL,
        0x50000ULL,
        0x30000ULL,
        0x10000ULL,
        0x3c3c00ULL,
        0x222200ULL,
        0x111100ULL,
        0x88880ULL,
        0x387840ULL,
        0x200020ULL,
        0x100010ULL,
        0x80008ULL,
        0x40004ULL,
        0x20002ULL,
        0x10001ULL
    },
    { // m = 12
        0xf0f000ULL,
        0x505000ULL,
        0x303000ULL,
        0x101000ULL,
        0xff0000ULL,
        0x550000ULL,
        0x330000ULL,
        0x110000ULL,
        0xf0000ULL,
        0x50000ULL,
        0x30000ULL,
        0x10000ULL,
        0x888800ULL,
        0x444400ULL,
        0x222200ULL,
        0x111100ULL,
        0x800080ULL,
        0x400040ULL,
        0x200020ULL,
        0x100010ULL,
        0x80008ULL,
        0x40004ULL,
        0x20002ULL,
        0x10001ULL
    },
    { // m = 13
        0x2222000ULL,
        0x3c3c000ULL,
        0x3636000ULL,
        0x3030000ULL,
        0x1010000ULL,
        0xff0000ULL,
        0x550000ULL,
        0x330000ULL,
        0x110000ULL,
        0xf0000ULL,
        0x50000ULL,
        0x30000ULL,
        0x10000ULL,
        0x3333000ULL,
        0x2aaa800ULL,
        0x2666400ULL,
        0x2000200ULL,
        0x1000100ULL,
        0x800080ULL,
        0x400040ULL,
        0x200020ULL,
        0x100010ULL,
        0x80008ULL,
        0x40004ULL,
        0x20002ULL,
        0x10001ULL
    },
    { // m = 14
        0xcccc000ULL,
        0x4444000ULL,
        0xf0f0000ULL,
        0x5050000ULL,
        0x3030000ULL,
        0x1010000ULL,
        0xff0000ULL,
        0x550000ULL,
        0x330000ULL,
        0x110000ULL,
        0xf0000ULL,
        0x50000ULL,
        0x30000ULL,
        0x10000ULL,
        0xaaaa000ULL,
        0x5555000ULL,
        0x8000800ULL,
        0x4000400ULL,
        0x2000200ULL,
        0x1000100ULL,
        0x800080ULL,
        0x400040ULL,
        0x200020ULL,
        0x100010ULL,
        0x80008ULL,
        0x40004ULL,
        0x20002ULL,
        0x10001ULL
    },
    { // m = 15
        0x2aaa8000ULL,
        0x33330000ULL,
        0x11110000ULL,
        0xf0f0000ULL,
        0x5050000ULL,
        0x3030000ULL,
        0x1010000ULL,
        0xff0000ULL,
        0x550000ULL,
        0x330000ULL,
        0x110000ULL,
        0xf0000ULL,
        0x50000ULL,
        0x30000ULL,
        0x10000ULL,
        0x3fffc000ULL,
        0x20002000ULL,
        0x10001000ULL,
        0x8000800ULL,
        0x4000400ULL,
        0x2000200ULL,
        0x1000100ULL,
        0x800080ULL,
        0x400040ULL,
        0x200020ULL,
        0x100010ULL,
        0x80008ULL,
        0x40004ULL,
        0x20002ULL,
        0x10001ULL
    },
    { // m = 16
        0xffff0000ULL,
        0x55550000ULL,
        0x33330000ULL,
        0x11110000ULL,
        0xf0f0000ULL,
        0x5050000ULL,
        0x3030000ULL,
        0x1010000ULL,
        0xff0000ULL,
        0x550000ULL,
        0x330000ULL,
        0x110000ULL,
        0xf0000ULL,
        0x50000ULL,
        0x30000ULL,
        0x10000ULL,
        0x80008000ULL,
        0x40004000ULL,
        0x20002000ULL,
        0x10001000ULL,
        0x8000800ULL,
        0x4000400ULL,
        0x2000200ULL,
        0x1000100ULL,
        0x800080ULL,
        0x400040ULL,
        0x200020ULL,
        0x100010ULL,
        0x80008ULL,
        0x40004ULL,
        0x20002ULL,
        0x10001ULL
    },
    { // m = 17
        0x200020000ULL,
        0x3fffc0000ULL,
        0x355560000ULL,
        0x333300000ULL,
        0x311120000ULL,
        0x30f0c0000ULL,
        0x305060000ULL,
        0x303000000ULL,
        0x301020000ULL,
        0x300fc0000ULL,
        0x300560000ULL,
        0x300300000ULL,
        0x300120000ULL,
        0x3000c0000ULL,
        0x300060000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0x300030000ULL,
        0x280028000ULL,
        0x240024000ULL,
        0x220022000ULL,
        0x210021000ULL,
        0x208020800ULL,
        0x204020400ULL,
        0x202020200ULL,
        0x201020100ULL,
        0x200820080ULL,
        0x200420040ULL,
        0x200220020ULL,
        0x200120010ULL,
        0x2000a0008ULL,
        0x200060004ULL,
        0x200000002ULL,
        0x100000001ULL
    },
    { // m = 18
        0xc000c0000ULL,
        0x400040000ULL,
        0xffff00000ULL,
        0x555500000ULL,
        0xf333c0000ULL,
        0x511140000ULL,
        0xf0f000000ULL,
        0x505000000ULL,
        0xf030c0000ULL,
        0x501040000ULL,
        0xf00f00000ULL,
        0x500500000ULL,
        0xf003c0000ULL,
        0x500140000ULL,
        0xf00000000ULL,
        0x500000000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0xa000a0000ULL,
        0x500050000ULL,
        0x880088000ULL,
        0x440044000ULL,
        0x820082000ULL,
        0x410041000ULL,
        0x808080800ULL,
        0x404040400ULL,
        0x802080200ULL,
        0x401040100ULL,
        0x800880080ULL,
        0x400440040ULL,
        0x800280020ULL,
        0x400140010ULL,
        0x800000008ULL,
        0x400000004ULL,
        0x200000002ULL,
        0x100000001ULL
    },
    { // m = 19
        0x2800280000ULL,
        0x3000300000ULL,
        0x1000100000ULL,
        0xffff00000ULL,
        0x2d55780000ULL,
        0x3333000000ULL,
        0x1111000000ULL,
        0xf0f000000ULL,
        0x2d05280000ULL,
        0x3303300000ULL,
        0x1101100000ULL,
        0xf00f00000ULL,
        0x2d00780000ULL,
        0x3300000000ULL,
        0x1100000000ULL,
        0xf00000000ULL,
        0x500000000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0x3c003c0000ULL,
        0x2200220000ULL,
        0x1100110000ULL,
        0x880088000ULL,
        0x3840384000ULL,
        0x2020202000ULL,
        0x1010101000ULL,
        0x808080800ULL,
        0x3804380400ULL,
        0x2002200200ULL,
        0x1001100100ULL,
        0x800880080ULL,
        0x3800780040ULL,
        0x2000000020ULL,
        0x1000000010ULL,
        0x800000008ULL,
        0x400000004ULL,
        0x200000002ULL,
        0x100000001ULL
    },
    { // m = 20
        0xf000f00000ULL,
        0x5000500000ULL,
        0x3000300000ULL,
        0x1000100000ULL,
        0xffff000000ULL,
        0x5555000000ULL,
        0x3333000000ULL,
        0x1111000000ULL,
        0xff0ff00000ULL,
        0x5505500000ULL,
        0x3303300000ULL,
        0x1101100000ULL,
        0xff00000000ULL,
        0x5500000000ULL,
        0x3300000000ULL,
        0x1100000000ULL,
        0xf00000000ULL,
        0x500000000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0x8800880000ULL,
        0x4400440000ULL,
        0x2200220000ULL,
        0x1100110000ULL,
        0x8080808000ULL,
        0x4040404000ULL,
        0x2020202000ULL,
        0x1010101000ULL,
        0x8008800800ULL,
        0x4004400400ULL,
        0x2002200200ULL,
        0x1001100100ULL,
        0x8000000080ULL,
        0x4000000040ULL,
        0x2000000020ULL,
        0x1000000010ULL,
        0x800000008ULL,
        0x400000004ULL,
        0x200000002ULL,
        0x100000001ULL
    },
    { // m = 21
        0x22002200000ULL,
        0x3c003c00000ULL,
        0x36003600000ULL,
        0x30003000000ULL,
        0x10001000000ULL,
        0xffff000000ULL,
        0x5555000000ULL,
        0x3333000000ULL,
        0x23113200000ULL,
        0x3cf0cc00000ULL,
        0x36506600000ULL,
        0x30300000000ULL,
        0x10100000000ULL,
        0xff00000000ULL,
        0x5500000000ULL,
        0x3300000000ULL,
        0x1100000000ULL,
        0xf00000000ULL,
        0x500000000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0x33003300000ULL,
        0x2a802a80000ULL,
        0x26402640000ULL,
        0x20202020000ULL,
        0x10101010000ULL,
        0x8080808000ULL,
        0x4040404000ULL,
        0x2020202000ULL,
        0x32013201000ULL,
        0x2a00aa00800ULL,
        0x26006600400ULL,
        0x20000000200ULL,
        0x10000000100ULL,
        0x8000000080ULL,
        0x4000000040ULL,
        0x2000000020ULL,
        0x1000000010ULL,
        0x800000008ULL,
        0x400000004ULL,
        0x200000002ULL,
        0x100000001ULL
    },
    { // m = 22
        0xcc00cc00000ULL,
        0x44004400000ULL,
        0xf000f000000ULL,
        0x50005000000ULL,
        0x30003000000ULL,
        0x10001000000ULL,
        0xffff000000ULL,
        0x5555000000ULL,
        0xcf33fc00000ULL,
        0x45115400000ULL,
        0xf0f00000000ULL,
        0x50500000000ULL,
        0x30300000000ULL,
        0x10100000000ULL,
        0xff00000000ULL,
        0x5500000000ULL,
        0x3300000000ULL,
        0x1100000000ULL,
        0xf00000000ULL,
        0x500000000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0xaa00aa00000ULL,
        0x55005500000ULL,
        0x80808080000ULL,
        0x40404040000ULL,
        0x20202020000ULL,
        0x10101010000ULL,
        0x8080808000ULL,
        0x4040404000ULL,
        0xa802a802000ULL,
        0x54015401000ULL,
        0x80000000800ULL,
        0x40000000400ULL,
        0x20000000200ULL,
        0x10000000100ULL,
        0x8000000080ULL,
        0x4000000040ULL,
        0x2000000020ULL,
        0x1000000010ULL,
        0x800000008ULL,
        0x400000004ULL,
        0x200000002ULL,
        0x100000001ULL
    },
    { // m = 23
        0x2a802a800000ULL,
        0x330033000000ULL,
        0x110011000000ULL,
        0xf000f000000ULL,
        0x50005000000ULL,
        0x30003000000ULL,
        0x10001000000ULL,
        0xffff000000ULL,
        0x2ad57f800000ULL,
        0x333300000000ULL,
        0x111100000000ULL,
        0xf0f00000000ULL,
        0x50500000000ULL,
        0x30300000000ULL,
        0x10100000000ULL,
        0xff00000000ULL,
        0x5500000000ULL,
        0x3300000000ULL,
        0x1100000000ULL,
        0xf00000000ULL,
        0x500000000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0x3fc03fc00000ULL,
        0x202020200000ULL,
        0x101010100000ULL,
        0x80808080000ULL,
        0x40404040000ULL,
        0x20202020000ULL,
        0x10101010000ULL,
        0x8080808000ULL,
        0x3f807f804000ULL,
        0x200000002000ULL,
        0x100000001000ULL,
        0x80000000800ULL,
        0x40000000400ULL,
        0x20000000200ULL,
        0x10000000100ULL,
        0x8000000080ULL,
        0x4000000040ULL,
        0x2000000020ULL,
        0x1000000010ULL,
        0x800000008ULL,
        0x400000004ULL,
        0x200000002ULL,
        0x100000001ULL
    },
    { // m = 24
        0xff00ff000000ULL,
        0x550055000000ULL,
        0x330033000000ULL,
        0x110011000000ULL,
        0xf000f000000ULL,
        0x50005000000ULL,
        0x30003000000ULL,
        0x10001000000ULL,
        0xffff00000000ULL,
        0x555500000000ULL,
        0x333300000000ULL,
        0x111100000000ULL,
        0xf0f00000000ULL,
        0x50500000000ULL,
        0x30300000000ULL,
        0x10100000000ULL,
        0xff00000000ULL,
        0x5500000000ULL,
        0x3300000000ULL,
        0x1100000000ULL,
        0xf00000000ULL,
        0x500000000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0x808080800000ULL,
        0x404040400000ULL,
        0x202020200000ULL,
        0x101010100000ULL,
        0x80808080000ULL,
        0x40404040000ULL,
        0x20202020000ULL,
        0x10101010000ULL,
        0x800000008000ULL,
        0x400000004000ULL,
        0x200000002000ULL,
        0x100000001000ULL,
        0x80000000800ULL,
        0x40000000400ULL,
        0x20000000200ULL,
        0x10000000100ULL,
        0x8000000080ULL,
        0x4000000040ULL,
        0x2000000020ULL,
        0x1000000010ULL,
        0x800000008ULL,
        0x400000004ULL,
        0x200000002ULL,
        0x100000001ULL
    },
    { // m = 25
        0x2020202000000ULL,
        0x3fc03fc000000ULL,
        0x3560356000000ULL,
        0x3300330000000ULL,
        0x3120312000000ULL,
        0x30c030c000000ULL,
        0x3060306000000ULL,
        0x3000300000000ULL,
        0x1000100000000ULL,
        0xffff00000000ULL,
        0x555500000000ULL,
        0x333300000000ULL,
        0x111100000000ULL,
        0xf0f00000000ULL,
        0x50500000000ULL,
        0x30300000000ULL,
        0x10100000000ULL,
        0xff00000000ULL,
        0x5500000000ULL,
        0x3300000000ULL,
        0x1100000000ULL,
        0xf00000000ULL,
        0x500000000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0x3030303000000ULL,
        0x2828282800000ULL,
        0x2424242400000ULL,
        0x2222222200000ULL,
        0x2121212100000ULL,
        0x20a0a0a080000ULL,
        0x2060606040000ULL,
        0x2000000020000ULL,
        0x1000000010000ULL,
        0x800000008000ULL,
        0x400000004000ULL,
        0x200000002000ULL,
        0x100000001000ULL,
        0x80000000800ULL,
        0x40000000400ULL,
        0x20000000200ULL,
        0x10000000100ULL,
        0x8000000080ULL,
        0x4000000040ULL,
        0x2000000020ULL,
        0x1000000010ULL,
        0x800000008ULL,
        0x400000004ULL,
        0x200000002ULL,
        0x100000001ULL
    },
    { // m = 26
        0xc0c0c0c000000ULL,
        0x4040404000000ULL,
        0xff00ff0000000ULL,
        0x5500550000000ULL,
        0xf3c0f3c000000ULL,
        0x5140514000000ULL,
        0xf000f00000000ULL,
        0x5000500000000ULL,
        0x3000300000000ULL,
        0x1000100000000ULL,
        0xffff00000000ULL,
        0x555500000000ULL,
        0x333300000000ULL,
        0x111100000000ULL,
        0xf0f00000000ULL,
        0x50500000000ULL,
        0x30300000000ULL,
        0x10100000000ULL,
        0xff00000000ULL,
        0x5500000000ULL,
        0x3300000000ULL,
        0x1100000000ULL,
        0xf00000000ULL,
        0x500000000ULL,
        0x300000000ULL,
        0x100000000ULL,
        0xa0a0a0a000000ULL,
        0x5050505000000ULL,
        0x8888888800000ULL,
        0x4444444400000ULL,
        0x8282828200000ULL,
        0x4141414100000ULL,
        0x8000000080000ULL,
        0x4000000040000ULL,
        0x2000000020000ULL,
        0x1000000010000ULL,
        0x800000008000ULL,
        0x400000004000ULL,
        0x200000002000ULL,
        0x100000001000ULL,
        0x80000000800ULL,
        0x40000000400ULL,
        0x20000000200ULL,
        0x10000000100ULL,
        0x8000000080ULL,
        0x4000000040ULL,
        0x2000000020ULL,
        0x1000000010ULL,
        0x800000008ULL,
        0x400000004ULL,
        0x200000002ULL,
        0x100000001ULL
    }
};

} // namespace sobol

